// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CAnimationParametersDialog, for altering a document's animation parameters
		\author Tim Shead (tshead@k-3d.com)
*/

#include "animationparameters.h"
#include "document.h"
#include "timestorage.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dspinbutton.h>
#include <k3dtimecontroller.h>

#include <k3diapplication.h>

#include <sdpgtkevent.h>

///////////////////////////////////////////////////////////////////////
// CAnimationDurationDataSource

class CAnimationDurationDataSource :
	public k3dSpinButtonDataSource
{
public:
	CAnimationDurationDataSource(CTimeStorage* const TimeStorage) :
		m_TimeStorage(TimeStorage),
		m_TimeController(TimeStorage)
	{
	}
	
	k3dIEventSource* EventSource()
	{
		return &m_TimeStorage->m_Duration;
	}
	
	gdouble Value()
	{
		return m_TimeController.Duration();
	}
	
	void PushEditing()
	{
	}
	
	void SetValue(const gdouble Value)
	{
		m_TimeController.SetDuration(Value);
	}
	
	void PopEditing()
	{
	}

private:
	CTimeStorage* const m_TimeStorage;
	k3dTimeController m_TimeController;
};

///////////////////////////////////////////////////////////////////////
// CAnimationFrameRateDataSource

class CAnimationFrameRateDataSource :
	public k3dSpinButtonDataSource
{
public:
	CAnimationFrameRateDataSource(CTimeStorage* const TimeStorage) :
		m_TimeStorage(TimeStorage),
		m_TimeController(TimeStorage)
	{
	}
	
	k3dIEventSource* EventSource()
	{
		return &m_TimeStorage->m_FrameRate;
	}
	
	gdouble Value()
	{
		return m_TimeController.FrameRate();
	}
	
	void PushEditing()
	{
	}
	
	void SetValue(const gdouble Value)
	{
		m_TimeController.SetFrameRate(Value);
	}
	
	void PopEditing()
	{
	}

private:
	CTimeStorage* const m_TimeStorage;
	k3dTimeController m_TimeController;
};

///////////////////////////////////////////////////////////////////////
// CAnimationTotalFramesDataSource

class CAnimationTotalFramesDataSource :
	public k3dSpinButtonDataSource,
	public k3dEventSource,
	public k3dEventObserver
{
public:
	CAnimationTotalFramesDataSource(CTimeStorage* const TimeStorage) :
		m_TimeStorage(TimeStorage),
		m_TimeController(TimeStorage)
	{
		m_TimeStorage->m_Duration.AddEventObserver(this, k3dIVariable::Event_Changed, "CAnimationTotalFramesDataSource");
		m_TimeStorage->m_FrameRate.AddEventObserver(this, k3dIVariable::Event_Changed, "CAnimationTotalFramesDataSource");
	}
	
	virtual ~CAnimationTotalFramesDataSource()
	{
		m_TimeStorage->m_Duration.RemoveEventObserver(this);
		m_TimeStorage->m_FrameRate.RemoveEventObserver(this);
	}
	
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
	{
		// Sanity checks ...
		g_return_val_if_fail(Source, SDP_E_FAIL);
		g_return_val_if_string_fail(Event, SDP_E_FAIL);
		
		sdpString event(Event);
		
		if(event == k3dIVariable::Event_Changed)
			{
				Notify(k3dIVariable::Event_Changed, 0);
				return SDP_S_OK;
			}
	
		return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
	}

	k3dIEventSource* EventSource()
	{
		return this;
	}
	
	gdouble Value()
	{
		return m_TimeController.TotalFrames();
	}
	
	void PushEditing()
	{
	}

	void SetValue(const gdouble Value)
	{
		const gulong frames = static_cast<gulong>(Value);
		m_TimeController.SetTotalFrames(frames);
	}

	void PopEditing()
	{
	}

private:
	CTimeStorage* const m_TimeStorage;
	k3dTimeController m_TimeController;
};

///////////////////////////////////////////////////////////////////////
// CAnimationStartFrameDataSource

class CAnimationStartFrameDataSource :
	public k3dSpinButtonDataSource
{
public:
	CAnimationStartFrameDataSource(CTimeStorage* const TimeStorage) :
		m_TimeStorage(TimeStorage),
		m_TimeController(TimeStorage)
	{
	}
	
	k3dIEventSource* EventSource()
	{
		return &m_TimeStorage->m_StartTime;
	}
	
	gdouble Value()
	{
		return m_TimeController.StartFrame();
	}
	
	void PushEditing()
	{
	}
	
	void SetValue(const gdouble Value)
	{
		const gulong frame = static_cast<gulong>(Value);
		m_TimeController.SetStartFrame(frame);
	}
	
	void PopEditing()
	{
	}

private:
	CTimeStorage* const m_TimeStorage;
	k3dTimeController m_TimeController;
};

///////////////////////////////////////////////////////////////////////
// CAnimationViewFrameDataSource

class CAnimationViewFrameDataSource :
	public k3dSpinButtonDataSource
{
public:
	CAnimationViewFrameDataSource(CTimeStorage* const TimeStorage) :
		m_TimeStorage(TimeStorage),
		m_TimeController(TimeStorage)
	{
	}
	
	k3dIEventSource* EventSource()
	{
		return &m_TimeStorage->m_ViewTime;
	}
	
	gdouble Value()
	{
		return m_TimeController.ViewFrame();
	}
	
	void PushEditing()
	{
	}

	void SetValue(const gdouble Value)
	{
		const gulong frame = static_cast<gulong>(Value);
		m_TimeController.SetViewFrame(frame);
	}
	
	void PopEditing()
	{
	}

private:
	CTimeStorage* const m_TimeStorage;
	k3dTimeController m_TimeController;
};

///////////////////////////////////////////////////////////////////////
// CAnimationEndFrameDataSource

class CAnimationEndFrameDataSource :
	public k3dSpinButtonDataSource
{
public:
	CAnimationEndFrameDataSource(CTimeStorage* const TimeStorage) :
		m_TimeStorage(TimeStorage),
		m_TimeController(TimeStorage)
	{
	}
	
	k3dIEventSource* EventSource()
	{
		return &m_TimeStorage->m_EndTime;
	}
	
	gdouble Value()
	{
		return m_TimeController.EndFrame();
	}
	
	void PushEditing()
	{
	}
	
	void SetValue(const gdouble Value)
	{
		const gulong frame = static_cast<gulong>(Value);
		m_TimeController.SetEndFrame(frame);
	}
	
	void PopEditing()
	{
	}

private:
	CTimeStorage* const m_TimeStorage;
	k3dTimeController m_TimeController;
};

///////////////////////////////////////////////////////////////////////
// CAnimationParametersDialog

const sdpString CAnimationParametersDialog::dialogname = "animationproperties";
const sdpString CAnimationParametersDialog::control_duration = "duration";
const sdpString CAnimationParametersDialog::control_framerate = "framerate";
const sdpString CAnimationParametersDialog::control_totalframes = "totalframes";
const sdpString CAnimationParametersDialog::control_firstframe = "firstframe";
const sdpString CAnimationParametersDialog::control_viewframe = "viewframe";
const sdpString CAnimationParametersDialog::control_lastframe = "lastframe";

CAnimationParametersDialog::CAnimationParametersDialog(sdpIUnknown* const ParentCommandNode, CTimeStorage* const TimeStorage, k3dIDocument* Document) :
	CAnimationParametersDialogBase(ParentCommandNode, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Document(Document),
	m_TimeStorage(TimeStorage)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(m_TimeStorage);

	g_return_if_fail(LoadGTKMLTemplate("animationparameters.gtkml"));

	// We want to be notified if the document goes away ...
	k3dIEventSource* document = dynamic_cast<k3dIEventSource*>(m_Document);
	g_return_if_fail(document);
	document->AddEventObserver(this, k3dIDocument::Event_Close, "CAnimationParametersDialog");

	// Connect spin buttons ...
	if(Getk3dSpinButton(control_duration))
		Getk3dSpinButton(control_duration)->Attach(new CAnimationDurationDataSource(m_TimeStorage), k3dApplication()->CommandRecorder(), m_Document->StateRecorder(), "Duration", 0, new k3dActionRedrawAll(m_Document, false, 0), 0);

	if(Getk3dSpinButton(control_framerate))
		Getk3dSpinButton(control_framerate)->Attach(new CAnimationFrameRateDataSource(m_TimeStorage), k3dApplication()->CommandRecorder(), m_Document->StateRecorder(), "Frame Rate", 0, new k3dActionRedrawAll(m_Document, false, 0), 0);

	if(Getk3dSpinButton(control_totalframes))
		Getk3dSpinButton(control_totalframes)->Attach(new CAnimationTotalFramesDataSource(m_TimeStorage), k3dApplication()->CommandRecorder(), m_Document->StateRecorder(), "Total Frames", 0, new k3dActionRedrawAll(m_Document, false, 0), 0);

	if(Getk3dSpinButton(control_firstframe))
		Getk3dSpinButton(control_firstframe)->Attach(new CAnimationStartFrameDataSource(m_TimeStorage), k3dApplication()->CommandRecorder(), m_Document->StateRecorder(), "First Frame", 0, new k3dActionRedrawAll(m_Document, false, 0), 0);

	if(Getk3dSpinButton(control_viewframe))
		Getk3dSpinButton(control_viewframe)->Attach(new CAnimationViewFrameDataSource(m_TimeStorage), k3dApplication()->CommandRecorder(), m_Document->StateRecorder(), "View Frame", 0, new k3dActionRedrawAll(m_Document, false, 0), 0);

	if(Getk3dSpinButton(control_lastframe))
		Getk3dSpinButton(control_lastframe)->Attach(new CAnimationEndFrameDataSource(m_TimeStorage), k3dApplication()->CommandRecorder(), m_Document->StateRecorder(), "Last Frame", 0, new k3dActionRedrawAll(m_Document, false, 0), 0);

	Show();
}

CAnimationParametersDialog::~CAnimationParametersDialog()
{
	// Cancel event notifications ...
	k3dIEventSource* document = dynamic_cast<k3dIEventSource*>(m_Document);
	g_return_if_fail(document);
	document->RemoveEventObserver(this);
}

sdpResult CAnimationParametersDialog::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	if(event == k3dIDocument::Event_Close)
		delete this;
	else
		return CAnimationParametersDialogBase::ObserveEvent(Source, Event, EventData, ObserverData);

	return SDP_S_OK;
}


