// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CApplication class, which encapsulates the running K-3D program
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "applicationframe.h"
#include "blackboxrecorder.h"
#include "clipboard.h"
#include "commandnodeinspector.h"
#include "dialogs.h"
#include "document.h"
#include "objectinfo.h"
#include "macrorecorder.h"
#include "renderfarm.h"
#include "scripteditor.h"
#include "shadereditor.h"
#include "shaders.h"
#include "tutorialmenu.h"
#include "tutorialmessage.h"
#include "tutorialrecorder.h"

#include <k3dcom.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dobjectinfocollection.h>
#include <k3doptions.h>
#include <k3dscripting.h>
#include <k3dstaterecorder.h>

#include <k3dicamera.h>
#include <k3discriptengine.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkutility.h>

#include <errno.h>
#include <time.h>
#include <fstream>
#include <iterator>

#ifdef SDPWIN32
#ifndef SDPMINGW32
#define popen _popen
#else
#include <stdio.h>
#endif
#endif // SDPWIN32

#ifdef SDPUNIX
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#endif // SDPUNUX

#include <sdpgl.h>

namespace
{

/// Writes program versioning information to a stream
class Version
{
public:
	friend std::ostream& operator<<(std::ostream& Stream, const Version& RHS)
	{
		Stream << "K-3D Version " << VERSION << std::endl;
		Stream << "Copyright (c) 1995-2001, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
		Stream << "Licensed by the GNU General Public License.  See the COPYING file for details." << std::endl;
		Stream << "K-3D Home Page: http://www.k-3d.com" << std::endl;
		Stream << "Bug reports to tshead@k-3d.com" << std::endl;
		
		return Stream;
	}
};

/// Writes program usage options to a stream
class Usage
{
	friend std::ostream& operator<<(std::ostream& Stream, const Usage& RHS)
	{
		Stream << "usage: k3d [options]" << std::endl;
		Stream << std::endl;
		Stream << "  -h, --help               prints this help information and exits" << std::endl;
		Stream << "      --basepath [path]    specifies the path for loading plugins, templates, etc." << std::endl;
		Stream << "      --dialogpath [path]  overrides the path for storing dialog templates [default: <basepath>/dialogtemplates]" << std::endl;
		Stream << "      --libpath [path]     overrides the path for loading plugin libraries [default: <basepath>/lib]" << std::endl;
		Stream << "      --scriptspath [path] overrides the path for loading scripts [default: <basepath>/scripts]" << std::endl;
		Stream << "      --shaderspath [path] overrides the path(s) for loading shaders [default: <basepath>/shaders]" << std::endl; 
		Stream << "      --sharepath [path]   overrides the path for loading shared data files [default: <basepath>/share]" << std::endl;
		Stream << "      --tutorialspath [path]  overrides the path for loading interactive tutorials [default: <basepath>/tutorials]" << std::endl;
		Stream << "      --optionspath [path] overrides the filepath for storing user options [default: ~/.k3d/options.k3d]" << std::endl;
		Stream << "  -n, --new                creates a new document after startup" << std::endl;
		Stream << "  -o, --open [file]        opens the given file after startup" << std::endl;
		Stream << "      --script [file]      play the given script after startup (use - for stdin)" << std::endl;
		Stream << "      --batch              operate in batch (no user intervention) mode" << std::endl;
		Stream << "      --exit               exits the program (useful after running scripts in batch mode)" << std::endl;
		Stream << "  -v, --verbose            prints diagnostic messages while running" << std::endl;
		Stream << "      --version            prints program version information and exits" << std::endl;
		Stream << "      --nosplash           disables the startup splash screen" << std::endl;
		Stream << "      --blackbox [file]    starts the Black Box Recorder, recording to [file]" << std::endl;
		
		return Stream;
	}
};

void HandleError(const sdpString Message, bool& Quit, bool& Error)
{
	std::cerr << Message << std::endl << std::endl;
	std::cerr << Usage() << std::endl << std::endl;
	
	Quit = true;
	Error = true;
}

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CApplication

const sdpString CApplication::uiname = "application";
const sdpString CApplication::control_tutorialmessage = "tutorialmessage";

CApplication::CApplication() :
	k3dCommandNode(0, uiname),
	m_ActiveDocument(0),
	m_TutorialMessage(0),
	m_Options("k3dml"),
	m_Modules(new CObjectModules(this)),
	m_Shaders(new CShaderCollection(this)),
	m_Verbose(false),
	m_BatchMode(false),
	m_ShowSplash(true),
	m_RenderFarm(new CRenderFarm()),
	m_Clipboard(new CClipboard())
{
	k3dRegisterApplication(this);
}

CApplication::~CApplication()
{
	// Notify clients ...
	Notify(k3dIApplication::Event_Close, 0);

	// Delete any remaining documents ...
	for(DocumentIterator doc = m_Documents.begin(); doc != m_Documents.end(); doc = m_Documents.begin())
		CloseDocument(*doc);

	// Zap each cached scripting engine ...
	for(ScriptEngineIterator engine = m_ScriptEngines.begin(); engine != m_ScriptEngines.end(); engine++)
		{
			sdpIDeletable* const deletable = dynamic_cast<sdpIDeletable*>(engine->second);
			if(deletable)
				delete deletable;
			else
				std::cerr << g_pretty_function() << ": couldn't delete scripting engine [" << engine->second->Language() << "]" << std::endl;
		}

	// Save user preferences ...
	SaveOptions();

	delete m_TutorialMessage;
	delete m_Modules;
	delete m_Shaders;
	delete m_Clipboard;
	delete m_RenderFarm;
}

int CApplication::Run(sdpStringArray& Options)
{
	bool quit = false;
	bool error = false;

	// Look for command-line options that will cause us to quit immediately ...
	ParseCommandLine1(Options, quit, error);
	if(quit)
		return error ? 1 : 0;

	// Look for command-line options that will configure our subsequent behavior ...
	ParseCommandLine2(Options, quit, error);
	if(quit)
		return error ? 1 : 0;

	// Make sure we have all resources required to run ...
	CheckDependencies(quit, error);
	if(quit)
		return error ? 1 : 0;

	// Create a splash screen as needed ...
	CSplashBox* const splashscreen = m_ShowSplash ? new CSplashBox() : 0;

	// Load user options ...
	LoadOptions();

	// Dynamically load external object modules ...
	LoadModules();

	// Dynamically load Ri Shaders ...
	m_Shaders->LoadShaders();

	// Close the splash screen and open the main window ...
	delete splashscreen;
	new CApplicationFrame(this);
	sdpGtkHandlePendingEvents();
	
	// Parse command-line options that control program execution ...
	ParseCommandLine3(Options, quit, error);
	if(quit)
		return error ? 1 : 0;

	// Main event-loop
	gtk_main();
	
	return 0;
}

void CApplication::ParseCommandLine1(sdpStringArray& CommandLineOptions, bool& Quit, bool& Error)
{
	// Initial conditions ...
	Quit = false;
	Error = false;
	
	for(sdpStringArray::iterator option = CommandLineOptions.begin(); option != CommandLineOptions.end(); )
		{
			if((*option) == "--version")
				{
					std::cout << Version() << std::endl << std::endl;
					
					Quit = true;
					option = CommandLineOptions.erase(option);
				}
			else
				{
					++option;
				}
		}
		
	for(sdpStringArray::iterator option = CommandLineOptions.begin(); option != CommandLineOptions.end(); )
		{
			if((*option) == "--help" || (*option) == "-h")
				{
					std::cout << Usage() << std::endl << std::endl;
					
					Quit = true;
					option = CommandLineOptions.erase(option);
				}
			else
				{
					++option;
				}
		}
}

void CApplication::ParseCommandLine2(sdpStringArray& CommandLineOptions, bool& Quit, bool& Error)
{
	// Initial conditions ...
	Quit = false;
	Error = false;

	// Make a copy of the original options and clear 'em out ...
	sdpStringArray options = CommandLineOptions;
	CommandLineOptions.clear();

	// For each command-line argument ...
	for(sdpStringArray::iterator option = options.begin(); option != options.end(); ++option)
		{
			if((*option) == "--verbose" || (*option) == "-v")
				{
					std::cout << Version() << std::endl << std::endl;
					m_Verbose = true;
				}
			else if((*option) == "--batch")
				{
					m_BatchMode = true;
				}
			else if((*option) == "--nosplash")
				{
					m_ShowSplash = false;
				}
			else if((*option) == "--basepath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --basepath option!", Quit, Error);
							return;
						}
						
					m_BasePath = (*option);
				}
			else if((*option) == "--dialogpath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --dialogpath option!", Quit, Error);
							return;
						}
						
					m_DialogTemplatesPath = (*option);
				}
			else if((*option) == "--libpath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --libpath option!", Quit, Error);
							return;
						}
						
					m_LibPath = (*option);
				}
			else if((*option) == "--scriptspath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --scriptspath option!", Quit, Error);
							return;
						}
						
					m_ScriptsPath = (*option);
				}
			else if((*option) == "--shaderspath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --shaderspath option!", Quit, Error);
							return;
						}
						
					m_ShadersPath = (*option);
				}
			else if((*option) == "--sharepath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --sharepath option!", Quit, Error);
							return;
						}
						
					m_SharePath = (*option);
				}
			else if((*option) == "--tutorialspath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --tutorialspath option!", Quit, Error);
							return;
						}
						
					m_TutorialsPath = (*option);
				}
			else if((*option) == "--optionspath")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a directory path with the --optionspath option!", Quit, Error);
							return;
						}
						
					m_OptionsPath = (*option);
				}
			else
				{
					CommandLineOptions.push_back((*option));
				}
		}
}

void CApplication::CheckDependencies(bool& Quit, bool& Error)
{
	// Initial conditions ...
	Quit = false;
	Error = false;

	// Make sure this platform supports dynamic-modules ...
	if(!g_module_supported())
		{
			HandleError("Host does not support dynamically-loaded modules.  You cannot run K-3D!!", Quit, Error);
			return;
		}

	// Check to be sure that the user has specified a basepath one-way-or-another ...
	if(0 == m_BasePath.size())
		{
			m_BasePath = sdpGetEnv("K3D_BASE_PATH");
			if(0 == m_BasePath.size())
				{
					std::cerr << "You MUST specify a path to the directory that" << std::endl;
					std::cerr << "contains the K-3D plugins & support files (usually" << std::endl;
					std::cerr << "the directory that contains the K-3D executable).  You can" << std::endl;
					std::cerr << "do this by setting the K3D_BASE_PATH environment variable, or" << std::endl;
					std::cerr << "running K-3D with the --basepath [path] command-line option." << std::endl << std::endl;

					Quit = true;
					Error = true;
					return;
				}
		}
		
	// Default any remaining unspecified paths ...
	if(0 == m_DialogTemplatesPath.size())
		m_DialogTemplatesPath = sdpGetEnv("K3D_DIALOGTEMPLATES_PATH");
	if(0 == m_DialogTemplatesPath.size())
		m_DialogTemplatesPath = sdpAppendPath(m_BasePath, "dialogtemplates");
		
	if(0 == m_LibPath.size())
		m_LibPath = sdpGetEnv("K3D_LIB_PATH");
	if(0 == m_LibPath.size())
		m_LibPath = sdpAppendPath(m_BasePath, "lib");
		
	if(0 == m_ScriptsPath.size())
		m_ScriptsPath = sdpGetEnv("K3D_SCRIPTS_PATH");
	if(0 == m_ScriptsPath.size())
		m_ScriptsPath = sdpAppendPath(m_BasePath, "scripts");

	if(0 == m_ShadersPath.size())		
		m_ShadersPath = sdpGetEnv("K3D_SHADERS_PATH");
	if(0 == m_ShadersPath.size())
		m_ShadersPath = sdpAppendPath(m_BasePath, "shaders");
		
	if(0 == m_SharePath.size())
		m_SharePath = sdpGetEnv("K3D_SHARE_PATH");
	if(0 == m_SharePath.size())
		m_SharePath = sdpAppendPath(m_BasePath, "share");
		
	if(0 == m_TutorialsPath.size())
		m_TutorialsPath = sdpGetEnv("K3D_TUTORIALS_PATH");
	if(0 == m_TutorialsPath.size())
		m_TutorialsPath = sdpAppendPath(m_BasePath, "tutorials");

	if(0 == m_OptionsPath.size())
		m_OptionsPath = sdpGetEnv("K3D_OPTIONS_PATH");
	if(0 == m_OptionsPath.size())
		m_OptionsPath = sdpAppendPath(sdpGetHomeDirectory(), ".k3d", "options.k3d");

	if(SDP_S_OK == Verbose())
		{
			std::cout << "Base path: " << BasePath() << std::endl;
			std::cout << "Dialog Templates path: " << DialogTemplatesPath() << std::endl;
			std::cout << "Lib path: " << LibPath() << std::endl;
			std::cout << "Scripts path: " << ScriptsPath() << std::endl;
			std::cout << "Shaders path(s): " << ShadersPath() << std::endl;
			std::cout << "Share path: " << SharePath() << std::endl;
			std::cout << "Tutorials path: " << TutorialsPath() << std::endl;
			std::cout << "Options path: " << OptionsPath() << std::endl;
			std::cout << std::endl;
		}

	// Test for an OpenGL implementation ...
	if(!sdpglExists(std::cerr))
		{
			HandleError("Could not connect to an OpenGL server.  You cannot run K-3D!!", Quit, Error);
			return;
		}

	// Display details about the OpenGL implementation ...
	if(SDP_S_OK == Verbose())
		sdpglConfiguration(std::cout);
}

void CApplication::ParseCommandLine3(sdpStringArray& CommandLineOptions, bool& Quit, bool& Error)
{
	// Initial conditions ...
	Quit = false;
	Error = false;

	// Check past user preferences for opening the tutorials menu ...
	bool showtutorialmenu = k3dOptionsGetTutorialMenu(Options());
	
	// Make a copy of the original options and clear 'em out ...
	sdpStringArray options = CommandLineOptions;
	CommandLineOptions.clear();

	// For each command-line argument ...
	for(sdpStringArray::iterator option = options.begin(); option != options.end(); ++option)
		{
			if((*option) == "--new" || (*option) == "-n")
				{
					const sdpString templatedocument = k3dOptionsGetTemplateDocument(Options());
					
					k3dIDocument* const document = CreateDocument(templatedocument);
					g_return_if_fail(document);
			
					if(0 == templatedocument.size())
						{
							k3d::Editing editing(document->Objects());
						
							// Create an axis object ...
							k3dIObject* const axesobject = document->Objects()->CreateObject(k3d::classid::Axes);
							g_return_if_fail(axesobject);

							// Create a default material object ...
							k3dIObject* const materialobject = document->Objects()->CreateObject(k3d::classid::Material);
							g_return_if_fail(materialobject);

							// Create a default light object ...
							k3dIObject* const lightobject = document->Objects()->CreateObject(k3d::classid::Light);
							g_return_if_fail(lightobject);
							k3dSetPosition(lightobject, sdpVector3(-20.0, 20.0, -30.0));

							// Create a default camera object ...
							k3dIObject* const cameraobject = document->Objects()->CreateObject(k3d::classid::Camera);
							g_return_if_fail(cameraobject);

							k3dSetPosition(cameraobject, sdpVector3(-20.0, 20.0, -20.0));
							k3dSetOrientation(cameraobject, sdpEulerAngles(sdpRadians(0.0), sdpRadians(37.0), sdpRadians(45.0), sdpEulerAngles::ZXYstatic));

							k3dICamera* const camera = dynamic_cast<k3dICamera*>(cameraobject);
							g_return_if_fail(camera);
							camera->ShowViewport();
						}
				}
			else if((*option) == "--open" || (*option) == "-o")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a K-3D document path with the --open option!", Quit, Error);
							return;
						}

					OpenDocument(*option);
				}
			else if((*option) == "--script")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a K-3D script path with the --script option!", Quit, Error);
							return;
						}
					
					bool result = false;
					if(*option == "-")
						{
							*option = "STDIN";
							result = k3d::ExecuteScript(std::cin, "STDIN");
						}
					else
						{
							std::ifstream file(*option);
							result = k3d::ExecuteScript(file, *option);
						}
					
					if(!result)
						{
							std::cerr << "Error executing script [ " << *option << " ]" << std::endl << std::endl;
							Quit = true;
							Error = true;
							return;
						}
				}
			else if((*option) == "--exit")
				{
					Quit = true;
					return;
				}
			else if((*option) == "--blackbox")
				{
					++option;
					if(option == options.end())
						{
							HandleError("You must supply a log file path with the --blackbox option!", Quit, Error);
							return;
						}
						
					BlackBoxRecorder(*option);
				}
			else if((*option) == "--tutorials")
				{
					showtutorialmenu = true;
				}
			else
				{
					CommandLineOptions.push_back(*option);
				}
		}

	// Test for unknown options ...
	if(CommandLineOptions.size())
		{
			std::cerr << "Unknown options: ";
			std::copy(CommandLineOptions.begin(), CommandLineOptions.end(), std::ostream_iterator<sdpString>(std::cerr, " "));
			std::cerr << std::endl << std::endl;
			
			std::cerr << Usage() << std::endl;
			Quit = true;
			Error = true;
			return;
		}

	// Open-up the tutorials menu, if requested ...
	if(showtutorialmenu)
		Tutorials();
}

////////////////////////////////////////////////////////////////////////////////////
// k3dIApplication overrides ...

sdpResult CApplication::Verbose()
{
	return m_Verbose ? SDP_S_OK : SDP_S_FALSE;
}

bool CApplication::BatchMode()
{
	return m_BatchMode;
}

sdpResult CApplication::Close()
{
	if(SDP_S_OK == Verbose())
		std::cout << "CApplication::Close()" << std::endl;

	// Kick us out of the main program loop ...
	gtk_main_quit();

	return SDP_S_OK;
}

k3dIDocument* CApplication::OpenDocument(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, 0);

	Notify(k3dIApplication::Event_OpenDocument, FilePath);

	// Create a new doc ...
	CDocument* document = new CDocument(this, new k3dStateRecorder());
	g_return_val_if_fail(document, 0);

	// Initialize it ...
	if(!document->Open(FilePath))
		{
			delete document;
			g_return_val_if_fail(0, 0);
		}

	// Add it to the doc list ...
	m_Documents.push_back(document);
	
	// Make it the new "active" document ...
	SetActiveDocument(document);

	return document;
}

k3dIDocument* CApplication::CreateDocument(SDPCSTRING TemplatePath)
{
//std::cout << "CApplication::CreateDocument()" << std::endl;

	Notify(k3dIApplication::Event_CreateDocument, 0);

	// Create a new doc ...
	CDocument* document = new CDocument(this, new k3dStateRecorder());
	g_return_val_if_fail(document, 0);

	// Initialize it ...
	if(!document->Create(TemplatePath))
		{
			delete document;
			g_return_val_if_fail(0, 0);
		}

	// Add it to the doc list ...
	m_Documents.push_back(document);

	// Make it the new "active" document ...
	SetActiveDocument(document);

	return document;
}

k3dIDocument* CApplication::ActiveDocument()
{
	return m_ActiveDocument;
}

sdpResult CApplication::SetActiveDocument(k3dIDocument* Document)
{
	// We allow Document to be NULL here, in case the user closes the last document

	// If this is the same document, we're done ...
	if(m_ActiveDocument == Document)
		return SDP_S_FALSE;

	m_ActiveDocument = Document;
	Notify(k3dIApplication::Event_ActiveDocumentChanged, 0);

	return SDP_S_OK;
}

sdpResult CApplication::CloseDocument(k3dIDocument* Document)
{
	// Sanity checks ...
	g_return_val_if_fail(Document, SDP_E_FAIL);

	Notify(k3dIApplication::Event_CloseDocument, 0);

	// Get the private document structure ...	
	CDocument* document = static_cast<CDocument*>(Document);
	
	// Find the document iterator ...
	DocumentIterator iterator = std::find(m_Documents.begin(), m_Documents.end(), document);
	g_return_val_if_fail(iterator != m_Documents.end(), SDP_E_FAIL);

	// Cleanup links to this document ...
	if(m_ActiveDocument == document)
		{
			// Activate a new document ...
			SetActiveDocument(0);

/*
			if(m_Documents.size())
				m_Documents.back()->OnActivate(this);
*/
		}
	
	// Remove it from the document list ...
	m_Documents.erase(iterator);

	// Delete the document ...
	delete document;
	
	return SDP_S_OK;
}

sdpxml::Document& CApplication::Options()
{
	return m_Options;
}

k3dIScriptEngine* CApplication::ScriptEngine(const sdp::ClassID Language)
{
	// Sanity checks ...
	g_return_val_if_fail(Language != sdp::ClassIDNull, 0);

	// Try to find an existing engine ...
	ScriptEngineIterator engineiterator = m_ScriptEngines.find(Language);
	if(engineiterator != m_ScriptEngines.end())
		return engineiterator->second;

	// Try to find the object info for this Language ...
	std::vector<k3dIObjectInfo*> objectinfocollection;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(objectinfocollection), k3d::SameClassID(Language));

	// If we don't support this object type, fail ...
	if(objectinfocollection.empty())
		{
			std::cerr << g_pretty_function() << ": requested unknown scripting engine [" << Language << "]" << std::endl;
			g_return_val_if_fail(0, 0);
		}

	// Create an instance of the object ...
	CObjectInfo* const objectinfo = dynamic_cast<CObjectInfo*>(*objectinfocollection.begin());
	g_return_val_if_fail(objectinfo, 0);

	sdpIUnknown* const unknown = objectinfo->CreateInstance()(Language, this);
	g_return_val_if_fail(unknown, 0);
	
	k3dIScriptEngine* const engine = dynamic_cast<k3dIScriptEngine*>(unknown);
	g_return_val_if_fail(engine, 0);
			
	// Add it to our cache for future requests ...
	m_ScriptEngines[Language] = engine;

	// Some debugging output ...
	if(SDP_S_OK == Verbose())
		std::cout << g_pretty_function() << ": created [" << engine->Language() << "] scripting engine" << std::endl;

	return engine;
}

k3dICommandRecorder* CApplication::CommandRecorder()
{
	return this;
}

sdpResult CApplication::RegisterCommandRecorder(k3dICommandRecorder* CommandRecorder)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandRecorder, SDP_E_FAIL);
	g_return_val_if_fail(std::find(m_CommandRecorders.begin(), m_CommandRecorders.end(), CommandRecorder) == m_CommandRecorders.end(), SDP_E_FAIL);
	
	m_CommandRecorders.insert(CommandRecorder);
	
	return SDP_S_OK;
}

sdpResult CApplication::UnRegisterCommandRecorder(k3dICommandRecorder* CommandRecorder)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandRecorder, SDP_E_FAIL);

	CommandRecorderIterator recorder = std::find(m_CommandRecorders.begin(), m_CommandRecorders.end(), CommandRecorder);
	g_return_val_if_fail(recorder != m_CommandRecorders.end(), SDP_E_FAIL);
	
	m_CommandRecorders.erase(recorder);

	return SDP_S_OK;
}

sdpResult CApplication::SetClipboard(SDPCSTRING Text)
{
	m_Clipboard->SetText(Text);
	g_return_val_if_fail(m_Clipboard->GrabSelection(), SDP_E_FAIL);
	
	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////////
// k3dICommandRecorder overrides ...

sdpResult CApplication::IsRecording()
{
	return m_CommandRecorders.size() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CApplication::RecordCommand(k3dICommandNode* CommandNode, k3dICommandNode::CommandType Type, SDPCSTRING CommandName, SDPCSTRING Arguments)
{
	if(SDP_S_OK != IsRecording())
		{
			std::cerr << "CApplication::RecordCommand(): [" << CommandName << "] [" << Arguments << "]" << std::endl;
			g_return_val_if_fail(0, SDP_E_FAIL);
		}
		
	g_return_val_if_fail(SDP_S_OK == IsRecording(), SDP_E_FAIL);

	for(CommandRecorderIterator recorder = m_CommandRecorders.begin(); recorder != m_CommandRecorders.end(); recorder++)
		if(SDP_S_OK == (*recorder)->IsRecording())
			(*recorder)->RecordCommand(CommandNode, Type, CommandName, Arguments);
	
	return SDP_S_OK;
}

SDPCSTRING CApplication::BasePath()
{
	return m_BasePath;
}

SDPCSTRING CApplication::DialogTemplatesPath()
{
	return m_DialogTemplatesPath;
}

SDPCSTRING CApplication::LibPath()
{
	return m_LibPath;
}

SDPCSTRING CApplication::ScriptsPath()
{
	return m_ScriptsPath;
}

SDPCSTRING CApplication::ShadersPath()
{
	return m_ShadersPath;
}

SDPCSTRING CApplication::SharePath()
{
	return m_SharePath;
}

SDPCSTRING CApplication::TutorialsPath()
{
	return m_TutorialsPath;
}

SDPCSTRING CApplication::OptionsPath()
{
	return m_OptionsPath;
}

k3dIObjectInfoCollection* CApplication::Objects()
{
	return m_Modules;
}

sdpsl::IShaderCollection* CApplication::Shaders()
{
	return m_Shaders;
}

k3dIRenderFarm* CApplication::RenderFarm()
{
	return m_RenderFarm;
}

void CApplication::Message(SDPCSTRING Message, SDPCSTRING Title)
{
	sdpStringArray buttons;
	buttons.push_back("OK");
	
	CMessageBox* const dialog = new CMessageBox(this, Message, Title, 1, buttons, "message.gtkml");
	dialog->Run();
}

guint CApplication::Query(SDPCSTRING Message, SDPCSTRING Title, const guint DefaultButton, const sdpStringArray Buttons)
{
	CMessageBox* const dialog = new CMessageBox(this, Message, Title, DefaultButton, Buttons, "query.gtkml");
	return dialog->Run();
}

void CApplication::Error(SDPCSTRING Message, SDPCSTRING Title)
{
	sdpStringArray buttons;
	buttons.push_back("OK");
	
	CMessageBox* const dialog = new CMessageBox(this, Message, Title, 1, buttons, "error.gtkml");
	dialog->Run();
}

sdpResult CApplication::DebugOutput(SDPCSTRING Message)
{
	std::cout << sdpString(Message) << std::endl;
	return SDP_S_OK;
}

CTutorialMessage* CApplication::TutorialMessageUI()
{
	if(0 == m_TutorialMessage)
		{
			m_TutorialMessage = new CTutorialMessage(m_TutorialMessage);
			g_return_val_if_fail(m_TutorialMessage, 0);
		}

	return m_TutorialMessage;
}

sdpResult CApplication::TutorialMessage(SDPCSTRING Message)
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_tutorialmessage, Message);

	CTutorialMessage* tutorialmessage = TutorialMessageUI();
	g_return_val_if_fail(tutorialmessage, SDP_E_FAIL);

	tutorialmessage->Update(Message);
	return tutorialmessage->Cancelled() ? SDP_S_FALSE : SDP_S_OK;
}

sdpResult CApplication::TutorialMouseDisplay(SDPCSTRING Message, TutorialMouseAction Action, GdkModifierType Modifiers)
{
	CTutorialMessage* tutorialmessage = TutorialMessageUI();
	g_return_val_if_fail(tutorialmessage, SDP_E_FAIL);

	tutorialmessage->Update(Message, Action, Modifiers);
	return tutorialmessage->Cancelled() ? SDP_S_FALSE : SDP_S_OK;
}

sdpResult CApplication::BrowserNavigate(SDPCSTRING URL)
{
	// Sanity checks ...
	g_return_val_if_string_fail(URL, SDP_E_FAIL);
	
	// Get the user's preferred HTML viewer ...
	sdpString commandline = k3dOptionsGetHTMLViewer(Options());
	g_return_val_if_fail(commandline.size(), SDP_E_FAIL);
	commandline.formattedreplace('%', "p", URL);

	// Fire-up the web browser ...
	if(SDP_S_OK == Verbose())
		std::cout << "CApplication::BrowserNavigate(): " << commandline << std::endl;
	
	popen(commandline, "r");

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////
// k3dICommandNode overrides

sdpResult CApplication::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	const sdpString name(Name);
	const sdpString arguments(Arguments);

	k3dRecordCommand(this, k3dICommandNode::SCRIPTING, Name, Arguments);

	if(name == control_tutorialmessage)
		{
			return TutorialMessage(Arguments);
		}
	else if(name == Command_OpenDocument)
		{
			k3dIDocument* const document = OpenDocument(Arguments);
			g_return_val_if_fail(document, SDP_E_FAIL);
			
			if(Result)
				*Result = document;
				
			return SDP_S_OK;
		}
	else if(name == Command_CreateDocument)
		{
			k3dIDocument* const document = CreateDocument(Arguments);
			g_return_val_if_fail(document, SDP_E_FAIL);
			
			if(Result)
				*Result = document;
				
			return SDP_S_OK;
		}
	else if(name == Command_Close)
		{
			return Close();
		}

	return k3dCommandNode::Command(Name, Arguments, Result);
}

void CApplication::LoadOptions()
{
	// Load the document (if it doesn't exist, create it)
	if(!sdpFileExists(m_OptionsPath))
		{
			if(!sdpFileExists(sdpDirectory(m_OptionsPath)))
				sdpCreateDirectory(sdpDirectory(m_OptionsPath));
				
			std::ofstream file(m_OptionsPath);
			file << "<k3dml/>" << std::endl;
		}
		
	m_Options.Load(m_OptionsPath);

	// And make sure it's really a K-3D Markup Language document ...
	if(m_Options.Name() != "k3dml")
		m_Options = sdpxml::Element("k3dml");
	
	// Save the current set of options ...
	SaveOptions();
}

void CApplication::SaveOptions()
{
	if(SDP_S_OK == Verbose())
		std::cout << "Saving options to: " << m_OptionsPath << std::endl;

	std::ofstream stream(m_OptionsPath);
	stream << m_Options << std::endl;
}

void CApplication::LoadModules()
{
	if(SDP_S_OK == Verbose())
		std::cout << "Loading modules from: " << LibPath() << std::endl;

	m_Modules->LoadModules(LibPath(), true);
}

void CApplication::Tutorials()
{
	new CTutorialMenu(this);
}

void CApplication::ScriptEditor()
{
	new CScriptEditor(this);
}

void CApplication::ShaderEditor()
{
	new CShaderEditor(this);
}

void CApplication::MacroRecorder()
{
	new CMacroRecorder(this);
}

void CApplication::TutorialRecorder()
{
	new CTutorialRecorder(this);
}

void CApplication::BlackBoxRecorder()
{
	sdpString filepath;
	if(!k3dGetFilePath("log", "Record Black Box Data:", true, 0, filepath))
		return;

	BlackBoxRecorder(filepath);
}

void CApplication::BlackBoxRecorder(SDPCSTRING LogFile)
{
	// Sanity checks ...
	g_assert_string(LogFile);

	new CBlackBoxRecorder(this, LogFile);
}

void CApplication::CommandNodeInspector()
{
	new CCommandNodeInspector(this);
}


