#ifndef APPLICATION_H
#define APPLICATION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CApplication class, which encapsulates the running K-3D program
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deventsource.h>
#include <k3dcommandnode.h>

#include <k3diapplication.h>
#include <k3dicommandrecorder.h>
#include <k3direnderfarm.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H

// Forward class references ...
class CClipboard;
class CDocument;
class CObjectModules;
class CRenderFarm;
class CShaderCollection;
class CTutorialMessage;

/////////////////////////////////////////////////////////////////////////////
// CApplication

/// Encapsulates the running K-3D server application
class CApplication :
	public k3dIApplication,
	public k3dICommandRecorder,
	public k3dCommandNode,
	public k3dEventSource
{
friend class CApplicationFrame;
public:
	typedef std::list<CDocument*> DocumentList;
	typedef DocumentList::iterator DocumentIterator;

	static const sdpString uiname;
	static const sdpString control_tutorialmessage;

public:
	CApplication();
	~CApplication();

	/// Starts the application event loop
	int Run(sdpStringArray& Options);

	// k3dIApplication overrides
	sdpResult Verbose();
	bool BatchMode();
	sdpResult Close();
	k3dIDocument* OpenDocument(SDPCSTRING FilePath);
	k3dIDocument* CreateDocument(SDPCSTRING TemplatePath);
	k3dIDocument* ActiveDocument();
	sdpResult SetActiveDocument(k3dIDocument* Document);
	sdpResult CloseDocument(k3dIDocument* Document);
	SDPCSTRING BasePath();
	SDPCSTRING DialogTemplatesPath();
	SDPCSTRING LibPath();
	SDPCSTRING ScriptsPath();
	SDPCSTRING ShadersPath();
	SDPCSTRING SharePath();
	SDPCSTRING TutorialsPath();
	SDPCSTRING OptionsPath();
	k3dIObjectInfoCollection* Objects();
	sdpsl::IShaderCollection* Shaders();
	k3dIRenderFarm* RenderFarm();
	sdpResult BrowserNavigate(SDPCSTRING URL);
	void Message(SDPCSTRING Message, SDPCSTRING Title);
	guint Query(SDPCSTRING Message, SDPCSTRING Title, const guint DefaultButton, const sdpStringArray Buttons);
	void Error(SDPCSTRING Message, SDPCSTRING Title);
	sdpResult DebugOutput(SDPCSTRING Message);
	sdpResult TutorialMessage(SDPCSTRING Message);
	sdpResult TutorialMouseDisplay(SDPCSTRING Message, TutorialMouseAction Action, GdkModifierType Modifiers);
	sdpxml::Document& Options();
	k3dIScriptEngine* ScriptEngine(const sdp::ClassID Language);
	k3dICommandRecorder* CommandRecorder();
	sdpResult RegisterCommandRecorder(k3dICommandRecorder* CommandRecorder);
	sdpResult UnRegisterCommandRecorder(k3dICommandRecorder* CommandRecorder);
	sdpResult SetClipboard(SDPCSTRING Text);

	// k3dICommandRecorder overrides ...
	sdpResult IsRecording();
	sdpResult RecordCommand(k3dICommandNode* CommandNode, k3dICommandNode::CommandType Type, SDPCSTRING CommandName, SDPCSTRING Arguments);

	// k3dICommandNode overrides ...
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	/// Display available tutorials
	void Tutorials();
	/// Opens a script editor
	void ScriptEditor();
	/// Opens a shader editor
	void ShaderEditor();
	/// Opens a macro recorder
	void MacroRecorder();
	/// Opens a tutorial recorder
	void TutorialRecorder();
	/// Prompts the user for a filepath and turns-on a black box recorder
	void BlackBoxRecorder();
	/// Turns-on a black box recorder
	void BlackBoxRecorder(SDPCSTRING LogFile);
	/// Opens a command node inspector
	void CommandNodeInspector();
	/// Saves changes to user preferences
	void SaveOptions();

private:
	/// Looks for command-line options that cause us to quit right away.  Returns false iff the program should quit.
	void ParseCommandLine1(sdpStringArray& CommandLineOptions, bool& Quit, bool& Error);
	/// Parses command line options a second time and dispatches commands
	void ParseCommandLine2(sdpStringArray& CommandLineOptions, bool& Quit, bool& Error);
	/// Tests for runtime capabilities that are needed in order for K-3D to function
	void CheckDependencies(bool& Quit, bool& Error);
	/// Parses command line options a third time and dispatches commnads
	void ParseCommandLine3(sdpStringArray& CommandLineOptions, bool& Quit, bool& Error);
	
	/// Loads user preferences at startup
	void LoadOptions();
	/// Loads external object modules
	void LoadModules();
	/// Returns a singleton UI for displaying tutorial information
	CTutorialMessage* TutorialMessageUI();

	/// Stores a pointer to the currently active document
	k3dIDocument* m_ActiveDocument;
	/// Provides a singleton UI for displaying tutorial messages
	CTutorialMessage* m_TutorialMessage;
	/// Stores the K-3D base path
	sdpString m_BasePath;
	/// Stores the K-3D dialogtemplates path
	sdpString m_DialogTemplatesPath;
	/// Stores the K-3D lib path
	sdpString m_LibPath;
	/// Stores the K-3D scripts path
	sdpString m_ScriptsPath;
	/// Stores the K-3D shaders path(s)
	sdpString m_ShadersPath;
	/// Stores the K-3D share path
	sdpString m_SharePath;
	/// Stores the K-3D tutorials path
	sdpString m_TutorialsPath;
	/// Stores the user options path
	sdpString m_OptionsPath;
	/// Stores user options as an XML document
	sdpxml::Document m_Options;
	/// Stores the collection of dynamically-loaded object modules
	CObjectModules* const m_Modules;
	/// Stores the collection of dynamically-loaded RenderMan Shaders
	CShaderCollection* const m_Shaders;
	/// Stores the collection of open documents
	DocumentList m_Documents;
	/// Stores the collection of cached scripting engines
	typedef std::map<sdp::ClassID, k3dIScriptEngine*> ScriptEngineMap;
	typedef ScriptEngineMap::iterator ScriptEngineIterator;
	ScriptEngineMap m_ScriptEngines;
	/// Stores collection of registered command recorders
	typedef std::set<k3dICommandRecorder*> CommandRecorders;
	typedef CommandRecorders::iterator CommandRecorderIterator;
	CommandRecorders m_CommandRecorders;
	/// Set to true when the user wants verbose reporting to stdout
	bool m_Verbose;
	/// Set to true when batch-mode is enabled (operate without user intervention)
	bool m_BatchMode;
	/// Controls whether the startup splash dialog will be shown
	bool m_ShowSplash;
	/// Stores a renderfarm object for rendering jobs/frames
	CRenderFarm* const m_RenderFarm;
	/// Provdes "clipboard" (X selection) functionality
	CClipboard* const m_Clipboard;
};

#endif // APPLICATION_H
