// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CApplicationFrame class, which implements the per-application UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include "applicationframe.h"
#include "application.h"
#include "document.h"
#include "objectinfo.h"
#include "optionsdialog.h"

#include <k3dcom.h>
#include <k3ddnd.h>
#include <k3dmisc.h>
#include <k3dobjectinfocollection.h>
#include <k3doptions.h>
#include <k3dscripting.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3diobjectcollection.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>
#include <sdpgtkutility.h>

#include <fstream>
#include <strstream>

#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

//////////////////////////////////////////////////////////////////////
// CAboutBox

const sdpString CAboutBox::dialogname = "aboutbox";

CAboutBox::CAboutBox() :
	CAboutBoxBase(k3dApplication(), dialogname, 0)
{
	g_return_if_fail(LoadGTKMLTemplate("about.gtkml"));

	if(SDP_S_OK == k3dApplication()->Verbose())
		Widget("test").Show();

	// Set our version and copyright notices ...
	Label("version").SetText("K-3D Version " + sdpString(VERSION));
	
	// Center it ...
	gtk_window_set_position(RootWindow(), GTK_WIN_POS_CENTER);
	
	Show();
	DoModal();
	delete this;
}

void CAboutBox::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ok")
		CancelModal();
	else if(Event->Name() == "test")
		OnTest();
	else		
		CAboutBoxBase::OnEvent(Event);
}

void ShowUINodeHierarchy(k3dICommandNode* CommandNode, gulong Level)
{
	// Sanity checks ...
	g_assert(CommandNode);
	
	sdpString indentation(2 * Level, ' ');
	
	std::cout << indentation << CommandNode->CommandNodeName() << std::endl;
	for(K3DCOMMANDNODEHANDLE uihandle = CommandNode->FirstCommandNode(); uihandle; uihandle = CommandNode->NextCommandNode(uihandle))
		ShowUINodeHierarchy(CommandNode->CommandNode(uihandle), Level+1);
}

void CAboutBox::OnTest()
{
	std::cout << "CAboutBox::OnTest()" << std::endl;

	k3dICommandNode* commandnode = dynamic_cast<k3dICommandNode*>(k3dApplication());
	g_return_if_fail(commandnode);
	
	ShowUINodeHierarchy(commandnode, 1);
}

//////////////////////////////////////////////////////////////////////
// CSplashBox

const sdpString CSplashBox::dialogname = "splashbox";

CSplashBox::CSplashBox() :
	CSplashBoxBase(k3dApplication(), dialogname, 0)
{
	g_return_if_fail(LoadGTKMLTemplate("splash.gtkml"));

	// Set our version and copyright notices ...
	Label("version").SetText("K-3D Version " + sdpString(VERSION));
	
	// Center it ...
	gtk_window_set_position(RootWindow(), GTK_WIN_POS_CENTER);

	// Make sure we're visible ...
	Show();

	// Give the dialog a chance to display itself ...
	for(int z = 0; z < 10; z++)
		{
			sdpSleep(10);
			sdpGtkHandlePendingEvents();
		}
}

//////////////////////////////////////////////////////////////////////
// CApplicationFrame

const sdpString CApplicationFrame::dialogname = "window";

const sdpString CApplicationFrame::control_filenew = "filenew";
const sdpString CApplicationFrame::control_fileopen = "fileopen";
const sdpString CApplicationFrame::control_fileexit = "fileexit";
const sdpString CApplicationFrame::control_toolsoptions = "toolsoptions";
const sdpString CApplicationFrame::control_toolsblackboxrecorder = "toolsblackboxrecorder";
const sdpString CApplicationFrame::control_toolsrecordmacro = "toolsrecordmacro";
const sdpString CApplicationFrame::control_toolsrecordtutorial = "toolsrecordtutorial";
const sdpString CApplicationFrame::control_toolseditshader = "toolseditshader";
const sdpString CApplicationFrame::control_toolseditcommandnodeinspector = "toolseditcommandnodeinspector";
const sdpString CApplicationFrame::control_toolseditscript = "toolseditscript";
const sdpString CApplicationFrame::control_toolsplayscript = "toolsplayscript";
const sdpString CApplicationFrame::control_helptutorials = "helptutorials";
const sdpString CApplicationFrame::control_helponline = "helponline";
const sdpString CApplicationFrame::control_helpabout = "helpabout";

const sdpString CApplicationFrame::control_highlighttool = "highlighttool";
const sdpString CApplicationFrame::control_highlightobject = "highlightobject";
const sdpString CApplicationFrame::control_opentool = "opentool";
const sdpString CApplicationFrame::control_createobject = "createobject";

const sdpString CApplicationFrame::control_dragbegin = "dragbegin";
const sdpString CApplicationFrame::control_dragmotion = "dragmotion";
const sdpString CApplicationFrame::control_dragdataget = "dragdataget";
const sdpString CApplicationFrame::control_dragdatadelete = "dragdatadelete";
const sdpString CApplicationFrame::control_dragdrop = "dragdrop";
const sdpString CApplicationFrame::control_dragend = "dragend";

CApplicationFrame::CApplicationFrame(CApplication* const Application) :
	CApplicationFrameBase(Application, dialogname, new k3dOptionsWindowGeometryStore(Application, 0)),
	m_Application(Application)
{
	g_return_if_fail(LoadGTKMLTemplate("applicationframe.gtkml"));
	
	// Setup our user actions ...
	AddCommand(control_filenew, control_filenew, CONTROL_MENU_ITEM);
	AddCommand(control_fileopen, control_fileopen, CONTROL_MENU_ITEM);
	AddCommand(control_fileexit, control_fileexit, CONTROL_MENU_ITEM);
	AddCommand(control_toolsoptions, control_toolsoptions, CONTROL_MENU_ITEM);
	AddCommand(control_toolsrecordmacro, control_toolsrecordmacro, CONTROL_MENU_ITEM);
	AddCommand(control_toolsrecordtutorial, control_toolsrecordtutorial, CONTROL_MENU_ITEM);
	AddCommand(control_toolseditcommandnodeinspector, control_toolseditcommandnodeinspector, CONTROL_MENU_ITEM);
	AddCommand(control_toolseditscript, control_toolseditscript, CONTROL_MENU_ITEM);
	AddCommand(control_toolsplayscript, control_toolsplayscript, CONTROL_MENU_ITEM);
	AddCommand(control_helponline, control_helponline, CONTROL_MENU_ITEM);
	AddCommand(control_helpabout, control_helpabout, CONTROL_MENU_ITEM);
	AddCommand(control_opentool, control_opentool, CONTROL_BUTTON);
	AddCommand(control_createobject, control_createobject, CONTROL_BUTTON);

	m_Application->AddEventObserver(this, k3dIApplication::Event_Close, "CApplicationFrame");
	m_Application->AddEventObserver(this, k3dIApplication::Event_ActiveDocumentChanged, "CApplicationFrame");

	// Set ourselves up for drag-n-drop ...
	ObjectsList().SetDragSource(GDK_BUTTON1_MASK, k3dDNDCreateObject(), 1, GDK_ACTION_COPY);

//	MapEvent("drag-begin", control_dragbegin, false, ObjectsList(), true);
//	MapEvent("drag-motion", control_dragmotion, false, ObjectsList(), true);
	MapEvent("drag-data-get", control_dragdataget, false, ObjectsList(), true);
//	MapEvent("drag-data-delete", control_dragdatadelete, false, ObjectsList(), true);
//	MapEvent("drag-drop", control_dragdrop, false, ObjectsList(), true);
//	MapEvent("drag-end", control_dragend, false, ObjectsList(), true);

	InitializeTrees();
	CallUpdateControls();
	Show();
}

CApplicationFrame::~CApplicationFrame()
{
	if(SDP_S_OK == m_Application->Verbose())
		std::cout << "CApplicationFrame::~CApplicationFrame()" << std::endl;

	m_Application->RemoveEventObserver(this);
}

sdpResult CApplicationFrame::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	sdpString event(Event);
	
	if(event == k3dIApplication::Event_ActiveDocumentChanged)
		{
			CallUpdateControls();
			return SDP_S_OK;
		}
	else if(event == k3dIApplication::Event_Close)
		{
			delete this;
			return SDP_S_OK;
		}

	return CApplicationFrameBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void CApplicationFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_filenew)
		OnFileNew();
	else if(Event->Name() == control_fileopen)
		OnFileOpen();
	else if(Event->Name() == control_fileexit)
		OnFileExit();
	else if(Event->Name() == "toolsclicked")
		OnToolsClicked(Event);
	else if(Event->Name() == "objectsclicked")
		OnObjectsClicked(Event);
	else if(Event->Name() == control_toolsblackboxrecorder)
		OnToolsBlackBoxRecorder();
	else if(Event->Name() == control_toolsrecordmacro)
		OnToolsRecordMacro();
	else if(Event->Name() == control_toolsrecordtutorial)
		OnToolsRecordTutorial();
	else if(Event->Name() == control_toolseditshader)
		OnToolsEditShader();
	else if(Event->Name() == control_toolseditcommandnodeinspector)
		OnToolsEditCommandNodeInspector();
	else if(Event->Name() == control_toolseditscript)
		OnToolsEditScript();
	else if(Event->Name() == control_toolsplayscript)
		OnToolsPlayScript();
	else if(Event->Name() == control_toolsoptions)
		OnToolsOptions();
	else if(Event->Name() == control_helptutorials)
		OnHelpTutorials();
	else if(Event->Name() == control_helponline)
		OnHelpOnline();
	else if(Event->Name() == control_helpabout)
		OnHelpAbout();
	else if(Event->Name() == control_opentool)
		OnOpenTool();
	else if(Event->Name() == control_createobject)
		OnCreateObject();
	else if(Event->Name() == control_dragdataget)
		OnDragDataGet(Event);
	else
		CApplicationFrameBase::OnEvent(Event);
}

void CApplicationFrame::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(SDP_S_OK == m_Application->Verbose())
		std::cout << "CApplicationFrame::OnDelete()" << std::endl;

	// Kill the application ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);
	
	if(SafeToExit())
		m_Application->Close();
}

// User wants to create a new document ...
void CApplicationFrame::OnFileNew()
{
//std::cout << "CApplicationFrame::OnFileNew()" << std::endl;

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_filenew, 0);
	
	// See if we have a template document, first ...
	const sdpString templatedocument = k3dOptionsGetTemplateDocument(m_Application->Options());
	
	sdpIUnknown* unknown = 0;
	g_return_if_fail(sdpSucceeded(m_Application->Command(k3dIApplication::Command_CreateDocument, templatedocument, &unknown)));
	g_return_if_fail(unknown);

	if(0 == templatedocument.size())
		{	
			k3dIDocument* const document = dynamic_cast<k3dIDocument*>(unknown);
			g_return_if_fail(document);

			k3d::Editing editing(document->Objects());

			// Create an axis object ...
			k3dIObject* const axesobject = document->Objects()->CreateObject(k3d::classid::Axes);
			g_return_if_fail(axesobject);

			// Create a default material object ...
			k3dIObject* const materialobject = document->Objects()->CreateObject(k3d::classid::Material);
			g_return_if_fail(materialobject);

			// Create a default light object ...
			k3dIObject* const lightobject = document->Objects()->CreateObject(k3d::classid::Light);
			g_return_if_fail(lightobject);
			k3dSetPosition(lightobject, sdpVector3(-20.0, 20.0, -30.0));

			// Create a default camera object ...
			k3dIObject* const cameraobject = document->Objects()->CreateObject(k3d::classid::Camera);
			g_return_if_fail(cameraobject);
			k3dSetPosition(cameraobject, sdpVector3(-20.0, 20.0, -20.0));
			k3dSetOrientation(cameraobject, sdpEulerAngles(sdpRadians(0.0), sdpRadians(37.0), sdpRadians(45.0), sdpEulerAngles::ZXYstatic));

			k3dICamera* const camera = dynamic_cast<k3dICamera*>(cameraobject);
			g_return_if_fail(camera);
			camera->ShowViewport();
		}
}

// User wants to open an existing document ...
void CApplicationFrame::OnFileOpen()
{
	sdpString filepath;
	if(!k3dGetFilePath("document", "Open K-3D Document:", false, 0, filepath))
		return;

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fileopen, filepath);
	g_return_if_fail(sdpSucceeded(m_Application->Command(k3dIApplication::Command_OpenDocument, filepath, 0)));
}

// User wants to exit the application ...
void CApplicationFrame::OnFileExit()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fileexit, 0);

	// Kill ourselves!
	bool exit = false;
	if(m_Application->BatchMode())
		exit = true;
	else
		exit = SafeToExit();
		
	if(exit)
		g_return_if_fail(sdpSucceeded(m_Application->Command(k3dIApplication::Command_Close, 0, 0)));
}

void CApplicationFrame::OnToolsBlackBoxRecorder()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolsblackboxrecorder, 0);
	m_Application->BlackBoxRecorder();
}

void CApplicationFrame::OnToolsRecordMacro()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolsrecordmacro, 0);
	m_Application->MacroRecorder();
}

void CApplicationFrame::OnToolsRecordTutorial()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolsrecordtutorial, 0);
	m_Application->TutorialRecorder();
}

void CApplicationFrame::OnToolsEditShader()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolseditshader, 0);
	m_Application->ShaderEditor();
}

void CApplicationFrame::OnToolsEditCommandNodeInspector()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolseditcommandnodeinspector, 0);
	m_Application->CommandNodeInspector();
}

void CApplicationFrame::OnToolsEditScript()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolseditscript, 0);
	m_Application->ScriptEditor();
}

void CApplicationFrame::OnToolsPlayScript()
{
	sdpString filepath;
	if(!k3dGetFilePath("script", "Play K-3D Script:", false, 0, filepath))
		return;

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolsplayscript, 0);

std::cout << filepath << std::endl;
	
	// Make it happen ...
	std::ifstream file(filepath);
	g_return_if_fail(k3d::ExecuteScript(file, filepath));
}

void CApplicationFrame::OnToolsOptions()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_toolsoptions, 0);

	COptionsDialog* dialog = new COptionsDialog(m_Application);
	g_return_if_fail(dialog);
}

void CApplicationFrame::OnHelpTutorials()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_helptutorials, 0);
	m_Application->Tutorials();
}

void CApplicationFrame::OnHelpOnline()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_helponline, 0);
	m_Application->BrowserNavigate("http://www.k-3d.com");
}

void CApplicationFrame::OnHelpAbout()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_helpabout, 0);

	CAboutBox* aboutbox = new CAboutBox();
	g_return_if_fail(aboutbox);
}

void CApplicationFrame::OnOpenTool()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_opentool, "");

	sdpGtkCList list = ToolsList();
	sdpGtkCList::Rows selection = list.GetSelectedRows();

	for(sdpGtkCList::RowIterator row = selection.begin(); row != selection.end(); row++)
		{
			CObjectInfo* const objectinfo = reinterpret_cast<CObjectInfo*>(list.GetRowData(*row));
			g_return_if_fail(objectinfo);
			
			CreateObject(objectinfo);
		}
}

void CApplicationFrame::OnCreateObject()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_createobject, "");

	sdpGtkCList list = ObjectsList();
	sdpGtkCList::Rows selection = list.GetSelectedRows();
	for(sdpGtkCList::RowIterator row = selection.begin(); row != selection.end(); row++)
		{
			CObjectInfo* const objectinfo = reinterpret_cast<CObjectInfo*>(list.GetRowData(*row));
			g_return_if_fail(objectinfo);
			
			k3dRecordStateChangeSet changeset(m_Application->ActiveDocument(), sdpString("Create ") + objectinfo->Name());
			CreateObject(objectinfo);
		}
	
	// Update the display ...
	k3dRedrawAll(m_Application->ActiveDocument(), false);
}

void CApplicationFrame::OnDragDataGet(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	sdpGtkEventWidgetDragDataGet* const event = static_cast<sdpGtkEventWidgetDragDataGet*>(Event);

	// We're going to generate a K3DML document ...	
	sdpxml::Document xmldocument("k3dml");
	
	sdpxml::Element& application = xmldocument.Append(sdpxml::Element("application"));
	sdpxml::Element& objects = application.Append(sdpxml::Element("objects"));
	
	// For each selected object ...	
	sdpGtkCList::Rows selection = ObjectsList().GetSelectedRows();
	for(sdpGtkCList::RowIterator row = selection.begin(); row != selection.end(); row++)
		{
			CObjectInfo* objectinfo = reinterpret_cast<CObjectInfo*>(ObjectsList().GetRowData(*row));
			g_assert(objectinfo);
			
			// Create an XML element for the object ...
			objects.Append(sdpxml::Element("object", "", sdpxml::Attribute("name", objectinfo->Name()), sdpxml::Attribute("class", sdpToString(objectinfo->ClassID()))));
		}

	// Buffer the data (since drag-and-drop is asynchronous) ...
	std::strstream stream;
	stream << xmldocument << std::ends;
	
	m_DNDBuffer = stream.str();
	stream.freeze(false);

	// Hand the data off to the drag-and-drop mechanism ...
	gtk_selection_data_set(event->Selection(), GDK_SELECTION_TYPE_STRING, 8, reinterpret_cast<const guchar*>(m_DNDBuffer.c_str()), m_DNDBuffer.size());
}

void CApplicationFrame::OnToolsClicked(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Get the row we clicked on ...
	int row;
	int column;
	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;
	ToolsList().GetHitInfo(gint(event->Event()->x), gint(event->Event()->y), &row, &column);

	if(-1 == row)
		return;
	
	k3dIObjectInfo* objectinfo = reinterpret_cast<k3dIObjectInfo*>(ToolsList().GetRowData(row));
	g_return_if_fail(objectinfo);
	
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_highlighttool, objectinfo->Name());

	if(event->Event()->button == 1 && event->Event()->type == GDK_2BUTTON_PRESS)
		{
			ToolsList().SelectRow(row, 0);
			OnOpenTool();
		}
}

void CApplicationFrame::OnObjectsClicked(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Get the row we clicked on ...
	int row;
	int column;
	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;
	ObjectsList().GetHitInfo(gint(event->Event()->x), gint(event->Event()->y), &row, &column);

	if(-1 == row)
		return;
		
	k3dIObjectInfo* objectinfo = reinterpret_cast<k3dIObjectInfo*>(ObjectsList().GetRowData(row));
	g_return_if_fail(objectinfo);
	
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_highlightobject, objectinfo->Name());

	if(event->Event()->button == 1 && event->Event()->type == GDK_2BUTTON_PRESS)
		{
			ObjectsList().SelectRow(row, 0);
			OnCreateObject();
		}
}

class CApplicationFrameBookmarks :
	public k3dIObjectInfoVisitor,
	public k3dIBookmarkVisitor
{
public:
	CApplicationFrameBookmarks(sdpGtkCList Tools, sdpGtkCList Objects) :
		m_ObjectInfo(0),
		m_Tools(Tools),
		m_Objects(Objects)
	{
	}

	virtual ~CApplicationFrameBookmarks()
	{
	}

	void VisitObjectInfo(k3dIObjectInfo* const ObjectInfo)
	{
		m_ObjectInfo = ObjectInfo;
		m_ObjectInfo->VisitBookmarks(this);
	}

	void VisitBookmark(SDPCSTRING Bookmark)
	{
		const sdpString bookmark(Bookmark);
		
		const bool incomplete = (SDP_S_OK == m_ObjectInfo->QueryIncomplete());
		const bool experimental = (SDP_S_OK == m_ObjectInfo->QueryExperimental());
		
		sdpString name(m_ObjectInfo->Name());
		if(experimental)
			name += " (Experimental)";
		else if(incomplete)
			name += " (Incomplete)";

		// Setup our colors
		GdkColor experimentalcolor;
		experimentalcolor.red = 0xffff;
		experimentalcolor.green = 0x0000;
		experimentalcolor.blue = 0x0000;	
		gdk_colormap_alloc_color(gdk_colormap_get_system(), &experimentalcolor, FALSE, TRUE);
												
		GdkColor incompletecolor;
		incompletecolor.red = 0x0000;
		incompletecolor.green = 0x0000;
		incompletecolor.blue = 0xffff;
		gdk_colormap_alloc_color(gdk_colormap_get_system(), &incompletecolor, FALSE, TRUE);										
												
		if(0 == bookmark.find("Tools/"))
			{
				const char* labels[] = { name };
				const gint row = m_Tools.Append(labels);
				m_Tools.SetRowData(row, m_ObjectInfo);
				
				if(experimental)
					m_Tools.SetForegroundColor(row, &experimentalcolor);
				else if(incomplete)
					m_Tools.SetForegroundColor(row, &incompletecolor);
			}
		else if(0 == bookmark.find("Objects/"))
			{
				const char* labels[] = { name };
				const gint row = m_Objects.Append(labels);
				m_Objects.SetRowData(row, m_ObjectInfo);
				
				if(experimental)
					m_Objects.SetForegroundColor(row, &experimentalcolor);
				else if(incomplete)
					m_Objects.SetForegroundColor(row, &incompletecolor);
			}
	}
	
private:
	k3dIObjectInfo* m_ObjectInfo;
	sdpGtkCList m_Tools;
	sdpGtkCList m_Objects;
};

void CApplicationFrame::InitializeTrees()
{
	sdpGtkCList tools = ToolsList();
	sdpGtkCList objects = ObjectsList();

	tools.Clear();
	objects.Clear();

	CApplicationFrameBookmarks visitor(tools, objects);
	k3dApplication()->Objects()->VisitObjectInfo(&visitor);

	tools.Sort();
	objects.Sort();
}

void CApplicationFrame::UpdateControls()
{
	// If there isn't an active document, disable the trees ...
	if(m_Application->m_ActiveDocument)
		{
			ToolsList().SetSensitive(true);
			ObjectsList().SetSensitive(true);
			Button(control_opentool).SetSensitive(true);
			Button(control_createobject).SetSensitive(true);
		}
	else
		{
			ToolsList().SetSensitive(false);
			ObjectsList().SetSensitive(false);
			Button(control_opentool).SetSensitive(false);
			Button(control_createobject).SetSensitive(false);
		}
		
	CApplicationFrameBase::UpdateControls();
}

sdpResult CApplicationFrame::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	if(Name == control_highlighttool)
		{
			// Get an object info from our arguments
			std::vector<k3dIObjectInfo*> objectinfocollection;
			k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(objectinfocollection), k3d::SameName(Arguments));
			g_return_val_if_fail(1 == objectinfocollection.size(), SDP_E_FAIL);
			
			k3dIObjectInfo* const objectinfo = *objectinfocollection.begin();

			// Find a row that matches ...
			sdpGtkCList list = ToolsList();
			gint row = list.FindRowFromData(objectinfo);

			// Move the pointer to the row ...
			list.InteractiveShow(k3dOptionsGetTutorialSpeed(m_Application->Options()), true);
			list.InteractiveWarpPointer(row, k3dOptionsGetTutorialSpeed(m_Application->Options()), true);
			list.SelectRow(row, 0);
			m_Application->TutorialMouseDisplay("Highlight Tool:", k3dIApplication::LMB_CLICK, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == control_highlightobject)
		{
			// Get an object info from our arguments
			std::vector<k3dIObjectInfo*> objectinfocollection;
			k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(objectinfocollection), k3d::SameName(Arguments));
			g_return_val_if_fail(1 == objectinfocollection.size(), SDP_E_FAIL);
			
			k3dIObjectInfo* const objectinfo = *objectinfocollection.begin();

			// Find a row that matches ...
			sdpGtkCList list = ObjectsList();
			gint row = list.FindRowFromData(objectinfo);

			// Move the pointer to the row ...
			list.InteractiveShow(k3dOptionsGetTutorialSpeed(m_Application->Options()), true);
			list.InteractiveWarpPointer(row, k3dOptionsGetTutorialSpeed(m_Application->Options()), true);
			list.SelectRow(row, 0);
			m_Application->TutorialMouseDisplay("Highlight Object:", k3dIApplication::LMB_CLICK, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == control_helptutorials)
		{
			m_Application->Tutorials();
			return SDP_S_OK;
		}
		
	return CApplicationFrameBase::Command(Name, Arguments, Result);
}

// Prompt the user to see if it's safe to exit ...
bool CApplicationFrame::SafeToExit()
{
	if(0 == m_Application->m_Documents.size())
		return true;

	// If all documents have been saved, then it's safe to close
	bool all_documents_saved = true;
	for(CApplication::DocumentIterator document = m_Application->m_Documents.begin(); document != m_Application->m_Documents.end(); document++)
		{
			if((*document)->StateRecorder()->UnsavedChanges())
				all_documents_saved = false;
		}

	if(all_documents_saved)
		return true;

	sdpStringArray buttons;
	buttons.push_back("Yes");
	buttons.push_back("No");
	if(1 == m_Application->Query("Exit K-3D?  Unsaved changes in all open documents will be lost (No Undo)", "Exit K-3D:", 1, buttons))
		return true;
		
	return false;
}

void CApplicationFrame::CreateObject(k3dIObjectInfo* const ObjectInfo)
{
	// Sanity checks ...
	g_assert(m_Application->ActiveDocument());
	g_assert(ObjectInfo);

	k3dICommandNode* const commandnode = dynamic_cast<k3dICommandNode*>(m_Application->ActiveDocument());
	g_return_if_fail(commandnode);

	sdpIUnknown* result = 0;
		{
			k3d::Editing editing(m_Application->ActiveDocument()->Objects());
			g_return_if_fail(sdpSucceeded(commandnode->Command(k3dIDocument::Command_CreateObject, ObjectInfo->Name(), &result)));
			g_return_if_fail(result);
		}

	k3dIObject* const object = dynamic_cast<k3dIObject*>(result);
	g_return_if_fail(object);
	
	object->EditObject();
}

