#ifndef APPLICATIONFRAME_H
#define APPLICATIONFRAME_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CApplicationFrame class, which implements the per-application UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>

// Forward class references ...
class CApplication;
class CScriptEditor;
class CScriptRecorder;
class k3dIEventSource;
class k3dIObjectInfo;

////////////////////////////////////////////////////////////////////////////
// CAboutBox

typedef k3dDialog CAboutBoxBase;

/// Creates a modal "about" box that displays program version and copyright
class CAboutBox :
	public CAboutBoxBase
{
public:
	CAboutBox();

	static const sdpString dialogname;

protected:
	void OnEvent(sdpGtkEvent* Event);
	void OnTest();
};

////////////////////////////////////////////////////////////////////////////
// CSplashBox

typedef k3dDialog CSplashBoxBase;

/// Creates a "splash" screen while the program is loading
class CSplashBox :
	public CSplashBoxBase
{
public:
	CSplashBox();

	static const sdpString dialogname;
};

////////////////////////////////////////////////////////////////////////////
// CApplicationFrame

typedef k3dDialog CApplicationFrameBase;

/// Implements the application user interface
class CApplicationFrame :
	public CApplicationFrameBase
{
public:
	CApplicationFrame(CApplication* const Application);
	~CApplicationFrame();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// k3dICommandNode overrides
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	static const sdpString dialogname;	
					
private:
	/// Overrides sdpGtkObjectContainer to handle custom events
	void OnEvent(sdpGtkEvent* Event);
	/// Called to see if its OK to delete the frame
	void OnDelete(sdpGtkEvent* Event);
	/// Called when the user clicks on the tools tree
	void OnToolsClicked(sdpGtkEvent* Event);
	/// Called when the user clicks on the objects tree
	void OnObjectsClicked(sdpGtkEvent* Event);
	/// Called when the user wants to create a new document
	void OnFileNew();
	/// Called when the user wants to open an existing document
	void OnFileOpen();
	/// Called to exit the program gracefully
	void OnFileExit();
	/// Called when the user wants to start the black box recorder for troubleshooting
	void OnToolsBlackBoxRecorder();
	/// Called when the user wants to record a macro (non-interactive script)
	void OnToolsRecordMacro();
	/// Called when the user wants to record a tutorial (interactive script)
	void OnToolsRecordTutorial();
	/// Called when the user wants to edit / compile shaders
	void OnToolsEditShader();
	/// Called when the user wants to open a command node inspector window
	void OnToolsEditCommandNodeInspector();
	/// Called when the user wants to edit / test scripts
	void OnToolsEditScript();
	/// Called when the user wants to play a script
	void OnToolsPlayScript();
	/// Called when the user wants to modify the program options
	void OnToolsOptions();
	/// Called to display available interactive tutorials
	void OnHelpTutorials();
	/// Called to open the K-3D website, http://www.k-3d.com
	void OnHelpOnline();
	/// Called to display an "about" box with program version and copyright information
	void OnHelpAbout();
	/// Called by the user to open a tool
	void OnOpenTool();
	/// Called by the user to create an instance of a K-3D object
	void OnCreateObject();
	/// Called when the user completes a drag-and-drop operation
	void OnDragDataGet(sdpGtkEvent* Event);

	void InitializeTrees();
	void UpdateControls();

	/// Prompt the user to see if it's safe to go exit the program
	bool SafeToExit();

	/// Returns the tools tree widget
	sdpGtkCList ToolsList() { return CList("tools"); }
	/// Returns the objects tree widget
	sdpGtkCList ObjectsList() { return CList("objects"); }

	/// Stores a pointer to the main application
	CApplication* const m_Application;

	/// Provides a storage buffer for data being transmitted via drag-and-drop
	sdpString m_DNDBuffer;

	/// Standard object-creation
	void CreateObject(k3dIObjectInfo* const ObjectInfo);

	// Events handled by this window
	static const sdpString control_filenew;
	static const sdpString control_fileopen;
	static const sdpString control_fileexit;
	static const sdpString control_toolsoptions;
	static const sdpString control_toolsblackboxrecorder;
	static const sdpString control_toolsrecordmacro;
	static const sdpString control_toolsrecordtutorial;
	static const sdpString control_toolseditshader;
	static const sdpString control_toolseditcommandnodeinspector;
	static const sdpString control_toolseditscript;
	static const sdpString control_toolsplayscript;
	static const sdpString control_helptutorials;
	static const sdpString control_helponline;
	static const sdpString control_helpabout;

	static const sdpString control_highlighttool;
	static const sdpString control_highlightobject;
	static const sdpString control_opentool;
	static const sdpString control_createobject;
	
	static const sdpString control_dragbegin;
	static const sdpString control_dragmotion;
	static const sdpString control_dragdataget;
	static const sdpString control_dragdatadelete;
	static const sdpString control_dragdrop;
	static const sdpString control_dragend;
};

#endif // APPLICATIONFRAME_H
