// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBlackBoxRecorder class, which records user actions in realtime for troubleshooting purposes
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "blackboxrecorder.h"

#include <k3dcom.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3discriptengine.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>

#include <strstream>

//////////////////////////////////////////////////////////////////////
// CBlackBoxRecorder

const sdpString CBlackBoxRecorder::dialogname = "blackboxrecorder";
const sdpString CBlackBoxRecorder::control_cancel = "cancel";

CBlackBoxRecorder::CBlackBoxRecorder(CApplication* Application, SDPCSTRING OutputFilePath) :
	CBlackBoxRecorderBase(Application, dialogname, new k3dOptionsWindowGeometryStore(Application, 0)),
	m_Application(Application),
	m_OutputFilepath(OutputFilePath)
{
	g_return_if_fail(LoadGTKMLTemplate("blackboxrecorder.gtkml"));

	// Sanity checks ...
	g_assert(m_Application);
	
	m_Application->AddEventObserver(this, k3dIApplication::Event_Close, 0);

	// Try opening the file ...
	m_OutputFile.open(m_OutputFilepath);
	g_return_if_fail(m_OutputFile.good());

	// Connect to the scripting interface ...
	m_ScriptEngine = k3dApplication()->ScriptEngine(k3d::classid::JavaScriptEngine);
	g_return_if_fail(m_ScriptEngine);

	// Register ourselves with the App as a consumer of command events ...
	g_return_if_fail(sdpSucceeded(k3dApplication()->RegisterCommandRecorder(this)));

	Show();
}

CBlackBoxRecorder::~CBlackBoxRecorder()
{
	m_Application->RemoveEventObserver(this);
	m_Application->UnRegisterCommandRecorder(this);
	m_ScriptEngine = 0;
}

//////////////////////////////////////////////////////////////////////
// k3dICommandRecorder overrides

sdpResult CBlackBoxRecorder::IsRecording()
{
	return SDP_S_OK;
}

sdpResult CBlackBoxRecorder::RecordCommand(k3dICommandNode* CommandNode, k3dICommandNode::CommandType Type, SDPCSTRING Name, SDPCSTRING Arguments)
{
	// We only want user interface commands ... !
	if(Type != k3dICommandNode::USERINTERFACE)
		return SDP_S_FALSE;

	// Sanity checks ...
	g_return_val_if_fail(CommandNode, SDP_E_FAIL);
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	sdpString buffer;
	gulong buffersize = 1024;
	m_ScriptEngine->ConvertCommand(CommandNode, Name, Arguments, buffer.buffer(buffersize), buffersize);
	buffer.resize(buffersize);

	std::cout << buffer << std::endl;
	std::cout.flush();

	m_OutputFile << buffer << std::endl;
	m_OutputFile.flush();

	return SDP_S_OK;
}

void CBlackBoxRecorder::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_cancel)
		OnCancel();
	else
		CBlackBoxRecorderBase::OnEvent(Event);
}

void CBlackBoxRecorder::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Turn this into a cancel ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	OnCancel();
}

void CBlackBoxRecorder::OnCancel()
{
	delete this;
}

sdpResult CBlackBoxRecorder::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIApplication::Event_Close)
		{
			OnCancel();
			return SDP_S_OK;
		}
	else
		return CBlackBoxRecorderBase::ObserveEvent(Source, Event, EventData, ObserverData);
}
