// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CClipboard class, which acts as a data source for the primary X selection
		\author Tim Shead (tshead@k-3d.com)
*/

#include "clipboard.h"
#include <sdpgtkevents.h>

#ifdef	SDPWIN32
#include <windows.h>
#endif

//////////////////////////////////////////////////////////////////////
// CClipboard

CClipboard::CClipboard() :
	CClipboardBase(),
	m_HaveSelection(false)
{
	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<window type=\"toplevel\" show=\"false\">"
				"<event signal=\"selection-get\" name=\"selectionget\"/>"
				"<event signal=\"selection-clear-event\" name=\"selectionclear\"/>"
			"</window>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, ""));
	
	// We want to handle X selections ...
	gtk_selection_add_target(RootWidget(), gdk_atom_intern("CLIPBOARD", false), GDK_SELECTION_TYPE_STRING, 0);
}

CClipboard::~CClipboard()
{
	RootWidget().Destroy();
}

void CClipboard::SetText(SDPCSTRING Text)
{
	m_Text = Text;
}

bool CClipboard::GrabSelection()
{
#ifdef SDPWIN32

	m_HaveSelection = false;

	g_return_val_if_fail(OpenClipboard(0), false);
	g_return_val_if_fail(EmptyClipboard(), false);

	HGLOBAL clipbuffer = GlobalAlloc(GMEM_DDESHARE, m_Text.size() + 1);
	g_return_val_if_fail(clipbuffer, false);

	char* buffer = static_cast<char*>(GlobalLock(clipbuffer));
	g_return_val_if_fail(buffer, false);
	
	strcpy(buffer, m_Text.c_str());
	GlobalUnlock(clipbuffer);

	g_return_val_if_fail(SetClipboardData(CF_TEXT, buffer), false);
	g_return_val_if_fail(CloseClipboard(), false);

	m_HaveSelection = true;
						
#else // SDPWIN32

	m_HaveSelection = gtk_selection_owner_set(RootWidget(), gdk_atom_intern("CLIPBOARD", false), GDK_CURRENT_TIME) ? true : false;

#endif // !SDPWIN32

	return HaveSelection();
}

bool CClipboard::HaveSelection() const
{
	return m_HaveSelection;
}

void CClipboard::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "selectionget")
		OnSelectionGet(Event);
	else if(Event->Name() == "selectionclear")
		OnSelectionClear(Event);
	else		
		CClipboardBase::OnEvent(Event);
}

void CClipboard::OnSelectionGet(sdpGtkEvent* Event)
{
	sdpGtkEventWidgetSelectionGet* event = static_cast<sdpGtkEventWidgetSelectionGet*>(Event);
	gtk_selection_data_set(event->Selection(), GDK_SELECTION_TYPE_STRING, 8, reinterpret_cast<const guchar*>(m_Text.c_str()), m_Text.size());
}

void CClipboard::OnSelectionClear(sdpGtkEvent* Event)
{
	m_HaveSelection = false;
}


