// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CMessageBox informational dialog
		\author Tim Shead (tshead@k-3d.com)
		\author Dan Erikson <derikson@montana.com>
*/

#include "dialogs.h"

#include <k3diapplication.h>

#include <sdpgtkevent.h>
#include <sdpgtkevents.h>

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

//////////////////////////////////////////////////////////////////////
// CMessageBox

const sdpString CMessageBox::dialogname         = "messagebox";
const sdpString CMessageBox::control_buttonbox  = "buttonbox";

CMessageBox::CMessageBox(k3dIApplication* const Application, SDPCSTRING Message, SDPCSTRING Title, guint DefaultButton, const sdpStringArray Buttons, SDPCSTRING DialogTemplate) :
	CMessageBoxBase(Application, dialogname, 0),
	m_Result(0),
	m_LastButton(0)
{
	// Sanity checks ...
	g_assert_string(DialogTemplate);

	// create the dialog box
	g_return_if_fail(LoadGTKMLTemplate(DialogTemplate));

	// Set our window title
	RootWindow().SetTitle(Title);
	// Set the dialog message
	Label("message").SetText(Message);

	// Center it ...
	gtk_window_set_position(RootWindow(), GTK_WIN_POS_CENTER);

	AddButtons(Buttons);

	// if there was a default button specified, then make it default
	if(DefaultButton)
		{
			sdpString defaultname = "button";
			defaultname += sdpToString(DefaultButton);
			Button(defaultname).GrabDefault();
		}
	Show();
}

guint CMessageBox::AddButton(SDPCSTRING Label)
{
	// if there is no label specified, then just return
	if(!sdpString(Label).size())
		return 0;

	// the name for the button will be "button" + the button number.  button1, button2, etc.
	const sdpString name = "button" + sdpToString(++m_LastButton);

	sdpGtkButton button;
	button.Create(Label);
	// make sure the button can default
	GTK_WIDGET_SET_FLAGS(button.Object(), GTK_CAN_DEFAULT);
	Container(control_buttonbox).Attach(button);
	MapObject(name, button.Object());

	// if the button is clicked, issue an event with the button's name as the event name
	MapEvent("clicked", name, false, button, true);

	// add a command with the button's name as the command name
	AddCommand(name, name, CONTROL_BUTTON);

	button.Show();
	return m_LastButton;
}

void CMessageBox::AddButtons(const sdpStringArray Buttons)
{
	// Create the buttons
	for(sdpStringArray::const_iterator button = Buttons.begin(); button != Buttons.end(); button++)
		AddButton(*button);
}

guint CMessageBox::Run()
{
	DoModal();
	guint result = m_Result;
	delete this;
	return result;
}

void CMessageBox::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	const sdpString name(Event->Name());
	const sdpString signal(Event->Signal());

	// if the name of the command starts with "button", then the user has clicked one of our buttons
	if(0 == name.find("button"))
		{
			// set m_Result to be the button number that was clicked
			m_Result = sdpFromString(name.right(name.size()-6), 0);

			// no longer modal.  we can return a value from Run() now.
			CancelModal();
		}
	else		
		CMessageBoxBase::OnEvent(Event);
}

void CMessageBox::OnDelete(sdpGtkEvent* Event)
{
	// we want Run() to return 0
	m_Result = 0;
	CancelModal();
}
