
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDocument class, which encapsulates an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "documentframe.h"
#include "hierarchy.h"
#include "objectinfo.h"
#include "objectcollection.h"
#include "timestorage.h"

#include <k3ddependencies.h>
#include <k3dfilefilters.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dobjectinfocollection.h>

#include <k3dianimate.h>
#include <k3difileformat.h>
#include <k3digeometryreadformat.h>
#include <k3digeometrywriteformat.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3dimaterial.h>
#include <k3dimouseeventobserver.h>
#include <k3diobjectinitializer.h>
#include <k3dipersist.h>
#include <k3distate.h>
#include <k3ditimeserver.h>
#include <k3ditransform.h>

#include <fstream>

#include <time.h>

#ifndef	SDPWIN32
#include <pwd.h>
#include <unistd.h>
#endif // !SDPWIN32

/////////////////////////////////////////////////////////////////////////////
// CDocument

gulong CDocument::m_DocumentNumber = 1;

CDocument::CDocument(CApplication* const Application, k3dIStateRecorder* const StateRecorder) :
	k3dCommandNode(Application, "document"),
	m_Hierarchy(new CHierarchy(this)),
	m_Objects(new CObjectCollection(this, m_Hierarchy)),
	m_Application(Application),
	m_MouseFocus(0),
	m_Creator("unknown"),
	m_CreationHost("localhost"),
	m_CreationDomain("localdomain"),
	m_Dying(false),
	m_PlaybackMode(STOP),
	m_AnimationTimer(0),
	m_IgnoreAnimationTimer(false),
	m_StateRecorder(StateRecorder),
	m_Time(new CTimeStorage(StateRecorder))
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocument::CDocument()" << std::endl;
	
	// Sanity checks ...
	g_assert(m_Application);
	g_assert(m_StateRecorder);

#ifndef	SDPWIN32
	// Record our creator ("oooohhhhhh!") ...
	passwd* const password = getpwuid(getuid());
	if(password)
		{
			if(password->pw_name)
				if(strlen(password->pw_name))
					m_Creator = password->pw_name;
		}
	
	// Record our creation host and domain ...
	gethostname(m_CreationHost.buffer(1024), 1024);
	m_CreationHost.resize(strlen(m_CreationHost));
	
//	getdomainname(m_CreationDomain.buffer(1024), 1024);
//	m_CreationDomain.resize(strlen(m_CreationDomain));
#endif

	// Record our creation date and time ...
	g_date_set_time(&m_CreationDate, time(0));
	g_get_current_time(&m_CreationTime);

	// Handle application events ...
	m_Application->AddEventObserver(this, k3dIApplication::Event_CreateDocument, "CDocument");
	m_Application->AddEventObserver(this, k3dIApplication::Event_OpenDocument, "CDocument");
	m_Application->AddEventObserver(this, k3dIApplication::Event_CloseDocument, "CDocument");
}

CDocument::~CDocument()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocument::~CDocument() " << m_DocumentTitle << std::endl;

	// Cancel application events ...
	m_Application->RemoveEventObserver(this);
	
	Notify(k3dIDocument::Event_Close, 0);
	
	// Update our "dying" flag ...
	m_Dying = true;

	// Kill the animation timer ...
	if(m_AnimationTimer)
		gtk_timeout_remove(m_AnimationTimer);

	/** \todo Figure out why we dump core if m_Objects is destroyed last */
	delete m_Objects;

	delete m_Time;
	delete m_StateRecorder;
	delete m_Hierarchy;
}

// Create a new, default document from scratch ...
bool CDocument::Create(SDPCSTRING TemplateDocumentPath)
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocument::Create()" << std::endl;
	
	// Set our default title ...
	m_DocumentTitle = "Untitled Document " + sdpToString(m_DocumentNumber++);

	// If we have a template document, load it ...
	const sdpString templatedocumentpath(TemplateDocumentPath);
	if(templatedocumentpath.size())
		{
			g_return_val_if_fail(Load(templatedocumentpath), false);
			k3dRedrawAll(this, false);
		}

	// Create the document frame ...
	CDocumentFrame* frame = new CDocumentFrame(this);
	g_return_val_if_fail(frame, false);

	// Create our animation timer ...
	m_AnimationTimer = gtk_timeout_add(33, RawAnimationTimer, gpointer(this));

	return true;
}

// Open a document from disk ...
bool CDocument::Open(SDPCSTRING DocumentPath)
{
	// Sanity checks ...
	g_assert_string(DocumentPath);

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocument::Open() loading " << DocumentPath << std::endl;

	// Load the document ...
	g_return_val_if_fail(Load(DocumentPath), false);

	// Record our path & title ...
	m_DocumentPath = DocumentPath;
	m_DocumentTitle = sdpFile(DocumentPath);

	// Force a display update ...
	k3dRedrawAll(this, false);

	// Create the document frame ...
	CDocumentFrame* frame = new CDocumentFrame(this);
	g_return_val_if_fail(frame, false);

	// Create our animation timer ...
	m_AnimationTimer = gtk_timeout_add(33, RawAnimationTimer, gpointer(this));

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocument::Open() done loading " << DocumentPath << std::endl;

	return true;
}

bool CDocument::Load(SDPCSTRING DocumentPath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(DocumentPath, false);

	// Try loading the file as XML ...
	sdpxml::Document doc("k3dml");
	g_return_val_if_fail(doc.Load(DocumentPath), false);

	// Make sure it's a K3D document ...
	g_return_val_if_fail(doc.Name() == "k3dml", false);

	// Look for the document element ...
	sdpxml::ElementPointer const document = sdpxml::FindElement(doc, sdpxml::SameName("document"));
	if(0 == document)
		return true;

	// Load away ...
	m_Objects->Load(doc, *document);
	m_Hierarchy->Load(m_Objects, *document);
	m_Time->Load(doc, *document);

	// Let all our objects know we're done loading ...
	m_Objects->LoadComplete();
	m_Time->LoadComplete();
		
	// Let all our object know that they're initialized ...
	m_Objects->InitComplete();
	
	return true;
}

bool CDocument::IsDying()
{
	return m_Dying;
}

SDPCSTRING CDocument::Title()
{
	return m_DocumentTitle;
}

///////////////////////////////////////////////////////////////
// k3dIDocument implementation

//////////////////////////////////////////////////////////////////////////////////////////
// k3dIDocumentMetaData implementation ...

SDPCSTRING CDocument::Creator()
{
	return m_Creator;
}

SDPCSTRING CDocument::CreationHost()
{
	return m_CreationHost;
}

SDPCSTRING CDocument::CreationDomain()
{
	return m_CreationDomain;
}

sdpResult CDocument::CreationDate(GDate* const Date)
{
	// Sanity checks ...
	g_return_val_if_fail(Date, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CDocument::CreationTime(GTimeVal* const Time)
{
	// Sanity checks ...
	g_return_val_if_fail(Time, SDP_E_FAIL);

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////////////////
// k3dICommandNode implementation

sdpResult CDocument::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	k3dRecordCommand(this, k3dICommandNode::SCRIPTING, Name, Arguments);
	
	const sdpString name(Name);
	const sdpString arguments(Arguments);
	
	if(name == Command_CreateObject)
		{
			std::vector<k3dIObjectInfo*> info;
			k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(info), k3d::SameName(Arguments));
			if(0 == info.size())
				{
					std::cerr << "No object class named " << Arguments << std::endl;
					g_return_val_if_fail(0, SDP_E_FAIL);
				}
			else if(info.size() > 1)
				{
					std::cerr << "Multiple object classes named " << Arguments << std::endl;
					g_return_val_if_fail(0, SDP_E_FAIL);
				}
			
			k3dIObject* const object = m_Objects->CreateObject((*info.begin())->ClassID());
			g_return_val_if_fail(object, SDP_E_FAIL);
			
			if(Result)
				*Result = object;
				
			return SDP_S_OK;
		}
	else if(name == Command_DeleteObject)
		{
			std::vector<k3dIObject*> objects;
			k3d::copy_if(Objects(), std::back_inserter(objects), k3d::SameName(arguments));
			g_return_val_if_fail(1 == objects.size(), SDP_E_FAIL);
			
			return m_Objects->DeleteObject(*objects.begin());
		}
	else if(name == Command_Save)
		{
			return Save(Arguments);
		}
	else if(name == Command_Close)
		{
			return k3dApplication()->CloseDocument(this);
		}
		
	return k3dCommandNode::Command(Name, Arguments, Result);
}

//////////////////////////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CDocument::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	const sdpString clientdata(ObserverData);
	
	// This little bit of trickery ensures that documents always have a unique UI node name ...
	if(event == k3dIApplication::Event_CreateDocument || event == k3dIApplication::Event_OpenDocument)
		{
			sdpString name = CommandNodeName();
			SetCommandNodeName(name + "*");
			return SDP_S_OK;
		}
	else if(event == k3dIApplication::Event_CloseDocument)
		{
			sdpString name = CommandNodeName();
			if(name[name.size()-1] == '*')
				SetCommandNodeName(name.left(name.size() - 1));
				
			return SDP_S_OK;
		}

	return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
}

// Called by the animation interval timer
gint CDocument::RawAnimationTimer(gpointer Data)
{
	// Call the object-instance version ...
	return ((CDocument*)Data)->AnimationTimer();
}

// Called by the animation interval timer
gint CDocument::AnimationTimer()
{
	// We only pay attention to every-other timer signal ...
	m_IgnoreAnimationTimer = !m_IgnoreAnimationTimer;
	if(m_IgnoreAnimationTimer)
		return true;

	switch(m_PlaybackMode)
		{
			case REVERSEPLAY:
				{
					// Calculate our new time ...
					gdouble newtime = m_Time->ViewTime() - m_Time->FrameDelta();
					if(newtime < m_Time->StartTime())
						{
							// We passed the beginning of the animation, so stop ...
							newtime = m_Time->StartTime();
							SetPlaybackMode(STOP);
						}

					// Set the new time ...
					m_Time->SetViewTime(newtime);

					// Redraw everything ...
					k3dRedrawAll(this, false);
				}
				break;

			case STOP:
				break;

			case PLAY:
				{
					// Calculate our new time ...
					gdouble newtime = m_Time->ViewTime() + m_Time->FrameDelta();
					if(newtime > m_Time->EndTime())
						{
							// We passed the end of the animation, so stop ...
							newtime = m_Time->EndTime();
							SetPlaybackMode(STOP);
						}

					// Set the new time ...
					m_Time->SetViewTime(newtime);

					// Redraw everything ...
					k3dRedrawAll(this, false);
				}
				break;

			case LOOPPLAY:
				{
					// Calculate our new time ...
					gdouble newtime = m_Time->ViewTime() + m_Time->FrameDelta();
					if(newtime > m_Time->EndTime())
						newtime = m_Time->StartTime();

					// Set the new time ...
					m_Time->SetViewTime(newtime);

					// Redraw everything ...
					k3dRedrawAll(this, false);
				}
				break;

			default:
				g_return_val_if_fail(0, true);
		}

	// Continue indefinitely ...
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// k3dIDocument implementation

k3dIObjectCollection* CDocument::Objects()
{
	return m_Objects;
}

k3dIHierarchy* CDocument::Hierarchy()
{
	return m_Hierarchy;
}

k3dIStateRecorder* CDocument::StateRecorder()
{
	return m_StateRecorder;
}

k3dITime* CDocument::Time()
{
	return m_Time;
}

SDPCSTRING CDocument::Path()
{
	return m_DocumentPath;
}

sdpResult CDocument::Import(SDPCSTRING FilePath, k3dIGeometryReadFormat* const FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_fail(FormatFilter, SDP_E_FAIL);

	// Give the format a chance to look at the file, first ...
	sdpResult result = FormatFilter->PreRead(FilePath);
	if(SDP_S_OK == result)
		{
			// Prompt the user for any options before reading ...
			result = FormatFilter->ReadOptions(FilePath);
			if(SDP_S_OK == result)
				{
					// Read the file ...
					result = FormatFilter->ReadFile(FilePath);
				}
		}

	return result;
}

sdpResult CDocument::Export(SDPCSTRING FilePath, k3dIGeometryWriteFormat* const FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_fail(FormatFilter, SDP_E_FAIL);

	// Give the format a chance to look at the file, first ...
	sdpResult result = FormatFilter->PreWrite(FilePath);
	if(SDP_S_OK == result)
		{
			// Prompt the user for any options before writing ...
			result = FormatFilter->WriteOptions(FilePath);
			if(SDP_S_OK == result)
				{
					// Write the file ...
					result = FormatFilter->WriteFile(FilePath);
				}
		}

	return result;
}

// Save the document to the given path ...
sdpResult CDocument::Save(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Try to open the file ...
	std::ofstream filestream(FilePath);
	g_return_val_if_fail(filestream.good(), SDP_E_FAIL);

	// Create our output document and dependencies objects ...
	sdpxml::Document xmldocument("k3dml");
	k3dDependencies dependencies;

	// Application-wide stuff ...
	xmldocument.Append(sdpxml::Element("application"));

	// The document itself ...
	sdpxml::Element& document = xmldocument.Append(sdpxml::Element("document"));

	// Time-related data ...
	g_return_val_if_fail(sdpSucceeded(m_Time->Save(document, &dependencies)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_Time->SaveComplete()), SDP_E_FAIL);
	
	// Save objects ...
	m_Objects->Save(document, &dependencies);
	m_Objects->SaveComplete();
	
	// Save the hierarchy ...
	m_Hierarchy->Save(document, &dependencies);

	// Save the XML ...
	filestream << xmldocument << std::endl;
	g_return_val_if_fail(filestream.good(), SDP_E_FAIL);
	
	// Record our new path & title ...
	m_DocumentPath = FilePath;
	m_DocumentTitle = sdpFile(FilePath);
	Notify(k3dIDocument::Event_TitleChanged, m_DocumentTitle);
	
	// Record save position in the undo/redo stack
	StateRecorder()->MarkSaved();

	return SDP_S_OK;
}

void CDocument::SetMouseFocus(k3dIMouseEventObserver* const Observer)
{
	if(Observer != m_MouseFocus)
		{
			m_MouseFocus = Observer;
			Notify(Event_MouseFocusChanged, 0);
		}
}

k3dIMouseEventObserver* CDocument::MouseFocus()
{
	return m_MouseFocus;
}

CDocument::PlaybackModeType CDocument::PlaybackMode()
{
	return m_PlaybackMode;
}

void CDocument::SetPlaybackMode(PlaybackModeType Mode)
{
	m_PlaybackMode = Mode;
	Notify(k3dIDocument::Event_PlaybackModeChanged, 0);
}


