#ifndef DOCUMENT_H
#define DOCUMENT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CDocument class, which encapsulates an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcommandnode.h>
#include <k3deventsource.h>
#include <k3deventobserver.h>
#include <k3dvariable.h>

#include <k3didocument.h>
#include <k3didocumentmetadata.h>
#include <k3ditime.h>

// Forward class declarations
class CApplication;
class CHierarchy;
class CObjectCollection;
class CObjectInfo;
class CTimeStorage;

class k3dIDependencies;
class k3dIStateRecorder;

class sdpxml::Element;

/////////////////////////////////////////////////////////////////////
// CDocument

/// Encapsulates an open K-3D document
class CDocument :
	public k3dIDocument,
	public k3dIDocumentMetaData,
	public k3dCommandNode,
	public k3dEventSource,
	public k3dEventObserver
{
friend class CAnimationParametersDialog;
friend class CAnimationTotalFramesDataSource;
friend class CAnimationStartFrameDataSource;
friend class CAnimationViewFrameDataSource;
friend class CAnimationEndFrameDataSource;
friend class CDocumentFrame;
friend class CObjectContext;

public:
	CDocument(CApplication* const Application, k3dIStateRecorder* const StateRecorder);
	~CDocument();

	// k3dIDocument implementation
	k3dIObjectCollection* Objects();
	k3dIHierarchy* Hierarchy();
	k3dIStateRecorder* StateRecorder();
	k3dITime* Time();
	
	SDPCSTRING Path();
	sdpResult Import(SDPCSTRING FilePath, k3dIGeometryReadFormat* const FormatFilter);
	sdpResult Export(SDPCSTRING FilePath, k3dIGeometryWriteFormat* const FormatFilter);
	sdpResult Save(SDPCSTRING FilePath);

	void SetMouseFocus(k3dIMouseEventObserver* const Observer);
	k3dIMouseEventObserver* MouseFocus();

	// k3dIDocumentMetaData implementation
	SDPCSTRING Creator();
	SDPCSTRING CreationHost();
	SDPCSTRING CreationDomain();
	sdpResult CreationDate(GDate* const Date);
	sdpResult CreationTime(GTimeVal* const Time);

	// k3dICommandNode implementation
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	/// Called to create a new document
	bool Create(SDPCSTRING TemplateDocumentPath);
	/// Called to open an existing document from an XML file
	bool Open(SDPCSTRING DocumentPath);

	/// Returns "true" if the document is in the process of closing itself
	bool IsDying();

	/// Stores a hierarchical view of document objects
	CHierarchy* const m_Hierarchy;
	/// Stores document objects ...
	CObjectCollection* const m_Objects;

	/// Returns the title of this document
	SDPCSTRING Title();

	/// Stores a pointer to the owning application
	CApplication* const m_Application;

	/// Stores the currently active object
	k3dIMouseEventObserver* m_MouseFocus;

	/// Enumerates the document's animation playback modes
	typedef enum
	{
		REVERSEPLAY,
		STOP,
		PLAY,
		LOOPPLAY,
	} PlaybackModeType;

	/// Returns the current animation playback mode
	PlaybackModeType PlaybackMode();
	/// Sets the current animation playback mode
	void SetPlaybackMode(PlaybackModeType Mode);

private:
	/// Stores the document creator
	sdpString m_Creator;
	/// Stores the host the document was created on
	sdpString m_CreationHost;
	/// Stores the domain the document was created on
	sdpString m_CreationDomain;
	/// Stores the date the document was created on
	GDate m_CreationDate;
	/// Stores the time the document was created
	GTimeVal m_CreationTime;

	/// Stores the full document filepath
	sdpString m_DocumentPath;
	/// Stores the document title
	sdpString m_DocumentTitle;
	/// Stores a number for uniquely naming new documents
	static gulong m_DocumentNumber;

	/// Set to "true" if the document is in the process of going-away
	bool m_Dying;

	/// Stores the current animation playback mode
	PlaybackModeType m_PlaybackMode;

	/// Called by the animation interval timer
	static gint RawAnimationTimer(gpointer Data);
	/// Called by the animation interval timer
	gint AnimationTimer();
	/// Stores the animation timer handler id
	guint m_AnimationTimer;
	/// Used to ignore every other timer signal (so the display can keep up)
	bool m_IgnoreAnimationTimer;

	/// Records changes made by the user for Undo / Redo purposes
	k3dIStateRecorder* const m_StateRecorder;
	/// Stores time-related data
	CTimeStorage* m_Time;

	/// Loads an XML document
	bool Load(SDPCSTRING Path);
};

#endif // DOCUMENT_H
