
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDocumentFrame class, which encapsulates an open K-3D document's UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include "animationparameters.h"
#include "application.h"
#include "document.h"
#include "documentframe.h"
#include "hierarchyframe.h"
#include "hierarchycontrol.h"

#include <k3dfilefilters.h>
#include <k3dfilterselector.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>
#include <k3dpathcollection.h>

#include <k3difileformat.h>
#include <k3digeometryreadformat.h>
#include <k3digeometrywriteformat.h>
#include <k3dimesh.h>
#include <k3dipath.h>
#include <k3dipoint.h>
#include <k3distate.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

namespace
{

class SmoothShade
{
public:
	SmoothShade(const gulong State) :
		m_State(State)
	{
	}

	void operator()(k3dIMesh* const Mesh)
	{
		Mesh->ShadePaths(m_State, 0.0);
	}

private:
	const gulong m_State;
};

class FlatShade
{
public:
	FlatShade(const gulong State) :
		m_State(State)
	{
	}

	void operator()(k3dIMesh* const Mesh)
	{
		Mesh->UnshadePaths(m_State);
	}

private:
	const gulong m_State;
};

class Weld
{
public:
	Weld(const gulong Criteria) :
		m_Criteria(Criteria)
	{
	}

	void operator()(k3dIMesh* const Mesh)
	{
		Mesh->WeldPoints(m_Criteria);
	}

private:
	const gulong m_Criteria;
};

class Unweld
{
public:
	Unweld(const gulong Criteria) :
		m_Criteria(Criteria)
	{
	}

	void operator()(k3dIMesh* const Mesh)
	{
		Mesh->UnweldPoints(m_Criteria);
	}

private:
	const gulong m_Criteria;
};

class FlipPathOrientation
{
public:
	void operator()(k3dIPathCollection* const PathCollection)
	{
		k3d::for_each_if(PathCollection, std::mem_fun(&k3dIPath::FlipOrientation), k3d::FlagIsSet<k3dIPath*>(k3dIState::VISIBLESELECTED));
	}
};

class AdvancePathFirstPoint
{
public:
	void operator()(k3dIPathCollection* const PathCollection)
	{
		k3d::for_each_if(PathCollection, std::mem_fun(&k3dIPath::AdvanceFirstPoint), k3d::FlagIsSet<k3dIPath*>(k3dIState::VISIBLESELECTED));
	}
};

} // namespace

//////////////////////////////////////////////////////////////////////
// CDocumentFrame

const sdpString CDocumentFrame::dialogname = "window";

const sdpString CDocumentFrame::control_filenewhierarchy = "filenewhierarchy";
const sdpString CDocumentFrame::control_fileimport = "fileimport";
const sdpString CDocumentFrame::control_fileexport = "fileexport";
const sdpString CDocumentFrame::control_filesave = "filesave";
const sdpString CDocumentFrame::control_filesaveas = "filesaveas";
const sdpString CDocumentFrame::control_fileclose = "fileclose";

const sdpString CDocumentFrame::control_editundo = "editundo";
const sdpString CDocumentFrame::control_editredo = "editredo";
const sdpString CDocumentFrame::control_editanimationparameters = "editanimationparameters";
const sdpString CDocumentFrame::control_editsmoothshade = "editsmoothshade";
const sdpString CDocumentFrame::control_editflatshade = "editflatshade";
const sdpString CDocumentFrame::control_editweld = "editweld";
const sdpString CDocumentFrame::control_editunweld = "editunweld";
const sdpString CDocumentFrame::control_editpathorientation = "editpathorientation";
const sdpString CDocumentFrame::control_editpathfirstpoint = "editpathfirstpoint";

const sdpString CDocumentFrame::control_undostack = "undostack";

const sdpString CDocumentFrame::control_rewind = "rewind";
const sdpString CDocumentFrame::control_reverseplay = "reverseplay";
const sdpString CDocumentFrame::control_stop = "stop";
const sdpString CDocumentFrame::control_play = "play";
const sdpString CDocumentFrame::control_loopplay = "loopplay";
const sdpString CDocumentFrame::control_fastforward = "fastforward";

const sdpString CDocumentFrame::control_animationscrollbar = "animationscrollbar";
const sdpString CDocumentFrame::control_animationframe = "animationframe";

CDocumentFrame::CDocumentFrame(CDocument* Document) :
	CDocumentFrameBase(Document, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Document(Document)
{
	// Sanity checks ...
	g_assert(m_Document);

	g_return_if_fail(LoadGTKMLTemplate("documentframe.gtkml"));

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocumentFrame::CDocumentFrame()" << std::endl;

	// Create the hierarchy control ...
	m_HierarchyControl = new CHierarchyControl(m_Document, this, Widget("hierarchy"));
	g_return_if_fail(m_HierarchyControl);

	// Create an event for our animation scrollbar ...
	sdpGtkAdjustment adjustment = Scrollbar(control_animationscrollbar).Adjustment();
	MapEvent("value-changed", control_animationscrollbar, false, adjustment, true);

	// Update our titlebar ...
	UpdateTitlebar();

	// Setup user actions
	AddCommand(control_filenewhierarchy, control_filenewhierarchy, CONTROL_MENU_ITEM);
	AddCommand(control_fileimport, control_fileimport, CONTROL_MENU_ITEM);
	AddCommand(control_fileexport, control_fileexport, CONTROL_MENU_ITEM);
	AddCommand(control_filesave, control_filesave, CONTROL_MENU_ITEM);
	AddCommand(control_filesaveas, control_filesaveas, CONTROL_MENU_ITEM);
	AddCommand(control_fileclose, control_fileclose, CONTROL_MENU_ITEM);
	AddCommand(control_editundo, control_editundo, CONTROL_MENU_ITEM);
	AddCommand(control_editredo, control_editredo, CONTROL_MENU_ITEM);
	AddCommand(control_editanimationparameters, control_editanimationparameters, CONTROL_MENU_ITEM);
	AddCommand(control_editsmoothshade, control_editsmoothshade, CONTROL_MENU_ITEM);
	AddCommand(control_editflatshade, control_editflatshade, CONTROL_MENU_ITEM);
	AddCommand(control_editweld, control_editweld, CONTROL_MENU_ITEM);
	AddCommand(control_editunweld, control_editunweld, CONTROL_MENU_ITEM);
	AddCommand(control_editpathorientation, control_editpathorientation, CONTROL_MENU_ITEM);
	AddCommand(control_editpathfirstpoint, control_editpathfirstpoint, CONTROL_MENU_ITEM);

	AddCommand(control_rewind, control_rewind, CONTROL_BUTTON);
	AddCommand(control_reverseplay, control_reverseplay, CONTROL_BUTTON);
	AddCommand(control_stop, control_stop, CONTROL_BUTTON);
	AddCommand(control_play, control_play, CONTROL_BUTTON);
	AddCommand(control_loopplay, control_loopplay, CONTROL_BUTTON);
	AddCommand(control_fastforward, control_fastforward, CONTROL_BUTTON);

	// Setup event notification ...
	m_Document->AddEventObserver(this, k3dIDocument::Event_Close, "CDocumentFrame");
	m_Document->AddEventObserver(this, k3dIDocument::Event_TitleChanged, "CDocumentFrame");
	m_Document->AddEventObserver(this, k3dIDocument::Event_PlaybackModeChanged, "CDocumentFrame");

	k3dIEventSource* staterecorder = dynamic_cast<k3dIEventSource*>(m_Document->StateRecorder());
	g_return_if_fail(staterecorder);
	staterecorder->AddEventObserver(this, k3dIStateRecorder::Event_StackChanged, "CDocumentFrame");
	staterecorder->AddEventObserver(this, k3dIStateRecorder::Event_Saved, "CDocumentFrame");

	k3dIEventSource* time = dynamic_cast<k3dIEventSource*>(m_Document->Time());
	g_return_if_fail(time);
	time->AddEventObserver(this, k3dITime::Event_ViewTimeChanged, "CDocumentFrame");

	CallUpdateControls();
	Show();
}

CDocumentFrame::~CDocumentFrame()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocumentFrame::~CDocumentFrame()" << std::endl;

	// Disable event notification ...
	Document()->RemoveEventObserver(this);

	k3dIEventSource* staterecorder = dynamic_cast<k3dIEventSource*>(m_Document->StateRecorder());
	g_return_if_fail(staterecorder);
	staterecorder->RemoveEventObserver(this);

	k3dIEventSource* time = dynamic_cast<k3dIEventSource*>(m_Document->Time());
	g_return_if_fail(time);
	time->RemoveEventObserver(this);

	delete m_HierarchyControl;
}

sdpResult CDocumentFrame::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	
	if(event == k3dIDocument::Event_Close)
		delete this;
	else if(event == k3dIDocument::Event_TitleChanged)
		UpdateTitlebar();
	else if(event == k3dIStateRecorder::Event_StackChanged)
		UpdateUndoStack();
	else if(event == k3dIStateRecorder::Event_Saved)
		UpdateUndoStack();
	else if(event == k3dIDocument::Event_PlaybackModeChanged)
		UpdatePlaybackMode();
	else if(event == k3dITime::Event_ViewTimeChanged)
		UpdateViewTime();
	else
		return CDocumentFrameBase::ObserveEvent(Source, Event, EventData, ObserverData);

	return SDP_S_OK;
}

bool CDocumentFrame::Load(sdpxml::Element& Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element.Name() == "documentframe", false);
	return true;
}

bool CDocumentFrame::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);
	
	Element.Append(sdpxml::Element("documentframe"));

	return true;
}

void CDocumentFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_filenewhierarchy)
		OnFileNewHierarchy();
	else if(Event->Name() == control_fileimport)
		OnFileImport();
	else if(Event->Name() == control_fileexport)
		OnFileExport();
	else if(Event->Name() == control_filesave)
		OnFileSave();
	else if(Event->Name() == control_filesaveas)
		OnFileSaveAs();
	else if(Event->Name() == control_fileclose)
		OnFileClose();
	else if(Event->Name() == control_editundo)
		OnEditUndo();
	else if(Event->Name() == control_editredo)
		OnEditRedo();
	else if(Event->Name() == control_editsmoothshade)
		OnEditSmoothShade();
	else if(Event->Name() == control_editflatshade)
		OnEditFlatShade();
	else if(Event->Name() == control_editweld)
		OnEditWeld();
	else if(Event->Name() == control_editunweld)
		OnEditUnweld();
	else if(Event->Name() == control_editpathorientation)
		OnEditFlipPathOrientation();
	else if(Event->Name() == control_editpathfirstpoint)
		OnEditAdvancePathFirstPoint();
	else if(Event->Name() == control_editanimationparameters)
		OnEditAnimationParameters();
	else if(Event->Name() == control_undostack)
		OnUndoStack(Event);
	else if(Event->Name() == control_rewind)
		OnRewind();
	else if(Event->Name() == control_reverseplay)
		OnReversePlay();
	else if(Event->Name() == control_stop)
		OnStop();
	else if(Event->Name() == control_play)
		OnPlay();
	else if(Event->Name() == control_loopplay)
		OnLoopPlay();
	else if(Event->Name() == control_fastforward)
		OnFastForward();
	else if(Event->Name() == control_animationscrollbar)
		OnAnimationScrollbar();
	else if(Event->Name() == control_animationframe)
		OnAnimationFrame();
	else
		CDocumentFrameBase::OnEvent(Event);
}

void CDocumentFrame::OnDelete(sdpGtkEvent* Event)
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocumentFrame::OnDelete()" << std::endl;

	// Cancel the normal window close ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	OnFileClose();
}

void CDocumentFrame::OnFileNewHierarchy()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_filenewhierarchy, 0);
	
	CHierarchyFrame* frame = new CHierarchyFrame(Document());
	g_return_if_fail(frame);
}

void CDocumentFrame::OnFileImport()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fileimport, 0);

	// Make sure we have some file formats to choose from ...
	std::vector<k3dIObjectInfo*> fileformats;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(fileformats), k3d::Implements(sdp::TypeInfo<k3dIGeometryReadFormat>()));
	if(0 == fileformats.size())
		{
			k3dApplication()->Error("No geometry import file filters available", "Import Geometry:");
			return;
		}

	// Prompt the user for a file ...
	sdpString filepath;
	if(!k3dGetFilePath("geometry", "Import Geometry:", false, 0, filepath))
		return;

	// Prompt the user to choose an import filter ...
	k3d::FilterSelector filterselector(Document(), "Select Import File Format:", filepath, sdp::TypeInfo<k3dIGeometryReadFormat>());
	filterselector.DoModal(RootWindow());
	if(!filterselector.OK())
		return;
	k3d::auto_ptr<k3dIGeometryReadFormat> filter(filterselector.Filter<k3dIGeometryReadFormat>());

	// Record undo/redo information ...
	k3dRecordStateChangeSet changeset(Document(), "Import " + filepath);

	// Import the file ...
	if(sdpFailed(Document()->Import(filepath, filter.get())))
		{
			SDPCSTRING message = "Error importing geometry file.  If you chose \"Automatic\" as your filter type,\n"
													"try choosing a specific filter that matches the type of file you're importing.";
													
			k3dApplication()->Error(message, "Import Geometry:");
			return;
		}
	
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnFileExport()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fileexport, 0);

	// Get the list of available export formats ...
	std::vector<k3dIObjectInfo*> fileformats;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(fileformats), k3d::Implements(sdp::TypeInfo<k3dIGeometryWriteFormat>()));
	if(0 == fileformats.size())
		{
			k3dApplication()->Error("No geometry export file filters available", "Export Geometry:");
			return;
		}

	// Prompt the user for a file to export ...
	sdpString filepath;
	if(!k3dGetFilePath("geometry", "Export Geometry:", true, 0, filepath))
		return;
	
	// Prompt the user to choose an export filter ...		
	k3d::FilterSelector filterselector(Document(), "Select Export File Format:", filepath, sdp::TypeInfo<k3dIGeometryWriteFormat>());
	filterselector.DoModal(RootWindow());
	if(!filterselector.OK())
		return;
	k3d::auto_ptr<k3dIGeometryWriteFormat> filter(filterselector.Filter<k3dIGeometryWriteFormat>());

	// Export the file ...
	if(sdpFailed(Document()->Export(filepath, filter.get())))
		{
			SDPCSTRING message = "Error exporting geometry file.  If you chose \"Automatic\" as your filter type,\n"
													"try choosing a specific filter that matches the file type you want to export.";
													
			k3dApplication()->Error(message, "Export Geometry:");
			return;
		}

	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnFileSave()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_filesave, 0);

	// If the user hasn't chosen a name, yet, turn it into a "Save As" ...
	sdpString path = Document()->Path();
	if(0 == path.size())
		OnFileSaveAs();
	else
		g_return_if_fail(sdpSucceeded(Document()->Command(k3dIDocument::Command_Save, path, 0)));
}

void CDocumentFrame::OnFileSaveAs()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_filesaveas, 0);

	sdpString filepath;
	if(!k3dGetFilePath("document", "Save K-3D Document As:", true, 0, filepath))
		return;

	g_return_if_fail(sdpSucceeded(Document()->Command(k3dIDocument::Command_Save, filepath, 0)));
}

void CDocumentFrame::OnFileClose()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CDocumentFrame::OnFileClose()" << std::endl;

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fileclose, 0);

	// Tell the document to close ...
	bool closedocument = false;

	// if we're in batch mode, we close without asking ...
	if(k3dApplication()->BatchMode())
		closedocument = true;
	else if(!Document()->StateRecorder()->UnsavedChanges())
		closedocument = true;
	else
		closedocument = SafeToClose();
	
	if(closedocument)
		g_return_if_fail(sdpSucceeded(Document()->Command(k3dIDocument::Command_Close, 0, 0)));
}

void CDocumentFrame::OnEditUndo()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editundo, 0);

	// Ignore it if there aren't any actions to undo ...
	if(!Document()->StateRecorder()->UndoCount())
		return;

	// Tell the document to undo the last action ...
	Document()->StateRecorder()->Undo();
	
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditRedo()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editredo, 0);

	// Ignore it if there aren't any actions to redo ...
	if(!Document()->StateRecorder()->RedoCount())
		return;

	// Tell the document to redo the last action ...
	Document()->StateRecorder()->Redo();
	
	k3dRedrawAll(m_Document, false);
}


void CDocumentFrame::OnEditSmoothShade()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editsmoothshade, 0);
	
	k3d::for_each_if<k3dIMesh>(m_Document->Objects(), SmoothShade(k3dIState::VISIBLESELECTED), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditFlatShade()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editflatshade, 0);

	k3d::for_each_if<k3dIMesh>(m_Document->Objects(), FlatShade(k3dIState::VISIBLESELECTED), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditWeld()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editweld, 0);

	k3d::for_each_if<k3dIMesh>(m_Document->Objects(), Weld(k3dIState::VISIBLESELECTED), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));
	k3dDeselectAll(m_Document);
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditUnweld()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editunweld, 0);

	k3d::for_each_if<k3dIMesh>(m_Document->Objects(), Unweld(k3dIState::VISIBLESELECTED), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));
	k3dDeselectAll(m_Document);
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditFlipPathOrientation()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editpathorientation, 0);
	
	k3d::for_each_if<k3dIPathCollection>(m_Document->Objects(), FlipPathOrientation(), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditAdvancePathFirstPoint()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editpathfirstpoint, 0);
	
	k3d::for_each_if<k3dIPathCollection>(m_Document->Objects(), AdvancePathFirstPoint(), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnEditAnimationParameters()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editanimationparameters, 0);
	new CAnimationParametersDialog(Document(), Document()->m_Time, Document());
}

void CDocumentFrame::OnFocusIn()
{
	// If the document is in the process of being closed, do nothing ...
	if(Document()->IsDying())
		return;

	// Activate the document ...
	k3dApplication()->SetActiveDocument(Document());
}

void CDocumentFrame::OnUndoStack(sdpGtkEvent* Event)
{
	// Get the row we clicked on ...
	int row;
	int column;
	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;
	gint rowselected = CList(control_undostack).GetHitInfo(gint(event->Event()->x), gint(event->Event()->y), &row, &column);

	// Cache a reference to the state recorder ...
	k3dIStateRecorder* staterecorder = Document()->StateRecorder();

	// If the user clicked on a row, call Undo() or Redo() until we reach the selected state ...
	if(rowselected)
		{
			while(gulong(row) < staterecorder->UndoCount())
				staterecorder->Undo();

			while(gulong(row) > staterecorder->UndoCount())
				staterecorder->Redo();
		}

	// Otherwise, the user must have clicked after the last row, so redo everything ...
	else
		{
			while(staterecorder->RedoCount())
				staterecorder->Redo();
		}

	// Update the display ...
	k3dRedrawAll(Document(), false);

	gtk_signal_emit_stop_by_name(CList(control_undostack).Object(), "button-press-event");
}

void CDocumentFrame::OnRewind()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_rewind, 0);

	k3dTimeController time(Document()->Time());
	time.SetViewFrame(time.StartFrame());
	
	Document()->SetPlaybackMode(CDocument::STOP);

	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnReversePlay()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_reverseplay, 0);
	Document()->SetPlaybackMode(CDocument::REVERSEPLAY);
}

void CDocumentFrame::OnStop()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_stop, 0);
	Document()->SetPlaybackMode(CDocument::STOP);
}

void CDocumentFrame::OnPlay()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_play, 0);
	Document()->SetPlaybackMode(CDocument::PLAY);
}

void CDocumentFrame::OnLoopPlay()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_loopplay, 0);
	Document()->SetPlaybackMode(CDocument::LOOPPLAY);
}

void CDocumentFrame::OnFastForward()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fastforward, 0);

	k3dTimeController time(Document()->Time());
	time.SetViewFrame(time.EndFrame());
	Document()->SetPlaybackMode(CDocument::STOP);

	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnAnimationScrollbar()
{
	// Update our scrollbar ...
	sdpGtkAdjustment adjustment = Scrollbar(control_animationscrollbar).Adjustment();
	
	k3dTimeController time(Document()->Time());
	time.SetViewFrame(gint(adjustment.Value()));
	
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::OnAnimationFrame()
{
	// Set the new frame ...
	k3dTimeController time(Document()->Time());
	time.SetViewFrame(atol(Entry(control_animationframe).GetText()));
	
	k3dRedrawAll(m_Document, false);
}

void CDocumentFrame::UpdateControls()
{
	UpdateUndoStack();
	UpdatePlaybackMode();
	UpdateViewTime();
	
	CDocumentFrameBase::UpdateControls();
}

void CDocumentFrame::UpdateTitlebar()
{
	// Get our title and stuff it in the titlebar ...
	RootWindow().SetTitle(Title());
}

sdpString CDocumentFrame::Title()
{
	// Create the title string ...
	sdpString title = Document()->Title();

	return title;
}

/// Populates the contents of the Undo Stack with a current list of Undo-able changes
class CDocumentFramePopulateUndoStack :
	public k3dIChangeSetVisitor
{
public:
	CDocumentFramePopulateUndoStack(sdpGtkCList List, k3dIStateRecorder* const StateRecorder) :
		m_List(List)
	{
		// Sanity checks ...
		g_assert(StateRecorder);
	
		// Setup our colors
		m_RedoForeground.red = 0x4444;
		m_RedoForeground.green = 0x4444;
		m_RedoForeground.blue = 0x4444;	
		gdk_colormap_alloc_color(gdk_colormap_get_system(), &m_RedoBackground, FALSE, TRUE);

		m_RedoBackground.red = 0xffff;
		m_RedoBackground.green = 0xffff;
		m_RedoBackground.blue = 0xaaaa;
		gdk_colormap_alloc_color(gdk_colormap_get_system(), &m_RedoBackground, FALSE, TRUE);
		
		// Update the list ...	
		m_List.Freeze();
		m_List.Clear();
		
		StateRecorder->VisitChangeSets(this);
		
		m_List.Thaw();
	}
	
	virtual ~CDocumentFramePopulateUndoStack()
	{
	}
	
	void VisitUndoChangeSet(SDPCSTRING UndoLabel, const bool Saved)
	{
		// If this is the channge set where we last saved, then append [Saved] to the name
		sdpString name(UndoLabel);
		if(Saved)
			name += " [Saved]";

		m_List.Append(name);
	}
	
	void VisitRedoChangeSet(SDPCSTRING RedoLabel, const bool Saved)
	{
		// If this is the channge set where we last saved, then append [Saved] to the name
		sdpString name(RedoLabel);
		if(Saved)
			name += " [Saved]";

		m_List.Append(name);
		
		m_List.SetForegroundColor(m_List.RowCount() - 1, &m_RedoForeground);
		m_List.SetBackgroundColor(m_List.RowCount() - 1, &m_RedoBackground);
	}
	
private:
	sdpGtkCList m_List;
	GdkColor m_RedoForeground;
	GdkColor m_RedoBackground;
};

void CDocumentFrame::UpdateUndoStack()
{
	// Setup the undo menu item ...
	sdpGtkMenuItem undoitem = MenuItem(control_editundo);
	GList* undolist = gtk_container_children(undoitem);
	sdpGtkLabel undolabel((GtkLabel*)undolist->data);

	if(Document()->StateRecorder()->UndoCount())
		{
			undolabel.SetText("Undo \"" + sdpString(Document()->StateRecorder()->NextUndoLabel()) + "\"");
			undolabel.SetSensitive(true);
		}
	else
		{
			undolabel.SetText("Can't Undo");
			undolabel.SetSensitive(false);
		}

	// Setup the redo menu item ...
	sdpGtkMenuItem redoitem = MenuItem(control_editredo);
	GList* redolist = gtk_container_children(redoitem);
	sdpGtkLabel redolabel((GtkLabel*)redolist->data);

	if(Document()->StateRecorder()->RedoCount())
		{
			redolabel.SetText("Redo \"" + sdpString(Document()->StateRecorder()->NextRedoLabel()) + "\"");
			redolabel.SetSensitive(true);
		}
	else
		{
			redolabel.SetText("Can't Redo");
			redolabel.SetSensitive(false);
		}

	CDocumentFramePopulateUndoStack populateundostack(CList(control_undostack), Document()->StateRecorder());
}

void CDocumentFrame::UpdatePlaybackMode()
{
	BlockAllEvents();

	// Set all our buttons to "un-toggled" ...
	ToggleButton(control_rewind).SetState(false);
	ToggleButton(control_reverseplay).SetState(CDocument::REVERSEPLAY == Document()->m_PlaybackMode);
	ToggleButton(control_stop).SetState(CDocument::STOP == Document()->m_PlaybackMode);
	ToggleButton(control_play).SetState(CDocument::PLAY == Document()->m_PlaybackMode);
	ToggleButton(control_loopplay).SetState(CDocument::LOOPPLAY == Document()->m_PlaybackMode);
	ToggleButton(control_fastforward).SetState(false);

	UnblockAllEvents();
}

void CDocumentFrame::UpdateViewTime()
{
	BlockAllEvents();

	k3dTimeController time(Document()->Time());

	// Display the frame number ...
	Entry(control_animationframe).SetText(sdpToString(time.ViewFrame()));

	// Update our scrollbar ...
	gulong pagesize = static_cast<gulong>(sdpRound(time.FrameRate()));

	sdpGtkAdjustment adjustment = Scrollbar(control_animationscrollbar).Adjustment();
	adjustment.SetLower(time.StartFrame());
	adjustment.SetUpper(time.EndFrame() + pagesize);
	adjustment.SetValue(time.ViewFrame());
	adjustment.SetStepIncrement(1);
	adjustment.SetPageIncrement(pagesize);
	adjustment.SetPageSize(pagesize);
	adjustment.Changed();
	adjustment.ValueChanged();
	
	UnblockAllEvents();
}

bool CDocumentFrame::SafeToClose()
{
	sdpStringArray buttons;
	buttons.push_back("Yes");
	buttons.push_back("No");
	
	return 1 == k3dApplication()->Query("Close the document?  Unsaved changes will be lost (No Undo)", "Close Document:", 1, buttons);
}


