#ifndef DOCUMENTFRAME_H
#define DOCUMENTFRAME_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CDocumentFrame class, which encapsulates an open K-3D document's UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>

// Forward class declarations
class CDocument;
class CHierarchyControl;

///////////////////////////////////////////////////////////
// CDocumentFrame

typedef k3dDialog CDocumentFrameBase;

/// Implements a document's user interface
class CDocumentFrame :
	public CDocumentFrameBase
{
public:
	CDocumentFrame(CDocument* Document);
	virtual ~CDocumentFrame();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	/// Called to load the frame from an XML document
	bool Load(sdpxml::Element& Element);
	/// Saves the frame to an XML document
	bool Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);

	/// Returns the owning document
	CDocument* Document() { return m_Document; }

	static const sdpString dialogname;

protected:
	/// Overridden from sdpGtkObjectContainer to handle events
	void OnEvent(sdpGtkEvent* Event);

	/// Called to see if its OK to delete the frame
	void OnDelete(sdpGtkEvent* Event);
	/// Called when the user wants to create a new hierarchy view
	void OnFileNewHierarchy();
	/// Called when the user wants to import geometry into the document
	void OnFileImport();
	/// Called when the user wants to export geometry from the document
	void OnFileExport();
	/// Called when the user wants to save the document
	void OnFileSave();
	/// Called when the user wants to save the document to a new file
	void OnFileSaveAs();
	/// Called when the user wants to close the document
	void OnFileClose();
	/// Called when the user wants to undo an action
	void OnEditUndo();
	/// Called when the user wants to redo an undone action
	void OnEditRedo();
	/// Called when the user wants to smooth-shade paths
	void OnEditSmoothShade();
	/// Called when the user wants to unshade smooth-shade paths
	void OnEditFlatShade();
	/// Called when the user wants to weld points together
	void OnEditWeld();
	/// Called when the user wants to unweld points
	void OnEditUnweld();
	/// Called when the user wants to flip path orientations
	void OnEditFlipPathOrientation();
	/// Called when the user wants to move path firstpoints
	void OnEditAdvancePathFirstPoint();
	/// Called when the user wants to edit animation parameters
	void OnEditAnimationParameters();
	/// Called when the frame receives the input focus
	void OnFocusIn();
	/// Called when the user hits the "rewind" button
	void OnRewind();
	/// Called when the user hits the "reverse play" button
	void OnReversePlay();
	/// Called when the user hits the "stop" button
	void OnStop();
	/// Called when the user hits the "play" button
	void OnPlay();
	/// Called when the user hits the "continuous play" button
	void OnLoopPlay();
	/// Called when the user hits the "fast forward" button
	void OnFastForward();
	/// Called when the user moves the animation scrollbar
	void OnAnimationScrollbar();
	/// Called when the user selects a new animation frame
	void OnAnimationFrame();
	/// Called when the user clicks in the undo stack to perform multiple undos/redos
	void OnUndoStack(sdpGtkEvent* Event);

	/// Returns the frame title
	sdpString Title();
	/// Updates the frame titlebar to display the frame title
	void UpdateTitlebar();

	void UpdateControls();

	/// Updates the undo/redo menu commands as needed
	void UpdateUndoStack();
	/// Updates the animation playback controls as needed
	void UpdatePlaybackMode();
	/// Updates the animation frame controls as needed
	void UpdateViewTime();

	/// Prompts the user to see if it's safe to close the document
	bool SafeToClose();

	/// Stores a pointer to the parent document
	CDocument* m_Document;

	/// Stores the hierarchy UI
	CHierarchyControl* m_HierarchyControl;

	// User actions ...
	static const sdpString control_filenewhierarchy;
	static const sdpString control_fileimport;
	static const sdpString control_fileexport;
	static const sdpString control_filesave;
	static const sdpString control_filesaveas;
	static const sdpString control_fileclose;
	static const sdpString control_editundo;
	static const sdpString control_editredo;
	static const sdpString control_editanimationparameters;
	static const sdpString control_editsmoothshade;
	static const sdpString control_editflatshade;
	static const sdpString control_editweld;
	static const sdpString control_editunweld;
	static const sdpString control_editpathorientation;
	static const sdpString control_editpathfirstpoint;

	static const sdpString control_undostack;

	static const sdpString control_rewind;
	static const sdpString control_reverseplay;
	static const sdpString control_stop;
	static const sdpString control_play;
	static const sdpString control_loopplay;
	static const sdpString control_fastforward;
	
	static const sdpString control_animationscrollbar;
	static const sdpString control_animationframe;
};

#endif // DOCUMENTFRAME_H
