// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CHierarchy class, which encapsulates a set of hierarchical relationships between K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include "hierarchy.h"

#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dstaterecorder.h>

#include <k3didependencies.h>
#include <k3didocument.h>
#include <k3dilocation.h>
#include <k3distate.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H

namespace
{

class SnapshotWorldCoordinates
{
public:
	SnapshotWorldCoordinates() :
		m_Matrix(sdpIdentity3D())
	{
	}

	void operator()(const CHierarchyRecord& Record)
	{
		switch(Record.Type())
			{
				case k3dIHierarchy::PUSH:
					if(Record.Transform())
						{
							m_Stack.push(m_Matrix);
							m_Matrix = m_Matrix * k3dGetTransformMatrix(Record.Transform(), k3dITransform::ANIMATED_FRAME);
						}
					break;

				case k3dIHierarchy::NODE:
					if(Record.LocationCollection())
						{
							for(K3DLOCATIONHANDLE locationhandle = Record.LocationCollection()->FirstLocation(); locationhandle; locationhandle = Record.LocationCollection()->NextLocation(locationhandle))
								{
									k3dILocation* const location = Record.LocationCollection()->Location(locationhandle);
									location->SetWorldXYZ(m_Matrix * location->AnimXYZ());
								}
								
							if(Record.Transform())
								m_Matrix = m_Matrix * k3dGetTransformMatrix(Record.Transform(), k3dITransform::ANIMATED_CHILD_FRAME);
						}
					break;

				case k3dIHierarchy::POP:
					if(Record.Transform())
						{
							m_Matrix = m_Stack.top();
							m_Stack.pop();
						}
					break;
					
				default:
					g_return_if_fail(0);
			}
	}
	
private:
	sdpMatrix4 m_Matrix;
	std::stack<sdpMatrix4> m_Stack;
};

class PreAnimate
{
public:
	PreAnimate(const gdouble RenderTime, const gdouble RenderDeltaTime) :
		m_RenderTime(RenderTime),
		m_RenderDeltaTime(RenderDeltaTime)
	{
	}

	void operator()(const CHierarchyRecord& Record)
	{
		// Skip invisible objects ...
		if(SDP_S_OK != Record.Object()->State()->QueryState(k3dIState::VISIBLE))
			return;
	
		// Reset times ...
		if(Record.TimeServer())
			Record.TimeServer()->SetAnimTime(m_RenderTime, m_RenderDeltaTime);

		// Reset transformations ...
		if(Record.Transform())
			{
				Record.Transform()->SetPosition(k3dITransform::ANIMATED_FRAME, Record.Transform()->Position(k3dITransform::CURRENT_FRAME));
				Record.Transform()->SetOrientation(k3dITransform::ANIMATED_FRAME, Record.Transform()->Orientation(k3dITransform::CURRENT_FRAME));
				Record.Transform()->SetScale(k3dITransform::ANIMATED_FRAME, Record.Transform()->Scale(k3dITransform::CURRENT_FRAME));

				Record.Transform()->SetPosition(k3dITransform::ANIMATED_CHILD_FRAME, Record.Transform()->Position(k3dITransform::CURRENT_CHILD_FRAME));
				Record.Transform()->SetOrientation(k3dITransform::ANIMATED_CHILD_FRAME, Record.Transform()->Orientation(k3dITransform::CURRENT_CHILD_FRAME));
				Record.Transform()->SetScale(k3dITransform::ANIMATED_CHILD_FRAME, Record.Transform()->Scale(k3dITransform::CURRENT_CHILD_FRAME));
			}

		// Reset locations ...
		if(Record.LocationCollection())
			for(K3DLOCATIONHANDLE locationhandle = Record.LocationCollection()->FirstLocation(); locationhandle; locationhandle = Record.LocationCollection()->NextLocation(locationhandle))
				{
					k3dILocation* const location = Record.LocationCollection()->Location(locationhandle);
					location->SetAnimXYZ(location->LocalXYZ());
				}

		// Pre-animate ...
		if(Record.Animate())
			Record.Animate()->PreAnimate();
	}

private:
	const gdouble m_RenderTime;
	const gdouble m_RenderDeltaTime;
};

class Animate
{
public:
	void operator()(const CHierarchyRecord& Record)
	{
		// Skip objects that aren't animated ...
		if(0 == Record.Animate())
			return;

		// Skip invisible objects ...
		if(SDP_S_OK != Record.Object()->State()->QueryState(k3dIState::VISIBLE))
			return;
	
		switch(Record.Type())
			{
				case k3dIHierarchy::PUSH:
					Record.Animate()->PushAnimate();
					break;
				case k3dIHierarchy::NODE:
					Record.Animate()->Animate();
					break;
				case k3dIHierarchy::POP:
					Record.Animate()->PopAnimate();
					break;
				default:
					g_return_if_fail(0);
			}
	}
};

class PostAnimate
{
public:
	void operator()(const CHierarchyRecord& Record)
	{
		// Skip invisible objects ...
		if(SDP_S_OK != Record.Object()->State()->QueryState(k3dIState::VISIBLE))
			return;
	
		if(Record.Animate())
			Record.Animate()->PostAnimate();
	}
};

} // namespace

//////////////////////////////////////////////////////////////
// CHierarchy

CHierarchy::CHierarchy(k3dIDocument* const Document) :
	m_Document(Document),
	m_HierarchyModified(false)
{
	// Sanity checks ...
	g_assert(m_Document);
}

CHierarchy::~CHierarchy()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// k3dIHierarchy implementation

k3dIHierarchy::NodeType CHierarchy::FirstHierarchyNode(K3DHIERARCHYNODEHANDLE* Handle)
{
	if(0 == m_Records.size())
		return COMPLETE;

	*Handle = K3DHIERARCHYNODEHANDLE(1);
	return m_Records.front().Type();
}

k3dIHierarchy::NodeType CHierarchy::NextHierarchyNode(K3DHIERARCHYNODEHANDLE* Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, ERROR);

	if(guint(*Handle) >= m_Records.size())
		return COMPLETE;

	*Handle = K3DHIERARCHYNODEHANDLE(guint(*Handle)+1);
	return m_Records[guint(*Handle)-1].Type();
}

k3dIHierarchy::NodeType CHierarchy::HierarchyNode(k3dIObject* Object, K3DHIERARCHYNODEHANDLE* Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Object, ERROR);

	// Find the PUSH record for this object ...
	HierarchyRecordIterator push = std::find(m_Records.begin(), m_Records.end(), CHierarchyRecord(k3dIHierarchy::PUSH, Object));
	g_return_val_if_fail(push != m_Records.end(), ERROR);
	
	// Return the data ...
	*Handle = K3DHIERARCHYNODEHANDLE((push - m_Records.begin()) + 1);
	return push->Type();
}

k3dIObject* CHierarchy::HierarchyObject(K3DHIERARCHYNODEHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);

	return m_Records[guint(Handle)-1].Object();
}

k3dIObject* CHierarchy::ObjectParent(k3dIObject* Object)
{
	HierarchyParentIterator parent = m_ParentMap.find(Object);
	return parent != m_ParentMap.end() ? parent->second : 0;
}

sdpResult CHierarchy::SetObjectParent(k3dIObject* Object, k3dIObject* NewParent)
{
	// Sanity checks ...
	g_return_val_if_fail(Object, SDP_E_FAIL);
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// If we're recording undos ...
	if(!m_HierarchyModified && m_Document->StateRecorder()->ChangeSet())
		{
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyRecords>(m_Records));
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyParentMap>(m_ParentMap));
			m_HierarchyModified = true;
		}

	// Cache the new parent for efficiency ...
	m_ParentMap[Object] = NewParent;

	// Get original records ...
	HierarchyRecordIterator oldpush(GetPush(Object));
	HierarchyRecordIterator oldpop(GetPop(Object));
	
	// Make a copy ...
	HierarchyRecords temp;
	std::copy(oldpush, oldpop+1, std::back_inserter(temp));

	// Remove original records ...
	m_Records.erase(oldpush, oldpop+1);
	
	// Add a new set ...
	HierarchyRecordIterator parentpop(m_Records.end());
	if(NewParent)
		parentpop = GetPop(NewParent);
		
	m_Records.insert(parentpop, temp.begin(), temp.end());

	// Let the world know!
	Notify(k3dIHierarchy::Event_Changed, "");
			
	return SDP_S_OK;
}

sdpResult CHierarchy::SetObjectSibling(k3dIObject* Object, k3dIObject* NewSibling)
{
	// Sanity checks ...
	g_return_val_if_fail(Object, SDP_E_FAIL);
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// If we're recording undos ...
	if(!m_HierarchyModified && m_Document->StateRecorder()->ChangeSet())
		{
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyRecords>(m_Records));
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyParentMap>(m_ParentMap));
			m_HierarchyModified = true;
		}

	// Cache the (potentially) new parent for efficiency ...
	m_ParentMap[Object] = NewSibling ? ObjectParent(NewSibling) : 0;

	// Get original records ...
	HierarchyRecordIterator oldpush(GetPush(Object));
	HierarchyRecordIterator oldpop(GetPop(Object));
	
	// Make a copy ...
	HierarchyRecords temp;
	std::copy(oldpush, oldpop+1, std::back_inserter(temp));

	// Remove original records ...
	m_Records.erase(oldpush, oldpop+1);
	
	// Add a new set ...
	HierarchyRecordIterator siblingpop(m_Records.begin());
	if(NewSibling)
		siblingpop = GetPop(NewSibling)+1;
		
	m_Records.insert(siblingpop, temp.begin(), temp.end());

	// Let the world know!
	Notify(k3dIHierarchy::Event_Changed, "");
			
	return SDP_S_OK;
}

void CHierarchy::SnapshotWorldCoordinates()
{
	std::for_each(m_Records.begin(), m_Records.end(), ::SnapshotWorldCoordinates());
}

void CHierarchy::Animate(const gdouble Time, const gdouble FrameDelta)
{
	std::for_each(m_Records.begin(), m_Records.end(), PreAnimate(Time, FrameDelta));
	std::for_each(m_Records.begin(), m_Records.end(), ::Animate());
	std::for_each(m_Records.begin(), m_Records.end(), PostAnimate());
}

void CHierarchy::Add(k3dIObject* const Object)
{
	// Sanity checks ...
	g_assert(Object);
	g_return_if_fail(Editing());

	// If we're recording undos ...
	if(!m_HierarchyModified && m_Document->StateRecorder()->ChangeSet())
		{
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyRecords>(m_Records));
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyParentMap>(m_ParentMap));
			m_HierarchyModified = true;
		}

	m_Records.push_back(CHierarchyRecord(PUSH, Object));
	m_Records.push_back(CHierarchyRecord(NODE, Object));
	m_Records.push_back(CHierarchyRecord(POP, Object));

	m_ParentMap[Object] = 0;

	// Let the world know!
	Notify(k3dIHierarchy::Event_Changed, "");
}

void CHierarchy::Remove(k3dIObject* const Object)
{
	// Sanity checks ...
	g_assert(Object);
	g_return_if_fail(Editing());

	// Make the actual changes ...
	RawRemove(Object);

	// Let the world know!
	Notify(k3dIHierarchy::Event_Changed, "");
}

void CHierarchy::RawRemove(k3dIObject* const Object)
{
	// Sanity checks ...
	g_assert(Object);

	g_return_if_fail(GetPush(Object) != m_Records.end());
	g_return_if_fail(GetNode(Object) != m_Records.end());
	g_return_if_fail(GetPop(Object) != m_Records.end());

	// If we're recording undos ...
	if(!m_HierarchyModified && m_Document->StateRecorder()->ChangeSet())
		{
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyRecords>(m_Records));
			m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<HierarchyParentMap>(m_ParentMap));
			m_HierarchyModified = true;
		}

	// Update our state ...
	m_Records.erase(GetPush(Object));
	m_Records.erase(GetNode(Object));
	m_Records.erase(GetPop(Object));
	m_ParentMap.erase(Object);
}

sdpResult CHierarchy::Save(sdpxml::Element& Document, k3dIDependencies* const Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& hierarchy = Document.Append(sdpxml::Element("hierarchy"));

	std::stack<sdpxml::ElementPointer> elementstack;
	elementstack.push(&hierarchy);

	for(HierarchyRecordIterator record = m_Records.begin(); record != m_Records.end(); ++record)
		{
			// Skip objects that aren't marked as saveable ...
			if(SDP_S_OK != record->Object()->State()->QueryState(k3dIState::SAVEABLE))
				continue;

			// Skip objects that don't support persistence ...
			if(!record->Persist())
				continue;

			if(k3dIHierarchy::PUSH == record->Type())
				elementstack.push(&elementstack.top()->Append(sdpxml::Element("object", "", sdpxml::Attribute("id", sdpToString(record->Object()->ObjectID())))));
			else if(k3dIHierarchy::POP == record->Type())
				elementstack.pop();
		}

	return SDP_S_OK;
}

void CHierarchy::LoadObject(k3dIObjectCollection* const ObjectCollection, sdpxml::Element& Element, k3dIObject* const Parent)
{
	// Sanity checks ...
	g_assert(ObjectCollection);
	g_assert(Element.Name() == "object");

	// Get the object ID ...
	const k3dIObject::ID id = sdpxml::GetAttribute(Element, "id", 0);
	g_return_if_fail(id);

	// Get the object ...
	std::vector<k3dIObject*> objects;
	k3d::copy_if(ObjectCollection, std::back_inserter(objects), k3d::SameID(id));
	g_return_if_fail(1 == objects.size());
	k3dIObject* const object = *objects.begin();
	
	// Get rid of any existing records ...
	RawRemove(object);	

	// Create records for insertion ...
	HierarchyRecords temp;
	temp.push_back(CHierarchyRecord(PUSH, object));
	temp.push_back(CHierarchyRecord(NODE, object));
	temp.push_back(CHierarchyRecord(POP, object));

	// Cache the parent for efficiency ...
	m_ParentMap[object] = Parent;

	// Insert records into the hierarchy ...
	HierarchyRecordIterator parentpop(m_Records.end());
	if(Parent)
		parentpop = GetPop(Parent);

	m_Records.insert(parentpop, temp.begin(), temp.end());		

	// Handle children
	for(sdpxml::ElementCollection::iterator element = Element.Children().begin(); element != Element.Children().end(); element++)
		LoadObject(ObjectCollection, *element, object);
}

sdpResult CHierarchy::Load(k3dIObjectCollection* const ObjectCollection, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectCollection);

	// Look for the hierarchy element ...
	sdpxml::ElementPointer hierarchy = sdpxml::FindElement(Element, sdpxml::SameName("hierarchy"));
	if(hierarchy)
		{
			// Load any records it contains ...
			for(sdpxml::ElementCollection::iterator element = hierarchy->Children().begin(); element != hierarchy->Children().end(); element++)
				LoadObject(ObjectCollection, *element, 0);
		}
			
	// As a final step, make sure that we generate records for any "unaccounted" objects that didn't show up in the XML ...
	typedef std::vector<k3dIObject*> Objects;
	Objects objects;
	k3d::copy(ObjectCollection, std::back_inserter(objects));
	for(Objects::iterator object = objects.begin(); object != objects.end(); ++object)
		{
			HierarchyRecordIterator push = GetPush(*object);
			if(push != m_Records.end())
				continue;
				
			m_Records.push_back(CHierarchyRecord(PUSH, *object));
			m_Records.push_back(CHierarchyRecord(NODE, *object));
			m_Records.push_back(CHierarchyRecord(POP, *object));
			
			std::cerr << "CHierarchy::Load() - Warning - generating records for object [" << (*object)->Name() << "]" << std::endl;
		}

	return SDP_S_OK;
}

void CHierarchy::Dump(std::ostream& Stream)
{
	Stream << g_pretty_function() << std::endl;

	static SDPCSTRING const indent = "   ";
	sdpString indentation;

	for(HierarchyRecordIterator record = m_Records.begin(); record != m_Records.end(); record++)
		{
			if(POP == record->Type())
				indentation = indentation.left(indentation.size() - strlen(indent));

			Stream << SDPCSTRING(indentation);

			switch(record->Type())
				{
					case COMPLETE:
						Stream << "COMPLETE - ";
						break;
					case PUSH:
						Stream << "PUSH - ";
						break;
					case NODE:
						Stream << "NODE - ";
						break;
					case POP:
						Stream << "POP - ";
						break;
					default:
						Stream << "UNKNOWN: " << gint(record->Type()) << " - ";
						break;
				}

			Stream << reinterpret_cast<void*>(record->Object()) << " - " << record->Object()->Name() << std::endl;

			if(PUSH == record->Type())
				indentation += indent;
		}
}

void CHierarchy::Validate()
{
	// The hierarchy size MUST be a multiple of three ...
	if(0 != (m_Records.size() % 3))
		std::cerr << "CHierarchy::Validate(): Illegal hierarchy size (not a multiple of 3): " << m_Records.size()  << std::endl;

	// Step through the hierarchy to make sure there are no over/under-flows or unknown record types ...
	int stack = 0;
	for(HierarchyRecordIterator record = m_Records.begin(); record != m_Records.end(); record++)
		{
			switch(record->Type())
				{
					case PUSH:
						stack++;
						break;
					case NODE:
						break;
					case POP:
						if(stack)
							{
								stack--;
							}
						else
							std::cerr << "CHierarchy::Validate(): Hierarchy stack underflow" << std::endl;
						break;
					default:
						std::cerr << "CHierarchy::Validate(): Unknown hierarchy record type: " << gint(record->Type()) << std::endl;
						break;
				}
		}

	if(stack)
		std::cerr << "CHierarchy::Validate(): Hierarchy stack overflow: " << stack << std::endl;
}

CHierarchy::HierarchyRecordIterator CHierarchy::GetPush(k3dIObject* const Object)
{
	// Sanity checks ...
	g_assert(Object);
	return std::find(m_Records.begin(), m_Records.end(), CHierarchyRecord(PUSH, Object));
}

CHierarchy::HierarchyRecordIterator CHierarchy::GetNode(k3dIObject* const Object)
{
	// Sanity checks ...
	g_assert(Object);
	return std::find(m_Records.begin(), m_Records.end(), CHierarchyRecord(NODE, Object));
}

CHierarchy::HierarchyRecordIterator CHierarchy::GetPop(k3dIObject* const Object)
{
	// Sanity checks ...
	g_assert(Object);
	return std::find(m_Records.begin(), m_Records.end(), CHierarchyRecord(POP, Object));
}

void CHierarchy::VisitHierarchyRecords(CHierarchyRecordVisitor& Visitor)
{
	for(HierarchyRecordIterator record = m_Records.begin(); record != m_Records.end(); record++)
		Visitor.VisitHierarchyRecord(*record);
}

void CHierarchy::PushEditing()
{
//std::cerr << g_pretty_function() << std::endl;
	if(!Editing())
		m_HierarchyModified = false;

	k3dEditable::PushEditing();
}

void CHierarchy::PopEditing()
{
//std::cerr << g_pretty_function() << std::endl;

	k3dEditable::PopEditing();

	// If we're recording undos ...
	k3dIStateChangeSet* const changeset = m_Document->StateRecorder()->ChangeSet();
	if(!Editing() && m_HierarchyModified && changeset)
		{
			changeset->RecordNewState(new k3dValueStateContainer<HierarchyRecords>(m_Records));
			changeset->RecordNewState(new k3dValueStateContainer<HierarchyParentMap>(m_ParentMap));
			changeset->RecordAction(new k3dActionNotify(this, k3dIHierarchy::Event_Changed, 0, 0));
			m_HierarchyModified = false;
		}
}

