#ifndef HIERARCHY_H
#define HIERARCHY_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CHierarchy class, which encapsulates a set of hierarchical relationships between K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deditable.h>
#include <k3deventsource.h>

#include <k3dianimate.h>
#include <k3dihierarchy.h>
#include <k3dilocationcollection.h>
#include <k3diobject.h>
#include <k3dipersist.h>
#include <k3ditimeserver.h>
#include <k3ditransform.h>

// Forward class declarations
class CDocument;

class k3dIDependencies;
class k3dIDocument;
class k3dIObjectCollection;
namespace sdpxml { class Element; }

///////////////////////////////////////////////////////////
// CHierarchyRecord

/// Stores the position of an object context within the hierarchy
class CHierarchyRecord
{
public:
	CHierarchyRecord(const k3dIHierarchy::NodeType Type, k3dIObject* const Object) :
		m_Type(Type),
		m_Animate(dynamic_cast<k3dIAnimate*>(Object)),
		m_LocationCollection(dynamic_cast<k3dILocationCollection*>(Object)),
		m_Object(Object),
		m_Persist(dynamic_cast<k3dIPersist*>(Object)),
		m_TimeServer(dynamic_cast<k3dITimeServer*>(Object)),
		m_Transform(dynamic_cast<k3dITransform*>(Object))
	{
		// Sanity checks ...
		g_assert(m_Object);
	}

	const k3dIHierarchy::NodeType Type() const
	{
		return m_Type;
	}
	
	k3dIAnimate* const Animate() const
	{
		return m_Animate;
	}
	
	k3dILocationCollection* const LocationCollection() const
	{
		return m_LocationCollection;
	}
	
	k3dIObject* const Object() const
	{
		return m_Object;
	}
	
	k3dIPersist* const Persist() const
	{
		return m_Persist;
	}
	
	k3dITimeServer* const TimeServer() const
	{
		return m_TimeServer;
	}
	
	k3dITransform* const Transform() const
	{
		return m_Transform;
	}

	friend bool operator==(const CHierarchyRecord& LHS, const CHierarchyRecord& RHS)
	{
		return (LHS.m_Type == RHS.m_Type) && (LHS.m_Object == RHS.m_Object);
	}
	
	friend bool operator!=(const CHierarchyRecord& LHS, const CHierarchyRecord& RHS)
	{
		return (LHS.m_Type != RHS.m_Type) || (LHS.m_Object != RHS.m_Object);
	}

private:
	/// Stores the record type: PUSH, NODE, or POP
	k3dIHierarchy::NodeType m_Type;
	
	/// Cached object interface for efficiency
	k3dIAnimate* m_Animate;
	/// Cached object interface for efficiency
	k3dILocationCollection* m_LocationCollection;
	/// Stores a reference to the object
	k3dIObject* m_Object;
	/// Cached object interface for efficiency
	k3dIPersist* m_Persist;
	/// Cached object interface for efficiency
	k3dITimeServer* m_TimeServer;
	/// Cached object interface for efficiency
	k3dITransform* m_Transform;
};

//////////////////////////////////////////////////////////////
// CHierarchyRecordVisitor

/// Abstract interface for objects that wish to enumerate through a collection of hierarchy records
class CHierarchyRecordVisitor
{
public:
	virtual ~CHierarchyRecordVisitor() {}
	
	virtual void VisitHierarchyRecord(const CHierarchyRecord& Record) = 0;
};

//////////////////////////////////////////////////////////////
// CHierarchy

/// Encapsulates a set of hierarchical relationships between K-3D objects
class CHierarchy :
	public k3dIHierarchy,
	public k3dEventSource,
	public k3dEditable
{
public:
	CHierarchy(k3dIDocument* const Document);
	~CHierarchy();
	
	// k3dIHierarchy implementation ...
	NodeType FirstHierarchyNode(K3DHIERARCHYNODEHANDLE* Handle);
	NodeType NextHierarchyNode(K3DHIERARCHYNODEHANDLE* Handle);
	NodeType HierarchyNode(k3dIObject* Object, K3DHIERARCHYNODEHANDLE* Handle);
	k3dIObject* HierarchyObject(K3DHIERARCHYNODEHANDLE Handle);

	k3dIObject* ObjectParent(k3dIObject* Object);
	sdpResult SetObjectParent(k3dIObject* Object, k3dIObject* NewParent);
	sdpResult SetObjectSibling(k3dIObject* Object, k3dIObject* NewSibling);

	void SnapshotWorldCoordinates();
	void Animate(const gdouble Time, const gdouble FrameDelta);

	// k3dIEditable implementation ...
	void PushEditing();
	void PopEditing();

	/// Enumerates through the collection of hierarchy records, in order
	void VisitHierarchyRecords(CHierarchyRecordVisitor& Visitor);

	/// Creates records for a new object
	void Add(k3dIObject* const Object);
	/// Removes records for the given object context (with undo support)
	void Remove(k3dIObject* const Object);
	
	/// Saves hierarchy data to a K-3D (XML) document
	sdpResult Save(sdpxml::Element& Document, k3dIDependencies* const Dependencies);
	/// Loads hierarchy data from a K-3D (XML) document
	sdpResult Load(k3dIObjectCollection* const ObjectCollection, sdpxml::Element& Element);
	
	/// Dumps the hierarchy to a stream for debugging
	void Dump(std::ostream& Stream);
	/// Performs sanity checks on the hierarchy for debugging
	void Validate();
	
private:
	/// Stores a reference to the owning document
	k3dIDocument* const m_Document;
	
	/// Stores a collection of hierarchy records
	typedef std::vector<CHierarchyRecord> HierarchyRecords;
	typedef HierarchyRecords::iterator HierarchyRecordIterator;
	/// Stores the hierarchy records
	HierarchyRecords m_Records;

	/// Stores a mapping of objects-to-parents for fast lookup
	typedef std::map<k3dIObject*, k3dIObject*> HierarchyParentMap;
	typedef HierarchyParentMap::iterator HierarchyParentIterator;
	/// Stores the mapping of objects to their parents for fast lookup
	HierarchyParentMap m_ParentMap;
	
	/// Set to true iff the hierarchy has been modified during the current editing session
	bool m_HierarchyModified;

	void LoadObject(k3dIObjectCollection* const ObjectCollection, sdpxml::Element& Element, k3dIObject* const Parent);
	void RawRemove(k3dIObject* const Object);

	/// Returns the "push" record for an object
	HierarchyRecordIterator GetPush(k3dIObject* const Object);
	/// Returns the "node" record for an object
	HierarchyRecordIterator GetNode(k3dIObject* const Object);
	/// Returns the "pop" record for an object
	HierarchyRecordIterator GetPop(k3dIObject* const Object);
};

#endif // HIERARCHY_H

