// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CHierarchyControl class, which encapsulates a hierarchical view of an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "hierarchycontrol.h"
#include "renameobjectdialog.h"

#include <k3ddependencies.h>
#include <k3ddnd.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3dihierarchy.h>
#include <k3dimouseeventobserver.h>
#include <k3diobject.h>
#include <k3diobjectcollection.h>
#include <k3diobjectinfo.h>
#include <k3dipersist.h>
#include <k3distate.h>

#include <sdpgtkclipboard.h>
#include <sdpgtkevents.h>
#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

#include <fstream>
#include <strstream>

///////////////////////////////////////////////////////////////////////
// Helper functions

static void DrawRectangle(const sdpRectangle& Rectangle)
{
	glBegin(GL_QUADS);
	glVertex2d(Rectangle.left, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.bottom);
	glVertex2d(Rectangle.left, Rectangle.bottom);
	glEnd();
}

static void DrawHollowRectangle(const sdpRectangle& Rectangle)
{
	glBegin(GL_LINE_LOOP);
	glVertex2d(Rectangle.left, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.bottom);
	glVertex2d(Rectangle.left, Rectangle.bottom);
	glEnd();
}

///////////////////////////////////////////////////////////////////////////////
// CHierarchyControlItem

const gdouble CHierarchyControlItem::m_ItemOffset(3.0);
const gdouble CHierarchyControlItem::m_InteriorBorder(1.0);
const gdouble CHierarchyControlItem::m_HorizontalPadding(3.0);
const gdouble CHierarchyControlItem::m_ParentBoxThickness(2.0);
const gdouble CHierarchyControlItem::m_SiblingLineThickness(2.0);

CHierarchyControlItem::CHierarchyControlItem(CHierarchyControl* const ParentControl, sdpGtkOpenGLDrawingArea* const DrawingArea, k3dIObject* const Object, const gulong Index, const gulong Generation) :
	m_ParentControl(ParentControl),
	m_DrawingArea(DrawingArea),
	m_Object(Object),
	m_Highlighted(false),
	m_Dragging(false),
	m_Index(Index),
	m_Generation(Generation),
	m_Size(DrawingArea->Width(), Height(DrawingArea)),
	m_Position(Generation * 1.5 * (DrawingArea->FontAscent() + DrawingArea->FontDescent()), Index * Height(DrawingArea)),
	m_State(NORMAL)
{
	// Sanity checks ...
	g_assert(m_ParentControl);
	g_assert(m_DrawingArea);
	g_assert(m_Object);

	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(Object);
	g_return_if_fail(eventserver);
	eventserver->AddEventObserver(this, k3dIState::Event_Changed, "CHierarchyControlItem");
	eventserver->AddEventObserver(this, k3dIObject::Event_Renamed, "CHierarchyControlItem");
}

CHierarchyControlItem::CHierarchyControlItem(const CHierarchyControlItem& Original) :
	m_ParentControl(Original.m_ParentControl),
	m_DrawingArea(Original.m_DrawingArea),
	m_Object(Original.m_Object),
	m_Highlighted(Original.m_Highlighted),
	m_Dragging(Original.m_Dragging),
	m_Index(Original.m_Index),
	m_Generation(Original.m_Generation),
	m_Size(Original.m_Size),
	m_Position(Original.m_Position),
	m_State(Original.m_State)
{
	// Sanity checks ...
	g_assert(m_ParentControl);
	g_assert(m_DrawingArea);
	g_assert(m_Object);

	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(Object());
	g_return_if_fail(eventserver);
	eventserver->AddEventObserver(this, k3dIState::Event_Changed, "CHierarchyControlItem");
	eventserver->AddEventObserver(this, k3dIObject::Event_Renamed, "CHierarchyControlItem");
};

CHierarchyControlItem::~CHierarchyControlItem()
{
	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(Object());
	g_return_if_fail(eventserver);
	eventserver->RemoveEventObserver(this);
}

sdpRectangle CHierarchyControlItem::BoundingBox() const
{
	return sdpRectangle(m_Position[0], m_Position[0] + m_Size[0], m_Position[1], m_Position[1] + m_Size[1]);
}

sdpRectangle CHierarchyControlItem::StandardBox() const
{
	sdpRectangle result = BoundingBox();

	result.bottom -= m_ItemOffset;
	result.Shrink(m_InteriorBorder);

	return result;
}

sdpRectangle CHierarchyControlItem::VisibilityBox() const
{
	sdpRectangle result = StandardBox();

	result.Shrink(m_InteriorBorder);
	result.right = result.left + result.Height();

	return result;
}

sdpRectangle CHierarchyControlItem::SelectionBox() const
{
	sdpRectangle result = VisibilityBox();

	result.left = result.right + m_HorizontalPadding;
	result.right = result.left + result.Height();

	return result;
}

sdpRectangle CHierarchyControlItem::TextBox() const
{
	sdpRectangle result = StandardBox();

	result.Shrink(m_InteriorBorder);
	result.left = SelectionBox().Right() + m_HorizontalPadding;

	return result;
}

sdpRectangle CHierarchyControlItem::ParentDropBox() const
{
	sdpRectangle result = StandardBox();

	result.bottom = (result.top + result.bottom) * 0.5;

	return result;
}

sdpRectangle CHierarchyControlItem::SiblingDropBox() const
{
	sdpRectangle result = StandardBox();

	result.top = (result.top + result.bottom) * 0.5;

	return result;
}

gdouble CHierarchyControlItem::Height(const sdpGtkOpenGLDrawingArea* const DrawingArea)
{
	// Sanity checks ...
	g_assert(DrawingArea);

	return DrawingArea->FontAscent() + DrawingArea->FontDescent() + (2 * m_InteriorBorder) + m_ItemOffset;
}

void CHierarchyControlItem::SetDropTarget(const sdpVector2 MouseCoordinates, const std::vector<CHierarchyControlItem*>& DropSources)
{
	if(SiblingDropBox().Contains(MouseCoordinates))
		{
			// Don't allow an object to become its own sibling ...
			for(gulong i = 0; i < DropSources.size(); i++)
				{
					if(DropSources[i] == this)
						return;

					if(DropSources[i]->m_Index == m_Index+1)
						return;
				}

			SetState(SIBLINGDROPTARGET);
		}
	else if(ParentDropBox().Contains(MouseCoordinates))
		{
			// Don't allow an object to become its own parent ...
			for(gulong i = 0; i < DropSources.size(); i++)
				if(DropSources[i] == this)
					return;

			SetState(PARENTDROPTARGET);
		}
}

////////////////////////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CHierarchyControlItem::ObserveEvent(k3dIEventSource* EventSource, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ClientData)
{
	// Sanity checks ...
	g_return_val_if_fail(EventSource, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIState::Event_Changed || event == k3dIObject::Event_Renamed)
		{
			m_ParentControl->RedrawControl();
			return SDP_S_OK;
		}

	return k3dEventObserver::ObserveEvent(EventSource, Event, EventData, ClientData);
}

void CHierarchyControlItem::Draw() const
{
	// If we're being dragged, enable transparency ...
	if(IsDragging())
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}

	// Should we be drawn in reverse colors?
	bool inverted = IsDragging() || IsHighlighted();

	// Draw the background ...
	if(inverted)
		{
			glColor4dv(m_ParentControl->m_HighlightedColor);
			DrawRectangle(BoundingBox());
		}

	glLineWidth(1.0f);
	glDisable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);

	// Draw the visible state ...
	glColor4dv(inverted ? m_ParentControl->m_HighlightedTextColor : m_ParentControl->m_TextColor);
	if(SDP_S_OK == Object()->State()->QueryState(k3dIState::VISIBLE))
		DrawRectangle(VisibilityBox());
	else
		DrawHollowRectangle(VisibilityBox());

	// Draw the selection state ...
	glColor4dv(inverted ? m_ParentControl->m_HighlightedTextColor : m_ParentControl->m_TextColor);
	if(SDP_S_OK == Object()->State()->QueryState(k3dIState::SELECTED))
		DrawRectangle(SelectionBox());
	else
		DrawHollowRectangle(SelectionBox());

	// Draw the text ...
	sdpRectangle textrectangle = TextBox();

	glColor4dv(inverted ? m_ParentControl->m_HighlightedTextColor : m_ParentControl->m_TextColor);
	glRasterPos2d(textrectangle.left, textrectangle.top + m_DrawingArea->FontAscent());
	sdpString name = Object()->Name();
	glCallLists(name.size(), GL_UNSIGNED_BYTE, (GLubyte*)name.c_str());

	// Cleanup if we're being dragged ...
	if(IsDragging())
		glDisable(GL_BLEND);

	// If we're a parent drop target ...
	if(IsParentDropTarget())
		{
			glColor4dv(m_ParentControl->m_HighlightedColor);

			glLineWidth(m_ParentBoxThickness);
			glEnable(GL_LINE_SMOOTH);
			glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

			DrawHollowRectangle(BoundingBox());
		}

	if(IsSiblingDropTarget())
		{
			glColor4dv(m_ParentControl->m_HighlightedColor);

			glLineWidth(m_SiblingLineThickness);
			glEnable(GL_LINE_SMOOTH);
			glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

			sdpRectangle rectangle(StandardBox());
			rectangle.top = rectangle.bottom;
			rectangle.bottom = rectangle.top + m_ParentBoxThickness;

			DrawRectangle(rectangle);
		}
}

void CHierarchyControlItem::EditObject() const
{
	m_Object->EditObject();
}

void CHierarchyControlItem::RenameObject() const
{
	CRenameObjectDialog* dialog = new CRenameObjectDialog(m_Object);
	g_return_if_fail(dialog);
}

///////////////////////////////////////////////////////////////////////////////
// CHierarchyControl

const sdpString CHierarchyControl::dialogname = "hierarchy";
const sdpString CHierarchyControl::highlightitem = "highlightitem";
const sdpString CHierarchyControl::highlightrange = "highlightrange";
const sdpString CHierarchyControl::togglehighlight = "togglehighlight";
const sdpString CHierarchyControl::togglevisible = "togglevisible";
const sdpString CHierarchyControl::toggleselected = "toggleselected";
const sdpString CHierarchyControl::editproperties = "editproperties";
const sdpString CHierarchyControl::setparent = "setparent";
const sdpString CHierarchyControl::setsibling = "setsibling";
const sdpString CHierarchyControl::control_edithighlighted = "edithighlighted";
const sdpString CHierarchyControl::control_renamehighlighted = "renamehighlighted";
const sdpString CHierarchyControl::control_deletehighlighted = "deletehighlighted";
const sdpString CHierarchyControl::control_savehighlighted = "savehighlighted";
const sdpString CHierarchyControl::control_cuthighlighted = "cuthighlighted";
const sdpString CHierarchyControl::control_duplicatehighlighted = "duplicatehighlighted";
const sdpString CHierarchyControl::control_copyhighlighted = "copyhighlighted";
const sdpString CHierarchyControl::control_paste = "paste";
const sdpString CHierarchyControl::control_dragdatareceived = "dragdatareceived";

CHierarchyControl::CHierarchyControl(k3dIDocument* Document, sdpIUnknown* ParentCommandNode, sdpGtkWidget Owner) :
	CHierarchyControlBase(ParentCommandNode, dialogname, 0),
	m_Document(Document)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Owner.Attached());

	g_return_if_fail(LoadGTKMLTemplate("hierarchycontrol.gtkml"));

	// Attach ourselves to the parent widget ...
	sdpGtkContainer(GTK_CONTAINER(Owner.Object())).Attach(RootWidget());

	// Extract our various resources ...
	m_BackgroundColor = sdpFromString<sdpVector4>(Resource("backgroundcolor"), sdpVector4(1, 1, 1, 1));
	m_TextColor = sdpFromString<sdpVector4>(Resource("textcolor"), sdpVector4(0, 0, 0, 1));
	m_HighlightedColor = sdpFromString<sdpVector4>(Resource("highlightedcolor"), sdpVector4(0, 0, 0.5, 0.5));
	m_HighlightedTextColor = sdpFromString<sdpVector4>(Resource("highlightedtextcolor"), sdpVector4(1, 1, 1, 1));

	// Setup OpenGL ...
	sdpGtkContainer container(Container("hierarchy"));

	if(!m_Widget.Create(container, true, true, 8, 8, 8, 0))
		if(!m_Widget.Create(container, true, true, 5, 5, 5, 0))
			if(!m_Widget.Create(container, false, false, 4, 4, 4, 0))
				std::cerr << "CHierarchyControl::CHierarchyControl(): Could not find useable OpenGL visual" << std::endl;

	if(m_Widget.Initialized())
		{
			m_Widget.InitializeFont();

			// Setup some events ...
			MapEvent("configure-event", "configure", false, m_Widget, true);
			MapEvent("expose-event", "expose", false, m_Widget, true);
			MapEvent("motion-notify-event", "mousemove", false, m_Widget, true);
			MapEvent("button-press-event", "buttondown", false, m_Widget, true);
			MapEvent("button-release-event", "buttonup", false, m_Widget, true);

			// Set ourselves up for drag-n-drop ...
			m_Widget.SetDragDestination(GtkDestDefaults(GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_HIGHLIGHT | GTK_DEST_DEFAULT_DROP), k3dDNDCreateObject(), 1, GDK_ACTION_COPY);
			MapEvent("drag-data-received", control_dragdatareceived, false, m_Widget, true);
		}

	// Setup user controls ...
	AddCommand(control_edithighlighted, control_edithighlighted, CONTROL_BUTTON);
	AddCommand(control_renamehighlighted, control_renamehighlighted, CONTROL_BUTTON);
	AddCommand(control_deletehighlighted, control_deletehighlighted, CONTROL_BUTTON);
	AddCommand(control_duplicatehighlighted, control_duplicatehighlighted, CONTROL_BUTTON);
	AddCommand(control_cuthighlighted, control_cuthighlighted, CONTROL_BUTTON);
	AddCommand(control_copyhighlighted, control_copyhighlighted, CONTROL_BUTTON);
	AddCommand(control_paste, control_paste, CONTROL_BUTTON);
//	AddCommand(control_savehighlighted, control_savehighlighted, CONTROL_BUTTON);

	UpdateRecords();
	Show();

	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(m_Document->Hierarchy());
	g_return_if_fail(eventserver);
	eventserver->AddEventObserver(this, k3dIHierarchy::Event_Changed, "CHierarchyControl");


// The Win32 port of GTK+ has some kind of refresh problem, so force one to get us going ...
#ifdef SDPWIN32
	RootWidget().QueueClear();
	sdpGtkHandlePendingEvents();
#endif
}

CHierarchyControl::~CHierarchyControl()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CHierarchyControl::~CHierarchyControl()" << std::endl;

	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(Document()->Hierarchy());
	g_return_if_fail(eventserver);
	eventserver->RemoveEventObserver(this);
}

class CHierarchyControlWarpDragObjects : public sdpGtkInteractiveWarpPointerCallback
{
public:
	CHierarchyControlWarpDragObjects(CHierarchyControl& Control, GdkModifierType Modifiers) : m_Control(Control), m_Modifiers(Modifiers) {}

	bool HandleInteractivePointerWarp(GtkWidget* const Widget, const gint XOffset, const gint YOffset)
	{
		sdpGtkWarpPointer(Widget, XOffset, YOffset);
		m_Control.RawMouseMove(Widget, m_Modifiers);

		return true;
	}

protected:
	CHierarchyControl& m_Control;
	GdkModifierType m_Modifiers;
};

sdpResult CHierarchyControl::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	if(Name == highlightitem)
		{
			HierarchyControlItemIterator item = GetItem(Arguments);
			g_return_val_if_fail(item != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Highlight Object:", k3dIApplication::LMB_CLICK, GdkModifierType(0));
			sdpVector2 target = item->BoundingBox().Center();
			Widget("hierarchy").InteractiveWarpPointer(gint(target[0]), gint(target[1]), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GdkModifierType(0));
			RawButtonUp(Widget("hierarchy"), 1, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == togglehighlight)
		{
			HierarchyControlItemIterator item = GetItem(Arguments);
			g_return_val_if_fail(item != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Toggle Highlight:", k3dIApplication::LMB_CLICK, GDK_CONTROL_MASK);
			sdpVector2 target = item->BoundingBox().Center();
			Widget("hierarchy").InteractiveWarpPointer(gint(target[0]), gint(target[1]), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GDK_CONTROL_MASK);
			RawButtonUp(Widget("hierarchy"), 1, GDK_CONTROL_MASK);

			return SDP_S_OK;
		}
	else if(Name == highlightrange)
		{
			HierarchyControlItemIterator item = GetItem(Arguments);
			g_return_val_if_fail(item != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Highlight Range:", k3dIApplication::LMB_CLICK, GDK_CONTROL_MASK);
			sdpVector2 target = item->BoundingBox().Center();
			Widget("hierarchy").InteractiveWarpPointer(gint(target[0]), gint(target[1]), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GDK_SHIFT_MASK);
			RawButtonUp(Widget("hierarchy"), 1, GDK_SHIFT_MASK);

			return SDP_S_OK;
		}
	else if(Name == togglevisible)
		{
			HierarchyControlItemIterator item = GetItem(Arguments);
			g_return_val_if_fail(item != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Toggle Visibility:", k3dIApplication::LMB_CLICK, GdkModifierType(0));
			sdpVector2 target = item->VisibilityBox().Center();
			Widget("hierarchy").InteractiveWarpPointer(gint(target[0]), gint(target[1]), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GdkModifierType(0));
			RawButtonUp(Widget("hierarchy"), 1, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == toggleselected)
		{
			HierarchyControlItemIterator item = GetItem(Arguments);
			g_return_val_if_fail(item != m_Items.end(), SDP_E_FAIL);

			sdpVector2 target = item->SelectionBox().Center();

			Widget("hierarchy").InteractiveWarpPointer(gint(target[0]), gint(target[1]), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);
			k3dApplication()->TutorialMouseDisplay("Toggle Selected/Unselected:", k3dIApplication::LMB_CLICK, GdkModifierType(0));

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GdkModifierType(0));
			RawButtonUp(Widget("hierarchy"), 1, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == editproperties)
		{
			HierarchyControlItemIterator item = GetItem(Arguments);
			g_return_val_if_fail(item != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Open Properties", k3dIApplication::LMB_DOUBLE_CLICK, GdkModifierType(0));
			sdpVector2 target = item->BoundingBox().Center();
			Widget("hierarchy").InteractiveWarpPointer(gint(target[0]), gint(target[1]), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GdkModifierType(0));
			RawButtonUp(Widget("hierarchy"), 1, GdkModifierType(0));
			RawButtonDown(Widget("hierarchy"), GDK_2BUTTON_PRESS, 1, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == setparent)
		{
			HierarchyControlItemIterator parent = GetItem(Arguments);
			g_return_val_if_fail(parent != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Reparent Object:", k3dIApplication::LMB_DRAG, GdkModifierType(0));

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GdkModifierType(0));

			sdpVector2 target = parent->ParentDropBox().Center();
			CHierarchyControlWarpDragObjects callback(*this, GdkModifierType(0));
			sdpGtkInteractiveWarpPointer(Widget("hierarchy"), gint(target[0]), gint(target[1]), callback, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonUp(Widget("hierarchy"), 1, GdkModifierType(0));

			return SDP_S_OK;
		}
	else if(Name == setsibling)
		{
			HierarchyControlItemIterator sibling = GetItem(Arguments);
			g_return_val_if_fail(sibling != m_Items.end(), SDP_E_FAIL);

			k3dApplication()->TutorialMouseDisplay("Reorder Object:", k3dIApplication::LMB_DRAG, GdkModifierType(0));

			RawButtonDown(Widget("hierarchy"), GDK_BUTTON_PRESS, 1, GdkModifierType(0));

			sdpVector2 target = sibling->SiblingDropBox().Center();
			CHierarchyControlWarpDragObjects callback(*this, GdkModifierType(0));
			sdpGtkInteractiveWarpPointer(Widget("hierarchy"), gint(target[0]), gint(target[1]), callback, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);

			RawButtonUp(Widget("hierarchy"), 1, GdkModifierType(0));

			return SDP_S_OK;
		}

	return CHierarchyControlBase::Command(Name, Arguments, Result);
}

// k3dIEventObserver implementation
sdpResult CHierarchyControl::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIHierarchy::Event_Changed)
		{
			UpdateRecords();
			UpdateScrollbars();

#ifdef SDPWIN32
			RedrawControl();
#endif // SDPWIN32

			return SDP_S_OK;
		}

	return CHierarchyControlBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

// Called to load the frame from a document ...
bool CHierarchyControl::Load(sdpxml::Element& Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element.Name() == "hierarchycontrol", false);

	return true;
}

HierarchyItems CHierarchyControl::HighlightedItems() const
{
	HierarchyItems results;

	for(ConstHierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		if(item->IsHighlighted())
			results.push_back(*item);

	return results;
}

bool CHierarchyControl::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	Element.Append(sdpxml::Element("hierarchycontrol"));

	return true;
}

k3dIDocument* CHierarchyControl::Document() const
{
	return m_Document;
}

void CHierarchyControl::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "configure")
		OnConfigure();
	else if(Event->Name() == "scrollbar")
		OnScrollbar();
	else if(Event->Name() == "expose")
		OnDraw();
	else if(Event->Name() == "mousemove")
		RawMouseMove(Event);
	else if(Event->Name() == "buttondown")
		RawButtonDown(Event);
	else if(Event->Name() == "buttonup")
		RawButtonUp(Event);
	else if(Event->Name() == control_edithighlighted)
		OnEditHighlighted();
	else if(Event->Name() == control_renamehighlighted)
		OnRenameHighlighted();
	else if(Event->Name() == control_deletehighlighted)
		OnDeleteHighlighted();
	else if(Event->Name() == control_savehighlighted)
		OnSaveHighlighted();
	else if(Event->Name() == control_duplicatehighlighted)
		OnDuplicateHighlighted();
	else if(Event->Name() == control_cuthighlighted)
		OnCutHighlighted();
	else if(Event->Name() == control_copyhighlighted)
		OnCopyHighlighted();
	else if(Event->Name() == control_paste)
		OnPaste();
	else if(Event->Name() == control_dragdatareceived)
		OnDragDataReceived(Event);
	else
		CHierarchyControlBase::OnEvent(Event);
}

void CHierarchyControl::OnEditHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_edithighlighted, 0);

	for(ConstHierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		if(item->IsHighlighted())
			item->EditObject();
}

void CHierarchyControl::OnRenameHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_renamehighlighted, 0);

	for(ConstHierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		if(item->IsHighlighted())
			item->RenameObject();
}

void CHierarchyControl::OnDeleteHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_deletehighlighted, 0);
	k3dRecordStateChangeSet changeset(Document(), "Delete Highlighted");

	DeleteHighlighted();

	k3dRedrawAll(Document(), false);
}

void CHierarchyControl::DeleteHighlighted()
{
	typedef std::vector<k3dIObject*> Objects;
	Objects objects;
	for(ConstHierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		if(item->IsHighlighted())
			objects.push_back(item->Object());

	k3d::PushEditing(Document()->Objects());
	for(Objects::iterator object = objects.begin(); object != objects.end(); object++)
		Document()->Objects()->DeleteObject(*object);
	k3d::PopEditing(Document()->Objects());
}

void CHierarchyControl::OnSaveHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_savehighlighted, 0);

	sdpString filepath;
	if(!k3dGetFilePath("objectlibrary", "Save Highlighted Objects:", true, 0, filepath))
		return;

	std::ofstream file(filepath);
	g_return_if_fail(file.good());

	CopyHighlighted(file);
}

void CHierarchyControl::OnDuplicateHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_duplicatehighlighted, 0);
	k3dRecordStateChangeSet changeset(Document(), "Duplicate Highlighted");

	CopyHighlighted();
	Paste();

	k3dRedrawAll(Document(), false);
}

void CHierarchyControl::OnCutHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_cuthighlighted, 0);
	k3dRecordStateChangeSet changeset(Document(), "Cut Highlighted");

	CutHighlighted();

	k3dRedrawAll(Document(), false);
}

void CHierarchyControl::OnCopyHighlighted()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_copyhighlighted, 0);
	CopyHighlighted();
}

void CHierarchyControl::OnPaste()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_paste, 0);
	k3dRecordStateChangeSet changeset(Document(), "Paste");

	Paste();

	k3dRedrawAll(Document(), false);
}

void CHierarchyControl::CutHighlighted()
{
	CopyHighlighted();
	DeleteHighlighted();
}

void CHierarchyControl::CopyHighlighted()
{
	std::strstream text;
	CopyHighlighted(text);

	k3dApplication()->SetClipboard(text.str());

	text.freeze(false);
}

void CHierarchyControl::CopyHighlighted(std::ostream& Stream)
{
	// Keep track of objects that have already been copied, to prevent duplicates ...
	typedef std::set<k3dIObject*> ObjectSet;
	ObjectSet writtenobjects;

	// Keep track of inter-object dependencies ...
	k3dDependencies dependencies;

	// We're going to generate a K3DML document ...
	sdpxml::Document xmldocument("k3dml");

	sdpxml::Element& document = xmldocument.Append(sdpxml::Element("document"));
	sdpxml::Element& objects = document.Append(sdpxml::Element("objects"));

	// For each highlighted object ...
	for(ConstHierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		{
			if(!item->IsHighlighted())
				continue;

			// If this object has already been copied, skip it ...
			k3dIObject* object = item->Object();
			if(writtenobjects.find(object) != writtenobjects.end())
				continue;
			writtenobjects.insert(object);

			// If the object doesn't support persistence, no-can-do ...
			k3dIPersist* persist = dynamic_cast<k3dIPersist*>(object);
			if(!persist)
				continue;

			// Create an XML element for the object ...
			sdpxml::Element& objectelement = objects.Append(sdpxml::Element("object", "",
				sdpxml::Attribute("name", object->Name()),
				sdpxml::Attribute("class", sdpToString(object->ObjectInfo()->ClassID())),
				sdpxml::Attribute("id", sdpToString(object->ObjectID()))));

			// Tell the object to save itself ...
			persist->Save(objectelement, &dependencies);
		}

	// For each dependent object ...
	for(ObjectSet::const_iterator objectiterator = dependencies.Objects().begin(); objectiterator != dependencies.Objects().end(); objectiterator++)
		{
			// If this object has already been copied, skip it ...
			k3dIObject* object = *objectiterator;
			if(writtenobjects.find(object) != writtenobjects.end())
				continue;
			writtenobjects.insert(object);

			// If the object doesn't support persistence, no-can-do ...
			k3dIPersist* persist = dynamic_cast<k3dIPersist*>(object);
			if(!persist)
				continue;

			// Create an XML element for the object ...
			sdpxml::Element& objectelement = objects.Append(sdpxml::Element("object", "",
				sdpxml::Attribute("name", object->Name()),
				sdpxml::Attribute("class", sdpToString(object->ObjectInfo()->ClassID())),
				sdpxml::Attribute("id", sdpToString(object->ObjectID()))));

			// Tell the object to save itself ...
			persist->Save(objectelement, &dependencies);
		}

	// Put the results in our output stream ...
	Stream << xmldocument << std::ends;
}

void CHierarchyControl::Paste()
{
	sdpGtkClipboardText clipboard;
	std::istrstream stream(clipboard.GetText().c_str());

	// Parse it as XML ...
	sdpxml::Document xmldocument("k3dml");
	g_return_if_fail(xmldocument.Load(stream, "Clipboard Text"));

	// Make sure we've got a K3DML document ...
	g_return_if_fail(xmldocument.Name() == "k3dml");

	// Look for document data ...
	sdpxml::ElementPointer document = sdpxml::FindElement(xmldocument, sdpxml::SameName("document"));
	if(!document)
		return;

	// Look for object data ...
	sdpxml::ElementPointer objects = sdpxml::FindElement(*document, sdpxml::SameName("objects"));
	if(!objects)
		return;

	// Keep track of any new objects we create ...
	typedef std::set<k3dIObject*> ObjectSet;
	ObjectSet newobjects;

	// For each element in the document ...
	for(sdpxml::ElementCollection::iterator objectelement = objects->Children().begin(); objectelement != objects->Children().end(); objectelement++)
		{
			// If it isn't an <object> element, skip it ...
			if(objectelement->Name() != "object")
				continue;

			// Let the document "thaw" the object ...
			k3dIObject* const object = Document()->Objects()->CreateObject(xmldocument, *objectelement);
			g_return_if_fail(object);

			newobjects.insert(object);
		}

	// Let objects know they're done being loaded ...
	for(ObjectSet::iterator object = newobjects.begin(); object != newobjects.end(); ++object)
		{
			// The object BETTER support k3dIPersist - if not, how did it end up on the clipboard?
			k3dIPersist* const persist = dynamic_cast<k3dIPersist*>(*object);
			g_return_if_fail(persist);
			persist->LoadComplete();
		}

	// Let objects know they're done being initialized ...
	std::for_each(newobjects.begin(), newobjects.end(), std::mem_fun(&k3dIObject::InitComplete));

	RedrawControl();
}

void CHierarchyControl::OnDragDataReceived(sdpGtkEvent* const Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Get the drag data ...
	sdpGtkEventWidgetDragDataReceived* const event = static_cast<sdpGtkEventWidgetDragDataReceived*>(Event);
	GtkSelectionData* const selection = event->Selection();
	std::istrstream stream(reinterpret_cast<char*>(selection->data));

	// Parse it as XML ...
	sdpxml::Document xmldocument("k3dml");
	g_return_if_fail(xmldocument.Load(stream, "Drag-and-drop text"));

	// Make sure we've got a K3DML document ...
	g_return_if_fail(xmldocument.Name() == "k3dml");

	// Look for application data ...
	sdpxml::ElementPointer application = sdpxml::FindElement(xmldocument, sdpxml::SameName("application"));
	if(!application)
		return;

	// Look for object data ...
	sdpxml::ElementPointer objects = sdpxml::FindElement(*application, sdpxml::SameName("objects"));
	if(!objects)
		return;

	// Record undos ...
	k3dRecordStateChangeSet changeset(Document(), "Drag-and-drop new object");

	// For each element in the object data ...
	for(sdpxml::ElementCollection::iterator objectelement = objects->Children().begin(); objectelement != objects->Children().end(); objectelement++)
		{
			// If it isn't an <object> element, skip it ...
			if(objectelement->Name() != "object")
				continue;

			sdp::ClassID classid(sdp::ClassIDNull);
			g_return_if_fail(sdpxml::ParseAttribute(*objectelement, "class", classid));
			g_return_if_fail(classid != sdp::ClassIDNull);

			k3dIObject* const object = Document()->Objects()->CreateObject(classid);
			g_return_if_fail(object);
		}

	k3dRedrawAll(Document(), false);
}

// Called when the display is resized ...
void CHierarchyControl::OnConfigure()
{
//std::cout << "CHierarchyControl::OnConfigure()" << std::endl;
	UpdateScrollbars();
}

void CHierarchyControl::OnScrollbar()
{
//std::cout << "CHierarchyControl::OnScrollbar()" << std::endl;

//std::cout << Adjustment("vscrollbar").Value() << std::endl;

	RedrawControl();
}

// Called to refresh the display ...
void CHierarchyControl::OnDraw()
{
//std::cout << "CHierarchyControl::OnDraw()" << std::endl;
	sdpglPushTrace(false);

	// If we're not completely initialized, we're done ...
	if(!m_Widget.Initialized())
		return;

	m_Widget.Begin();

	// Use the widget's font ...
	glListBase(m_Widget.FontBase());

	// Get the widget sizes ...
	gulong width = m_Widget.Width();
	gulong height = m_Widget.Height();
	glViewport(0, 0, width, height);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, width, height, 0, -2, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslated(-Adjustment("hscrollbar").Value(), -Adjustment("vscrollbar").Value(), 0);

	// Clear the background ...
	glClearColor(m_BackgroundColor[0], m_BackgroundColor[1], m_BackgroundColor[2], 0);
	glClear(GL_COLOR_BUFFER_BIT);

	// Setup drawing options ...
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	// Draw items ...
	DrawItems();

	// Draw drag items ...
	DrawDragItems();

	// Cleanup ...
	m_Widget.SwapBuffers();
	m_Widget.End();

	sdpglPopTrace();
}

void CHierarchyControl::DrawItems()
{
	gdouble top = Adjustment("vscrollbar").Value();
	gdouble bottom = top + m_Widget.Height();

	for(ConstHierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		{
			if(item->BoundingBox().bottom < top)
				continue;

			if(item->BoundingBox().top > bottom)
				continue;

			item->Draw();
		}
}

void CHierarchyControl::DrawDragItems()
{
	for(ConstHierarchyDragIterator item = m_DragItems.begin(); item != m_DragItems.end(); item++)
		item->second.Draw();
}

void CHierarchyControl::RedrawControl()
{
	Widget("hierarchy").QueueDraw();
}

void CHierarchyControl::UpdateRecords()
{
std::cerr << g_pretty_function() << std::endl;

	m_Items.clear();

	k3dIHierarchy* hierarchy = Document()->Hierarchy();
	g_return_if_fail(hierarchy);

	// For each node in the hierarchy tree ...
	gulong count = 0;
	gulong generation = 0;
	K3DHIERARCHYNODEHANDLE handle;
	k3dIObject* object;
	for(k3dIHierarchy::NodeType type = hierarchy->FirstHierarchyNode(&handle); k3dIHierarchy::COMPLETE != type; type = hierarchy->NextHierarchyNode(&handle))
		{
			object = hierarchy->HierarchyObject(handle);

			switch(type)
				{
					case k3dIHierarchy::PUSH:
						generation++;
						break;
					case k3dIHierarchy::NODE:
						if(SDP_S_OK == object->State()->QueryState(k3dIState::USERVIEWABLE))
							m_Items.push_back(CHierarchyControlItem(this, &m_Widget, object, count++, generation-1));
						break;
					case k3dIHierarchy::POP:
						generation--;
						break;
					case k3dIHierarchy::COMPLETE:
						break;
					case k3dIHierarchy::ERROR:
						break;
				}
		}
}

void CHierarchyControl::UpdateScrollbars()
{
//std::cout << "CHierarchyControl::UpdateScrollbars()" << std::endl;

	if(!m_Widget.Initialized())
		return;

	gdouble itemheight = 0;
	gulong widgetwidth = m_Widget.Width();
	gulong widgetheight = m_Widget.Height();

	sdpRectangle boundingbox(DBL_MAX, -DBL_MAX, DBL_MAX, -DBL_MAX);
	if(m_Items.size())
		{
			itemheight = m_Items.front().BoundingBox().Height();

			for(HierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
				{
					sdpRectangle itemboundingbox = item->BoundingBox();
					boundingbox.left = sdpMin(boundingbox.left, itemboundingbox.left);
					boundingbox.right = sdpMax(boundingbox.right, itemboundingbox.right);
					boundingbox.top = sdpMin(boundingbox.top, itemboundingbox.top);
					boundingbox.bottom = sdpMax(boundingbox.bottom, itemboundingbox.bottom);
				}
		}
	else
		{
			boundingbox.left = 0.0;
			boundingbox.top = 0.0;
			boundingbox.right = widgetwidth;
			boundingbox.bottom = widgetheight;
		}

	sdpGtkAdjustment hadjustment = Adjustment("hscrollbar");
	sdpGtkAdjustment vadjustment = Adjustment("vscrollbar");

	hadjustment.SetUpper(boundingbox.Width());
	vadjustment.SetUpper(boundingbox.Height());

	hadjustment.SetPageSize(widgetwidth);
	vadjustment.SetPageSize(widgetheight);

	hadjustment.SetStepIncrement(itemheight * 0.5);
	vadjustment.SetStepIncrement(itemheight * 0.5);

	hadjustment.Changed();
	vadjustment.Changed();
}

HierarchyControlItemIterator CHierarchyControl::HitTest(sdpVector2 LogicalCoordinates)
{
	HierarchyControlItemIterator result;
	for(result = m_Items.begin(); result != m_Items.end(); result++)
		if(result->BoundingBox().Contains(LogicalCoordinates))
			break;

	return result;
}

HierarchyControlItemIterator CHierarchyControl::GetItem(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpString name(Name);
	HierarchyControlItemIterator result;
	for(result = m_Items.begin(); result != m_Items.end(); result++)
		{
			if(result->Object()->Name() == name)
				break;
		}

	return result;
}

sdpVector2 CHierarchyControl::ScreenToLogical(const sdpVector2 ScreenCoordinates)
{
	sdpVector2 result(ScreenCoordinates);

	result[0] += Adjustment("hscrollbar").Value();
	result[1] += Adjustment("vscrollbar").Value();

	return result;
}

void CHierarchyControl::OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	sdpVector2 logicalmouse = ScreenToLogical(CurrentMouse);

	HierarchyControlItemIterator item = HitTest(logicalmouse);
	if(item == m_Items.end())
		return;

	if(item->VisibilityBox().Contains(logicalmouse))
		return;

	if(item->SelectionBox().Contains(logicalmouse))
		return;

	if(Modifiers & GDK_SHIFT_MASK)
		{
			k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, highlightrange, item->Object()->Name());

/*
			bool highlight = false;
			for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(): i++)
				{
					if((item == i) || ())
						highlight = !highlight;

					if(highlight)
						i->SetState(CHierarchyControlItem::HIGHLIGHT);
				}
*/
		}
	else if(Modifiers & GDK_CONTROL_MASK)
		{
			k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, togglehighlight, item->Object()->Name());
			item->SetState(item->IsHighlighted() ? CHierarchyControlItem::NORMAL : CHierarchyControlItem::HIGHLIGHT);
		}
	else
		{
			k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, highlightitem, item->Object()->Name());

			for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(); i++)
				i->SetState(i->Object() == item->Object() ? CHierarchyControlItem::HIGHLIGHT : CHierarchyControlItem::NORMAL);
		}

	RedrawControl();
}

void CHierarchyControl::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	sdpVector2 logicalmouse = ScreenToLogical(CurrentMouse);

	HierarchyControlItemIterator item = HitTest(logicalmouse);
	if(item == m_Items.end())
		{
			for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(); i++)
				i->SetState(CHierarchyControlItem::NORMAL);

			RedrawControl();
		}
	else
		{
			if(item->VisibilityBox().Contains(logicalmouse))
				{
					k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, togglevisible, item->Object()->Name());

					k3dDeepSetState(item->Object(), k3dIState::VISIBLE, SDP_S_OK == item->Object()->State()->QueryState(k3dIState::VISIBLE) ? false : true);
					k3dRedrawAll(Document(), false);
				}
			else if(item->SelectionBox().Contains(logicalmouse))
				{
					k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, toggleselected, item->Object()->Name());

					k3dDeepSetState(item->Object(), k3dIState::SELECTED, SDP_S_OK == item->Object()->State()->QueryState(k3dIState::SELECTED) ? false : true);

					k3dIMouseEventObserver* const mouseobserver = dynamic_cast<k3dIMouseEventObserver*>(item->Object());
					if(mouseobserver)
						{
							if(SDP_S_OK == item->Object()->State()->QueryState(k3dIState::SELECTED))
								{
									Document()->SetMouseFocus(mouseobserver);
								}
							else
								{
									if(Document()->MouseFocus() == mouseobserver)
										Document()->SetMouseFocus(0);
								}
						}

					k3dRedrawAll(Document(), false);
				}
		}
}

void CHierarchyControl::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	sdpVector2 logicalmouse = ScreenToLogical(CurrentMouse);

	HierarchyControlItemIterator target = HitTest(logicalmouse);
	if(target == m_Items.end())
		return;

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, editproperties, target->Object()->Name());

	for(HierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		item->SetState(item->Object() == target->Object() ? CHierarchyControlItem::HIGHLIGHT : CHierarchyControlItem::NORMAL);

	RedrawControl();

	target->EditObject();
}

void CHierarchyControl::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	for(HierarchyControlItemIterator item = m_Items.begin(); item != m_Items.end(); item++)
		{
			if(item->IsHighlighted())
				{
					m_DragItems.push_back(HierarchyDragItem(item, *item));
					m_DragItems.back().second.SetState(CHierarchyControlItem::DRAG);
				}
		}

	RedrawControl();
}

void CHierarchyControl::MarkDropTarget(HierarchyControlItemIterator Target, const sdpVector2 LogicalMouse)
{
	std::vector<CHierarchyControlItem*> dropsources;
	for(HierarchyDragIterator item = m_DragItems.begin(); item != m_DragItems.end(); item++)
		dropsources.push_back(&(*item->first));

	Target->SetDropTarget(LogicalMouse, dropsources);
}

void CHierarchyControl::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	if(0 == m_DragItems.size())
		return;

	// Move drag items ...
	for(HierarchyDragIterator item = m_DragItems.begin(); item != m_DragItems.end(); item++)
		{
			sdpVector2 newposition = item->first->Position() + (CurrentMouse - StartMouse);
			item->second.SetPosition(newposition);
		}

	// Find a drop target ...
	sdpVector2 logicalmouse = ScreenToLogical(CurrentMouse);

	HierarchyControlItemIterator target = HitTest(logicalmouse);
	if(target == m_Items.end())
		return;

	// Let it know it's the chosen one ...
	for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(); i++)
		i->SetState(CHierarchyControlItem::NORMAL);

	MarkDropTarget(target, logicalmouse);

	RedrawControl();
}

void CHierarchyControl::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	if(0 == m_DragItems.size())
		return;

	// Find a drop target ...
	const sdpVector2 logicalmouse = ScreenToLogical(CurrentMouse);
	const HierarchyControlItemIterator target = HitTest(logicalmouse);
	if(target != m_Items.end())
		{
			// Let it know it's the chosen one ...
			MarkDropTarget(target, logicalmouse);

			k3dIHierarchy* const hierarchy = Document()->Hierarchy();
			g_return_if_fail(hierarchy);

			if(target->IsParentDropTarget())
				{
					k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, setparent, target->Object()->Name());

					k3dStartStateChangeSet(Document());
					k3d::PushEditing(hierarchy);

					for(HierarchyDragIterator dragitem = m_DragItems.begin(); dragitem != m_DragItems.end(); dragitem++)
						hierarchy->SetObjectParent(dragitem->second.Object(), target->Object());

					k3d::PopEditing(hierarchy);
					k3dFinishStateChangeSet(Document(), "Reparent object(s)");

					k3dRedrawAll(Document(), false);
				}
			else if(target->IsSiblingDropTarget())
				{
					k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, setsibling, target->Object()->Name());

					k3dStartStateChangeSet(Document());
					k3d::PushEditing(hierarchy);

					for(HierarchyDragIterator dragitem = m_DragItems.begin(); dragitem != m_DragItems.end(); dragitem++)
						hierarchy->SetObjectSibling(dragitem->second.Object(), target->Object());

					k3d::PopEditing(hierarchy);
					k3dFinishStateChangeSet(Document(), "Reorder object(s)");

					k3dRedrawAll(Document(), false);
				}
		}

	// We shouldn't have any drop targets at this point ...
	for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(); i++)
		i->SetState(CHierarchyControlItem::NORMAL);

	m_DragItems.clear();
	RootWidget().QueueDraw();
}

void CHierarchyControl::OnRButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	// Highlight the object under the pointer before proceeding ...
	sdpVector2 logicalmouse = ScreenToLogical(CurrentMouse);

	HierarchyControlItemIterator item = HitTest(logicalmouse);
	if(item == m_Items.end())
		return;

	if(Modifiers & GDK_SHIFT_MASK)
		{
			k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, highlightrange, item->Object()->Name());

/*
			bool highlight = false;
			for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(): i++)
				{
					if((item == i) || ())
						highlight = !highlight;

					if(highlight)
						i->SetState(CHierarchyControlItem::HIGHLIGHT);
				}
*/
		}
	else if(Modifiers & GDK_CONTROL_MASK)
		{
			k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, togglehighlight, item->Object()->Name());
			item->SetState(item->IsHighlighted() ? CHierarchyControlItem::NORMAL : CHierarchyControlItem::HIGHLIGHT);
		}
	else
		{
			k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, highlightitem, item->Object()->Name());

			for(HierarchyControlItemIterator i = m_Items.begin(); i != m_Items.end(); i++)
				i->SetState(i->Object() == item->Object() ? CHierarchyControlItem::HIGHLIGHT : CHierarchyControlItem::NORMAL);
		}

	RedrawControl();

	g_return_if_fail(m_Popup.Load(sdpAppendPath(k3dApplication()->DialogTemplatesPath(), "hierarchycontextmenu.gtkml"), k3dOptionsGetPreferredLanguage(k3dApplication()->Options())));

	MapEvent("activate", control_edithighlighted, false, m_Popup.Object(control_edithighlighted), true);
	MapEvent("activate", control_renamehighlighted, false, m_Popup.Object(control_renamehighlighted), true);
	MapEvent("activate", control_deletehighlighted, false, m_Popup.Object(control_deletehighlighted), true);
	MapEvent("activate", control_duplicatehighlighted, false, m_Popup.Object(control_duplicatehighlighted), true);
	MapEvent("activate", control_cuthighlighted, false, m_Popup.Object(control_cuthighlighted), true);
	MapEvent("activate", control_copyhighlighted, false, m_Popup.Object(control_copyhighlighted), true);
	MapEvent("activate", control_paste, false, m_Popup.Object(control_paste), true);
	MapEvent("activate", control_savehighlighted, false, m_Popup.Object(control_savehighlighted), true);

	m_Popup.RootWidget().Show();
	m_Popup.RootMenu().Popup(3);
}


