#ifndef HIERARCHYCONTROL_H
#define HIERARCHYCONTROL_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CHierarchyControl class, which encapsulates a hierarchical view of an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>
#include <k3deventobserver.h>

#include <sdpgtkmouseinput.h>
#include <sdpgtkopengldrawingarea.h>

// Forward class declarations
class CHierarchyControl;
class CHierarchyControlItem;
class k3dIDocument;
class k3dIDependencies;
class sdpxml::Element;

///////////////////////////////////////////////////////////
// CHierarchyControlItem

/// Encapsulates a single visible item within the hierarchy control
class CHierarchyControlItem : public k3dEventObserver
{
public:
	CHierarchyControlItem(CHierarchyControl* const ParentControl, sdpGtkOpenGLDrawingArea* const DrawingArea, k3dIObject* const Object, const gulong Index, const gulong Generation);
	CHierarchyControlItem(const CHierarchyControlItem& Original);
	
	virtual ~CHierarchyControlItem();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	/// Returns the K-3D object represented by this item
	k3dIObject* Object() const { return m_Object; }

	/// Gives the item a chance to render itself
	void Draw() const;

	/// Returns the item's position
	sdpVector2 Position() const { return m_Position; }
	/// Sets the item's position
	void SetPosition(const sdpVector2 Position) { m_Position = Position; }

	/// Opens the UI of the K-3D object represented by this item
	void EditObject() const;
	/// Opens a dialog for renaming the K-3D object represented by this item
	void RenameObject() const;

	/// Returns this item's bounding box (in logical coordinates)
	sdpRectangle BoundingBox() const;
	/// Returns this item's standard box (in logical coordinates)
	sdpRectangle StandardBox() const;
	/// Returns this item's visibility check box (in logical coordinates)
	sdpRectangle VisibilityBox() const;
	/// Returns this item's selection check box (in logical coordinates)
	sdpRectangle SelectionBox() const;
	/// Returns this item's text box (in logical coordinates)
	sdpRectangle TextBox() const;
	/// Returns this item's parent-drop-target box (in logical coordinates)
	sdpRectangle ParentDropBox() const;
	/// Returns this item's sibling-drop-target box (in logical coordinates)
	sdpRectangle SiblingDropBox() const;

	/// Enumerates item states
	typedef enum
	{
		NORMAL,
		HIGHLIGHT,
		DRAG,
		PARENTDROPTARGET,
		SIBLINGDROPTARGET,
	} ItemState;

	void SetState(const ItemState State) { m_State = State; }

	bool IsNormal() const { return NORMAL == m_State; }
	bool IsHighlighted() const { return HIGHLIGHT == m_State; }
	bool IsDragging() const { return DRAG == m_State; }
	bool IsParentDropTarget() const { return PARENTDROPTARGET == m_State; }
	bool IsSiblingDropTarget() const { return SIBLINGDROPTARGET == m_State; }

	void SetDropTarget(const sdpVector2 MouseCoordinates, const std::vector<CHierarchyControlItem*>& DropSources);

protected:
	static const gdouble m_ItemOffset;
	static const gdouble m_InteriorBorder;
	static const gdouble m_HorizontalPadding;
	static const gdouble m_ParentBoxThickness;
	static const gdouble m_SiblingLineThickness;

	/// Computes the standard item height, based on drawing-area font size ...
	static gdouble Height(const sdpGtkOpenGLDrawingArea* const DrawingArea);

	/// Stores the control that owns this item
	CHierarchyControl* const m_ParentControl;
	/// Stores the OpenGL area that we'll be drawing into
	sdpGtkOpenGLDrawingArea* const m_DrawingArea;
	/// Stores the K-3D object that this item represents
	k3dIObject* const m_Object;
	/// Stores the highlighted state of this item
	bool m_Highlighted;
	/// Stores the drag state of this item
	bool m_Dragging;
	/// Stores this item's 0-based index
	gulong m_Index;
	/// Stores this item's generation
	gulong m_Generation;
	/// Stores the (pixel) size of this item
	sdpVector2 m_Size;
	/// Stores the (pixel) position of this item, relative to the parent control
	sdpVector2 m_Position;
	/// Stores the item state enumeration
	ItemState m_State;
};

typedef std::list<CHierarchyControlItem> HierarchyItems;
typedef HierarchyItems::iterator HierarchyControlItemIterator;
typedef HierarchyItems::const_iterator ConstHierarchyControlItemIterator;

typedef std::pair<HierarchyControlItemIterator, CHierarchyControlItem> HierarchyDragItem;
typedef std::list<HierarchyDragItem> HierarchyDragItems;
typedef HierarchyDragItems::iterator HierarchyDragIterator;
typedef HierarchyDragItems::const_iterator ConstHierarchyDragIterator;

///////////////////////////////////////////////////////////
// CHierarchyControl

typedef k3dDialog CHierarchyControlBase;

class CHierarchyControl : public CHierarchyControlBase,
											public sdpGtkMouseInput
{
friend class CHierarchyControlWarpDragObjects;
public:
	CHierarchyControl(k3dIDocument* Document, sdpIUnknown* ParentCommandNode, sdpGtkWidget Owner);
	virtual ~CHierarchyControl();
	
	// k3dICommandNode overrides ...
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// Initialization ...
	bool Load(sdpxml::Element& Element);
	
	// Returns a collection of currently-highlighted items ...
	HierarchyItems HighlightedItems() const;

	// Persistence ...
	bool Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	
	/// Returns the parent document
	k3dIDocument* Document() const;

	/// Called to cause a redraw of the entire control
	void RedrawControl();
	
	/// Background color
	sdpVector4 m_BackgroundColor;
	/// The normal item text color
	sdpVector4 m_TextColor;
	/// The color of selected items
	sdpVector4 m_HighlightedColor;
	/// The selected item text color
	sdpVector4 m_HighlightedTextColor;

protected:
	void ClearItems();

	// Event handlers ...
	void OnEvent(sdpGtkEvent* Event);
	
	void OnEditHighlighted();
	void OnRenameHighlighted();
	void OnDeleteHighlighted();
	void OnSaveHighlighted();
	void OnDuplicateHighlighted();
	void OnCutHighlighted();
	void OnCopyHighlighted();
	void OnPaste();
	
	void OnDragDataReceived(sdpGtkEvent* const Event);
	
	void DeleteHighlighted();
	void CutHighlighted();
	void CopyHighlighted();
	void CopyHighlighted(std::ostream& Stream);
	void Paste();
	void OnConfigure();
	void OnScrollbar();
	void OnDraw();

	/// Draws all regular items
	void DrawItems();
	/// Draws any items being dragged
	void DrawDragItems();

	/// Called to update our data when the document hierarchy is altered
	void UpdateRecords();
	/// Called to update the size / position of scrollbars
	void UpdateScrollbars();

	/// Tells a control item that we'd like to make it a drop target (it could refuse)
	void MarkDropTarget(HierarchyControlItemIterator Target, const sdpVector2 LogicalMouse);

	// sdpGtkMouseInput overrides ...
	void OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);

	/// Returns the hierarchy control record at the given screen coordinates (if any)
	HierarchyControlItemIterator HitTest(sdpVector2 ScreenCoordinates);
	/// Returns a hierarchy control record by object name
	HierarchyControlItemIterator GetItem(SDPCSTRING Name);
	/// Converts screen (scrolled) coordinates to logical (unscrolled) coordinates
	sdpVector2 ScreenToLogical(const sdpVector2 ScreenCoordinates);

	/// Stores a reference to the owning document
	k3dIDocument* m_Document;
	/// OpenGL context for drawing
	sdpGtkOpenGLDrawingArea m_Widget;
	/// Stores the collection of items for the tree view
	HierarchyItems m_Items;
	/// Stores the collection of items being dragged by the mouse
	HierarchyDragItems m_DragItems;

	/// Stores the item context menu
	sdpGtkObjectContainer m_Popup;

	static const sdpString dialogname;
	static const sdpString highlightitem;
	static const sdpString highlightrange;
	static const sdpString togglehighlight;
	static const sdpString togglevisible;
	static const sdpString toggleselected;
	static const sdpString editproperties;
	static const sdpString setparent;
	static const sdpString setsibling;
	static const sdpString control_edithighlighted;
	static const sdpString control_renamehighlighted;
	static const sdpString control_deletehighlighted;
	static const sdpString control_savehighlighted;
	static const sdpString control_duplicatehighlighted;
	static const sdpString control_cuthighlighted;
	static const sdpString control_copyhighlighted;
	static const sdpString control_paste;
	static const sdpString control_dragdatareceived;
};

#endif // !HIERARCHYCONTROL_H

