// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CHierarchyFrame class, which displays a hierarchical view of an open K-3D document, using CHierarchyControl
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "hierarchyframe.h"
#include "hierarchycontrol.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <sdpgtkevents.h>

// We have an unfortunate clash with X
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

///////////////////////////////////////////////////////////////////////////////
// CHierarchyFrame

const sdpString CHierarchyFrame::dialogname = "hierarchywindow";
const sdpString CHierarchyFrame::control_fileclose = "fileclose";

CHierarchyFrame::CHierarchyFrame(CDocument* Document) :
	CHierarchyFrameBase(Document, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Document(Document)
{
	// Sanity checks ...
	g_assert(m_Document);
	
	g_return_if_fail(LoadGTKMLTemplate("hierarchyframe.gtkml"));
	
	// Create the hierarchy control ...
	m_HierarchyControl = new CHierarchyControl(Document, this, Widget("hierarchy"));
	g_return_if_fail(m_HierarchyControl);

	m_Document->AddEventObserver(this, k3dIDocument::Event_Close, "CHierarchyFrame");
			
	UpdateTitlebar();
	CallUpdateControls();
	Show();
}


CHierarchyFrame::~CHierarchyFrame()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CHierarchyFrame::~CHierarchyFrame()" << std::endl;

	Document()->RemoveEventObserver(this);

	if(m_HierarchyControl)
		{
			delete m_HierarchyControl;
			m_HierarchyControl = 0;
		}
}

sdpResult CHierarchyFrame::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIDocument::Event_Close)
		{
			delete this;
			return SDP_S_OK;
		}
	
	return CHierarchyFrameBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

// Called to load the frame from a document ...
bool CHierarchyFrame::Load(sdpxml::Element& Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element.Name() == "hierarchyframe", false);

	return true;
}

bool CHierarchyFrame::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& hierarchyframe = Element.Append(sdpxml::Element("hierarchyframe"));
	
	return m_HierarchyControl->Save(hierarchyframe, Dependencies);
}

void CHierarchyFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_fileclose)
		OnFileClose();
	else
		CHierarchyFrameBase::OnEvent(Event);
}

void CHierarchyFrame::OnDelete(sdpGtkEvent* Event)
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CHierarchyFrame::OnDelete()" << std::endl;

	// Don't let it happen ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);
	
	OnFileClose();
}

void CHierarchyFrame::OnFileClose()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CHierarchyFrame::OnFileClose()" << std::endl;

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_fileclose, 0);
	delete this;
}

void CHierarchyFrame::OnFocusIn()
{
	// If the document is in the process of being closed, do nothing ...
	if(m_Document->IsDying())
		return;

	// Activate the document ...
	k3dApplication()->SetActiveDocument(m_Document);
}

// Reset the window titlebar ...
void CHierarchyFrame::UpdateTitlebar()
{
	// Get our title and stuff it in the titlebar ...
	RootWindow().SetTitle(Title());
}

// Our window title is the same as the document ...
sdpString CHierarchyFrame::Title()
{
	return sdpString(m_Document->Title());
}


