// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CMacroRecorder class, which provides a UI for interactively recording scripts
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "macrorecorder.h"

#include <k3dcom.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3discriptengine.h>
#include <sdpgtkevents.h>

//////////////////////////////////////////////////////////////////////
// CMacroRecorder

const sdpString CMacroRecorder::dialogname = "scriptrecorder";
const sdpString CMacroRecorder::control_record = "record";
const sdpString CMacroRecorder::control_stop = "stop";

CMacroRecorder::CMacroRecorder(CApplication* Application) :
	CMacroRecorderBase(Application, dialogname),
	m_ScriptEngine(k3dApplication()->ScriptEngine(k3d::classid::JavaScriptEngine)),
	m_Recording(true)
{
	// Sanity checks ...
	g_assert(m_ScriptEngine);
	
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("macrorecorder.gtkml"));
	
	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(k3dApplication());
	if(eventserver)
		eventserver->AddEventObserver(this, k3dIApplication::Event_Close, 0);
		
	k3dApplication()->RegisterCommandRecorder(this);

	ToggleButton(control_record).SetState(m_Recording);

	Show();
}

CMacroRecorder::~CMacroRecorder()
{
	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(k3dApplication());
	if(eventserver)
		eventserver->RemoveEventObserver(this);
		
	k3dApplication()->UnRegisterCommandRecorder(this);
	m_ScriptEngine = 0;
}

//////////////////////////////////////////////////////////////////////
// k3dICommandRecorder overrides

sdpResult CMacroRecorder::IsRecording()
{
	return m_Recording ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CMacroRecorder::RecordCommand(k3dICommandNode* CommandNode, k3dICommandNode::CommandType Type, SDPCSTRING Name, SDPCSTRING Arguments)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandNode, SDP_E_FAIL);
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	// If we aren't recording ...
	if(!m_Recording)
		return SDP_S_FALSE;
			
	// Skip all UI events ...
	if(Type == k3dICommandNode::USERINTERFACE)
		return SDP_S_FALSE;

	sdpString buffer;
	gulong buffersize = 1024;
	m_ScriptEngine->ConvertCommand(CommandNode, Name, Arguments, buffer.buffer(buffersize), buffersize);
	buffer.resize(buffersize);
	buffer += "\n";

	gint position = TextControl().GetLength();
	TextControl().InsertText(buffer, buffer.size(), &position);
	
	return SDP_S_OK;
}

sdpString CMacroRecorder::GetTitlebarText()
{
	sdpString result = "Macro Recorder: " + CMacroRecorderBase::GetTitlebarText();
	
	if(m_Recording)
		result += " [recording]";
		
	return result;
}

void CMacroRecorder::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_record)
		OnRecord();
	else if(Event->Name() == control_stop)
		OnStop();
	else if(Event->Name() == control_play)
		OnPlay();
	else
		CMacroRecorderBase::OnEvent(Event);
}

void CMacroRecorder::OnRecord()
{
	m_Recording = ToggleButton(control_record).GetState();

	if(m_Recording)
		{
			Widget(control_stop).SetSensitive(true);
			Widget(control_play).SetSensitive(false);
		}
	else
		{
			Widget(control_stop).SetSensitive(false);
			Widget(control_play).SetSensitive(true);
		}
		
	UpdateTitlebar();
}

void CMacroRecorder::OnStop()
{
	ToggleButton(control_record).SetState(false);
}

sdpResult CMacroRecorder::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIApplication::Event_Close)
		{
			OnFileClose();
			return SDP_S_OK;
		}
	else
		return CMacroRecorderBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

