// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CObjectCollection, an implementation of k3dIObjectCollection
		\author Tim Shead (tshead@k-3d.com)
*/

#include "hierarchy.h"
#include "objectcollection.h"
#include "objectinfo.h"

#include <k3dmisc.h>
#include <k3dobjectinfocollection.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dianimate.h>
#include <k3didocument.h>
#include <k3dilocationcollection.h>
#include <k3dimouseeventobserver.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>
#include <k3diobjectinitializer.h>
#include <k3dipathcollection.h>
#include <k3dipersist.h>
#include <k3distate.h>
#include <k3ditimeserver.h>
#include <k3ditransform.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H

namespace
{

/// Provides an implementation of k3dIObjectInitializer
class ObjectInitializer :
	public k3dIObjectInitializer
{
public:
	ObjectInitializer(const k3dIObject::ID ObjectID, k3dIDocument* const Document, k3dIObjectInfo* const ObjectInfo) :
		m_ObjectID(ObjectID),
		m_Document(Document),
		m_ObjectInfo(ObjectInfo)
	{
		// Sanity checks ...
		g_assert(m_ObjectID);
		g_assert(m_Document);
		g_assert(m_ObjectInfo);
	}

	// k3dIObjectInitializer implementation ...
	k3dIObject::ID ObjectID()
	{
		return m_ObjectID;
	}

	k3dIDocument* Document()
	{
		return m_Document;
	}

	k3dIObjectInfo* ObjectInfo()
	{
		return m_ObjectInfo;
	}

private:
	const k3dIObject::ID m_ObjectID;
	k3dIDocument* const m_Document;
	k3dIObjectInfo* const m_ObjectInfo;
};

/// Serializes K-3D objects
class SaveObject
{
public:
	SaveObject(sdpxml::Element& Element, k3dIDependencies* const Dependencies) :
		m_Element(Element),
		m_Dependencies(Dependencies),
		m_Count(0)
	{
		// Sanity checks ...
		g_assert(Dependencies);
	}

	void operator()(k3dIObject* const Object)
	{
		// Skip the object if it isn't saveable ...
		if(SDP_S_OK != Object->State()->QueryState(k3dIState::SAVEABLE))
			return;

		// Save the object if it supports persistence ...
		k3dIPersist* const persist = dynamic_cast<k3dIPersist*>(Object);
		if(!persist)
			return;

		// Create an XML element for the object ...
		sdpxml::Element& objectelement = m_Element.Append(sdpxml::Element("object", "",
			sdpxml::Attribute("name", Object->Name()),
			sdpxml::Attribute("class", sdpToString(Object->ObjectInfo()->ClassID())),
			sdpxml::Attribute("id", sdpToString(Object->ObjectID()))));
		persist->Save(objectelement, m_Dependencies);

		m_Count++;
	}

private:
	sdpxml::Element& m_Element;
	k3dIDependencies* m_Dependencies;
	gulong m_Count;
};

/// Notifies K-3D objects that serialization is complete
class SaveComplete
{
public:
	void operator()(k3dIObject* const Object)
	{
		// Skip the object if it isn't saveable ...
		if(SDP_S_OK != Object->State()->QueryState(k3dIState::SAVEABLE))
			return;

		// Notify the object, if it supports persistence ...
		k3dIPersist* const persist = dynamic_cast<k3dIPersist*>(Object);
		if(persist)
			persist->SaveComplete();
	}
};

} // namespace

/////////////////////////////////////////////////////////////////////////
// CObjectCollection

CObjectCollection::CObjectCollection(k3dIDocument* const Document, CHierarchy* const Hierarchy) :
	m_Document(Document),
	m_Hierarchy(Hierarchy),
	m_CollectionChanged(false)
{
	// Sanity checks ...
	g_assert(m_Document);
	g_assert(m_Hierarchy);
}

CObjectCollection::~CObjectCollection()
{
	// Give objects a chance to shut down ...
	std::for_each(m_Objects.begin(), m_Objects.end(), std::mem_fun(&k3dIObject::ObjectDeleted));

	// Zap objects ...
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); ++object)
		{
			if(SDP_S_OK == k3dApplication()->Verbose())
				std::cout << "Destroying object [" << (*object)->Name() << "]" << std::endl;

			// Zap that object!
			sdpIDeletable* const deletable = dynamic_cast<sdpIDeletable*>(*object);
			delete deletable;
		}
}

k3dIObject* CObjectCollection::CreateObject(const sdp::ClassID& ClassID)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), 0);

	// Try to find the object info for this ClassID ...
	std::vector<k3dIObjectInfo*> info;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(info), k3d::SameClassID(ClassID));

	// If we don't support this object type, fail ...
	g_return_val_if_fail(1 == info.size(), 0);
	CObjectInfo* const objectinfo = dynamic_cast<CObjectInfo*>(*info.begin());
	g_return_val_if_fail(objectinfo, 0);

	// Create the object instance ...
	k3dIObject* const object = RawCreateObject(objectinfo, NextUniqueObjectID());
	g_return_val_if_fail(object, 0);

	// Initialize the object as a new object ...
	k3dIPersist* const persist = dynamic_cast<k3dIPersist*>(object);
	if(persist)
		g_return_val_if_fail(sdpSucceeded(persist->InitNew()), 0);

	// Let the object know that it's fully initialized ...
	g_return_val_if_fail(sdpSucceeded(object->InitComplete()), 0);

	// Notify clients that a New! Object! Has Arrived!
	Notify(Event_Changed, 0);

	return object;
}

k3dIObject* CObjectCollection::CreateObject(sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element.Name() == "object", 0);

	// Get the object's class ID ...
	const sdp::ClassID classid = sdpxml::GetAttribute(Element, "class", sdp::ClassIDNull);
	g_return_val_if_fail(classid != sdp::ClassIDNull, 0);

	// Get the object name, and make sure we don't end up with any duplicates ...
	const sdpString objectname = k3dGetUniqueName(this, sdpxml::GetAttribute<sdpString>(Element, "name", ""));
	g_return_val_if_fail(objectname.size(), 0);

	// Get the object ID ...
	const k3dIObject::ID objectid = sdpxml::GetAttribute(Element, "id", 0);
	g_return_val_if_fail(objectid, 0);

	// Make sure we don't end up with any duplicate IDs ...
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); ++object)
		{
			if((*object)->ObjectID() != objectid)
				continue;

			(*object)->SetObjectID(NextUniqueObjectID());
			break;
		}

	// Get the object's class info ...
	std::vector<k3dIObjectInfo*> info;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(info), k3d::SameClassID(classid));

	// If we don't support this object type, fail ...
	g_return_val_if_fail(1 == info.size(), 0);
	CObjectInfo* const objectinfo = dynamic_cast<CObjectInfo*>(*info.begin());
	g_return_val_if_fail(objectinfo, 0);

	// Create the object instance ...
	k3dIObject* const object = RawCreateObject(objectinfo, objectid);
	g_return_val_if_fail(object, 0);

	// Tell the object to load itself ...
	k3dIPersist* const persist = dynamic_cast<k3dIPersist*>(object);
	g_return_val_if_fail(persist, 0);	// If the object doesn't support k3dIPersist, how did it get into the document?
	g_return_val_if_fail(sdpSucceeded(persist->Load(Document, Element)), 0);

	// Notify clients that a New! Object! Has Arrived!
	Notify(Event_Changed, 0);

	return object;
}

k3dIObject* CObjectCollection::RawCreateObject(k3dIObjectInfo* const ObjectInfo, const k3dIObject::ID ObjectID)
{
	// Sanity checks ...
	g_assert(ObjectInfo);
	g_assert(ObjectID);

	// Create an object initializer ...
	ObjectInitializer initializer(ObjectID, m_Document, ObjectInfo);

	// Create an instance of the object ...
	sdpIUnknown* const unknown = dynamic_cast<CObjectInfo*>(ObjectInfo)->CreateInstance()(ObjectInfo->ClassID(), &initializer);
	g_return_val_if_fail(unknown, 0);

	// All K-3D objects have to support k3dIObject!
	k3dIObject* const object = dynamic_cast<k3dIObject*>(unknown);
	g_return_val_if_fail(object, 0);

	// Add this object to the hierarchy
	k3d::PushEditing(m_Document->Objects());
	m_Hierarchy->Add(object);
	k3d::PopEditing(m_Document->Objects());

	// If we're recording undos ...
	if(!m_CollectionChanged)
		{
			k3dIStateChangeSet* const changeset = m_Document->StateRecorder()->ChangeSet();
			if(changeset)
				{
					m_CollectionChanged = true;

					changeset->RecordOldState(new k3dNewInstanceStateContainer<sdpIDeletable>(dynamic_cast<sdpIDeletable*>(object), changeset));
					changeset->RecordOldState(new k3dValueStateContainer<Objects>(m_Objects));
				}
		}

	// Add the object to our collection ...
	m_Objects.push_back(object);

	return object;
}

bool CObjectCollection::DeleteObject(k3dIObject* const Object)
{
//	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << g_pretty_function() << " [" << Object->Name() << "]" << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(Object, false);
	g_return_val_if_fail(Editing(), false);

	// If this object has the mouse focus, take it away ...
	if(m_Document->MouseFocus() && (m_Document->MouseFocus() == dynamic_cast<k3dIMouseEventObserver*>(Object)))
		m_Document->SetMouseFocus(0);

	// Do the real work ...
	if(m_Document->StateRecorder()->ChangeSet())
		return RawDeleteObject(Object);

	return RawDestroyObject(Object);
}

bool CObjectCollection::RawDeleteObject(k3dIObject* const Object)
{
//	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << g_pretty_function() << " [" << Object->Name() << "]" << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(Object, false);

	// Make sure this is actually an object that we can delete ...
	sdpIDeletable* const deletable = dynamic_cast<sdpIDeletable*>(Object);
	g_return_val_if_fail(deletable, false);

	// Remove this object from the hierarchy ...
	m_Hierarchy->Remove(Object);

	// If we're recording undos ...
	if(!m_CollectionChanged)
		{
			k3dIStateChangeSet* const changeset = m_Document->StateRecorder()->ChangeSet();
			if(changeset)
				{
					m_CollectionChanged = true;

					changeset->RecordOldState(new k3dDeleteInstanceStateContainer<sdpIDeletable>(deletable, changeset));
					changeset->RecordOldState(new k3dValueStateContainer<Objects>(m_Objects));
				}
		}

	// Get rid of our reference ...
	m_Objects.erase(std::remove(m_Objects.begin(), m_Objects.end(), Object), m_Objects.end());

	// Let the object know that it's getting deleted ...
	Object->ObjectDeleted();

	// Notify clients ...
	Notify(Event_Changed, 0);

	return true;
}

bool CObjectCollection::RawDestroyObject(k3dIObject* const Object)
{
//	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << g_pretty_function() << " [" << Object->Name() << "]" << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(Object, false);

	// Make sure this is actually an object that we can delete ...
	sdpIDeletable* const deletable = dynamic_cast<sdpIDeletable*>(Object);
	g_return_val_if_fail(deletable, false);

	// Remove this object from the hierarchy ...
	m_Hierarchy->Remove(Object);

	// Get rid of our reference ...
	m_Objects.erase(std::remove(m_Objects.begin(), m_Objects.end(), Object), m_Objects.end());

	// Let the object know that it's getting deleted ...
	Object->ObjectDeleted();

	// Notify clients ...
	Notify(Event_Changed, 0);

	// Do the deed ...
// Shouldn't delete for the deletion to be 'undone'
// (have to do some kind of garbage collection with undo level)
//	delete deletable;

	return true;
}

void CObjectCollection::VisitObjects(k3dIObjectVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	std::for_each(m_Objects.begin(), m_Objects.end(), sdp::MakeVisitorAdapter(Visitor, &k3dIObjectVisitor::VisitObject));
}

K3DOBJECTHANDLE CObjectCollection::FirstObject()
{
	return K3DOBJECTHANDLE(m_Objects.size() ? 1 : 0);
}

K3DOBJECTHANDLE CObjectCollection::NextObject(K3DOBJECTHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= m_Objects.size(), 0);

	return K3DOBJECTHANDLE(gulong(Handle) < m_Objects.size() ? gulong(Handle)+1 : 0);
}

k3dIObject* CObjectCollection::Object(K3DOBJECTHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= m_Objects.size(), 0);

	return m_Objects[gulong(Handle)-1];
}

bool CObjectCollection::InitComplete()
{
	std::for_each(m_Objects.begin(), m_Objects.end(), std::mem_fun(&k3dIObject::InitComplete));
	return true;
}

bool CObjectCollection::Load(sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Load objects
	sdpxml::ElementPointer const objects = sdpxml::FindElement(Element, sdpxml::SameName("objects"));
	if(!objects)
		return true;

	// For each object in the document ...
	for(sdpxml::ElementCollection::iterator object = objects->Children().begin(); object != objects->Children().end(); object++)
		g_return_val_if_fail(CreateObject(Document, *object), false);

	return true;
}

bool CObjectCollection::LoadComplete()
{
	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); ++object)
		{
			k3dIPersist* const persist = dynamic_cast<k3dIPersist*>(*object);
			if(persist)
				persist->LoadComplete();
		}
	return true;
}

bool CObjectCollection::Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies)
{
	// Sanity checks ...
	g_return_val_if_fail(Dependencies, false);

	// Make it happen ...
	sdpxml::Element& objects = Element.Append(sdpxml::Element("objects"));

	// Save every object ...
	std::for_each(m_Objects.begin(), m_Objects.end(), ::SaveObject(objects, Dependencies));
	return true;
}

bool CObjectCollection::SaveComplete()
{
	std::for_each(m_Objects.begin(), m_Objects.end(), ::SaveComplete());
	return true;
}

void CObjectCollection::PushEditing()
{
	if(!Editing())
		m_CollectionChanged = false;

	k3dEditable::PushEditing();
	k3d::PushEditing(m_Hierarchy);
}

void CObjectCollection::PopEditing()
{
	k3d::PopEditing(m_Hierarchy);
	k3dEditable::PopEditing();

	if(!Editing() && m_CollectionChanged)
		{
			k3dIStateChangeSet* const changeset = m_Document->StateRecorder()->ChangeSet();
			if(changeset)
				{
					changeset->RecordAction(new k3dActionNotify(this, Event_Changed, 0, 0));
					changeset->RecordNewState(new k3dValueStateContainer<Objects>(m_Objects));
				}
		}

}

const k3dIObject::ID CObjectCollection::NextUniqueObjectID()
{
	k3dIObject::ID maxid = 0;

	for(ObjectIterator object = m_Objects.begin(); object != m_Objects.end(); ++object)
		if((*object)->ObjectID() > maxid)
			maxid = (*object)->ObjectID();

	return maxid + 1;
}


