// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CObjectInfo class, which encapsulates information about a class of K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include "objectinfo.h"

#include <k3dmisc.h>
#include <k3diapplication.h>

#include <strstream>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H

/////////////////////////////////////////////////////////////////////////////
// CObjectInfo

class InsertInterfaces
{
public:
	InsertInterfaces(CObjectInfo::InterfaceList& Interfaces) :
		m_Interfaces(Interfaces)
	{
	}
	
	void operator()(sdpxml::Element& Element)
	{
   		const sdpString interfacetype = sdpxml::GetAttribute<sdpString>(Element, "type", "");
		g_return_if_fail(interfacetype.size());
		m_Interfaces.push_back(interfacetype);
	}
	
private:
	CObjectInfo::InterfaceList& m_Interfaces;
};

class InsertBookmarks
{
public:
	InsertBookmarks(CObjectInfo::Bookmarks& Bookmarks) :
		m_Bookmarks(Bookmarks)
	{
	}
	
	void operator()(sdpxml::Element& Element)
	{
		const sdpString location = sdpxml::GetAttribute<sdpString>(Element, "location", "");
		g_return_if_fail(location.size());
		m_Bookmarks.push_back(location);
	}
	
private:
	CObjectInfo::Bookmarks& m_Bookmarks;
};

CObjectInfo::CObjectInfo(k3dIObjectModule* const Module, sdpxml::Element& Element, sdpComCreateInstanceProc CreateInstance) :
	m_Module(Module),
	m_ClassID(sdp::ClassIDNull),
	m_CreateInstance(CreateInstance),
	m_Incomplete(false),
	m_Experimental(false)
{
	// Sanity checks ...
	g_assert(m_Module);
	g_assert(m_CreateInstance);

	// Store object information ...
	sdpxml::ParseAttribute(Element, "clsid", m_ClassID);
	sdpxml::ParseAttribute(Element, "name", m_Name);

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "  Loading object " << m_Name << std::endl;

	// Look for a short description of the object ...
	sdpxml::ElementPointer const shortdescription = sdpxml::FindElement(Element, sdpxml::SameName("shortdescription"));
	if(shortdescription)
		m_ShortDescription = shortdescription->Text();

	// Look for interfaces supported by this object ...
	sdpxml::ForEach(Element, sdpxml::SameName("interface"), InsertInterfaces(m_Interfaces));

	// Look for bookmarks ...
	sdpxml::ForEach(Element, sdpxml::SameName("bookmark"), InsertBookmarks(m_Bookmarks));

	// See if this is an incomplete object ...
	sdpxml::ParseAttribute(Element, "incomplete", m_Incomplete);

	// See if this is an experimental object ...
	sdpxml::ParseAttribute(Element, "experimental", m_Experimental);
}

CObjectInfo::~CObjectInfo()
{
}

///////////////////////////////////////////////////////////////////////////////////
// k3dIObjectInfo implementation

k3dIObjectModule* CObjectInfo::Module()
{
	return m_Module;
}

const sdp::ClassID& CObjectInfo::ClassID()
{
	return m_ClassID;
}

SDPCSTRING CObjectInfo::Name()
{
	return m_Name;
}

SDPCSTRING CObjectInfo::ShortDescription()
{
	return m_ShortDescription;
}

sdpResult CObjectInfo::QueryObjectInterface(const sdp::ITypeInfo& InterfaceType)
{
	// If the list is empty (likely) we're done ...
	if(!m_Interfaces.size())
		return SDP_S_FALSE;

	// Iterate through available interfaces ...
	InterfaceIterator i = std::find(m_Interfaces.begin(), m_Interfaces.end(), sdpString(InterfaceType.RTTI().name()));
	if(i != m_Interfaces.end())
		return SDP_S_OK;

	// Couldn't find it ...
	return SDP_S_FALSE;
}

sdpResult CObjectInfo::QueryIncomplete()
{
	return m_Incomplete ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CObjectInfo::QueryExperimental()
{
	return m_Experimental ? SDP_S_OK : SDP_S_FALSE;
}

void CObjectInfo::VisitBookmarks(k3dIBookmarkVisitor* const BookmarkVisitor)
{
	// Sanity checks ...
	g_return_if_fail(BookmarkVisitor);
	
	for(BookmarkIterator bookmark = m_Bookmarks.begin(); bookmark != m_Bookmarks.end(); bookmark++)
		BookmarkVisitor->VisitBookmark(*bookmark);
}

//////////////////////////////////////////////////////////////////////////////////////
// CObjectModule

CObjectModule::CObjectModule(k3dIApplication* const Application, SDPCSTRING FilePath, SDPCSTRING FileName, GModule* const Module, sdpComInitializeModuleProc const InitializeModuleProc, sdpComModuleInfoProc const ModuleInfoProc, sdpComCreateInstanceProc const CreateInstanceProc) :
	m_Application(Application),
	m_Path(sdpDirectory(FilePath)),
	m_Name(FileName),
	m_Module(Module),
	m_InitializeModuleProc(InitializeModuleProc),
	m_ModuleInfoProc(ModuleInfoProc),
	m_CreateInstanceProc(CreateInstanceProc)
{
	// Sanity checks ...
	g_assert(m_Application);
	g_assert(m_InitializeModuleProc);
	g_assert(m_ModuleInfoProc);
	g_assert(m_CreateInstanceProc);

	// Note ... we don't check FilePath, FileName, or Module, since the module may be statically-linked

	// Initialize the module ...
	m_InitializeModuleProc(this);
}

CObjectModule::~CObjectModule()
{
	// Unload the module ...
	if(m_Module)
		g_module_close(m_Module);
}

class CObjectModuleLoadObject
{
public:
	CObjectModuleLoadObject(CObjectModule* const Module, CObjectModule::ObjectInfoCollection& Objects, const sdpComCreateInstanceProc CreateInstanceProc) :
		m_Module(Module),
		m_Objects(Objects),
		m_CreateInstanceProc(CreateInstanceProc)
	{
	}
	
	void operator()(sdpxml::Element& Element)
	{
		// Sanity checks ...
		g_assert(Element.Name() == "object");

		// Create an object info record ...
		CObjectInfo* const newobjectinfo = new CObjectInfo(m_Module, Element, m_CreateInstanceProc);		

		// Double-check that we don't have a duplicate CLSID ...
		for(CObjectModule::ObjectInfoIterator objectinfo = m_Objects.begin(); objectinfo != m_Objects.end(); objectinfo++)
			{
				if((*objectinfo)->ClassID() == newobjectinfo->ClassID())
					{
						std::cerr << "Object type [" << newobjectinfo->Name() << "] duplicates the Class ID of object type [" << (*objectinfo)->Name() << "] and will not be loaded." << std::endl;
						delete newobjectinfo;
						return;
					}
			}		
		
		m_Objects.push_back(newobjectinfo);
	}

private:
	CObjectModule* const m_Module;
	CObjectModule::ObjectInfoCollection& m_Objects;
	const sdpComCreateInstanceProc m_CreateInstanceProc;
};

void CObjectModule::LoadObjects(CObjectModule::ObjectInfoCollection& Objects)
{
	// Load the sdpComModuleInfo document and parse it using XML ...
	sdpxml::Document doc("sdpcommoduleinfo");
	std::istrstream moduleinfo(const_cast<char*>(m_ModuleInfoProc()));
	g_return_if_fail(doc.Load(moduleinfo, ""));

	// Make sure we've got an sdpComModuleInfo document ...
	g_return_if_fail(doc.Name() == "sdpcommoduleinfo");

	// Get information about the library ...
	const sdpString name = sdpxml::GetAttribute<sdpString>(doc, "name", "");
	const sdpString version = sdpxml::GetAttribute<sdpString>(doc, "version", "");
	
	g_return_if_fail(name.size());
	g_return_if_fail(version.size());

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "Loading module " << name << " Version " << version << std::endl;

	// Iterate through its' children and load their info ...
	CObjectModuleLoadObject functor(this, Objects, m_CreateInstanceProc);
	sdpxml::ForEach(doc, sdpxml::SameName("object"), functor);

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << std::endl;
}

////////////////////////////////////////////////
// k3dIObjectModuleInitializer implementation ...

/* Implemented below ...
k3dIApplication* CObjectModule::Application()
{
	return m_Application;
}
*/

////////////////////////////////////////////////
// k3dIObjectModule implementation ...

k3dIApplication* CObjectModule::Application()
{
	return m_Application;
}

SDPCSTRING CObjectModule::Name()
{
	return m_Name;
}

SDPCSTRING CObjectModule::Path()
{
	return m_Path;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// CObjectModules

CObjectModules::CObjectModules(k3dIApplication* const Application) :
	m_Application(Application)
{
	// Sanity checks ...
	g_assert(m_Application);
}

CObjectModules::~CObjectModules()
{
	std::for_each(m_Objects.begin(), m_Objects.end(), sdp::DeleteObject());
	std::for_each(m_Modules.begin(), m_Modules.end(), sdp::DeleteObject());
}

gulong CObjectModules::ObjectInfoCount()
{
	return m_Objects.size();
}

void CObjectModules::VisitObjectInfo(k3dIObjectInfoVisitor* const ObjectInfoVisitor)
{
	// Sanity checks ...
	g_return_if_fail(ObjectInfoVisitor);
	
	for(CObjectModule::ObjectInfoIterator objectinfo = m_Objects.begin(); objectinfo != m_Objects.end(); objectinfo++)
		ObjectInfoVisitor->VisitObjectInfo(*objectinfo);
}

#define LOADSTATICMODULE(ModuleName) \
	extern sdpComInitializeModuleProc ModuleName##Initialize; \
	extern sdpComModuleInfoProc ModuleName##Info; \
	extern sdpComCreateInstanceProc ModuleName##CreateInstance; \
	m_Modules.push_back(new CObjectModule(m_Application, 0, 0, 0, ModuleName##Initialize, ModuleName##Info, ModuleName##CreateInstance));

// Load libraries within the given directory ...
void CObjectModules::LoadModules(SDPCSTRING Directory, bool Recursive)
{
#ifdef K3D_HAVE_STATIC_PLUGINS

	// Handle (optional) statically-linked plugin modules
//	LOADSTATICMODULE(BitmapModule);
	LOADSTATICMODULE(CoreModule);
	LOADSTATICMODULE(GeometryModule);
	LOADSTATICMODULE(JavaScriptModule);
	LOADSTATICMODULE(SFXModule);
	LOADSTATICMODULE(ToolModule);
//	LOADSTATICMODULE(TransformerModule);

/** \todo Figure out how to handle the optional modules when statically-linking */
/*
#ifdef K3D_HAVE_EXPERIMENTAL
//	LOADSTATICMODULE(ExperimentalModule);
#endif // K3D_HAVE_EXPERIMENTAL

#ifdef K3D_HAVE_PLIB
//	LOADSTATICMODULE(PLIBModule);
#endif // K3D_HAVE_PLIB
*/

#endif // K3D_HAVE_STATIC_PLUGINS

	// Load dynamically-linked plugin modules ...
	sdpForEachFile(Directory, *this, Recursive);

	// Load all of the objects ...
	for(ObjectModuleIterator module = m_Modules.begin(); module != m_Modules.end(); module++)
		(*module)->LoadObjects(m_Objects);
}

// For each file within the original directory ...
bool CObjectModules::HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName)
{
	// Skip files that don't end in ".so" ...
	sdpString filename(FileName);
	if(filename.size() < 4)
		return true;

	if(filename.right(3) != ".so")
		return true;

	// Try to open the file as a module ...
	GModule* module = g_module_open(FilePath, GModuleFlags(0));
	if(!module)
		{
			std::cout << "Library " << FileName << ": " << g_module_error() << std::endl;
			return true;
		}

	// Look for the required sdpCom entry points ...
	sdpComInitializeModuleProc initializemodule = 0;
	g_module_symbol(module, "sdpComInitializeModule", reinterpret_cast<gpointer*>(&initializemodule));
	
	sdpComModuleInfoProc moduleinfo = 0;
	g_module_symbol(module, "sdpComModuleInfo", reinterpret_cast<gpointer*>(&moduleinfo));

	sdpComCreateInstanceProc classprocedure = 0;
	g_module_symbol(module, "sdpComCreateInstance", reinterpret_cast<gpointer*>(&classprocedure));

	// Print warnings ...
	if(!initializemodule)
		if(SDP_S_OK == k3dApplication()->Verbose())
			std::cout << "Module " << FileName << ": missing sdpComInitializeModule entry point" << std::endl;

	if(!moduleinfo)
		if(SDP_S_OK == k3dApplication()->Verbose())
			std::cout << "Module " << FileName << ": missing sdpComModuleInfo entry point" << std::endl;

	if(!classprocedure)
		if(SDP_S_OK == k3dApplication()->Verbose())
			std::cout << "Module " << FileName << ": missing sdpComCreateInstance entry point" << std::endl;

	// If we're missing any of the sdpCom entry points, we're done ...
	if(!(initializemodule && moduleinfo && classprocedure))
		return true;

	// Store the library ...
	m_Modules.push_back(new CObjectModule(m_Application, FilePath, FileName, module, initializemodule, moduleinfo, classprocedure));

	return true;
}

