#ifndef OBJECTINFO_H
#define OBJECTINFO_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CObjectInfo class, which encapsulates information about a class of K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3diobjectinfo.h>
#include <k3diobjectinfocollection.h>
#include <k3diobjectmodule.h>
#include <k3diobjectmoduleinitializer.h>

#include <gmodule.h>

// Forward class declarations
class k3dIApplication;
namespace sdpxml { class Element; }

//////////////////////////////////////////////////////////////////////////////////////
// CObjectInfo

class CObjectInfo :
	public k3dIObjectInfo
{
public:
	CObjectInfo(k3dIObjectModule* const Module, sdpxml::Element& Element, sdpComCreateInstanceProc CreateInstance);
	virtual ~CObjectInfo();

	// k3dIObjectInfo overrides ...
	k3dIObjectModule* Module();
	const sdp::ClassID& ClassID();
	SDPCSTRING Name();
	SDPCSTRING ShortDescription();
	sdpResult QueryObjectInterface(const sdp::ITypeInfo& Interface);
	sdpResult QueryIncomplete();
	sdpResult QueryExperimental();
	void VisitBookmarks(k3dIBookmarkVisitor* const BookmarkVisitor);

	/// Returns the sdpCom class-instantiation procedure for this class type
	sdpComCreateInstanceProc CreateInstance() { return m_CreateInstance; }

	/// A collection of sdpCOM interfaces supported by this object type
	typedef std::vector<sdpString> InterfaceList;
	typedef InterfaceList::iterator InterfaceIterator;

	/// A collection of bookmarks for this object
	typedef std::vector<sdpString> Bookmarks;
	typedef Bookmarks::iterator BookmarkIterator;

protected:
	/// The library that owns this object
	k3dIObjectModule* const m_Module;
	/// The object Class ID
	sdp::ClassID m_ClassID;
	/// The object name
	sdpString m_Name;
	/// A short description of the object
	sdpString m_ShortDescription;
	/// The function for creating objects of this type
	sdpComCreateInstanceProc const m_CreateInstance;
	
	/// Stores interfaces supported by this object	
	InterfaceList m_Interfaces;
	/// Stores bookmarks defined for this object
	Bookmarks m_Bookmarks;
	/// Stores whether this is an incomplete object
	bool m_Incomplete;
	/// Stores whether this is an experimental object
	bool m_Experimental;
};

//////////////////////////////////////////////////////////////////////////////////////
// CObjectModule

class CObjectModule :
	public k3dIObjectModuleInitializer,
	public k3dIObjectModule
{
public:
	CObjectModule(k3dIApplication* const Application, SDPCSTRING FilePath, SDPCSTRING FileName, GModule* const Module, sdpComInitializeModuleProc const InitializeModuleProc, sdpComModuleInfoProc const ModuleInfoProc, sdpComCreateInstanceProc const CreateInstanceProc);
	virtual ~CObjectModule();

	// k3dIObjectModuleInitializer implementation ...
	k3dIApplication* Application();

	// k3dIObjectModule implementation ...
	SDPCSTRING Name();
	SDPCSTRING Path();

	typedef std::vector<CObjectInfo*> ObjectInfoCollection;
	typedef ObjectInfoCollection::iterator ObjectInfoIterator;
	void LoadObjects(ObjectInfoCollection& ObjectInfoSet);

protected:
	void LoadObject(sdpxml::Element& Element);

	// Data ...
	k3dIApplication* const m_Application;
	const sdpString m_Path;
	const sdpString m_Name;
	GModule* const m_Module;
	sdpComInitializeModuleProc const m_InitializeModuleProc;
	sdpComModuleInfoProc const m_ModuleInfoProc;
	sdpComCreateInstanceProc const m_CreateInstanceProc;
};

//////////////////////////////////////////////////////////////////////////////////////
// CObjectModules

class CObjectModules :
	public k3dIObjectInfoCollection
{
public:
	CObjectModules(k3dIApplication* const Application);
	virtual ~CObjectModules();

	// k3dIObjectInfoCollection implementation
	gulong ObjectInfoCount();
	void VisitObjectInfo(k3dIObjectInfoVisitor* const ObjectInfoVisitor);

	// Initialization ...
	void LoadModules(SDPCSTRING Directory, bool Recursive);

	// Functor-type operator for use with sdpForEachFile() ...
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName);

protected:
	k3dIApplication* const m_Application;

	typedef std::vector<CObjectModule*> ObjectModules;
	typedef ObjectModules::iterator ObjectModuleIterator;
	ObjectModules m_Modules;

	CObjectModule::ObjectInfoCollection m_Objects;
};

#endif // OBJECTINFO_H
