// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the COptionsDialog, which provides a UI for changing global user options
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "optionsdialog.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <sdpgtkevents.h>

#include <fstream>

///////////////////////////////////////////////////////////////////////
// COptionsDialog

const sdpString COptionsDialog::dialogname = "options";
const sdpString COptionsDialog::control_showtutorialmenu = "showtutorials";
const sdpString COptionsDialog::control_restorewindows = "restorewindows";
const sdpString COptionsDialog::control_templatedocument = "templatedocument";
const sdpString COptionsDialog::control_choosetemplatedocument = "choosetemplatedocument";
const sdpString COptionsDialog::control_preferredlanguage = "preferredlanguage";
const sdpString COptionsDialog::control_defaultengine = "defaultengine";
const sdpString COptionsDialog::control_htmlviewer = "htmlviewer";
const sdpString COptionsDialog::control_bitmapviewer = "bitmapviewer";
const sdpString COptionsDialog::control_mpegviewer = "mpegviewer";
const sdpString COptionsDialog::control_slpreprocessor = "slpreprocessor";
const sdpString COptionsDialog::control_slparser = "slparser";
const sdpString COptionsDialog::control_renderfarmpath = "renderfarmpath";

COptionsDialog::COptionsDialog(CApplication* const Application) :
	COptionsDialogBase(Application, dialogname, new k3dOptionsWindowGeometryStore(Application, 0)),
	m_Application(Application)
{
	// Sanity checks ...
	g_assert(m_Application);
	
	g_return_if_fail(LoadGTKMLTemplate("options.gtkml"));
	
	// We want to know if the default engine combo-box changes ...
	MapEvent("changed", control_defaultengine, false, Combo(control_defaultengine).Entry(), true);
	
	// Show ourselves as a modal dialog ...
	CallUpdateControls();
	Show();
	DoModal();
	
	// Delete ourselves when we're done ...
	delete this;
	
	// Ensure that any changes get saved ...
	m_Application->SaveOptions();
}

void COptionsDialog::UpdateControls()
{
	// UI-related options ...
	ToggleButton(control_showtutorialmenu).SetState(k3dOptionsGetTutorialMenu(k3dApplication()->Options()));
	ToggleButton(control_restorewindows).SetState(k3dOptionsGetRestoreWindows(k3dApplication()->Options()));
	Entry(control_templatedocument).SetText(k3dOptionsGetTemplateDocument(k3dApplication()->Options()));
	Entry(control_preferredlanguage).SetText(k3dOptionsGetPreferredLanguage(k3dApplication()->Options()));

	// Default render engine ...
	sdpString defaultenginetype;
	sdpString defaultengine;
	sdpString defaultenginename;
	k3dOptionsGetDefaultRenderEngine(k3dApplication()->Options(), defaultenginetype, defaultengine, defaultenginename);

	// Build a list of available engines ...
	sdpStringArray enginenames;
	k3dOptionsRenderEngines engines;
	k3dOptionsGetRenderEngines(k3dApplication()->Options(), engines);
	for(k3dOptionsRenderEngines::iterator engine = engines.begin(); engine != engines.end(); engine++)
		{
			enginenames.push_back(engine->Name());
/*
			// Look for the custom render engine ...
			if((engine->Type() == "ri") && (engine->Engine() == "custom"))
				{
					Entry(control_customenginevisible).SetText(engine->VisibleRender());
					Entry(control_customenginehidden).SetText(engine->Render());
				}
*/
		}

	Combo(control_defaultengine).SetList(enginenames);
	Combo(control_defaultengine).SetEntryText(defaultenginename);
			
	// Helper applications ...
	Entry(control_htmlviewer).SetText(k3dOptionsGetHTMLViewer(k3dApplication()->Options()));
	Entry(control_bitmapviewer).SetText(k3dOptionsGetBitmapViewer(k3dApplication()->Options()));
	Entry(control_mpegviewer).SetText(k3dOptionsGetMPEGViewer(k3dApplication()->Options()));
	Entry(control_slpreprocessor).SetText(k3dOptionsGetSLPreprocessor(k3dApplication()->Options()));
	Entry(control_slparser).SetText(k3dOptionsGetSLParser(k3dApplication()->Options()));

	// Paths ...
	Entry(control_renderfarmpath).SetText(k3dOptionsGetRenderFarmPath(k3dApplication()->Options()));
	
	COptionsDialogBase::UpdateControls();
}

void COptionsDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_showtutorialmenu)
		OnShowTutorialMenu();
	else if(Event->Name() == control_restorewindows)
		OnRestoreWindows();
	else if(Event->Name() == control_templatedocument)
		OnTemplateDocument();
	else if(Event->Name() == control_choosetemplatedocument)
		OnChooseTemplateDocument();
	else if(Event->Name() == control_preferredlanguage)
		OnPreferredLanguage();
	else if(Event->Name() == control_htmlviewer)
		OnHTMLViewer();
	else if(Event->Name() == control_bitmapviewer)
		OnBitmapViewer();
	else if(Event->Name() == control_mpegviewer)
		OnMPEGViewer();
	else if(Event->Name() == control_slpreprocessor)
		OnSLPreprocessor();
	else if(Event->Name() == control_slparser)
		OnSLParser();
	else if(Event->Name() == control_defaultengine)
		OnDefaultEngine();
	else if(Event->Name() == control_renderfarmpath)
		OnRenderFarmPath();
	else
		COptionsDialogBase::OnEvent(Event);
}

void COptionsDialog::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't allow the GTK+ tree to go away just yet ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	// Force updates of all our text controls (since it seems that "focus-out" doesn't get generated when the dialog is deleted)
	OnTemplateDocument();
	OnPreferredLanguage();
	OnHTMLViewer();
	OnBitmapViewer();
	OnMPEGViewer();
	OnSLPreprocessor();
	OnSLParser();
//	OnCustomEngineVisible();
//	OnCustomEngineHidden();
	OnRenderFarmPath();
	
	// Cancel modal operation ....
	CancelModal();
}

void COptionsDialog::OnShowTutorialMenu()
{
	k3dOptionsSetTutorialMenu(k3dApplication()->Options(), ToggleButton(control_showtutorialmenu).GetState());
}

void COptionsDialog::OnRestoreWindows()
{
	k3dOptionsSetRestoreWindows(k3dApplication()->Options(), ToggleButton(control_restorewindows).GetState());
}

void COptionsDialog::OnTemplateDocument()
{
	k3dOptionsSetTemplateDocument(k3dApplication()->Options(), Entry(control_templatedocument).GetText());
}

void COptionsDialog::OnChooseTemplateDocument()
{
	sdpString filepath;
	if(!k3dGetFilePath("document", "Choose K-3D Template Document:", false, k3dOptionsGetTemplateDocument(k3dApplication()->Options()), filepath))
		return;
		
	Entry(control_templatedocument).SetText(filepath);
	k3dOptionsSetTemplateDocument(k3dApplication()->Options(), filepath);
}

void COptionsDialog::OnPreferredLanguage()
{
	k3dOptionsSetPreferredLanguage(k3dApplication()->Options(), Entry(control_preferredlanguage).GetText());
}

void COptionsDialog::OnHTMLViewer()
{
	k3dOptionsSetHTMLViewer(k3dApplication()->Options(), Entry(control_htmlviewer).GetText());
}

void COptionsDialog::OnBitmapViewer()
{
	k3dOptionsSetBitmapViewer(k3dApplication()->Options(), Entry(control_bitmapviewer).GetText());
}

void COptionsDialog::OnMPEGViewer()
{
	k3dOptionsSetMPEGViewer(k3dApplication()->Options(), Entry(control_mpegviewer).GetText());
}

void COptionsDialog::OnSLPreprocessor()
{
	k3dOptionsSetSLPreprocessor(k3dApplication()->Options(), Entry(control_slpreprocessor).GetText());
}

void COptionsDialog::OnSLParser()
{
	k3dOptionsSetSLParser(k3dApplication()->Options(), Entry(control_slparser).GetText());
}

void COptionsDialog::OnDefaultEngine()
{
	// Get the selected engine name ...
	const sdpString defaultenginename = Combo(control_defaultengine).GetEntryText();
	g_return_if_fail(defaultenginename.size());

	// Get a list of available engines ...
	k3dOptionsRenderEngines engines;
	k3dOptionsGetRenderEngines(k3dApplication()->Options(), engines);
	for(k3dOptionsRenderEngines::iterator engine = engines.begin(); engine != engines.end(); engine++)
		{
			// Look for the default engine by name ...
			if(defaultenginename == engine->Name())
				{
					k3dOptionsSetDefaultRenderEngine(k3dApplication()->Options(), engine->Type(), engine->Engine());
					return;
				}
		}
}

void COptionsDialog::OnRenderFarmPath()
{
	k3dOptionsSetRenderFarmPath(k3dApplication()->Options(), Entry(control_renderfarmpath).GetText());
}


