// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRenameObjectDialog class, which provides a UI for renaming a K-3D object
		\author Tim Shead (tshead@k-3d.com)
*/

#include "renameobjectdialog.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3dieventsource.h>

#include <sdpgtkevents.h>

///////////////////////////////////////////////////////////////////////
// CRenameObjectDialog

const sdpString CRenameObjectDialog::dialogname = "rename";
const sdpString CRenameObjectDialog::control_oldname = "oldname";
const sdpString CRenameObjectDialog::control_newname = "newname";
const sdpString CRenameObjectDialog::control_ok = "ok";
const sdpString CRenameObjectDialog::control_cancel = "cancel";

CRenameObjectDialog::CRenameObjectDialog(k3dIObject* const Object) :
	CRenameObjectDialogBase(Object, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Sanity checks ...
	g_assert(m_Object);

	g_return_if_fail(LoadGTKMLTemplate("renameobject.gtkml"));

	// We want to know if our object changes name or goes away ...
	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(m_Object);
	g_return_if_fail(eventserver);
	eventserver->AddEventObserver(this, k3dIObject::Event_Renamed, "CRenameObjectDialog");
	eventserver->AddEventObserver(this, k3dIObject::Event_Deleted, "CRenameObjectDialog");

	AddCommand(control_newname, control_newname, CONTROL_EDITABLE);
	AddCommand(control_ok, control_ok, CONTROL_BUTTON);
	AddCommand(control_cancel, control_cancel, CONTROL_BUTTON);

	CallUpdateControls();
	Show();
}

CRenameObjectDialog::~CRenameObjectDialog()
{
	// Disable further event notifications
	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(m_Object);
	g_return_if_fail(eventserver);
	eventserver->RemoveEventObserver(this);
}

sdpResult CRenameObjectDialog::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIObject::Event_Renamed)
		{
			CallUpdateControls();
			return SDP_S_OK;
		}
	else if(event == k3dIObject::Event_Deleted)
		{
			delete this;
			return SDP_S_OK;
		}

	return CRenameObjectDialogBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void CRenameObjectDialog::UpdateControls()
{
	Entry(control_oldname).SetText(m_Object->Name());
	Widget(control_oldname).SetSensitive(false);
	
	Entry(control_newname).SetText(m_Object->Name());
	
	CRenameObjectDialogBase::UpdateControls();
}

void CRenameObjectDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_newname)
		OnNewName();
	else if(Event->Name() == control_ok)
		OnOK();
	else if(Event->Name() == control_cancel)
		OnCancel();
	else
		CRenameObjectDialogBase::OnEvent(Event);
}

void CRenameObjectDialog::OnNewName()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_newname, Entry(control_newname).GetText());
}

void CRenameObjectDialog::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't let it happen ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	// Turn it into a "cancel", instead ...
	OnCancel();
}

void CRenameObjectDialog::OnOK()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_ok, 0);
	k3dRecordStateChangeSet changeset(m_Object->Document(), "Rename Object");
	
	m_Object->SetName(Entry(control_newname).GetText());
	delete this;
}

void CRenameObjectDialog::OnCancel()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_cancel, 0);
	
	delete this;
}


