// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRenderFarm class, which encapsulates a virtual render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include "renderfarm.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3diapplication.h>

#include <errno.h>

//////////////////////////////////////////////////////////////////////////////////
// CRenderFarm

CRenderFarm::CRenderFarm()
{
}

CRenderFarm::~CRenderFarm()
{
	std::for_each(m_RenderJobs.begin(), m_RenderJobs.end(), sdp::DeleteObject());
}

// k3dIRenderFarm overrides ...
k3dIRenderJob* CRenderFarm::CreateJob(SDPCSTRING JobName, guchar Priority)
{
	// Sanity checks ...
	g_return_val_if_string_fail(JobName, 0);

	// Ensure that the job gets a unique name ...
	gulong index = 0;
	sdpString jobname(JobName);
	while(sdpFileExists(sdpAppendPath(Path(), jobname)))
		jobname = sdpString(JobName) + "-" + sdpToString(index++);
		
	// Create the job ...
	CRenderJob* renderjob = new CRenderJob(this, jobname, Priority);
	g_return_val_if_fail(renderjob, 0);
	m_RenderJobs.push_back(renderjob);
	
	return renderjob;
}

sdpResult CRenderFarm::StartJob(k3dIRenderJob* Job)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);

	// Make sure it really is one of our jobs ...
	CRenderJob* job = static_cast<CRenderJob*>(Job);
	g_return_val_if_fail(std::find(m_RenderJobs.begin(), m_RenderJobs.end(), job) != m_RenderJobs.end(), SDP_E_FAIL);

	// Update control files for the job and all its frames ...
	g_return_val_if_fail(job->WriteControlFiles(), false);
	
	// Mark the job ready-to-run ...
	g_return_val_if_fail(job->MarkReady(), false);

	// Start the local rendering process ...
	const sdpString commandline = "renderjob " + sdpString(k3dApplication()->OptionsPath()) + " " + sdpString(job->Path()) + " &";

	errno = 0;
	int result = system(commandline);
	if(-1 == result || (127 == result && errno))
		{
			std::cerr << "CRenderFarm::StartJob(): Error starting render job " << commandline << std::endl;
			return SDP_E_FAIL;
		}
			
	return SDP_S_OK;	
}

SDPCSTRING CRenderFarm::Path()
{
	return k3dOptionsGetRenderFarmPath(k3dApplication()->Options());
}

