// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRenderFrame class, which encapsulates a frame to be rendered as part of a job
		\author Tim Shead (tshead@k-3d.com)
*/

#include "renderframe.h"
#include "renderjob.h"

#include <k3dmisc.h>

#include <fstream>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H

CRenderFrame::CRenderFrame(CRenderJob* Job, SDPCSTRING Name) :
	m_Job(Job),
	m_Name(Name)
{
	// Sanity checks ...
	g_assert(Job);
	g_assert_string(Name);
	
	// Store our new path ...
	m_Path = sdpAppendPath(Job->Path(), m_Name);
	
	// Create the directory ...
	if(!sdpCreateDirectory(m_Path))
		std::cerr << "CRenderFrame::CRenderFrame(): Error creating frame directory \"" << m_Path << "\"" << std::endl;
}

//////////////////////////////////////////////////////////////////////////////////////
// k3dIRenderFrame overrides ...
			
k3dIRenderJob* CRenderFrame::Job()
{
	return m_Job;
}

sdpResult CRenderFrame::AddInputFile(SDPCSTRING Name, SDPSTRING Buffer, gulong* BufferLength)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	g_return_val_if_fail(Buffer, SDP_E_FAIL);
	g_return_val_if_fail(BufferLength, SDP_E_FAIL);

	// Make sure the filepath is unique ...
	gulong index = 0;
	sdpString name(Name);
	while(sdpFileExists(sdpAppendPath(Path(), name)))
		name = sdpString(Name) + "-" + sdpToString(index++);
		
	m_InputFiles.push_back(name);	
		
	return k3dReturnString(sdpAppendPath(Path(), name), Buffer, BufferLength);
}

sdpResult CRenderFrame::AddOutputFile(SDPCSTRING Name, SDPSTRING Buffer, gulong* BufferLength)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	g_return_val_if_fail(Buffer, SDP_E_FAIL);
	g_return_val_if_fail(BufferLength, SDP_E_FAIL);

	// Make sure the filepath is unique ...
	gulong index = 0;
	sdpString name(Name);
	while(sdpFileExists(sdpAppendPath(Path(), name)))
		name = sdpString(Name) + "-" + sdpToString(index++);
		
	m_OutputFiles.push_back(name);	
		
	return k3dReturnString(sdpAppendPath(Path(), name), Buffer, BufferLength);
}

sdpResult CRenderFrame::AddRenderOperation(SDPCSTRING EngineType, SDPCSTRING Engine, SDPCSTRING SourceFilePath, bool VisibleRender)
{
	// Sanity checks ...
	g_return_val_if_string_fail(EngineType, SDP_E_FAIL);
	g_return_val_if_string_fail(Engine, SDP_E_FAIL);
	g_return_val_if_string_fail(SourceFilePath, SDP_E_FAIL);
	
	m_RenderOperations.push_back(RenderOperation(EngineType, Engine, SourceFilePath, VisibleRender));
	
	return SDP_S_OK;
}

sdpResult CRenderFrame::AddCopyOperation(SDPCSTRING SourceFilePath, SDPCSTRING TargetFilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(SourceFilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(TargetFilePath, SDP_E_FAIL);

	m_CopyOperations.push_back(CopyOperation(SourceFilePath, TargetFilePath));	
	
	return SDP_S_OK;
}

sdpResult CRenderFrame::AddViewOperation(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	
	m_ViewOperations.push_back(FilePath);

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// Helper methods ...

bool CRenderFrame::WriteControlFile()
{
	// Create the empty control file ...
	std::ofstream file(sdpAppendPath(Path(), ControlFile));
	g_return_val_if_fail(file.good(), false);
	
	// Create the XML document ...
	sdpxml::Document document("k3dml");
	sdpxml::Element& frame = document.Append(sdpxml::Element("frame"));

	for(InputFileIterator inputfile = m_InputFiles.begin(); inputfile != m_InputFiles.end(); inputfile++)
		frame.Append(sdpxml::Element("inputfile", "", sdpxml::Attribute("path", *inputfile)));

	for(OutputFileIterator outputfile = m_OutputFiles.begin(); outputfile != m_OutputFiles.end(); outputfile++)
		frame.Append(sdpxml::Element("outputfile", "", sdpxml::Attribute("path", *outputfile)));
	
	for(RenderOperationIterator renderop = m_RenderOperations.begin(); renderop != m_RenderOperations.end(); renderop++)
		frame.Append(sdpxml::Element("renderoperation", "",
			sdpxml::Attribute("type", renderop->Type()),
			sdpxml::Attribute("engine", renderop->Engine()),
			sdpxml::Attribute("sourcepath", renderop->SourcePath()),
			sdpxml::Attribute("visiblerender", sdpToString(renderop->VisibleRender()))));
	
	for(CopyOperationIterator copyop = m_CopyOperations.begin(); copyop != m_CopyOperations.end(); copyop++)
		frame.Append(sdpxml::Element("copyoperation", "",
			sdpxml::Attribute("from", copyop->first),
			sdpxml::Attribute("to", copyop->second)));
				
	for(ViewOperationIterator viewop = m_ViewOperations.begin(); viewop != m_ViewOperations.end(); viewop++)
		frame.Append(sdpxml::Element("viewoperation", "", sdpxml::Attribute("path", *viewop)));
	
	// Write the document ...
	file << document << std::endl;
	g_return_val_if_fail(file.good(), false);
	
	return true;
}

bool CRenderFrame::MarkReady()
{
	// Create the empty "ready" status file ...
	std::ofstream file(sdpAppendPath(Path(), StatusReadyFile));
	g_return_val_if_fail(file.good(), false);
	
	return true;
}



