#ifndef RENDERFRAME_H
#define RENDERFRAME_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CRenderFrame class, which encapsulates a frame to be rendered as part of a job
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3direnderframe.h>

// Forward class declarations
class CRenderJob;

//////////////////////////////////////////////////////////////////////////////////
// CRenderFrame

class CRenderFrame : public k3dIRenderFrame
{
public:
	CRenderFrame(CRenderJob* Job, SDPCSTRING FrameName);
	virtual ~CRenderFrame() {}
	
	// k3dIRenderFrame overrides ...
	k3dIRenderJob* Job();
	sdpResult AddInputFile(SDPCSTRING Name, SDPSTRING Buffer, gulong* BufferLength);
	sdpResult AddOutputFile(SDPCSTRING Name, SDPSTRING Buffer, gulong* BufferLength);
	sdpResult AddRenderOperation(SDPCSTRING EngineType, SDPCSTRING Engine, SDPCSTRING SourceFilePath, bool VisibleRender);
	sdpResult AddCopyOperation(SDPCSTRING SourceFilePath, SDPCSTRING TargetFilePath);
	sdpResult AddViewOperation(SDPCSTRING FilePath);

	SDPCSTRING Name() { return m_Name; }
	SDPCSTRING Path() { return m_Path; }
			
	/// Called to write the control file for this frame (an XML document which describes how the frame should be rendered)
	bool WriteControlFile();
	/// Called to write the ready status file for this frame (an empty file that indicates that the frame is ready for rendering)
	bool MarkReady();

protected:
	CRenderJob* m_Job;
	sdpString m_Name;
	sdpString m_Path;
	
	/// Stores an array of input files used during rendering
	typedef sdpStringArray InputFiles;
	typedef InputFiles::iterator InputFileIterator;
	InputFiles m_InputFiles;
	/// Stores an array of output files generated by rendering
	typedef sdpStringArray OutputFiles;
	typedef OutputFiles::iterator OutputFileIterator;
	OutputFiles m_OutputFiles;
	
	/// Defines a render operation to be performed by the render farm
	class RenderOperation
	{
	public:
		RenderOperation(SDPCSTRING Type, SDPCSTRING Engine, SDPCSTRING SourcePath, bool VisibleRender) :
			m_Type(Type),
			m_Engine(Engine),
			m_SourcePath(SourcePath),
			m_VisibleRender(VisibleRender)
		{
			// Sanity checks ...
			g_assert_string(Type);
			g_assert_string(Engine);
			g_assert_string(SourcePath);
		}
	
		SDPCSTRING Type() const { return m_Type; }
		SDPCSTRING Engine() const { return m_Engine; }
		SDPCSTRING SourcePath() const { return m_SourcePath; }
		bool VisibleRender() const { return m_VisibleRender; }
	
	protected:
		sdpString m_Type;
		sdpString m_Engine;
		sdpString m_SourcePath;
		bool m_VisibleRender;
	};
	typedef std::vector<RenderOperation> RenderOperations;
	typedef RenderOperations::iterator RenderOperationIterator;
	/// Stores an array of render operations to perform
	RenderOperations m_RenderOperations;
		
	/// Defines a pair of filepaths for a post-render copy operation
	typedef std::pair<sdpString, sdpString> CopyOperation;
	typedef std::vector<CopyOperation> CopyOperations;
	typedef CopyOperations::iterator CopyOperationIterator;
	/// Stores an array of post-render copy operations to perform
	CopyOperations m_CopyOperations;
	
	/// Stores an array of image files to view once rendering is complete
	typedef sdpStringArray ViewOperations;
	typedef ViewOperations::iterator ViewOperationIterator;
	ViewOperations m_ViewOperations;
};

//////////////////////////////////////////////////////////////////////////////////
// FrameArray

typedef std::list<CRenderFrame*> RenderFrameList;
typedef RenderFrameList::iterator RenderFrameIterator;

#endif // RENDERFRAME_H
