// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRenderJob class, which encapsulates a job to be rendered by the render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include "renderjob.h"
#include "renderfarm.h"

#include <k3diapplication.h>
#include <k3direnderfarm.h>
#include <fstream>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // SDPXML_H

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// CRenderJob

CRenderJob::CRenderJob(CRenderFarm* RenderFarm, SDPCSTRING Name, guchar Priority) :
	m_RenderFarm(RenderFarm),
	m_Name(Name),
	m_Priority(Priority)
{
	// Sanity checks ...
	g_assert(RenderFarm);
	g_assert_string(Name);
	
	// Store our new path ...
	m_Path = sdpAppendPath(m_RenderFarm->Path(), m_Name);
	
	// Create the directory ...
	if(!sdpCreateDirectory(Path()))
		std::cerr << "CRenderJob::CRenderJob(): Error creating job directory \"" << Path() << "\"" << std::endl;
}

CRenderJob::~CRenderJob()
{
	std::for_each(m_Frames.begin(), m_Frames.end(), sdp::DeleteObject());
}

////////////////////////////////////////////////////////////////////////////////////
// k3dIRenderJob overrides ...
			
k3dIRenderFarm* CRenderJob::Farm()
{
	return m_RenderFarm;
}

k3dIRenderFrame* CRenderJob::CreateFrame(SDPCSTRING FrameName)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FrameName, 0);
	
	// Ensure that we end up with a unique name ...
	gulong index = 0;
	sdpString framename(FrameName);
	while(sdpFileExists(sdpAppendPath(Path(), framename)))
		framename = sdpString(FrameName) + "-" + sdpToString(index++);
	
	// Create the frame ...
	CRenderFrame* renderframe = new CRenderFrame(this, framename);
	g_assert(renderframe);
	m_Frames.push_back(renderframe);
	
	return renderframe;
}

//////////////////////////////////////////////////////////////////////////////
// Helper methods ...

bool CRenderJob::WriteControlFiles()
{
	// Create a control file for each frame ...
	std::for_each(m_Frames.begin(), m_Frames.end(), std::mem_fun(&CRenderFrame::WriteControlFile));

	// Create the empty control file ...
	std::ofstream file(sdpAppendPath(Path(), ControlFile));
	g_return_val_if_fail(file.good(), false);
	
	// Create the XML document ...
	sdpxml::Document document("k3dml");
	sdpxml::Element& job = document.Append(sdpxml::Element("job"));
	sdpxml::SetAttribute(job, sdpxml::Attribute("priority", sdpToString(m_Priority)));
	
	// Write the document ...
	file << document << std::endl;
	g_return_val_if_fail(file.good(), false);
	
	return true;
}

bool CRenderJob::MarkReady()
{
	// Mark each frame as "ready" ...
	std::for_each(m_Frames.begin(), m_Frames.end(), std::mem_fun(&CRenderFrame::MarkReady));

	// Create the empty "ready" status file ...
	std::ofstream file(sdpAppendPath(Path(), StatusReadyFile));
	g_return_val_if_fail(file.good(), false);
	
	return true;
}


