// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements classes for encapsulating information about a Pixar RenderMan Shading Language shader
		\author Tim Shead (tshead@k-3d.com)
*/

#include "shaders.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3diapplication.h>

#include <fstream>
#include <strstream>

/*
#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H
*/

////////////////////////////////////////////////////////////////
// CShaderArgument

CShaderArgument::CShaderArgument(sdpsl::IShaderArgument* const ShaderArgument) :
	m_Name(ShaderArgument->Name()),
	m_StorageClass(ShaderArgument->StorageClass()),
	m_Type(ShaderArgument->Type()),
	m_ExtendedType(ShaderArgument->ExtendedType()),
	m_ArrayCount(ShaderArgument->ArrayCount()),
	m_Space(ShaderArgument->Space()),
	m_Output(ShaderArgument->Output()),
	m_Value(ShaderArgument->Value())
{
}

/////////////////////////////////////////////////////////////////////////////
// sdpsl::IShaderArgument implementation

SDPCSTRING CShaderArgument::Name()
{
	return m_Name;
}

sdpsl::IShaderArgument::SLStorageClass CShaderArgument::StorageClass()
{
	return m_StorageClass;
}

sdpsl::IShaderArgument::SLType CShaderArgument::Type()
{
	return m_Type;
}

sdpsl::IShaderArgument::SLExtendedType CShaderArgument::ExtendedType()
{
	return m_ExtendedType;
}

gulong CShaderArgument::ArrayCount()
{
	return m_ArrayCount;
}

SDPCSTRING CShaderArgument::Space()
{
	return m_Space;
}

bool CShaderArgument::Output()
{
	return m_Output;
}

SDPCSTRING CShaderArgument::Value()
{
	return m_Value;
}

////////////////////////////////////////////////////////////////
// CShader

CShader::CShader(sdpsl::IShader* const Shader) :
	m_SourcePath(Shader->SourcePath()),
	m_Name(Shader->Name()),
	m_Authors(Shader->Authors()),
	m_Copyright(Shader->Copyright()),
	m_Description(Shader->Description()),
	m_Type(Shader->Type()),
	m_ExtendedType(Shader->ExtendedType())
{
	Shader->VisitArguments(this);
}

CShader::~CShader()
{
	std::for_each(m_Arguments.begin(), m_Arguments.end(), sdp::DeleteObject());
}

/////////////////////////////////////////////////////////////////////////////////////
// sdpsl::IShader implementation ...

SDPCSTRING CShader::SourcePath()
{
	return m_SourcePath;
}

SDPCSTRING CShader::Name()
{
	return m_Name;
}

SDPCSTRING CShader::Authors()
{
	return m_Authors;
}

SDPCSTRING CShader::Copyright()
{
	return m_Copyright;
}

SDPCSTRING CShader::Description()
{
	return m_Description;
}

sdpsl::IShader::SLType CShader::Type()
{
	return m_Type;
}

sdpsl::IShader::SLExtendedType CShader::ExtendedType()
{
	return m_ExtendedType;
}

gulong CShader::ArgumentCount()
{
	return m_Arguments.size();
}

void CShader::VisitArguments(sdpsl::IShaderArgumentVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	
	for(ArgumentIterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		Visitor->VisitShaderArgument(*argument);
}

// sdpsl::IShaderArgumentVisitor implementation
void CShader::VisitShaderArgument(sdpsl::IShaderArgument* const Argument)
{
	// Sanity checks ...
	g_return_if_fail(Argument);
	
	// Create a new CShaderArgument object ...
	m_Arguments.push_back(new CShaderArgument(Argument));
}

///////////////////////////////////////////////////////////////////////////////////
// CShaderCollection

CShaderCollection::CShaderCollection(k3dIApplication* const Application) :
	m_Application(Application)
{
	// Sanity checks ...
	g_assert(m_Application);
}

CShaderCollection::~CShaderCollection()
{
	std::for_each(m_Shaders.begin(), m_Shaders.end(), sdp::DeleteObject());
}

void CShaderCollection::VisitShaders(sdpsl::IShaderVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	for(ShaderIterator shader = m_Shaders.begin(); shader != m_Shaders.end(); shader++)
		Visitor->VisitShader(*shader);
}

void CShaderCollection::VisitShader(sdpsl::IShader* const Shader)
{
	// Create a shader info object ...
	m_Shaders.push_back(new CShader(Shader));
}

// Load shaders from the user's SHADERS environment variable ...
bool CShaderCollection::LoadShaders()
{
	// Get the user's SHADERS path ...
	sdpString shaderspath = k3dApplication()->ShadersPath();

	// Parse all the paths within SHADERS ...
	sdpString shaderdirectory;
	
	sdpString::size_type index = shaderspath.find(G_SEARCHPATH_SEPARATOR);
	while(index != sdpString::npos)
		{
			shaderdirectory = shaderspath.left(index);
			shaderspath = shaderspath.right(shaderspath.size() - (index + 1));

			if(shaderdirectory.size())
				LoadShaders(shaderdirectory);

			index = shaderspath.find(G_SEARCHPATH_SEPARATOR);
		}

	shaderdirectory = shaderspath;
	if(shaderdirectory.size())
		LoadShaders(shaderdirectory);

	return true;
}

// Load shaders from the given directory ...
bool CShaderCollection::LoadShaders(SDPCSTRING Directory)
{
	// Sanity checks ...
	g_assert_string(Directory);

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "Loading shaders from " << Directory << std::endl;
	
	return sdpForEachFile(Directory, *this, false);
}

// Called by sdpForEachFile ...
bool CShaderCollection::HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName)
{
	// Sanity checks ...
	const sdpString shadersourcepath(FilePath);
	g_return_val_if_fail(shadersourcepath.size(), true);

	// Skip files that aren't shader source code ...
	if(shadersourcepath.right(3) != ".sl")
		return true;

	// Some basic tracing ...
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "  Loading shader " << shadersourcepath << " ... " << std::endl;

	// Look for an existing metafile ...
	const sdpString metafilepath = sdpsl::MetafilePath(shadersourcepath);
	if(sdpFileExists(metafilepath))
		{
			if(SDP_S_OK == k3dApplication()->Verbose()) 
		    	std::cout << "\tLoading shader metafile ..." << std::endl;
			
			std::ifstream metafilestream(metafilepath);
			g_return_val_if_fail(sdpsl::ParseShaderMetafile(metafilestream, shadersourcepath, this), true);
		}
	// Didn't find it, so parse the source code ...
	else
		{
			if(SDP_S_OK == k3dApplication()->Verbose())
				std::cout << "\tParsing shader source code ..." << std::endl;

 			// Get the user's preferred shading-language preprocessor ...
 			sdpString preprocessorcommand = k3dOptionsGetSLPreprocessor(k3dApplication()->Options());
 			preprocessorcommand.formattedreplace('%', "p", shadersourcepath);
 
 			if(SDP_S_OK == k3dApplication()->Verbose())
 				std::cout << "\t" << preprocessorcommand << std::endl;
 
 			// Make it happen ...
 			std::strstream preprocessedshader;
 			g_return_val_if_fail(sdpRunProcess(preprocessorcommand, preprocessedshader), true);
 			preprocessedshader << std::ends;
 
			g_return_val_if_fail(sdpsl::ParseShaderSource(preprocessedshader, shadersourcepath, this), true);
		}

	return true;
}


