#ifndef SHADERS_H
#define SHADERS_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares classes for encapsulating information about a Pixar RenderMan Shading Language shader
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPSLPARSE_H
#include <sdpsl.h>
#endif // !SDPSLPARSE_H

// Forward class declarations
class k3dIApplication;

////////////////////////////////////////////////////////////////////
// CShaderArgument

/// Encapsulates a single Pixar RenderMan Shading Language shader argument
class CShaderArgument :
	public sdpsl::IShaderArgument
{
public:
	CShaderArgument(sdpsl::IShaderArgument* const ShaderArgument);
	virtual ~CShaderArgument() {}

	// sdpsl::IShaderArgument implementation
	SDPCSTRING Name();
	SLStorageClass StorageClass();
	SLType Type();
	SLExtendedType ExtendedType();
	gulong ArrayCount();
	SDPCSTRING Space();
	bool Output();
	SDPCSTRING Value();

protected:
	/// Stores the argument name
	const sdpString m_Name;
	/// Stores the argument storage class
	const sdpsl::IShaderArgument::SLStorageClass m_StorageClass;
	/// Stores the argument type
	const sdpsl::IShaderArgument::SLType m_Type;
	/// Stores the extended argument type
	const sdpsl::IShaderArgument::SLExtendedType m_ExtendedType;
	/// Stores the argument array-count
	const gulong m_ArrayCount;
	/// Stores the argument space
	const sdpString m_Space;
	/// Stores whether this is an output (part of the light-message system) argument
	const bool m_Output;
	/// Stores the argument value
	const sdpString m_Value;
};

////////////////////////////////////////////////////////////////////
// CShader

/// Encapsulates a single Pixar RenderMan Shading Language shader
class CShader :
	public sdpsl::IShader,
	public sdpsl::IShaderArgumentVisitor
{
public:
	CShader(sdpsl::IShader* const Shader);
	virtual ~CShader();

	// sdpsl::IShader implementation
	SDPCSTRING SourcePath();
	SDPCSTRING Name();
	SDPCSTRING Authors();
	SDPCSTRING Copyright();
	SDPCSTRING Description();
	SLType Type();
	SLExtendedType ExtendedType();
	gulong ArgumentCount();
	void VisitArguments(sdpsl::IShaderArgumentVisitor* const Visitor);
	
	// sdpsl::IShaderArgumentVisitor implementation
	void VisitShaderArgument(sdpsl::IShaderArgument* const Argument);

protected:
	/// Stores the shader source filepath
	const sdpString m_SourcePath;
	/// Stores the shader name (could be totally unrelated to the filepath)
	const sdpString m_Name;
	/// Stores the shader author(s)
	const sdpString m_Authors;
	/// Stores the shader copyright
	const sdpString m_Copyright;
	/// Stores a short description of the shader
	const sdpString m_Description;
	/// Stores the shader type
	const sdpsl::IShader::SLType m_Type;
	/// Stores the extended shader type
	const sdpsl::IShader::SLExtendedType m_ExtendedType;
	/// Stores the shader arguments
	typedef std::vector<CShaderArgument*> Arguments;
	typedef Arguments::iterator ArgumentIterator;
	Arguments m_Arguments;
};

//////////////////////////////////////////////////////////////////////////////////////
// CShaderCollection

/// Stores a collection of shaders
class CShaderCollection :
	public sdpsl::IShaderCollection,
	public sdpsl::IShaderVisitor
{
public:
	CShaderCollection(k3dIApplication* const Application);
	virtual ~CShaderCollection();

	// sdpsl::IShaderCollection implementation ...
	void VisitShaders(sdpsl::IShaderVisitor* const Visitor);

	// sdpsl::IShaderVisitor implementation ...
	void VisitShader(sdpsl::IShader* const Shader);

	/// Functor operator for use with sdpForEachFile()
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName);

	/// Loads shaders based on the user's environment
	bool LoadShaders();

	/// Loads shaders from the given directory
	bool LoadShaders(SDPCSTRING Directory);

protected:
	k3dIApplication* const m_Application;
	
	typedef std::vector<CShader*> Shaders;
	typedef Shaders::iterator ShaderIterator;
	Shaders m_Shaders;
};

#endif // SHADERS_H

