
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTimeStorage class, which provides an implementation of k3dITime
		\author Tim Shead (tshead@k-3d.com)
*/

#include "timestorage.h"

/////////////////////////////////////////////////////////////////////////////
// CTimeStorage

CTimeStorage::CTimeStorage(k3dIStateRecorder* const StateRecorder) :
//	k3dCommandNode(Application, "document"),
	m_StartTime("starttime", 0.0, StateRecorder, new k3dVariableRange<gdouble>(0.0, 1.0, 0)),
	m_ViewTime("viewtime", 0.0, StateRecorder, new k3dVariableRange<gdouble>(0.0, 1.0, 0)),
	m_EndTime("endtime", 1.0, StateRecorder, new k3dVariableRange<gdouble>(0.0, 1.0, 0)), // We have to adjust this after everything's initialized ...
	m_Duration("duration", 4.0, StateRecorder, new k3dVariableMinimum<gdouble>(0.0, 0)),
	m_FrameRate("framerate", 30.0, StateRecorder, new k3dVariableMinimum<gdouble>(0.0, 0)),
	m_RenderTime(0.0),
	m_RenderDeltaTime(0.0)
{
	// Keep the end time at the beginning of the last frame ...		
	k3dSetValue(m_EndTime, 1.0 - FrameDelta());

	AddVariable(&m_StartTime);
	AddVariable(&m_ViewTime);
	AddVariable(&m_EndTime);
	AddVariable(&m_Duration);
	AddVariable(&m_FrameRate);
}

CTimeStorage::~CTimeStorage()
{
}

/////////////////////////////////////////////////////////////////////////////
// sdpITime overrides ...

gdouble CTimeStorage::StartTime()
{
	return m_StartTime.Value();
}

gdouble CTimeStorage::ViewTime()
{
	return m_ViewTime.Value();
}

gdouble CTimeStorage::EndTime()
{
	return m_EndTime.Value();
}

gdouble CTimeStorage::Duration()
{
	return m_Duration.Value();
}

gdouble CTimeStorage::FrameRate()
{
	return m_FrameRate.Value();
}

gdouble CTimeStorage::FrameDelta()
{
	return 1.0 / (m_Duration.Value() * m_FrameRate.Value());
}

gdouble CTimeStorage::RenderTime()
{
	return m_RenderTime;
}

gdouble CTimeStorage::RenderDeltaTime()
{
	return m_RenderDeltaTime;
}

sdpResult CTimeStorage::SetStartTime(const gdouble Time)
{
	// Clamp 0.o <= starttime <= endtime
	const gdouble time = sdpClamp(0.0, Time, m_EndTime.Value());
	if(time != m_StartTime.Value())
		{
			k3dSetValue(m_StartTime, time);
			Notify(Event_StartTimeChanged, 0);
			
			SetViewTime(StartTime());
		}

	return SDP_S_OK;
}

sdpResult CTimeStorage::SetViewTime(const gdouble Time)
{
	// Clamp starttime <= viewtime <= endtime
	const gdouble time = sdpClamp(m_StartTime.Value(), Time, m_EndTime.Value());
	if(time != m_ViewTime.Value())
		{
			k3dSetValue(m_ViewTime, time);	
			Notify(Event_ViewTimeChanged, 0);
		}

	return SDP_S_OK;
}

sdpResult CTimeStorage::SetEndTime(const gdouble Time)
{
	// clamp starttime <= endtime <= 1.0
	const gdouble time = sdpClamp(m_StartTime.Value(), Time, 1.0);
	if(time != m_EndTime.Value())
		{
			k3dSetValue(m_EndTime, time);
			Notify(Event_EndTimeChanged, 0);
			
			SetViewTime(EndTime());
		}

	return SDP_S_OK;
}

sdpResult CTimeStorage::SetDuration(const gdouble Duration)
{
	const gdouble duration = sdpMax(0.0, Duration);
	if(duration != m_Duration.Value())
		{
			k3dSetValue(m_Duration, duration);
			Notify(Event_DurationChanged, 0);
			
			SetStartTime(0.0);
			SetViewTime(0.0);
			SetEndTime(1.0 - FrameDelta());
		}

	return SDP_S_OK;
}

sdpResult CTimeStorage::SetFrameRate(const gdouble FrameRate)
{
	const gdouble framerate = sdpMax(0.0, FrameRate);
	if(framerate != m_FrameRate.Value())
		{
			k3dSetValue(m_FrameRate, framerate);
			Notify(Event_FrameRateChanged, 0);
			
			SetStartTime(0.0);
			SetViewTime(0.0);
			SetEndTime(1.0 - FrameDelta());
		}

	return SDP_S_OK;
}

sdpResult CTimeStorage::SetRenderTime(const gdouble Time)
{
	m_RenderTime = Time;
	return SDP_S_OK;
}

sdpResult CTimeStorage::SetRenderDeltaTime(const gdouble Time)
{
	m_RenderDeltaTime = Time;
	return SDP_S_OK;
}

sdpResult CTimeStorage::Load(sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	sdpxml::ElementPointer time = sdpxml::FindElement(Element, sdpxml::SameName("time"));
	if(!time)
		return SDP_S_FALSE;

	g_return_val_if_fail(sdpSucceeded(k3dVariableContainer::Load(Document, *time)), SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CTimeStorage::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_return_val_if_fail(Dependencies, SDP_E_FAIL);
	
	sdpxml::Element& time = Element.Append(sdpxml::Element("time"));
	g_return_val_if_fail(sdpSucceeded(k3dVariableContainer::Save(time, Dependencies)), SDP_E_FAIL);
	
	return SDP_S_OK;
}


