// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTutorialMenu class, which provides a UI for running tutorials
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "tutorialmenu.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dscripting.h>

#include <sdpgtkevents.h>

#include <fstream>

//////////////////////////////////////////////////////////////////////
// CTutorialMenu

const sdpString CTutorialMenu::dialogname = "tutorialmenu";
const sdpString CTutorialMenu::control_tutorials = "tutorials";
const sdpString CTutorialMenu::control_play = "play";
const sdpString CTutorialMenu::control_cancel = "cancel";

CTutorialMenu::CTutorialMenu(CApplication* Application) :
	CTutorialMenuBase(Application, dialogname, new k3dOptionsWindowGeometryStore(Application, 0)),
	m_Application(Application)
{
	// Sanity checks ...
	g_assert(m_Application);

	g_return_if_fail(LoadGTKMLTemplate("tutorialmenu.gtkml"));
	
	// Connect the vscrollbar to our list of available tutorials ...
	CList(control_tutorials).SetVAdjustment(Range("vscrollbar").Adjustment());

	sdpGtkCList list = CList(control_tutorials);
	g_return_if_fail(list.Attached());

	// We will be generating a collection of strings to insert into our list ...
	sdpStringArray listentries;

	// Get a collection of available tutorial files ...
	sdpFileContainer tutorialfiles;
	sdpGetFiles(k3dApplication()->TutorialsPath(), false, tutorialfiles);

	// Look for an index file ...
	sdpFileIterator indexfile;
	for(indexfile = tutorialfiles.begin(); indexfile != tutorialfiles.end(); indexfile++)
		if(sdpFile(*indexfile) == "tutorialindex.k3d")
			break;

	// If we got one ...
	if(indexfile != tutorialfiles.end())
		{
			// Load it as XML ...
			sdpxml::Document index("k3dml");
			index.Load(*indexfile);
			
			// Remove it from the list of tutorial files ...
			tutorialfiles.erase(indexfile);
			
			// Get the set of indexed tutorials ...
			gulong tutorialnumber = 1;
			sdpxml::ElementPointerCollection indexedfiles = sdpxml::XPathSelect(index, index, "application/tutorials/tutorial");
			for(sdpxml::ElementPointerCollection::iterator indexedfile = indexedfiles.begin(); indexedfile != indexedfiles.end(); indexedfile++)
				{
					sdpString title = sdpxml::GetAttribute<sdpString>(**indexedfile, "title", "");
					title = sdpToString(tutorialnumber++) + ". " + title;
					
					sdpString filepath = sdpxml::GetAttribute<sdpString>(**indexedfile, "path", "");
					filepath = sdpAppendPath(k3dApplication()->TutorialsPath(), filepath);

					// Make sure the file exists ...
					for(sdpFileIterator file = tutorialfiles.begin(); file != tutorialfiles.end(); file++)
						{
							if(sdpFile(*file) == sdpFile(filepath))
								{
									tutorialfiles.erase(file);
									break;
								}
						}

					listentries.push_back(title);
					m_PathMap[title] = filepath;
				}
		}

	for(sdpFileIterator file = tutorialfiles.begin(); file != tutorialfiles.end(); file++)
		{
			listentries.push_back(sdpFile(*file));
			m_PathMap[sdpFile(*file)] = *file;
		}

	// Insert data into the list ...
	for(sdpStringArray::iterator listentry = listentries.begin(); listentry != listentries.end(); listentry++)
		list.Append(*listentry);

	// Setup recordable interactive events ...
	m_Application->AddEventObserver(this, k3dIApplication::Event_Close, "CTutorialMenu");

	Show();
}

CTutorialMenu::~CTutorialMenu()
{
	m_Application->RemoveEventObserver(this);
}

sdpResult CTutorialMenu::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIApplication::Event_Close)
		{
			delete this;
			return SDP_S_OK;
		}

	return CTutorialMenuBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void CTutorialMenu::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_play)
		OnPlay();
	else if(Event->Name() == control_cancel)
		OnCancel();
	else if(Event->Name() == control_tutorials)
		OnTutorials(Event);
	else
		CTutorialMenuBase::OnEvent(Event);
}

void CTutorialMenu::OnPlay()
{
	const sdpGtkCList::Rows selection = CList(control_tutorials).GetSelectedRows();
	g_return_if_fail(selection.size() == 1);

	const sdpString tutorial = CList(control_tutorials).GetText(selection[0], 0);
	const sdpString filepath = m_PathMap[tutorial];

	delete this;

	std::ifstream file(filepath);	
	k3d::ExecuteScript(file, filepath);
}

void CTutorialMenu::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't allow the GTK+ tree to go away just yet ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	// Turn it into a "cancel", instead ...
	OnCancel();
}

void CTutorialMenu::OnCancel()
{
	delete this;
}

void CTutorialMenu::OnTutorials(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;

	// If it isn't the LMB, forget it ...
	if(event->Event()->button != 1 || event->Event()->type != GDK_2BUTTON_PRESS)
		{
			event->SetResult(true);
			return;
		}

	// Turn it into a "play" event ...
	OnPlay();
}


