// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTutorialMessage informational dialog
		\author Tim Shead (tshead@k-3d.com)
*/

#include "tutorialmessage.h"

#include <k3dmisc.h>
#include <k3dmouseeventobserver.h>
#include <k3doptions.h>

#include <k3diapplication.h>

#include <sdpgtkevents.h>
#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

#include <iomanip>

#ifndef	SDPWIN32
#include <gdk/gdkx.h>
#endif //SDPWIN32

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

//////////////////////////////////////////////////////////////////////
// CTutorialMessage

const sdpString CTutorialMessage::dialogname = "tutorialmessage";
const sdpString CTutorialMessage::control_close = "close";
const sdpString CTutorialMessage::control_keypress = "keypress";
const sdpString CTutorialMessage::control_cruisecontrol = "cruisecontrol";
const sdpString CTutorialMessage::control_speed = "speed";
const sdpString CTutorialMessage::control_currentspeed = "currentspeed";

CTutorialMessage::CTutorialMessage(CTutorialMessage*& OwnerReference) :
	CTutorialMessageBase(k3dApplication(), dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_OwnerReference(OwnerReference),
	m_CurrentSpeed(k3dOptionsGetTutorialSpeed(k3dApplication()->Options())),
	m_Cancelled(false)
{
	g_return_if_fail(LoadGTKMLTemplate("tutorialmessage.gtkml"));

	SetMouseDiagram("mouse_dim");
	SetModifiersDiagram("modifiers_dim");	

	BlockAllEvents();
	
	// If we're in batch mode, enable cruise control ...
	ToggleButton(control_cruisecontrol).SetState(k3dApplication()->BatchMode());

	// Setup the tutorial speed control ...
	Adjustment(control_speed).SetValue(log10(m_CurrentSpeed));
	
	std::strstream buffer;
	buffer << "x " << std::setprecision(2) << m_CurrentSpeed << std::ends;	
	Label(control_currentspeed).SetText(buffer.str());
	buffer.freeze(false);
	
	UnblockAllEvents();

	gtk_window_set_position(RootWindow(), GTK_WIN_POS_CENTER);
	Show();
}

CTutorialMessage::~CTutorialMessage()
{
	m_OwnerReference = 0;
}

void CTutorialMessage::Update(SDPCSTRING Message)
{
	// Make sure we're visible and focused ...
	RootWidget().Realize();
	
	GdkWindow* window = GTK_WIDGET(RootWidget().Object())->window;
	g_return_if_fail(window);

	gdk_window_show(window);
	gdk_window_raise(window);
// 	XSetInputFocus(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(window), RevertToNone, GDK_CURRENT_TIME);

	SetMessage(Message);
	SetMouseDiagram("mouse_dim");
	SetModifiersDiagram("modifiers_dim");

	Widget("continue").SetSensitive(true);
	Widget("quit").SetSensitive(true);

	Widget("continue").InteractiveShow(m_CurrentSpeed, true);
	Widget("continue").InteractiveWarpPointer(m_CurrentSpeed, false, false);

	if(ToggleButton(control_cruisecontrol).GetState())
		Button("continue").InteractiveActivate();
	else
		DoModal();
}

static sdpString MouseDiagram(k3dIApplication::TutorialMouseAction Action)
{
	switch(Action)
		{
			case k3dIApplication::MOUSE_DONE:
				return "mouse_dim";
			case k3dIApplication::MB_NONE:
				return "mouse";
			case k3dIApplication::LMB_DRAG:
				return "mouse_lmb";
			case k3dIApplication::LMB_CLICK:
				return "mouse_lmb_1";
			case k3dIApplication::LMB_DOUBLE_CLICK:
				return "mouse_lmb_2";
			case k3dIApplication::MMB_DRAG:
				return "mouse_mmb";
			case k3dIApplication::MMB_CLICK:
				return "mouse_mmb_1";
			case k3dIApplication::MMB_DOUBLE_CLICK:
				return "mouse_mmb_2";
			case k3dIApplication::RMB_DRAG:
				return "mouse_rmb";
			case k3dIApplication::RMB_CLICK:
				return "mouse_rmb_1";
			case k3dIApplication::RMB_DOUBLE_CLICK:
				return "mouse_rmb_2";
			case k3dIApplication::LMBRMB_DRAG:
				return "mouse_lmb_rmb";
		}
		
	return sdpString();
}

static sdpString ModifiersDiagram(GdkModifierType Modifiers)
{
	sdpString result = "modifiers";
	
	if(k3dMouseEventObserver::ShiftDown(Modifiers))
		result += "_shift";
		
	if(k3dMouseEventObserver::ControlDown(Modifiers))
		result += "_ctrl";
		
	if(k3dMouseEventObserver::Mod1Down(Modifiers))
		result += "_alt";
	
	return result;
}

static sdpString MouseDescription(k3dIApplication::TutorialMouseAction Action)
{
	sdpString result;
	
	switch(Action)
		{
			case k3dIApplication::MOUSE_DONE:
				break;
			case k3dIApplication::MB_NONE:
				result = "Drag the mouse.";
				break;
			case k3dIApplication::LMB_DRAG:
				result = "Hold down the Left Mouse Button (LMB) and drag";
				break;
			case k3dIApplication::LMB_CLICK:
				result = "Click the Left Mouse Button (LMB)";
				break;
			case k3dIApplication::LMB_DOUBLE_CLICK:
				result = "Double-click the Left Mouse Button (LMB)";
				break;
			case k3dIApplication::MMB_DRAG:
				result = "Hold down the Middle Mouse Button (MMB) and drag";
				break;
			case k3dIApplication::MMB_CLICK:
				result = "Click the Middle Mouse Button (MMB)";
				break;
			case k3dIApplication::MMB_DOUBLE_CLICK:
				result = "Double-click the Middle Mouse Button (MMB)";
				break;
			case k3dIApplication::RMB_DRAG:
				result = "Hold down the Right Mouse Button (RMB) and drag";
				break;
			case k3dIApplication::RMB_CLICK:
				result = "Click the Right Mouse Button (RMB)";
				break;
			case k3dIApplication::RMB_DOUBLE_CLICK:
				result = "Double-click the Right Mouse Button (RMB)";
				break;
			case k3dIApplication::LMBRMB_DRAG:
				result = "Hold the the Left (LMB) and Right (RMB) mouse buttons and drag";
				break;
			default:
				g_return_val_if_fail(0, result);	// Did you add a mouse action without telling me?
		}
		
	return result;
}

static sdpString ModifiersDescription(GdkModifierType Modifiers)
{
	sdpString result;
	
	bool shift = k3dMouseEventObserver::ShiftDown(Modifiers);
	bool control = k3dMouseEventObserver::ControlDown(Modifiers);
	bool alt = k3dMouseEventObserver::Mod1Down(Modifiers);

	switch(shift + control + alt)
		{
			case 0:
				break;
			case 1:
				if(shift)
					result = "Hold down SHIFT";
				else if(control)
					result = "Hold down CTRL";
				else if(alt)
					result = "Hold down ALT";
				break;
			case 2:
				result += "Hold down ";
				if(shift)
					result += "SHIFT";
				else if(control)
					result += "CTRL";
				result += " and ";
				if(control)
					result += "CTRL";
				else if(alt)
					result += "ALT";
				break;
			case 3:
				result = "Hold down SHIFT, CTRL, and ALT";
				break;
			default:
				g_return_val_if_fail(0, result);
		}

	return result;
}

static sdpString ActionDescription(SDPCSTRING Message, k3dIApplication::TutorialMouseAction Action, GdkModifierType Modifiers)
{
	sdpString result;

	sdpString message(Message);
	sdpString modifiersdescription = ModifiersDescription(Modifiers);
	sdpString mousedescription = MouseDescription(Action);

	if(message.size())
		result += message + "\n";
			
	if(modifiersdescription.size())
		result += modifiersdescription + ".  ";
	
	if(mousedescription.size())
		result += mousedescription;
	
	result += ".";
	
	return result;
}

void CTutorialMessage::Update(SDPCSTRING Message, k3dIApplication::TutorialMouseAction Action, GdkModifierType Modifiers)
{
	gdk_window_show(GTK_WIDGET(RootWidget().Object())->window);
	gdk_window_raise(GTK_WIDGET(RootWidget().Object())->window);

	SetMessage(ActionDescription(Message, Action, Modifiers));
	SetMouseDiagram(MouseDiagram(Action));
	SetModifiersDiagram(ModifiersDiagram(Modifiers));

	Widget("continue").SetSensitive(false);
	Widget("quit").SetSensitive(false);

	switch(Action)
		{
			case k3dIApplication::MOUSE_DONE:
			case k3dIApplication::MB_NONE:
			case k3dIApplication::LMB_DRAG:
			case k3dIApplication::MMB_DRAG:
			case k3dIApplication::RMB_DRAG:
			case k3dIApplication::LMBRMB_DRAG:
				break;
			case k3dIApplication::LMB_CLICK:
			case k3dIApplication::LMB_DOUBLE_CLICK:
			case k3dIApplication::MMB_CLICK:
			case k3dIApplication::MMB_DOUBLE_CLICK:
			case k3dIApplication::RMB_CLICK:
			case k3dIApplication::RMB_DOUBLE_CLICK:
				{
//					sdpGtkEvent* timeout = MapEvent("timeout", "reset", RootWidget(), false);
//					static_cast<sdpGtkEventTimeout*>(timeout)->SetDuration(2000);
//					timeout->Connect();
				}
				break;
			default:
				g_return_if_fail(0);	// Did you add a mouse action without telling me?
		}
}

bool CTutorialMessage::Cancelled()
{
	return m_Cancelled;
}

////////////////////////////////////////////////////////////////////////
// k3dICommandNode overrides

sdpResult CTutorialMessage::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	if(Name == control_close)
		{
			delete this;
			return SDP_S_OK;
		}

	return CTutorialMessageBase::Command(Name, Arguments, Result);
}

void CTutorialMessage::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "continue")
		OnContinue();
	else if(Event->Name() == "quit")
		OnQuit();
	else if(Event->Name().left(5) == "mouse")
		OnMouseEvent(Event);
	else if(Event->Name() == control_cruisecontrol)
		OnCruiseControl();
	else if(Event->Name() == control_keypress)
		OnKeyPress(Event);
	else if(Event->Name() == control_speed)
		OnSpeed();
	else
		CTutorialMessageBase::OnEvent(Event);
}

void CTutorialMessage::OnDelete(sdpGtkEvent* const Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	// Kill the application ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);
	
	delete this;
}

void CTutorialMessage::OnKeyPress(sdpGtkEvent* const Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	// If we aren't in Cruise Control mode, we don't care ...
	if(!ToggleButton(control_cruisecontrol).GetState())
		return;
		
	// Turn-off Cruise Control ...
	ToggleButton(control_cruisecontrol).SetState(false);
	
	// Let the user know ...
	k3dApplication()->Message("Cruise Control disabled", "Tutorial:");
}

void CTutorialMessage::OnMouseEvent(sdpGtkEvent* const Event)
{
	static_cast<sdpGtkEventTimeout*>(Event)->SetResult(false);

	SetMouseDiagram(Event->Name());
}

void CTutorialMessage::OnCruiseControl()
{
	// If the user turned-off cruise control, we don't care ...
	if(!ToggleButton(control_cruisecontrol).GetState())
		return;
		
	// Give the user a chance to cancel ...
	sdpStringArray buttons;
	buttons.push_back("Yes");
	buttons.push_back("No");
	guint result = k3dApplication()->Query("Enable Cruise Control?  Once you hit \"Continue\" the tutorial will run automatically without stopping.\nPress any key while the mouse pointer is over the Tutorial window to stop.", "Tutorial:", 1, buttons);
	if(1 != result)
		ToggleButton(control_cruisecontrol).SetState(false);
}

void CTutorialMessage::OnSpeed()
{
	m_CurrentSpeed = pow(10.0, Adjustment(control_speed).Value());
	std::strstream buffer;
	buffer << "x " << std::setprecision(2) << m_CurrentSpeed << std::ends;
	Label(control_currentspeed).SetText(buffer.str());
	buffer.freeze(false);
	
	k3dOptionsSetTutorialSpeed(k3dApplication()->Options(), m_CurrentSpeed);
}

void CTutorialMessage::OnContinue()
{
	// To prevent the user from clicking too soon ...
	if(!IsModal())
		return;
		
	Widget("continue").SetSensitive(false);
	Widget("quit").SetSensitive(false);

	CancelModal();
}

void CTutorialMessage::OnQuit()
{
	if(!IsModal())
		return;

	m_Cancelled = true;

	CancelModal();
	delete this;
}

void CTutorialMessage::SetMessage(SDPCSTRING Message)
{
	// Prevent needless updates!
	if(m_CurrentMessage == Message)
		return;

	m_CurrentMessage = Message;
	
	Text("text").DeleteText(0, -1);

// The Win32 GTK+ port has a little glitch with clearing text from the text widget ...
#ifdef SDPWIN32
	Text("text").QueueClear();
	sdpGtkHandlePendingEvents();
#endif

	gint position = 0;
	Text("text").InsertText(Message, strlen(Message), &position);
}

void CTutorialMessage::SetMouseDiagram(SDPCSTRING	MouseDiagram)
{
	// Sanity checks ...
	g_assert_string(MouseDiagram);

	// Avoid unnecessary updates ...
	if(m_CurrentMouseDiagram == MouseDiagram)
		return;

	Pixmap(MouseDiagram).Show();
	
	if(m_CurrentMouseDiagram.size())
		Pixmap(m_CurrentMouseDiagram).Hide();
		
	m_CurrentMouseDiagram = MouseDiagram;
}

void CTutorialMessage::SetModifiersDiagram(SDPCSTRING ModifiersDiagram)
{
	// Sanity checks ...
	g_return_if_string_fail(ModifiersDiagram);
	
	// Avoid unnecessary updates ...
	if(m_CurrentModifiersDiagram == ModifiersDiagram)
		return;
		
	Pixmap(ModifiersDiagram).Show();
	
	if(m_CurrentModifiersDiagram.size())
		Pixmap(m_CurrentModifiersDiagram).Hide();
	
	m_CurrentModifiersDiagram = ModifiersDiagram;
}


