#ifndef K3DIAPPLICATION_H
#define K3DIAPPLICATION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares k3dIApplication, an abstract interface for the global K-3D Application object
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

#include <gdk/gdktypes.h>

// Forward class references
class k3dICommandNode;
class k3dICommandRecorder;
class k3dIDocument;
class k3dIObjectInfoCollection;
class k3dIRenderFarm;
class k3dIScriptEngine;

namespace sdpxml { class Document; }
namespace sdpsl { class IShaderCollection; }

/// Abstract interface for the global K-3D Application object
class k3dIApplication :
	public virtual sdpIUnknown
{
public:
	/// Specifies verbose debugging output
	virtual sdpResult Verbose() = 0;
	/// Specifies batch-mode (no user intervention)
	virtual bool BatchMode() = 0;

	/// Closes K-3D
	virtual sdpResult Close() = 0;

	/// Opens a K-3D document
	virtual k3dIDocument* OpenDocument(SDPCSTRING FilePath) = 0;
	/// Creates a new K-3D document
	virtual k3dIDocument* CreateDocument(SDPCSTRING TemplatePath) = 0;
	/// Returns the currently active K-3D document
	virtual k3dIDocument* ActiveDocument() = 0;
	/// Sets the currently active K-3D document
	virtual sdpResult SetActiveDocument(k3dIDocument* Document) = 0;
	/// Closes an open K-3D document
	virtual sdpResult CloseDocument(k3dIDocument* Document) = 0;

	/// Returns the absolute path to the K-3D base directory (contains bin, dialogtemplates, lib, scripts, share, tutorials, etc.)
	virtual SDPCSTRING BasePath() = 0;
	/// Returns the absolute path to the K-3D dialogtemplates directory
	virtual SDPCSTRING DialogTemplatesPath() = 0;
	/// Returns the absolute path to the K-3D lib (including plugins) directory
	virtual SDPCSTRING LibPath() = 0;
	/// Returns the absolute path to the K-3D scripts directory
	virtual SDPCSTRING ScriptsPath() = 0;
	/// Returns the absolute path to the K-3D shaders directory(s)
	virtual SDPCSTRING ShadersPath() = 0;
	/// Returns the absolute path to the K-3D shared-data directory
	virtual SDPCSTRING SharePath() = 0;
	/// Returns the absolute path to the K-3D tutorials directory
	virtual SDPCSTRING TutorialsPath() = 0;
	/// Returns the absolute path to the user's K-3D options directory
	virtual SDPCSTRING OptionsPath() = 0;
	
	/// Displays a URL in the user's preferred web browser
	virtual sdpResult BrowserNavigate(SDPCSTRING URL) = 0;
	/// Displays an informational message in a modal dialog box
	virtual void Message(SDPCSTRING Message, SDPCSTRING Title) = 0;
	/**
		 \brief Prompts the user to choose one of selections in a modal dialog box
		 \param Message the text displayed by the message box.
		 \param Title the title of the dialog box.
		 \param DefaultButton one-based index of the button that is selected by default.  If DefaultButton is 0, then no button is selected by default.
		 \return the number of the button selected by the user
	*/
	virtual guint Query(SDPCSTRING Message, SDPCSTRING Title, const guint DefaultButton, const sdpStringArray Buttons) = 0;
	/// Displays an error message in a modal dialog box
	virtual void Error(SDPCSTRING Message, SDPCSTRING Title) = 0;
	/// Handles debug output (usually by sending it to stdout)
	virtual sdpResult DebugOutput(SDPCSTRING Message) = 0;
	
	/// Enumerates mouse actions that can be displayed diagrammatically for tutorial purposes
	typedef enum
	{
		MOUSE_DONE,
		MB_NONE,
		LMB_DRAG,
		LMB_CLICK,
		LMB_DOUBLE_CLICK,
		MMB_DRAG,
		MMB_CLICK,
		MMB_DOUBLE_CLICK,
		RMB_DRAG,
		RMB_CLICK,
		RMB_DOUBLE_CLICK,
		LMBRMB_DRAG,
	} TutorialMouseAction;

	/// Displays a message (for tutorial purposes) in a dialog box
	virtual sdpResult TutorialMessage(SDPCSTRING Message) = 0;
	/// Displays an iconic mouse (for tutorial purposes) in a dialog box
	virtual sdpResult TutorialMouseDisplay(SDPCSTRING Message, TutorialMouseAction Action, GdkModifierType Modifiers) = 0;
	
	/// Returns the XML document that contains user options
	virtual sdpxml::Document& Options() = 0;

	/// Returns the collection of available plugin objects
	virtual k3dIObjectInfoCollection* Objects() = 0;
	/// Returns the collection of available RenderMan shaders
	virtual sdpsl::IShaderCollection* Shaders() = 0;
	/// Returns the current render farm
	virtual k3dIRenderFarm* RenderFarm() = 0;
	/// Returns a scripting engine object that supports the requested language
	virtual k3dIScriptEngine* ScriptEngine(const sdp::ClassID Language) = 0;
		
	/// Returns the root command-recording object
	virtual k3dICommandRecorder* CommandRecorder() = 0;
	/// Registers a command recording object to receive notification of activity
	virtual sdpResult RegisterCommandRecorder(k3dICommandRecorder* CommandRecorder) = 0;
	/// Unregisters the command-recording object
	virtual sdpResult UnRegisterCommandRecorder(k3dICommandRecorder* CommandRecorder) = 0;

	/// Makes the given text available as the X primary selection
	virtual sdpResult SetClipboard(SDPCSTRING Text) = 0;

	/// Command-node command to close the application
	static SDPCSTRING Command_Close;
	/// Command-node command to create a new document
	static SDPCSTRING Command_CreateDocument;
	/// Command-node command to open an existing document
	static SDPCSTRING Command_OpenDocument;

	/// K-3D event sent when the application closes
	static SDPCSTRING Event_Close;
	/// K-3D event sent when a document is created
	static SDPCSTRING Event_CreateDocument;
	/// K-3D event sent when a document is opened
	static SDPCSTRING Event_OpenDocument;
	/// K-3D event sent when a document is closed
	static SDPCSTRING Event_CloseDocument;
	/// K-3D event sent whenever a document becomes active (or the last document is deleted)
	static SDPCSTRING Event_ActiveDocumentChanged;

protected:
	virtual ~k3dIApplication() {}
};

#endif // K3DIAPPLICATION_H

