#ifndef K3DICAMERA_H
#define K3DICAMERA_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares k3dICamera, an abstract interface for an object capable of rendering a 2D view of the 3D workspace
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

// Forward class references
class k3dIViewVolume;

/// Abstract interface for an object capable of rendering a 2D view of the 3D workspace
class k3dICamera :
	public virtual sdpIUnknown
{
public:
	/// Makes the camera viewport visible
	virtual sdpResult ShowViewport() = 0;
	/// Closes the camera viewport
	virtual sdpResult HideViewport() = 0;
	/// Returns the width of the camera viewport (could be zero)
	virtual gulong Width() = 0;
	/// Returns the height of the camera viewport (could be zero)
	virtual gulong Height() = 0;
	/// Requests a redraw of the camera viewport
	virtual sdpResult Redraw(bool Synchronous) = 0;
	virtual sdpResult RenderPreview() = 0;
	virtual sdpResult RenderFrame() = 0;
	virtual sdpResult RenderAnimation() = 0;
	virtual sdpResult SnapshotCameraCoordinates() = 0;
	virtual sdpResult SnapshotScreenCoordinates() = 0;
	virtual k3dIViewVolume* ViewVolume() = 0;

	/// Enumerates the possible axes that can the "active"
	typedef enum
	{
		XAXIS,
		YAXIS,
		ZAXIS,
	} Axis;

	/// Returns the currently active axis
	virtual Axis ActiveAxis() = 0;
	/// Sets the current active axis
	virtual void SetActiveAxis(const Axis ActiveAxis) = 0;

	/// Enumerates the possible transformation types that can be "active"
	typedef enum
	{
		POSITION,
		ORIENTATION,
		SCALE
	} TransformType;

	/// Returns the currently active transformation type
	virtual TransformType ActiveTransformType() = 0;
	/// Sets the current active transformation type
	virtual void SetActiveTransformType(const TransformType ActiveTransformType) = 0;

	/// Enumerates the possible coordinate systems that can be "active"
	typedef enum
	{
		MODEL,
		WORLD,
		SCREEN,
	} CoordinateSystem;

	/// Returns the currently active coordinate system
	virtual CoordinateSystem ActiveCoordinateSystem() = 0;
	/// Sets the current coordinate system
	virtual void SetActiveCoordinateSystem(const CoordinateSystem System) = 0;

	/// Enumerates the possible edit modes than can be "active"
	typedef enum
	{
		POINT,
		POLYGON,
	} EditMode;

	/// Returns the currently active edit mode
	virtual EditMode ActiveEditMode() = 0;
	/// Sets the current active edit mode
	virtual void SetActiveEditMode(const EditMode Mode) = 0;

	/// Enumerates the possible "snap" modes that can be "active"
	typedef enum
	{
		FREEHAND,
		SNAP,
	} SnapMode;

	/// Returns the currently active snap mode
	virtual SnapMode ActiveSnapMode() = 0;
	/// Sets the current active snap mode
	virtual void SetActiveSnapMode(const SnapMode Mode) = 0;

	static SDPCSTRING Event_ActiveAxisChanged;
	static SDPCSTRING Event_ActiveTransformationChanged;
	static SDPCSTRING Event_ActiveCoordinateSystemChanged;
	static SDPCSTRING Event_ActiveEditModeChanged;
	static SDPCSTRING Event_ActiveSnapModeChanged;

protected:
	virtual ~k3dICamera() {}
};

#endif // K3DICAMERA_H


