#ifndef K3DINURBSMESH_H
#define K3DINURBSMESH_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares k3dINurbsMesh, an abstract interface for geometric objects containing NURBS points and NURBS paths
		\author Romain Behar with NURBS students (romainbehar@yahoo.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

// Forward class declarations
class k3dIObject;
class k3dIDocument;
class k3dITransform;
class k3dINurbsPathCollection;
class k3dINurbsControlPointCollection;
class k3dINurbsPath;
class k3dINurbsControlPoint;

/// Abstract interface for geometric objects containing NURBS points and NURBS paths
class k3dINurbsMesh :
	public virtual sdpIUnknown
{
public:
	/// Returns the parent document
	virtual k3dIDocument* Document() = 0;
	/// Returns an interface for accessing NURBS paths
	virtual k3dINurbsPathCollection* Paths() = 0;
	/// Creates a new NURBS path
	virtual k3dINurbsPath* CreatePath() = 0;
	/// Deletes an existing NURBS path
	virtual sdpResult DeletePath(k3dINurbsPath* const Path, const bool DeletePoints) = 0;
	/// Returns an interface for accessing points
	virtual k3dINurbsControlPointCollection* Points() = 0;
	/// Creates a new NURBS point
	virtual k3dINurbsControlPoint* CreatePoint() = 0;
	/// Deletes an existing NURBS point
	virtual sdpResult DeletePoint(k3dINurbsControlPoint* const Point) = 0;
	/// Sets smooth-shading for any paths that meet the given state and angle criteria
	virtual sdpResult ShadePaths(const gulong Criteria, const gdouble Angle) = 0;
	/// Removes smooth-shading from any paths that meet the given state criteria
	virtual sdpResult UnshadePaths(const gulong Criteria) = 0;
	virtual sdpResult WeldPoints(const gulong Criteria) = 0;
	virtual sdpResult UnweldPoints(const gulong Criteria) = 0;
	/// Forces a recalculation of path normals, including smooth-shading
	virtual sdpResult CalculateNormals() = 0;
	/// Returns the dimensions of all points that meet the given criteria
	virtual sdpResult Dimensions(const gulong Criteria, double Min[3], double Max[3]) = 0;
	/// Returns the mesh color (for use in editor views)
	virtual const sdpVector3& Color() const = 0;
	/// Sets the mesh color (for use in editor views)
	virtual void SetColor(const sdpVector3& Color) = 0;

protected:
	virtual ~k3dINurbsMesh() {}
};

#endif // K3DINURBSMESH_H
