#ifndef K3DISTATERECORDER_H
#define K3DISTATERECORDER_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dIStateRecorder and k3dIStateRecorderVisitor interfaces, which provide functionality for recording undo/redo -able state changes
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

// Forward class declarations
class k3dIAction;

/// Implemented by objects that encapsulate a change in state
class k3dIStateContainer
{
public:
	virtual ~k3dIStateContainer()
	{
	}

	/// When called, the derived object should restore the state it encapsulates
	virtual void RestoreState() = 0;
};

/// Abstract interface for an object that stores an atomic set of state changes that can be undone / redone
class k3dIStateChangeSet
{
public:
	virtual ~k3dIStateChangeSet()
	{
	}

	/// Records an original state that should be restored as part of an "undo" operation	
	virtual void RecordOldState(k3dIStateContainer* const OldState) = 0;
	/// Records a chain of actions to perform after old states are restored as part of an "undo" operation
	virtual void RecordOldAction(k3dIAction* const Action) = 0;
	/// Records a chain of actions to perform after an "undo" or a "redo" operation
	virtual void RecordAction(k3dIAction* const Action) = 0;
	/// Records a chain of actions to perform after new states have been restored as part of a "redo" operation
	virtual void RecordNewAction(k3dIAction* const Action) = 0;
	/// Records a new state that should be restored as part of a "redo" operation
	virtual void RecordNewState(k3dIStateContainer* const NewState) = 0;

	/// Restores original states, runs actions recorded with RecordOldAction(), then actions recorded with RecordAction()	
	virtual void Undo() = 0;
	/// Restores new states, runs actions recorded with RecordNewAction(), then actions recorded with RecordAction()
	virtual void Redo() = 0;
};

/// Implemented by objects that want to retrieve information about available undo/redo change sets
class k3dIChangeSetVisitor
{
public:
	virtual void VisitUndoChangeSet(SDPCSTRING UndoLabel, const bool Saved) = 0;
	virtual void VisitRedoChangeSet(SDPCSTRING RedoLabel, const bool Saved) = 0;
};

/// Encapsulates functionality for recording undo/redo -able state changes
class k3dIStateRecorder :
	public virtual sdpIUnknown
{
public:
	/// Called by clients to register a change set for recording subsequent state changes
	virtual void SetCurrentChangeSet(k3dIStateChangeSet* const ChangeSet) = 0;
	/// Called by clients to see if changes in state should be recorded
	virtual k3dIStateChangeSet* ChangeSet() = 0;
	/// Called by clients once a set of changes is complete, to make them a part of the undo/redo stack (the recorder assumes responsibility for the lifetime of the change set)
	virtual void RecordChangeSet(k3dIStateChangeSet* const ChangeSet, SDPCSTRING ChangeSetLabel) = 0;

	/// Called to mark the current state as saved
	virtual void MarkSaved() = 0;
	/// Returns true iff the recorder contains unsaved changes
	virtual bool UnsavedChanges() = 0;

	/// Returns the number of change sets left that can be undone
	virtual gulong UndoCount() = 0;
	/// Returns the number of undone change set that can be redone
	virtual gulong RedoCount() = 0;

	/// Returns the label of the state change set that will be undone with the next call to Undo()
	virtual SDPCSTRING NextUndoLabel() = 0;
	/// Returns the label of the state change set that will be redone with the next call to Redo()
	virtual SDPCSTRING NextRedoLabel() = 0;

	/// Called to undo the last change set
	virtual void Undo() = 0;
	/// Called to redo the most-recently-undone change set
	virtual void Redo() = 0;
	
	/// Enumerates through the entire collection of change sets
	virtual void VisitChangeSets(k3dIChangeSetVisitor* const Visitor) = 0;

	// K-3D events ...
	static SDPCSTRING Event_StackChanged;
	static SDPCSTRING Event_Undo;
	static SDPCSTRING Event_Redo;
	static SDPCSTRING Event_Saved;
};

#endif // K3DISTATERECORDER_H

