#ifndef K3DACTION_H
#define K3DACTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Provides a default implementation of k3dIAction for further derivation
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3diaction.h>

// Forward class declarations
class k3dICommandNode;
class k3dIDocument;

////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dAction

/// Provides an implementation of k3dIAction for further derivation
class k3dAction :
	public k3dIAction
{
public:
	k3dAction(k3dIAction* const NextAction);
	~k3dAction();
	
protected:
	/// Call this in derived-class DoAction() implementations
	void DoNextAction();
	
private:
	k3dIAction* const m_NextAction;
};

///////////////////////////////////////////////////////////////////////////
// k3dActionRedrawAll

/// k3dIAction implementation that redraws the screen
class k3dActionRedrawAll :
	public k3dAction
{
public:
	k3dActionRedrawAll(k3dIDocument* const Document, const bool Synchronous, k3dIAction* const NextAction);
	void DoAction();
	
protected:
	k3dIDocument* const m_Document;
	const bool m_Synchronous;
};

///////////////////////////////////////////////////////////////////////////
// k3dActionSendCommand

/// k3dIAction implementation that sends a command to a k3dICommandNode object
class k3dActionSendCommand :
	public k3dAction
{
public:
	k3dActionSendCommand(k3dICommandNode* const Target, SDPCSTRING Command, SDPCSTRING Arguments, k3dIAction* const NextAction);
	void DoAction();
	
protected:
	k3dICommandNode* const m_Target;
	const sdpString m_Command;
	const sdpString m_Arguments;
};

///////////////////////////////////////////////////////////////////////////
// k3dActionMethodCall

/// k3dIAction implementation that calls an arbitrary method on another object
template<class Type>
class k3dActionMethodCall :
	public k3dAction
{
public:
	explicit k3dActionMethodCall(Type* const Object, void (Type::*Method)(), k3dIAction* const NextAction) :
		k3dAction(NextAction),
		m_Object(Object),
		m_Method(Method)
	{
		// Sanity checks ...
		g_assert(m_Object);
		g_assert(m_Method);
	}

	void DoAction()
	{
		(m_Object->*m_Method)();
		DoNextAction();
	}

private:
	Type* const m_Object;
	void (Type::*m_Method)();
};

///////////////////////////////////////////////////////////////////////////
// k3dActionDebugMessage

/// k3dIAction implementation that writes a string to stderr for debugging purposes
class k3dActionDebugMessage :
	public k3dAction
{
public:
	k3dActionDebugMessage(SDPCSTRING Message, k3dIAction* const NextAction);
	void DoAction();

private:
	const sdpString m_Message;
};

///////////////////////////////////////////////////////////////////////////
// k3dActionUpdate

/// k3dIAction implementation that calls an object's Update(bool) method
template<class Type>
class k3dActionUpdate :
	public k3dAction
{
public:
	k3dActionUpdate(Type* const Object, const bool RecordState, k3dIAction* const NextAction) :
		k3dAction(NextAction),
		m_Object(Object),
		m_RecordState(RecordState)
	{
		// Sanity checks ...
		g_assert(m_Object);
	}
	
	void DoAction()
	{
		m_Object->Update(m_RecordState);
		DoNextAction();	
	}
	
private:
	Type* const m_Object;
	const bool m_RecordState;
};

///////////////////////////////////////////////////////////////////////////
// k3dActionUpdate2

/// k3dIAction implementation that calls an object's Update(bool, bool) method
template<class Type>
class k3dActionUpdate2 :
	public k3dAction
{
public:
	k3dActionUpdate2(Type* const Object, const bool RecordState, const bool CreateGeometry, k3dIAction* const NextAction) :
		k3dAction(NextAction),
		m_Object(Object),
		m_RecordState(RecordState),
		m_CreateGeometry(CreateGeometry)
	{
		// Sanity checks ...
		g_assert(m_Object);
	}
	
	void DoAction()
	{
		m_Object->Update(m_RecordState, m_CreateGeometry);
		DoNextAction();	
	}
	
private:
	Type* const m_Object;
	const bool m_RecordState;
	const bool m_CreateGeometry;
};

#endif // K3DACTION_H

