#ifndef K3DBOUNDINGBOX_H
#define K3DBOUNDINGBOX_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dBoundingBox2 and k3dBoundingBox3 classes, which encapsulate rectangular and cubic regions, respectively
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

//////////////////////////////////////////////////////////////////////////////////
// k3dBoundingBox2

class k3dBoundingBox2
{
public:
	k3dBoundingBox2() :
		m_Min(DBL_MAX),
		m_Max(-DBL_MAX)
	{
	}
		
	k3dBoundingBox2(const sdpVector2 Corner1, const sdpVector2 Corner2) :
		m_Min(sdpVectorMin(Corner1, Corner2)),
		m_Max(sdpVectorMax(Corner1, Corner2))
	{
	}
	
	void Expand(const sdpVector2 Point)
	{
		m_Min = sdpVectorMin(m_Min, Point);
		m_Max = sdpVectorMax(m_Max, Point);
	}
	
	bool IsDefined() const
	{
		return (m_Min[0] <= m_Max[0]) && (m_Min[1] <= m_Max[1]) && (m_Min[2] <= m_Max[2]);
	}

	bool Contains(const sdpVector2 Point) const
	{
		return (m_Min[0] <= Point[0]) && (Point[0] <= m_Max[0]) && (m_Min[1] <= Point[1]) && (Point[1] <= m_Max[1]);
	}

	sdpVector2 Min() const
	{
		return m_Min;
	}
	
	sdpVector2 Max() const
	{
		return m_Max;
	}
	
	sdpVector2 Center() const
	{
		return 0.5 * (m_Min + m_Max);
	}
	
	sdpVector2 Size() const
	{
		return m_Max - m_Min;
	}

	friend k3dBoundingBox2 k3dIntersect(const k3dBoundingBox2& A, const k3dBoundingBox2& B)
	{
		return k3dBoundingBox2(sdpVectorMax(A.m_Min, B.m_Min), sdpVectorMin(A.m_Max, B.m_Max));
	}

protected:
	sdpVector2 m_Min;
	sdpVector2 m_Max;
};

//////////////////////////////////////////////////////////////////////////////////
// k3dBoundingBox3

class k3dBoundingBox3
{
public:
	k3dBoundingBox3() :
		m_Min(DBL_MAX),
		m_Max(-DBL_MAX)
	{
	}
	
	k3dBoundingBox3(const sdpVector3 Corner1, const sdpVector3 Corner2) :
		m_Min(sdpVectorMin(Corner1, Corner2)),
		m_Max(sdpVectorMax(Corner1, Corner2))
	{
	}

	void Expand(const sdpVector3 Point)
	{
		m_Min = sdpVectorMin(m_Min, Point);
		m_Max = sdpVectorMax(m_Max, Point);
	}
	
	/// Returns true iff the current bounding box is defined (m_Min < m_Max)
	bool IsDefined() const
	{
		return (m_Min[0] <= m_Max[0]) && (m_Min[1] <= m_Max[1]) && (m_Min[2] <= m_Max[2]);
	}
	
	/// Tests to see if the given point is within the current bounds
	bool Contains(const sdpVector3 Point) const
	{
		return (m_Min[0] <= Point[0]) && (Point[0] <= m_Max[0]) &&
				 (m_Min[1] <= Point[1]) && (Point[1] <= m_Max[1]) &&
				 (m_Min[2] <= Point[2]) && (Point[2] <= m_Max[2]);
	}

	/// Returns the minimum bounds
	sdpVector3 Min() const
	{
		return m_Min;
	}
	
	/// Returns the maximum bounds
	sdpVector3 Max() const
	{
		return m_Max;
	}
	
	/// Returns the geometric center
	sdpVector3 Center() const
	{
		return 0.5 * (m_Min + m_Max);
	}
	
	/// Returns the size of the bounds
	sdpVector3 Size() const
	{
		return m_Max - m_Min;
	}

	friend k3dBoundingBox3 k3dIntersect(const k3dBoundingBox3& A, const k3dBoundingBox3& B)
	{
		return k3dBoundingBox3(sdpVectorMax(A.m_Min, B.m_Min), sdpVectorMin(A.m_Max, B.m_Max));
	}

protected:
	sdpVector3 m_Min;
	sdpVector3 m_Max;
};

#endif // K3DBOUNDINGBOX_H
