// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements utility classes / templates / functions related to creating new geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dbuildgeometry.h>
#include <k3dcom.h>
#include <k3dmisc.h>

#include <k3didocument.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dimouseeventobserver.h>
#include <k3diobject.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>
#include <k3dipoint.h>
#include <k3distate.h>
#include <k3distaterecorder.h>

namespace
{
	
const sdpVector3 g_NewColor(0, 0, 1);
const sdpVector3 g_DefaultColor(0, 0, 0);

class SetManagedState
{
public:
	void operator()(k3dIMesh* const Mesh)
	{
		k3d::Editing editing(Mesh);
		
		Mesh->SetColor(g_NewColor);
		dynamic_cast<k3dIObject*>(Mesh)->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE, false);
	}

	void operator()(k3d::NewMeshProxy* const Mesh)
	{
		operator()(*Mesh);
	}
			
	void operator()(k3d::EngineMeshProxy* const Mesh)
	{
		operator()(Mesh->Actual());
	}
	
	void operator()(k3dIPath* const Path)
	{
		k3d::Editing editing(Path);
		
		Path->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE, false);
		Path->Materials()->SetMaterial(k3dDefaultMaterial(Path->Mesh()->Document()));
	}
	
	void operator()(k3d::NewPathProxy* const Path)
	{
		operator()(*Path);
	}
	
	void operator()(k3d::EnginePathProxy* const Path)
	{
		operator()(Path->Actual());
	}
	
	void operator()(k3dIPoint* const Point)
	{
		Point->State()->SetState(k3dIState::SELECTABLE | k3dIState::SAVEABLE, false);
	}
	
	void operator()(k3d::NewPointProxy* const Point)
	{
		operator()(*Point);
	}
	
	void operator()(k3d::EnginePointProxy* const Point)
	{
		operator()(Point->Actual());
	}
	
	void operator()(k3d::EngineEdgeProxy* const Edge)
	{
	}
};

class SetFinishedState
{
public:
	void operator()(k3dIMesh* const Mesh)
	{
		k3d::Editing editing(Mesh);
		
		Mesh->SetColor(g_DefaultColor);
		dynamic_cast<k3dIObject*>(Mesh)->State()->SetState(k3dIState::SELECTED | k3dIState::SELECTABLE | k3dIState::SAVEABLE, true);
	}
	
	void operator()(k3d::NewMeshProxy* const Mesh)
	{
		operator()(*Mesh);
	}
	
	void operator()(k3d::EngineMeshProxy* const Mesh)
	{
		operator()(Mesh->Actual());
	}
	
	void operator()(k3dIPath* const Path)
	{
		Path->State()->SetState(k3dIState::SELECTED | k3dIState::SELECTABLE | k3dIState::SAVEABLE, true);
	}
	
	void operator()(k3d::NewPathProxy* const Path)
	{
		operator()(*Path);
	}
	
	void operator()(k3d::EnginePathProxy* const Path)
	{
		operator()(Path->Actual());
	}
	
	void operator()(k3dIPoint* const Point)
	{
		Point->State()->SetState(k3dIState::SELECTED | k3dIState::SELECTABLE | k3dIState::SAVEABLE, true);
	}
	
	void operator()(k3d::NewPointProxy* const Point)
	{
		operator()(*Point);
	}
	
	void operator()(k3d::EnginePointProxy* const Point)
	{
		operator()(Point->Actual());
	}
	
	void operator()(k3d::EngineEdgeProxy* const Edge)
	{
	}
};

gulong GridColumn(const gulong Column, const gulong Columns, const bool StitchSide)
{
	return Column % (StitchSide ? Columns : Columns+1);
}

gulong GridRow(const gulong Row, const gulong Rows, const bool StitchTop)
{
	return Row % (StitchTop ? Rows : Rows+1);
}

void AttachGridPoint(k3dIPath* const Path, k3d::NewPointProxy* const Point)
{
	// Sanity checks ...
	g_return_if_fail(Path);
	g_return_if_fail(Point);
	
	// Attach the point to the path ...
	k3dIPathPoint* const pathpoint = Path->Points()->AddPoint(*Point);
	g_return_if_fail(pathpoint);
	
	// Store texture coordinates ...
	k3d::Editing editing(pathpoint);
	pathpoint->Location()->SetTextureXYZ(Point->Parameters());
	pathpoint->Location()->SetImplicitTextureXYZ(Point->Parameters());
}

bool IsCubeSurfaceIndex(const gulong XSegments, const gulong YSegments, const gulong ZSegments, const gulong X, const gulong Y, const gulong Z)
{
	// Sanity checks ...
	g_return_val_if_fail(XSegments, false);
	g_return_val_if_fail(YSegments, false);
	g_return_val_if_fail(ZSegments, false);
	
	return ((0 == X) || (XSegments == X)) || ((0 == Y) || (YSegments == Y)) || ((0 == Z) || (ZSegments == Z));
}

void AttachCubePoint(k3dIPath* const Path, k3dIPoint* const Point, const gdouble S, const gdouble T)
{
	// Sanity checks ...
	g_return_if_fail(Path);
	g_return_if_fail(Point);

	// Attach the point to the path ...
	k3dIPathPoint* const pathpoint = Path->Points()->AddPoint(Point);
	g_return_if_fail(pathpoint);

	// Calculate texture coordinates ...
	const sdpVector3 texturecoords(S, T, 0.0);

	// Store point texture coordinates (yes, we end up creating discontinuities) ...
	Point->Location()->SetTextureXYZ(texturecoords);
	Point->Location()->SetImplicitTextureXYZ(texturecoords);

	// Store pathpoint texture coordinates
	k3d::Editing editing(pathpoint);
	pathpoint->Location()->SetTextureXYZ(texturecoords);
	pathpoint->Location()->SetImplicitTextureXYZ(texturecoords);
}

const sdpString NewGeometryDescription(const gulong PointCount, const gulong PathCount, const gulong MeshCount)
{
	sdpString result;

	if(PointCount)
		{
			result += sdpToString(PointCount) + " Point";
			if(PointCount > 1)
				result += "s";
		}

	if(PathCount)
		{
			if(result.size())
				result += ", ";

			result += sdpToString(PathCount) + " Path";
			if(PathCount > 1)
				result += "s";
		}

	if(MeshCount)
		{
			if(result.size())
				result += ", ";

			result += sdpToString(MeshCount) + " Mesh";
			if(MeshCount > 1)
				result += "s";
		}

	if(result.size())
		result = "Created " + result;
	
	return result;
}

} // namespace

namespace k3d
{

//////////////////////////////////////////////////////////////////////////////
// NewMeshProxy

NewMeshProxy::NewMeshProxy(k3dIMesh* const Mesh) :
	m_Mesh(Mesh)
{
	// Sanity checks ...
	g_assert(m_Mesh);
	g_assert(dynamic_cast<k3dIObject*>(m_Mesh));
}

void NewMeshProxy::PushEditing()
{
	k3d::PushEditing(m_Mesh);
}

void NewMeshProxy::PopEditing()
{
	k3d::PopEditing(m_Mesh);
}

//////////////////////////////////////////////////////////////////////////////
// NewPathProxy

NewPathProxy::NewPathProxy(k3dIPath* const Path, const sdpVector3 Parameters) :
	m_Path(Path),
	m_Parameters(Parameters)
{
	// Sanity checks ...
	g_assert(m_Path);
}

void NewPathProxy::PushEditing()
{
	k3d::PushEditing(m_Path);
}

void NewPathProxy::PopEditing()
{
	k3d::PopEditing(m_Path);
}

//////////////////////////////////////////////////////////////////////////////
// NewPointProxy

NewPointProxy::NewPointProxy(k3dIPoint* const Point, const sdpVector3 Parameters) :
	m_Point(Point),
	m_Parameters(Parameters)
{
	// Sanity checks ...
	g_assert(m_Point);
}

void NewPointProxy::PushEditing()
{
	k3d::PushEditing(m_Point);
}

void NewPointProxy::PopEditing()
{
	k3d::PopEditing(m_Point);
}

//////////////////////////////////////////////////////////////////////////////////
// NewGeometryCollection

NewGeometryCollection::NewGeometryCollection() :
	m_Changes(0)
{
}

NewGeometryCollection::~NewGeometryCollection()
{
	Clear();
}

void NewGeometryCollection::Clear()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), sdp::DeleteObject());
	std::for_each(m_Paths.begin(), m_Paths.end(), sdp::DeleteObject());
	std::for_each(m_Points.begin(), m_Points.end(), sdp::DeleteObject());

	m_Meshes.clear();
	m_Paths.clear();
	m_Points.clear();
}

void NewGeometryCollection::StartChangeSet(k3dIDocument* const Document)
{
	// Sanity checks ...
	g_return_if_fail(Document);
	g_return_if_fail(!m_Changes);
	
	m_Changes = k3dStartStateChangeSet(Document);
	
	k3d::PushEditing(Document->Objects());
}

void NewGeometryCollection::FinishChangeSet(k3dIDocument* const Document)
{
	// Sanity checks ...
	g_return_if_fail(Document);
	g_return_if_fail(m_Changes);

	k3d::PopEditing(Document->Objects());

	Document->StateRecorder()->SetCurrentChangeSet(0);
}

// k3dIEditable implementation
void NewGeometryCollection::PushEditing()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), std::mem_fun(&NewMeshProxy::PushEditing));
	std::for_each(m_Paths.begin(), m_Paths.end(), std::mem_fun(&NewPathProxy::PushEditing));
	std::for_each(m_Points.begin(), m_Points.end(), std::mem_fun(&NewPointProxy::PushEditing));
}

void NewGeometryCollection::PopEditing()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), std::mem_fun(&NewMeshProxy::PopEditing));
	std::for_each(m_Paths.begin(), m_Paths.end(), std::mem_fun(&NewPathProxy::PopEditing));
	std::for_each(m_Points.begin(), m_Points.end(), std::mem_fun(&NewPointProxy::PopEditing));
}

NewMeshProxy* NewGeometryCollection::Insert(k3dIMesh* const Mesh)
{
	// Sanity checks ...
	g_return_val_if_fail(Mesh, 0);
	
	m_Meshes.push_back(new NewMeshProxy(Mesh));
	return m_Meshes.back();
}

NewPathProxy* NewGeometryCollection::Insert(k3dIPath* const Path, const sdpVector3 Parameters)
{
	// Sanity checks ...
	g_return_val_if_fail(Path, 0);
	
	m_Paths.push_back(new NewPathProxy(Path, Parameters));
	return m_Paths.back();
}

NewPointProxy* NewGeometryCollection::Insert(k3dIPoint* const Point, const sdpVector3 Parameters)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	
	m_Points.push_back(new NewPointProxy(Point, Parameters));
	return m_Points.back();
}

void NewGeometryCollection::Manage()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), SetManagedState());
	std::for_each(m_Paths.begin(), m_Paths.end(), SetManagedState());
	std::for_each(m_Points.begin(), m_Points.end(), SetManagedState());
}

void NewGeometryCollection::Cancel()
{
	if(m_Changes)
		{
			m_Changes->Undo();
			delete m_Changes;
			m_Changes = 0;
		}

	Clear();		
}

void NewGeometryCollection::Finish(k3dIDocument* const Document, SDPCSTRING ChangeSetName)
{
	// Sanity checks ...
	g_return_if_fail(Document);

	std::for_each(m_Meshes.begin(), m_Meshes.end(), SetFinishedState());
	std::for_each(m_Paths.begin(), m_Paths.end(), SetFinishedState());
	std::for_each(m_Points.begin(), m_Points.end(), SetFinishedState());
	
	if(m_Changes)
		{
			Document->StateRecorder()->RecordChangeSet(m_Changes, ChangeSetName);
			m_Changes = 0;
		}
		
	for(NewMeshes::iterator mesh = m_Meshes.begin(); mesh != m_Meshes.end(); ++mesh)
		{
			k3dIMouseEventObserver* const observer = dynamic_cast<k3dIMouseEventObserver*>((*mesh)->Object());
			if(observer)
				{
					Document->SetMouseFocus(observer);
					break;
				}
		}

	Clear();		
}

const sdpString NewGeometryCollection::Description() const 
{
	return NewGeometryDescription(m_Points.size(), m_Paths.size(), m_Meshes.size());
}

////////////////////////////////////////////////////////////////////////////////////////////////
// CreateNewMesh

void CreateNewMesh(k3dIDocument* const Document, SDPCSTRING Name, NewGeometryCollection& Collection)
{
	// Sanity checks ...
	g_return_if_fail(Document);
	g_return_if_string_fail(Name);
	
	k3dIObject* const meshobject = Document->Objects()->CreateObject(k3d::classid::Mesh);
	g_return_if_fail(meshobject);
	
	k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(meshobject);
	g_return_if_fail(mesh);
	
	meshobject->SetName(Name);
	
	Collection.Insert(mesh);
}

////////////////////////////////////////////////////////////////////////////////////////////////
// CreateNewPath

void CreateNewPath(k3dIMesh* const Mesh, const gulong PointCount, NewGeometryCollection& Collection)
{
	// Sanity checks ...
	g_return_if_fail(Mesh);

	// Create the new path ...
	k3dIPath* const path = Mesh->CreatePath();
	g_return_if_fail(path);

	// Insert the path into the collection and set it up for further modification ...
	NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(0, 0, 0));
	g_return_if_fail(pathproxy);
	pathproxy->PushEditing();

	// Create and attach points ...
	for(gulong i = 0; i < PointCount; i++)
		{
			// Create the point ...
			k3dIPoint* const point = Mesh->CreatePoint();
			g_return_if_fail(point);

			// Setup some default parameters ...
			const sdpVector3 parameters(static_cast<gdouble>(i) / static_cast<gdouble>(PointCount), 0.0, 0.0);
			
			// Insert the point into the collection and set it up for further modification ...
			NewPointProxy* const pointproxy = Collection.Insert(point, parameters);
			g_return_if_fail(pointproxy);
			pointproxy->PushEditing();

			// Attach it to the path ...
			k3dIPathPoint* const pathpoint = path->Points()->AddPoint(point);
			g_return_if_fail(pathpoint);
			
			// Setup some default texture coordinates ...
			point->Location()->SetTextureXYZ(parameters);
			point->Location()->SetImplicitTextureXYZ(parameters);
			
			k3d::PushEditing(pathpoint);
			pathpoint->Location()->SetTextureXYZ(parameters);
			pathpoint->Location()->SetImplicitTextureXYZ(parameters);
			k3d::PopEditing(pathpoint);
		}
}

///////////////////////////////////////////////////////////////////////////////////////////////
// CreateNewPointsPaths

void CreateNewPointsPaths(k3dIMesh* const Mesh, const gulong PointCount, const gulong PathCount, NewGeometryCollection& Collection)
{
	// Sanity checks ...
	g_return_if_fail(Mesh);
	
	// Create new points ...
	for(gulong i = 0; i < PointCount; i++)
		{
			// Create the point ...
			k3dIPoint* const point = Mesh->CreatePoint();
			g_return_if_fail(point);

			// Setup some default parameters ...
			const sdpVector3 parameters(static_cast<gdouble>(i) / static_cast<gdouble>(PointCount-1), 0.0, 0.0);

			// Insert the new point into the collection and set it up for further modification ...
			NewPointProxy* const pointproxy = Collection.Insert(point, parameters);
			g_return_if_fail(pointproxy);
			pointproxy->PushEditing();			

			// Setup default texture coordinates ...
			point->Location()->SetTextureXYZ(parameters);
			point->Location()->SetImplicitTextureXYZ(parameters);
		}

	// Create new paths ...
	for(gulong i = 0; i < PathCount; i++)
		{
			// Create the path ...
			k3dIPath* const path = Mesh->CreatePath();
			g_return_if_fail(path);

			// Setup some default parameters
			const sdpVector3 parameters(static_cast<gdouble>(i) / static_cast<gdouble>(PathCount-1), 0.0, 0.0);

			// Insert the new path into the collection and set it up for further modification ...
			NewPathProxy* const pathproxy = Collection.Insert(path, parameters);
			g_return_if_fail(pathproxy);
			pathproxy->PushEditing();
		}
}


///////////////////////////////////////////////////////////////////////////////////////////////
// CreateNewGrid

void CreateNewGrid(k3dIMesh* const Mesh, const gulong Rows, const gulong Columns, const bool StitchTop, const bool StitchSide, NewGeometryCollection& Collection)
{
	// Sanity checks ...
	g_return_if_fail(Mesh);
	g_return_if_fail(Rows);
	g_return_if_fail(Columns);

	// Calculate the number of paths along each axis ...
	const gulong pathrows = Rows;
	const gulong pathcolumns = Columns;

	// Create paths ...
	typedef std::vector<k3dIPath*> PathRow;
	typedef std::vector<PathRow> PathGrid;
	PathGrid paths(pathrows, PathRow(pathcolumns));
	for(gulong pathrow = 0; pathrow < pathrows; pathrow++)
		{
			for(gulong pathcolumn = 0; pathcolumn < pathcolumns; pathcolumn++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);
					paths[pathrow][pathcolumn] = path;
					
					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(pathcolumn, pathrow, 0));
					g_return_if_fail(pathproxy);
					pathproxy->PushEditing();
				}
		}

	// Calculate the actual number of points that need to be created along each axis ...
	const gulong pointrows = pathrows + (StitchTop ? 0 : 1);
	const gulong pointcolumns = pathcolumns + (StitchSide ? 0 : 1);
	
	// Create points ...
	typedef std::vector<NewPointProxy*> PointRow;
	typedef std::vector<PointRow> GridPoints;
	GridPoints points(pointrows, PointRow(pointcolumns));
	
	for(gulong pointrow = 0; pointrow < pointrows; pointrow++)
		{
			for(gulong pointcolumn = 0; pointcolumn < pointcolumns; pointcolumn++)
				{
					// Create a point ...
					k3dIPoint* const point = Mesh->CreatePoint();
					g_return_if_fail(point);

					// Setup some default parameters / texture coordinates ...
					const sdpVector3 parameters(gdouble(pointcolumn) / gdouble(StitchSide ? pointcolumns : pointcolumns-1), gdouble(pointrow) / gdouble(StitchTop ? pointrows : pointrows-1), 0);

					// Insert the point into the collection and set it up for further modification ...
					NewPointProxy* const pointproxy = Collection.Insert(point, parameters);
					g_return_if_fail(pointproxy);
					pointproxy->PushEditing();
					
					points[pointrow][pointcolumn] = pointproxy;
					
					(*pointproxy)->Location()->SetTextureXYZ(parameters);
					(*pointproxy)->Location()->SetImplicitTextureXYZ(parameters);				
					
				}
		}

	// Attach points to paths ...
	for(gulong pathrow = 0; pathrow < pathrows; pathrow++)
		{
			for(gulong pathcolumn = 0; pathcolumn < pathcolumns; pathcolumn++)
				{
					// Attach four points in clockwise (for our LH coordinate system) order ...
					k3dIPath* const path = paths[pathrow][pathcolumn];
					
					AttachGridPoint(path, points[GridRow(pathrow, pathrows, StitchTop)][GridColumn(pathcolumn, pathcolumns, StitchSide)]);
					AttachGridPoint(path, points[GridRow(pathrow, pathrows, StitchTop)][GridColumn(pathcolumn+1, pathcolumns, StitchSide)]);
					AttachGridPoint(path, points[GridRow(pathrow+1, pathrows, StitchTop)][GridColumn(pathcolumn+1, pathcolumns, StitchSide)]);
					AttachGridPoint(path, points[GridRow(pathrow+1, pathrows, StitchTop)][GridColumn(pathcolumn, pathcolumns, StitchSide)]);
				}
		}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CreateNewCube

void CreateNewCube(k3dIMesh* const Mesh, const gulong XSegments, const gulong YSegments, const gulong ZSegments, NewGeometryCollection& Collection)
{
	// Sanity checks ...
	g_return_if_fail(Mesh);
	g_return_if_fail(XSegments);
	g_return_if_fail(YSegments);
	g_return_if_fail(ZSegments);
	
	// Calculate points required ...
	const gulong xpoints = XSegments + 1;
	const gulong ypoints = YSegments + 1;
	const gulong zpoints = ZSegments + 1;
	
	// Create points ...
	typedef std::vector<k3dIPoint*> PointRow;
	typedef std::vector<PointRow> PointGrid;
	typedef std::vector<PointGrid> CubePoints;
	CubePoints points(zpoints, PointGrid(ypoints, PointRow(xpoints)));
	
	for(gulong z = 0; z < zpoints; z++)
		{
			for(gulong y = 0; y < ypoints; y++)
				{
					for(gulong x = 0; x < xpoints; x++)
						{
							if(!IsCubeSurfaceIndex(XSegments, YSegments, ZSegments, x, y, z))
								{
									points[z][y][x] = 0;
									continue;
								}
								
							// Create a point ...
							k3dIPoint* const point = Mesh->CreatePoint();
							g_return_if_fail(point);
							
							// Insert it into the collection and set it up for further modification ...
							const sdpVector3 parameters(gdouble(x) / gdouble(XSegments), gdouble(y) / gdouble(YSegments), gdouble(z) / gdouble(ZSegments));
							NewPointProxy* const pointproxy = Collection.Insert(point, parameters);
							pointproxy->PushEditing();
							
							points[z][y][x] = point;
						}
				}
		}
		
	// Create paths ...
	// Front face ...
	for(gulong y = 0; y < YSegments; y++)
		{
			for(gulong x = 0; x < XSegments; x++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);
					
					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(x, y, 0));
					pathproxy->PushEditing();

					AttachCubePoint(path, points[0][y+1][x], gdouble(x) / XSegments, 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[0][y+1][x+1], gdouble(x+1) / XSegments, 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[0][y][x+1], gdouble(x+1) / XSegments, 1.0 - (gdouble(y) / YSegments));
					AttachCubePoint(path, points[0][y][x], gdouble(x) / XSegments, 1.0 - (gdouble(y) / YSegments));
				}
		}

	// Back face ...
	for(gulong y = 0; y < YSegments; y++)
		{
			for(gulong x = 0; x < XSegments; x++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);

					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(x, y, 1));
					pathproxy->PushEditing();

					AttachCubePoint(path, points[ZSegments][y+1][x+1], 1.0 - (gdouble(x+1) / XSegments), 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[ZSegments][y+1][x], 1.0 - (gdouble(x) / XSegments), 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[ZSegments][y][x], 1.0 - (gdouble(x) / XSegments), 1.0 - (gdouble(y) / YSegments));
					AttachCubePoint(path, points[ZSegments][y][x+1], 1.0 - (gdouble(x+1) / XSegments), 1.0 - (gdouble(y) / YSegments));
				}
		}

	// Left face ...
	for(gulong y = 0; y < YSegments; y++)
		{
			for(gulong z = 0; z < ZSegments; z++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);

					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(z, y, 2));
					pathproxy->PushEditing();

					AttachCubePoint(path, points[z+1][y+1][0], 1.0 - (gdouble(z+1) / ZSegments), 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[z][y+1][0], 1.0 - (gdouble(z) / ZSegments), 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[z][y][0], 1.0 - (gdouble(z) / ZSegments), 1.0 - (gdouble(y) / YSegments));
					AttachCubePoint(path, points[z+1][y][0], 1.0 - (gdouble(z+1) / ZSegments), 1.0 - (gdouble(y) / YSegments));
				}
		}

	// Right face ...
	for(gulong y = 0; y < YSegments; y++)
		{
			for(gulong z = 0; z < ZSegments; z++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);

					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(z, y, 3));
					pathproxy->PushEditing();

					AttachCubePoint(path, points[z][y+1][XSegments], gdouble(z) / ZSegments, 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[z+1][y+1][XSegments], gdouble(z+1) / ZSegments, 1.0 - (gdouble(y+1) / YSegments));
					AttachCubePoint(path, points[z+1][y][XSegments], gdouble(z+1) / ZSegments, 1.0 - (gdouble(y) / YSegments));
					AttachCubePoint(path, points[z][y][XSegments], gdouble(z) / ZSegments, 1.0 - (gdouble(y) / YSegments));
				}
		}

	// Top face ...
	for(gulong z = 0; z < ZSegments; z++)
		{
			for(gulong x = 0; x < XSegments; x++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);

					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(x, z, 4));
					pathproxy->PushEditing();

					AttachCubePoint(path, points[z+1][YSegments][x], gdouble(x) / XSegments, 1.0 - (gdouble(z+1) / ZSegments));
					AttachCubePoint(path, points[z+1][YSegments][x+1], gdouble(x+1) / XSegments, 1.0 - (gdouble(z+1) / ZSegments));
					AttachCubePoint(path, points[z][YSegments][x+1], gdouble(x+1) / XSegments, 1.0 - (gdouble(z) / ZSegments));
					AttachCubePoint(path, points[z][YSegments][x], gdouble(x) / XSegments, 1.0 - (gdouble(z) / ZSegments));
				}
		}

	// Bottom face ...
	for(gulong z = 0; z < ZSegments; z++)
		{
			for(gulong x = 0; x < XSegments; x++)
				{
					// Create a path ...
					k3dIPath* const path = Mesh->CreatePath();
					g_return_if_fail(path);

					// Insert it into the collection and set it up for further modification ...
					NewPathProxy* const pathproxy = Collection.Insert(path, sdpVector3(x, z, 5));
					pathproxy->PushEditing();

					AttachCubePoint(path, points[z+1][0][x+1], gdouble(x+1) / XSegments, gdouble(z+1) / ZSegments);
					AttachCubePoint(path, points[z+1][0][x], gdouble(x) / XSegments, gdouble(z+1) / ZSegments);
					AttachCubePoint(path, points[z][0][x], gdouble(x) / XSegments, gdouble(z) / ZSegments);
					AttachCubePoint(path, points[z][0][x+1], gdouble(x+1) / XSegments, gdouble(z) / ZSegments);
				}
		}
}

//////////////////////////////////////////////////////////////////////////////////
// EngineMeshProxy
	
EngineMeshProxy::EngineMeshProxy(GeometryEngine& Engine) :
	m_Engine(Engine),
	m_Actual(0),
	m_Original(false)
{
}

EngineMeshProxy::EngineMeshProxy(GeometryEngine& Engine, k3dIMesh* const Mesh) :
	m_Engine(Engine),
	m_Actual(Mesh),
	m_Original(true)
{
	// Sanity checks ...
	g_assert(m_Actual);
}

EngineMeshProxy::~EngineMeshProxy()
{
}

void EngineMeshProxy::PushEditing()
{
	k3d::PushEditing(Actual());
}

void EngineMeshProxy::PopEditing()
{
	k3d::PopEditing(Actual());
}

void EngineMeshProxy::BuildGeometry()
{
}

void EngineMeshProxy::Synchronize()
{
	// Make sure our underlying object exists (if it doesn't already)
	Actual();
	PopEditing();
}

void EngineMeshProxy::UpdateGeometry()
{
}

void EngineMeshProxy::Cancel()
{
}

k3dIMesh* EngineMeshProxy::Actual() const
{
	if(0 == m_Actual)
		{
			k3dIObject* const object = Engine().Document()->Objects()->CreateObject(k3d::classid::Mesh);
			g_assert(object);

			m_Actual = dynamic_cast<k3dIMesh*>(object);
			g_assert(m_Actual);
		}
		
	return m_Actual;
}
			
///////////////////////////////////////////////////////////////////////////////////
// EnginePathProxy

EnginePathProxy::EnginePathProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh) :
	m_Engine(Engine),
	m_Mesh(Mesh),
	m_Actual(0),
	m_Original(false)
{
}
		
EnginePathProxy::EnginePathProxy(GeometryEngine& Engine, k3dIPath* const Path) :
	m_Engine(Engine),
	m_Mesh(Engine.Mesh(Path->Mesh())),
	m_Actual(Path),
	m_Original(true)
{
	// Sanity checks ...
	g_assert(m_Actual);

	// Build our collection of owned-points ...
	k3dIPathPointCollection* const pathpoints = m_Actual->Points();
	for(K3DPATHPOINTHANDLE handle = pathpoints->FirstPathPoint(); handle; handle = pathpoints->NextPathPoint(handle))
		m_Points.push_back(&m_Engine.Point(pathpoints->PathPoint(handle)->Point()));
		
	// Build our collection of owned-edges ...
	const gulong pointcount = m_Points.size();
	for(gulong i = 0; i < pointcount; i++)
		m_Edges.push_back(&m_Engine.Edge(*m_Points[(i+0)%pointcount], *m_Points[(i+1)%pointcount], true));
}

EnginePathProxy::~EnginePathProxy()
{
}
			
void EnginePathProxy::PushEditing()
{
	k3d::PushEditing(Actual());
}

void EnginePathProxy::PopEditing()
{
	k3d::PopEditing(Actual());
}

void EnginePathProxy::BuildGeometry()
{
}

void EnginePathProxy::Synchronize()
{
	// Make sure our underlying object exists (if it doesn't already)
	Actual();
	PopEditing();
}

void EnginePathProxy::UpdateGeometry()
{
}

void EnginePathProxy::Cancel()
{
}

k3dIPath* EnginePathProxy::Actual() const
{
	if(0 == m_Actual)
		{
			Mesh().PushEditing();
			m_Actual = Mesh().Actual()->CreatePath();
			g_assert(m_Actual);
		}
	
	return m_Actual;
}

///////////////////////////////////////////////////////////////////////////////////
// EnginePointProxy

EnginePointProxy::EnginePointProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh) :
	m_Engine(Engine),
	m_Mesh(Mesh),
	m_Actual(0),
	m_Original(false)
{
}
		
EnginePointProxy::EnginePointProxy(GeometryEngine& Engine, k3dIPoint* const Point) :
	m_Engine(Engine),
	m_Mesh(Engine.Mesh(Point->Mesh())),
	m_Actual(Point),
	m_Original(true)
{
	// Sanity checks ...
	g_assert(m_Actual);
}

EnginePointProxy::~EnginePointProxy()
{
}
			
void EnginePointProxy::PushEditing()
{
	k3d::PushEditing(Actual());
}

void EnginePointProxy::PopEditing()
{
	k3d::PopEditing(Actual());
}

void EnginePointProxy::BuildGeometry()
{
}

void EnginePointProxy::Synchronize()
{
	// Make sure our underlying object exists (if it doesn't already)
	Actual();
	PopEditing();
}

void EnginePointProxy::UpdateGeometry()
{
}

void EnginePointProxy::Cancel()
{
}

k3dIPoint* EnginePointProxy::Actual() const
{
	if(0 == m_Actual)
		{
			Mesh().PushEditing();
			m_Actual = Mesh().Actual()->CreatePoint();
			g_assert(m_Actual);
		}
		
	return m_Actual;
}

//////////////////////////////////////////////////////////////////////////////////
// EngineEdgeProxy

EngineEdgeProxy::EngineEdgeProxy(GeometryEngine& Engine, EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original) :
	m_Engine(Engine),
	m_Point1(Point1),
	m_Point2(Point2),
	m_Original(Original)
{
}

EngineEdgeProxy::~EngineEdgeProxy()
{
}

void EngineEdgeProxy::PushEditing()
{
}

void EngineEdgeProxy::PopEditing()
{
}

void EngineEdgeProxy::BuildGeometry()
{
}

void EngineEdgeProxy::UpdateGeometry()
{
}

void EngineEdgeProxy::Synchronize()
{
}

void EngineEdgeProxy::Cancel()
{
}

//////////////////////////////////////////////////////////////////////////////////
// GeometryEngine

GeometryEngine::GeometryEngine(
		k3dIDocument* const Document,
		IEngineMeshProxyFactory* const MeshFactory,
		IEnginePathProxyFactory* const PathFactory,
		IEnginePointProxyFactory* const PointFactory,
		IEngineEdgeProxyFactory* const EdgeFactory) :
	m_Document(Document),
	m_Changes(0),
	m_MeshFactory(MeshFactory),
	m_PathFactory(PathFactory),
	m_PointFactory(PointFactory),
	m_EdgeFactory(EdgeFactory),
	m_AllowNewGeometry(false)
{
	// Sanity checks ...
	g_assert(m_Document);
	g_assert(m_MeshFactory.get());
	g_assert(m_PathFactory.get());
	g_assert(m_PointFactory.get());
	g_assert(m_EdgeFactory.get());
}

GeometryEngine::~GeometryEngine()
{
	Clear();
}

void GeometryEngine::Clear()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), sdp::DeleteObject());
	m_Meshes.clear();
	m_MeshMap.clear();
	
	std::for_each(m_Paths.begin(), m_Paths.end(), sdp::DeleteObject());
	m_Paths.clear();
	m_PathMap.clear();
	
	std::for_each(m_Points.begin(), m_Points.end(), sdp::DeleteObject());
	m_Points.clear();
	m_PointMap.clear();
	
	std::for_each(m_Edges.begin(), m_Edges.end(), sdp::DeleteObject());
	m_Edges.clear();
	m_EdgeMap.clear();
}

void GeometryEngine::PushEditing()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), std::mem_fun(&EngineMeshProxy::PushEditing));
	std::for_each(m_Paths.begin(), m_Paths.end(), std::mem_fun(&EnginePathProxy::PushEditing));
	std::for_each(m_Points.begin(), m_Points.end(), std::mem_fun(&EnginePointProxy::PushEditing));
	std::for_each(m_Edges.begin(), m_Edges.end(), std::mem_fun(&EngineEdgeProxy::PushEditing));
}

void GeometryEngine::PopEditing()
{
	std::for_each(m_Meshes.begin(), m_Meshes.end(), std::mem_fun(&EngineMeshProxy::PopEditing));
	std::for_each(m_Paths.begin(), m_Paths.end(), std::mem_fun(&EnginePathProxy::PopEditing));
	std::for_each(m_Points.begin(), m_Points.end(), std::mem_fun(&EnginePointProxy::PopEditing));
	std::for_each(m_Edges.begin(), m_Edges.end(), std::mem_fun(&EngineEdgeProxy::PopEditing));
}

void GeometryEngine::StartChangeSet()
{
std::cerr << g_pretty_function() << std::endl;

	// Sanity checks ...
	g_return_if_fail(!m_Changes);
	
	m_Changes = k3dStartStateChangeSet(m_Document);
}

void GeometryEngine::BuildGeometry()
{
std::cerr << g_pretty_function() << std::endl;

	PushEditing();

	// We operate on copies of our collections, since they may grow/shrink as we're going ...
	MeshCollection meshes(m_Meshes.begin(), m_Meshes.end());
	PathCollection paths(m_Paths.begin(), m_Paths.end());
	PointCollection points(m_Points.begin(), m_Points.end());
	EdgeCollection edges(m_Edges.begin(), m_Edges.end());

	m_AllowNewGeometry = true;
	std::for_each(meshes.begin(), meshes.end(), std::mem_fun(&EngineMeshProxy::BuildGeometry));
	std::for_each(paths.begin(), paths.end(), std::mem_fun(&EnginePathProxy::BuildGeometry));
	std::for_each(points.begin(), points.end(), std::mem_fun(&EnginePointProxy::BuildGeometry));
	std::for_each(edges.begin(), edges.end(), std::mem_fun(&EngineEdgeProxy::BuildGeometry));
	m_AllowNewGeometry = false;
	
	PopEditing();
}

void GeometryEngine::Synchronize()
{
std::cerr << g_pretty_function() << std::endl;

	// We operate on copies of our collections, since they may grow/shrink as we're going ...
	MeshCollection meshes(m_Meshes.begin(), m_Meshes.end());
	PathCollection paths(m_Paths.begin(), m_Paths.end());
	PointCollection points(m_Points.begin(), m_Points.end());
	EdgeCollection edges(m_Edges.begin(), m_Edges.end());

	m_AllowNewGeometry = true;
	std::for_each(meshes.begin(), meshes.end(), std::mem_fun(&EngineMeshProxy::Synchronize));
	std::for_each(paths.begin(), paths.end(), std::mem_fun(&EnginePathProxy::Synchronize));
	std::for_each(points.begin(), points.end(), std::mem_fun(&EnginePointProxy::Synchronize));
	std::for_each(edges.begin(), edges.end(), std::mem_fun(&EngineEdgeProxy::Synchronize));
	m_AllowNewGeometry = false;
}

void GeometryEngine::FinishChangeSet()
{
std::cerr << g_pretty_function() << std::endl;

	// Sanity checks ...
	g_return_if_fail(m_Changes);

	m_Document->StateRecorder()->SetCurrentChangeSet(0);
}

void GeometryEngine::UpdateGeometry()
{
	PushEditing();

	std::for_each(m_Meshes.begin(), m_Meshes.end(), std::mem_fun(&EngineMeshProxy::UpdateGeometry));
	std::for_each(m_Paths.begin(), m_Paths.end(), std::mem_fun(&EnginePathProxy::UpdateGeometry));
	std::for_each(m_Points.begin(), m_Points.end(), std::mem_fun(&EnginePointProxy::UpdateGeometry));
	std::for_each(m_Edges.begin(), m_Edges.end(), std::mem_fun(&EngineEdgeProxy::UpdateGeometry));
	
	PopEditing();
}

void GeometryEngine::Manage()
{
std::cerr << g_pretty_function() << std::endl;

	sdp::for_each_if(m_Meshes.begin(), m_Meshes.end(), SetManagedState(), std::not1(std::mem_fun(&EngineMeshProxy::Original)));
	sdp::for_each_if(m_Paths.begin(), m_Paths.end(), SetManagedState(), std::not1(std::mem_fun(&EnginePathProxy::Original)));
	sdp::for_each_if(m_Points.begin(), m_Points.end(), SetManagedState(), std::not1(std::mem_fun(&EnginePointProxy::Original)));
	sdp::for_each_if(m_Edges.begin(), m_Edges.end(), SetManagedState(), std::not1(std::mem_fun(&EngineEdgeProxy::Original)));
}

void GeometryEngine::Cancel()
{
std::cerr << g_pretty_function() << std::endl;

	if(m_Changes)
		{
			m_Changes->Undo();
			delete m_Changes;
			m_Changes = 0;
		}

	// Give everyone a chance to cleanup ...
	std::for_each(m_Meshes.begin(), m_Meshes.end(), std::mem_fun(&EngineMeshProxy::Cancel));
	std::for_each(m_Paths.begin(), m_Paths.end(), std::mem_fun(&EnginePathProxy::Cancel));
	std::for_each(m_Points.begin(), m_Points.end(), std::mem_fun(&EnginePointProxy::Cancel));
	std::for_each(m_Edges.begin(), m_Edges.end(), std::mem_fun(&EngineEdgeProxy::Cancel));

	// Get rid of any generated proxies ... we have to jump through some hoops since we have containers of pointers-to-objects
	sdp::for_each_if(m_Meshes.begin(), m_Meshes.end(), sdp::DeleteAndNullObject(), std::not1(std::mem_fun(&EngineMeshProxy::Original)));
	m_Meshes.erase(std::remove(m_Meshes.begin(), m_Meshes.end(), static_cast<EngineMeshProxy*>(0)), m_Meshes.end());
	
	sdp::for_each_if(m_Paths.begin(), m_Paths.end(), sdp::DeleteAndNullObject(), std::not1(std::mem_fun(&EnginePathProxy::Original)));
	m_Paths.erase(std::remove(m_Paths.begin(), m_Paths.end(), static_cast<EnginePathProxy*>(0)), m_Paths.end());
	
	sdp::for_each_if(m_Points.begin(), m_Points.end(), sdp::DeleteAndNullObject(), std::not1(std::mem_fun(&EnginePointProxy::Original)));
	m_Points.erase(std::remove(m_Points.begin(), m_Points.end(), static_cast<EnginePointProxy*>(0)), m_Points.end());

	sdp::for_each_if(m_Edges.begin(), m_Edges.end(), sdp::DeleteAndNullObject(), std::not1(std::mem_fun(&EngineEdgeProxy::Original)));
	m_Edges.erase(std::remove(m_Edges.begin(), m_Edges.end(), static_cast<EngineEdgeProxy*>(0)), m_Edges.end());
}

void GeometryEngine::Finish(SDPCSTRING ChangeSetName)
{
std::cerr << g_pretty_function() << std::endl;

	// Sanity checks ...
	g_return_if_string_fail(ChangeSetName);

	sdp::for_each_if(m_Meshes.begin(), m_Meshes.end(), SetFinishedState(), std::not1(std::mem_fun(&EngineMeshProxy::Original)));
	sdp::for_each_if(m_Paths.begin(), m_Paths.end(), SetFinishedState(), std::not1(std::mem_fun(&EnginePathProxy::Original)));
	sdp::for_each_if(m_Points.begin(), m_Points.end(), SetFinishedState(), std::not1(std::mem_fun(&EnginePointProxy::Original)));
	sdp::for_each_if(m_Edges.begin(), m_Edges.end(), SetFinishedState(), std::not1(std::mem_fun(&EngineEdgeProxy::Original)));

	/// Record state changes for undo/redo purposes
	if(m_Changes)
		{
			m_Document->StateRecorder()->RecordChangeSet(m_Changes, ChangeSetName);
			m_Changes = 0;
		}
}

EngineMeshProxy& GeometryEngine::Mesh()
{
	// Sanity checks ...
	g_assert(m_AllowNewGeometry);
	
	// Create a new proxy and set it up for further modification ...
	EngineMeshProxy* const proxy = m_MeshFactory->CreateMeshProxy(*this);
	m_Meshes.push_back(proxy);
	proxy->PushEditing();
	
	return *proxy;
}

EngineMeshProxy& GeometryEngine::Mesh(k3dIMesh* const Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);
	
	EngineMeshProxy* result = m_MeshMap[Mesh];
	if(!result)
		{
			result = m_MeshFactory->CreateMeshProxy(*this, Mesh);
			m_Meshes.push_back(result);
			m_MeshMap[Mesh] = result;
		}
		
	return *result;
}

EnginePathProxy& GeometryEngine::Path(EngineMeshProxy& Mesh)
{
	// Sanity checks ...
	g_assert(m_AllowNewGeometry);
	
	// Create a new proxy and set it up for further modification ...
	EnginePathProxy* const proxy = m_PathFactory->CreatePathProxy(*this, Mesh);
	m_Paths.push_back(proxy);
	proxy->PushEditing();
	
	return *proxy;
}

EnginePathProxy& GeometryEngine::Path(k3dIPath* const Path)
{
	// Sanity checks ...
	g_assert(Path);
	
	EnginePathProxy* result = m_PathMap[Path];
	if(!result)
		{
			result = m_PathFactory->CreatePathProxy(*this, Path);
			m_Paths.push_back(result);
			m_PathMap[Path] = result;
		}
		
	return *result;
}

EnginePointProxy& GeometryEngine::Point(EngineMeshProxy& Mesh)
{
	// Sanity checks ...
	g_assert(m_AllowNewGeometry);
	
	// Create a new proxy and set it up for further modification ...
	EnginePointProxy* const proxy = m_PointFactory->CreatePointProxy(*this, Mesh);
	m_Points.push_back(proxy);
	proxy->PushEditing();
	
	return *proxy;
}

EnginePointProxy& GeometryEngine::Point(k3dIPoint* const Point)
{
	// Sanity checks ...
	g_assert(Point);
	
	EnginePointProxy* result = m_PointMap[Point];
	if(!result)
		{
			result = m_PointFactory->CreatePointProxy(*this, Point);
			m_Points.push_back(result);
			m_PointMap[Point] = result;
		}
		
	return *result;
}

EngineEdgeProxy& GeometryEngine::Edge(EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original)
{
	EdgeMap::key_type const key = std::make_pair(&Point1 < &Point2 ? &Point1 : &Point2, &Point1 < &Point2 ? &Point2 : &Point1);
	EngineEdgeProxy* result = m_EdgeMap[key];

	if(0 == result)
		{
			result = m_EdgeFactory->CreateEdgeProxy(*this, Point1, Point2, Original);
			m_Edges.push_back(result);
			m_EdgeMap[key] = result;
		}

	return *result;
}

const sdpString GeometryEngine::Description() const
{
	return NewGeometryDescription(
		std::count_if(m_Points.begin(), m_Points.end(), std::not1(std::mem_fun(&EnginePointProxy::Original))),
		std::count_if(m_Paths.begin(), m_Paths.end(), std::not1(std::mem_fun(&EnginePathProxy::Original))),
		std::count_if(m_Meshes.begin(), m_Meshes.end(), std::not1(std::mem_fun(&EngineMeshProxy::Original)))
		);
}

std::ostream& operator<<(std::ostream& Stream, const GeometryEngine& Engine)
{
	Stream << "Meshes: ";
		Stream << std::count_if(Engine.m_Meshes.begin(), Engine.m_Meshes.end(), std::mem_fun(&EngineMeshProxy::Original)) << " (orig) ";
		Stream << std::count_if(Engine.m_Meshes.begin(), Engine.m_Meshes.end(), std::not1(std::mem_fun(&EngineMeshProxy::Original))) << " (new) " << std::endl;
		
	Stream << "Paths: ";
		Stream << std::count_if(Engine.m_Paths.begin(), Engine.m_Paths.end(), std::mem_fun(&EnginePathProxy::Original)) << " (orig) ";
		Stream << std::count_if(Engine.m_Paths.begin(), Engine.m_Paths.end(), std::not1(std::mem_fun(&EnginePathProxy::Original))) << " (new) " << std::endl;

	Stream << "Points: ";
		Stream << std::count_if(Engine.m_Points.begin(), Engine.m_Points.end(), std::mem_fun(&EnginePointProxy::Original)) << " (orig) ";
		Stream << std::count_if(Engine.m_Points.begin(), Engine.m_Points.end(), std::not1(std::mem_fun(&EnginePointProxy::Original))) << " (new) " << std::endl;

	Stream <<"Edges: ";
		Stream << std::count_if(Engine.m_Edges.begin(), Engine.m_Edges.end(), std::mem_fun(&EngineEdgeProxy::Original)) << " (orig) ";
		Stream << std::count_if(Engine.m_Edges.begin(), Engine.m_Edges.end(), std::not1(std::mem_fun(&EngineEdgeProxy::Original))) << " (new) " << std::endl;
	
	return Stream;
}

} // namespace k3d



