#ifndef K3DBUILDGEOMETRY_H
#define K3DBUILDGEOMETRY_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares convenience containers, templates, and functions for creating new geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

#include <k3dselection.h>

#include <k3dieditable.h>
#include <k3dimesh.h>
#include <k3diobject.h>
#include <k3dipath.h>
#include <k3dipoint.h>

#include <memory>

// Forward class declarations
//class k3dIMesh;
//class k3dIPath;
//class k3dIPoint;
class k3dIStateChangeSet;

namespace k3d
{

/// Proxy for a newly-created k3dIMesh object	
class NewMeshProxy :
	public k3dIEditable
{
public:
	NewMeshProxy(k3dIMesh* const Mesh);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	k3dIMesh* operator->() const
	{
		return m_Mesh;
	}
	
	operator k3dIMesh*() const
	{
		return m_Mesh;
	}
	
	k3dIObject* Object() const
	{
		return dynamic_cast<k3dIObject*>(m_Mesh);
	}

private:
	/// Stores a reference to the k3dIMesh this object proxies for
	k3dIMesh* const m_Mesh;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Proxy for a newly-created k3dIPath object
class NewPathProxy :
	public k3dIEditable
{
public:
	NewPathProxy(k3dIPath* const Path, const sdpVector3 Parameters);
	
	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	k3dIPath* operator->() const
	{
		return m_Path;
	}
	
	operator k3dIPath*() const
	{
		return m_Path;
	}
	
	const sdpVector3& Parameters() const
	{
		return m_Parameters;
	}
	
private:
	/// Stores a reference to the k3dIPath this object proxies for
	k3dIPath* const m_Path;	
	/// Stores s,t parameter coordinates for this path
	const sdpVector3 m_Parameters;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Proxy for a newly-created k3dIPoint object
class NewPointProxy :
	public k3dIEditable
{
public:
	NewPointProxy(k3dIPoint* const Point, const sdpVector3 Parameters);
	
	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	k3dIPoint* operator->() const
	{
		return m_Point;
	}
	
	operator k3dIPoint*() const
	{
		return m_Point;
	}
	
	const sdpVector3& Parameters() const
	{
		return m_Parameters;
	}
	
private:
	/// Stores a reference to the k3dIPath this object proxies for
	k3dIPoint* const m_Point;	
	/// Stores s,t parameter coordinates for this point
	const sdpVector3 m_Parameters;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Convenience container for working with newly-created geometry
class NewGeometryCollection :
	public k3dIEditable
{
public:
	NewGeometryCollection();
	~NewGeometryCollection();
	
	/// Clears the contents of the collection
	void Clear();

	/// Begins (and stores) a new k3dIChangeSet to record undo/redo data
	void StartChangeSet(k3dIDocument* const Document);
	/// Finishes recording undo/redo data associated with this collection
	void FinishChangeSet(k3dIDocument* const Document);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	/// Adds a newly-created k3dIMesh object to this container
	NewMeshProxy* Insert(k3dIMesh* const Mesh);
	/// Adds a newly-created k3dIPath object to this container
	NewPathProxy* Insert(k3dIPath* const Path, const sdpVector3 Parameters);
	/// Adds a newly-created k3dIPoint object to this container
	NewPointProxy* Insert(k3dIPoint* const Point, const sdpVector3 Parameters);

	/// Puts all of the objects in this container in a "managed" (not-selectable, not-savable, special-color) state
	void Manage();
	/// Discards all of the objects inserted in this container (and undoes any document changes)
	void Cancel();
	/// Puts all of the objects in this container in a "finished" (selectable, savable, default-color) state, and record the (optional) changeset
	void Finish(k3dIDocument* const Document, SDPCSTRING ChangeSetName);

	typedef std::list<NewMeshProxy*> NewMeshes;
	const NewMeshes& Meshes() const
	{
		return m_Meshes;
	}
	
	typedef std::list<NewPathProxy*> NewPaths;
	const NewPaths& Paths() const
	{
		return m_Paths;
	}
	
	typedef std::list<NewPointProxy*> NewPoints;
	const NewPoints& Points() const
	{
		return m_Points;
	}
	
	/// Returns a human-readable description of the collection
	const sdpString Description() const;

private:
	/// Optional storage for recording changes made to the document to create the new geometry
	k3dIStateChangeSet* m_Changes;

	/// Stores references to newly-created k3dIMesh objects for easier manipulation
	NewMeshes m_Meshes;
	/// Stores references to newly-created k3dIPath objects for easier manipulation
	NewPaths m_Paths;
	/// Stores references to newly-created k3dIPoint objects for easier manipulation
	NewPoints m_Points;
};

/// Creates a new mesh object, naming it and inserting it into a NewGeometryCollection
void CreateNewMesh(k3dIDocument* const Document, SDPCSTRING Name, NewGeometryCollection& Collection);
/// Creates a new path object, automatically creating points parameterized in the range (0 - 1), 0, 0 and inserts everything into a NewGeometryCollection
void CreateNewPath(k3dIMesh* const Mesh, const gulong PointCount, NewGeometryCollection& Collection);
/// Creates new point and path objects, parameterized in the range (0 - 1), 0, 0 and inserts them into a NewGeometryCollection
void CreateNewPointsPaths(k3dIMesh* const Mesh, const gulong PointCount, const gulong PathCount, NewGeometryCollection& Collection);
/// Creates a new collection of paths, topologically connected to form a grid, parameterized in the range (0 - 1), (0 - 1), 0 and inserts them into a NewGeometryCollection
void CreateNewGrid(k3dIMesh* const Mesh, const gulong Rows, const gulong Columns, const bool StitchTop, const bool StitchSide, NewGeometryCollection& Collection);
/// Creates a new collection of paths, topologically connected to form a cube, parameterized in the range (0 - 1), (0 - 1), (0 - 1) and inserts them into a NewGeometryCollection
void CreateNewCube(k3dIMesh* const Mesh, const gulong XSegments, const gulong YSegments, const gulong ZSegments, NewGeometryCollection& Collection);

// Forward class declaration
class GeometryEngine;
class EngineMeshProxy;
class EnginePathProxy;
class EnginePointProxy;
class EngineEdgeProxy;

/// Proxy for a k3dIMesh object (new or existing) within a GeometryEngine
class EngineMeshProxy :
	public k3dIEditable
{
public:
	EngineMeshProxy(GeometryEngine& Engine);
	EngineMeshProxy(GeometryEngine& Engine, k3dIMesh* const Mesh);
	
	virtual ~EngineMeshProxy();

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();
	
	/// Override this in derived classes to add/remove geometry from the parent GeometryEngine
	virtual void BuildGeometry();
	/// Override this in derived classes to modify geometry within the parent GeometryEngine
	virtual void UpdateGeometry();
	/// Override this in derived classes to synchronize the state of the proxy with the document
	virtual void Synchronize();
	/// Override this to free resources during a cancel operation
	virtual void Cancel();
	
	GeometryEngine& Engine() const
	{
		return m_Engine;
	}

	k3dIMesh* Actual() const;
			
	bool Original() const
	{
		return m_Original;
	}

protected:
	typedef std::vector<EngineMeshProxy*> MeshCollection;
	typedef std::vector<EnginePathProxy*> PathCollection;
	typedef std::vector<EnginePointProxy*> PointCollection;
	typedef std::vector<EngineEdgeProxy*> EdgeCollection;

private:	
	/// Stores a reference to the owning GeometryEngine
	GeometryEngine& m_Engine;
	/// Stores a reference to the k3dIMesh this object proxies for
	mutable k3dIMesh* m_Actual;
	/// Set to true iff this k3dIMesh is an original from the document, false if it was created by GeometryEngine::BuildGeometry()
	const bool m_Original;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Proxy for a k3dIPath object (new or existing) within a GeometryEngine
class EnginePathProxy :
	public k3dIEditable
{
public:
	EnginePathProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh);
	EnginePathProxy(GeometryEngine& Engine, k3dIPath* const Path);
	
	virtual ~EnginePathProxy();
	
	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();
	
	/// Override this in derived classes to add/remove geometry from the parent GeometryEngine
	virtual void BuildGeometry();
	/// Override this in derived classes to modify geometry within the parent GeometryEngine
	virtual void UpdateGeometry();
	/// Override this in derived classes to synchronize the state of the proxy with the document
	virtual void Synchronize();
	/// Override this to free resources during a cancel operation
	virtual void Cancel();

	GeometryEngine& Engine() const
	{
		return m_Engine;
	}

	EngineMeshProxy& Mesh() const
	{
		return m_Mesh;
	}

	k3dIPath* Actual() const;
	
	bool Original() const
	{
		return m_Original;
	}

protected:
	typedef std::vector<EngineMeshProxy*> MeshCollection;
	typedef std::vector<EnginePathProxy*> PathCollection;
	typedef std::vector<EnginePointProxy*> PointCollection;
	typedef std::vector<EngineEdgeProxy*> EdgeCollection;
	
	/// Stores the collection of original points associated with this path (iff it's an original path)
	PointCollection m_Points;
	/// Stores the collection of original edges associated with this path (iff it's an original path)
	EdgeCollection m_Edges;

private:
	/// Stores a reference to the owning GeometryEngine
	GeometryEngine& m_Engine;
	/// Stores a reference to the owning EngineMeshProxy
	EngineMeshProxy& m_Mesh;
	/// Stores a reference to the k3dIPath this object proxies for
	mutable k3dIPath* m_Actual;	
	/// Set to true iff this k3dIPath is an original from the document, false if it was created by GeometryEngine::BuildGeometry()
	const bool m_Original;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Proxy for a k3dIPoint object (new or existing) within a GeometryEngine
class EnginePointProxy :
	public k3dIEditable
{
public:
	EnginePointProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh);
	EnginePointProxy(GeometryEngine& Engine, k3dIPoint* const Point);
	
	virtual ~EnginePointProxy();
	
	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();
	
	/// Override this in derived classes to add/remove geometry from the parent GeometryEngine
	virtual void BuildGeometry();
	/// Override this in derived classes to modify geometry within the parent GeometryEngine
	virtual void UpdateGeometry();
	/// Override this in derived classes to synchronize the state of the proxy with the document
	virtual void Synchronize();
	/// Override this to free resources during a cancel operation
	virtual void Cancel();

	GeometryEngine& Engine() const
	{
		return m_Engine;
	}

	EngineMeshProxy& Mesh() const
	{
		return m_Mesh;
	}

	k3dIPoint* Actual() const;
			
	bool Original() const
	{
		return m_Original;
	}

protected:
	typedef std::vector<EngineMeshProxy*> MeshCollection;
	typedef std::vector<EnginePathProxy*> PathCollection;
	typedef std::vector<EnginePointProxy*> PointCollection;
	typedef std::vector<EngineEdgeProxy*> EdgeCollection;

private:	
	/// Stores a reference to the owning GeometryEngine
	GeometryEngine& m_Engine;
	/// Stores a reference to the owning EngineMeshProxy
	EngineMeshProxy& m_Mesh;
	/// Stores a reference to the k3dIPoint this object proxies for
	mutable k3dIPoint* m_Actual;
	/// Set to true iff this k3dIPoint is an original from the document, false if it was created by GeometryEngine::BuildGeometry()
	const bool m_Original;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Proxy for an edge connecting two k3dIPoint objects (new or existing) within a GeometryEngine
class EngineEdgeProxy :
	public k3dIEditable
{
public:
	EngineEdgeProxy(GeometryEngine& Engine, EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original);
	virtual ~EngineEdgeProxy();
	
	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	/// Override this in derived classes to add/remove geometry from the parent GeometryEngine
	virtual void BuildGeometry();
	/// Override this in derived classes to modify geometry within the parent GeometryEngine
	virtual void UpdateGeometry();
	/// Override this in derived classes to synchronize the state of the proxy with the document
	virtual void Synchronize();
	/// Override this to free resources during a cancel operation
	virtual void Cancel();

	GeometryEngine& Engine() const
	{
		return m_Engine;
	}

	EnginePointProxy& Point1() const 
	{
		return m_Point1;
	}
	
	EnginePointProxy& Point2() const 
	{
		return m_Point2;
	}

	bool Original() const
	{
		return m_Original;
	}

/*
	friend bool operator<(const EngineEdgeProxy& LHS, const EngineEdgeProxy& RHS)
	{
		if(&LHS.Point1() != &RHS.Point1())
			return &LHS.Point1() < &RHS.Point1();
			
		return &LHS.Point2() < &RHS.Point2();
	}
	
	friend bool operator==(const EngineEdgeProxy& LHS, const EngineEdgeProxy& RHS)
	{
		return (&LHS.Point1() == &RHS.Point1()) && (&LHS.Point2() == &RHS.Point2());
	}
*/

protected:
	typedef std::vector<EngineMeshProxy*> MeshCollection;
	typedef std::vector<EnginePathProxy*> PathCollection;
	typedef std::vector<EnginePointProxy*> PointCollection;
	typedef std::vector<EngineEdgeProxy*> EdgeCollection;

private:	
	/// Stores a reference to the owning GeometryEngine
	GeometryEngine& m_Engine;
	/// Stores a reference to the first point
	EnginePointProxy& m_Point1;
	/// Stores a reference to the second point
	EnginePointProxy& m_Point2;
	/// Set to true iff this edge is an "original" that should survive cancellations
	const bool m_Original;
	/// Set to true iff this object is in an editing state
	bool m_Editing;
};

/// Abstract interface for an object that can create EngineMeshProxy objects (Factory Design Pattern)
class IEngineMeshProxyFactory
{
public:
	virtual ~IEngineMeshProxyFactory() {}
	
	virtual EngineMeshProxy* CreateMeshProxy(GeometryEngine& Engine) = 0;
	virtual EngineMeshProxy* CreateMeshProxy(GeometryEngine& Engine, k3dIMesh* const Mesh) = 0;
};

/// Abstract interface for an object that can create EnginePathProxy objects (Factory Design Pattern)
class IEnginePathProxyFactory
{
public:
	virtual ~IEnginePathProxyFactory() {}
	
	virtual EnginePathProxy* CreatePathProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh) = 0;
	virtual EnginePathProxy* CreatePathProxy(GeometryEngine& Engine, k3dIPath* const Path) = 0;
};

/// Abstract interface for an object that can create EnginePointProxy objects (Factory Design Pattern)
class IEnginePointProxyFactory
{
public:
	virtual ~IEnginePointProxyFactory() {}
	
	virtual EnginePointProxy* CreatePointProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh) = 0;
	virtual EnginePointProxy* CreatePointProxy(GeometryEngine& Engine, k3dIPoint* const Point) = 0;
};

/// Abstract interface for an object that can create EngineEdgeProxy objects (Factory Design Pattern)
class IEngineEdgeProxyFactory
{
public:
	virtual ~IEngineEdgeProxyFactory() {}
	
	virtual EngineEdgeProxy* CreateEdgeProxy(GeometryEngine& Engine, EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original) = 0;
};
			
/// Default implementation of IEngineMeshProxyFactory
template<class MeshProxyType>
class EngineMeshProxyFactory :
	public IEngineMeshProxyFactory
{
public:
	EngineMeshProxy* CreateMeshProxy(GeometryEngine& Engine)
	{
		return new MeshProxyType(Engine);
	}
	
	EngineMeshProxy* CreateMeshProxy(GeometryEngine& Engine, k3dIMesh* const Mesh)
	{
		return new MeshProxyType(Engine, Mesh);
	}
};

/// Default implementation of IEnginePathProxyFactory
template<class PathProxyType>
class EnginePathProxyFactory :
	public IEnginePathProxyFactory
{
public:
	EnginePathProxy* CreatePathProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh)
	{
		return new PathProxyType(Engine, Mesh);
	}
	
	EnginePathProxy* CreatePathProxy(GeometryEngine& Engine, k3dIPath* const Path)
	{
		return new PathProxyType(Engine, Path);
	}
};

/// Default implementation of IEnginePointProxyFactory
template<class PointProxyType>
class EnginePointProxyFactory :
	public IEnginePointProxyFactory
{
public:
	EnginePointProxy* CreatePointProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh)
	{
		return new PointProxyType(Engine, Mesh);
	}
	
	EnginePointProxy* CreatePointProxy(GeometryEngine& Engine, k3dIPoint* const Point)
	{
		return new PointProxyType(Engine, Point);
	}
};

/// Default implementation of IEngineEdgeProxyFactory
template<class EdgeProxyType>
class EngineEdgeProxyFactory :
	public IEngineEdgeProxyFactory
{
public:
	EngineEdgeProxy* CreateEdgeProxy(GeometryEngine& Engine, EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original)
	{
		return new EdgeProxyType(Engine, Point1, Point2, Original);
	}
};

/// Commonly-used implementation of IEngineMeshProxyFactory that constructs proxies with a reference to a parent object
template<typename ParentType, class MeshProxyType>
class EngineMeshProxyFactory2 :
	public IEngineMeshProxyFactory
{
public:
	EngineMeshProxyFactory2(ParentType& Parent) :
		m_Parent(Parent)
	{
	}

	EngineMeshProxy* CreateMeshProxy(GeometryEngine& Engine)
	{
		return new MeshProxyType(m_Parent, Engine);
	}
	
	EngineMeshProxy* CreateMeshProxy(GeometryEngine& Engine, k3dIMesh* const Mesh)
	{
		return new MeshProxyType(m_Parent, Engine, Mesh);
	}
	
private:
	ParentType& m_Parent;
};

/// Commonly-used implementation of IEnginePathProxyFactory that constructs proxies with a reference to a parent object
template<typename ParentType, class PathProxyType>
class EnginePathProxyFactory2 :
	public IEnginePathProxyFactory
{
public:
	EnginePathProxyFactory2(ParentType& Parent) :
		m_Parent(Parent)
	{
	}

	EnginePathProxy* CreatePathProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh)
	{
		return new PathProxyType(m_Parent, Engine, Mesh);
	}
	
	EnginePathProxy* CreatePathProxy(GeometryEngine& Engine, k3dIPath* const Path)
	{
		return new PathProxyType(m_Parent, Engine, Path);
	}
	
private:
	ParentType& m_Parent;
};

/// Commonly-used implementation of IEnginePointProxyFactory that constructs proxies with a reference to a parent object
template<typename ParentType, class PointProxyType>
class EnginePointProxyFactory2 :
	public IEnginePointProxyFactory
{
public:
	EnginePointProxyFactory2(ParentType& Parent) :
		m_Parent(Parent)
	{
	}

	EnginePointProxy* CreatePointProxy(GeometryEngine& Engine, EngineMeshProxy& Mesh)
	{
		return new PointProxyType(m_Parent, Engine, Mesh);
	}
	
	EnginePointProxy* CreatePointProxy(GeometryEngine& Engine, k3dIPoint* const Point)
	{
		return new PointProxyType(m_Parent, Engine, Point);
	}
	
private:
	ParentType& m_Parent;
};

/// Commonly-used implementation of IEngineEdgeProxyFactory that constructs proxies with a reference to a parent object
template<typename ParentType, class EdgeProxyType>
class EngineEdgeProxyFactory2 :
	public IEngineEdgeProxyFactory
{
public:
	EngineEdgeProxyFactory2(ParentType& Parent) :
		m_Parent(Parent)
	{
	}

	EngineEdgeProxy* CreateEdgeProxy(GeometryEngine& Engine, EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original)
	{
		return new EdgeProxyType(m_Parent, Engine, Point1, Point2, Original);
	}
	
private:
	ParentType& m_Parent;
};

/// Advanced container for creating / modifying  geometry based on an existing selection
class GeometryEngine :
	public k3dIEditable
{
public:
	GeometryEngine(
			k3dIDocument* const Document,
			IEngineMeshProxyFactory* const MeshFactory,
			IEnginePathProxyFactory* const PathFactory,
			IEnginePointProxyFactory* const PointFactory,
			IEngineEdgeProxyFactory* const EdgeFactory
		);
		
	~GeometryEngine();

	/// Discards all objects in this container (so it can be reused)
	void Clear();
	
	/// Returns a reference to the owning document
	k3dIDocument* Document() const
	{
		return m_Document;
	}

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	/// Begins a new change set for recording document modifications
	void StartChangeSet();
	/// Causes contained objects to create new geometry as-needed
	void BuildGeometry();
	/// Synchronizes modifications to the geometry collections with the document
	void Synchronize();
	/// Completes recording of document modifications
	void FinishChangeSet();
	/// Causes contained objects to update existing/new geometry as-needed
	void UpdateGeometry();
	/// Puts all of the new objects in this container in a "managed" (new geometry created, not-selectable, not-savable, special-color) state
	void Manage();
	/// Discards all of the new objects inserted in this container (and undoes any document changes)
	void Cancel();
	/// Puts all of the new objects in this container in a "finished" (selectable, savable, default-color) state, and record the (optional) changeset
	void Finish(SDPCSTRING ChangeSetName);

	friend class Inserter/* :
		public std::iterator<std::output_iterator_tag, void, void, void, void>*/
	{
	public:
		explicit Inserter(GeometryEngine& Engine) :
			m_Engine(Engine)
		{
		}
		
		Inserter& operator=(k3dIMesh* const Mesh)
		{
			m_Engine.Mesh(Mesh);
			return *this;
		}
		
		Inserter& operator=(k3dIPath* const Path)
		{
			m_Engine.Path(Path);
			return *this;
		}
		
		Inserter& operator=(k3dIPoint* const Point)
		{
			m_Engine.Point(Point);
			return *this;
		}
		
		Inserter& operator*()
		{
			return *this;
		}
		
		Inserter& operator++()
		{
			return *this;
		}
		
		Inserter operator++(int)
		{
			return *this;
		}
		
	private:
		GeometryEngine& m_Engine;
	};

	/// Returns a proxy for a new k3dIMesh object
	EngineMeshProxy& Mesh();
	/// Returns a proxy for an existing k3dIMesh object
	EngineMeshProxy& Mesh(k3dIMesh* const Mesh);
	/// Returns a proxy for a new k3dIPath object
	EnginePathProxy& Path(EngineMeshProxy& Mesh);
	/// Returns a proxy for an existing k3dIPath object
	EnginePathProxy& Path(k3dIPath* const Path);
	/// Returns a proxy for a new k3dIPoint object
	EnginePointProxy& Point(EngineMeshProxy& Mesh);
	/// Returns a proxy for an existing k3dIPoint object
	EnginePointProxy& Point(k3dIPoint* const Point);
	/// Returns a proxy object for an edge connecting two EnginePointProxy objects
	EngineEdgeProxy& Edge(EnginePointProxy& Point1, EnginePointProxy& Point2, const bool Original);

	/// Returns a human-readable description of the collection
	const sdpString Description() const;

	friend std::ostream& operator<<(std::ostream& Stream, const GeometryEngine& Engine);

private:
	/// Stores a reference to the owning document
	k3dIDocument* const m_Document;

	/// Optional storage for recording changes made to a document to create new geometry
	k3dIStateChangeSet* m_Changes;

	/// Stores existing/new mesh proxies
	typedef std::vector<EngineMeshProxy*> MeshCollection;
	MeshCollection m_Meshes;
	
	typedef std::map<k3dIMesh*, EngineMeshProxy*> MeshMap;
	MeshMap m_MeshMap;

	/// Stores existing/new path proxies
	typedef std::vector<EnginePathProxy*> PathCollection;
	PathCollection m_Paths;
	
	typedef std::map<k3dIPath*, EnginePathProxy*> PathMap;
	PathMap m_PathMap;

	/// Stores existing/new point proxies
	typedef std::vector<EnginePointProxy*> PointCollection;
	PointCollection m_Points;
	
	typedef std::map<k3dIPoint*, EnginePointProxy*> PointMap;
	PointMap m_PointMap;

	/// Stores existing/new edge proxies
	typedef std::vector<EngineEdgeProxy*> EdgeCollection;
	EdgeCollection m_Edges;
	
	typedef std::map<std::pair<EnginePointProxy*, EnginePointProxy*>, EngineEdgeProxy*> EdgeMap;
	EdgeMap m_EdgeMap;
	
	/// Stores a factory object for creating mesh proxies
	const std::auto_ptr<IEngineMeshProxyFactory> m_MeshFactory;
	/// Stores a factory object for creating path proxies
	const std::auto_ptr<IEnginePathProxyFactory> m_PathFactory;
	/// Stores a factory object for creating point proxies
	const std::auto_ptr<IEnginePointProxyFactory> m_PointFactory;
	/// Stores a factory object for creating edge proxies
	const std::auto_ptr<IEngineEdgeProxyFactory> m_EdgeFactory;

	/// Set to true only during BuildGeometry() and Synchronize() calls
	bool m_AllowNewGeometry;
};

} // namespace k3d

#endif // K3DBUILDGEOMETRY_H

