// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dChannelButton class, which provides a UI for representing objects that support k3dIChannel
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dchannelbutton.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dieditable.h>
#include <k3dieventsource.h>
#include <k3diobject.h>

#include <sdpgtkevents.h>
#include <sdpgtkutility.h>
#include <sdpgtkcolorselector.h>

// We have an unfortunate class with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

///////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dChannelButtonSetColor

const sdpString k3dChannelButtonSetColor::control_close = "close";
const sdpString k3dChannelButtonSetColor::control_colorchanged = "colorchanged";
const sdpString k3dChannelButtonSetColor::control_colorselection = "colorselection";

k3dChannelButtonSetColor::k3dChannelButtonSetColor(sdpIUnknown* const ParentCommandNode, k3dIChannel* const Channel) :
	k3dChannelButtonSetColorBase(ParentCommandNode, "setcolor", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Channel(Channel)
{
	// Sanity checks ...
	g_assert(m_Channel);

	std::istrstream uitemplate( 
		"<gtkml>"
			"<window type=\"toplevel\" show=\"false\">"
				"<vbox homogeneous=\"false\">"
					"<colorselection name=\"colorselection\">"
						"<event signal=\"color-changed\" name=\"colorchanged\"/>"
					"</colorselection>"
					"<hbuttonbox layout=\"end\">"
						"<button name=\"close\">Close"
							"<event signal=\"clicked\" name=\"close\"/>"
						"</button>"
					"</hbuttonbox>"
				"</vbox>"
			"</window>"
		"</gtkml>");

	g_return_if_fail(Load(uitemplate, "k3dChannelButtonSetColor"));

	AddCommand(control_close, control_close, CONTROL_BUTTON);

	ColorSelection(control_colorselection).GetColor(m_LastColor);
	ColorSelection(control_colorselection).SetUpdatePolicy(GTK_UPDATE_DISCONTINUOUS);
	
	// We want to be notified if the channel goes away ...
	k3dIEventSource* channel = dynamic_cast<k3dIEventSource*>(m_Channel);
	g_return_if_fail(channel);
	channel->AddEventObserver(this, k3dIObject::Event_Deleted, "k3dChannelButtonSetColor");
	
	// Set our title ...
	RootWindow().SetTitle("Set [" + sdpString(dynamic_cast<k3dIObject*>(m_Channel)->Name()) + "] Color");
	
	// Make ourselves visible ...
	RootWidget().Show();
}
	
k3dChannelButtonSetColor::~k3dChannelButtonSetColor()
{
	// Cancel event notifications
	k3dIEventSource* channel = dynamic_cast<k3dIEventSource*>(m_Channel);
	g_return_if_fail(channel);
	channel->RemoveEventObserver(this);
}

sdpResult k3dChannelButtonSetColor::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	const sdpString name(Name);
	const sdpString arguments(Arguments);
	
	if(name == control_colorchanged)
		{
			const sdpVector3 color(sdpFromString<sdpVector3>(arguments, sdpVector3()));
			
			ColorSelection(control_colorselection).InteractiveSetColor(color, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
			OnColorChanged(color);
			
			return SDP_S_OK;
		}
	
	return k3dChannelButtonSetColorBase::Command(Name, Arguments, Result);
}
	
sdpResult k3dChannelButtonSetColor::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);
	if(event == k3dIObject::Event_Deleted)
		{
			delete this;
			return SDP_S_OK;
		}
	
	return k3dChannelButtonSetColorBase::ObserveEvent(Source, Event, EventData, ObserverData);
}
	
void k3dChannelButtonSetColor::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_close)
		OnClose();
	else if(Event->Name() == control_colorchanged)
		RawOnColorChanged();
	else
		k3dChannelButtonSetColorBase::OnEvent(Event);
}

void k3dChannelButtonSetColor::OnClose()
{
	// Record this as a user action ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_close, 0);
	k3dChannelButtonSetColorBase::OnClose();
}

void k3dChannelButtonSetColor::RawOnColorChanged()
{
	sdpVector3 newcolor;
	ColorSelection(control_colorselection).GetColor(newcolor);
	
	if(newcolor != m_LastColor)
		{
			m_LastColor = newcolor;
			OnColorChanged(newcolor);
		}
}

void k3dChannelButtonSetColor::OnColorChanged(const sdpVector3 NewColor)
{
	k3dIObject* object = dynamic_cast<k3dIObject*>(m_Channel);
	g_return_if_fail(object);
	
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_colorchanged, sdpToString(NewColor));
	k3dRecordStateChangeSet undo(object->Document(), "Set [" + sdpString(object->Name()) + "] Color");
	
	k3d::Editing editing(m_Channel);
	k3dChannelSetConstantValue(m_Channel, NewColor);
}

//////////////////////////////////////////////////////////////////////
// k3dChannelButton

sdpVector3 k3dChannelButton::m_BackgroundColor(0.8, 0.8, 0.8);
sdpVector3 k3dChannelButton::m_NormalColor(0.0, 0.0, 0.0);

const sdpString k3dChannelButton::control_startvalue = "startvalue";
const sdpString k3dChannelButton::control_endvalue = "endvalue";
const sdpString k3dChannelButton::control_quickedit = "quickedit";
const sdpString k3dChannelButton::control_edit = "edit";

k3dChannelButton::k3dChannelButton(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dChannelButtonBase(CommandNodeParent, CommandNodeName),
	m_Channel(0),
	m_ShowCurve(false)
{
	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<hbox homogeneous=\"false\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"
				"<frame shadowtype=\"out\" expand=\"true\" fill=\"true\" name=\"startvalueframe\">"
					"<entry name=\"startvalue\" charactersize=\"6 1.5\">"
						"<event signal=\"button-press-event\" name=\"startvaluefocusin\" after=\"true\"/>"
						"<event signal=\"focus-in-event\" name=\"startvaluefocusin\" after=\"true\"/>"
						"<event signal=\"focus-out-event\" name=\"startvaluefocusout\"/>"
					"</entry>"
				"</frame>"
				"<frame shadowtype=\"out\" expand=\"true\" fill=\"true\" name=\"endvalueframe\">"
					"<entry name=\"endvalue\" charactersize=\"6 1.5\">"
						"<event signal=\"button-press-event\" name=\"endvaluefocusin\" after=\"true\"/>"
						"<event signal=\"focus-in-event\" name=\"endvaluefocusin\" after=\"true\"/>"
						"<event signal=\"focus-out-event\" name=\"endvaluefocusout\"/>"
					"</entry>"
				"</frame>"
				"<button expand=\"true\" fill=\"true\" name=\"quickedit\">"
					"<event signal=\"clicked\" name=\"quickedit\"/>"
					"<eventbox name=\"value\" size=\"16 16\"/>"
				"</button>"
				"<button name=\"edit\">"
					"<event signal=\"clicked\" name=\"edit\"/>"
					"<arrow direction=\"right\" shadowtype=\"in\"/>"
				"</button>"
			"</hbox>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, "k3dChannelButton"));

	// Make sure the edit buttons can't get the focus (makes tabbing difficult)
	Widget(control_edit).Object()->flags &= ~GTK_CAN_FOCUS;
	
	AddCommand(control_startvalue, control_startvalue, CONTROL_EDITABLE);
	AddCommand(control_endvalue, control_endvalue, CONTROL_EDITABLE);
	AddCommand(control_edit, control_edit, CONTROL_BUTTON);
	AddCommand(control_quickedit, control_quickedit, CONTROL_BUTTON);
}

k3dChannelButton::~k3dChannelButton()
{
//std::cout << "k3dChannelButton::~k3dChannelButton()" << std::endl;

	// Let the channel know we ain't interested ... !
	if(m_Channel)
		{
			k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(m_Channel);
			if(eventserver)
				eventserver->RemoveEventObserver(this);
		}
	
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Disconnect our OpenGL context ...
	if(m_ValueWidget.Initialized())
		m_ValueWidget.Destroy();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();
}

bool k3dChannelButton::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element.Name() == "k3dchannelbutton");

	sdpGtkMarkAttribute(Document, Element, "showcurve");
	sdpxml::ParseAttribute(Element, "showcurve", m_ShowCurve);

	return true;
}

bool k3dChannelButton::Attach(k3dIChannel* Channel)
{
	// Sanity checks ...
	g_assert(Channel);
	g_assert(!m_Channel); // You only get to attach once!

	// Store a reference to our channel ...
	m_Channel = Channel;

	// We want to know when the channel is modified ...
	k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(Channel);
	if(eventserver)
		eventserver->AddEventObserver(this, k3dIChannel::Event_Changed, "k3dChannelButton");
	
	// We force ShowCurve on if it's a color channel ...
	if(k3dIChannel::DATA_COLOR == m_Channel->Datatype())
		m_ShowCurve = true;

	// Update the display ...
	Update();
	
	return true;
}

const sdpString k3dChannelButton::CustomType() const
{
	return sdpString("k3dChannelButton");
}

sdpResult k3dChannelButton::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	sdpString name(Name);
	
	if(name == control_startvalue)
		{
			g_return_val_if_fail(InteractiveSetText(control_startvalue, Arguments, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true), SDP_E_FAIL);
			OnStartValueFocusOut();
			return SDP_S_OK;
		}
	else if(name == control_endvalue)
		{
			g_return_val_if_fail(InteractiveSetText(control_endvalue, Arguments, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true), SDP_E_FAIL);
			OnEndValueFocusOut();
			return SDP_S_OK;
		}

	return k3dChannelButtonBase::Command(Name, Arguments, Result);
}

void k3dChannelButton::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "configurevalue" || Event->Name() == "exposevalue")
		OnValueDraw();
	else if(Event->Name() == "startvaluefocusin")
		OnStartValueFocusIn();
	else if(Event->Name() == "startvaluefocusout")
		OnStartValueFocusOut();
	else if(Event->Name() == "endvaluefocusin")
		OnEndValueFocusIn();
	else if(Event->Name() == "endvaluefocusout")
		OnEndValueFocusOut();
	else if(Event->Name() == control_edit)
		OnEdit();
	else if(Event->Name() == control_quickedit)
		OnQuickEdit();
	else if(Event->Name() == "destroy")
		OnDestroy();
	else
		k3dChannelButtonBase::OnEvent(Event);
}

void k3dChannelButton::OnDestroy()
{
//std::cout << "k3dChannelButton::OnDestroy()" << std::endl;

	DisconnectAllEvents();

	if(m_ValueWidget.Initialized())
		m_ValueWidget.Destroy();

	Clear();
}

void k3dChannelButton::OnValueDraw()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	m_ValueWidget.Begin();
	
	// Get the widget sizes ...
	gulong width = m_ValueWidget.Width();
	gulong height = m_ValueWidget.Height();
	glViewport(0, 0, width, height);

	if(k3dIChannel::DATA_COLOR == m_Channel->Datatype())
		{
			// Setup the viewing projection ...
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0, 1, -1, 1, -1, 1);
	
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			// Setup drawing options ...
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_LIGHTING);
			glShadeModel(GL_FLAT);

			// Make it happen ...
			glBegin(GL_QUADS);
	
			sdpVector3 color;
			gdouble error = 0.5 / gdouble(width);
			gulong granularity = 1;

			gdouble x2;
			gdouble x1 = 0.0;		
			for(gulong i = 0; i < width+granularity; i += granularity)
				{
					x2 = gdouble(i) / gdouble(width);		
		
					m_Channel->Value3(x2, error, color);

//std::cout << x2 << " " << error << " " << color << std::endl;

					glColor3dv(color);
					glVertex2d(x1, -1);
					glVertex2d(x1, 1);
					glVertex2d(x2, 1);
					glVertex2d(x2, -1);
			
					x1 = x2;
				}

			glEnd();
		}
	else
		{
			// Clear the background ...
			glClearColor(m_BackgroundColor[0], m_BackgroundColor[1], m_BackgroundColor[2], 0.0);
			glClear(GL_COLOR_BUFFER_BIT);
			
			// Get the range of node positions from the channel ...
			gdouble min, max;
			m_Channel->CurveRange(&min, &max);
			
			// Setup the viewing projection ...
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0, 1, min, max, -1, 1);
	
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			// Setup drawing options ...
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_LIGHTING);
			glShadeModel(GL_FLAT);
			glColor3dv(m_NormalColor);	

			// Make it happen ...
			glBegin(GL_LINE_STRIP);

			gdouble x;
			gdouble value;
			gdouble error = 0.5 / gdouble(width);
			gulong granularity = 1;

			for(gulong i = 0; i < width+granularity; i += granularity)
				{
					x = gdouble(i) / gdouble(width);		
					value = m_Channel->Value1(x, error);

					glVertex2d(x, value);
				}

			glEnd();
		}
	
	m_ValueWidget.SwapBuffers();
	m_ValueWidget.End();
}

void k3dChannelButton::OnStartValueFocusIn()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	// Make sure everything's selected ...
	Editable(control_startvalue).SelectRegion(0, -1);
}

void k3dChannelButton::OnStartValueFocusOut()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	k3dRecordStateChangeSet undo(dynamic_cast<k3dIObject*>(m_Channel)->Document(), "Set first value " + Editable(control_startvalue).GetText());
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_startvalue, Editable(control_startvalue).GetText());

	// Get the new value from the control ...
	double newvalue = k3dConvertStringDouble(m_Channel->Datatype(), Editable(control_startvalue).GetText(0, -1));
	
	k3dIBezierChannel* const bezierchannel = dynamic_cast<k3dIBezierChannel*>(m_Channel);
	g_return_if_fail(bezierchannel);
	
	bezierchannel->Editable()->PushEditing();
	
	// Set values for the initial two nodes ...	
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(bezierchannel, nodes);
	
	for(gulong i = 0; (i < 2) && (i < nodes.size()); i++)
		{
			sdpVector2 position;
			nodes[i]->Position(position);
			nodes[i]->SetPosition(sdpVector2(position[0], newvalue));
		}
		
	bezierchannel->Editable()->PopEditing();	
		
	k3dRedrawAll(dynamic_cast<k3dIObject*>(m_Channel)->Document(), false);
}

void k3dChannelButton::OnEndValueFocusIn()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	// Make sure everything's selected ...
	Editable(control_endvalue).SelectRegion(0, -1);
}

void k3dChannelButton::OnEndValueFocusOut()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	k3dRecordStateChangeSet undo(dynamic_cast<k3dIObject*>(m_Channel)->Document(), "Set last value " + Editable(control_endvalue).GetText());
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_endvalue, Editable(control_endvalue).GetText());

	// Get the new value from the control ...
	double newvalue = k3dConvertStringDouble(m_Channel->Datatype(), Editable(control_endvalue).GetText(0, -1));

	k3dIBezierChannel* const bezierchannel = dynamic_cast<k3dIBezierChannel*>(m_Channel);
	g_return_if_fail(bezierchannel);
	
	bezierchannel->Editable()->PushEditing();

	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(m_Channel, nodes);
	
	for(gulong i = 0; i < 2; i++)
		{
			sdpVector2 position;
			nodes[nodes.size() - i - 1]->Position(position);
			nodes[nodes.size() - i - 1]->SetPosition(sdpVector2(position[0], newvalue));
		}

	bezierchannel->Editable()->PopEditing();

	k3dRedrawAll(dynamic_cast<k3dIObject*>(m_Channel)->Document(), false);
}

void k3dChannelButton::OnEdit()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	k3dIObject* object = dynamic_cast<k3dIObject*>(m_Channel);
	g_return_if_fail(object);

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_edit, 0);

	object->EditObject();
}

void k3dChannelButton::OnQuickEdit()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	k3dIObject* object = dynamic_cast<k3dIObject*>(m_Channel);
	g_return_if_fail(object);

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_quickedit, 0);

	switch(m_Channel->Datatype())
		{
			case k3dIChannel::DATA_SCALAR:
			case k3dIChannel::DATA_TIME:
			case k3dIChannel::DATA_ANGLE:
			case k3dIChannel::DATA_DISTANCE:
			case k3dIChannel::DATA_AREA:
			case k3dIChannel::DATA_VOLUME:
			case k3dIChannel::DATA_MASS:
			case k3dIChannel::DATA_FORCE:
			case k3dIChannel::DATA_PRESSURE:
				{
					// Tell the channel to open its UI ...
					object->EditObject();
					break;
				}

			case k3dIChannel::DATA_COLOR:
				{
					new k3dChannelButtonSetColor(dynamic_cast<k3dICommandNode*>(m_Channel), m_Channel);
					break;
				}

			case k3dIChannel::DATA_ROTATIONXYZ:
			case k3dIChannel::DATA_SCALEXYZ:
				g_return_if_fail(0);
		}
}

/// Counts the number of nodes in a Bezier channel
class k3dChannelButtonNodeCount : public k3dIBezierChannelNodeVisitor
{
public:
	k3dChannelButtonNodeCount() :
		m_Count(0)
	{
	}
	
	sdpResult VisitBezierChannelNode(k3dIBezierChannelNode* const Node)
	{
		m_Count++;
		return SDP_S_OK;
	}

	const gulong Count() const
	{
		return m_Count;
	}

private:
	gulong m_Count;
};

void k3dChannelButton::Update()
{
	// Make sure we've got a channel, first!
	g_return_if_fail(m_Channel);

	// Control what's visible and what isn't ...
	if(m_ShowCurve)
		{
			Widget("startvalueframe").Hide();
			Widget("endvalueframe").Hide();
			Widget("quickedit").Show();
		}
	else
		{
			Widget("startvalueframe").Show();
			
			k3dIBezierChannel* channel = dynamic_cast<k3dIBezierChannel*>(m_Channel);
			if(channel)
				{
					k3dChannelButtonNodeCount count;
					channel->VisitNodes(&count);
					
					if(count.Count() > 1)
						Widget("endvalueframe").Show();
					else
						Widget("endvalueframe").Hide();
				}
			else
				{
					Widget("endvalueframe").Show();
				}
				
			Widget("quickedit").Hide();
		}

	// Attach the OpenGL drawing area for previewing the shape of the channel ...
	if(!m_ValueWidget.Attached())
		{
			sdpGtkContainer container(Container("value"));
			if(!m_ValueWidget.Create(container, true, true, 8, 8, 8, 0))
				if(!m_ValueWidget.Create(container, true, true, 5, 5, 5, 0))
					if(!m_ValueWidget.Create(container, false, false, 4, 4, 4, 0))
						std::cerr << "k3dChannelButton::Update(): Could not find useable OpenGL visual" << std::endl;

			if(m_ValueWidget.Initialized())
				{	
					MapEvent("configure-event", "configurevalue", false, m_ValueWidget, true);
					MapEvent("expose-event", "exposevalue", false, m_ValueWidget, true);
				}
		}

	// If we're showing the curve on the face of the button, force an update ...
	if(m_ShowCurve)
		{
			Widget("value").QueueDraw();
			return;
		}

	// Otherwise, display the new value ...
	switch(m_Channel->Datatype())
		{
			case k3dIChannel::DATA_SCALAR:
			case k3dIChannel::DATA_TIME:
			case k3dIChannel::DATA_ANGLE:
			case k3dIChannel::DATA_DISTANCE:
			case k3dIChannel::DATA_AREA:
			case k3dIChannel::DATA_VOLUME:
			case k3dIChannel::DATA_MASS:
			case k3dIChannel::DATA_FORCE:
			case k3dIChannel::DATA_PRESSURE:
				{
					Entry(control_startvalue).SetText(k3dConvertDoubleString(m_Channel->Datatype(), m_Channel->Value1(0.0, 0.1), 3));
					Entry(control_endvalue).SetText(k3dConvertDoubleString(m_Channel->Datatype(), m_Channel->Value1(1.0, 0.1), 3));
					break;
				}

			case k3dIChannel::DATA_COLOR:
			case k3dIChannel::DATA_ROTATIONXYZ:
			case k3dIChannel::DATA_SCALEXYZ:
				g_return_if_fail(0);	// Should never happen!
		}
}

sdpResult k3dChannelButton::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
//std::cout << "k3dChannelButton::ObserveEvent(" << sdpString(Event) << ", " << sdpString(EventData) << ", " << sdpString(ClientData) << ")" << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIChannel::Event_Changed)
		{
			Update();
			return SDP_S_OK;
		}

	return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
}

