// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dCheckButton class, which provides a MVC UI for boolean data sources
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dcheckbutton.h>
#include <k3dstaterecorder.h>

#include <k3dicommandrecorder.h>
#include <k3distaterecorder.h>

#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

///////////////////////////////////////////////////////////////////////////
// k3dCheckButtonVariableDataSource

k3dCheckButtonVariableDataSource::k3dCheckButtonVariableDataSource(k3dVariable<bool>& Source) :
	m_Source(Source)
{
}

k3dIEventSource* k3dCheckButtonVariableDataSource::EventSource()
{
	return dynamic_cast<k3dIEventSource*>(static_cast<k3dIVariable*>(&m_Source));
}

bool k3dCheckButtonVariableDataSource::Value()
{
	return m_Source.Value();
}

void k3dCheckButtonVariableDataSource::PushEditing()
{
	m_Source.PushEditing();
}

void k3dCheckButtonVariableDataSource::SetValue(const bool Value)
{
	m_Source.SetValue(Value);
}

void k3dCheckButtonVariableDataSource::PopEditing()
{
	m_Source.PopEditing();
}

//////////////////////////////////////////////////////////////////////
// k3dCheckButton

const sdpString k3dCheckButton::control_toggle = "toggle";

k3dCheckButton::k3dCheckButton(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dCheckButtonBase(CommandNodeParent, CommandNodeName),
	m_DataSource(0)
{
	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<checkbutton name=\"toggle\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"
				"<event signal=\"toggled\" name=\"toggle\"/>"
			"</checkbutton>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, ""));

	RootWidget().Show();

	AddCommand(control_toggle, control_toggle, CONTROL_TOGGLE_BUTTON);
}

k3dCheckButton::~k3dCheckButton()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();

	// Cancel any event notification from the data source
	if(m_DataSource)
		{
			k3dIEventSource* datasource = m_DataSource->EventSource();
			if(datasource)
				datasource->RemoveEventObserver(this);
		}

	delete m_DataSource;
}

const sdpString k3dCheckButton::CustomType() const
{
	return sdpString("k3dCheckButton");
}

sdpResult k3dCheckButton::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);
	if(event == k3dIVariable::Event_Changed)
		Update();
	else
		k3dCheckButtonBase::ObserveEvent(Source, Event, EventData, ObserverData);

	return SDP_S_OK;
}

bool k3dCheckButton::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element.Name() == "k3dcheckbutton");

	const sdpString text = sdpGtkInternationalText(Document, Element);
	if(text.size())
		{
			sdpGtkLabel label;
			label.Create(text);
			label.SetAlignment(0.0, 0.5);
			label.Show();

			Container(control_toggle).Attach(label);
		}

	return true;
}

bool k3dCheckButton::Attach(k3dCheckButtonDataSource* const DataSource, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	// Sanity checks ...
	g_return_val_if_fail(DataSource, false);

	g_return_val_if_fail(k3dCheckButtonBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);

	// Store references ...
	m_DataSource = DataSource;

	// Update the display ...
	Update();

	// We want to be notified if the data source changes ...
	k3dIEventSource* const datasource = m_DataSource->EventSource();
	g_return_val_if_fail(datasource, false);

	datasource->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dCheckButton");

	return true;
}

bool k3dCheckButton::Attach(k3dVariable<bool>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	return Attach(new k3dCheckButtonVariableDataSource(Variable), CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction);
}

void k3dCheckButton::Update()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Display the current value ...
	ToggleButton(control_toggle).SetState(m_DataSource->Value());
}

void k3dCheckButton::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_toggle)
		OnToggle();
	else if(Event->Name() == "destroy")
		OnDestroy();
	else
		k3dCheckButtonBase::OnEvent(Event);
}

void k3dCheckButton::OnDestroy()
{
	DisconnectAllEvents();
	Clear();
}

void k3dCheckButton::OnToggle()
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Get the control value ...
	bool newvalue = ToggleButton(control_toggle).GetState();

	// If the value hasn't changed, we're done ...
	if(newvalue == m_DataSource->Value())
		return;

	// Record the command for posterity (tutorials) ...
	if(m_CommandRecorder)
		if(SDP_S_OK == m_CommandRecorder->IsRecording())
			m_CommandRecorder->RecordCommand(this, k3dICommandNode::USERINTERFACE, control_toggle, 0);

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		m_StateRecorder->SetCurrentChangeSet(new k3dStateChangeSet());

	// Update everything with the new value ...
	PushEditing();
	SetValue(newvalue);
	PopEditing();

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		{
			m_StateRecorder->RecordChangeSet(m_StateRecorder->ChangeSet(), newvalue ? m_StateChangeName + " On" : m_StateChangeName + " Off");
			m_StateRecorder->SetCurrentChangeSet(0);
		}
}

void k3dCheckButton::SetValue(const bool Value)
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Let the datasource know that we're going to be making changes ...
	m_DataSource->PushEditing();

	// Update storage with the new value ...
	m_DataSource->SetValue(Value);

	// Let the data source know we're done ...
	m_DataSource->PopEditing();

	// Notify any interested parties that things changed ...
	if(m_ChangeAction)
		m_ChangeAction->DoAction();
}

void k3dCheckButton::PushEditing()
{
	// Notify interested parties ...
	if(m_PushEditingAction)
		m_PushEditingAction->DoAction();

}

void k3dCheckButton::PopEditing()
{
	// Notify interested parties ...
	if(m_PopEditingAction)
		m_PopEditingAction->DoAction();
}


