#ifndef K3DCHECKBUTTON_H
#define K3DCHECKBUTTON_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dCheckButton class, which provides a standard MVC UI for boolean data that supports tutorials, undo/redo, etc.
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcontrol.h>
#include <k3dvariable.h>
#include <sdpgtkloopevents.h>

// Forward class declarations
class k3dICommandRecorder;
class k3dIStateRecorder;
class sdpGtkIObjectContainer;
class sdpxml::Element;

///////////////////////////////////////////////////////////////////////////
// k3dCheckButtonDataSource

/// Abstract interface for an object that can act as a data source for a k3dCheckButton (i.e. the "model" in model-view-controller)
class k3dCheckButtonDataSource :
	public k3dIEditable
{
public:
	virtual ~k3dCheckButtonDataSource() {}

	virtual k3dIEventSource* EventSource() = 0;
	virtual bool Value() = 0;
	virtual void SetValue(const bool Value) = 0;
};

///////////////////////////////////////////////////////////////////////////
// k3dCheckButtonVariableDataSource

/// An implementation of k3dCheckButtonDataSource for use with k3dVariable<bool>
class k3dCheckButtonVariableDataSource :
	public k3dCheckButtonDataSource
{
public:
	k3dCheckButtonVariableDataSource(k3dVariable<bool>& Source);

	// k3dCheckButtonDataSource implementation
	k3dIEventSource* EventSource();
	bool Value();
	void SetValue(const bool Value);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

private:
	k3dVariable<bool>& m_Source;
};

///////////////////////////////////////////////////////////////////////////
// k3dCheckButton

typedef k3dControl k3dCheckButtonBase;

/// Provides a UI for manipulating boolean quantities (i.e. the view and the controller from model-view-controller)
class k3dCheckButton :
	public k3dCheckButtonBase
{
public:
	k3dCheckButton(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName);
	~k3dCheckButton();

	/// Called by the framework when instantiating the button from a GTKML (XML) document
	bool Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element);

	/// Attaches the button to the object it fronts for
	bool Attach(k3dCheckButtonDataSource* const Source, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);
	bool Attach(k3dVariable<bool>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);

	const sdpString CustomType() const;

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

protected:
	/// Called to handle SDPGTK events
	void OnEvent(sdpGtkEvent* Event);

	/// Called when the GTK+ widgets are about to disappear
	void OnDestroy();
	/// Called when the user toggles the button
	void OnToggle();

	/// Called prior to making one-or-more changes to the underlying data source
	void PushEditing();

	/// Called when the the button needs to be updated
	void Update();

	/// Called to set the contents of the button
	void SetValue(const bool Value);

	/// Called once all changes to the underlying data source are complete
	void PopEditing();

	/// Storeas a reference to the underlying data object
	k3dCheckButtonDataSource* m_DataSource;

	static const sdpString control_toggle;
};

#endif // K3DCHECKBUTTON_H
