// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dChooser class, which provides a MVC UI for boolean data sources
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dicommandrecorder.h>
#include <k3distaterecorder.h>

#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

//////////////////////////////////////////////////////////////////////
// k3dChooser

const sdpString k3dChooser::control_menu = "menu";
const sdpString k3dChooser::control_value = "value";
const sdpString k3dChooser::control_valuetoggled = "toggled";

k3dChooser::k3dChooser(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dChooserBase(CommandNodeParent, CommandNodeName),
	m_DataSource(0)
{
}

k3dChooser::~k3dChooser()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();

	// Cancel any event notification from the data source
	if(m_DataSource)
		{
			k3dIEventSource* datasource = m_DataSource->EventSource();
			if(datasource)
				datasource->RemoveEventObserver(this);
		}

	delete m_DataSource;
}

const sdpString k3dChooser::CustomType() const
{
	return sdpString("k3dChooser");
}

extern gint gtk_option_menu_button_press (GtkWidget* widget, GdkEventButton *event);

sdpResult k3dChooser::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	g_return_val_if_fail(m_DataSource, SDP_E_FAIL);

	sdpString name(Name);
	sdpString arguments(Arguments);

	if(name == control_value)
		{
			// Convert the arguments string into the desired value ...
			int value;
			m_DataSource->ConvertFromString(arguments, value);

			// Find the matching choice record ...
			ChoiceIterator choice;
			for(choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
				if(choice->Value() == value)
					break;
			g_return_val_if_fail(choice != m_Choices.end(), SDP_E_FAIL);

			// Figure out what to do, depending on our UI type ...
			switch(m_Type)
				{
					case OPTIONMENU:
						InteractiveMoveTo(control_menu, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);

						GdkEventButton event;
						memset(&event, 0, sizeof(event));
						event.type = GDK_BUTTON_PRESS;
						event.button = 1;
						event.time = GDK_CURRENT_TIME;
						gtk_widget_event(Widget(control_menu), reinterpret_cast<GdkEvent*>(&event));

						MenuItem(sdpToString(value)).InteractiveWarpPointer(k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);
						MenuItem(sdpToString(value)).InteractiveActivate();

						gtk_menu_popdown(Menu("actualmenu"));

						break;
					case RADIOBUTTONS:
						InteractiveToggleButton(sdpToString(value), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
						break;
					default:
						g_return_val_if_fail(0, SDP_E_FAIL);
				}

			return SDP_S_OK;
		}

	return k3dChooserBase::Command(Name, Arguments, Result);
}

sdpResult k3dChooser::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);
	if(event == k3dIVariable::Event_Changed)
		Update();
	else
		k3dChooserBase::ObserveEvent(Source, Event, EventData, ObserverData);

	return SDP_S_OK;
}

bool k3dChooser::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element.Name() == "k3dchooser");

	// Figure out which type of chooser to create ...
	const sdpString typetext = sdpxml::GetAttribute<sdpString>(Element, "type", "");
	sdpGtkMarkAttribute(Document, Element, "type");

	if(typetext == "optionmenu")
		m_Type = OPTIONMENU;
	else if(typetext == "radiobuttons")
		m_Type = RADIOBUTTONS;
	else
		{
			std::cerr << g_pretty_function() << ": k3dchooser type not specified at " << sdpxml::FileReference(Document, Element) << std::endl;
			m_Type = OPTIONMENU;
		}

	// Populate our list of choices ...
	for(sdpxml::ElementCollection::iterator element = Element.Children().begin(); element != Element.Children().end(); element++)
		{
			if(element->Name() != "k3dchoice")
				{
					std::cerr << g_pretty_function() << ": unknown element <" << element->Name() << "> at " << sdpxml::FileReference(Document, *element) << " will be ignored" << std::endl;
					continue;
				}

			const gint value = sdpxml::GetAttribute(*element, "value", 0);
			sdpGtkMarkAttribute(Document, *element, "value");

			m_Choices.push_back(Choice(sdpGtkInternationalText(Document, *element), value));
		}

	// Create and load our UI template ...
	std::strstream uitemplate;

	uitemplate <<
		"<gtkml>"
			"<eventbox>"
				"<event signal=\"destroy\" name=\"destroy\"/>";

	switch(m_Type)
		{
			case OPTIONMENU:
				{
					uitemplate << "<optionmenu name=\"" << control_menu << "\"><menu name=\"actualmenu\">";
					for(ChoiceIterator choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
						{
							uitemplate << "<menuitem name=\"" << choice->Value() << "\">" << choice->Name();
							uitemplate << "<event signal=\"activate\" name=\"" << control_value << " " << choice->Value() << "\"/>";
							uitemplate << "</menuitem>";
						}
					uitemplate << "</menu></optionmenu>";
				}
				break;
			case RADIOBUTTONS:
				{
					uitemplate << "<vbox homogeneous=\"false\">";
					for(ChoiceIterator choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
						{
							uitemplate << "<radiobutton name=\"" << choice->Value() << "\" group=\"group\">" << choice->Name();
							uitemplate << "<event signal=\"toggled\" name=\"" << control_valuetoggled << " " << choice->Value() << "\"/>";
							uitemplate << "</radiobutton>";
						}
					uitemplate << "</vbox>";
				}
				break;
			default:
				g_return_val_if_fail(0, false);
		}

	uitemplate <<
			"</eventbox>"
		"</gtkml>\n";

	g_return_val_if_fail(Load(uitemplate, ""), false);

	RootWidget().Show();

	return true;
}

bool k3dChooser::Attach(k3dChooserDataSource* const DataSource, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	// Sanity checks ...
	g_return_val_if_fail(DataSource, false);

	g_return_val_if_fail(k3dChooserBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);

	// Store references ...
	m_DataSource = DataSource;

	// Update the display ...
	Update();

	// We want to be notified if the data source changes ...
	k3dIEventSource* datasource = m_DataSource->EventSource();
	g_return_val_if_fail(datasource, false);
	datasource->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dChooser");

	return true;
}

void k3dChooser::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(0 == Event->Name().find(control_value))
		OnValue(Event);
	else if(0 == Event->Name().find(control_valuetoggled))
		OnValueToggled(Event);
	else if(Event->Name() == "destroy")
		OnDestroy();
	else
		k3dChooserBase::OnEvent(Event);
}

void k3dChooser::OnDestroy()
{
	DisconnectAllEvents();
	Clear();
}

void k3dChooser::OnValue(sdpGtkEvent* Event)
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Get the new value ...
	const int newvalue = atoi(Event->Name().right(Event->Name().size() - control_value.size()));

	// If the value hasn't changed, we're done ...
	if(newvalue == m_DataSource->Value())
		return;

	// Convert the new value to text ...
	sdpString textvalue;
	m_DataSource->ConvertToString(newvalue, textvalue);

	// Find the matching choice record for this value ...
	ChoiceIterator choice;
	for(choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
		if(choice->Value() == newvalue)
			break;
	g_return_if_fail(choice != m_Choices.end());

	// Record the command for posterity (tutorials) ...
	if(m_CommandRecorder)
		if(SDP_S_OK == m_CommandRecorder->IsRecording())
			m_CommandRecorder->RecordCommand(this, k3dICommandNode::USERINTERFACE, control_value, textvalue);

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		m_StateRecorder->SetCurrentChangeSet(new k3dStateChangeSet());

	// Update everything with the new value ...
	PushEditing();
	SetValue(newvalue);
	PopEditing();

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		{
			m_StateRecorder->RecordChangeSet(m_StateRecorder->ChangeSet(), m_StateChangeName + " \"" + choice->Name() + "\"");
			m_StateRecorder->SetCurrentChangeSet(0);
		}
}

void k3dChooser::OnValueToggled(sdpGtkEvent* Event)
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Make sure we're picking-up the button that was toggled "on" ... not the one toggled "off" ...
	const int newvalue = atoi(Event->Name().right(Event->Name().size() - control_valuetoggled.size()));
	if(!ToggleButton(sdpToString(newvalue)).GetState())
		return;

	// If the value hasn't changed, we're done ...
	if(newvalue == m_DataSource->Value())
		return;

	// Convert the new value to text ...
	sdpString textvalue;
	m_DataSource->ConvertToString(newvalue, textvalue);

	// Find the matching choice record for this value ...
	ChoiceIterator choice;
	for(choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
		if(choice->Value() == newvalue)
			break;
	g_return_if_fail(choice != m_Choices.end());

	// Record the command for posterity (tutorials) ...
	if(m_CommandRecorder)
		if(SDP_S_OK == m_CommandRecorder->IsRecording())
			m_CommandRecorder->RecordCommand(this, k3dICommandNode::USERINTERFACE, control_value, textvalue);

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		m_StateRecorder->SetCurrentChangeSet(new k3dStateChangeSet());

	// Update everything with the new value ...
	PushEditing();
	SetValue(newvalue);
	PopEditing();

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		{
			m_StateRecorder->RecordChangeSet(m_StateRecorder->ChangeSet(), m_StateChangeName + " \"" + choice->Name() + "\"");
			m_StateRecorder->SetCurrentChangeSet(0);
		}
}

void k3dChooser::Update()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Get the current value ...
	const int currentvalue = m_DataSource->Value();

	// Block events for a moment ...
	BlockAllEvents();

	// Display the current value ...
	switch(m_Type)
		{
			case OPTIONMENU:
				{
					for(ChoiceIterator choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
						if(choice->Value() == currentvalue)
							{
								OptionMenu(control_menu).SetHistory(choice - m_Choices.begin());
								break;
							}
				}
				break;
			case RADIOBUTTONS:
				{
					for(ChoiceIterator choice = m_Choices.begin(); choice != m_Choices.end(); choice++)
						ToggleButton(sdpToString(choice->Value())).SetState(choice->Value() == currentvalue);
				}
				break;
			default:
				g_return_if_fail(0);
		}

	// Restore events ...
	UnblockAllEvents();
}

void k3dChooser::SetValue(const int Value)
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Update storage with the new value ...
	m_DataSource->SetValue(Value);

	// Notify any interested parties that things changed ...
	if(m_ChangeAction)
		m_ChangeAction->DoAction();
}

void k3dChooser::PushEditing()
{
	// Notify interested parties ...
	if(m_PushEditingAction)
		m_PushEditingAction->DoAction();
}

void k3dChooser::PopEditing()
{
	// Notify interested parties ...
	if(m_PopEditingAction)
		m_PopEditingAction->DoAction();
}


