#ifndef K3DCHOOSER_H
#define K3DCHOOSER_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dChooser class, which provides a standard MVC UI for boolean data that supports tutorials, undo/redo, etc.
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcontrol.h>
#include <k3dvariable.h>
#include <sdpgtkloopevents.h>

// Forward class declarations
class k3dICommandRecorder;
class k3dIStateRecorder;
class sdpGtkIObjectContainer;
class sdpxml::Element;

///////////////////////////////////////////////////////////////////////////
// k3dChooserDataSource

/// Abstract interface for objects that can act as a datasource for a k3dChooser (i.e. the "model" in model-view-controller)
class k3dChooserDataSource
{
public:
	virtual ~k3dChooserDataSource() {}

	virtual k3dIEventSource* EventSource() = 0;

	virtual int Value() = 0;
	virtual void SetValue(const int Value) = 0;

	virtual bool ConvertToString(const int& Value, sdpString& String) const = 0;
	virtual bool ConvertFromString(const sdpString& String, int& Value) const = 0;
};

///////////////////////////////////////////////////////////////////////////
// k3dChooserVariableDataSource

/// Implementation of k3dChooserDataSource for use with k3dVariable<> objects and enumerated types
template<class Type>
class k3dChooserVariableDataSource :
	public k3dChooserDataSource
{
public:
	k3dChooserVariableDataSource(k3dVariable<Type>& Source) :
		m_Source(Source)
	{
	}

	k3dIEventSource* EventSource()
	{
		return dynamic_cast<k3dIEventSource*>(static_cast<k3dIVariable*>(&m_Source));
	}

	int Value()
	{
		return static_cast<int>(m_Source.Value());
	}

	void SetValue(const int Value)
	{
		m_Source.PushEditing();
		m_Source.SetValue(static_cast<Type>(Value));
		m_Source.PopEditing();
	}

	bool ConvertToString(const int& Value, sdpString& String) const
	{
		String = sdpToString(static_cast<Type>(Value));
		return true;
	}

	bool ConvertFromString(const sdpString& String, int& Value) const
	{
		Type value = sdpFromString(String, static_cast<Type>(0));
		Value = static_cast<int>(value);

		return true;
	}

private:
	k3dVariable<Type>& m_Source;
};

///////////////////////////////////////////////////////////////////////////
// k3dChooser

typedef k3dControl k3dChooserBase;

/// Provides a UI for editing objects that support k3dIChannel
class k3dChooser :
	public k3dChooserBase
{
public:
	k3dChooser(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName);
	~k3dChooser();

	/// Called by the framework when instantiating the button from a GTKML (XML) document
	bool Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element);

	/// Attaches the button to the object it fronts for
	bool Attach(k3dChooserDataSource* const Source, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);

	const sdpString CustomType() const;

	// k3dICommandNode overrides
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

protected:
	/// Called to handle SDPGTK events
	void OnEvent(sdpGtkEvent* Event);

	/// Called when the GTK+ widgets are about to disappear
	void OnDestroy();
	void OnValue(sdpGtkEvent* Event);
	void OnValueToggled(sdpGtkEvent* Event);

	/// Called when the the button needs to be updated
	void Update();

	/// Called prior to making one-or-more changes to the underlying data source
	void PushEditing();

	/// Called to modify the data source
	void SetValue(const int Value);

	/// Called once all changes to the underlying data source are complete
	void PopEditing();

	/// Storeas a reference to the underlying data object
	k3dChooserDataSource* m_DataSource;

	/// Encapsulates a single available user choice
	class Choice
	{
	public:
		Choice() :
			m_Name(""),
			m_Value(0)
		{
		}

		Choice(SDPCSTRING Name, const int Value) :
			m_Name(Name),
			m_Value(Value)
		{
		}

		const sdpString& Name() { return m_Name; }
		const int Value() { return m_Value; }

	protected:
		sdpString m_Name;
		int m_Value;
	};

	/// A collection of choices
	typedef std::vector<Choice> Choices;
	typedef Choices::iterator ChoiceIterator;
	Choices m_Choices;

	/// Enumerates available chooser types
	typedef enum
	{
		OPTIONMENU,
		RADIOBUTTONS
	} ChooserType;

	ChooserType m_Type;

	static const sdpString control_menu;
	static const sdpString control_value;
	static const sdpString control_valuetoggled;
};

#endif // K3DCHOOSER_H
