#ifndef K3DCOLLECTIONS_H
#define K3DCOLLECTIONS_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPTYPES_H
#include <sdptypes.h>
#endif // !SDPTYPES_H

#include <k3diobject.h>
#include <k3diobjectinfo.h>

namespace k3d
{

/// Tests a K-3D object to see if the requested state is set
template<typename Object>
class FlagIsSet
{
public:
	FlagIsSet(const gulong State) :
		m_State(State)
	{
	}

	bool operator()(Object O)
	{
		return SDP_S_OK == O->State()->QueryState(m_State);
	}

private:
	const gulong m_State;
};

/// Tests a K-3D object to see if the requested state is clear
template<typename Object>
class FlagIsClear
{
public:
	FlagIsClear(const gulong State) :
		m_State(State)
	{
	}

	bool operator()(Object O)
	{
		return SDP_S_FALSE == O->State()->QueryState(m_State);
	}

private:
	const gulong m_State;
};

/// Tests a K-3D object to see if it implements the requested interface
class Implements
{
public:
	Implements(const sdp::ITypeInfo& InterfaceType) :
		m_InterfaceType(InterfaceType)
	{
	}
	
	bool operator()(sdpIUnknown* const Object)
	{
		return m_InterfaceType.IsImplementedBy(Object);
	}

	bool operator()(k3dIObjectInfo* const ObjectInfo)
	{
		return SDP_S_OK == ObjectInfo->QueryObjectInterface(m_InterfaceType);
	}
	
private:
	const sdp::ITypeInfo& m_InterfaceType;
};

/// Tests a K-3D object to see if it has the requested name
class SameName
{
public:
	SameName(const sdpString Name) :
		m_Name(Name)
	{
		// Sanity checks ...
		g_assert(m_Name.size());
	}
	
	bool operator()(k3dIObject* const Object)
	{
		return Object->Name() == m_Name;
	}
	
	bool operator()(k3dIObjectInfo* const ObjectInfo)
	{
		return ObjectInfo->Name() == m_Name;
	}

private:
	const sdpString m_Name;
};

/// Tests a K-3D object to see if it has the requested ID
class SameID
{
public:
	SameID(const k3dIObject::ID ObjectID) :
		m_ID(ObjectID)
	{
	}
	
	bool operator()(k3dIObject* const Object)
	{
		return Object->ObjectID() == m_ID;
	}
	
private:
	const gulong m_ID;
};

/// Tests a k3dIObjectInfo to see if it matches the requested sdp::ClassID
class SameClassID
{
public:
	SameClassID(const sdp::ClassID ClassID) :
		m_ClassID(ClassID)
	{
	}
	
	bool operator()(k3dIObjectInfo* const ObjectInfo)
	{
		return ObjectInfo->ClassID() == m_ClassID;
	}

private:
	const sdp::ClassID m_ClassID;
};

/*
/// Applies a functor to objects in hierarchy order
template<class Functor>
bool k3dForEachObject(k3dIHierarchy* Hierarchy, gulong Criteria, Functor& ObjectFunctor)
{
	// Sanity checks ...
	g_assert(Hierarchy);

	// For each node in the hierarchy ...
	K3DHIERARCHYNODEHANDLE handle;
	k3dIObject* object;
	for(k3dIHierarchy::NodeType type = Hierarchy->FirstHierarchyNode(&handle); k3dIHierarchy::COMPLETE != type; type = Hierarchy->NextHierarchyNode(&handle))
		{
			object = Hierarchy->HierarchyObject(handle);
			
			// Skip the object if it doesn't fit our criteria ...
			if(SDP_S_OK != object->State()->QueryState(Criteria))
				continue;
			
			// Apply the functor ...
			if(!ObjectFunctor.HandleObject(type, object))
				return false;
		}

	return true;
}

/// Applies a functor to every point in the given document
template<class Functor>
bool k3dForEachPoint(k3dIDocument* Document, gulong Criteria, Functor& PointFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// Declare some stuff up front for efficiency ...
	k3dIObject* object;
	k3dIPointCollection* pointserver;
	K3DPOINTHANDLE pointhandle;
	k3dIPoint* point;

	// For each object in the document ...
	for(K3DOBJECTHANDLE objecthandle = Document->FirstObject(); objecthandle; objecthandle = Document->NextObject(objecthandle))
		{
			object = Document->Object(objecthandle);

			// Skip it if it doesn't contain points ...
			pointserver = dynamic_cast<k3dIPointCollection*>(object);
			if(pointserver)
				{
					// For each point in the object ...
					for(pointhandle = pointserver->FirstPoint(); pointhandle; pointhandle = pointserver->NextPoint(pointhandle))
						{
							point = pointserver->Point(pointhandle);

							// Skip points that don't match our criteria ...
							if(SDP_S_OK != point->State()->QueryState(Criteria))
								continue;

							// Apply the user function to the point ...
							if(!PointFunctor.HandlePoint(point))
								return false;
						}
				}
		}

	return true;
}

/// Applies a functor to every path in the given document
template<class Functor>
bool k3dForEachPath(k3dIDocument* Document, gulong Criteria, Functor& PathFunctor)
{
	// Sanity checks ...
	g_assert(Document);

	// Declare some stuff up front for efficiency ...
	k3dIObject* object;
	k3dIPathCollection* pathserver;
	K3DPATHHANDLE pathhandle;
	k3dIPath* path;

	// For each object in the document ...
	for(K3DOBJECTHANDLE objecthandle = Document->FirstObject(); objecthandle; objecthandle = Document->NextObject(objecthandle))
		{
			object = Document->Object(objecthandle);

			// Skip it if it doesn't contain paths ...
			pathserver = dynamic_cast<k3dIPathCollection*>(object);
			if(pathserver)
				{
					// For each path ...
					for(pathhandle = pathserver->FirstPath(); pathhandle; pathhandle = pathserver->NextPath(pathhandle))
						{
							path = pathserver->Path(pathhandle);

							// Skip paths that don't match our criteria ...
							if(SDP_S_OK != path->State()->QueryState(Criteria))
								continue;

							// Apply the user function to the path ...
							if(!PathFunctor.HandlePath(path))
								return false;
						}
				}
		}

	return true;
}
*/

} // namespace k3d

#endif // K3DCOLLECTIONS_H
