// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements k3dCommandNode, which provides a default implementation of k3dICommandNode
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcommandnode.h>

k3dCommandNode::k3dCommandNode(sdpIUnknown* const Parent, SDPCSTRING Name) :
	m_Name(Name),
	m_Parent(0)
{
	SetParentCommandNode(Parent);	
}

k3dCommandNode::~k3dCommandNode()
{
//std::cout << "k3dCommandNode::~k3dCommandNode() [" << CommandNodeName() << "]" << std::endl;

	if(m_Parent)
		m_Parent->RemoveCommandNode(this);
}

/////////////////////////////////////////////////////////////////////////////////////////////
// k3dICommandNode overrides ...

sdpResult k3dCommandNode::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	std::cout << g_pretty_function() << " [" << CommandNodeName() << "] unhandled command: [" << Name << "] [" << Arguments << "]" << std::endl;
	return SDP_E_FAIL;
}

SDPCSTRING k3dCommandNode::CommandNodeName()
{
	return m_Name;
}

k3dICommandNode* k3dCommandNode::CommandNodeParent()
{
	return m_Parent;
}

static k3dICommandNode* CommandNodeByName(k3dICommandNode* Node, SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert(Node);
	g_assert_string(Name);
	
	sdpString name(Name);
	g_assert(sdpString::npos == name.find("/"));
	
	if(name == ".")
		return Node;
	else if(name == "..")
		return Node->CommandNodeParent();
		
	for(K3DCOMMANDNODEHANDLE handle = Node->FirstCommandNode(); handle; handle = Node->NextCommandNode(handle))
		{
			k3dICommandNode* node = Node->CommandNode(handle);
			if(name == node->CommandNodeName())
				return node;	
		}
		
	return 0;
}

k3dICommandNode* k3dCommandNode::CommandNodeByPath(SDPCSTRING Path)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Path, 0);
	
	sdpString path(Path);
	path.trim();
	
	k3dICommandNode* result = this;
	if(path.substr(0, 1) == "/")
		{
			path = path.right(path.size()-1);
			result = GetRootCommandNode();
		}

	while(path.size())
		{
			sdpString name;
			sdpString::size_type slash = path.find("/");
			if(slash != sdpString::npos)
				{
					name = path.left(slash);
					path = path.right(path.size() - (slash+1));
				}
			else
				{
					name = path;
					path.clear();
				}
			
			if(name.size())	
				result = CommandNodeByName(result, name);
				
			if(0 == result)
				break;
		}

	return result;
}

K3DCOMMANDNODEHANDLE k3dCommandNode::FirstCommandNode()
{
	return m_Children.size() ? K3DCOMMANDNODEHANDLE(1) : K3DCOMMANDNODEHANDLE(0);
}

K3DCOMMANDNODEHANDLE k3dCommandNode::NextCommandNode(K3DCOMMANDNODEHANDLE CommandNode)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandNode, 0);
	g_return_val_if_fail(gulong(CommandNode) <= m_Children.size(), 0);
	
	return gulong(CommandNode) < m_Children.size() ? K3DCOMMANDNODEHANDLE(gulong(CommandNode)+1) : 0;
}

k3dICommandNode* k3dCommandNode::CommandNode(K3DCOMMANDNODEHANDLE CommandNode)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandNode, 0);
	g_return_val_if_fail(gulong(CommandNode) <= m_Children.size(), 0);
	
	return m_Children[gulong(CommandNode)-1];
}

sdpResult k3dCommandNode::AddCommandNode(k3dICommandNode* CommandNode)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandNode, SDP_E_FAIL);

//std::cout << CommandNodeName() << ": +" << CommandNode->CommandNodeName() << std::endl;	
			
	// Make sure the interface and name are both unique
	for(CommandNodeIterator ui = m_Children.begin(); ui != m_Children.end(); ui++)
		{
			if(CommandNode == (*ui))
				{
					std::cerr << "k3dCommandNode::AddCommandNode() - [" << CommandNodeName() << "] can't add node [" << CommandNode->CommandNodeName() << "] twice!" << std::endl;
					return SDP_E_FAIL;
				}
				
			if(CommandNode->CommandNodeName() == sdpString((*ui)->CommandNodeName()))
				{
					std::cerr << "k3dCommandNode::AddCommandNode() - [" << CommandNodeName() << "] can't add 2nd node named [" << CommandNode->CommandNodeName() << "] !" << std::endl;
					return SDP_E_FAIL;
				}
		}
	
	m_Children.push_back(CommandNode);
	return SDP_S_OK;
}

sdpResult k3dCommandNode::RemoveCommandNode(k3dICommandNode* CommandNode)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandNode, SDP_E_FAIL);
	
	// Find the UI node to erase ...
	CommandNodeIterator ui = std::find(m_Children.begin(), m_Children.end(), CommandNode);
	if(ui == m_Children.end())
		{
			std::cerr << "k3dCommandNode::RemoveCommandNode() - [" << CommandNodeName() << "] can't remove unknown node [" << CommandNode->CommandNodeName() << "] !" << std::endl;
			return SDP_E_FAIL;
		}
	
	m_Children.erase(ui);
	return SDP_S_OK;
}

void k3dCommandNode::SetCommandNodeName(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);
	
	m_Name = Name;
}

void k3dCommandNode::SetParentCommandNode(sdpIUnknown* const Parent)
{
	if(m_Parent)
		{
			m_Parent->RemoveCommandNode(this);
			m_Parent = 0;
		}

	if(0 == Parent)
		return;

	m_Parent = dynamic_cast<k3dICommandNode*>(Parent);

	if(m_Parent)
		m_Parent->AddCommandNode(this);
}

k3dICommandNode* k3dCommandNode::GetRootCommandNode()
{
	k3dICommandNode* result = this;
	while(result->CommandNodeParent())
		result = result->CommandNodeParent();
		
	return result;
}
