// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dDialog class, which provides a standard dialog implementation
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>

#include <sdpgtkevents.h>

const sdpString k3dDialog::control_delete = "k3dDialog::delete";

k3dDialog::k3dDialog(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName, k3dWindowGeometryStore* const WindowGeometryStore) :
	k3dDialogBase(ParentCommandNode, CommandNodeName),
	m_WindowGeometryStore(WindowGeometryStore)
{
}

bool k3dDialog::LoadGTKMLTemplate(SDPCSTRING Template)
{
	// Sanity checks ...
	g_assert_string(Template);
	
	// Load the template ...
	g_return_val_if_fail(Load(sdpAppendPath(k3dApplication()->DialogTemplatesPath(), Template), k3dOptionsGetPreferredLanguage(k3dApplication()->Options())), false);
	
	// Map some default events ...
	g_return_val_if_fail(MapEvent("delete-event", control_delete, false, Root(), true), false);

	return true;
}

bool k3dDialog::LoadGTKMLTemplate(SDPCSTRING Subdirectory, SDPCSTRING Template)
{
	// Sanity checks ...
	g_assert_string(Subdirectory);
	g_assert_string(Template);
	
	g_return_val_if_fail(LoadGTKMLTemplate(sdpAppendPath(Subdirectory, Template)), false);
	
	return true;
}

void k3dDialog::Show()
{
	RootWidget().Show();

	// Set our screen position ...
	if(m_WindowGeometryStore)
		{
			gint left = 0;
			gint top = 0;
			guint width = 0;
			guint height = 0;
			if(m_WindowGeometryStore->GetWindowGeometry(k3dCommandNodePath(this), left, top, width, height))
				gdk_window_move_resize(static_cast<GtkWidget*>(RootWidget())->window, left, top, width, height);
		}
}

k3dDialog::~k3dDialog()
{
	// Serialize our screen geometry (but not in batch mode, since it has a bad habit of trashing the user's stored settings) ...
	if(!k3dApplication()->BatchMode() && m_WindowGeometryStore)
		{
			gint left = 0;
			gint top = 0;
			gdk_window_get_root_origin(static_cast<GtkWidget*>(RootWidget())->window, &left, &top);

			gint width = 0;
			gint height = 0;
			gdk_window_get_size(static_cast<GtkWidget*>(RootWidget())->window, &width, &height);
			
			m_WindowGeometryStore->SetWindowGeometry(k3dCommandNodePath(this), left, top, width, height);
		}
	delete m_WindowGeometryStore;

	// Make sure we're not modal ...
	if(IsModal())
		CancelModal();

	// No more events from this point forward ...
	DisconnectAllEvents();

	// Get rid of user actions
	ClearCommands();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();

	// Get rid of our widget pointers ...
	Clear();
}

void k3dDialog::CloseDialog()
{
	delete this;
}

void k3dDialog::CallUpdateControls()
{
	BlockAllEvents();
	UpdateControls();
	UnblockAllEvents();
}

void k3dDialog::UpdateControls()
{
}

// Custom event handling ...
void k3dDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ok")
		OnOK();
	else if(Event->Name() == "cancel")
		OnCancel();
	else if(Event->Name() == "close")
		OnClose();
	else if(Event->Name() == control_delete)
		OnDelete(Event);
	else
		k3dDialogBase::OnEvent(Event);
}

// Called when the user hits the standard "OK" button ...
void k3dDialog::OnOK()
{
	CloseDialog();
}

// Called when the user hits the standard "Cancel" button ...
void k3dDialog::OnCancel()
{
	CloseDialog();
}

// Called when the user hits the standard "Close" button ...
void k3dDialog::OnClose()
{
	OnCancel();
}

void k3dDialog::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't allow the GTK+ tree to go away just yet ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	// Turn it into a "close", instead ...
	OnClose();
}

