// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dEditControl class, which provides a UI for string quantities
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3deditcontrol.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dicommandrecorder.h>
#include <k3distaterecorder.h>

#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

//////////////////////////////////////////////////////////////////////
// k3dEditControl

const sdpString k3dEditControl::control_value = "value";
const sdpString k3dEditControl::control_value_focus_in = "value_focus_in";
const sdpString k3dEditControl::control_value_focus_out = "value_focus_out";

k3dEditControl::k3dEditControl(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dEditControlBase(CommandNodeParent, CommandNodeName),
	m_DataSource(0)
{
	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<entry name=\"value\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"
				"<event signal=\"focus-in-event\" name=\"value_focus_in\" after=\"true\"/>"
				"<event signal=\"focus-out-event\" name=\"value_focus_out\"/>"
			"</entry>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, ""));

	RootWidget().Show();

//	AddCommand(control_value, control_value, CONTROL_EDITABLE);
}

k3dEditControl::~k3dEditControl()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();

	// Cancel any event notification from the data source
	if(m_DataSource)
		m_DataSource->RemoveEventObserver(this);
						
//	delete m_DataSource;
}

const sdpString k3dEditControl::CustomType() const
{
	return sdpString("k3dEditControl");
}

sdpResult k3dEditControl::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIVariable::Event_Changed)
		{
			Update();
			return SDP_S_OK;
		}
	else if(event == k3dIVariable::Event_Deleted)
		{
			if(m_DataSource)
				{
					m_DataSource->RemoveEventObserver(this);
					m_DataSource = 0;
				}
				
			return SDP_S_OK;
		}
		
	return k3dEditControlBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

sdpResult k3dEditControl::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	if(Name == control_value)
		{
			g_return_val_if_fail(InteractiveSetText(control_value, Arguments, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true), SDP_E_FAIL);
			OnValueFocusOut();
			return SDP_S_OK;
		}

	return k3dEditControlBase::Command(Name, Arguments, Result);
}

bool k3dEditControl::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element.Name() == "k3deditcontrol");

	return true;
}

bool k3dEditControl::Create()
{
	return true;
}

bool k3dEditControl::Attach(k3dEditControlDataSource* const DataSource, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	// Sanity checks ...
	g_return_val_if_fail(DataSource, false);

	g_return_val_if_fail(k3dEditControlBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);
	
	// Store references ...
	m_DataSource = DataSource;
	
	// Update the display ...
	Update();
	
	// We want to be notified if the data source changes ...
	m_DataSource->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dEditControl");
	m_DataSource->AddEventObserver(this, k3dIVariable::Event_Deleted, "k3dEditControl");

	return true;
}

void k3dEditControl::Update()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);
	
	// Display the current value ...
	Editable(control_value).SetText(m_DataSource->Value());
}

void k3dEditControl::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_value_focus_in)
		OnValueFocusIn();
	else if(Event->Name() == control_value_focus_out)
		OnValueFocusOut();
	else if(Event->Name() == "destroy")
		OnDestroy();
	else
		k3dEditControlBase::OnEvent(Event);
}

void k3dEditControl::OnDestroy()
{
	DisconnectAllEvents();

	Clear();
}

void k3dEditControl::OnValueFocusIn()
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Make sure everything's selected ...
	Editable(control_value).SelectRegion(0, -1);
}

void k3dEditControl::OnValueFocusOut()
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);
	
	// Get rid of the selection ...
	Editable(control_value).SelectRegion(0, 0);
	
	// Get the control text ...
	const sdpString newvalue = Editable(control_value).GetText();

	// If the value hasn't changed, we're done ...
	if(newvalue == m_DataSource->Value())
		return;

	// Record the command for posterity (tutorials) ...
	if(m_CommandRecorder)
		if(SDP_S_OK == m_CommandRecorder->IsRecording())
			m_CommandRecorder->RecordCommand(this, k3dICommandNode::USERINTERFACE, control_value, newvalue);

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		m_StateRecorder->SetCurrentChangeSet(new k3dStateChangeSet());
									
	// Update everything with the new value ...
	SetValue(newvalue);	
	
	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		{
			m_StateRecorder->RecordChangeSet(m_StateRecorder->ChangeSet(), m_StateChangeName + " \"" + newvalue + "\"");
			m_StateRecorder->SetCurrentChangeSet(0);
		}
}

void k3dEditControl::SetValue(const sdpString Value)
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Update storage with the new value ...
	m_DataSource->SetValue(Value);

	// Notify any interested parties that things changed ...
	if(m_ChangeAction)
		m_ChangeAction->DoAction();
}


