
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements k3dEventSource, a default implementation of the k3dIEventSource interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deventsource.h>

#include <k3dieventobserver.h>

#include <algorithm>
#include <iostream>

////////////////////////////////////////////////////////////////////////////
// class k3dEventSource

k3dEventSource::k3dEventSource()
{
}

k3dEventSource::~k3dEventSource()
{
	// Check for clients that never removed themselves ... could be trouble, Sheriff!
	if(0 == m_Observers.size())
		return;
	
	std::cerr << "k3dEventSource::~k3dEventSource() - still contains clients for:" << std::endl;
	for(ObserverIterator client = m_Observers.begin(); client != m_Observers.end(); client++)
		std::cerr << "\t [" << client->Event() << "] [" << client->ObserverData() << "]" << std::endl;
}

sdpResult k3dEventSource::AddEventObserver(k3dIEventObserver* const EventObserver, SDPCSTRING Event, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(EventObserver, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	// Add the new client ...
	m_Observers.push_back(Observer(EventObserver, Event, ObserverData));

	return SDP_S_OK;
}

sdpResult k3dEventSource::RemoveEventObserver(k3dIEventObserver* const EventObserver)
{
//std::cout << "k3dEventSource::RemoveEventObserver()" << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(EventObserver, SDP_E_FAIL);

	// Delete any clients that match ...
	const gulong eventcount = m_Observers.size();
	for(ObserverIterator client = m_Observers.begin(); client != m_Observers.end(); )
		{
			if(client->EventObserver() == EventObserver)
				client = m_Observers.erase(client);
			else
				client++;	
		}

	for(ObserverIterator broadcastclient = m_BroadcastObservers.begin(); broadcastclient != m_BroadcastObservers.end(); )
		{
			if(broadcastclient->EventObserver() == EventObserver)
				broadcastclient = m_BroadcastObservers.erase(broadcastclient);
			else
				broadcastclient++;	
		}

	// This client wasn't registered!
	if(eventcount == m_Observers.size())
		{
			std::cerr << "k3dEventSource::RemoveEventObserver() - Unknown client " << EventObserver << std::endl;
			return SDP_E_FAIL;
		}

	return SDP_S_OK;
}

sdpResult k3dEventSource::RemoveEventObserverEvent(k3dIEventObserver* EventObserver, SDPCSTRING Event, SDPCSTRING ObserverData)
{
//std::cout << "k3dEventSource::RemoveEventObserverEvent(" << EventObserver << ", \"" << Event << "\", \"" << ObserverData << "\");" << std::endl;

	// Sanity checks ...
	g_return_val_if_fail(EventObserver, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	// Erase any clients that match ...
	const gulong eventcount = m_Observers.size();
	for(ObserverIterator client = m_Observers.begin(); client != m_Observers.end(); )
		{
			if(client->EventObserver() == EventObserver && client->Event() == Event && client->ObserverData() == ObserverData)
				client = m_Observers.erase(client);
			else
				client++;	
		}

	for(ObserverIterator broadcastclient = m_BroadcastObservers.begin(); broadcastclient != m_BroadcastObservers.end(); )
		{
			if(broadcastclient->EventObserver() == EventObserver && broadcastclient->Event() == Event && broadcastclient->ObserverData() == ObserverData)
				broadcastclient = m_BroadcastObservers.erase(broadcastclient);
			else
				broadcastclient++;	
		}
	
	// This client isn't registered!
	if(eventcount == m_Observers.size())
		{
			std::cerr << "k3dEventSource::RemoveEventObserverEvent() - Unknown client [" << EventObserver << "] [" << Event << "] [" << ObserverData << "]" << std::endl;
			return SDP_E_FAIL;
		}

	return SDP_S_OK;
}

void k3dEventSource::Notify(SDPCSTRING Event, SDPCSTRING EventData)
{
	// Sanity checks ...
	g_return_if_string_fail(Event);

	// Make a copy of clients to notify ...
	m_BroadcastObservers.clear();
	for(ObserverIterator client = m_Observers.begin(); client != m_Observers.end(); client++)
		m_BroadcastObservers.push_back(*client);

	// For each client (the loop is a little convoluted since a client might potentially remove itself in response to an event) ...
	while(m_BroadcastObservers.size())
		{
			Observer client = m_BroadcastObservers.front();
			m_BroadcastObservers.erase(m_BroadcastObservers.begin());

			if(client.Event() != Event)
				continue;

			client.EventObserver()->ObserveEvent(this, Event, EventData, client.ObserverData());
		}
}

void k3dEventSource::NotifyAndTrace(SDPCSTRING Event, SDPCSTRING EventData)
{
	// Sanity checks ...
	g_return_if_string_fail(Event);

	std::cout << "k3dEventSource::NotifyAndTrace(): [" << sdpString(Event) << "] [" << sdpString(EventData) << "]" << std::endl;

	// Make a copy of clients to notify ...
	m_BroadcastObservers.clear();
	for(ObserverIterator client = m_Observers.begin(); client != m_Observers.end(); client++)
		m_BroadcastObservers.push_back(*client);

	// For each client (the loop is a little convoluted since a client might potentially remove itself in response to an event) ...
	while(m_BroadcastObservers.size())
		{
			Observer client = m_BroadcastObservers.front();
			m_BroadcastObservers.erase(m_BroadcastObservers.begin());

			if(client.Event() != Event)
				continue;

			std::cout << "  [" << sdpString(client.ObserverData()) << "]" << std::endl;

			client.EventObserver()->ObserveEvent(this, Event, EventData, client.ObserverData());
		}
}


