#ifndef K3DEVENTSERVER_H
#define K3DEVENTSERVER_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares k3dEventSource, a default implementation of the k3dIEventSource interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3daction.h>
#include <k3dieventsource.h>

////////////////////////////////////////////////////////////////////////////////////////////
// k3dEventSource

/// A default implementation of the k3dIEventSource interface
class k3dEventSource :
	public k3dIEventSource
{
friend class k3dActionNotify;
public:
	k3dEventSource();
	virtual ~k3dEventSource();

	// k3dIEventSource overrides ...
	sdpResult AddEventObserver(k3dIEventObserver* const EventObserver, SDPCSTRING Event, SDPCSTRING ObserverData);
	sdpResult RemoveEventObserver(k3dIEventObserver* const EventObserver);
	sdpResult RemoveEventObserverEvent(k3dIEventObserver* const EventObserver, SDPCSTRING Event, SDPCSTRING ObserverData);

protected:
	/// Called to transmit an event to waiting clients
	void Notify(SDPCSTRING Event, SDPCSTRING EventData);
	/// Called to transmit an event to waiting clients, with tracing
	void NotifyAndTrace(SDPCSTRING Event, SDPCSTRING EventData);

	/// Encapsulates a client's request to be notified of an event
	class Observer
	{
	public:
		Observer(k3dIEventObserver* const EventObserver, SDPCSTRING Event, SDPCSTRING ObserverData) :
			m_EventObserver(EventObserver),
			m_Event(Event),
			m_ObserverData(ObserverData)
		{
			// Sanity checks ...
			g_assert(m_EventObserver);
			g_assert_string(m_Event);
		}
		
		k3dIEventObserver* const EventObserver() const
		{
			return m_EventObserver;
		}
		
		const sdpString& Event() const
		{
			return m_Event;
		}
		
		const sdpString& ObserverData() const
		{
			return m_ObserverData;
		}

	protected:
		k3dIEventObserver* const m_EventObserver;
		const sdpString m_Event;
		const sdpString m_ObserverData;
	};

	/// A collection of clients requesting event notification
	typedef std::list<Observer> Observers;
	typedef Observers::iterator ObserverIterator;

private:
	/// Stores clients requesting event notification
	Observers m_Observers;
	/// Temporary storage while broadcasting an event
	Observers m_BroadcastObservers;
};

/////////////////////////////////////////////////////////////////////////////////////////////////
// k3dActionNotify

/// An implementation of k3dAction that causes a k3dEventSource to fire an event
class k3dActionNotify :
	public k3dAction
{
public:
	k3dActionNotify(k3dEventSource* const EventSource, const sdpString EventName, const sdpString EventData, k3dAction* const NextAction) :
		k3dAction(NextAction),
		m_EventSource(EventSource),
		m_EventName(EventName),
		m_EventData(EventData)
	{
		// Sanity checks ...
		g_assert(m_EventSource);
		g_assert(m_EventName.size());
	}
	
	void DoAction()
	{
		m_EventSource->Notify(m_EventName, m_EventData);
		DoNextAction();
	}

private:
	k3dEventSource* const m_EventSource;
	const sdpString m_EventName;
	const sdpString m_EventData;
};

#endif // K3DEVENTSERVER_H

