#ifndef K3DFILEFILTERS_H
#define K3DFILEFILTERS_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares convenience functions for working with file filter objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dobjectinfocollection.h>

#include <k3difileformat.h>
#include <k3diobjectcollection.h>

// Forward class declarations
class k3dIDocument;
class k3dIObject;

namespace k3d
{

/// Attempts to create a new file format object of the requested type
template<typename Interface>
Interface* GetFileFilter(k3dIDocument* const Document, const sdpString FilterName)
{
	// Sanity checks ...
	g_return_val_if_fail(Document, 0);
	g_return_val_if_fail(FilterName.size(), 0);

   	// Try to find an object with the same name ...
  	typedef std::vector<k3dIObjectInfo*> ObjectInfoCollection;
  	ObjectInfoCollection filters;
   	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(filters), k3d::SameName(FilterName));
	g_return_val_if_fail(1 == filters.size(), 0);

	// Create the filter
	k3d::Editing editing(Document->Objects());
	k3d::auto_ptr<Interface> filter(Document->Objects()->CreateObject(filters.front()->ClassID()));
	g_return_val_if_fail(filter.get(), 0);

	// Make sure the object is complete ...
	k3dIFileFormat* const fileformat = dynamic_cast<k3dIFileFormat*>(filter.get());
	g_return_val_if_fail(fileformat, 0);

	return filter.release();
}

/// Attempts to find a new file format object that can read/write the given file
template<typename Interface>
Interface* GetAutoFilter(k3dIDocument* const Document, const sdpString FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Document, 0);
	g_return_val_if_fail(FilePath.size(), 0);

   	// Get the set of objects that support the requested interface ...
  	typedef std::vector<k3dIObjectInfo*> ObjectInfoCollection;
  	ObjectInfoCollection filters;
   	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(filters), k3d::Implements(sdp::TypeInfo<Interface>()));

	// Create filters and test them against the actual file ...
	k3d::Editing editing(Document->Objects());
	for(ObjectInfoCollection::iterator filterinfo = filters.begin(); filterinfo != filters.end(); ++filterinfo)
		{
			// Create a file filter object ...
			k3d::auto_ptr<Interface> filter(Document->Objects()->CreateObject((*filterinfo)->ClassID()));
			if(!filter.get())
				{
					std::cerr << g_pretty_function() << ": " << (*filterinfo)->Name() << " doesn't support promised interface!" << std::endl;
					continue;
				}

			// If the object supports this file type, we're done ...
			k3dIFileFormat* const fileformat = dynamic_cast<k3dIFileFormat*>(filter.get());
			if(!fileformat)
				{
					std::cerr << g_pretty_function() << ": " << (*filterinfo)->Name() << " doesn't support k3dIFileFormat!" << std::endl;
					continue;
				}
			
			if(fileformat->QueryCanHandle(FilePath))
				return filter.release();
		}

	// Couldn't find a filter to fit this filetype ...
	return 0;
}

} // namespace k3d

#endif // K3DFILEFILTERS_H
