// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the FilterSelector class, which provides the user with a standard UI for selecting file import/export filters
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dfilefilters.h>
#include <k3dfilterselector.h>

#include <k3diapplication.h>
#include <k3difileformat.h>

#include <sdpgtkevents.h>

namespace k3d
{

//////////////////////////////////////////////////////////////////////
// FilterSelector

FilterSelector::FilterSelector(k3dIDocument* const Document, const sdpString Title, const sdpString FilePath, const sdp::ITypeInfo& FilterType) :
	m_Document(Document),
	m_FilePath(FilePath),
	m_OK(false)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(FilePath.size());
	
	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<window title=\"Select File Format:\" type=\"dialog\" show=\"false\" allowgrow=\"false\">"
				"<vbox homogeneous=\"false\">"
					"<label labelpadding=\"6 6\">Select a file format.  You may select \"Automatic\"\n"
						"to attempt automatic selection based on file name / contents.</label>"
					"<combo homogeneous=\"false\" name=\"filter\"/>"
					"<hbuttonbox layout=\"end\">"
						"<button name=\"ok\">OK"
							"<event signal=\"clicked\" name=\"ok\"/>"
						"</button>"
						"<button name=\"cancel\">Cancel"
							"<event signal=\"clicked\" name=\"cancel\"/>"
						"</button>"
					"</hbuttonbox>"
				"</vbox>"
			"</window>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, "FilterSelector"));
	
	// Set our title ...
	RootWindow().SetTitle(Title);

	// Get the list of available formats ...
	typedef std::vector<k3dIObjectInfo*> FileFormats;
	FileFormats fileformats;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(fileformats), k3d::Implements(FilterType));

	// Our special "key" value for automatic filter selection ...
	const sdpString automatic("Automatic");

	// Create a mapping from human-readable strings to file filter names ...
	m_Choices[automatic] = sdpString();
	for(FileFormats::iterator format = fileformats.begin(); format != fileformats.end(); ++format)
		m_Choices[(*format)->ShortDescription()] = (*format)->Name();
			
	// Create a collection of human-readable choices for the user ...
	sdpStringArray combochoices;
	combochoices.push_back(automatic);
	for(FileFormats::iterator format = fileformats.begin(); format != fileformats.end(); ++format)
		combochoices.push_back((*format)->ShortDescription());
	
	Combo("filter").SetList(combochoices);
	Combo("filter").SetEntryText(automatic);

	RootWidget().Show();
}

FilterSelector::~FilterSelector()
{
	RootWidget().Destroy();
}

bool FilterSelector::OK() const
{
	return m_OK;
}

bool FilterSelector::Cancelled() const
{
	return !m_OK;
}

void FilterSelector::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "ok")
		OnOK();
	else if(Event->Name() == "cancel")
		OnCancel();
	else if(Event->Name() == "delete")
		OnDelete(Event);
	else
		FilterSelectorBase::OnEvent(Event);
}

void FilterSelector::OnOK()
{
	m_OK = true;
	CancelModal();
}

void FilterSelector::OnCancel()
{
	m_OK = false;
	CancelModal();
}

void FilterSelector::OnDelete(sdpGtkEvent* const Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't allow the GTK+ tree to go away just yet ...
	static_cast<sdpGtkEventWidgetDeleteEvent*>(Event)->SetResult(true);

	// Turn it into a "cancel", instead ...
	OnCancel();
}

const sdpString FilterSelector::FilterChoice()
{
	return Combo("filter").GetEntryText();
}

} // namespace k3d


