
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dFrameRange class for working with ranges of numbered files
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dframerange.h>
#include <iomanip>

k3dFrameRange::k3dFrameRange(const sdpString File)
{
	// Sanity checks ...
	g_assert(File.size());

	sdpString::size_type index;
	sdpString buffer(File);

	// Get the post-digits string ...
	index = buffer.find_last_of("0123456789");
	if(sdpString::npos != index)
		{
			m_PostDigits = buffer.right(buffer.size() - (index+1));
			buffer = buffer.left(index+1);
		}
		
	// Get the post-digit
	index = buffer.find_last_not_of("0123456789");
	if(sdpString::npos != index)
		{
			m_FrameDigits = buffer.size() - (index+1);
			m_PreDigits = buffer.left(index+1);
		}
	else
		{
			m_FrameDigits = buffer.size();
		}
		
	// Find the first available frame ...
	gulong maxframe = MaxFrame();
	for(gulong i = 0; i < maxframe; i++)
		{
			sdpString file;
			Frame(i, file);
			
			if(sdpFileExists(file))
				{
					m_FirstFrame = i;
					break;
				}
		}

	// Find the last available frame ...
	for(gulong j = m_FirstFrame + 1; j < maxframe; j++)
		{
			sdpString file;
			Frame(j, file);
			
			if(!sdpFileExists(file))
				{
					m_LastFrame = j - 1;
					break;
				}
		}
}

k3dFrameRange::k3dFrameRange(const sdpString File, const gulong FirstFrame, const gulong LastFrame) :
	m_FirstFrame(FirstFrame),
	m_LastFrame(LastFrame)
{
	// Sanity checks ...
	g_assert(File.size());
	
	sdpString::size_type index;
	sdpString buffer(File);

	// Get the post-digits string ...
	index = buffer.find_last_of("0123456789");
	if(sdpString::npos != index)
		{
			m_PostDigits = buffer.right(buffer.size() - (index+1));
			buffer = buffer.left(index+1);
		}
	
	// Get the digits ...
	index = buffer.find_last_not_of("0123456789");
	if(sdpString::npos != index)
		{
			m_FrameDigits = buffer.size() - (index+1);
			m_PreDigits = buffer.left(index+1);
		}
	else
		{
			m_FrameDigits = buffer.size();
		}
}

k3dFrameRange::k3dFrameRange(const sdpString PreDigits, const sdpString PostDigits, const gulong FirstFrame, const gulong LastFrame, const gulong FrameDigits) :
	m_PreDigits(PreDigits),
	m_PostDigits(PostDigits),
	m_FirstFrame(FirstFrame),
	m_LastFrame(LastFrame),
	m_FrameDigits(FrameDigits)
{
	// Sanity checks ...
	g_assert(FirstFrame <= LastFrame);
	g_assert(FrameDigits);
}

bool k3dFrameRange::Frame(const gulong FrameNumber, sdpString& Result) const
{
	// Sanity checks ...
	g_return_val_if_fail(FrameNumber <= MaxFrame(), false);
	
	std::strstream buffer;
	buffer << m_PreDigits << std::setfill('0') << std::setw(m_FrameDigits) << sdpToString(FrameNumber) << m_PostDigits;
	Result = buffer.str();
	buffer.freeze(false);
	
	return true;
}

gulong k3dFrameRange::MaxFrame() const
{
	return gulong(pow(10, m_FrameDigits) - 1);
}


