#ifndef K3DHIERARCHYCOLLECTION_H
#define K3DHIERARCHYCOLLECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcollections.h>
#include <k3dihierarchy.h>

namespace k3d
{

/// Applies a functor to members of a hierarchy, in hierarchy-order
template<typename Functor>
Functor for_each(k3dIHierarchy* const Hierarchy, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(Hierarchy, F);

	// Move through the hierarchy records ...
	K3DHIERARCHYNODEHANDLE handle;
	for(k3dIHierarchy::NodeType type = Hierarchy->FirstHierarchyNode(&handle); type != k3dIHierarchy::COMPLETE; type = Hierarchy->NextHierarchyNode(&handle))
		F(Hierarchy->HierarchyObject(handle), type);
	
	return F;
}

/// Applies a functor to every child of the given object
template<typename Functor>
Functor for_each_child(k3dIHierarchy* const Hierarchy, k3dIObject* const Parent, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(Hierarchy, F);
	g_return_val_if_fail(Parent, F);

	// Move through the parent's child records ...
	gulong depth = 0;
	K3DHIERARCHYNODEHANDLE handle;
	for(k3dIHierarchy::NodeType type = Hierarchy->GetHierarchyNode(Parent, &handle); ; type = Hierarchy->NextHierarchyNode(&handle))
		{
			switch(type)
				{
					case k3dIHierarchy::PUSH:
						depth++;
						break;

					case k3dIHierarchy::NODE:
						if(2 == depth)
							F(Hierarchy->GetHierarchyObject(handle));
						break;

					case k3dIHierarchy::POP:
						depth--;
							if(0 == depth)
								return F;
						break;

					case k3dIHierarchy::COMPLETE:
					case k3dIHierarchy::ERROR:
						g_return_val_if_fail(0, F);
				}
	}

	g_return_val_if_fail(0, F);
}

/// Applies a functor to every descendant of the given object
template<typename Functor>
Functor for_each_descendant(k3dIHierarchy* const Hierarchy, k3dIObject* const Parent, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(Hierarchy, F);
	g_return_val_if_fail(Parent, F);

	// Move through the parent's child records ...
	gulong depth = 0;
	K3DHIERARCHYNODEHANDLE handle;
	for(k3dIHierarchy::NodeType type = Hierarchy->HierarchyNode(Parent, &handle); ; type = Hierarchy->NextHierarchyNode(&handle))
		{
			switch(type)
				{
					case k3dIHierarchy::PUSH:
						depth++;
						break;

					case k3dIHierarchy::NODE:
						if(depth > 1)
							F(Hierarchy->HierarchyObject(handle));
						break;

					case k3dIHierarchy::POP:
						depth--;
						if(0 == depth)
							return F;
						break;

					case k3dIHierarchy::COMPLETE:
					case k3dIHierarchy::ERROR:
						g_return_val_if_fail(0, F);
				}
		}

	g_return_val_if_fail(0, F);
}

} // namespace k3d

#endif // K3DHIERARCHYCOLLECTION_H
