
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dLocation and k3dLocationCollection classes, which provide default implementations of k3dILocation and k3dILocationCollection, respectively
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dlocationcollection.h>
#include <k3distate.h>
#include <sdpxml.h>


////////////////////////////////////////////////////////////////
// k3dLocation

k3dLocation::k3dLocation() :
	k3dState(k3dIState::VISIBLE)
{
}

////////////////////////////////////////////////////////////////
// k3dIState overrides ...

sdpResult k3dLocation::SetState(gulong Mask, bool State)
{
	RawSetState(Mask, State);
	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////
// k3dILocation overrides ...

k3dIState* k3dLocation::State()
{
	return this;
}

const sdpVector3& k3dLocation::LocalXYZ()
{
	return m_LocalXYZ;
}

const sdpVector3& k3dLocation::AnimXYZ()
{
	return m_AnimXYZ;
}

const sdpVector3& k3dLocation::WorldXYZ()
{
	return m_WorldXYZ;
}

const sdpVector3& k3dLocation::CameraXYZ()
{
	return m_CameraXYZ;
}

const sdpVector3& k3dLocation::ScreenXYZ()
{
	return m_ScreenXYZ;
}

const sdpVector3& k3dLocation::NormalXYZ()
{
	return m_NormalXYZ;
}

const sdpVector3& k3dLocation::TextureXYZ()
{
	return m_TextureXYZ;
}

const sdpVector3& k3dLocation::ImplicitTextureXYZ()
{
	return m_ImplicitTextureXYZ;
}

void k3dLocation::SetLocalXYZ(const sdpVector3& Coords)
{
	m_LocalXYZ = Coords;
}

void k3dLocation::SetAnimXYZ(const sdpVector3& Coords)
{
	m_AnimXYZ = Coords;
}

void k3dLocation::SetWorldXYZ(const sdpVector3& Coords)
{
	m_WorldXYZ = Coords;
}

void k3dLocation::SetCameraXYZ(const sdpVector3& Coords)
{
	m_CameraXYZ = Coords;
}

void k3dLocation::SetScreenXYZ(const sdpVector3& Coords)
{
	m_ScreenXYZ = Coords;
}

void k3dLocation::SetNormalXYZ(const sdpVector3& Coords)
{
	m_NormalXYZ = Coords;
}

void k3dLocation::SetTextureXYZ(const sdpVector3& Coords)
{
	m_TextureXYZ = Coords;
}

void k3dLocation::SetImplicitTextureXYZ(const sdpVector3& Coords)
{	
	m_ImplicitTextureXYZ = Coords;
}


////////////////////////////////////////////////////////////////
// k3dLocationCollection

k3dLocationCollection::k3dLocationCollection()
{
}

k3dLocationCollection::~k3dLocationCollection()
{
	std::for_each(m_Locations.begin(), m_Locations.end(), sdp::DeleteObject());
}

////////////////////////////////////////////////////////////////
// k3dILocationCollection overrides

void k3dLocationCollection::VisitLocations(k3dILocationVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	
	std::for_each(m_Locations.begin(), m_Locations.end(), sdp::VisitorAdapter<k3dILocationVisitor, k3dILocation*>(Visitor, &k3dILocationVisitor::VisitLocation));
}

gulong k3dLocationCollection::LocationCount()
{
	return m_Locations.size();
}


K3DLOCATIONHANDLE k3dLocationCollection::FirstLocation()
{
	return K3DLOCATIONHANDLE(m_Locations.size() ? 1 : 0);

	return 0;
}


K3DLOCATIONHANDLE k3dLocationCollection::NextLocation(K3DLOCATIONHANDLE Location)
{
	g_assert(Location);

	LocationContainer::size_type i = (LocationContainer::size_type)Location + 1;
	return K3DLOCATIONHANDLE(i <= m_Locations.size() ? i : 0);
}


k3dILocation* k3dLocationCollection::Location(K3DLOCATIONHANDLE Location)
{
	g_assert(Location);

	return m_Locations[LocationContainer::size_type(Location) - 1];
}


/////////////////////////////////////////////////////////////////////////////
// Helper functions

sdpResult k3dLocationCollection::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	// Don't bother if we don't have any locations to save ...
	if(!m_Locations.size())
		return SDP_S_FALSE;

	sdpxml::Element& locations = Element.Append(sdpxml::Element("locations"));

	for(LocationIterator location = m_Locations.begin(); location != m_Locations.end(); location++)
		locations.Append(sdpxml::Element("location", "", sdpxml::Attribute("position", sdpToString((*location)->m_LocalXYZ))));

	return SDP_S_OK;
}

class k3dLocationCollectionLoadLocation
{
public:
	k3dLocationCollectionLoadLocation(k3dLocationCollection* const Client) :
		m_Client(Client)
	{
		// Sanity checks ...
		g_assert(m_Client);
	}

	void operator()(sdpxml::Element& Element)
	{
		// Sanity checks ...
		g_return_if_fail(Element.Name() != "location");

g_assert_not_reached();
//		sdpxml::ParseAttribute(Element, "position", m_Client->m_Locations[m_Index].m_LocalXYZ);
	}

private:
	k3dLocationCollection* const m_Client;
};


sdpResult k3dLocationCollection::Load(sdpxml::Element& Element)
{
	sdpxml::ElementPointer const locations = sdpxml::FindElement(Element, sdpxml::SameName("locations"));
	if(!locations)
		return SDP_S_OK;

	k3dLocationCollectionLoadLocation functor(this);
	std::for_each(locations->Children().begin(), locations->Children().end(), functor);

	return SDP_S_OK;
}

