#ifndef K3DLOCATIONCOLLECTION_H
#define K3DLOCATIONCOLLECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dLocation and k3dLocationCollection classes, which provide default implementations of k3dILocation and k3dILocationCollection, respectively
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcollections.h>
#include <k3dstate.h>

#include <k3dilocation.h>
#include <k3dilocationcollection.h>

namespace k3d
{

/// Applies a functor to every k3dILocation in a collection
template<typename Functor>
Functor for_each(k3dILocationCollection* const LocationCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(LocationCollection, F);

	// For each object in the document ...
	for(K3DLOCATIONHANDLE handle = LocationCollection->FirstLocation(); handle; handle = LocationCollection->NextLocation(handle))
		F(LocationCollection->Location(handle));

	return F;
}
	
/// Applies a functor to every k3dILocation in a collection that satisfies the predicate
template<typename Functor, typename Predicate>
Functor for_each_if(k3dILocationCollection* const LocationCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(LocationCollection, F);

	// For each object in the document ...
	for(K3DLOCATIONHANDLE handle = LocationCollection->FirstLocation(); handle; handle = LocationCollection->NextLocation(handle))
		{
			k3dILocation* const object = LocationCollection->Location(handle);
		
			if(P(object))
				F(object);
		}

	return F;
}

/// Copies k3dILocation references from a k3dILocationCollection into a generic container
template<typename InsertIterator>
void copy(k3dILocationCollection* const LocationCollection, InsertIterator D)
{
	// Sanity checks ...
	g_return_if_fail(LocationCollection);

	// For each object in the document ...
	for(K3DLOCATIONHANDLE handle = LocationCollection->FirstLocation(); handle; handle = LocationCollection->NextLocation(handle))
		*D++ = LocationCollection->Location(handle);
}

/// Copies k3dILocation references that satisfy a predicate from a k3dILocationCollection into a generic container
template<typename InsertIterator, typename Predicate>
void copy_if(k3dILocationCollection* const LocationCollection, InsertIterator D, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(LocationCollection);

	// For each object in the document ...
	for(K3DLOCATIONHANDLE handle = LocationCollection->FirstLocation(); handle; handle = LocationCollection->NextLocation(handle))
		{
			k3dILocation* const object = LocationCollection->Location(handle);
			if(P(object))
				*D++ = object;
		}
}

} // namespace k3d

///////////////////////////////////////////////////////////////////
// k3dLocation

class k3dLocation :
	public k3dILocation,
	public k3dState
{
//friend class k3dLocationCollection;
public:
	k3dLocation();

	// k3dIState implementation ...
	sdpResult SetState(const gulong Mask, const bool State);

	// k3dILocation implementation ...
	k3dIState* State();
	const sdpVector3& LocalXYZ();
	const sdpVector3& AnimXYZ();
	const sdpVector3& WorldXYZ();
	const sdpVector3& CameraXYZ();
	const sdpVector3& ScreenXYZ();
	const sdpVector3& NormalXYZ();
	const sdpVector3& TextureXYZ();
	const sdpVector3& ImplicitTextureXYZ();
	void SetLocalXYZ(const sdpVector3& Coords);
	void SetAnimXYZ(const sdpVector3& Coords);
	void SetWorldXYZ(const sdpVector3& Coords);
	void SetCameraXYZ(const sdpVector3& Coords);
	void SetScreenXYZ(const sdpVector3& Coords);
	void SetNormalXYZ(const sdpVector3& Coords);
	void SetTextureXYZ(const sdpVector3& Coords);
	void SetImplicitTextureXYZ(const sdpVector3& Coords);

	sdpVector3 m_LocalXYZ;
	sdpVector3 m_AnimXYZ;
	sdpVector3 m_WorldXYZ;
	sdpVector3 m_CameraXYZ;
	sdpVector3 m_ScreenXYZ;
	sdpVector3 m_NormalXYZ;
	sdpVector3 m_TextureXYZ;
	sdpVector3 m_ImplicitTextureXYZ;
};

inline bool operator < (const k3dLocation& A, const k3dLocation& B) { return false; };
inline bool operator == (const k3dLocation& A, const k3dLocation& B) { return false; };

////////////////////////////////////////////////////////////////////
// k3dLocationCollection

class k3dLocationCollection :
	public k3dILocationCollection
{
public:
	k3dLocationCollection();
	virtual ~k3dLocationCollection();

	// k3dILocationCollection overrides
	void VisitLocations(k3dILocationVisitor* const Visitor);
	gulong LocationCount();
	K3DLOCATIONHANDLE FirstLocation();
	K3DLOCATIONHANDLE NextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* Location(K3DLOCATIONHANDLE Location);

	// Helper functions
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpxml::Element& Element);

protected:
	typedef std::vector<k3dLocation*> LocationContainer;
	typedef LocationContainer::iterator LocationIterator;
	LocationContainer m_Locations;
};

#endif // K3DLOCATIONCOLLECTION_H
