// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dMesh class, which implements k3dIMesh, and encapsulates a collection of points, paths, NURBS curves, material, and shading information
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dboundingbox.h>
#include <k3dmesh.h>
#include <k3dmisc.h>
#include <k3dpoint.h>
#include <k3dpath.h>
#include <k3dpathpoint.h>
#include <k3drishadersetup.h>
#include <k3dshadingnode.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3didocument.h>
#include <k3dieditorengine.h>
#include <k3dieditormaterial.h>
#include <k3dimaterial.h>
#include <k3direnderframe.h>
#include <k3direnderjob.h>
#include <k3dirimaterial.h>
#include <k3distaterecorder.h>

#include <sdpgl.h>

#include <strstream>

//////////////////////////////////////////////////////////////////////////
// k3dMesh

const sdpVector3 k3dMesh::m_SelectedColor(1, 1, 1);
const sdpVector3 k3dMesh::m_PathNormalColor(1, 1, 0);

k3dMesh::k3dMesh(sdpIUnknown* Initializer) :
	k3dMeshBase(Initializer),
	m_CurrentPointID(0),
	m_CurrentPathID(0),
	m_CurrentPathPointID(0),
	m_Color("color", sdpVector3(0, 0, 0), Document()->StateRecorder(), 0),
	m_RenderMode("rendermode", NORMAL, Document()->StateRecorder(), 0),
	m_Material("material", Document(), Document()->StateRecorder(), 0),
	m_PointCollectionChanged(false)
{
	AddVariable(&m_Color);
	AddVariable(&m_RenderMode);
	AddVariable(&m_Material);
}

k3dMesh::~k3dMesh()
{
	std::for_each(m_ShadingNodes.begin(), m_ShadingNodes.end(), sdp::DeleteObject());
	std::for_each(m_Paths.begin(), m_Paths.end(), sdp::DeleteObject());
	std::for_each(m_Points.begin(), m_Points.end(), sdp::DeleteObject());
}

////////////////////////////////////////////////////////////////////////////
// k3dIObject implementation

/*
k3dIState* k3dMesh::State()
{
	return this;
}
*/

/*
////////////////////////////////////////////////////////////////////////////
// k3dIState implementation

sdpResult k3dMesh::SetState(gulong Mask, bool State)
{
	// If we're changing our selection state, make it a deep select of all points/paths/curves ...
	if(Mask & k3dIState::SELECTED)
		{
			NURBSCurveIterator curveend = m_NURBSCurves.end();
			for(NURBSCurveIterator curve = m_NURBSCurves.begin(); curve != curveend; curve++)
				(*curve)->SetState(k3dIState::SELECTED, State);

			PathIterator pathend = m_Paths.end();
			for(PathIterator path = m_Paths.begin(); path != pathend; path++)
				(*path)->SetState(k3dIState::SELECTED, State);

			PointIterator pointend = m_Points.end();
			for(PointIterator point = m_Points.begin(); point != pointend; point++)
				(*point)->SetState(k3dIState::SELECTED, State);
		}

	return k3dMeshBase::SetState(Mask, State);
}
*/

///////////////////////////////////////////////////////////////
// k3dIMesh implementation

k3dIDocument* k3dMesh::Document()
{
	return k3dMeshBase::Document();
}

k3dIPathCollection* k3dMesh::Paths()
{
	return this;
}

k3dIPath* k3dMesh::CreatePath()
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), 0);

	// Create a new path object ...
	k3dPath* const path = new k3dPath(this, NextPathID());

	// Record undo/redo data ...
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_PathCollectionChanged)
				{
					m_PathCollectionChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<PathCollection>(m_Paths));
				}

			changeset->RecordOldState(new k3dNewInstanceStateContainer<k3dPath>(path, changeset));
		}

	// Maintain our collection ...
	m_Paths.push_back(path);

	return path;
}

sdpResult k3dMesh::DeletePath(k3dIPath* const Path)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Get the path ...
	const PathIterator pathiterator = std::find(m_Paths.begin(), m_Paths.end(), Path);
	g_return_val_if_fail(pathiterator != m_Paths.end(), SDP_E_FAIL);

	// Record undo/redo data ...
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_PathCollectionChanged)
				{
					m_PathCollectionChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<PathCollection>(m_Paths));
				}

			changeset->RecordOldState(new k3dDeleteInstanceStateContainer<k3dPath>(*pathiterator, changeset));
		}

	// Maintain our collection ...
	m_Paths.erase(pathiterator);

	return SDP_S_OK;
}

k3dIPointCollection* k3dMesh::Points()
{
	return this;
}

k3dIPoint* k3dMesh::CreatePoint()
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), 0);

	// Create a new point object ...
	k3dPoint* const point = new k3dPoint(this, NextPointID());

	// Record undo/redo data ...
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_PointCollectionChanged)
				{
					m_PointCollectionChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<PointCollection>(m_Points));
				}

			changeset->RecordOldState(new k3dNewInstanceStateContainer<k3dPoint>(point, changeset));
		}

	// Maintain our collection ...
	m_Points.push_back(point);

	return point;
}

sdpResult k3dMesh::DeletePoint(k3dIPoint* const Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Get the point iterator ...
	PointIterator point = std::find(m_Points.begin(), m_Points.end(), Point);
	g_return_val_if_fail(point != m_Points.end(), SDP_E_FAIL);

	// Update any paths that reference this point ...
	for(PathIterator path = m_Paths.begin(); path != m_Paths.end(); path++)
		{
			(*path)->PushEditing();
			(*path)->RemovePoint(Point);
			(*path)->PopEditing();
		}

	// Record undo/redo data ...
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_PointCollectionChanged)
				{
					m_PointCollectionChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<PointCollection>(m_Points));
				}

			changeset->RecordOldState(new k3dDeleteInstanceStateContainer<k3dPoint>(*point, changeset));
		}

	// Maintain our collection ...
	m_Points.erase(point);

	return SDP_S_OK;
}

class k3dMeshCalculateNormal
{
public:
	class k3dMeshResetPathpoint
	{
	public:
		k3dMeshResetPathpoint(k3dPath* Path) : m_Path(Path) {}

		void operator()(k3dPathPoint* PathPoint)
		{
			PathPoint->m_NormalXYZ = m_Path->m_Normal;
		}

		k3dPath* m_Path;
	};

	void operator()(k3dPath* Path)
	{
		// Calculate the path's normal ...
		Path->CalculateNormal();
		Path->m_Normal.Normalize();

		// Reset pathpoint normals to match the path's (flat shading) ...
		k3dMeshResetPathpoint functor(Path);
		std::for_each(Path->m_PathPoints.begin(), Path->m_PathPoints.end(), functor);
	}
};

class k3dMeshSetState
{
public:
	k3dMeshSetState(gulong Mask, bool State) { m_Mask = Mask; m_State = State; }

	void operator()(k3dPoint* Point)
	{
		Point->SetState(m_Mask, m_State);
	}

	gulong m_Mask;
	bool m_State;
};

sdpResult k3dMesh::ShadePaths(const gulong State, const gdouble Angle)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Start by getting rid of any existing shading ...
	UnshadePaths(State);

	// Setup paths by calculating normals ...
	k3dMeshCalculateNormal normalfunctor;
	std::for_each(m_Paths.begin(), m_Paths.end(), normalfunctor);

	// Setup points by clearing the USER1 state ...
	k3dMeshSetState statefunctor(k3dIState::USER1, false);
	std::for_each(m_Points.begin(), m_Points.end(), statefunctor);

	// Temporary storage ...
	std::vector<k3dPathPoint*> pathpointarray;
	k3dShadingNode* node;
	gulong i, pathpointcount;

/*
	CLayer* layer;
	k3dMesh* mesh;
	k3dPath *path, *path2;
	k3dPathPoint *pathpoint, *pathpoint2;
	k3dPoint3D* point;
	k3dShadingNode *node;

	gulong i, count;

	sdpVector3 normal1, normal2;


	if(Angle)
		{
			double angle = fabs(cos(KRadians(MaxAngle)));

			for(layer = m_Layers.front(); layer; layer = layer->m_Next)
				{
					for(mesh = layer->m_Meshes.front(); mesh; mesh = mesh->m_Next)
						{
							for(path = mesh->m_Paths.front(); path; path = path->m_Next)
								{
									if(SDP_S_OK != path->QueryState(State)) continue;

									normal1 = path->m_Normal;
									normal1.Normalize();

									for(pathpoint = path->m_PathPoints.front(); pathpoint; pathpoint = pathpoint->m_Next)
										{
											if(SDP_S_OK == pathpoint->m_Point->QueryState(k3dIState::USER1)) continue;

											count = 1;
											pathpointarray.clear();

											for(path2 = path->m_Next; path2; path2 = path2->m_Next)
												{
													if(SDP_S_OK != path2->QueryState(State)) continue;

													normal2 = path2->m_Normal;
													normal2.Normalize();

													if(fabs(normal1 * normal2) >= angle)	// Seems bass-ackward, but cos() is that way ...
														{
    													for(pathpoint2 = path2->m_PathPoints.front(); pathpoint2; pathpoint2 = pathpoint2->m_Next)
    														{
    															if(pathpoint2->m_Point == pathpoint->m_Point)
																		{
																			pathpointarray.push_back(pathpoint2);
    																	count++;
    																}
    														}
														}
												}

											if(count > 1)
												{
													node = mesh->CreateShadingNode();
													node->m_PointCount = count;
													node->m_PointArray = (k3dPathPoint**)malloc(count * sizeof(k3dPathPoint*));

													node->m_PointArray[0] = pathpoint;

													for(i = 1; i < count; i++)
														{
															node->m_PointArray[i] = pathpointarray[i - 1];
														}
												}
										}
								}
						}
				}
		}
	else
*/
		{
			// For each path ...
			PathIterator pathend = m_Paths.end();
			for(PathIterator path = m_Paths.begin(); path != pathend; path++)
				{
					// Skip paths that don't match our criteria ...
					if(SDP_S_OK != (*path)->QueryState(State))
						continue;

					// For each pathpoint ...
					k3dPath::PathPointIterator pathpointend = (*path)->m_PathPoints.end();
					for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != pathpointend; pathpoint++)
						{
							// If this point is already smooth-shaded, skip it ...
							if(SDP_S_OK == (*pathpoint)->m_Point->QueryState(k3dIState::USER1))
								continue;

							pathpointarray.clear();
							pathpointarray.push_back(*pathpoint);

							// For each path following the one we're looking at ...
							PathIterator path2 = path;
							for(path2++; path2 != pathend; path2++)
								{
									// Skip paths that don't match our criteria
									if(SDP_S_OK != (*path2)->QueryState(State))
										continue;

									// For each pathpoint ...
									k3dPath::PathPointIterator pathpointend2 = (*path2)->m_PathPoints.end();
    							for(k3dPath::PathPointIterator pathpoint2 = (*path2)->m_PathPoints.begin(); pathpoint2 != pathpointend2; pathpoint2++)
    								{
											// If it matches the path point we're looking at ...
    									if((*pathpoint2)->m_Point == (*pathpoint)->m_Point)
												{
													// Mark it ...
													(*pathpoint2)->m_Point->SetState(k3dIState::USER1, true);
													// Store it ...
													pathpointarray.push_back(*pathpoint2);
    										}
    								}
								}

							// If we found some pathpoints to shade together ...
							pathpointcount = pathpointarray.size();
							if(pathpointcount > 1)
								{
									// Mark this pathpoint ...
									(*pathpoint)->m_Point->SetState(k3dIState::USER1, true);

									// Create a new shading node ...
									node = new k3dShadingNode();

									for(i = 0; i < pathpointcount; i++)
										node->m_PathPoints.push_back(pathpointarray[i]);

									// Stash it ...
									m_ShadingNodes.push_back(node);
								}
						}
				}
		}

	return SDP_S_OK;
}

sdpResult k3dMesh::UnshadePaths(const gulong State)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// For each shading node ...
	ShadingNodeIterator nextnode;
	ShadingNodeIterator end = m_ShadingNodes.end();
	for(ShadingNodeIterator node = m_ShadingNodes.begin(); node != end; node = nextnode)
		{
			nextnode = node;
			nextnode++;

			// Remove any paths that match our criteria ...
			(*node)->RemovePaths(State);

			// If the node is empty, get rid of it ...
			if(0 == (*node)->PathPointCount())
				{
					delete (*node);
					m_ShadingNodes.erase(node);
				}
		}

	return SDP_S_OK;
}

bool operator<(const sdpVector3& A, const sdpVector3& B)
{
	if(A[0] == B[0])
		{
			if(A[1] == B[1])
				return A[2] < B[2];
			else
				return A[1] < B[1];
		}

	return A[0] < B[0];
}

sdpResult k3dMesh::WeldPoints(const gulong Criteria)
{
	PushEditing();

	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Eliminate shading
	UnshadePaths(0);

	// Create a mapping of coordinates to pathpoints-that-share-them ...
	typedef std::map<sdpVector3, k3dPath::PathPoints> CoordPathPointMap;

	CoordPathPointMap coordpathpointmap;
	PathIterator path;
	for(path = m_Paths.begin(); path != m_Paths.end(); path++)
		{
			for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != (*path)->m_PathPoints.end(); pathpoint++)
				{
					if((*pathpoint)->m_Point->AdvancedQuery(Criteria))
						{
							coordpathpointmap[(*pathpoint)->m_Point->m_LocalXYZ].push_back(*pathpoint);
						}
				}
		}

	// Concatenate any points that share coordinates ...
	typedef std::set<k3dPoint*> PointSet;
	PointSet doomedpaths;

	for(CoordPathPointMap::iterator coordpathpoint = coordpathpointmap.begin(); coordpathpoint != coordpathpointmap.end(); coordpathpoint++)
		{
			for(k3dPath::PathPoints::iterator pathpoint = coordpathpoint->second.begin() + 1; pathpoint != coordpathpoint->second.end(); pathpoint++)
				{
					if((*pathpoint)->m_Point != coordpathpoint->second.front()->m_Point)
						{
							doomedpaths.insert((*pathpoint)->m_Point);
							(*pathpoint)->m_Point = coordpathpoint->second.front()->m_Point;
						}
				}
		}

	// Concatenate any adjacent pathpoints that share points ...
	k3dPath::PathPoints doomedpathpoints;
	for(path = m_Paths.begin(); path != m_Paths.end(); path++)
		{
			// Skip single-point paths ...
			if(1 == (*path)->m_PathPoints.size())
				continue;

			k3dPoint* lastpoint = 0;
			for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != (*path)->m_PathPoints.end(); pathpoint++)
				{
					if((*pathpoint)->m_Point->AdvancedQuery(Criteria))
						{
							if((*pathpoint)->m_Point == lastpoint)
								{
									doomedpathpoints.push_back(*pathpoint);
								}

							lastpoint = (*pathpoint)->m_Point;
						}
				}
		}

	// Get rid of our sacrificial lambs ...
	for(PointSet::iterator point = doomedpaths.begin(); point != doomedpaths.end(); point++)
		DeletePoint(*point);

	for(k3dPath::PathPoints::iterator pathpoint = doomedpathpoints.begin(); pathpoint != doomedpathpoints.end(); pathpoint++)
		(*pathpoint)->m_Path->RemovePathPoint(*pathpoint);

	PopEditing();

	return SDP_S_OK;
}

sdpResult k3dMesh::UnweldPoints(const gulong Criteria)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Eliminate shading
	UnshadePaths(0);

	// Create a mapping of selected-points to paths-that-share-them ...
	typedef std::map<k3dPoint*, std::vector<k3dPathPoint*> > PointPathMap;

	PointPathMap pointpathmap;
	for(PathIterator path = m_Paths.begin(); path != m_Paths.end(); path++)
		{
			for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != (*path)->m_PathPoints.end(); pathpoint++)
				{
					if((*pathpoint)->m_Point->AdvancedQuery(Criteria))
						{
							pointpathmap[(*pathpoint)->m_Point].push_back(*pathpoint);
						}
				}
		}

	// Create duplicate points so paths no longer share ...
	for(PointPathMap::iterator pointpath = pointpathmap.begin(); pointpath != pointpathmap.end(); pointpath++)
		{
			for(std::vector<k3dPathPoint*>::iterator pathpoint = pointpath->second.begin() + 1; pathpoint != pointpath->second.end(); pathpoint++)
				{
					k3dPoint* const newpoint = static_cast<k3dPoint*>(CreatePoint());

					newpoint->m_LocalXYZ = (*pathpoint)->m_Point->m_LocalXYZ;
					newpoint->m_TextureXYZ = (*pathpoint)->m_Point->m_TextureXYZ;
					newpoint->m_ImplicitTextureXYZ = (*pathpoint)->m_Point->m_ImplicitTextureXYZ;

					(*pathpoint)->m_Point = newpoint;
				}
		}

	return SDP_S_OK;
}

// Close a mesh (fill in holes)
/*gint operator < (const sdpVector3& a, const sdpVector3& b)
{
if(a.n[0] != b.n[0])
	return a.n[0] < b.n[0];
else
	if(a.n[1] != b.n[1])
		return a.n[1] < b.n[1];
	else
		return a.n[2] < b.n[2];
}
*/

typedef std::pair<gulong, gulong> Edge;
class MeshEdge :
	public Edge
{
public:
	MeshEdge(const gulong Index1, const gulong Index2, k3dMesh* const mesh, gulong first) :
		Edge(Index1 < Index2 ? Index1 : Index2, Index1 < Index2 ? Index2 : Index1),
		m_Mesh(mesh)
	{
		m_First = first;
	}

	gulong m_First;
	k3dMesh* const m_Mesh;
};

class nextedge
{
public:
	nextedge(MeshEdge const CurrentEdge, gulong const CurrentPoint) :
		m_CurrentEdge(CurrentEdge),
		m_CurrentPoint(CurrentPoint)
	{
	}

	bool operator()(MeshEdge const edge)
	{
		return (m_CurrentEdge != edge) && ((m_CurrentPoint == edge.first) || (m_CurrentPoint == edge.second));
	}

private:
	MeshEdge const m_CurrentEdge;
	gulong const m_CurrentPoint;
};


sdpResult k3dMesh::CloseMesh(const gulong Criteria)
{
	PushEditing();

	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Eliminate shading
	UnshadePaths(0);

	// Count how many faces each edge belongs to ...
	std::map<MeshEdge, gulong> edgetest;

	// Index points first ...
	std::map<k3dPoint*, gulong> pointlookup;
	std::map<gulong, k3dPoint*> indexlookup;
	gulong pointindex = 0;
	for(PointIterator point = m_Points.begin(); point != m_Points.end(); point++)
		{
			if((*point)->AdvancedQuery(Criteria))
				{
					pointlookup[*point] = pointindex;
					indexlookup[pointindex] = *point;
					pointindex++;
				}
		}

	// Loop through paths and store edges ...
	PathIterator path;
	for(path = m_Paths.begin(); path != m_Paths.end(); path++)
		{
			k3dPoint* firstpoint = (*((*path)->m_PathPoints.begin()))->m_Point;
			for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != (*path)->m_PathPoints.end(); pathpoint++)
				{
					if(!(*pathpoint)->m_Point->AdvancedQuery(Criteria))
						continue;

					// Create an edge ...
					k3dPoint* point1 = (*pathpoint)->m_Point;
					k3dPoint* point2;
					if((pathpoint+1) == (*path)->m_PathPoints.end())
						point2 = firstpoint;
					else
						point2 = (*(pathpoint+1))->m_Point;

					const gulong index1 = pointlookup[point1];
					const gulong index2 = pointlookup[point2];
					const MeshEdge edge(index1, index2, (*path)->m_Mesh, index1);

					// Count how many times we came across this edge ...
					if(edgetest.find(edge) != edgetest.end())
						edgetest[edge]++;
					else
						edgetest[edge] = 1;
				}
		}

	// Check edge count ...
	std::map<MeshEdge, gulong>::iterator test = edgetest.begin();
	bool hollow = false;
	std::set<MeshEdge> edges;
	for(; test != edgetest.end(); test++)
		switch((*test).second)
			{
				case 1:
					// Remember there's a hole and save lonely edge ...
					hollow = true;
					edges.insert((*test).first);
				break;
				case 2:
					// This edge's fine
				break;
				default:
					std::cerr << "An edge belongs to more than 2 faces??";
			}

	// Fill in holes ...
	gulong edgenumber = edges.size();
	while(hollow && edges.size())
		{
			// Pick out an edge and build a closing path ...
			std::set<MeshEdge>::iterator edge = edges.begin();
			gulong firstpoint = (*edge).first;
			gulong currentpoint = (*edge).second;

			// Start in the right direction ...
			if((*edge).m_First == firstpoint)
				std::swap(firstpoint, currentpoint);

			k3dPath* const newpath = static_cast<k3dPath*>((*edge).m_Mesh->CreatePath());
			edges.erase(edge);

			newpath->Points()->AddPoint(indexlookup[firstpoint]);
			newpath->Points()->AddPoint(indexlookup[currentpoint]);
			while(true)
				{
					edge = std::find_if(edges.begin(), edges.end(), nextedge(*edge, currentpoint));
					if(edge == edges.end())
						break;

					if((*edge).first == currentpoint)
						currentpoint = (*edge).second;
					else
						currentpoint = (*edge).first;

					edges.erase(edge);

					newpath->AddPoint(indexlookup[currentpoint]);
				}

			// Set default material
			//k3dIMaterial* defaultmaterial = k3dDefaultMaterial(Document());
			//newpath->Materials()->SetMaterial(defaultmaterial);

			// Quit if edge number didn't decrease
			if(edges.size() == edgenumber)
				break;
			else
				edgenumber = edges.size();
		}

	PopEditing();

	return SDP_S_OK;
}

sdpResult k3dMesh::CalculateNormals()
{
	// Calculate path normals ...
	k3dMeshCalculateNormal functor;
	std::for_each(m_Paths.begin(), m_Paths.end(), functor);

	// Temp storage ...
	sdpVector3 normal;
	k3dShadingNode::PathPointIterator pathpoint;
	k3dShadingNode::PathPointIterator pathpointend;

	// Integrate pathpoint vectors for smooth shading ...
	ShadingNodeIterator nodeend = m_ShadingNodes.end();
	for(ShadingNodeIterator node = m_ShadingNodes.begin(); node != nodeend; node++)
		{
			normal = 0.0;

			pathpointend = (*node)->m_PathPoints.end();
			for(pathpoint = (*node)->m_PathPoints.begin(); pathpoint != pathpointend; pathpoint++)
				normal += (*pathpoint)->Path()->m_Normal;

			for(pathpoint = (*node)->m_PathPoints.begin(); pathpoint != pathpointend; pathpoint++)
				(*pathpoint)->m_NormalXYZ = normal;
		}
	return SDP_S_OK;
}

// Return the dimensions of every point within the mesh that fits the given criteria ...
sdpResult k3dMesh::Dimensions(gulong Criteria, double Min[3], double Max[3])
{
	Max[0] = Max[1] = Max[2] = -DBL_MAX;
	Min[0] = Min[1] = Min[2] = DBL_MAX;

	gulong count = 0;
	gulong end = m_Points.size();
	for(gulong i = 0; i < end; i++)
		{
			if(SDP_S_OK != m_Points[i]->State()->QueryState(Criteria))
				continue;

			count++;

			if(m_Points[i]->m_LocalXYZ[0] > Max[0])
				Max[0] = m_Points[i]->m_LocalXYZ[0];
			if(m_Points[i]->m_LocalXYZ[0] < Min[0])
				Min[0] = m_Points[i]->m_LocalXYZ[0];
			if(m_Points[i]->m_LocalXYZ[1] > Max[1])
				Max[1] = m_Points[i]->m_LocalXYZ[1];
			if(m_Points[i]->m_LocalXYZ[1] < Min[1])
				Min[1] = m_Points[i]->m_LocalXYZ[1];
			if(m_Points[i]->m_LocalXYZ[2] > Max[2])
				Max[2] = m_Points[i]->m_LocalXYZ[2];
			if(m_Points[i]->m_LocalXYZ[2] < Min[2])
				Min[2] = m_Points[i]->m_LocalXYZ[2];
		}

	if(!count)
		{
			Min[0] = Min[1] = Min[2] = 0.0;
			Max[0] = Max[1] = Max[2] = 0.0;
		}

	return SDP_S_OK;
}

const sdpVector3& k3dMesh::Color() const
{
	return m_Color.Value();
}

void k3dMesh::SetColor(const sdpVector3& Color)
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	m_Color.SetValue(Color);
}

///////////////////////////////////////////////////////////////
// k3dIPathCollection implementation

void k3dMesh::VisitPaths(k3dIPathVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	std::for_each(m_Paths.begin(), m_Paths.end(), sdp::MakeVisitorAdapter(Visitor, &k3dIPathVisitor::VisitPath));
}

sdpResult k3dMesh::AddPath(k3dIPath* const Path)
{
	g_return_val_if_fail(0, SDP_S_OK); // Not implemented for meshes
}

sdpResult k3dMesh::RemovePath(k3dIPath* const Path)
{
	g_return_val_if_fail(0, SDP_S_OK); // Not implemented for meshes
}

gulong k3dMesh::PathCount()
{
	return m_Paths.size();
}

K3DPATHHANDLE k3dMesh::FirstPath()
{
	return K3DPATHHANDLE(m_Paths.size() ? 1 : 0);
}

K3DPATHHANDLE k3dMesh::NextPath(K3DPATHHANDLE Path)
{
	// Sanity checks ...
	g_return_val_if_fail(Path, 0);
	g_return_val_if_fail(gulong(Path) <= m_Paths.size(), 0);

	return K3DPATHHANDLE(gulong(Path) < m_Paths.size() ? gulong(Path)+1 : 0);
}

k3dIPath* k3dMesh::Path(K3DPATHHANDLE Path)
{
	// Sanity checks ...
	g_return_val_if_fail(Path, 0);
	g_return_val_if_fail(gulong(Path) <= m_Paths.size(), 0);

	return m_Paths[gulong(Path)-1];
}

/////////////////////////////////////////////////////////////////
// k3dIPointCollection implementation

void k3dMesh::VisitPoints(k3dIPointVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	std::for_each(m_Points.begin(), m_Points.end(), sdp::MakeVisitorAdapter(Visitor, &k3dIPointVisitor::VisitPoint));
}

sdpResult k3dMesh::AddPoint(k3dIPoint* Point)
{
	g_return_val_if_fail(0, SDP_S_OK); // Not implemented for meshes ...
}

sdpResult k3dMesh::RemovePoint(k3dIPoint* Point)
{
	g_return_val_if_fail(0, SDP_S_OK); // Not implemented for meshes ...
}

gulong k3dMesh::PointCount()
{
	return m_Points.size();
}

K3DPOINTHANDLE k3dMesh::FirstPoint()
{
	return K3DPOINTHANDLE(m_Points.size() ? 1 : 0);
}

K3DPOINTHANDLE k3dMesh::NextPoint(K3DPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_Points.size(), 0);

	return K3DPOINTHANDLE(gulong(Point) < m_Points.size() ? gulong(Point)+1 : 0);
}

k3dIPoint* k3dMesh::Point(K3DPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_Points.size(), 0);

	return m_Points[gulong(Point)-1];
}

////////////////////////////////////////////////////////
// k3dILocationCollection implementation ...

void k3dMesh::VisitLocations(k3dILocationVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	std::for_each(m_Points.begin(), m_Points.end(), sdp::MakeVisitorAdapter(Visitor, &k3dILocationVisitor::VisitLocation));
}

gulong k3dMesh::LocationCount()
{
	return PointCount();
}

K3DLOCATIONHANDLE k3dMesh::FirstLocation()
{
	return K3DLOCATIONHANDLE(m_Points.size() ? 1 : 0);
}

K3DLOCATIONHANDLE k3dMesh::NextLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_Points.size(), 0);

	return K3DLOCATIONHANDLE(gulong(Location) < m_Points.size() ? gulong(Location)+1 : 0);
}

k3dILocation* k3dMesh::Location(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_Points.size(), 0);

	return m_Points[gulong(Location)-1];
}

////////////////////////////////////////////////////////
// k3dIEditorRender implementation ...

/* Keep thinking I'll find a use for this...

	static GLubyte halftonea[] = {
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55
		};

	static GLubyte halftoneb[] = {
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa,
		0x55, 0x55, 0x55, 0x55, 0xaa, 0xaa, 0xaa, 0xaa
		}

	glPolygonStipple(halftone);
	glEnable(GL_POLYGON_STIPPLE);

*/

sdpResult k3dMesh::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	StoreOpenGLAttributes attributes;

	// Draw points if in point mode ...
	if(k3dICamera::POINT == Engine->Camera()->ActiveEditMode())
		{
			// Draw selected geometry ...
			glColor3dv(const_cast<sdpVector3&>(m_SelectedColor));
			EditorDrawPoints(k3dIState::VISIBLESELECTED, 0);

			// Draw unselected geometry ...
			glColor3dv(const_cast<sdpVector3&>(m_Color.Value()));
			EditorDrawPoints(k3dIState::VISIBLE, k3dIState::SELECTED);
		}

	// Draw solid geometry ...
	if(SDP_S_OK == Engine->EditorQuerySolid())
		{
			// Offset solid geometry, so it's cleanly visible ...
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonOffset(1.0, 1.0);

			if(SDP_S_OK == Engine->EditorQueryTextured())
				EditorDrawSolidTexturedPaths(Engine, k3dIState::VISIBLE, 0);
			else
				EditorDrawSolidPaths(Engine, k3dIState::VISIBLE, 0);

			glDisable(GL_POLYGON_OFFSET_FILL);
		}

	// Draw path orientations ...
	if(SDP_S_OK == Engine->EditorQueryOrientation())
		{
			glColor3dv(const_cast<sdpVector3&>(m_SelectedColor));
			EditorDrawOrientations(k3dIState::VISIBLESELECTED, 0);
		}

	// Draw wireframe geometry ...
	if(SDP_S_OK == Engine->EditorQueryWireframe())
		{
			// Draw selected geometry ...
			glColor3dv(const_cast<sdpVector3&>(m_SelectedColor));
			EditorDrawWireframePaths(Engine, k3dIState::VISIBLESELECTED, 0);

			// Draw unselected geometry ...
			glColor3dv(const_cast<sdpVector3&>(m_Color.Value()));
			EditorDrawWireframePaths(Engine, k3dIState::VISIBLE, k3dIState::SELECTED);
		}

	// Draw path normals ...
	if(SDP_S_OK == Engine->EditorQueryNormals())
		{
			glColor3dv(const_cast<sdpVector3&>(m_PathNormalColor));
			EditorDrawNormals(k3dIState::VISIBLESELECTED, 0);
		}

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, this, 1.0);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// Rendering helpers ...

void k3dMesh::EditorDrawPoints(const gulong Set, const gulong Unset)
{
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	// For each point ...
	glBegin(GL_POINTS);

	const gulong size = m_Points.size();
	for(gulong i = 0; i < size; i++)
		{
			if(m_Points[i]->AdvancedQuery(Set, Unset))
				glVertex3dv(m_Points[i]->m_AnimXYZ.n);
		}

	glEnd();
}

void k3dMesh::EditorDrawNormals(const gulong Set, const gulong Unset)
{
	StoreOpenGLAttributes attributes;

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glLineWidth(1.0f);

	double normallength, finallength;

	sdpVector3 center, coords;

	// For each path ...
	const gulong size = m_Paths.size();
	for(gulong i = 0; i < size; i++)
		{
			// Ignore paths with two-or-fewer points ...
			if(m_Paths[i]->PointCount() <= 2)
				continue;

			// Ignore paths that don't fit our criteria ...
			if(m_Paths[i]->AdvancedQuery(Set, Unset))
				{
					normallength = m_Paths[i]->m_Normal.Length();
					if(normallength <= 0.0)
						continue;

					// Calculate bounding box for this path
					k3dBoundingBox3 boundingbox;
					for(gulong j = 0; j < m_Paths[i]->m_PathPoints.size(); j++)
						boundingbox.Expand(m_Paths[i]->m_PathPoints[j]->m_Point->m_AnimXYZ);

					finallength = 0.5 * boundingbox.Size().Length();

					normallength = finallength / normallength;

					center = boundingbox.Center();
					coords = m_Paths[i]->m_Normal * normallength;
					coords += center;

					glBegin(GL_LINES);
					glVertex3dv(center);
					glVertex3dv(coords);
					glEnd();
				}
		}
}

void k3dMesh::EditorDrawOrientations(const gulong Set, const gulong Unset)
{
	StoreOpenGLAttributes attributes;

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glLineWidth(1.0f);

	// For each path ...
	const gulong size = m_Paths.size();
	for(gulong i = 0; i < size; i++)
		{
			if(m_Paths[i]->AdvancedQuery(Set, Unset))
				m_Paths[i]->DrawOpenGLOrientation();
		}
}

void k3dMesh::EditorDrawWireframePaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset)
{
	// Wireframe drawing ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	if(SDP_S_OK == Engine->EditorQueryWireframeCulling())
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	const gulong size = m_Paths.size();
	for(gulong i = 0; i < size; i++)
		{
			if(m_Paths[i]->AdvancedQuery(Set, Unset))
				m_Paths[i]->DrawOpenGLWireframe();
		}
}

void k3dMesh::EditorDrawSolidPaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset)
{
	static GLubyte halftone[] = {
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55
		};

	static GLfloat black[4] = { 0, 0, 0, 1 };
	static GLfloat white[4] = { 1, 1, 1, 1 };

	k3dIMaterial* material = 0;
	k3dIMaterial* lastmaterial = reinterpret_cast<k3dIMaterial*>(-1);
	k3dIEditorMaterial* editormaterial = 0;

	// Solid drawing ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	if(SDP_S_OK == Engine->EditorQuerySolidCulling())
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	// Draw paths ...
	const gulong size = m_Paths.size();
	for(gulong i = 0; i < size; i++)
		{
			if(m_Paths[i]->AdvancedQuery(Set, Unset))
				{
					// See if we need to set up a new material ...
					material = m_Paths[i]->m_Material;
					if(material != lastmaterial)
						{
							lastmaterial = material;
							if(material) // This path has a material assigned ...
								{
									editormaterial = dynamic_cast<k3dIEditorMaterial*>(material);
									if(editormaterial)
										editormaterial->EditorMaterialSetup(Engine);

									glDisable(GL_POLYGON_STIPPLE);
								}
							else // No material assigned, so create a default stippled effect ...
								{
									glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, black);
									glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, white);
									glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, black);
									glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, black);

									glPolygonStipple(halftone);
									glEnable(GL_POLYGON_STIPPLE);
								}
						}

					// Draw the path ...
					m_Paths[i]->DrawOpenGLSolid();
				}
		}
}

void k3dMesh::EditorDrawSolidTexturedPaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset)
{
	static GLubyte halftone[] = {
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55,
		0xaa, 0xaa, 0xaa, 0xaa, 0x55, 0x55, 0x55, 0x55
		};

	static GLfloat black[4] = { 0, 0, 0, 1 };
	static GLfloat white[4] = { 1, 1, 1, 1 };

	k3dIMaterial* material = 0;
	k3dIMaterial* lastmaterial = reinterpret_cast<k3dIMaterial*>(-1);
	k3dIEditorMaterial* editormaterial = 0;

	// Solid drawing ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_1D);
	glEnable(GL_TEXTURE_2D);

	if(SDP_S_OK == Engine->EditorQuerySolidCulling())
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);


	// Draw paths ...
	const gulong size = m_Paths.size();
	for(gulong i = 0; i < size; i++)
		{
			if(m_Paths[i]->AdvancedQuery(Set, Unset))
				{
					// See if we need to set up a new material ...
					material = m_Paths[i]->m_Material;
					if(material != lastmaterial)
						{
							lastmaterial = material;
							if(material) // This path has a material assigned ...
								{
									editormaterial = dynamic_cast<k3dIEditorMaterial*>(material);
									if(editormaterial)
										editormaterial->EditorMaterialSetup(Engine);

									glDisable(GL_POLYGON_STIPPLE);
								}
							else // No material assigned, so create a default stippled effect ...
								{
									glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, black);
									glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, white);
									glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, black);
									glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, black);

									glPolygonStipple(halftone);
									glEnable(GL_POLYGON_STIPPLE);
								}
						}

					// Draw the path ...
					m_Paths[i]->DrawOpenGLSolidTextured();
				}
		}
}

///////////////////////////////////////////////////////////////
// k3dIMaterialServer implementation ...

k3dIMaterial* k3dMesh::Material()
{
	return m_Material.Interface();
}

sdpResult k3dMesh::SetMaterial(k3dIMaterial* Material)
{
	// We allow NULL here, as a synonym for the default material ...
	m_Material.SetObject(Material);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditable implementation

void k3dMesh::PushEditing()
{
	if(!Editing())
		{
			m_PointCollectionChanged = false;
			m_PathCollectionChanged = false;
		}

	k3dMeshBase::PushEditing();
	m_Color.PushEditing();
}

void k3dMesh::PopEditing()
{
	m_Color.PopEditing();
	k3dMeshBase::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_PointCollectionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<PointCollection>(m_Points));

					if(m_PathCollectionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<PathCollection>(m_Paths));
				}
		}

}

///////////////////////////////////////////////////////////////
// k3dIRiRender implementation

sdpResult k3dMesh::RiPushRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	k3dMeshBase::RiPushRender(Frame, Engine);

	if(m_RiSolid)
		Engine->RiSolidBegin(k3dIRiEngine::RI_PRIMITIVE);

	return SDP_S_OK;
}

sdpResult k3dMesh::RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	// Setup our material interfaces ...
	k3dIMaterial* material = 0;
	k3dIMaterial* lastmaterial = (k3dIMaterial*)-1;
	k3dIRiMaterial* rimaterial = 0;

	switch(m_RenderMode.Value())
	{
		case SUBDIVISION:
			{
				// Get the mesh material ...
				material = m_Material.Interface();
				if(material)
					{
						rimaterial = dynamic_cast<k3dIRiMaterial*>(material);
						if(rimaterial)
							{
								rimaterial->RiMaterialSetup(Frame,Engine);
							}
					}
				RiSubdivisionMesh(Engine, rimaterial);
			}
			break;

		default:
			{
				// For each path ...
				k3dPath* path;
				for(gulong i = 0; i < m_Paths.size(); i++)
					{
						path = m_Paths[i];

						// Skip hidden or unrenderable paths
						if(SDP_S_OK != path->QueryState(k3dIState::VISIBLE))
							continue;

						// Get the path material ...
						material = path->m_Material;
						if(material)
							{
								if(material != lastmaterial)
									{
										rimaterial = dynamic_cast<k3dIRiMaterial*>(material);
										if(rimaterial)
											{
												lastmaterial = material;
												rimaterial->RiMaterialSetup(Frame, Engine);
											}
									}
							}

						// Render the path ...
						RiPath(Engine, rimaterial, path);
					}
			}
	}

	return SDP_S_OK;
}

// Render a path ...
void k3dMesh::RiPath(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path)
{
	switch(Path->m_PathPoints.size())
		{
			case 1:
			case 2:
				return;

			case 3:
				RiPath3(Engine, Material, Path);
				return;

			case 4:
				RiPath4(Engine, Material, Path);
				return;

			default:
				RiPathN(Engine, Material, Path);
				return;
		}
}

// Render a three-sided path as a bilinear patch ...
void k3dMesh::RiPath3(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path)
{
	// Arrays of coordinates ...
	k3dRiRealArray* points = new k3dRiRealArray("point");
	k3dRiRealArray* normals = new k3dRiRealArray("normal");
	k3dRiRealArray* scoords = new k3dRiRealArray("float");
	k3dRiRealArray* tcoords = new k3dRiRealArray("float");

	k3dPathPoint* middlepoint = 0;
	for(gulong i = 0; i < 3; i++)
		{
			k3dPathPoint* pathpoint = Path->m_PathPoints[i];

			if(1 == i)
				middlepoint = pathpoint;

			points->Append(pathpoint->m_Point->m_AnimXYZ);
			normals->Append(pathpoint->m_NormalXYZ);
			scoords->Append(pathpoint->m_TextureXYZ[0]);
			tcoords->Append(pathpoint->m_TextureXYZ[1]);
		}

	points->Append(middlepoint->m_Point->m_AnimXYZ);
	normals->Append(middlepoint->m_NormalXYZ);
	scoords->Append(middlepoint->m_TextureXYZ[0]);
	tcoords->Append(middlepoint->m_TextureXYZ[1]);

	k3dRiParameterList parameters;
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_P, points));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_N, normals));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_S, scoords));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_T, tcoords));

	Engine->RiPatchV(k3dIRiEngine::RI_BILINEAR, parameters.ParameterCount(), parameters.Tokens(), parameters.Values(), parameters.ValueCounts());
}

// Render a four-sided path as a bilinear patch ...
void k3dMesh::RiPath4(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path)
{
	static gint bilinearpatchlookup[4] = {0, 1, 3, 2};

	// Arrays of coordinates ...
	k3dRiRealArray* points = new k3dRiRealArray("point");
	k3dRiRealArray* normals = new k3dRiRealArray("normal");
	k3dRiRealArray* scoords = new k3dRiRealArray("float");
	k3dRiRealArray* tcoords = new k3dRiRealArray("float");

	for(gulong i = 0; i < 4; i++)
		{
			k3dPathPoint* pathpoint = Path->m_PathPoints[bilinearpatchlookup[i]];

			points->Append(pathpoint->m_Point->m_AnimXYZ);
			normals->Append(pathpoint->m_NormalXYZ);
			scoords->Append(pathpoint->m_TextureXYZ[0]);
			tcoords->Append(pathpoint->m_TextureXYZ[1]);
		}

	k3dRiParameterList parameters;
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_P, points));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_N, normals));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_S, scoords));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_T, tcoords));

	Engine->RiPatchV(k3dIRiEngine::RI_BILINEAR, parameters.ParameterCount(), parameters.Tokens(), parameters.Values(), parameters.ValueCounts());
}

// Render a greater-than-four-sided path ...
void k3dMesh::RiPathN(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path)
{
	// Arrays of coordinates ...
	k3dRiRealArray* points = new k3dRiRealArray("point");
	k3dRiRealArray* normals = new k3dRiRealArray("normal");
	k3dRiRealArray* scoords = new k3dRiRealArray("float");
	k3dRiRealArray* tcoords = new k3dRiRealArray("float");

	for(gulong i = 0; i < Path->m_PathPoints.size(); i++)
		{
			k3dPathPoint* pathpoint = Path->m_PathPoints[i];

			points->Append(pathpoint->m_Point->m_AnimXYZ);
			normals->Append(pathpoint->m_NormalXYZ);
			scoords->Append(pathpoint->m_TextureXYZ[0]);
			tcoords->Append(pathpoint->m_TextureXYZ[1]);
		}

	k3dRiParameterList parameters;
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_P, points));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_N, normals));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_S, scoords));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_T, tcoords));

	gulong pointcount = Path->m_PathPoints.size();
	Engine->RiGeneralPathV(1, &pointcount, parameters.ParameterCount(), parameters.Tokens(), parameters.Values(), parameters.ValueCounts());
}

// Render a subdivisionmesh
void k3dMesh::RiSubdivisionMesh(k3dIRiEngine* const Engine, k3dIRiMaterial* Material)
{
	// If we don't have any points or paths, we're done ...
	if(!m_Points.size())
		return;

	if(!m_Paths.size())
		return;

	// Iterators (#@$! M$ compiler!!!)
	PointIterator point;
	PathIterator path;

	// Build a map of point IDs to indexes so that we can build the path table.
	std::map<k3dPoint*, gulong> indexmap;
	for(point = m_Points.begin(); point != m_Points.end(); point++)
	{
		gulong index = indexmap.size();
		indexmap[*point] = index;
	}

	// Build the Vertex Counts array
	std::vector<gulong> vertexcounts;
	for(path = m_Paths.begin(); path != m_Paths.end(); path++)
		vertexcounts.push_back((*path)->m_PathPoints.size());

	// Build the Vertices array
	std::vector<gulong> vertices;
	for(path = m_Paths.begin(); path != m_Paths.end(); path++)
		{
			for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != (*path)->m_PathPoints.end(); pathpoint++)
				vertices.push_back(indexmap[(*pathpoint)->m_Point]);
		}

	// Arrays of coordinates ...
	k3dRiRealArray* points = new k3dRiRealArray("point");
	k3dRiRealArray* scoords = new k3dRiRealArray("float");
	k3dRiRealArray* tcoords = new k3dRiRealArray("float");

	for(point = m_Points.begin(); point != m_Points.end(); point++)
		{
			points->Append((*point)->m_AnimXYZ);
			scoords->Append((*point)->m_TextureXYZ[0]);
			tcoords->Append((*point)->m_TextureXYZ[1]);

		}

	k3dRiParameterList parameters;
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_P, points));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_S, scoords));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_T, tcoords));

	Engine->RiSubdivisionMeshV(sdpString("catmull-clark"), m_Paths.size(), &vertexcounts[0], &vertices[0], 0, 0, 0, 0, 0, parameters.ParameterCount(), parameters.Tokens(), parameters.Values(), parameters.ValueCounts());
}


sdpResult k3dMesh::RiPopRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	if(m_RiSolid)
		Engine->RiSolidEnd();

	return k3dMeshBase::RiPopRender(Frame, Engine);
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult k3dMesh::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	// Call the base-class implementation ...
	g_return_val_if_fail(sdpSucceeded(k3dMeshBase::Save(Element, Dependencies)), SDP_E_FAIL);

	// Save our geometry ...
	g_return_val_if_fail(sdpSucceeded(SaveGeometry(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dMesh::SaveGeometry(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...

	g_assert(Dependencies);

	// Save points ...
	g_return_val_if_fail(sdpSucceeded(SavePoints(Element, Dependencies)), SDP_E_FAIL);
	// Save paths ...
	g_return_val_if_fail(sdpSucceeded(SavePaths(Element, Dependencies)), SDP_E_FAIL);
	// Save shading data ...
	g_return_val_if_fail(sdpSucceeded(SaveShading(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dMesh::SavePoints(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& points = Element.Append(sdpxml::Element("points"));

	for(PointIterator point = m_Points.begin(); point != m_Points.end(); point++)
		(*point)->Save(points, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::SavePaths(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& paths = Element.Append(sdpxml::Element("paths"));

	for(PathIterator path = m_Paths.begin(); path != m_Paths.end(); path++)
		(*path)->Save(paths, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::SaveShading(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& shading = Element.Append(sdpxml::Element("shading"));

	for(ShadingNodeIterator node = m_ShadingNodes.begin(); node != m_ShadingNodes.end(); node++)
		(*node)->Save(shading, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::Load(sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Call the base-class implementation ...
	g_return_val_if_fail(sdpSucceeded(k3dMeshBase::Load(Document, Element)), SDP_E_FAIL);

	// Load geometry ...
	g_return_val_if_fail(sdpSucceeded(LoadGeometry(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dMesh::LoadGeometry(sdpxml::Element& Element)
{
	PushEditing();

	// Sanity checks ...
	g_return_val_if_fail(sdpSucceeded(LoadPoints(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(LoadPaths(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(LoadShading(Element)), SDP_E_FAIL);

	PopEditing();

	return SDP_S_OK;
}

/// Functor for loading points from an XML document
class k3dMeshLoadPoint
{
public:
	k3dMeshLoadPoint(k3dMesh* const Mesh) :
		m_Mesh(Mesh)
	{
		// Sanity checks ...
		g_assert(m_Mesh);
	}

	void operator()(sdpxml::Element& Element)
	{
		k3dPoint* const point = static_cast<k3dPoint*>(m_Mesh->CreatePoint());
		point->Load(Element);
	}

private:
	k3dMesh* const m_Mesh;
};

sdpResult k3dMesh::LoadPoints(sdpxml::Element& Element)
{
	const sdpxml::ElementCollection::iterator points = std::find_if(Element.Children().begin(), Element.Children().end(), sdpxml::SameName("points"));
	if(points == Element.Children().end())
		return SDP_S_FALSE;

	k3dMeshLoadPoint functor(this);
	std::for_each(points->Children().begin(), points->Children().end(), functor);

	return SDP_S_OK;
}

/// Functor for loading paths from an XML document
class k3dMeshLoadPath
{
public:
	k3dMeshLoadPath(k3dMesh* const Mesh) :
		m_Mesh(Mesh)
	{
		// Sanity checks ...
		g_assert(m_Mesh);
	}

	void operator()(sdpxml::Element& Element)
	{
		k3dPath* const path = static_cast<k3dPath*>(m_Mesh->CreatePath());
		path->Load(Element, m_Mesh);
	}

private:
	k3dMesh* const m_Mesh;
};

sdpResult k3dMesh::LoadPaths(sdpxml::Element& Element)
{
	const sdpxml::ElementCollection::iterator paths = std::find_if(Element.Children().begin(), Element.Children().end(), sdpxml::SameName("paths"));
	if(paths == Element.Children().end())
		return SDP_S_FALSE;

	k3dMeshLoadPath functor(this);
	std::for_each(paths->Children().begin(), paths->Children().end(), functor);

	return SDP_S_OK;
}

/// Functor object used to load shading nodes
class k3dMeshLoadShadingNode
{
public:
	k3dMeshLoadShadingNode(k3dMesh* const Mesh) :
		m_Mesh(Mesh)
	{
		// Sanity checks ...
		g_assert(m_Mesh);
	}

	void operator()(sdpxml::Element& Element)
	{
		// Create a new shading node ...
		k3dShadingNode* const node = new k3dShadingNode();
		m_Mesh->m_ShadingNodes.push_back(node);

		node->Load(Element, m_Mesh);
	}

private:
	k3dMesh* const m_Mesh;
};

sdpResult k3dMesh::LoadShading(sdpxml::Element& Element)
{
	const sdpxml::ElementCollection::iterator shading = std::find_if(Element.Children().begin(), Element.Children().end(), sdpxml::SameName("shading"));
	if(shading == Element.Children().end())
		return SDP_S_FALSE;

	k3dMeshLoadShadingNode functor(this);
	std::for_each(shading->Children().begin(), shading->Children().end(), functor);

	return SDP_S_OK;
}

sdpResult k3dMesh::LoadComplete()
{
	// Call the base-class version ...
	g_return_val_if_fail(sdpSucceeded(k3dMeshBase::LoadComplete()), SDP_E_FAIL);

	// For each path ...
	for(PathIterator path = m_Paths.begin(); path != m_Paths.end(); path++)
		(*path)->LoadComplete();

	return SDP_S_OK;
}

/*
k3dShadingNode* k3dMesh::CreateShadingNode()
{
	k3dShadingNode* newnode = m_ShadingNodes.create();
	g_return_val_if_fail(newnode, NULL);

	newnode->m_Mesh = this;

	return newnode;
}

void k3dMesh::DestroyShadingNode(k3dShadingNode* Node)
{
	g_assert(Node);
	m_ShadingNodes.erase(Node);
}

void k3dMesh::MoveSelectedToLast()
{
	k3dPath* nextpath;
	k3dPath* lastpath = m_Paths.back();
	for(k3dPath* path = m_Paths.front(); path != lastpath; path = nextpath)
		{
			nextpath = m_Paths[i]->m_Next;

			if(SDP_S_OK != m_Paths[i]->QueryState(k3dIState::VISIBLESELECTED)) continue;
			m_Paths[i]->SetState(k3dIState::VISIBLESELECTED, FALSE);

			m_Paths.moveback(path);
		}
}

void k3dMesh::UpdateNumbers()
{
	if(m_Sheet)
		m_Sheet->m_MeshStatsPage->UpdateNumbers();
}

void k3dMesh::DeepSelect(bool State)
{
	m_ObjectContext->Object()->State()->SetState(k3dIState::SELECTED, State);

	ForEachPath(path->SetState(k3dIState::SELECTED, State));
	ForEachPoint(m_Points[i]->SetState(k3dIState::SELECTED, State));
}

void k3dMesh::DeleteIfEmpty()
{
	if(m_Points.size()) return;
	if(m_Paths.size()) return;

	// If the document is being closed, all bets are off ...
	if(SDP_S_OK == Document()->DocumentClosing()) return;

	sdpString title;
	title.format("%s Mesh:", m_ObjectContext->Name());
	if(SDP_S_OK != m_IO->Query("The Mesh is empty.  Delete? (No undo)", title)) return;

	k3dDeleteObject(this);
}


#define ONETHIRD	0.333333333
#define TWOTHIRDS	0.666666667


void CDocument::OnEditPathsPatches()
{
	CLayer* layer;
	k3dMesh* mesh;
	k3dPath* path;
	k3dPoint *point1, *point2, *point3, *point4;
	k3dPathPoint *pathpoint, *firstpathpoint, *nextpathpoint, *lastpathpoint;
	k3dPathPoint *controlpoint1, *controlpoint2, *controlpoint3, *controlpoint4;

	k3dBoundingBox3 bbox, texturebbox;
	sdpVector3 coords[4], texturecoords[4];
	sdpVector3 center, texturecenter;

	for(layer = m_LayerList.front(); layer; layer = layer->m_Next)
		{
			for(mesh = layer->m_Meshes.front(); mesh; mesh = mesh->m_Next)
				{
					for(path = mesh->m_Paths.front(); path; path = m_Paths[i]->m_Next)
						{
							if(SDP_S_OK != m_Paths[i]->QueryState(k3dIState::VISIBLE)) continue;
							if(SDP_S_OK != m_Paths[i]->QueryState(k3dIState::SELECTED)) continue;
							if(SDP_S_OK != m_Paths[i]->QueryClosed()) continue;
							if(SDP_S_OK != m_Paths[i]->QueryPath()) continue;
							if(4 != m_Paths[i]->GetPointCount()) continue;

							m_Paths[i]->SetClosed(TRUE);
							m_Paths[i]->SetBezierPatch();

							gulong i = 0;
							bbox.Init();
							texturebbox.Init();
							firstpathpoint = m_Paths[i]->m_PathPoints.front();

							for(pathpoint = m_Paths[i]->m_PathPoints.front(); pathpoint;)
								{
									nextpathpoint = pathpoint->m_Next;

									bbox.Expand(pathpoint->m_Point->m_LocalXYZ);
									texturebbox.Expand(pathpoint->m_TextureXYZ);

									coords[i] = pathpoint->m_Point->m_LocalXYZ;
									texturecoords[i] = pathpoint->m_TextureXYZ;

									controlpoint1 = m_Paths[i]->CreatePathPoint(NULL);
									controlpoint2 = m_Paths[i]->CreatePathPoint(NULL);

									point1 = mesh->CreatePoint();
									point2 = mesh->CreatePoint();

									controlpoint1->m_Point = point1;
									controlpoint2->m_Point = point2;


									if(nextpathpoint)
										{
											point1->m_LocalXYZ = point1->m_LocalXYZ = pathpoint->m_Point->m_LocalXYZ + ONETHIRD *	(nextpathpoint->m_Point->m_LocalXYZ - pathpoint->m_Point->m_LocalXYZ);
											point2->m_LocalXYZ = point2->m_LocalXYZ = pathpoint->m_Point->m_LocalXYZ + TWOTHIRDS * (nextpathpoint->m_Point->m_LocalXYZ - pathpoint->m_Point->m_LocalXYZ);
											controlpoint1->m_TextureXYZ = pathpoint->m_TextureXYZ + ONETHIRD *	(nextpathpoint->m_TextureXYZ - pathpoint->m_TextureXYZ);
											controlpoint2->m_TextureXYZ = pathpoint->m_TextureXYZ + TWOTHIRDS * (nextpathpoint->m_TextureXYZ - pathpoint->m_TextureXYZ);
										}
									else
										{
											point1->m_LocalXYZ = point1->m_LocalXYZ = pathpoint->m_Point->m_LocalXYZ + ONETHIRD *	(firstpathpoint->m_Point->m_LocalXYZ - pathpoint->m_Point->m_LocalXYZ);
											point2->m_LocalXYZ = point2->m_LocalXYZ = pathpoint->m_Point->m_LocalXYZ + TWOTHIRDS * (firstpathpoint->m_Point->m_LocalXYZ - pathpoint->m_Point->m_LocalXYZ);
											controlpoint1->m_TextureXYZ = pathpoint->m_TextureXYZ + ONETHIRD *	(firstpathpoint->m_TextureXYZ - pathpoint->m_TextureXYZ);
											controlpoint2->m_TextureXYZ = pathpoint->m_TextureXYZ + TWOTHIRDS * (firstpathpoint->m_TextureXYZ - pathpoint->m_TextureXYZ);
										}

									pathpoint->m_Next = controlpoint1;
									controlpoint1->m_Next = controlpoint2;
									controlpoint2->m_Next = nextpathpoint;

									pathpoint = nextpathpoint;
									lastpathpoint = controlpoint2;
									i++;
								}

							controlpoint1 = m_Paths[i]->CreatePathPoint(NULL);
							controlpoint2 = m_Paths[i]->CreatePathPoint(NULL);
							controlpoint3 = m_Paths[i]->CreatePathPoint(NULL);
							controlpoint4 = m_Paths[i]->CreatePathPoint(NULL);

							point1 = mesh->CreatePoint();
							point2 = mesh->CreatePoint();
							point3 = mesh->CreatePoint();
							point4 = mesh->CreatePoint();

							center = bbox.Center();
							texturecenter = texturebbox.Center();

							point1->m_LocalXYZ = point1->m_LocalXYZ = ((coords[0] - center) * ONETHIRD) + center;
							point2->m_LocalXYZ = point2->m_LocalXYZ = ((coords[1] - center) * ONETHIRD) + center;
							point3->m_LocalXYZ = point3->m_LocalXYZ = ((coords[2] - center) * ONETHIRD) + center;
							point4->m_LocalXYZ = point4->m_LocalXYZ = ((coords[3] - center) * ONETHIRD) + center;

							controlpoint1->m_TextureXYZ = ((texturecoords[0] - texturecenter) * ONETHIRD) + texturecenter;
							controlpoint2->m_TextureXYZ = ((texturecoords[1] - texturecenter) * ONETHIRD) + texturecenter;
							controlpoint3->m_TextureXYZ = ((texturecoords[2] - texturecenter) * ONETHIRD) + texturecenter;
							controlpoint4->m_TextureXYZ = ((texturecoords[3] - texturecenter) * ONETHIRD) + texturecenter;

							controlpoint1->m_Point = point1;
							controlpoint2->m_Point = point2;
							controlpoint3->m_Point = point3;
							controlpoint4->m_Point = point4;

							lastpathpoint->m_Next = controlpoint1;
							controlpoint1->m_Next = controlpoint2;
							controlpoint2->m_Next = controlpoint3;
							controlpoint3->m_Next = controlpoint4;
							controlpoint4->m_Next = NULL;
						}
				}
		}

	RedrawAllView(false);
}

void CDocument::OnEditPathsSplines()
{
	CLayer* layer;
	k3dMesh* mesh;
	k3dPath* path;
	k3dPoint *point1, *point2;
	k3dPathPoint *pathpoint, *firstpathpoint, *nextpathpoint, *controlpoint1, *controlpoint2;

	for(layer = m_LayerList.front(); layer; layer = layer->m_Next)
		{
			for(mesh = layer->m_Meshes.front(); mesh; mesh = mesh->m_Next)
				{
					for(path = mesh->m_Paths.front(); path; path = m_Paths[i]->m_Next)
						{
							if(SDP_S_OK != m_Paths[i]->QueryState(k3dIState::VISIBLE)) continue;
							if(SDP_S_OK != m_Paths[i]->QueryState(k3dIState::SELECTED)) continue;
							if(SDP_S_OK != m_Paths[i]->QueryPath()) continue;

							m_Paths[i]->SetBezierSpline();

							firstpathpoint = m_Paths[i]->m_PathPoints.front();

							for(pathpoint = m_Paths[i]->m_PathPoints.front(); pathpoint;)
								{
									nextpathpoint = pathpoint->m_Next;

									controlpoint1 = m_Paths[i]->CreatePathPoint(NULL);
									controlpoint2 = m_Paths[i]->CreatePathPoint(NULL);

									controlpoint1->m_Path = path;
									controlpoint2->m_Path = path;

									point1 = mesh->CreatePoint();
									point2 = mesh->CreatePoint();

									controlpoint1->m_Point = point1;
									controlpoint2->m_Point = point2;

									if(nextpathpoint)
										{
											point1->m_LocalXYZ.n[0] = point1->m_LocalXYZ.n[0] = pathpoint->m_Point->m_LocalXYZ.n[0] + 0.2 * (nextpathpoint->m_Point->m_LocalXYZ.n[0] - pathpoint->m_Point->m_LocalXYZ.n[0]);
											point1->m_LocalXYZ.n[1] = point1->m_LocalXYZ.n[1] = pathpoint->m_Point->m_LocalXYZ.n[1] + 0.2 * (nextpathpoint->m_Point->m_LocalXYZ.n[1] - pathpoint->m_Point->m_LocalXYZ.n[1]);
											point1->m_LocalXYZ.n[2] = point1->m_LocalXYZ.n[2] = pathpoint->m_Point->m_LocalXYZ.n[2] + 0.2 * (nextpathpoint->m_Point->m_LocalXYZ.n[2] - pathpoint->m_Point->m_LocalXYZ.n[2]);
											point2->m_LocalXYZ.n[0] = point2->m_LocalXYZ.n[0] = pathpoint->m_Point->m_LocalXYZ.n[0] + 0.8 * (nextpathpoint->m_Point->m_LocalXYZ.n[0] - pathpoint->m_Point->m_LocalXYZ.n[0]);
											point2->m_LocalXYZ.n[1] = point2->m_LocalXYZ.n[1] = pathpoint->m_Point->m_LocalXYZ.n[1] + 0.8 * (nextpathpoint->m_Point->m_LocalXYZ.n[1] - pathpoint->m_Point->m_LocalXYZ.n[1]);
											point2->m_LocalXYZ.n[2] = point2->m_LocalXYZ.n[2] = pathpoint->m_Point->m_LocalXYZ.n[2] + 0.8 * (nextpathpoint->m_Point->m_LocalXYZ.n[2] - pathpoint->m_Point->m_LocalXYZ.n[2]);
										}
									else
										{
											point1->m_LocalXYZ.n[0] = point1->m_LocalXYZ.n[0] = pathpoint->m_Point->m_LocalXYZ.n[0] + 0.2 * (firstpathpoint->m_Point->m_LocalXYZ.n[0] - pathpoint->m_Point->m_LocalXYZ.n[0]);
											point1->m_LocalXYZ.n[1] = point1->m_LocalXYZ.n[1] = pathpoint->m_Point->m_LocalXYZ.n[1] + 0.2 * (firstpathpoint->m_Point->m_LocalXYZ.n[1] - pathpoint->m_Point->m_LocalXYZ.n[1]);
											point1->m_LocalXYZ.n[2] = point1->m_LocalXYZ.n[2] = pathpoint->m_Point->m_LocalXYZ.n[2] + 0.2 * (firstpathpoint->m_Point->m_LocalXYZ.n[2] - pathpoint->m_Point->m_LocalXYZ.n[2]);
											point2->m_LocalXYZ.n[0] = point2->m_LocalXYZ.n[0] = pathpoint->m_Point->m_LocalXYZ.n[0] + 0.8 * (firstpathpoint->m_Point->m_LocalXYZ.n[0] - pathpoint->m_Point->m_LocalXYZ.n[0]);
											point2->m_LocalXYZ.n[1] = point2->m_LocalXYZ.n[1] = pathpoint->m_Point->m_LocalXYZ.n[1] + 0.8 * (firstpathpoint->m_Point->m_LocalXYZ.n[1] - pathpoint->m_Point->m_LocalXYZ.n[1]);
											point2->m_LocalXYZ.n[2] = point2->m_LocalXYZ.n[2] = pathpoint->m_Point->m_LocalXYZ.n[2] + 0.8 * (firstpathpoint->m_Point->m_LocalXYZ.n[2] - pathpoint->m_Point->m_LocalXYZ.n[2]);
										}

									pathpoint->m_Next = controlpoint1;
									controlpoint1->m_Next = controlpoint2;
									controlpoint2->m_Next = nextpathpoint;

									pathpoint = nextpathpoint;
								}
						}
				}
		}

	RedrawAllViews(false);
}

void CDocument::OnEditSplinesPatches()
{
}

void CDocument::OnEditSplinesPaths()
{
}
*/

k3dIPoint::ID k3dMesh::NextPointID()
{
	// Increment the point ID ...
	k3dIPoint::ID result = ++m_CurrentPointID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dIPath::ID k3dMesh::NextPathID()
{
	// Increment the path ID ...
	const k3dIPath::ID result = ++m_CurrentPathID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dIPathPoint::ID k3dMesh::NextPathPointID()
{
	// Increment the pathpoint ID ...
	const k3dIPathPoint::ID result = ++m_CurrentPathPointID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dPoint* k3dMesh::FindPoint(k3dIPoint::ID PointID)
{
	// Sanity checks ...
	g_assert(PointID);

	for(PointIterator point = m_Points.begin(); point != m_Points.end(); point++)
		if((*point)->PointID() == PointID)
			return (*point);

	// Couldn't find it!!!
	g_return_val_if_fail(0, 0);
	return 0;
}

k3dPathPoint* k3dMesh::FindPathPoint(k3dIPathPoint::ID PathPointID)
{
	// Sanity checks ...
	g_assert(PathPointID);

	for(PathIterator path = m_Paths.begin(); path != m_Paths.end(); path++)
		for(k3dPath::PathPointIterator pathpoint = (*path)->m_PathPoints.begin(); pathpoint != (*path)->m_PathPoints.end(); pathpoint++)
			if((*pathpoint)->PathPointID() == PathPointID)
				return (*pathpoint);

	// Couldn't find it!!!
	g_return_val_if_fail(0, 0);
	return 0;
}


