#ifndef K3DMESH_H
#define K3DMESH_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dMesh class, which implements k3dIMesh, and encapsulates a collection of points, paths, NURBS curves, material, and shading information
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
		\author Paul Gregory (pgregory@aqsis.com)
*/

#include <k3dtransformobject.h>
#include <k3dvariable.h>

#include <k3dimesh.h>
#include <k3dipath.h>
#include <k3dipathcollection.h>
#include <k3dipathpoint.h>
#include <k3dipoint.h>
#include <k3dipointcollection.h>
#include <k3dilocationcollection.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>

// Forward class declarations ...
class k3dPoint;
class k3dPath;
class k3dPathPoint;
class k3dShadingNode;
class k3dIRiEngine;
class k3dIRiMaterial;

/////////////////////////////////////////////////////////////////////////////////////
// k3dMesh

typedef k3dTransformObject k3dMeshBase;

/// Provides a default implementation of k3dIMesh
class k3dMesh :
	public k3dMeshBase,
	public k3dIMesh,
	public k3dIPathCollection,
	public k3dIPointCollection,
	public k3dILocationCollection,
	public k3dIMaterialCollection
{
friend class k3dMeshLoadShadingNode;
public:
	k3dMesh(sdpIUnknown* Initializer);
	~k3dMesh();

	// k3dIMesh implementation
	k3dIDocument* Document();
	k3dIPathCollection* Paths();
	k3dIPath* CreatePath();
	sdpResult DeletePath(k3dIPath* const Path);
	k3dIPointCollection* Points();
	k3dIPoint* CreatePoint();
	sdpResult DeletePoint(k3dIPoint* const Point);
	sdpResult ShadePaths(const gulong Criteria, const gdouble Angle);
	sdpResult UnshadePaths(const gulong Criteria);
	sdpResult WeldPoints(const gulong Criteria);
	sdpResult UnweldPoints(const gulong Criteria);
	sdpResult CloseMesh(const gulong Criteria);
	sdpResult CalculateNormals();
	sdpResult Dimensions(const gulong Criteria, double Min[3], double Max[3]);
	const sdpVector3& Color() const;
	void SetColor(const sdpVector3& Color);

	// k3dIObject implementation
//	k3dIState* State();

	// k3dIState implementation
//	sdpResult SetState(gulong Mask, bool State);

	// k3dIPathCollection implementation
	void VisitPaths(k3dIPathVisitor* const Visitor);
	sdpResult AddPath(k3dIPath* const Path);
	sdpResult RemovePath(k3dIPath* const Path);

	gulong PathCount();
	K3DPATHHANDLE FirstPath();
	K3DPATHHANDLE NextPath(K3DPATHHANDLE Path);
	k3dIPath* Path(K3DPATHHANDLE Path);

	// k3dIPointCollection implementation
	void VisitPoints(k3dIPointVisitor* const Visitor);
	sdpResult AddPoint(k3dIPoint* const Point);
	sdpResult RemovePoint(k3dIPoint* const Point);

	gulong PointCount();
	K3DPOINTHANDLE FirstPoint();
	K3DPOINTHANDLE NextPoint(K3DPOINTHANDLE Point);
	k3dIPoint* Point(K3DPOINTHANDLE Point);

	// k3dILocationCollection implementation
	void VisitLocations(k3dILocationVisitor* const Visitor);
	gulong LocationCount();
	K3DLOCATIONHANDLE FirstLocation();
	K3DLOCATIONHANDLE NextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* Location(K3DLOCATIONHANDLE Location);

	// k3dIEditorRender implementation
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3dIRiRender implementation
	sdpResult RiPushRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiPopRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);

	// k3dIPersist implementation
	sdpResult Load(sdpxml::Document& Document, sdpxml::Element& Element);
	sdpResult LoadComplete();
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);

	// k3dIMaterialCollection implementation
	k3dIMaterial* Material();
	sdpResult SetMaterial(k3dIMaterial* const Material);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	/// Returns a point based on its point ID
	k3dPoint* FindPoint(const k3dIPoint::ID PointID);
	/// Returns a pathpoint based on its pathpoint ID
	k3dPathPoint* FindPathPoint(const k3dIPathPoint::ID PathPointID);

	/// Returns the next unique point ID
	k3dIPoint::ID NextPointID();
	/// Returns the next unique path ID
	k3dIPath::ID NextPathID();
	/// Returns the next unique pathpoint ID
	k3dIPathPoint::ID NextPathPointID();

	/// Enumerates available RenderMan "Render Modes" for this mesh
	typedef enum
	{
		NORMAL,
		SUBDIVISION,
	} RenderMode;

	friend std::ostream& operator<<(std::ostream& Stream, const RenderMode& Value)
	{
		switch(Value)
			{
				case NORMAL:
					Stream << "normal";
					break;
				case SUBDIVISION:
					Stream << "subdivision";
					break;
			}

		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, RenderMode& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "normal")
			Value = NORMAL;
		else if(text == "subdivision")
			Value = SUBDIVISION;
		else
			std::cerr << g_pretty_function() << ": unknown value [" << text << "]" << std::endl;

		return Stream;
	}

protected:
	/// Stores the current point ID
	k3dIPoint::ID m_CurrentPointID;
	/// Stores the current path ID
	k3dIPath::ID m_CurrentPathID;
	/// Stores the current pathpoint ID
	k3dIPathPoint::ID m_CurrentPathPointID;

	/// Stores the color for drawing unselected geometry
	k3dVariable<sdpVector3> m_Color;

	/// Store the selected render mode.
	k3dVariable<RenderMode> m_RenderMode;

	/// Stores our assigned material
	k3dObjectInterfaceVariable<k3dIMaterial> m_Material;

	typedef std::vector<k3dPoint*> PointCollection;
	typedef PointCollection::iterator PointIterator;
	/// Stores the array of point objects
	PointCollection m_Points;
	/// True iff the point collection has changed during the current edit session
	bool m_PointCollectionChanged;

	typedef std::vector<k3dPath*> PathCollection;
	typedef PathCollection::iterator PathIterator;
	/// Stores the array of path objects
	PathCollection m_Paths;
	/// True iff the path collection has changed during the current edit session
	bool m_PathCollectionChanged;

	typedef std::list<k3dShadingNode*> ShadingNodeCollection;
	typedef ShadingNodeCollection::iterator ShadingNodeIterator;
	/// Stores the array of shading node objects
	ShadingNodeCollection m_ShadingNodes;

	/// Saves geometry to a K-3D (XML) document
	sdpResult SaveGeometry(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	/// Saves points to a K-3D (XML) document
	sdpResult SavePoints(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	/// Saves paths to a K-3D (XML) document
	sdpResult SavePaths(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	/// Saves shading nodes to a K-3D (XML) document
	sdpResult SaveShading(sdpxml::Element& Element, k3dIDependencies* Dependencies);

	/// Loads geometry from a K-3D (XML) document
	sdpResult LoadGeometry(sdpxml::Element& Element);
	/// Loads points from a K-3D (XML) document
	sdpResult LoadPoints(sdpxml::Element& Element);
	/// Loads paths from a K-3D (XML) document
	sdpResult LoadPaths(sdpxml::Element& Element);
	/// Loads shading nodes from a K-3D (XML) document
	sdpResult LoadShading(sdpxml::Element& Element);

	/// Draws points
	inline void EditorDrawPoints(const gulong Set, const gulong Unset);
	/// Draws path normals
	inline void EditorDrawNormals(const gulong Set, const gulong Unset);
	/// Draws path orientations
	inline void EditorDrawOrientations(const gulong Set, const gulong Unset);
	/// Draws wireframe paths
	inline void EditorDrawWireframePaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset);
	/// Draws solid paths
	inline void EditorDrawSolidPaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset);
	/// Draws solid, textured paths
	inline void EditorDrawSolidTexturedPaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset);

	/// Stores an array of points for RenderMan rendering
	typedef std::vector<sdpVector3> RtPointArray;
	/// Stores an array of scalars for RenderMan rendering
	typedef std::vector<gdouble> RtFloatArray;

	/// Draws a path
	inline void RiPath(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path);
	/// Draws a 3-sided path
	inline void RiPath3(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path);
	/// Draws a 4-sided path
	inline void RiPath4(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path);
	/// Draws a greater-than-4 sided path
	inline void RiPathN(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dPath* const Path);
	/// Draws the whole mesh as a SubdivisionMesh
	inline void	RiSubdivisionMesh(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material);

	/// Stores the color for drawing selected geometry
	static const sdpVector3 m_SelectedColor;
	/// Stores the color for drawing path normals
	static const sdpVector3 m_PathNormalColor;
};

#endif // K3DMESH_H
