
// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements a kitchen-junk-drawer-ful of convenience functions, templates, and classes
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>
#include <k3dpathcollection.h>
#include <k3dpathpointcollection.h>
#include <k3dpointcollection.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dibezierchannelvaluenode.h>
#include <k3dicamera.h>
#include <k3dicommandrecorder.h>
#include <k3dicurve.h>
#include <k3didocument.h>
#include <k3dieditable.h>
#include <k3dihierarchy.h>
#include <k3diknot.h>
#include <k3diknotcollection.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3dimaterial.h>
#include <k3dimesh.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>
#include <k3diobjectinfocollection.h>
#include <k3diobjectmodule.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>
#include <k3dipoint.h>
#include <k3diriengine.h>
#include <k3distate.h>
#include <k3disnaptarget.h>
#include <k3ditransform.h>
#include <k3diviewvolume.h>

#include <sdpgtkfileselector.h>
#include <sdpgl.h>
#include <sdpsl.h>
#include <sdpunits.h>

#include <fstream>
#include <iomanip>
#include <strstream>

static k3dIApplication* g_Application = 0;

void k3dRegisterApplication(k3dIApplication* const Application)
{
	// Sanity checks ...
	g_assert(Application);

	if(g_Application)
		g_assert(Application == g_Application);
	else
		g_Application = Application;
}

k3dIApplication* k3dApplication()
{
	// Sanity checks ...
	g_assert(g_Application);

	return g_Application;
}

gdouble k3dConvertStringDouble(const k3dIChannel::DataType Data, SDPCSTRING Input)
{
	// Parse the input string into value & units ...
	gdouble inputvalue = 0.0;
	sdpString inputunits;

	std::istrstream stream(Input);
	stream >> inputvalue >> inputunits;

	try
		{
			switch(Data)
				{
					case k3dIChannel::DATA_ANGLE:
						if(0 == inputunits.size())
							inputunits = "deg";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::AngleUnits().Unit(inputunits)), sdp::AngleUnits().Unit("rad")).Value();
					case k3dIChannel::DATA_AREA:
						if(0 == inputunits.size())
							inputunits = "m^2";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::AreaUnits().Unit(inputunits)), sdp::AreaUnits().Unit("m^2")).Value();
					case k3dIChannel::DATA_COLOR:
						break;
					case k3dIChannel::DATA_DISTANCE:
						if(0 == inputunits.size())
							inputunits = "m";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::DistanceUnits().Unit(inputunits)), sdp::DistanceUnits().Unit("m")).Value();
					case k3dIChannel::DATA_FORCE:
						if(0 == inputunits.size())
							inputunits = "N";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::ForceUnits().Unit(inputunits)), sdp::ForceUnits().Unit("N")).Value();
					case k3dIChannel::DATA_MASS:
						if(0 == inputunits.size())
							inputunits = "Kg";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::MassUnits().Unit(inputunits)), sdp::MassUnits().Unit("Kg")).Value();
					case k3dIChannel::DATA_PRESSURE:
						if(0 == inputunits.size())
							inputunits = "Pa";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::PressureUnits().Unit(inputunits)), sdp::PressureUnits().Unit("Pa")).Value();
					case k3dIChannel::DATA_ROTATIONXYZ:
						break;
					case k3dIChannel::DATA_SCALAR:
						return inputvalue;
					case k3dIChannel::DATA_SCALEXYZ:
						break;
					case k3dIChannel::DATA_TIME:
						if(0 == inputunits.size())
							inputunits = "s";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::TimeUnits().Unit(inputunits)), sdp::TimeUnits().Unit("s")).Value();
					case k3dIChannel::DATA_VOLUME:
						if(0 == inputunits.size())
							inputunits = "m^3";
						return sdp::Measurement(sdp::Measurement(inputvalue, sdp::VolumeUnits().Unit(inputunits)), sdp::VolumeUnits().Unit("m^3")).Value();
				}
		}
	catch(std::exception& e)
		{
			std::cerr << e.what() << std::endl;
		}

	g_return_val_if_fail(0, 0.0);
}

sdpString k3dConvertDoubleString(const k3dIChannel::DataType Data, const gdouble InputValue, const gulong Precision)
{
	std::ostrstream buffer;
	buffer << setiosflags(ios::fixed);
	buffer.precision(Precision);

	switch(Data)
		{
			case k3dIChannel::DATA_ANGLE:
				buffer << sdp::Measurement(sdp::Measurement(InputValue, sdp::AngleUnits().Unit("rad")), sdp::AngleUnits().Unit("deg"));
				break;
			case k3dIChannel::DATA_AREA:
				buffer << sdp::Measurement(InputValue, sdp::AreaUnits().Unit("m^2"));
				break;
			case k3dIChannel::DATA_COLOR:
				break;
			case k3dIChannel::DATA_DISTANCE:
				buffer << sdp::Measurement(InputValue, sdp::DistanceUnits().Unit("m"));
				break;
			case k3dIChannel::DATA_FORCE:
				buffer << sdp::Measurement(InputValue, sdp::ForceUnits().Unit("N"));
				break;
			case k3dIChannel::DATA_MASS:
				buffer << sdp::Measurement(InputValue, sdp::MassUnits().Unit("Kg"));
				break;
			case k3dIChannel::DATA_PRESSURE:
				buffer << sdp::Measurement(InputValue, sdp::PressureUnits().Unit("Pa"));
				break;
			case k3dIChannel::DATA_ROTATIONXYZ:
				break;
			case k3dIChannel::DATA_SCALAR:
				buffer << InputValue;
				break;
			case k3dIChannel::DATA_SCALEXYZ:
				break;
			case k3dIChannel::DATA_TIME:
				buffer << sdp::Measurement(InputValue, sdp::TimeUnits().Unit("s"));
				break;
			case k3dIChannel::DATA_VOLUME:
				buffer << sdp::Measurement(InputValue, sdp::VolumeUnits().Unit("m^3"));
				break;
		}

	buffer << std::ends;
	const sdpString result(buffer.str());
	buffer.freeze(false);
	return result;
}

sdpString k3dUpdateDoubleString(const k3dIChannel::DataType Data, SDPCSTRING Input, gdouble& OutputValue, const gulong Precision)
{
	OutputValue = k3dConvertStringDouble(Data, Input);
	return k3dConvertDoubleString(Data, OutputValue, Precision);
}

// Return the path to an object's library ...
SDPCSTRING k3dModulePath(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	return Object->ObjectInfo()->Module()->Path();
}

class k3dGetShadersVisitor :
	public sdpsl::IShaderVisitor
{
public:
	k3dGetShadersVisitor(k3dShaderSet& Results) :
		m_Results(Results)
	{
	}

	void VisitShader(sdpsl::IShader* const Shader)
	{
		m_Results.insert(Shader);
	}

private:
	k3dShaderSet& m_Results;
};

void k3dGetShaders(sdpsl::IShaderCollection* const Collection, k3dShaderSet& Results)
{
	// Sanity checks ...
	g_assert(Collection);

	k3dGetShadersVisitor visitor(Results);
	Collection->VisitShaders(&visitor);
}

class k3dGetShaderArgumentsVisitor :
	public sdpsl::IShaderArgumentVisitor
{
public:
	k3dGetShaderArgumentsVisitor(k3dShaderArgumentSet& Results) :
		m_Results(Results)
	{
	}

	void VisitShaderArgument(sdpsl::IShaderArgument* const Argument)
	{
		m_Results.insert(Argument);
	}

private:
	k3dShaderArgumentSet& m_Results;
};

void k3dGetShaderArguments(sdpsl::IShader* const Shader, k3dShaderArgumentSet& Results)
{
	// Sanity checks ...
	g_assert(Shader);

	k3dGetShaderArgumentsVisitor visitor(Results);
	Shader->VisitArguments(&visitor);
}

bool k3dUpdateShader(sdpsl::IShader* const Shader, const sdpString RenderEngineType, const sdpString RenderEngine)
{
	// Sanity checks ...
	g_return_val_if_fail(Shader, false);
	g_return_val_if_fail(RenderEngineType.size(), false);
	g_return_val_if_fail(RenderEngine.size(), false);

	// Get the set of available render engines
	k3dOptionsRenderEngines engines;
	k3dOptionsGetRenderEngines(k3dApplication()->Options(), engines);

	// For each engine ...
	for(k3dOptionsRenderEngines::iterator engine = engines.begin(); engine != engines.end(); engine++)
		{
			// If it's the wrong type, skip it ...
			if(engine->Type() != RenderEngineType)
				continue;

			// If it's the wrong engine, skip it ...
			if(engine->Engine() != RenderEngine)
				continue;

			// Figure out the full path to the compiled shader ...
			const sdpString shadersourcepath(Shader->SourcePath());
			sdpString shaderbinarypath(engine->ShaderBinary());
			shaderbinarypath.formattedreplace('%', "p", shadersourcepath.substr(0, shadersourcepath.rfind(".")).c_str());

			std::cout << "\tUpdating [" << shadersourcepath << "] to [" << shaderbinarypath << "]" << std::endl;

			// Get the last modification time of the shader source ...
			GTime sourcemodificationtime = 0;
			g_return_val_if_fail(sdpFileModificationTime(shadersourcepath, sourcemodificationtime), false);

			// Get the last modification time of the shader binary (if it exists) ...
			GTime binarymodificationtime = 0;
			sdpFileModificationTime(shaderbinarypath, binarymodificationtime);

			// If the source is newer (or the binary doesn't exist), try to compile ...
			if(sourcemodificationtime > binarymodificationtime)
				{
#ifndef SDPWIN32
					// Create a command line that can compile the shader ...
					sdpString commandline = "cd " + sdpDirectory(shadersourcepath) + "; " + engine->CompileShader();
					commandline.formattedreplace('%', "p", shadersourcepath);

					if(SDP_S_OK == k3dApplication()->Verbose())
						std::cout << "\t" << commandline << std::endl;

					// Make it happen ...
					std::strstream buffer;
					g_return_val_if_fail(sdpRunProcess(commandline, buffer), false);
					buffer << std::ends;

#else // !SDPWIN32
					// Get current dir
					sdpString k3dCurrentDirHere = sdpGetCurrentDirectory();

					// Create a line for cd shadersourcepath ...
					sdpString commandline = sdpDirectory(shadersourcepath);

					if(SDP_S_OK == k3dApplication()->Verbose())
					std::cout << "\t cd " << commandline << std::endl;

					// Change temporary do shaders dir
					SetCurrentDirectory(commandline);

					// Create a command line to compile the shader ...
					commandline = engine->CompileShader();
					commandline.formattedreplace('%', "p", shadersourcepath);

					if(SDP_S_OK == k3dApplication()->Verbose())
						std::cout << "\t" << commandline << std::endl;

					// Make it happen ...
					std::strstream buffer;
					bool CompileResultIsOK = sdpRunProcess(commandline, buffer);
					buffer << std::ends;

					if(SDP_S_OK == k3dApplication()->Verbose())
						std::cout << "\t cd " << k3dCurrentDirHere << std::endl;

					// Restore k3d dir
					SetCurrentDirectory(k3dCurrentDirHere);

					if (!(CompileResultIsOK)) return false;
#endif // SDPWIN32

				}
			return true;
		}

	return false;
}

void k3dDeepSetState(sdpIUnknown* const Object, const gulong Mask, const bool State)
{
	// Sanity checks ...
	g_assert(Object);

	// Set our state ...
	k3dIObject* const object = dynamic_cast<k3dIObject*>(Object);
	if(object)
		object->State()->SetState(Mask, State);

	// If this is a mesh, perform a deep state change of points / paths ...
	k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(Object);
	if(mesh)
		{
			k3dIPathCollection* const paths = mesh->Paths();
			for(K3DPATHHANDLE path = paths->FirstPath(); path; path = paths->NextPath(path))
				paths->Path(path)->State()->SetState(Mask, State);

			k3dIPointCollection* const points = mesh->Points();
			for(K3DPOINTHANDLE point = points->FirstPoint(); point; point = points->NextPoint(point))
				points->Point(point)->State()->SetState(Mask, State);
		}
}

void k3dDeepSetState(k3dIPath* const Path, const gulong Mask, const bool State)
{
	// Sanity checks ...
	g_assert(Path);

	// Set our state ...
	Path->State()->SetState(Mask, State);

	// Set the state of any of our points ...
	k3dIPathPointCollection* const pathpoints = Path->Points();
	for(K3DPATHPOINTHANDLE pathpoint = pathpoints->FirstPathPoint(); pathpoint; pathpoint = pathpoints->NextPathPoint(pathpoint))
		pathpoints->PathPoint(pathpoint)->Point()->State()->SetState(Mask, State);
}

namespace
{

/// Functor object for deselecting points
class DeselectPoint
{
public:
	void operator()(k3dIPoint* const Point)
	{
		Point->State()->SetState(k3dIState::SELECTED, false);
	}
};

/// Functor object for deselecting paths
class DeselectPath
{
public:
	void operator()(k3dIPath* const Path)
	{
		Path->State()->SetState(k3dIState::SELECTED, false);
	}
};

/// Functor object for deselecting objects
class DeselectObject
{
public:
	void operator()(k3dIObject* const Object)
	{
		Object->State()->SetState(k3dIState::SELECTED, false);

		k3dIPathCollection* const pathcollection = dynamic_cast<k3dIPathCollection*>(Object);
		if(pathcollection)
			k3d::for_each(pathcollection, DeselectPath());

		k3dIPointCollection* const pointcollection = dynamic_cast<k3dIPointCollection*>(Object);
		if(pointcollection)
			k3d::for_each(pointcollection, DeselectPoint());
	}
};

} // namespace

void k3dDeselectAll(k3dIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);

	// Deselect objects ...
	k3d::for_each(Document->Objects(), DeselectObject());
}

/// Destroys an object
sdpResult k3dDeleteObject(sdpIUnknown* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Make sure it's a K-3D object!
	k3dIObject* object = dynamic_cast<k3dIObject*>(Object);
	g_return_val_if_fail(object, SDP_E_FAIL);

	// Zap it ...
	k3d::Editing editing(object->Document()->Objects());
	object->Document()->Objects()->DeleteObject(object);

	return SDP_S_OK;
}

bool k3dSetPosition(sdpIUnknown* const Object, const sdpVector3 Position)
{
	// Sanity check ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* const transform = dynamic_cast<k3dITransform*>(Object);
	g_return_val_if_fail(transform, false);

	// If so, position it ...
	k3d::Editing editing(transform);
	transform->SetPosition(k3dITransform::CURRENT_FRAME, sdpVector3(Position));

	return true;
}

bool k3dSetOrientation(sdpIUnknown* const Object, const sdpEulerAngles Orientation)
{
	// Sanity check ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* const transform = dynamic_cast<k3dITransform*>(Object);
	g_return_val_if_fail(transform, false);

	// If so, orient it ...
	k3d::Editing editing(transform);
	transform->SetOrientation(k3dITransform::CURRENT_FRAME, sdpAngleAxis(sdpQuaternion(Orientation)));

	return true;
}

bool k3dSetScale(sdpIUnknown* const Object, const sdpVector3 Scale)
{
	// Sanity check ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* const transform = dynamic_cast<k3dITransform*>(Object);
	g_return_val_if_fail(transform, false);

	// If so, orient it ...
	k3d::Editing editing(transform);
	transform->SetScale(k3dITransform::CURRENT_FRAME, sdpVector3(Scale));

	return true;
}

sdpString k3dGetUniqueName(k3dIObjectCollection* const ObjectCollection, const sdpString Name)
{
	// See if the name already exists ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			if(Name == ObjectCollection->Object(handle)->Name())
				{
					// Recursively see if a new name works ...
					return k3dGetUniqueName(ObjectCollection, Name + "*");
				}
		}

	return Name;
}

bool k3dSetName(sdpIUnknown* const Object, const sdpString ObjectName)
{
	// Sanity checks ...
	k3dIObject* const object = dynamic_cast<k3dIObject*>(Object);
	g_return_val_if_fail(object, false);

	// If so, rename it ...
	object->SetName(ObjectName);

	return true;
}

// Sets a channel's maximum range of values (if possible)
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const gdouble MinValue, const gdouble MaxValue)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dIBezierChannel* channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	channel->SetRange1(MinValue, MaxValue);

	return true;
}

// Sets a channel's maximum range of values (if possible)
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const sdpVector3 MinValue, const sdpVector3 MaxValue)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dIBezierChannel* channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	channel->SetRange3(const_cast<sdpVector3&>(MinValue), const_cast<sdpVector3&>(MaxValue));

	return true;
}

/// Used to gather a collection of channel nodes
class k3dBezierChannelNodesVisitor : public k3dIBezierChannelNodeVisitor
{
public:
	k3dBezierChannelNodesVisitor(k3dBezierChannelNodeArray& Nodes) :
		m_Nodes(Nodes)
	{
	}

	sdpResult VisitBezierChannelNode(k3dIBezierChannelNode* const Node)
	{
		// Sanity checks ...
		g_return_val_if_fail(Node, SDP_E_FAIL);

		m_Nodes.push_back(Node);
		return SDP_S_OK;
	}

private:
	k3dBezierChannelNodeArray& m_Nodes;
};

bool k3dBezierChannelNodes(sdpIUnknown* Channel, k3dBezierChannelNodeArray& Nodes)
{
	// Sanity checks ...
	g_assert(Channel);

	// See if it's really a Bezier channel ...
	k3dIBezierChannel* channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	// Visit all of the channel's nodes ...
	k3dBezierChannelNodesVisitor visitor(Nodes);
	g_return_val_if_fail(sdpSucceeded(channel->VisitNodes(&visitor)), false);

	return true;
}

// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const gdouble Value)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	// Get rid of all segments ...
	channel->SetSegmentCount(0);

	// Reset the remaining node positions ...
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);

	for(k3dBezierChannelNodeArray::iterator node = nodes.begin(); node != nodes.end(); node++)
		(*node)->SetPosition(sdpVector2(0.0, Value));

	return true;
}

// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const sdpVector3 Value)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	// Get rid of all segments ...
	channel->SetSegmentCount(0);

	// Reset the remaining node positions ...
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);

	for(k3dBezierChannelNodeArray::iterator node = nodes.begin(); node != nodes.end(); node++)
		{
			(*node)->SetPosition(sdpVector2(0.0, 0.0));

			k3dIBezierChannelValueNode* valuenode = dynamic_cast<k3dIBezierChannelValueNode*>(*node);
			if(valuenode)
				valuenode->SetValue3(const_cast<sdpVector3&>(Value));
		}

	return true;
}

// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const gdouble Value1, const gdouble Value2)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	// We want just a single segment ...
	channel->SetSegmentCount(1);

	// Setup all four nodes ...
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);
	g_return_val_if_fail(4 == nodes.size(), false);

	k3dBezierChannelNodeArray::iterator node(nodes.begin());
	(*node)->SetPosition(sdpVector2(0.0, Value1));

	node++;
	(*node)->SetPosition(sdpLerp(sdpVector2(0.0, Value1), sdpVector2(1.0, Value2), 0.3));

	node++;
	(*node)->SetPosition(sdpLerp(sdpVector2(0.0, Value1), sdpVector2(1.0, Value2), 0.7));

	node++;
	(*node)->SetPosition(sdpVector2(1.0, Value2));

	return true;
}

// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const sdpVector3 Value1, const sdpVector3 Value2)
{
	// Sanity checks ...
	g_assert(Channel);

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(Channel);
	if(!channel)
		return false;

	// We want just a single segment ...
	channel->SetSegmentCount(1);

	// Setup all four nodes ...
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);
	g_return_val_if_fail(4 == nodes.size(), false);

	k3dBezierChannelNodeArray::iterator node(nodes.begin());
	(*node)->SetPosition(sdpVector2(0.0, 0.0));
	dynamic_cast<k3dIBezierChannelValueNode*>(*node)->SetValue3(const_cast<sdpVector3&>(Value1));

	node++;
	(*node)->SetPosition(sdpLerp(sdpVector2(0.0, 0.0), sdpVector2(1.0, 1.0), 0.3));

	node++;
	(*node)->SetPosition(sdpLerp(sdpVector2(0.0, 0.0), sdpVector2(1.0, 1.0), 0.7));

	node++;
	(*node)->SetPosition(sdpVector2(1.0, 1.0));
	dynamic_cast<k3dIBezierChannelValueNode*>(*node)->SetValue3(const_cast<sdpVector3&>(Value2));

	return true;
}

bool k3dMakeBezier(k3dICurve* Curve)
{
	// Sanity checks ...
	g_assert(Curve);

	const gulong pointcount = Curve->Points()->PointCount();
	g_return_val_if_fail(pointcount > 1, false);

	// A Bezier curve is a NURBS curve where the order and number of control points are equal ...
	Curve->SetOrder(pointcount);

	// ... and the knot values are equal multiplicities of zero and one ...
	k3dIKnotCollection* const knots = Curve->Knots();
	g_assert(knots);

	gulong index = 0;
	for(K3DKNOTHANDLE knothandle = knots->FirstKnot(); knothandle; knothandle = knots->NextKnot(knothandle))
		{
			gdouble value = index++ < pointcount ? 0.0 : 1.0;
			knots->Knot(knothandle)->SetValue(value);
		}

	return true;
}

bool k3dGetDimensions(k3dIObject* Object, gulong Criteria, sdpVector3& Min, sdpVector3& Max)
{
	// Sanity checks ...
	g_assert(Object);

	// If it's a mesh, ask the mesh for the results ...
	k3dIMesh* mesh = dynamic_cast<k3dIMesh*>(Object);
	if(mesh)
		{
			mesh->Dimensions(Criteria, Min, Max);
			return true;
		}

	// If it's a location-server, call our specialized function for handling it ...
	k3dILocationCollection* locationserver = dynamic_cast<k3dILocationCollection*>(Object);
	if(locationserver)
		k3dGetDimensions(locationserver, Criteria, Min, Max);

	// Otherwise, give up!
	Min = 0.0;
	Max = 0.0;

	return false;
}

// Return the min/max model coordinates for every location in the given object that meets the given criteria ...
bool k3dGetDimensions(k3dILocationCollection* LocationCollection, gulong Criteria, sdpVector3& Min, sdpVector3& Max)
{
	// Sanity checks ...
	g_assert(LocationCollection);

	// Reset our results ...
	Min = DBL_MAX;
	Max = -DBL_MAX;

	// Iterate all locations ...
	sdpVector3 coords;
	k3dILocation* location;

	for(K3DLOCATIONHANDLE locationhandle = LocationCollection->FirstLocation(); locationhandle; locationhandle = LocationCollection->NextLocation(locationhandle))
		{
			location = LocationCollection->Location(locationhandle);

			// Skip locations that don't match our criteria ...
			if(SDP_S_OK != location->State()->QueryState(Criteria))
				continue;

			// Get the location's old coordinates ...
			coords = location->LocalXYZ();

			// Compare against our min/max coordinates ...
			if(coords[0] < Min[0])
				Min[0] = coords[0];
			if(coords[0] > Max[0])
				Max[0] = coords[0];
			if(coords[1] < Min[1])
				Min[1] = coords[1];
			if(coords[1] > Max[1])
				Max[1] = coords[1];
			if(coords[2] < Min[2])
				Min[2] = coords[2];
			if(coords[2] > Max[2])
				Max[2] = coords[2];
		}

	return true;
}

void k3dCalculateNormals(k3dIObjectCollection* const ObjectCollection)
{
	// Sanity checks ...
	g_return_if_fail(ObjectCollection);
	k3d::for_each<k3dIMesh>(ObjectCollection, std::mem_fun(&k3dIMesh::CalculateNormals));
}

bool k3dIsInside(sdpVector2 Point, k3dIPath* Path)
{
	// Sanity check ...
	g_assert(Path);

	sdpVector2Array array;
	sdpVector3 screencoords;

	// For each point ...
	k3dIPathPointCollection* points = Path->Points();
	for(K3DPATHPOINTHANDLE pathpointhandle = points->FirstPathPoint(); pathpointhandle; pathpointhandle = points->NextPathPoint(pathpointhandle))
		{
			screencoords = points->PathPoint(pathpointhandle)->Point()->Location()->ScreenXYZ();
			array.push_back(sdpVector2(screencoords, 2));	// Lose the Z coordinate ...
		}

	return sdpIsInside(Point, array);
}

sdpMatrix4 k3dGetTransformMatrix(k3dITransform* Transform, const k3dITransform::ReferenceFrame Frame)
{
	// Sanity checks ...
	g_assert(Transform);
	return sdpTranslation3D(Transform->Position(Frame)) * sdpRotation3D(Transform->Orientation(Frame)) * sdpScaling3D(Transform->Scale(Frame));
}

sdpMatrix4 k3dGetObjectToWorldMatrix(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	k3dIHierarchy* hierarchy = Object->Document()->Hierarchy();

	// Start out with an identity matrix ...
	sdpMatrix4 result = sdpIdentity3D();

	// Create a stack of k3dITransform-enabled objects ...
	std::vector<k3dITransform*> transformstack;

	// Start with the given object and move to the root of the hierarchy ...
	k3dITransform* transform;
	for(; Object; Object = hierarchy->ObjectParent(Object))
		{
			transform = dynamic_cast<k3dITransform*>(Object);
			if(transform)
				transformstack.push_back(transform);
		}

	// Concatenate parent self- and child-matrices in order from root -> leaf ...
	while(transformstack.size() > 1)
		{
			result = result * k3dGetTransformMatrix(transformstack.back(), k3dITransform::ANIMATED_FRAME);
			result = result * k3dGetTransformMatrix(transformstack.back(), k3dITransform::ANIMATED_CHILD_FRAME);

			transformstack.pop_back();
		}
	// Concatenate the given object's self-matrix only ...
	while(transformstack.size() > 0)
		{
			result = result * k3dGetTransformMatrix(transformstack.back(), k3dITransform::ANIMATED_FRAME);

			transformstack.pop_back();
		}

	return result;
}

sdpMatrix4 k3dGetWorldToObjectMatrix(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Return theinverse of the object-to-world matrix ...
	return k3dGetObjectToWorldMatrix(Object).Inverse();
}

sdpMatrix4 k3dGetObjectToObjectMatrix(k3dIObject* From, k3dIObject* To)
{
	// Sanity checks ...
	g_assert(From);
	g_assert(To);

	// Get the matrix from object-to-world ...
	sdpMatrix4 from = k3dGetObjectToWorldMatrix(From);

	// Get the matrix from world-to-object ...
	sdpMatrix4 to = k3dGetWorldToObjectMatrix(To);

	// return the concatenated result ...
	return from * to;
}

sdpMatrix4 k3dGetWorldToCameraMatrix(k3dICamera* Camera)
{
	// Sanity checks ...
	g_assert(Camera);

	// Convert the camera to an object ...
	k3dIObject* object = dynamic_cast<k3dIObject*>(Camera);
	g_assert(object);

	sdpMatrix4 result = k3dGetWorldToObjectMatrix(object);

	return result;
}

sdpMatrix4 k3dGetCameraToWorldMatrix(k3dICamera* Camera)
{
	// Sanity checks ...
	g_assert(Camera);

	// Convert the camera to an object ...
	k3dIObject* object = dynamic_cast<k3dIObject*>(Camera);
	g_assert(object);

	sdpMatrix4 result = k3dGetObjectToWorldMatrix(object);

	return result;
}

sdpMatrix4 k3dGetObjectToCameraMatrix(k3dIObject* Object, k3dICamera* Camera)
{
	// Sanity checks ...
	g_assert(Object);
	g_assert(Camera);

	// Convert the camera to an object ...
	k3dIObject* object = dynamic_cast<k3dIObject*>(Camera);
	g_assert(object);

	sdpMatrix4 result = k3dGetObjectToObjectMatrix(Object, object);

	return result;
}

sdpMatrix4 k3dGetCameraToObjectMatrix(k3dICamera* Camera, k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Camera);
	g_assert(Object);

	// Convert the camera to an object ...
	k3dIObject* object = dynamic_cast<k3dIObject*>(Camera);
	g_assert(object);

	sdpMatrix4 result = k3dGetObjectToObjectMatrix(object, Object);

	return result;
}

sdpVector3 k3dGetWorldPosition(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	return k3dGetObjectToWorldMatrix(Object) * sdpVector3(0, 0, 0);
}

void k3dPushOpenGLMatrix(const sdpMatrix4& Matrix)
{
	double glmatrix[16];
	Matrix.Transpose().CopyArray(glmatrix);
	glMultMatrixd(glmatrix);
}

void k3dPushRiMatrix(const sdpMatrix4& Matrix, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	sdpMatrix4 matrix = Matrix.Transpose();
	Engine->RiConcatTransform((gdouble*)matrix);
}

sdpString k3dCommandNodePath(k3dICommandNode* const CommandNode)
{
	sdpString result;
	for(k3dICommandNode* node = CommandNode; node; node = node->CommandNodeParent())
		{
			if(result.size())
				result = "/" + result;

			result = node->CommandNodeName() + result;
		}

	result = "/" + result;

	return result;
}

bool k3dGetFilePath(SDPCSTRING Type, SDPCSTRING Prompt, bool PromptOverwrite, SDPCSTRING OldPath, sdpString& FilePath)
{
	// Sanity checks ...
	g_assert_string(Type);
	g_assert_string(Prompt);

	// Get the user's path option, defaulting to their home directory if necessary ...
	sdpxml::Element& startpathelement = sdpxml::SafeXPathSelect(k3dApplication()->Options(), k3dApplication()->Options(), "application/paths/path[@type=\"" + sdpString(Type) + "\"]", sdpxml::Element("path", sdpGetHomeDirectory(), sdpxml::Attribute("type", Type)));
	sdpString startpath = startpathelement.Text();

	// Last-but-not-least, see if we already have a script path ...
	sdpString oldpath(OldPath);
	if(oldpath.size())
		startpath = sdpDirectory(oldpath);

	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	// Prompt the user for a filepath ...
	sdpGtkFileSelector fileselector(Prompt, startpath);
	fileselector.DoModal();
	if(!fileselector.OK())
		return false;

	// Make sure we're not overwriting Precious Memories of The Way We Were ...
	if(PromptOverwrite)
		{
			if(sdpFileExists(fileselector.FilePath()))
				{
					sdpStringArray buttons;
					buttons.push_back("Yes");
					buttons.push_back("No");
					if(1 != k3dApplication()->Query("Overwrite " + fileselector.FilePath() + "?", Prompt, 1, buttons))
						return false;
				}
		}

	// Store the path for posterity ...
	startpathelement.Text() = sdpDirectory(fileselector.FilePath());

	// Return the path ...
	FilePath = fileselector.FilePath();

	return true;
}

sdpVector2 k3dTransformMouseCoords(k3dICamera* Camera, const sdpVector2 MouseCoords, gdouble Left, gdouble Right, gdouble Top, gdouble Bottom)
{
	// Sanity Checks ...
	g_assert(Camera);

	gdouble width = Camera->Width();
	gdouble height = Camera->Height();

	gdouble mousex = sdpClamp<gdouble>(0, width, MouseCoords[0]);
	gdouble mousey = sdpClamp<gdouble>(0, height, MouseCoords[1]);

	if(0 == width)
		width = 1;

	if(0 == height)
		height = 1;

	return sdpVector2(sdpLerp(Left, Right, mousex / width), sdpLerp(Top, Bottom, mousey / height));
}

sdpVector2 k3dTransformNScaleMouseCoords(k3dICamera* Camera, const sdpVector2 MouseCoords, gdouble Left, gdouble Right, gdouble Top, gdouble Bottom, gdouble scale)
{
	// Sanity Checks ...
	g_assert(Camera);

	gdouble width = Camera->Width();
	gdouble height = Camera->Height();

	// Translate the window inside a bigger one (TODO : add "scale < 1" support)
	gdouble newx = MouseCoords[0] + width / scale;
	gdouble newy = MouseCoords[1] + height / scale;
	width *= scale;
	height *= scale;

	if(0 == width)
		width = 1;

	if(0 == height)
		height = 1;

	gdouble mousex = sdpClamp<gdouble>(0, width, newx);
	gdouble mousey = sdpClamp<gdouble>(0, height, newy);

	return sdpVector2(sdpLerp(Left, Right, mousex / width), sdpLerp(Top, Bottom, mousey / height));
}

sdpVector3 k3dMouseToCameraRay(k3dICamera* Camera, const sdpVector2 MouseCoords)
{
	// Sanity checks ...
	g_assert(Camera);

	// Get the camera frustum ...
//	const bool orthographic = Camera->ViewVolume()->Orthographic();
	const gdouble nearplane = Camera->ViewVolume()->NearPlane();
//	const gdouble farplane = Camera->ViewVolume()->FarPlane();
	const gdouble left = Camera->ViewVolume()->Left();
	const gdouble right = Camera->ViewVolume()->Right();
	const gdouble top = Camera->ViewVolume()->Top();
	const gdouble bottom = Camera->ViewVolume()->Bottom();

	// Convert mouse coordinates (pixels) to screen coordinates ...
	sdpVector2 screencoords = k3dTransformMouseCoords(Camera, MouseCoords, left * nearplane, right * nearplane, top * nearplane, bottom * nearplane);

	sdpVector3 result(screencoords, nearplane);
	result.Normalize();

	return result;
}

void k3dMouseToWorldRay(k3dICamera* Camera, const sdpVector2 MouseCoords, sdpVector3& RayOrigin, sdpVector3& RayDirection)
{
	// Sanity checks ...
	g_assert(Camera);

	// Get the mouse ray in camera coordinates ...
	sdpVector3 ray = k3dMouseToCameraRay(Camera, MouseCoords);

	// Transform it to the world frame ...
	sdpMatrix4 cameratoworld = k3dGetCameraToWorldMatrix(Camera);
	RayOrigin = cameratoworld * sdpVector3(0, 0, 0);
	RayDirection = (cameratoworld * ray) - RayOrigin;
}

bool k3dSnap(k3dIDocument* const Document, const sdpVector3 SourceCoordinates, sdpVector3& SnapCoordinates, sdpString& SnapDescription)
{
	// Sanity checks ...
	g_assert(Document);

	typedef std::set<k3dISnapTarget*> SnapSet;

	// Get any visible objects that support snapping ...
	SnapSet visibleobjects;
	k3d::copy_if<k3dISnapTarget>(Document->Objects(), std::inserter(visibleobjects, visibleobjects.begin()), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));

	// Get any selected objects that support snapping ...
	SnapSet selectedobjects;
	k3d::copy_if<k3dISnapTarget>(Document->Objects(), std::inserter(selectedobjects, selectedobjects.begin()), k3d::FlagIsSet<k3dIObject*>(k3dIState::SELECTED));

	// Subtract selected objects from visible objects (so they don't snap to themselves) ...
	SnapSet snapobjects;
	std::set_difference(visibleobjects.begin(), visibleobjects.end(), selectedobjects.begin(), selectedobjects.end(), std::inserter(snapobjects, snapobjects.begin()));

	// Iterate through snap objects, looking for the best (closest) snap ...
	bool result = false;
	gdouble distance = DBL_MAX;

	for(SnapSet::iterator snapobject = snapobjects.begin(); snapobject != snapobjects.end(); ++snapobject)
		{
			sdpVector3 coordinates(SourceCoordinates);
			sdpString description;
			if(SDP_S_OK != (*snapobject)->Snap(coordinates, description))
				continue;

			if((coordinates - SourceCoordinates).Length2() < distance)
				{
					result = true;
					distance = (coordinates - SourceCoordinates).Length2();

					SnapCoordinates = coordinates;
					SnapDescription = description;
				}
		}

	return result;
}

namespace
{

class RedrawAll
{
public:
	RedrawAll(const bool Synchronous) :
		m_Synchronous(Synchronous)
	{
	}

	void operator()(k3dIObject* const Object)
	{
		k3dICamera* const camera = dynamic_cast<k3dICamera*>(Object);
		if(camera)
			camera->Redraw(m_Synchronous);
	}

protected:
	const bool m_Synchronous;
};

} // namespace

void k3dRedrawAll(k3dIDocument* const Document, const bool Synchronous)
{
	// Sanity checks ...
	g_return_if_fail(Document);
	k3d::for_each(Document->Objects(), RedrawAll(Synchronous));
}

k3dIStateChangeSet*  k3dStartStateChangeSet(k3dIDocument* const Document)
{
	// Sanity checks ...
	g_assert(Document);

	k3dIStateChangeSet* const newchangeset = new k3dStateChangeSet();
	Document->StateRecorder()->SetCurrentChangeSet(newchangeset);

	return newchangeset;
}

void k3dCancelStateChangeSet(k3dIDocument* const Document)
{
	// Sanity checks ...
	g_assert(Document);

	// Make sure we're actually recording a changeset ...
	k3dIStateChangeSet* const changeset = Document->StateRecorder()->ChangeSet();
	g_return_if_fail(changeset);

	// Update the document ...
	Document->StateRecorder()->SetCurrentChangeSet(0);

	// Undo any changes up to this point, and delete leftover data ...
	changeset->Undo();
	delete changeset;
}

void k3dFinishStateChangeSet(k3dIDocument* const Document, SDPCSTRING ChangeSetName)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(ChangeSetName);

	Document->StateRecorder()->RecordChangeSet(Document->StateRecorder()->ChangeSet(), ChangeSetName);
	Document->StateRecorder()->SetCurrentChangeSet(0);
}

/////////////////////////////////////////////////////////////////////////////
// k3dRecordStateChangeSet

k3dRecordStateChangeSet::k3dRecordStateChangeSet(k3dIDocument* const Document, SDPCSTRING Label) :
	m_Document(Document),
	m_Label(Label)
{
	// Sanity checks ...
	g_assert(m_Document);
	g_assert(m_Label.size());

	k3dStartStateChangeSet(m_Document);
}

k3dRecordStateChangeSet::~k3dRecordStateChangeSet()
{
	k3dFinishStateChangeSet(m_Document, m_Label);
}

void k3dRecordCommand(k3dICommandNode* CommandNode, k3dICommandNode::CommandType Type, SDPCSTRING CommandName, SDPCSTRING Arguments)
{
	// Sanity checks ...
	g_assert(CommandNode);
	g_assert_string(CommandName);

	if(SDP_S_OK == k3dApplication()->CommandRecorder()->IsRecording())
		k3dApplication()->CommandRecorder()->RecordCommand(CommandNode, Type, CommandName, Arguments);
}

sdpResult k3dReturnString(SDPCSTRING Result, SDPSTRING Buffer, gulong* BufferLength)
{
	// Sanity checks ...
	g_return_val_if_fail(BufferLength, SDP_E_FAIL);

	sdpString result(Result);
	gulong inputlength = *BufferLength;
	*BufferLength = result.size()+1;

	if(result.size()+1 > inputlength)
		return SDP_S_BUFFER_TOO_SMALL;

	if(Buffer)
		strcpy(Buffer, result.c_str());

	return SDP_S_OK;
}

k3dIMaterial* k3dDefaultMaterial(k3dIDocument* const Document)
{
	// Sanity checks ...
	g_return_val_if_fail(Document, 0);

	std::vector<k3dIObject*> materials;
	k3d::copy_if(Document->Objects(), std::back_inserter(materials), k3d::Implements(sdp::TypeInfo<k3dIMaterial>()));

	if(materials.size())
		{
			k3dIMaterial* const material = dynamic_cast<k3dIMaterial*>(*materials.rbegin());
			g_return_val_if_fail(material, 0);

			return material;
		}

	return 0;
}

namespace k3d
{

void PushEditing(sdpIUnknown* const Editable)
{
	k3dIEditable* const editable = dynamic_cast<k3dIEditable*>(Editable);
	g_return_if_fail(editable);

	editable->PushEditing();
}

void PopEditing(sdpIUnknown* const Editable)
{
	k3dIEditable* const editable = dynamic_cast<k3dIEditable*>(Editable);
	g_return_if_fail(editable);

	editable->PopEditing();
}

//////////////////////////////////////////////////////////////////////////////////////////
// Editing

Editing::Editing(sdpIUnknown* const Editable) :
	m_Editable(dynamic_cast<k3dIEditable*>(Editable))
{
	if(m_Editable)
		m_Editable->PushEditing();
}

Editing::~Editing()
{
	if(m_Editable)
		m_Editable->PopEditing();
}

} // namespace k3d


