#ifndef K3DMISC_H
#define K3DMISC_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares a kitchen-junk-drawer-ful of convenience functions, templates, and classes
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dichannel.h>
#include <k3dicommandnode.h>
#include <k3ditransform.h>

// Forward class declarations
class k3dIApplication;
class k3dIBezierChannelNode;
class k3dICamera;
class k3dICurve;
class k3dIDocument;
class k3dIEditable;
class k3dILocation;
class k3dILocationCollection;
class k3dIMaterial;
class k3dIMesh;
class k3dIObject;
class k3dIObjectCollection;
class k3dIObjectInfo;
class k3dIPath;
class k3dIPathPoint;
class k3dIPoint;
class k3dIRiEngine;
class k3dIStateChangeSet;
namespace sdpxml { class Element; }

// Convenience collections
typedef std::vector<k3dIBezierChannelNode*> k3dBezierChannelNodeArray;
typedef std::vector<k3dIPoint*> k3dPointArray;
typedef std::vector<k3dIPathPoint*> k3dPathPointArray;

/// Initializes access to the global singleton k3dIApplication object - must be called once-per-module
void k3dRegisterApplication(k3dIApplication* const Application);
/// Provides access to the global singleton k3dIApplication object
k3dIApplication* k3dApplication();

/// Returns a complete command node path
sdpString k3dCommandNodePath(k3dICommandNode* const CommandNode);

/// Prompts the user for a filepath, checking for old choices, and storing the current choice for reuse
bool k3dGetFilePath(SDPCSTRING Type, SDPCSTRING Prompt, bool PromptOverwrite, SDPCSTRING OldPath, sdpString& FilePath);

/// Checks to see if a screen coordinate is contained within a path's screen coordinates
bool k3dIsInside(sdpVector2 Point, k3dIPath* Path);

/// Called to update path normals throughout a document
void k3dCalculateNormals(k3dIObjectCollection* const ObjectCollection);

/// Performs a deep (points and paths, too) state change on an object
void k3dDeepSetState(sdpIUnknown* const Object, const gulong Mask, const bool State);
/// Performs a deep (points, too) state change on a path
void k3dDeepSetState(k3dIPath* const Path, const gulong Mask, const bool State);

/// Deselects everything in the document
void k3dDeselectAll(k3dIDocument* Document);

/// Parses a scalar value from a string, converting units as needed
gdouble k3dConvertStringDouble(const k3dIChannel::DataType Data, SDPCSTRING Input);
/// Converts a double to a string, displaying units-of-measure
sdpString k3dConvertDoubleString(const k3dIChannel::DataType Data, const gdouble InputValue, const gulong Precision);
/// Parsers a scalar value from a string, converting units and returning a string with the converted value and units
sdpString k3dUpdateDoubleString(const k3dIChannel::DataType Data, SDPCSTRING Input, gdouble& OutputValue, const gulong Precision);

/// Returns the absolute path to an object's module
SDPCSTRING k3dModulePath(k3dIObject* Object);

/// Deletes an object (this is undo-able)
sdpResult k3dDeleteObject(sdpIUnknown* Object);

/// Sets an object's position (if possible)
bool k3dSetPosition(sdpIUnknown* const Object, const sdpVector3 Position);
/// Sets an object's orientation (if possible)
bool k3dSetOrientation(sdpIUnknown* const Object, const sdpEulerAngles Orientation);
/// Sets an object's scale (if possible)
bool k3dSetScale(sdpIUnknown* const Object, const sdpVector3 Scale);

/// Returns a unique object name based on the one supplied
sdpString k3dGetUniqueName(k3dIObjectCollection* const ObjectCollection, const sdpString Name);
/// Sets an object's name (if possible)
bool k3dSetName(sdpIUnknown* const Object, const sdpString ObjectName);

/// Sets a channel's maximum range of values
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const gdouble MinValue, const gdouble MaxValue);
/// Sets a channel's maximum range of values
bool k3dChannelSetValueRange(sdpIUnknown* Channel, const sdpVector3 MinValue, const sdpVector3 MaxValue);
/// Returns a channel's set of available nodes
bool k3dBezierChannelNodes(sdpIUnknown* Channel, k3dBezierChannelNodeArray& Nodes);
/// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const gdouble Value);
/// Sets a channel to a constant value (if possible)
bool k3dChannelSetConstantValue(sdpIUnknown* Channel, const sdpVector3 Value);
/// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const gdouble Value1, const gdouble Value2);
/// Sets a channel to a linear range of values (if possible)
bool k3dChannelSetLinearValue(sdpIUnknown* Channel, const sdpVector3 Value1, const gdouble Value2);

/// Forces a NURBS curve into a Bezier configuration
bool k3dMakeBezier(k3dICurve* Curve);

/// Returns the dimensions of an arbitrary object whose points/locations meet the given criteria
bool k3dGetDimensions(k3dIObject* Object, gulong Criteria, sdpVector3& Min, sdpVector3& Max);
/// Returns the min/max coordinates of all points within the given location server that fit the given criteria
bool k3dGetDimensions(k3dILocationCollection* LocationCollection, gulong Criteria, sdpVector3& Min, sdpVector3& Max);

///////////////////////////////////////////////////////////////////////////////
// Shaders

namespace sdpsl
{
class IShader;
class IShaderArgument;
class IShaderCollection;
}

/// A collection of sdpsl::IShader objects
typedef std::set<sdpsl::IShader*> k3dShaderSet;
/// A collection of sdpsl::IShaderArgument objects
typedef std::set<sdpsl::IShaderArgument*> k3dShaderArgumentSet;

/// Returns a set of sdpsl::IShader objects
void k3dGetShaders(sdpsl::IShaderCollection* const Collection, k3dShaderSet& Results);
/// Returns a set of sdpsl::IShaderArgument objects
void k3dGetShaderArguments(sdpsl::IShader* const Shader, k3dShaderArgumentSet& Results);

/// Brings a shader up-to-date by compiling it
bool k3dUpdateShader(sdpsl::IShader* const Shader, const sdpString RenderEngineType, const sdpString RenderEngine);

/// Returns a transformation matrix for the requested object and reference frame
sdpMatrix4 k3dGetTransformMatrix(k3dITransform* Transform, const k3dITransform::ReferenceFrame Frame);
/// Returns a complete, concatenated, object-to-world transformation matrix for an object (includes all its' parents)
sdpMatrix4 k3dGetObjectToWorldMatrix(k3dIObject* Object);
/// Returns a complete, concatenated, world-to-object transformation matrix for an object (includes all its' parents)
sdpMatrix4 k3dGetWorldToObjectMatrix(k3dIObject* Object);
/// Returns a complete, concatenated transformation matrix to transform from one object space to another
sdpMatrix4 k3dGetObjectToObjectMatrix(k3dIObject* From, k3dIObject* To);
/// Returns a complete, concatenated world-to-camera transformation matrix for the given camera
sdpMatrix4 k3dGetWorldToCameraMatrix(k3dICamera* Camera);
/// Returns a complete, concatenated camera-to-world transformation matrix for the given camera
sdpMatrix4 k3dGetCameraToWorldMatrix(k3dICamera* Camera);
/// Returns a complete, concatenated transformation matrix to transform from an object's space to the given camera's space
sdpMatrix4 k3dGetObjectToCameraMatrix(k3dIObject* Object, k3dICamera* Camera);
/// Returns a complete, concatenated transformation matrix to transform from a camera's space to the given object's space
sdpMatrix4 k3dGetCameraToObjectMatrix(k3dICamera* Camera, k3dIObject* Object);

/// Returns the position of an object in world coordinates
sdpVector3 k3dGetWorldPosition(k3dIObject* Context);

/// Pushes a matrix onto the OpenGL matrix stack
void k3dPushOpenGLMatrix(const sdpMatrix4& Matrix);
/// Pushes a matrix onto the RenderMan Interface matrix stack
void k3dPushRiMatrix(const sdpMatrix4& Matrix, k3dIRiEngine* const Engine);

/// Transform mouse coordinates into some other reference frame
sdpVector2 k3dTransformMouseCoords(k3dICamera* Camera, const sdpVector2 MouseCoords, gdouble Left, gdouble Right, gdouble Top, gdouble Bottom);
/// Transform mouse coordinates into some other reference frame with scaling
sdpVector2 k3dTransformNScaleMouseCoords(k3dICamera* Camera, const sdpVector2 MouseCoords, gdouble Left, gdouble Right, gdouble Top, gdouble Bottom, gdouble scale);
/// Transform mouse coordinates to a normalized ray originating at the camera, within the camera's reference frame
sdpVector3 k3dMouseToCameraRay(k3dICamera* Camera, const sdpVector2 MouseCoords);
/// Transform mouse coordinates to a normalized ray originating at the camera, within the world reference frame
void k3dMouseToWorldRay(k3dICamera* Camera, const sdpVector2 MouseCoords, sdpVector3& RayOrigin, sdpVector3& RayDirection);

/// "Snaps" a point against other snap targets in the document, returning true iff a snapped value was found
bool k3dSnap(k3dIDocument* const Document, const sdpVector3 SourceCoordinates, sdpVector3& SnapCoordinates, sdpString& SnapDescription);

/// Redraws all cameras
void k3dRedrawAll(k3dIDocument* const Document, const bool Synchronous);

/// Starts recording a state change set for undo/redo purposes
k3dIStateChangeSet* k3dStartStateChangeSet(k3dIDocument* const Document);
/// Cancels recording a state change set for undo/redo purposes
void k3dCancelStateChangeSet(k3dIDocument* const Document);
/// Finishes recording a state change set for undo/redo purposes
void k3dFinishStateChangeSet(k3dIDocument* const Document, SDPCSTRING ChangeSetName);

/// Provides an exception-safe way to record a state change for UNDO purposes
class k3dRecordStateChangeSet
{
public:
	k3dRecordStateChangeSet(k3dIDocument* const Document, SDPCSTRING Label);
	~k3dRecordStateChangeSet();

private:
	k3dIDocument* const m_Document;
	const sdpString m_Label;
};

/// Records a user action for use in a macro
void k3dRecordCommand(k3dICommandNode* const CommandNode, const k3dICommandNode::CommandType Type, SDPCSTRING Name, SDPCSTRING Arguments);

/// Puts data into a buffer in a standardized way for use with interface methods that return string results
sdpResult k3dReturnString(SDPCSTRING Result, SDPSTRING Buffer, gulong* BufferLength);

/// Returns the default material for application to new objects (could return NULL!)
k3dIMaterial* k3dDefaultMaterial(k3dIDocument* const Document);

namespace k3d
{

/// Begins editing an object
void PushEditing(sdpIUnknown* const Editable);
/// Completes editing an object
void PopEditing(sdpIUnknown* const Editable);

/// Return- and exception-safe class for controlling an editable object's state
class Editing
{
public:
	Editing(sdpIUnknown* const Editable);
	~Editing();
	
private:
	k3dIEditable* const m_Editable;
};
	
/// Provides std::auto_ptr-like behavior, but handles K-3D objects (i.e. it knows how to destroy them correctly, through their owning document)
template<class Interface>
class auto_ptr
{
public:
	explicit auto_ptr(Interface* const Object = 0) :
		m_Object(0)
	{
		reset(Object);
	}

	auto_ptr(sdpIUnknown* const Object = 0) :
		m_Object(0)
	{
		reset(dynamic_cast<Interface*>(Object));
	}
	
	~auto_ptr()
	{
		reset(0);
	}

	Interface* operator->() const
	{
		return m_Object;
	}
	
	Interface* get() const
	{
		return m_Object;
	}
	
	Interface* release()
	{
		Interface* result = m_Object;
		m_Object = 0;
		return result;
	}
	
	void reset(Interface* const Object)
	{
		if(Object != m_Object)
			{
				k3dIObject* const object = dynamic_cast<k3dIObject*>(m_Object);
				if(object)
					{
						k3d::Editing editing(object->Document()->Objects());
						object->Document()->Objects()->DeleteObject(object);
					}

				m_Object = Object;
			}
	}

private:
	Interface* m_Object;
};
	
} // namespace k3d

#endif // K3DMISC_H
