// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Defines the Module class, which provides useful boilerplate for implementing K-3D object modules
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dcom.h>
#include <k3dmisc.h>
#include <k3dmodule.h>
#include <k3diobjectmoduleinitializer.h>

#include <strstream>

namespace k3d
{

Module::Module(const sdpString ModuleName, const sdpString ModuleVersion) :
	m_ModuleInfo("sdpcommoduleinfo")
{
	// Sanity checks ...
	g_assert(ModuleName.size());
	g_assert(ModuleVersion.size());
	
	// Setup our module info XML document ...
	m_ModuleInfo.Attributes().push_back(sdpxml::Attribute("name", ModuleName));
	m_ModuleInfo.Attributes().push_back(sdpxml::Attribute("version", ModuleVersion));
}

Module::~Module()
{
}

sdpResult Module::InitializeModule(sdpIUnknown* Initializer)
{
	// Sanity checks ...
	g_return_val_if_fail(Initializer, SDP_E_FAIL);
	
	// Get our real initializer object ...
	k3dIObjectModuleInitializer* const initializer = dynamic_cast<k3dIObjectModuleInitializer*>(Initializer);
	g_return_val_if_fail(initializer, SDP_E_FAIL);
	
	// Store data ...
	k3dRegisterApplication(initializer->Application());
	
	return SDP_S_OK;
}

SDPCSTRING Module::ModuleInfo()
{
	if(0 == m_ModuleInfoText.size())
		{
			std::strstream stream;
			stream << m_ModuleInfo << std::ends;
			m_ModuleInfoText = stream.str();
			stream.freeze(false);
		}
		
	return m_ModuleInfoText;
}

sdpxml::Element& Module::RegisterObject(const sdp::ClassID& ClassID, const sdpString Name)
{
	// Sanity checks ...
	g_assert(ClassID != sdp::ClassIDNull);
	g_assert(Name.size());

	m_ModuleInfo.Children().push_back(sdpxml::Element("object", "", sdpxml::Attribute("clsid", sdpToString(ClassID)), sdpxml::Attribute("name", Name)));
	
	return m_ModuleInfo.Children().back();
}

void Module::RegisterInterface(sdpxml::Element& Object, const sdp::ITypeInfo& InterfaceType)
{
	Object.Children().push_back(sdpxml::Element("interface", "", sdpxml::Attribute("type", InterfaceType.RTTI().name())));
}

void Module::RegisterShortDescription(sdpxml::Element& Object, const sdpString Description)
{
	// Sanity checks ...
	g_assert(Description.size());
	Object.Children().push_back(sdpxml::Element("shortdescription", Description));	
}

void Module::RegisterBookmark(sdpxml::Element& Object, const sdpString Bookmark)
{
	// Sanity checks ...
	g_assert(Bookmark.size());
	Object.Children().push_back(sdpxml::Element("bookmark", "", sdpxml::Attribute("location", Bookmark)));
}

void Module::RegisterIncomplete(sdpxml::Element& Object)
{
	Object.Attributes().push_back(sdpxml::Attribute("incomplete", sdpToString(true)));
}

void Module::RegisterExperimental(sdpxml::Element& Object)
{
	Object.Attributes().push_back(sdpxml::Attribute("experimental", sdpToString(true)));
}

} // namespace k3d



