// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dNurbsControlPoint class, which implements a geometric nurbs point
		\author Romain Behar with NURBS students (romainbehar@yahoo.com)
*/

#include <k3dnurbscontrolpoint.h>
#include <k3dnurbsmesh.h>
#include <k3dmisc.h>
#include <k3dstaterecorder.h>

#include <k3didocument.h>

/////////////////////////////////////////////////////////////////////////
// k3dNurbsControlPoint

k3dNurbsControlPoint::k3dNurbsControlPoint(k3dNurbsMesh* const Mesh, const ID PointID) :
	k3dState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::USERVIEWABLE | k3dIState::SAVEABLE),
	m_Mesh(Mesh),
	m_PointID(PointID)
{
}

k3dNurbsControlPoint::~k3dNurbsControlPoint()
{
/*
	// Reset active point
	if(m_Mesh->m_Layer->m_Document->m_ActivePoint == this)
		m_Mesh->m_Layer->m_Document->m_ActivePoint = 0;
*/
}

sdpResult k3dNurbsControlPoint::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& point = Element.Append(sdpxml::Element("point", "",
		sdpxml::Attribute("position", sdpToString(m_LocalXYZ)),
		sdpxml::Attribute("weight", sdpToString(m_Weight)),
		sdpxml::Attribute("texture", sdpToString(m_TextureXYZ)),
		sdpxml::Attribute("itexture", sdpToString(m_ImplicitTextureXYZ)),
		sdpxml::Attribute("id", sdpToString(m_PointID))));

	k3dState::Save(point, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dNurbsControlPoint::Load(sdpxml::Element& Element)
{
	// The point ID is required ...
	g_return_val_if_fail(sdpxml::ParseAttribute(Element, "id", m_PointID), SDP_E_FAIL);

	// Load the point coordinates ...
	sdpxml::ParseAttribute(Element, "position", m_LocalXYZ);

	// Load the weight point...
	sdpxml::ParseAttribute(Element, "weight", m_Weight);

	// Load texture coordinates ...
	sdpxml::ParseAttribute(Element, "texture", m_TextureXYZ);
	sdpxml::ParseAttribute(Element, "itexture", m_ImplicitTextureXYZ);

	// Load our state ...
	k3dState::Load(Element);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// k3dINurbsControlPoint implementation

k3dIState* k3dNurbsControlPoint::State()
{
	return this;
}

k3dILocation* k3dNurbsControlPoint::Location()
{
	return this;
}

/////////////////////////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
k3dINurbsMesh* k3dNurbsControlPoint::Mesh()
{
	return m_Mesh;
}

const k3dINurbsControlPoint::ID k3dNurbsControlPoint::PointID()
{
	return m_PointID;
}


///////////////////////////////////////////////////////////////////////////
// k3dIEditable implementation ...

void k3dNurbsControlPoint::PushEditing()
{
	if(!Editing())
		{
			m_LocalXYZChanged = false;
			//m_WeightChanged = false;
			m_TextureXYZChanged = false;
		}

	k3dEditable::PushEditing();
}

void k3dNurbsControlPoint::PopEditing()
{
	k3dEditable::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_LocalXYZChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_LocalXYZ));

					//if(m_WeightChanged)
					//	changeset->RecordNewState(new k3dValueStateContainer<gdouble>(m_Weight));

					if(m_TextureXYZChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_TextureXYZ));
				}
		}
}

///////////////////////////////////////////////////////////////////////////
// k3dIState implementation ...

sdpResult k3dNurbsControlPoint::SetState(gulong Mask, bool State)
{
	// Perform an undo-able state change ...
	RawSetState(Mask, State, m_Mesh->Document()->StateRecorder());
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// k3dILocation implementation ...

/*
k3dIState* k3dPoint::State()
{
	return this;
}
*/

const sdpVector3& k3dNurbsControlPoint::LocalXYZ()
{
	return m_LocalXYZ;
}

const gdouble k3dNurbsControlPoint::Weight()
{
	return m_Weight;
}

const sdpVector3& k3dNurbsControlPoint::AnimXYZ()
{
	return m_AnimXYZ;
}

const sdpVector3& k3dNurbsControlPoint::WorldXYZ()
{
	return m_WorldXYZ;
}

const sdpVector3& k3dNurbsControlPoint::CameraXYZ()
{
	return m_CameraXYZ;
}

const sdpVector3& k3dNurbsControlPoint::ScreenXYZ()
{
	return m_ScreenXYZ;
}

const sdpVector3& k3dNurbsControlPoint::NormalXYZ()
{
	return m_NormalXYZ;
}

const sdpVector3& k3dNurbsControlPoint::TextureXYZ()
{
	return m_TextureXYZ;
}

const sdpVector3& k3dNurbsControlPoint::ImplicitTextureXYZ()
{
	return m_ImplicitTextureXYZ;
}

void k3dNurbsControlPoint::SetLocalXYZ(const sdpVector3& Coords)
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
	if(changeset && !m_LocalXYZChanged)
		{
			m_LocalXYZChanged = true;
			changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_LocalXYZ));
		}

	m_LocalXYZ = Coords;
}

void k3dNurbsControlPoint::SetWeight(const gdouble PtWeight)
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	/*k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
	if(changeset && !m_WeightChanged)
		{
			m_WeightChanged = true;
			changeset->RecordOldState(new k3dValueStateContainer<gdouble>(m_Weight));
		}
	*/
	m_Weight = PtWeight;
}

void k3dNurbsControlPoint::SetAnimXYZ(const sdpVector3& Coords)
{
	m_AnimXYZ = Coords;
}

void k3dNurbsControlPoint::SetWorldXYZ(const sdpVector3& Coords)
{
	m_WorldXYZ = Coords;
}

void k3dNurbsControlPoint::SetCameraXYZ(const sdpVector3& Coords)
{
	m_CameraXYZ = Coords;
}

void k3dNurbsControlPoint::SetScreenXYZ(const sdpVector3& Coords)
{
	m_ScreenXYZ = Coords;
}

void k3dNurbsControlPoint::SetNormalXYZ(const sdpVector3& Coords)
{
}

void k3dNurbsControlPoint::SetTextureXYZ(const sdpVector3& Coords)
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
	if(changeset && !m_TextureXYZChanged)
		{
			m_TextureXYZChanged = true;
			changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_TextureXYZ));
		}

	m_TextureXYZ = Coords;
}

void k3dNurbsControlPoint::SetImplicitTextureXYZ(const sdpVector3& Coords)
{
	m_ImplicitTextureXYZ = Coords;
}

