#ifndef K3DNURBSCONTROLPOINTCOLLECTION_H
#define K3DNURBSCONTROLPOINTCOLLECTION_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Romain Behar with NURBS students (romainbehar@yahoo.com)
*/

#include <k3dcollections.h>
#include <k3dinurbscontrolpointcollection.h>

namespace k3d
{

/// Applies a functor to every k3dINurbsControlPoint in a collection
template<typename Functor>
Functor for_each(k3dINurbsControlPointCollection* const NurbsControlPointCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(NurbsControlPointCollection, F);

	// For each object in the document ...
	for(K3DNURBSCONTROLPOINTHANDLE handle = NurbsControlPointCollection->FirstPoint(); handle; handle = NurbsControlPointCollection->NextPoint(handle))
		F(NurbsControlPointCollection->Point(handle));

	return F;
}

/// Applies a functor to every k3dINurbsControlPoint in a collection that satisfies the predicate
template<typename Functor, typename Predicate>
Functor for_each_if(k3dINurbsControlPointCollection* const NurbsControlPointCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(NurbsControlPointCollection, F);

	// For each object in the document ...
	for(K3DNURBSCONTROLPOINTHANDLE handle = NurbsControlPointCollection->FirstPoint(); handle; handle = NurbsControlPointCollection->NextPoint(handle))
		{
			k3dINurbsControlPoint* const object = NurbsControlPointCollection->Point(handle);

			if(P(object))
				F(object);
		}

	return F;
}

/// Copies k3dINurbsControlPoint references from a k3dINurbsControlPointCollection into a generic container
template<typename InsertIterator>
void copy(k3dINurbsControlPointCollection* const NurbsControlPointCollection, InsertIterator D)
{
	// Sanity checks ...
	g_return_if_fail(NurbsControlPointCollection);

	// For each object in the document ...
	for(K3DNURBSCONTROLPOINTHANDLE handle = NurbsControlPointCollection->FirstPoint(); handle; handle = NurbsControlPointCollection->NextPoint(handle))
		*D++ = NurbsControlPointCollection->Point(handle);
}

/// Copies k3dINurbsControlPoint references that satisfy a predicate from a k3dINurbsControlPointCollection into a generic container
template<typename InsertIterator, typename Predicate>
void copy_if(k3dINurbsControlPointCollection* const NurbsControlPointCollection, InsertIterator D, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(NurbsControlPointCollection);

	// For each object in the document ...
	for(K3DNURBSCONTROLPOINTHANDLE handle = NurbsControlPointCollection->FirstPoint(); handle; handle = NurbsControlPointCollection->NextPoint(handle))
		{
			k3dINurbsControlPoint* const object = NurbsControlPointCollection->Point(handle);
			if(P(object))
				*D++ = object;
		}
}

} // namespace k3d

#endif // K3DNURBSCONTROLPOINTCOLLECTION_H
