#ifndef K3DNURBSMESH_H
#define K3DNURBSMESH_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dNurbsMesh class, which implements k3dINurbsMesh, and encapsulates a collection of NURBS points, NURBS paths, and NURBS curves
		\author Romain Behar with NURBS students (romainbehar@yahoo.com)
*/

#include <k3dtransformobject.h>
#include <k3dvariable.h>

#include <k3dinurbsmesh.h>
#include <k3dinurbspath.h>
#include <k3dinurbspathcollection.h>
#include <k3dinurbspathpoint.h>
#include <k3dinurbscontrolpoint.h>
#include <k3dinurbscontrolpointcollection.h>
#include <k3dilocationcollection.h>
//#include <k3dimaterial.h>
//#include <k3dimaterialcollection.h>

// Forward class declarations ...
class k3dNurbsControlPoint;
class k3dNurbsPath;
class k3dNurbsPathPoint;
//class k3dShadingNode;
//class k3dIRiEngine;
//class k3dIRiMaterial;

/////////////////////////////////////////////////////////////////////////////////////
// k3dNurbsMesh

typedef k3dTransformObject k3dMeshBase;

/// Provides a default implementation of k3dINurbsMesh
class k3dNurbsMesh :
	public k3dMeshBase,
	public k3dINurbsMesh,
	public k3dINurbsPathCollection,
	public k3dINurbsControlPointCollection,
	public k3dILocationCollection//,
//	public k3dIMaterialCollection
{
friend class k3dNurbsMeshLoadShadingNode;
public:
	k3dNurbsMesh(sdpIUnknown* Initializer);
	~k3dNurbsMesh();

	// k3dINurbsMesh implementation
	k3dIDocument* Document();
	k3dINurbsPathCollection* Paths();
	k3dINurbsPath* CreatePath();
	sdpResult DeletePath(k3dINurbsPath* const Path, const bool DeletePoints);
	k3dINurbsControlPointCollection* Points();
	k3dINurbsControlPoint* CreatePoint();
	sdpResult DeletePoint(k3dINurbsControlPoint* const Point);
	//sdpResult ShadePaths(const gulong Criteria, const gdouble Angle);
	//sdpResult UnshadePaths(const gulong Criteria);
	sdpResult WeldPoints(const gulong Criteria);
	sdpResult UnweldPoints(const gulong Criteria);
	//sdpResult CalculateNormals();
	sdpResult Dimensions(const gulong Criteria, double Min[3], double Max[3]);
	const sdpVector3& Color() const;
	void SetColor(const sdpVector3& Color);

	// k3dIObject implementation
//	k3dIState* State();

	// k3dIState implementation
//	sdpResult SetState(gulong Mask, bool State);

	// k3dINurbsPathCollection implementation
	void VisitPaths(k3dIPathVisitor* const Visitor);
	sdpResult AddPath(k3dINurbsPath* const Path);
	sdpResult RemovePath(k3dINurbsPath* const Path);

	gulong PathCount();
	K3DNURBSPATHHANDLE FirstPath();
	K3DNURBSPATHHANDLE NextPath(K3DNURBSPATHHANDLE Path);
	k3dINurbsPath* Path(K3DNURBSPATHHANDLE Path);

	// k3dINurbsControlPointCollection implementation
	void VisitPoints(k3dINurbsControlPointVisitor* const Visitor);
	sdpResult AddPoint(k3dINurbsControlPoint* const Point);
	sdpResult RemovePoint(k3dINurbsControlPoint* const Point);

	gulong PointCount();
	K3DNURBSCONTROLPOINTHANDLE FirstPoint();
	K3DNURBSCONTROLPOINTHANDLE NextPoint(K3DNURBSCONTROLPOINTHANDLE Point);
	k3dINurbsControlPoint* Point(K3DNURBSCONTROLPOINTHANDLE Point);

	// k3dILocationCollection implementation
	void VisitLocations(k3dILocationVisitor* const Visitor);
	gulong LocationCount();
	K3DLOCATIONHANDLE FirstLocation();
	K3DLOCATIONHANDLE NextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* Location(K3DLOCATIONHANDLE Location);

	// k3dIEditorRender implementation
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3dIRiRender implementation
/*
	sdpResult RiPushRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiPopRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
*/

	// k3dIPersist implementation
	sdpResult Load(sdpxml::Document& Document, sdpxml::Element& Element);
	sdpResult LoadComplete();
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);

	// k3dIMaterialCollection implementation
	//k3dIMaterial* Material();
	//sdpResult SetMaterial(k3dIMaterial* const Material);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	/// Returns a nurbs point based on its point ID
	k3dNurbsControlPoint* FindPoint(const k3dINurbsControlPoint::ID PointID);
	/// Returns a nurbs pathpoint based on its nurbs pathpoint ID
	k3dNurbsPathPoint* FindPathPoint(const k3dINurbsPathPoint::ID PathPointID);

	/// Returns the next unique nurbs point ID
	k3dINurbsControlPoint::ID NextPointID();
	/// Returns the next unique nurbs path ID
	k3dINurbsPath::ID NextPathID();
	/// Returns the next unique nurbs pathpoint ID
	k3dINurbsPathPoint::ID NextPathPointID();

	/// Enumerates available RenderMan "Render Modes" for this mesh
	typedef enum
	{
		NORMAL,
		SUBDIVISION,
	} RenderMode;

	friend std::ostream& operator<<(std::ostream& Stream, const RenderMode& Value)
	{
		switch(Value)
			{
				case NORMAL:
					Stream << "normal";
					break;
				case SUBDIVISION:
					Stream << "subdivision";
					break;
			}

		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, RenderMode& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "normal")
			Value = NORMAL;
		else if(text == "subdivision")
			Value = SUBDIVISION;
		else
			std::cerr << g_pretty_function() << ": unknown value [" << text << "]" << std::endl;

		return Stream;
	}

protected:
	/// Stores the current nurbs point ID
	k3dINurbsControlPoint::ID m_CurrentPointID;
	/// Stores the current nurbs path ID
	k3dINurbsPath::ID m_CurrentPathID;
	/// Stores the current nurbs pathpoint ID
	k3dINurbsPathPoint::ID m_CurrentPathPointID;

	/// Stores the color for drawing unselected geometry
	k3dVariable<sdpVector3> m_Color;

	/// Store the selected render mode.
	k3dVariable<RenderMode> m_RenderMode;

	/// Stores our assigned material
	//k3dObjectInterfaceVariable<k3dIMaterial> m_Material;

	typedef std::vector<k3dNurbsControlPoint*> PointCollection;
	typedef PointCollection::iterator PointIterator;
	/// Stores the array of point objects
	PointCollection m_Points;
	/// True iff the point collection has changed during the current edit session
	bool m_PointCollectionChanged;

	typedef std::vector<k3dNurbsPath*> PathCollection;
	typedef PathCollection::iterator PathIterator;
	/// Stores the array of path objects
	PathCollection m_Paths;
	/// True iff the nurbs path collection has changed during the current edit session
	bool m_PathCollectionChanged;

	//typedef std::list<k3dShadingNode*> ShadingNodeCollection;
	//typedef ShadingNodeCollection::iterator ShadingNodeIterator;
	/// Stores the array of shading node objects
	//ShadingNodeCollection m_ShadingNodes;

	/// Saves geometry to a K-3D (XML) document
	sdpResult SaveGeometry(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	/// Saves points to a K-3D (XML) document
	sdpResult SavePoints(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	/// Saves paths to a K-3D (XML) document
	sdpResult SavePaths(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	/// Saves shading nodes to a K-3D (XML) document
	//sdpResult SaveShading(sdpxml::Element& Element, k3dIDependencies* Dependencies);

	/// Loads geometry from a K-3D (XML) document
	sdpResult LoadGeometry(sdpxml::Element& Element);
	/// Loads points from a K-3D (XML) document
	sdpResult LoadPoints(sdpxml::Element& Element);
	/// Loads paths from a K-3D (XML) document
	sdpResult LoadPaths(sdpxml::Element& Element);
	/// Loads shading nodes from a K-3D (XML) document
	//sdpResult LoadShading(sdpxml::Element& Element);

/*
	/// Draws points
	inline void EditorDrawPoints(const gulong Set, const gulong Unset);
	/// Draws path normals
	inline void EditorDrawNormals(const gulong Set, const gulong Unset);
	/// Draws path orientations
	inline void EditorDrawOrientations(const gulong Set, const gulong Unset);
	/// Draws wireframe paths
	inline void EditorDrawWireframePaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset);
	/// Draws solid paths
	inline void EditorDrawSolidPaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset);
	/// Draws solid, textured paths
	inline void EditorDrawSolidTexturedPaths(k3dIEditorEngine* const Engine, const gulong Set, const gulong Unset);
*/

	/// Stores an array of points for RenderMan rendering
	typedef std::vector<sdpVector3> RtPointArray;
	/// Stores an array of scalars for RenderMan rendering
	typedef std::vector<gdouble> RtFloatArray;

/*
	/// Draws a path
	inline void RiPath(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dNurbsPath* const Path);
	/// Draws a 3-sided path
	inline void RiPath3(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dNurbsPath* const Path);
	/// Draws a 4-sided path
	inline void RiPath4(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dNurbsPath* const Path);
	/// Draws a greater-than-4 sided path
	inline void RiPathN(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material, k3dNurbsPath* const Path);
	/// Draws the whole mesh as a SubdivisionMesh
	inline void RiSubdivisionMesh(k3dIRiEngine* const Engine, k3dIRiMaterial* const Material);
*/

	/// Stores the color for drawing selected geometry
	static const sdpVector3 m_SelectedColor;
	/// Stores the color for drawing path normals
	static const sdpVector3 m_PathNormalColor;
};

#endif // K3DNURBSMESH_H
