// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dNurbsPath class, which encapsulates a geometric path
		\author Romain Behar with NURBS students (romainbehar@yahoo.com)
*/

#include <k3dnurbsmesh.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dnurbspath.h>

#include <k3distaterecorder.h>
#include <k3didocument.h>
#include <k3dimaterial.h>

/////////////////////////////////////////////////////////////////////////
// k3dNurbsPath

k3dNurbsPath::k3dNurbsPath(k3dNurbsMesh* const Mesh, const ID PathID) :
	k3dState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::USERVIEWABLE | k3dIState::SAVEABLE),
	m_Mesh(Mesh),
	m_PathID(PathID)//,
	//m_Material(0)
{
}

k3dNurbsPath::~k3dNurbsPath()
{
	std::for_each(m_PathPoints.begin(), m_PathPoints.end(), sdp::DeleteObject());
}

void k3dNurbsPath::PushEditing()
{
	if(!Editing())
		{
			m_PathPointCollectionChanged = false;
			//m_MaterialChanged = false;
			//m_NodalChanged = false;
		}

	k3dEditable::PushEditing();

}

void k3dNurbsPath::PopEditing()
{
	k3dEditable::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_PathPointCollectionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<PathPoints>(m_PathPoints));

					//if(m_MaterialChanged)
					//	changeset->RecordNewState(new k3dValueStateContainer<k3dIMaterial*>(m_Material));

					//if(m_NodalChanged)
						//changeset->RecordNewState(new k3dValueStateContainer<gfloat*>(m_Nodal));
				}
		}

}

sdpResult k3dNurbsPath::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	k3dIObject::ID materialid = 0;
	//if(m_Material)
	//	materialid = dynamic_cast<k3dIObject*>(m_Material)->ObjectID();

	sdpxml::Element& path = Element.Append(sdpxml::Element("path", "",
		sdpxml::Attribute("material", sdpToString(materialid)),
		//sdpxml::Attribute("nodal", sdpToString(nodal)),
		sdpxml::Attribute("id", sdpToString(m_PathID))));


	// Write properties ...
	k3dState::Save(path, Dependencies);

	// Write the pathpoints ...
	if(m_PathPoints.size())
		{
			sdpxml::Element& pathpoints = path.Append(sdpxml::Element("pathpoints"));

			for(PathPointIterator pathpoint = m_PathPoints.begin(); pathpoint != m_PathPoints.end(); pathpoint++)
				(*pathpoint)->Save(pathpoints, Dependencies);
		}

	return SDP_S_OK;
}

class k3dPathLoadPathPoint
{
public:
	k3dPathLoadPathPoint(k3dNurbsPath* const Path) :
		m_Path(Path)
	{
		// Sanity checks ...
		g_assert(m_Path);
	}

	void operator()(sdpxml::Element& Element)
	{
		// Create a new pathpoint and push it into the pathpoint list ...
		k3dNurbsPathPoint* const pathpoint = new k3dNurbsPathPoint(m_Path);
		m_Path->m_PathPoints.push_back(pathpoint);

		// Load the pathpoint and attach it to its point ...
		pathpoint->Load(Element);
	}

private:
	k3dNurbsPath* const m_Path;
};

sdpResult k3dNurbsPath::Load(sdpxml::Element& Element, k3dNurbsMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	// Read the ID (required) ...
	g_return_val_if_fail(sdpxml::ParseAttribute(Element, "id", m_PathID), SDP_E_FAIL);

	// Load the nodal vector... (on ne sait pas si c'est comme a et si c'est ici qu'il faut charger le vecteur nodal
	//sdpxml::ParseAttribute(Element, "nodal", m_Nodal);

	// Read properties ...
	k3dState::Load(Element);

	// Read path material ...
	const k3dIObject::ID materialid = sdpxml::GetAttribute<k3dIObject::ID>(Element, "material", 0);

	// Store the material ID for binding later ...
	//m_Material = reinterpret_cast<k3dIMaterial*>(materialid);

	// Load pathpoints
	sdpxml::ElementPointer pathpoints = sdpxml::FindElement(Element, sdpxml::SameName("pathpoints"));
	if(pathpoints)
		{
			k3dPathLoadPathPoint functor(this);
			sdpxml::ForEach(*pathpoints, sdpxml::SameName("pathpoint"), functor);
		}

	return SDP_S_OK;
}

sdpResult k3dNurbsPath::LoadComplete()
{
/*
	// Bind our material ID to a material ...
	const k3dIObject::ID materialid = reinterpret_cast<k3dIObject::ID>(m_Material);
	m_Material = 0;

	if(materialid)
		{
			std::vector<k3dIObject*> materialobjects;
			k3d::copy_if(m_Mesh->Document()->Objects(), std::back_inserter(materialobjects), k3d::SameID(materialid));

			if(1 == materialobjects.size())
				m_Material = dynamic_cast<k3dIMaterial*>(*materialobjects.begin());
		}
*/

	// Give pathpoints a chance to bind to points ...
	for(PathPointIterator pathpoint = m_PathPoints.begin(); pathpoint != m_PathPoints.end(); pathpoint++)
		(*pathpoint)->LoadComplete(m_Mesh);

	return SDP_S_OK;
}

//////////////////////////////////////////
// k3dINurbsPath implementation

k3dINurbsMesh* k3dNurbsPath::Mesh()
{
	return m_Mesh;
}

const k3dINurbsPath::ID k3dNurbsPath::PathID()
{
	return m_PathID;
}

k3dINurbsPathPointCollection* k3dNurbsPath::Points()
{
	return this;
}

k3dILocationCollection* k3dNurbsPath::Locations()
{
	return this;
}

/*
k3dIMaterialCollection* k3dNurbsPath::Materials()
{
	return this;
}
*/

k3dIState* k3dNurbsPath::State()
{
	return this;
}


/*
sdpResult k3dNurbsPath::PathNormal(double Normal[3])
{
	m_Normal.CopyArray(Normal);
	return SDP_S_OK;
}

sdpResult k3dNurbsPath::SetPathNormal(double Normal[3])
{
	m_Normal = Normal;
	return SDP_S_OK;
}

sdpResult k3dNurbsPath::FlipOrientation()
{
	if(0 == m_PathPoints.size())
		return SDP_S_OK;

	std::rotate(m_PathPoints.begin(), m_PathPoints.begin()+1, m_PathPoints.end());
	std::reverse(m_PathPoints.begin(), m_PathPoints.end());
	return SDP_S_OK;
}

sdpResult k3dNurbsPath::AdvanceFirstPoint()
{
	if(0 == m_PathPoints.size())
		return SDP_S_OK;

	std::rotate(m_PathPoints.begin(), m_PathPoints.begin()+1, m_PathPoints.end());
	return SDP_S_OK;
}
*/


/*
const gfloat* k3dNurbsPath::Nodal()
{
	return m_Nodal;
}

void k3dNurbsPath::SetNodal(const gfloat* VectorNodal)
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
	/ *if(changeset && !m_NodalChanged)
		{
			m_NodalChanged = true;
			changeset->RecordOldState(new k3dValueStateContainer<gfloat*>(m_Nodal));
		}
	* /
	m_Nodal = VectorNodal;
}
*/

/////////////////////////////////////////
// k3dIPathPointCollection implementation

void k3dNurbsPath::VisitPathPoints(k3dIPathPointVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	std::for_each(m_PathPoints.begin(), m_PathPoints.end(), sdp::MakeVisitorAdapter(Visitor, &k3dIPathPointVisitor::VisitPathPoint));
}

k3dINurbsPathPoint* k3dNurbsPath::AddPoint(k3dINurbsControlPoint* const Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(Editing(), 0);

	k3dNurbsPathPoint* pathpoint = new k3dNurbsPathPoint(this, static_cast<k3dNurbsControlPoint*>(Point), m_Mesh->NextPathPointID());
	g_return_val_if_fail(pathpoint, 0);
	m_PathPoints.push_back(pathpoint);

	return pathpoint;
}

sdpResult k3dNurbsPath::RemovePoint(k3dINurbsControlPoint* const Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	for(PathPointIterator pathpoint = m_PathPoints.begin(); pathpoint != m_PathPoints.end(); pathpoint++)
		{
			if((*pathpoint)->Point() == Point)
				{
					k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
					if(changeset)
						{
							changeset->RecordOldState(new k3dDeleteInstanceStateContainer<k3dNurbsPathPoint>(*pathpoint, changeset));

							if(!m_PathPointCollectionChanged)
								{
									m_PathPointCollectionChanged = true;
									changeset->RecordOldState(new k3dValueStateContainer<PathPoints>(m_PathPoints));
								}
						}

					m_PathPoints.erase(pathpoint);
					return SDP_S_OK;
				}
		}

	return SDP_S_FALSE;
}

sdpResult k3dNurbsPath::RemovePathPoint(k3dINurbsPathPoint* const PathPoint)
{
	// Sanity check ...
	g_return_val_if_fail(PathPoint, SDP_E_FAIL);
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	k3dNurbsPathPoint* pathpoint = static_cast<k3dNurbsPathPoint*>(PathPoint);  //!!!!!!!!!!!!!!!!!!!!!!!
	PathPointIterator iterator = std::find(m_PathPoints.begin(), m_PathPoints.end(), pathpoint);
	g_return_val_if_fail(iterator != m_PathPoints.end(), SDP_E_FAIL);

	m_PathPoints.erase(iterator);
	delete pathpoint;

	return SDP_S_OK;
}
gulong k3dNurbsPath::PointCount()
{
	return m_PathPoints.size();
}

K3DNURBSPATHPOINTHANDLE k3dNurbsPath::FirstPathPoint()
{
	return K3DNURBSPATHPOINTHANDLE(m_PathPoints.size() ? 1 : 0);
}

K3DNURBSPATHPOINTHANDLE k3dNurbsPath::NextPathPoint(K3DNURBSPATHPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_PathPoints.size(), 0);

	return K3DNURBSPATHPOINTHANDLE(gulong(Point) < m_PathPoints.size() ? gulong(Point)+1 : 0);
}

k3dINurbsPathPoint* k3dNurbsPath::PathPoint(K3DNURBSPATHPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_PathPoints.size(), 0);

	return m_PathPoints[gulong(Point)-1];
}

/////////////////////////////////////////
// k3dILocationCollection implementation

void k3dNurbsPath::VisitLocations(k3dILocationVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	std::for_each(m_PathPoints.begin(), m_PathPoints.end(), sdp::MakeVisitorAdapter(Visitor, &k3dILocationVisitor::VisitLocation));
}

gulong k3dNurbsPath::LocationCount()
{
	return m_PathPoints.size();
}

K3DLOCATIONHANDLE k3dNurbsPath::FirstLocation()
{
	return K3DLOCATIONHANDLE(m_PathPoints.size() ? 1 : 0);
}

K3DLOCATIONHANDLE k3dNurbsPath::NextLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_PathPoints.size(), 0);

	return K3DLOCATIONHANDLE(gulong(Location) < m_PathPoints.size() ? gulong(Location)+1 : 0);
}

k3dILocation* k3dNurbsPath::Location(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_PathPoints.size(), 0);

	return m_PathPoints[gulong(Location)-1]->m_Point;
}
/*
/////////////////////////////////////////
// k3dIMaterialCollection implementation

k3dIMaterial* k3dPath::Material()
{
	return m_Material;
}

sdpResult k3dPath::SetMaterial(k3dIMaterial* const Material)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	if(!m_MaterialChanged)
		{
			m_MaterialChanged = true;

			k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
			if(changeset)
				changeset->RecordOldState(new k3dValueStateContainer<k3dIMaterial*>(m_Material));
		}

	m_Material = Material;

	return SDP_S_OK;
}
*/



/////////////////////////////////////////
// k3dIState implementation

sdpResult k3dNurbsPath::SetState(gulong Mask, bool State)
{
	// Perform an undo-able state change ...
	RawSetState(Mask, State, m_Mesh->Document()->StateRecorder());
	return SDP_S_OK;
}

