#ifndef K3DNURBSPATH_H
#define K3DNURBSPATH_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dNurbsPath class, which encapsulates a geometric nurbs path
		\author Romain Behar with NURBS students (romainbehar@yahoo.com)
*/

#include <k3deditable.h>
#include <k3dnurbscontrolpoint.h>
#include <k3dnurbspathpoint.h>
#include <k3dstate.h>

#include <k3dilocationcollection.h>
#include <k3dimaterialcollection.h>
#include <k3dinurbspath.h>
#include <k3dinurbspathpointcollection.h>

#include <sdpgl.h>

// Forward class declarations
class k3dNurbsMesh;

///////////////////////////////////////////////////////////////////////////////////////////
// k3dNurbsPath

/// Encapsulates a closed path / Bezier spline / Bezier patch
class k3dNurbsPath :
	public k3dINurbsPath,
	public k3dINurbsPathPointCollection,
	public k3dILocationCollection,
	//public k3dIMaterialCollection,
	public k3dEditable,
	public k3dState
{
public:
	k3dNurbsPath(k3dNurbsMesh* const Mesh, ID PathID);
	virtual ~k3dNurbsPath();

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	// k3dIPath implementation
	k3dINurbsMesh* Mesh();
	k3dINurbsPathPointCollection* Points();
	k3dILocationCollection* Locations();
	//k3dIMaterialCollection* Materials();
	k3dIState* State();
	//sdpResult PathNormal(double Normal[3]);
	//sdpResult SetPathNormal(double Normal[3]);
	//sdpResult FlipOrientation();
	//sdpResult AdvanceFirstPoint();
	//const gdouble*  Nodal();

	//void SetNodal(const gdouble* VectorNodal);


	const ID PathID();

	// k3dINurbsPathPointCollection implementation
	void VisitPathPoints(k3dIPathPointVisitor* const Visitor);
	k3dINurbsPathPoint* AddPoint(k3dINurbsControlPoint* const Point);
	sdpResult RemovePoint(k3dINurbsControlPoint* const Point);
	sdpResult RemovePathPoint(k3dINurbsPathPoint* const PathPoint);

	gulong PointCount();
	K3DNURBSPATHPOINTHANDLE FirstPathPoint();
	K3DNURBSPATHPOINTHANDLE NextPathPoint(K3DNURBSPATHPOINTHANDLE Point);
	k3dINurbsPathPoint* PathPoint(K3DNURBSPATHPOINTHANDLE Point);

	// k3dILocationCollection implementation
	void VisitLocations(k3dILocationVisitor* const Visitor);
	gulong LocationCount();
	K3DLOCATIONHANDLE FirstLocation();
	K3DLOCATIONHANDLE NextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* Location(K3DLOCATIONHANDLE Location);

	// k3dIMaterialCollection implementation
	//k3dIMaterial* Material();
	//sdpResult SetMaterial(k3dIMaterial* const Material);

	// k3dIState implementation
	sdpResult SetState(const gulong Mask, const bool State);

	// k3dIPersist implementation
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpxml::Element& Element, k3dNurbsMesh* Mesh);
	sdpResult LoadComplete();

	/// Returns the nurbs path center point (as an average)
	sdpVector3 CenterPoint()
	{
		sdpVector3 result;

		const gulong size = m_PathPoints.size();
		for(gulong i = 0; i < size; i++)
			result += m_PathPoints[i]->m_Point->m_AnimXYZ;

		if(size)
			result /= size;

		return result;
	}

	/// Draws a visual representation of the path's orientation
	void DrawOpenGLOrientation()
	{
		// Ignore paths with two-or-fewer points ...
		if(m_PathPoints.size() < 3)
			return;

		// Get our first segment ...
		const sdpVector3 center = CenterPoint();
		const sdpVector3 p0 = m_PathPoints[0]->m_Point->m_AnimXYZ;
		const sdpVector3 p2 = m_PathPoints[1]->m_Point->m_AnimXYZ;
		const sdpVector3 p1 = sdpLerp(p0, p2, 0.5);

		const sdpVector3 p3 = sdpLerp(center, p0, 0.8);
		const sdpVector3 p5 = sdpLerp(center, p2, 0.8);
		const sdpVector3 p4 = sdpLerp(p5, p3, 0.2);

		glBegin(GL_LINES);

		glVertex3dv(const_cast<sdpVector3&>(p3));
		glVertex3dv(const_cast<sdpVector3&>(p5));

		glVertex3dv(const_cast<sdpVector3&>(p5));
		glVertex3dv(sdpLerp(p4, p1, 0.5));

		glEnd();
	}

	/// Draws the wireframe representation of the nurbs path
	void DrawOpenGLWireframe()
	{
		glBegin(GL_POLYGON);
		for(gulong i = 0; i < m_PathPoints.size(); i++)
			glVertex3dv(m_PathPoints[i]->m_Point->m_AnimXYZ);
		glEnd();
	}

	/// Draws a solid representation of the nurbs path
	void DrawOpenGLSolid()
	{
		glBegin(GL_POLYGON);
		for(gulong i = 0; i < m_PathPoints.size(); i++)
			{
				glNormal3dv(m_PathPoints[i]->m_NormalXYZ);
				glVertex3dv(m_PathPoints[i]->m_Point->m_AnimXYZ);
			}
		glEnd();
	}

	/// Draws a solid, textured representation of the nurbs path
	void DrawOpenGLSolidTextured()
	{
		glBegin(GL_POLYGON);
		for(gulong i = 0; i < m_PathPoints.size(); i++)
			{
				glTexCoord3dv(m_PathPoints[i]->m_TextureXYZ);
				glNormal3dv(m_PathPoints[i]->m_NormalXYZ);
				glVertex3dv(m_PathPoints[i]->m_Point->m_AnimXYZ);
			}
		glEnd();
	}

	/// Recalculates the nurbs path's normal vector
	void CalculateNormal()
	{
		// Reset the normal to zero ...
		m_Normal = 0.0;

		// Skip degenerate paths ...
		gulong pointcount = m_PathPoints.size();
		if(pointcount < 3)
			return;

		// Integrate the path normal from its' edge normals ...
		sdpVector3 normal1;
		sdpVector3 normal2;

		for(gulong i = 2; i < pointcount; i++)
			{
				normal1 = m_PathPoints[i]->m_Point->m_AnimXYZ - m_PathPoints[i-1]->m_Point->m_AnimXYZ;
				normal2 = m_PathPoints[i-2]->m_Point->m_AnimXYZ - m_PathPoints[i-1]->m_Point->m_AnimXYZ;

				m_Normal += normal1 ^ normal2;
			}

		normal1 = m_PathPoints[0]->m_Point->m_AnimXYZ - m_PathPoints[pointcount-1]->m_Point->m_AnimXYZ;
		normal2 = m_PathPoints[pointcount-2]->m_Point->m_AnimXYZ - m_PathPoints[pointcount-1]->m_Point->m_AnimXYZ;

		m_Normal += normal1 ^ normal2;
	}

	/// Stores the parent nurbs mesh that owns this nurbs path
	k3dNurbsMesh* m_Mesh;
	/// Stores the unique (within the parent mesh) path ID
	ID m_PathID;
	/// Stores the material assigned to this path
	//k3dIMaterial* m_Material;
	/// Stores the path normal
	sdpVector3 m_Normal;
	/// Stores the nodal vector
	//gdouble*  m_Nodal;
	/// A collection of k3dNurbsPathPoint structures contained by the path
	typedef std::vector<k3dNurbsPathPoint*> PathPoints;
	typedef PathPoints::iterator PathPointIterator;
	PathPoints m_PathPoints;

	/// Set to true iff the collection of pathpoints has been modified during the current edit session
	bool m_PathPointCollectionChanged;
	/// Set to true iff the material has been modified during the current edit session
	//bool m_MaterialChanged;
	// Set to true iff the nodal vector has been modified during the current edit session
	//bool m_NodalChanged;

};

#endif // K3DNURBSPATH_H
