
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements k3dObject, a default implementation of the k3dIObject interface, for use as a base class for K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dobject.h>
#include <k3dstaterecorder.h>
#include <k3dvariable.h>

#include <k3diapplication.h>
#include <k3dieventobserver.h>
#include <k3dihierarchy.h>
#include <k3dimouseeventobserver.h>
#include <k3diobjectinfo.h>
#include <k3diobjectinitializer.h>
#include <k3distaterecorder.h>

#include <algorithm>
#include <iostream>

////////////////////////////////////////////////////////////////////////////
// class k3dObject

const sdpString k3dObject::Event_UserMessage1 = "k3dObject::Event_UserMessage1";

k3dObject::k3dObject(sdpIUnknown* Initializer) :
	k3dCommandNode(0, "object"),
	k3dState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::USERVIEWABLE)
{
	// Sanity checks ...
	g_assert(Initializer);
	
	k3dIObjectInitializer* const initializer = dynamic_cast<k3dIObjectInitializer*>(Initializer);
	g_assert(initializer);
	
	m_ObjectID = initializer->ObjectID();
	g_assert(m_ObjectID);
	
	m_ObjectInfo = initializer->ObjectInfo();
	g_assert(m_ObjectInfo);

	m_Document = initializer->Document();
	g_assert(m_Document);

	m_Hierarchy = initializer->Document()->Hierarchy();
	g_assert(m_Hierarchy);

	m_Name = k3dGetUniqueName(Document()->Objects(), "Untitled " + sdpString(ObjectInfo()->Name()));
			
	SetCommandNodeName(m_Name);
	SetParentCommandNode(dynamic_cast<k3dICommandNode*>(Document()));
}

k3dObject::~k3dObject()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "Destroying object [" << Name() << "]" << std::endl;

	// Notify our clients that we're going away ...
	Notify(k3dIObject::Event_Destroyed, 0);
}

////////////////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult k3dObject::InitComplete()
{
	return SDP_S_OK;
}

const k3dIObject::ID k3dObject::ObjectID()
{
	return m_ObjectID;
}

void k3dObject::SetObjectID(const ID ObjectID)
{
	// Sanity checks ...
	g_assert(ObjectID);
	
	m_ObjectID = ObjectID;
	Notify(Event_IDChanged, sdpToString(m_ObjectID));
}

SDPCSTRING k3dObject::Name()
{
	return m_Name;
}

sdpResult k3dObject::SetName(SDPCSTRING Name)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	// Make sure it's a unique name ...
	const sdpString newname = k3dGetUniqueName(Document()->Objects(), Name);

	// Record the state change ...
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			changeset->RecordOldState(new k3dValueStateContainer<sdpString>(m_Name));
			changeset->RecordOldAction(new k3dActionNotify(this, k3dIObject::Event_Renamed, m_Name, 0));
		}
		
	m_Name = newname;

	if(changeset)
		{
			changeset->RecordNewState(new k3dValueStateContainer<sdpString>(m_Name));
			changeset->RecordNewAction(new k3dActionNotify(this, k3dIObject::Event_Renamed, m_Name, 0));
		}
		
	// Notify our clients that we got renamed ...
	Notify(k3dIObject::Event_Renamed, m_Name);

	// Reposition ourselves in the command-node hierarchy
	SetCommandNodeName(m_Name);
	SetParentCommandNode(dynamic_cast<k3dICommandNode*>(Document()));

	return SDP_S_OK;
}

k3dIState* k3dObject::State()
{
	return this;
}

sdpResult k3dObject::EditObject()
{
	return SDP_S_OK;
}

sdpResult k3dObject::ObjectDeleted()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "Deleting object [" << Name() << "]" << std::endl;

	// Notify our clients that we're going away ...
	Notify(k3dIObject::Event_Deleted, 0);
	
	// Remove ourselves from the command-node tree ...
	SetParentCommandNode(0);

	return SDP_S_OK;
}

sdpResult k3dObject::ObjectUnDeleted()
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "Un-Deleting object [" << Name() << "]" << std::endl;

	// Notify our clients that we're going away ...
	Notify(k3dIObject::Event_UnDeleted, 0);
	
	// Add ourselves to the command-node tree ...
	SetParentCommandNode(dynamic_cast<k3dICommandNode*>(Document()));

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dIState overrides ...

sdpResult k3dObject::SetState(gulong Mask, bool State)
{
	// If we're being deselected and we're the active object ...
	if(Mask & k3dIState::SELECTED)
		{
			if(!State)
				{
					if(Document()->MouseFocus() == dynamic_cast<k3dIMouseEventObserver*>(this))
						Document()->SetMouseFocus(0);
				}
		}

	RawSetState(Mask, State);
	
	if(Mask & k3dIState::SELECTED)
		Notify(State ? k3dIState::Event_Selected : k3dIState::Event_Deselected, "");
		
	if(Mask & k3dIState::VISIBLE)
		Notify(State ? k3dIState::Event_Visible : k3dIState::Event_Hidden, "");
	
	Notify(k3dIState::Event_Changed, "");
	
	return SDP_S_OK;
}

