#ifndef K3DOBJECTCOLLECTION_H
#define K3DOBJECTCOLLECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcollections.h>
#include <k3diobjectcollection.h>

namespace k3d
{

/// Applies a functor to every k3dIObject in a collection that supports a specific interface
template<typename Interface, typename Functor>
Functor for_each(k3dIObjectCollection* const ObjectCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectCollection, F);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			// Test for the requested interface ...
			Interface* const i = dynamic_cast<Interface*>(ObjectCollection->Object(handle));
			if(i)
				F(i);
		}

	return F;
}
	
/// Applies a functor to every k3dIObject in a collection
template<typename Functor>
Functor for_each(k3dIObjectCollection* const ObjectCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectCollection, F);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		F(ObjectCollection->Object(handle));

	return F;
}
	
/// Applies a functor to every k3dIObject in a collection that satisfies a specific interface and the predicate
template<typename Interface, typename Functor, typename Predicate>
Functor for_each_if(k3dIObjectCollection* const ObjectCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectCollection, F);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			k3dIObject* const object = ObjectCollection->Object(handle);
			if(P(object))
				{
					// Test for the requested interface ...
					Interface* const i = dynamic_cast<Interface*>(ObjectCollection->Object(handle));
					if(i)
						F(i);
				}
		}

	return F;
}

/// Applies a functor to every k3dIObject in a collection that satisfies the predicate
template<typename Functor, typename Predicate>
Functor for_each_if(k3dIObjectCollection* const ObjectCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectCollection, F);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			k3dIObject* const object = ObjectCollection->Object(handle);
		
			if(P(object))
				F(object);
		}

	return F;
}

/// Copies k3dIObject references from a k3dIObjectCollection into a generic container if they support the requested interface
template<typename Interface, typename InsertIterator>
void copy(k3dIObjectCollection* const ObjectCollection, InsertIterator D)
{
	// Sanity checks ...
	g_return_if_fail(ObjectCollection);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			// Test for the requested interface ...
			Interface* const i = dynamic_cast<Interface*>(ObjectCollection->Object(handle));
			if(i)
				*D++ = i;
		}
}

/// Copies k3dIObject references from a k3dIObjectCollection into a generic container
template<typename InsertIterator>
void copy(k3dIObjectCollection* const ObjectCollection, InsertIterator D)
{
	// Sanity checks ...
	g_return_if_fail(ObjectCollection);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		*D++ = ObjectCollection->Object(handle);
}

/// Copies k3dIObject references that satisfy a predicate from a k3dIObjectCollection into a generic container, if they support the requested interface
template<typename Interface, typename InsertIterator, typename Predicate>
void copy_if(k3dIObjectCollection* const ObjectCollection, InsertIterator D, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(ObjectCollection);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			k3dIObject* const object = ObjectCollection->Object(handle);
			if(P(object))
				{
					// Test for the requested interface ...
					Interface* const i = dynamic_cast<Interface*>(ObjectCollection->Object(handle));
					if(i)
						*D++ = i;
				}
		}
}

/// Copies k3dIObject references that satisfy a predicate from a k3dIObjectCollection into a generic container
template<typename InsertIterator, typename Predicate>
void copy_if(k3dIObjectCollection* const ObjectCollection, InsertIterator D, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(ObjectCollection);

	// For each object in the document ...
	for(K3DOBJECTHANDLE handle = ObjectCollection->FirstObject(); handle; handle = ObjectCollection->NextObject(handle))
		{
			k3dIObject* const object = ObjectCollection->Object(handle);
			if(P(object))
				*D++ = object;
		}
}

} // namespace k3d

#endif // K3DOBJECTCOLLECTION_H

