// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dObjectDialog class, which provides a standard dialog implementation for K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dobject.h>
#include <k3dobjectdialog.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3dimouseeventobserver.h>

#include <sdpgtkevents.h>

const sdpString k3dObjectDialog::control_setfocus = "setfocus";
const sdpString k3dObjectDialog::control_ok = "ok";
const sdpString k3dObjectDialog::control_cancel = "cancel";
const sdpString k3dObjectDialog::control_close = "close";

k3dObjectDialog::k3dObjectDialog(k3dObject* const Object, const bool GrabMouseFocus, const bool AutoDestroy, const bool ShowObjectName, SDPCSTRING CommandNodeName, k3dWindowGeometryStore* const WindowGeometryStore) :
	k3dObjectDialogBase(Object, CommandNodeName, WindowGeometryStore),
	m_Object(Object),
	m_AutoDestroy(AutoDestroy),
	m_ShowObjectName(ShowObjectName)
{
	// Sanity checks ...
	g_assert(m_Object);
	
	m_Object->AddEventObserver(this, k3dIObject::Event_Renamed, "k3dObjectDialog");
	m_Object->AddEventObserver(this, k3dIObject::Event_Deleted, "k3dObjectDialog");
	m_Object->AddEventObserver(this, k3dIObject::Event_UpdateUI, "k3dObjectDialog");
	m_Object->AddEventObserver(this, k3dObject::Event_UserMessage1, "k3dObjectDialog");
	
	if(GrabMouseFocus)
		{
			k3dObjectDialog::GrabMouseFocus();
			k3dRedrawAll(m_Object->Document(), false);
		}
}

bool k3dObjectDialog::LoadGTKMLTemplate(SDPCSTRING Subdirectory, SDPCSTRING Template)
{
	// Sanity checks ...
	g_assert_string(Subdirectory);
	g_assert_string(Template);
	
	// Load the template ...
	g_return_val_if_fail(k3dObjectDialogBase::LoadGTKMLTemplate(Subdirectory, Template), false);
	
	// Update the titlebar ...
	if(m_ShowObjectName)
		ObjectRenamed(m_Object->Name());
	
	return true;
}

k3dObjectDialog::~k3dObjectDialog()
{
	m_Object->RemoveEventObserver(this);
}

/////////////////////////////////////////////////////////////
// k3dICommandNode overrides
sdpResult k3dObjectDialog::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_assert_string(Name);

	if(Name == control_ok)
		{
			return InteractiveActivateButton(control_ok, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
		}
	else if(Name == control_cancel)
		{
			return InteractiveActivateButton(control_cancel, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
		}
	else if(Name == control_close)
		{
			return InteractiveActivateButton(control_close, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
		}

	return k3dObjectDialogBase::Command(Name, Arguments, Result);
}

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

void k3dObjectDialog::ObjectRenamed(SDPCSTRING NewName)
{
	// If we're not open, we're done ...
	g_return_if_fail(Root());
	
	RootWindow().SetTitle(NewName);
}

void k3dObjectDialog::ObjectDeleted()
{
	// Since the object is already going away, we don't need to destroy it ourselves ...
	m_AutoDestroy = false;
	delete this;
}

////////////////////////////////////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult k3dObjectDialog::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	const sdpString eventdata(EventData);
	
	if(event == k3dIObject::Event_Renamed)
		{
			if(m_ShowObjectName)
				ObjectRenamed(m_Object->Name());
	
			return SDP_S_OK;
		}
	else if(event == k3dIObject::Event_Deleted)
		{
			ObjectDeleted();
			return SDP_S_OK;
		}
	else if(event == k3dIObject::Event_UpdateUI)
		{
			CallUpdateControls();
			return SDP_S_OK;
		}
	else if(event == k3dObject::Event_UserMessage1)
		{
			sdpGtkEntry entry(Entry("usermessage1"));
			if(entry.Attached())
				entry.SetText(eventdata);
			return SDP_S_OK;
		}
	
	return k3dObjectDialogBase::ObserveEvent(Source, Event, EventData, ObserverData);
}
	
void k3dObjectDialog::OnOK()
{
	// Record this as a user action ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_ok, 0);
	k3dObjectDialogBase::OnOK();
}

void k3dObjectDialog::OnCancel()
{
	// Record this as a user action ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_cancel, 0);
	k3dObjectDialogBase::OnCancel();
}

void k3dObjectDialog::OnClose()
{
	// Record this as a user action ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_close, 0);
	k3dObjectDialogBase::OnClose();
}

void k3dObjectDialog::CloseDialog()
{
	// If AutoDestroy is enabled, kill our owning object (we'll be notified and destroy ourselves, so override the default implementation) 
	if(m_AutoDestroy)
		k3dDeleteObject(m_Object);
	else
		k3dObjectDialogBase::CloseDialog();
}

void k3dObjectDialog::GrabMouseFocus()
{
	k3dIMouseEventObserver* const observer = dynamic_cast<k3dIMouseEventObserver*>(m_Object);
	g_return_if_fail(observer);
		
	// Make our object the active object ...
	m_Object->Document()->SetMouseFocus(observer);
}

