#ifndef K3DOBJECTINFOCOLLECTION_H
#define K3DOBJECTINFOCOLLECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcollections.h>
#include <k3diobjectinfocollection.h>

namespace
{
	
template<typename Functor>
class for_each_visitor :
	public k3dIObjectInfoVisitor
{
public:
	for_each_visitor(Functor& F) :
		m_F(F)
	{
	}
	
	void VisitObjectInfo(k3dIObjectInfo* const ObjectInfo)
	{
		m_F(ObjectInfo);
	}

private:
	Functor& m_F;
};
	
template<typename Functor, typename Predicate>
class for_each_if_visitor :
	public k3dIObjectInfoVisitor
{
public:
	for_each_if_visitor(Functor& F, Predicate& P) :
		m_F(F),
		m_P(P)
	{
	}
	
	void VisitObjectInfo(k3dIObjectInfo* const ObjectInfo)
	{
		if(m_P(ObjectInfo))
			m_F(ObjectInfo);
	}

private:
	Functor& m_F;
	Predicate& m_P;
};

template<typename InsertIterator>
class copy_visitor :
	public k3dIObjectInfoVisitor
{
public:
	copy_visitor(InsertIterator& I) :
		m_I(I)
	{
	}
	
	void VisitObjectInfo(k3dIObjectInfo* const ObjectInfo)
	{
		*m_I++ = ObjectInfo;
	}

private:
	InsertIterator& m_I;
};
	
template<typename InsertIterator, typename Predicate>
class copy_if_visitor :
	public k3dIObjectInfoVisitor
{
public:
	copy_if_visitor(InsertIterator& I, Predicate& P) :
		m_I(I),
		m_P(P)
	{
	}
	
	void VisitObjectInfo(k3dIObjectInfo* const ObjectInfo)
	{
		if(m_P(ObjectInfo))
			*m_I++ = ObjectInfo;
	}

private:
	InsertIterator& m_I;
	Predicate& m_P;
};

} // namespace

namespace k3d
{

/// Applies a functor to every k3dIObjectInfo in a collection
template<typename Functor>
Functor for_each(k3dIObjectInfoCollection* const ObjectInfoCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectInfoCollection, F);

	for_each_visitor<Functor> v(F);
	ObjectInfoCollection->VisitObjectInfo(&v);

	return F;
}

/// Applies a functor to every k3dIObjectInfo in a collection that satisfies the predicate
template<typename Functor, typename Predicate>
Functor for_each_if(k3dIObjectInfoCollection* const ObjectInfoCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(ObjectInfoCollection, F);

	for_each_if_visitor<Functor, Predicate> v(F, P);
	ObjectInfoCollection->VisitObjectInfo(&v);

	return F;
}

/// Copies k3dIObjectInfo references from a k3dIObjectInfoCollection into a generic container
template<typename InsertIterator>
void copy(k3dIObjectInfoCollection* const ObjectInfoCollection, InsertIterator I)
{
	// Sanity checks ...
	g_return_if_fail(ObjectInfoCollection);

	copy_visitor<InsertIterator> v(I);
	ObjectInfoCollection->VisitObjectInfo(&v);
}

/// Copies k3dIObjectInfo references that satisfy a predicate from a k3dIObjectInfoCollection into a generic container
template<typename InsertIterator, typename Predicate>
void copy_if(k3dIObjectInfoCollection* const ObjectInfoCollection, InsertIterator I, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(ObjectInfoCollection);

	copy_if_visitor<InsertIterator, Predicate> v(I, P);
	ObjectInfoCollection->VisitObjectInfo(&v);
}

} // namespace k3d

#endif // K3DOBJECTINFOCOLLECTION_H

