// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dObjectSelector class, which provides a user interface for either creating new, or selecting existing, K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dobjectinfocollection.h>
#include <k3dobjectselector.h>
#include <k3doptions.h>
#include <k3dvariable.h>

#include <k3diapplication.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>

#include <sdpgtkevents.h>

////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dObjectSelectorInterfaceFilter

k3dObjectSelectorInterfaceFilter::k3dObjectSelectorInterfaceFilter(sdp::ITypeInfo* const InterfaceType) :
	m_InterfaceType(InterfaceType)
{
	// Sanity checks ...
	assert(m_InterfaceType);
}

k3dObjectSelectorInterfaceFilter::~k3dObjectSelectorInterfaceFilter()
{
	delete m_InterfaceType;
}

bool k3dObjectSelectorInterfaceFilter::Allow(k3dIObjectInfo* const ObjectInfo)
{
	return SDP_S_OK == ObjectInfo->QueryObjectInterface(*m_InterfaceType) ? true : false;
}

bool k3dObjectSelectorInterfaceFilter::Allow(k3dIObject* const Object)
{
	return m_InterfaceType->IsImplementedBy(Object);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dObjectSelector

const sdpString k3dObjectSelector::control_optionmenu = "optionmenu";
const sdpString k3dObjectSelector::control_menu = "menu";
const sdpString k3dObjectSelector::control_selectnone = "selectnone";
const sdpString k3dObjectSelector::control_selectobject = "selectobject";
const sdpString k3dObjectSelector::control_newobject = "newobject";
const sdpString k3dObjectSelector::control_editobject = "editobject";

const sdpString k3dObjectSelector::data_object = "k3dObjectSelector::object";
const sdpString k3dObjectSelector::data_objectinfo = "k3dObjectSelector::objectinfo";

k3dObjectSelector::k3dObjectSelector(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName) :
	k3dObjectSelectorBase(ParentCommandNode, CommandNodeName),
	m_ObjectContainer(0),
	m_AllowNone(true),
	m_Filter(0),
	m_ObjectContainerEvents(0),
	m_DocumentEvents(0)
{
	// Create our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<hbox homogeneous=\"false\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"
				"<optionmenu name=\"optionmenu\" expand=\"true\" fill=\"true\">"
					"<menu name=\"menu\"/>"
				"</optionmenu>"
				"<button name=\"editobject\">"
					"<event signal=\"clicked\" name=\"editobject\"/>"
					"<arrow direction=\"right\" shadowtype=\"in\"/>"
				"</button>"
			"</hbox>"
		"</gtkml>\n");

	// Load our template ...
	g_return_if_fail(Load(uitemplate, ""));

	AddCommand(control_editobject, control_editobject, CONTROL_BUTTON);

	m_ObjectContainerEvents = new k3dEventConnector(this, k3dIVariable::Event_Changed, "k3dObjectSelector", new k3dEventConnector(this, k3dIVariable::Event_Deleted, "k3dObjectSelector", 0));
	m_DocumentEvents = new k3dEventConnector(this, k3dIObjectCollection::Event_Changed, "k3dObjectSelector", 0);
}

k3dObjectSelector::~k3dObjectSelector()
{
	// Get rid of our filter ...
	delete m_Filter;

	// Cancel event notification ...
	delete m_ObjectContainerEvents;
	delete m_DocumentEvents;

	// No more SDPGTK events from this point forward ...
	DisconnectAllEvents();

	// Clear out our storage ...
	Clear();
}

/// Called by the framework when being instantiated from a GTKML (XML) document
bool k3dObjectSelector::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);

	g_assert(Element.Name() == "k3dobjectselector");

	return true;
}

// UI creation ...
bool k3dObjectSelector::Attach(k3dObjectVariable* const ObjectContainer,
	k3dICommandRecorder* const CommandRecorder,
	k3dIStateRecorder* const StateRecorder,
	SDPCSTRING StateChangeName,
	k3dIAction* const PushEditingAction,
	k3dIAction* const ChangeAction,
	k3dIAction* const PopEditingAction,
	const bool AllowNone,
	k3dObjectSelectorFilter* const Filter)
{
	// Sanity checks ...
	g_assert(ObjectContainer);

	// Attach the base class ...
	g_return_val_if_fail(k3dObjectSelectorBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);

	// Store data ...
	m_ObjectContainer = ObjectContainer;
	m_AllowNone = AllowNone;
	m_Filter = Filter;

	// Enable event notification ...
	m_ObjectContainerEvents->Attach(m_ObjectContainer);
	m_DocumentEvents->Attach(dynamic_cast<k3dIEventSource*>(m_ObjectContainer->Document()->Objects()));

	// Update ourselves ...
	UpdateSelection();

	return true;
}

const sdpString k3dObjectSelector::CustomType() const
{
	return sdpString("k3dObjectSelector");
}

sdpResult k3dObjectSelector::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	const sdpString name(Name);
	const sdpString arguments(Arguments);

	if(name == control_newobject)
		{
			// Look for an object info with the correct name ...
			typedef std::vector<k3dIObjectInfo*> ObjectInfoCollection;
			ObjectInfoCollection objectinfocollection;
			k3d::copy(k3dApplication()->Objects(), std::back_inserter(objectinfocollection));

			k3dIObjectInfo* objectinfo = 0;
			for(ObjectInfoCollection::iterator info = objectinfocollection.begin(); info != objectinfocollection.end(); info++)
				{
					if((*info)->Name() == arguments)
						{
							objectinfo = *info;
							break;
						}
				}
			g_return_val_if_fail(objectinfo, SDP_E_FAIL);

			// Find the correct menu item ...
			MenuItemIterator menuitem;
			for(menuitem = m_MenuItems.begin(); menuitem != m_MenuItems.end(); menuitem++)
				{
					if((0 == menuitem->GetData(data_object)) && (objectinfo == menuitem->GetData(data_objectinfo)))
						break;
				}
			g_return_val_if_fail(menuitem != m_MenuItems.end(), SDP_E_FAIL);

			InteractiveMoveTo(control_optionmenu, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);

			GdkEventButton event;
			memset(&event, 0, sizeof(event));
			event.type = GDK_BUTTON_PRESS;
			event.button = 1;
			event.time = GDK_CURRENT_TIME;
			gtk_widget_event(Widget(control_optionmenu), reinterpret_cast<GdkEvent*>(&event));

			menuitem->InteractiveWarpPointer(k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);
			menuitem->InteractiveActivate();
			gtk_menu_popdown(Menu(control_menu));

			return SDP_S_OK;
		}
	else if(name == control_editobject)
		{
			InteractiveActivateButton(control_editobject, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
			return SDP_S_OK;
		}
	else if(name == control_selectobject)
		{
			// Find an object with the correct name ...
			k3dIObject* object = 0;
			for(K3DOBJECTHANDLE handle = m_ObjectContainer->Document()->Objects()->FirstObject(); handle; handle = m_ObjectContainer->Document()->Objects()->NextObject(handle))
				{
					if(m_ObjectContainer->Document()->Objects()->Object(handle)->Name() == arguments)
						{
							object = m_ObjectContainer->Document()->Objects()->Object(handle);
							break;
						}
				}
			g_return_val_if_fail(object, SDP_E_FAIL);

			// Find the correct menu item ...
			MenuItemIterator menuitem;
			for(menuitem = m_MenuItems.begin(); menuitem != m_MenuItems.end(); menuitem++)
				{
					if((object == menuitem->GetData(data_object)) && (0 == menuitem->GetData(data_objectinfo)))
						break;
				}
			g_return_val_if_fail(menuitem != m_MenuItems.end(), SDP_E_FAIL);

			InteractiveMoveTo(control_optionmenu, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);

			GdkEventButton event;
			memset(&event, 0, sizeof(event));
			event.type = GDK_BUTTON_PRESS;
			event.button = 1;
			event.time = GDK_CURRENT_TIME;
			gtk_widget_event(Widget(control_optionmenu), reinterpret_cast<GdkEvent*>(&event));

			menuitem->InteractiveWarpPointer(k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);
			menuitem->InteractiveActivate();
			gtk_menu_popdown(Menu(control_menu));

			return SDP_S_OK;
		}
	else if(name == control_selectnone)
		{
			// As a sanity check, make sure we support this ...
			g_return_val_if_fail(m_AllowNone, SDP_E_FAIL);

			// Find the correct menu item ...
			MenuItemIterator menuitem;
			for(menuitem = m_MenuItems.begin(); menuitem != m_MenuItems.end(); menuitem++)
				{
					if((0 == menuitem->GetData(data_object)) && (0 == menuitem->GetData(data_objectinfo)))
						break;
				}
			g_return_val_if_fail(menuitem != m_MenuItems.end(), SDP_E_FAIL);

			InteractiveMoveTo(control_optionmenu, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);

			GdkEventButton event;
			memset(&event, 0, sizeof(event));
			event.type = GDK_BUTTON_PRESS;
			event.button = 1;
			event.time = GDK_CURRENT_TIME;
			gtk_widget_event(Widget(control_optionmenu), reinterpret_cast<GdkEvent*>(&event));

			menuitem->InteractiveWarpPointer(k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true, false);
			menuitem->InteractiveActivate();
			gtk_menu_popdown(Menu(control_menu));

			return SDP_S_OK;
		}

	return k3dObjectSelectorBase::Command(Name, Arguments, Result);
}

sdpResult k3dObjectSelector::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIVariable::Event_Changed)
		{
			UpdateSelection();

			// Notify any interested parties that things changed ...
			if(m_ChangeAction)
				m_ChangeAction->DoAction();

			return SDP_S_OK;
		}
	else if(event == k3dIObjectCollection::Event_Changed)
		{
			UpdateSelection();
			return SDP_S_OK;
		}
	else if(event == k3dIVariable::Event_Deleted)
		{
			m_ObjectContainerEvents->Detach();
			m_ObjectContainer = 0;

			return SDP_S_OK;
		}

	return k3dObjectSelectorBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void k3dObjectSelector::UpdateSelection()
{
	// If we haven't been attached, yet, we're done ...
	if(0 == m_ObjectContainer)
		return;

	// Delete any leftover events ...
	for(EventIterator event = m_Events.begin(); event != m_Events.end(); event++)
		{
			(*event)->Disconnect();
			DeleteEvent(*event);
		}
	m_Events.clear();

	// Delete any leftover menu items ...
	for(MenuItemIterator item = m_MenuItems.begin(); item != m_MenuItems.end(); item++)
		item->Destroy();
	m_MenuItems.clear();

	// Get our menu object ...
	sdpGtkMenu menu = Menu(control_menu);
	menu.Show();

	// Keep track of the menu item index so we can select one ...
	guint index = 0;
	guint selecteditem = 0;

	// Insert a "none" choice ...
	if(m_AllowNone)
		{
			// Create the "none" option ...
			sdpGtkMenuItem menuitem;
			menuitem.Create("--None--");
			menuitem.Show();
			menuitem.SetData(data_object, 0);
			menuitem.SetData(data_objectinfo, 0);
			menu.Append(menuitem);

			m_MenuItems.push_back(menuitem);
			m_Events.push_back(MapEvent("activate", "none", false, menuitem, true));

			index++;
		}

	// Insert choices for creating new objects ...
	typedef std::vector<k3dIObjectInfo*> ObjectInfoCollection;
	ObjectInfoCollection objectinfocollection;
	k3d::copy(k3dApplication()->Objects(), std::back_inserter(objectinfocollection));

	for(ObjectInfoCollection::iterator objectinfo = objectinfocollection.begin(); objectinfo != objectinfocollection.end(); objectinfo++)
		{
			// Filter choices ...
			if(m_Filter)
				if(!m_Filter->Allow(*objectinfo))
					continue;

			// Create a "new" menu item ...
			sdpGtkMenuItem menuitem;
			menuitem.Create("--New " + sdpString((*objectinfo)->Name()) + "--");
			menuitem.Show();
			menuitem.SetData(data_object, 0);
			menuitem.SetData(data_objectinfo, (*objectinfo));
			menu.Append(menuitem);

			m_MenuItems.push_back(menuitem);

			m_Events.push_back(MapEvent("activate", "new-" + sdpToString(*objectinfo), false, menuitem, true));

			index++;
		}

	// Insert choices for selecting existing objects ...
	for(K3DOBJECTHANDLE handle = m_ObjectContainer->Document()->Objects()->FirstObject(); handle; handle = m_ObjectContainer->Document()->Objects()->NextObject(handle))
		{
			k3dIObject* const object = m_ObjectContainer->Document()->Objects()->Object(handle);

			// Filter available choices ...
			if(m_Filter)
				if(!m_Filter->Allow(object))
					continue;

			// Create a menu item for the object ...
			sdpGtkMenuItem menuitem;
			menuitem.Create(object->Name());
			menuitem.Show();
			menuitem.SetData(data_object, object);
			menuitem.SetData(data_objectinfo, 0);
			menu.Append(menuitem);

			m_MenuItems.push_back(menuitem);

			m_Events.push_back(MapEvent("activate", "select-" + sdpToString(object), false, menuitem, true));

			if(m_ObjectContainer->Object())
				if(object->ObjectID() == m_ObjectContainer->Object()->ObjectID())
					selecteditem = index;

			index++;
		}

	// Add the menu to our optionmenu ...
	OptionMenu(control_optionmenu).SetMenu(menu);

	// Select the current item ...
	OptionMenu(control_optionmenu).SetHistory(selecteditem);
}

// Custom event handling ...
void k3dObjectSelector::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "destroy")
		OnDestroy();
	else if(Event->Name() == "none")
		OnNone();
	else if(Event->Name() == control_editobject)
		OnEdit();
	else if(Event->Name().find("select-") != sdpString::npos)
		OnSelect(Event->Name());
	else if(Event->Name().find("new-") != sdpString::npos)
		OnNew(Event->Name());
	else
		k3dObjectSelectorBase::OnEvent(Event);
}

std::istream& operator>>(std::istream& Stream, k3dIObject*& RHS)
{
	gulong temp = 0;
	Stream >> std::hex >> temp;
	RHS = reinterpret_cast<k3dIObject*>(temp);

	return Stream;
}

std::istream& operator>>(std::istream& Stream, k3dIObjectInfo*& RHS)
{
	gulong temp = 0;
	Stream >> std::hex >> temp;
	RHS = reinterpret_cast<k3dIObjectInfo*>(temp);

	return Stream;
}

// Called when we're going away ...
void k3dObjectSelector::OnDestroy()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clear out our storage ...
	Clear();
}

// Called when the user chooses "new" ...
void k3dObjectSelector::OnNew(sdpString EventName)
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);

	// Create a new object ...
	const sdpString objectinfotext = EventName.right(EventName.size() - 4);
	k3dIObjectInfo* objectinfo = reinterpret_cast<k3dIObjectInfo*>(strtol(objectinfotext, 0, 16));

	k3dRecordStateChangeSet undo(m_ObjectContainer->Document(), "Create New " + sdpString(objectinfo->Name()));
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_newobject, objectinfo->Name());

	k3d::PushEditing(m_ObjectContainer->Document()->Objects());
	k3dIObject* object = m_ObjectContainer->Document()->Objects()->CreateObject(objectinfo->ClassID());
	g_return_if_fail(object);

	SetObject(object);
	object->EditObject();
	k3d::PopEditing(m_ObjectContainer->Document()->Objects());
}

// Called when the user chooses "none" ...
void k3dObjectSelector::OnNone()
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);

	if(m_StateRecorder)
		k3dStartStateChangeSet(m_ObjectContainer->Document());

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_selectnone, 0);

	if(m_StateRecorder)
		k3dFinishStateChangeSet(m_ObjectContainer->Document(), "Select None");

	k3d::PushEditing(m_ObjectContainer->Document()->Objects());
	SetObject(0);
	k3d::PopEditing(m_ObjectContainer->Document()->Objects());
}

// Called when the user clicks the "edit" button ...
void k3dObjectSelector::OnEdit()
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editobject, 0);

	k3d::PushEditing(m_ObjectContainer->Document()->Objects());
	if(m_ObjectContainer->Object())
		m_ObjectContainer->Object()->EditObject();
	k3d::PopEditing(m_ObjectContainer->Document()->Objects());
}

// Called when the user clicks on our list of existing objects ...
void k3dObjectSelector::OnSelect(sdpString EventName)
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);

	const sdpString objecttext = EventName.right(EventName.size() - 7);
	k3dIObject* const object = reinterpret_cast<k3dIObject*>(strtol(objecttext, 0, 16));

	if(m_StateRecorder)
		k3dStartStateChangeSet(m_ObjectContainer->Document());

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_selectobject, object->Name());

	k3d::PushEditing(m_ObjectContainer->Document()->Objects());
	SetObject(object);
	k3d::PopEditing(m_ObjectContainer->Document()->Objects());

	if(m_StateRecorder)
		k3dFinishStateChangeSet(m_ObjectContainer->Document(), "Select " + sdpString(object->Name()));

	k3dRedrawAll(m_ObjectContainer->Document(), false);
}

void k3dObjectSelector::SetObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_return_if_fail(m_ObjectContainer);

	// Set the new object ...
	m_ObjectContainer->SetObject(Object);

	// Notify any interested parties that things changed ...
	if(m_ChangeAction)
		m_ChangeAction->DoAction();
}



