#ifndef K3DOBJECTSELECTOR_H
#define K3DOBJECTSELECTOR_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dObjectSelector class, which provides a user interface for either creating new, or selecting existing, K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcontrol.h>
#include <k3didocument.h>
#include <k3diobject.h>

// Forward class references ...
class k3dEventConnector;
class k3dObjectVariable;

///////////////////////////////////////////////////////////////////////////
// k3dObjectSelectorFilter

/// Called by k3dObjectSelector to filter available user choices
class k3dObjectSelectorFilter
{
public:
	virtual ~k3dObjectSelectorFilter() {}
	
	virtual bool Allow(k3dIObjectInfo* const ObjectInfo) = 0;
	virtual bool Allow(k3dIObject* const Object) = 0;
};

///////////////////////////////////////////////////////////////////////////
// k3dObjectSelectorInterfaceFilter

/// An implementation of k3dObjectSelectorFilter that only displays objects that implement the given interface
class k3dObjectSelectorInterfaceFilter :
	public k3dObjectSelectorFilter
{
public:
	k3dObjectSelectorInterfaceFilter(sdp::ITypeInfo* const InterfaceType);
	~k3dObjectSelectorInterfaceFilter();

	bool Allow(k3dIObjectInfo* const ObjectInfo);
	bool Allow(k3dIObject* const Object);

private:
	sdp::ITypeInfo* const m_InterfaceType;
};

///////////////////////////////////////////////////////////////////////////
// k3dObjectSelector

typedef k3dControl k3dObjectSelectorBase;

/// Provides a user interface for selecting new & existing objects
class k3dObjectSelector :
	public k3dObjectSelectorBase
{
public:
	k3dObjectSelector(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName);
	~k3dObjectSelector();

	/// Called by the framework when being instantiated from a GTKML (XML) document
	bool Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element);
	
	/// Attaches the control to the object variable it fronts for
	bool Attach(k3dObjectVariable* const ObjectContainer,
		k3dICommandRecorder* const CommandRecorder,
		k3dIStateRecorder* const StateRecorder,
		SDPCSTRING StateChangeName,
		k3dIAction* const PushEditingAction,
		k3dIAction* const ChangeAction,
		k3dIAction* const PopEditingAction,
		const bool AllowNone,
		k3dObjectSelectorFilter* const Filter);

	const sdpString CustomType() const;

	// k3dICommandNode overrides
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

protected:
	/// Called to build the list of available objects and display the current selection
	void UpdateSelection();
	
	/// Custom event-handling ...
	void OnEvent(sdpGtkEvent* Event);

	/// Called when the UI is about to go away
	void OnDestroy();
	/// Called when the user wants to create a new object
	void OnNew(sdpString EventName);
	/// Called when the user decides to select no object
	void OnNone();
	/// Called when the user wants to edit an existing object
	void OnEdit();
	/// Called when the user selects an existing object
	void OnSelect(sdpString EventName);
	/// Called to set the selected objects
	void SetObject(k3dIObject* const Object);

	/// Stores the underlying object container
	k3dObjectVariable* m_ObjectContainer;
	/// Stores whether the user is allowed to choose no object
	bool m_AllowNone;
	/// Stores an object for filtering available user choices
	k3dObjectSelectorFilter* m_Filter;
	/// Manages K-3D events
	k3dEventConnector* m_ObjectContainerEvents;
	k3dEventConnector* m_DocumentEvents;
	/// Stores the menu of available options separately from the rest of the interface (since it can change a lot)
	sdpGtkObjectContainer m_Menu;
	
	/// Declares a list of menu items
	typedef std::list<sdpGtkMenuItem> MenuItems;
	typedef MenuItems::iterator MenuItemIterator;
	/// Stores menu items for subsequent cleanup
	MenuItems m_MenuItems;
	
	/// Declares a list of SDPGTK events
	typedef std::vector<sdpGtkEvent*> Events;
	typedef Events::iterator EventIterator;
	/// Stores SDPGTK events for subsequent cleanup
	Events m_Events;

	static const sdpString control_optionmenu;
	static const sdpString control_menu;
	static const sdpString control_selectnone;
	static const sdpString control_selectobject;
	static const sdpString control_newobject;
	static const sdpString control_editobject;
	
	static const sdpString data_object;
	static const sdpString data_objectinfo;
};

#endif // K3DOBJECTSELECTOR_H

