// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPath class, which encapsulates a geometric path
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dmesh.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dpath.h>

#include <k3distaterecorder.h>
#include <k3didocument.h>
#include <k3dimaterial.h>

/////////////////////////////////////////////////////////////////////////
// k3dPath

k3dPath::k3dPath(k3dMesh* const Mesh, const ID PathID) :
	k3dState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::USERVIEWABLE | k3dIState::SAVEABLE),
	m_Mesh(Mesh),
	m_PathID(PathID),
	m_Material(0)
{
}

k3dPath::~k3dPath()
{
	std::for_each(m_PathPoints.begin(), m_PathPoints.end(), sdp::DeleteObject());
}

void k3dPath::PushEditing()
{
	if(!Editing())
		{
			m_PathPointCollectionChanged = false;
			m_MaterialChanged = false;
		}

	k3dEditable::PushEditing();

}

void k3dPath::PopEditing()
{
	k3dEditable::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_PathPointCollectionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<PathPoints>(m_PathPoints));

					if(m_MaterialChanged)
						changeset->RecordNewState(new k3dValueStateContainer<k3dIMaterial*>(m_Material));
				}
		}

}

sdpResult k3dPath::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	k3dIObject::ID materialid = 0;
	if(m_Material)
		materialid = dynamic_cast<k3dIObject*>(m_Material)->ObjectID();

	sdpxml::Element& path = Element.Append(sdpxml::Element("path", "",
		sdpxml::Attribute("material", sdpToString(materialid)),
		sdpxml::Attribute("id", sdpToString(m_PathID))));

	// Write properties ...
	k3dState::Save(path, Dependencies);

	// Write the pathpoints ...
	if(m_PathPoints.size())
		{
			sdpxml::Element& pathpoints = path.Append(sdpxml::Element("pathpoints"));

			for(PathPointIterator pathpoint = m_PathPoints.begin(); pathpoint != m_PathPoints.end(); pathpoint++)
				(*pathpoint)->Save(pathpoints, Dependencies);
		}

	return SDP_S_OK;
}

class k3dPathLoadPathPoint
{
public:
	k3dPathLoadPathPoint(k3dPath* const Path) :
		m_Path(Path)
	{
		// Sanity checks ...
		g_assert(m_Path);
	}

	void operator()(sdpxml::Element& Element)
	{
		// Create a new pathpoint and push it into the pathpoint list ...
		k3dPathPoint* const pathpoint = new k3dPathPoint(m_Path);
		m_Path->m_PathPoints.push_back(pathpoint);

		// Load the pathpoint and attach it to its point ...
		pathpoint->Load(Element);
	}

private:
	k3dPath* const m_Path;
};

sdpResult k3dPath::Load(sdpxml::Element& Element, k3dMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	// Read the ID (required) ...
	g_return_val_if_fail(sdpxml::ParseAttribute(Element, "id", m_PathID), SDP_E_FAIL);

	// Read properties ...
	k3dState::Load(Element);

	// Read path material ...
	const k3dIObject::ID materialid = sdpxml::GetAttribute<k3dIObject::ID>(Element, "material", 0);

	// Store the material ID for binding later ...
	m_Material = reinterpret_cast<k3dIMaterial*>(materialid);

	// Load pathpoints
	sdpxml::ElementPointer pathpoints = sdpxml::FindElement(Element, sdpxml::SameName("pathpoints"));
	if(pathpoints)
		{
			k3dPathLoadPathPoint tmfunctor(this);
			sdpxml::Element Parent = *pathpoints;

			// The functor seems not to work (both under Win32 and Linux, gcc upgrade?)...
			//sdpxml::ForEach(*pathpoints, sdpxml::SameName("pathpoint"), functor);
			// Have to rewrite the function below:
			sdpxml::SameName samen("pathpoint");
			for(sdpxml::ElementCollection::iterator child = Parent.Children().begin(); child != Parent.Children().end(); child++)
				{
					// Create a new pathpoint and push it into the pathpoint list ...
					k3dPathPoint* const pathpoint = new k3dPathPoint(this);
					m_PathPoints.push_back(pathpoint);

					// Load the pathpoint and attach it to its point ...
					pathpoint->Load(*child);
				}
		}

	return SDP_S_OK;
}

sdpResult k3dPath::LoadComplete()
{
	// Bind our material ID to a material ...
	const k3dIObject::ID materialid = reinterpret_cast<k3dIObject::ID>(m_Material);
	m_Material = 0;

	if(materialid)
		{
			std::vector<k3dIObject*> materialobjects;
			k3d::copy_if(m_Mesh->Document()->Objects(), std::back_inserter(materialobjects), k3d::SameID(materialid));

			if(1 == materialobjects.size())
				m_Material = dynamic_cast<k3dIMaterial*>(*materialobjects.begin());
		}

	// Give pathpoints a chance to bind to points ...
	for(PathPointIterator pathpoint = m_PathPoints.begin(); pathpoint != m_PathPoints.end(); pathpoint++)
		(*pathpoint)->LoadComplete(m_Mesh);

	return SDP_S_OK;
}

//////////////////////////////////////////
// k3dIPath implementation

k3dIMesh* k3dPath::Mesh()
{
	return m_Mesh;
}

const k3dIPath::ID k3dPath::PathID()
{
	return m_PathID;
}

k3dIPathPointCollection* k3dPath::Points()
{
	return this;
}

k3dILocationCollection* k3dPath::Locations()
{
	return this;
}

k3dIMaterialCollection* k3dPath::Materials()
{
	return this;
}

k3dIState* k3dPath::State()
{
	return this;
}

sdpResult k3dPath::PathNormal(double Normal[3])
{
	m_Normal.CopyArray(Normal);
	return SDP_S_OK;
}

sdpResult k3dPath::SetPathNormal(double Normal[3])
{
	m_Normal = Normal;
	return SDP_S_OK;
}

sdpResult k3dPath::FlipOrientation()
{
	if(0 == m_PathPoints.size())
		return SDP_S_OK;

	std::rotate(m_PathPoints.begin(), m_PathPoints.begin()+1, m_PathPoints.end());
	std::reverse(m_PathPoints.begin(), m_PathPoints.end());
	return SDP_S_OK;
}

sdpResult k3dPath::AdvanceFirstPoint()
{
	if(0 == m_PathPoints.size())
		return SDP_S_OK;

	std::rotate(m_PathPoints.begin(), m_PathPoints.begin()+1, m_PathPoints.end());
	return SDP_S_OK;
}

/////////////////////////////////////////
// k3dIPathPointCollection implementation

void k3dPath::VisitPathPoints(k3dIPathPointVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	std::for_each(m_PathPoints.begin(), m_PathPoints.end(), sdp::MakeVisitorAdapter(Visitor, &k3dIPathPointVisitor::VisitPathPoint));
}

k3dIPathPoint* k3dPath::AddPoint(k3dIPoint* const Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(Editing(), 0);

	k3dPathPoint* pathpoint = new k3dPathPoint(this, static_cast<k3dPoint*>(Point), m_Mesh->NextPathPointID());
	g_return_val_if_fail(pathpoint, 0);
	m_PathPoints.push_back(pathpoint);

	return pathpoint;
}

sdpResult k3dPath::RemovePoint(k3dIPoint* const Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	for(PathPointIterator pathpoint = m_PathPoints.begin(); pathpoint != m_PathPoints.end(); pathpoint++)
		{
			if((*pathpoint)->Point() == Point)
				{
					k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
					if(changeset)
						{
							changeset->RecordOldState(new k3dDeleteInstanceStateContainer<k3dPathPoint>(*pathpoint, changeset));

							if(!m_PathPointCollectionChanged)
								{
									m_PathPointCollectionChanged = true;
									changeset->RecordOldState(new k3dValueStateContainer<PathPoints>(m_PathPoints));
								}
						}

					m_PathPoints.erase(pathpoint);
					return SDP_S_OK;
				}
		}

	return SDP_S_FALSE;
}

sdpResult k3dPath::RemovePathPoint(k3dIPathPoint* const PathPoint)
{
	// Sanity check ...
	g_return_val_if_fail(PathPoint, SDP_E_FAIL);
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	k3dPathPoint* pathpoint = static_cast<k3dPathPoint*>(PathPoint);
	PathPointIterator iterator = std::find(m_PathPoints.begin(), m_PathPoints.end(), pathpoint);
	g_return_val_if_fail(iterator != m_PathPoints.end(), SDP_E_FAIL);

	m_PathPoints.erase(iterator);
	delete pathpoint;

	return SDP_S_OK;
}
gulong k3dPath::PointCount()
{
	return m_PathPoints.size();
}

K3DPATHPOINTHANDLE k3dPath::FirstPathPoint()
{
	return K3DPATHPOINTHANDLE(m_PathPoints.size() ? 1 : 0);
}

K3DPATHPOINTHANDLE k3dPath::NextPathPoint(K3DPATHPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_PathPoints.size(), 0);

	return K3DPATHPOINTHANDLE(gulong(Point) < m_PathPoints.size() ? gulong(Point)+1 : 0);
}

k3dIPathPoint* k3dPath::PathPoint(K3DPATHPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_PathPoints.size(), 0);

	return m_PathPoints[gulong(Point)-1];
}

/////////////////////////////////////////
// k3dILocationCollection implementation

void k3dPath::VisitLocations(k3dILocationVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	std::for_each(m_PathPoints.begin(), m_PathPoints.end(), sdp::MakeVisitorAdapter(Visitor, &k3dILocationVisitor::VisitLocation));
}

gulong k3dPath::LocationCount()
{
	return m_PathPoints.size();
}

K3DLOCATIONHANDLE k3dPath::FirstLocation()
{
	return K3DLOCATIONHANDLE(m_PathPoints.size() ? 1 : 0);
}

K3DLOCATIONHANDLE k3dPath::NextLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_PathPoints.size(), 0);

	return K3DLOCATIONHANDLE(gulong(Location) < m_PathPoints.size() ? gulong(Location)+1 : 0);
}

k3dILocation* k3dPath::Location(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_PathPoints.size(), 0);

	return m_PathPoints[gulong(Location)-1]->m_Point;
}

/////////////////////////////////////////
// k3dIMaterialCollection implementation

k3dIMaterial* k3dPath::Material()
{
	return m_Material;
}

sdpResult k3dPath::SetMaterial(k3dIMaterial* const Material)
{
	// Sanity checks ...
	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	if(!m_MaterialChanged)
		{
			m_MaterialChanged = true;

			k3dIStateChangeSet* const changeset = m_Mesh->Document()->StateRecorder()->ChangeSet();
			if(changeset)
				changeset->RecordOldState(new k3dValueStateContainer<k3dIMaterial*>(m_Material));
		}

	m_Material = Material;

	return SDP_S_OK;
}

/////////////////////////////////////////
// k3dIState implementation

sdpResult k3dPath::SetState(gulong Mask, bool State)
{
	// Perform an undo-able state change ...
	RawSetState(Mask, State, m_Mesh->Document()->StateRecorder());
	return SDP_S_OK;
}

