// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPathPoint class, which encapsulates the many-to-many relationship between k3dIPoint and k3dIPath/k3dICurve
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpathpoint.h>
#include <k3dpoint.h>
#include <k3dmesh.h>
#include <k3dmisc.h>
#include <k3dstaterecorder.h>

#include <k3didocument.h>

/////////////////////////////////////////////////////////////////////////
// k3dPathPoint

k3dPathPoint::k3dPathPoint() :
	m_Path(0),
	m_Point(0),
	m_PathPointID(0)
{
}

k3dPathPoint::k3dPathPoint(k3dPath* const Path) :
	m_Path(Path),
	m_Point(0),
	m_PathPointID(0)
{
	// Sanity checks ...
	g_assert(m_Path);
}

k3dPathPoint::k3dPathPoint(k3dPoint* const Point, const ID PathPointID) :
	m_Path(0),
	m_Point(Point),
	m_PathPointID(PathPointID)
{
	// Sanity checks ...
	g_assert(m_Point);
	g_assert(m_PathPointID);
}

k3dPathPoint::k3dPathPoint(k3dPath* const Path, k3dPoint* const Point, const ID PathPointID) :
	m_Path(Path),
	m_Point(Point),
	m_PathPointID(PathPointID)
{
	// Sanity checks ...
	g_assert(m_Path);
	g_assert(m_Point);
	g_assert(m_PathPointID);
}

k3dPathPoint::~k3dPathPoint()
{
}

///////////////////////////////////////////////////////
// k3dIPathPoint implementation

k3dIPoint* k3dPathPoint::Point()
{
	return m_Point;
}

k3dIState* k3dPathPoint::State()
{
	return m_Point->State();
}

k3dILocation* k3dPathPoint::Location()
{
	return this;
}

const k3dIPathPoint::ID k3dPathPoint::PathPointID()
{
	return m_PathPointID;
}

///////////////////////////////////////////////
// k3dILocation implementation

/*
k3dIState* k3dPathPoint::State()
{
	return m_Point->State();
}
*/

const sdpVector3& k3dPathPoint::LocalXYZ()
{
	return m_Point->LocalXYZ();
}

const sdpVector3& k3dPathPoint::AnimXYZ()
{
	return m_Point->AnimXYZ();
}

const sdpVector3& k3dPathPoint::WorldXYZ()
{
	return m_Point->WorldXYZ();
}

const sdpVector3& k3dPathPoint::CameraXYZ()
{
	return m_Point->CameraXYZ();
}

const sdpVector3& k3dPathPoint::ScreenXYZ()
{
	return m_Point->ScreenXYZ();
}

const sdpVector3& k3dPathPoint::NormalXYZ()
{
	return m_NormalXYZ;
}

const sdpVector3& k3dPathPoint::TextureXYZ()
{
	return m_TextureXYZ;
}

const sdpVector3& k3dPathPoint::ImplicitTextureXYZ()
{
	return m_ImplicitTextureXYZ;
}

void k3dPathPoint::SetLocalXYZ(const sdpVector3& Coords)
{
	m_Point->SetLocalXYZ(Coords);
}

void k3dPathPoint::SetAnimXYZ(const sdpVector3& Coords)
{
	m_Point->SetAnimXYZ(Coords);
}

void k3dPathPoint::SetWorldXYZ(const sdpVector3& Coords)
{
	m_Point->SetWorldXYZ(Coords);
}

void k3dPathPoint::SetCameraXYZ(const sdpVector3& Coords)
{
	m_Point->SetCameraXYZ(Coords);
}

void k3dPathPoint::SetScreenXYZ(const sdpVector3& Coords)
{
	m_Point->SetScreenXYZ(Coords);
}

void k3dPathPoint::SetNormalXYZ(const sdpVector3& Coords)
{
	m_NormalXYZ = Coords;
}

void k3dPathPoint::SetTextureXYZ(const sdpVector3& Coords)
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	k3dIStateChangeSet* const changeset = m_Point->Mesh()->Document()->StateRecorder()->ChangeSet();
	if(changeset && !m_TextureXYZChanged)
		{
			m_TextureXYZChanged = true;
			changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_TextureXYZ));
		}

	m_TextureXYZ = Coords;
}

void k3dPathPoint::SetImplicitTextureXYZ(const sdpVector3& Coords)
{
	m_ImplicitTextureXYZ = Coords;
}

/////////////////////////////////////////////////////////////////////////////////////////
// k3dIEditable implemenation

void k3dPathPoint::PushEditing()
{
	if(!Editing())
		m_TextureXYZChanged = false;

	k3dEditable::PushEditing();
}

void k3dPathPoint::PopEditing()
{
	k3dEditable::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = m_Point->Mesh()->Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_TextureXYZChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_TextureXYZ));
				}
		}
}

sdpResult k3dPathPoint::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	Element.Append(sdpxml::Element("pathpoint", "",
		sdpxml::Attribute("id", sdpToString(m_PathPointID)),
		sdpxml::Attribute("pointid", sdpToString(m_Point->PointID())),
		sdpxml::Attribute("texture", sdpToString(m_TextureXYZ)),
		sdpxml::Attribute("itexture", sdpToString(m_ImplicitTextureXYZ))));

	return SDP_S_OK;
}

sdpResult k3dPathPoint::Load(sdpxml::Element& Element)
{
	// Load the pathpoint ID (required) ...
	m_PathPointID = sdpxml::GetAttribute<k3dIPathPoint::ID>(Element, "id", 0);
	g_return_val_if_fail(m_PathPointID, SDP_E_FAIL);

	// Grab the point handle (required) and store it for binding later ...
	const k3dIPoint::ID pointid = sdpxml::GetAttribute<k3dIPoint::ID>(Element, "pointid", 0);
	g_return_val_if_fail(pointid, SDP_E_FAIL);
	m_Point = reinterpret_cast<k3dPoint*>(pointid);

	// Load texture coordinates ...
	sdpxml::ParseAttribute(Element, "texture", m_TextureXYZ);
	sdpxml::ParseAttribute(Element, "itexture", m_ImplicitTextureXYZ);

	return SDP_S_OK;
}

sdpResult k3dPathPoint::LoadComplete(k3dMesh* Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	// Find the point that matches our pointid ...
	m_Point = Mesh->FindPoint(reinterpret_cast<k3dIPoint::ID>(m_Point));
	g_return_val_if_fail(m_Point, SDP_E_FAIL);

	return SDP_S_OK;
}


