#ifndef K3DPATHPOINTCOLLECTION_H
#define K3DPATHPOINTCOLLECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcollections.h>
#include <k3dipathpointcollection.h>

namespace k3d
{

/// Iterator adapter for pulling k3dIPoint data from a k3dIPathPoint
template<typename InsertIterator>
class PointAdapterIterator/* :
	public std::iterator<std::output_iterator_tag, void, void, void, void>*/
{
public:
	explicit PointAdapterIterator(InsertIterator Iterator) :
		m_Iterator(Iterator)
	{
	}
	
	PointAdapterIterator& operator=(k3dIPathPoint* const PathPoint)
	{
		m_Iterator = PathPoint->Point();
		return *this;
	}
	
	PointAdapterIterator& operator*()
	{
		return *this;
	}
	
	PointAdapterIterator& operator++()
	{
		m_Iterator++;
		return *this;
	}
	
	PointAdapterIterator operator++(int)
	{
		++m_Iterator;
		return *this;
	}
	
private:
	InsertIterator m_Iterator;
};

/// Convenience function for instantiating PointAdapterIterator objects
template<typename InsertIterator>
PointAdapterIterator<InsertIterator> PointAdapter(InsertIterator Iterator)
{
	return PointAdapterIterator<InsertIterator>(Iterator);
}

/// Applies a functor to every k3dIPathPoint in a collection
template<typename Functor>
Functor for_each(k3dIPathPointCollection* const PathPointCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(PathPointCollection, F);

	// For each object in the document ...
	for(K3DPATHPOINTHANDLE handle = PathPointCollection->FirstPathPoint(); handle; handle = PathPointCollection->NextPathPoint(handle))
		F(PathPointCollection->PathPoint(handle));

	return F;
}
	
/// Applies a functor to every k3dIPathPoint in a collection that satisfies the predicate
template<typename Functor, typename Predicate>
Functor for_each_if(k3dIPathPointCollection* const PathPointCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(PathPointCollection, F);

	// For each object in the document ...
	for(K3DPATHPOINTHANDLE handle = PathPointCollection->FirstPathPoint(); handle; handle = PathPointCollection->NextPathPoint(handle))
		{
			k3dIPathPoint* const object = PathPointCollection->PathPoint(handle);
		
			if(P(object))
				F(object);
		}

	return F;
}

/// Copies k3dIPathPoint references from a k3dIPathPointCollection into a generic container
template<typename InsertIterator>
void copy(k3dIPathPointCollection* const PathPointCollection, InsertIterator D)
{
	// Sanity checks ...
	g_return_if_fail(PathPointCollection);

	// For each object in the document ...
	for(K3DPATHPOINTHANDLE handle = PathPointCollection->FirstPathPoint(); handle; handle = PathPointCollection->NextPathPoint(handle))
		*D++ = PathPointCollection->PathPoint(handle);
}

/// Copies k3dIPathPoint references that satisfy a predicate from a k3dIPathPointCollection into a generic container
template<typename InsertIterator, typename Predicate>
void copy_if(k3dIPathPointCollection* const PathPointCollection, InsertIterator D, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(PathPointCollection);

	// For each object in the document ...
	for(K3DPATHPOINTHANDLE handle = PathPointCollection->FirstPathPoint(); handle; handle = PathPointCollection->NextPathPoint(handle))
		{
			k3dIPathPoint* const object = PathPointCollection->PathPoint(handle);
			if(P(object))
				*D++ = object;
		}
}

} // namespace k3d

#endif // K3DPATHPOINTCOLLECTION_H
