#ifndef K3DPOINTCOLLECTION_H
#define K3DPOINTCOLLECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares several convenience templates for iterating over collections
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcollections.h>
#include <k3dipointcollection.h>

namespace k3d
{

/// Applies a functor to every k3dIPoint in a collection
template<typename Functor>
Functor for_each(k3dIPointCollection* const PointCollection, Functor F)
{
	// Sanity checks ...
	g_return_val_if_fail(PointCollection, F);

	// For each object in the document ...
	for(K3DPOINTHANDLE handle = PointCollection->FirstPoint(); handle; handle = PointCollection->NextPoint(handle))
		F(PointCollection->Point(handle));

	return F;
}
	
/// Applies a functor to every k3dIPoint in a collection that satisfies the predicate
template<typename Functor, typename Predicate>
Functor for_each_if(k3dIPointCollection* const PointCollection, Functor F, Predicate P)
{
	// Sanity checks ...
	g_return_val_if_fail(PointCollection, F);

	// For each object in the document ...
	for(K3DPOINTHANDLE handle = PointCollection->FirstPoint(); handle; handle = PointCollection->NextPoint(handle))
		{
			k3dIPoint* const object = PointCollection->Point(handle);
		
			if(P(object))
				F(object);
		}

	return F;
}

/// Copies k3dIPoint references from a k3dIPointCollection into a generic container
template<typename InsertIterator>
void copy(k3dIPointCollection* const PointCollection, InsertIterator D)
{
	// Sanity checks ...
	g_return_if_fail(PointCollection);

	// For each object in the document ...
	for(K3DPOINTHANDLE handle = PointCollection->FirstPoint(); handle; handle = PointCollection->NextPoint(handle))
		*D++ = PointCollection->Point(handle);
}

/// Copies k3dIPoint references that satisfy a predicate from a k3dIPointCollection into a generic container
template<typename InsertIterator, typename Predicate>
void copy_if(k3dIPointCollection* const PointCollection, InsertIterator D, Predicate P)
{
	// Sanity checks ...
	g_return_if_fail(PointCollection);

	// For each object in the document ...
	for(K3DPOINTHANDLE handle = PointCollection->FirstPoint(); handle; handle = PointCollection->NextPoint(handle))
		{
			k3dIPoint* const object = PointCollection->Point(handle);
			if(P(object))
				*D++ = object;
		}
}

} // namespace k3d

#endif // K3DPOINTCOLLECTION_H
