// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dPositionController class, which provides a UI for manually placing objects that support k3dITransform
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dpositioncontroller.h>
#include <k3dicommandrecorder.h>
#include <k3distaterecorder.h>

#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

//////////////////////////////////////////////////////////////////////
// k3dPositionController

const sdpString k3dPositionController::control_x = "x";
const sdpString k3dPositionController::control_y = "y";
const sdpString k3dPositionController::control_z = "z";
const sdpString k3dPositionController::control_reset = "reset";

k3dPositionController::k3dPositionController(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dPositionControllerBase(CommandNodeParent, CommandNodeName),
	m_DataSource(0),
	m_X(0),
	m_Y(0),
	m_Z(0)
{
	m_X = new k3dSpinButton(this, control_x);
	m_Y = new k3dSpinButton(this, control_y);
	m_Z = new k3dSpinButton(this, control_z);

	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<vbox homogeneous=\"false\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"

				"<eventbox name=\"x_container\"/>"
				"<eventbox name=\"y_container\"/>"
				"<eventbox name=\"z_container\"/>"
				
				"<button name=\"reset\">Reset Position"
					"<event signal=\"clicked\" name=\"reset\"/>"
				"</button>"

			"</vbox>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, ""));

	// Attach the spin buttons ...
	Container("x_container").Attach(m_X->RootWidget());
	Container("y_container").Attach(m_Y->RootWidget());
	Container("z_container").Attach(m_Z->RootWidget());
	
	AddCommand(control_reset, control_reset, CONTROL_BUTTON);
	
	RootWidget().Show();
}

k3dPositionController::~k3dPositionController()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();
		
	delete m_X;
	delete m_Y;
	delete m_Z;
}

const sdpString k3dPositionController::CustomType() const
{
	return sdpString("k3dPositionController");
}

bool k3dPositionController::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	
	g_assert(Element.Name() == "k3dpositioncontroller");

	return true;
}

class k3dPositionControllerDataSource :
	public k3dSpinButtonDataSource,
	public k3dEventSource,
	public k3dEventObserver
{
public:
	k3dPositionControllerDataSource(k3dIDocument* const Document, k3dITransform* const DataSource, const k3dITransform::ReferenceFrame Frame, const unsigned int Index) :
		m_Document(Document),
		m_DataSource(DataSource),
		m_Frame(Frame),
		m_Index(Index)
	{
		// Sanity checks ...
		g_assert(m_Document);
		g_assert(m_DataSource);
		g_assert(m_Index < 3);
		
		k3dIEventSource* const eventserver = dynamic_cast<k3dIEventSource*>(m_DataSource);
		g_assert(eventserver);
		eventserver->AddEventObserver(this, k3dITransform::Event_PositionChanged, 0);
	}
	
	virtual ~k3dPositionControllerDataSource()
	{
		k3dIEventSource* const eventserver = dynamic_cast<k3dIEventSource*>(m_DataSource);
		g_assert(eventserver);
		eventserver->RemoveEventObserver(this);
	}

	k3dIEventSource* EventSource()
	{
//		return dynamic_cast<k3dIEventSource*>(m_DataSource);
		return this;
	}
	
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
	{
		// Sanity checks ...
		g_return_val_if_fail(Source, SDP_E_FAIL);
		g_return_val_if_string_fail(Event, SDP_E_FAIL);
		
		const sdpString event(Event);
		if(event == k3dITransform::Event_PositionChanged)
			{
				Notify(k3dIVariable::Event_Changed, 0);
				return SDP_S_OK;
			}
		
		return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
	}	
	
	gdouble Value()
	{
		return m_DataSource->Position(m_Frame)[m_Index];
	}

	void PushEditing()
	{
		k3d::PushEditing(m_DataSource);
	}
			
	void SetValue(const gdouble Value)
	{
		sdpVector3 coords = m_DataSource->Position(m_Frame);
		coords[m_Index] = Value;
		m_DataSource->SetPosition(m_Frame, coords);
	}

	void PopEditing()
	{
		k3d::PopEditing(m_DataSource);
	}
			
private:
	k3dIDocument* const m_Document;
	k3dITransform* const m_DataSource;
	const k3dITransform::ReferenceFrame m_Frame;
	const unsigned int m_Index;
};

bool k3dPositionController::Attach(k3dIDocument* const Document, k3dITransform* const DataSource, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	// Sanity checks ...
	g_return_val_if_fail(Document, false);
	g_return_val_if_fail(DataSource, false);
	g_return_val_if_fail(k3dPositionControllerBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);
	
	// Store references ...
	m_Document = Document;
	m_DataSource = DataSource;

	g_return_val_if_fail(m_X->Create(k3dIChannel::DATA_DISTANCE, 3, 0.1), false);
	g_return_val_if_fail(m_Y->Create(k3dIChannel::DATA_DISTANCE, 3, 0.1), false);
	g_return_val_if_fail(m_Z->Create(k3dIChannel::DATA_DISTANCE, 3, 0.1), false);

	g_return_val_if_fail(m_X->Attach(new k3dPositionControllerDataSource(m_Document, m_DataSource, k3dITransform::CURRENT_FRAME, 0), CommandRecorder, StateRecorder, "X Position", 0, new k3dActionRedrawAll(Document, false, 0), 0), false);
	g_return_val_if_fail(m_Y->Attach(new k3dPositionControllerDataSource(m_Document, m_DataSource, k3dITransform::CURRENT_FRAME, 1), CommandRecorder, StateRecorder, "Y Position", 0, new k3dActionRedrawAll(Document, false, 0), 0), false);
	g_return_val_if_fail(m_Z->Attach(new k3dPositionControllerDataSource(m_Document, m_DataSource, k3dITransform::CURRENT_FRAME, 2), CommandRecorder, StateRecorder, "Z Position", 0, new k3dActionRedrawAll(Document, false, 0), 0), false);

	return true;
}

void k3dPositionController::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "destroy")
		OnDestroy();
	else if(Event->Name() == control_reset)
		OnReset();
	else
		k3dPositionControllerBase::OnEvent(Event);
}

void k3dPositionController::OnDestroy()
{
	DisconnectAllEvents();

	Clear();
}

void k3dPositionController::OnReset()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_reset, 0);

	k3dStartStateChangeSet(m_Document);
	k3d::PushEditing(m_DataSource);

	sdpVector3 origin;
	m_DataSource->SetPosition(k3dITransform::CURRENT_FRAME, origin);
	
	k3d::PopEditing(m_DataSource);
	k3dFinishStateChangeSet(m_Document, "Reset Position");
	
	k3dRedrawAll(m_Document, false);
}

