
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dRiEngine class, which provides an implementation of k3dIRiEngine
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3driengine.h>
#include <k3diapplication.h>
#include <numeric>

/// Helper class for handling token formatting
class FormatToken
{
public:
	FormatToken(SDPCSTRING Token) :
		m_Token(Token)
	{
	}
	
	friend std::ostream& operator << (std::ostream& Stream, const FormatToken& Arg)
	{
		Stream << "\"" << Arg.m_Token << "\"";
		return Stream;
	}

protected:
	SDPCSTRING m_Token;
};

/// Helper class for handling string formatting
class FormatString
{
public:
	FormatString(SDPCSTRING String) :
		m_String(String)
	{
	}
	
	friend std::ostream& operator << (std::ostream& Stream, const FormatString& Arg)
	{
		Stream << "\"" << Arg.m_String << "\"";
		return Stream;
	}

protected:
	SDPCSTRING m_String;
};

/// Helper class for handling integer array formatting
class FormatIntegerArray
{
public:
	FormatIntegerArray(gpointer const ArrayValues, const gulong ArrayCount) :
		m_ArrayValues(static_cast<gint*>(ArrayValues)),
		m_ArrayCount(ArrayCount)
	{
	}

	friend std::ostream& operator << (std::ostream& Stream, const FormatIntegerArray& Arg)
	{
		Stream << "[ ";
		
		for(gulong i = 0; i < Arg.m_ArrayCount; i++)
			Stream << Arg.m_ArrayValues[i] << " ";
		
		Stream << "]";
		
		return Stream;
	}

private:
	gint* const m_ArrayValues;
	const gulong m_ArrayCount;
};

/// Helper class for handling gulong array formatting
class FormatUnsignedLongArray
{
public:
	FormatUnsignedLongArray(gpointer const ArrayValues, const gulong ArrayCount) :
		m_ArrayValues(static_cast<gulong*>(ArrayValues)),
		m_ArrayCount(ArrayCount)
	{
	}

	friend std::ostream& operator << (std::ostream& Stream, const FormatUnsignedLongArray& Arg)
	{
		Stream << "[ ";
		
		for(gulong i = 0; i < Arg.m_ArrayCount; i++)
			Stream << Arg.m_ArrayValues[i] << " ";
		
		Stream << "]";
		
		return Stream;
	}

private:
	gulong* const m_ArrayValues;
	const gulong m_ArrayCount;
};

/// Helper class for handling real array formatting
class FormatRealArray
{
public:
	FormatRealArray(gpointer const ArrayValues, const gulong ArrayCount) :
		m_ArrayValues(static_cast<gdouble*>(ArrayValues)),
		m_ArrayCount(ArrayCount)
	{
	}

	friend std::ostream& operator << (std::ostream& Stream, const FormatRealArray& Arg)
	{
		Stream << "[ ";
		
		for(gulong i = 0; i < Arg.m_ArrayCount; i++)
			Stream << Arg.m_ArrayValues[i] << " ";
		
		Stream << "]";
		
		return Stream;
	}

private:
	gdouble* const m_ArrayValues;
	const gulong m_ArrayCount;
};

/// Helper class for handling string array formatting
class FormatStringArray
{
public:
	FormatStringArray(gpointer const ArrayValues, const gulong ArrayCount) :
		m_ArrayValues(const_cast<SDPCSTRING*>(static_cast<SDPSTRING*>(ArrayValues))),
		m_ArrayCount(ArrayCount)
	{
	}

	friend std::ostream& operator << (std::ostream& Stream, const FormatStringArray& Arg)
	{
		Stream << "[ ";
		
		for(gulong i = 0; i < Arg.m_ArrayCount; i++)
			Stream << FormatString(Arg.m_ArrayValues[i]) << " ";
		
		Stream << "]";
		
		return Stream;
	}

private:
	SDPCSTRING* m_ArrayValues;
	const gulong m_ArrayCount;
};

/// Helper class for handling parameter list formatting
class FormatParameters
{
public:
	FormatParameters(const gulong ParameterCount, SDPCSTRING* Tokens, gpointer* Values, gulong* ValueCounts, k3dRiEngine::Declarations& Declarations) :
		m_ParameterCount(ParameterCount),
		m_Tokens(Tokens),
		m_Values(Values),
		m_ValueCounts(ValueCounts),
		m_Declarations(Declarations)
	{
	}

	friend std::ostream& operator << (std::ostream& Stream, const FormatParameters& Arg)
	{
		for(gulong i = 0; i < Arg.m_ParameterCount; i++)
			{
				Stream << FormatToken(Arg.m_Tokens[i]) << " ";
				
				// Lookup the token type ...
				const sdpString type = Arg.m_Declarations[Arg.m_Tokens[i]];
				
				if(type.contains("integer"))
					{
						Stream << FormatIntegerArray(Arg.m_Values[i], Arg.m_ValueCounts[i]) << " ";
					}
				else if(type.contains("float") || type.contains("point") || type.contains("hpoint") || type.contains("vector") || type.contains("normal") || type.contains("color") || type.contains("matrix"))
					{
						Stream << FormatRealArray(Arg.m_Values[i], Arg.m_ValueCounts[i]) << " ";
					}
				else if(type.contains("string"))
					{
						Stream << FormatStringArray(Arg.m_Values[i], Arg.m_ValueCounts[i]) << " ";
					}
				else
					{
						std::cerr << "k3dRiEngine::FormatParameters(): Unknown type [" << Arg.m_Tokens[i] << "] [" << type << "]" << std::endl;
					}
			}
		
		return Stream;
	}
			
protected:
	const gulong m_ParameterCount;
	SDPCSTRING* m_Tokens;
	gpointer* m_Values;
	gulong* m_ValueCounts;
	k3dRiEngine::Declarations& m_Declarations;
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dRiEngine

k3dRiEngine::k3dRiEngine() :
	m_Stream(&std::cout),
	m_ObjectHandle(1),
	m_LightHandle(1)
{
	// Setup our built-in parameters ...
	m_Declarations["P"] = "point";
	m_Declarations["N"] = "point";
	m_Declarations["s"] = "float";
	m_Declarations["t"] = "float";
	m_Declarations["visibility"] = "integer";
	m_Declarations["truedisplacement"] = "integer";
	m_Declarations["casts_shadows"] = "string";
	m_Declarations["emissioncolor"] = "color";
}

k3dRiEngine::~k3dRiEngine()
{
}

sdpResult k3dRiEngine::SetOutputStream(std::ostream* const Stream)
{
	// Sanity checks ...
	g_assert(Stream);
	g_return_val_if_fail(Stream->good(), SDP_E_FAIL);

	m_Stream = Stream;
	
	return SDP_S_OK;
}

sdpResult k3dRiEngine::WriteRIBHeader()
{
	RiStructure("RenderMan RIB-Structure 1.0");
	(*m_Stream) << "version 3.03" << std::endl;
	
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDeclare(SDPCSTRING Name, SDPCSTRING Type)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	g_return_val_if_string_fail(Type, SDP_E_FAIL);

	// Save the declaration for later ...
	m_Declarations[Name] = Type;

	(*m_Stream) << "Declare " << FormatToken(Name) << " " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFrameBegin(const gulong FrameNumber)
{
	(*m_Stream) << "FrameBegin " << FrameNumber << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFrameEnd()
{
	(*m_Stream) << "FrameEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiWorldBegin()
{
	(*m_Stream) << "WorldBegin" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiWorldEnd()
{
	(*m_Stream) << "WorldEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFormat(const gulong XResolution, const gulong YResolution, const gdouble AspectRatio)
{
	(*m_Stream) << "Format " << XResolution << " " << YResolution << " " << AspectRatio << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFrameAspectRatio(gdouble AspectRatio)
{
	(*m_Stream) << "FrameAspectRatio " << AspectRatio << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiScreenWindow(gdouble Left, gdouble Right, gdouble Bottom, gdouble Top)
{
	(*m_Stream) << "ScreenWindow " << Left << " " << Right << " " << Bottom << " " << Top << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCropWindow(gdouble XMin, gdouble XMax, gdouble YMin, gdouble YMax)
{
	(*m_Stream) << "CropWindow " << XMin << " " << XMax << " " << YMin << " " << YMax << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiProjectionV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Projection " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiClipping(gdouble NearPlane, gdouble FarPlane)
{
	(*m_Stream) << "Clipping " << NearPlane << " " << FarPlane << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDepthOfField(gdouble FStop, gdouble FocalLength, gdouble FocalDistance)
{
	(*m_Stream) << "DepthOfField " << FStop << " " << FocalLength << " " << FocalDistance << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiShutter(gdouble OpenTime, gdouble CloseTime)
{
	(*m_Stream) << "Shutter " << OpenTime << " " << CloseTime << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPixelFilter(SDPCSTRING FilterName, gdouble XWidth, gdouble YWidth)
{
	(*m_Stream) << "PixelFilter " << FormatToken(FilterName) << " " << XWidth << " " << YWidth << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPixelVariance(gdouble Variation)
{
	(*m_Stream) << "PixelVariance " << Variation << std::endl;
	return SDP_S_OK;
}
sdpResult k3dRiEngine::RiPixelSamples(gdouble XSamples, gdouble YSamples)
{
	(*m_Stream) << "PixelSamples " << XSamples << " " << YSamples << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiExposure(gdouble Gain, gdouble Gamma)
{
	(*m_Stream) << "Exposure " << Gain << " " << Gamma << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiImagerV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Imager " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiQuantize(SDPCSTRING Type, gint One, gint QMin, gint QMax, gdouble Amplitude)
{
	(*m_Stream) << "Quantize " << FormatToken(Type) << " " << One << " " << QMin << " " << QMax << " " << Amplitude << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDisplayV(SDPCSTRING Name, SDPCSTRING Type, SDPCSTRING Mode, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Display " << FormatToken(Name) << " " << FormatString(Type) << " " << FormatString(Mode) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiHiderV(SDPCSTRING Type, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Hider " << FormatToken(Type) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiColorSamples(gulong ParameterCount, gdouble nRGB[], gdouble RGBn[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiRelativeDetail(gdouble RelativeDetail)
{
	(*m_Stream) << "RelativeDetail " << RelativeDetail << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiOptionV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Option " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAttributeBegin()
{
	(*m_Stream) << "AttributeBegin" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAttributeEnd()
{
	(*m_Stream) << "AttributeEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiColor(gdouble Color[3])
{
	(*m_Stream) << "Color " << sdpVector3(Color) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiOpacity(gdouble Color[3])
{
	(*m_Stream) << "Opacity " << sdpVector3(Color) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTextureCoordinates(gdouble S1, gdouble T1, gdouble S2, gdouble T2, gdouble S3, gdouble T3, gdouble S4, gdouble T4)
{
	(*m_Stream) << "TextureCoordinates " << S1 << " " << T1 << " " << S2 << " " << T2 << " " << S3 << " " << T3 << " " << S4 << " " << T4 << std::endl;
	return SDP_S_OK;
}

gulong k3dRiEngine::RiLightSourceV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	gulong lighthandle = m_LightHandle++;
	
	(*m_Stream) << "LightSource " << FormatToken(Name) << " " << lighthandle << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return lighthandle;
}

gulong k3dRiEngine::RiAreaLightSourceV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	gulong lighthandle = m_LightHandle++;
	
	(*m_Stream) << "AreaLightSource " << FormatToken(Name) << " " << lighthandle << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return lighthandle;
}

sdpResult k3dRiEngine::RiIlluminate(gulong LightHandle, bool OnOff)
{
	(*m_Stream) << "Illuminate " << LightHandle << " " << OnOff << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSurfaceV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Surface " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAtmosphereV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Atmosphere " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiInteriorV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Interior " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiExteriorV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Exterior " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiShadingRate(gdouble Size)
{
	(*m_Stream) << "ShadingRate " << Size << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiShadingInterpolation(SDPCSTRING Type)
{
	(*m_Stream) << "ShadingInterpolation " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiMatte(bool OnOff)
{
	(*m_Stream) << "Matte " << OnOff << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiBound(gdouble Bound[6])
{
	(*m_Stream) << "Bound " << FormatRealArray(Bound, 6) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDetail(gdouble Bound[6])
{
	(*m_Stream) << "Detail " << FormatRealArray(Bound, 6) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDetailRange(const gdouble MinVis, const gdouble LowTran, const gdouble UpTran, const gdouble MaxVis)
{
	(*m_Stream) << "DetailRange " << MinVis << " " << LowTran << " " << UpTran << " " << MaxVis << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiGeometricApproximation(SDPCSTRING Type, gdouble Value)
{
	(*m_Stream) << "GeometricApproximation " << FormatToken(Type) << " " << Value << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiGeometricRepresentation(SDPCSTRING Type)
{
	(*m_Stream) << "GeometricRepresentation " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiOrientation(SDPCSTRING Orientation)
{
	(*m_Stream) << "Orientation " << FormatToken(Orientation) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiReverseOrientation()
{
	(*m_Stream) << "ReverseOrientation" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSides(const gulong Sides)
{
	(*m_Stream) << "Sides " << Sides << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiIdentity()
{
	(*m_Stream) << "Identity" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTransform(gdouble Transform[16])
{
	(*m_Stream) << "Transform " << FormatRealArray(Transform, 16) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiConcatTransform(gdouble Transform[16])
{
	(*m_Stream) << "ConcatTransform " << FormatRealArray(Transform, 16) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPerspective(gdouble FieldOfView)
{
	(*m_Stream) << "Perspective " << FieldOfView << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTranslate(gdouble DX, gdouble DY, gdouble DZ)
{
	(*m_Stream) << "Translate " << DX << " " << DY << " " << DZ << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiRotate(gdouble Angle, gdouble DX, gdouble DY, gdouble DZ)
{
	(*m_Stream) << "Rotate " << Angle << " " << DX << " " << DY << " " << DZ << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiScale(gdouble DX, gdouble DY, gdouble DZ)
{
	(*m_Stream) << "Scale " << DX << " " << DY << " " << DZ << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSkew(gdouble Angle, gdouble DX1, gdouble DY1, gdouble DZ1, gdouble DX2, gdouble DY2, gdouble DZ2)
{
	(*m_Stream) << "Skew " << Angle << " " << DX1 << " " << DY1 << " " << DZ1 << " " << DX2 << " " << DY2 << " " << DZ2 << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDeformationV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Deformation " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDisplacementV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Displacement " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCoordinateSystem(SDPCSTRING Space)
{
	(*m_Stream) << "CoordinateSystem " << FormatToken(Space) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCoordSysTransform(SDPCSTRING Space)
{
	(*m_Stream) << "CoordSysTransform " << FormatToken(Space) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTransformBegin()
{
	(*m_Stream) << "TransformBegin" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTransformEnd()
{
	(*m_Stream) << "TransformEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAttributeV(SDPCSTRING Name, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Attribute " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPathV(const gulong VertexCount, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Path " << VertexCount << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiGeneralPathV(const gulong LoopCount, const gulong VertexCounts[], const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	// Calculate the total number of vertices ...
	gulong vertexcount = 0;
	for(gulong i = 0; i < LoopCount; i++)
		vertexcount += VertexCounts[i];

	(*m_Stream) << "GeneralPath " << FormatUnsignedLongArray(const_cast<gulong*>(VertexCounts), LoopCount) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPointsPathsV(gulong PathCount, gulong VertexCounts[], gulong Vertexes[], const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiPointsGeneralPathsV(gulong PathCount, gulong LoopCounts[], gulong VertexCounts[], gulong Vertices[], const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiBasis(gdouble UBasis[16], const gulong UStep, gdouble VBasis[16], const gulong VStep)
{
	(*m_Stream) << "Basis " << FormatRealArray(UBasis, 16) << " " << UStep << " " << FormatRealArray(VBasis, 16) << " " << VStep << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPatchV(SDPCSTRING Type, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Patch " << FormatToken(Type) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPatchMeshV(SDPCSTRING Type, const gulong UCount, SDPCSTRING UWrap, const gulong VCount,	SDPCSTRING VWrap, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiNuPatchV(const gulong UCount, const gulong UOrder, gdouble UKnot[], const gdouble UMin, const gdouble UMax, const gulong VCount, const gulong VOrder, gdouble VKnot[], const gdouble VMin, const gdouble VMax, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiTrimCurve(gulong LoopCount, gulong CurveCounts[], gulong Order[], gdouble Knot[], gdouble AMin[], gdouble AMax[], gulong ParameterCount, gdouble U[], gdouble V[], gdouble W[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiSphereV(gdouble Radius, gdouble ZMin, gdouble ZMax, gdouble ThetaMax, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Sphere " << Radius << " " << ZMin << " " << ZMax << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiConeV(gdouble Height, gdouble Radius, gdouble ThetaMax, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Cone " << Height << " " << Radius << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCylinderV(gdouble Radius, gdouble ZMin, gdouble ZMax, gdouble ThetaMax,	const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Cylinder " << Radius << " " << ZMin << " " << ZMax << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiHyperboloidV(gdouble Point1[3], gdouble Point2[3], gdouble ThetaMax, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Hyperboloid " << sdpVector3(Point1) << " " << sdpVector3(Point2) << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiParaboloidV(gdouble RMax, gdouble ZMin, gdouble ZMax, gdouble ThetaMax, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Paraboloid " << RMax << " " << ZMin << " " << ZMax << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDiskV(gdouble Height, gdouble Radius, gdouble ThetaMax, 	const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Disk " << Height << " " << Radius << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTorusV(gdouble MajorRadius, gdouble MinorRadius, gdouble PhiMin,	gdouble PhiMax, gdouble ThetaMax, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Torus " << MajorRadius << " " << MinorRadius << " " << PhiMin << " " << PhiMax << " " << ThetaMax << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCurvesV(SDPCSTRING Degree, const gulong ParameterCountCurves, const gulong ParameterCountVerts[], SDPCSTRING Wrap, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiGeometryV(SDPCSTRING Type, const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	(*m_Stream) << "Geometry " << FormatToken(Type) << " " << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSolidBegin(SDPCSTRING Type)
{
	(*m_Stream) << "SolidBegin " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSolidEnd()
{
	(*m_Stream) << "SolidEnd" << std::endl;
	return SDP_S_OK;
}

gulong k3dRiEngine::RiObjectBegin()
{
	gulong objecthandle = m_ObjectHandle++;
	(*m_Stream) << "ObjectBegin " << objecthandle << std::endl;
	
	return objecthandle;
}

sdpResult k3dRiEngine::RiObjectEnd()
{
	(*m_Stream) << "ObjectEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiObjectInstance(gulong Handle)
{
	(*m_Stream) << "ObjectInstance " << Handle << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiMotionBeginV(gulong ParameterCount, gdouble Times[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiMotionEnd()
{
	(*m_Stream) << "MotionEnd" << std::endl; 
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiErrorIgnore(gint Code, gint Severity, SDPCSTRING Message)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiErrorPrint(gint Code, gint Severity, SDPCSTRING Message)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiErrorAbort(gint Code, gint Severity, SDPCSTRING Message)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiComment(SDPCSTRING Comment)
{
	(*m_Stream) << "#" << Comment << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiNewline()
{
	(*m_Stream) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiReadArchive(SDPCSTRING Archive)
{
	(*m_Stream) << "ReadArchive " << Archive << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiStructure(SDPCSTRING Structure)
{
	(*m_Stream) << "##" << Structure << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSubdivisionMeshV(SDPCSTRING Scheme, const gulong PathCount, gulong VertexCounts[], gulong Vertices[], gulong TagCount, SDPSTRING Tags[], gulong ArgCounts[], gulong IntegerArgs[], gdouble FloatArgs[], const gulong ParameterCount, SDPCSTRING Tokens[], gpointer Values[], gulong ValueCounts[])
{
	// Calculate the number of vertices in the Vertexes[] array.
	// From RiSpec 3.2 - VertexCounts length is PathCount,
	//                   Vertexes length is sum of all values in VertexCounts.
	std::vector<gulong> vertexcounts(VertexCounts, VertexCounts + PathCount);
	const gulong vertexcount = std::accumulate(vertexcounts.begin(), vertexcounts.end(), 0);

/*
	// Calculate the number of points required by finding the largest vertex index.
	std::vector<gulong> vertices(Vertices, Vertices + vertexcount);
	const gulong pointcount = (*std::max_element(vertices.begin(), vertices.end())) + 1;
*/

	// Write out the basics ...
	(*m_Stream) << "SubdivisionMesh " << FormatToken(Scheme) << " " << FormatUnsignedLongArray(VertexCounts, PathCount) << " " << FormatUnsignedLongArray(Vertices, vertexcount) << " ";
	
	// If tags are specified, output them with their arguments.
	if(TagCount > 0)
		{
			// Calculate the number of arguments in the IntegerArgs and FloatArgs arrays.
			// From RiSpec 3.2 - ArgCounts length is TagCount*2
			//                   IntegerArgs length is sum of all even numbered entries in ArgCounts 
			//                   FloatArgs length is sum of all odd numbered entries in ArgCounts 
			gulong integercount = 0;
			gulong floatcount = 0;
			for(gulong i = 0; i < TagCount; i++)
				{
					integercount += ArgCounts[(i*2)];
					floatcount += ArgCounts[(i*2)+1];
				}
		
			(*m_Stream) << " " << FormatStringArray(Tags, TagCount) << " " << FormatUnsignedLongArray(ArgCounts, TagCount * 2) << " " <<  FormatIntegerArray(IntegerArgs, integercount) << " " << FormatRealArray(FloatArgs, floatcount) << " ";
		}

	// Output the parameters
	(*m_Stream) << FormatParameters(ParameterCount, Tokens, Values, ValueCounts, m_Declarations) << std::endl;

	return SDP_S_OK;
}



