
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements convenience classes for sending rendering parameters to a k3dIRiEngine interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3drishadersetup.h>
#include <iostream>

///////////////////////////////////////////////////////////////////////////
// k3dRiArray

k3dRiArray::k3dRiArray(const sdpString Type) :
	m_Type(Type)
{
	// Sanity checks ...
	g_assert(m_Type.size());
}

k3dRiArray::~k3dRiArray()
{
}

const sdpString& k3dRiArray::Type() const
{
	return m_Type;
}

///////////////////////////////////////////////////////////////////////////
// k3dRiIntegerArray

k3dRiIntegerArray::k3dRiIntegerArray(const sdpString Type) :
	k3dRiArray(Type)
{
}

k3dRiIntegerArray::k3dRiIntegerArray(const sdpString Type, const gint Value) :
	k3dRiArray(Type)
{
	Append(Value);
}

void k3dRiIntegerArray::Append(const gint Value)
{
	m_Values.push_back(Value);
}

gpointer k3dRiIntegerArray::Values() const
{
	return m_Values.size() ? static_cast<gpointer>(const_cast<gint*>(&m_Values[0])) : 0;
}

gulong k3dRiIntegerArray::ValueCount() const
{
	return m_Values.size();
}

///////////////////////////////////////////////////////////////////////////
// k3dRiRealArray

k3dRiRealArray::k3dRiRealArray(const sdpString Type) :
	k3dRiArray(Type)
{
}

k3dRiRealArray::k3dRiRealArray(const sdpString Type, const gdouble Value) :
	k3dRiArray(Type)
{
	Append(Value);
}

k3dRiRealArray::k3dRiRealArray(const sdpString Type, const sdpVector3 Value) :
	k3dRiArray(Type)
{
	Append(Value);
}

void k3dRiRealArray::Append(const gdouble Value)
{
	m_Values.push_back(Value);
}

void k3dRiRealArray::Append(const sdpVector3 Value)
{
	m_Values.push_back(Value[0]);
	m_Values.push_back(Value[1]);
	m_Values.push_back(Value[2]);
}

gpointer k3dRiRealArray::Values() const
{
	return m_Values.size() ? static_cast<gpointer>(const_cast<gdouble*>(&m_Values[0])) : 0;
}

gulong k3dRiRealArray::ValueCount() const
{
	return m_Values.size();
}

///////////////////////////////////////////////////////////////////////////
// k3dRiStringArray

k3dRiStringArray::k3dRiStringArray(const sdpString Type) :
	k3dRiArray(Type)
{
}

k3dRiStringArray::k3dRiStringArray(const sdpString Type, const sdpString Value) :
	k3dRiArray(Type)
{
	Append(Value);
}

void k3dRiStringArray::Append(const sdpString Value)
{
	m_Values.push_back(Value);
	m_ValuePointers.push_back(const_cast<SDPSTRING>(static_cast<SDPCSTRING>(m_Values.back())));
}

gpointer k3dRiStringArray::Values() const
{
	return m_ValuePointers.size() ? gpointer(&m_ValuePointers[0]) : 0;
}

gulong k3dRiStringArray::ValueCount() const
{
	return m_ValuePointers.size();
}

///////////////////////////////////////////////////////////////////////////
// k3dRiTokenValuePair

k3dRiTokenValuePair::k3dRiTokenValuePair(const sdpString Token, k3dRiArray* const Values) :
	m_Token(Token),
	m_Values(Values)
{
	// Sanity checks ...
	g_assert(m_Token.size());
	g_assert(m_Values);
}

k3dRiTokenValuePair::~k3dRiTokenValuePair()
{
	delete m_Values;
}

const sdpString& k3dRiTokenValuePair::Token() const
{
	return m_Token;
}

const sdpString& k3dRiTokenValuePair::Type() const
{
	return m_Values->Type();
}

gpointer k3dRiTokenValuePair::Values() const
{
	return m_Values->Values();
}

gulong k3dRiTokenValuePair::ValueCount() const
{
	return m_Values->ValueCount();
}

///////////////////////////////////////////////////////////////////////////
// k3dRiParameterList

k3dRiParameterList::k3dRiParameterList()
{
}

k3dRiParameterList::k3dRiParameterList(k3dRiTokenValuePair* const Pair)
{
	// Sanity checks ...
	g_assert(Pair);
	
	AddTokenValuePair(Pair);
}

k3dRiParameterList::~k3dRiParameterList()
{
	std::for_each(m_Pairs.begin(), m_Pairs.end(), sdp::DeleteObject());
}

void k3dRiParameterList::AddTokenValuePair(k3dRiTokenValuePair* const Pair)
{
	// Sanity checks ...
	g_assert(Pair);

	// Don't allow duplicate tokens!!!
	for(PairArray::iterator pair = m_Pairs.begin(); pair != m_Pairs.end(); pair++)
		if((*pair)->Token() == Pair->Token())
			{
				std::cerr << "k3dRiParameterList::AddTokenValuePair() - Duplicate token-value pair [" << Pair->Token() << "] will not be added" << std::endl;
				g_return_if_fail(0);
			}
			
	m_Tokens.push_back(const_cast<SDPSTRING>(static_cast<SDPCSTRING>(Pair->Token())));
	m_Values.push_back(Pair->Values());
	m_ValueCounts.push_back(Pair->ValueCount());
	
	m_Pairs.push_back(Pair);
}

// Declares parameters and their types ...
void k3dRiParameterList::DeclareParameters(k3dIRiEngine* const Engine) const
{
	// Sanity checks ...
	g_assert(Engine);

	// Declare each parameter ...
	for(PairArray::const_iterator pair = m_Pairs.begin(); pair != m_Pairs.end(); pair++)
		Engine->RiDeclare((*pair)->Token(), (*pair)->Type());
}

void k3dRiParameterList::SetAttributes(SDPCSTRING Name, k3dIRiEngine* const Engine) const
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(Engine);

	// If we don't have any parameters, we're done!
	if(!ParameterCount())
		return;

	// Set attributes ...
	Engine->RiAttributeV(Name, ParameterCount(), Tokens(), Values(), ValueCounts());
}

void k3dRiParameterList::SetOptions(SDPCSTRING Name, k3dIRiEngine* const Engine) const
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(Engine);

	// If we don't have any parameters, we're done!
	if(!ParameterCount())
		return;

	// Set options ...
	Engine->RiOptionV(Name, ParameterCount(), Tokens(), Values(), ValueCounts());
}

gulong k3dRiParameterList::ParameterCount() const
{
	return m_Pairs.size();
}

SDPCSTRING* k3dRiParameterList::Tokens() const
{
	return m_Pairs.size() ? &m_Tokens[0] : 0;
}

gpointer* k3dRiParameterList::Values() const
{
	return m_Pairs.size() ? (gpointer*)&m_Values[0] : 0;
}

gulong* k3dRiParameterList::ValueCounts() const
{
	return m_Pairs.size() ? const_cast<gulong*>(&m_ValueCounts[0]) : 0;
}

/////////////////////////////////////////////////////////////////////////////
// k3dRiShader

// Constructor for use with a ShaderSelector ...
k3dRiShader::k3dRiShader()
{
}

// Constructor for stand-alone use ...
k3dRiShader::k3dRiShader(SDPCSTRING ShaderName)
{
	SetShader(ShaderName);
}

// Explicitly set the shader name ...
void k3dRiShader::SetShader(SDPCSTRING ShaderName)
{
	m_ShaderName = ShaderName;
}

//////////////////////////////////////////////////////////////////////
// k3dRiLightShader

k3dRiLightShader::k3dRiLightShader()
{
}

k3dRiLightShader::k3dRiLightShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

gulong k3dRiLightShader::SetupLight(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiLightSourceV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

gulong k3dRiLightShader::SetupAreaLight(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiAreaLightSourceV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

//////////////////////////////////////////////////////////////////////
// k3dRiVolumeShader

k3dRiVolumeShader::k3dRiVolumeShader()
{
}

k3dRiVolumeShader::k3dRiVolumeShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiVolumeShader::SetupInterior(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiInteriorV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

sdpResult k3dRiVolumeShader::SetupExterior(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiExteriorV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

sdpResult k3dRiVolumeShader::SetupAtmosphere(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiAtmosphereV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

//////////////////////////////////////////////////////////////////////
// k3dRiTransformationShader

k3dRiTransformationShader::k3dRiTransformationShader()
{
}

k3dRiTransformationShader::k3dRiTransformationShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiTransformationShader::Setup(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiDeformationV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

//////////////////////////////////////////////////////////////////////
// k3dRiDisplacementShader

k3dRiDisplacementShader::k3dRiDisplacementShader()
{
}

k3dRiDisplacementShader::k3dRiDisplacementShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiDisplacementShader::Setup(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiDisplacementV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

//////////////////////////////////////////////////////////////////////
// k3dRiSurfaceShader

k3dRiSurfaceShader::k3dRiSurfaceShader()
{
}

k3dRiSurfaceShader::k3dRiSurfaceShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiSurfaceShader::Setup(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiSurfaceV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}

//////////////////////////////////////////////////////////////////////
// k3dRiImagerShader

k3dRiImagerShader::k3dRiImagerShader()
{
}

k3dRiImagerShader::k3dRiImagerShader(SDPCSTRING ShaderName)
	: k3dRiShader(ShaderName)
{
}

sdpResult k3dRiImagerShader::Setup(k3dIRiEngine* const Engine)
{
	DeclareParameters(Engine);
	return Engine->RiImagerV(m_ShaderName, ParameterCount(), Tokens(), Values(), ValueCounts());
}


