#ifndef K3DRISHADERSETUP_H
#define K3DRISHADERSETUP_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares convenience classes for sending rendering parameters to a k3dIRiEngine interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dicamera.h>
#include <k3driengine.h>

///////////////////////////////////////////////////////////////////////////
// k3dRiArray

/// Abstract interface for an array of values
class k3dRiArray
{
public:
	k3dRiArray(const sdpString Type);
	virtual ~k3dRiArray();
	
	const sdpString& Type() const;
	virtual gpointer Values() const = 0;
	virtual gulong ValueCount() const = 0;
	
protected:
	const sdpString m_Type;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiIntegerArray

class k3dRiIntegerArray :
	public k3dRiArray
{
public:
	k3dRiIntegerArray(const sdpString Type);
	k3dRiIntegerArray(const sdpString Type, const gint Value);
	
	void Append(const gint Value);
	
	gpointer Values() const;
	gulong ValueCount() const;

protected:
	std::vector<gint> m_Values;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiRealArray

class k3dRiRealArray :
	public k3dRiArray
{
public:
	k3dRiRealArray(const sdpString Type);
	k3dRiRealArray(const sdpString Type, const gdouble Value);
	k3dRiRealArray(const sdpString Type, const sdpVector3 Value);
	
	void Append(const gdouble Value);
	void Append(const sdpVector3 Value);
	
	gpointer Values() const;
	gulong ValueCount() const;
	
protected:
	std::vector<gdouble> m_Values;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiStringArray

class k3dRiStringArray :
	public k3dRiArray
{
public:
	k3dRiStringArray(const sdpString Type);
	k3dRiStringArray(const sdpString Type, const sdpString Value);

	void Append(const sdpString Value);

	gpointer Values() const;
	gulong ValueCount() const;
	
protected:
	std::vector<sdpString> m_Values;
	std::vector<SDPSTRING> m_ValuePointers;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiTokenValuePair

class k3dRiTokenValuePair
{
public:
	k3dRiTokenValuePair(const sdpString Token, k3dRiArray* const Values);
	virtual ~k3dRiTokenValuePair();

	const sdpString& Token() const;
	const sdpString& Type() const;
	gpointer Values() const;
	gulong ValueCount() const;

protected:
	const sdpString m_Token;
	k3dRiArray* const m_Values;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiParameterList

/// Encapsulates a set of token-value pairs
class k3dRiParameterList
{
public:
	k3dRiParameterList();
	k3dRiParameterList(k3dRiTokenValuePair* const Pair);
	
	virtual ~k3dRiParameterList();

	void AddTokenValuePair(k3dRiTokenValuePair* const Pair);

	/// Declares the type of each parameter to RenderMan
	void DeclareParameters(k3dIRiEngine* const Engine) const;
	/// Passes the parameters to RenderMan as options
	void SetOptions(SDPCSTRING Name, k3dIRiEngine* const Engine) const;
	/// Passes the parameters to RenderMan as attributes
	void SetAttributes(SDPCSTRING Name, k3dIRiEngine* const Engine) const;

	/// Returns the number of token-value pairs
	gulong ParameterCount() const;
	/// Returns a pointer to the array of parameter tokens
	SDPCSTRING* Tokens() const;
	/// Returns a pointer to the array of parameter values
	gpointer* Values() const;
	/// Returns a pointer to the array of parameter value counts
	gulong* ValueCounts() const;

protected:
	/// Stores the array of tokens to be passed to RenderMan
	typedef std::vector<SDPSTRING> TokenArray;
	TokenArray m_Tokens;
	/// Stores the array of values to be passed to RenderMan
	typedef std::vector<gpointer> ValueArray;
	ValueArray m_Values;
	/// Stores the array of value counts to be passed to RenderMan
	typedef std::vector<gulong> ValueCountArray;
	ValueCountArray m_ValueCounts;

	/// Stores the array of token-value pairs
	typedef std::vector<k3dRiTokenValuePair*> PairArray;
	PairArray m_Pairs;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiShader

/// Encapsulates a RenderMan Interface shader: a shader name, and a collection of parameters
class k3dRiShader : public k3dRiParameterList
{
public:
	// Default constructor, normally used with a ShaderSelector ...
	k3dRiShader();

	// Constructor for use as a stand-alone object ...
	k3dRiShader(SDPCSTRING ShaderName);

	/// Sets the name of the shader
	void SetShader(SDPCSTRING ShaderName);
	/// Returns the shader name
	SDPCSTRING ShaderName() { return m_ShaderName; }

protected:
	sdpString m_ShaderName;
};

////////////////////////////////////////////////////////////////////////////
// k3dRiLightShader

/// Encapsulates a RenderMan Interface light shader
class k3dRiLightShader : public k3dRiShader
{
public:
	k3dRiLightShader();
	k3dRiLightShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to RenderMan as a normal light source
	gulong SetupLight(k3dIRiEngine* const Engine);
	/// Sends the parameter list to RenderMan as an area light source
	gulong SetupAreaLight(k3dIRiEngine* const Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiVolumeShader

/// Encapsulates a RenderMan Interface volume shader
class k3dRiVolumeShader : public k3dRiShader
{
public:
	k3dRiVolumeShader();
	k3dRiVolumeShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to RenderMan as an interior volume
	sdpResult SetupInterior(k3dIRiEngine* const Engine);
	/// Sends the parameter list to RenderMan as an exterior volume
	sdpResult SetupExterior(k3dIRiEngine* const Engine);
	/// Sends the parameter list to RenderMan as an atmosphere shader
	sdpResult SetupAtmosphere(k3dIRiEngine* const Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiTransformationShader

/// Encapsulates a RenderMan Interface transformation shader
class k3dRiTransformationShader : public k3dRiShader
{
public:
	k3dRiTransformationShader();
	k3dRiTransformationShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to RenderMan as a transformation shader
	sdpResult Setup(k3dIRiEngine* const Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiDisplacementShader

/// Encapsulates a RenderMan Interface displacement shader
class k3dRiDisplacementShader : public k3dRiShader
{
public:
	k3dRiDisplacementShader();
	k3dRiDisplacementShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to RenderMan as a displacement shader
	sdpResult Setup(k3dIRiEngine* const Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiSurfaceShader

/// Encapsulates a RenderMan Interface surface shader
class k3dRiSurfaceShader : public k3dRiShader
{
public:
	k3dRiSurfaceShader();
	k3dRiSurfaceShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to RenderMan as a surface shader
	sdpResult Setup(k3dIRiEngine* const Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiImagerShader

/// Encapsulates a RenderMan Interface imaging shader
class k3dRiImagerShader : public k3dRiShader
{
public:
	k3dRiImagerShader();
	k3dRiImagerShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to RenderMan as an imaging shader
	sdpResult Setup(k3dIRiEngine* const Engine);
};

#endif // K3DRISHADERSETUP_H
