#ifndef K3DSELECTION_H
#define K3DSELECTION_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares convenience containers, templates, and functions for dealing with the current selection 
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3diobjectcollection.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>
#include <k3dipathcollection.h>
#include <k3distate.h>

#include <iterator>

// Forward class declarations
class k3dIDocument;

namespace k3d
{

/// Proxy for a selected k3dILocation and its original value	
class SelectedLocationProxy
{
public:
	SelectedLocationProxy(k3dILocation* const Location, const sdpVector3& OldXYZ) :
		m_Location(Location),
		m_OldXYZ(OldXYZ),
		m_NewXYZ(OldXYZ)
	{
		// Sanity checks ...
		g_assert(m_Location);
	}

	SelectedLocationProxy(k3dILocation* const Location, const sdpVector3& OldXYZ, const sdpVector3& NewXYZ) :
		m_Location(Location),
		m_OldXYZ(OldXYZ),
		m_NewXYZ(NewXYZ)
	{
		// Sanity checks ...
		g_assert(m_Location);
	}

	k3dILocation* operator->() const
	{
		return m_Location;
	}

	operator k3dILocation*() const
	{
		return m_Location;
	}

	const sdpVector3& OldXYZ() const
	{
		return m_OldXYZ;
	}
	
	sdpVector3& NewXYZ()
	{
		return m_NewXYZ;
	}
	
private:
	/// Stores a reference to the k3dILocation this object proxies for
	k3dILocation* m_Location;
	/// Stores the original position of the k3dILocation this object proxies for
	sdpVector3 m_OldXYZ;
	/// Scratch space for clients to store data (handy for tools that lerp between an object and its transformed state)
	sdpVector3 m_NewXYZ;
};

/// Convenience typedef for an array of SelectedLocationProxy objects (the preferred way to store a collection of them, due to locality-of-reference)
typedef std::vector<SelectedLocationProxy> SelectedLocationArray;

/// Inserts SelectedLocationProxy objects into a collection for all selected locations in a document
template<typename OutputIterator>
void GetSelectedLocations(k3dIObjectCollection* const Objects, OutputIterator Collection)
{
	// Sanity checks ...
	g_return_if_fail(Objects);

	// For each object in the document ...
	for(K3DOBJECTHANDLE objecthandle = Objects->FirstObject(); objecthandle; objecthandle = Objects->NextObject(objecthandle))
		{
			k3dILocationCollection* const locationcollection = dynamic_cast<k3dILocationCollection*>(Objects->Object(objecthandle));
			if(!locationcollection)
				continue;

			for(K3DLOCATIONHANDLE locationhandle = locationcollection->FirstLocation(); locationhandle; locationhandle = locationcollection->NextLocation(locationhandle))
				{
					k3dILocation* const location = locationcollection->Location(locationhandle);
					if(SDP_S_OK == location->State()->QueryState(k3dIState::VISIBLESELECTED))
						{
							*Collection++ = SelectedLocationProxy(location, location->LocalXYZ());
						}
				}
		}
}

/// Proxy for a selected k3dIPath and its original value	
class SelectedPathProxy
{
public:
	SelectedPathProxy(k3dIPath* const Path) :
		m_Path(Path)
	{
		// Sanity checks ...
		g_assert(m_Path);
	}

	k3dIPath* operator->() const
	{
		return m_Path;
	}

	operator k3dIPath*() const
	{
		return m_Path;
	}

private:
	/// Stores a reference to the k3dIPath this object proxies for
	k3dIPath* m_Path;
};

/// Convenience typedef for an array of SelectedPathProxy objects (the preferred way to store a collection of them, due to locality-of-reference)
typedef std::vector<SelectedPathProxy> SelectedPathArray;

/// Inserts SelectedPathProxy objects into a collection for all selected locations in a document
template<typename OutputIterator>
void GetSelectedPaths(k3dIObjectCollection* const Objects, OutputIterator Collection)
{
	// Sanity checks ...
	g_return_if_fail(Objects);

	// For each object in the document ...
	for(K3DOBJECTHANDLE objecthandle = Objects->FirstObject(); objecthandle; objecthandle = Objects->NextObject(objecthandle))
		{
			k3dIPathCollection* const pathcollection = dynamic_cast<k3dIPathCollection*>(Objects->Object(objecthandle));
			if(!pathcollection)
				continue;

			for(K3DPATHHANDLE pathhandle = pathcollection->FirstPath(); pathhandle; pathhandle = pathcollection->NextPath(pathhandle))
				{
					k3dIPath* const path = pathcollection->Path(pathhandle);
					if(SDP_S_OK == path->State()->QueryState(k3dIState::VISIBLESELECTED))
						{
							*Collection++ = SelectedPathProxy(path);
						}
				}
		}
}

/// Calculates / caches data about the dimensions of a set of points
class Dimensions
{
public:
	Dimensions();

	/// Adds a point to the collection, updating statistics as needed
	void Insert(const sdpVector3 Coords);
		
	/// Returns the number of locations in the input data
	gulong LocationCount() const;
	/// Returns the minimum coordinates (along each axis) of the input data
	sdpVector3 Min() const;
	/// Returns the maximum coordinates (along each axis) of the input data
	sdpVector3 Max() const;
	/// Returns the geometric center of the input data
	sdpVector3 Center() const;
	/// Returns the average of the input data
	sdpVector3 Average() const;
	/// Returns the maximum extents (furthest distance from the origin, along each axis) of the input data
	sdpVector3 Extents() const;
	/// Returns the size of a bounding-box that can hold the input data
	sdpVector3 Size() const;
	/// Returns the computed normal vector for the collection of points in the container
	sdpVector3 Normal() const;
	
	/// Insertion iterator for initializing the collection from STL-like algorithms
	friend class Inserter/* :
		public std::iterator<std::output_iterator_tag, void, void, void, void>*/
	{
	public:
		explicit Inserter(Dimensions& Container) :
			m_Container(Container)
		{
		}
		
		Inserter& operator=(const SelectedLocationProxy& Location)
		{
			m_Container.Insert(Location.OldXYZ());
			return *this;
		}

		Inserter& operator=(k3dIPathPoint* const PathPoint)
		{
			m_Container.Insert(PathPoint->Location()->LocalXYZ());
			return *this;
		}
		
		Inserter& operator*()
		{
			return *this;
		}
		
		Inserter& operator++()
		{
			return *this;
		}
		
		Inserter operator++(int)
		{
			return *this;
		}
		
	private:
		Dimensions& m_Container;
	};

private:
	typedef std::vector<sdpVector3> Locations;
	Locations m_Locations;
	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Total;
};

} // namespace k3d

#endif // K3DSELECTION_H

