// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderArgumentsPage class, which provides a UI for editing a set of shader arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dshaderargumentspage.h>

#include <k3ditexture.h>

#include <sdpgtkevents.h>

#include <iostream>

/// Button factory
static k3dShaderArgumentsPageButton* CreateButton(k3dShaderInstanceArgument* const Argument, k3dICommandNode* const ParentCommandNode)
{
	// Sanity checks ...
	g_assert(Argument);

	switch(Argument->Type())
		{
			case sdpsl::IShaderArgument::POINT:
			case sdpsl::IShaderArgument::VECTOR:
			case sdpsl::IShaderArgument::NORMAL:
			case sdpsl::IShaderArgument::MATRIX:
			case sdpsl::IShaderArgument::STRING:
				if(sdpsl::IShaderArgument::EX_TEXTURE == Argument->ExtendedType())
					return new k3dShaderArgumentsPageTextureButton(static_cast<k3dShaderInstanceTextureArgument*>(Argument), ParentCommandNode);
				else
					return new k3dShaderArgumentsPageStringButton(static_cast<k3dShaderInstanceStringArgument*>(Argument), ParentCommandNode);
			case sdpsl::IShaderArgument::FLOAT:
			case sdpsl::IShaderArgument::COLOR:
				return new k3dShaderArgumentsPageChannelButton(static_cast<k3dShaderInstanceChannelArgument*>(Argument), ParentCommandNode);
		}
		
	std::cerr << "CreateButton(): Unknown type [" << Argument->Type() << "] for [" << Argument->Name() << "]" << std::endl;
	return 0;
}

//////////////////////////////////////////////////////////////////////
// k3dShaderArgumentsPageChannelButton

k3dShaderArgumentsPageChannelButton::k3dShaderArgumentsPageChannelButton(k3dShaderInstanceChannelArgument* const Argument, sdpIUnknown* ParentCommandNode) :
	m_Button(ParentCommandNode, Argument->Name()),
	m_Channel(Argument->Channel())
{
	// Sanity checks ...
	g_assert(m_Channel);
}

bool k3dShaderArgumentsPageChannelButton::Create(GtkWidget* Widget)
{
	// Sanity checks ...
	g_assert(Widget);

	// Attach the channel button to its parent GTK+ container ...
	sdpGtkContainer container(GTK_CONTAINER(Widget));
	container.Attach(m_Button.RootWidget());

	// Attach the channel button to the channel ...
	g_return_val_if_fail(m_Button.Attach(m_Channel), false);

	return true;
}

//////////////////////////////////////////////////////////////////////
// k3dShaderArgumentsPageStringButton

k3dShaderArgumentsPageStringButton::k3dShaderArgumentsPageStringButton(k3dShaderInstanceStringArgument* const Argument, sdpIUnknown* ParentCommandNode) :
	m_Button(ParentCommandNode, Argument->Name()),
	m_String(Argument->Value())
{
}

bool k3dShaderArgumentsPageStringButton::Create(GtkWidget* Widget)
{
	// Sanity checks ...
	g_assert(Widget);

	// Create the button ...
	m_Button.Create();

	// Attach the button to its parent GTK+ container ...
	sdpGtkContainer container(GTK_CONTAINER(Widget));
	container.Attach(m_Button.RootWidget());

	// Attach the button to the data source ...
	g_return_val_if_fail(m_Button.Attach(&m_String, 0, 0, 0, 0, 0, 0), false);

	return true;
}

//////////////////////////////////////////////////////////////////////
// k3dShaderArgumentsPageTextureButton

k3dShaderArgumentsPageTextureButton::k3dShaderArgumentsPageTextureButton(k3dShaderInstanceTextureArgument* const Argument, sdpIUnknown* ParentCommandNode) :
	m_Selector(ParentCommandNode, Argument->Name()),
	m_ObjectContainer(&Argument->TextureContainer())
{
	// Sanity checks ...
	g_assert(m_ObjectContainer);
}

k3dShaderArgumentsPageTextureButton::~k3dShaderArgumentsPageTextureButton()
{
//std::cout << "k3dShaderArgumentsPageTextureButton::~k3dShaderArgumentsPageTextureButton()" << std::endl;
}

bool k3dShaderArgumentsPageTextureButton::Create(GtkWidget* Widget)
{
	// Sanity checks ...
	g_assert(Widget);
	
	// Attach the object selector to its parent GTK+ container ...
	sdpGtkContainer container(GTK_CONTAINER(Widget));
	container.Attach(m_Selector.RootWidget());
	
	// Attach the object selector to the object ... (k3dITexture)
	g_return_val_if_fail(m_Selector.Attach(m_ObjectContainer, 0, 0, 0, 0, 0, 0, true, new k3dObjectSelectorInterfaceFilter(new sdp::TypeInfo<k3dITexture>())), false);
	return true;
}

//////////////////////////////////////////////////////////////////////
// k3dShaderArgumentsPage

k3dShaderArgumentsPage::k3dShaderArgumentsPage(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName) :
	k3dShaderArgumentsPageBase(ParentCommandNode, CommandNodeName),
	m_Document(0),
	m_ShaderInstance(0)
{
	// Create our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<scrolledwindow>"
				"<frame shadowtype=\"none\" name=\"frame\">"
				"</frame>"
			"</scrolledwindow>"
		"</gtkml>\n");

	// Load our template ...
	g_return_if_fail(Load(uitemplate, ""));
}

k3dShaderArgumentsPage::~k3dShaderArgumentsPage()
{
	// If we were attached before, we don't need to hear about any more events ...
	if(m_ShaderInstance)
		m_ShaderInstance->RemoveEventObserver(this);

	// Let our buttons go ...
	ClearButtons();
}

bool k3dShaderArgumentsPage::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	
	
	return true;
}

bool k3dShaderArgumentsPage::Attach(k3dIDocument* const Document, k3dShaderInstance* const ShaderInstance)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ShaderInstance);
	g_assert(0 == m_ShaderInstance);	// You only get to Attach() once!

	// Store data ...
	m_Document = Document;
	m_ShaderInstance = ShaderInstance;
	m_ShaderInstance->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dShaderArgumentsPage");

	UpdateButtons();
	
	return true;
}

void k3dShaderArgumentsPage::UpdateButtons()
{
	// Sanity checks ...
	g_assert(m_ShaderInstance);	// You forgot to call Attach()!!!

	// Get rid of any old stuff ...
	ClearButtons();

	if(m_Table.Attached())
		m_Table.Destroy();

	// If the current shader instance doesn't have any arguments, we're done ...
	if(0 == m_ShaderInstance->m_Arguments.size())
		return;

	// Create the table control for organizing buttons & lables ...
	sdpGtkContainer frame(Container("frame"));
	g_return_if_fail(m_Table.Create(m_ShaderInstance->m_Arguments.size(), 2, false));
	frame.Attach(m_Table);

	// For each argument ...
	gulong row = 0;
	for(k3dShaderInstance::Arguments::const_iterator argument = m_ShaderInstance->m_Arguments.begin(); argument != m_ShaderInstance->m_Arguments.end(); argument++)
		{
			// Create something to attach our button to ...
			sdpGtkEventBox box;
			box.Create();
			box.Show();
			m_Table.Attach(box, 0, 1, row, row+1, GtkAttachOptions(GTK_EXPAND | GTK_FILL), GtkAttachOptions(GTK_EXPAND | GTK_FILL), 0, 0);

			// Create a button ...
			k3dShaderArgumentsPageButton* const button = CreateButton(*argument, this);
			g_return_if_fail(button);
			g_return_if_fail(button->Create(box));
			m_Buttons.push_back(button);

			sdpGtkLabel label;
			label.Create((*argument)->Name());
			label.Show();
			m_Table.Attach(label, 1, 2, row, row+1, GTK_FILL, GTK_FILL, 4, 0);
			
			row++;
		}

	m_Table.Show();		
}

void k3dShaderArgumentsPage::ClearButtons()
{
	// Get rid of our buttons ...
	std::for_each(m_Buttons.begin(), m_Buttons.end(), sdp::DeleteObject());
	m_Buttons.clear();
}

const sdpString k3dShaderArgumentsPage::CustomType() const
{
	return sdpString("k3dShaderArgumentsPage");
}

sdpResult k3dShaderArgumentsPage::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	sdpString event(Event);
	
	if(event == k3dIVariable::Event_Changed)
		{
			UpdateButtons();
			return SDP_S_OK;
		}
	
	return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
}


