// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderInfo class, which provides a user interface for choosing Pixar RenderMan shaders
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dshaderinstance.h>
#include <k3dshaderinfo.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3ditexture.h>

#include <sdpgtkevents.h>

const sdpString k3dShaderInfo::control_authors = "authors";
const sdpString k3dShaderInfo::control_copyright = "copyright";
const sdpString k3dShaderInfo::control_description = "description";

k3dShaderInfo::k3dShaderInfo(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName) :
	k3dShaderInfoBase(ParentCommandNode, CommandNodeName),
	m_Document(0),
	m_ShaderInstance(0)
{
//std::cout << "k3dShaderInfo::k3dShaderInfo()" << std::endl;

	// Create our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<vbox homogeneous=\"false\">"
				"<table columns=\"2\" rows=\"3\" homogeneous=\"false\">"
					"<label tableposition=\"0 1 0 1\" labelpadding=\"2\">Description:</label>"
					"<entry tableposition=\"1 2 0 1\" name=\"description\"/>"
					"<label tableposition=\"0 1 1 2\" labelpadding=\"2\">Author(s):</label>"
					"<entry tableposition=\"1 2 1 2\" name=\"authors\"/>"
					"<label tableposition=\"0 1 2 3\" labelpadding=\"2\">Copyright:</label>"
					"<entry tableposition=\"1 2 2 3\" name=\"copyright\"/>"
				"</table>"
			"</vbox>"
		"</gtkml>\n");

	// Load our template ...
	g_return_if_fail(Load(uitemplate, ""));
}

k3dShaderInfo::~k3dShaderInfo()
{
//std::cout << "k3dShaderInfo::~k3dShaderInfo()" << std::endl;

	if(m_ShaderInstance)
		m_ShaderInstance->RemoveEventObserver(this);
}

sdpResult k3dShaderInfo::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	sdpString event(Event);
	
	if(event == k3dIVariable::Event_Changed)
		{
			UpdateShader();
			return SDP_S_OK;
		}
	
	return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
}

/// Called by the framework when being instantiated from a GTKML (XML) document
bool k3dShaderInfo::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	
	g_assert(Element.Name() == "k3dshaderinfo");

	return true;
}

// UI creation ...
bool k3dShaderInfo::Attach(k3dIDocument* Document, k3dShaderInstance* ShaderInstance)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ShaderInstance);

	// You only get to Attach() me once!
	g_assert(!m_Document);
	g_assert(!m_ShaderInstance);

	// Store data ...
	m_Document = Document;
	m_ShaderInstance = ShaderInstance;

	// We want to know when the shader changes ...
	m_ShaderInstance->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dShaderInfo");

	// Show the currently selected shader ...
	UpdateShader();

	// Make everything visible ...
	RootWidget().Show();

	return true;
}

const sdpString k3dShaderInfo::CustomType() const
{
	return sdpString("k3dShaderInfo");
}

void k3dShaderInfo::UpdateShader()
{
	// Sanity checks ...
	g_return_if_fail(m_ShaderInstance);
	
	if(m_ShaderInstance->Shader())
		{
			Editable(control_authors).SetText(m_ShaderInstance->Shader()->Authors());
			Editable(control_copyright).SetText(m_ShaderInstance->Shader()->Copyright());
			Editable(control_description).SetText(m_ShaderInstance->Shader()->Description());
		}
	else
		{
			Editable(control_authors).SetText("");
			Editable(control_copyright).SetText("");
			Editable(control_description).SetText("");
		}
}



