// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderInstance class, which manages an instance of a Pixar RenderMan shader and its arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcom.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3drishadersetup.h>
#include <k3dshaderinstance.h>

#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3diritexture.h>
#include <k3ditexture.h>

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceArgument

k3dShaderInstanceArgument::k3dShaderInstanceArgument(sdpsl::IShaderArgument* const Argument) :
	m_Name(Argument->Name()),
	m_Type(Argument->Type()),
	m_ExtendedType(Argument->ExtendedType())
{
	// Sanity checks ...
	g_assert(m_Name.size());
}

k3dShaderInstanceArgument::k3dShaderInstanceArgument(sdpxml::Element& Element) :
	m_Name(sdpxml::GetAttribute<sdpString>(Element, "name", "")),
#ifndef	SDPWIN32
	m_Type(sdpFromString(sdpxml::GetAttribute<sdpString>(Element, "type", ""), sdpsl::IShaderArgument::FLOAT)),
	m_ExtendedType(sdpFromString(sdpxml::GetAttribute<sdpString>(Element, "extendedtype", ""), sdpsl::IShaderArgument::EX_FLOAT))
#else
	m_Type(sdpsl::IShaderArgument::FLOAT),
	m_ExtendedType(sdpsl::IShaderArgument::EX_FLOAT)
#endif // SDPWIN32
{
#ifdef	SDPWIN32
	const sdpString type = sdpxml::GetAttribute<sdpString>(Element, "type", "");
	const sdpString extendedtype = sdpxml::GetAttribute<sdpString>(Element, "extendedtype", "");
	m_Type = sdpFromString(type, sdpsl::IShaderArgument::FLOAT);
	m_ExtendedType = sdpFromString(extendedtype, sdpsl::IShaderArgument::EX_FLOAT);
#endif	//SDPWIN32

	// Sanity checks ...
	g_assert(m_Name.size());
}

k3dShaderInstanceArgument::~k3dShaderInstanceArgument()
{
}

// Partial sdpsl::IShaderArgument implementation
SDPCSTRING k3dShaderInstanceArgument::Name() const
{
	return m_Name;
}

sdpsl::IShaderArgument::SLType k3dShaderInstanceArgument::Type() const
{
	return m_Type;
}

sdpsl::IShaderArgument::SLExtendedType k3dShaderInstanceArgument::ExtendedType() const
{
	return m_ExtendedType;
}

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceStringArgument

k3dShaderInstanceStringArgument::k3dShaderInstanceStringArgument(sdpsl::IShaderArgument* const Argument) :
	k3dShaderInstanceArgument(Argument),
	m_Value(Argument->Name(), Argument->Value(), 0, 0)
{
}

k3dShaderInstanceStringArgument::k3dShaderInstanceStringArgument(sdpxml::Element& Element) :
	k3dShaderInstanceArgument(Element),
	m_Value(sdpxml::GetAttribute<sdpString>(Element, "name", ""), sdpxml::GetAttribute<sdpString>(Element, "value", ""), 0, 0)
{
	// Sanity checks ...
	const sdpString typestring = sdpxml::GetAttribute<sdpString>(Element, "type", "");
	g_assert(typestring.size());
	
	const sdpsl::IShaderArgument::SLType type = sdpFromString(typestring, sdpsl::IShaderArgument::STRING);
	
	g_assert(type == sdpsl::IShaderArgument::STRING);
}

k3dVariable<sdpString>& k3dShaderInstanceStringArgument::Value()
{
	return m_Value;
}

void k3dShaderInstanceStringArgument::InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader)
{
	Shader.AddTokenValuePair(new k3dRiTokenValuePair(Name(), new k3dRiStringArray("string", m_Value.Value())));
}
	
sdpResult k3dShaderInstanceStringArgument::Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);
	
	Element.Append(sdpxml::Element("argument", "",
		sdpxml::Attribute("name", Name()),
		sdpxml::Attribute("type", sdpToString(Type())),
		sdpxml::Attribute("extendedtype", sdpToString(ExtendedType())),
		sdpxml::Attribute("value", m_Value.Value())));
	
	return SDP_S_OK;
}

sdpResult k3dShaderInstanceStringArgument::LoadComplete(k3dIDocument* const Document)
{
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceChannelArgument

k3dShaderInstanceChannelArgument::k3dShaderInstanceChannelArgument(k3dIDocument* const Document, sdpsl::IShaderArgument* const Argument) :
	k3dShaderInstanceArgument(Argument),
	m_Value("value", Document, Document->StateRecorder(), 0)
{
	// Create the channel object ...
	g_return_if_fail(m_Value.Create(k3d::classid::Channel, Argument->Name(), false, false, true));

	k3d::Editing editing(m_Value.Interface());			

	// Setup the channel datatype and value, based on our argument type ...
	switch(Argument->Type())
		{
			case sdpsl::IShaderArgument::FLOAT:
				{
					m_Value->SetDatatype(k3dIChannel::DATA_SCALAR);
					k3dChannelSetConstantValue(m_Value.Interface(), sdpFromString<gdouble>(Argument->Value(), 0.0));
					
					return;
				}
			case sdpsl::IShaderArgument::COLOR:
				{
					m_Value->SetDatatype(k3dIChannel::DATA_COLOR);
					k3dChannelSetConstantValue(m_Value.Interface(), sdpFromString<sdpVector3>(Argument->Value(), sdpVector3()));
					
					return;
				}
			case sdpsl::IShaderArgument::POINT:
			case sdpsl::IShaderArgument::VECTOR:
			case sdpsl::IShaderArgument::NORMAL:
				{
					std::cerr << g_pretty_function() << ": Unsupported argument type [" << Argument->Type() << "] for [" << Argument->Name() << "]" << std::endl;
					return;
				}
			case sdpsl::IShaderArgument::MATRIX:
			case sdpsl::IShaderArgument::STRING:
			default:
				g_return_if_fail(0);
		}
}

k3dShaderInstanceChannelArgument::k3dShaderInstanceChannelArgument(k3dIDocument* const Document, sdpxml::Element& Element) :
	k3dShaderInstanceArgument(Element),
	m_Value("value", Document, Document->StateRecorder(), 0)
{
	// Sanity checks ...
	const sdpString typestring = sdpxml::GetAttribute<sdpString>(Element, "type", "");
	g_assert(typestring.size());
	const sdpsl::IShaderArgument::SLType type = sdpFromString(typestring, sdpsl::IShaderArgument::FLOAT);
	g_assert(type == sdpsl::IShaderArgument::FLOAT || type == sdpsl::IShaderArgument::COLOR);

	sdpxml::ParseAttribute(Element, "value", m_ValueID);
}

k3dIChannel* k3dShaderInstanceChannelArgument::Channel()
{
	return m_Value.Interface();
}

void k3dShaderInstanceChannelArgument::InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader)
{
	switch(Type())
		{
			case sdpsl::IShaderArgument::FLOAT:
				{
					const gdouble value = m_Value.Interface()->Value1(Time, DeltaTime);
					Shader.AddTokenValuePair(new k3dRiTokenValuePair(Name(), new k3dRiRealArray("float", value)));
					break;
				}
			case sdpsl::IShaderArgument::COLOR:
				{
					sdpVector3 value;
					m_Value.Interface()->Value3(Time, DeltaTime, value);
					Shader.AddTokenValuePair(new k3dRiTokenValuePair(Name(), new k3dRiRealArray("color", value)));
					break;
				}
			default:
				g_return_if_fail(0);
		}
}
	
sdpResult k3dShaderInstanceChannelArgument::Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies)
{
	// Sanity checks ...
	
	g_assert(Dependencies);
	
	if(m_Value.Object())
		Element.Append(sdpxml::Element("argument", "",
			sdpxml::Attribute("name", Name()),
			sdpxml::Attribute("type", sdpToString(Type())),
			sdpxml::Attribute("extendedtype", sdpToString(ExtendedType())),
			sdpxml::Attribute("value", sdpToString(m_Value.Object()->ObjectID()))));
	else
		Element.Append(sdpxml::Element("argument", "",
			sdpxml::Attribute("name", Name()),
			sdpxml::Attribute("type", sdpToString(Type())),
			sdpxml::Attribute("extendedtype", sdpToString(ExtendedType()))));
	
	return SDP_S_OK;
}

sdpResult k3dShaderInstanceChannelArgument::LoadComplete(k3dIDocument* const Document)
{
	// Get our object by ID ...
	std::vector<k3dIObject*> objects;
	k3d::copy_if(Document->Objects(), std::back_inserter(objects), k3d::SameID(m_ValueID));
	g_return_val_if_fail(1 == objects.size(), SDP_E_FAIL);

	// Attach the channel object ...
	g_return_val_if_fail(sdpSucceeded(m_Value.SetObject(*objects.begin())), SDP_E_FAIL);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceTextureArgument

k3dShaderInstanceTextureArgument::k3dShaderInstanceTextureArgument(k3dIDocument* const Document, sdpsl::IShaderArgument* const Argument) :
	k3dShaderInstanceArgument(Argument),
	m_Value("value", Document, Document->StateRecorder(), 0)
{
}

k3dShaderInstanceTextureArgument::k3dShaderInstanceTextureArgument(k3dIDocument* const Document, sdpxml::Element& Element) :
	k3dShaderInstanceArgument(Element),
	m_Value("value", Document, Document->StateRecorder(), 0)
{
	// Sanity checks ...
	const sdpString typestring = sdpxml::GetAttribute<sdpString>(Element, "type", "");
	g_assert(typestring.size());
	
	const sdpString extendedtypestring = sdpxml::GetAttribute<sdpString>(Element, "extendedtype", "");
	g_assert(extendedtypestring.size());
	
	const sdpsl::IShaderArgument::SLType type = sdpFromString(typestring, sdpsl::IShaderArgument::STRING);
	const sdpsl::IShaderArgument::SLExtendedType extendedtype = sdpFromString(extendedtypestring, sdpsl::IShaderArgument::EX_STRING);
	
	g_assert(type == sdpsl::IShaderArgument::STRING && extendedtype == sdpsl::IShaderArgument::EX_TEXTURE);

	sdpxml::ParseAttribute(Element, "value", m_ValueID);
}

k3dObjectInterfaceVariable<k3dITexture>& k3dShaderInstanceTextureArgument::TextureContainer()
{
	return m_Value;
}

void k3dShaderInstanceTextureArgument::InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader)
{
	if(m_Value.Interface())
		{
			k3dIRiTexture* const texture = dynamic_cast<k3dIRiTexture*>(m_Value.Interface());
			if(texture)
				{
					Shader.AddTokenValuePair(new k3dRiTokenValuePair(Name(), new k3dRiStringArray("string", texture->RiTexturePath(Frame, Engine))));
					return;
				}
		}
		
	Shader.AddTokenValuePair(new k3dRiTokenValuePair(Name(), new k3dRiStringArray("string", "")));
}
	
sdpResult k3dShaderInstanceTextureArgument::Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);
	
	if(m_Value.Object())
		Element.Append(sdpxml::Element("argument", "",
			sdpxml::Attribute("name", Name()),
			sdpxml::Attribute("type", sdpToString(Type())),
			sdpxml::Attribute("extendedtype", sdpToString(ExtendedType())),
			sdpxml::Attribute("value", sdpToString(m_Value.Object()->ObjectID()))));
	else
		Element.Append(sdpxml::Element("argument", "",
			sdpxml::Attribute("name", Name()),
			sdpxml::Attribute("type", sdpToString(Type())),
			sdpxml::Attribute("extendedtype", sdpToString(ExtendedType()))));
	
	return SDP_S_OK;
}

sdpResult k3dShaderInstanceTextureArgument::LoadComplete(k3dIDocument* const Document)
{
	// If we don't have an object ID (it could happen - user didn't assign a texture), we're done ...
	if(0 == m_ValueID)
		return SDP_S_OK;

	// Get our object by ID ...
	std::vector<k3dIObject*> objects;
	k3d::copy_if(Document->Objects(), std::back_inserter(objects), k3d::SameID(m_ValueID));
	g_return_val_if_fail(1 == objects.size(), SDP_E_FAIL);

	// Attach the channel object ...
	g_return_val_if_fail(sdpSucceeded(m_Value.SetObject(*objects.begin())), SDP_E_FAIL);

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////
// k3dShaderInstance

k3dShaderInstance::k3dShaderInstance(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder) :
	k3dVariableCommon(Name, StateRecorder),
	m_Document(Document),
	m_Shader(0),
	m_PrimaryTexture(0)
{
	// Sanity checks ...
	g_assert(m_Document);
}

k3dShaderInstance::~k3dShaderInstance()
{
	Clear();
}

bool k3dShaderInstance::SetShader(SDPCSTRING ShaderName)
{
	// Sanity checks ...
	g_assert_string(ShaderName);

	// Get a list of available shaders	
	k3dShaderSet shaders;
	k3dGetShaders(k3dApplication()->Shaders(), shaders);

	// Test 'em against the requested shader name ...
	const sdpString shadername(ShaderName);
	for(k3dShaderSet::iterator shader = shaders.begin(); shader != shaders.end(); shader++)
		if((*shader)->Name() == shadername)
			return SetShader(*shader);
			
	// Couldn't find a shader that matched ...
	std::cerr << "k3dShaderInstance::SetShader(): No shader named \"" << ShaderName << "\" available!" << std::endl;
	return false;
}

bool k3dShaderInstance::SetShader(sdpsl::IShader* const Shader)
{
	// If it's the same one we've already got, we're done ...
	if(Shader == m_Shader)
		return true;

	// Get rid of old data ...
	Clear();

	// Set the shader info (we allow nulls) ...
	m_Shader = Shader;

	// Allocate new data ...
	g_return_val_if_fail(Allocate(), false);

	// Let interested parties know that we've changed ...
	Notify(Event_Changed, 0);

	return true;
}

sdpsl::IShader* k3dShaderInstance::Shader() const
{
	return m_Shader;
}

k3dITexture* k3dShaderInstance::PrimaryTexture() const
{
	return m_PrimaryTexture ? static_cast<k3dShaderInstanceTextureArgument*>(m_PrimaryTexture)->TextureContainer().Interface() : 0;
}

void k3dShaderInstance::SetPrimaryTexture(k3dShaderInstanceArgument* const Argument)
{
	// Sanity checks ...
	g_return_if_fail(sdpsl::IShaderArgument::EX_TEXTURE == Argument->ExtendedType());
	g_return_if_fail(std::find(m_Arguments.begin(), m_Arguments.end(), Argument) != m_Arguments.end());

	m_PrimaryTexture = Argument;
}

void k3dShaderInstance::InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader)
{
	// Sanity checks ...
	g_assert(Frame);
	g_assert(Engine);

	// Insert the shader name ...
	Shader.SetShader(m_Shader ? m_Shader->Name() : "");

	// Insert the shader arguments ...
	for(ArgumentIterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		(*argument)->InsertInto(Frame, Engine, Time, DeltaTime, Shader);
}

sdpResult k3dShaderInstance::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const
{
	// Sanity checks ...
	g_assert(Dependencies);

	sdpxml::Element& shader = Element.Append(sdpxml::Element("variable", "",
		sdpxml::Attribute("name", m_Name),
		sdpxml::Attribute("shadername", m_Shader ? m_Shader->Name() : "")));
	
	for(Arguments::const_iterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		(*argument)->Save(shader, Dependencies);
	
	return SDP_S_OK;
}

class k3dShaderInstanceLoadArguments
{
public:
	k3dShaderInstanceLoadArguments(k3dIDocument* const Document, k3dShaderInstance::Arguments& ShaderArguments) :
		m_Document(Document),
		m_Arguments(ShaderArguments)
	{
		// Sanity checks ...
		g_assert(m_Document);
	}
	
	void operator()(sdpxml::Element& Element)
	{
		// Get the argument name ...
		const sdpString name = sdpxml::GetAttribute<sdpString>(Element, "name", "");
		g_return_if_fail(name.size());
	
		// Get the argument type and extended type ...
		const sdpString type = sdpxml::GetAttribute<sdpString>(Element, "type", "");
		g_return_if_fail(type.size());
		
		const sdpString extendedtype = sdpxml::GetAttribute<sdpString>(Element, "extendedtype", "");
//		g_return_if_fail(extendedtype.size());

		// Create our own argument ...
		switch(sdpFromString(type, sdpsl::IShaderArgument::FLOAT))
			{
				case sdpsl::IShaderArgument::POINT:
				case sdpsl::IShaderArgument::VECTOR:
				case sdpsl::IShaderArgument::NORMAL:
				case sdpsl::IShaderArgument::MATRIX:
					std::cerr << "k3dShaderInstanceLoadArguments::HandleElement(): Unsupported extended shader argument type [" << type << "] for [" << name << "]" << std::endl;
					break;
				case sdpsl::IShaderArgument::FLOAT:
				case sdpsl::IShaderArgument::COLOR:
					m_Arguments.push_back(new k3dShaderInstanceChannelArgument(m_Document, Element));
					break;
				case sdpsl::IShaderArgument::STRING:
					if(sdpFromString(extendedtype, sdpsl::IShaderArgument::EX_STRING) == sdpsl::IShaderArgument::EX_TEXTURE)
						m_Arguments.push_back(new k3dShaderInstanceTextureArgument(m_Document, Element));
					else
						m_Arguments.push_back(new k3dShaderInstanceStringArgument(Element));
					break;
				default:
					std::cerr << "k3dShaderInstanceLoadArguments::HandleElement(): Unknown shader argument type [" << extendedtype << "] for [" << name << "]" << std::endl;
					break;
			}
	}
	
protected:
	k3dIDocument* const m_Document;
	k3dShaderInstance::Arguments& m_Arguments;
};

sdpResult k3dShaderInstance::Load(sdpxml::Element& Element)
{
	// <shader> is obsolete, and should be disallowed by 0.4 ...
	g_return_val_if_fail((Element.Name() == "variable") || (Element.Name() == "shader"), SDP_E_FAIL);

	// Pull out the name ...
	sdpString name;
	g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), SDP_E_FAIL);
	g_return_val_if_fail(name == m_Name, SDP_E_FAIL);

	// Load the shader name ...
	sdpString shadername = sdpxml::GetAttribute<sdpString>(Element, "shadername", "");
	
	// We allow empty shader names, so exit quietly if we didn't get one ...
	if(0 == shadername.size())
		return SDP_S_OK;

	// Get the set of available shaders ...
	k3dShaderSet shaders;
	k3dGetShaders(k3dApplication()->Shaders(), shaders);

	for(k3dShaderSet::iterator shader = shaders.begin(); shader != shaders.end(); shader++)
		if((*shader)->Name() == shadername)
			{
				m_Shader = (*shader);
				break;
			}

	// No info, so we're done ...
	if(0 == m_Shader)
		{
			std::cerr << "k3dShaderInstance::Load() - shader [" << shadername << "] not available!" << std::endl;
			return SDP_S_FALSE;
		}

	// Load the shader arguments ...
	k3dShaderInstanceLoadArguments functor(m_Document, m_Arguments);
	sdpxml::ForEach(Element, sdpxml::SameName("argument"), functor);

	return SDP_S_OK;
}

sdpResult k3dShaderInstance::LoadComplete()
{
	for(ArgumentIterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		(*argument)->LoadComplete(m_Document);
	
	SetDefaultPrimaryTexture();
	
	return SDP_S_OK;
}

bool k3dShaderInstance::Allocate()
{
	// If we don't have a shader, we're done ...
	if(0 == m_Shader)
		return true;

	// Get all our shader arguments as a collection ...
	k3dShaderArgumentSet arguments;
	k3dGetShaderArguments(m_Shader, arguments);

	// For each shader argument ...
	for(k3dShaderArgumentSet::iterator argument = arguments.begin(); argument != arguments.end(); argument++)
		{
			// Create our own argument ...
			switch((*argument)->Type())
				{
					case sdpsl::IShaderArgument::POINT:
					case sdpsl::IShaderArgument::VECTOR:
					case sdpsl::IShaderArgument::NORMAL:
					case sdpsl::IShaderArgument::MATRIX:
						std::cerr << "k3dShaderInstance::Allocate(): Unsupported shader argument type [" << (*argument)->Type() << "] for [" << (*argument)->Name() << "]" << std::endl;
						break;
					case sdpsl::IShaderArgument::FLOAT:
					case sdpsl::IShaderArgument::COLOR:
						m_Arguments.push_back(new k3dShaderInstanceChannelArgument(m_Document, *argument));
						break;
					case sdpsl::IShaderArgument::STRING:
						if(sdpsl::IShaderArgument::EX_TEXTURE == (*argument)->ExtendedType())
							m_Arguments.push_back(new k3dShaderInstanceTextureArgument(m_Document, *argument));
						else
							m_Arguments.push_back(new k3dShaderInstanceStringArgument(*argument));
						break;
					default:
						std::cerr << "k3dShaderInstance::Allocate(): Unknown shader argument type [" << (*argument)->Type() << "] for [" << (*argument)->Name() << "]" << std::endl;
						break;
				}
		}
	
	SetDefaultPrimaryTexture();
	
	return true;
}

void k3dShaderInstance::Clear()
{
	m_Shader = 0;
	m_PrimaryTexture = 0;

	std::for_each(m_Arguments.begin(), m_Arguments.end(), sdp::DeleteObject());
	m_Arguments.clear();
}

void k3dShaderInstance::SetDefaultPrimaryTexture()
{
	for(ArgumentIterator argument = m_Arguments.begin(); argument != m_Arguments.end(); argument++)
		if(sdpsl::IShaderArgument::EX_TEXTURE == (*argument)->ExtendedType())
			{
				m_PrimaryTexture = (*argument);
				break;
			}
}

