#ifndef K3DSHADERINSTANCE_H
#define K3DSHADERINSTANCE_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dShaderInstance class, which manages an instance of a Pixar RenderMan shader and its arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deventsource.h>
#include <k3dvariable.h>

#include <sdpsl.h>

// Forward class declarations
class k3dRiShader;

class k3dIChannel;
class k3dIDependencies;
class k3dIDocument;
class k3dIRenderFrame;
class k3dIRiEngine;
class sdpsl::IShader;
class k3dITexture;

class sdpxml::Element;

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceArgument

class k3dShaderInstanceArgument
{
public:
	k3dShaderInstanceArgument(sdpsl::IShaderArgument* const Argument);
	k3dShaderInstanceArgument(sdpxml::Element& Element);
	virtual ~k3dShaderInstanceArgument();

	// Partial sdpsl::IShaderArgument implementation
	SDPCSTRING Name() const;
	sdpsl::IShaderArgument::SLType Type() const;
	sdpsl::IShaderArgument::SLExtendedType ExtendedType() const;
	
	/// Initializes a parameter list with our current state
	virtual void InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader) = 0;

	/// Serializes the argument data
	virtual sdpResult Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies) = 0;
	/// Initializes the argument data following a load
	virtual sdpResult LoadComplete(k3dIDocument* const Document) = 0;
									
protected:
	const sdpString m_Name;
#ifndef	SDPWIN32
	const sdpsl::IShaderArgument::SLType m_Type;
	const sdpsl::IShaderArgument::SLExtendedType m_ExtendedType;
#else	// SDPWIN32
		  sdpsl::IShaderArgument::SLType m_Type;
		  sdpsl::IShaderArgument::SLExtendedType m_ExtendedType;
#endif	// SDPWIN32
};

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceStringArgument

class k3dShaderInstanceStringArgument :
	public k3dShaderInstanceArgument
{
public:
	k3dShaderInstanceStringArgument(sdpsl::IShaderArgument* const Argument);
	k3dShaderInstanceStringArgument(sdpxml::Element& Element);

	/// Returns the argument value	
	k3dVariable<sdpString>& Value();

	void InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader);
	
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies);
	sdpResult LoadComplete(k3dIDocument* const Document);

protected:
	k3dVariable<sdpString> m_Value;
};

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceChannelArgument

class k3dShaderInstanceChannelArgument :
	public k3dShaderInstanceArgument
{
public:
	k3dShaderInstanceChannelArgument(k3dIDocument* const Document, sdpsl::IShaderArgument* const Argument);
	k3dShaderInstanceChannelArgument(k3dIDocument* const Document, sdpxml::Element& Element);
	
	/// Returns the k3dIChannel value of this argument
	k3dIChannel* Channel();

	void InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader);
	
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies);
	sdpResult LoadComplete(k3dIDocument* const Document);

protected:
	/// Stores the channel for this argument
	k3dChannelVariable m_Value;
	/// Temporarily stores the channel object ID at load-time
	k3dIObject::ID m_ValueID;
};

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstanceTextureArgument

class k3dShaderInstanceTextureArgument :
	public k3dShaderInstanceArgument
{
public:
	k3dShaderInstanceTextureArgument(k3dIDocument* const Document, sdpsl::IShaderArgument* const Argument);
	k3dShaderInstanceTextureArgument(k3dIDocument* const Document, sdpxml::Element& Element);

	/// Returns the k3dITexture value	of this argument
	k3dObjectInterfaceVariable<k3dITexture>& TextureContainer();

	void InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader);
	
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies);
	sdpResult LoadComplete(k3dIDocument* const Document);

protected:
	/// Stores the texture for this argument
	k3dObjectInterfaceVariable<k3dITexture> m_Value;
	/// Temporarily stores the texture object ID at load-time
	k3dIObject::ID m_ValueID;
};

///////////////////////////////////////////////////////////////////////////
// k3dShaderInstance

/// Manages an instance of a shader and its arguments
class k3dShaderInstance :
	public k3dVariableCommon
{
public:
	k3dShaderInstance(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder);
	~k3dShaderInstance();

	/// Sets the current shader by name (releases old data, channels, etc.)
	bool SetShader(SDPCSTRING ShaderName);
	/// Sets the current shader (releases old data, channels, etc.)
	bool SetShader(sdpsl::IShader* const Shader);
	/// Returns the current shader
	sdpsl::IShader* Shader() const ;
	/// Returns the primary (for Editor & OpenGL rendering purposes) texture argument
	k3dITexture* PrimaryTexture() const;

	/// Sets the primary texture
	void SetPrimaryTexture(k3dShaderInstanceArgument* const Argument);

	/// Called to initialize a parameter list with our current state
	void InsertInto(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine, const gdouble Time, const gdouble DeltaTime, k3dRiShader& Shader);

	// k3dIVariable implementation ...
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const;
	sdpResult Load(sdpxml::Element& Element);
	sdpResult LoadComplete();

protected:
friend class k3dShaderArgumentsPage;
friend class k3dShaderInstanceLoadArguments;

	/// Allocates data based on the selected shader
	bool Allocate();
	/// Releases all data
	void Clear();

	/// Determines a primary (for Editor & OpenGL rendering purposes) texture argument
	void SetDefaultPrimaryTexture();

	/// Stores a reference to the owning document
	k3dIDocument* const m_Document;
	/// Stores the current shader info
	sdpsl::IShader* m_Shader;
	/// Store the primary texture (for use with Editor & OpenGL rendering) if any ...
	k3dShaderInstanceArgument* m_PrimaryTexture;
	/// Stores a set of shader arguments
	typedef std::vector<k3dShaderInstanceArgument*> Arguments;
	typedef Arguments::iterator ArgumentIterator;
	/// Stores the individual arguments ...
	Arguments m_Arguments;
};

#endif // K3DSHADERINSTANCE_H

