// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderSelector class, which provides a user interface for choosing Pixar RenderMan shaders
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dshaderinstance.h>
#include <k3dshaderselector.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3ditexture.h>
#include <k3diobjectcollection.h>

#include <sdpgtkevents.h>

const sdpString k3dShaderSelector::control_toggleselection = "toggleselection";
const sdpString k3dShaderSelector::control_currentselection = "currentselection";
const sdpString k3dShaderSelector::control_arrow = "arrow";
const sdpString k3dShaderSelector::control_shaders = "shaders";
const sdpString k3dShaderSelector::control_select = "select";

k3dShaderSelector::k3dShaderSelector(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName) :
	k3dShaderSelectorBase(ParentCommandNode, CommandNodeName),
	m_Document(0),
	m_ShaderInstance(0)
{
//std::cout << "k3dShaderSelector::k3dShaderSelector()" << std::endl;

	// Create our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<vbox homogeneous=\"false\">"
				"<button name=\"toggleselection\">"
					"<event signal=\"clicked\" name=\"toggleselection\"/>"
					"<hbox homogeneous=\"false\">"
						"<label expand=\"true\" fill=\"true\" name=\"currentselection\" labelpadding=\"2\">Select Shader:</label>"
						"<arrow name=\"arrow\" direction=\"down\"/>"
					"</hbox>"
				"</button>"
				"<scrolledwindow expand=\"true\" fill=\"true\" show=\"false\" name=\"shaders\" hscrollbar=\"never\">"
					"<clist columns=\"1\" name=\"shaderlist\" charactersize=\"2 10\">"
						"<event signal=\"button-press-event\" name=\"select\"/>"
					"</clist>"
				"</scrolledwindow>"
			"</vbox>"
		"</gtkml>\n");

	// Load our template ...
	g_return_if_fail(Load(uitemplate, ""));
}

k3dShaderSelector::~k3dShaderSelector()
{
//std::cout << "k3dShaderSelector::~k3dShaderSelector()" << std::endl;

	if(m_ShaderInstance)
		m_ShaderInstance->RemoveEventObserver(this);
}

sdpResult k3dShaderSelector::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	if(event == k3dIVariable::Event_Changed)
		{
			UpdateShader();
			return SDP_S_OK;
		}

	return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
}

sdpResult k3dShaderSelector::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	sdpString name(Name);
	sdpString arguments(Arguments);

	// We need to select a new shader ...
	if(name == control_select)
		{
			// Make the shader list visible ...
			if(0 == (Widget(control_shaders).Object()->flags & GTK_VISIBLE))
				InteractiveActivateButton(control_toggleselection, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);

			// Start with the "none" shader ...
			sdpsl::IShader* shader = 0;

			// If a shader name was supplied, look for it ...
			if(arguments.size())
				{
					// Get the collection of shader information ...
					k3dShaderSet shaders;
					k3dGetShaders(k3dApplication()->Shaders(), shaders);

					for(k3dShaderSet::iterator shaderiterator = shaders.begin(); shaderiterator != shaders.end(); shaderiterator++)
						{
							if((*shaderiterator)->Name() == arguments)
								{
									shader = (*shaderiterator);
									break;
								}
						}

					// Couldn't find a shader by that name ...
					g_return_val_if_fail(shader, SDP_E_FAIL);
				}

			// Look for a row that matches the shader ...
			sdpGtkCList list = CList("shaderlist");
			const gint row = list.FindRowFromData(shader);
			g_return_val_if_fail(-1 != row, SDP_E_FAIL);

			// Move the pointer to the row ...
			list.InteractiveShow(k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
			list.InteractiveWarpPointer(row, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
			list.SelectRow(row, 0);
			OnSelect(row);

			return SDP_S_OK;
		}

	return k3dShaderSelectorBase::Command(Name, Arguments, Result);
}

/// Called by the framework when being instantiated from a GTKML (XML) document
bool k3dShaderSelector::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);

	g_assert(Element.Name() == "k3dshaderselector");

	return true;
}

// UI creation ...
bool k3dShaderSelector::Attach(k3dIDocument* Document, k3dShaderInstance* ShaderInstance, bool AllowNone, sdpsl::IShader::SLType Type)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ShaderInstance);

	// You only get to Attach() me once!
	g_assert(!m_Document);
	g_assert(!m_ShaderInstance);

	// Store data ...
	m_Document = Document;
	m_ShaderInstance = ShaderInstance;

	// We want to know when the shader changes ...
	m_ShaderInstance->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dShaderSelector");

	// Get the collection of available shaders ...
	k3dShaderSet shaders;
	k3dGetShaders(k3dApplication()->Shaders(), shaders);

	// Sort the shaders by name ...
	typedef std::multimap<sdpString, sdpsl::IShader*> SortedShaders;
	SortedShaders sortedshaders;
	for(k3dShaderSet::iterator shader = shaders.begin(); shader != shaders.end(); shader++)
		sortedshaders.insert(std::make_pair(sdpString((*shader)->Name()), (*shader)));

	// Get our list object ...
	sdpGtkCList list = CList("shaderlist");
	list.Freeze();
	list.Clear();

	// If we give the user the option of selecting nothing ...
	if(AllowNone)
		{
			// Create the "none" option ...
			gint row = list.Append("--None--");
			list.SetRowData(row, 0);
		}

	for(SortedShaders::iterator sortedshader = sortedshaders.begin(); sortedshader != sortedshaders.end(); sortedshader++)
		{
			// Skip shaders that don't match the requested type ...
			if(sortedshader->second->Type() != Type)
				continue;

			gint row = list.Append(sortedshader->second->Name());
			list.SetRowData(row, sortedshader->second);
		}

	list.Thaw();

	// Show the currently selected shader ...
	UpdateShader();

	// Make everything visible ...
	RootWidget().Show();

	return true;
}

const sdpString k3dShaderSelector::CustomType() const
{
	return sdpString("k3dShaderSelector");
}

// Custom event handling ...
void k3dShaderSelector::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_toggleselection)
		OnToggleSelection();
	else if(Event->Name() == control_select)
		OnSelect(Event);
	else
		k3dShaderSelectorBase::OnEvent(Event);
}

void k3dShaderSelector::OnToggleSelection()
{
	if(Widget(control_shaders).Object()->flags & GTK_VISIBLE)
		HideSelection();
	else
		ShowSelection();
}

void k3dShaderSelector::ShowSelection()
{
	if(Widget(control_shaders).Object()->flags & GTK_VISIBLE)
		return;

	Widget(control_shaders).Show();
	ScrolledWindow(control_shaders).HorizontalAdjustment().SetValue(0.0);
	ScrolledWindow(control_shaders).HorizontalAdjustment().SetLower(0.0);
	ScrolledWindow(control_shaders).HorizontalAdjustment().SetUpper(0.0);

	Arrow(control_arrow).Set(GTK_ARROW_UP, GTK_SHADOW_NONE);
}

void k3dShaderSelector::HideSelection()
{
	if(0 == (Widget(control_shaders).Object()->flags & GTK_VISIBLE))
		return;

	Widget(control_shaders).Hide();
	Arrow(control_arrow).Set(GTK_ARROW_DOWN, GTK_SHADOW_NONE);
}

void k3dShaderSelector::OnSelect(sdpGtkEvent* Event)
{
	sdpGtkCList list = CList("shaderlist");

	// Get the row we clicked on ...
	gint row, column;
	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;
	list.GetHitInfo(gint(event->Event()->x), gint(event->Event()->y), &row, &column);

	if(-1 == row)
		return;

	OnSelect(row);
}

void k3dShaderSelector::OnSelect(const gint Row)
{
	if(-1 == Row)
		return;

	sdpGtkCList list = CList("shaderlist");
	sdpsl::IShader* shader = reinterpret_cast<sdpsl::IShader*>(list.GetRowData(Row));
	const sdpString shadername = shader ? shader->Name() : "";

	k3dRecordCommand(this, USERINTERFACE, control_select, shadername);

	k3d::PushEditing(m_Document->Objects());
	SetShader(shader);
	k3d::PopEditing(m_Document->Objects());
	OnToggleSelection();
}

void k3dShaderSelector::SetShader(sdpsl::IShader* Shader)
{
	// Set the new shader handle ...
	m_ShaderInstance->SetShader(Shader);
	k3dRedrawAll(m_Document, false);
}

void k3dShaderSelector::UpdateShader()
{
	// Sanity checks ...
	g_return_if_fail(m_ShaderInstance);

	if(m_ShaderInstance->Shader())
		Label(control_currentselection).SetText(m_ShaderInstance->Shader()->Name());
	else
		Label(control_currentselection).SetText("--None--");
}



