// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dSpinButton class, which provides a UI for numeric quantities
		\author Tim Shead <tshead@k-3d.com>
		\author Dan Erikson <derikson@montana.com>
*/

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dspinbutton.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dicommandrecorder.h>
#include <k3distaterecorder.h>

#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

#include <iomanip>

///////////////////////////////////////////////////////////////////////////
// k3dVariableFormatDataType

k3dVariableFormatDataType::k3dVariableFormatDataType(const k3dIChannel::DataType DataType, const gulong Precision) :
	m_DataType(DataType),
	m_Precision(Precision)
{
}

k3dVariableFormatDataType::~k3dVariableFormatDataType()
{
}

bool k3dVariableFormatDataType::FormattedOutput(const gdouble& Value, sdpString& Output)
{
	Output = k3dConvertDoubleString(m_DataType, Value, m_Precision);
	return true;
}

bool k3dVariableFormatDataType::ParsedInput(const sdpString& Input, gdouble& InputValue)
{
	if(0 == Input.size())
		return true;
		
	if("+" == Input.substr(0, 1))
		InputValue += k3dConvertStringDouble(m_DataType, Input.right(Input.size()-1));
	else
		InputValue = k3dConvertStringDouble(m_DataType, Input);
	return true;
}

///////////////////////////////////////////////////////////////////////////
// k3dVariableFormatPrecision

k3dVariableFormatPrecision::k3dVariableFormatPrecision(const gulong Precision) :
	m_Precision(Precision)
{
}

k3dVariableFormatPrecision::~k3dVariableFormatPrecision()
{
}

bool k3dVariableFormatPrecision::FormattedOutput(const gdouble& Value, sdpString& Output)
{
	std::strstream buffer;
	buffer << std::setprecision(m_Precision) << Value;
	Output = buffer.str();
	buffer.freeze(false);

	return true;
}

bool k3dVariableFormatPrecision::ParsedInput(const sdpString& Input, gdouble& InputValue)
{
	InputValue = atof(Input);
	return true;
}

//////////////////////////////////////////////////////////////////////
// k3dSpinButton

const sdpString k3dSpinButton::control_value = "value";
const sdpString k3dSpinButton::control_up = "up";
const sdpString k3dSpinButton::control_down = "down";

k3dSpinButton::k3dSpinButton(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dSpinButtonBase(CommandNodeParent, CommandNodeName),
	m_DataSource(0),
	m_Format(0),
	m_StepIncrement(0.01),
	m_StartScrollingEvent(0),
	m_FasterScrollingEvent(0),
	m_FastestScrollingEvent(0),
	m_ScrollingEvent(0),
	m_ScrollMode(NONE)
{
	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<hbox homogeneous=\"false\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"

				"<entry expand=\"true\" fill=\"true\" name=\"value\">"
					"<event signal=\"button-press-event\" name=\"value_focus_in\" after=\"true\"/>"
					"<event signal=\"focus-in-event\" name=\"value_focus_in\" after=\"true\"/>"
					"<event signal=\"focus-out-event\" name=\"value_focus_out\"/>"
					"<event signal=\"activate\" name=\"value_activate\"/>"
				"</entry>"

				"<vbox homogeneous=\"true\">"
					"<button fill=\"true\" name=\"up\">"
						"<event signal=\"pressed\" name=\"up_pressed\"/>"
						"<event signal=\"released\" name=\"up_released\"/>"
						"<arrow direction=\"up\"/>"
					"</button>"
					"<button fill=\"true\" name=\"down\">"
						"<event signal=\"pressed\" name=\"down_pressed\"/>"
						"<event signal=\"released\" name=\"down_released\"/>"
						"<arrow direction=\"down\"/>"
					"</button>"
				"</vbox>"

			"</hbox>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, ""));

	// Make sure the spin buttons can't get the focus (makes tabbing difficult)
	Widget(control_up).Object()->flags &= ~GTK_CAN_FOCUS;
	Widget(control_down).Object()->flags &= ~GTK_CAN_FOCUS;

	RootWidget().Show();

//	AddCommand(control_value, control_value, CONTROL_EDITABLE);
}

k3dSpinButton::~k3dSpinButton()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();

	// Cancel any event notification from the data source
	if(m_DataSource)
		{
			k3dIEventSource* const datasource = m_DataSource->EventSource();
			if(datasource)
				datasource->RemoveEventObserver(this);
		}
						
	delete m_DataSource;
}

const sdpString k3dSpinButton::CustomType() const
{
	return sdpString("k3dSpinButton");
}

sdpResult k3dSpinButton::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	sdpString event(Event);
	if(event == k3dIVariable::Event_Changed)
		Update();
	else
		k3dSpinButtonBase::ObserveEvent(Source, Event, EventData, ObserverData);
	
	return SDP_S_OK;
}

sdpResult k3dSpinButton::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	if(Name == control_value)
		{
			g_return_val_if_fail(InteractiveSetText(control_value, Arguments, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true), SDP_E_FAIL);
			OnValueFocusOut();
			return SDP_S_OK;
		}

	return k3dSpinButtonBase::Command(Name, Arguments, Result);
}

bool k3dSpinButton::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	
	g_assert(Element.Name() == "k3dspinbutton");

	// Get the requested precision ...
	const gulong precision = sdpxml::GetAttribute<gulong>(Element, "precision", 0);
	sdpGtkMarkAttribute(Document, Element, "precision");

	// Get the requested data type ...
	const sdpString datatypetext = sdpxml::GetAttribute<sdpString>(Element, "datatype", "");
	sdpGtkMarkAttribute(Document, Element, "datatype");
	
	if(datatypetext.size())
		{
			k3dIChannel::DataType datatype = k3dIChannel::DATA_SCALAR;
		
			if(datatypetext == "scalar")
				datatype = k3dIChannel::DATA_SCALAR;
			else if(datatypetext == "time")
				datatype = k3dIChannel::DATA_TIME;
			else if(datatypetext == "angle")
				datatype = k3dIChannel::DATA_ANGLE;
			else if(datatypetext == "distance")
				datatype = k3dIChannel::DATA_DISTANCE;
			else if(datatypetext == "area")
				datatype = k3dIChannel::DATA_AREA;
			else if(datatypetext == "volume")
				datatype = k3dIChannel::DATA_VOLUME;
			else if(datatypetext == "mass")
				datatype = k3dIChannel::DATA_MASS;
			else if(datatypetext == "force")
				datatype = k3dIChannel::DATA_FORCE;
			else if(datatypetext == "pressure")
				datatype = k3dIChannel::DATA_PRESSURE;
			else if(datatypetext == "color")
				datatype = k3dIChannel::DATA_COLOR;
			else if(datatypetext == "rotation")
				datatype = k3dIChannel::DATA_ROTATIONXYZ;
			else if(datatypetext == "scale")
				datatype = k3dIChannel::DATA_SCALEXYZ;
			else
				g_return_val_if_fail(0, false);	// Unknown data type!

			m_Format = new k3dVariableFormatDataType(datatype, precision);
		}
	else
		{
			m_Format = new k3dVariableFormatPrecision(precision);
		}

	sdpGtkMarkAttribute(Document, Element, "stepincrement");
	sdpxml::ParseAttribute(Element, "stepincrement", m_StepIncrement);

	return true;
}

bool k3dSpinButton::Create(k3dIChannel::DataType Datatype, const gulong Precision, const gdouble Increment)
{
	m_Format = new k3dVariableFormatDataType(Datatype, Precision);
	m_StepIncrement = Increment;
	
	return true;
}

bool k3dSpinButton::Attach(k3dSpinButtonDataSource* const DataSource, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	// Sanity checks ...
	g_return_val_if_fail(DataSource, false);

	g_return_val_if_fail(k3dSpinButtonBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);
	
	// Store references ...
	m_DataSource = DataSource;
	
	// Update the display ...
	Update();
	
	// We want to be notified if the data source changes ...
	k3dIEventSource* datasource = m_DataSource->EventSource();
	g_return_val_if_fail(datasource, false);
	datasource->AddEventObserver(this, k3dIVariable::Event_Changed, "k3dSpinButton");

	return true;
}

bool k3dSpinButton::Attach(k3dVariable<gdouble>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	return Attach(new k3dSpinButtonVariableDataSource(Variable), CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction);
}

void k3dSpinButton::Update()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);
	g_return_if_fail(m_Format);
	
	// Format the current datasource value ...
	sdpString outputtext;
	g_return_if_fail(m_Format->FormattedOutput(m_DataSource->Value(), outputtext));

	// Put it into our edit control ...
	Editable(control_value).SetText(outputtext);
}

void k3dSpinButton::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "scrolling")
		OnScrolling();
	else if(Event->Name() == "start_scrolling")
		OnStartScrolling();
	else if(Event->Name() == "faster_scrolling")
		OnFasterScrolling();
	else if(Event->Name() == "fastest_scrolling")
		OnFastestScrolling();
	else if(Event->Name() == "value_focus_in")
		OnValueFocusIn();
	else if(Event->Name() == "value_focus_out")
		OnValueFocusOut();
	else if(Event->Name() == "value_activate")
		OnValueActivate();
	else if(Event->Name() == "up_pressed")
		OnUpPressed();
	else if(Event->Name() == "up_released")
		OnUpReleased();
	else if(Event->Name() == "down_pressed")
		OnDownPressed();
	else if(Event->Name() == "down_released")
		OnDownReleased();
	else if(Event->Name() == "destroy")
		OnDestroy();
	else
		k3dSpinButtonBase::OnEvent(Event);
}

void k3dSpinButton::OnDestroy()
{
	DisconnectAllEvents();
	Clear();
}

void k3dSpinButton::OnValueFocusIn()
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Make sure everything's selected ...
	Editable(control_value).SelectRegion(0, -1);
}

void k3dSpinButton::OnValueFocusOut()
{
	OnValueActivate();
}

void k3dSpinButton::OnValueActivate()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);
	
	// Get rid of the selection ...
	Editable(control_value).SelectRegion(0, 0);

	// If the value has changed, record it ...
	const gdouble newvalue = EditControlValue();
	if(newvalue != m_DataSource->Value())
		{
			PushEditing();
			SetValue(newvalue);
			PopEditing();
		}
}

void k3dSpinButton::StartScrolling(const ScrollMode Mode)
{
	m_ScrollMode = Mode;

	m_StartScrollingEvent = static_cast<sdpGtkEventTimeout*>(MapEvent("timeout", "start_scrolling", false, RootWidget(), false));
	m_StartScrollingEvent->SetDuration(200);

	m_FasterScrollingEvent = static_cast<sdpGtkEventTimeout*>(MapEvent("timeout", "faster_scrolling", false, RootWidget(), false));
	m_FasterScrollingEvent->SetDuration(400);

	m_FastestScrollingEvent = static_cast<sdpGtkEventTimeout*>(MapEvent("timeout", "fastest_scrolling", false, RootWidget(), false));
	m_FastestScrollingEvent->SetDuration(600);

	m_StartScrollingEvent->Connect();
	m_FasterScrollingEvent->Connect();
	m_FastestScrollingEvent->Connect();
}

void k3dSpinButton::CancelScrolling()
{
	m_ScrollMode = NONE;
	
	if(m_StartScrollingEvent)
		{
			m_StartScrollingEvent->Disconnect();
			DeleteEvent(m_StartScrollingEvent);
			m_StartScrollingEvent = 0;
		}

	if(m_FasterScrollingEvent)
		{
			m_FasterScrollingEvent->Disconnect();
			DeleteEvent(m_FasterScrollingEvent);
			m_FasterScrollingEvent = 0;
		}

	if(m_FastestScrollingEvent)
		{
			m_FastestScrollingEvent->Disconnect();
			DeleteEvent(m_FastestScrollingEvent);
			m_FastestScrollingEvent = 0;
		}

	if(m_ScrollingEvent)
		{
			m_ScrollingEvent->Disconnect();
			DeleteEvent(m_ScrollingEvent);
			m_ScrollingEvent = 0;
		}
}

void k3dSpinButton::OnStartScrolling()
{
	if(m_StartScrollingEvent)
		{
			m_StartScrollingEvent->Disconnect();
			DeleteEvent(m_StartScrollingEvent);
			m_StartScrollingEvent = 0;
		}

	m_ScrollingEvent = static_cast<sdpGtkEventTimeout*>(MapEvent("timeout", "scrolling", false, RootWidget(), false));
	m_ScrollingEvent->SetDuration(60);
	m_ScrollingEvent->Connect();
}

void k3dSpinButton::OnFasterScrolling()
{
	if(m_FasterScrollingEvent)
		{
			m_FasterScrollingEvent->Disconnect();
			DeleteEvent(m_FasterScrollingEvent);
			m_FasterScrollingEvent = 0;
		}
		
	if(m_ScrollingEvent)
		m_ScrollingEvent->SetDuration(35);
}

void k3dSpinButton::OnFastestScrolling()
{
	if(m_FastestScrollingEvent)
		{
			m_FastestScrollingEvent->Disconnect();
			DeleteEvent(m_FastestScrollingEvent);
			m_FastestScrollingEvent = 0;
		}
		
	if(m_ScrollingEvent)
		m_ScrollingEvent->SetDuration(10);
}

void k3dSpinButton::OnScrolling()
{
	switch(m_ScrollMode)
		{
			case NONE:
				CancelScrolling();
				break;
			case INCREMENT:
				Increment();
				break;
			case DECREMENT:
				Decrement();
				break;
			default:
				g_return_if_fail(0);
		}
}

void k3dSpinButton::OnUpPressed()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Let everyone know we're going to be making changes ...
	PushEditing();
	
	// Make sure any manual edits get recorded ...
	const gdouble newvalue = EditControlValue();
	if(newvalue != m_DataSource->Value())
		SetValue(newvalue);

	// Get the current value ...
	m_ScrollStartValue = m_DataSource->Value();

	// Increment the current value ...
	Increment();

	// Start the scrolling process ...
	StartScrolling(INCREMENT);	
}

void k3dSpinButton::OnUpReleased()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// We're done making changes ...
	PopEditing();

	// Cancel any scrolling ...
	CancelScrolling();
}

void k3dSpinButton::OnDownPressed()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Let the world know we're going to be making changes ...
	PushEditing();
	
	// Make sure any manual edits get recorded ...
	const gdouble newvalue = EditControlValue();
	if(newvalue != m_DataSource->Value())
		SetValue(newvalue);

	// Get the current value ...
	m_ScrollStartValue = m_DataSource->Value();

	// Decrement the current value ...
	Decrement();
	
	// Start the scrolling process ...
	StartScrolling(DECREMENT);	
}

void k3dSpinButton::OnDownReleased()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// We're done making changes ...
	PopEditing();
	
	// Cancel any scrolling ...
	CancelScrolling();
}

gdouble k3dSpinButton::EditControlValue()
{
	// Get the control text ...
	const sdpString inputtext = Editable(control_value).GetText();

	// Convert it into a double ...
	gdouble result = m_DataSource->Value();
	m_Format->ParsedInput(inputtext, result);

	return result;
}

void k3dSpinButton::PushEditing()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		m_StateRecorder->SetCurrentChangeSet(new k3dStateChangeSet());

	// Let the datasource know that we're going to be making changes ...
	m_DataSource->PushEditing();

	// Notify interested parties ...
	if(m_PushEditingAction)
		m_PushEditingAction->DoAction();
}

void k3dSpinButton::Increment()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Update everything ...
	SetValue(m_DataSource->Value() + m_StepIncrement);
}

void k3dSpinButton::Decrement()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Update everything ...
	SetValue(m_DataSource->Value() - m_StepIncrement);
}

void k3dSpinButton::SetValue(const gdouble& Value)
{
	// Make sure we've got some storage, first!
	g_return_if_fail(m_DataSource);

	// Update storage with the new value ...
	m_DataSource->SetValue(Value);

	// Notify any interested parties that things changed ...
	if(m_ChangeAction)
		m_ChangeAction->DoAction();
}

void k3dSpinButton::PopEditing()
{
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	// Let the data source know we're done ...
	m_DataSource->PopEditing();

	// Record the command for posterity (tutorials) ...
	if(m_CommandRecorder)
		if(SDP_S_OK == m_CommandRecorder->IsRecording())
			m_CommandRecorder->RecordCommand(this, k3dICommandNode::USERINTERFACE, control_value, sdpToString(m_DataSource->Value()));
	
	// Turn this into an undo/redo -able event ...
	if(m_StateRecorder)
		{
			sdpString formattedvalue;
			g_return_if_fail(m_Format->FormattedOutput(m_DataSource->Value(), formattedvalue));

			m_StateRecorder->RecordChangeSet(m_StateRecorder->ChangeSet(), m_StateChangeName + "\"" + formattedvalue + "\"");
			m_StateRecorder->SetCurrentChangeSet(0);
		}
	
	// Notify interested parties ...
	if(m_PopEditingAction)
		m_PopEditingAction->DoAction();
}

