#ifndef K3DSPINBUTTON_H
#define K3DSPINBUTTON_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dSpinButton class, which provides a standard MVC UI for scalar values
		\author Tim Shead (tshead@k-3d.com)
		\author Dan Erikson <derikson@montana.com>
*/

#include <k3dcontrol.h>
#include <k3dvariable.h>

#include <k3dichannel.h>
#include <k3dieditable.h>

#include <sdpgtkloopevents.h>

// Forward class declarations
class k3dICommandRecorder;
class k3dIStateRecorder;
class sdpGtkIObjectContainer;
class sdpxml::Element;

///////////////////////////////////////////////////////////////////////////
// k3dVariableFormatDataType

/// Implementation of k3dControlFormat<> for displaying floating-point data with real-world units and fixed precision
class k3dVariableFormatDataType :
	public k3dControlFormat<gdouble>
{
public:
	k3dVariableFormatDataType(const k3dIChannel::DataType DataType, const gulong Precision);
	virtual ~k3dVariableFormatDataType();

	bool FormattedOutput(const gdouble& Value, sdpString& Output);
	bool ParsedInput(const sdpString& Input, gdouble& InputValue);

private:
	const k3dIChannel::DataType m_DataType;
	const gulong m_Precision;
};

///////////////////////////////////////////////////////////////////////////
// k3dVariableFormatPrecision

/// Implementation of k3dControlFormat<> for displaying floating-point data with fixed precision
class k3dVariableFormatPrecision :
	public k3dControlFormat<gdouble>
{
public:
	k3dVariableFormatPrecision(const gulong Precision);
	virtual ~k3dVariableFormatPrecision();

	bool FormattedOutput(const gdouble& Value, sdpString& Output);
	bool ParsedInput(const sdpString& Input, gdouble& InputValue);

private:
	const gulong m_Precision;
};

///////////////////////////////////////////////////////////////////////////
// k3dSpinButtonDataSource

/// Abstract interface for objects that can act as a datasource for a k3dSpinButton (i.e. the "model" in model-view-controller)
class k3dSpinButtonDataSource :
	public k3dIEditable
{
public:
	virtual ~k3dSpinButtonDataSource() {}
	
	virtual k3dIEventSource* EventSource() = 0;
	
	virtual gdouble Value() = 0;
	virtual void SetValue(const gdouble Value) = 0;
	
};

///////////////////////////////////////////////////////////////////////////
// k3dSpinButtonVariableDataSource

/// Implementation of k3dSpinButtonDataSource for working with objects of type k3dVariable<gdouble>
class k3dSpinButtonVariableDataSource :
	public k3dSpinButtonDataSource
{
public:
	k3dSpinButtonVariableDataSource(k3dVariable<gdouble>& Source) :
		m_Source(Source)
	{
	}

	k3dIEventSource* EventSource()
	{
		return dynamic_cast<k3dIEventSource*>(&m_Source);
	}

	gdouble Value()
	{
		return m_Source.Value();
	}

	void PushEditing()
	{
		m_Source.PushEditing();
	}
			
	void SetValue(const gdouble Value)
	{
		m_Source.SetValue(Value);
	}
	
	void PopEditing()
	{
		m_Source.PopEditing();
	}
	
private:
	k3dVariable<gdouble>& m_Source;
};

///////////////////////////////////////////////////////////////////////////
// k3dSpinButton

typedef k3dControl k3dSpinButtonBase;

/// Provides a UI for displaying/editing scalar values (i.e. the "view" and the "controller" in model-view-controller) 
class k3dSpinButton :
	public k3dSpinButtonBase
{
public:
	k3dSpinButton(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName);
	virtual ~k3dSpinButton();

	/// Called by the framework when instantiating the button from a GTKML (XML) document
	bool Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element);
	/// Called to set spin button parameters manually
	bool Create(k3dIChannel::DataType Datatype, const gulong Precision, const gdouble Increment);

	/// Attaches the button to a generic data source
	bool Attach(k3dSpinButtonDataSource* const Source, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);
	/// Attaches the button to a k3dVariable<gdouble>
	bool Attach(k3dVariable<gdouble>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);

	const sdpString CustomType() const;

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// k3dICommandNode implementation
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

private:
	/// Called to handle SDPGTK events
	void OnEvent(sdpGtkEvent* Event);
	
	/// Called when the GTK+ widgets are about to disappear
	void OnDestroy();
	/// Called when the user tabs into / clicks on the edit control
	void OnValueFocusIn();
	/// Called when the user tabs out of the edit control
	void OnValueFocusOut();
	/// Called when the user hits the "enter" key
	void OnValueActivate();
	/// Called when the user presses the "up" button
	void OnUpPressed();
	/// Called when the user releases the "up" button
	void OnUpReleased();
	/// Called when the user presses the "down" button
	void OnDownPressed();
	/// Called when the user releases the "down" button
	void OnDownReleased();

	/// Enumerates scroll "modes"
	typedef enum
	{
		NONE,
		INCREMENT,
		DECREMENT
	} ScrollMode;

	/// Called when the user begins scrolling by holding-down one of the arrow buttons
	void StartScrolling(const ScrollMode Mode);
	/// Called when the user finishs scrolling by releasing an arrow button
	void CancelScrolling();
	
	void OnStartScrolling();
	void OnFasterScrolling();
	void OnFastestScrolling();
	void OnScrolling();
			
	/// Called when the contents of the edit control need to be updated (typically due to outside modifications of the datasource) 
	void Update();

	/// Returns the value of the edit control contents
	gdouble EditControlValue();
	
	/// Called prior to making one-or-more changes to the underlying data source
	void PushEditing();	
	/// Called to increment the current value
	void Increment();
	/// Called to decrement the current value
	void Decrement();
	/// Called to modify the underlying data source
	void SetValue(const gdouble& Value);
	/// Called once all changes to the underlying data source are complete
	void PopEditing();

	/// Storeas a reference to the underlying data object
	k3dSpinButtonDataSource* m_DataSource;
	/// Stores a reference to the (optional) formatting object
	k3dControlFormat<gdouble>* m_Format;
	/// Stores the increment used to modify the button's value using the spin controls
	gdouble m_StepIncrement;

	/// Stores the original value at the time that scrolling begins
	gdouble m_ScrollStartValue;
	/// Stores the event that gets scrolling started
	sdpGtkEventTimeout* m_StartScrollingEvent;
	/// Stores the event that gets scrolling going faster
	sdpGtkEventTimeout* m_FasterScrollingEvent;
	/// Stores the event that gets scrolling going at high speed
	sdpGtkEventTimeout* m_FastestScrollingEvent;
	/// Stores the event that drives continuous scrolling
	sdpGtkEventTimeout* m_ScrollingEvent;
	/// Stores the current scroll mode	
	ScrollMode m_ScrollMode;
	
	static const sdpString control_value;
	static const sdpString control_up;
	static const sdpString control_down;
};

#endif // K3DSPINBUTTON_H

