// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dState class, which provides a default implementation of k3dIState
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dstate.h>
#include <k3dstaterecorder.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // SDPXML_H

////////////////////////////////////////////////////////////////////////////////////
// k3dState

k3dState::k3dState(const gulong State) :
	m_StateFlags(State)
{
}

////////////////////////////////////////////////////////////////////////////////////
// k3dIState overrides ...

sdpResult k3dState::QueryState(const gulong State)
{
	return ((m_StateFlags & State) == State) ? SDP_S_OK : SDP_S_FALSE;
}

void k3dState::RawSetState(const gulong Mask, const bool State, k3dIStateRecorder* const StateRecorder)
{
	// If nothing changed, we're done ...
	const gulong changed = State ? ((Mask^m_StateFlags) & Mask) : (Mask & m_StateFlags);
	if(!changed)
		return;

	// If we're recording state changes ...
	if(StateRecorder->ChangeSet())
		StateRecorder->ChangeSet()->RecordOldState(new k3dValueStateContainer<gulong>(m_StateFlags));
		
	RawSetState(Mask, State);
	
	if(StateRecorder->ChangeSet())
		StateRecorder->ChangeSet()->RecordNewState(new k3dValueStateContainer<gulong>(m_StateFlags));
}

void k3dState::RawSetState(const gulong Mask, const bool State)
{
	if(State)
		m_StateFlags |= Mask;
	else
		m_StateFlags &= ~Mask;
}

sdpResult k3dState::Save(sdpxml::Element& Element, k3dIDependencies* const Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	Element.Children().push_back(sdpxml::Element("state"));
	sdpxml::Element& state = Element.Children().back();

	if(m_StateFlags & VISIBLE)
		sdpxml::SetAttribute(state, sdpxml::Attribute("visible", "true"));
			
	if(m_StateFlags & SELECTED)
		sdpxml::SetAttribute(state, sdpxml::Attribute("selected", "true"));
			
	if(m_StateFlags & SELECTABLE)
		sdpxml::SetAttribute(state, sdpxml::Attribute("selectable", "true"));
			
	if(m_StateFlags & USERVIEWABLE)
		sdpxml::SetAttribute(state, sdpxml::Attribute("userviewable", "true"));
			
	return SDP_S_OK;
}

sdpResult k3dState::Load(sdpxml::Element& Element)
{
	// Reset our state ...
	m_StateFlags = 0;

	// Implied by the fact that we're being loaded ... duh!
	RawSetState(SAVEABLE, true);

	// See what we find in the file ...
	sdpxml::ElementPointer state = sdpxml::FindElement(Element, sdpxml::SameName("state"));
	if(!state)
		return SDP_S_FALSE;

	RawSetState(VISIBLE, sdpxml::GetAttribute(*state, "visible", false));
	RawSetState(SELECTED, sdpxml::GetAttribute(*state, "selected", false));
	RawSetState(SELECTABLE, sdpxml::GetAttribute(*state, "selectable", false));
	RawSetState(USERVIEWABLE, sdpxml::GetAttribute(*state, "userviewable", false));

	return SDP_S_OK;
}

