// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dStateRecorder implementation of k3dIStateRecorder
		\author Tim Shead (tshead@k-3d.com)
		\author Dan Erikson (derikson@montana.com)
*/

#include <k3dmisc.h>
#include <k3dstaterecorder.h>

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dStateChangeSet

k3dStateChangeSet::k3dStateChangeSet()
{
}

k3dStateChangeSet::~k3dStateChangeSet()
{
	std::for_each(m_OldStates.begin(), m_OldStates.end(), sdp::DeleteObject());
	std::for_each(m_OldActions.begin(), m_OldActions.end(), sdp::DeleteObject());
	std::for_each(m_Actions.begin(), m_Actions.end(), sdp::DeleteObject());
	std::for_each(m_NewActions.begin(), m_NewActions.end(), sdp::DeleteObject());
	std::for_each(m_NewStates.begin(), m_NewStates.end(), sdp::DeleteObject());
}

void k3dStateChangeSet::RecordOldState(k3dIStateContainer* const OldState)
{
	// Sanity checks ...
	g_return_if_fail(OldState);
	m_OldStates.push_back(OldState);
}

void k3dStateChangeSet::RecordOldAction(k3dIAction* const Action)
{
	// Sanity checks ...
	g_return_if_fail(Action);
	m_OldActions.push_back(Action);
}

void k3dStateChangeSet::RecordAction(k3dIAction* const Action)
{
	// Sanity checks ...
	g_return_if_fail(Action);
	m_Actions.push_back(Action);
}

void k3dStateChangeSet::RecordNewAction(k3dIAction* const Action)
{
	// Sanity checks ...
	g_return_if_fail(Action);
	m_NewActions.push_back(Action);
}

void k3dStateChangeSet::RecordNewState(k3dIStateContainer* const NewState)
{
	// Sanity checks ...
	g_return_if_fail(NewState);
	m_NewStates.push_back(NewState);
}

void k3dStateChangeSet::Undo()
{
	std::for_each(m_OldStates.begin(), m_OldStates.end(), std::mem_fun(&k3dIStateContainer::RestoreState));
	std::for_each(m_OldActions.begin(), m_OldActions.end(), std::mem_fun(&k3dIAction::DoAction));
	std::for_each(m_Actions.begin(), m_Actions.end(), std::mem_fun(&k3dIAction::DoAction));
}

void k3dStateChangeSet::Redo()
{
	std::for_each(m_NewStates.begin(), m_NewStates.end(), std::mem_fun(&k3dIStateContainer::RestoreState));
	std::for_each(m_NewActions.begin(), m_NewActions.end(), std::mem_fun(&k3dIAction::DoAction));
	std::for_each(m_Actions.begin(), m_Actions.end(), std::mem_fun(&k3dIAction::DoAction));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dStateRecorder

k3dStateRecorder::k3dStateRecorder() :
	m_CurrentChangeSet(0),
	m_RedoPosition(m_ChangeSets.end()),
	m_SavedPosition(m_ChangeSets.end())
{
}

k3dStateRecorder::~k3dStateRecorder()
{
	std::for_each(m_ChangeSets.begin(), m_ChangeSets.end(), sdp::DeleteObject());
	
	// Sanity checks ...
	g_return_if_fail(!m_CurrentChangeSet);
}

void k3dStateRecorder::SetCurrentChangeSet(k3dIStateChangeSet* const ChangeSet)
{
std::cerr << __PRETTY_FUNCTION__ << " " << ChangeSet << std::endl;
	
	// Sanity checks ...
	g_return_if_fail(m_CurrentChangeSet || ChangeSet);
	g_return_if_fail(!(m_CurrentChangeSet && ChangeSet));
	
	m_CurrentChangeSet = ChangeSet;
}

k3dIStateChangeSet* k3dStateRecorder::ChangeSet()
{
	return m_CurrentChangeSet;
}

void k3dStateRecorder::RecordChangeSet(k3dIStateChangeSet* const ChangeSet, SDPCSTRING ChangeSetLabel)
{
	// Sanity checks ...
	g_return_if_fail(ChangeSet);
	g_return_if_string_fail(ChangeSetLabel);
	
	// Recording a state change creates a "branch" in the state change tree, so trim any data in the old branch ...
	for(ChangeSetIterator victim = m_RedoPosition; victim != m_ChangeSets.end(); victim++)
		{
			if(m_SavedPosition == victim)
				{
					m_SavedPosition = m_ChangeSets.end();
					break;
				}
		}
	
	std::for_each(m_RedoPosition, m_ChangeSets.end(), sdp::DeleteObject());
	m_ChangeSets.erase(m_RedoPosition, m_ChangeSets.end());

	// Store the new change set ...
	m_ChangeSets.push_back(new ChangeSetRecord(ChangeSet, ChangeSetLabel));
	m_RedoPosition = m_ChangeSets.end();

	Notify(Event_StackChanged, 0);
}

void k3dStateRecorder::MarkSaved()
{
	if(m_RedoPosition != m_ChangeSets.begin())
		m_SavedPosition = --ChangeSetIterator(m_RedoPosition);
		
	Notify(Event_Saved, 0);
}

bool k3dStateRecorder::UnsavedChanges()
{
	// Either we haven't changed anything or we undid all our changes
	if(m_RedoPosition == m_ChangeSets.begin())
		return m_SavedPosition != m_ChangeSets.end();

	// See if the current state and the saved state match ...
	return m_SavedPosition != --ChangeSetIterator(m_RedoPosition);
}

gulong k3dStateRecorder::UndoCount()
{
	return std::distance(m_ChangeSets.begin(), m_RedoPosition);
}

gulong k3dStateRecorder::RedoCount()
{
	return std::distance(m_RedoPosition, m_ChangeSets.end());
}

SDPCSTRING k3dStateRecorder::NextUndoLabel()
{
	return m_RedoPosition != m_ChangeSets.begin() ? static_cast<SDPCSTRING>((*(--ChangeSetIterator(m_RedoPosition)))->Label()) : static_cast<SDPCSTRING>(0);
}

SDPCSTRING k3dStateRecorder::NextRedoLabel()
{
	return m_RedoPosition != m_ChangeSets.end() ? static_cast<SDPCSTRING>((*m_RedoPosition)->Label()) : static_cast<SDPCSTRING>(0);
}

void k3dStateRecorder::Undo()
{
	// Should never happen in the middle of recording a state change set ...
	g_return_if_fail(!m_CurrentChangeSet);

	// Make sure we have actions to undo ...
	g_return_if_fail(m_RedoPosition != m_ChangeSets.begin());

	// Undo the most recent change set
	m_RedoPosition--;
	(*m_RedoPosition)->ChangeSet()->Undo();

	// Let the gentry know ...
	Notify(Event_StackChanged, 0);
	Notify(Event_Undo, 0);
}

void k3dStateRecorder::Redo()
{
	// Should never happen in the middle of recording a state change set ...
	g_return_if_fail(!m_CurrentChangeSet);

	// Make sure we have actions to redo ...
	g_return_if_fail(m_RedoPosition != m_ChangeSets.end());

	// Redo the most recently undone change set ...
	(*m_RedoPosition)->ChangeSet()->Redo();
	m_RedoPosition++;

	// Let the gentry know ...
	Notify(Event_StackChanged, 0);
	Notify(Event_Redo, 0);
}

void k3dStateRecorder::VisitChangeSets(k3dIChangeSetVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	for(ChangeSetIterator undoset = m_ChangeSets.begin(); undoset != m_RedoPosition; undoset++)
		Visitor->VisitUndoChangeSet((*undoset)->Label(), undoset == m_SavedPosition);
	
	for(ChangeSetIterator redoset = m_RedoPosition; redoset != m_ChangeSets.end(); redoset++)
		Visitor->VisitRedoChangeSet((*redoset)->Label(), redoset == m_SavedPosition);
}


