#ifndef K3DSTATERECORDER_H
#define K3DSTATERECORDER_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dStateRecorder implementation of k3dIStateRecorder
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deventsource.h>
#include <k3distaterecorder.h>

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dValueStateContainer

/// Provides an implementation of k3dIStateContainer for storing an object's value
template<class Type>
class k3dValueStateContainer :
	public k3dIStateContainer
{
public:
	k3dValueStateContainer(Type& Instance) :
		m_Instance(Instance),
		m_InstanceValue(Instance)
	{
	}
	
	k3dValueStateContainer(Type& Instance, const Type InstanceValue) :
		m_Instance(Instance),
		m_InstanceValue(InstanceValue)
	{
	}
	
	void RestoreState()
	{
		m_Instance = m_InstanceValue;
	}
	
private:
	Type& m_Instance;
	const Type m_InstanceValue;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dNewInstanceStateContainer

/// Provides an implementation of k3dIStateContainer for storing instances of objects that are newly created	
template<class Type>
class k3dNewInstanceStateContainer :
	public k3dIStateContainer
{
public:
	k3dNewInstanceStateContainer(Type* const Instance, k3dIStateChangeSet* const ChangeSet) :
		m_Instance(Instance),
		m_InstanceOwned(false)
	{
		// Sanity checks ...
		g_assert(ChangeSet);
		g_return_if_fail(Instance); // This isn't a huge problem, since deleting a NULL pointer isn't an error, but it still shouldn't happen
		
		ChangeSet->RecordOldState(new k3dValueStateContainer<bool>(m_InstanceOwned, true));
		ChangeSet->RecordNewState(new k3dValueStateContainer<bool>(m_InstanceOwned, false));
	}
	
	~k3dNewInstanceStateContainer()
	{
		if(m_InstanceOwned)
			delete m_Instance;
	}

	void RestoreState()
	{
	}

private:
	Type* const m_Instance;
	bool m_InstanceOwned;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dDeleteInstanceStateContainer

/// Provides an implementation of k3dIStateContainer for storing instances of objects that have been deleted	
template<class Type>
class k3dDeleteInstanceStateContainer :
	public k3dIStateContainer
{
public:
	k3dDeleteInstanceStateContainer(Type* const Instance, k3dIStateChangeSet* const ChangeSet) :
		m_Instance(Instance),
		m_InstanceOwned(true)
	{
		// Sanity checks ...
		g_assert(ChangeSet);
		g_return_if_fail(Instance); // This isn't a huge problem, since deleting a NULL pointer isn't an error, but it still shouldn't happen
		
		ChangeSet->RecordOldState(new k3dValueStateContainer<bool>(m_InstanceOwned, false));
		ChangeSet->RecordNewState(new k3dValueStateContainer<bool>(m_InstanceOwned, true));
	}
	
	~k3dDeleteInstanceStateContainer()
	{
		if(m_InstanceOwned)
			delete m_Instance;
	}

	void RestoreState()
	{
	}

private:
	Type* const m_Instance;
	bool m_InstanceOwned;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dStateChangeSet

/// Provides an implementation of k3dIStateChangeSet
class k3dStateChangeSet :
	public k3dIStateChangeSet
{
public:
	k3dStateChangeSet();
	~k3dStateChangeSet();
	
	// k3dIStateChangeSet implementation
	void RecordOldState(k3dIStateContainer* const OldState);
	void RecordOldAction(k3dIAction* const Action);
	void RecordAction(k3dIAction* const Action);
	void RecordNewAction(k3dIAction* const Action);
	void RecordNewState(k3dIStateContainer* const NewState);
	
	void Undo();
	void Redo();
	
private:
	typedef std::vector<k3dIStateContainer*> StateCollection;
	StateCollection m_OldStates;
	StateCollection m_NewStates;
	
	typedef std::vector<k3dIAction*> ActionCollection;
	ActionCollection m_OldActions;
	ActionCollection m_Actions;
	ActionCollection m_NewActions;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dStateRecorder

/// Provides an implementation of k3dIStateRecorder
class k3dStateRecorder :
	public k3dIStateRecorder,
	public k3dEventSource
{
public:
	k3dStateRecorder();
	~k3dStateRecorder();

	// k3dIStateRecorder implementation ...
	void SetCurrentChangeSet(k3dIStateChangeSet* const ChangeSet);
	k3dIStateChangeSet* ChangeSet();
	void RecordChangeSet(k3dIStateChangeSet* const ChangeSet, SDPCSTRING ChangeSetLabel);

	void MarkSaved();
	bool UnsavedChanges();

	gulong UndoCount();
	gulong RedoCount();

	SDPCSTRING NextUndoLabel();
	SDPCSTRING NextRedoLabel();

	void Undo();
	void Redo();
	
	void VisitChangeSets(k3dIChangeSetVisitor* const Visitor);
	
private:
	/// Associates a name with a change set
	class ChangeSetRecord
	{
	public:
		ChangeSetRecord(k3dIStateChangeSet* const ChangeSet, SDPCSTRING Label) :
			m_ChangeSet(ChangeSet),
			m_Label(Label)
		{
			// Sanity checks ...
			g_assert(m_ChangeSet);
			g_assert(m_Label.size());
		}
		
		~ChangeSetRecord()
		{
			delete m_ChangeSet;
		}
		
		k3dIStateChangeSet* ChangeSet() const
		{
			return m_ChangeSet;
		}
		
		const sdpString& Label() const
		{
			return m_Label;
		}
		
	private:
		k3dIStateChangeSet* const m_ChangeSet;
		sdpString m_Label;
	};

	/// Stores a collection of change sets
	typedef std::list<ChangeSetRecord*> ChangeSetCollection;
	typedef ChangeSetCollection::iterator ChangeSetIterator;
	/// Stores change sets
	ChangeSetCollection m_ChangeSets;
	
	/// Stores the current change set
	k3dIStateChangeSet* m_CurrentChangeSet;
	/// Stores an iterator that partitions the m_ChangeSets collection into undo-able and redo-able change sets
	ChangeSetIterator m_RedoPosition;
	/// Stores an iterator that points to the last saved change set
	ChangeSetIterator m_SavedPosition;
};

#endif // K3DSTATERECORDER_H

