// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTargetController class, which provides a UI for manually scaling objects that support k3dITransform
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dobjectselector.h>
#include <k3dtargetcontroller.h>
#include <k3dicommandrecorder.h>
#include <k3distaterecorder.h>

#include <sdpgtkloopevents.h>
#include <sdpgtkutility.h>

//////////////////////////////////////////////////////////////////////
// k3dTargetController

const sdpString k3dTargetController::control_object = "object";
const sdpString k3dTargetController::control_reset = "reset";

k3dTargetController::k3dTargetController(sdpIUnknown* CommandNodeParent, SDPCSTRING CommandNodeName) :
	k3dTargetControllerBase(CommandNodeParent, CommandNodeName),
	m_DataSource(0),
	m_Selector(0)
{
	m_Selector = new k3dObjectSelector(this, control_object);

	// Create and load our UI template ...
	std::istrstream uitemplate(
		"<gtkml>"
			"<vbox homogeneous=\"false\">"
				"<event signal=\"destroy\" name=\"destroy\"/>"

				"<eventbox name=\"object_container\"/>"
/*				
				"<button name=\"reset\">Reset Scale"
					"<event signal=\"clicked\" name=\"reset\"/>"
				"</button>"
*/
			"</vbox>"
		"</gtkml>\n");

	g_return_if_fail(Load(uitemplate, ""));

	// Attach the spin buttons ...
	Container("object_container").Attach(m_Selector->RootWidget());

/*	
	AddCommand(control_reset, control_reset, CONTROL_BUTTON);
*/
	
	RootWidget().Show();
}

k3dTargetController::~k3dTargetController()
{
	// No more events from this point forward ...
	DisconnectAllEvents();

	// Clean-up the GTK+ tree ...
	if(Root())
		RootWidget().Destroy();
		
	// Delete spin buttons ...
	delete m_Selector;
}

const sdpString k3dTargetController::CustomType() const
{
	return sdpString("k3dTargetController");
}

bool k3dTargetController::Create(sdpGtkIObjectContainer* ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	
	g_assert(Element.Name() == "k3dtargetcontroller");

	return true;
}

///////////////////////////////////////////////////////////////////////////
// k3dTargetControllerFilter

/// An implementation of k3dObjectSelectorFilter that only displays targetable objects 
class k3dTargetControllerFilter :
	public k3dObjectSelectorFilter
{
public:
	k3dTargetControllerFilter(k3dIObject* const Self) :
		m_Self(Self)
	{
		// Sanity checks ...
		g_assert(m_Self);
	}

	bool Allow(k3dIObjectInfo* const ObjectInfo)
	{
		// Sanity checks ...
		g_assert(ObjectInfo);
		
		return false;
	}
	
	bool Allow(k3dIObject* const Object)
	{
		// Sanity checks ...
		g_assert(Object);
		
		// Don't allow us to select ourselves ...
		if(Object == m_Self)
			return false;
		
		// Only allow transformable objects ...
		k3dITransform* const transform = dynamic_cast<k3dITransform*>(Object);
		if(!transform)
			return false;

		return true;
	}

private:
	k3dIObject* const m_Self;
};



bool k3dTargetController::Attach(k3dIObject* const Object, k3dObjectVariable* const DataSource, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	// Sanity checks ...
	g_return_val_if_fail(Object, false);
	g_return_val_if_fail(DataSource, false);
	g_return_val_if_fail(k3dTargetControllerBase::Attach(CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);
	
	// Store references ...
	m_Document = Object->Document();
	m_DataSource = DataSource;

	// k3dITransform
	g_return_val_if_fail(m_Selector->Attach(m_DataSource, CommandRecorder, StateRecorder, "Target", 0, new k3dActionRedrawAll(m_Document, false, 0), 0, true, new k3dTargetControllerFilter(Object)), false);

	return true;
}

void k3dTargetController::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "destroy")
		OnDestroy();
	else if(Event->Name() == control_reset)
		OnReset();
	else
		k3dTargetControllerBase::OnEvent(Event);
}

void k3dTargetController::OnDestroy()
{
	DisconnectAllEvents();

	Clear();
}

void k3dTargetController::OnReset()
{
/*
	// Sanity checks ...
	g_return_if_fail(m_DataSource);

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_reset, 0);
	k3dRecordStateChangeSet undo(m_Document, "Reset Scale");

	sdpVector3 scale(1, 1, 1);
	m_DataSource->SetScale(k3dITransform::CURRENT_FRAME, scale);
	
	k3dRedrawAll(m_Document, false);
*/
}

