// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTextEditor class, which provides an extensible UI for editing textfiles
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dtexteditor.h>

#include <k3dmisc.h>
#include <k3diapplication.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>

#include <fstream>
#include <strstream>

//////////////////////////////////////////////////////////////////////
// k3dTextEditor

k3dTextEditor::k3dTextEditor(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName) :
	k3dTextEditorBase(ParentCommandNode, CommandNodeName, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Changed(false)
{
}

bool k3dTextEditor::LoadGTKMLTemplate(SDPCSTRING Template)
{
	// Sanity checks ...
	g_assert_string(Template);
	
	// Load the dialog template ...
	g_return_val_if_fail(k3dTextEditorBase::LoadGTKMLTemplate(Template), false);

	// Put something in the titlebar ...
	UpdateTitlebar();
			
	// Attach scrollbars to the text widget ...
	GtkText* text = TextControl();
	Scrollbar("vscrollbar").SetAdjustment(text->vadj);
	
	return true;
}

k3dTextEditor::~k3dTextEditor()
{
	RescueChanges();
}

sdpGtkText k3dTextEditor::TextControl()
{
	return Text("text");
}

// We have an unfortunate class with X
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

void k3dTextEditor::UpdateTitlebar()
{
	sdpString title = GetTitlebarText();
	RootWindow().SetTitle(title);
}

sdpString k3dTextEditor::GetTitlebarText()
{
	sdpString result = GetTitle();

	if(m_Changed)
		result += " [changed]";

	return result;
}

sdpString k3dTextEditor::GetTitle()
{
	if(m_FilePath.size())
		return sdpFile(m_FilePath);
		
	return sdpString("Untitled file");
}

void k3dTextEditor::Load(sdpString Filepath)
{
	g_assert(Filepath.size());
	
	// Try opening the file ...
	std::ifstream file(Filepath);
	if(!file.good())
		{
			k3dApplication()->Error("Error opening file!", "Open File:");
			return;
		}

	m_FilePath = Filepath;	
	Load(file);
}

void k3dTextEditor::Load(std::istream& Stream)
{
	// Get rid of old text ...	
	TextControl().DeleteText(0, -1);
	
	// Load the file into our text widget ...
	std::strstream script;
	Stream.get(*script.rdbuf(), '\0');
	script << std::ends;
		
	gint position = 0;
	TextControl().InsertText(script.str(), strlen(script.str()), &position);
	
	script.freeze(false);
	
	m_Changed = false;
	UpdateTitlebar();
}

void k3dTextEditor::Save(sdpString Filepath)
{
	// Sanity checks ...
	g_assert(Filepath.size());
	
	// Try opening the file ...
	std::ofstream file(Filepath);
	if(!file.good())
		{
			k3dApplication()->Error("Error saving file!", "Save File:");
			return;
		}
		
	m_FilePath = Filepath;
	Save(file);
}

void k3dTextEditor::Save(std::ostream& Stream)
{
	// Sanity checks ...
	g_assert(Stream.good());
	
	// Get the script text from our widget ...
	sdpString script = TextControl().GetText();
	Stream.write(script, script.size());
	
	m_Changed = false;
	UpdateTitlebar();
}

void k3dTextEditor::RescueChanges()
{
	if(!m_Changed)
		return;
	
	sdpStringArray buttons;
	buttons.push_back("Yes");
	buttons.push_back("No");
	if(1 != k3dApplication()->Query("The file has changed.  Save modifications?", GetTitle() + ":", 1, buttons))
		return;
		
	OnFileSave();
}

void k3dTextEditor::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "changed")
		OnChanged();
	else if(Event->Name() == "filenew")
		OnFileNew();
	else if(Event->Name() == "fileopen")
		OnFileOpen();
	else if(Event->Name() == "filesave")
		OnFileSave();
	else if(Event->Name() == "filesaveas")
		OnFileSaveAs();
	else if(Event->Name() == "filerevert")
		OnFileRevert();
	else if(Event->Name() == "fileclose")
		OnFileClose();
	else if(Event->Name() == "delete")
		OnDelete(Event);
	else
		k3dTextEditorBase::OnEvent(Event);
}

void k3dTextEditor::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Don't allow the GTK+ tree to go away just yet ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(true);

	// Turn it into a "close", instead ...
	OnFileClose();
}

void k3dTextEditor::OnFileNew()
{
	RescueChanges();

	TextControl().DeleteText(0, -1);
	m_FilePath.clear();
	
	UpdateTitlebar();
}

void k3dTextEditor::OnFileOpen()
{
	RescueChanges();

	sdpString filepath;
	if(!k3dGetFilePath("script", "Open File:", false, m_FilePath, filepath))
		return;
	
	Load(filepath);
}

void k3dTextEditor::OnFileSave()
{
	// If we don't have a filepath, prompt the user ...
	if(m_FilePath.size())
		Save(m_FilePath);
	else
		OnFileSaveAs();
}

void k3dTextEditor::OnFileRevert()
{
	if(0 == m_FilePath.size())
		return;
		
	if(!m_Changed)
		return;
		
	sdpStringArray buttons;
	buttons.push_back("Yes");
	buttons.push_back("No");
	if(1 != k3dApplication()->Query("The file has changed.  Revert to previously saved version?", GetTitle() + ":", 1, buttons))
		return;
	
	Load(m_FilePath);	
}

void k3dTextEditor::OnFileClose()
{
	delete this;
}

void k3dTextEditor::OnChanged()
{
	if(!m_Changed)
		{
			m_Changed = true;
			UpdateTitlebar();
		}
}


