// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTimeController facade which provides high-level, frame-oriented manipulation of a document's k3dITime interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include "k3dtimecontroller.h"

#include <k3didocument.h>
#include <k3ditime.h>

k3dTimeController::k3dTimeController(k3dITime* const Time) :
	m_Time(Time)
{
	// Sanity checks ...
	g_assert(m_Time);
}

gulong k3dTimeController::StartFrame() const
{
	return RealToUser(InternalToReal(m_Time->StartTime()));
}

gulong k3dTimeController::ViewFrame() const
{
	return RealToUser(InternalToReal(m_Time->ViewTime()));
}

gulong k3dTimeController::EndFrame() const
{
	return RealToUser(InternalToReal(m_Time->EndTime()));
}

gulong k3dTimeController::TotalFrames() const
{
	return gulong(floor(RealDuration()-1.0)) + 1;
}

gdouble k3dTimeController::Duration() const
{
	return m_Time->Duration();
}

gdouble k3dTimeController::FrameRate() const
{
	return m_Time->FrameRate();
}

sdpResult k3dTimeController::SetStartFrame(const gulong Frame)
{
	return m_Time->SetStartTime(RealToInternal(UserToReal(Frame)));
}

sdpResult k3dTimeController::SetViewFrame(const gulong Frame)
{
	return m_Time->SetViewTime(RealToInternal(UserToReal(Frame)));
}

sdpResult k3dTimeController::SetEndFrame(const gulong Frame)
{
	return m_Time->SetEndTime(RealToInternal(UserToReal(Frame)));
}

sdpResult k3dTimeController::SetTotalFrames(const gulong Frames)
{
	m_Time->SetDuration(gdouble(Frames) / FrameRate());
			
	return SDP_S_OK;
}

sdpResult k3dTimeController::SetDuration(const gdouble Duration)
{
	m_Time->SetDuration(Duration);
			
	return SDP_S_OK;
}

sdpResult k3dTimeController::SetFrameRate(const gdouble Rate)
{
	m_Time->SetFrameRate(Rate);
	
	return SDP_S_OK;
}

gdouble k3dTimeController::UserToReal(const gulong UserFrame) const
{
	return sdpMin(TotalFrames(), UserFrame) - 1.0;
}

gulong k3dTimeController::RealToUser(const gdouble RealFrame) const
{
	return gulong(floor(RealFrame)) + 1;
}

gdouble k3dTimeController::RealToInternal(const gdouble RealTime) const
{
	return sdpClamp(0.0, gdouble(RealTime) / RealDuration(), 1.0);
}

gdouble k3dTimeController::InternalToReal(const gdouble InternalTime) const
{
	return InternalTime * RealDuration();
}

gdouble k3dTimeController::RealDuration() const
{
	return m_Time->FrameRate() * m_Time->Duration();
}



