// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTransformObject class, which serves as a base class for K-3D objects that have position, orientation, scale, and can render themselves
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3dstaterecorder.h>
#include <k3dtransformobject.h>

#include <k3dicamera.h>
#include <k3didocument.h>
#include <k3dieditorengine.h>
#include <k3dihierarchy.h>
#include <k3dilocationcollection.h>
#include <k3diriengine.h>
#include <k3distaterecorder.h>
#include <k3divariable.h>

#include <sdpgtkwidget.h>
#include <sdpgl.h>

//////////////////////////////////////////////////////////////////////////////////
// k3dTransformObject ...

k3dTransformObject::k3dTransformObject(sdpIUnknown* Initializer) :
	k3dTransformObjectBase(Initializer),
	k3dMouseEventObserver("LMB-drag to rotate/translate/scale Object; Control-LMB-drag to constrain to Active Axis"),
	m_TransformType(NONE),
	m_Target("target", Document(), Document()->StateRecorder(), 0),
	m_CurrentPositionChanged(false),
	m_ChildPositionChanged(false),
	m_CurrentOrientationChanged(false),
	m_ChildOrientationChanged(false),
	m_CurrentScaleChanged(false),
	m_ChildScaleChanged(false)
{
	AddVariable(&m_Target);
}


k3dTransformObject::~k3dTransformObject()
{
}

///////////////////////////////////////////////////////////////////
// k3dIPersist implementation ...

sdpResult k3dTransformObject::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Dependencies);

	g_return_val_if_fail(sdpSucceeded(k3dTransformObjectBase::Save(Element, Dependencies)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dTransform::Save(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dTransformObject::Load(sdpxml::Document& Document, sdpxml::Element& Element)
{
	g_return_val_if_fail(sdpSucceeded(k3dTransformObjectBase::Load(Document, Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dTransform::Load(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////
// k3dIEditorRender implementation ...

sdpResult k3dTransformObject::EditorPushRender(k3dIEditorEngine* const Engine)
{
	sdpglTrace("k3dTransformObject::EditorPushRender()");

	// Push a new set of matrices onto the stack ...
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	k3dPushOpenGLMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_FRAME));

	return k3dEditorRender::EditorPushRender(Engine);
}

sdpResult k3dTransformObject::EditorPushChildRender(k3dIEditorEngine* const Engine)
{
	sdpglTrace("k3dTransformObject::EditorPushChildRender()");

	// Push a new set of matrices onto the stack ...
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	k3dPushOpenGLMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_CHILD_FRAME));

	return k3dEditorRender::EditorPushChildRender(Engine);
}

sdpResult k3dTransformObject::EditorPopChildRender(k3dIEditorEngine* const Engine)
{
	sdpglTrace("k3dTransformObject::EditorPopChildRender()");

	// Pop old matrices off the stack ...
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	return k3dEditorRender::EditorPopChildRender(Engine);
}

sdpResult k3dTransformObject::EditorPopRender(k3dIEditorEngine* const Engine)
{
	sdpglTrace("k3dTransformObject::EditorPopRender()");

	// Pop old matrices off the stack ...
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	return k3dEditorRender::EditorPopRender(Engine);
}

/////////////////////////////////////////////////////////////////////
// k3dIRiRender implementation ...

sdpResult k3dTransformObject::RiPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	Engine->RiTransformBegin();
	
	sdpMatrix4 matrix = k3dGetObjectToWorldMatrix(this).Transpose();
	Engine->RiTransform(matrix);
	Engine->RiCoordinateSystem(sdpString("object_") + Name());
	
	Engine->RiTransformEnd();
	
	return SDP_S_OK;
}

sdpResult k3dTransformObject::RiPushRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	k3dRiRender::RiPushRender(Frame, Engine);
	k3dPushRiMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_FRAME), Engine);

	return SDP_S_OK;
}

sdpResult k3dTransformObject::RiPushChildRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	k3dRiRender::RiPushChildRender(Frame, Engine);
	k3dPushRiMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_CHILD_FRAME), Engine);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// k3dITransform implementation ...

const sdpAngleAxis& k3dTransformObject::Orientation(const ReferenceFrame Frame)
{
	if((ANIMATED_FRAME == Frame) && m_Target.Object())
		{
			k3dIObject* const parentobject = Document()->Hierarchy()->ObjectParent(this);
			const sdpVector3 thisposition = (parentobject ? k3dGetObjectToWorldMatrix(parentobject) : sdpIdentity3D()) * sdpTranslation3D(m_Position[ANIMATED_FRAME]) * sdpVector3(0, 0, 0);
			const sdpVector3 targetposition = k3dGetObjectToWorldMatrix(m_Target.Object()) * sdpVector3(0, 0, 0);
			const sdpVector3 targetspherical = (targetposition - thisposition).Spherical();

			m_TargetOrientation = sdpAngleAxis(sdpQuaternion(sdpEulerAngles(0, -targetspherical[2], targetspherical[1], sdpEulerAngles::ZXYstatic)));
			return m_TargetOrientation;
		}
	
	return k3dTransform::Orientation(Frame);
}

void k3dTransformObject::SetPosition(const k3dITransform::ReferenceFrame Frame, const sdpVector3& Position)
{
	// Are we recording actions right now?
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
	 	{
			if(!m_CurrentPositionChanged && (k3dITransform::CURRENT_FRAME == Frame))
				{
					g_return_if_fail(Editing());
					
					m_CurrentPositionChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_Position[Frame]));
					changeset->RecordAction(new k3dActionNotify(this, k3dITransform::Event_PositionChanged, 0, 0));
				}
			else if(!m_ChildPositionChanged && (k3dITransform::CURRENT_CHILD_FRAME == Frame))
				{
					g_return_if_fail(Editing());
					
					m_ChildPositionChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_Position[Frame]));
					changeset->RecordAction(new k3dActionNotify(this, k3dITransform::Event_ChildPositionChanged, 0, 0));
				}
		}

	// Make the change ...
	k3dTransform::SetPosition(Frame, Position);

	// Notify clients ...
	if(k3dITransform::CURRENT_FRAME == Frame)
		Notify(k3dITransform::Event_PositionChanged, 0);
	else if(k3dITransform::CURRENT_CHILD_FRAME == Frame)
		Notify(k3dITransform::Event_ChildPositionChanged, 0);
}

void k3dTransformObject::SetOrientation(const k3dITransform::ReferenceFrame Frame, const sdpAngleAxis& Orientation)
{
	// Are we recording actions right now?
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_CurrentOrientationChanged && (k3dITransform::CURRENT_FRAME == Frame))
				{
					g_return_if_fail(Editing());
					
					m_CurrentOrientationChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<sdpAngleAxis>(m_Orientation[Frame]));
					changeset->RecordAction(new k3dActionNotify(this, k3dITransform::Event_OrientationChanged, 0, 0));
				}
			else if(!m_ChildOrientationChanged && (k3dITransform::CURRENT_CHILD_FRAME == Frame))
				{
					g_return_if_fail(Editing());
					
					m_ChildOrientationChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<sdpAngleAxis>(m_Orientation[Frame]));
					changeset->RecordAction(new k3dActionNotify(this, k3dITransform::Event_ChildOrientationChanged, 0, 0));
				}
		}

	// Make the change ...
	k3dTransform::SetOrientation(Frame, Orientation);

	// Notify clients ...
	if(k3dITransform::CURRENT_FRAME == Frame)
		Notify(k3dITransform::Event_OrientationChanged, 0);
	else if(k3dITransform::CURRENT_CHILD_FRAME == Frame)
		Notify(k3dITransform::Event_ChildOrientationChanged, 0);
}

void k3dTransformObject::SetScale(const k3dITransform::ReferenceFrame Frame, const sdpVector3& Scale)
{
	// Are we recording actions right now?
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_CurrentScaleChanged && (k3dITransform::CURRENT_FRAME == Frame))
				{
					g_return_if_fail(Editing());
					
					m_CurrentScaleChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_Scale[Frame]));
					changeset->RecordAction(new k3dActionNotify(this, k3dITransform::Event_ScaleChanged, 0, 0));
				}
			else if(!m_ChildScaleChanged && (k3dITransform::CURRENT_CHILD_FRAME == Frame))
				{
					g_return_if_fail(Editing());
					
					m_ChildScaleChanged = true;
					changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_Scale[Frame]));
					changeset->RecordAction(new k3dActionNotify(this, k3dITransform::Event_ChildScaleChanged, 0, 0));
				}
		}

	// Make the change ...
	k3dTransform::SetScale(Frame, Scale);

	// Notify clients ...
	if(k3dITransform::CURRENT_FRAME == Frame)
		Notify(k3dITransform::Event_ScaleChanged, 0);
	else if(k3dITransform::CURRENT_CHILD_FRAME == Frame)
		Notify(k3dITransform::Event_ChildScaleChanged, 0);
}

/////////////////////////////////////////////////////////////////////////////
// k3dISnapTarget implementation ...

sdpResult k3dTransformObject::Snap(sdpVector3& Coordinates, sdpString& Description)
{
	// Convert coordinates to our frame ...
	const sdpVector3 worldcoordinates = k3dGetWorldToObjectMatrix(this) * Coordinates;

	// Snap coordinates ...
	if(worldcoordinates.Length2() > 10)
		return SDP_S_FALSE;

	// Convert coordinates back to the world frame ...
	Coordinates = k3dGetObjectToWorldMatrix(this) * sdpVector3();
	Description = "Object Center Point";

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// k3dIEditable implementation ...

void k3dTransformObject::PushEditing()
{
	if(!Editing())
		{
			m_CurrentPositionChanged = false;
			m_ChildPositionChanged = false;
			m_CurrentOrientationChanged = false;
			m_ChildOrientationChanged = false;
			m_CurrentScaleChanged = false;
			m_ChildScaleChanged = false;	
		}

	k3dEditable::PushEditing();
}

void k3dTransformObject::PopEditing()
{
	k3dEditable::PopEditing();

	if(!Editing())
		{
			// Are we recording actions right now?
			k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_CurrentPositionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_Position[CURRENT_FRAME]));
						
					if(m_ChildPositionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_Position[CURRENT_CHILD_FRAME]));
						
					if(m_CurrentOrientationChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpAngleAxis>(m_Orientation[CURRENT_FRAME]));
						
					if(m_ChildOrientationChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpAngleAxis>(m_Orientation[CURRENT_CHILD_FRAME]));
						
					if(m_CurrentScaleChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_Scale[CURRENT_FRAME]));
						
					if(m_ChildScaleChanged)
						changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_Scale[CURRENT_CHILD_FRAME]));
				}
		}
}

/////////////////////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation ...

sdpResult k3dTransformObject::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	switch(Camera->ActiveTransformType())
		{
			case k3dICamera::POSITION:
				k3dStartStateChangeSet(Document());
				m_TransformType = TRANSLATEFREE;
				break;
			case k3dICamera::ORIENTATION:
				k3dStartStateChangeSet(Document());
				m_TransformType = ROTATEFREE;
				break;
			case k3dICamera::SCALE:
				k3dStartStateChangeSet(Document());
				m_TransformType = SCALEFREE;
				break;
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}

	PushEditing();
	
	return SDP_S_OK;
}

sdpResult k3dTransformObject::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	switch(m_TransformType)
		{
			case NONE:
				break;

			case ROTATEX:
			case ROTATEY:
			case ROTATEZ:
			case ROTATEFREE:
				Rotate(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			case TRANSLATEX:
			case TRANSLATEY:
			case TRANSLATEZ:
			case TRANSLATEFREE:
				Translate(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			case POSITIVESCALEX:
			case NEGATIVESCALEX:
			case POSITIVESCALEY:
			case NEGATIVESCALEY:
			case POSITIVESCALEZ:
			case NEGATIVESCALEZ:
			case SCALEFREE:
				Scale(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}

	return SDP_S_OK;
}

sdpResult k3dTransformObject::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	
	switch(Camera->ActiveTransformType())
		{
			case k3dICamera::POSITION:
				k3dFinishStateChangeSet(Document(), "Position");
				break;
			case k3dICamera::ORIENTATION:
				k3dFinishStateChangeSet(Document(), "Orientation");
				break;
			case k3dICamera::SCALE:
				k3dFinishStateChangeSet(Document(), "Scale");
				break;
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}
	m_TransformType = NONE;

	return SDP_S_OK;
}

void k3dTransformObject::Translate(k3dICamera* const Camera, const GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Camera);
	g_return_if_fail(Camera->Width());
	g_return_if_fail(Camera->Height());

	// If the mouse didn't really move, we're done ...
	if(LastMouse == CurrentMouse)
		return;

	// Turn mouse coordinates into a vector in world coordinates ...
	sdpVector3 lastorigin, lastdirection, currentorigin, currentdirection;
	k3dMouseToWorldRay(Camera, LastMouse, lastorigin, lastdirection);
	k3dMouseToWorldRay(Camera, CurrentMouse, currentorigin, currentdirection);

	// Get our position in world coordinates ...
	sdpVector3 worldposition = k3dGetWorldPosition(this);

	// Calculate the mouse's intersection with a plane, based on the active axis ...
	sdpVector3 lastintersection, currentintersection;

	if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), lastorigin, lastdirection, lastintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dICamera::YAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), lastorigin, lastdirection, lastintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dICamera::ZAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), lastorigin, lastdirection, lastintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					default:
						g_return_if_fail(0);
				}
		}
	else
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), lastorigin, lastdirection, lastintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dICamera::YAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), lastorigin, lastdirection, lastintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dICamera::ZAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), lastorigin, lastdirection, lastintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					default:
						g_return_if_fail(0);
				}
		}

	// Calculate our change in world coordinates ...
	sdpVector3 worlddelta = currentintersection - lastintersection;

	// Constrain movement to the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						worlddelta[1] = worlddelta[2] = 0;
						break;
					case k3dICamera::YAXIS:
						worlddelta[0] = worlddelta[2] = 0;
						break;
					case k3dICamera::ZAXIS:
						worlddelta[0] = worlddelta[1] = 0;
						break;
					default:
						g_return_if_fail(0);
				}
		}

	SetPosition(k3dITransform::CURRENT_FRAME, Position(k3dITransform::CURRENT_FRAME) + worlddelta);
	k3dRedrawAll(Document(), false);
}

static sdpVector3 Trackball(sdpVector2 ScreenCoords)
{
	// Convert screen coordinates to trackball coordinates ...
	sdpVector3 trackball(ScreenCoords, 0);
	gdouble radius = trackball.Length();
	if(radius < 1)
		trackball[2] = -sqrt(1 - radius * radius);
	else
		trackball.Normalize();

	return trackball;
}

void k3dTransformObject::Rotate(k3dICamera* const Camera, const GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Camera);
	g_return_if_fail(Camera->Width());
	g_return_if_fail(Camera->Height());

	// If the mouse didn't really move, we're done ...
	if(LastMouse == CurrentMouse)
		return;

	// Convert mouse coordinates to screen coordinates ...
	sdpVector2 lastmouse, currentmouse;
	gdouble width = Camera->Width();
	gdouble height = Camera->Height();
	if(width > height)
		{
			lastmouse = k3dTransformNScaleMouseCoords(Camera, LastMouse, -width / height, width / height, 1, -1, 2);
			currentmouse = k3dTransformNScaleMouseCoords(Camera, CurrentMouse, -width / height, width / height, 1, -1, 2);
		}
	else
		{
			lastmouse = k3dTransformNScaleMouseCoords(Camera, LastMouse, -1, 1, height / width, -height / width, 2);
			currentmouse = k3dTransformNScaleMouseCoords(Camera, CurrentMouse, -1, 1, height / width, -height / width, 2);
		}

	// Convert screen coordinates to "trackball" coordinates ...
	static const gdouble trackballsize = 1 / 0.9;
	sdpVector3 last = Trackball(lastmouse * trackballsize);
	sdpVector3 current = Trackball(currentmouse * trackballsize);

	// Transform "trackball" coordinates to model coordinates ...
	k3dIObject* parent = Hierarchy()->ObjectParent(this);
	sdpMatrix4 cameratoworld = parent ? k3dGetCameraToObjectMatrix(Camera, parent) : k3dGetCameraToWorldMatrix(Camera);
	last = (cameratoworld * last) - (cameratoworld * sdpVector3(0, 0, 0));
	current = (cameratoworld * current) - (cameratoworld * sdpVector3(0, 0, 0));

	// Convert the change in mouse position to a quaternion ...
	sdpQuaternion delta(last * current, last ^ current);

	// Constrain movement to the active axis ...
	if(ControlDown(Modifiers))
		{
			gdouble w = sdpClamp(-1.0, delta.w, 1.0);
			gdouble phi = acos(w);
			//gdouble phi = acos(delta.w) / 2;
			//delta.w = cos(phi);

			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						if(delta.v.n[0] >= 0)
							delta.v = sdpVector3(sin(phi), 0, 0);
						else
							delta.v = sdpVector3(-sin(phi), 0, 0);
					break;
					case k3dICamera::YAXIS:
						if(delta.v.n[1] >= 0)
							delta.v = sdpVector3(0, sin(phi), 0);
						else
							delta.v = sdpVector3(0, -sin(phi), 0);
					break;
					case k3dICamera::ZAXIS:
						if(delta.v.n[2] >= 0)
							delta.v = sdpVector3(0, 0, sin(phi));
						else
							delta.v = sdpVector3(0, 0, -sin(phi));
					break;
				}
		}

	// "Add" the quaternion to our existing orientation ...
	const sdpAngleAxis currentrotation = Orientation(k3dITransform::CURRENT_FRAME);
	SetOrientation(k3dITransform::CURRENT_FRAME, sdpAngleAxis(delta * sdpQuaternion(currentrotation)));

	k3dRedrawAll(Document(), false);
}

void k3dTransformObject::Scale(k3dICamera* const Camera, const GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Camera);
	g_return_if_fail(Camera->Width());
	g_return_if_fail(Camera->Height());

	// If the mouse didn't really move we're done ...
	if(LastMouse == CurrentMouse)
		return;

	// Convert mouse coordinates to screen coordinates ...
	sdpVector2 lastmouse, currentmouse;
	gdouble width = Camera->Width();
	gdouble height = Camera->Height();
	if(width > height)
		{
			lastmouse = k3dTransformMouseCoords(Camera, LastMouse, -width / height, width / height, 1, -1);
			currentmouse = k3dTransformMouseCoords(Camera, CurrentMouse, -width / height, width / height, 1, -1);
		}
	else
		{
			lastmouse = k3dTransformMouseCoords(Camera, LastMouse, -1, 1, height / width, -height / width);
			currentmouse = k3dTransformMouseCoords(Camera, CurrentMouse, -1, 1, height / width, -height / width);
		}

	const sdpVector2 deltamouse = currentmouse - lastmouse;

	sdpVector3 delta;
	if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						delta = sdpVector3(deltamouse[0], 0, 0);
						break;
					case k3dICamera::YAXIS:
						delta = sdpVector3(0, deltamouse[1], 0);
						break;
					case k3dICamera::ZAXIS:
						delta = sdpVector3(0, 0, deltamouse[0]);
						break;
					default:
						g_return_if_fail(0);
				}
		}
	else
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						delta = sdpVector3(0, deltamouse[1], deltamouse[0]);
						break;
					case k3dICamera::YAXIS:
						delta = sdpVector3(deltamouse[0], 0, deltamouse[1]);
						break;
					case k3dICamera::ZAXIS:
						delta = sdpVector3(deltamouse[0], deltamouse[1], 0);
						break;
					default:
						g_return_if_fail(0);
				}
		}

	SetScale(k3dITransform::CURRENT_FRAME, k3dTransform::Scale(k3dITransform::CURRENT_FRAME) + delta);
	k3dRedrawAll(Document(), false);
}

// Draw a single transformation handle
static void DrawTransformationHandle(k3dIEditorEngine* const Engine, GLUquadricObj* Quadric, double Radius, GLfloat Color[3], GLfloat EmissionColor[3], SDPCSTRING AxisLabel)
{
	// Draw axis labels ...
	glDisable(GL_LIGHTING);
	
	static GLfloat labelcolor[3] = { 0.3, 0.3, 0.3 };
	glColor3fv(labelcolor);

	gdouble labelposition = Radius * 1.1;

	glListBase(Engine->EditorFontBase());

	sdpString positivelabel = sdpString("+") + AxisLabel;
	glRasterPos3d(0, 0, labelposition);
	glCallLists(positivelabel.size(), GL_UNSIGNED_BYTE, (GLubyte*)positivelabel.c_str());

	sdpString negativelabel = sdpString("-") + AxisLabel;
	glRasterPos3d(0, 0, -labelposition);
	glCallLists(negativelabel.size(), GL_UNSIGNED_BYTE, (GLubyte*)negativelabel.c_str());

	// Draw the circle ...
	glColor3fv(Color);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, Color);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, EmissionColor);

	gluQuadricDrawStyle(Quadric, GLenum(GLU_SILHOUETTE));
	gluQuadricNormals(Quadric, GLenum(GLU_SMOOTH));
	gluDisk(Quadric, 0.0, Radius, 32, 1);

	// Draw the axis ...
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, -Radius);
	glVertex3d(0.0, 0.0, Radius);
	glEnd();

	// Draw the arrows ...
	glEnable(GL_LIGHTING);
	gluQuadricDrawStyle(Quadric, GLenum(GLU_FILL));

	// Draw the positive-direction arrow ...
	glPushMatrix();
	glTranslated(0.0, 0.0, Radius);
	gluCylinder(Quadric, Radius * 0.03, Radius * 0.001, Radius * 0.15, 8, 1);
	glPopMatrix();

	// Draw the negative-direction arrow ...
	glPushMatrix();
	glTranslated(0.0, 0.0, -Radius);
	glRotated(180.0, 1.0, 0.0, 0.0);
	gluCylinder(Quadric, Radius * 0.03, Radius * 0.001, Radius * 0.15, 8, 1);
	glPopMatrix();
}

// Draw all three transformation handles
void k3dTransformObject::DrawTransformationHandles(k3dIEditorEngine* const Engine, gdouble Radius)
{
	StoreOpenGLAttributes attributes;

	// Setup solid, untextured, lit rendering ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	// Create a quadric object for drawing arrows (cones) ...
	GLUquadricObj* const quadric = gluNewQuadric();
	g_return_if_fail(quadric);

	glMatrixMode(GL_MODELVIEW);

	// X axis
	static GLfloat xcolor[3] = { 1, 0, 0 };
	static GLfloat xemissioncolor[3] = { 0.5, 0, 0 };
	glPushMatrix();
	glRotated(90.0, 0.0, 1.0, 0.0);
	DrawTransformationHandle(Engine, quadric, Radius, xcolor, xemissioncolor, "X");
	glPopMatrix();

	// Y axis
	static GLfloat ycolor[3] = { 1, 1, 0 };
	static GLfloat yemissioncolor[3] = { 0.5, 0.5, 0 };
	glPushMatrix();
	glRotated(-90.0, 1.0, 0.0, 0.0);
	DrawTransformationHandle(Engine, quadric, Radius, ycolor, yemissioncolor, "Y");
	glPopMatrix();

	// Z axis
	static GLfloat zcolor[3] = { 0, 1, 0 };
	static GLfloat zemissioncolor[3] = { 0, 0.5, 0 };
	DrawTransformationHandle(Engine, quadric, Radius, zcolor, zemissioncolor, "Z");

	gluDeleteQuadric(quadric);
}

void k3dTransformObject::DrawTransformationHandles(k3dIEditorEngine* const Engine, k3dILocationCollection* LocationCollection, gdouble MinimumRadius)
{
	// Sanity checks ...
	g_assert(Engine);
	g_assert(LocationCollection);

	if(0 == LocationCollection->LocationCount())
		{
			DrawTransformationHandles(Engine, MinimumRadius);
			return;
		}

	sdpVector3 min;
	sdpVector3 max;
	k3dGetDimensions(LocationCollection, 0, min, max);
	min.Abs();
	max.Abs();

	sdpVector3 size(sdpMax(min[0], max[0]), sdpMax(min[1], max[1]), sdpMax(min[2], max[2]));

	DrawTransformationHandles(Engine, sdpMax(MinimumRadius, size.Length()));
}

void k3dTransformObject::DrawBoundingBox(k3dIEditorEngine* const Engine, const gdouble Radius)
{
	// Sanity checks ...
	g_assert(Engine);

	const gdouble radius = fabs(Radius);

	DrawBoundingBox(Engine, k3dBoundingBox3(sdpVector3(-radius, -radius, -radius), sdpVector3(radius, radius, radius)));
}

void k3dTransformObject::DrawBoundingBox(k3dIEditorEngine* const Engine, k3dILocationCollection* const LocationCollection, const gdouble MinimumRadius)
{
	// Sanity checks ...
	g_assert(Engine);
	g_assert(LocationCollection);
}

void k3dTransformObject::DrawBoundingBox(k3dIEditorEngine* const Engine, const k3dBoundingBox3 Box)
{
	// Sanity checks ...
	g_assert(Engine);

	StoreOpenGLAttributes attributes;

	// Setup wireframe, untextured, unlit rendering ...
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	
	glColor3d(0, 0, 0);
	
	glBegin(GL_LINE_LOOP);
	glVertex3d(Box.Min()[0], Box.Min()[1], Box.Min()[2]);
	glVertex3d(Box.Max()[0], Box.Min()[1], Box.Min()[2]);
	glVertex3d(Box.Max()[0], Box.Min()[1], Box.Max()[2]);
	glVertex3d(Box.Min()[0], Box.Min()[1], Box.Max()[2]);
	glEnd();	
	
	glBegin(GL_LINE_LOOP);
	glVertex3d(Box.Min()[0], Box.Max()[1], Box.Min()[2]);
	glVertex3d(Box.Max()[0], Box.Max()[1], Box.Min()[2]);
	glVertex3d(Box.Max()[0], Box.Max()[1], Box.Max()[2]);
	glVertex3d(Box.Min()[0], Box.Max()[1], Box.Max()[2]);
	glEnd();
	
	glBegin(GL_LINES);
	glVertex3d(Box.Min()[0], Box.Min()[1], Box.Min()[2]);
	glVertex3d(Box.Min()[0], Box.Max()[1], Box.Min()[2]);
	
	glVertex3d(Box.Max()[0], Box.Min()[1], Box.Min()[2]);
	glVertex3d(Box.Max()[0], Box.Max()[1], Box.Min()[2]);
	
	glVertex3d(Box.Max()[0], Box.Min()[1], Box.Max()[2]);
	glVertex3d(Box.Max()[0], Box.Max()[1], Box.Max()[2]);

	glVertex3d(Box.Min()[0], Box.Min()[1], Box.Max()[2]);
	glVertex3d(Box.Min()[0], Box.Max()[1], Box.Max()[2]);
	glEnd();

	glBegin(GL_LINE_STRIP);
	glVertex3d(Box.Min()[0], Box.Min()[1], Box.Min()[2]);
	glVertex3d(0.5 * (Box.Min()[0] + Box.Max()[0]), Box.Min()[1], Box.Max()[2]);
	glVertex3d(Box.Max()[0], Box.Min()[1], Box.Min()[2]);
	glEnd();
}


