#ifndef K3DTRANSFORMOBJECT_H
#define K3DTRANSFORMOBJECT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dTransformObject class, which serves as a base class for K-3D objects that have position, orientation, scale, and can render themselves
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dboundingbox.h>
#include <k3deditable.h>
#include <k3deditorrender.h>
#include <k3dmouseeventobserver.h>
#include <k3dpersistobject.h>
#include <k3drirender.h>
#include <k3dtransform.h>
#include <k3dvariable.h>

#include <k3disnaptarget.h>

// Forward class references ...
class k3dILocationCollection;

/////////////////////////////////////////////////////////////////////////////////////////
// k3dTransformObject

typedef k3dPersistObject k3dTransformObjectBase;

/// A base class for K-3D objects that have position, orientation, scale, and can render themselves
class k3dTransformObject :
	public k3dTransformObjectBase,
	public k3dTransform,
	public k3dEditorRender,
	public k3dMouseEventObserver,
	public k3dRiRender,
	public k3dISnapTarget,
	public k3dEditable
{
public:
	k3dTransformObject(sdpIUnknown* Initializer);
	virtual ~k3dTransformObject();

	// k3dIPersist implementation
	sdpResult Load(sdpxml::Document& Document, sdpxml::Element& Element);
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);

	// k3dIEditorRender implementation
	sdpResult EditorPushRender(k3dIEditorEngine* const Engine);
	sdpResult EditorPushChildRender(k3dIEditorEngine* const Engine);
	sdpResult EditorPopChildRender(k3dIEditorEngine* const Engine);
	sdpResult EditorPopRender(k3dIEditorEngine* const Engine);

	// k3dIRiRender implementation
	sdpResult RiPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiPushRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiPushChildRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);

	// k3dITransform implementation
	const sdpAngleAxis& Orientation(const ReferenceFrame Frame);
	void SetPosition(const ReferenceFrame Frame, const sdpVector3& Position);
	void SetOrientation(const ReferenceFrame Frame, const sdpAngleAxis& Orientation);
	void SetScale(const ReferenceFrame Frame, const sdpVector3& Scale);

	// k3dIMouseEventObserver implementation
	sdpResult OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

	// k3dISnapTarget implementation
	sdpResult Snap(sdpVector3& Coordinates, sdpString& Description);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

protected:
	/// Draws a standard set of "handles" with which the user can interactively transform the object
	void DrawTransformationHandles(k3dIEditorEngine* const Engine, const gdouble Radius);
	/// Draws a standard set of "handles" with which the user can interactively transform the object
	void DrawTransformationHandles(k3dIEditorEngine* const Engine, k3dILocationCollection* const LocationCollection, const gdouble MinimumRadius);
	
	/// Draws a standard bounding box to make it easier to visualize an object
	void DrawBoundingBox(k3dIEditorEngine* const Engine, const gdouble Radius);
	/// Draws a standard bounding box to make it easier to visualize an object
	void DrawBoundingBox(k3dIEditorEngine* const Engine, k3dILocationCollection* const LocationCollection, const gdouble MinimumRadius);
	/// Draws a standard bounding box to make it easier to visualize an object
	void DrawBoundingBox(k3dIEditorEngine* const Engine, const k3dBoundingBox3 Box);
	
	// Interactive transformations ...
	typedef enum
	{
		NONE,
		ROTATEX,
		ROTATEY,
		ROTATEZ,
		ROTATEFREE,
		TRANSLATEX,
		TRANSLATEY,
		TRANSLATEZ,
		TRANSLATEFREE,
		POSITIVESCALEX,
		NEGATIVESCALEX,
		POSITIVESCALEY,
		NEGATIVESCALEY,
		POSITIVESCALEZ,
		NEGATIVESCALEZ,
		SCALEFREE
	} TransformType;

	TransformType m_TransformType;
	
	k3dObjectInterfaceVariable<k3dITransform> m_Target;
	sdpAngleAxis m_TargetOrientation;

	bool m_CurrentPositionChanged;
	bool m_ChildPositionChanged;
	
	bool m_CurrentOrientationChanged;
	bool m_ChildOrientationChanged;
	
	bool m_CurrentScaleChanged;
	bool m_ChildScaleChanged;

	void Rotate(k3dICamera* const Camera, const GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void Translate(k3dICamera* const Camera, const GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void Scale(k3dICamera* const Camera, const GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
};

#endif // K3DTRANSFORMOBJECT_H
