// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTriangleMesh class, which efficiently encapsulates an implicit/procedural surface composed of triangles
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "k3dtrianglemesh.h"

#include <k3distaterecorder.h>
#include <k3dieditorengine.h>
#include <k3dieditormaterial.h>
#include <k3dimaterial.h>
#include <k3dirimaterial.h>
#include <k3drishadersetup.h>
#include <k3didocument.h>
#include <k3dboundingbox.h>
#include <k3dmisc.h>

#include <strstream>

//////////////////////////////////////////////////////////////////////////
// k3dTriangleMesh

// Static member variables
short k3dTriangleMesh::m_RiBilinearPatchLookup[4] = {0, 1, 3, 2};

sdpVector3 k3dTriangleMesh::m_SelectedColor = 1.0;
sdpVector3 k3dTriangleMesh::m_DefaultColor = 0.0;

k3dTriangleMesh::k3dTriangleMesh(sdpIUnknown* Initializer) :
	k3dTriangleMeshBase(Initializer),
	m_Color("color", m_DefaultColor, Document()->StateRecorder(), 0),
	m_Material("material", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_Color);
	AddVariable(&m_Material);
}

k3dTriangleMesh::~k3dTriangleMesh()
{
}

void k3dTriangleMesh::ReserveTriangles(gulong TriangleCount)
{
	m_Triangles.resize(TriangleCount);
}

void k3dTriangleMesh::SetFlatShading()
{
	m_SmoothShading = false;
}

void k3dTriangleMesh::SetSmoothShading()
{
	m_SmoothShading = true;
}

void k3dTriangleMesh::CreateTriangleGrid(const gulong XSegments, const gulong YSegments)
{
	// Sanity checks ...
	g_assert(XSegments);
	g_assert(YSegments);

	const gulong xpoints = XSegments+1;
	const gulong ypoints = YSegments+1;

	m_Locations.clear();
	m_Locations.resize(m_Locations.size() + (xpoints * ypoints));
	m_Triangles.clear();
	m_Triangles.resize(m_Triangles.size() + (XSegments * YSegments * 2));

	gulong triangleindex = 0;
	for(gulong y = 0; y < YSegments; y++)
		{
			for(gulong x = 0; x < XSegments; x++)
				{
					Triangle(triangleindex++).SetPoints(y*xpoints + x, y*xpoints + x+1, (y+1)*xpoints + x+1);
					Triangle(triangleindex++).SetPoints(y*xpoints + x, (y+1)*xpoints + x+1, (y+1)*xpoints + x);
				}
		}
}

void k3dTriangleMesh::CalculateTriangleNormals()
{
	sdpVector3 edge1, edge2;
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			edge1 = m_Locations[i->m_Points[1]]->m_AnimXYZ - m_Locations[i->m_Points[0]]->m_AnimXYZ;
			edge2 = m_Locations[i->m_Points[2]]->m_AnimXYZ - m_Locations[i->m_Points[0]]->m_AnimXYZ;

			i->m_Normal = edge1 ^ edge2;
			i->m_Normal.Normalize();
		}
}

void k3dTriangleMesh::CalculateLocationNormals()
{
	// Start by calculating all of the triangle normals ...
	CalculateTriangleNormals();

	// Zero out each location's normal ...
	LocationIterator locationend = m_Locations.end();
	for(LocationIterator location = m_Locations.begin(); location != locationend; location++)
		(*location)->m_NormalXYZ = 0.0;

	// For each triangle ...
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator triangle = m_Triangles.begin(); triangle != triangleend; triangle++)
		{
			// Add the triangle's normal to the location ...
			m_Locations[triangle->m_Points[0]]->m_NormalXYZ += triangle->m_Normal;
			m_Locations[triangle->m_Points[1]]->m_NormalXYZ += triangle->m_Normal;
			m_Locations[triangle->m_Points[2]]->m_NormalXYZ += triangle->m_Normal;
		}
}

////////////////////////////////////////////////////////
// k3dIEditorRender overrides ...

sdpResult k3dTriangleMesh::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dTriangleMeshBase::InitNew()), SDP_E_FAIL);

	// Always start ourselves out with the default material assigned ...
	g_return_val_if_fail(sdpSucceeded(SetMaterial(k3dDefaultMaterial(Document()))), SDP_E_FAIL);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////
// k3dIEditorRender overrides ...

sdpResult k3dTriangleMesh::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	StoreOpenGLAttributes attributes;

	// If solid rendering has been requested ...
	if(SDP_S_OK == Engine->EditorQuerySolid())
		{
			// Setup the material ...
			k3dIMaterial* material = m_Material.Interface();
			if(material)
				{
					k3dIEditorMaterial* editormaterial = dynamic_cast<k3dIEditorMaterial*>(material);
					if(editormaterial)
						{
							editormaterial->EditorMaterialSetup(Engine);

							if(SDP_S_OK ==  Engine->EditorQueryTextured())
								EditorRenderSolidTextured(Engine);
							else
								EditorRenderSolid(Engine);
						}
				}
			else
				{
					// This is our fallback position if we don't have an assigned material ...
					EditorRenderWireframe(Engine);
				}
		}
	else
		{
			// The render engine only wants wireframe rendering
			EditorRenderWireframe(Engine);
		}

//	if(k3dICamera::POINT == Camera->ActiveEditMode())
		EditorRenderPoints(Engine);

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, this, 1.0);

	return SDP_S_OK;
}

void k3dTriangleMesh::EditorRenderWireframe(k3dIEditorEngine* const Engine)
{
	// Setup for wireframe paths ...
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	// For each triangle ...
	TriangleIterator end = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != end; i++)
		{
			glBegin(GL_LINE_LOOP);
			glVertex3dv(m_Locations[i->m_Points[0]]->m_AnimXYZ);
			glVertex3dv(m_Locations[i->m_Points[1]]->m_AnimXYZ);
			glVertex3dv(m_Locations[i->m_Points[2]]->m_AnimXYZ);
			glEnd();
		}
}

void k3dTriangleMesh::EditorRenderSolid(k3dIEditorEngine* const Engine)
{
	if(m_SmoothShading)
		{
			CalculateLocationNormals();
			EditorRenderSolidSmooth(Engine);
		}
	else
		{
			CalculateTriangleNormals();
			EditorRenderSolidFlat(Engine);
		}
}

void k3dTriangleMesh::EditorRenderSolidFlat(k3dIEditorEngine* const Engine)
{
	// Triangles ...
	glBegin(GL_TRIANGLES);
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			glNormal3dv(i->m_Normal);

			glVertex3dv(m_Locations[i->m_Points[0]]->m_AnimXYZ);
			glVertex3dv(m_Locations[i->m_Points[1]]->m_AnimXYZ);
			glVertex3dv(m_Locations[i->m_Points[2]]->m_AnimXYZ);
		}
	glEnd();
}

void k3dTriangleMesh::EditorRenderSolidSmooth(k3dIEditorEngine* const Engine)
{
	// Triangles ...
	glBegin(GL_TRIANGLES);
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			glNormal3dv(m_Locations[i->m_Points[0]]->m_NormalXYZ);
			glVertex3dv(m_Locations[i->m_Points[0]]->m_AnimXYZ);
			glNormal3dv(m_Locations[i->m_Points[1]]->m_NormalXYZ);
			glVertex3dv(m_Locations[i->m_Points[1]]->m_AnimXYZ);
			glNormal3dv(m_Locations[i->m_Points[2]]->m_NormalXYZ);
			glVertex3dv(m_Locations[i->m_Points[2]]->m_AnimXYZ);
		}
	glEnd();
}

void k3dTriangleMesh::EditorRenderSolidTextured(k3dIEditorEngine* const Engine)
{
	if(m_SmoothShading)
		{
			CalculateLocationNormals();
			EditorRenderSolidTexturedSmooth(Engine);
		}
	else
		{
			CalculateTriangleNormals();
			EditorRenderSolidTexturedFlat(Engine);
		}
}

void k3dTriangleMesh::EditorRenderSolidTexturedFlat(k3dIEditorEngine* const Engine)
{
	glBegin(GL_TRIANGLES);

	TriangleIterator end = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != end; i++)
		{
			glNormal3dv(i->m_Normal);

			glTexCoord3dv(m_Locations[i->m_Points[0]]->m_TextureXYZ);
			glVertex3dv(m_Locations[i->m_Points[0]]->m_AnimXYZ);
			glTexCoord3dv(m_Locations[i->m_Points[1]]->m_TextureXYZ);
			glVertex3dv(m_Locations[i->m_Points[1]]->m_AnimXYZ);
			glTexCoord3dv(m_Locations[i->m_Points[2]]->m_TextureXYZ);
			glVertex3dv(m_Locations[i->m_Points[2]]->m_AnimXYZ);
		}

	glEnd();
}

void k3dTriangleMesh::EditorRenderSolidTexturedSmooth(k3dIEditorEngine* const Engine)
{
	glBegin(GL_TRIANGLES);

	TriangleIterator end = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != end; i++)
		{
			glNormal3dv(m_Locations[i->m_Points[0]]->m_NormalXYZ);
			glTexCoord3dv(m_Locations[i->m_Points[0]]->m_TextureXYZ);
			glVertex3dv(m_Locations[i->m_Points[0]]->m_AnimXYZ);
			glNormal3dv(m_Locations[i->m_Points[1]]->m_NormalXYZ);
			glTexCoord3dv(m_Locations[i->m_Points[1]]->m_TextureXYZ);
			glVertex3dv(m_Locations[i->m_Points[1]]->m_AnimXYZ);
			glNormal3dv(m_Locations[i->m_Points[2]]->m_NormalXYZ);
			glTexCoord3dv(m_Locations[i->m_Points[2]]->m_TextureXYZ);
			glVertex3dv(m_Locations[i->m_Points[2]]->m_AnimXYZ);
		}

	glEnd();
}

void k3dTriangleMesh::EditorRenderPoints(k3dIEditorEngine* const Engine)
{
	glBegin(GL_POINTS);

	LocationIterator end = m_Locations.end();
	for(LocationIterator location = m_Locations.begin(); location != end; location++)
		glVertex3dv((*location)->m_AnimXYZ);

	glEnd();
}

///////////////////////////////////////////////////////////////
// k3dIMaterialServer overrides ...

k3dIMaterial* k3dTriangleMesh::Material()
{
	return m_Material.Interface();
}

sdpResult k3dTriangleMesh::SetMaterial(k3dIMaterial* Material)
{
	// We allow NULL here, as a synonym for the default material ...
	m_Material.SetObject(Material);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIRiRender overrides

sdpResult k3dTriangleMesh::RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	// Setup our material interfaces ...
	k3dIMaterial* material = m_Material.Interface();
	if(material)
		{
			k3dIRiMaterial* rimaterial = dynamic_cast<k3dIRiMaterial*>(material);
			if(rimaterial)
				{
					// Setup the material ...
					rimaterial->RiMaterialSetup(Frame, Engine);

					// Render each triangle ...
					TriangleIterator triangleend = m_Triangles.end();
					for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
						RiTriangle(Engine, *i);
				}
		}

	return SDP_S_OK;
}

// Render a triangle as a bilinear patch ...
void k3dTriangleMesh::RiTriangle(k3dIRiEngine* const Engine, k3dTriangle& Triangle)
{
	// Arrays of coordinates ...
	k3dRiRealArray* points = new k3dRiRealArray("point");
	k3dRiRealArray* normals = new k3dRiRealArray("normal");
//	k3dRiRealArray* scoords = new k3dRiRealArray("float");
//	k3dRiRealArray* tcoords = new k3dRiRealArray("float");

	for(gulong i = 0; i < 3; i++)
		{
			points->Append(m_Locations[Triangle.m_Points[i]]->m_AnimXYZ);
			normals->Append(m_SmoothShading ? m_Locations[Triangle.m_Points[i]]->m_NormalXYZ : Triangle.m_Normal);
//			scoords->Append(pathpoint->m_TextureXYZ[0]);
//			tcoords->Append(pathpoint->m_TextureXYZ[1]);
		}

	points->Append(m_Locations[Triangle.m_Points[1]]->m_AnimXYZ);
	normals->Append(m_SmoothShading ? m_Locations[Triangle.m_Points[1]]->m_NormalXYZ : Triangle.m_Normal);
//	scoords->Append(middlepoint->m_TextureXYZ[0]);
//	tcoords->Append(middlepoint->m_TextureXYZ[1]);

	k3dRiParameterList parameters;
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_P, points));
	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_N, normals));
//	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_S, scoords));
//	parameters.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_T, tcoords));

	Engine->RiPatchV(k3dIRiEngine::RI_BILINEAR, parameters.ParameterCount(), parameters.Tokens(), parameters.Values(), parameters.ValueCounts());
}

