// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dUserInterfaceElement class, which provides a standard command-handling for GTK+ controls / dialogs / windows
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3duserinterfaceelement.h>

#include <k3dchannelbutton.h>
#include <k3dcheckbutton.h>
#include <k3dchooser.h>
#include <k3dobjectselector.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>
#include <k3dshaderinfo.h>
#include <k3dshaderselector.h>
#include <k3dshaderargumentspage.h>
#include <k3dspinbutton.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>

const sdpString k3dUserInterfaceElement::control_highlight = "highlight";

k3dUserInterfaceElement::k3dUserInterfaceElement(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName) :
	k3dCommandNode(ParentCommandNode, CommandNodeName)
{
}

k3dUserInterfaceElement::~k3dUserInterfaceElement()
{
}

////////////////////////////////////////////////////////////////////////
// k3dICommandNode overrides

void k3dUserInterfaceElement::AddCommand(SDPCSTRING CommandName, SDPCSTRING ControlName, const CommandControlType& ControlType)
{
	// Sanity checks ...
	g_assert_string(CommandName);
	g_assert_string(ControlName);

	// Enforce the unique-ness of entries ...
	if(CommandName == control_highlight)
		{
			std::cerr << "k3dUserInterfaceElement::AddCommand(" << CommandName << ", " << ControlName << ") - attempt to add reserved action will be ignored" << std::endl;
			g_return_if_fail(0);
		}

	if(m_CommandEntries.find(CommandName) != m_CommandEntries.end())
		{
			std::cerr << "k3dUserInterfaceElement::AddCommand(" << CommandName << ", " << ControlName << ") - attempt to add duplicate action will be ignored" << std::endl;
			g_return_if_fail(0);
		}

	// Make sure the control exists and is the correct type ...
	switch(ControlType)
		{
			case CONTROL_BUTTON:
				{
					sdpGtkButton button = Button(ControlName);
					g_return_if_fail(button.Attached());
				}
				break;
			case CONTROL_TOGGLE_BUTTON:
				{
					sdpGtkToggleButton togglebutton = ToggleButton(ControlName);
					g_return_if_fail(togglebutton.Attached());
				}
				break;
			case CONTROL_MENU_ITEM:
				{
					sdpGtkMenuItem menuitem = MenuItem(ControlName);
					g_return_if_fail(menuitem.Attached());
				}
				break;
			case CONTROL_EDITABLE:
				{
					sdpGtkEditable editable = Editable(ControlName);
					g_return_if_fail(editable.Attached());
				}
				break;
			case CONTROL_COMBO:
				{
					sdpGtkCombo combo = Combo(ControlName);
					g_return_if_fail(combo.Attached());
				}
				break;
			case CONTROL_NONE:
				break;
		}

	m_CommandEntries[CommandName] = CommandEntry(ControlName, ControlType);
}

sdpResult k3dUserInterfaceElement::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
//std::cout << "k3dUserInterfaceElement::Command(" << Name << ", " << Arguments << ")" << std::endl;

	// Sanity checks ...
	g_assert_string(Name);
	
	// If this is a highlight action ...
	if(Name == control_highlight)
		{
			InteractiveHighlight(Arguments, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
			return SDP_S_OK;
		}
	
	// If this is a user control action ...
	CommandIterator commanditerator = m_CommandEntries.find(Name);
	if(commanditerator != m_CommandEntries.end())
		{
			CommandEntry& entry = commanditerator->second;

			switch(entry.ControlType())
				{
					case CONTROL_BUTTON:
						return InteractiveActivateButton(entry.ControlName(), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
					case CONTROL_TOGGLE_BUTTON:
						return InteractiveToggleButton(entry.ControlName(), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
					case CONTROL_MENU_ITEM:
						return InteractiveActivateMenuItem(entry.ControlName(), k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
					case CONTROL_EDITABLE:
						return InteractiveSetText(entry.ControlName(), Arguments, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true) ? SDP_S_OK : SDP_E_FAIL;
					case CONTROL_COMBO:
						std::cout << "k3dUserInterfaceElement::Command(" << Name << ", " << Arguments << "): CONTROL_COMBO not currently handled!" << std::endl;
						break;
					case CONTROL_NONE:
						break;
				}
				
			g_return_val_if_fail(0, SDP_E_FAIL);	// Did you add a new control action type without telling me?
		}

	return k3dCommandNode::Command(Name, Arguments, Result);
}

void k3dUserInterfaceElement::ClearCommands()
{
	m_CommandEntries.clear();
}

sdpGtkObject* k3dUserInterfaceElement::CreateCustomObject(sdpxml::Element& Element)
{
	// Sanity checks ...
	if(Element.Name() == "k3dchannelbutton")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dChannelButton(this, name);
		}
	else if(Element.Name() == "k3dcheckbutton")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dCheckButton(this, name);
		}
	else if(Element.Name() == "k3dchooser")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dChooser(this, name);
		}
	else if(Element.Name() == "k3dobjectselector")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dObjectSelector(this, name);
		}
	else if(Element.Name() == "k3dshaderinfo")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dShaderInfo(this, name);
		}
	else if(Element.Name() == "k3dshaderselector")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dShaderSelector(this, name);
		}
	else if(Element.Name() == "k3dshaderargumentspage")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dShaderArgumentsPage(this, name);
		}
	else if(Element.Name() == "k3dspinbutton")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dSpinButton(this, name);
		}
	else if(Element.Name() == "k3dpositioncontroller")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dPositionController(this, name);
		}
	else if(Element.Name() == "k3dorientationcontroller")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dOrientationController(this, name);
		}
	else if(Element.Name() == "k3dscalecontroller")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dScaleController(this, name);
		}
	else if(Element.Name() == "k3dtargetcontroller")
		{
			sdpString name;
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), 0);
			
			return new k3dTargetController(this, name);
		}
	
	return sdpGtkObjectContainer::CreateCustomObject(Element);
}

// Custom control access ...
k3dChannelButton* k3dUserInterfaceElement::Getk3dChannelButton(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dChannelButton", 0);

	return static_cast<k3dChannelButton*>(object);
}

k3dCheckButton* k3dUserInterfaceElement::Getk3dCheckButton(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dCheckButton", 0);

	return static_cast<k3dCheckButton*>(object);
}

k3dChooser* k3dUserInterfaceElement::Getk3dChooser(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dChooser", 0);

	return static_cast<k3dChooser*>(object);
}

k3dObjectSelector* k3dUserInterfaceElement::Getk3dObjectSelector(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dObjectSelector", 0);

	return static_cast<k3dObjectSelector*>(object);
}

k3dShaderInfo* k3dUserInterfaceElement::Getk3dShaderInfo(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dShaderInfo", 0);

	return static_cast<k3dShaderInfo*>(object);
}

k3dShaderSelector* k3dUserInterfaceElement::Getk3dShaderSelector(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dShaderSelector", 0);

	return static_cast<k3dShaderSelector*>(object);
}

k3dShaderArgumentsPage* k3dUserInterfaceElement::Getk3dShaderArgumentsPage(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dShaderArgumentsPage", 0);

	return static_cast<k3dShaderArgumentsPage*>(object);
}

k3dSpinButton* k3dUserInterfaceElement::Getk3dSpinButton(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dSpinButton", 0);

	return static_cast<k3dSpinButton*>(object);
}

k3dPositionController* k3dUserInterfaceElement::Getk3dPositionController(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dPositionController", 0);

	return static_cast<k3dPositionController*>(object);
}

k3dOrientationController* k3dUserInterfaceElement::Getk3dOrientationController(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dOrientationController", 0);

	return static_cast<k3dOrientationController*>(object);
}

k3dScaleController* k3dUserInterfaceElement::Getk3dScaleController(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dScaleController", 0);

	return static_cast<k3dScaleController*>(object);
}

k3dTargetController* k3dUserInterfaceElement::Getk3dTargetController(SDPCSTRING Name)
{
	// Sanity checks ...
	g_assert_string(Name);

	sdpGtkObject* object = CustomObject(Name);
	g_return_val_if_fail(object, 0);
	g_return_val_if_fail(object->CustomType() == "k3dTargetController", 0);

	return static_cast<k3dTargetController*>(object);
}

bool k3dUserInterfaceElement::Connectk3dChannelButton(SDPCSTRING Name, k3dIChannel* const Channel)
{
	k3dChannelButton* button = Getk3dChannelButton(Name);
	g_return_val_if_fail(button, false);
	
	g_return_val_if_fail(button->Attach(Channel), false);
	
	return true;
}

bool k3dUserInterfaceElement::Connectk3dCheckButton(SDPCSTRING Name, k3dVariable<bool>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	k3dCheckButton* const button = Getk3dCheckButton(Name);
	g_return_val_if_fail(button, false);
	
	g_return_val_if_fail(button->Attach(Variable, CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);
	
	return true;
}

bool k3dUserInterfaceElement::Connectk3dSpinButton(SDPCSTRING Name, k3dVariable<gdouble>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction)
{
	k3dSpinButton* const button = Getk3dSpinButton(Name);
	g_return_val_if_fail(button, false);
	
	g_return_val_if_fail(button->Attach(Variable, CommandRecorder, StateRecorder, StateChangeName, PushEditingAction, ChangeAction, PopEditingAction), false);
	
	return true;
}


