#ifndef K3DUSERINTERFACEELEMENT_H
#define K3DUSERINTERFACEELEMENT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dUserInterfaceElement class, which provides a standard functionality for controls, dialogs, and windows, including automated command-handling
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcommandnode.h>
#include <k3dvariable.h>

#include <sdpgtkobjectcontainer.h>
#include <map>

// Forward class declarations ...
class k3dControlChangeHandler;
class k3dChannelButton;
class k3dCheckButton;
class k3dChooser;
class k3dObjectSelector;
class k3dShaderInfo;
class k3dShaderSelector;
class k3dShaderArgumentsPage;
class k3dSpinButton;

class k3dPositionController;
class k3dOrientationController;
class k3dScaleController;
class k3dTargetController;

class k3dIAction;
class k3dIChannel;
class k3dICommandRecorder;
class k3dIStateRecorder;

///////////////////////////////////////////////////////////////////////////
// k3dUserInterfaceElement

typedef sdpGtkObjectContainer k3dUserInterfaceElementBase;

class k3dUserInterfaceElement :
	public k3dUserInterfaceElementBase,
	public k3dCommandNode
{
public:
	k3dUserInterfaceElement(sdpIUnknown* ParentCommandNode, SDPCSTRING CommandNodeName);
	virtual ~k3dUserInterfaceElement();
	
	// k3dICommandNode overrides
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

protected:
	/// Enumerates the types of SDPGTK controls that can be interactively driven by user actions (macros)
	typedef enum
	{
		CONTROL_BUTTON,
		CONTROL_TOGGLE_BUTTON,
		CONTROL_MENU_ITEM,
		CONTROL_EDITABLE,
		CONTROL_COMBO,
		CONTROL_NONE,
	} CommandControlType;

	/// Stores a user action that can be interactively driven by user actions (macros)
	void AddCommand(SDPCSTRING CommandName, SDPCSTRING ControlName, const CommandControlType& ControlType);
	
	/// Removes all commands
	void ClearCommands();

        /// Allows us to instantiate UI elements using custom tags in the GTKML source document
	sdpGtkObject* CreateCustomObject(sdpxml::Element& Element);
	
	// Custom control access ...
	k3dChannelButton* Getk3dChannelButton(SDPCSTRING Name);
	k3dCheckButton* Getk3dCheckButton(SDPCSTRING Name);
	k3dChooser* Getk3dChooser(SDPCSTRING Name);
	k3dObjectSelector* Getk3dObjectSelector(SDPCSTRING Name);
	k3dShaderInfo* Getk3dShaderInfo(SDPCSTRING Name);
	k3dShaderSelector* Getk3dShaderSelector(SDPCSTRING Name);
	k3dShaderArgumentsPage* Getk3dShaderArgumentsPage(SDPCSTRING Name);
	k3dSpinButton* Getk3dSpinButton(SDPCSTRING Name);
	
	k3dPositionController* Getk3dPositionController(SDPCSTRING Name);
	k3dOrientationController* Getk3dOrientationController(SDPCSTRING Name);
	k3dScaleController* Getk3dScaleController(SDPCSTRING Name);
	k3dTargetController* Getk3dTargetController(SDPCSTRING Name);
	
	// Simplified control initialization ...
	bool Connectk3dChannelButton(SDPCSTRING Name, k3dIChannel* const Channel);
	bool Connectk3dCheckButton(SDPCSTRING Name, k3dVariable<bool>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);
	bool Connectk3dSpinButton(SDPCSTRING Name, k3dVariable<gdouble>& Variable, k3dICommandRecorder* const CommandRecorder, k3dIStateRecorder* const StateRecorder, SDPCSTRING StateChangeName, k3dIAction* const PushEditingAction, k3dIAction* const ChangeAction, k3dIAction* const PopEditingAction);
	
private:
	/// Internal class for storing commands
	class CommandEntry
	{
	public:
		CommandEntry()
		{
		}

		CommandEntry(SDPCSTRING ControlName, const CommandControlType& ControlType)
		{
			m_ControlName = ControlName;
			m_ControlType = ControlType;
		}

		const sdpString& ControlName() { return m_ControlName; }
		const CommandControlType& ControlType() { return m_ControlType; }

	private:
		sdpString m_ControlName;
		CommandControlType m_ControlType;
	};

	/// A collection of user actions
	typedef std::map<sdpString, CommandEntry> CommandEntries;
	typedef CommandEntries::iterator CommandIterator;
	/// Stores a collection of user actions
	CommandEntries m_CommandEntries;
	
	/// Control highlighting is predefined
	static const sdpString control_highlight;
};

#endif // K3DUSERINTERFACEELEMENT_H

