
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements objects for storing and manipulating serializable data in a standardized way
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dcom.h>
#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dvariable.h>

#include <k3dibezierchannel.h>
#include <k3dichannel.h>
#include <k3didependencies.h>
#include <k3didocument.h>
#include <k3dieditable.h>
#include <k3dieventsource.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>
#include <k3distate.h>

////////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableCommon

k3dVariableCommon::k3dVariableCommon(SDPCSTRING Name, k3dIStateRecorder* StateRecorder) :
	m_Name(Name),
	m_StateRecorder(StateRecorder),
	m_ValueChanged(false)
{
	// Sanity checks ...
	g_assert_string(Name);

	// We allow NULL for StateRecorder, since it's optional
}

k3dVariableCommon::~k3dVariableCommon()
{
	Notify(Event_Deleted, 0);
}

SDPCSTRING k3dVariableCommon::Name() const
{
	return m_Name;
}

void k3dVariableCommon::PushEditing()
{
	// This would suggest that you're calling PushEditing() twice-in-a-row without an intervening call to PopEditing() ...
	g_return_if_fail(!m_ValueChanged);

	k3dEditable::PushEditing();
	m_ValueChanged = false;
}

void k3dVariableCommon::PopEditing()
{
	m_ValueChanged = false;
	k3dEditable::PopEditing();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dObjectVariable

k3dObjectVariable::k3dObjectVariable(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector) :
	k3dObjectVariableBase(Name, StateRecorder),
	m_Document(Document),
	m_Object(0),
	m_ObjectID(0),
	m_EventConnector(EventConnector)
{
	// Sanity checks ...
	g_assert_string(Name);
	g_assert(m_Document);
}

k3dObjectVariable::~k3dObjectVariable()
{
	ReleaseObject();
	delete m_EventConnector;
}

/// Stores k3dObjectVariable state changes for undo/redo functionality
class k3dObjectVariableStateChange : public k3dIStateContainer
{
public:
	k3dObjectVariableStateChange(k3dObjectVariable* const Variable, k3dIObject* const Object) :
		m_Variable(Variable),
		m_Object(Object)
	{
	}
	
	void RestoreState()
	{
		m_Variable->SetObject(m_Object);
	}
	
private:
	k3dObjectVariable* const m_Variable;
	k3dIObject* const m_Object;
};

sdpResult k3dObjectVariable::SetObject(sdpIUnknown* Object)
{
	if(m_Document->StateRecorder()->ChangeSet())
		m_Document->StateRecorder()->ChangeSet()->RecordOldState(new k3dObjectVariableStateChange(this, m_Object));

	// Release any previous object ...
	ReleaseObject();

	// We allow the caller to assign a null object ...
	if(Object)
		{
			k3dIObject* const object = dynamic_cast<k3dIObject*>(Object);
			g_return_val_if_fail(object, SDP_E_FAIL);
		
			// Store the new object ...
			m_Object = object;

			// We want the new object to notify us if it goes away ...
			k3dIEventSource* const eventserver = dynamic_cast<k3dIEventSource*>(m_Object);
			g_return_val_if_fail(eventserver, SDP_E_FAIL);
			eventserver->AddEventObserver(this, k3dIObject::Event_Deleted, "k3dObjectVariable");

			// Connect to events on behalf of our owner ...
			if(m_EventConnector)
				m_EventConnector->Attach(eventserver);
		}

	if(m_Document->StateRecorder()->ChangeSet())
		m_Document->StateRecorder()->ChangeSet()->RecordNewState(new k3dObjectVariableStateChange(this, m_Object));

	// Let the world know that we've changed ...
	Notify(k3dIVariable::Event_Changed, 0);

	return SDP_S_OK;
}

bool k3dObjectVariable::Create(sdp::ClassID ClassID, SDPCSTRING Name, bool Edit, bool UserVisible, bool Saveable)
{
	// Sanity checks ...
	g_return_val_if_fail(ClassID != sdp::ClassIDNull, false);

	// Create the object ...
	k3dIObject* object = m_Document->Objects()->CreateObject(ClassID);
	g_return_val_if_fail(object, false);

	g_return_val_if_fail(sdpSucceeded(object->SetName(Name)), false);
	g_return_val_if_fail(sdpSucceeded(object->State()->SetState(k3dIState::USERVIEWABLE, UserVisible)), false);
	g_return_val_if_fail(sdpSucceeded(object->State()->SetState(k3dIState::SAVEABLE, Saveable)), false);
	if(Edit)
		object->EditObject();

	// Attach ourselves to it ...
	g_return_val_if_fail(SDP_S_OK == SetObject(object), false);

	return true;
}

k3dIDocument* k3dObjectVariable::Document() const
{
	return m_Document;
}

void k3dObjectVariable::ReleaseObject()
{
	// Cancel any event notifications ...
	if(m_Object)
		{
			k3dIEventSource* eventserver = dynamic_cast<k3dIEventSource*>(m_Object);
			g_return_if_fail(eventserver);
			eventserver->RemoveEventObserver(this);
		}

	if(m_EventConnector)
		m_EventConnector->Detach();
	
	m_Object = 0;
}

sdpResult k3dObjectVariable::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const
{
	// Sanity checks ...
	g_assert(Dependencies);

	const k3dIObject::ID id = m_Object ? m_Object->ObjectID() : 0;
	Element.Children().push_back(sdpxml::Element("variable", "", sdpxml::Attribute("name", m_Name), sdpxml::Attribute("value", sdpToString(id))));
	
	if(m_Object)
		Dependencies->RegisterObject(m_Object);

	return SDP_S_OK;
}

sdpResult k3dObjectVariable::Load(sdpxml::Element& Element)
{
	// The test for "object" is for backward compatibility with documents prior to version 0.2 ... it should be removed by version 0.4
	g_assert((Element.Name() == "object") || (Element.Name() == "variable"));

	// Pull out the name ...
	sdpString name;
	g_return_val_if_fail(sdpxml::ParseAttribute(Element, "name", name), SDP_E_FAIL);
	g_return_val_if_fail(name == m_Name, SDP_E_FAIL);

	// Pull out the id ...
	sdpString id;
	if(!sdpxml::ParseAttribute(Element, "value", id))
		{
			// This is for backward compatibility with documents prior to version 0.2 ... it should be removed by version 0.4
			g_return_val_if_fail(sdpxml::ParseAttribute(Element, "id", id), SDP_E_FAIL);
			std::cout << "k3dObjectVariable::Load() - loading obsolete object ID attribute" << std::endl;
		}

	// Store it (temporarily) ...
	m_ObjectID = atol(id);

	return SDP_S_OK;
}

sdpResult k3dObjectVariable::LoadComplete()
{
	// Skip unbound objects (that have an ID of zero) ...
	if(0 == m_ObjectID)
		return SDP_S_OK;

	// Find a document object with the same ID ...
	std::vector<k3dIObject*> objects;
	k3d::copy_if(m_Document->Objects(), std::back_inserter(objects), k3d::SameID(m_ObjectID));	
	g_return_val_if_fail(1 == objects.size(), SDP_E_FAIL);

	// Bind it to our internal object ...
	SetObject(*objects.begin());

	return SDP_S_OK;
}

sdpResult k3dObjectVariable::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	sdpString event(Event);

	// Release our object ...
	if(event == k3dIObject::Event_Deleted)
		{
			SetObject(0);
			return SDP_S_OK;
		}

	return k3dEventObserver::ObserveEvent(Source, Event, EventData, ObserverData);
}

/////////////////////////////////////////////////////////////
// k3dObjectVariable


//////////////////////////////////////////////////////////////
// k3dChannelVariable

k3dChannelVariable::k3dChannelVariable(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector) :
	k3dChannelVariableBase(Name, Document, StateRecorder, EventConnector)	
{
}

///////////////////////////////////////////////////////////////
// k3dScalarChannel

k3dScalarChannel::k3dScalarChannel(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector) :
	k3dScalarChannelBase(Name, Document, StateRecorder, EventConnector)
{
}

bool k3dScalarChannel::Create(SDPCSTRING Name, double FirstValue, bool ShowTime, double MinY, double MaxY, k3dIChannel::DataType Data, bool Saveable)
{
	g_return_val_if_fail(k3dScalarChannelBase::Create(k3d::classid::Channel, Name, false, false, Saveable), false);
	
	k3dIBezierChannel* const bezierchannel = dynamic_cast<k3dIBezierChannel*>(m_Interface);
	g_return_val_if_fail(bezierchannel, false);
	
	k3d::Editing editing(m_Interface);
	g_return_val_if_fail(sdpSucceeded(m_Interface->SetDatatype(Data)), false);
	k3dChannelSetValueRange(m_Interface, MinY, MaxY);
	k3dChannelSetConstantValue(m_Interface, FirstValue);

	return true;
}

///////////////////////////////////////////////////////////////
// k3dColorChannel

k3dColorChannel::k3dColorChannel(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector) :
	k3dColorChannelBase(Name, Document, StateRecorder, EventConnector)
{
}

bool k3dColorChannel::Create(SDPCSTRING Name, double FirstColor[3], bool ShowTime, bool Saveable)
{
	g_return_val_if_fail(k3dColorChannelBase::Create(k3d::classid::Channel, Name, false, false, Saveable), false);
	
	k3dIBezierChannel* const bezierchannel = dynamic_cast<k3dIBezierChannel*>(m_Interface);
	g_return_val_if_fail(bezierchannel, false);
	
	k3d::Editing editing(m_Interface);
	g_return_val_if_fail(sdpSucceeded(m_Interface->SetDatatype(k3dIChannel::DATA_COLOR)), false);
	k3dChannelSetConstantValue(m_Interface, FirstColor);

	return true;
}


