#ifndef K3DVARIABLE_H
#define K3DVARIABLE_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares objects for storing and manipulating serializable data in a standardized way
		\author Tim Shead (tshead@k-3d.com)
*/

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // SDPXML_H

#include <k3deditable.h>
#include <k3deventobserver.h>
#include <k3deventsource.h>
#include <k3dstaterecorder.h>

#include <k3dichannel.h>
#include <k3diobject.h>
#include <k3divariable.h>

#include <strstream>

// Forward class declarations
class k3dEventConnector;

///////////////////////////////////////////////////////////
// k3dVariableCommon

/// Abstract class for variable objects that stores the variable name  and state recorder, and implements event-handling functionality
class k3dVariableCommon :
	public k3dIVariable,
	public k3dEditable,
	public k3dEventSource
{
public:
	k3dVariableCommon(SDPCSTRING Name, k3dIStateRecorder* const StateRecorder);
	virtual ~k3dVariableCommon();

	/// Returns the variable name
	SDPCSTRING Name() const;

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

protected:
	/// Stores the human-readable name of this variable
	const sdpString m_Name;
	/// Stores an optional k3dIStateRecorder interface for recording undo/redo information
	k3dIStateRecorder* const m_StateRecorder;
	/// Set to true iff the variable state has changed during the current edit session
	bool m_ValueChanged;
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableConstraint

/// Provides a type-safe template for applying one-or-more constraints on the value of a variable
template<class Type>
class k3dVariableConstraint
{
public:
	k3dVariableConstraint(k3dVariableConstraint<Type>* const NextConstraint) :
		m_NextConstraint(NextConstraint)
	{
		// We allow NULL for NextContstraint because it's optional ...
	}

	virtual ~k3dVariableConstraint()
	{
		delete m_NextConstraint;
	}

	virtual void Constrain(const Type& Value, Type& ConstrainedValue) = 0;

protected:
	k3dVariableConstraint<Type>* const m_NextConstraint;
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableMinimum

/// Provides a constraint on the minimum value of a variable
template<class Type>
class k3dVariableMinimum :
	public k3dVariableConstraint<Type>
{
public:
	k3dVariableMinimum(const Type& Minimum, k3dVariableConstraint<Type>* const NextConstraint) :
		k3dVariableConstraint<Type>(NextConstraint),
		m_Minimum(Minimum)
	{
	}

	void Constrain(const Type& Value, Type& ConstrainedValue)
	{
		ConstrainedValue = sdpMax(Value, m_Minimum);

		if(m_NextConstraint)
			m_NextConstraint->Constrain(ConstrainedValue, ConstrainedValue);
	}

private:
	const Type m_Minimum;
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableMaximum

/// Provides a constraint on the maximum value of a variable
template<class Type>
class k3dVariableMaximum :
	public k3dVariableConstraint<Type>
{
public:
	k3dVariableMaximum(const Type& Maximum, k3dVariableConstraint<Type>* const NextConstraint) :
		k3dVariableConstraint<Type>(NextConstraint),
		m_Maximum(Maximum)
	{
	}

	void Constrain(const Type& Value, Type& ConstrainedValue)
	{
		ConstrainedValue = sdpMin(Value, m_Maximum);

		if(m_NextConstraint)
			m_NextConstraint->Constrain(ConstrainedValue, ConstrainedValue);
	}

private:
	const Type m_Maximum;
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableRange

/// Clamps a variable to a range of values
template<class Type>
class k3dVariableRange :
	public k3dVariableConstraint<Type>
{
public:
	k3dVariableRange(const Type& Minimum, const Type& Maximum, k3dVariableConstraint<Type>* const NextConstraint) :
		k3dVariableConstraint<Type>(NextConstraint),
		m_Minimum(Minimum),
		m_Maximum(Maximum)
	{
	}

	void Constrain(const Type& Value, Type& ConstrainedValue)
	{
		ConstrainedValue = sdpClamp(m_Minimum, Value, m_Maximum);

		if(m_NextConstraint)
			m_NextConstraint->Constrain(ConstrainedValue, ConstrainedValue);
	}

private:
	const Type m_Minimum;
	const Type m_Maximum;
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dVariableTruncate

/// Forces a floating-point value to integers by truncating fractional data
class k3dVariableTruncate :
	public k3dVariableConstraint<gdouble>
{
public:
	k3dVariableTruncate(k3dVariableConstraint<gdouble>* const NextConstraint) :
		k3dVariableConstraint<gdouble>(NextConstraint)
	{
	}

	void Constrain(const gdouble& Value, gdouble& ConstrainedValue)
	{
		ConstrainedValue = glong(Value);

		if(m_NextConstraint)
			m_NextConstraint->Constrain(ConstrainedValue, ConstrainedValue);
	}
};

////////////////////////////////////////////////////////////////////////////////////////
// k3dVariable

typedef k3dVariableCommon k3dVariableBase;

/// Provides type-safe access to a variable that stores any type that's serializable
template<class Type>
class k3dVariable :
	public k3dVariableBase
{
public:
	k3dVariable(SDPCSTRING Name, const Type& Value, k3dIStateRecorder* StateRecorder, k3dVariableConstraint<Type>* Constraint) :
		k3dVariableBase(Name, StateRecorder),
		m_Value(Value),
		m_Constraint(Constraint)
	{
		// Note - we allow a NULL for Constraint because it's optional
	}

	~k3dVariable()
	{
		delete m_Constraint;
	}

	/// Sets the variable's contents to the given value
	void SetValue(const Type& Value)
	{
		// Sanity checks ...
		g_return_if_fail(Editing());

		// If the value hasn't changed, we're done ...
		if(Value == m_Value)
			return;

		Type newvalue(Value);

		// Apply constraints to the input value as appropriate ...
		if(m_Constraint)
			m_Constraint->Constrain(newvalue, newvalue);

		// Record the original state, provided the constrained value is new ...
		if(m_StateRecorder && m_StateRecorder->ChangeSet() && !m_ValueChanged && newvalue != m_Value)
			{
				m_StateRecorder->ChangeSet()->RecordOldState(new k3dValueStateContainer<Type>(m_Value));
				m_ValueChanged = true;
			}

		// Store the new value ...
		m_Value = newvalue;

		// Notify clients that we've changed ...
		Notify(Event_Changed, 0);
	}

	/// Returns the variable's contents by value
	const Type& Value() const
	{
		return m_Value;
	}

	void PopEditing()
	{
		// Record the new state ...
		if(m_StateRecorder && m_StateRecorder->ChangeSet() && m_ValueChanged)
			{
				m_StateRecorder->ChangeSet()->RecordNewState(new k3dValueStateContainer<Type>(m_Value));
				m_StateRecorder->ChangeSet()->RecordAction(new k3dActionNotify(this, Event_Changed, 0, 0));
			}

		k3dVariableBase::PopEditing();
	}

	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const
	{
		// Sanity checks ...
		g_assert(Dependencies);

		Element.Children().push_back(sdpxml::Element("variable", "", sdpxml::Attribute("name", m_Name), sdpxml::Attribute("value", sdpToString(m_Value))));
		return SDP_S_OK;
	}

	sdpResult Load(sdpxml::Element& Element)
	{
		// Sanity checks ...
		g_return_val_if_fail(Element.Name() == "variable", SDP_E_FAIL);

		// Pull out the name ...
		const sdpString name = sdpxml::GetAttribute<sdpString>(Element, "name", "");
		g_return_val_if_fail(name == m_Name, SDP_E_FAIL);

		// Pull out the value ...
		const sdpString textvalue = sdpxml::GetAttribute<sdpString>(Element, "value", "");

		// Convert and store it ...
		m_Value = sdpFromString(textvalue, m_Value);

		return SDP_S_OK;
	}

	sdpResult LoadComplete()
	{
		return SDP_S_OK;
	}

protected:
	/// Stores the contents of the variable
	Type m_Value;
	/// Stores an optional constraint object for controlling this variable's value
	k3dVariableConstraint<Type>* const m_Constraint;
};

/// Convenience function for setting k3dVariable values in one step
template<class Type>
void k3dSetValue(k3dVariable<Type>& Storage, const Type& Value)
{
	Storage.PushEditing();
	Storage.SetValue(Value);
	Storage.PopEditing();
}

/////////////////////////////////////////////////////////////
// k3dObjectVariable

typedef k3dVariableCommon k3dObjectVariableBase;

/// A variable object that stores a reference to another K-3D object
class k3dObjectVariable :
	public k3dObjectVariableBase,
	public k3dEventObserver
{
public:
	k3dObjectVariable(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector);
	~k3dObjectVariable();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	virtual sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const;
	virtual sdpResult Load(sdpxml::Element& Element);
	virtual sdpResult LoadComplete();

	/// Attaches an existing K-3D object to the variable
	virtual sdpResult SetObject(sdpIUnknown* Object);
	/// Returns a k3dIObject interface to the attached object
	k3dIObject* Object() const { return m_Object; }

	/// Creates a new K-3D object by sdpCom ClassID and attaches it
	virtual bool Create(sdp::ClassID ClassID, SDPCSTRING Name, bool Edit, bool Show, bool Saveable);

	/// Returns the owning document
	k3dIDocument* Document() const;

protected:
	/// Frees any existing references to a K-3D object
	virtual void ReleaseObject();

	/// Stores the owning document
	k3dIDocument* const m_Document;
	/// Stores the attached object's k3dIObject interface
	k3dIObject* m_Object;
	/// Temporarily stores an object ID while loading
	k3dIObject::ID m_ObjectID;

	/// Stores a chain of event connectors to facilitate communication with the attached object
	k3dEventConnector* const m_EventConnector;
};

/////////////////////////////////////////////////////////////
// k3dObjectInterfaceVariable

typedef k3dObjectVariable k3dObjectInterfaceVariableBase;

/// Variable object that references an object in a type-safe manner using the given interface
template<class InterfaceClass>
class k3dObjectInterfaceVariable :
	public k3dObjectInterfaceVariableBase
{
public:
	k3dObjectInterfaceVariable(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector) :
		k3dObjectInterfaceVariableBase(Name, Document, StateRecorder, EventConnector),
		m_Interface(0)
	{
	}

	/// Overrides the base-class Create() to obtain an interface pointer of the correct type to the new K-3D object
	bool Create(sdp::ClassID ClassID, SDPCSTRING Name, bool Edit, bool UserVisible, bool Saveable)
	{
		g_return_val_if_fail(k3dObjectInterfaceVariableBase::Create(ClassID, Name, Edit, UserVisible, Saveable), false);

		m_Interface = dynamic_cast<InterfaceClass*>(m_Object);
		g_return_val_if_fail(m_Interface, false);

		return true;
	}

	/// Overrides the base-class SetObject() to obtain our extra interface pointer
	sdpResult SetObject(sdpIUnknown* Object)
	{
		// Call the default implementation ...
		g_return_val_if_fail(sdpSucceeded(k3dObjectInterfaceVariableBase::SetObject(Object)), SDP_E_FAIL);

		// Get the desired interface ...
		if(m_Object)
			{
				m_Interface = dynamic_cast<InterfaceClass*>(m_Object);
				g_return_val_if_fail(m_Interface, SDP_E_FAIL);
			}

		return SDP_S_OK;
	}

	/// Overrides the base-class ReleaseObject() so we can release our extra interface pointer
	void ReleaseObject()
	{
		// Call the default implementation ...
		k3dObjectInterfaceVariableBase::ReleaseObject();

		// Release our extra interface ...
		if(m_Interface)
			m_Interface = 0;
	}

	/// Returns the interface to the attached K-3D object
	InterfaceClass* Interface() { return m_Interface; }
	/// Returns the interface to the attached K-3D object (short form)
	InterfaceClass* operator->() { return m_Interface; }

protected:
	/// Stores a pointer to the desired interface type on the underlying object
	InterfaceClass* m_Interface;
};

//////////////////////////////////////////////////////////////
// k3dChannelVariable

typedef k3dObjectInterfaceVariable<k3dIChannel> k3dChannelVariableBase;

/// A convenience base-class for variables that store a K-3D channel object
class k3dChannelVariable :
	public k3dChannelVariableBase
{
public:
	k3dChannelVariable(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector);
};

///////////////////////////////////////////////////////////////
// k3dScalarChannel

typedef k3dChannelVariable k3dScalarChannelBase;

/// A convenience class for variables that store a K-3D channel object storing scalar values
class k3dScalarChannel :
	public k3dScalarChannelBase
{
public:
	k3dScalarChannel(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector);

	/// Creates the K-3D channel object and initializes it to store scalar values
	bool Create(SDPCSTRING Name, double FirstValue, bool ShowTime, double MinY, double MaxY, k3dIChannel::DataType Data, bool Saveable);
};

///////////////////////////////////////////////////////////////
// k3dColorChannel

typedef k3dChannelVariable k3dColorChannelBase;

/// A convenience class for variables that store a K-3D channel object storing color values
class k3dColorChannel :
	public k3dColorChannelBase
{
public:
	k3dColorChannel(SDPCSTRING Name, k3dIDocument* const Document, k3dIStateRecorder* const StateRecorder, k3dEventConnector* const EventConnector);

	/// Creates the K-3D channel object and initializes it to store color values
	bool Create(SDPCSTRING Name, double FirstColor[3], bool ShowTime, bool Saveable);
};

#endif // K3DVARIABLE_H

