
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dVariableContainer class, which encapsulates a set of k3dVariable objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dvariablecontainer.h>
#include <k3divariable.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // SDPXML_H

//////////////////////////////////////////////////////////////////////////////////
// k3dVariableContainer

k3dVariableContainer::~k3dVariableContainer()
{
}

//////////////////////////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult k3dVariableContainer::InitNew()
{
	return SDP_S_OK;
}

sdpResult k3dVariableContainer::Load(sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Create a mapping of elements to choose from ...
	typedef std::map<sdpString, sdpxml::ElementPointer> ElementMap;
	typedef ElementMap::iterator ElementIterator;
	ElementMap elementmap;
		
	// Insert <variable> tags ...
	sdpxml::ElementPointer variables = sdpxml::FindElement(Element, sdpxml::SameName("variables"));
	if(variables)
		{
			for(sdpxml::ElementCollection::iterator element = variables->Children().begin(); element != variables->Children().end(); element++)
				{
					if(element->Name() != "variable")
						continue;
						
					const sdpString name = sdpxml::GetAttribute<sdpString>(*element, "name", "");
					if(0 == name.size())
						continue;
						
					if(elementmap.find(name) != elementmap.end())
						std::cerr << "k3dVariableContainer::LoadVariables() - duplicate <variable> [" << name << "] at " << sdpxml::FileReference(Document, *element) << " will be ignored" << std::endl;
					else
						elementmap[name] = &(*element);
				}
		}

	/** \todo Insert <object> tags ... this is for backwards-compatibility with documents prior to version 0.2 ... it should be gone after 0.2 is release */
	sdpxml::ElementPointer objects = sdpxml::FindElement(Element, sdpxml::SameName("objects"));
	if(objects)
		{
			for(sdpxml::ElementCollection::iterator element = objects->Children().begin(); element != objects->Children().end(); element++)
				{
					if(element->Name() != "object")
						continue;
						
					const sdpString name = sdpxml::GetAttribute<sdpString>(*element, "name", "");
					if(0 == name.size())
						continue;
						
					if(elementmap.find(name) != elementmap.end())
						std::cerr << "k3dVariableContainer::LoadVariables() - duplicate <object> [" << name << "] at " << sdpxml::FileReference(Document, *element) << " will be ignored" << std::endl;
					else
						elementmap[name] = &(*element);
						
					std::cout << "k3dVariableContainer::LoadVariables() - loading variable from obsolete <object> tag" << std::endl;
				}
		}

	/** \todo Insert <shader> tags ... this is for backwards-compatibility with documents prior to version 0.2 ... it should be gone after 0.2 is release */
	for(sdpxml::ElementCollection::iterator element = Element.Children().begin(); element != Element.Children().end(); element++)
		{
			if(element->Name() != "shader")
				continue;
				
			const sdpString name = sdpxml::GetAttribute<sdpString>(*element, "name", "");
			if(0 == name.size())
				continue;
		
			if(elementmap.find(name) != elementmap.end())
				std::cerr << "k3dVariableContainer::LoadVariables() - duplicate <shader> [" << name << "] at " << sdpxml::FileReference(Document, *element) << " will be ignored" << std::endl;
			else		
				elementmap[name] = &(*element);
				
			std::cout << "k3dVariableContainer::LoadVariables() - loading variable from obsolete <shader> tag" << std::endl;
		}

	// If we didn't find anything, we're done ...
	if(0 == elementmap.size())
		return SDP_S_FALSE;

	// For each variable ...
	for(VariableIterator variable = m_Variables.begin(); variable != m_Variables.end(); variable++)
		{
			// See if it has an element ...
			sdpxml::ElementPointer variableelement = elementmap[(*variable)->Name()];
			if(!variableelement)
				continue;

			// Load it up!
			(*variable)->Load(*variableelement);
			elementmap.erase((*variable)->Name());
		}

	// Look for unused data ...
	if(elementmap.size())
		{
			std::cout << "k3dVariableContainer::LoadVariables() - couldn't find data for " << elementmap.size() << " variables:" << std::endl;
			for(ElementIterator element = elementmap.begin(); element != elementmap.end(); element++)
				std::cout << "\t" << element->first << std::endl;
			std::cout << "This usually means that you're loading an older document with newer plugins, and isn't cause for concern." << std::endl << std::endl;
		}

	return SDP_S_OK;
}

sdpResult k3dVariableContainer::LoadComplete()
{
	// Reconnect any objects ...
	for(VariableIterator variable = m_Variables.begin(); variable != m_Variables.end(); variable++)
		(*variable)->LoadComplete();

	return SDP_S_OK;
}

sdpResult k3dVariableContainer::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_return_val_if_fail(Dependencies, SDP_E_FAIL);

	// See if we have any variables to save ...
	if(!m_Variables.size())
		return SDP_S_FALSE;

	// Save variables ...
	Element.Children().push_back(sdpxml::Element("variables"));
	sdpxml::Element& variables = Element.Children().back();

	for(VariableIterator variable = m_Variables.begin(); variable != m_Variables.end(); variable++)
		(*variable)->Save(variables, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dVariableContainer::SaveComplete()
{
	return SDP_S_OK;
}

class k3dVariableContainerUniqueName
{
public:
	k3dVariableContainerUniqueName(SDPCSTRING Name) :
		m_Name(Name)
	{
	}
	
	void operator()(k3dIVariable* Variable)
	{
		if(m_Name == Variable->Name())
			std::cerr << "Adding duplicate variable [" << m_Name << "] to object" << std::endl;
	}
	
protected:
	sdpString m_Name;
};

void k3dVariableContainer::AddVariable(k3dIVariable* Variable)
{
	// Sanity checks ...
	g_assert(Variable);

	// Make sure the variable name is unique ...
	std::for_each(m_Variables.begin(), m_Variables.end(), k3dVariableContainerUniqueName(Variable->Name()));

	// Add this variable to our list ...
	m_Variables.push_back(Variable);
}



