// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dWindowGeometryStore and derived classes, which encapsulate persisting window geometry (position & size) to various storage types
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dwindowgeometrystore.h>

#include <k3diapplication.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // !SDPXML_H

//////////////////////////////////////////////////////////////////////////////////
// k3dWindowGeometryStore

k3dWindowGeometryStore::k3dWindowGeometryStore(k3dWindowGeometryStore* const NextStore) :
	m_NextStore(NextStore)
{
}

k3dWindowGeometryStore::~k3dWindowGeometryStore()
{
	delete m_NextStore;
}

//////////////////////////////////////////////////////////////////////////////////
// k3dOptionsWindowGeometryStore

k3dOptionsWindowGeometryStore::k3dOptionsWindowGeometryStore(k3dIApplication* const Application, k3dWindowGeometryStore* const NextStore) :
	k3dWindowGeometryStore(NextStore),
	m_Application(Application)
{
	// Sanity checks ...
	g_assert(m_Application);
}

void k3dOptionsWindowGeometryStore::SetWindowGeometry(SDPCSTRING WindowName, const gint Left, const gint Top, const guint Width, const guint Height)
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "k3dOptionsWindowGeometryStore::SetWindowGeometry: " << WindowName << std::endl;

	// Sanity checks ...
	g_assert_string(WindowName);

	// If the user has disabled restoring window size & position, we're done ...
	if(!k3dOptionsGetRestoreWindows(k3dApplication()->Options()))
		return;

	// Get an element in user options where we can store the data ...
	sdpxml::Element& window = sdpxml::SafeXPathSelect(k3dApplication()->Options(), k3dApplication()->Options(), "application/userinterface/windows/window[@name=\"" + sdpString(WindowName) + "\"]", sdpxml::Element("window", "", sdpxml::Attribute("name", WindowName)));
	
	// Save that data ... !
	sdpxml::SetAttribute(window, sdpxml::Attribute("geometry", sdpToString(sdpVector4(Left, Top, Width, Height))));

	// Give any downstream stores a crack at the data ...
	if(m_NextStore)
		m_NextStore->SetWindowGeometry(WindowName, Left, Top, Width, Height);
}

bool k3dOptionsWindowGeometryStore::GetWindowGeometry(SDPCSTRING WindowName, gint& Left, gint& Top, guint& Width, guint& Height)
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "k3dOptionsWindowGeometryStore::GetWindowGeometry: " << WindowName << std::endl;

	// Sanity checks ...
	g_assert_string(WindowName);

	// If the user has disabled restoring window size & position, we're done ...
	if(!k3dOptionsGetRestoreWindows(k3dApplication()->Options()))
		return false;

	// Give any downstream stores first crack at specifying the size & position ...
	if(m_NextStore)
		if(m_NextStore->GetWindowGeometry(WindowName, Left, Top, Width, Height))
			return true;

	// Get an element in user options where we can grab the data ...
	sdpxml::Element& window = sdpxml::SafeXPathSelect(k3dApplication()->Options(), k3dApplication()->Options(), "application/userinterface/windows/window[@name=\"" + sdpString(WindowName) + "\"]", sdpxml::Element("window", "", sdpxml::Attribute("name", WindowName)));

	// Get the data ...
	const sdpVector4 geometry = sdpxml::GetAttribute<sdpVector4>(window, "geometry", sdpVector4());
			
	const gint left = static_cast<gint>(geometry[0]);
	const gint top = static_cast<gint>(geometry[1]);
	const guint width = static_cast<guint>(geometry[2]);
	const guint height = static_cast<guint>(geometry[3]);
	
	if((0 == width) || (0 == height))
		return false;
	
	Left = left;
	Top = top;
	Width = width;
	Height = height;
	
	return true;
}
	
//////////////////////////////////////////////////////////////////////////////////
// k3dPersistWindowGeometryStore

k3dPersistWindowGeometryStore::k3dPersistWindowGeometryStore(k3dWindowGeometryStore* const NextStore) :
	k3dWindowGeometryStore(NextStore)
{
}

void k3dPersistWindowGeometryStore::SetWindowGeometry(SDPCSTRING WindowName, const gint Left, const gint Top, const guint Width, const guint Height)
{
	m_Left = Left;
	m_Top = Top;
	m_Width = Width;
	m_Height = Height;
	
	if(m_NextStore)
		m_NextStore->SetWindowGeometry(WindowName, Left, Top, Width, Height);
}

bool k3dPersistWindowGeometryStore::GetWindowGeometry(SDPCSTRING WindowName, gint& Left, gint& Top, guint& Width, guint& Height)
{
	if(m_NextStore)
		if(m_NextStore->GetWindowGeometry(WindowName, Left, Top, Width, Height))
			return true;

	Left = m_Left;
	Top = m_Top;
	Width = m_Width;
	Height = m_Height;
	
	return true;
}


