// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBitmap K-3D object, which encapsulates bitmap images
		\author Tim Shead (tshead@k-3d.com)
*/

#include "bitmap.h"

#include <k3dfilefilters.h>
#include <k3dfilterselector.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dibitmapreadformat.h>
#include <k3dibitmapwriteformat.h>
#include <k3difileformat.h>
#include <k3diobjectinfo.h>
#include <k3direnderframe.h>
#include <k3divariable.h>

#include <sdpgtkevent.h>
#include <sdpgtkfileselector.h>

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

/////////////////////////////////////////////////////////
// CBitmap

CBitmap::CBitmap(sdpIUnknown* Initializer) :
	CBitmapBase(Initializer),
	m_ImagePath("imagepath", "", Document()->StateRecorder(), 0)
{
	AddVariable(&m_ImagePath);
}

CBitmap::~CBitmap()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CBitmap::EditObject()
{
	new CBitmapDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIBitmap implementation

sdpResult CBitmap::NewImage(const gulong Width, const gulong Height)
{
	// Sanity checks ...
	g_return_val_if_fail(Width, SDP_E_FAIL);
	g_return_val_if_fail(Height, SDP_E_FAIL);

	// Get rid of existing data ...
	k3dSetValue<sdpString>(m_ImagePath, "");
	m_Image.Allocate(Width, Height);

	Notify(k3dIBitmap::Event_Changed, 0);

	return SDP_S_OK;
}

sdpResult CBitmap::LoadImage(SDPCSTRING FilePath, k3dIBitmapReadFormat* const FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_fail(FormatFilter, SDP_E_FAIL);

	k3dSetValue<sdpString>(m_ImagePath, FilePath);

	return RawLoadFile(FilePath, FormatFilter) ? SDP_S_OK : SDP_E_FAIL;
}

bool CBitmap::RawLoadFile(const sdpString FilePath, k3dIBitmapReadFormat* const FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_fail(FilePath.size(), false);
	g_return_val_if_fail(FormatFilter, false);

	bool result = false;

	// Read the file ...
	gulong width = 0;
	gulong height = 0;
	if(sdpSucceeded(FormatFilter->PreRead(this, FilePath, &width, &height)))
		{
			m_Image.Allocate(width, height);
			result = sdpSucceeded(FormatFilter->ReadFile(this, FilePath));

			Notify(k3dIBitmap::Event_Changed, 0);
		}

	return result;
}

sdpResult CBitmap::SaveImage(SDPCSTRING FilePath, k3dIBitmapWriteFormat* const FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_fail(FormatFilter, SDP_E_FAIL);

	return RawSaveFile(FilePath, FormatFilter) ? SDP_S_OK : SDP_E_FAIL;
}

bool CBitmap::RawSaveFile(const sdpString FilePath, k3dIBitmapWriteFormat* const FormatFilter)
{
	// Sanity checks ...
	g_return_val_if_fail(FilePath.size(), false);
	g_return_val_if_fail(FormatFilter, false);

	if(sdpFailed(FormatFilter->PreWrite(this, FilePath)))
		return false;

	return sdpSucceeded(FormatFilter->WriteFile(this, FilePath));
}

k3dPixel* CBitmap::Data()
{
	return m_Image.Data();
}

gulong CBitmap::Width()
{
	return m_Image.Width();
}

gulong CBitmap::Height()
{
	return m_Image.Height();
}

sdpResult CBitmap::FlipVertical()
{
	// Sanity checks ...
	g_return_val_if_fail(Data() && Width() && Height(), SDP_E_FAIL);

	m_Image.FlipVertical();

	Notify(k3dIBitmap::Event_Changed, 0);

	return SDP_S_OK;
}

sdpResult CBitmap::FlipHorizontal()
{
	// Sanity checks ...
	g_return_val_if_fail(Data() && Width() && Height(), SDP_E_FAIL);

	m_Image.FlipHorizontal();

	Notify(k3dIBitmap::Event_Changed, 0);

	return SDP_E_FAIL;
}

sdpResult CBitmap::Scale(gulong NewWidth, gulong NewHeight)
{
	// Sanity checks ...
	g_return_val_if_fail(Data() && Width() && Height(), SDP_E_FAIL);

	m_Image.Scale(NewWidth, NewHeight, k3dImageBuffer::NEAREST);

	Notify(k3dIBitmap::Event_Changed, 0);

	return SDP_E_FAIL;
}

////////////////////////////////////////////////////////////////////////////
// k3dIEditorTexture implementation

sdpResult CBitmap::EditorTexturePreRender(k3dIEditorEngine* const Engine)
{
	// If we don't have any data, we're done ...
	if(!Data())
		return SDP_S_FALSE;

	// Create our OpenGL data as needed ...
	if(!m_OpenGLData.Data() || !m_OpenGLData.Width() || !m_OpenGLData.Height())
		{
			m_OpenGLData.Copy(Data(), Width(), Height());
			m_OpenGLData.ConvertForOpenGL(k3dImageBuffer::NEAREST);
		}

	return SDP_S_OK;
}

sdpResult CBitmap::EditorTextureSetup(k3dIEditorEngine* const Engine)
{
	// If we don't have any data, we're done ...
	if(!Data())
		{
			glDisable(GL_TEXTURE_2D);
			return SDP_S_OK;
		}

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

//	GLint texturecomponents = m_UseLuma.Value() ? 1 : 3;
//	texturecomponents += m_UseAlpha.Value() ? 1 : 0;
	GLint texturecomponents = 3;

	glTexImage2D(GL_TEXTURE_2D,	// Reserved
		0,												// Detail level
		texturecomponents,				// Components used for modulating
		m_OpenGLData.Width(),			// Bitmap width
		m_OpenGLData.Height(),		// Bitmap height
		0,												// Border size
		GL_RGBA,									// GL_RGB or GL_RGBA
		GL_UNSIGNED_BYTE,					// Data type of pixel values
		m_OpenGLData.Data()				// Pixel values
		);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);

/*
	if(m_Antialias.Value())
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		}
	else
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		}

	if(m_Perspective.Value())
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	else
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);
*/

	glEnable(GL_TEXTURE_2D);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////
// k3dIRiTexture implementation

sdpResult CBitmap::RiTextureStartJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CBitmap::RiTexturePreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	gulong buffersize = 256;
	sdpResult result = Frame->AddInputFile("texture", m_RiTexturePath.buffer(buffersize), &buffersize);
	if(SDP_S_BUFFER_TOO_SMALL == result)
		result = Frame->AddInputFile("texture", m_RiTexturePath.buffer(buffersize), &buffersize);
	g_return_val_if_fail(SDP_S_OK == result, SDP_E_FAIL);

	m_RiTexturePath.resize(strlen(m_RiTexturePath));
	m_RiTexturePath += ".tif";

	k3d::auto_ptr<k3dIBitmapWriteFormat> filter(k3d::GetAutoFilter<k3dIBitmapWriteFormat>(Document(), m_RiTexturePath));
	return SaveImage(m_RiTexturePath, filter.get());
}

sdpResult CBitmap::RiTexturePostRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CBitmap::RiTextureFinishJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

SDPCSTRING CBitmap::RiTexturePath(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, 0);
	g_return_val_if_fail(Engine, 0);

	//return m_RiTexturePath;
	return sdpMakePosixFilePath(m_RiTexturePath);
}

///////////////////////////////////////////////////////////////
// k3dIAnimate implementation

sdpResult CBitmap::PostAnimate()
{
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CBitmap::LoadComplete()
{
	g_return_val_if_fail(sdpSucceeded(CBitmapBase::LoadComplete()), SDP_E_FAIL);

	// Reload the bitmap image ...
	if(m_ImagePath.Value().size())
		{
			k3d::auto_ptr<k3dIBitmapReadFormat> filter(k3d::GetAutoFilter<k3dIBitmapReadFormat>(Document(), m_ImagePath.Value()));
			RawLoadFile(m_ImagePath.Value(), filter.get());
		}

	return SDP_S_OK;
}

sdpResult CBitmap::InitComplete()
{
	return CBitmapBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CBitmapDialog

const sdpString CBitmapDialog::dialogname = "properties";
const sdpString CBitmapDialog::control_load = "load";
const sdpString CBitmapDialog::control_save = "save";
const sdpString CBitmapDialog::control_view = "view";

CBitmapDialog::CBitmapDialog(CBitmap* const Object) :
	CBitmapDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object),
	m_BitmapChangedEvent(Object, this, k3dIBitmap::Event_Changed, 0, 0)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("bitmaplib", "bitmap.gtkml"));

	AddCommand(control_load, control_load, CONTROL_BUTTON);
	AddCommand(control_save, control_save, CONTROL_BUTTON);
	AddCommand(control_view, control_view, CONTROL_BUTTON);

	// Setup OpenGL for rendering previews ...
	sdpGtkContainer rgbcontainer(Container("rgbpreview"));

	if(!m_RGBWidget.Create(rgbcontainer, true, true, 8, 8, 8, 0))
		{
			if(!m_RGBWidget.Create(rgbcontainer, true, true, 5, 5, 5, 0))
				{
					if(!m_RGBWidget.Create(rgbcontainer, false, false, 4, 4, 4, 0))
						{
							std::cerr << "CBitmapDialog::CBitmapDialog(): Could not find useable OpenGL visual for RGB preview" << std::endl;
						}
				}
		}

	if(m_RGBWidget.Initialized())
		{
			MapEvent("configure-event", "configurergb", false, m_RGBWidget, true);
			MapEvent("expose-event", "exposergb", false, m_RGBWidget, true);
		}

	sdpGtkContainer alphacontainer(Container("alphapreview"));

	if(!m_AlphaWidget.Create(alphacontainer, true, true, 8, 8, 8, 0))
		{
			if(!m_AlphaWidget.Create(alphacontainer, true, true, 5, 5, 5, 0))
				{
					if(!m_AlphaWidget.Create(alphacontainer, false, false, 4, 4, 4, 0))
						{
							std::cerr << "CBitmapDialog::CBitmapDialog(): Could not find useable OpenGL visual for Alpha preview" << std::endl;
						}
				}
		}

	if(m_AlphaWidget.Initialized())
		{
			MapEvent("configure-event", "configurealpha", false, m_AlphaWidget, true);
			MapEvent("expose-event", "exposealpha", false, m_AlphaWidget, true);
		}

	CallUpdateControls();
	Show();
}

sdpResult CBitmapDialog::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	if(event == k3dIBitmap::Event_Changed)
		{
			CallUpdateControls();
			return SDP_S_OK;
		}

	return CBitmapDialogBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void CBitmapDialog::UpdateControls()
{
	if(m_Object->m_ImagePath.Value().empty())
		Label("pathlabel").SetText("<Choose Bitmap>");
	else
		Label("pathlabel").SetText(sdpFile(m_Object->m_ImagePath.Value()));

	// If the object has no data, we're done ...
	if(!m_Object->Data())
		return;

	// Create the preview image ...
	m_PreviewImage.Copy(m_Object->m_Image);
	m_PreviewImage.ConvertForOpenGL(128, 128, k3dImageBuffer::NEAREST);

	// Get image data ...
	gulong imagewidth = m_Object->Width();
	gulong imageheight = m_Object->Height();

	gdouble imageratio = 1.0;
	if(imageheight)
		imageratio = gdouble(imagewidth) / gdouble(imageheight);

	// Update our frame aspect ratios to match the image ...
	AspectFrame("rgbframe").Set(0.5, 0.5, imageratio, false);
	AspectFrame("alphaframe").Set(0.5, 0.5, imageratio, false);

	// Update our information fields ...
	Label("width").SetText("Width: " + sdpToString(imagewidth));
	Label("height").SetText("Height: " + sdpToString(imageheight));

	CBitmapDialogBase::UpdateControls();
}

void CBitmapDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "configurergb" || Event->Name() == "exposergb")
		DrawRGB();
	else if(Event->Name() == "configurealpha" || Event->Name() == "exposealpha")
		DrawAlpha();
	else if(Event->Name() == "load")
		OnLoad();
	else if(Event->Name() == "save")
		OnSave();
	else if(Event->Name() == "view")
		OnView();
	else
		CBitmapDialogBase::OnEvent(Event);
}

void CBitmapDialog::DrawRGB()
{
	// If we haven't been initialized yet, we're done ...
	if(!m_RGBWidget.Initialized())
		return;

	// Get the preview image data ...
	k3dPixel* imagedata = m_PreviewImage.Data();
	gulong imagewidth = m_PreviewImage.Width();
	gulong imageheight = m_PreviewImage.Height();

	// If we don't have any data, we're done ...
	if(!imagedata || !imagewidth || !imageheight)
		return;

	// Start OpenGL rendering ...
	m_RGBWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_RGBWidget.Width();
	gulong height = m_RGBWidget.Height();

	glViewport(0, 0, width, height);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, imagewidth, imageheight, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagedata);

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBegin(GL_POLYGON);
		glTexCoord2d(0.0, 0.0); glVertex2d(-1.0, 1.0);
		glTexCoord2d(1.0, 0.0); glVertex2d(1.0, 1.0);
		glTexCoord2d(1.0, 1.0); glVertex2d(1.0, -1.0);
		glTexCoord2d(0.0, 1.0); glVertex2d(-1.0, -1.0);
	glEnd();

	glFlush();

	m_RGBWidget.SwapBuffers();
	m_RGBWidget.End();
}

void CBitmapDialog::DrawAlpha()
{
	// If we haven't been initialized yet, we're done ...
	if(!m_AlphaWidget.Initialized())
		return;

	// Get image data ...
	k3dPixel* imagedata = m_PreviewImage.Data();
	gulong imagewidth = m_PreviewImage.Width();
	gulong imageheight = m_PreviewImage.Height();

	// If we don't have any data, we're done ...
	if(!imagedata || !imagewidth || !imageheight)
		return;

	// Start OpenGL rendering ...
	m_AlphaWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_AlphaWidget.Width();
	gulong height = m_AlphaWidget.Height();

	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, imagewidth, imageheight, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagedata);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBegin(GL_POLYGON);
		glTexCoord2d(0.0, 0.0); glVertex2d(-1.0, 1.0);
		glTexCoord2d(1.0, 0.0); glVertex2d(1.0, 1.0);
		glTexCoord2d(1.0, 1.0); glVertex2d(1.0, -1.0);
		glTexCoord2d(0.0, 1.0); glVertex2d(-1.0, -1.0);
	glEnd();

	glFlush();

	m_AlphaWidget.SwapBuffers();
	m_AlphaWidget.End();
}

void CBitmapDialog::OnLoad()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_load, 0);

	sdpString filepath;
	if(!k3dGetFilePath("bitmap", "Load Bitmap File:", false, m_Object->m_ImagePath.Value(), filepath))
		return;

	// Prompt the user to choose an export filter ...
	k3d::FilterSelector filterselector(m_Object->Document(), "Select Import File Format:", filepath, sdp::TypeInfo<k3dIBitmapReadFormat>());
	filterselector.DoModal(RootWindow());
	if(!filterselector.OK())
		return;

	k3d::auto_ptr<k3dIBitmapReadFormat> filter(filterselector.Filter<k3dIBitmapReadFormat>());
	m_Object->LoadImage(filepath, filter.get());

	// Update the display ...
	k3dRedrawAll(m_Object->Document(), false);
}

void CBitmapDialog::OnSave()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_save, 0);

	sdpString filepath;
	if(!k3dGetFilePath("bitmap", "Save Bitmap File:", true, m_Object->m_ImagePath.Value(), filepath))
		return;

	// Prompt the user to choose an export filter ...
	k3d::FilterSelector filterselector(m_Object->Document(), "Select Export File Format:", filepath, sdp::TypeInfo<k3dIBitmapWriteFormat>());
	filterselector.DoModal(RootWindow());
	if(!filterselector.OK())
		return;

	k3d::auto_ptr<k3dIBitmapWriteFormat> filter(filterselector.Filter<k3dIBitmapWriteFormat>());
	m_Object->SaveImage(filepath, filter.get());
}

void CBitmapDialog::OnView()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_view, 0);

	// Get the user's preferred bitmap viewer ...
	sdpString viewcommand = k3dOptionsGetBitmapViewer(k3dApplication()->Options());
	g_return_if_fail(viewcommand.size());

	viewcommand.formattedreplace('%', "p", m_Object->m_ImagePath.Value());
	viewcommand += " &";

	system(viewcommand);
}


