#ifndef BITMAP_H
#define BITMAP_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CBitmap K-3D object, which encapsulates bitmap images
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deventconnector.h>
#include <k3dobjectdialog.h>
#include <k3dpersistobject.h>
#include <k3dvariable.h>
#include <k3dimagebuffer.h>

#include <k3dibitmap.h>
#include <k3ditexture.h>
#include <k3dieditortexture.h>
#include <k3diritexture.h>
#include <k3ditime.h>

#include <sdpgtkopengldrawingarea.h>

// Forward class declarations ...
class CBitmap;

/////////////////////////////////////////////////////////////////////////////
// CBitmapDialog

typedef k3dObjectDialog CBitmapDialogBase;

/// Provides the user interface for CBitmap objects
class CBitmapDialog :
	public CBitmapDialogBase
{
public:
	CBitmapDialog(CBitmap* const Object);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

protected:
	CBitmap* const m_Object;
	k3dEventConnector m_BitmapChangedEvent;
	
	k3dImageBuffer m_PreviewImage;
	sdpGtkOpenGLDrawingArea m_RGBWidget;
	sdpGtkOpenGLDrawingArea m_AlphaWidget;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void DrawRGB();
	void DrawAlpha();
	void OnLoad();
	void OnSave();
	void OnView();

	void UpdateControls();

	static const sdpString dialogname;
	static const sdpString control_load;
	static const sdpString control_save;
	static const sdpString control_view;
};

///////////////////////////////////////////////////////////////////////////////
// CBitmap

typedef k3dPersistObject CBitmapBase;

class CBitmap :
	public CBitmapBase,
	public k3dIBitmap,
	public k3dITexture,
	public k3dIEditorTexture,
	public k3dIRiTexture
{
friend class CBitmapDialog;
public:
	CBitmap(sdpIUnknown* Initializer);
	~CBitmap();

	// k3dIBitmap implementation
	sdpResult NewImage(const gulong Width, const gulong Height);
	sdpResult LoadImage(SDPCSTRING FilePath, k3dIBitmapReadFormat* const FormatFilter);
	sdpResult SaveImage(SDPCSTRING FilePath, k3dIBitmapWriteFormat* const FormatFilter);

	k3dPixel* Data();
	gulong Width();
	gulong Height();	

	sdpResult FlipVertical();
	sdpResult FlipHorizontal();
	sdpResult Scale(const gulong NewWidth, const gulong NewHeight);

	// k3dIObject implementation
	sdpResult EditObject();

	// k3dIEditorTexture implementation
	sdpResult EditorTexturePreRender(k3dIEditorEngine* const Engine);
	sdpResult EditorTextureSetup(k3dIEditorEngine* const Engine);

	// k3dIRiTexture implementation
	sdpResult RiTextureStartJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine);
	sdpResult RiTexturePreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiTexturePostRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiTextureFinishJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine);
	
	SDPCSTRING RiTexturePath(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);

	// k3dIAnimate implementation ...
	sdpResult PostAnimate();

	//k3dIPersist implementation
	sdpResult LoadComplete();
	sdpResult InitComplete();

private:
	/// Stores the bitmap image path
	k3dVariable<sdpString> m_ImagePath;
	/// Stores the loaded bitmap image
	k3dImageBuffer m_Image;
	/// Stores the bitmap image after making it suitable for use with OpenGL
	k3dImageBuffer m_OpenGLData;

	typedef enum
	{
		CLAMP = 1,
		TILE = 2,
	} RepeatType;

	/// Stores the absolute path to the saved TIFF version of this bitmap
	sdpString m_RiTexturePath;

	/// Does the real work of loading a file ...
	bool RawLoadFile(const sdpString FilePath, k3dIBitmapReadFormat* const FormatFilter);
	/// Does the real work of saving a file ...
	bool RawSaveFile(const sdpString FilePath, k3dIBitmapWriteFormat* const FormatFilter);
				
	/// Called whenever we need to dump our buffer and start over ...
	void UpdateBuffer();
};

#endif // BITMAP_H
