// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CNetPBMReader K-3D object, which reads and writes JPEG images
		\author Tim Shead (tshead@k-3d.com)
*/

#include "netpbmreader.h"

#include <k3dibitmap.h>
#include <k3dmisc.h>

#include <k3diapplication.h>

#include <stdio.h>
#include <fstream>
#include <strstream>

CNetPBMReader::CNetPBMReader(sdpIUnknown* Initializer) :
	CNetPBMReaderBase(Initializer),
	m_Formats("netpbm")
{
}

CNetPBMReader::~CNetPBMReader()
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool CNetPBMReader::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpxml::ElementPointer reader = GetReader(FilePath);
	if(!reader)
		return false;

	const sdpString format = sdpxml::GetAttribute<sdpString>(*reader, "format", "");
	g_return_val_if_fail(format.size(), false);

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "Identified " << format << " bitmap image" << std::endl;

	return true;
}

///////////////////////////////////////////////////////////////
// k3dIBitmapReadFormat implementation

sdpResult CNetPBMReader::PreRead(k3dIBitmap* const Bitmap, SDPCSTRING FilePath, gulong* const Width, gulong* const Height)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_fail(Width, SDP_E_FAIL);
	g_return_val_if_fail(Height, SDP_E_FAIL);

	// Convert the source file to a ppm file ...
	g_return_val_if_fail(ConvertToPNM(FilePath, m_PNMFile), SDP_E_FAIL);

	// Read the PNM file header to get its dimensions ...
	std::ifstream file(m_PNMFile);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);

	gulong width = 0;
	gulong height = 0;
	ImageType type;

	g_return_val_if_fail(ReadPNMHeader(file, width, height, type), SDP_E_FAIL);
	g_return_val_if_fail(width, SDP_E_FAIL);
	g_return_val_if_fail(height, SDP_E_FAIL);

	// Return 'em ...
	*Width = width;
	*Height = height;

	return SDP_S_OK;
}

sdpResult CNetPBMReader::ReadOptions(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CNetPBMReader::ReadFile(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_fail(m_PNMFile.size(), SDP_E_FAIL);

	// Open the file ... (binary mode to avoid auto-conversion of std::endl)
	std::ifstream file(m_PNMFile, ios::in | ios::bin);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);

	// Load it up ...
	g_return_val_if_fail(ReadPNM(file, Bitmap), SDP_E_FAIL);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper functions

bool CNetPBMReader::LoadFormats()
{
	if(m_Formats.Children().size())
		return true;

	const sdpString formatfile = sdpAppendPath(k3dApplication()->SharePath(), "netpbm.conf");
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << g_pretty_function() << ": Loading " << formatfile << std::endl;

	if(!m_Formats.Load(formatfile))
		{
			std::cerr << g_pretty_function() << ": Error loading " << formatfile << " configuration file" << std::endl;
			return false;
		}

	return true;
}

sdpxml::ElementPointer CNetPBMReader::GetReader(SDPCSTRING FilePath)
{
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << "CNetPBMReader: obtaining reader for [" << FilePath << "]" << std::endl;

	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, 0);

	// Load our external format description file ...
	g_return_val_if_fail(LoadFormats(), 0);

	// Make sure we've got something ...
	g_return_val_if_fail(m_Formats.Children().size(), 0);
	g_return_val_if_fail(m_Formats.Name() == "netpbm", 0);

	sdpxml::ElementPointer readers = sdpxml::FindElement(m_Formats, sdpxml::SameName("readers"));
	g_return_val_if_fail(readers, 0);

//#ifndef SDPWIN32
	// First, see if the filetype can be identified by /etc/magic through the "file" command ...
	sdpString commandline = "file -b \"" + sdpString(FilePath) + "\"";

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << commandline << std::endl;

	std::ostrstream filetypebuffer;
	if(sdpRunProcess(commandline, filetypebuffer))
		{
			sdpString filetype = filetypebuffer.str();
			filetypebuffer.freeze(false);

			// For each netpbm reader ...
			for(sdpxml::ElementCollection::iterator reader = readers->Children().begin(); reader != readers->Children().end(); reader++)
				{
					// See if it has a "magic" regular expression ...
					const sdpString magic = sdpxml::GetAttribute<sdpString>(*reader, "magic", "");
					if(0 == magic.size())
						continue;

					sdpRegEx expression;
					if(!expression.Create(magic))
						{
							std::cerr << "Error compiling magic regular expression " << magic << std::endl;
							continue;
						}

					// See if the "magic" regular expression has any matches against the output of the "file" command ...
					sdpRegExMatchArray matches;
					expression.Search(filetype, matches);

					if(matches.size() > 0)
						return &(*reader);
				}
		}
//#endif // !SDPWIN32

	// Second try ... look at filename for standard file extensions ...
	for(sdpxml::ElementCollection::iterator reader = readers->Children().begin(); reader != readers->Children().end(); reader++)
		{
			// See if it has a "magic" regular expression ...
			const sdpString filename = sdpxml::GetAttribute<sdpString>(*reader, "filename", "");
			if(0 == filename.size())
				continue;

			sdpRegEx expression;
			if(!expression.Create(filename, false))
				{
					std::cerr << "Error compiling filename regular expression " << filename << std::endl;
					continue;
				}

			// See if the filename regular expression has any matches against the input filename ...
			sdpRegExMatchArray matches;
			expression.Search(FilePath, matches);

			if(matches.size() > 0)
				return &(*reader);
		}

	return 0;
}

bool CNetPBMReader::ConvertToPNM(SDPCSTRING FilePath, sdpString& PNMFilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	PNMFilePath.clear();

	sdpxml::ElementPointer reader = GetReader(FilePath);
	g_return_val_if_fail(reader, false);

	sdpString command = sdpxml::GetAttribute<sdpString>(*reader, "command", "");
	g_return_val_if_fail(command.size(), false);

	command.formattedreplace('%', "p", FilePath);

	sdpString pnmfile = sdpAppendPath(sdpGetTempDirectory(), "netpbmreader.pnm");
	sdpString commandline = command + " > " + pnmfile;

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << commandline << std::endl;

	g_return_val_if_fail(sdpRunProcess(commandline, std::cout), false);

	PNMFilePath = pnmfile;
	return true;
}

bool CNetPBMReader::ReadPNMHeader(std::istream& Stream, gulong& Width, gulong& Height, ImageType& Type)
{
	// Sanity checks ...
	g_return_val_if_fail(Stream.good(), false);

	// A buffer for reading lines ...
	sdpString linebuffer;

	// Check for magic characters at the beginning of the file ...
	Stream.getline(linebuffer.buffer(1024), 1024);
	if(Stream.eof())
		return false;

	if(linebuffer.substr(0, 2) == "P4")
		Type = PBM;
	else if(linebuffer.substr(0, 2) == "P5")
		Type = PGM;
	else if(linebuffer.substr(0, 2) == "P6")
		Type = PPM;
	else
		return false;

	// Detect (optional) comments here ...
	while(Stream.peek() == '#')
		Stream.getline(linebuffer.buffer(1024), 1024);

	// Get the file dimensions ...
	Stream >> Width >> Height;
	if(Stream.eof())
		g_return_val_if_fail(0, false);

	// If it's a bitmap image, we're done ...
	if(PBM == Type)
		{
			Stream >> std::ws;
			return true;
		}

	// Detect (optional) comments here ...
	while(Stream.peek() == '#')
		Stream.getline(linebuffer.buffer(1024), 1024);

	// Get the maximum per-channel value for the image ...
	gulong maxvalue = 0;
	Stream >> maxvalue >> std::ws;
	if(Stream.eof())
		g_return_val_if_fail(0, false);

	return true;
}

bool CNetPBMReader::ReadPNM(std::istream& Stream, k3dIBitmap* Bitmap)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Bitmap);

	gulong sourcewidth = 0;
	gulong sourceheight = 0;
	ImageType sourcetype;

	g_return_val_if_fail(ReadPNMHeader(Stream, sourcewidth, sourceheight, sourcetype), false);
	g_return_val_if_fail(sourcewidth, false);
	g_return_val_if_fail(sourceheight, false);

	// Get the destination bitmap buffer ...
	k3dPixel* destination = Bitmap->Data();
	gulong destinationwidth = Bitmap->Width();
	gulong destinationheight = Bitmap->Height();

	// Sanity checks ...
	g_return_val_if_fail(destination, false);
	g_return_val_if_fail(sourcewidth == destinationwidth, false);
	g_return_val_if_fail(sourceheight == destinationheight, false);

	switch(sourcetype)
		{
			case PBM:
				return ReadPBM(Stream, destination, sourcewidth, sourceheight);
			case PGM:
				return ReadPGM(Stream, destination, sourcewidth, sourceheight);
			case PPM:
				return ReadPPM(Stream, destination, sourcewidth, sourceheight);
			default:
				g_return_val_if_fail(0, false);
		}

	return false;
}

bool CNetPBMReader::ReadPBM(std::istream& Stream, k3dPixel* Destination, gulong Width, gulong Height)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Destination);
	g_assert(Width);
	g_assert(Height);

	gchar packed, pixel;

	// For each source row ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					if(0 == x%8)
						{
							Stream.get(packed);
							g_return_val_if_fail(!Stream.eof(), false);
						}

					pixel = (packed >> (7 - (x%8))) & 0x01;

					Destination->Red = pixel * 255;
					Destination->Green = pixel * 255;
					Destination->Blue = pixel * 255;
					Destination->Alpha = 255;

					Destination++;
				}
		}
	return true;
}

bool CNetPBMReader::ReadPGM(std::istream& Stream, k3dPixel* Destination, gulong Width, gulong Height)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Destination);
	g_assert(Width);
	g_assert(Height);

	gchar grey;

	// For each source row ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					Stream.get(grey);
					g_return_val_if_fail(!Stream.eof(), false);

					Destination->Red = grey;
					Destination->Green = grey;
					Destination->Blue = grey;
					Destination->Alpha = 255;

					Destination++;
				}
		}
	return true;
}

bool CNetPBMReader::ReadPPM(std::istream& Stream, k3dPixel* Destination, gulong Width, gulong Height)
{
	// Sanity checks ...
	g_assert(Stream.good());
	g_assert(Destination);
	g_assert(Width);
	g_assert(Height);

	gchar red, green, blue;

	// For each source row ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					Stream.get(red);
					Stream.get(green);
					Stream.get(blue);
					g_return_val_if_fail(!Stream.eof(), false);

					Destination->Red = red;
					Destination->Green = green;
					Destination->Blue = blue;
					Destination->Alpha = 255;

					Destination++;
				}
		}
	return true;
}


