// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CPBMWriter K-3D object, which writes bitmap images using the NetPBM tools
		\author Tim Shead (tshead@k-3d.com)
*/

#include "netpbmwriter.h"

#include <k3dibitmap.h>
#include <k3dmisc.h>

#include <k3diapplication.h>

#include <stdio.h>
#include <fstream>
#include <strstream>

CNetPBMWriter::CNetPBMWriter(sdpIUnknown* Initializer) :
	CNetPBMWriterBase(Initializer),
	m_Formats("netpbm")
{
}

CNetPBMWriter::~CNetPBMWriter()
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat overrides

bool CNetPBMWriter::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	if(!GetWriter(FilePath))
		return false;

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// k3dIBitmapWriteFormat implementation

sdpResult CNetPBMWriter::PreWrite(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	g_return_val_if_fail(GetWriter(FilePath), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CNetPBMWriter::WriteOptions(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	g_return_val_if_fail(GetWriter(FilePath), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CNetPBMWriter::WriteFile(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, SDP_E_FAIL);
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	sdpxml::ElementPointer writer = GetWriter(FilePath);
	g_return_val_if_fail(writer, SDP_E_FAIL);

	// Save the image as a PPM file ...
	sdpString ppmfilepath = sdpAppendPath(sdpGetTempDirectory(), "netpbmwriter.ppm");
	g_return_val_if_fail(SaveAsPPM(Bitmap, ppmfilepath), SDP_E_FAIL);

	// Convert the PPM file to the desired end file format ...
	sdpString command = sdpxml::GetAttribute<sdpString>(*writer, "command", "");
	g_return_val_if_fail(command.size(), false);

	command.formattedreplace('%', "p", ppmfilepath);
	sdpString commandline = command + " > " + sdpString(FilePath);

	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << commandline << std::endl;

	g_return_val_if_fail(sdpRunProcess(commandline, std::cout), SDP_E_FAIL);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper functions

bool CNetPBMWriter::LoadFormats()
{
	if(m_Formats.Children().size())
		return true;

	const sdpString formatfile = sdpAppendPath(k3dApplication()->SharePath(), "netpbm.conf");
	if(SDP_S_OK == k3dApplication()->Verbose())
		std::cout << g_pretty_function() << ": Loading " << formatfile << std::endl;

	if(!m_Formats.Load(formatfile))
		{
			std::cerr << g_pretty_function() << ": Error loading " << formatfile << " configuration file" << std::endl;
			return false;
		}

	return true;
}

sdpxml::ElementPointer CNetPBMWriter::GetWriter(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, 0);

	// Load our external format description file ...
	g_return_val_if_fail(LoadFormats(), 0);

	// Make sure we've got something ...
	g_return_val_if_fail(m_Formats.Children().size(), 0);
	g_return_val_if_fail(m_Formats.Name() == "netpbm", 0);

	sdpxml::ElementPointer writers = sdpxml::FindElement(m_Formats, sdpxml::SameName("writers"));
	g_return_val_if_fail(writers, 0);

	for(sdpxml::ElementCollection::iterator writer = writers->Children().begin(); writer != writers->Children().end(); writer++)
		{
			// See if it has a "filename" regular expression ...
			const sdpString filename = sdpxml::GetAttribute<sdpString>(*writer, "filename", "");
			if(0 == filename.size())
				continue;

			sdpRegEx expression;
			if(!expression.Create(filename, false))
				{
					std::cerr << "Error compiling filename regular expression " << filename << std::endl;
					continue;
				}

			// See if the filename regular expression has any matches against the input filename ...
			sdpRegExMatchArray matches;
			expression.Search(FilePath, matches);

			if(matches.size() > 0)
				return &(*writer);
		}

	return 0;
}

bool CNetPBMWriter::SaveAsPPM(k3dIBitmap* Bitmap, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Bitmap, false);
	g_return_val_if_string_fail(FilePath, false);

	// Get data from the bitmap ...
	k3dPixel* data = Bitmap->Data();
	gulong width = Bitmap->Width();
	gulong height = Bitmap->Height();

	g_return_val_if_fail(data, false);

	// Create the output stream ... (prefer binary mode, due to Win32 endl)
	std::ofstream file(FilePath, ios::out | ios::bin);
	g_return_val_if_fail(file.good(), false);

	return SaveAsPPM(data, width, height, file);
}

bool CNetPBMWriter::SaveAsPPM(k3dPixel* Data, gulong Width, gulong Height, std::ostream& Stream)
{
	// Write our magic headers ...
	Stream << "P6" << std::endl;
	Stream << Width << " " << Height << std::endl;
	Stream << "255" << std::endl;

	// Write data ...
	for(gulong y = 0; y < Height; y++)
		{
			for(gulong x = 0; x < Width; x++)
				{
					Stream << Data->Red << Data->Green << Data->Blue;
					Data++;
				}
		}

	return true;
}

