// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CAxes K-3D object, which provides a configurable set of axes to help in visualizing the 3D workspace
		\author Tim Shead (tshead@k-3d.com)
*/

#include "axes.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>
#include <k3dieditorengine.h>
#include <k3dicamera.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

double CAxes::m_XAxisColor[4] = {1.0, 0.0, 0.0, 1.0};
double CAxes::m_YAxisColor[4] = {1.0, 1.0, 0.0, 1.0};
double CAxes::m_ZAxisColor[4] = {0.0, 1.0, 0.0, 1.0};
double CAxes::m_GridColor[4] = {0.4, 0.4, 0.4, 1.0};

CAxes::CAxes(sdpIUnknown* Initializer) :
	CAxesBase(Initializer),
	m_Axes("axes", true, Document()->StateRecorder(), 0),
	m_XYPlane("xyplane", false, Document()->StateRecorder(), 0),
	m_YZPlane("yzplane", false, Document()->StateRecorder(), 0),
	m_XZPlane("xzplane", true, Document()->StateRecorder(), 0),
	m_GridSize("gridsize", 2.0, Document()->StateRecorder(), 0),
	m_GridCount("gridcount", 5, Document()->StateRecorder(), new k3dVariableMinimum<gdouble>(1.0, new k3dVariableTruncate(0)))
{
	AddVariable(&m_Axes);
	AddVariable(&m_XYPlane);
	AddVariable(&m_YZPlane);
	AddVariable(&m_XZPlane);
	AddVariable(&m_GridSize);
	AddVariable(&m_GridCount);
}

CAxes::~CAxes()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CAxes::EditObject()
{
	new CAxesDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CAxes::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	sdpglTrace("CAxes::EditorRender()");

	StoreOpenGLAttributes attributes;

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glDisable(GL_CULL_FACE);

	glLineWidth(1.0f);
	glDisable(GL_LINE_STIPPLE);

	gdouble size = m_GridCount.Value() * m_GridSize.Value();

	// Draw axes and labels
	if(m_Axes.Value())
		{
			// Draw X axis
			glColor3dv(m_XAxisColor);
			glBegin(GL_LINE_LOOP);
			glVertex3d(-size, 0.0, 0.0);
			glVertex3d(size, 0.0, 0.0);
			glEnd();

			// Draw Y axis
			glColor3dv(m_YAxisColor);
			glBegin(GL_LINE_LOOP);
			glVertex3d(0.0, -size, 0.0);
			glVertex3d(0.0, size, 0.0);
			glEnd();

			// Draw Z axis
			glColor3dv(m_ZAxisColor);
			glBegin(GL_LINE_LOOP);
			glVertex3d(0.0, 0.0, -size);
			glVertex3d(0.0, 0.0, size);
			glEnd();


			// Draw axis labels ...
			glColor3dv(m_GridColor);

			gdouble labelposition = size * 1.1;
			
			glListBase(Engine->EditorFontBase());
			
			glRasterPos3d(labelposition, 0, 0);
			glCallLists(2, GL_UNSIGNED_BYTE, (GLubyte*)"+X");
			glRasterPos3d(0, labelposition, 0);
			glCallLists(2, GL_UNSIGNED_BYTE, (GLubyte*)"+Y");
			glRasterPos3d(0, 0, labelposition);
			glCallLists(2, GL_UNSIGNED_BYTE, (GLubyte*)"+Z");

			glRasterPos3d(-labelposition, 0, 0);
			glCallLists(2, GL_UNSIGNED_BYTE, (GLubyte*)"-X");
			glRasterPos3d(0, -labelposition, 0);
			glCallLists(2, GL_UNSIGNED_BYTE, (GLubyte*)"-Y");
			glRasterPos3d(0, 0, -labelposition);
			glCallLists(2, GL_UNSIGNED_BYTE, (GLubyte*)"-Z");
		}

	// Setup grid color
	glColor3dv(m_GridColor);

	// Draw XY plane
	if(m_XYPlane.Value())
		{
			glBegin(GL_LINES);
			for(gdouble x = -1.0 * m_GridCount.Value(); x <= m_GridCount.Value(); x++)
				{
					glVertex3d(x * m_GridSize.Value(), -size, 0.0);
					glVertex3d(x * m_GridSize.Value(), size, 0.0);
				}
			for(gdouble y = -1.0 * m_GridCount.Value(); y <= m_GridCount.Value(); y++)
				{
					glVertex3d(-size, y * m_GridSize.Value(), 0.0);
					glVertex3d(size, y * m_GridSize.Value(), 0.0);
				}
			glEnd();
		}

	// Draw YZ plane
	if(m_YZPlane.Value())
		{
			glBegin(GL_LINES);
			for(gdouble z = -1.0 * m_GridCount.Value(); z <= m_GridCount.Value(); z++)
				{
					glVertex3d(0.0, -size, z * m_GridSize.Value());
					glVertex3d(0.0, size, z * m_GridSize.Value());
				}
			for(gdouble y = -1.0 * m_GridCount.Value(); y <= m_GridCount.Value(); y++)
				{
					glVertex3d(0.0, y * m_GridSize.Value(), -size);
					glVertex3d(0.0, y * m_GridSize.Value(), size);
				}
			glEnd();
		}

	// Draw XZ plane
	if(m_XZPlane.Value())
		{
			glBegin(GL_LINES);
			for(gdouble x = -1.0 * m_GridCount.Value(); x <= m_GridCount.Value(); x++)
				{
					glVertex3d(x * m_GridSize.Value(), 0.0, -size);
					glVertex3d(x * m_GridSize.Value(), 0.0, size);
				}
			for(gdouble z = -1.0 * m_GridCount.Value(); z <= m_GridCount.Value(); z++)
				{
					glVertex3d(-size, 0.0, z * m_GridSize.Value());
					glVertex3d(size, 0.0, z * m_GridSize.Value());
				}
			glEnd();
		}

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, size);
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// k3dISnapTarget overrides
sdpResult CAxes::Snap(sdpVector3& Coordinates, sdpString& Description)
{
	// Convert coordinates to our frame ...
	sdpVector3 coordinates = k3dGetWorldToObjectMatrix(this) * Coordinates;

	// Snap coordinates ...
	coordinates[0] = sdpRound(coordinates[0] / m_GridSize.Value()) * m_GridSize.Value();
	coordinates[1] = sdpRound(coordinates[1] / m_GridSize.Value()) * m_GridSize.Value();
	coordinates[2] = sdpRound(coordinates[2] / m_GridSize.Value()) * m_GridSize.Value();

	// Convert coordinates back to world frame ...
	Coordinates = k3dGetObjectToWorldMatrix(this) * coordinates;
	Description = "Axes Grid";

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CAxesDialog

const sdpString CAxesDialog::dialogname = "properties";
const sdpString CAxesDialog::control_axes = "axes";
const sdpString CAxesDialog::control_xyplane = "xyplane";
const sdpString CAxesDialog::control_yzplane = "yzplane";
const sdpString CAxesDialog::control_xzplane = "xzplane";
const sdpString CAxesDialog::control_gridsize = "gridsize";
const sdpString CAxesDialog::control_gridcount = "gridcount";
const sdpString CAxesDialog::control_position = "position";
const sdpString CAxesDialog::control_orientation = "orientation";
const sdpString CAxesDialog::control_scale = "scale";
const sdpString CAxesDialog::control_target = "target";

CAxesDialog::CAxesDialog(CAxes* const Object) :
	CAxesDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "axes.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_gridsize, m_Object->m_GridSize, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Axes Grid Size", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dSpinButton(control_gridcount, m_Object->m_GridCount, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Axes Grid Count", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Connect check buttons
	Connectk3dCheckButton(control_axes, m_Object->m_Axes, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Display Axes", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_xyplane, m_Object->m_XYPlane, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Display XY Plane", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_yzplane, m_Object->m_YZPlane, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Display YZ Plane", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_xzplane, m_Object->m_XZPlane, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Display XZ Plane", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}

void CAxesDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	CAxesDialogBase::OnEvent(Event);
}

