// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CCamera K-3D object, which controls the position, orientation, and display of a CCameraFrame within the 3D workspace
		\author Tim Shead <tshead@k-3d.com>
		\author Giuseppe Zompatori <mailus@tin.it>
*/

#include "camera.h"

#include <k3dboundingbox.h>
#include <k3dchooser.h>
#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpathcollection.h>
#include <k3dpathpointcollection.h>
#include <k3dpointcollection.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>
#include <k3dshaderargumentspage.h>
#include <k3dshaderinfo.h>
#include <k3dshaderselector.h>
#include <k3dviewvolume.h>

#include <k3diapplication.h>
#include <k3dihierarchy.h>
#include <k3dimesh.h>
#include <k3dipoint.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>
#include <k3direnderfarm.h>
#include <k3direnderjob.h>
#include <k3direnderframe.h>
#include <k3distaterecorder.h>
#include <k3ditexture.h>

#include <sdpgtkevent.h>
#include <sdpgtkfileselector.h>
#include <sdpgtkutility.h>


#include <gdk/gdkprivate.h>

#if defined SDPWIN32
#include <gdkwin32.h>
#else // SDPWIN32
#include <GL/glx.h>
#endif // !SDPWIN32

#include <strstream>

namespace
{

const sdpVector3 SelectionColor(0.215, 0.215, 0.8);

/// Sets k3dILocation camera coordinates
class SnapshotCameraCoordinates :
	public k3dILocationVisitor
{
public:
	SnapshotCameraCoordinates(const sdpMatrix4& Matrix) :
		m_Matrix(Matrix)
	{
	}

	void operator()(k3dIObject* const Object)
	{
		// Skip objects that don't support locations ...
		k3dILocationCollection* const locations = dynamic_cast<k3dILocationCollection*>(Object);
		if(locations)
			locations->VisitLocations(this);
	}

	void VisitLocation(k3dILocation* const Location)
	{
		Location->SetCameraXYZ(m_Matrix * Location->WorldXYZ());
	}

private:
	sdpMatrix4 m_Matrix;
};

/// Sets screen coordinates for each k3dILocation
class SnapshotScreenCoordinates :
	k3dILocationVisitor
{
public:
	SnapshotScreenCoordinates(k3dIViewVolume* const ViewVolume, const gdouble WindowPixelWidth, const gdouble WindowPixelHeight) :
		m_ViewVolume(ViewVolume),
		m_WindowPixelWidth(WindowPixelWidth),
		m_WindowPixelHeight(WindowPixelHeight)
	{
		// Sanity checks ...
		g_assert(m_ViewVolume);
	}

	void operator()(k3dIObject* const Object)
	{
		// Skip objects that don't support locations ...
		k3dILocationCollection* const locations = dynamic_cast<k3dILocationCollection*>(Object);
		if(locations)
			locations->VisitLocations(this);
	}

	void VisitLocation(k3dILocation* const Location)
	{
		sdpVector3 coords = Location->CameraXYZ();
		m_ViewVolume->CameraToScreen(coords, m_WindowPixelWidth, m_WindowPixelHeight);
		Location->SetScreenXYZ(coords);
	}

private:
	k3dIViewVolume* const m_ViewVolume;
	const gdouble m_WindowPixelWidth;
	const gdouble m_WindowPixelHeight;
};

/// Functor for selecting a point
class SelectSinglePoint :
	public k3dIPointVisitor
{
public:
	SelectSinglePoint(const sdpVector2 Mouse, const bool SelectMode) :
		m_Mouse(Mouse),
		m_SelectMode(SelectMode),
		m_PointSelected(false)
	{
	}

	bool PointSelected() const
	{
		return m_PointSelected;
	}

	void operator()(k3dIPointCollection* const PointCollection)
	{
		// If we already found something, we're done
		if(m_PointSelected)
			return;

		PointCollection->VisitPoints(this);
	}

	void VisitPoint(k3dIPoint* const Point)
	{
		// If we already found something, we're done ...
		if(m_PointSelected)
			return;

		const sdpVector3 screencoords = Point->Location()->ScreenXYZ();
		if((fabs(screencoords[0] - m_Mouse[0]) < 4.0) && (fabs(screencoords[1] - m_Mouse[1]) < 4.0))
			{
				Point->State()->SetState(k3dIState::SELECTED, m_SelectMode);
				m_PointSelected = true;
			}
/*
				if(selectmode)
					Document()->SetActivePoint(point);
				else
					if(point == activepoint)
						Document()->SetActivePoint(NULL);
*/
	}

private:
	const sdpVector2 m_Mouse;
	const bool m_SelectMode;
	bool m_PointSelected;
};

/// Functor for selecting paths
class SelectSinglePath :
	public k3dIPathVisitor
{
public:
	SelectSinglePath(const sdpVector2 Mouse, const bool SelectMode) :
		m_Mouse(Mouse),
		m_SelectMode(SelectMode),
		m_PathSelected(false)
	{
	}

	bool PathSelected() const
       	{
		return m_PathSelected;
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		// If we already found something, we're done
		if(m_PathSelected)
			return;

		PathCollection->VisitPaths(this);
	}

	void VisitPath(k3dIPath* const Path)
	{
		// If we already found something, we're done ...
		if(m_PathSelected)
			return;

		// If the mouse isn't inside the path, skip it ...
		if(!k3dIsInside(m_Mouse, Path))
			return;

		// Select the path ...
		k3dDeepSetState(Path, k3dIState::SELECTED, m_SelectMode);
		m_PathSelected = true;
	}

private:
	const sdpVector2 m_Mouse;
	const bool m_SelectMode;
	bool m_PathSelected;
};

/// Functor for selecting a mesh object
class SelectSingleMesh
{
public:
	SelectSingleMesh(k3dIDocument* const Document, const sdpVector2 Mouse, const bool SelectMode) :
		m_Document(Document),
		m_Mouse(Mouse),
		m_SelectMode(SelectMode),
		m_MeshSelected(false)
	{
		// Sanity checks ...
		g_assert(m_Document);
	}

	bool MeshSelected() const
	{
		return m_MeshSelected;
	}

	void operator()(k3dIMesh* const Mesh)
	{
		// If we already found something, we're done ...
		if(m_MeshSelected)
			return;

		// Calculate the mesh's screen bounding box ...
		k3dBoundingBox2 boundingbox;

		sdpVector3 screencoords;
		k3dIPointCollection* const points = Mesh->Points();
		for(K3DPOINTHANDLE pointhandle = points->FirstPoint(); pointhandle; pointhandle = points->NextPoint(pointhandle))
			{
				screencoords = points->Point(pointhandle)->Location()->ScreenXYZ();
				boundingbox.Expand(sdpVector2(screencoords, 2));
			}

		// If the mouse click is outside the screen bounding box we're done ...
		if(!boundingbox.Contains(m_Mouse))
			return;

		// For each path in the mesh ...
		k3dIPathCollection* const paths = Mesh->Paths();
		for(K3DPATHHANDLE pathhandle = paths->FirstPath(); pathhandle; pathhandle = paths->NextPath(pathhandle))
			{
				k3dIPath* const path = paths->Path(pathhandle);
				if(SDP_S_OK != path->State()->QueryState(k3dIState::VISIBLESELECTABLE))
					continue;

				// If the mouse click is outside the path, skip it ...
				if(!k3dIsInside(m_Mouse, path))
					continue;

				// Select/deselect the mesh ...
				k3dDeepSetState(Mesh, k3dIState::SELECTED, m_SelectMode);

				// If selecting, make the mesh the active object ...
				if(m_SelectMode)
					m_Document->SetMouseFocus(dynamic_cast<k3dIMouseEventObserver*>(Mesh));

				// We're done ...
				m_MeshSelected = true;
				return;
			}

		return;
	}

private:
	k3dIDocument* const m_Document;
	const sdpVector2 m_Mouse;
	const bool m_SelectMode;
	bool m_MeshSelected;
};

/// Functor object for selecting points with a rubber-band box
class SelectBoxPoints :
	public k3dIPointVisitor
{
public:
	SelectBoxPoints(const sdpVector2& UpperLeft, const sdpVector2& LowerRight, const bool SelectMode) :
		m_BoundingBox(UpperLeft, LowerRight),
		m_SelectMode(SelectMode),
		m_Selected(false)
	{
std::cout << m_SelectMode << std::endl;
	}

	const bool Selected() const
	{
		return m_Selected;
	}

	void operator()(k3dIPointCollection* const PointCollection)
	{
		PointCollection->VisitPoints(this);
	}

	void VisitPoint(k3dIPoint* const Point)
	{
		if(m_BoundingBox.Contains(sdpVector2(Point->Location()->ScreenXYZ(), 2)))
			{
				m_Selected = true;
				Point->State()->SetState(k3dIState::SELECTED, m_SelectMode);
			}
	}

private:
	const k3dBoundingBox2 m_BoundingBox;
	const bool m_SelectMode;
	bool m_Selected;
};

/// Functor object for selecting paths with a rubber-band box
class SelectBoxPaths :
	public k3dIPathVisitor
{
public:
	SelectBoxPaths(const sdpVector2& UpperLeft, const sdpVector2& LowerRight, const bool SelectMode, const bool SelectPartial) :
		m_BoundingBox(UpperLeft, LowerRight),
		m_SelectMode(SelectMode),
		m_SelectPartial(SelectPartial),
		m_Selected(false)
	{
	}

	const bool Selected() const
	{
		return m_Selected;
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		PathCollection->VisitPaths(this);
	}

	void VisitPath(k3dIPath* const Path)
	{
		bool someinside = false;
		bool allinside = true;

		// For each point in the path ...
		k3dIPathPointCollection* const points = Path->Points();
		for(K3DPATHPOINTHANDLE pointhandle = points->FirstPathPoint(); pointhandle; pointhandle = points->NextPathPoint(pointhandle))
			{
				if(m_BoundingBox.Contains(sdpVector2(points->PathPoint(pointhandle)->Point()->Location()->ScreenXYZ(), 2)))
					{
						someinside = true;
						if(m_SelectPartial)
							break;
					}
				else
					{
						allinside = false;
						if(!m_SelectPartial)
							return;
					}
			}

		// If none of the points were inside the bounding-box, we're done ...
		if(!someinside)
			return;

		// If select partial is off, and we didn't get every point, we're done ...
		if(!allinside && !m_SelectPartial)
			return;

		k3dDeepSetState(Path, k3dIState::SELECTED, m_SelectMode);
		m_Selected = true;
	}

private:
	const k3dBoundingBox2 m_BoundingBox;
	const bool m_SelectMode;
	const bool m_SelectPartial;
	bool m_Selected;
};

/// Functor object for selecting points with a lasso
class SelectLassoPoints :
	public k3dIPointVisitor
{
public:
	SelectLassoPoints(const sdpVector2Array& LassoPoints, const bool SelectMode) :
		m_LassoPoints(LassoPoints),
		m_SelectMode(SelectMode),
		m_Selected(false)
	{
	}

	const bool Selected() const
	{
		return m_Selected;
	}

	void operator()(k3dIPointCollection* const PointCollection)
	{
		PointCollection->VisitPoints(this);
	}

	void VisitPoint(k3dIPoint* const Point)
	{
		if(sdpIsInside(sdpVector2(Point->Location()->ScreenXYZ(), 2), m_LassoPoints))
			{
				m_Selected = true;
				Point->State()->SetState(k3dIState::SELECTED, m_SelectMode);
			}
	}

private:
	const sdpVector2Array& m_LassoPoints;
	const bool m_SelectMode;
	bool m_Selected;
};

/// Functor object for selecting paths with a lasso
class SelectLassoPaths :
	public k3dIPathVisitor
{
public:
	SelectLassoPaths(const sdpVector2Array& LassoPoints, const bool SelectMode, const bool SelectPartial) :
		m_LassoPoints(LassoPoints),
		m_SelectMode(SelectMode),
		m_SelectPartial(SelectPartial),
		m_Selected(false)
	{
	}

	const bool Selected() const
	{
		return m_Selected;
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		PathCollection->VisitPaths(this);
	}

	void VisitPath(k3dIPath* const Path)
	{
		bool someinside = false;
		bool allinside = true;

		// For each point in the path ...
		k3dIPathPointCollection* const points = Path->Points();
		for(K3DPATHPOINTHANDLE pointhandle = points->FirstPathPoint(); pointhandle; pointhandle = points->NextPathPoint(pointhandle))
			{
				if(sdpIsInside(sdpVector2(points->PathPoint(pointhandle)->Point()->Location()->ScreenXYZ(), 2), m_LassoPoints))
					{
						someinside = true;
						if(m_SelectPartial)
							break;
					}
				else
					{
						allinside = false;
						if(!m_SelectPartial)
							return;
					}
			}

		// If we couldn't find anything, we're done ...
		if(!someinside)
			return;

		// If select partial is off, and we didn't get every point, we're done ...
		if(!allinside && !m_SelectPartial)
			return;

		k3dDeepSetState(Path, k3dIState::SELECTED, m_SelectMode);
		m_Selected = true;
	}

private:
	const sdpVector2Array m_LassoPoints;
	const bool m_SelectMode;
	const bool m_SelectPartial;
	bool m_Selected;
};

/// Creates an XOR GC in the special selection color
GdkGC* SelectionGC(sdpGtkWidget& Widget)
{
	// Sanity checks ...
	g_return_val_if_fail(Widget.Attached(), 0);

	// Create the rubber-band color ...
	GdkColor color;
	color.red = static_cast<gushort>(SelectionColor[0] * 0xffff);
	color.green = static_cast<gushort>(SelectionColor[1] * 0xffff);
	color.blue = static_cast<gushort>(SelectionColor[2] * 0xffff);
	gdk_color_alloc(gdk_colormap_get_system(), &color);

	// Create an XOR gc for drawing ...
	GdkGC* gc = gdk_gc_new(((GtkWidget*)Widget)->window);
	gdk_gc_set_foreground(gc, &color);
	gdk_gc_set_function(gc, GDK_XOR);

	return gc;
}

/// XORs a rubber-band rectangle into a widget
void RubberBand(sdpGtkWidget& Widget, sdpRectangle Rectangle)
{
	// Sanity checks ...
	g_return_if_fail(Widget.Attached());

	// Get our GC ...
	GdkGC* const gc = SelectionGC(Widget);

	// Normalize coordinates ...
	Rectangle.Normalize();

	// Draw the rectangle ...
	gdk_draw_rectangle(((GtkWidget*)Widget)->window, gc, 0, gint(Rectangle.Left()), gint(Rectangle.Top()), gint(Rectangle.Width()), gint(Rectangle.Height()));

	// Cleanup ...
	gdk_gc_destroy(gc);
}

/// Sets the current mouse input focus, based on the contents of a k3dObjectVariable
class FocusChangeAction :
	public k3dAction
{
public:
	FocusChangeAction(k3dIDocument* const Document, const k3dObjectVariable& Storage, k3dIAction* const NextAction) :
		k3dAction(NextAction),
		m_Document(Document),
		m_Storage(Storage)
	{
		// Sanity checks ...
		g_assert(m_Document);
	}

	void DoAction()
	{
		m_Document->SetMouseFocus(m_Storage.Object() ? dynamic_cast<k3dIMouseEventObserver*>(m_Storage.Object()) : 0);
		DoNextAction();
	}

private:
	k3dIDocument* const m_Document;
	const k3dObjectVariable& m_Storage;
};

} // namespace

//////////////////////////////////////////////////////////////////////////////////////////////////////////
// CCamera

gdouble CCamera::m_EditorBackgroundColor[4] = {0.8, 0.8, 0.8, 1.0};
gdouble CCamera::m_EditorNormalColor[4] = {0.3, 0.3, 0.1, 1.0};
gdouble CCamera::m_EditorSelectedColor[4] = {1.0, 1.0, 1.0, 1.0};
gdouble CCamera::m_EditorUnselectedColor[4] = {0.0, 0.0, 0.0, 1.0};
gdouble CCamera::m_EditorControlPointColor[4] = {0.0, 0.0, 1.0, 1.0};
gdouble CCamera::m_EditorFirstColor[4] = {1.0, 0.0, 1.0, 1.0};
gdouble CCamera::m_EditorActiveColor[4] = {1.0, 0.0, 0.0, 1.0};

GLfloat CCamera::m_PointSize = 4.0f;
gulong CCamera::m_ClickSensitivity = 4;
gdouble CCamera::m_PositionSensitivity = -0.1;
gdouble CCamera::m_OrientationSensitivity = 1.0;
gdouble CCamera::m_ZoomSign = -1.0;

bool CCamera::m_BoxSelectPartial = false;
bool CCamera::m_LassoSelectPartial = false;

CCamera::CCamera(sdpIUnknown* Initializer) :
	CCameraBase(Initializer),
	m_Frame(0),
	m_Time(Document()->Time()),
	m_DisplayEngine(&m_EditorEngine),
	m_RenderEngine(&m_RiEngine),

	// Non-persistent data ...
	m_Focus("focus", Document(), 0, 0),

	// Camera data ...
	m_ActiveAxis("activeaxis", ZAXIS, 0, 0),
	m_ActiveTransformType("activetransformtype", POSITION, 0, 0),
	m_ActiveCoordinateSystem("activecoordinatesystem", WORLD, 0, 0),
	m_ActiveEditMode("activeeditmode", POLYGON, 0, 0),
	m_ActiveSnapMode("activesnapmode", FREEHAND, 0, 0),

	// Viewport options ...
	m_ViewportVisible("viewportvisible", true, Document()->StateRecorder(), 0),

	// Output options ...
	m_OutputPixelWidth("outputpixelwidth", 320, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_OutputPixelHeight("outputpixelheight", 240, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_OutputPixelAspectRatio("outputpixelaspectratio", 1.0, Document()->StateRecorder(), 0),
	m_WindowPixelAspectRatio("windowpixelaspectratio", 1.0, Document()->StateRecorder(), 0),
	m_Orthographic("orthographic", false, Document()->StateRecorder(), 0),
	m_Field("field", EVENFIELDFIRST, Document()->StateRecorder(), 0),

	// Editor options ...
	m_EditorWireframe("editorwireframe", true, Document()->StateRecorder(), 0),
	m_EditorSolid("editorsolid", true, Document()->StateRecorder(), 0),
	m_EditorFog("editorfog", false, Document()->StateRecorder(), 0),
	m_EditorFogNear("editorfognear", 30.0, Document()->StateRecorder(), 0),
	m_EditorFogFar("editorfogfar", 100.0, Document()->StateRecorder(), 0),
	m_EditorSmoothing("editorsmoothing", true, Document()->StateRecorder(), 0),
	m_EditorTexture("editortexture", false, Document()->StateRecorder(), 0),
	m_EditorTwoSided("editortwosided", false, Document()->StateRecorder(), 0),
	m_EditorWireframeTwoSided("editorwireframetwosided", true, Document()->StateRecorder(), 0),
	m_EditorOrientation("editororientation", true, Document()->StateRecorder(), 0),
	m_EditorNormals("editornormals", false, Document()->StateRecorder(), 0),
	m_EditorWireframePatchDivisions("editorwireframepatchdivisions", 4, Document()->StateRecorder(), 0),
	m_EditorSolidPatchDivisions("editorsolidpatchdivisions", 12, Document()->StateRecorder(), 0),

	// Ri-renderer options ...
	m_RiRenderEngine("riengine", "", Document()->StateRecorder(), 0),
	m_RiXSamples("rixsamples", 1, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RiYSamples("riysamples", 1, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RiPixelFilterName("ripixelfiltername", k3dIRiEngine::RI_GAUSSIAN, Document()->StateRecorder(), 0),
	m_RiPixelFilterWidth("ripixelfilterwidth", 2.0, Document()->StateRecorder(), 0),
	m_RiPixelFilterHeight("ripixelfilterheight", 2.0, Document()->StateRecorder(), 0),
	m_RiTwoSided("ritwosided", true, Document()->StateRecorder(), 0),
	m_RiShadingRate("rishadingrate", 5.0, Document()->StateRecorder(), 0),
	m_RiShadingInterpolation("rishadinginterpolation", k3dIRiEngine::RI_CONSTANT, Document()->StateRecorder(), 0),
	m_RiGridSize("rigridsize", 256.0, Document()->StateRecorder(), 0),
	m_RiXBucketSize("rixbucketsize", 16, Document()->StateRecorder(), 0),
	m_RiYBucketSize("riybucketsize", 16, Document()->StateRecorder(), 0),
	m_RiEyeSplits("rieyesplits", 10.0, Document()->StateRecorder(), 0),
	m_RiAtmosphere("riatmosphere", false, Document()->StateRecorder(), 0),
	m_RiAtmosphereShader("riatmosphereshader", Document(), Document()->StateRecorder()),
	m_RiDepthOfField("ridepthoffield", false, Document()->StateRecorder(), 0),
	m_RiMotionBlur("rimotionblur", false, Document()->StateRecorder(), 0),
	m_RiFieldRender("rifieldrender", false, Document()->StateRecorder(), 0),
	m_RiRadiosity("riradiosity", false, Document()->StateRecorder(), 0),
	m_RiRadiositySteps("riradiositysteps", 100, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RiRadiosityPatchSamples("riradiositypatchsamples", 1, Document()->StateRecorder(), 0),
	m_RiRadiosityPatchSize("riradiositypatchsize", 10, Document()->StateRecorder(), 0),
	m_RiRadiosityElementSize("riradiosityelementsize", 5, Document()->StateRecorder(), 0),
	m_RiRadiosityMinimumSize("riradiosityminimumsize", 0.5, Document()->StateRecorder(), 0),
	m_RiTextureMemory("ritexturememory", 1024, Document()->StateRecorder(), 0),
	m_RiEnableLerp("rienablelerp", true, Document()->StateRecorder(), 0),
	m_RiEnableGaussian("rienablelegaussian", true, Document()->StateRecorder(), 0),

	// Owned objects ...
	m_Focallength("focallength", Document(), Document()->StateRecorder(), 0),
	m_Focalplane("focalplane", Document(), Document()->StateRecorder(), 0),
	m_Fstop("fstop", Document(), Document()->StateRecorder(), 0),
	m_Shutteropen("shutteropen", Document(), Document()->StateRecorder(), 0),
	m_Shutterclose("shutterclose", Document(), Document()->StateRecorder(), 0),
	m_Exposure("exposure", Document(), Document()->StateRecorder(), 0),
	m_Gamma("gamma", Document(), Document()->StateRecorder(), 0),
	m_Xdirection("xdirection", Document(), Document()->StateRecorder(), 0),
	m_Ydirection("ydirection", Document(), Document()->StateRecorder(), 0),
	m_Zdirection("zdirection", Document(), Document()->StateRecorder(), 0),
	m_FarPlane("farplane", Document(), Document()->StateRecorder(), 0),
	m_NearPlane("nearplane", Document(), Document()->StateRecorder(), 0)
{
	// Viewport options ...
	AddVariable(&m_ViewportVisible);

	// Camera options ...
	AddVariable(&m_ActiveAxis);
	AddVariable(&m_ActiveTransformType);
	AddVariable(&m_ActiveCoordinateSystem);
	AddVariable(&m_ActiveEditMode);
	AddVariable(&m_ActiveSnapMode);

	// Output options ...
	AddVariable(&m_OutputPixelWidth);
	AddVariable(&m_OutputPixelHeight);
	AddVariable(&m_OutputPixelAspectRatio);
	AddVariable(&m_WindowPixelAspectRatio);
	AddVariable(&m_Orthographic);
	AddVariable(&m_Field);

	// Editor options ...
	AddVariable(&m_EditorSolid);
	AddVariable(&m_EditorFog);
	AddVariable(&m_EditorFogNear);
	AddVariable(&m_EditorFogFar);
	AddVariable(&m_EditorSmoothing);
	AddVariable(&m_EditorTexture);
	AddVariable(&m_EditorTwoSided);
	AddVariable(&m_EditorOrientation);
	AddVariable(&m_EditorNormals);
	AddVariable(&m_EditorWireframePatchDivisions);
	AddVariable(&m_EditorSolidPatchDivisions);

	// Ri-renderer options ...
	AddVariable(&m_RiRenderEngine);
	AddVariable(&m_RiXSamples);
	AddVariable(&m_RiYSamples);
	AddVariable(&m_RiPixelFilterName);
	AddVariable(&m_RiPixelFilterWidth);
	AddVariable(&m_RiPixelFilterHeight);
	AddVariable(&m_RiTwoSided);
	AddVariable(&m_RiShadingRate);
	AddVariable(&m_RiShadingInterpolation);
	AddVariable(&m_RiGridSize);
	AddVariable(&m_RiXBucketSize);
	AddVariable(&m_RiYBucketSize);
	AddVariable(&m_RiEyeSplits);
	AddVariable(&m_RiAtmosphere);
	AddVariable(&m_RiAtmosphereShader);
	AddVariable(&m_RiDepthOfField);
	AddVariable(&m_RiMotionBlur);
	AddVariable(&m_RiFieldRender);
	AddVariable(&m_RiTextureMemory);
	AddVariable(&m_RiEnableLerp);
	AddVariable(&m_RiEnableGaussian);
	AddVariable(&m_RiRadiosity);
	AddVariable(&m_RiRadiositySteps);
	AddVariable(&m_RiRadiosityPatchSamples);
	AddVariable(&m_RiRadiosityPatchSize);
	AddVariable(&m_RiRadiosityElementSize);
	AddVariable(&m_RiRadiosityMinimumSize);

	// Owned objects ...
	AddVariable(&m_Focallength);
	AddVariable(&m_Focalplane);
	AddVariable(&m_Fstop);
	AddVariable(&m_Shutteropen);
	AddVariable(&m_Shutterclose);
	AddVariable(&m_Exposure);
	AddVariable(&m_Gamma);
	AddVariable(&m_Xdirection);
	AddVariable(&m_Ydirection);
	AddVariable(&m_Zdirection);
	AddVariable(&m_FarPlane);
	AddVariable(&m_NearPlane);

	k3dIEventSource* const documenteventserver = dynamic_cast<k3dIEventSource*>(Document());
	g_return_if_fail(documenteventserver);
	documenteventserver->AddEventObserver(this, k3dIDocument::Event_MouseFocusChanged, "CCamera");
}

CCamera::~CCamera()
{
//	std::cout << "Destroying CCamera" << std::endl;
	k3dIEventSource* const documenteventserver = dynamic_cast<k3dIEventSource*>(Document());
	g_return_if_fail(documenteventserver);
	documenteventserver->RemoveEventObserver(this);
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CCamera::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	if(event == k3dIDocument::Event_MouseFocusChanged)
		{
			m_Focus.SetObject(Document()->MouseFocus() ? dynamic_cast<k3dIObject*>(Document()->MouseFocus()) : 0);
			return SDP_S_OK;
		}

	return CCameraBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CCamera::EditObject()
{
	new CCameraDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate implementation

sdpResult CCamera::PostAnimate()
{
	SetupViewVolume();

	return CCameraBase::PostAnimate();
}

// sdpsl::IShaderCollection implementation
void CCamera::VisitShaders(sdpsl::IShaderVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	if(m_RiAtmosphereShader.Shader())
		Visitor->VisitShader(m_RiAtmosphereShader.Shader());
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender implementation ...

sdpResult CCamera::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	// Don't draw ourself for ourself ...
	if(Engine == &m_EditorEngine)
		return SDP_S_FALSE;

	sdpglTrace("CCamera::EditorRender()");

	StoreOpenGLAttributes attributes;

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glColor3d(0, 0, 0);
	glLineWidth(1.0f);
	glDisable(GL_LINE_STIPPLE);

	// Our dimensions
	const gdouble bodylength = 0.5 * 0.5;
	const gdouble bodywidth = 0.25 * 0.5;
	const gdouble bodyheight = 0.25 * 0.5;
	const gdouble lenslength = 0.25 * 0.5;
	const gdouble lenswidth = 0.15 * 0.5;
	const gdouble lensheight = 0.1 * 0.5;
	const gdouble lensoffset = 0.05;
	const gdouble filmradius = 0.2;
	const gdouble filmwidth = 0.125 * 0.5;

	// Draw the camera body ...
	glBegin(GL_LINE_LOOP);
	glVertex3d(-bodywidth, bodyheight, bodylength);
	glVertex3d(bodywidth, bodyheight, bodylength);
	glVertex3d(bodywidth, bodyheight, -bodylength);
	glVertex3d(-bodywidth, bodyheight, -bodylength);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(-bodywidth, -bodyheight, bodylength);
	glVertex3d(bodywidth, -bodyheight, bodylength);
	glVertex3d(bodywidth, -bodyheight, -bodylength);
	glVertex3d(-bodywidth, -bodyheight, -bodylength);
	glEnd();

	glBegin(GL_LINES);
	glVertex3d(-bodywidth, bodyheight, bodylength);
	glVertex3d(-bodywidth, -bodyheight, bodylength);
	glVertex3d(bodywidth, bodyheight, bodylength);
	glVertex3d(bodywidth, -bodyheight, bodylength);
	glVertex3d(bodywidth, bodyheight, -bodylength);
	glVertex3d(bodywidth, -bodyheight, -bodylength);
	glVertex3d(-bodywidth, bodyheight, -bodylength);
	glVertex3d(-bodywidth, -bodyheight, -bodylength);
	glEnd();

	// Draw the camera lens ...
	glBegin(GL_LINE_LOOP);
	glVertex3d(-lenswidth, lensheight, bodylength);
	glVertex3d(lenswidth, lensheight, bodylength);
	glVertex3d(lenswidth, -lensheight, bodylength);
	glVertex3d(-lenswidth, -lensheight, bodylength);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(-lenswidth - lensoffset, lensheight + lensoffset, bodylength + lenslength);
	glVertex3d(lenswidth + lensoffset, lensheight + lensoffset, bodylength + lenslength);
	glVertex3d(lenswidth + lensoffset, -lensheight - lensoffset, bodylength + lenslength);
	glVertex3d(-lenswidth - lensoffset, -lensheight - lensoffset, bodylength + lenslength);
	glEnd();

	glBegin(GL_LINES);
	glVertex3d(-lenswidth, lensheight, bodylength);
	glVertex3d(-lenswidth - lensoffset, lensheight + lensoffset, bodylength + lenslength);
	glVertex3d(lenswidth, lensheight, bodylength);
	glVertex3d(lenswidth + lensoffset, lensheight + lensoffset, bodylength + lenslength);
	glVertex3d(lenswidth, -lensheight, bodylength);
	glVertex3d(lenswidth + lensoffset, -lensheight - lensoffset, bodylength + lenslength);
	glVertex3d(-lenswidth, -lensheight, bodylength);
	glVertex3d(-lenswidth - lensoffset, -lensheight - lensoffset, bodylength + lenslength);
	glEnd();

	// Draw the film can ...
	glBegin(GL_LINE_LOOP);
	glVertex3d(-filmwidth, bodyheight, -bodylength);
	glVertex3d(-filmwidth, bodyheight, -bodylength + filmradius);
	glVertex3d(-filmwidth, bodyheight + 0.8 * filmradius, -bodylength + 0.8 * filmradius);
	glVertex3d(-filmwidth, bodyheight + filmradius, -bodylength);
	glVertex3d(-filmwidth, bodyheight + 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(-filmwidth, bodyheight, -bodylength - filmradius);
	glVertex3d(-filmwidth, bodyheight - 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(-filmwidth, bodyheight - filmradius, -bodylength);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(filmwidth, bodyheight, -bodylength);
	glVertex3d(filmwidth, bodyheight, -bodylength + filmradius);
	glVertex3d(filmwidth, bodyheight + 0.8 * filmradius, -bodylength + 0.8 * filmradius);
	glVertex3d(filmwidth, bodyheight + filmradius, -bodylength);
	glVertex3d(filmwidth, bodyheight + 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(filmwidth, bodyheight, -bodylength - filmradius);
	glVertex3d(filmwidth, bodyheight - 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(filmwidth, bodyheight - filmradius, -bodylength);
	glEnd();

	glBegin(GL_LINES);
	glVertex3d(-filmwidth, bodyheight, -bodylength + filmradius);
	glVertex3d(filmwidth, bodyheight, -bodylength + filmradius);
	glVertex3d(-filmwidth, bodyheight + 0.8 * filmradius, -bodylength + 0.8 * filmradius);
	glVertex3d(filmwidth, bodyheight + 0.8 * filmradius, -bodylength + 0.8 * filmradius);
	glVertex3d(-filmwidth, bodyheight + filmradius, -bodylength);
	glVertex3d(filmwidth, bodyheight + filmradius, -bodylength);
	glVertex3d(-filmwidth, bodyheight + 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(filmwidth, bodyheight + 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(-filmwidth, bodyheight, -bodylength - filmradius);
	glVertex3d(filmwidth, bodyheight, -bodylength - filmradius);
	glVertex3d(-filmwidth, bodyheight - 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(filmwidth, bodyheight - 0.8 * filmradius, -bodylength - 0.8 * filmradius);
	glVertex3d(-filmwidth, bodyheight - filmradius, -bodylength);
	glVertex3d(filmwidth, bodyheight - filmradius, -bodylength);
	glEnd();

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, 1.0);

	return CCameraBase::EditorRender(Engine);
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CCamera::InitComplete()
{
	k3d::PushEditing(Document()->Objects());

	if(0 == m_Focallength.Object())
		g_return_val_if_fail(m_Focallength.Create("Camera focal length", 1.6, true, 0.0, 5000.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Focalplane.Object())
		g_return_val_if_fail(m_Focalplane.Create("Camera focus distance", 35.0, true, 0, ULONG_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Fstop.Object())
		g_return_val_if_fail(m_Fstop.Create("Camera f-stop", 0.3, true, 0.01, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Shutteropen.Object())
		g_return_val_if_fail(m_Shutteropen.Create("Camera shutter-opened", 0.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Shutterclose.Object())
		g_return_val_if_fail(m_Shutterclose.Create("Camera shutter-closed", 0.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Exposure.Object())
		g_return_val_if_fail(m_Exposure.Create("Camera exposure", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Gamma.Object())
		g_return_val_if_fail(m_Gamma.Create("Camera gamma", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Xdirection.Object())
		g_return_val_if_fail(m_Xdirection.Create("Camera X direction", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Ydirection.Object())
		g_return_val_if_fail(m_Ydirection.Create("Camera Y direction", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Zdirection.Object())
		g_return_val_if_fail(m_Zdirection.Create("Camera Z direction", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_FarPlane.Object())
		g_return_val_if_fail(m_FarPlane.Create("Camera farplane", 100000.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_NearPlane.Object())
		g_return_val_if_fail(m_NearPlane.Create("Camera nearplane", 0.2, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);

	if(0 == m_RiRenderEngine.Value().size())
		{
			sdpString type, engine, name;
			k3dOptionsGetDefaultRenderEngine(k3dApplication()->Options(), type, engine, name);
			k3dSetValue(m_RiRenderEngine, engine);
		}

	if(m_ViewportVisible.Value())
		ShowViewport();

	k3d::PopEditing(Document()->Objects());
	return CCameraBase::InitComplete();
}

///////////////////////////////////////////////////////////////
// k3dICamera implementation

sdpResult CCamera::ShowViewport()
{
	// If it's already open, we're done ...
	if(m_Frame)
		return SDP_S_OK;

	// Setup our rendering engines ...
	g_return_val_if_fail(sdpSucceeded(m_EditorEngine.SetCamera(this)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_RiEngine.SetCamera(this)), SDP_E_FAIL);

	k3dSetValue(m_ViewportVisible, true);

	m_Frame = new CCameraFrame(this);
	g_return_val_if_fail(m_Frame, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CCamera::HideViewport()
{
	k3dSetValue(m_ViewportVisible, false);

	delete m_Frame;
	m_Frame = 0;

	return SDP_S_OK;
}

gulong CCamera::Width()
{
	sdpGtkOpenGLDrawingArea* viewport = DrawingArea();
	if(viewport)
		return viewport->Width();

	return 0;
}

gulong CCamera::Height()
{
	sdpGtkOpenGLDrawingArea* viewport = DrawingArea();
	if(viewport)
		return viewport->Height();

	return 0;
}

sdpResult CCamera::Redraw(bool Synchronous)
{
	if(Synchronous)
		return Redraw();

	if(DrawingArea())
		DrawingArea()->QueueDraw();

	return SDP_S_OK;
}

sdpResult CCamera::Redraw()
{
	// Sanity checks ...
	g_return_val_if_fail(m_DisplayEngine, SDP_E_FAIL);

	return m_DisplayEngine->Redraw();
}

sdpGtkOpenGLDrawingArea* CCamera::DrawingArea()
{
	if(m_Frame)
		return &m_Frame->DrawingArea();

	return 0;
}

sdpResult CCamera::RenderPreview()
{
	// Sanity checks ...
	g_return_val_if_fail(m_RenderEngine, SDP_E_FAIL);

	return m_RenderEngine->RenderPreview();
}

sdpResult CCamera::RenderFrame()
{
	// Sanity checks ...
	g_return_val_if_fail(m_RenderEngine, SDP_E_FAIL);

	return m_RenderEngine->RenderFrame();
}

sdpResult CCamera::RenderAnimation()
{
	// Sanity checks ...
	g_return_val_if_fail(m_RenderEngine, SDP_E_FAIL);

	return m_RenderEngine->RenderAnimation();
}

void CCamera::SetupViewVolume()
{
	sdpMatrix4 matrix(k3dGetObjectToWorldMatrix(this));
	sdpEulerAngles angles(matrix, sdpEulerAngles::ZXYstatic);

	m_CurrentOrientation = angles;
	m_CurrentPosition = sdpExtractTranslation(matrix);
}


// Calculate camera coordinates ...
sdpResult CCamera::SnapshotCameraCoordinates()
{
	Document()->Hierarchy()->SnapshotWorldCoordinates();
	k3d::for_each(Document()->Objects(), ::SnapshotCameraCoordinates(k3dGetWorldToCameraMatrix(this)));

	return SDP_S_OK;
}

// Calculate screen coordinates ...
sdpResult CCamera::SnapshotScreenCoordinates()
{
	SnapshotCameraCoordinates();
	k3d::for_each(Document()->Objects(), ::SnapshotScreenCoordinates(ViewVolume(), Width(), Height()));

	return SDP_S_OK;
}

k3dIViewVolume* CCamera::ViewVolume()
{
	return this;
}

k3dICamera::Axis CCamera::ActiveAxis()
{
	return m_ActiveAxis.Value();
}

void CCamera::SetActiveAxis(const Axis ActiveAxis)
{
	m_ActiveAxis.SetValue(ActiveAxis);
}

k3dICamera::TransformType CCamera::ActiveTransformType()
{
	return m_ActiveTransformType.Value();
}

void CCamera::SetActiveTransformType(const k3dICamera::TransformType ActiveTransformType)
{
	m_ActiveTransformType.SetValue(ActiveTransformType);
}

k3dICamera::CoordinateSystem CCamera::ActiveCoordinateSystem()
{
	return m_ActiveCoordinateSystem.Value();
}

void CCamera::SetActiveCoordinateSystem(const CoordinateSystem System)
{
	m_ActiveCoordinateSystem.SetValue(System);
}

k3dICamera::EditMode CCamera::ActiveEditMode()
{
	return m_ActiveEditMode.Value();
}

void CCamera::SetActiveEditMode(const EditMode Mode)
{
	m_ActiveEditMode.SetValue(Mode);
}

k3dICamera::SnapMode CCamera::ActiveSnapMode()
{
	return m_ActiveSnapMode.Value();
}

void CCamera::SetActiveSnapMode(const SnapMode Mode)
{
	m_ActiveSnapMode.SetValue(Mode);
}

///////////////////////////////////////////////////////////////
// k3dIViewport implementation ...

const gdouble CCamera::OutputPixelWidth()
{
	return m_OutputPixelWidth.Value();
}

const gdouble CCamera::OutputPixelHeight()
{
	return m_OutputPixelHeight.Value();
}

const gdouble CCamera::OutputPixelAspectRatio()
{
	return m_OutputPixelAspectRatio.Value();
}

const gdouble CCamera::WorldAspectRatio()
{
	g_return_val_if_fail(OutputPixelHeight(), 0.0);
	return (OutputPixelWidth() / OutputPixelHeight()) * OutputPixelAspectRatio();
}

const gdouble CCamera::WindowPixelWidth(const gdouble WindowPixelHeight)
{
	return WindowPixelHeight * WindowAspectRatio();
}

const gdouble CCamera::WindowPixelHeight(const gdouble WindowPixelWidth)
{
	return WindowPixelWidth / WindowAspectRatio();
}

const gdouble CCamera::WindowPixelAspectRatio()
{
	return m_WindowPixelAspectRatio.Value();
}

const gdouble CCamera::WindowAspectRatio()
{
	g_return_val_if_fail(WindowPixelAspectRatio(), 0.0);
	return WorldAspectRatio() / WindowPixelAspectRatio();
}

///////////////////////////////////////////////////////////////
// k3dIViewVolume implementation ...

k3dIViewport* CCamera::Viewport()
{
	return this;
}

const bool CCamera::Orthographic()
{
	return m_Orthographic.Value();
}

const gdouble CCamera::NearPlane()
{
	return m_NearPlane->Value1(m_CurrentTime, m_DeltaTime);
}

const gdouble CCamera::FarPlane()
{
	return m_FarPlane->Value1(m_CurrentTime, m_DeltaTime);
}

const gdouble CCamera::Left()
{
	if(Orthographic())
		return -0.5 * WorldAspectRatio() * OrthographicWorldHeight();

	return -0.5 * (WorldAspectRatio() / PerspectiveFocalLength());
}

const gdouble CCamera::Right()
{
	if(Orthographic())
		return 0.5 * WorldAspectRatio() * OrthographicWorldHeight();

	return 0.5 * (WorldAspectRatio() / PerspectiveFocalLength());
}

const gdouble CCamera::Top()
{
	if(Orthographic())
		return 0.5 * OrthographicWorldHeight();

	return 0.5 * (1.0 / PerspectiveFocalLength());
}

const gdouble CCamera::Bottom()
{
	if(Orthographic())
		return -0.5 * OrthographicWorldHeight();

	return -0.5 * (1.0 / PerspectiveFocalLength());
}

void CCamera::CameraToScreen(gdouble Coords[3], const gdouble WindowPixelWidth, const gdouble WindowPixelHeight)
{
	if(Orthographic())
		{
			const gdouble px = (2.0 * Coords[0]) / (OrthographicWorldHeight() * WorldAspectRatio());
			const gdouble py = (2.0 * Coords[1]) / OrthographicWorldHeight();

			Coords[0] = (px * WindowPixelWidth * 0.5) + (WindowPixelWidth * 0.5);
			Coords[1] = (WindowPixelHeight * 0.5) - (py * WindowPixelHeight * 0.5);
			Coords[2] = 0.0;
		}

	// Avoid divide-by-zero ...
	if(0.0 == Coords[2])
		return;

	const gdouble px = ((Coords[0] * PerspectiveFocalLength()) / Coords[2]) / (WorldAspectRatio() * 0.5);
	const gdouble py = ((Coords[1] * PerspectiveFocalLength()) / Coords[2]) / (0.5);

	Coords[0] = (px * WindowPixelWidth * 0.5) + (WindowPixelWidth * 0.5);
	Coords[1] = (WindowPixelHeight * 0.5) - (py * WindowPixelHeight * 0.5);
	Coords[2] = 0.0;
}

const gdouble CCamera::OrthographicWorldHeight()
{
	return 20.0 / m_Focallength->Value1(m_CurrentTime, m_DeltaTime);
}

const gdouble CCamera::PerspectiveFocalLength()
{
	return m_Focallength->Value1(m_CurrentTime, m_DeltaTime);
}

const gdouble CCamera::HorizontalFOV()
{
	return 2.0 * atan2(0.5 * WorldAspectRatio(), PerspectiveFocalLength());
}

const gdouble CCamera::VerticalFOV()
{
	return 2.0 * atan2(0.5 * 1.0, PerspectiveFocalLength());
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CCamera::OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	const bool shift = ShiftDown(Modifiers);
	const bool selectmode = !ControlDown(Modifiers);
	const bool pointmode = POINT == ActiveEditMode();
	const bool polygonmode = POLYGON == ActiveEditMode();

	const sdpVector2 currentmouse(CurrentX, CurrentY);

	SnapshotScreenCoordinates();

	if(pointmode)
		{
			if(shift)
				{
					k3dRecordStateChangeSet changeset(Document(), "Select Path");
					k3d::for_each_if<k3dIPathCollection>(Document()->Objects(), ::SelectSinglePath(currentmouse, selectmode), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
				}
			else
				{
					k3dRecordStateChangeSet changeset(Document(), "Select Point");
					k3d::for_each_if<k3dIPointCollection>(Document()->Objects(), ::SelectSinglePoint(currentmouse, selectmode), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
				}
		}
	else if(polygonmode)
		{
			if(shift)
				{
					k3dRecordStateChangeSet changeset(Document(), "Select Object");
					k3d::for_each_if<k3dIMesh>(Document()->Objects(), ::SelectSingleMesh(Document(), currentmouse, selectmode), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
				}
			else
				{
					k3dRecordStateChangeSet changeset(Document(), "Select Path");
					k3d::for_each_if<k3dIPathCollection>(Document()->Objects(), ::SelectSinglePath(currentmouse, selectmode), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
				}
		}
	else
		{
			g_return_val_if_fail(0, SDP_S_OK);
		}

	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnLButtonDoubleClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	EditObject();
	return SDP_S_OK;
}

sdpResult CCamera::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	if(0 == DrawingArea())
		return SDP_S_FALSE;

	m_LassoPoints.clear();
	m_LassoPoints.push_back(sdpVector2(CurrentX, CurrentY));

	if(!ShiftDown(Modifiers))
		{
			m_RubberBand = sdpRectangle(CurrentX, CurrentX, CurrentY, CurrentY);
			RubberBand(*DrawingArea(), m_RubberBand);
		}

	return SDP_S_OK;
}

sdpResult CCamera::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	if(0 == DrawingArea())
		return SDP_S_FALSE;

	// Erase the existing rubber-band box ...
	if(!ShiftDown(Modifiers))
		RubberBand(*DrawingArea(), m_RubberBand);

	if(ShiftDown(Modifiers))	// Lasso selection
		{
			if((sdpVector2(CurrentX, CurrentY) - m_LassoPoints[m_LassoPoints.size()-1]).Length2() > 10)
				{
					m_LassoPoints.push_back(sdpVector2(CurrentX, CurrentY));

					GdkGC* const gc = SelectionGC(*DrawingArea());

					gdk_draw_line(((GtkWidget*)*DrawingArea())->window,
												gc,
												gint(m_LassoPoints[m_LassoPoints.size() - 2][0]),
												gint(m_LassoPoints[m_LassoPoints.size() - 2][1]),
												gint(m_LassoPoints[m_LassoPoints.size() - 1][0]),
												gint(m_LassoPoints[m_LassoPoints.size() - 1][1]));

					gdk_gc_destroy(gc);
				}
		}
	else	// Rubber-band selection ...
		{
			m_RubberBand = sdpRectangle(StartX, LastX, StartY, LastY);
			RubberBand(*DrawingArea(), m_RubberBand);
		}

	return SDP_S_OK;
}

sdpResult CCamera::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	const bool lasso = ShiftDown(Modifiers);
	const bool selectmode = !ControlDown(Modifiers);
	const bool pointmode = POINT == ActiveEditMode();
	const bool polygonmode = POLYGON == ActiveEditMode();

	const sdpVector2 currentmouse(CurrentX, CurrentY);
	const sdpVector2 startmouse(StartX, StartY);

	SnapshotScreenCoordinates();

	if(lasso)
		{
			k3dRecordStateChangeSet changeset(Document(), "Lasso Select");

			if(pointmode)
				k3d::for_each_if<k3dIPointCollection>(Document()->Objects(), ::SelectLassoPoints(m_LassoPoints, selectmode), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
			else if(polygonmode)
				k3d::for_each_if<k3dIPathCollection>(Document()->Objects(), ::SelectLassoPaths(m_LassoPoints, selectmode, m_LassoSelectPartial), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
			else
				g_return_val_if_fail(0, SDP_E_FAIL);
		}
	else
		{
			k3dRecordStateChangeSet changeset(Document(), "Box Select");

			if(pointmode)
				::SelectBoxPoints selection = k3d::for_each_if<k3dIPointCollection>(Document()->Objects(), ::SelectBoxPoints(startmouse, currentmouse, selectmode), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
			else if(polygonmode)
				::SelectBoxPaths selection = k3d::for_each_if<k3dIPathCollection>(Document()->Objects(), ::SelectBoxPaths(startmouse, currentmouse, selectmode, m_BoxSelectPartial), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTABLE));
			else
				g_return_val_if_fail(0, SDP_E_FAIL);
		}

	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

// User clicked the right mouse button, so deselect everything ...
sdpResult CCamera::OnRButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	k3dRecordStateChangeSet changeset(Document(), "Deselect All");
	k3dDeselectAll(Document());
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnRButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
//	::SetCursor(NULL);

	return SDP_S_OK;
}

sdpResult CCamera::OnRButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	if(ShiftDown(Modifiers) && ControlDown(Modifiers))
	{
//	::SetCursor(m_ZoomCursor);

		gdouble zoomfactor;
		const gdouble focallength = m_Focallength->Value1(0.0, 0.1);

		if(CurrentY > LastY)
			{
				if(m_ZoomSign == -1.0)
					zoomfactor = pow(.995, (double)(CurrentY - LastY));
				else
					zoomfactor = pow(1.005, (double)(CurrentY - LastY));
			}
		else
			{
				if(m_ZoomSign == -1.0)
					zoomfactor = pow(1.005, (double)(LastY - CurrentY));
				else
					zoomfactor = pow(.995, (double)(LastY - CurrentY));
			}

		k3d::Editing editing(m_Focallength.Interface());
		k3dChannelSetConstantValue(m_Focallength.Interface(), focallength * zoomfactor);
	}
	else if(ShiftDown(Modifiers))
		{
//					::SetCursor(m_DollyCursor);

			const gdouble focallength = m_Focallength->Value1(m_CurrentTime, m_DeltaTime);

			sdpVector3 offset;
			k3dIObject* parent = Hierarchy()->ObjectParent(this);
			sdpMatrix4 matCameraToParent = parent? k3dGetCameraToObjectMatrix(this,parent) : k3dGetCameraToWorldMatrix(this);
			sdpVector3 xaxis = (matCameraToParent * sdpVector3(1,0,0))-(matCameraToParent * sdpVector3(0,0,0));
			sdpVector3 yaxis = (matCameraToParent * sdpVector3(0,1,0))-(matCameraToParent * sdpVector3(0,0,0));
			offset -= gdouble(CurrentX - LastX) * (-m_PositionSensitivity / focallength) * xaxis;
			offset -= gdouble(CurrentY - LastY) * (m_PositionSensitivity / focallength) * yaxis;
			m_Position[k3dITransform::CURRENT_FRAME] += offset;
		}
	else if(ControlDown(Modifiers))
		{
//					::SetCursor(m_TiltCursor);

			// We want mouse coordinates relative to the center of the window ...
			sdpVector2 currentmouse(CurrentX - (Width() >> 1), (Height() >> 1) - CurrentY);
			sdpVector2 lastmouse(LastX - (Width() >> 1), (Height() >> 1) - LastY);
			const gdouble deltatheta = currentmouse.Angle() - lastmouse.Angle();
			sdpQuaternion quaternion(m_Orientation[k3dITransform::CURRENT_FRAME]);
			sdpEulerAngles angles(quaternion, sdpEulerAngles::ZXYstatic);
			angles[0] -= deltatheta;
			m_Orientation[k3dITransform::CURRENT_FRAME] = angles;
		}
	else
	{
//					::SetCursor(m_PanCursor);

			// Calculate our field-of-view (based on a perspective view) ...
			const gdouble deltax = sdp::Pi() / gdouble(Width());
			const gdouble deltay = sdp::Pi() / gdouble(Height());

			sdpQuaternion quaternion(m_Orientation[k3dITransform::CURRENT_FRAME]);
			sdpMatrix4 matCameraToWorld=k3dGetCameraToWorldMatrix(this);
			k3dIObject* parent = Hierarchy()->ObjectParent(this);
			sdpMatrix4 matWorldToParent = sdpIdentity3D();
			sdpMatrix4 matParentToWorld = sdpIdentity3D();
			if(parent) matWorldToParent = k3dGetWorldToObjectMatrix(parent);
			if(parent) matParentToWorld = k3dGetObjectToWorldMatrix(parent);

			// Get the rotation centre in world coordinates, just default to the world origin for now.
			sdpVector3 centre = sdpVector3(0,0,0);
			// Get the camera offset position.
			sdpVector3 pos = m_Position[k3dITransform::CURRENT_FRAME];
			pos = matParentToWorld * pos;
			// Move so that the requested rotation centre is at the origin of camera space
			pos -= centre;
			// Rotate about the local x axis using the y mouse movement.
			sdpVector3 xaxis = matCameraToWorld * sdpVector3(1,0,0);
			xaxis -= (matCameraToWorld * sdpVector3(0,0,0));
			sdpAngleAxis axrot(-(gdouble(LastY - CurrentY) * deltay), xaxis);
			pos = sdpMatrix4(sdpRotation3D(axrot)) * pos;
			quaternion = sdpQuaternion(axrot)*quaternion;
			// Now rotate about the world y axis using the mouse x movement.
			sdpVector3 yaxis = matWorldToParent * sdpVector3(0,1,0);
			yaxis -= (matWorldToParent * sdpVector3(0,0,0));
			sdpAngleAxis ayrot = sdpAngleAxis(-(gdouble(LastX - CurrentX) * deltax), yaxis);
			pos = sdpMatrix4(sdpRotation3D(ayrot)) * pos;
			quaternion = sdpQuaternion(ayrot)*quaternion;
			// Move back by the centre offset.
			pos += centre;
			// and convert back into parent coordinates.
			pos = matWorldToParent * pos;
			// Put the new position and orientation of the camera back.
			m_Position[k3dITransform::CURRENT_FRAME] = matWorldToParent * pos;
			m_Orientation[k3dITransform::CURRENT_FRAME] = quaternion;
		}

	// Update the view ...
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnMButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
//	::SetCursor(m_ZoomCursor);

	gdouble zoomfactor;
	const gdouble focallength = m_Focallength->Value1(0.0, 0.1);

	if(CurrentY > LastY)
		{
			if(m_ZoomSign == -1.0)
				zoomfactor = pow(.995, (double)(CurrentY - LastY));
			else
				zoomfactor = pow(1.005, (double)(CurrentY - LastY));
		}
	else
		{
			if(m_ZoomSign == -1.0)
				zoomfactor = pow(1.005, (double)(LastY - CurrentY));
			else
				zoomfactor = pow(.995, (double)(LastY - CurrentY));
		}

	k3d::Editing editing(m_Focallength.Interface());
	k3dChannelSetConstantValue(m_Focallength.Interface(), focallength * zoomfactor);
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CCamera::OnMButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{

//	::SetCursor(NULL);

/*
	const gdouble focallength = m_Focallength->Value1(0.0, 0.1);
	k3dRecordStateChangeSet changeset(Document(), "Camera Zoom");
	k3d::Editing editing(m_Focallength.Interface());
	k3dChannelSetConstantValue(m_Focallength.Interface(), focallength);
	k3dRedrawAll(Document(), false);
*/

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CCameraDialog

const sdpString CCameraDialog::dialogname = "properties";

const sdpString CCameraDialog::control_editorwireframe = "editorwireframe";
const sdpString CCameraDialog::control_editorsolid = "editorsolid";
const sdpString CCameraDialog::control_editorfog = "editorfog";
const sdpString CCameraDialog::control_editorfognear = "editorfognear";
const sdpString CCameraDialog::control_editorfogfar = "editorfogfar";
const sdpString CCameraDialog::control_editortexture = "editortexture";
const sdpString CCameraDialog::control_editortwosided = "editortwosided";
const sdpString CCameraDialog::control_editorwireframetwosided = "editorwireframetwosided";
const sdpString CCameraDialog::control_editororientation = "editororientation";
const sdpString CCameraDialog::control_editornormals = "editornormals";

const sdpString CCameraDialog::control_riengine = "riengine";
const sdpString CCameraDialog::control_rixsamples = "rixsamples";
const sdpString CCameraDialog::control_riysamples = "riysamples";
const sdpString CCameraDialog::control_ripixelfiltername = "ripixelfiltername";
const sdpString CCameraDialog::control_ripixelfilterwidth = "ripixelfilterwidth";
const sdpString CCameraDialog::control_ritexturememory = "ritexturememory";
const sdpString CCameraDialog::control_rienablelerp = "rienablelerp";
const sdpString CCameraDialog::control_rienablegaussian = "rienablegaussian";
const sdpString CCameraDialog::control_ripixelfilterheight = "ripixelfilterheight";
const sdpString CCameraDialog::control_ritwosided = "ritwosided";
const sdpString CCameraDialog::control_rishadingrate = "rishadingrate";
const sdpString CCameraDialog::control_rishadinginterpolation = "rishadinginterpolation";
const sdpString CCameraDialog::control_rigridsize = "rigridsize";
const sdpString CCameraDialog::control_rixbucketsize = "rixbucketsize";
const sdpString CCameraDialog::control_riybucketsize = "riybucketsize";
const sdpString CCameraDialog::control_rieyesplits = "rieyesplits";
const sdpString CCameraDialog::control_riatmosphere = "riatmosphere";
const sdpString CCameraDialog::control_ridepthoffield = "ridepthoffield";
const sdpString CCameraDialog::control_riradiosity = "riradiosity";
const sdpString CCameraDialog::control_riradiositysteps = "riradiositysteps";
const sdpString CCameraDialog::control_showviewport = "showviewport";
const sdpString CCameraDialog::control_hideviewport = "hideviewport";

const sdpString CCameraDialog::control_focalplane = "focalplane";
const sdpString CCameraDialog::control_fstop = "fstop";
const sdpString CCameraDialog::control_riexposure = "riexposure";
const sdpString CCameraDialog::control_rigamma = "rigamma";
const sdpString CCameraDialog::control_riatmosphereshaderselector = "riatmosphereshaderselector";
const sdpString CCameraDialog::control_riatmosphereshadercontainer = "riatmosphereshadercontainer";

const sdpString CCameraDialog::control_outputpixelwidth = "outputpixelwidth";
const sdpString CCameraDialog::control_outputpixelheight = "outputpixelheight";
const sdpString CCameraDialog::control_outputpixelaspectratio = "outputpixelaspectratio";
const sdpString CCameraDialog::control_windowpixelaspectratio = "windowpixelaspectratio";
const sdpString CCameraDialog::control_farplane = "farplane";
const sdpString CCameraDialog::control_nearplane = "nearplane";
const sdpString CCameraDialog::control_orthographic = "orthographic";
const sdpString CCameraDialog::control_focallength = "focallength";

const sdpString CCameraDialog::control_position = "position";
const sdpString CCameraDialog::control_orientation = "orientation";
const sdpString CCameraDialog::control_scale = "scale";
const sdpString CCameraDialog::control_target = "target";

CCameraDialog::CCameraDialog(CCamera* const Object) :
	CCameraDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "camera.gtkml"));

	AddCommand(control_riengine, control_riengine, CONTROL_COMBO);
	AddCommand(control_ripixelfiltername, control_ripixelfiltername, CONTROL_COMBO);
	AddCommand(control_rishadinginterpolation, control_rishadinginterpolation, CONTROL_COMBO);
	AddCommand(control_showviewport, control_showviewport, CONTROL_BUTTON);
	AddCommand(control_hideviewport, control_hideviewport, CONTROL_BUTTON);

	// Setup check buttons ...
	Connectk3dCheckButton(control_editorwireframe, m_Object->m_EditorWireframe, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Wireframe Paths", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_editorsolid, m_Object->m_EditorSolid, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Solid Paths", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_editorfog, m_Object->m_EditorFog, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Fog", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_editortexture, m_Object->m_EditorTexture, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Textures", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_rienablelerp, m_Object->m_RiEnableLerp, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Enables lerp interpolation for textures", 0, 0, 0);
	Connectk3dCheckButton(control_rienablegaussian, m_Object->m_RiEnableGaussian, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Enables gaussian filtering for textures", 0, 0, 0);
	Connectk3dCheckButton(control_editortwosided, m_Object->m_EditorTwoSided, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Two-Sided Paths", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_editorwireframetwosided, m_Object->m_EditorWireframeTwoSided, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Two-Sided Wireframes", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_editororientation, m_Object->m_EditorOrientation, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Path Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_editornormals, m_Object->m_EditorNormals, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Normals", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Connectk3dCheckButton(control_ritwosided, m_Object->m_RiTwoSided, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Two-sided RenderMan Surfaces", 0, 0, 0);
	Connectk3dCheckButton(control_riatmosphere, m_Object->m_RiAtmosphere, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera RenderMan Atmosphere", 0, 0, 0);
	Connectk3dCheckButton(control_ridepthoffield, m_Object->m_RiDepthOfField, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera RenderMan Depth-Of-Field", 0, 0, 0);
	Connectk3dCheckButton(control_riradiosity, m_Object->m_RiRadiosity, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera RenderMan Radiosity", 0, 0, 0);
	Connectk3dCheckButton(control_orthographic, m_Object->m_Orthographic, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orthographic View", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup spin buttons ...
	Connectk3dSpinButton(control_outputpixelwidth, m_Object->m_OutputPixelWidth, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Output Width", 0, 0, 0);
	Connectk3dSpinButton(control_outputpixelheight, m_Object->m_OutputPixelHeight, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Output Height", 0, 0, 0);
	Connectk3dSpinButton(control_outputpixelaspectratio, m_Object->m_OutputPixelAspectRatio, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Output Aspect Ratio", 0, 0, 0);
	Connectk3dSpinButton(control_windowpixelaspectratio, m_Object->m_WindowPixelAspectRatio, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Screen Aspect Ratio", 0, 0, 0);
	Connectk3dSpinButton(control_editorfognear, m_Object->m_EditorFogNear, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Fog Near", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dSpinButton(control_editorfogfar, m_Object->m_EditorFogFar, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Editor Fog Far", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dSpinButton(control_rixsamples, m_Object->m_RiXSamples, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera X Samples", 0, 0, 0);
	Connectk3dSpinButton(control_riysamples, m_Object->m_RiYSamples, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Y Samples", 0, 0, 0);
	Connectk3dSpinButton(control_ripixelfilterwidth, m_Object->m_RiPixelFilterWidth, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Pixel Filter Width", 0, 0, 0);
	Connectk3dSpinButton(control_ripixelfilterheight, m_Object->m_RiPixelFilterHeight, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Pixel Filter Height", 0, 0, 0);
	Connectk3dSpinButton(control_rishadingrate, m_Object->m_RiShadingRate, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Shading Rate", 0, 0, 0);
	Connectk3dSpinButton(control_rigridsize, m_Object->m_RiGridSize, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Grid Size", 0, 0, 0);
	Connectk3dSpinButton(control_rixbucketsize, m_Object->m_RiXBucketSize, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera X Bucket Size", 0, 0, 0);
	Connectk3dSpinButton(control_riybucketsize, m_Object->m_RiYBucketSize, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Y Bucket Size", 0, 0, 0);
	Connectk3dSpinButton(control_rieyesplits, m_Object->m_RiEyeSplits, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Eye splits", 0, 0, 0);
	Connectk3dSpinButton(control_riradiositysteps, m_Object->m_RiRadiositySteps, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Radiosity Steps", 0, 0, 0);
	Connectk3dSpinButton(control_ritexturememory, m_Object->m_RiTextureMemory, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Amount of memory reserved for textures by the render engine", 0, 0, 0);

	// Setup channel buttons ...
	Connectk3dChannelButton(control_farplane, m_Object->m_FarPlane.Interface());
	Connectk3dChannelButton(control_nearplane, m_Object->m_NearPlane.Interface());
	Connectk3dChannelButton(control_focalplane, m_Object->m_Focalplane.Interface());
	Connectk3dChannelButton(control_fstop, m_Object->m_Fstop.Interface());
	Connectk3dChannelButton(control_riexposure, m_Object->m_Exposure.Interface());
	Connectk3dChannelButton(control_rigamma, m_Object->m_Gamma.Interface());
	Connectk3dChannelButton(control_focallength, m_Object->m_Focallength.Interface());

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup atmosphere shader selector
	if(Getk3dShaderSelector("riatmosphereshaderselector"))
		Getk3dShaderSelector("riatmosphereshaderselector")->Attach(m_Object->Document(), &m_Object->m_RiAtmosphereShader, true, sdpsl::IShader::VOLUME);

	if(Getk3dShaderInfo("riatmosphereshaderinfo"))
		Getk3dShaderInfo("riatmosphereshaderinfo")->Attach(m_Object->Document(), &m_Object->m_RiAtmosphereShader);

	if(Getk3dShaderArgumentsPage("riatmospherearguments"))
		Getk3dShaderArgumentsPage("riatmospherearguments")->Attach(m_Object->Document(), &m_Object->m_RiAtmosphereShader);

	// Setup render engine options ...
	const sdpString selectedengine = m_Object->m_RiRenderEngine.Value();
	sdpString selectedenginename;

	// Build a list of available engines ...
	sdpStringArray enginenames;
	k3dOptionsRenderEngines engines;
	k3dOptionsGetRenderEngines(k3dApplication()->Options(), engines);
	for(k3dOptionsRenderEngines::iterator engine = engines.begin(); engine != engines.end(); engine++)
		{
			enginenames.push_back(engine->Name());

			if(selectedengine == engine->Engine())
				selectedenginename = engine->Name();
		}

	Combo(control_riengine).SetList(enginenames);
	Combo(control_riengine).SetEntryText(selectedenginename);

	MapEvent("changed", control_riengine, false, Combo(control_riengine).Entry(), true);

	// Setup pixel filter options ...
	sdpStringArray pixelfilters;
	pixelfilters.push_back(k3dIRiEngine::RI_BOX);
	pixelfilters.push_back(k3dIRiEngine::RI_CATMULL_ROM);
	pixelfilters.push_back(k3dIRiEngine::RI_GAUSSIAN);
	pixelfilters.push_back(k3dIRiEngine::RI_SINC);
	pixelfilters.push_back(k3dIRiEngine::RI_TRIANGLE);
	Combo(control_ripixelfiltername).SetList(pixelfilters);
	MapEvent("changed", control_ripixelfiltername, false, Combo(control_ripixelfiltername).Entry(), true);

	// Setup shading interpolation options ...
	sdpStringArray shadinginterpolations;
	shadinginterpolations.push_back(k3dIRiEngine::RI_CONSTANT);
	shadinginterpolations.push_back(k3dIRiEngine::RI_SMOOTH);
	Combo(control_rishadinginterpolation).SetList(shadinginterpolations);
	MapEvent("changed", control_rishadinginterpolation, false, Combo(control_rishadinginterpolation).Entry(), true);

	CallUpdateControls();
	Show();
}

void CCameraDialog::UpdateControls()
{
	Combo(control_ripixelfiltername).SetEntryText(m_Object->m_RiPixelFilterName.Value());
	Combo(control_rishadinginterpolation).SetEntryText(m_Object->m_RiShadingInterpolation.Value());

	CCameraDialogBase::UpdateControls();
}

void CCameraDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_riengine)
		OnRiEngine();
	else if(Event->Name() == control_ripixelfiltername)
		OnRiPixelFilterName();
	else if(Event->Name() == control_rishadinginterpolation)
		OnRiShadingInterpolation();
	else if(Event->Name() == control_showviewport)
		OnShowViewport();
	else if(Event->Name() == control_hideviewport)
		OnHideViewport();
	else
		CCameraDialogBase::OnEvent(Event);
}

void CCameraDialog::OnRiEngine()
{
	const sdpString selectedenginename = Combo(control_riengine).GetEntryText();

	k3dOptionsRenderEngines engines;
	k3dOptionsGetRenderEngines(k3dApplication()->Options(), engines);
	for(k3dOptionsRenderEngines::iterator engine = engines.begin(); engine != engines.end(); engine++)
		{
			if(selectedenginename == engine->Name())
				{
					k3dSetValue(m_Object->m_RiRenderEngine, engine->Engine());
					return;
				}
		}

	// Shouldn't ever happen!
	g_return_if_fail(0);
}

void CCameraDialog::OnRiPixelFilterName()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_ripixelfiltername, Combo(control_ripixelfiltername).GetEntryText());
	k3dSetValue<sdpString>(m_Object->m_RiPixelFilterName, Combo(control_ripixelfiltername).GetEntryText());
}

void CCameraDialog::OnRiShadingInterpolation()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_rishadinginterpolation, Combo(control_rishadinginterpolation).GetEntryText());
	k3dSetValue<sdpString>(m_Object->m_RiShadingInterpolation, Combo(control_rishadinginterpolation).GetEntryText());
}

void CCameraDialog::OnShowViewport()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_showviewport, 0);
	m_Object->ShowViewport();
}

void CCameraDialog::OnHideViewport()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_hideviewport, 0);
	m_Object->HideViewport();
}

///////////////////////////////////////////////////////////////////////////////
// CCameraFrame

const sdpString CCameraFrame::dialogname = "viewport";
const sdpString CCameraFrame::control_viewportcontainer = "viewportcontainer";
const sdpString CCameraFrame::control_focus = "focus";
const sdpString CCameraFrame::control_interactiveusage = "interactiveusage";

const sdpString CCameraFrame::control_editdelete = "editdelete";
/*
const sdpString CCameraFrame::control_editpointmode = "editpointmode";
const sdpString CCameraFrame::control_editsnapmode = "editsnapmode";
const sdpString CCameraFrame::control_editxaxis = "editxaxis";
const sdpString CCameraFrame::control_edityaxis = "edityaxis";
const sdpString CCameraFrame::control_editzaxis = "editzaxis";
const sdpString CCameraFrame::control_editposition = "editposition";
const sdpString CCameraFrame::control_editorientation = "editorientation";
const sdpString CCameraFrame::control_editscale = "editscale";
*/

const sdpString CCameraFrame::control_close = "close";
const sdpString CCameraFrame::control_setfocus = "setfocus";
const sdpString CCameraFrame::control_renderpreview = "renderpreview";
const sdpString CCameraFrame::control_renderframe = "renderframe";
const sdpString CCameraFrame::control_renderanimation = "renderanimation";
const sdpString CCameraFrame::control_defaultview = "defaultview";
const sdpString CCameraFrame::control_negativezview = "-zview";
const sdpString CCameraFrame::control_positivezview = "+zview";
const sdpString CCameraFrame::control_negativexview = "-xview";
const sdpString CCameraFrame::control_positivexview = "+xview";
const sdpString CCameraFrame::control_negativeyview = "-yview";
const sdpString CCameraFrame::control_positiveyview = "+yview";

const sdpString CCameraFrame::control_mousemove = "mousemove";

const sdpString CCameraFrame::control_lbuttonclick = "lbuttonclick";
const sdpString CCameraFrame::control_lbuttondoubleclick = "lbuttondoubleclick";
const sdpString CCameraFrame::control_lbuttonstartdrag = "lbuttonstartdrag";
const sdpString CCameraFrame::control_lbuttondrag = "lbuttondrag";
const sdpString CCameraFrame::control_lbuttonenddrag = "lbuttonenddrag";

const sdpString CCameraFrame::control_mbuttonclick = "mbuttonclick";
const sdpString CCameraFrame::control_mbuttondoubleclick = "mbuttondoubleclick";
const sdpString CCameraFrame::control_mbuttonstartdrag = "mbuttonstartdrag";
const sdpString CCameraFrame::control_mbuttondrag = "mbuttondrag";
const sdpString CCameraFrame::control_mbuttonenddrag = "mbuttonenddrag";

const sdpString CCameraFrame::control_rbuttonclick = "rbuttonclick";
const sdpString CCameraFrame::control_rbuttondoubleclick = "rbuttondoubleclick";
const sdpString CCameraFrame::control_rbuttonstartdrag = "rbuttonstartdrag";
const sdpString CCameraFrame::control_rbuttondrag = "rbuttondrag";
const sdpString CCameraFrame::control_rbuttonenddrag = "rbuttonenddrag";

const sdpString CCameraFrame::control_lrbuttonstartdrag = "lrbuttonstartdrag";
const sdpString CCameraFrame::control_lrbuttondrag = "lrbuttondrag";
const sdpString CCameraFrame::control_lrbuttonenddrag = "lrbuttonenddrag";

CCameraFrame::CCameraFrame(CCamera* const Object) :
	CCameraFrameBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object),
	m_DocumentEvents(0)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "cameraframe.gtkml"));

	// Attach object selectors ...
	if(Getk3dObjectSelector(control_focus))
		Getk3dObjectSelector(control_focus)->Attach(
			&m_Object->m_Focus,
			k3dApplication()->CommandRecorder(),
			0,
			"Select Focus",
			0,
			new FocusChangeAction(m_Object->Document(), m_Object->m_Focus, new k3dActionRedrawAll(m_Object->Document(), false, 0)),
			0,
			true,
			new k3dObjectSelectorInterfaceFilter(new sdp::TypeInfo<k3dIMouseEventObserver>()));

	// Setup choosers ...
	if(Getk3dChooser("activeaxis"))
		Getk3dChooser("activeaxis")->Attach(new k3dChooserVariableDataSource<k3dICamera::Axis>(m_Object->m_ActiveAxis), k3dApplication()->CommandRecorder(), 0, "", 0, 0, 0);

	if(Getk3dChooser("activetransformtype"))
		Getk3dChooser("activetransformtype")->Attach(new k3dChooserVariableDataSource<k3dICamera::TransformType>(m_Object->m_ActiveTransformType), k3dApplication()->CommandRecorder(), 0, "", 0, 0, 0);

/*
	if(Getk3dChooser("activecoordinatesystem"))
		Getk3dChooser("activecoordinatesystem")->Attach(new k3dChooserVariableDataSource<k3dICamera::CoordinateSystem>(m_Object->m_ActiveCoordinateSystem), k3dApplication()->CommandRecorder(), 0, "", 0, 0, 0);
*/

	if(Getk3dChooser("activeeditmode"))
		Getk3dChooser("activeeditmode")->Attach(new k3dChooserVariableDataSource<k3dICamera::EditMode>(m_Object->m_ActiveEditMode), k3dApplication()->CommandRecorder(), 0, "", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dChooser("activesnapmode"))
		Getk3dChooser("activesnapmode")->Attach(new k3dChooserVariableDataSource<k3dICamera::SnapMode>(m_Object->m_ActiveSnapMode), k3dApplication()->CommandRecorder(), 0, "", 0, 0, 0);

	AddCommand(control_close, control_close, CONTROL_MENU_ITEM);
	AddCommand(control_renderpreview, control_renderpreview, CONTROL_MENU_ITEM);
	AddCommand(control_renderframe, control_renderframe, CONTROL_MENU_ITEM);
	AddCommand(control_renderanimation, control_renderanimation, CONTROL_MENU_ITEM);
	AddCommand(control_defaultview, control_defaultview, CONTROL_MENU_ITEM);
	AddCommand(control_negativezview, control_negativezview, CONTROL_MENU_ITEM);
	AddCommand(control_positivezview, control_positivezview, CONTROL_MENU_ITEM);
	AddCommand(control_negativexview, control_negativexview, CONTROL_MENU_ITEM);
	AddCommand(control_positivexview, control_positivexview, CONTROL_MENU_ITEM);
	AddCommand(control_negativeyview, control_negativeyview, CONTROL_MENU_ITEM);
	AddCommand(control_positiveyview, control_positiveyview, CONTROL_MENU_ITEM);

	AddCommand(control_editdelete, control_editdelete, CONTROL_MENU_ITEM);
/*
	AddCommand(control_editpointmode, control_editpointmode, CONTROL_MENU_ITEM);
	AddCommand(control_editsnapmode, control_editsnapmode, CONTROL_MENU_ITEM);
	AddCommand(control_editxaxis, control_editxaxis, CONTROL_MENU_ITEM);
	AddCommand(control_edityaxis, control_edityaxis, CONTROL_MENU_ITEM);
	AddCommand(control_editzaxis, control_editzaxis, CONTROL_MENU_ITEM);
	AddCommand(control_editposition, control_editposition, CONTROL_MENU_ITEM);
	AddCommand(control_editorientation, control_editorientation, CONTROL_MENU_ITEM);
	AddCommand(control_editscale, control_editscale, CONTROL_MENU_ITEM);
*/

	// Create the OpenGL drawing area ...
	sdpGtkContainer container(Container(control_viewportcontainer));
	if(!m_DrawingArea.Create(container, true, true, 8, 8, 8, 16))
		{
			std::cerr << "CCameraFrame::CCameraFrame(): Could not find useable OpenGL visual" << std::endl;
			if(!m_DrawingArea.Create(container, true, true, 5, 5, 5, 16))
				{
					std::cerr << "CCameraFrame::CCameraFrame(): Could not find useable OpenGL visual" << std::endl;
					if(!m_DrawingArea.Create(container, false, false, 4, 4, 4, 16))
						{
							std::cerr << "CCameraFrame::CCameraFrame(): Could not find useable OpenGL visual" << std::endl;
						}
				}
		}

	if(m_DrawingArea.Initialized())
		{
			g_return_if_fail(m_DrawingArea.InitializeFont());

			// Setup some events ...
			MapEvent("configure-event", "configure", false, m_DrawingArea, true);
			MapEvent("expose-event", "expose", false, m_DrawingArea, true);
			MapEvent("motion-notify-event", "mousemove", false, m_DrawingArea, true);
			MapEvent("button-press-event", "buttondown", false, m_DrawingArea, true);
			MapEvent("button-release-event", "buttonup", false, m_DrawingArea, true);
		}

// The Win32 port of GTK+ has some kind of refresh problem, so force one to get us going ...
#ifdef SDPWIN32
	RootWidget().QueueClear();
	sdpGtkHandlePendingEvents();
#endif

	// Give ourselves a chance to adjust to the camera's output window specs ...
	UpdateViewportDimensions();

	// We want to know if any of the output-related variables change ...
	m_Object->m_OutputPixelWidth.AddEventObserver(this, k3dIVariable::Event_Changed, "viewportdimensions");
	m_Object->m_OutputPixelHeight.AddEventObserver(this, k3dIVariable::Event_Changed, "viewportdimensions");
	m_Object->m_OutputPixelAspectRatio.AddEventObserver(this, k3dIVariable::Event_Changed, "viewportdimensions");
	m_Object->m_WindowPixelAspectRatio.AddEventObserver(this, k3dIVariable::Event_Changed, "viewportdimensions");

	m_DocumentEvents = new k3dEventConnector(this, k3dIDocument::Event_MouseFocusChanged, "CCameraFrame", 0);
	m_DocumentEvents->Attach(dynamic_cast<k3dIEventSource*>(m_Object->Document()));

	// Make ourselves visible ...
	Show();
}

CCameraFrame::~CCameraFrame()
{
	k3dSetValue(m_Object->m_ViewportVisible, false);
	m_Object->m_Frame = 0;

	// Cleanup event notification ...
	m_Object->m_OutputPixelWidth.RemoveEventObserver(this);
	m_Object->m_OutputPixelHeight.RemoveEventObserver(this);
	m_Object->m_OutputPixelAspectRatio.RemoveEventObserver(this);
	m_Object->m_WindowPixelAspectRatio.RemoveEventObserver(this);

	delete m_DocumentEvents;
}

static void MouseCommand(GtkWidget* Widget, k3dIApplication::TutorialMouseAction Action, GdkModifierType Modifiers, sdpVector2 CurrentMouse)
{
	// Sanity checks ...
	g_return_if_fail(Widget);

	k3dApplication()->TutorialMouseDisplay("", Action, Modifiers);
	sdpGtkWarpPointer(Widget, gint(CurrentMouse[0]), gint(CurrentMouse[1]));
	sdpGtkHandlePendingEvents();
	sdpGtkSleep(20);
}

// k3dICommandNode implementation ...
sdpResult CCameraFrame::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	// Extract state information from the arguments ...
	std::istrstream arguments(Arguments);
	gint temp = 0;
	sdpVector2 currentmouse;
	sdpVector2 lastmouse;
	sdpVector2 startmouse;
	arguments >> temp >> currentmouse >> lastmouse >> startmouse;
	GdkModifierType modifiers = GdkModifierType(temp);

	// Convert the mouse coordinates from percentages to pixels ...
	sdpVector2 viewportsize = sdpVector2(m_Object->Width(), m_Object->Height());
	currentmouse = sdpProduct(currentmouse, viewportsize);
	lastmouse = sdpProduct(lastmouse, viewportsize);
	startmouse = sdpProduct(startmouse, viewportsize);

	const gdouble tutorialspeed = k3dOptionsGetTutorialSpeed(k3dApplication()->Options());
	const gulong delay = static_cast<gulong>(1200 / tutorialspeed);

	GtkWidget* widget = GTK_WIDGET(m_DrawingArea.Object());

	sdpString name(Name);
	if(name == control_mousemove)
		{
			OnMouseMove(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MB_NONE, modifiers, currentmouse);
		}
	else if(name == control_lbuttonclick)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_lbuttondoubleclick)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonDoubleClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_DOUBLE_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_lbuttonstartdrag)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lbuttondrag)
		{
			OnLButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lbuttonenddrag)
		{
			OnLButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_mbuttonclick)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnMButtonClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_mbuttondoubleclick)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnMButtonDoubleClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MMB_DOUBLE_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_mbuttonstartdrag)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnMButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_mbuttondrag)
		{
			OnMButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::MMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_mbuttonenddrag)
		{
			OnMButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::MMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_rbuttonclick)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnRButtonClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::RMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_rbuttondoubleclick)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnRButtonDoubleClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::RMB_DOUBLE_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_rbuttonstartdrag)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnRButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::RMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_rbuttondrag)
		{
			OnRButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::RMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_rbuttonenddrag)
		{
			OnRButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::RMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_lrbuttonstartdrag)
		{
			m_DrawingArea.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLRButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMBRMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lrbuttondrag)
		{
			OnLRButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMBRMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lrbuttonenddrag)
		{
			OnLRButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMBRMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if (name == control_editdelete)
		{
			OnEditDelete();
		}
	else
		{
			return CCameraFrameBase::Command(Name, Arguments, Result);
		}

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CCameraFrame::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	const sdpString clientdata(ObserverData);

	if(event == k3dIVariable::Event_Changed && clientdata == "viewportdimensions")
		{
			UpdateViewportDimensions();
			return SDP_S_OK;
		}
	else if(event == k3dIDocument::Event_MouseFocusChanged)
		{
			Entry(control_interactiveusage).SetText(m_Object->Document()->MouseFocus() ? m_Object->Document()->MouseFocus()->Usage() : "");
			return SDP_S_OK;
		}

	return CCameraFrameBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

sdpGtkOpenGLDrawingArea& CCameraFrame::DrawingArea()
{
	return m_DrawingArea;
}

void CCameraFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "configure")
		OnConfigureEvent(Event);
	else if(Event->Name() == "expose")
		OnExposeEvent(Event);
	else if(Event->Name() == "mousemove")
		RawMouseMove(Event);
	else if(Event->Name() == "buttondown")
		RawButtonDown(Event);
	else if(Event->Name() == "buttonup")
		RawButtonUp(Event);
	else if(Event->Name() == control_close)
		OnClose();
	else if(Event->Name() == control_renderpreview)
		OnRenderPreview();
	else if(Event->Name() == control_renderframe)
		OnRenderFrame();
	else if(Event->Name() == control_renderanimation)
		OnRenderAnimation();
	else if(Event->Name() == control_defaultview)
		OnDefaultView();
	else if(Event->Name() == control_negativezview)
		OnAxisView(Event->Name());
	else if(Event->Name() == control_positivezview)
		OnAxisView(Event->Name());
	else if(Event->Name() == control_negativexview)
		OnAxisView(Event->Name());
	else if(Event->Name() == control_positivexview)
		OnAxisView(Event->Name());
	else if(Event->Name() == control_negativeyview)
		OnAxisView(Event->Name());
	else if(Event->Name() == control_positiveyview)
		OnAxisView(Event->Name());
	else if(Event->Name() == control_editdelete)
		OnEditDelete();
	else
		CCameraFrameBase::OnEvent(Event);
}

void CCameraFrame::OnClose()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_close, "");
	CCameraFrameBase::OnClose();
}

void CCameraFrame::OnFocusIn()
{
	// We override the base-class implementation here, so the camera
	// doesn't grab the document's active object context ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_setfocus, "");
	k3dApplication()->SetActiveDocument(m_Object->Document());
}

void CCameraFrame::OnRenderPreview()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_renderpreview, "");
	g_return_if_fail(sdpSucceeded(m_Object->RenderPreview()));
}

// We have an unfortunate class with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

void CCameraFrame::OnRenderFrame()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_renderframe, "");
	g_return_if_fail(sdpSucceeded(m_Object->RenderFrame()));
}

void CCameraFrame::OnRenderAnimation()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_renderanimation, "");
	g_return_if_fail(sdpSucceeded(m_Object->RenderAnimation()));
}

void CCameraFrame::OnDefaultView()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_defaultview, "");
	k3dRecordStateChangeSet undo(m_Object->Document(), "Camera Default View");

	k3dSetPosition(m_Object, sdpVector3(-20.0, 20.0, -20.0));
	k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(37.0), sdpRadians(45.0), sdpEulerAngles::ZXYstatic));
	k3dRedrawAll(m_Object->Document(), false);
}

void CCameraFrame::OnAxisView(SDPCSTRING Axis)
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, Axis, "");

	if(Axis == control_negativezview)
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Camera -Z View");
			k3dSetPosition(m_Object, sdpVector3(0.0, 0.0, -34.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(0.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == control_positivezview)
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Camera +Z View");
			k3dSetPosition(m_Object, sdpVector3(0.0, 0.0, 34.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(sdpVector3(0.0, 180.0, 0.0)), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == control_negativexview)
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Camera -X View");
			k3dSetPosition(m_Object, sdpVector3(-34.0, 0.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(90.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == control_positivexview)
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Camera +X View");
			k3dSetPosition(m_Object, sdpVector3(34.0, 0.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(0.0), sdpRadians(-90.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == control_negativeyview)
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Camera -Y View");
			k3dSetPosition(m_Object, sdpVector3(0.0, -34.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(-90.0), sdpRadians(0.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else if(Axis == control_positiveyview)
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Camera +Y View");
			k3dSetPosition(m_Object, sdpVector3(0.0, 34.0, 0.0));
			k3dSetOrientation(m_Object, sdpEulerAngles(sdpRadians(90.0), sdpRadians(0.0), sdpRadians(0.0), sdpEulerAngles::XYZstatic));
		}
	else
		{
			g_return_if_fail(0);	// Unknown view!!!
		}

	k3dRedrawAll(m_Object->Document(), false);
}

// Called when the view is resized ...
void CCameraFrame::OnConfigureEvent(sdpGtkEvent* Event)
{
	m_Object->Redraw();
}

// Called when the view gets uncovered / needs to be redrawn ...
void CCameraFrame::OnExposeEvent(sdpGtkEvent* Event)
{
	m_Object->Redraw();
}

class CCameraFrameMouseEvent
{
public:
	virtual sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) = 0;
};

class CCameraFrameMouseDragEvent
{
public:
	virtual sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) = 0;
};

static void HandleEvent(CCameraFrame* Frame, CCameraFrameMouseEvent& Functor, SDPCSTRING Command, GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	// Sanity checks ...
	g_assert(Frame);

	// Setup mouse coordinates ...
	sdpVector2 currentmouse = k3dTransformMouseCoords(Frame->Object(), CurrentMouse, 0, 1, 0, 1);

	// Dispatch it ...
	sdpResult result = SDP_S_FALSE;
	k3dIMouseEventObserver* const mousefocus = Frame->Object()->Document()->MouseFocus();
	if(mousefocus)
		result = Functor.HandleEvent(mousefocus, Frame->Object(), Modifiers, CurrentMouse);

	if(SDP_S_OK != result)
		result = Functor.HandleEvent(Frame->Object(), Frame->Object(), Modifiers, CurrentMouse);

	// Record the event ...
	if(SDP_S_OK == result)
		k3dRecordCommand(Frame, k3dICommandNode::USERINTERFACE, Command, sdpToString(gint(Modifiers)) + " " + sdpToString(currentmouse));
}

static void HandleDragEvent(CCameraFrame* Frame, CCameraFrameMouseDragEvent& Functor, SDPCSTRING Command, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Frame);

	// Setup mouse coordinates ...
	sdpVector2 currentmouse = k3dTransformMouseCoords(Frame->Object(), CurrentMouse, 0, 1, 0, 1);
	sdpVector2 lastmouse = k3dTransformMouseCoords(Frame->Object(), LastMouse, 0, 1, 0, 1);
	sdpVector2 startmouse = k3dTransformMouseCoords(Frame->Object(), StartMouse, 0, 1, 0, 1);

	// Dispatch it ...
	sdpResult result = SDP_S_FALSE;
	k3dIMouseEventObserver* const mousefocus = Frame->Object()->Document()->MouseFocus();
	if(mousefocus)
		result = Functor.HandleDragEvent(mousefocus, Frame->Object(), Modifiers, CurrentMouse, LastMouse, StartMouse);

	if(SDP_S_OK != result)
		result = Functor.HandleDragEvent(Frame->Object(), Frame->Object(), Modifiers, CurrentMouse, LastMouse, StartMouse);

	// Record the event ...
	if(SDP_S_OK == result)
		k3dRecordCommand(Frame, k3dICommandNode::USERINTERFACE, Command, sdpToString(gint(Modifiers)) + " " + sdpToString(currentmouse) + " " + sdpToString(lastmouse) + " " + sdpToString(startmouse));
}

class CCameraFrameCallOnMouseMove : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnMouseMove(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnLButtonClick : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnLButtonClick(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnLButtonDoubleClick : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnLButtonDoubleClick(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnLButtonStartDrag : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnLButtonStartDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnLButtonDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnLButtonDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnLButtonEndDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnLButtonEndDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnMButtonClick : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnMButtonClick(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnMButtonDoubleClick : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnMButtonDoubleClick(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnMButtonStartDrag : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnMButtonStartDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnMButtonDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnMButtonDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnMButtonEndDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnMButtonEndDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnRButtonClick : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnRButtonClick(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnRButtonDoubleClick : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnRButtonDoubleClick(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnRButtonStartDrag : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnRButtonStartDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnRButtonDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnRButtonDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnRButtonEndDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnRButtonEndDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnLRButtonStartDrag : public CCameraFrameMouseEvent
{
public:
	sdpResult HandleEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse) { return Observer->OnLRButtonStartDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1]); }
};

class CCameraFrameCallOnLRButtonDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnLRButtonDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

class CCameraFrameCallOnLRButtonEndDrag : public CCameraFrameMouseDragEvent
{
public:
	sdpResult HandleDragEvent(k3dIMouseEventObserver* const Observer, k3dICamera* Camera, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse) { return Observer->OnLRButtonEndDrag(Camera, Modifiers, CurrentMouse[0], CurrentMouse[1], LastMouse[0], LastMouse[1], StartMouse[0], StartMouse[1]); }
};

void CCameraFrame::OnMouseMove(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnMouseMove functor;
	HandleEvent(this, functor, control_mousemove, Modifiers, CurrentMouse);
}

void CCameraFrame::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnLButtonClick functor;
	HandleEvent(this, functor, control_lbuttonclick, Modifiers, CurrentMouse);
}

void CCameraFrame::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnLButtonDoubleClick functor;
	HandleEvent(this, functor, control_lbuttondoubleclick, Modifiers, CurrentMouse);
}

void CCameraFrame::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnLButtonStartDrag functor;
	HandleEvent(this, functor, control_lbuttonstartdrag, Modifiers, CurrentMouse);
}

void CCameraFrame::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnLButtonDrag functor;
	HandleDragEvent(this, functor, control_lbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnLButtonEndDrag functor;
	HandleDragEvent(this, functor, control_lbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnMButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnMButtonClick functor;
	HandleEvent(this, functor, control_mbuttonclick, Modifiers, CurrentMouse);
}

void CCameraFrame::OnMButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnMButtonDoubleClick functor;
	HandleEvent(this, functor, control_mbuttondoubleclick, Modifiers, CurrentMouse);
}

void CCameraFrame::OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnMButtonStartDrag functor;
	HandleEvent(this, functor, control_mbuttonstartdrag, Modifiers, CurrentMouse);
}

void CCameraFrame::OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnMButtonDrag functor;
	HandleDragEvent(this, functor, control_mbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnMButtonEndDrag functor;
	HandleDragEvent(this, functor, control_mbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnRButtonClick functor;
	HandleEvent(this, functor, control_rbuttonclick, Modifiers, CurrentMouse);
}

void CCameraFrame::OnRButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnRButtonDoubleClick functor;
	HandleEvent(this, functor, control_rbuttondoubleclick, Modifiers, CurrentMouse);
}

void CCameraFrame::OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnRButtonStartDrag functor;
	HandleEvent(this, functor, control_rbuttonstartdrag, Modifiers, CurrentMouse);
}

void CCameraFrame::OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnRButtonDrag functor;
	HandleDragEvent(this, functor, control_rbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnRButtonEndDrag functor;
	HandleDragEvent(this, functor, control_rbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CCameraFrameCallOnLRButtonStartDrag functor;
	HandleEvent(this, functor, control_lrbuttonstartdrag, Modifiers, CurrentMouse);
}

void CCameraFrame::OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnLRButtonDrag functor;
	HandleDragEvent(this, functor, control_lrbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	CCameraFrameCallOnLRButtonEndDrag functor;
	HandleDragEvent(this, functor, control_lrbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CCameraFrame::UpdateViewportDimensions()
{
	AspectFrame(control_viewportcontainer).Set(0.5, 0.5, m_Object->ViewVolume()->Viewport()->WindowAspectRatio(), false);
}

namespace
{

template<typename InsertIterator, typename PredicateType>
class PathInserterType
{
public:
	PathInserterType(InsertIterator Iterator, PredicateType Predicate) :
		m_Iterator(Iterator),
		m_Predicate(Predicate)
	{
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		k3d::copy_if(PathCollection, m_Iterator, m_Predicate);
	}

private:
	InsertIterator m_Iterator;
	PredicateType m_Predicate;
};

template<typename InsertIterator, typename PredicateType>
PathInserterType<InsertIterator, PredicateType> PathInserter(InsertIterator Iterator, PredicateType Predicate)
{
	return PathInserterType<InsertIterator, PredicateType>(Iterator, Predicate);
}

template<typename InsertIterator, typename PredicateType>
class PointInserterType
{
public:
	PointInserterType(InsertIterator Iterator, PredicateType Predicate) :
		m_Iterator(Iterator),
		m_Predicate(Predicate)
	{
	}

	void operator()(k3dIPointCollection* const PointCollection)
	{
		k3d::copy_if(PointCollection, m_Iterator, m_Predicate);
	}

private:
	InsertIterator m_Iterator;
	PredicateType m_Predicate;
};

template<typename InsertIterator, typename PredicateType>
PointInserterType<InsertIterator, PredicateType> PointInserter(InsertIterator Iterator, PredicateType Predicate)
{
	return PointInserterType<InsertIterator, PredicateType>(Iterator, Predicate);
}

template<typename PointContainerType>
class RemovePathPoint
{
public:
	RemovePathPoint(PointContainerType& Container) :
		m_Container(Container)
	{
	}

	void operator()(k3dIPathPoint* const PathPoint)
	{
		m_Container.erase(PathPoint->Point());
	}

private:
	PointContainerType& m_Container;
};

template<typename PointContainerType>
class RemovePathPoints
{
public:
	RemovePathPoints(PointContainerType& Container) :
		m_Container(Container)
	{
	}

	void operator()(k3dIPath* const Path)
	{
		k3d::for_each(Path->Points(), RemovePathPoint<PointContainerType>(m_Container));
	}

private:
	PointContainerType& m_Container;
};

template<typename PointContainerType>
class PathPointRemover
{
public:
	PathPointRemover(PointContainerType& Container) :
		m_Container(Container)
	{
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		k3d::for_each_if(PathCollection, RemovePathPoints<PointContainerType>(m_Container), k3d::FlagIsClear<k3dIPath*>(k3dIState::SELECTED));
	}

private:
	PointContainerType& m_Container;
};

} // namespace

void CCameraFrame::OnEditDelete()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_editdelete, 0);

	// Record undo/redo data ...
	k3dRecordStateChangeSet changeset(m_Object->Document(), "Delete selected");

	// Delete selected objects ...
	typedef std::vector<k3dIObject*> Objects;
	Objects objects;
	k3d::copy_if(m_Object->Document()->Objects(), std::back_inserter(objects), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTED));

	k3d::PushEditing(m_Object->Document()->Objects());
	for(Objects::iterator object = objects.begin(); object != objects.end(); object++)
		m_Object->Document()->Objects()->DeleteObject(*object);
	k3d::PopEditing(m_Object->Document()->Objects());

	// Get the set of selected paths ...
	typedef std::set<k3dIPath*>Paths;
	Paths paths;
	k3d::for_each_if<k3dIPathCollection>(m_Object->Document()->Objects(), PathInserter(std::inserter(paths, paths.begin()), k3d::FlagIsSet<k3dIPath*>(k3dIState::VISIBLESELECTED)), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));

	// Get the set of selected points ...
	typedef std::set<k3dIPoint*> Points;
	Points points;
	k3d::for_each_if<k3dIPointCollection>(m_Object->Document()->Objects(), PointInserter(std::inserter(points, points.begin()), k3d::FlagIsSet<k3dIPoint*>(k3dIState::VISIBLESELECTED)), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));

	// If we *aren't* in point-edit mode, make sure we don't delete any points that are in use by un-selected paths
	if(k3dICamera::POINT != m_Object->ActiveEditMode())
		k3d::for_each<k3dIPathCollection>(m_Object->Document()->Objects(), PathPointRemover<Points>(points));

	// Keep track of which meshes are going to end-up modified ...
	typedef std::set<k3dIMesh*> Meshes;
	Meshes meshes;
	for(Paths::iterator path = paths.begin(); path !=paths.end(); ++path)
		meshes.insert((*path)->Mesh());
	for(Points::iterator point = points.begin(); point != points.end(); ++point)
		meshes.insert((*point)->Mesh());

	// Zap those babies!
	std::for_each(meshes.begin(), meshes.end(), &k3d::PushEditing);

	for(Paths::iterator path = paths.begin(); path !=paths.end(); ++path)
		(*path)->Mesh()->DeletePath(*path);
	for(Points::iterator point = points.begin(); point != points.end(); ++point)
		(*point)->Mesh()->DeletePoint(*point);

	std::for_each(meshes.begin(), meshes.end(), &k3d::PopEditing);

/*
	// If we're in point-edit mode ...
	if(k3dIDocument::POINT == Document()->ActiveEditMode())
		{
			// Delete the (leftover) set of selected points ...
			k3dPointArray points;
			k3dGetPoints(m_Document, k3dIState::VISIBLESELECTED, points);

			Meshes meshes;
			for(k3dPointArray::iterator point = points.begin(); point != points.end(); point++)
				meshes.insert((*point)->Mesh());

			std::for_each(meshes.begin(), meshes.end(), &k3d::PushEditing);

			for(k3dPointArray::iterator point = points.begin(); point != points.end(); point++)
				(*point)->Mesh()->DeletePoint(*point);

*/

	k3dRedrawAll(m_Object->Document(), false);
}

/*
void CCameraFrame::UpdateActiveEditMode()
{
	BlockAllEvents();

	CheckMenuItem(control_editpointmode).SetState(k3dIDocument::POINT == Document()->ActiveEditMode());
	ToggleButton("toolbarpointmode").SetState(k3dIDocument::POINT == Document()->ActiveEditMode());

	UnblockAllEvents();
}

void CCameraFrame::UpdateActiveSnapMode()
{
	BlockAllEvents();

	CheckMenuItem(control_editsnapmode).SetState(k3dIDocument::SNAP == Document()->ActiveSnapMode());
	ToggleButton("toolbarsnapmode").SetState(k3dIDocument::SNAP == Document()->ActiveSnapMode());

	UnblockAllEvents();
}

void CCameraFrame::UpdateActiveAxis()
{
	BlockAllEvents();

	CheckMenuItem(control_editxaxis).SetState(k3dIDocument::XAXIS == Document()->ActiveAxis());
	CheckMenuItem(control_edityaxis).SetState(k3dIDocument::YAXIS == Document()->ActiveAxis());
	CheckMenuItem(control_editzaxis).SetState(k3dIDocument::ZAXIS == Document()->ActiveAxis());

	ToggleButton("toolbarxaxis").SetState(k3dIDocument::XAXIS == Document()->ActiveAxis());
	ToggleButton("toolbaryaxis").SetState(k3dIDocument::YAXIS == Document()->ActiveAxis());
	ToggleButton("toolbarzaxis").SetState(k3dIDocument::ZAXIS == Document()->ActiveAxis());

	UnblockAllEvents();
}

void CCameraFrame::UpdateActiveTransformation()
{
	BlockAllEvents();

	CheckMenuItem(control_editposition).SetState(k3dIDocument::POSITION == Document()->ActiveTransformType());
	CheckMenuItem(control_editorientation).SetState(k3dIDocument::ORIENTATION == Document()->ActiveTransformType());
	CheckMenuItem(control_editscale).SetState(k3dIDocument::SCALE == Document()->ActiveTransformType());

	ToggleButton("toolbarposition").SetState(k3dIDocument::POSITION == Document()->ActiveTransformType());
	ToggleButton("toolbarorientation").SetState(k3dIDocument::ORIENTATION == Document()->ActiveTransformType());
	ToggleButton("toolbarscale").SetState(k3dIDocument::SCALE == Document()->ActiveTransformType());

	UnblockAllEvents();
}
*/



