#ifndef CAMERA_H
#define CAMERA_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CCamera K-3D object, which controls the position, orientation, and display of a CCameraView within the 3D workspace
		\author Tim Shead (tshead@k-3d.com)
		\author Giuseppe Zompatori <mailus@tin.it>
*/

#include <k3dlocationcollection.h>
#include <k3dshaderinstance.h>
#include <k3dtransformobject.h>
#include <k3dvariable.h>

#include <k3dobjectdialog.h>

#include <k3dicamera.h>
#include <k3ditime.h>
#include <k3diviewport.h>
#include <k3diviewvolume.h>

#include <sdpgtkmouseinput.h>
#include <sdpgtkopengldrawingarea.h>

#include "editorengine.h"
#include "riengine.h"

// Forward class declarations
class k3dEventConnector;
class k3dIViewVolume;

/////////////////////////////////////////////////////////////////////////////
// CCameraDialog

typedef k3dObjectDialog CCameraDialogBase;

class CCameraDialog :
	public CCameraDialogBase
{
public:
	CCameraDialog(CCamera* const Object);

protected:
	CCamera* const m_Object;

	void UpdateControls();

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnRiEngine();
	void OnRiPixelFilterName();
	void OnRiShadingInterpolation();

	void OnShowViewport();
	void OnHideViewport();

	static const sdpString dialogname;
	static const sdpString control_editorwireframe;
	static const sdpString control_editorsolid;
	static const sdpString control_editorfog;
	static const sdpString control_editorfognear;
	static const sdpString control_editorfogfar;
	static const sdpString control_editortexture;
	static const sdpString control_editortwosided;
	static const sdpString control_editorwireframetwosided;
	static const sdpString control_editororientation;
	static const sdpString control_editornormals;
	static const sdpString control_riengine;
	static const sdpString control_rixsamples;
	static const sdpString control_riysamples;
	static const sdpString control_ripixelfiltername;
	static const sdpString control_ripixelfilterwidth;
	static const sdpString control_ripixelfilterheight;
	static const sdpString control_ritwosided;
	static const sdpString control_rishadingrate;
	static const sdpString control_rishadinginterpolation;
	static const sdpString control_rigridsize;
	static const sdpString control_rixbucketsize;
	static const sdpString control_riybucketsize;
	static const sdpString control_rieyesplits;
	static const sdpString control_riatmosphere;
	static const sdpString control_ridepthoffield;
	static const sdpString control_riradiosity;
	static const sdpString control_riradiositysteps;
	static const sdpString control_ritexturememory;
	static const sdpString control_rienablelerp;
	static const sdpString control_rienablegaussian;
	static const sdpString control_showviewport;
	static const sdpString control_hideviewport;
	static const sdpString control_focalplane;
	static const sdpString control_fstop;
	static const sdpString control_riexposure;
	static const sdpString control_rigamma;
	static const sdpString control_riatmosphereshaderselector;
	static const sdpString control_riatmosphereshadercontainer;
	static const sdpString control_outputpixelwidth;
	static const sdpString control_outputpixelheight;
	static const sdpString control_outputpixelaspectratio;
	static const sdpString control_windowpixelaspectratio;
	static const sdpString control_farplane;
	static const sdpString control_nearplane;
	static const sdpString control_focallength;
	static const sdpString control_orthographic;
	static const sdpString control_position;
	static const sdpString control_orientation;
	static const sdpString control_scale;
	static const sdpString control_target;
};

///////////////////////////////////////////////////////////
// CCameraFrame

typedef k3dObjectDialog CCameraFrameBase;

class CCameraFrame :
	public CCameraFrameBase,
	public sdpGtkMouseInput
{
public:
	CCameraFrame(CCamera* const Object);
	~CCameraFrame();

	CCamera* Object() { return m_Object; }

	// k3dICommandNode overrides ...
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// Accesses the viewport
	sdpGtkOpenGLDrawingArea& DrawingArea();

private:
	CCamera* const m_Object;
	sdpGtkOpenGLDrawingArea m_DrawingArea;
	k3dEventConnector* m_DocumentEvents;

	// Event handlers ...
	void OnEvent(sdpGtkEvent* Event);
	
	void OnClose();
	void OnFocusIn();
	void OnRenderPreview();
	void OnRenderFrame();
	void OnRenderAnimation();
	void OnDefaultView();
	void OnAxisView(SDPCSTRING Axis);
	void OnConfigureEvent(sdpGtkEvent* Event);
	void OnExposeEvent(sdpGtkEvent* Event);

	/// Called to delete selected geometry
	void OnEditDelete();
/*
	/// Called when the user toggles between point- and path-edit mode
	void OnEditPointMode();
	/// Called when the user toggles between freehand- and snap- mode
	void OnEditSnapMode();
	/// Called when the user makes the X axis active
	void OnEditXAxis();
	/// Called when the user makes the Y axis active
	void OnEditYAxis();
	/// Called when the user makes the Z axis active
	void OnEditZAxis();
	/// Called when the user wants to modify positions
	void OnEditPosition();
	/// Called when the user wants to modify orientations
	void OnEditOrientation();
	/// Called when the user wants to modify scales
	void OnEditScale(); 
*/

	// sdpGtkMouseInput overrides ...
	void OnMouseMove(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	void OnMButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	void OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	void OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	
	void UpdateViewportDimensions();
	
/*
	/// Updates the active axis menu commands as needed
	void UpdateActiveAxis();
	/// Updates the active transformation menu commands as needed
	void UpdateActiveTransformation();
	/// Updates the active coordinate system menu commands as needed
	void UpdateActiveCoordinateSystem();
	/// Updates the edit-mode menu commands as needed
	void UpdateActiveEditMode();
	/// Updates the snap-mode controls as needed
	void UpdateActiveSnapMode();
*/

	static const sdpString dialogname;
	static const sdpString control_viewportcontainer;
	static const sdpString control_focus;
	static const sdpString control_interactiveusage;

	static const sdpString control_editdelete;
/*
	static const sdpString control_editpointmode;
	static const sdpString control_editsnapmode;
	static const sdpString control_editxaxis;
	static const sdpString control_edityaxis;
	static const sdpString control_editzaxis;
	static const sdpString control_editposition;
	static const sdpString control_editorientation;
	static const sdpString control_editscale;
*/
		
	static const sdpString control_close;
	static const sdpString control_setfocus;
	static const sdpString control_renderpreview;
	static const sdpString control_renderframe;
	static const sdpString control_renderanimation;
	static const sdpString control_defaultview;
	static const sdpString control_negativezview;
	static const sdpString control_positivezview;
	static const sdpString control_negativexview;
	static const sdpString control_positivexview;
	static const sdpString control_negativeyview;
	static const sdpString control_positiveyview;
	
	static const sdpString control_mousemove;

	static const sdpString control_lbuttonclick;
	static const sdpString control_lbuttondoubleclick;
	static const sdpString control_lbuttonstartdrag;
	static const sdpString control_lbuttondrag;
	static const sdpString control_lbuttonenddrag;

	static const sdpString control_mbuttonclick;
	static const sdpString control_mbuttondoubleclick;
	static const sdpString control_mbuttonstartdrag;
	static const sdpString control_mbuttondrag;
	static const sdpString control_mbuttonenddrag;

	static const sdpString control_rbuttonclick;
	static const sdpString control_rbuttondoubleclick;
	static const sdpString control_rbuttonstartdrag;
	static const sdpString control_rbuttondrag;
	static const sdpString control_rbuttonenddrag;

	static const sdpString control_lrbuttonstartdrag;
	static const sdpString control_lrbuttondrag;
	static const sdpString control_lrbuttonenddrag;
};

///////////////////////////////////////////////////////////
// CCamera

typedef k3dTransformObject CCameraBase;

class CCamera :
	public CCameraBase,
	public k3dICamera,
	public k3dIViewport,
	public k3dIViewVolume,
	sdpsl::IShaderCollection
{
friend class CEditorEngine;
friend class COpenGLEngine;
friend class CRiEngine;
friend class CCameraDialog;
friend class CCameraFrame;
public:
	CCamera(sdpIUnknown* Initializer);
	virtual ~CCamera();

	CCameraFrame* m_Frame;

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// k3dIObject implementation
	sdpResult EditObject();

	// k3dIAnimate implementation ...
	sdpResult PostAnimate();

	// sdpsl::IShaderCollection implementation
	void VisitShaders(sdpsl::IShaderVisitor* const Visitor);

	// k3dIEditorRender implementation ...
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3dICamera implementation ...
	sdpResult ShowViewport();
	sdpResult HideViewport();
	gulong Width();
	gulong Height();
	sdpResult Redraw(bool Synchronous);
	sdpResult RenderPreview();
	sdpResult RenderFrame();
	sdpResult RenderAnimation();
	sdpResult SnapshotCameraCoordinates();
	sdpResult SnapshotScreenCoordinates();
	k3dIViewVolume* ViewVolume();
	Axis ActiveAxis();
	void SetActiveAxis(const Axis ActiveAxis);
	k3dICamera::TransformType ActiveTransformType();
	void SetActiveTransformType(const k3dICamera::TransformType ActiveTransformType);
	CoordinateSystem ActiveCoordinateSystem();
	void SetActiveCoordinateSystem(const CoordinateSystem System);
	EditMode ActiveEditMode();
	void SetActiveEditMode(const EditMode Mode);
	SnapMode ActiveSnapMode();
	void SetActiveSnapMode(const SnapMode Mode);

	// k3dIViewport implementation ...
	const gdouble OutputPixelWidth();
	const gdouble OutputPixelHeight();
	const gdouble OutputPixelAspectRatio();
	const gdouble WorldAspectRatio();
	const gdouble WindowPixelWidth(const gdouble WindowPixelHeight);
	const gdouble WindowPixelHeight(const gdouble WindowPixelWidth);
	const gdouble WindowPixelAspectRatio();
	const gdouble WindowAspectRatio();

	// k3dIViewVolume implementation ...
	k3dIViewport* Viewport();
	const bool Orthographic();
	const gdouble NearPlane();
	const gdouble FarPlane();
	const gdouble Left();
	const gdouble Right();
	const gdouble Top();
	const gdouble Bottom();
	void CameraToScreen(gdouble Coords[3], const gdouble WindowPixelWidth, const gdouble WindowPixelHeight);

	// k3dIMouseEventObserver implementation
	sdpResult OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDoubleClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnRButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnRButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnRButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnMButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnMButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

	// k3dIPersist implementation
	sdpResult InitComplete();

	/// Redraws the viewport
	sdpResult Redraw();
	/// Returns the viewport
	sdpGtkOpenGLDrawingArea* DrawingArea();

protected:
	// Cache some document interfaces ...
	k3dITime* const m_Time;

	// Rendering engines ...
	CEditorEngine m_EditorEngine;
	CRiEngine m_RiEngine;

	k3dIRenderEngine* m_DisplayEngine;
	k3dIRenderEngine* m_RenderEngine;	

	/// Enumerates field-dominance values for interlaced rendering
	typedef enum
	{
		ODDFIELDFIRST,
		EVENFIELDFIRST,
	} Field;

	friend std::ostream& operator<<(std::ostream& Stream, const Field& Value)
	{
		switch(Value)
			{
				case ODDFIELDFIRST:
					Stream << "odd";
					break;
				case EVENFIELDFIRST:
					Stream << "even";
					break;
			}
	
		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, Field& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "odd")
			Value = ODDFIELDFIRST;
		else if(text == "even")
			Value = EVENFIELDFIRST;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}

	// Provide stream operators for k3dICamera enumerations
	friend std::ostream& operator<<(std::ostream& Stream, const k3dICamera::Axis& Value)
	{
		switch(Value)
			{
				case XAXIS:
					Stream << "xaxis";
					break;
				case YAXIS:
					Stream << "yaxis";
					break;
				case ZAXIS:
					Stream << "zaxis";
					break;
			}
		return Stream;
	}
	
	friend std::ostream& operator<<(std::ostream& Stream, const k3dICamera::TransformType& Value)
	{
		switch(Value)
			{
				case POSITION:
					Stream << "position";
					break;
				case ORIENTATION:
					Stream << "orientation";
					break;
				case SCALE:
					Stream << "scale";
					break;
			}
		return Stream;
	}
	
	friend std::ostream& operator<<(std::ostream& Stream, const k3dICamera::CoordinateSystem& Value)
	{
		switch(Value)
			{
				case MODEL:
					Stream << "model";
					break;
				case WORLD:
					Stream << "world";
					break;
				case SCREEN:
					Stream << "screen";
					break;
			}
		return Stream;
	}
	
	friend std::ostream& operator<<(std::ostream& Stream, const k3dICamera::EditMode& Value)
	{
		switch(Value)
			{
				case POINT:
					Stream << "point";
					break;
				case POLYGON:
					Stream << "polygon";
					break;
			}
		return Stream;
	}
	
	friend std::ostream& operator<<(std::ostream& Stream, const k3dICamera::SnapMode& Value)
	{
		switch(Value)
			{
				case FREEHAND:
					Stream << "freehand";
					break;
				case SNAP:
					Stream << "snap";
					break;
			}
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, k3dICamera::Axis& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "xaxis")
			Value = XAXIS;
		else if(text == "yaxis")
			Value = YAXIS;
		else if(text == "zaxis")
			Value = ZAXIS;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, k3dICamera::TransformType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "position")
			Value = POSITION;
		else if(text == "orientation")
			Value = ORIENTATION;
		else if(text == "scale")
			Value = SCALE;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, k3dICamera::CoordinateSystem& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "model")
			Value = MODEL;
		else if(text == "world")
			Value = WORLD;
		else if(text == "screen")
			Value = SCREEN;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, k3dICamera::EditMode& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "point")
			Value = POINT;
		else if(text == "polygon")
			Value = POLYGON;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, k3dICamera::SnapMode& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "freehand")
			Value = FREEHAND;
		else if(text == "snap")
			Value = SNAP;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	// Non-persistent data ...
	k3dObjectVariable m_Focus;

	// Camera data ...
	k3dVariable<k3dICamera::Axis> m_ActiveAxis;
	k3dVariable<k3dICamera::TransformType> m_ActiveTransformType;
	k3dVariable<k3dICamera::CoordinateSystem> m_ActiveCoordinateSystem;
	k3dVariable<k3dICamera::EditMode> m_ActiveEditMode;
	k3dVariable<k3dICamera::SnapMode> m_ActiveSnapMode;

	// Viewport data ...
	k3dVariable<bool> m_ViewportVisible;

	// Output data ...
	k3dVariable<gdouble> m_OutputPixelWidth;
	k3dVariable<gdouble> m_OutputPixelHeight;
	k3dVariable<gdouble> m_OutputPixelAspectRatio;
	k3dVariable<gdouble> m_WindowPixelAspectRatio;
	k3dVariable<bool> m_Orthographic;
	k3dVariable<Field> m_Field;

	// Editor render data ...
	k3dVariable<bool> m_EditorWireframe;
	k3dVariable<bool> m_EditorSolid;
	k3dVariable<bool> m_EditorFog;
	k3dVariable<gdouble> m_EditorFogNear;
	k3dVariable<gdouble> m_EditorFogFar;
	k3dVariable<bool> m_EditorSmoothing;
	k3dVariable<bool> m_EditorTexture;
	k3dVariable<bool> m_EditorTwoSided;
	k3dVariable<bool> m_EditorWireframeTwoSided;
	k3dVariable<bool> m_EditorOrientation;
	k3dVariable<bool> m_EditorNormals;
	k3dVariable<gulong> m_EditorWireframePatchDivisions;
	k3dVariable<gulong> m_EditorSolidPatchDivisions;

	// RIB render data ...
	k3dVariable<sdpString> m_RiRenderEngine;
	k3dVariable<gdouble> m_RiXSamples;
	k3dVariable<gdouble> m_RiYSamples;
	k3dVariable<sdpString> m_RiPixelFilterName;
	k3dVariable<gdouble> m_RiPixelFilterWidth;
	k3dVariable<gdouble> m_RiPixelFilterHeight;
	k3dVariable<bool> m_RiTwoSided;
	k3dVariable<gdouble> m_RiShadingRate;
	k3dVariable<sdpString> m_RiShadingInterpolation;
	k3dVariable<gdouble> m_RiGridSize;
	k3dVariable<gdouble> m_RiXBucketSize;
	k3dVariable<gdouble> m_RiYBucketSize;
	k3dVariable<gdouble> m_RiEyeSplits;
	k3dVariable<bool> m_RiAtmosphere;
	k3dShaderInstance m_RiAtmosphereShader;
	k3dVariable<bool> m_RiDepthOfField;
	k3dVariable<bool> m_RiMotionBlur;
	k3dVariable<bool> m_RiFieldRender;
	k3dVariable<bool> m_RiRadiosity;
	k3dVariable<gdouble> m_RiRadiositySteps;
	k3dVariable<gulong> m_RiRadiosityPatchSamples;
	k3dVariable<gdouble> m_RiRadiosityPatchSize;
	k3dVariable<gdouble> m_RiRadiosityElementSize;
	k3dVariable<gdouble> m_RiRadiosityMinimumSize;
	k3dVariable<gdouble> m_RiTextureMemory;
	k3dVariable<bool> m_RiEnableLerp;
	k3dVariable<bool> m_RiEnableGaussian;

	// Owned objects ...
	k3dScalarChannel m_Focallength;
	k3dScalarChannel m_Focalplane;
	k3dScalarChannel m_Fstop;
	k3dScalarChannel m_Shutteropen;
	k3dScalarChannel m_Shutterclose;
	k3dScalarChannel m_Exposure;
	k3dScalarChannel m_Gamma;
	k3dScalarChannel m_Xdirection;
	k3dScalarChannel m_Ydirection;
	k3dScalarChannel m_Zdirection;
	k3dScalarChannel m_FarPlane;
	k3dScalarChannel m_NearPlane;

	// Colors ...
	static gdouble m_EditorBackgroundColor[4];
	static gdouble m_EditorNormalColor[4];
	static gdouble m_EditorSelectedColor[4];
	static gdouble m_EditorUnselectedColor[4];
	static gdouble m_EditorControlPointColor[4];
	static gdouble m_EditorFirstColor[4];
	static gdouble m_EditorActiveColor[4];

	// User preferences ...
	static GLfloat m_PointSize;
	static gulong m_ClickSensitivity;
	static gdouble m_PositionSensitivity;
	static gdouble m_OrientationSensitivity;
	static gdouble m_ZoomSign;
	static bool m_LassoSelectPartial;
	static bool m_BoxSelectPartial;

	/// Current camera position in world coordinates
	sdpVector3 m_CurrentPosition;
	/// Current camera orientation in world coordinates
	sdpAngleAxis m_CurrentOrientation;

	/// Calculates parameters related to the viewing frustum
	void SetupViewVolume();

	// Selection related ...
	sdpVector2Array m_LassoPoints;
	sdpRectangle m_RubberBand;

	// Our OpenGL context ...
	sdpGtkOpenGLDrawingArea m_OpenGLWidget;
	
	// Helper functions for doing view calculations
	const gdouble OrthographicWorldHeight();
	const gdouble PerspectiveFocalLength();
	const gdouble HorizontalFOV();
	const gdouble VerticalFOV();
};


#endif // CAMERA_H

