// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CChannel class which implements a data object that supports k3dIChannel
		\author Tim Shead (tshead@k-3d.com)
*/

#include <channel.h>

#include <k3dmisc.h>
#include <k3dmouseeventobserver.h>
#include <k3doptions.h>
#include <k3dstaterecorder.h>

#include <k3diapplication.h>
#include <k3dibezierchannelnode.h>
#include <k3dibezierchannelvaluenode.h>
#include <k3distate.h>
#include <k3ditime.h>

#include <sdpgtkcolorselector.h>
#include <sdpgtkevents.h>
#include <sdpgtkutility.h>

#include <iterator>

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

//////////////////////////////////////////////////////////////////////////
// CChannelNode

CChannelNode::CChannelNode(CChannel* const Channel, const sdpVector2 Position) :
	m_Position(Position),
	m_Channel(Channel),
	m_PositionChanged(false)
{
}

CChannelNode::~CChannelNode()
{
}

sdpResult CChannelNode::Position(sdpVector2& Position)
{
	Position = m_Position;
	return SDP_S_OK;
}

sdpResult CChannelNode::SetPosition(const sdpVector2& Position)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Channel->Editing(), SDP_E_FAIL);
	
	RawSetPosition(Position);	
	
	return SDP_S_OK;
}

void CChannelNode::RawSetPosition(const sdpVector2& Position)
{
	sdpVector2 position(Position);
	m_Channel->ConstrainNodePosition(this, position);
	
	UnconstrainedSetPosition(position);
}

void CChannelNode::UnconstrainedSetPosition(const sdpVector2& Position)
{
	k3dIStateChangeSet* const changeset = m_Channel->Document()->StateRecorder()->ChangeSet();
	if(changeset && !m_PositionChanged)
		{
			changeset->RecordOldState(new k3dValueStateContainer<sdpVector2>(m_Position));
			m_PositionChanged = true;
		}

	m_Position = Position;
}

CChannel* const CChannelNode::Channel() const
{
	return m_Channel;
}
	
CChannelValueNode* CChannelNode::ValueNode()
{
	return 0;
}

void CChannelNode::Deleted()
{
	Notify(k3dIBezierChannelNode::Event_Deleted, 0);
}

void CChannelNode::UnDeleted()
{
	Notify(k3dIBezierChannelNode::Event_Undeleted, 0);
}

void CChannelNode::PushEditing()
{
	if(!Editing())
		m_PositionChanged = false;

	k3dEditable::PushEditing();
}

void CChannelNode::PopEditing()
{
	k3dEditable::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = m_Channel->Document()->StateRecorder()->ChangeSet();
			if(changeset && m_PositionChanged)
				changeset->RecordNewState(new k3dValueStateContainer<sdpVector2>(m_Position));
		}
}

sdpResult CChannelNode::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const
{
	// Sanity checks ...
	g_assert(Dependencies);

	Element.Append(sdpxml::Element("node", "", sdpxml::Attribute("coords", sdpToString(m_Position))));

	return SDP_S_OK;
}

sdpResult CChannelNode::Load(sdpxml::Element& Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element.Name() == "node", SDP_E_FAIL);

	sdpxml::ParseAttribute(Element, "coords", m_Position);

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////////
// CChannelValueNode

CChannelValueNode::CChannelValueNode(CChannel* const Channel, const sdpVector2 Position, const sdpVector3 Value) :
	CChannelValueNodeBase(Channel, Position),
	m_Value(Value)
{
}

CChannelValueNode* CChannelValueNode::ValueNode()
{
	return this;
}

void CChannelValueNode::PushEditing()
{
	if(!Editing())
		m_ValueChanged = false;

	CChannelValueNodeBase::PushEditing();
}

void CChannelValueNode::PopEditing()
{
	CChannelValueNodeBase::PopEditing();

	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = m_Channel->Document()->StateRecorder()->ChangeSet();
			if(changeset && m_ValueChanged)
				changeset->RecordNewState(new k3dValueStateContainer<sdpVector3>(m_Value));
		}
}

const sdpVector3 CChannelValueNode::Value() const
{
	return m_Value;
}

sdpResult CChannelValueNode::Value1(gdouble& Value)
{
	Value = m_Value[0];
	return SDP_S_OK;
}

sdpResult CChannelValueNode::Value2(sdpVector2& Value)
{
	Value[0] = m_Value[0];
	Value[1] = m_Value[1];
	return SDP_S_OK;
}

sdpResult CChannelValueNode::Value3(sdpVector3& Value)
{
	// Sanity checks ...
	Value = m_Value;
	return SDP_S_OK;
}

sdpResult CChannelValueNode::SetValue1(const gdouble Value)
{
	return SetValue3(sdpVector3(Value, 0, 0));
}

sdpResult CChannelValueNode::SetValue2(const sdpVector2& Value)
{
	return SetValue3(sdpVector3(Value[0], Value[1], 0));
}

sdpResult CChannelValueNode::SetValue3(const sdpVector3& Value)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Channel->Editing(), SDP_E_FAIL);

	k3dIStateChangeSet* const changeset = m_Channel->Document()->StateRecorder()->ChangeSet();
	if(changeset && !m_ValueChanged)
		{
			changeset->RecordOldState(new k3dValueStateContainer<sdpVector3>(m_Value));
			m_ValueChanged = true;
		}

	m_Value = Value;
	return SDP_S_OK;
}

sdpResult CChannelValueNode::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const
{
	// Sanity checks ...
	g_assert(Dependencies);

	Element.Append(sdpxml::Element("valuenode", "", sdpxml::Attribute("coords", sdpToString(m_Position)), sdpxml::Attribute("value", sdpToString(m_Value))));

	return SDP_S_OK;
}

sdpResult CChannelValueNode::Load(sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(Element.Name() == "valuenode");

	sdpxml::ParseAttribute(Element, "coords", m_Position);
	sdpxml::ParseAttribute(Element, "value", m_Value);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////
// CChannel

const gulong CChannel::m_MaxIterations = 5000;

CChannel::CChannel(sdpIUnknown* Initializer) :
	CChannelBase(Initializer),
	m_DataType("datatype", k3dIChannel::DATA_SCALAR, Document()->StateRecorder(), 0),
	m_MinValue("minvalue", -DBL_MAX, Document()->StateRecorder(), 0),
	m_MaxValue("maxvalue", DBL_MAX, Document()->StateRecorder(), 0),
	m_MinCoords("mincoords", sdpVector2(0.0, -DBL_MAX), Document()->StateRecorder(), 0),
	m_MaxCoords("maxcoords", sdpVector2(1.0, DBL_MAX), Document()->StateRecorder(), 0),
	m_Cycles("cycles", 1.0, Document()->StateRecorder(), 0),
	m_Phase("phase", 0.0, Document()->StateRecorder(), 0)
{
	AddVariable(&m_DataType);
	AddVariable(&m_MinValue);
	AddVariable(&m_MaxValue);
	AddVariable(&m_MinCoords);
	AddVariable(&m_MaxCoords);
	AddVariable(&m_Cycles);
	AddVariable(&m_Phase);

	// We always have at least one node ...
	CreateSegment(sdpVector2());
	
	// Make sure this object doesn't show up in the hierarchy ...
	RawSetState(k3dIState::USERVIEWABLE, false);
}

CChannel::~CChannel()
{
	std::for_each(m_Nodes.begin(), m_Nodes.end(), sdp::DeleteObject());
}

CChannelNode* CChannel::CreateNode(const sdpVector2 Position)
{
	// Constrain the new coordinates
	const sdpVector2 coords(sdpClamp(m_MinCoords.Value()[0], Position[0], m_MaxCoords.Value()[0]), sdpClamp(m_MinCoords.Value()[1], Position[1], m_MaxCoords.Value()[1]));

	// Create the node ...
	CChannelNode* const newnode = new CChannelNode(this, coords);
	return newnode;
}

CChannelValueNode* CChannel::CreateValueNode(const sdpVector2 Position, const sdpVector3 Value)
{
	// Constrain the new coordinates
	sdpVector2 coords(sdpClamp(m_MinCoords.Value()[0], Position[0], m_MaxCoords.Value()[0]), sdpClamp(m_MinCoords.Value()[1], Position[1], m_MaxCoords.Value()[1]));
	if(k3dIChannel::DATA_COLOR == m_DataType.Value())
		coords[1] = sdpRound(coords[1]);

	// Create the node ...
	CChannelValueNode* const newnode = new CChannelValueNode(this, coords, Value);
	return newnode;
}

NodeIterator CChannel::GetInsertPosition(const sdpVector2 Position)
{
	// If we don't have anything yet, it's easy ...
	if(0 == m_Nodes.size())
		return m_Nodes.end();
		
	// We have one other node, which is also easy ...
	if(1 == m_Nodes.size())
		return (Position[0] < m_Nodes.front()->m_Position[0]) ? m_Nodes.begin() : m_Nodes.end();

	// Before the first node ...
	if(Position[0] < m_Nodes.front()->m_Position[0])
		return m_Nodes.begin();
		
	// After the last node ...
	if(Position[0] >= m_Nodes.back()->m_Position[0])
		return m_Nodes.end();

	// Iterate through what's left ...
	for(NodeIterator node = m_Nodes.begin() + 3; node != m_Nodes.end(); node++)
		{
			if(!(*node)->ValueNode())
				continue;
				
			if(Position[0] < (*node)->m_Position[0])
				return node-1;
		}
		
	// Should never happen!
	g_return_val_if_fail(0, m_Nodes.end());
	return m_Nodes.end();
}

CChannelNode* CChannel::CreateSegment(const sdpVector2 Position)
{
	static const gdouble neardelta = 0.3;
	static const gdouble fardelta = 0.7;

	// If we're recording undos ...
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_NodeCollectionChanged)
				{
					changeset->RecordOldState(new k3dValueStateContainer<NodeCollection>(m_Nodes));
					m_NodeCollectionChanged = true;
				}
		}
	
	// Keep track of what we create ...
	NodeCollection newnodes;

	// Figure out where we're gonna put the new node(s) ...
	NodeIterator insertposition = GetInsertPosition(Position);

	// Create a new value value ...
	CChannelNode* newnode = CreateValueNode(Position, sdpVector3(1, 1, 1));
	newnodes.push_back(newnode);
	
	// Create either 0 or two other nodes, depending on where we are ...	
	if(0 == m_Nodes.size())
		{
			// This is was the first node inserted into the curve, so we're done ...
		}
	else if(insertposition == m_Nodes.begin())
		{
			// We were inserted to the left of any other nodes ...
			CChannelNode* const right=*(insertposition);

			newnodes.push_back(CreateNode(sdpVector2(sdpLerp<gdouble>(newnode->m_Position[0], right->m_Position[0], neardelta), newnode->m_Position[1])));
			newnodes.push_back(CreateNode(sdpVector2(sdpLerp<gdouble>(newnode->m_Position[0], right->m_Position[0], fardelta), right->m_Position[1])));
		}
	else if(insertposition == m_Nodes.end())
		{
			// We were inserted to the right of any other nodes ...
			CChannelNode* const left=*(insertposition-1);
						
			newnodes.insert(newnodes.begin(), CreateNode(sdpVector2(sdpLerp<gdouble>(newnode->m_Position[0], left->m_Position[0], neardelta), newnode->m_Position[1])));
			newnodes.insert(newnodes.begin(), CreateNode(sdpVector2(sdpLerp<gdouble>(newnode->m_Position[0], left->m_Position[0], fardelta), left->m_Position[1])));
		}
	else
		{
			// We were inserted between two other nodes ...
			CChannelNode* const valueleft=*(insertposition - 2);
			CChannelNode* const valueright=*(insertposition + 1);

			newnodes.insert(newnodes.begin(), CreateNode(sdpVector2(sdpLerp<gdouble>(newnode->m_Position[0], valueleft->m_Position[0], neardelta), newnode->m_Position[1])));
			newnodes.push_back(CreateNode(sdpVector2(sdpLerp<gdouble>(newnode->m_Position[0], valueright->m_Position[0], neardelta), newnode->m_Position[1])));
		}
	
	// Insert the new nodes into the collection ...
	m_Nodes.insert(insertposition, newnodes.begin(), newnodes.end());

	// Put the new nodes under undo/redo control ...
	if(changeset)
		for(NodeCollection::iterator newnode = newnodes.begin(); newnode != newnodes.end(); newnode++)
			changeset->RecordOldState(new k3dNewInstanceStateContainer<CChannelNode>(*newnode, changeset));

	// Constrain existing control nodes ...
	newnode->RawSetPosition(const_cast<sdpVector2&>(Position));

	return newnode;
}

void CChannel::RawDeleteNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_if_fail(Node);

	NodeIterator iterator = std::find(m_Nodes.begin(), m_Nodes.end(), Node);
	g_return_if_fail(iterator != m_Nodes.end());
	
	k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
	if(changeset)
		{
			if(!m_NodeCollectionChanged)
				{
					changeset->RecordOldState(new k3dValueStateContainer<NodeCollection>(m_Nodes));
					m_NodeCollectionChanged = true;
				}
		}
			
	m_Nodes.erase(iterator);

	if(changeset)
		changeset->RecordOldState(new k3dDeleteInstanceStateContainer<CChannelNode>(Node, changeset));
		
	Node->Deleted();
}

void CChannel::DeleteNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_if_fail(Node);
	
	// If there's only one node left, we ain't gonna delete it ...
	g_return_if_fail(m_Nodes.size() > 1);

	// Build a set of related nodes to be deleted ...
	NodeIterator node = std::find(m_Nodes.begin(), m_Nodes.end(), Node);
	g_return_if_fail(node != m_Nodes.end());
			
	typedef std::set<CChannelNode*> NodeSet;
	NodeSet nodes;
	
	if((*node)->ValueNode())
		{
			if(IsFirstNode(*node))
				{
					nodes.insert(*node);
					nodes.insert(*(node+1));
					nodes.insert(*(node+2));
				}
			else if(IsLastNode(*node))
				{
					nodes.insert(*node);
					nodes.insert(*(node-1));
					nodes.insert(*(node-2));
				}
			else	
				{
					nodes.insert(*node);
					nodes.insert(*(node-1));
					nodes.insert(*(node+1));
				}
		}
	else
		{
			CChannelNode* const valuenode = GetValueNode(*node);
			g_return_if_fail(valuenode);
			
			nodes.insert(*node);
			nodes.insert(valuenode);
			
			if(IsFirstNode(valuenode) || IsLastNode(valuenode))
				nodes.insert(GetNeighborControlNode(*node));
			else
				nodes.insert(GetOppositeControlNode(*node));
		}
		
	// Delete the identified nodes ...
	for(NodeSet::iterator victim = nodes.begin(); victim != nodes.end(); victim++)
		RawDeleteNode(*victim); 
}

void CChannel::ClearNodes()
{
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node = m_Nodes.begin())
		RawDeleteNode(*node);
}

NodeIterator CChannel::GetNodeIterator(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_val_if_fail(Node, m_Nodes.end());
	
	NodeIterator result = std::find(m_Nodes.begin(), m_Nodes.end(), Node);
	g_return_val_if_fail(result != m_Nodes.end(), m_Nodes.end());
	
	return result;
}

CChannelNode* CChannel::GetPrevNode(CChannelNode* const Node)
{
	const NodeIterator node(GetNodeIterator(Node));
	g_return_val_if_fail(node != m_Nodes.end(), 0);
	
	return IsFirstNode(*node) ? 0 : *(node-1);
}

CChannelNode* CChannel::GetNextNode(CChannelNode* const Node)
{
	const NodeIterator node(GetNodeIterator(Node));
	g_return_val_if_fail(node != m_Nodes.end(), 0);
	
	return IsLastNode(*node) ? 0 : *(node+1);
}

CChannelNode* CChannel::GetOppositeControlNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_val_if_fail(Node, 0);
	g_return_val_if_fail(!Node->ValueNode(), 0);
	g_return_val_if_fail(m_Nodes.size() >= 4, 0);
	
	// Get the iterator for this node ...
	const NodeIterator node(std::find(m_Nodes.begin(), m_Nodes.end(), Node));
	g_return_val_if_fail(node != m_Nodes.begin(), 0);
	g_return_val_if_fail(node != m_Nodes.end(), 0);
	g_return_val_if_fail(node+1 != m_Nodes.end(), 0);
	
	// Find the closest value node and return the control node on the opposite side ...
	const NodeIterator left(node-1);
	const NodeIterator right(node+1);
	
	if((*left)->ValueNode())
		{
			if(left == m_Nodes.begin())
				{
					return *(m_Nodes.end() - 2);
				}
			else
				{
					return *(left-1);
				}
		}
	else if((*right)->ValueNode())
		{
			if(right+1 == m_Nodes.end())
				{
					return *(m_Nodes.begin() + 1);
				}
			else
				{
					return *(right+1);
				}
		}
			
	// Something went wrong ...
	g_return_val_if_fail(0, 0);
	return 0;
}

CChannelNode* CChannel::GetValueNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_val_if_fail(Node, 0);
	g_return_val_if_fail(!Node->ValueNode(), 0);
	g_return_val_if_fail(m_Nodes.size() >= 4, 0);
	
	// Get the iterator for this node ...
	NodeIterator node = std::find(m_Nodes.begin(), m_Nodes.end(), Node);
	g_return_val_if_fail(node != m_Nodes.begin(), 0);
	g_return_val_if_fail(node != m_Nodes.end(), 0);
	g_return_val_if_fail(node+1 != m_Nodes.end(), 0);
	
	// Find the closest value node ...
	const NodeIterator left(node-1);
	const NodeIterator right(node+1);
	if((*left)->ValueNode())
		return *left;
	else if((*right)->ValueNode())
		return *right;
		
	// Something went wrong ...
	g_return_val_if_fail(0, 0);
	return 0;
}

CChannelNode* CChannel::GetNeighborControlNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_val_if_fail(Node, 0);
	g_return_val_if_fail(!Node->ValueNode(), 0);
	g_return_val_if_fail(m_Nodes.size() >= 4, 0);
	
	// Get the iterator for this node ...
	NodeIterator node = std::find(m_Nodes.begin(), m_Nodes.end(), Node);
	g_return_val_if_fail(node != m_Nodes.begin(), 0);
	g_return_val_if_fail(node != m_Nodes.end(), 0);
	g_return_val_if_fail(node+1 != m_Nodes.end(), 0);
	
	// Find the closest value node ...
	const NodeIterator left(node-1);
	const NodeIterator right(node+1);
	if((*left)->ValueNode())
		return *right;
	else if((*right)->ValueNode())
		return *left;
		
	// Something went wrong ...
	g_return_val_if_fail(0, 0);
	return 0;
}

bool CChannel::IsFirstNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_val_if_fail(Node, false);
	
	return GetNodeIterator(Node) == m_Nodes.begin();
}

bool CChannel::IsLastNode(CChannelNode* const Node)
{
	// Sanity checks ...
	g_return_val_if_fail(Node, false);
	
	return GetNodeIterator(Node)+1 == m_Nodes.end();
}

gulong CChannel::GetElementsPerNode()
{
	switch(m_DataType.Value())
		{
			case k3dIChannel::DATA_SCALAR:
			case k3dIChannel::DATA_TIME:
			case k3dIChannel::DATA_ANGLE:
			case k3dIChannel::DATA_DISTANCE:
			case k3dIChannel::DATA_AREA:
			case k3dIChannel::DATA_VOLUME:
			case k3dIChannel::DATA_MASS:
			case k3dIChannel::DATA_FORCE:
			case k3dIChannel::DATA_PRESSURE:
				return 1;
			case k3dIChannel::DATA_COLOR:
			case k3dIChannel::DATA_ROTATIONXYZ:
			case k3dIChannel::DATA_SCALEXYZ:
				return 3;
			default:
				std::cerr << "CChannel::GetElementsPerNode() - Unknown datatype: " << m_DataType.Value() << std::endl;
				g_return_val_if_fail(0, 0);
		}

	return 0;
}

sdpRectangle CChannel::Extents()
{
	sdpRectangle result(0, 1, -DBL_MAX, DBL_MAX);

	// For each node ...
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		{
			result.top = sdpMax(result.top, (*node)->m_Position[1]);
			result.bottom = sdpMin(result.bottom, (*node)->m_Position[1]);
		}

	return result;
}

sdpRectangle CChannel::ValueNodeExtents()
{
	sdpRectangle result(0, 1, -DBL_MAX, DBL_MAX);

	// For each node ...
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		{
			if(0 == (*node)->ValueNode())
				continue;
				
			result.top = sdpMax(result.top, (*node)->m_Position[1]);
			result.bottom = sdpMin(result.bottom, (*node)->m_Position[1]);
		}

	return result;
}

void CChannel::Straighten()
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	// Get the maximum extents of the curve ...
	sdpRectangle extents = ValueNodeExtents();

	// Take a guess at whether the slope of the result should be positive or negative ...
	if(m_Nodes.front()->m_Position[1] > m_Nodes.back()->m_Position[1])
		std::swap(extents.top, extents.bottom);

	const sdpVector2 left = sdpVector2(extents.Left(), extents.Bottom());
	const sdpVector2 right = sdpVector2(extents.Right(), extents.Top());

	if(k3dIChannel::DATA_COLOR == m_DataType.Value())
		{
			gulong count = 0;
			for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
				{
					const gdouble position = count % 3;
					const gdouble odd = (count / 3) % 2;
				
					sdpVector2 coords(sdpLerp(left, right, gdouble(count) / gdouble(m_Nodes.size()-1)));
					coords[1] = sdpLerp(odd, 1.0 - odd, position / 3.0);
					
					(*node)->RawSetPosition(coords);
				
					count++;
				}
		}
	else
		{
			// Space the nodes out evenly to create a linear curve ...
			gulong count = 0;
			for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
				{
					(*node)->RawSetPosition(sdpLerp(left, right, gdouble(count) / gdouble(m_Nodes.size()-1)));
					count++;
				}
		}
}

void CChannel::MirrorX()
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	// Get the curve extents ...
	sdpRectangle extents = ValueNodeExtents();
	const gdouble reflection = extents.Center()[0];

	// Mirror each node's value ...
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		(*node)->UnconstrainedSetPosition(sdpVector2(reflection - ((*node)->m_Position[0] - reflection), (*node)->m_Position[1]));

	// Get everything back in order ...
	if(Document()->StateRecorder()->ChangeSet())
		{
			Document()->StateRecorder()->ChangeSet()->RecordOldState(new k3dValueStateContainer<NodeCollection>(m_Nodes));
			m_NodeCollectionChanged = true;
		}				
		
	std::reverse(m_Nodes.begin(), m_Nodes.end());
}

void CChannel::MirrorY()
{
	// Sanity checks ...
	g_return_if_fail(Editing());

	// Get the curve extents ...
	sdpRectangle extents = ValueNodeExtents();
	const gdouble reflection = extents.Center()[1];

	// Mirror each node's value ...
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		(*node)->RawSetPosition(sdpVector2((*node)->m_Position[0], reflection - ((*node)->m_Position[1] - reflection)));
}

void CChannel::ConstrainNodePosition(CChannelNode* const Node, sdpVector2& Position)
{
	// Sanity checks ...
	g_return_if_fail(Node);

	// Get the node iterator ...
	const NodeIterator node(std::find(m_Nodes.begin(), m_Nodes.end(), Node));
	g_return_if_fail(node != m_Nodes.end());
	const glong nodeindex(node - m_Nodes.begin());

	// Get the preceeding and following range of nodes ...
	NodeCollection preceeding;
	NodeCollection following;
	for(glong i = 1; i < 3; i++)
		{
			if(nodeindex - i >= 0)
				preceeding.insert(preceeding.begin(), m_Nodes[nodeindex-i]);
				
			if(nodeindex + i < glong(m_Nodes.size()))
				following.insert(following.begin(), m_Nodes[nodeindex+i]);
		}

	// Constrain the new position ...
	sdpVector2 coords(sdpClamp(m_MinCoords.Value()[0], Position[0], m_MaxCoords.Value()[0]), sdpClamp(m_MinCoords.Value()[1], Position[1], m_MaxCoords.Value()[1]));

/*
	// Value nodes are constrained to { 0, 1 } when we're a color node ...
	if(SDP_S_OK == (*node)->ValueNode() && k3dIChannel::DATA_COLOR == m_DataType.Value())
		coords[1] = sdpRound(coords[1]);
*/

	// If this is the value node, restrict its' neighboring tangent vectors ...
	if((*node)->ValueNode())
		{
			for(NodeIterator left = preceeding.begin(); left != preceeding.end(); left++)
				{
					if(coords[0] < (*left)->m_Position[0])
						(*left)->SetPosition(sdpVector2(coords[0], (*left)->m_Position[1]));
				}

			for(NodeIterator right = following.begin(); right != following.end(); right++)
				{
					if(coords[0] > (*right)->m_Position[0])
						(*right)->SetPosition(sdpVector2(coords[0], (*right)->m_Position[1]));
				}
		}
	// Otherwise, it's a control node ...
	else
		{
			for(NodeIterator left = preceeding.begin(); left != preceeding.end(); left++)
				{
					if(coords[0] < (*left)->m_Position[0] && (*left)->ValueNode())
						(*left)->SetPosition(sdpVector2(coords[0], (*left)->m_Position[1]));
				}

			for(NodeIterator right = following.begin(); right != following.end(); right++)
				{
					if(coords[0] > (*right)->m_Position[0] && (*right)->ValueNode())
						(*right)->SetPosition(sdpVector2(coords[0], (*right)->m_Position[1]));
				}
		}

	Position = coords;
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CChannel::EditObject()
{
	new CChannelDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIChannel implementation

sdpResult CChannel::SetDatatype(k3dIChannel::DataType Data)
{
	// Sanity checks ...
//	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	k3dSetValue(m_DataType, Data);

	if(k3dIChannel::DATA_COLOR == Data)
		SetRange3(sdpVector3(0.0), sdpVector3(1.0));

	return SDP_S_OK;
}

k3dIChannel::DataType CChannel::Datatype()
{
	return m_DataType.Value();
}

sdpResult CChannel::CurveRange(gdouble* Min, gdouble* Max)
{
	// Sanity checks ...
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	
	*Min = DBL_MAX;
	*Max = -DBL_MAX;

	// For each node ...
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		{
			*Min = sdpMin(*Min, (*node)->m_Position[1]);
			*Max = sdpMax(*Max, (*node)->m_Position[1]);
		}

	return SDP_S_OK;
}

gdouble CChannel::Value1(const gdouble Time, const gdouble MaxError)
{
	// We skip checking the data type here for efficiency ...
	
	// Clamp time to [0.0 1.0]
	gdouble time = sdpClamp(0.0, Time, 1.0);

	// Offset our time frequency and phase ...
	time *= m_Cycles.Value();
	time += m_Phase.Value();
	if(time > 1.0)
		time -= floor(time);

	// Cache the first and last nodes for efficiency ...
	const NodeIterator begin(m_Nodes.begin());
	const NodeIterator end(m_Nodes.end()-1);

	// If we're before the beginning of the first node, that's the value ...
	if(time <= (*begin)->m_Position[0])
		return (*begin)->m_Position[1];

	// If we're past the end of the last node, that's the value ...
	if(time >= (*end)->m_Position[0])
		return (*end)->m_Position[1];

	// If we get this far, we know we have at least one segment (at least 4 nodes), so find the node that contains our time ...
	NodeIterator node;
	for(node = begin + 3; node != end; node += 3)
		if(time <= (*node)->m_Position[0]) break;

	// Mark the beginning and end nodes of this node ...
	const sdpVector2 a((*(node-3))->m_Position);
	const sdpVector2 d((*(node))->m_Position);

	// If the two nodes are adjacent, we're done ...
	const gdouble dx = d[0] - a[0];
	if(0.0 == dx)
		return d[1];

	// Get the two "inbetween" nodes ...
	const sdpVector2 b((*(node-2))->m_Position);
	const sdpVector2 c((*(node-1))->m_Position);

	// Find the bezier interpolation between the two nodes ...
	//
	// Basically, we choose a Bezier curve length u that we think might be close,
	// then calculate the x value at that u.  We compare x to our desired
	// time dt, and, if it exceeds the maximum error, adjust u a portion of the
	// distance and repeat.  We also keep track of the number of iterations and,
	// if they exceed a threshold, abort.
	//
	// If anyone has a better idea, I'd love to hear it!

	gdouble dt = (time - a[0]) / dx;
	gdouble u = dt;
	gdouble x;
	gdouble u2;
	gdouble omu;
	gdouble omu2;

	gulong count = 0;
	while(1)
		{
			u2 = u * u;
			omu = 1.0 - u;
			omu2 = omu * omu;

			x = omu * omu2 * a[0] + 3 * u * omu2 * b[0] + 3 * u2 * omu * c[0] + u * u2 * d[0];
			x = (x - a[0]) / dx;

			if(fabs(x - dt) < MaxError)
				break;

			u += 0.4 * (dt - x);

			if(++count > m_MaxIterations)
				{
					std::cout << "CChannel::GetValue1() couldn't converge on a solution in " << m_MaxIterations << "iterations!" << std::endl;
					break;
				}
		}

	return omu * omu2 * a[1] + 3 * u * omu2 * b[1] + 3 * u2 * omu * c[1] + u * u2 * d[1];
}

sdpResult CChannel::Value2(const gdouble Time, const gdouble MaxError, double Value[2])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CChannel::Value3(const gdouble Time, const gdouble MaxError, double Value[3])
{
	// We skip checking the data type here for efficiency ...
	
	// Clamp time to [0.0 1.0]
	gdouble time = sdpClamp(0.0, Time, 1.0);

	// Offset our time frequency and phase ...
	time *= m_Cycles.Value();
	time += m_Phase.Value();
	if(time > 1.0)
		time -= floor(time);

	// Cache the first and last nodes for efficiency ...
	const NodeIterator begin(m_Nodes.begin());
	const NodeIterator end(m_Nodes.end()-1);

	// If we're before the beginning of the first node, that's the value ...
	if(time <= (*begin)->m_Position[0])
		{
			(*begin)->ValueNode()->Value().CopyArray(Value);
			return SDP_S_OK;
		}

	// If we're past the end of the last node, that's the value ...
	if(time >= (*end)->m_Position[0])
		{
			(*end)->ValueNode()->Value().CopyArray(Value);
			return SDP_S_OK;
		}

	// If we get this far, we know we have at least one segment (at least 4 nodes), so find the node that contains our time ...
	NodeIterator node;
	for(node = begin + 3; node != end; node += 3)
		if(time <= (*node)->m_Position[0]) break;

	// Mark the beginning and end nodes of this node ...
	const sdpVector2 a((*(node-3))->m_Position);
	const sdpVector2 d((*(node))->m_Position);

	const sdpVector3 avalue((*(node-3))->ValueNode()->Value());
	const sdpVector3 dvalue((*(node))->ValueNode()->Value());

	// If the two nodes are adjacent, we're done ...
	const gdouble dx = d[0] - a[0];
	if(0.0 == dx)
		{
			dvalue.CopyArray(Value);
			return SDP_S_OK;
		}

	// Get the mix position ...
	gdouble mix = Value1(Time, MaxError);

	// Adjust the mix for our "sawtooth" pattern ...
	if(a[1] > d[1])
		mix = 1.0 - mix;

	// Linear interpolate the final value ...
	sdpVector3 result(sdpLerp<sdpVector3>(avalue, dvalue, mix));
	result.CopyArray(Value);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditable implementation

void CChannel::PushEditing()
{
	if(!Editing())
		m_NodeCollectionChanged = false;

	k3dEditable::PushEditing();
}

void CChannel::PopEditing()
{
	k3dEditable::PopEditing();
	if(!Editing())
		{
			k3dIStateChangeSet* const changeset = Document()->StateRecorder()->ChangeSet();
			if(changeset)
				{
					if(m_NodeCollectionChanged)
						changeset->RecordNewState(new k3dValueStateContainer<NodeCollection>(m_Nodes));
					
					changeset->RecordAction(new k3dActionNotify(this, k3dIChannel::Event_Changed, 0, 0));
				}
			Notify(k3dIChannel::Event_Changed, 0);
		}


}

///////////////////////////////////////////////////////////////
// k3dIBezierChannel implementation

k3dIEditable* CChannel::Editable()
{
	return this;
}

sdpResult CChannel::SetRange1(const gdouble Min, const gdouble Max)
{
	// Sanity checks ...
//	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Make sure the data type really is one-dimensional ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);

	k3dSetValue(m_MinValue, sdpVector3(Min));
	k3dSetValue(m_MaxValue, sdpVector3(Max));

	k3dSetValue(m_MinCoords, sdpVector2(0.0, Min));
	k3dSetValue(m_MaxCoords, sdpVector2(1.0, Max));

	return SDP_S_OK;
}

sdpResult CChannel::SetRange2(const gdouble Min[2], const gdouble Max[2])
{
	// Sanity checks ...
//	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Make sure the data type really is two-dimensional ...
	g_return_val_if_fail(2 == GetElementsPerNode(), SDP_E_FAIL);

	k3dSetValue(m_MinValue, sdpVector3(Min));
	k3dSetValue(m_MaxValue, sdpVector3(Max));

	k3dSetValue(m_MinCoords, sdpVector2(0.0, 0.0));
	k3dSetValue(m_MaxCoords, sdpVector2(1.0, 1.0));

	return SDP_S_OK;
}

sdpResult CChannel::SetRange3(const gdouble Min[3], const gdouble Max[3])
{
	// Sanity checks ...
//	g_return_val_if_fail(Editing(), SDP_E_FAIL);

	// Make sure the data type really is three-dimensional ...
	g_return_val_if_fail(3 == GetElementsPerNode(), SDP_E_FAIL);

	k3dSetValue(m_MinValue, sdpVector3(Min));
	k3dSetValue(m_MaxValue, sdpVector3(Max));

	k3dSetValue(m_MinCoords, sdpVector2(0.0, 0.0));
	k3dSetValue(m_MaxCoords, sdpVector2(1.0, 1.0));

	return SDP_S_OK;
}

sdpResult CChannel::Range1(gdouble* Min, gdouble* Max)
{
	// Sanity checks ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	
	*Min = m_MinValue.Value()[0];
	*Max = m_MaxValue.Value()[0];
	
	return SDP_S_OK;
}

sdpResult CChannel::Range2(gdouble Min[2], gdouble Max[2])
{
	// Sanity checks ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	
	Min[0] = m_MinValue.Value()[0];
	Min[1] = m_MinValue.Value()[1];
	
	Max[0] = m_MaxValue.Value()[0];
	Max[1] = m_MaxValue.Value()[1];
	
	return SDP_S_OK;
}

sdpResult CChannel::Range3(gdouble Min[3], gdouble Max[3])
{
	// Sanity checks ...
	g_return_val_if_fail(1 == GetElementsPerNode(), SDP_E_FAIL);
	g_return_val_if_fail(Min, SDP_E_FAIL);
	g_return_val_if_fail(Max, SDP_E_FAIL);
	
	Min[0] = m_MinValue.Value()[0];
	Min[1] = m_MinValue.Value()[1];
	Min[2] = m_MinValue.Value()[2];
	
	Max[0] = m_MaxValue.Value()[0];
	Max[1] = m_MaxValue.Value()[1];
	Max[2] = m_MaxValue.Value()[2];
	
	return SDP_S_OK;
}

sdpResult CChannel::SetSegmentCount(const gulong SegmentCount)
{
	// Sanity checks ...
//	g_return_val_if_fail(Editing(), SDP_E_FAIL);

/*
	// If we're recording undos ...
	if(Document()->StateRecorder()->ChangeSet())
		Document()->StateRecorder()->ChangeSet()->RecordOldState(new CChannelNodesStateChange(this));
*/

	ClearNodes();
	for(gulong i = 0; i < SegmentCount+1; i++)
		CreateSegment(sdpVector2(gdouble(i) / gdouble(SegmentCount)));

/*
	// If we're recording undos ...
	if(Document()->StateRecorder()->ChangeSet())
		Document()->StateRecorder()->ChangeSet()->RecordNewState(new CChannelNodesStateChange(this));
*/

	return SDP_S_OK;
}

sdpResult CChannel::VisitNodes(k3dIBezierChannelNodeVisitor* Visitor)
{
	// Sanity checks ...
	g_return_val_if_fail(Visitor, SDP_E_FAIL);
	
	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		if(SDP_S_OK != Visitor->VisitBezierChannelNode(*node))
			return SDP_S_FALSE;
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CChannel::Save(sdpxml::Element& Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	
	g_assert(Dependencies);

	g_return_val_if_fail(sdpSucceeded(CChannelBase::Save(Element, Dependencies)), SDP_E_FAIL);

	sdpxml::Element& nodes = Element.Append(sdpxml::Element("nodes"));

	for(NodeIterator node = m_Nodes.begin(); node != m_Nodes.end(); node++)
		(*node)->Save(nodes, Dependencies);

	return SDP_S_OK;
}

sdpResult CChannel::Load(sdpxml::Document& Document, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_return_val_if_fail(sdpSucceeded(CChannelBase::Load(Document, Element)), SDP_E_FAIL);

	sdpxml::ElementPointer nodes = sdpxml::FindElement(Element, sdpxml::SameName("nodes"));
	if(!nodes)
		return OldFashionedLoad(Element);	// Try to do a backwards-compatible load ...

	// Because we already have one node by default, we need to start from scratch ...
	ClearNodes();

	for(sdpxml::ElementCollection::iterator element = nodes->Children().begin(); element != nodes->Children().end(); element++)
		{
			if(element->Name() == "valuenode")
				{
					CChannelValueNode* const node = new CChannelValueNode(this, sdpVector2(), sdpVector3());
					m_Nodes.push_back(node);
					node->Load(*element);
				}
			else if(element->Name() == "node")
				{
					CChannelNode* const node = new CChannelNode(this, sdpVector2());
					m_Nodes.push_back(node);
					node->Load(*element);
				}
			else
				{
					std::cerr << "CChannel::Load() - Unknown element \"" << element->Name() << "\" will be ignored ... " << std::endl;
				}
		}

	return SDP_S_OK;
}

sdpResult CChannel::OldFashionedLoad(sdpxml::Element& Element)
{
g_assert_not_reached();
/*
	// Sanity checks ...
	g_return_val_if_fail(sdpSucceeded(CChannelBase::Load(Element)), SDP_E_FAIL);

	sdpxml::ElementPointer segmentselement = sdpxml::FindElement(Element, sdpxml::SameName("segments"));
	if(!segmentselement)
		return SDP_S_OK;

	// Nag the user ...
	std::cerr << "CChannel::OldFashionedLoad(): Loading deprecated data format.  Be sure to resave your document to update to the new format!!!" << std::endl;

	// Load all of the nodes from all of the segments ...
	sdpxml::ElementPointerCollection segments = sdpxml::XPathSelect(segmentselement, "segment");
		
	sdpxml::ElementCollection nodes;
	for(sdpxml::ElementCollection::iterator segment = segments.begin(); segment != segments.end(); segment++)
		{
			sdpxml::ElementCollection nodeelements = sdpXmlXPathSelect((*segment), "node");
			std::copy(nodeelements.begin(), nodeelements.end(), std::back_inserter(nodes));
		}

	// Hopefully, we found *something* ...
	g_return_val_if_fail(nodes.size() >= 3, SDP_E_FAIL);

	// Get rid of the (unused) first and last nodes ...
	nodes.erase(nodes.begin());
	nodes.erase(nodes.end()-1);

	// Because we already have one node by default, we need to start from scratch ...
	ClearNodes();

	// Start creating nodes from scratch ...
	gulong index(0);
	for(sdpxml::ElementCollection::iterator node = nodes.begin(); node != nodes.end(); node++)
		{
			sdpVector2 coords;
			sdpXmlGetAttribute(*node, "coords", coords);
		
			if(0 == index % 3)
				{
					sdpVector3 value;
					sdpXmlGetAttribute(*node, "value", value);
					
					m_Nodes.push_back(CreateValueNode(coords, value));
				}
			else
				{
					m_Nodes.push_back(CreateNode(coords));
				}
		
			index++;
		}
*/

	return SDP_S_OK;
}

sdpResult CChannel::InitComplete()
{
	// If for any reason we don't have the correct number of nodes, reset ourselves to a sensible fallback position ...
	if(0 == m_Nodes.size())
		{
			std::cerr << "CChannel::InitComplete() - no nodes loaded!  Supplying a default ..." << std::endl;
			SetSegmentCount(0);
		}
	else if(0 != ((m_Nodes.size() - 1) % 3))
		{
			std::cerr << "CChannel::InitComplete() - incorrect number of nodes loaded!  Resetting to default ..." << std::endl;
			SetSegmentCount(0);
		}

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
// CChannelDialogSetNodeColor

const sdpString CChannelDialogSetNodeColor::control_close = "close";
const sdpString CChannelDialogSetNodeColor::control_colorchanged = "colorchanged";
const sdpString CChannelDialogSetNodeColor::control_colorselection = "colorselection";

CChannelDialogSetNodeColor::CChannelDialogSetNodeColor(sdpIUnknown* const ParentCommandNode, CChannelValueNode* const Node) :
	CChannelDialogSetNodeColorBase(ParentCommandNode, "setnodecolor", 0),
	m_Node(Node)
{
	// Sanity checks ...
	g_assert(m_Node);

	std::istrstream uitemplate( 
		"<gtkml>"
			"<window type=\"toplevel\" show=\"false\">"
				"<vbox homogeneous=\"false\">"
					"<colorselection name=\"colorselection\">"
						"<event signal=\"color-changed\" name=\"colorchanged\"/>"
					"</colorselection>"
					"<hbuttonbox layout=\"end\">"
						"<button name=\"close\">Close"
							"<event signal=\"clicked\" name=\"close\"/>"
						"</button>"
					"</hbuttonbox>"
				"</vbox>"
			"</window>"
		"</gtkml>");

	g_return_if_fail(Load(uitemplate, "CChannelDialogSetNodeColor"));

	AddCommand(control_close, control_close, CONTROL_BUTTON);

	// Initialize the color selection widget ...
	BlockAllEvents();
	ColorSelection(control_colorselection).SetColor(sdpVector3(m_Node->Value()));
	ColorSelection(control_colorselection).GetColor(m_LastColor);
	UnblockAllEvents();

	ColorSelection(control_colorselection).SetUpdatePolicy(GTK_UPDATE_DISCONTINUOUS);

	// We want to be notified if the node goes away ...
	m_Node->AddEventObserver(this, k3dIBezierChannelNode::Event_Deleted, "CChannelDialogSetNodeColor");
	
	// Set our title ...
	RootWindow().SetTitle("Set Node Color");
	
	// Make ourselves visible ...
	Show();
}
	
CChannelDialogSetNodeColor::~CChannelDialogSetNodeColor()
{
	// Cancel event notifications
	m_Node->RemoveEventObserver(this);
}

sdpResult CChannelDialogSetNodeColor::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	const sdpString name(Name);
	const sdpString arguments(Arguments);
	
	if(name == control_colorchanged)
		{
			const sdpVector3 color(sdpFromString<sdpVector3>(arguments, sdpVector3(0, 0, 0)));
			
			ColorSelection(control_colorselection).InteractiveSetColor(color, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);
			OnColorChanged(color);
					
			return SDP_S_OK;
		}
	
	return CChannelDialogSetNodeColorBase::Command(Name, Arguments, Result);
}
	
sdpResult CChannelDialogSetNodeColor::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	if(event == k3dIBezierChannelNode::Event_Deleted)
		{
			delete this;
			return SDP_S_OK;
		}
	
	return CChannelDialogSetNodeColorBase::ObserveEvent(Source, Event, EventData, ObserverData);
}
	
void CChannelDialogSetNodeColor::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == control_close)
		OnClose();
	else if(Event->Name() == control_colorchanged)
		RawOnColorChanged();
	else
		CChannelDialogSetNodeColorBase::OnEvent(Event);
}

void CChannelDialogSetNodeColor::OnClose()
{
	// Record this as a user action ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_close, 0);
	CChannelDialogSetNodeColorBase::OnClose();
}

void CChannelDialogSetNodeColor::RawOnColorChanged()
{
	sdpVector3 newcolor;
	ColorSelection(control_colorselection).GetColor(newcolor);
	
	if(newcolor != m_LastColor)
		{
			m_LastColor = newcolor;
			OnColorChanged(newcolor);
		}
}

void CChannelDialogSetNodeColor::OnColorChanged(const sdpVector3 NewColor)
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_colorchanged, sdpToString(NewColor));
	k3dRecordStateChangeSet undo(m_Node->Channel()->Document(), "Set Color Value");
	k3d::Editing editing(m_Node->Channel());
	
	m_Node->SetValue3(NewColor);
}

/////////////////////////////////////////////////////////////////////////////
// CChannelDialog

sdpVector3 CChannelDialog::m_BackgroundColor(0.8, 0.8, 0.8);
sdpVector3 CChannelDialog::m_NormalColor(0.0, 0.0, 0.0);
sdpVector3 CChannelDialog::m_ValueNodeColor(1.0, 1.0, 0.0);
sdpVector3 CChannelDialog::m_SelectedColor(1.0, 1.0, 1.0);
sdpVector3 CChannelDialog::m_TangentColor(0.0, 0.0, 1.0);
sdpVector3 CChannelDialog::m_SelectorColor(0.215, 0.215, 0.8);
sdpVector3 CChannelDialog::m_CursorColor(1.0, 0.0, 0.0);
sdpVector3 CChannelDialog::m_TimeColor(0.0, 1.0, 0.0);
sdpRectangle CChannelDialog::m_Frustum(-1, 1, 1, -1);

const sdpString CChannelDialog::dialogname = "properties";

const sdpString CChannelDialog::control_cycles = "cycles";
const sdpString CChannelDialog::control_phase = "phase";
const sdpString CChannelDialog::control_straightenchannel = "straightenchannel";
const sdpString CChannelDialog::control_mirrorxchannel = "mirrorxchannel";
const sdpString CChannelDialog::control_mirrorychannel = "mirrorychannel";
const sdpString CChannelDialog::control_straightenselected = "straightenselected";
const sdpString CChannelDialog::control_deleteselected = "deleteselected";
const sdpString CChannelDialog::control_resetview = "resetview";

const sdpString CChannelDialog::control_mousemove = "mousemove";
const sdpString CChannelDialog::control_lbuttondown = "lbuttondown";
const sdpString CChannelDialog::control_lbuttonclick = "lbuttonclick";
const sdpString CChannelDialog::control_lbuttondoubleclick = "lbuttondoubleclick";
const sdpString CChannelDialog::control_lbuttonstartdrag = "lbuttonstartdrag";
const sdpString CChannelDialog::control_lbuttondrag = "lbuttondrag";
const sdpString CChannelDialog::control_lbuttonenddrag = "lbuttonenddrag";
const sdpString CChannelDialog::control_mbuttonclick = "mbuttonclick";
const sdpString CChannelDialog::control_mbuttondoubleclick = "mbuttondoubleclick";
const sdpString CChannelDialog::control_mbuttonstartdrag = "mbuttonstartdrag";
const sdpString CChannelDialog::control_mbuttondrag = "mbuttondrag";
const sdpString CChannelDialog::control_mbuttonenddrag = "mbuttonenddrag";
const sdpString CChannelDialog::control_rbuttonclick = "rbuttonclick";
const sdpString CChannelDialog::control_rbuttondoubleclick = "rbuttondoubleclick";
const sdpString CChannelDialog::control_rbuttonstartdrag = "rbuttonstartdrag";
const sdpString CChannelDialog::control_rbuttondrag = "rbuttondrag";
const sdpString CChannelDialog::control_rbuttonenddrag = "rbuttonenddrag";
const sdpString CChannelDialog::control_lrbuttonstartdrag = "lrbuttonstartdrag";
const sdpString CChannelDialog::control_lrbuttondrag = "lrbuttondrag";
const sdpString CChannelDialog::control_lrbuttonenddrag = "lrbuttonenddrag";

CChannelDialog::CChannelDialog(CChannel* const Object) :
	CChannelDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "channel.gtkml"));

	// Reset our view so it's properly scaled & centered ...
	ResetView();

	AddCommand(control_straightenchannel, control_straightenchannel, CONTROL_BUTTON);
	AddCommand(control_mirrorxchannel, control_mirrorxchannel, CONTROL_BUTTON);
	AddCommand(control_mirrorychannel, control_mirrorychannel, CONTROL_BUTTON);
	AddCommand(control_straightenselected, control_straightenselected, CONTROL_BUTTON);
	AddCommand(control_deleteselected, control_deleteselected, CONTROL_BUTTON);
	AddCommand(control_resetview, control_resetview, CONTROL_BUTTON);

	// Attach spin buttons ...
	Connectk3dSpinButton(control_phase, m_Object->m_Phase, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Channel Phase", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Create the OpenGL widgets ...
	sdpGtkContainer curvecontainer(Container("curve"));
	if(!m_CurveWidget.Create(curvecontainer, true, true, 8, 8, 8, 0))
		if(!m_CurveWidget.Create(curvecontainer, true, true, 5, 5, 5, 0))
			if(!m_CurveWidget.Create(curvecontainer, false, false, 4, 4, 4, 0))
				std::cerr << "CChannelDialog::CChannelDialog(): Could not find useable OpenGL visual" << std::endl;

	sdpGtkContainer colorcontainer(Container("color"));
	if(!m_ColorWidget.Create(colorcontainer, true, true, 8, 8, 8, 0))
		if(!m_ColorWidget.Create(colorcontainer, true, true, 5, 5, 5, 0))
			if(!m_ColorWidget.Create(colorcontainer, false, false, 4, 4, 4, 0))
				std::cerr << "CChannelDialog::CChannelDialog(): Could not find useable OpenGL visual" << std::endl;

	// Setup some events ...
	if(m_CurveWidget.Initialized())
		{
			MapEvent("configure-event", "configurecurve", false, m_CurveWidget, true);
			MapEvent("expose-event", "exposecurve", false, m_CurveWidget, true);
			MapEvent("motion-notify-event", "mousemovecurve", false, m_CurveWidget, true);
			MapEvent("button-press-event", "buttondowncurve", false, m_CurveWidget, true);
			MapEvent("button-release-event", "buttonupcurve", false, m_CurveWidget, true);
		}
	
	if(m_ColorWidget.Initialized())
		{
			MapEvent("configure-event", "configurecolor", false, m_ColorWidget, true);
			MapEvent("expose-event", "exposecolor", false, m_ColorWidget, true);
		}

	// We want to be notified when the time changes ...
	k3dIEventSource* time = dynamic_cast<k3dIEventSource*>(m_Object->Document()->Time());
	time->AddEventObserver(this, k3dITime::Event_ViewTimeChanged, "CChannelDialog");

	m_Object->AddEventObserver(this, k3dIChannel::Event_Changed, "CChannelDialog");

	// Only make the color preview bar visible if the channel is, indeed, a color channel ...
	if(k3dIChannel::DATA_COLOR == m_Object->Datatype())
		Widget("colorframe").Show();
	else
		Widget("colorframe").Hide();

	// Initialize the cursor ...
	SetCursorX(0.0);

	CallUpdateControls();
	Show();
	
// The Win32 port of GTK+ has some kind of refresh problem, so force one to get us going ...
#ifdef SDPWIN32
	RootWidget().QueueClear();
	sdpGtkHandlePendingEvents();
#endif
}

CChannelDialog::~CChannelDialog()
{
	// We don't want any more events ...
	k3dIEventSource* time = dynamic_cast<k3dIEventSource*>(m_Object->Document()->Time());
	time->RemoveEventObserver(this);
}

static void MouseCommand(GtkWidget* Widget, k3dIApplication::TutorialMouseAction Action, GdkModifierType Modifiers, sdpVector2 CurrentMouse)
{
	// Sanity checks ...
	g_return_if_fail(Widget);
	
	k3dApplication()->TutorialMouseDisplay("", Action, Modifiers);
	sdpGtkWarpPointer(Widget, gint(CurrentMouse[0]), gint(CurrentMouse[1]));
	sdpGtkHandlePendingEvents();
	sdpGtkSleep(20);
}

// k3dICommandNode implementation ...
sdpResult CChannelDialog::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	// Extract state information from the arguments ...
	std::istrstream arguments(Arguments);
	gint temp = 0;
	sdpVector2 currentmouse;
	sdpVector2 lastmouse;
	sdpVector2 startmouse;
	arguments >> temp >> currentmouse >> lastmouse >> startmouse;
	GdkModifierType modifiers = GdkModifierType(temp);

	// Convert the mouse from world coordinates to screen ...
	currentmouse = Screen(currentmouse);
	lastmouse = Screen(lastmouse);
	startmouse = Screen(startmouse);

	k3dIApplication* const application = k3dApplication();

	const gdouble tutorialspeed = k3dOptionsGetTutorialSpeed(application->Options());	
	const gulong delay = static_cast<gulong>(1200 / tutorialspeed);
	
	GtkWidget* widget = GTK_WIDGET(m_CurveWidget.Object());

	sdpString name(Name);
	
/*
	if(name == control_mousemove)
		{
			OnMouseMove(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MB_NONE, modifiers, currentmouse);
		}
	else*/ if(name == control_lbuttondown)
		{
			m_CurveWidget.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonDown(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
/*
	else if(name == control_lbuttonclick)
		{
			m_Viewport.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
*/
	else if(name == control_lbuttondoubleclick)
		{
			m_CurveWidget.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonDoubleClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_DOUBLE_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_lbuttonstartdrag)
		{
			m_CurveWidget.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lbuttondrag)
		{
			OnLButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lbuttonenddrag)
		{
			OnLButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
/*
	else if(name == control_mbuttonclick)
		{
			m_Viewport.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnMButtonClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_mbuttondoubleclick)
		{
			m_Viewport.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnMButtonDoubleClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MMB_DOUBLE_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_mbuttonstartdrag)
		{
			m_Viewport.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnMButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::MMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_mbuttondrag)
		{
			OnMButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::MMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_mbuttonenddrag)
		{
			OnMButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::MMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
*/
	else if(name == control_rbuttonclick)
		{
			m_CurveWidget.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnRButtonClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::RMB_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
/*
	else if(name == control_rbuttondoubleclick)
		{
			m_Viewport.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnRButtonDoubleClick(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::RMB_DOUBLE_CLICK, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
*/
	else if(name == control_rbuttonstartdrag)
		{
			m_CurveWidget.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnRButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::RMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_rbuttondrag)
		{
			OnRButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::RMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_rbuttonenddrag)
		{
			OnRButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::RMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else if(name == control_lrbuttonstartdrag)
		{
			m_CurveWidget.InteractiveWarpPointer(gint(currentmouse[0]), gint(currentmouse[1]), tutorialspeed, true, false);
			OnLRButtonStartDrag(modifiers, currentmouse);
			MouseCommand(widget, k3dIApplication::LMBRMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lrbuttondrag)
		{
			OnLRButtonDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMBRMB_DRAG, modifiers, currentmouse);
		}
	else if(name == control_lrbuttonenddrag)
		{
			OnLRButtonEndDrag(modifiers, currentmouse, lastmouse, startmouse);
			MouseCommand(widget, k3dIApplication::LMBRMB_DRAG, modifiers, currentmouse);
			sdpGtkSleep(delay);
		}
	else
		{
			return CChannelDialogBase::Command(Name, Arguments, Result);
		}

	return SDP_S_OK;
}

sdpResult CChannelDialog::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIChannel::Event_Changed)
		{
			// Update the current selection (it may refer to nodes that have gone away) ...
			SelectedNodeCollection a(m_Object->m_Nodes.begin(), m_Object->m_Nodes.end());
			SelectedNodeCollection b(m_Selected);
			m_Selected.clear();
			
			std::set_intersection(a.begin(), a.end(), b.begin(), b.end(), std::inserter(m_Selected, m_Selected.end()));
		
			RedrawAll();
			UpdateControls();
			return SDP_S_OK;
		}
	else if(event == k3dITime::Event_ViewTimeChanged)
		{
			RedrawAll();
			UpdateControls();
			return SDP_S_OK;
		}
			
	return CChannelDialogBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void CChannelDialog::UpdateControls()
{
	CChannelDialogBase::UpdateControls();
}

sdpVector2 CChannelDialog::World(sdpVector2 ScreenPosition)
{
	// Get the dimensions of the curve window ...
	gulong width = m_CurveWidget.Width();
	gulong height = m_CurveWidget.Height();

	sdpVector2 result;

	// Convert pixel coordinates to "screen" coordinates ...
	// (Make sure we handle (possible) minimized / hidden widget)
	if(width)
		result[0] = sdpLerp(m_Frustum.Left(), m_Frustum.Right(), ScreenPosition[0] / width);

	if(height)
		result[1] = sdpLerp(m_Frustum.Top(), m_Frustum.Bottom(), ScreenPosition[1] / height);

	// Scale the results to account for pan/zoom ...
	result = sdpProduct(result, m_Scale) + m_Origin;
	result[1] *= m_VScale;

	return result;
}

sdpVector2 CChannelDialog::Screen(sdpVector2 WorldPosition)
{
	// Get the dimensions of the curve window ...
	gulong width = m_CurveWidget.Width();
	gulong height = m_CurveWidget.Height();

	// Convert world coordinates to "screen" coordinates ...
	WorldPosition[1] *= (1.0 / m_VScale);
	sdpVector2 result = sdpProduct((WorldPosition - m_Origin), sdpVector2(1.0 / m_Scale[0], 1.0 / m_Scale[1]));

	// Convert to pixel coordinates ...
	result[0] = width * (result[0] - m_Frustum.Left()) / m_Frustum.Width();
	result[1] = height * (1 - ((result[1] - m_Frustum.Bottom()) / m_Frustum.Height()));

	return result;
}

CChannelNode* CChannelDialog::HitTest(sdpVector2 Mouse)
{
	const static gdouble tolerance = 6.0;

	// For each node ...
	for(NodeIterator node = m_Object->m_Nodes.begin(); node != m_Object->m_Nodes.end(); node++)
		{
			// Check to see if we hit anything ...
			sdpVector2 coords(Screen((*node)->m_Position));

			if((fabs(Mouse[0] - coords[0]) < tolerance) && (fabs(Mouse[1] - coords[1]) < tolerance))
				return *node;
		}

	return 0;
}

void CChannelDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "configurecurve" || Event->Name() == "exposecurve")
		DrawCurve();
	else if(Event->Name() == "configurecolor" || Event->Name() == "exposecolor")
		DrawColor();
	else if(Event->Name() == "mousemovecurve")
		RawMouseMove(Event);
	else if(Event->Name() == "buttondowncurve")
		RawButtonDown(Event);
	else if(Event->Name() == "buttonupcurve")
		RawButtonUp(Event);
	else if(Event->Name() == "vscale")
		OnVScale();
	else if(Event->Name() == control_resetview)
		OnResetView();
	else if(Event->Name() == control_straightenchannel)
		OnChannelStraighten();
	else if(Event->Name() == control_mirrorxchannel)
		OnChannelMirrorX();
	else if(Event->Name() == control_mirrorychannel)
		OnChannelMirrorY();
	else if(Event->Name() == control_straightenselected)
		OnStraightenSelected();
	else if(Event->Name() == control_deleteselected)
		OnDeleteSelected();
	else
		CChannelDialogBase::OnEvent(Event);
}

void CChannelDialog::RecordEvent(SDPCSTRING Command, GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	// Sanity checks ...
	g_assert_string(Command);

	// Setup mouse coordinates ...
	sdpVector2 currentmouse(World(CurrentMouse));

	// Record the event ...		
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, Command, sdpToString(gint(Modifiers)) + " " + sdpToString(currentmouse));
}

void CChannelDialog::RecordDragEvent(SDPCSTRING Command, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert_string(Command);

	// Setup mouse coordinates ...
	sdpVector2 currentmouse(World(CurrentMouse));
	sdpVector2 lastmouse(World(LastMouse));
	sdpVector2 startmouse(World(StartMouse));

	// Record the event ...		
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, Command, sdpToString(gint(Modifiers)) + " " + sdpToString(currentmouse) + " " + sdpToString(lastmouse) + " " + sdpToString(startmouse));
}

void CChannelDialog::OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	RecordEvent(control_lbuttondown, Modifiers, CurrentMouse);

	CChannelNode* const node = HitTest(CurrentMouse);
	if(node)
		{
			if(k3dMouseEventObserver::ShiftDown(Modifiers))
				{
					m_Selected.insert(node);
					if(node->ValueNode())
						{
							if(!m_Object->IsFirstNode(node))
								m_Selected.insert(m_Object->GetPrevNode(node));
								
							if(!m_Object->IsLastNode(node))
								m_Selected.insert(m_Object->GetNextNode(node));
						}
				}
			else
				{
					SetCursorXY(node->m_Position[0], node->m_Position[1]);
					m_Selected.insert(node);
				}
		}
	else
		{
			SetCursorX(World(CurrentMouse)[0]);
		}

	RedrawAll();
}

void CChannelDialog::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
}

void CChannelDialog::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	RecordEvent(control_lbuttondoubleclick, Modifiers, CurrentMouse);

	// Convert mouse coordinates to world coordinates ...
	const sdpVector2 worldcoords = World(CurrentMouse);

	// See if the user clicked on an existing node, first ...
	CChannelNode* const node = HitTest(CurrentMouse);
	if(node)
		{
			SetCursorXY(worldcoords[0], worldcoords[1]);
			EditNode(*node);
		}
	// Otherwise, create a new node ...
	else
		{
			k3dRecordStateChangeSet undo(m_Object->Document(), "Create Node(s)");
			k3d::Editing editing(m_Object->Editable());

			// Create the node ...			
			CChannelNode* const newnode = m_Object->CreateSegment(worldcoords);
			g_return_if_fail(newnode);

			// Select it ...
			m_Selected.insert(newnode);

			// Position the cursor ...
			SetCursorXY(worldcoords[0], worldcoords[1]);

			// Edit that baby!
			EditNode(*newnode);
		}
}

void CChannelDialog::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	RecordEvent(control_lbuttonstartdrag, Modifiers, CurrentMouse);
	
	if(m_Selected.size())
		{
			k3dStartStateChangeSet(m_Object->Document());
			m_Object->PushEditing();
		}
}

void CChannelDialog::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	RecordDragEvent(control_lbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);

	// Update the cursor position ...
	SetCursorX(World(CurrentMouse)[0]);

	// Calculate the amount of mouse movement in world coordinates ...
	const sdpVector2 delta(World(CurrentMouse) - World(LastMouse));
	
	// For each node ...
	for(SelectedNodeIterator node = m_Selected.begin(); node != m_Selected.end(); node++)
		{
			// Move the node ...
			(*node)->RawSetPosition((*node)->m_Position + delta);
			
			// See if the user's holding down the shift key ...
			if(!k3dMouseEventObserver::ShiftDown(Modifiers))
				continue;
				
			// If it's a control point and shift is down, maintain our slope ...
			if(!(*node)->ValueNode())
				{
					CChannelNode* valuenode = m_Object->GetValueNode(*node);
					g_return_if_fail(valuenode);
					
					const sdpVector2 controldelta((*node)->m_Position - valuenode->m_Position);
					
					CChannelNode* othernode = m_Object->GetOppositeControlNode(*node);
					g_return_if_fail(othernode);
					
					CChannelNode* othervaluenode = m_Object->GetValueNode(othernode);
					g_return_if_fail(othervaluenode);
					
					othernode->RawSetPosition(othervaluenode->m_Position - controldelta);
				}
			// If it's a value node and shift is down, constrain the first and last nodes (good for looping animations) ...
			else
				{
					CChannelNode* const firstnode = m_Object->m_Nodes.front();
					CChannelNode* const lastnode = m_Object->m_Nodes.back();
				
					if(*node == firstnode)
						lastnode->RawSetPosition(sdpVector2(lastnode->m_Position[0], firstnode->m_Position[1]));
					else if(*node == lastnode)
						firstnode->RawSetPosition(sdpVector2(firstnode->m_Position[0], lastnode->m_Position[1]));
				}
		}
		
	RedrawAll();
}

void CChannelDialog::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	RecordDragEvent(control_lbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);

	if(m_Selected.size())
		{
			m_Object->PopEditing();
			k3dFinishStateChangeSet(m_Object->Document(), "Move node(s)");
		}
}

void CChannelDialog::OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	RecordEvent(control_rbuttonclick, Modifiers, CurrentMouse);
	DeselectAll();
	RedrawAll();
}

void CChannelDialog::OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	RecordEvent(control_rbuttonstartdrag, Modifiers, CurrentMouse);
}

void CChannelDialog::OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	RecordDragEvent(control_rbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);

	sdpVector2 delta = World(CurrentMouse) - World(LastMouse);
	m_Origin += delta;
	RedrawAll();
}

void CChannelDialog::OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	RecordDragEvent(control_rbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CChannelDialog::OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	RecordEvent(control_lrbuttonstartdrag, Modifiers, CurrentMouse);
}

void CChannelDialog::OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	RecordDragEvent(control_lrbuttondrag, Modifiers, CurrentMouse, LastMouse, StartMouse);

	gdouble zoomfactor = CurrentMouse[1] > LastMouse[1] ? pow(1.005, CurrentMouse[1] - LastMouse[1]) : pow(0.995, LastMouse[1] - CurrentMouse[1]);
	m_Scale *= zoomfactor;	
	RedrawAll();
}

void CChannelDialog::OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	RecordDragEvent(control_lrbuttonenddrag, Modifiers, CurrentMouse, LastMouse, StartMouse);
}

void CChannelDialog::OnVScale()
{
	m_VScale = pow(10.0, Adjustment("vscale").Value());
	RedrawAll();
}

void CChannelDialog::OnResetView()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_resetview, 0);
	ResetView();
	RedrawAll();
}

void CChannelDialog::OnChannelStraighten()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_straightenchannel, 0);
	k3dRecordStateChangeSet undo(m_Object->Document(), "Straighten Channel");
	k3d::Editing editing(m_Object);

	m_Object->Straighten();
}

void CChannelDialog::OnChannelMirrorX()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_mirrorxchannel, 0);
	k3dRecordStateChangeSet undo(m_Object->Document(), "Channel Mirror X");
	k3d::Editing editing(m_Object);
	
	m_Object->MirrorX();
}

void CChannelDialog::OnChannelMirrorY()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_mirrorychannel, 0);
	k3dRecordStateChangeSet undo(m_Object->Document(), "Channel Mirror Y");
	k3d::Editing editing(m_Object);

	m_Object->MirrorY();
}

void CChannelDialog::OnStraightenSelected()
{
	k3dRecordStateChangeSet undo(m_Object->Document(), "Straighten selected segment(s)");
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_straightenselected, 0);
	k3d::Editing editing(m_Object);

	for(SelectedNodeIterator node = m_Selected.begin(); node != m_Selected.end(); node++)
		{
			if((*node)->ValueNode())
				continue;
				
			CChannelNode* const a = m_Object->GetValueNode(*node);
			CChannelNode* const b = (*node);
			CChannelNode* const c = m_Object->GetNeighborControlNode(b);
			CChannelNode* const d = m_Object->GetValueNode(c);

			sdpVector2 aposition;
			a->Position(aposition);
			
			sdpVector2 dposition;
			d->Position(dposition);

			b->RawSetPosition(sdpLerp(aposition, dposition, 0.3));
			c->RawSetPosition(sdpLerp(aposition, dposition, 0.7));
		}
}

void CChannelDialog::OnDeleteSelected()
{
	// If we only have one node to begin with, we're done ...
	if(1 == m_Object->m_Nodes.size())
		{
			k3dApplication()->Error("Can't delete the last node!", "Delete Selected:");
			return;
		}

	k3dRecordStateChangeSet undo(m_Object->Document(), "Delete selected segment(s)");
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_deleteselected, 0);
	k3d::Editing editing(m_Object);
	
	for(SelectedNodeIterator node = m_Selected.begin(); node != m_Selected.end(); node++)
		m_Object->DeleteNode(*node);
}

void CChannelDialog::EditNode(CChannelNode& Node)
{
	// If it isn't a value node, we're done ...
	g_return_if_fail(Node.ValueNode());
		
	// If it's a color ...
	if(k3dIChannel::DATA_COLOR == m_Object->Datatype())
		new CChannelDialogSetNodeColor(m_Object, Node.ValueNode());
}

void CChannelDialog::ResetView()
{
	// Get the maximum extents of the curve ...
	sdpRectangle extents = m_Object->Extents();

	// Make sure we can always see the X axis ...
	if(extents.bottom > 0)
		extents.bottom = 0;
		
	if(extents.top < 0)
		extents.top = 0;

	// Make sure we can handle curves that are straight across the origin ...
	if(extents.top == extents.bottom)
		extents.top += 1.0;

	// Initialize the scale & origin ...
	m_VScale = 1.0;
	m_Scale = sdpVector2(extents.Width(), extents.Height()) * 0.55;
	m_Origin = extents.Center();
}

void CChannelDialog::DrawCurve()
{
//std::cout << "CChannelDialog::DrawCurve()" << std::endl;

	// If we're not completely initialized, we're done ...
	if(!m_CurveWidget.Initialized())
		return;

	m_CurveWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_CurveWidget.Width();
	gulong height = m_CurveWidget.Height();
	glViewport(0, 0, width, height);

	// Clear the background ...
	glClearColor(m_BackgroundColor[0], m_BackgroundColor[1], m_BackgroundColor[2], 1.0);
	glClear(GL_COLOR_BUFFER_BIT);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(m_Frustum.Left(), m_Frustum.Right(), m_Frustum.Bottom(), m_Frustum.Top(), -1, 1);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScaled(1.0 / m_Scale[0], 1.0 / m_Scale[1], 1.0);
	glTranslated(-m_Origin[0], -m_Origin[1], 0.0);
	glScaled(1, 1.0 / m_VScale, 1.0);

	// Setup drawing options ...
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);


	// Make it happen ...
	DrawGrid();
	DrawCursor();
	DrawTime();
	DrawPreview();
	DrawControlCurve();
	DrawTangents();
	DrawNodes();

	// Cleanup ...
	m_CurveWidget.SwapBuffers();
	m_CurveWidget.End();
}

void CChannelDialog::DrawGrid()
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glColor3dv(m_BackgroundColor * 0.8);

	// Draw the grid ...
	const static int xdivisions = 10;
	const static int ydivisions = 10;

	glBegin(GL_LINES);

	for(int i = 0; i <= xdivisions; i++)
		{
			glVertex2d(gdouble(i) / gdouble(xdivisions), 0.0);
			glVertex2d(gdouble(i) / gdouble(xdivisions), 1.0);
		}

	for(int j = 0; j <= ydivisions; j++)
		{
			glVertex2d(0.0, gdouble(j) / gdouble(ydivisions));
			glVertex2d(1.0, gdouble(j) / gdouble(ydivisions));
		}

	glEnd();

	gdouble left = World(sdpVector2(0, 0))[0];
	gdouble top = World(sdpVector2(0, 0))[1];
	gdouble right = World(sdpVector2(0, m_CurveWidget.Width()))[0];
	gdouble bottom = World(sdpVector2(0, m_CurveWidget.Height()))[1];

	// Draw the origin / axes ...
	glColor3dv(m_BackgroundColor * 0.6);
	glLineWidth(2.0f);

	glBegin(GL_LINES);

	glVertex2d(0.0, top);
	glVertex2d(0.0, bottom);

	glVertex2d(left, 0.0);
	glVertex2d(right, 0.0);

	glEnd();

	glPopAttrib();
}

void CChannelDialog::DrawPreview()
{
	// Don't do it if we don't need to!
	if((m_Object->m_Cycles.Value() == 1.0) && (m_Object->m_Phase.Value() == 0.0))
		return;

	glColor3dv(m_BackgroundColor * 0.7);

	glBegin(GL_LINE_STRIP);
						
	// Sample the curve ...
	const static gulong stepcount = 200;
	const static gdouble error = 0.5 / stepcount;
	
	for(gulong i = 0; i < stepcount; i++)
		{
			gdouble delta = gdouble(i) / gdouble(stepcount-1);
			glVertex2d(delta, m_Object->Value1(delta, error));
		}
		
	glEnd();
}

void CChannelDialog::DrawControlCurve()
{
	// Setup Bezier curve drawing
	const static int stepsize = 50;

	GLdouble controlpoints[4][3];
	memset(controlpoints, 0, sizeof(controlpoints));

	glEnable(GL_MAP1_VERTEX_3);

	glColor3dv(m_NormalColor);

	// Draw initial straight line node before first node
	NodeIterator node(m_Object->m_Nodes.begin());
	g_return_if_fail(node != m_Object->m_Nodes.end());
	
	glBegin(GL_LINES);
	glVertex2d(-10000.0, (*node)->m_Position[1]);
	glVertex2d((*node)->m_Position[0], (*node)->m_Position[1]);
	glEnd();

	// Draw straight line node after last node
	node = m_Object->m_Nodes.end() - 1;
	glBegin(GL_LINES);
	glVertex2d((*node)->m_Position[0], (*node)->m_Position[1]);
	glVertex2d(10000.0, (*node)->m_Position[1]);
	glEnd();

	// Draw curve segments
	for(node = m_Object->m_Nodes.begin(); node != m_Object->m_Nodes.end(); )
		{
			controlpoints[0][0] = (*node)->m_Position[0];
			controlpoints[0][1] = (*node)->m_Position[1];
			node++;
			if(node == m_Object->m_Nodes.end())
				return;
			
			controlpoints[1][0] = (*node)->m_Position[0];
			controlpoints[1][1] = (*node)->m_Position[1];
			node++;
			g_return_if_fail(node != m_Object->m_Nodes.end());
			
			controlpoints[2][0] = (*node)->m_Position[0];
			controlpoints[2][1] = (*node)->m_Position[1];
			node++;
			g_return_if_fail(node != m_Object->m_Nodes.end());
			
			controlpoints[3][0] = (*node)->m_Position[0];
			controlpoints[3][1] = (*node)->m_Position[1];
			
			glMap1d(GL_MAP1_VERTEX_3, 0.0, 1.0, 3, 4, &controlpoints[0][0]);

			glBegin(GL_LINE_STRIP);
			for(int j = 0; j <= stepsize; j++)
				glEvalCoord1d(GLdouble(j) / GLdouble(stepsize));
			glEnd();
		}
}

void CChannelDialog::DrawTangents()
{
	glColor3dv(m_TangentColor);
	glBegin(GL_LINES);

	gulong index = 0;
	for(NodeIterator node = m_Object->m_Nodes.begin(); node != m_Object->m_Nodes.end(); node++)
		{
			if(1 == index % 3)
				{
					sdpVector2 a((*(node-1))->m_Position);
					sdpVector2 b((*node)->m_Position);
					glVertex2dv(a);
					glVertex2dv(b);
				}
			else if(2 == index % 3)
				{
					sdpVector2 a((*node)->m_Position);
					sdpVector2 b((*(node+1))->m_Position);
					glVertex2dv(a);
					glVertex2dv(b);
				}
				
			index++;
		}

	glEnd();
}

void CChannelDialog::DrawNodes()
{
	// Set ourselves up to draw some big round points ...
	glEnable(GL_POINT_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);

	for(NodeIterator node = m_Object->m_Nodes.begin(); node != m_Object->m_Nodes.end(); node++)
		DrawNode(*node);
}

void CChannelDialog::DrawNode(CChannelNode* Node)
{
	// Sanity checks ...
	g_assert(Node);

	glPointSize(Node->ValueNode() ? 6.0 : 5.0);
	glColor3dv(m_Selected.count(Node) ? m_SelectedColor : (Node->ValueNode() ? m_ValueNodeColor : m_NormalColor));

	sdpVector2 coords(Node->m_Position);
	
	glBegin(GL_POINTS);
	glVertex2dv(coords);
	glEnd();
}

void CChannelDialog::DrawCursor()
{
	gdouble top = World(sdpVector2(0, 0))[1];
	gdouble bottom = World(sdpVector2(0, m_CurveWidget.Height()))[1];

	glColor3dv(m_CursorColor);
	glBegin(GL_LINES);

	// Vertical cursor ...
	glVertex2d(m_Cursor[0], top);
	glVertex2d(m_Cursor[0], bottom);

	glEnd();
}

void CChannelDialog::DrawTime()
{
	gdouble top = World(sdpVector2(0, 0))[1];
	gdouble bottom = World(sdpVector2(0, m_CurveWidget.Height()))[1];

	glColor3dv(m_TimeColor);
	glBegin(GL_LINES);

	// Vertical cursor ...
	glVertex2d(m_Object->m_CurrentTime, top);
	glVertex2d(m_Object->m_CurrentTime, bottom);

	glEnd();
}

void CChannelDialog::DrawColor()
{
	// If we're not completely initialized, we're done ...
	if(!m_ColorWidget.Initialized())
		return;

	m_ColorWidget.Begin();

	// Get the widget sizes ...
	gulong width = m_ColorWidget.Width();
	gulong height = m_ColorWidget.Height();
	glViewport(0, 0, width, height);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(m_Frustum.Left(), m_Frustum.Right(), -1, 1, -1, 1);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScaled(1.0 / m_Scale[0], 1.0, 1.0);
	glTranslated(-m_Origin[0], 0.0, 0.0);

	// Setup drawing options ...
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	// Make it happen ...
	glBegin(GL_QUADS);
	
	sdpVector3 color1;
	gdouble error = 0.01;
	gulong granularity = 2;

	gdouble x2;
	gdouble x1 = World(sdpVector2(0, 0))[0];		
	for(gulong i = 0; i < width+granularity; i += granularity)
		{
			x2 = World(sdpVector2(gdouble(i+1), 0.0))[0];		
		
			m_Object->Value3(x2, error, color1);

			glColor3dv(color1);
			glVertex2d(x1, -1);
			glVertex2d(x1, 1);
			glVertex2d(x2, 1);
			glVertex2d(x2, -1);
			
			x1 = x2;
		}

	glEnd();

	// Cleanup ...
	m_ColorWidget.SwapBuffers();
	m_ColorWidget.End();
}

void CChannelDialog::DeselectAll()
{
	m_Selected.clear();
}

void CChannelDialog::RedrawAll()
{
	Widget("curve").QueueDraw();
	Widget("color").QueueDraw();
}

void CChannelDialog::SetCursorX(const gdouble X)
{
	// Constrain the cursor position
	m_Cursor[0] = sdpClamp(0.0, X, 1.0);

	Entry("cursorx").SetText(sdpToString(m_Cursor[0]));
	Entry("cursory").SetText("******");
	Entry("cursorvalue").SetText(sdpToString(m_Object->Value1(m_Cursor[0], m_Object->m_DeltaTime)));
}

void CChannelDialog::SetCursorXY(const gdouble X, const gdouble Y)
{
	// Constrain the cursor position
	m_Cursor[0] = sdpClamp(0.0, X, 1.0);
	m_Cursor[1] = Y;

	Entry("cursorx").SetText(sdpToString(m_Cursor[0]));
	Entry("cursory").SetText(sdpToString(m_Cursor[1]));
	Entry("cursorvalue").SetText(sdpToString(m_Object->Value1(m_Cursor[0], m_Object->m_DeltaTime)));
}


