#ifndef CHANNEL_H
#define CHANNEL_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CChannel class which implements a data object that supports k3dIChannel
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deditable.h>
#include <k3dobjectdialog.h>
#include <k3dpersistobject.h>
#include <k3dvariable.h>

#include <k3didependencies.h>
#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dibezierchannelvaluenode.h>
#include <k3dichannel.h>

#include <sdpgtkmouseinput.h>
#include <sdpgtkopengldrawingarea.h>

// Forward class declarations ...
class CChannel;
class CChannelNode;
class CChannelValueNode;

/// Stores a collection of nodes
typedef std::vector<CChannelNode*> NodeCollection;
typedef NodeCollection::iterator NodeIterator;

//////////////////////////////////////////////////////////////////////////////////////////////////
// CChannelDialogSetNodeColor

typedef k3dDialog CChannelDialogSetNodeColorBase;

/// Provides a UI for setting a channel to a single color
class CChannelDialogSetNodeColor :
	public CChannelDialogSetNodeColorBase
{
public:
	CChannelDialogSetNodeColor(sdpIUnknown* const ParentCommandNode, CChannelValueNode* const Node);
	~CChannelDialogSetNodeColor();

	// k3dICommandNode implementation
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

protected:	
	void OnEvent(sdpGtkEvent* Event);
	
	void OnClose();
	void RawOnColorChanged();
	void OnColorChanged(const sdpVector3 NewColor);

private:
	CChannelValueNode* const m_Node;
	sdpVector3 m_LastColor;

	static const sdpString control_close;
	static const sdpString control_colorchanged;
	static const sdpString control_colorselection;
};

/////////////////////////////////////////////////////////////////////////////
// CChannelDialog

typedef k3dObjectDialog CChannelDialogBase;

/// The main user interface for a CChannel object
class CChannelDialog :
	public CChannelDialogBase,
	public sdpGtkMouseInput
{
public:
	CChannelDialog(CChannel* const Object);
	~CChannelDialog();

	/// Called to create the UI
	bool Create();

	/// Called to force a redraw of the UI
	void RedrawAll();

private:
	// k3dICommandNode implementation
	sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	void UpdateControls();
	
	/// Dispatches custom events
	void OnEvent(sdpGtkEvent* Event);

	/// Records mouse events for tutorials
	void RecordEvent(SDPCSTRING Command, GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	/// Records mouse events for tutorials
	void RecordDragEvent(SDPCSTRING Command, GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	// sdpGtkMouseInput implementation ...
	void OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	/// Called when the user moves the vertical scale slider
	void OnVScale();
	/// Called when the user wants to reset the view to its default
	void OnResetView();
	/// Called to straighten the entire channel
	void OnChannelStraighten();
	/// Called to mirror the entire channel
	void OnChannelMirrorX();
	/// Called to mirror the entire channel
	void OnChannelMirrorY();
	/// Called to straighten selected segments
	void OnStraightenSelected();
	/// Called to delete selected nodes
	void OnDeleteSelected();

	/// Called to edit the value of a node
	void EditNode(CChannelNode& Node);
	/// Called to reset the scale & origin of the display
	void ResetView();
	/// Called to deselect everything
	void DeselectAll();

	/// Called to draw the curve graph
	void DrawCurve();
	/// Called to draw the color preview
	void DrawColor();
	/// Called to draw a background grid
	void DrawGrid();
	/// Called to draw a preview of the final Bezier curve, including frequency, phase, and modulation effects
	void DrawPreview();
	/// Called to draw the Bezier control curve
	void DrawControlCurve();
	/// Called to draw tangents to the Bezier control curve
	void DrawTangents();
	/// Called to draw handles at all the control points
	void DrawNodes();
	/// Called to draw a handle at a control node
	void DrawNode(CChannelNode* Node);
	/// Called to draw the current cursor position
	void DrawCursor();
	/// Called to draw the current time position
	void DrawTime();

	/// Converts "screen" coordinates to "world" coordinates
	sdpVector2 World(sdpVector2 ScreenCoords);
	/// Converts "world" coordinates to "screen" coordinates
	sdpVector2 Screen(sdpVector2 WorldCoords);
	/// Tests to see if the given mouse coordinates hit near a channel node
	CChannelNode* HitTest(const sdpVector2 CurrentMouse);

	/// Update the position of the grid cursor
	void SetCursorX(const gdouble X);
	/// Update the position of the grid cursor
	void SetCursorXY(const gdouble X, const gdouble Y);

	/// Stores the owning object
	CChannel* const m_Object;

	/// Stores the OpenGL context for the control curve graph
	sdpGtkOpenGLDrawingArea m_CurveWidget;
	/// Stores the OpenGL context for the color preview bar
	sdpGtkOpenGLDrawingArea m_ColorWidget;

	/// Stores the coordinates of the user interface cursor
	sdpVector2 m_Cursor;

	// Frustum
	static sdpRectangle m_Frustum;

	// Zoom / pan parameters ...
	sdpVector2 m_Origin;
	sdpVector2 m_Scale;
	gdouble m_VScale;

	/// Stores currently-selected nodes
	typedef std::set<CChannelNode*> SelectedNodeCollection;
	typedef SelectedNodeCollection::iterator SelectedNodeIterator;
	SelectedNodeCollection m_Selected;

	// Colors ...
	static sdpVector3 m_BackgroundColor;
	static sdpVector3 m_NormalColor;
	static sdpVector3 m_ValueNodeColor;
	static sdpVector3 m_SelectedColor;
	static sdpVector3 m_TangentColor;
	static sdpVector3 m_SelectorColor;
	static sdpVector3 m_CursorColor;
	static sdpVector3 m_TimeColor;

	static const sdpString dialogname;
	static const sdpString control_cycles;
	static const sdpString control_phase;
	static const sdpString control_straightenchannel;
	static const sdpString control_mirrorxchannel;
	static const sdpString control_mirrorychannel;
	static const sdpString control_straightenselected;
	static const sdpString control_deleteselected;
	static const sdpString control_resetview;

	static const sdpString control_mousemove;
	static const sdpString control_lbuttondown;
	static const sdpString control_lbuttonclick;
	static const sdpString control_lbuttondoubleclick;
	static const sdpString control_lbuttonstartdrag;
	static const sdpString control_lbuttondrag;
	static const sdpString control_lbuttonenddrag;
	static const sdpString control_mbuttonclick;
	static const sdpString control_mbuttondoubleclick;
	static const sdpString control_mbuttonstartdrag;
	static const sdpString control_mbuttondrag;
	static const sdpString control_mbuttonenddrag;
	static const sdpString control_rbuttonclick;
	static const sdpString control_rbuttondoubleclick;
	static const sdpString control_rbuttonstartdrag;
	static const sdpString control_rbuttondrag;
	static const sdpString control_rbuttonenddrag;
	static const sdpString control_lrbuttonstartdrag;
	static const sdpString control_lrbuttondrag;
	static const sdpString control_lrbuttonenddrag;
};

//////////////////////////////////////////////////////////////////////////
// CChannelNode

/// Encapsulates a single node (a 2-dimensional control-point location) within a CChannel
class CChannelNode :
	public k3dIBezierChannelNode,
	public k3dEditable,
	public k3dEventSource
{
public:
	CChannelNode(CChannel* const Channel, const sdpVector2 Position);
	virtual ~CChannelNode();

	// k3dIBezierChannelNode implementation
	sdpResult Position(sdpVector2& Position);
	sdpResult SetPosition(const sdpVector2& Position);

	// k3dIEditable implementation
	virtual void PushEditing();
	virtual void PopEditing();
		
	/// Internal implementation of SetPosition() that doesn't call PushEditing() / PopEditing()
	void RawSetPosition(const sdpVector2& Position);
	/// Internal implementation of SetPosition() that doesn't enforce any constraints
	void UnconstrainedSetPosition(const sdpVector2& Position);

	/// Provides access to the owning channel
	CChannel* const Channel() const;
	/// Provides a fast test for value nodes
	virtual CChannelValueNode* ValueNode();

	/// Saves the node to a K-3D XML document stream
	virtual sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const;
	/// Loads the node from a K-3D XML document
	virtual sdpResult Load(sdpxml::Element& Element);

	/// Lets the node know that it's been deleted
	void Deleted();
	
	/// Lets the node know that it's been undeleted
	void UnDeleted();

	/// Stores the node location
	sdpVector2 m_Position;

protected:
	/// Stores the owning channel
	CChannel* const m_Channel;
	/// Set to true iff the position has changed
	bool m_PositionChanged;
};

//////////////////////////////////////////////////////////////////////////
// CChannelValueNode

typedef CChannelNode CChannelValueNodeBase;

/// Encapsulates a single value node (a 2-dimensional location and 1, 2, or 3-dimensional value) within a CChannel
class CChannelValueNode :
	public CChannelValueNodeBase,
	public k3dIBezierChannelValueNode
{
public:
	CChannelValueNode(CChannel* const Channel, const sdpVector2 Position, const sdpVector3 Value);

	CChannelValueNode* ValueNode();

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();
		
	/// Provides quick access to node value
	const sdpVector3 Value() const;

	// k3dIBezierChannelValueNode implementation
	sdpResult Value1(gdouble& Value);
	sdpResult Value2(sdpVector2& Value);
	sdpResult Value3(sdpVector3& Value);
	sdpResult SetValue1(const gdouble Value);
	sdpResult SetValue2(const sdpVector2& Value);
	sdpResult SetValue3(const sdpVector3& Value);

	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies) const;
	sdpResult Load(sdpxml::Element& Element);

	/// Stores the node value
	sdpVector3 m_Value;
	
protected:
	/// Set to true iff the value has changed
	bool m_ValueChanged;
};

//////////////////////////////////////////////////////////////////////
// CChannel

typedef k3dPersistObject CChannelBase;

/// Encapsulates a data value that can change over time (supports k3dIChannel)
class CChannel :
	public CChannelBase,
	public k3dIChannel,
	public k3dIBezierChannel,
	public k3dEditable
{
friend class CChannelDialog;
public:
	CChannel(sdpIUnknown* Initializer);
	~CChannel();

	// k3dIChannel implementation
	sdpResult SetDatatype(const k3dIChannel::DataType Data);
	k3dIChannel::DataType Datatype();
	sdpResult CurveRange(gdouble* Min, gdouble* Max);
	gdouble Value1(const gdouble Time, const gdouble MaxError);
	sdpResult Value2(const gdouble Time, const gdouble MaxError, double Value[2]);
	sdpResult Value3(const gdouble Time, const gdouble MaxError, double Value[3]);

	// k3dIEditable implementation
	void PushEditing();
	void PopEditing();

	// k3dIBezierChannel implementation ...
	k3dIEditable* Editable();
	sdpResult SetRange1(const gdouble Min, const gdouble Max);
	sdpResult SetRange2(const gdouble Min[2], const gdouble Max[2]);
	sdpResult SetRange3(const gdouble Min[3], const gdouble Max[3]);
	sdpResult Range1(gdouble* Min, gdouble* Max);
	sdpResult Range2(gdouble Min[2], gdouble Max[2]);
	sdpResult Range3(gdouble Min[3], gdouble Max[3]);
	sdpResult SetSegmentCount(const gulong SegmentCount);
	sdpResult VisitNodes(k3dIBezierChannelNodeVisitor* const Visitor);

	// k3dIObject implementation
	sdpResult InitComplete();
	sdpResult EditObject();
 
	// k3dIPersist implementation
	sdpResult Load(sdpxml::Document& Document, sdpxml::Element& Element);
	sdpResult Save(sdpxml::Element& Element, k3dIDependencies* Dependencies);
	
	/// Creates a node	
	CChannelNode* CreateNode(const sdpVector2 Coords);
	/// Creates a value node
	CChannelValueNode* CreateValueNode(const sdpVector2 Coords, const sdpVector3 Value);
	/// Returns the position in the nodes collection to insert a new node
	NodeIterator GetInsertPosition(const sdpVector2 Coords);
	/// Creates a segment
	CChannelNode* CreateSegment(const sdpVector2 Coords);
	/// Deletes a node
	void DeleteNode(CChannelNode* const Node);
	
	/// Returns the given node's iterator
	NodeIterator GetNodeIterator(CChannelNode* const Node);
	/// Returns the previous (when available) control node
	CChannelNode* GetPrevNode(CChannelNode* const Node);
	/// Returns the next (when available) control node
	CChannelNode* GetNextNode(CChannelNode* const Node);
	/// Returns the opposite (from the nearest value node) control node
	CChannelNode* GetOppositeControlNode(CChannelNode* const Node);
	/// Returns the value node associated (closest) to the given control node
	CChannelNode* GetValueNode(CChannelNode* const Node);
	/// Returns the control node closest to the given node
	CChannelNode* GetNeighborControlNode(CChannelNode* const Node);

	/// Returns true iff the given node is the first node
	bool IsFirstNode(CChannelNode* const Node);
	/// Returns true iff the given node is the last node
	bool IsLastNode(CChannelNode* const Node);
	
	/// Constrains node positions
	void ConstrainNodePosition(CChannelNode* const Node, sdpVector2& Position);

	static const sdpString command_straighten;
	static const sdpString command_mirrorx;
	static const sdpString command_mirrory;	

protected:
	/// Returns the number of scalar values (1, 2, or 3) per node, based on the current data type
	gulong GetElementsPerNode();
	/// Returns the maximum extents of the curve
	sdpRectangle Extents();
	/// Returns the maximum extents of the value nodes (not the normals)
	sdpRectangle ValueNodeExtents();
	/// Called to straighten the entire channel
	void Straighten();
	/// Called to mirror the entire channel along its X axis
	void MirrorX();
	/// Called to mirror the entire channel along its Y axis
	void MirrorY();

	/// Deletes a node without any constraints
	void RawDeleteNode(CChannelNode* const Node);
	/// Deletes all nodes
	void ClearNodes();
	
	/// Implements backwards-compatible (pre 0.1.26.0) loading ... should be removed prior to 0.4
	sdpResult OldFashionedLoad(sdpxml::Element& Element);

	friend std::ostream& operator<<(std::ostream& Stream, const k3dIChannel::DataType& Value)
	{
		switch(Value)
			{
				case DATA_SCALAR:
					Stream << "scalar";
					break;
				case DATA_TIME:
					Stream << "time";
					break;
				case DATA_ANGLE:
					Stream << "angle";
					break;
				case DATA_DISTANCE:
					Stream << "distance";
					break;
				case DATA_AREA:
					Stream << "area";
					break;
				case DATA_VOLUME:
					Stream << "volume";
					break;
				case DATA_MASS:
					Stream << "mass";
					break;
				case DATA_FORCE:
					Stream << "force";
					break;
				case DATA_PRESSURE:
					Stream << "pressure";
					break;
				case DATA_COLOR:
					Stream << "color";
					break;
				case DATA_ROTATIONXYZ:
					Stream << "orientation";
					break;
				case DATA_SCALEXYZ:
					Stream << "scale";
					break;
			}
			
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, k3dIChannel::DataType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "scalar")
			Value = DATA_SCALAR;
		else if(text == "time")
			Value = DATA_TIME;
		else if(text == "angle")
			Value = DATA_ANGLE;
		else if(text == "distance")
			Value = DATA_DISTANCE;
		else if(text == "area")
			Value = DATA_AREA;
		else if(text == "volume")
			Value = DATA_VOLUME;
		else if(text == "mass")
			Value = DATA_MASS;
		else if(text == "force")
			Value = DATA_FORCE;
		else if(text == "pressure")
			Value = DATA_PRESSURE;
		else if(text == "color")
			Value = DATA_COLOR;
		else if(text == "orientation")
			Value = DATA_ROTATIONXYZ;
		else if(text == "scale")
			Value = DATA_SCALEXYZ;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
			
		return Stream;
	}


	/// Stores the channel's data type
	k3dVariable<k3dIChannel::DataType> m_DataType;
	/// Stores the minimum allowable data value
	k3dVariable<sdpVector3> m_MinValue;
	/// Stores the maximum allowable data value
	k3dVariable<sdpVector3> m_MaxValue;
	/// Stores the minimum allowable data coordinates
	k3dVariable<sdpVector2> m_MinCoords;
	/// Stores the maximum allowable data coordinates
	k3dVariable<sdpVector2> m_MaxCoords;
	/// Stores the number of times the channel should repeat during the animation
	k3dVariable<gdouble> m_Cycles;
	/// Stores the phase offset between the channel and the animation
	k3dVariable<gdouble> m_Phase;

	/// Stores channel nodes
	NodeCollection m_Nodes;
	/// Set to true iff the channel's collection of nodes has changed during the current editing session
	bool m_NodeCollectionChanged;

	/// Controls the maximum number of iterations when calculating the current channel value, to prevent infinite loops
	static const gulong m_MaxIterations;
};

#endif // CHANNEL_H

