// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDisk K-3D object, which renders a RenderMan disk primitive
		\author Tim Shead (tshead@k-3d.com)
*/

#include "disk.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>
#include <k3dieditorengine.h>
#include <k3dimaterial.h>
#include <k3diriengine.h>
#include <k3dirimaterial.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

CDisk::CDisk(sdpIUnknown* Initializer) :
	CDiskBase(Initializer),
	m_Material("material", Document(), Document()->StateRecorder(), 0),
	m_Radius("radius", Document(), Document()->StateRecorder(), 0),
	m_Height("height", Document(), Document()->StateRecorder(), 0),
	m_SweepAngle("sweepangle", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_Material);

	AddVariable(&m_Radius);
	AddVariable(&m_Height);
	AddVariable(&m_SweepAngle);

	m_Material.SetObject(k3dDefaultMaterial(Document()));
}

CDisk::~CDisk()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CDisk::EditObject()
{
	new CDiskDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CDisk::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	sdpglTrace("CDisk::EditorRender()");

	const gdouble radius = m_Radius->Value1(m_CurrentTime, m_DeltaTime);
	const gdouble height = m_Height->Value1(m_CurrentTime, m_DeltaTime);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glEnable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glDisable(GL_CULL_FACE);

	const gulong nsides = 360;
	const gulong sweepangle = (gulong)sdpDegrees(m_SweepAngle->Value1(m_CurrentTime, m_DeltaTime));

	glBegin(GL_POLYGON);
	if(sweepangle < 360)
		glVertex3d(0, height, 0);

	for(gulong i = 360-sweepangle; i < 360; i++)
		{
			glVertex3d(radius*cos(2*M_PI/(gdouble)nsides*(gdouble)i), height, radius*sin(2*M_PI/(gdouble)nsides*(gdouble)i));
		}
	glEnd();

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, radius);

	return SDP_S_OK;
}

sdpResult CDisk::RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	// If we don't have a material, we're done ...
	if(0 == m_Material.Interface())
		return SDP_S_FALSE;
	k3dIRiMaterial* const rimaterial = dynamic_cast<k3dIRiMaterial*>(m_Material.Object());
	if(0 == rimaterial)
		return SDP_S_FALSE;

	const gdouble radius = m_Radius->Value1(m_CurrentTime, m_DeltaTime);
	const gdouble height = m_Height->Value1(m_CurrentTime, m_DeltaTime);
	const gdouble sweepangle = sdpDegrees(m_SweepAngle->Value1(m_CurrentTime, m_DeltaTime));

	Engine->RiTransformBegin();

	sdpMatrix4 matrix = sdpRotation3D(sdpRadians(-90.0), sdpVector3(1, 0, 0)).Transpose();
	Engine->RiConcatTransform(matrix);

	rimaterial->RiMaterialSetup(Frame, Engine);
	Engine->RiDiskV(height, radius, sweepangle, 0, 0, 0, 0);

	Engine->RiTransformEnd();

	return SDP_S_OK;
}

k3dIMaterial* CDisk::Material()
{
	return m_Material.Interface();
}

sdpResult CDisk::SetMaterial(k3dIMaterial* Material)
{
	// We allow NULL here, as a synonym for the default material ...
	m_Material.SetObject(dynamic_cast<k3dIObject*>(Material));

	return SDP_S_OK;
}

sdpResult CDisk::InitComplete()
{
	if(0 == m_Radius.Object())
		g_return_val_if_fail(m_Radius.Create("Radius", 5.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Height.Object())
		g_return_val_if_fail(m_Height.Create("Height", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_SweepAngle.Object())
		g_return_val_if_fail(m_SweepAngle.Create("Sweep Angle", sdpRadians(360.0), true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);

	return CDiskBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CDiskDialog

const sdpString CDiskDialog::dialogname = "properties";
const sdpString CDiskDialog::control_radius = "radius";
const sdpString CDiskDialog::control_height = "height";
const sdpString CDiskDialog::control_sweepangle = "sweepangle";
const sdpString CDiskDialog::control_position = "position";
const sdpString CDiskDialog::control_orientation = "orientation";
const sdpString CDiskDialog::control_scale = "scale";
const sdpString CDiskDialog::control_target = "target";

CDiskDialog::CDiskDialog(CDisk* const Object) :
	CDiskDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "disk.gtkml"));

	// Connect spin buttons
	Connectk3dChannelButton(control_radius, m_Object->m_Radius.Interface());
	Connectk3dChannelButton(control_height, m_Object->m_Height.Interface());
	Connectk3dChannelButton(control_sweepangle, m_Object->m_SweepAngle.Interface());

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}

void CDiskDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	CDiskDialogBase::OnEvent(Event);
}

