// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CEditorEngine class, which is responsible for rendering editor-mode graphics into a camera
		\author Tim Shead (tshead@k-3d.com)
*/

#include "camera.h"

#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dviewvolume.h>

#include <k3dichannel.h>
#include <k3didocument.h>
#include <k3dieditorlight.h>
#include <k3dieditormaterial.h>
#include <k3dieditortexture.h>
#include <k3ditexture.h>

namespace
{

class PreRender
{
public:
	PreRender(k3dIEditorEngine* const Engine) :
		m_Engine(Engine)
	{
	}

	void operator()(k3dIObject* const Object)
	{
		k3dIEditorRender* const render = dynamic_cast<k3dIEditorRender*>(Object);
		if(render)
			render->EditorPreRender(m_Engine);

		k3dIEditorLight* const light = dynamic_cast<k3dIEditorLight*>(Object);
		if(light)
			light->EditorLightPreRender(m_Engine);

		k3dIEditorTexture* const texture = dynamic_cast<k3dIEditorTexture*>(Object);
		if(texture)
			texture->EditorTexturePreRender(m_Engine);

		k3dIEditorMaterial* const material = dynamic_cast<k3dIEditorMaterial*>(Object);
		if(material)
			material->EditorMaterialPreRender(m_Engine);
	}
	
private:
	k3dIEditorEngine* const m_Engine;
};

class LightSetup
{
public:
	LightSetup(k3dIEditorEngine* const Engine) :
		m_Engine(Engine),
		m_LightIndex(0)
	{
	}

	void operator()(k3dIEditorLight* const Light)
	{
		Light->EditorLightSetup(m_Engine, m_LightIndex++);
	}

private:
	k3dIEditorEngine* const m_Engine;
	gulong m_LightIndex;
};

class Render
{
public:
	Render(k3dIEditorEngine* const Engine) :
		m_Engine(Engine)
	{
	}

	void operator()(k3dIObject* const Object, const k3dIHierarchy::NodeType Type)
	{
		k3dIEditorRender* const render = dynamic_cast<k3dIEditorRender*>(Object);
		if(!render)
			return;

		switch(Type)
			{
				case k3dIHierarchy::PUSH:
					render->EditorPushRender(m_Engine);
					break;

				case k3dIHierarchy::NODE:
					render->EditorRender(m_Engine);
					render->EditorPushChildRender(m_Engine);
					break;

				case k3dIHierarchy::POP:
					render->EditorPopChildRender(m_Engine);
					render->EditorPopRender(m_Engine);
					break;

				default:
					g_return_if_fail(0); // Should never happen!
			}
	}

private:
	k3dIEditorEngine* const m_Engine;
};

class PostRender
{
public:
	PostRender(k3dIEditorEngine* const Engine) :
		m_Engine(Engine)
	{
	}

	void operator()(k3dIEditorRender* const Render)
	{
		Render->EditorPostRender(m_Engine);
	}

private:
	k3dIEditorEngine* const m_Engine;
};

} // namespace

//////////////////////////////////////////////////////////////////////////////////////
// CEditorEngine

CEditorEngine::CEditorEngine() :
	m_Camera(0)
{
}

CEditorEngine::~CEditorEngine()
{
}

/////////////////////////////////////////////////////////////////
// k3dIRenderEngine overrides ...

sdpResult CEditorEngine::SetCamera(k3dICamera* Camera)
{
	// Sanity checks ...
	g_return_val_if_fail(Camera, SDP_E_FAIL);

	m_Camera = (CCamera*)Camera;

	return SDP_S_OK;
}

sdpResult CEditorEngine::Redraw()
{
	// Sanity checks ...
	g_return_val_if_fail(m_Camera, SDP_E_FAIL);
	
	if(0 == m_Camera->DrawingArea())
		return SDP_S_FALSE;

	// Render ...
	m_Camera->DrawingArea()->Begin();
	EditorRender();
	m_Camera->DrawingArea()->SwapBuffers();
	m_Camera->DrawingArea()->End();

	return SDP_S_OK;
}

sdpResult CEditorEngine::RenderPreview()
{
	return SDP_S_FALSE;
}

sdpResult CEditorEngine::RenderFrame()
{
	return SDP_S_FALSE;
}

sdpResult CEditorEngine::RenderAnimation()
{
	return SDP_S_FALSE;
}

///////////////////////////////////////////////////////////////
// k3dIEditorEngine overrides ...

k3dICamera* CEditorEngine::Camera()
{
	return m_Camera;
}

sdpResult CEditorEngine::EditorQueryWireframe()
{
	return m_Camera->m_EditorWireframe.Value() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQuerySolid()
{
	return m_Camera->m_EditorSolid.Value() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryShaded()
{
	return m_Camera->m_EditorSmoothing.Value() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryTextured()
{
	return m_Camera->m_EditorTexture.Value() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryOrientation()
{
	return m_Camera->m_EditorOrientation.Value() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQueryNormals()
{
	return m_Camera->m_EditorNormals.Value() ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CEditorEngine::EditorQuerySolidCulling()
{
	return m_Camera->m_EditorTwoSided.Value() ? SDP_S_FALSE : SDP_S_OK;
}

sdpResult CEditorEngine::EditorQueryWireframeCulling()
{
	return m_Camera->m_EditorWireframeTwoSided.Value() ? SDP_S_FALSE : SDP_S_OK;
}

gulong CEditorEngine::EditorQueryWireframePatchDivisions()
{
	return m_Camera->m_EditorWireframePatchDivisions.Value();
}

gulong CEditorEngine::EditorQuerySolidPatchDivisions()
{
	return m_Camera->m_EditorSolidPatchDivisions.Value();
}

gulong CEditorEngine::EditorFontBase()
{
	return m_Camera->DrawingArea()->FontBase();
}

///////////////////////////////////////////////////////////////
// Helper functions ...

void CEditorEngine::EditorRender()
{
	sdpglTrace("CEditorEngine::EditorRender()");

	RenderOptions();
	PreRender();
	Render();
	PostRender();

	glFlush();
}


void CEditorEngine::RenderOptions()
{
	sdpglTrace("CEditorEngine::RenderOptions()");

	// Set point size ...
	glPointSize(m_Camera->m_PointSize);

	// Setup stippling for path normals / bezier control lines ...
	glLineStipple(2, 0xaaaa);
	glDisable(GL_LINE_STIPPLE);
	glDisable(GL_POLYGON_STIPPLE);

	// Setup culling ...
	glFrontFace(GL_CW);
	glCullFace(GL_BACK);

	// Setup dithering ...
	glDisable(GL_DITHER);

	// Setup antialiasing ...
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_BLEND);
	glLineWidth(1.0f);

	// Disable stencil ...
	glDisable(GL_STENCIL_TEST);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	glStencilMask(0x00);

	// Enable path shading ...
	glShadeModel(GL_SMOOTH);

	// Normalization ...
	glEnable(GL_NORMALIZE);

	// Set Z buffer options ...
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);

	// Normalization ...
	glEnable(GL_NORMALIZE);

	// Prepare additional bezier options ...
	glEnable(GL_MAP2_TEXTURE_COORD_2);
	glEnable(GL_AUTO_NORMAL);

	// Prepare texture options ...
	if(m_Camera->m_EditorTexture.Value())
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);


	// Prepare bezier options ...
	glEnable(GL_MAP1_VERTEX_3);
	glEnable(GL_MAP2_VERTEX_3);

	// Calculate current time & animate the document ...
	m_Camera->m_Time->SetRenderTime(m_Camera->m_Time->ViewTime());
	m_Camera->m_Time->SetRenderDeltaTime(1.0 / m_Camera->m_Time->Duration() * m_Camera->m_Time->FrameRate());
	m_Camera->Document()->Hierarchy()->Animate(m_Camera->m_Time->ViewTime(), 1.0 / m_Camera->m_Time->Duration() * m_Camera->m_Time->FrameRate());

	// Setup texture matrix
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();

	// Setup viewport ...
	glViewport(0, 0, m_Camera->Width(), m_Camera->Height());

	// Setup projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	k3dOpenGLViewVolumeSetup(m_Camera->ViewVolume());
	
	// Setup modelview matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glScaled(1.0, 1.0, -1.0);
	glRotated(-sdpDegrees(m_Camera->m_CurrentOrientation.angle), m_Camera->m_CurrentOrientation.axis[0], m_Camera->m_CurrentOrientation.axis[1], m_Camera->m_CurrentOrientation.axis[2]);
	glTranslated(-m_Camera->m_CurrentPosition[0], -m_Camera->m_CurrentPosition[1], -m_Camera->m_CurrentPosition[2]);

	// Clear background ...
	glClearDepth(1.0);
	glClearColor(GLfloat(m_Camera->m_EditorBackgroundColor[0]), GLfloat(m_Camera->m_EditorBackgroundColor[1]), GLfloat(m_Camera->m_EditorBackgroundColor[2]), 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Setup fog ...
	if(m_Camera->m_EditorFog.Value())
		{
			glEnable(GL_FOG);

			GLfloat fogdata[4];
			fogdata[0] = (GLfloat)m_Camera->m_EditorBackgroundColor[0];
			fogdata[1] = (GLfloat)m_Camera->m_EditorBackgroundColor[1];
			fogdata[2] = (GLfloat)m_Camera->m_EditorBackgroundColor[2];
			fogdata[3] = 1.0f;

			glFogfv(GL_FOG_COLOR, fogdata);
			glFogf(GL_FOG_START, static_cast<GLfloat>(m_Camera->m_EditorFogNear.Value()));
			glFogf(GL_FOG_END, static_cast<GLfloat>(m_Camera->m_EditorFogFar.Value()));
			glHint(GL_FOG_HINT, GL_NICEST);
			glFogi(GL_FOG_MODE, GL_LINEAR);
		}
	else
		{
			glDisable(GL_FOG);
		}

	// Setup lights ...
	LightSetup();

	// If we're supposed to draw path normals, make sure they're up-to-date ...
	if(m_Camera->m_EditorNormals.Value() || m_Camera->m_EditorSolid.Value())
		k3dCalculateNormals(m_Camera->Document()->Objects());
}

void CEditorEngine::PreRender()
{
	sdpglTrace("CEditorEngine::PreRender()");
	k3d::for_each(m_Camera->Document()->Objects(), ::PreRender(this));
}

void CEditorEngine::LightSetup()
{
	sdpglTrace("CEditorEngine::LightSetup()");

	// Enable lighting ...
	glEnable(GL_LIGHTING);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

	// Make sure all lights are turned off initially ...
	GLint maxlights;
	glGetIntegerv(GL_MAX_LIGHTS, &maxlights);
	for(GLint i = 0; i < maxlights; i++)
		glDisable(GLenum(GL_LIGHT0 + i));

	k3d::for_each<k3dIEditorLight>(m_Camera->Document()->Objects(), ::LightSetup(this));
}

void CEditorEngine::Render()
{
	sdpglTrace("CEditorEngine::Render()");
	k3d::for_each(m_Camera->Document()->Hierarchy(), ::Render(this));
}

void CEditorEngine::PostRender()
{
	sdpglTrace("CEditorEngine::PostRender()");
	k3d::for_each<k3dIEditorRender>(m_Camera->Document()->Objects(), ::PostRender(this));
}


