// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CLight K-3D object, which encapsulates a light source within the 3D workspace
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "light.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3drishadersetup.h>
#include <k3dshaderargumentspage.h>
#include <k3dshaderinfo.h>
#include <k3dshaderselector.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>
#include <k3dichannel.h>
#include <k3diobjectcollection.h>
#include <k3ditexture.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

CLight::CLight(sdpIUnknown* Initializer) :
	CLightBase(Initializer),
	m_Type("type", POINT, Document()->StateRecorder(), 0),
	m_CastShadows("castshadows", false, Document()->StateRecorder(), 0),
	m_Intensity("intensity", Document(), Document()->StateRecorder(), 0),
	m_Color("color", Document(), Document()->StateRecorder(), 0),
	m_Shader("shader", Document(), Document()->StateRecorder())
{
	AddVariable(&m_Type);
	AddVariable(&m_CastShadows);
	AddVariable(&m_Intensity);
	AddVariable(&m_Color);
	AddVariable(&m_Shader);
}

CLight::~CLight()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CLight::EditObject()
{
	new CLightDialog(this);
	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////
// sdpsl::IShaderCollection implementation

void CLight::VisitShaders(sdpsl::IShaderVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	if(m_Shader.Shader())
		Visitor->VisitShader(m_Shader.Shader());
}

////////////////////////////////////////////////////////////////
// k3dIEditorRender implementation ...

sdpResult CLight::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	sdpglTrace("CLight::EditorRender()");

	StoreOpenGLAttributes attributes;

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glColor3d(1.0, 1.0, 1.0);
	glLineWidth(1.0f);
	glDisable(GL_LINE_STIPPLE);

	switch(m_Type.Value())
		{
			case AMBIENT:
				EditorRenderAmbient();
				break;
			case POINT:
				EditorRenderPoint();
				break;
			case DISTANT:
				EditorRenderDistant();
				break;
			case SPOTLIGHT:
				EditorRenderSpotlight();
				break;
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, 1.0);

	return SDP_S_OK;
}

void CLight::EditorRenderAmbient()
{
	sdpglTrace("CLight::EditorRenderAmbient()");

	EditorRenderPoint();
}

void CLight::EditorRenderPoint()
{
	sdpglTrace("CLight::EditorRenderPoint()");

	// Draw a nice "furball"
	glBegin(GL_LINES);

	sdpVector3 coords(0, 0, 0);

	glVertex3d(coords[0] + 1.0, coords[1], coords[2]);
	glVertex3d(coords[0] - 1.0, coords[1], coords[2]);
	glVertex3d(coords[0], coords[1] + 1.0, coords[2]);
	glVertex3d(coords[0], coords[1] - 1.0, coords[2]);
	glVertex3d(coords[0], coords[1], coords[2] + 1.0);
	glVertex3d(coords[0], coords[1], coords[2] - 1.0);

	glVertex3d(coords[0] + 0.4, coords[1] + 0.4, coords[2] + 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] - 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] + 0.4, coords[2] + 0.4);
	glVertex3d(coords[0] + 0.4, coords[1] - 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] + 0.4, coords[1] + 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] - 0.4, coords[2] + 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] + 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] + 0.4, coords[1] - 0.4, coords[2] + 0.4);

	glEnd();
}

void CLight::EditorRenderSpotlight()
{
	sdpglTrace("CLight::EditorRenderSpotlight()");

	GLUquadricObj* quadric = gluNewQuadric();
	g_assert(quadric);

	// Draw the cone of illumination ...
//			glScaled(m_CurrentWidthPlane, m_CurrentHeightPlane, m_CurrentNearPlane);
//			glRotated(45.0, 0.0, 0.0, 1.0);
	glScaled(1.0, 1.0, -1.0);
	glTranslated(0.0, 0.0, -1.0);

	gluQuadricDrawStyle(quadric, GLenum(GLU_LINE));
	gluCylinder(quadric, sqrt(2.0), 0.001, 1.0, 8, 1);
	gluDeleteQuadric(quadric);

	// Draw the axis ...
	glBegin(GL_LINES);

	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, 0.0, -5.0);

	glEnd();
}

void CLight::EditorRenderDistant()
{
	sdpglTrace("CLight::EditorRenderDistant()");

	// Draw the "tail" of our arrow ...
	glBegin(GL_LINES);

	glVertex3d(0, 0, 0);
	glVertex3d(0, 0, -5);

	glEnd();

	// Draw the "head" of our arrow as a cone ...
	GLUquadricObj* quadric = gluNewQuadric();
	g_assert(quadric);

	gluQuadricDrawStyle(quadric, GLenum(GLU_LINE));
	gluCylinder(quadric, 0.4, 0.001, 0.8, 8, 1);
	gluDeleteQuadric(quadric);
}

///////////////////////////////////////////////////////////////
// k3dILight implementation


/////////////////////////////////////////////////////////////////
// k3dIEditorLight implementation
sdpResult CLight::EditorLightPreRender(k3dIEditorEngine* const Engine)
{
	sdpglTrace("CLight::EditorLightPreRender()");

	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::EditorLightSetup(k3dIEditorEngine* const Engine, gulong Light)
{
	sdpglTrace("CLight::EditorLightSetup()");

	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return Setup(Light);
}

sdpResult CLight::Setup(gulong Light)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	// We don't render ambient lights!
	if(AMBIENT == m_Type.Value())
		return SDP_S_FALSE;

	// Setup our own transformation matrix ...
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	k3dPushOpenGLMatrix(k3dGetObjectToWorldMatrix(this));

	// Offset our light index ...
	GLenum light = GLenum(Light + GL_LIGHT0);

	// Enable the light ...
	glEnable(light);

	// Setup intensity / attenuation ...
	gdouble intensity = 1.0 / m_Intensity->Value1(m_CurrentTime, m_DeltaTime);
	glLightf(light, GL_CONSTANT_ATTENUATION, 0.0f);
	glLightf(light, GL_LINEAR_ATTENUATION, 0.0f);
	glLightf(light, GL_QUADRATIC_ATTENUATION, GLfloat(intensity));

	// Setup color ...
	sdpVector3 color;
	m_Color->Value3(m_CurrentTime, m_DeltaTime, color);

	GLfloat glcolor[4];
	glcolor[0] = (GLfloat)color[0];
	glcolor[1] = (GLfloat)color[1];
	glcolor[2] = (GLfloat)color[2];
	glcolor[3] = 1.0f;

	glLightfv(light, GL_AMBIENT, glcolor);
	glLightfv(light, GL_DIFFUSE, glcolor);
	glLightfv(light, GL_SPECULAR, glcolor);

	// Setup light based on type
	if(DISTANT == m_Type.Value())
		{
			GLfloat gllightpos[] = { 0.0f, 0.0f, -1.0f, 0.0f };
			glLightfv(light, GL_POSITION, gllightpos);
		}
	else if(SPOTLIGHT == m_Type.Value())
		{
			// Set spotlight position
			GLfloat gllightpos[4];
			gllightpos[0] = 0.0f;
			gllightpos[1] = 0.0f;
			gllightpos[2] = 0.0f;
			gllightpos[3] = 1.0f;
			glLightfv(light, GL_POSITION, gllightpos);

			// Set spotlight angle & exponent
//			glLightf(Light, GL_SPOT_CUTOFF, (GLfloat)m_Coneangle->GetValue1(m_CurrentTime, m_DeltaTime));
//			glLightf(Light, GL_SPOT_EXPONENT, (GLfloat)m_Spotlightfalloff->GetValue1(m_CurrentTime, m_DeltaTime));

			const static GLfloat lightdirection[3] = { 0.0f, 0.0f, -1.0f };
			glLightfv(light, GL_SPOT_DIRECTION, lightdirection);
		}
	else if(POINT == m_Type.Value())
		{
			// Set point light position
			GLfloat gllightpos[4];
			gllightpos[0] = 0.0f;
			gllightpos[1] = 0.0f;
			gllightpos[2] = 0.0f;
			gllightpos[3] = 1.0f;
			glLightfv(light, GL_POSITION, gllightpos);

			// Make sure it doesn't become a spotlight
			glLightf(light, GL_SPOT_CUTOFF, 180.0f);
		}

	glPopMatrix();
	return SDP_S_OK;
}

/////////////////////////////////////////////////////
// k3dIRiLight implementation

sdpResult CLight::RiLightStartJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::RiLightPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::RiLightSetup(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	// If the user hasn't selected a shader, we're done ...
	if(0 == m_Shader.Shader())
		return SDP_S_FALSE;

	// Setup our own transformation matrix ...
	Engine->RiTransformBegin();
	Engine->RiIdentity();
	k3dPushRiMatrix(k3dGetObjectToWorldMatrix(this), Engine);

	// Setup light attributes ...
	k3dRiParameterList lightattributes(new k3dRiTokenValuePair("shadows", new k3dRiStringArray("string", m_CastShadows.Value() ? "on" : "off")));
	lightattributes.SetAttributes("light", Engine);

	// Create our light shader helper object ...
	k3dRiLightShader light;
	m_Shader.InsertInto(Frame, Engine, m_CurrentTime, m_DeltaTime, light);

	// Setup light "from" and "to" positions ...
	const sdpVector3 lightfrom(0, 0, 0);
	const sdpVector3 lightto = lightfrom + sdpVector3(0.0, 0.0, 1.0);
	light.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_FROM, new k3dRiRealArray("point", lightfrom)));
	light.AddTokenValuePair(new k3dRiTokenValuePair(k3dIRiEngine::RI_TO, new k3dRiRealArray("point", lightto)));

	// Fire it all off to the rendering engine ...
	light.SetupLight(Engine);

	// Pop our transformation matrix
	Engine->RiTransformEnd();

	return SDP_S_OK;
}

sdpResult CLight::RiLightPostRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::RiLightFinishJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

SDPCSTRING const CLight::Key_LightType = "LightType";
SDPCSTRING const CLight::Key_Ambient = "Ambient";
SDPCSTRING const CLight::Key_Point = "Point";
SDPCSTRING const CLight::Key_Distant = "Distant";
SDPCSTRING const CLight::Key_Spotlight = "Spotlight";
SDPCSTRING const CLight::Key_LightIntensityParameter = "LightIntensityParameter";
SDPCSTRING const CLight::Key_LightColorParameter = "LightColorParameter";

void CLight::GetShaderLightType()
{
	sdpString lighttype = Key_Point;
/*
	k3dIParameter* parameter = m_Light.Interface()->FindParameter(Key_LightType);
	if(parameter) lighttype = parameter->Value();

	if(lighttype == Key_Ambient)
		{
			m_Type() = AMBIENT;
		}
	else if(lighttype == Key_Point)
		{
			m_Type() = POINT;
		}
	else if(lighttype == Key_Distant)
		{
			m_Type() = DISTANT;
		}
	else if(lighttype == Key_Spotlight)
		{
			m_Type() = SPOTLIGHT;
		}
	else
		{
			g_assert(0);	// Unknown light type - default to "point"
			m_Type() = POINT;
		}
*/
}

void CLight::GetShaderLightIntensity()
{
/*
	// Release any existing references to intensity ...
	m_Intensity.Release();

	// See if the shader specifies which parameter controls intensity ...
	sdpString lightintensity;
	k3dIParameter* parameter = m_Light.Interface()->FindParameter(Key_LightIntensityParameter);

	// If not, look for a parameter named "intensity" ...
	lightintensity = parameter ? parameter->Value() : "intensity";

	// Find the parameter ...
	for(sdpsl::IShaderSelectorArgument* argument = m_Light->FirstArgument(); argument; argument = m_Light->NextArgument(argument))
		{
			if(lightintensity == argument->Name())
				{
					m_Intensity.Attach((sdpIUnknown*)argument->GetValue());
					return;
				}
		}

	OnModify();
*/
}

void CLight::GetShaderLightColor()
{
/*
	// Release any existing references to color ...
	m_Color.Release();

	// See if the shader specifies which parameter controls color ...
	sdpString lightcolor;
	k3dIParameter* parameter = m_Light.Interface()->FindParameter(Key_LightColorParameter);

	// If not, look for a parameter named "lightcolor" ...
	lightcolor = parameter ? parameter->Value() : "lightcolor";

	// Find the parameter ...
	for(sdpsl::IShaderSelectorArgument* argument = m_Light->FirstArgument(); argument; argument = m_Light->NextArgument(argument))
		{
			if(lightcolor == argument->Name())
				{
					m_Color.Attach((sdpIUnknown*)argument->GetValue());
					return;
				}
		}

	OnModify();
*/
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CLight::InitComplete()
{
	k3d::PushEditing(Document()->Objects());

	if(0 == m_Shader.Shader())
		m_Shader.SetShader("k3d_pointlight");

	if(0 == m_Intensity.Object())
		g_return_val_if_fail(m_Intensity.Create("Light intensity", 1200.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Color.Object())
		g_return_val_if_fail(m_Color.Create("Light color", sdpVector3(1.0), true, true), SDP_E_FAIL);

	GetShaderLightType();
	GetShaderLightIntensity();
	GetShaderLightColor();

	k3d::PopEditing(Document()->Objects());
	return CLightBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CLightDialog

const sdpString CLightDialog::dialogname = "properties";
const sdpString CLightDialog::control_castshadows = "castshadows";
const sdpString CLightDialog::control_position = "position";
const sdpString CLightDialog::control_orientation = "orientation";
const sdpString CLightDialog::control_scale = "scale";
const sdpString CLightDialog::control_target = "target";

CLightDialog::CLightDialog(CLight* const Object) :
	CLightDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "light.gtkml"));

	// Setup check buttons ...
	Connectk3dCheckButton(control_castshadows, m_Object->m_CastShadows, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Cast Shadows", 0, 0, 0);

	// Setup shader selectors ...
	if(Getk3dShaderSelector("shaderselector"))
		Getk3dShaderSelector("shaderselector")->Attach(m_Object->Document(), &m_Object->m_Shader, true, sdpsl::IShader::LIGHT);

	if(Getk3dShaderInfo("shaderinfo"))
		Getk3dShaderInfo("shaderinfo")->Attach(m_Object->Document(), &m_Object->m_Shader);

	// Setup shader arguments ...
	if(Getk3dShaderArgumentsPage("shaderarguments"))
		Getk3dShaderArgumentsPage("shaderarguments")->Attach(m_Object->Document(), &m_Object->m_Shader);

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}

void CLightDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	CLightDialogBase::OnEvent(Event);
}


