#ifndef LIGHT_H
#define LIGHT_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Defines the CLight K-3D object, which encapsulates a light source within the 3D workspace
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dobjectdialog.h>
#include <k3dshaderinstance.h>
#include <k3dtransformobject.h>
#include <k3dvariable.h>

#include <k3dilight.h>
#include <k3dieditorlight.h>
#include <k3dirilight.h>

// Forward class declarations
class CLight;

/////////////////////////////////////////////////////////////////////////////
// CLightDialog

typedef k3dObjectDialog CLightDialogBase;

/// Encapsulates the Light UI
class CLightDialog : public CLightDialogBase
{
public:
	CLightDialog(CLight* const Object);

protected:
	/// Stores a pointer to the owning object
	CLight* const m_Object;

	/// Handles our custom events
	void OnEvent(sdpGtkEvent* Event);

	static const sdpString dialogname;
	static const sdpString control_castshadows;
	static const sdpString control_position;
	static const sdpString control_orientation;
	static const sdpString control_scale;
	static const sdpString control_target;
};

///////////////////////////////////////////////////////////////////////////////
// CLight

typedef k3dTransformObject CLightBase;

class CLight :
	public CLightBase,
	public k3dILight,
	public k3dIEditorLight,
	public k3dIRiLight,
	public sdpsl::IShaderCollection
{
friend class CLightDialog;
public:
	CLight(sdpIUnknown* Initializer);
	~CLight();

	// k3dIObject implementation
	sdpResult EditObject();

	// k3dIEventObserver implementation

	// sdpsl::IShaderCollection implementation
	void VisitShaders(sdpsl::IShaderVisitor* const Visitor);

	// k3dIEditorRender implementation
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3dIEditorLight implementation
	sdpResult EditorLightPreRender(k3dIEditorEngine* const Engine);
	sdpResult EditorLightSetup(k3dIEditorEngine* const Engine, gulong Light);
	sdpResult Setup(gulong Light);

	// k3dIRiLight implementation
	sdpResult RiLightStartJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine);
	sdpResult RiLightPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiLightSetup(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiLightPostRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiLightFinishJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine);

	// k3dIPersist implementation
	sdpResult InitComplete();

protected:
	void GetShaderLightType();
	void GetShaderLightIntensity();
	void GetShaderLightColor();

	void EditorRenderAmbient();
	void EditorRenderPoint();
	void EditorRenderDistant();
	void EditorRenderSpotlight();

	typedef enum
	{
		AMBIENT = 1,
		POINT = 2,
		DISTANT = 3,
		SPOTLIGHT = 4,
	} LightType;
		
	friend std::ostream& operator<<(std::ostream& Stream, const LightType& Value)
	{
		switch(Value)
			{
				case AMBIENT:
					Stream << "ambient";
					break;
				case POINT:
					Stream << "point";
					break;
				case DISTANT:
					Stream << "distant";
					break;
				case SPOTLIGHT:
					Stream << "spotlight";
					break;
			}
			
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, LightType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "ambient")
			Value = AMBIENT;
		else if(text == "point")
			Value = POINT;
		else if(text == "distant")
			Value = DISTANT;
		else if(text == "spotlight")
			Value = SPOTLIGHT;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}


	static SDPCSTRING const Key_LightType;
	static SDPCSTRING const Key_Ambient;
	static SDPCSTRING const Key_Point;
	static SDPCSTRING const Key_Distant;
	static SDPCSTRING const Key_Spotlight;
	static SDPCSTRING const Key_LightIntensityParameter;
	static SDPCSTRING const Key_LightColorParameter;

	k3dVariable<LightType> m_Type;
	k3dVariable<bool> m_CastShadows;
//	k3dScalarChannel m_Target;
	k3dScalarChannel m_Intensity;
	k3dColorChannel m_Color;
	k3dShaderInstance m_Shader;
};

#endif // LIGHT_H
