// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the Material K-3D object, which controls the surface appearance of rendered geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "material.h"

#include <k3dchannelbutton.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>
#include <k3drishadersetup.h>
#include <k3dshaderargumentspage.h>
#include <k3dshaderinfo.h>
#include <k3dshaderselector.h>
#include <k3dspinbutton.h>

#include <k3diapplication.h>
#include <k3dichannel.h>
#include <k3dimaterialcollection.h>
#include <k3dipath.h>
#include <k3dipathcollection.h>

#include <sdpgl.h>

namespace
{

/// Functor object for applying a material to paths / objects
class AssignMaterial :
	public k3dIPathVisitor
{
public:
	AssignMaterial(k3dIMaterial* const Material) :
		m_Material(Material)
	{
	}

	void VisitPath(k3dIPath* const Path)
	{
		if(SDP_S_OK == Path->State()->QueryState(k3dIState::VISIBLESELECTED))
			{
				k3d::Editing editing(Path);
				Path->Materials()->SetMaterial(m_Material);
			}
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		PathCollection->VisitPaths(this);
	}

	void operator()(k3dIMaterialCollection* const MaterialCollection)
	{
		MaterialCollection->SetMaterial(m_Material);
	}

private:
	k3dIMaterial* const m_Material;
};

/// Functor object for selecting paths / objects that match the given material
class SelectMaterial :
	public k3dIPathVisitor
{
public:
	SelectMaterial(k3dIMaterial* const Material) :
		m_Material(Material)
	{
		// Sanity checks ...
		g_assert(m_Material);
	}

	void VisitPath(k3dIPath* const Path)
	{
		if(m_Material == Path->Materials()->Material())
			Path->State()->SetState(k3dIState::VISIBLESELECTED, true);
	}

	void operator()(k3dIPathCollection* const PathCollection)
	{
		PathCollection->VisitPaths(this);
	}

	void operator()(k3dIMaterialCollection* const MaterialCollection)
	{
		if(m_Material == MaterialCollection->Material())
			{
				k3dIState* const state = dynamic_cast<k3dIState*>(MaterialCollection);
				if(state)
					state->SetState(k3dIState::VISIBLESELECTED, true);
			}
	}

private:
	k3dIMaterial* const m_Material;
};

} // namespace

////////////////////////////////////////////////////////////////////////////////////
// CMaterial

GLfloat CMaterial::m_BlackData[4] = {0.0f, 0.0f, 0.0f, 1.0f};
GLfloat CMaterial::m_WhiteData[4] = {1.0f, 1.0f, 1.0f, 1.0f};

const sdpString CMaterial::shadowtype_none = "None";
const sdpString CMaterial::shadowtype_opaque = "Opaque";
const sdpString CMaterial::shadowtype_opacity = "Opacity";
const sdpString CMaterial::shadowtype_shaded = "Shaded";


CMaterial::CMaterial(sdpIUnknown* Initializer) :
	CMaterialBase(Initializer),
	m_SurfaceShader("surface", Document(), Document()->StateRecorder()),
	m_DisplacementShader("displacement", Document(), Document()->StateRecorder()),
	m_ShadowType("shadowtype", shadowtype_opacity, Document()->StateRecorder(), 0),
	m_TrueDisplacement("truedisplacement", false, Document()->StateRecorder(), 0),
	m_TrueDisplacementBounds("truedisplacementbounds", 1.0, Document()->StateRecorder(), 0),
	m_Matte("matte", false, Document()->StateRecorder(), 0),
	m_CameraRays("camerarays", true, Document()->StateRecorder(), 0),
	m_ReflectionRays("reflectionrays", true, Document()->StateRecorder(), 0),
	m_Color("color", Document(), Document()->StateRecorder(), 0),
	m_Opacity("opacity", Document(), Document()->StateRecorder(), 0),
	m_LuminousColor("luminouscolor", Document(), Document()->StateRecorder(), 0),
	m_Luminosity("luminosity", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_SurfaceShader);
	AddVariable(&m_DisplacementShader);
	AddVariable(&m_ShadowType);
	AddVariable(&m_TrueDisplacement);
	AddVariable(&m_TrueDisplacementBounds);
	AddVariable(&m_Matte);
	AddVariable(&m_CameraRays);
	AddVariable(&m_ReflectionRays);
	AddVariable(&m_Color);
	AddVariable(&m_Opacity);
	AddVariable(&m_LuminousColor);
	AddVariable(&m_Luminosity);

	m_EditorTexture = 0;
	m_RiTexture = 0;
}

CMaterial::~CMaterial()
{
	m_Color.RemoveEventObserver(this);
	m_Opacity.RemoveEventObserver(this);
	m_LuminousColor.RemoveEventObserver(this);
	m_Luminosity.RemoveEventObserver(this);
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CMaterial::EditObject()
{
	new CMaterialDialog(this);
	return SDP_S_OK;
}

namespace
{

class RemoveFromCollection
{
public:
	void operator()(k3dIMaterialCollection* const Collection)
	{
		k3d::Editing editing(Collection);
		Collection->SetMaterial(0);
	}
};

} // namespace

sdpResult CMaterial::ObjectDeleted()
{
	MaterialCollections collections(GetMaterialCollections());
	std::for_each(collections.begin(), collections.end(), RemoveFromCollection());

	return CMaterialBase::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CMaterial::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);

	const sdpString event(Event);
	if(event == k3dIChannel::Event_Changed)
		{
			k3dRedrawAll(Document(), false);
			return SDP_S_OK;
		}

	return CMaterialBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

///////////////////////////////////////////////////////////////
// sdpsl::IShaderCollection implementation

void CMaterial::VisitShaders(sdpsl::IShaderVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

	if(m_SurfaceShader.Shader())
		Visitor->VisitShader(m_SurfaceShader.Shader());

	if(m_DisplacementShader.Shader())
		Visitor->VisitShader(m_DisplacementShader.Shader());
}

///////////////////////////////////////////////////////////////
// k3dIMaterial implementation

sdpResult CMaterial::SetAmbient(double Color[3])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CMaterial::SetDiffuse(double Color[3])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CMaterial::SetSpecular(double Color[3])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CMaterial::SetEmissive(double Color[3])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CMaterial::SetSpecularSize(double Size)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

sdpResult CMaterial::SetTransparency(double Transparency)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
}

/////////////////////////////////////////////////////////////////
// k3dIEditorMaterial implementation

sdpResult CMaterial::EditorMaterialPreRender(k3dIEditorEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	if(SDP_S_OK != Engine->EditorQueryTextured())
		return SDP_S_OK;

	k3dITexture* const texture = m_SurfaceShader.PrimaryTexture();
	m_EditorTexture = texture ? dynamic_cast<k3dIEditorTexture*>(texture) : 0;

	return SDP_S_OK;
}

sdpResult CMaterial::EditorMaterialSetup(k3dIEditorEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);

	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, m_BlackData);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, m_DiffuseData);
	glColor4fv(m_DiffuseData);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, m_WhiteData);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (GLfloat)64.0);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, m_EmissiveData);

	if((SDP_S_OK == Engine->EditorQueryTextured()) && m_EditorTexture)
		m_EditorTexture->EditorTextureSetup(Engine);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////
// k3dIRiMaterial implementation

sdpResult CMaterial::RiMaterialStartJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialSetup(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	// If the user hasn't selected a shader, use NULL ...
	if(0 == m_SurfaceShader.Shader())
		{
			Engine->RiSurfaceV("null", 0, 	0, 0, 0);
			return SDP_S_FALSE;
		}

	// Setup the surface shader ...
	RiMaterialDefaultsSetup(Engine);
	k3dRiSurfaceShader surface;
	m_SurfaceShader.InsertInto(Frame, Engine, m_CurrentTime, m_DeltaTime, surface);
	surface.Setup(Engine);

	// Setup the displacement shader (only if the user has selected a shader) ...
	if(m_DisplacementShader.Shader())
		{
			k3dRiDisplacementShader displacement;
			m_DisplacementShader.InsertInto(Frame, Engine, m_CurrentTime, m_DeltaTime, displacement);
			displacement.Setup(Engine);
		}

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialPostRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialFinishJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialDefaultsSetup(k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	// Set "render" attributes ...
	k3dRiParameterList renderattributes;

	// Set visibility parameters
	int visibility = 0;
	if(m_CameraRays.Value())
		visibility |= 1;

	if(m_ReflectionRays.Value())
		visibility |= 2;

	if(shadowtype_none != m_ShadowType.Value())
		visibility |= 4;

	renderattributes.AddTokenValuePair(new k3dRiTokenValuePair("visibility", new k3dRiIntegerArray("integer", visibility)));

	// Set shadow-casting parameters
	if(shadowtype_none == m_ShadowType.Value())
		renderattributes.AddTokenValuePair(new k3dRiTokenValuePair("casts_shadows", new k3dRiStringArray("string", "none")));
	else if(shadowtype_opaque == m_ShadowType.Value())
		renderattributes.AddTokenValuePair(new k3dRiTokenValuePair("casts_shadows", new k3dRiStringArray("string", "opaque")));
	else if(shadowtype_opacity == m_ShadowType.Value())
		renderattributes.AddTokenValuePair(new k3dRiTokenValuePair("casts_shadows", new k3dRiStringArray("string", "Os")));
	else if(shadowtype_shaded == m_ShadowType.Value())
		renderattributes.AddTokenValuePair(new k3dRiTokenValuePair("casts_shadows", new k3dRiStringArray("string", "shade")));
	else
		std::cerr << "CMaterial::RiMaterialDefaultsSetup(): Unknown shadow type [" << m_ShadowType.Value() << "]" << std::endl;

	// Enable / disable true displacement
	renderattributes.AddTokenValuePair(new k3dRiTokenValuePair("truedisplacement", new k3dRiIntegerArray("integer", m_TrueDisplacement.Value() ? 1 : 0)));

	// Make the render attributes happen ...
	renderattributes.SetAttributes("render", Engine);

	// Setup displacement bounds ...
	k3dRiParameterList displacementattributes;
	displacementattributes.AddTokenValuePair(new k3dRiTokenValuePair("sphere", new k3dRiRealArray("float", m_TrueDisplacementBounds.Value())));
	displacementattributes.AddTokenValuePair(new k3dRiTokenValuePair("coordinatesystem", new k3dRiStringArray("string", "world")));
	displacementattributes.SetAttributes("displacementbound", Engine);

	// Set "radiosity" attributes ...
	k3dRiParameterList radiosityattributes;

	// Set the radiant color
	sdpVector3 color;
	m_LuminousColor->Value3(m_CurrentTime, m_DeltaTime, color);
	color *= m_Luminosity->Value1(m_CurrentTime, m_DeltaTime);
	radiosityattributes.AddTokenValuePair(new k3dRiTokenValuePair("emissioncolor", new k3dRiRealArray("color", color)));

	radiosityattributes.SetAttributes("radiosity", Engine);

	// Set base color
	m_Color->Value3(m_CurrentTime, m_DeltaTime, color);
	Engine->RiColor(color);

	// Set opacity
	sdpVector3 opacity = m_Opacity->Value1(m_CurrentTime, m_DeltaTime);
	Engine->RiOpacity(opacity);

	// Set the matte attribute
	Engine->RiMatte(m_Matte.Value() ? 1 : 0);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate implementation

sdpResult CMaterial::PostAnimate()
{
	// Cache some current values for speed ...
	sdpVector3 color;

	m_Color->Value3(m_CurrentTime, m_DeltaTime, color);
	m_DiffuseData[0] = (GLfloat)color[0];
	m_DiffuseData[1] = (GLfloat)color[1];
	m_DiffuseData[2] = (GLfloat)color[2];
	m_DiffuseData[3] = (GLfloat)m_Opacity->Value1(m_CurrentTime, m_DeltaTime);

	m_LuminousColor->Value3(m_CurrentTime, m_DeltaTime, color);
	m_EmissiveData[0] = (GLfloat)color[0];
	m_EmissiveData[1] = (GLfloat)color[1];
	m_EmissiveData[2] = (GLfloat)color[2];
	m_EmissiveData[3] = 1.0f;

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CMaterial::InitComplete()
{
	k3d::PushEditing(Document()->Objects());

	if(0 == m_SurfaceShader.Shader())
		m_SurfaceShader.SetShader("k3d_plastic");

	if(0 == m_Color.Object())
		g_return_val_if_fail(m_Color.Create("Material diffuse color", sdpVector3(1.0), true, true), SDP_E_FAIL);

	if(0 == m_Opacity.Object())
		g_return_val_if_fail(m_Opacity.Create("Material opacity", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	if(0 == m_LuminousColor.Object())
		g_return_val_if_fail(m_LuminousColor.Create("Material luminous color", sdpVector3(0.0), true, true), SDP_E_FAIL);

	if(0 == m_Luminosity.Object())
		g_return_val_if_fail(m_Luminosity.Create("Material luminous intensity", 0.5, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	m_Color.AddEventObserver(this, k3dIChannel::Event_Changed, "CMaterial");
	m_Opacity.AddEventObserver(this, k3dIChannel::Event_Changed, "CMaterial");
	m_LuminousColor.AddEventObserver(this, k3dIChannel::Event_Changed, "CMaterial");
	m_Luminosity.AddEventObserver(this, k3dIChannel::Event_Changed, "CMaterial");

	k3d::PopEditing(Document()->Objects());
	return CMaterialBase::InitComplete();
}

namespace
{

/// Functor object for locating paths / objects that use a specific material
class GetMaterialServers :
	public k3dIObjectVisitor,
	public k3dIPathVisitor
{
public:
	GetMaterialServers(k3dIMaterial* const Material, CMaterial::MaterialCollections& Results) :
		m_Material(Material),
		m_Results(Results)
	{
	}

	void VisitObject(k3dIObject* const Object)
	{
		k3dIMaterialCollection* const materialcollection = dynamic_cast<k3dIMaterialCollection*>(Object);
		if(materialcollection)
			{
				if(materialcollection->Material() == m_Material)
					m_Results.push_back(materialcollection);
			}

		k3dIPathCollection* const pathcollection = dynamic_cast<k3dIPathCollection*>(Object);
		if(pathcollection)
			pathcollection->VisitPaths(this);
	}

	void VisitPath(k3dIPath* const Path)
	{
		if(Path->Materials()->Material() == m_Material)
			m_Results.push_back(Path->Materials());
	}

private:
	k3dIMaterial* const m_Material;
	CMaterial::MaterialCollections& m_Results;
};

} // namespace

CMaterial::MaterialCollections CMaterial::GetMaterialCollections()
{
	MaterialCollections results;
	GetMaterialServers functor(this, results);
	Document()->Objects()->VisitObjects(&functor);

	return results;
}

/////////////////////////////////////////////////////////////////////////////
// CMaterialDialog

const sdpString CMaterialDialog::dialogname = "properties";
const sdpString CMaterialDialog::control_shadowtype = "shadowtype";
const sdpString CMaterialDialog::control_truedisplacement = "truedisplacement";
const sdpString CMaterialDialog::control_truedisplacementbounds = "truedisplacementbounds";
const sdpString CMaterialDialog::control_matte = "matte";
const sdpString CMaterialDialog::control_camerarays = "camerarays";
const sdpString CMaterialDialog::control_reflectionrays = "reflectionrays";
const sdpString CMaterialDialog::control_assign = "assign";
const sdpString CMaterialDialog::control_remove = "remove";
const sdpString CMaterialDialog::control_select = "select";
const sdpString CMaterialDialog::control_color = "color";
const sdpString CMaterialDialog::control_opacity = "opacity";
const sdpString CMaterialDialog::control_luminouscolor = "luminouscolor";
const sdpString CMaterialDialog::control_luminosity = "luminosity";

CMaterialDialog::CMaterialDialog(CMaterial* const Object) :
	CMaterialDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "material.gtkml"));

//	AddCommand(control_shadowtype, control_shadowtype, CONTROL_EDITABLE);
	AddCommand(control_assign, control_assign, CONTROL_BUTTON);
	AddCommand(control_remove, control_remove, CONTROL_BUTTON);
	AddCommand(control_select, control_select, CONTROL_BUTTON);

	// Setup check buttons ...
	Connectk3dCheckButton(control_truedisplacement, m_Object->m_TrueDisplacement, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "True Displacment", 0, 0, 0);
	Connectk3dCheckButton(control_matte, m_Object->m_Matte, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Matte", 0, 0, 0);
	Connectk3dCheckButton(control_camerarays, m_Object->m_CameraRays, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Camera Rays", 0, 0, 0);
	Connectk3dCheckButton(control_reflectionrays, m_Object->m_ReflectionRays, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Reflection Rays", 0, 0, 0);

	// Connect spin buttons ...
	if(Getk3dSpinButton(control_truedisplacementbounds))
		Getk3dSpinButton(control_truedisplacementbounds)->Attach(m_Object->m_TrueDisplacementBounds, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "True Displacement Bounds", 0, 0, 0);

	// Setup shader selectors ...
	if(Getk3dShaderSelector("surfaceselector"))
		Getk3dShaderSelector("surfaceselector")->Attach(m_Object->Document(), &m_Object->m_SurfaceShader, true, sdpsl::IShader::SURFACE);

	if(Getk3dShaderSelector("displacementselector"))
		Getk3dShaderSelector("displacementselector")->Attach(m_Object->Document(), &m_Object->m_DisplacementShader, true, sdpsl::IShader::DISPLACEMENT);

	// Setup shader argument pages ...
	if(Getk3dShaderArgumentsPage("surfacearguments"))
		Getk3dShaderArgumentsPage("surfacearguments")->Attach(m_Object->Document(), &m_Object->m_SurfaceShader);

	if(Getk3dShaderArgumentsPage("displacementarguments"))
		Getk3dShaderArgumentsPage("displacementarguments")->Attach(m_Object->Document(), &m_Object->m_DisplacementShader);

	// Setup shader info ...
	if(Getk3dShaderInfo("surfaceinfo"))
		Getk3dShaderInfo("surfaceinfo")->Attach(m_Object->Document(), &m_Object->m_SurfaceShader);

	if(Getk3dShaderInfo("displacementinfo"))
		Getk3dShaderInfo("displacementinfo")->Attach(m_Object->Document(), &m_Object->m_DisplacementShader);

	// Setup channel buttons ...
	Connectk3dChannelButton(control_color, m_Object->m_Color.Interface());
	Connectk3dChannelButton(control_opacity, m_Object->m_Opacity.Interface());
	Connectk3dChannelButton(control_luminouscolor, m_Object->m_LuminousColor.Interface());
	Connectk3dChannelButton(control_luminosity, m_Object->m_Luminosity.Interface());

	// Initialize the "shadowtype" combo box ...
	sdpStringArray shadowtypes;
	shadowtypes.push_back(CMaterial::shadowtype_none);
	shadowtypes.push_back(CMaterial::shadowtype_opaque);
	shadowtypes.push_back(CMaterial::shadowtype_opacity);
	shadowtypes.push_back(CMaterial::shadowtype_shaded);

	sdpGtkCombo shadowtype = Combo(control_shadowtype);
	shadowtype.SetList(shadowtypes);
	shadowtype.SetEntryText(m_Object->m_ShadowType.Value());
	MapEvent("changed", control_shadowtype, false, shadowtype.Entry(), true);

	CallUpdateControls();
	Show();
}

void CMaterialDialog::OnEvent(sdpGtkEvent* Event)
{
cerr << "Material select in" << endl;
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_shadowtype)
		OnShadowType();
	else if(Event->Name() == control_assign)
		OnAssign();
	else if(Event->Name() == control_remove)
		OnRemove();
	else if(Event->Name() == control_select)
		OnSelect();
	else
		CMaterialDialogBase::OnEvent(Event);
cerr << "Material select out" << endl;
}

void CMaterialDialog::OnShadowType()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_shadowtype, Combo(control_shadowtype).GetEntryText());
	k3dSetValue<sdpString>(m_Object->m_ShadowType, Combo(control_shadowtype).GetEntryText());
}

void CMaterialDialog::OnAssign()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_assign, 0);
	k3dStartStateChangeSet(m_Object->Document());

	k3d::for_each_if<k3dIMaterialCollection>(m_Object->Document()->Objects(), AssignMaterial(m_Object), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTED));
	k3d::for_each_if<k3dIPathCollection>(m_Object->Document()->Objects(), AssignMaterial(m_Object), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));

	k3dFinishStateChangeSet(m_Object->Document(), "Assign Material [" + sdpString(m_Object->Name()) + "]");

	k3dRedrawAll(m_Object->Document(), false);
}

void CMaterialDialog::OnRemove()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_remove, 0);
	k3dStartStateChangeSet(m_Object->Document());

	k3d::for_each_if<k3dIMaterialCollection>(m_Object->Document()->Objects(), AssignMaterial(0), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLESELECTED));
	k3d::for_each_if<k3dIPathCollection>(m_Object->Document()->Objects(), AssignMaterial(0), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));

	k3dFinishStateChangeSet(m_Object->Document(), "Remove Material");

	k3dRedrawAll(m_Object->Document(), false);
}

void CMaterialDialog::OnSelect()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_select, 0);

	k3d::for_each<k3dIMaterialCollection>(m_Object->Document()->Objects(), SelectMaterial(m_Object));
	k3d::for_each<k3dIPathCollection>(m_Object->Document()->Objects(), SelectMaterial(m_Object));

	k3dRedrawAll(m_Object->Document(), false);
}


