#ifndef MATERIAL_H
#define MATERIAL_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the Material K-3D object, which controls the surface appearance of rendered geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpersistobject.h>
#include <k3dshaderinstance.h>
#include <k3dobjectdialog.h>

#include <k3dimaterial.h>
#include <k3dieditormaterial.h>
#include <k3dieditortexture.h>
#include <k3dirimaterial.h>
#include <k3diritexture.h>
#include <k3ditexture.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

// Forward class declarations
class CMaterial;
class k3dIMaterialCollection;

/////////////////////////////////////////////////////////////////////////////
// CMaterialDialog

typedef k3dObjectDialog CMaterialDialogBase;

class CMaterialDialog :
	public CMaterialDialogBase
{
public:
	CMaterialDialog(CMaterial* const Object);

protected:
	CMaterial* const m_Object;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnShadowType();
	void OnAssign();
	void OnRemove();
	void OnSelect();

	static const sdpString dialogname;
	static const sdpString control_shadowtype;
	static const sdpString control_truedisplacement;
	static const sdpString control_truedisplacementbounds;
	static const sdpString control_matte;
	static const sdpString control_camerarays;
	static const sdpString control_reflectionrays;
	static const sdpString control_assign;
	static const sdpString control_remove;
	static const sdpString control_select;
	static const sdpString control_color;
	static const sdpString control_opacity;
	static const sdpString control_luminouscolor;
	static const sdpString control_luminosity;
};


//////////////////////////////////////////////////////////////////////
// CMaterial
typedef k3dPersistObject CMaterialBase;

/// Controls the surface appearance of rendered geometry
class CMaterial :
	public CMaterialBase,
	public k3dIMaterial,
	public k3dIEditorMaterial,
	public k3dIRiMaterial,
	public sdpsl::IShaderCollection
{
friend class CMaterialDialog;
public:
	CMaterial(sdpIUnknown* Initializer);
	~CMaterial();

	// k3dIObject implementation
	sdpResult EditObject();
	sdpResult ObjectDeleted();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// sdpsl::IShaderCollection implementation
	void VisitShaders(sdpsl::IShaderVisitor* const Visitor);

	// k3dIMaterial implementation
	sdpResult SetAmbient(double Color[3]);
	sdpResult SetDiffuse(double Color[3]);
	sdpResult SetSpecular(double Color[3]);
	sdpResult SetEmissive(double Color[3]);
	sdpResult SetSpecularSize(double Size); 
	sdpResult SetTransparency(double Transparency);

	// k3dIEditorMaterial implementation
	sdpResult EditorMaterialPreRender(k3dIEditorEngine* const Engine);
	sdpResult EditorMaterialSetup(k3dIEditorEngine* const Engine);

	// k3dIRiMaterial implementation
	sdpResult RiMaterialStartJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine);
	sdpResult RiMaterialPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiMaterialSetup(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiMaterialPostRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);
	sdpResult RiMaterialFinishJob(k3dIRenderJob* const Job, k3dIRiEngine* const Engine);

	// k3dIAnimate implementation
	sdpResult PostAnimate();

	// k3dIPersist implementation
	sdpResult InitComplete();

	/// Stores a collection of material collections
	typedef std::vector<k3dIMaterialCollection*> MaterialCollections;
	typedef MaterialCollections::iterator MaterialCollectionIterator;
	/// Returns a collection of material servers that reference this material
	MaterialCollections GetMaterialCollections();

protected:
	k3dShaderInstance m_SurfaceShader;
	k3dShaderInstance m_DisplacementShader;

	k3dVariable<sdpString> m_ShadowType;
	k3dVariable<bool> m_TrueDisplacement;
	k3dVariable<gdouble> m_TrueDisplacementBounds;
	k3dVariable<bool> m_Matte;
	k3dVariable<bool> m_CameraRays;
	k3dVariable<bool> m_ReflectionRays;

	k3dColorChannel m_Color;
	k3dScalarChannel m_Opacity;
	k3dColorChannel m_LuminousColor;
	k3dScalarChannel m_Luminosity;

	GLfloat m_AmbientData[4];
	GLfloat m_DiffuseData[4];
	GLfloat m_SpecularData[4];
	GLfloat m_EmissiveData[4];
	static GLfloat m_BlackData[4];
	static GLfloat m_WhiteData[4];

	k3dIEditorTexture* m_EditorTexture;
	k3dIRiTexture* m_RiTexture;

	sdpResult RiMaterialDefaultsSetup(k3dIRiEngine* const Engine);
	
	static const sdpString shadowtype_none;
	static const sdpString shadowtype_opaque;
	static const sdpString shadowtype_opacity;
	static const sdpString shadowtype_shaded;
};

#endif // MATERIAL_H
