// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTorus K-3D object, which renders a  RenderMan torus primitive
		\author Tim Shead (tshead@k-3d.com)
*/

#include "torus.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>
#include <k3dieditorengine.h>
#include <k3dimaterial.h>
#include <k3diriengine.h>
#include <k3dirimaterial.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

CTorus::CTorus(sdpIUnknown* Initializer) :
	CTorusBase(Initializer),
	m_Material("material", Document(), Document()->StateRecorder(), 0),
	m_MajorRadius("majorradius", Document(), Document()->StateRecorder(), 0),
	m_MinorRadius("minorradius", Document(), Document()->StateRecorder(), 0),
	m_PhiMin("phimin", Document(), Document()->StateRecorder(), 0),
	m_PhiMax("phimax", Document(), Document()->StateRecorder(), 0),
	m_SweepAngle("sweepangle", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_Material);
	
	AddVariable(&m_MajorRadius);
	AddVariable(&m_MinorRadius);
	AddVariable(&m_PhiMin);
	AddVariable(&m_PhiMax);
	AddVariable(&m_SweepAngle);
	
	m_Material.SetObject(k3dDefaultMaterial(Document()));
}

CTorus::~CTorus()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTorus::EditObject()
{
	new CTorusDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CTorus::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	sdpglTrace("CTorus::EditorRender()");

	const gdouble majorradius = fabs(m_MajorRadius->Value1(m_CurrentTime, m_DeltaTime));
	const gdouble minorradius = fabs(m_MinorRadius->Value1(m_CurrentTime, m_DeltaTime));

	// We're lazy, so just draw a bounding-box right now ...
	DrawBoundingBox(Engine, k3dBoundingBox3(sdpVector3(-(majorradius + minorradius), -minorradius, -(majorradius + minorradius)), sdpVector3(majorradius + minorradius, minorradius, majorradius + minorradius)));	

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, majorradius + minorradius);

	return SDP_S_OK;	
}

sdpResult CTorus::RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);
	
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;
		
	// If we don't have a material, we're done ...
	if(0 == m_Material.Interface())
		return SDP_S_FALSE;
	k3dIRiMaterial* const rimaterial = dynamic_cast<k3dIRiMaterial*>(m_Material.Object());
	if(0 == rimaterial)
		return SDP_S_FALSE;	
		
	const gdouble majorradius = fabs(m_MajorRadius->Value1(m_CurrentTime, m_DeltaTime));
	const gdouble minorradius = fabs(m_MinorRadius->Value1(m_CurrentTime, m_DeltaTime));
	const gdouble phimin = sdpDegrees(m_PhiMin->Value1(m_CurrentTime, m_DeltaTime));	
	const gdouble phimax = sdpDegrees(m_PhiMax->Value1(m_CurrentTime, m_DeltaTime));	
	const gdouble sweepangle = sdpDegrees(m_SweepAngle->Value1(m_CurrentTime, m_DeltaTime));

	Engine->RiTransformBegin();
	
	sdpMatrix4 matrix = sdpRotation3D(sdpRadians(-90.0), sdpVector3(1, 0, 0)).Transpose();
	Engine->RiConcatTransform(matrix);
	
	rimaterial->RiMaterialSetup(Frame, Engine);
	Engine->RiTorusV(majorradius, minorradius, phimin, phimax, sweepangle, 0, 0, 0, 0);
	
	Engine->RiTransformEnd();
		
	return SDP_S_OK;
}

k3dIMaterial* CTorus::Material()
{
	return m_Material.Interface();
}

sdpResult CTorus::SetMaterial(k3dIMaterial* Material)
{
	// We allow NULL here, as a synonym for the default material ...
	m_Material.SetObject(dynamic_cast<k3dIObject*>(Material));

	return SDP_S_OK;
}

sdpResult CTorus::InitComplete()
{
	if(0 == m_MajorRadius.Object())
		g_return_val_if_fail(m_MajorRadius.Create("Major Radius", 10.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_MinorRadius.Object())
		g_return_val_if_fail(m_MinorRadius.Create("Minor Radius", 5.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_PhiMin.Object())
		g_return_val_if_fail(m_PhiMin.Create("Min. Phi", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_PhiMax.Object())
		g_return_val_if_fail(m_PhiMax.Create("Max. Phi", sdpRadians(360.0), true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_SweepAngle.Object())
		g_return_val_if_fail(m_SweepAngle.Create("Sweep Angle", sdpRadians(360.0), true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);

	return CTorusBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CTorusDialog

const sdpString CTorusDialog::dialogname = "properties";
const sdpString CTorusDialog::control_majorradius = "majorradius";
const sdpString CTorusDialog::control_minorradius = "minorradius";
const sdpString CTorusDialog::control_phimin = "phimin";
const sdpString CTorusDialog::control_phimax = "phimax";
const sdpString CTorusDialog::control_sweepangle = "sweepangle";
const sdpString CTorusDialog::control_position = "position";
const sdpString CTorusDialog::control_orientation = "orientation";
const sdpString CTorusDialog::control_scale = "scale";
const sdpString CTorusDialog::control_target = "target";

CTorusDialog::CTorusDialog(CTorus* const Object) :
	CTorusDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("corelib", "torus.gtkml"));

	// Connect spin buttons
	Connectk3dChannelButton(control_majorradius, m_Object->m_MajorRadius.Interface());
	Connectk3dChannelButton(control_minorradius, m_Object->m_MinorRadius.Interface());
	Connectk3dChannelButton(control_phimin, m_Object->m_PhiMin.Interface());
	Connectk3dChannelButton(control_phimax, m_Object->m_PhiMax.Interface());
	Connectk3dChannelButton(control_sweepangle, m_Object->m_SweepAngle.Interface());

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}

void CTorusDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	CTorusDialogBase::OnEvent(Event);
}

