//
//  The BREP Library.
//  Copyright (C) 1996 Philippe Bekaert
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// C++ version: Romain Behar (romainbehar@yahoo.com), Feb 2002

#include "brep.h"
#include <stdio.h>


// Prints an informational, warning, error, fatal error  message
void BrepInfo(void* client_data, char* routine, char* text);
void BrepWarning(void* client_data, char* routine, char* text);
void BrepError(void* client_data, char* routine, char* text);
void BrepFatal(void* client_data, char* routine, char* text);


// Connect the wing to its starting vertex.
// The wing must have been connected to its contour before
extern void BrepConnectWingToVertex(BREP_WING *wing);



// Creates a new empty contour within the face
BREP_CONTOUR::BREP_CONTOUR(BREP_FACE* containingface)
{
	wings = NULL;
	face = containingface;

	if(containingface != NULL)
		face->ConnectContour(this);
}

// Various actions to be performed when a contour is specified completely. There
// is no iteration over the edges since the edges may occur in two contours. For
// edges, there are two constructor callbacks: CreateEdge, which is called when first
// inserting the edge in a contour, and CloseEdge callback, which is called when the
// edge is inserted a second time in a contour
void BREP_CONTOUR::CloseCallback(const bool done = false)
{
	// Notify the user that the contour is complete
	if(!done)
		DataCloseCallback(true);
}

// Release all storage associated with a contour and its contours,
// including edges and vertices if not used in other contours as well
BREP_CONTOUR::~BREP_CONTOUR()
{
	// Disconnect the contour from the containing face
	face->DisconnectContour(this);

	// Destroy its wings
	DestroyWings();
}


// Connect the wing as last wing in the contour
void BREP_CONTOUR::ConnectWing(BREP_WING* wing)
{
	wing->contour = this;

	if(wings == NULL)
		{
			// First wing in contour
			wing->next = wing->prev = wing;
			wings = wing;
		}
	else
		{
			// Not first wing
			wing->next = wings;
			wing->prev = wings->prev;
			wing->prev->next = wing->next->prev = wing;
		}
}

// Disconnect the wing from the contour. Use with care: the contour
// might contain holes after disconnecting the wing! wing->next
// and wing->prev are left untouched as these fields are needed
// when disconnecting an edge from its vertices.
void BREP_CONTOUR::DisconnectWing(BREP_WING* wing)
{
	if(wing->next == NULL || wing->prev == NULL)
		BrepError(wing->edge, "BrepDisconnectWingFromContour", "wing improperly connected to contour");

	if(wings == wing)
		{
			// The wing is the first wing in the contour
			if(wing->next == wing)
				// It is the only wing
				wings = NULL;
			else
				// Make the next wing the first wing of the contour
				wings = wing->next;
		}

	// Endpoint of wing->prev will in general not be the same as
	// startpoint of wing->next! A hole might be created in the contour
	if(wing->next != NULL)
		wing->next->prev = wing->prev;
	if(wing->prev != NULL)
		wing->prev->next = wing->next;

	wing->contour = NULL;
	wing->next = wing->prev = NULL;
}


// Creates a wing into the contour from vertex1 to vertex2. The
// inside of the face bounded by the contour is supposed to be on the
// left side. The wing being created is supposed to be the new last
// one in a contour, i.e. its starting vertex 'vertex1' must be the
// endvertex of the most recently created wing in the contour (if
// non-empty). It is the responsibility of the user to close the
// contour, i.e., the last edge in a contour should connect to the
// first one
BREP_WING* BREP_CONTOUR::CreateWing(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2)
{
	if((wings != NULL) && BrepEdgeOtherWing(wings->prev)->vertex != vertex1)
		{
			BrepError(this, "BrepCreateWing", "the starting point of a new wing should be the endpoint of the previous wing");
			return NULL;
		}

	// First create a wing that still needs to be connected with a contour
	BREP_WING* wing = BrepCreateWingWithoutContour(vertex1, vertex2);

	// Complain if the edge is not used in opposite sense in both directions
	if(wing->vertex != vertex1)
		BrepError(wing->edge, "BrepCreateWing", "edge should be used in opposite sense by the contours sharing it");

	// Connect the wing to the contour
	ConnectWing(wing);

	// Connect the wing to its start vertex
	// after it has been connected in a contour
	BrepConnectWingToVertex(wing);

	// Close the edge
	(wing->edge)->CloseCallback();

	return wing;
}


// Destroys the wings
void BREP_CONTOUR::DestroyWings()
{
	BREP_WING* first = wings;

	if(first == NULL)
		return;

	BREP_WING* prev;
	for(BREP_WING* wing = first->prev; wing != first; wing = prev)
		{
			prev = wing->prev;
			BrepDestroyWing(wing);
		}

	BrepDestroyWing(first);
}


// Execute func for every wing in the contour
void BREP_CONTOUR::IterateWings(void (*func)(BREP_WING*))
{
	BREP_WING* first = wings;

	if(first == NULL)
		return;

	BREP_WING* next = first;

	do
		{
			BREP_WING* wing = next;
			next = wing->next;
			func(wing);
		}
	while(next != first);
}

void BREP_CONTOUR::IterateWings1A(void (*func)(BREP_WING*, void*), void* parm)
{
	BREP_WING* first = wings;

	if(first == NULL)
		return;

	BREP_WING* next = first;

	do
		{
			BREP_WING* wing = next;
			next = wing->next;
			func(wing, parm);
		}
	while(next != first);
}

// Execute func for every vertex in the contour
void BREP_CONTOUR::IterateVertices(void (*func)(BREP_VERTEX*))
{
	BREP_WING* first = wings;

	if(first == NULL)
		return;

	BREP_WING* next = first;

	do
		{
			BREP_WING* wing = next;
			next = wing->next;
			func(wing->vertex);
		}
	while(next != first);
}


// Remove a vertex from a (closed) contour, the wings containing the vertex are
// removed as well (and their edges if not used in other contours as
// well). The vertex itself will also be removed when it is no longer
// used. If it is not the only vertex in the contour, a new wing connecting
// the neighbooring vertices is created, so the contour remains a loop
void BREP_CONTOUR::RemoveVertex(BREP_VERTEX *vertex)
{
	// Check for empty contour
	if(wings == NULL)
		{
			BrepError(this, "BrepContourRemoveVertex", "empty contour");
			return;
		}

	// If vertex is the only vertex in the contour, remove its self-edge
	if(wings->next == wings)
		if(wings->vertex != vertex)
			{
				BrepError(this, "BrepContourRemoveVertex", "vertex not belonging to contour or contour is not closed");
				return;
			}
		else
			BrepDestroyWing(wings);

	// Else, replace the wing arriving at and the wing leaving the
	// last occurence of the vertex in the contour by a single wing
	// connecting the previous and next vertex

	// Look for the last occurence of the vertex in the contour
	BREP_WING* wing = wings->prev;
	while(wing->vertex != vertex && wing != wings)
		wing = wing->prev;

	if(wing->vertex != vertex)
		BrepError(this, "BrepContourRemoveVertex", "vertex doesn't occur in contour");
	else
		BrepJoinWings(wing->prev, wing);
}


// Edge.cpp: routines for manipulating BREP_EDGEs and BREP_WINGs

// Creates an edge connecting vertex1 to vertex2.
// It is not connected to a contour
BREP_EDGE::BREP_EDGE(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2)
{
	wing1 = new BREP_WING(vertex1);
	wing2 = new BREP_WING(vertex2);

	wing1->edge = wing2->edge = this;
}

// Release all storage associated with an edge and its vertices
// if not used in other edges as well. The given edge must already
// be disconnected from its contours
BREP_EDGE::~BREP_EDGE()
{
	// Check if the edge is still being used in a contour.
	// If so, warn the world
	if(wing1->contour || wing2->contour)
		BrepWarning(this, "~BREP_EDGE", "Edge is still used in a contour");
}


// Initialize the wing (it still needs to be connected with a contour)
BREP_WING::BREP_WING(BREP_VERTEX* v) :
	vertex(v)
{
	prev = next = NULL;
	edge = NULL;
	contour = NULL;
}



// Creates a new empty face within the shell
BREP_FACE::BREP_FACE(BREP_SHELL* containingshell)
{
	shell = containingshell;

	if(shell != NULL)
		shell->ConnectFace(this);
}

// Various actions to be performed when a face is specified completely
void BREP_FACE::CloseCallback(const bool done = false)
{
	// Notify the user that the contours of the face
	// have been completely specified
	BREP_CONTOUR_RING::iterator contour;
	for(contour = outer_contour.begin(); contour != outer_contour.end(); contour++)
		(*contour)->CloseCallback();

	// Notify the user that the face is complete
	if(!done)
		DataCloseCallback(true);
}

// Release all storage associated with a face and its contours,
// including edges and vertices if not used in other faces as well
BREP_FACE::~BREP_FACE()
{
	// Disconnect the face from the containing shell
	if(shell != NULL)
		shell->DisconnectFace(this);

	// Destroy its contours
	BREP_CONTOUR_RING::iterator contour;
	for(contour = outer_contour.begin(); contour != outer_contour.end(); contour++)
		delete (*contour);
}

// Connect a contour
void BREP_FACE::ConnectContour(BREP_CONTOUR* contour)
{
	if(contour != NULL)
		outer_contour.push_back(contour);
}

// Disconnect a contour
void BREP_FACE::DisconnectContour(BREP_CONTOUR* contour)
{
	BREP_CONTOUR_RING::iterator contouriterator;

	contouriterator = find(outer_contour.begin(), outer_contour.end(), contour);
	if(contouriterator != outer_contour.end())
		{
			(*contouriterator)->face = NULL;
			outer_contour.erase(contouriterator);
		}
	//else
		// BrepInfo(this, "BrepDisconnectContourFromFace", "outer contour disconnected from face");
}



// Creates a new empty shell within the solid
BREP_SHELL::BREP_SHELL(BREP_SOLID* solid)
{
	if(solid != NULL)
		solid->ConnectShell(this);
}

// Perform various actions to be performed when a shell
// has been completely specified
void BREP_SHELL::CloseCallback(const bool done = false)
{
	// Notify the user that the faces of the shell have been completely specified
	BREP_FACE_RING::iterator face;
	for(face = faces.begin(); face != faces.end(); face++)
		(*face)->CloseCallback();

	// Notify the user that the shell is complete
	if(!done)
		DataCloseCallback(true);
}

// Release all memory associated to a shell and its contained faces
BREP_SHELL::~BREP_SHELL()
{
	// Disconnect the shell from the containing solid
	if(solid != NULL)
		solid->DisconnectShell(this);

	// Destroy its faces
	BREP_FACE_RING::iterator face;
	for(face = faces.begin(); face != faces.end(); face++)
		delete (*face);
}


// Connect a face to the shell
void BREP_SHELL::ConnectFace(BREP_FACE* face)
{
	faces.push_back(face);
}

// Disconnect the face from its containing shell
void BREP_SHELL::DisconnectFace(BREP_FACE* face)
{
	// Remove it from the list
	BREP_FACE_RING::iterator faceiterator;

	faceiterator = find(faces.begin(), faces.end(), face);
	if(faceiterator != faces.end())
		faces.erase(faceiterator);

	face->shell = NULL;
}


// Creates an empty solid
BREP_SOLID::BREP_SOLID()
{
	vertices = NULL;
}

// Performs various actions to be done when the boundary
// representation of the solid has been completely specified
void BREP_SOLID::CloseCallback(const bool done = false)
{
	// Notify the user that the shells are complete
	//BREP_SHELL_RING::iterator shell;
	//for(shell = shells.begin(); shell != shells.end(); shell++)
	//	(*shell)->CloseCallback();

	// Notify the user that the solid is complete
	if(!done)
		DataCloseCallback(true);
}

// Release all memory associated with the solid
BREP_SOLID::~BREP_SOLID()
{
	// Inverse operations of BrepCreateSolid, and in reverse order

	// Destroy shells in the solid
	BREP_SHELL_RING::iterator shell;
	for(shell = shells.begin(); shell != shells.end(); shell++)
		{
			delete (*shell);
			shells.erase(shell);
		}

	// Next dispose of the vertex octree
	BrepDestroyVertices(vertices);
}


// Connect a shell to the solid
void BREP_SOLID::ConnectShell(BREP_SHELL* shell)
{
	shell->solid = this;
	shells.push_back(shell);
}

// Disconnect a shell from the solid
void BREP_SOLID::DisconnectShell(BREP_SHELL* shell)
{
	BREP_SHELL_RING::iterator shelliterator;

	shelliterator = find(shells.begin(), shells.end(), shell);
	if(shelliterator != shells.end())
		shells.erase(shelliterator);

	shell->solid = NULL;
}


// Iterate though all contours inside the solid
void BREP_SOLID::IterateContours(void (*func)(BREP_CONTOUR*))
{
	BREP_SHELL_RING::iterator shell;
	for(shell = shells.begin(); shell != shells.end(); shell++)
		{
			BREP_FACE_RING::iterator face;
			for(face = (*shell)->faces.begin(); face != (*shell)->faces.end(); face++)
				{
					BREP_CONTOUR_RING::iterator contour;
					for(contour = (*face)->outer_contour.begin(); contour != (*face)->outer_contour.end(); contour++)
						func(*contour);
				}
		}
}


// wing1 and wing2 should be two wings belonging to the same contour.
// This routine creates a new edge connecting the start vertex of wing1
// and wing2, splitting the contour in two. A pointer to the newly created
// wing, connecting the start vertex of wing1 to the start vertex of wing2,
// is returned
BREP_WING* BrepMakeEdgeSplitContour(BREP_WING *wing1, BREP_WING *wing2)
{
  // Test whether the wings belong to the same contour
  if (wing1->contour != wing2->contour) {
    BrepError(wing1->edge, "BrepMakeEdgeSplitContour", "wings must belong to the same contour");
    return NULL;
  }

	// Create an edge connecting the two start vertices
	BREP_EDGE* edge = CreateEdge(wing1->vertex, wing2->vertex);
	BREP_WING* winga = edge->wing1;
	BREP_WING* wingb = edge->wing2;

  // connect the wings of the newly created edge to the contour,
//   * effectively breaking the contour in two loops: one starting
//   * with wing1 and one starting with wing2.
  winga->prev = wing1->prev;
  winga->next = wing2;
  wingb->prev = wing2->prev;
  wingb->next = wing1;
  winga->prev->next = winga->next->prev = winga;
  wingb->prev->next = wingb->next->prev = wingb;
  winga->contour = wingb->contour = wing1->contour;

  // Let the old contour point to the wing1 loop
  wing1->contour->wings = wing1;

  // Create a new contour for the wing2 loop
  wing2->contour = CreateContour(wing1->contour->face);
  wing2->contour->wings = wing2;

	// Let all wings in the wing2 loop point to the new contour
	BREP_WING* wing = wing2;
  do {
    wing->contour = wing2->contour;
    wing = wing->next;
  } while (wing != wing2);

  // Connect the newly created wings to their start vertices
  BrepConnectWingToVertex(winga);
  BrepConnectWingToVertex(wingb);

	// Notify the user that the edge has
	// been connected into two contours
	edge->CloseCallback();

	// Notify the user that two new contours have been completely specified
	//BrepUpdateContour(wing1->contour);
	(wing2->contour)->CloseCallback();
	(wing1->contour)->CloseCallback();

  // Return a pointer to the wing connecting wing1->vertex to wing2->vertex
  return winga;
}

// Wing1 and wing2 are two wings belonging to a different contour. This
// routine creates a new edge connecting the starting vertices of
// wing1 and wing2 and merges to two contours to one. A pointer to
// the newly created wing, connecting the start vertex of wing1 to the
// start vertex of wing2, is returned
BREP_WING *BrepMakeEdgeJoinContours(BREP_WING *wing1, BREP_WING *wing2)
{
  // Test whether the wings belong to a different contour
  if (wing1->contour == wing2->contour) {
    BrepError(wing1->edge, "BrepMakeEdgeJoinContours", "wings must belong to a different contour");
    return NULL;
  }

  // Test whether the wings belong to the same face
  if (wing1->contour->face != wing2->contour->face) {
    BrepWarning(wing1->edge, "BrepMakeEdgeJoinContours", "joining contours from different faces");
    // It's not an error ...
  }

	// Create an edge connecting the two start vertices
	BREP_EDGE* edge = CreateEdge(wing1->vertex, wing2->vertex);
	BREP_WING* winga = edge->wing1;
	BREP_WING* wingb = edge->wing2;

  // connect the wings of the newly created edge to the contour to which
//   * wing1 belongs.
  winga->prev = wing1->prev;
  winga->next = wing2;
  wingb->prev = wing2->prev;
  wingb->next = wing1;
  winga->prev->next = winga->next->prev = winga;
  wingb->prev->next = wingb->next->prev = wingb;

	// Unless the contour of wing2 is the outer contour of its face,
	// "move" its wings to the contour of wing1, and destroy the contour
  if (wing2->contour != wing2->contour->face->outer_contour.front()) {
    winga->contour = wingb->contour = wing1->contour;

    // Disconnect wing2->contour from its wings and destroy it
    wing2->contour->wings = NULL;
	delete (wing2->contour);

    // Let all wings in the former contour of wing2 point to the contour of wing1
	BREP_WING* wing;
    for (wing = wing2; wing->contour != wing1->contour; wing = wing->next)
      wing->contour = wing1->contour;
  } else {
    winga->contour = wingb->contour = wing2->contour;

    // Disconnect wing1->contour from its wings and destroy it
    wing1->contour->wings = NULL;
	delete (wing1->contour);

    // Let all wings in the former contour of wing2 point to the contour of wing1
	BREP_WING* wing;
    for (wing = wing1; wing->contour != wing2->contour; wing = wing->next)
      wing->contour = wing2->contour;
  }

  // Connect the newly created wings to their start vertices
  BrepConnectWingToVertex(winga);
  BrepConnectWingToVertex(wingb);

	// Notify the user that the edge has
	// been connected into two contours
	edge->CloseCallback();

	// Notify the user that the contour has been updated.
	//BrepUpdateContour(wing1->contour);
	(winga->contour)->CloseCallback();

  // Return a pointer to the wing connecting wing1->vertex to wing2->vertex
  return winga;
}

// Wing is a wing of an edge that is used in two different contours. This routine
// deletes the edge to which the wing belongs, inserting the reaminder of
// the "other" contour sharing the edge into the contour to which the wing
// belongs. The other contour is destroyed and the first is "closed" again
// in order to notify that it changed. Returns the wing from the first contour
// that comes before the first wing moved from the second to the first, or
// NULL if the first contour was a selfloop. In that case, the empty contour
// is deleted. In case of errors, wing is returned. Warning: if the face to which
// the other contour belongs, contains holes, these holes should probably be moved
// to the face to which 'wing' belongs.
BREP_WING *BrepDeleteEdgeJoinContours(BREP_WING *wing)
{
  BREP_WING *owing = BrepEdgeOtherWing(wing);
  BREP_WING *prev_wing = wing->prev, *next_wing = wing->next;
  BREP_CONTOUR *contour = wing->contour, *ocontour = owing->contour;

  if (!ocontour || ocontour == wing->contour) {
    BrepWarning(wing->edge, "BrepDeleteEdgeJoinContours", "Edge not enclosed in two different contours");
    return wing;
  }

  if (!prev_wing || !next_wing || !contour) {
    BrepError(wing->edge, "BrepDeleteEdgeJoinContours", "wing not properly connected in contour");
    return wing;
  }

	if(wing->prev == wing)
		{
			// Remove the selfloop from the other contour
			if(owing->next == owing)
				delete ocontour;
			else
				BrepDestroyWing(owing);

			// Delete the first contour (that contained only this wing)
			delete contour;

			return NULL;
		}

  // Disconnect and delete the wing from its contour
  BrepDestroyWing(wing);

  if (owing->next != owing) {
    // move all the wings of ocontour to contour, insert after prev_wing and before
//     * next_wing
    BREP_WING *wing1 = owing->next, *wing2 = owing->prev;
    owing->next = owing->prev = owing;
    ocontour->wings = owing;	// Will be destroyed soon ...

    prev_wing->next = wing1;
    wing1->prev = prev_wing;

    next_wing->prev = wing2;
    wing2->next = next_wing;

    for (wing=wing1; wing!=wing2; wing=wing->next)
      wing->contour = contour;
    wing2->contour = contour;
  }

	// Destroy the other contour
	delete ocontour;

	// Notify that the first contour has changed.
	contour->CloseCallback();

  // Last wing before the first inserted from the other contour.
  return prev_wing;
}

// 'wing' is a wing of an edge that is used two times (in opposite direction)
// in the same contour, i.o.w. a seam. This routine deletes this edge.
// If the wings happen to be consequtive wings, the wings are just deleted.
// If these are the only two wings in the contour, the contour is deleted as
// well and a NULL wing pointer is returned. If they are not consequtive, the
// contour is split in two contours. The wing before 'wing' will remain in
// the original contour and will be returned.
BREP_WING *BrepDeleteEdgeSplitContour(BREP_WING* wing)
{
  BREP_WING *owing;

  if (!wing || !wing->contour || BrepEdgeOtherWing(wing)->contour != wing->contour) {
    BrepWarning(wing->edge, "BrepDeleteEdgeSplitContour", "argument should be a seam");
    return wing;
  }

  owing = BrepEdgeOtherWing(wing);
	if(owing == wing->next || owing == wing->prev)
		{
			// Consequtive wings, the seam is a notch or a slit
			if(wing->next->next == wing)
				{
					// They are the only two wings in the contour
					delete (wing->contour);
					return NULL;
				} else {
      // Simply destroy the two wings. No new contour is split off.
      BREP_WING *prevwing = (owing->next == wing) ? owing->prev : wing->prev;
      BrepDestroyWing(wing);
      BrepDestroyWing(owing);
      (prevwing->contour)->CloseCallback();
      return prevwing;
    }
  } else {
    BREP_CONTOUR *contour = wing->contour, *split_contour;
    BREP_WING *first1 = owing->next, *last1 = wing->prev,
              *first2 = wing->next, *last2 = owing->prev;

    // Disconnect wing and owing from the contour + destroy
    BrepDestroyWing(wing);
    BrepDestroyWing(owing);

    // Close the loop first1 ... last1
    first1->prev = last1;
    last1->next = first1;
    contour->wings = last1;
	contour->CloseCallback();

    // Close the second loop 'first2' ... 'last2' and create a separate contour
    split_contour = CreateContour(contour->face);
    first2->prev = last2;
    last2->next = first2;
    split_contour->wings = first2;

	// For all wings in split_contour ...
	if(split_contour != NULL && split_contour->wings != NULL)
		{
			BREP_WING* first = split_contour->wings;
			BREP_WING* iwing;
			BREP_WING* next;
			for(iwing = first, next = iwing->next; iwing != NULL; iwing = (next == first) ? NULL : next, next = next->next)
				iwing->contour = split_contour;
		}

			split_contour->CloseCallback();

			return last1;
		}

  return wing;	// We never get here ....
}

// Creates a slit, connecting the two specified vertices,
// in the given face. A slit is a new contour containing
// two wings: one from v1 to v2 and the second
// from v2 back to v1. A pointer to the newly created
// wing, connecting v1 to v2, is returned
BREP_WING* BrepMakeSlit(BREP_FACE *face, BREP_VERTEX *v1, BREP_VERTEX *v2)
{
	// Create a new contour in the face
	BREP_CONTOUR* contour = CreateContour(face);

	// Create an edge connecting the two vertices
	BREP_EDGE* edge = CreateEdge(v1, v2);
	BREP_WING* winga = edge->wing1;
	BREP_WING* wingb = edge->wing2;

  // connect the wings of the newly created edge to the newly created
//   * contour and to each other
  winga->prev = winga->next = wingb;
  wingb->prev = wingb->next = winga;
  winga->contour = wingb->contour = contour;
  contour->wings = winga;

  // Connect the newly created wings to their start vertices
  BrepConnectWingToVertex(winga);
  BrepConnectWingToVertex(wingb);

	// Notify the user that the edge has
	// been connected into two contours
	edge->CloseCallback();

	// Notify the user that the new contour has been completely specified
	contour->CloseCallback();

  // Return the wing connecting wing->vertex to vertex
  return winga;
}

// Creates a notch, connecting the start vertex of the wing with
// the specified vertex, and inserts it into the contour to
// which the wing belongs. A pointer to the newly created
// wing, from the start vertex of the specified wing to the
// specified vertex, is returned
BREP_WING* BrepMakeNotch(BREP_WING* wing, BREP_VERTEX* vertex)
{
	// Create an edge connecting the start vertex
	// of the wing to the specified vertex
	BREP_EDGE* edge = CreateEdge(wing->vertex, vertex);
	BREP_WING* winga = edge->wing1;
	BREP_WING* wingb = edge->wing2;

  // Connect the wings of the newly created edge to the contour
  winga->prev = wing->prev;
  winga->next = wingb;
  wingb->prev = winga;
  wingb->next = wing;
  winga->prev->next = winga->next->prev = winga;
  wingb->prev->next = wingb->next->prev = wingb;
  winga->contour = wingb->contour = wing->contour;

  // Connect the newly created wings to their start vertices
  BrepConnectWingToVertex(winga);
  BrepConnectWingToVertex(wingb);

  // notify the user that the edge has been connected into two
	// contours.
	edge->CloseCallback();

  // Notify the user that the contour has been updated.
  // BrepUpdateContour(wing->contour);

  // Return the wing connecting wing->vertex to vertex
  return winga;
}


// Creates a wing from vertex1 to vertex2. First looks for an
// existing edge connecting the vertices. Creates such an edge if
// if there is not yet one. Returns a pointer to an unused wing in
// the edge. If there is no unused wing in this edge, complains, and
// creates a second edge between the two vertices.
// The wing still needs to be connected in a contour
BREP_WING* BrepCreateWingWithoutContour(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2)
{
	BREP_WING *wing;

	// First look if there is already an edge
	// connecting the two vertices
	BREP_EDGE* edge = BrepFindEdge(vertex1, vertex2);
	if(edge == NULL)
		{
			// There was not yet such an edge
			// Create such an edge
			edge = CreateEdge(vertex1, vertex2);

			// The wing to be used is the first wing --- vertex1
			// is the startvertex for the wing to be created
			wing = edge->wing1;
		}
	else
		{
			// There was already such an edge
			if(edge->wing1->vertex == vertex1)
				// Vertex1 is startvertex in the first wing?
				wing = edge->wing1;
			else if(edge->wing2->vertex == vertex1)
				// Vertex1 is startvertex in the second wing
				wing = edge->wing2;
			else
				{
					// Shouldn't happen
					BrepFatal(edge, "BrepCreateWingWithoutContour", "something impossible wrong here");
					// Useless, but compiler warning
					return NULL;
				}

			// Is the wing found already used in a contour?
			if(wing->contour)
				{
					BrepInfo(edge, "BrepCreateWing", "attempt to use an edge two times in the same sense - creating a duplicate edge");
					edge = CreateEdge(vertex1, vertex2);
					wing = edge->wing1;
				}
		}

	return wing;
}


// Splits only one wing. If a full edge is to be split, also its
// wing in the other contour (if any) needs to be split at the same vertex.
// (BrepSplitEdge does this).
// Returns the wing leaving at the inserted vertex
BREP_WING* BrepSplitWing(BREP_WING* wing, BREP_VERTEX* vertex)
{
	// Create two new wings, not connected to the contour
	BREP_WING* wing1 = BrepCreateWingWithoutContour(wing->vertex, vertex);
	BREP_WING* wing2 = BrepCreateWingWithoutContour(vertex, wing->next->vertex);

	// Now connect the pair of consecutive wings to the contour
	wing1->contour = wing2->contour = wing->contour;

	wing1->next = wing2;
	wing2->prev = wing1;

	if(wing->prev != wing)
		{
			// Wing is not the only wing in the contour
			wing1->prev = wing->prev;
			wing2->next = wing->next;
		}
	else
		{
			// Wing is the only wing in the contour
			wing1->prev = wing2;
			wing2->next = wing1;
		}

	// Destroy the original wing, leaves the contour with
	// a hole in it or even a contour without wings ...
	BrepDestroyWing(wing);

	// ... fill the hole
	wing1->prev->next = wing1;
	wing2->next->prev = wing2;

	// And make sure the contour knows about the news wings
	if(wing1->contour->wings == NULL)
		wing1->contour->wings = wing1;

	// Finally, connect the new wings to their startvertex
	BrepConnectWingToVertex(wing1);
	BrepConnectWingToVertex(wing2);

	// Close the edges
	wing1->edge->CloseCallback();
	wing2->edge->CloseCallback();

	return wing2;
}

// Joins two wings at there common endpoint. The wings are supposed to
// be consecutive wings, i.e. wing1->next = wing2 (and wing2->prev = wing1)
// in a closed contour. The common vertex is wing2->vertex,
// the start vertex of the second wing (which is supposed to be
// the endpoint of the first wing as well)
BREP_WING* BrepJoinWings(BREP_WING* wing1, BREP_WING* wing2)
{
	// Create a new wing connecting the startpoint
	// of wing1 with the endpoint of wing2
	BREP_WING* wing = BrepCreateWingWithoutContour(wing1->vertex, wing2->next->vertex);

	// Connect it to the contour
	wing->contour = wing1->contour;

	if(wing1->prev != wing2)
		{
			// The two wings were not the only two in the contour
			wing->prev = wing1->prev;
			wing->next = wing2->next;
		}
	else
		{
			// The two wings were the only two in the contour
			wing->prev = wing->next = wing;
		}

	// Destroy the two wings, this leaves a hole in the contour,
	// or if the wings were the only two, a contour without wings ...
	BrepDestroyWing(wing1);
	BrepDestroyWing(wing2);

	// ... fill the hole
	wing->prev->next = wing;
	wing->next->prev = wing;

	// Make sure the contour knows about the new wing
	if(wing->contour->wings == NULL)
		wing->contour->wings = wing;

	// Connect the new wing to its starvertex
	BrepConnectWingToVertex(wing);

	// Close the edge
	wing->edge->CloseCallback();

	return wing;
}

// Join two edges at their common endpoint. The edges must
// be used in the same contours and must share an endpoint
void BrepJoinEdges(BREP_EDGE *edge1, BREP_EDGE *edge2)
{
	bool do_wing0 = (edge1->wing1->contour != NULL);
	bool do_wing1 = (edge1->wing2->contour != NULL);

  // First join the wings on the edge1->wing1 side
  if (do_wing0) {
    if (edge1->wing1->next == edge2->wing1)
      BrepJoinWings(edge1->wing1, edge2->wing1);
    else if (edge1->wing1->next == edge2->wing2)
      BrepJoinWings(edge1->wing1, edge2->wing2);
    else if (edge1->wing1->prev == edge2->wing1)
      BrepJoinWings(edge2->wing1, edge1->wing1);
    else if (edge1->wing1->prev == edge2->wing2)
      BrepJoinWings(edge2->wing2, edge1->wing1);
    else
      BrepError(edge1, "BrepJoinEdges", "the two edges don't fit on the first side");
  }

  // Then join the wings on the edge1->wing2 side
  if (do_wing1) {
    if (edge1->wing2->next == edge2->wing1)
      BrepJoinWings(edge1->wing2, edge2->wing1);
    else if (edge1->wing2->next == edge2->wing2)
      BrepJoinWings(edge1->wing2, edge2->wing2);
    else if (edge1->wing2->prev == edge2->wing1)
      BrepJoinWings(edge2->wing1, edge1->wing2);
    else if (edge1->wing2->prev == edge2->wing2)
      BrepJoinWings(edge2->wing2, edge1->wing2);
    else
      BrepError(edge1, "BrepJoinEdges", "the two edges don't fit on the second side");
  }
}

// Replace the starting vertex of 'wing' by 'newvertex'. Assumes
// that 'wing' is correctly connected in a contour and it is not
// the only wing (that means: the contour is not a selfloop).
// Returns a pointer to the new wing, which is leaving at 'newvertex'
BREP_WING *BrepWingReplaceVertex(BREP_WING *wing, BREP_VERTEX *newvertex)
{
	if(wing->prev == NULL)
		{
			BrepError(wing->edge, "BrepWingReplaceVertex", "wing should be properly connected to a contour");
			return NULL;
		}

	if(wing->prev == wing)
		{
			BrepError(wing->edge, "BrepWingReplaceVertex", "wing should not be the only wing in the contour");
			return NULL;
		}

	// Create two new wings that will replace wing->prev and wing
	BREP_WING* wing1 = BrepCreateWingWithoutContour(wing->prev->vertex, newvertex);
	BREP_WING* wing2 = BrepCreateWingWithoutContour(newvertex, wing->next->vertex);

	// Connect the pair of new wings to the contour
	wing1->contour = wing2->contour = wing->contour;

	wing1->next = wing2;
	wing2->prev = wing1;

	if(wing->prev->prev != wing)
		{
			// There are more than two wings in the contour
			wing1->prev = wing->prev->prev;
			wing2->next = wing->next;
		}
	else
		{
			// There are only two wings in the contour: wing
			// and its predecessor, and they will be destroyed
			wing1->prev = wing2;
			wing2->next = wing1;
		}

	// Destroy the original wing and its predecessor in the contour, leaves
	// the contour with a hole in it or even a contour without wings ...
	BrepDestroyWing(wing->prev);
	BrepDestroyWing(wing);

	// ... fill the hole
	wing1->prev->next = wing1;
	wing2->next->prev = wing2;

	// And make sure the contour knows about the news wings.
	if(wing1->contour->wings == NULL)
		wing1->contour->wings = wing1;

	// Finally, connect the new wings to their startvertex
	BrepConnectWingToVertex(wing1);
	BrepConnectWingToVertex(wing2);

	// Close the edges
	wing1->edge->CloseCallback();

	return wing2;
}

// Returns the other wing in the edge
BREP_WING* BrepEdgeOtherWing(BREP_WING* wing)
{
	return (wing == wing->edge->wing1 ?
		wing->edge->wing2 :
		wing->edge->wing1);
}

// Iterate over a ring of wings
void BrepIterateWings(BREP_WING *wings, void (*func)(BREP_WING *))
{
  BREP_WING *first_wing, *wing, *next;

  if (wings) {
    next = first_wing = wings;
    do {
      wing = next;
      next = wing->next;
      func(wing);
    } while (next != first_wing);
  }
}


// Remove a wing from a contour, release the storage associated with the
// edge and its vertices if it is not used in other contours as well.
// Inverse of BrepCreateWing(), but also destroys unused vertices
void BrepDestroyWing(BREP_WING* wing)
{
	// First disconnect the wing from its starting vertex
	wing->vertex->DisconnectWing(wing);

	// Then disconnect the wing from its contour
	wing->contour->DisconnectWing(wing);

	// Destroy the edge if not being used in any contour anymore
	if(wing->edge->wing1->contour == NULL && wing->edge->wing2->contour == NULL)
		delete wing->edge;
}



// Connect the wing to its starting vertex
// (the wing must have been connected to its contour before)
void BrepConnectWingToVertex(BREP_WING* wing)
{
	if(wing != NULL)
		(wing->vertex)->wing_ring.push_back(wing);
}


// Looks whether an edge connecting the two vertices already
// exists, returns it if it exists, returns NULL if not
BREP_EDGE* BrepFindEdge(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2)
{
	BREP_WING_RING ring;
	BREP_WING_RING::iterator wingiterator;

	// There is an edge if there is a wing from vertex2 to vertex1 ...
	ring = vertex2->wing_ring;
	for(wingiterator = ring.begin(); wingiterator != ring.end(); wingiterator++)
		{
			BREP_WING* wing = *wingiterator;
			if(BrepEdgeOtherWing(wing)->vertex == vertex1)
				return wing->edge;
		}

	// ... or from vertex1 to vertex2
	ring = vertex1->wing_ring;
	for(wingiterator = ring.begin(); wingiterator != ring.end(); wingiterator++)
		{
			BREP_WING* wing = *wingiterator;
			if(BrepEdgeOtherWing(wing)->vertex == vertex1)
				return wing->edge;
		}

	// No edge connecting the two vertices was found
	return NULL;
}

// Iterator over all wings (included in a contour)
// connecting the two given vertices
void BrepIterateWingsWithVertex(BREP_VERTEX *vertex1, BREP_VERTEX *vertex2,
				void (*func)(BREP_WING*))
{
	BREP_WING_RING::iterator wingiterator;

	for(wingiterator = vertex2->wing_ring.begin(); wingiterator != vertex2->wing_ring.end(); wingiterator++)
		func(*wingiterator);

	for(wingiterator = vertex1->wing_ring.begin(); wingiterator != vertex1->wing_ring.end(); wingiterator++)
		func(*wingiterator);
}


// Returns the next wing with same starting vertex as wing.
// The order is counterclockwise if the vertex is used in
// outer contours only and clockwise if used in inner contours
// only. Returns a null pointer if the current wing is not
// properly connected to possible other wings sharing the
// vertex, or the boundary of a partial shell was reached
BREP_WING* BrepNextWingLeavingVertex(BREP_WING* wing)
{
	if(!wing->prev)
		return NULL;

	BREP_WING* next_wing = BrepEdgeOtherWing(wing->prev);

	if(next_wing->contour)
		return next_wing;

	return NULL;
}



// Creates a new vertex - no edges in which the vertex is used are specified,
// this is done when creating an edge
BREP_VERTEX::BREP_VERTEX()
{
	// A vertex is completely specified by its 3D coordinates
	//CloseCallback();
}

// Release all storage associated with the vertex. The vertex is supposed not
// to be used anymore in any edge. Disconnecting vertices and edges happens
// automatically when destroying edges
BREP_VERTEX::~BREP_VERTEX()
{
	// check whether the vertex is still being used in edges
	if(!wing_ring.empty())
		BrepInfo(this, "BrepDestroyVertex", "vertex still being used in edges!");
}


// Disconnect the wing from its starting vertex
// (the wing must be properly connected to a contour)
void BREP_VERTEX::DisconnectWing(BREP_WING* wing)
{
	// Sanity check...
	// Looks whether the wing is referenced in
	// the wing ring of its starting vertex
	BREP_WING_RING::iterator thewing;
	thewing = find(wing_ring.begin(), wing_ring.end(), wing);
	if(thewing == wing_ring.end())
		{
			BrepError(wing->edge, "BrepDisconnectWingFromVertex", "wing not connected to the vertex");
			return;
		}

	// Remove the wing element from the vertex wing ring
	wing_ring.erase(thewing);
}



// Attaches an existing BREP_VERTEX to the vertex octree
void BrepAttachVertex(BREP_VERTEX* vertex, BREP_VERTEX_OCTREE* vertices)
{
	vertices->AddWithDuplicates(vertex, VertexCompare);
}

// Looks up a vertex in the vertex octree, return NULL if not found
BREP_VERTEX* BrepFindVertex(BREP_VERTEX* vertex_data, BREP_VERTEX_OCTREE* vertices)
{
	return vertices->Find(vertex_data, VertexCompare);
}

// Looks up the first vertex in the vertex octree at the location as specified in the
// given vertex data. There may be multiple vertices at the same location (e.g. having
// a different normal and/or a different name). These vertices should normally be
// stored as a suboctree of the octree containing all vertices. A pointer to the
// top of this suboctree is returned if there are vertices at the given location.
// NULL is returned if there are no vertices at the given location
BREP_VERTEX_OCTREE* BrepFindVertexAtLocation(BREP_VERTEX* vertex_data, BREP_VERTEX_OCTREE* vertices)
{
	return vertices->FindSubtree(vertex_data, VertexCompareLocation);
}

// Calls func for each BREP_VERTEX in the octree
void BrepIterateVertices(BREP_VERTEX_OCTREE* vertices, void (*func)(BREP_VERTEX*))
{
	if(vertices != NULL)
		vertices->Iterate(func);
}

// Destroys all the vertices and the vertex octree
static void BrepDestroyVertex(BREP_VERTEX* vertex)
{ delete vertex; }
void BrepDestroyVertices(BREP_VERTEX_OCTREE* vertices)
{
	BrepIterateVertices(vertices, BrepDestroyVertex);
	delete vertices;
}

static void BrepIterateWingsWithVertices(BREP_VERTEX *v2, BREP_VERTEX_OCTREE *v1octree,
					 void (*func)(BREP_WING*))
{
	v1octree->Iterate2A(BrepIterateWingsWithVertex, v2, (void (*)(BREP_WING*))func);
}

// Iterator over all edge-wings between vertices at
// the locations specified by v1data and v2data
void BrepIterateWingsBetweenLocations(BREP_VERTEX* v1, BREP_VERTEX* v2,
	BREP_VERTEX_OCTREE* vertices, void (*func)(BREP_WING*))
{
	BREP_VERTEX_OCTREE* v1octree = BrepFindVertexAtLocation(v1, vertices);
	if(!v1octree)
		return;

	BREP_VERTEX_OCTREE* v2octree = BrepFindVertexAtLocation(v2, vertices);
	if(!v2octree)
		return;

	v2octree->Iterate2A(BrepIterateWingsWithVertices, v1octree, func);
}


// The default message callback routines ignore the client data
// Default callback for showing informational messages
static void brep_info(void *client_data, char *message)
{
#ifndef DEBUG
  // Don't show purely informational messages
#else // DEBUG
	std::cerr << message << std::endl;
#endif // DEBUG
}

// Default callback for showing warning messages
static void brep_warning(void *client_data, char *message)
{
	std::cerr << "warning: " << message << std::endl;
}

// Default callback for showing error messages
static void brep_error(void *client_data, char *message)
{
	std::cerr << "error: " << message << std::endl;
}

// Default callback for showing fatal error messages
static void brep_fatal(void *client_data, char *message)
{
	std::cerr << "fatal error: " << message << std::endl;
	exit(-1);
}

static BREP_MSG_CALLBACK_FUNC brep_info_callback_func = brep_info,
                              brep_warning_callback_func = brep_warning,
                              brep_error_callback_func = brep_error,
                              brep_fatal_callback_func = brep_fatal;

// The SetCallback functions return the previously set callback function
BREP_MSG_CALLBACK_FUNC BrepSetInfoCallback(BREP_MSG_CALLBACK_FUNC func)
{
  BREP_MSG_CALLBACK_FUNC old_func = brep_info_callback_func;
  brep_info_callback_func = func;
  return old_func;
}

BREP_MSG_CALLBACK_FUNC BrepSetWarningCallback(BREP_MSG_CALLBACK_FUNC func)
{
  BREP_MSG_CALLBACK_FUNC old_func = brep_warning_callback_func;
  brep_warning_callback_func = func;
  return old_func;
}

BREP_MSG_CALLBACK_FUNC BrepSetErrorCallback(BREP_MSG_CALLBACK_FUNC func)
{
  BREP_MSG_CALLBACK_FUNC old_func = brep_error_callback_func;
  brep_error_callback_func = func;
  return old_func;
}

BREP_MSG_CALLBACK_FUNC BrepSetFatalCallback(BREP_MSG_CALLBACK_FUNC func)
{
  BREP_MSG_CALLBACK_FUNC old_func = brep_fatal_callback_func;
  brep_fatal_callback_func = func;
  return old_func;
}

// Prints an informational message
void BrepInfo(void* client_data, char* routine, char* text)
{
	char buf[BREP_MAX_MESSAGE_LENGTH];

	sprintf(buf, text);

#ifdef DEBUG
	if(routine != NULL)
		sprintf(buf+strlen(buf), " (%s)", routine);
#endif

	brep_info_callback_func(client_data, buf);
}

// Prints a warning message
void BrepWarning(void* client_data, char* routine, char* text)
{
	char buf[BREP_MAX_MESSAGE_LENGTH];

	sprintf(buf, "%s", text);

#ifdef DEBUG
	if(routine != NULL)
		sprintf(buf+strlen(buf), " (%s)", routine);
#endif

	brep_warning_callback_func(client_data, buf);
}

// Prints an error message
void BrepError(void* client_data, char* routine, char* text)
{
	char buf[BREP_MAX_MESSAGE_LENGTH];

	sprintf(buf, "%s", text);

#ifdef DEBUG
	if(routine != NULL)
		sprintf(buf+strlen(buf), " (%s)", routine);
#endif

	brep_error_callback_func(client_data, buf);
}

// Prints a fatal error message
void BrepFatal(void* client_data, char* routine, char* text)
{
	char buf[BREP_MAX_MESSAGE_LENGTH];

	sprintf(buf, "%s", text);

#ifdef DEBUG
	if(routine != NULL)
		sprintf(buf+strlen(buf), " (%s)", routine);
#endif

	brep_fatal_callback_func(client_data, buf);
}


