//
//  The BREP Library.
//  Copyright (C) 1996 Philippe Bekaert
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// C++ version: Romain Behar (romainbehar@yahoo.com), Feb 2002

// brep.h: An edge-based Boundary Representation of solids, similar
//         to the winged-edge data structure

#ifndef _BREP_H_
#define _BREP_H_

#include <list>
#include <algorithm>
#include <stdlib.h>


// ****************** The Topological Data Structure ********************
//
// References:
//
// - Kevin Weiler, "Edge-based Data Structures for Solid Modeling in
//   Curved-Surface Environments", IEEE Computer graphics and Applications,
//   January, 1985
// - Martti Mantyla and Reijo Sulonen, "GWB: A Solid Modeler with Euler
//   Operators", IEEE Computer graphics and Applications, September 1982
// - Tony C. Woo, "A Combinatorial Analysis of Boundary Data Structure
//   Schemata", IEEE Computer graphics and Applications, March 1985
// - Andrew S. Glassner, "Maintaining Winged-Edge Models", Graphics Gems II,
//   James Arvo (editor), Academic Press, 1991
// - Mark Segal and Carlo H. Sequin, "Partitioning Polyhedral Objects into
//   Nonintersecting Parts", IEEE Computer graphics and Applications, January,
//   1988


// The boundary of a solid consists of one or more SHELLs
class BREP_SOLID;

// A shell is basically a collection of connected faces representing an
// orientable two-manifold. This means: a two-dimensional surface (in 3D
// space here), where each point on the surface has a neighbourhood
// topologically equivalent to an open disk. The surface must be orientable
// and closed. Orientable means that it is two-sided. There must be a clear
// distinction between its inside and its outside. The data structure
// here only represent the (adjacency) topology of a surface. It is up
// to the user of this data structure and library funtions to fill in
// the geometry information and make sure that it is geometrically
// consistent. That means that the assigned shape should be consistent with
// the topology

class BREP_SHELL;
typedef std::list<BREP_SHELL*> BREP_SHELL_RING;

// A face is a bounded open subset of a plane, bounded by one or more
// contours. One contour defines the outside of the face. Other contours
// describe its cavities. The contours describing cavities are called
// rings. The orientation of the outer contour must be counterclockwise as
// seen from the direction of the normal to the containing plane. Rings
// must be oriented clockwise. The cross product of a contour edge direction
// and the plane normal is always a vector that points to the outside of
// a face. Or in other words, the inside of a face is always on the left side
// of a contour. The geometric data supplied by the user must conform with
// this rule. The rings, if any, are pointed to by the outer contour.
// When building a face, the first contour being specified must be the
// outer contour (at least if your application relies on this
// interpretation)

class BREP_FACE;
typedef std::list<BREP_FACE*> BREP_FACE_RING;

// A contour represents a single closed planar polygonal curve (a loop).
// Self-intersecting contours are not allowed. Different contours may only
// intersect at common vertices or edges if at all. Here, we provide pointers
// to WINGs instead of EDGEs in order to have implicit information about
// the orientation of an edge as used in a contour. This facilitates
// iterating over the edges or vertices in a contour in a consistent manner.
// When building a contour, the edges must be specified in the right order:
// the inside of the face is always to the left of an edge. In a two-manifold,
// outer contours do not touch inner contours (they don't intersect
// and do not share vertices or edges). However, a vertex may be
// used several times in a contour and two outer or inner contours
// can share an edge, but only in opposite direction. If you need multiple
// pairs of contours sharing the same vertex, you probably want
// topologically different vertices at the same place in 3D space. Same if
// you believe you need inner and outer contours sharing a vertex

class BREP_CONTOUR;
typedef std::list<BREP_CONTOUR*> BREP_CONTOUR_RING;

// An edge is a unique connection between two vertices. There can be no more
// than one edge between two given vertices in a valid model. An edge can be
// used in two contours. The use of an edge in a contour is called a wing

class BREP_EDGE;

// A wing contains the starting vertex of the edge in the contour. The
// other vertex is the starting vertex in the other wing, since the edge should
// be used in opposite sense by the other contour sharing the edge.
// The invariant is: wing->vertex == OtherWingInEdge(wing->prev)->vertex:
// the starting point in a wing is the endpoint of the edge specified in
// wing->prev.
// The data structure contains two more pointers than a winged edge data
// structure, but the code to operate on it is significantly simpler, cfr.
// [Weiler 1985]. The data structure is more compact than in [Glassner 1991]

class BREP_WING;

// A doubly linked ring of wings indicates in what wings a vertex is being used.
// The code becomes a lot simpler and more efficient when explicitely
// remembering with each vertex what wings leave at the vertex. It also allows
// to deal easier with partial shells: shells not enclosing a region into space,
// such as a collection of coplanar faces, which is often used in synthetic
// environments

typedef std::list<BREP_WING*> BREP_WING_RING;

// A vertex is basically a point in 3D space

class BREP_VERTEX;

// Trees with 8 children per node

template <class T>
class OctreeNode
{
public:
	OctreeNode(T newelement)
	{
		pelement = newelement;
		for(int i = 0; i < 8; i++)
			child[i] = NULL;
	}

	~OctreeNode()
	{
		for(int i = 0; i < 8; i++)
			delete child[i];
	}

	// Allows duplicate elements in the octree
	OctreeNode* AddWithDuplicates(T newelement, int(*nodecmp)(T pelem1, T pelem2))
	{
		OctreeNode* o = NULL;
		OctreeNode* p = this;
		int cmp = 8;

		while(p != NULL)
			{
				cmp = nodecmp(p->pelement, newelement);

				// Allow duplicate items in the octree
				o = p;
				p = p->child[cmp%8];
			}

		// P is null, o is the octree node where we should add the new child
		p = new OctreeNode(newelement);

		if(o != NULL)
			{
				// Add the new node to the old octree
				o->child[cmp%8] = p;
				return this;
			}
		else
			// New octree node
			return p;
	}

	// This routine looks up the element in the octree using the element
	// comparison function specified. If the element is found, a pointer to the
	// OCTREE cell where the element is stored is returned. If it is not found,
	// NULL is returned. The nodecmp function returns 8 if the elements are
	// equal and the branch to be investigated if they are different.
	OctreeNode* FindSubtree(T element, int (*nodecmp)(T, T))
	{
		OctreeNode* p = this;

		while(p != NULL)
			{
				int cmp = nodecmp(p->pelement, element);
				if (cmp >= 8)
					return p;
				else
					p = p->child[cmp];
			}

		return NULL;
	}

	// The nodecmp function returns 8 if the elements are equal
	// and the branch to be investigated if not
	T Find(T element, int (*nodecmp)(T, T))
	{
		OctreeNode* p = FindSubtree(element, nodecmp);

		if(p != NULL)
			return p->pelement;

		return NULL;
	}

	void Iterate(void (*func)(T))
	{
		for(int i = 0; i < 8; i++)
			if(child[i] != NULL)
				child[i]->Iterate(func);

		func(pelement);
	}

	void Iterate2A(void (*func)(T, OctreeNode*, void (*)(BREP_WING*)), OctreeNode* parm1, void (*func2)(BREP_WING*))
	{
		for(int i = 0; i < 8; i++)
			child[i]->Iterate2A(func, parm1, func2);

		func(pelement, parm1, func2);
	}

	// OK
	void Iterate2A(void (*func)(T, T, void (*)(BREP_WING*)), T parm1, void (*func3)(BREP_WING*))
	{
		for(int i = 0; i < 8; i++)
			child[i]->Iterate2A(func, parm1, func3);

		func(pelement, parm1, func3);
	}

	void Iterate2A(T v1, T v2, void (*func2)(BREP_WING*))
	{
		for(int i = 0; i < 8; i++)
			child[i]->Iterate2A(func, parm1, func2);

		func2(pelement, parm1, func2);
	}

	void DeleteElements()
	{
		for(int i = 0; i < 8; i++)
			child[i]->DeleteElements();

		delete pelement;
	}


	T pelement;
	OctreeNode* child[8];
};


typedef OctreeNode<BREP_VERTEX*> BREP_VERTEX_OCTREE;


// ********************** Callback Functions **************************
// In order to maintain other than topological data with the entities,
// e.g. geometrical data such as coordinates of a vertex, plane equation
// of a face etc ..., a Close callback function is associated with
// each entity:
// it is called when the topological entity is completely specified.
// A vertex is completely specified when it is created, an edge is
// completely specified each time it is included in a contour.
// Contours, faces, shells and solids are considered completely
// specified only when the user lets it know to the library by calling
// CloseCallback().
// A Close callback is a good moment to compute plane equations
// for faces, to compute bounding boxes, to verify etc...
//
// Default callback functions: none. If a callback function is not set,
// no action is taken and nothing is changed to the data
// of an entity.

class BrepStruct
{
public:
	virtual void DataCloseCallback(bool done) = 0;

protected:
	virtual ~BrepStruct() {}
};


// Classes


class SOLID;
class BREP_SOLID :
	public BrepStruct
{
friend class SOLID;
public:
	BREP_SOLID();
	// Releases solid, also its vertices
	// if not used in other solids as well
	~BREP_SOLID();

	// Performs various actions to be done when the boundary
	// representation of the solid has been completely specified
	// (automatically calls shells->CloseCallback() for its
	// contained shells and so on)
	void CloseCallback(const bool done = false);

	// Connect a shell to the solid
	void ConnectShell(BREP_SHELL* shell);

	// Disconnect a shell from the solid
	void DisconnectShell(BREP_SHELL* shell);

	// Iterate through all contours inside the solid
	void IterateContours(void (*func)(BREP_CONTOUR*));

	//
	BREP_SHELL_RING shells;

	// Vertices, sorted in an octree for fast lookup
	BREP_VERTEX_OCTREE* vertices;
};


class SHELL;
class BREP_SHELL :
	public BrepStruct
{
friend class SHELL;
public:
	// Creates a new empty shell within the solid if solid is not a NULL
	// pointer. You will need to call CloseCallback() yourself to
	// close the shell and/or ConnectShell() to connect it
	// to a solid if you don't specify a solid to contain the shell with
	// this function
	BREP_SHELL(BREP_SOLID* solid);
	~BREP_SHELL();

	// Only needed for shells that were created
	// without specifying the enclosing solid
	void CloseCallback(const bool done = false);

	// Connect a face to the shell
	void ConnectFace(BREP_FACE* face);

	// Disconnect the face from its containing shell
	void DisconnectFace(BREP_FACE* face);

	BREP_FACE_RING faces;

	// Backpointer to the containing solid
	BREP_SOLID* solid;
};


class FACE;
class BREP_FACE :
	public BrepStruct
{
friend class FACE;
public:
	// Creates a new empty face within the shell
	BREP_FACE(BREP_SHELL* shell);
	// Release face and its contours, including
	// edges if not used in other faces as well
	~BREP_FACE();
	void CloseCallback(const bool done = false);

	// Connect a contour (the first one is supposed to be the outer contour
	void ConnectContour(BREP_CONTOUR* contour);
	// Disconnect a contour
	void DisconnectContour(BREP_CONTOUR* contour);

	// Contours form a doubly linked ring
	// in order to facilitate the splitting
	// and merging of contours. There is
	// one ring per face, containing the
	// outer contour as well as the rings
	BREP_CONTOUR_RING outer_contour;

	// Backpointer to the containing shell
	BREP_SHELL* shell;
};


class CONTOUR;
class BREP_CONTOUR :
	public BrepStruct
{
public:
	// Creates a new empty contour within the face.
	// 'face' cannot be a NULL pointer
	BREP_CONTOUR(BREP_FACE* face);
	~BREP_CONTOUR();

	void CloseCallback(const bool done = false);

	// Connect a wing
	void ConnectWing(BREP_WING* wing);

	// Disconnect a wing
	void DisconnectWing(BREP_WING* wing);

	// Create a wing into the contour
	BREP_WING* CreateWing(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2);

	// Destroy the wings
	void DestroyWings();

	// Remove a vertex (the vertex is not deleted)
	void RemoveVertex(BREP_VERTEX* vertex);

	// Execute a function for every wing:
	// iterate over the wings in the contour
	// (when properly built, the inside of the face
	// containing the contour is always on the left side of an edge)
	void IterateWings(void (*func)(BREP_WING*));
	void IterateWings1A(void (*func)(BREP_WING*, void*), void* parm);

	// Iterates over the vertices in a contour.
	// Same order as when iterating over the edges
	void IterateVertices(void (*func)(BREP_VERTEX*));

	BREP_WING* wings;

	// Backpointer to the containing face
	BREP_FACE* face;
};

extern BREP_CONTOUR* CreateContour(BREP_FACE* face);


class BREP_WING
{
public:
	BREP_WING(BREP_VERTEX* v);
	~BREP_WING() {}

	// Start vertex
	BREP_VERTEX* vertex;
	// ClockWise and CounterClockWise
	// next wing within outer contours
	BREP_WING* prev;
	BREP_WING* next;

	// Pointer to the containing edge
	BREP_EDGE* edge;

	// Backpointer to the containing contour
	BREP_CONTOUR* contour;

};


class BREP_EDGE
{
public:
	// Creates an edge connecting vertex1 to vertex2
	// (it is not connected to a contour)
	BREP_EDGE(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2);
	~BREP_EDGE();

	// Actions to be performed when an edge has been specified completely,
	// i.e. it is enclosed in two contours
	void CloseCallback(const bool done = false) {}

	// Two wings
	BREP_WING* wing1;
	BREP_WING* wing2;
};

extern BREP_EDGE* CreateEdge(BREP_VERTEX* v1, BREP_VERTEX* v2);


class BREP_VERTEX
{
public:
	// Creates a new vertex. The new vertex is not installed it in a
	// vertex octree as vertices do not have to be sorted in an octree
	BREP_VERTEX();
	~BREP_VERTEX();

	// Disconnect the wing from its starting vertex
	void DisconnectWing(BREP_WING* wing);

	// Iterate over the wings leaving at the vertex, in arbitrary order
	void IterateWings(void (*func)(BREP_WING*))
	{
		BREP_WING_RING::iterator wing;

		for(wing = wing_ring.begin(); wing != wing_ring.end(); wing++)
			func(*wing);
	}

	// Ring of wings leaving at the vertex
	BREP_WING_RING wing_ring;
};

extern int VertexCompare(BREP_VERTEX* v1, BREP_VERTEX* v2);
extern int VertexCompareLocation(BREP_VERTEX* brepv1, BREP_VERTEX* brepv2);


// Callback routines for communicating informational, warning, error and
// fatal error messages. First parameter (if non-null) is a pointer
// to the client data of the topological entity that caused the
// error (use and identification of this info is left to the user of
// this library). The second argument is the message itself
typedef void (*BREP_MSG_CALLBACK_FUNC)(void* client_data, char* message);

// No error message from the library is longer than this
#define BREP_MAX_MESSAGE_LENGTH 200

// The SetCallback functions return the previously set callback function
extern BREP_MSG_CALLBACK_FUNC BrepSetInfoCallback(BREP_MSG_CALLBACK_FUNC f);
extern BREP_MSG_CALLBACK_FUNC BrepSetWarningCallback(BREP_MSG_CALLBACK_FUNC f);
extern BREP_MSG_CALLBACK_FUNC BrepSetErrorCallback(BREP_MSG_CALLBACK_FUNC f);
extern BREP_MSG_CALLBACK_FUNC BrepSetFatalCallback(BREP_MSG_CALLBACK_FUNC f);



//********************** Contructors *****************************

extern BREP_WING* BrepCreateWingWithoutContour(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2);


//*********************** Modifiers *******************************

// Splits only one wing. If a full edge is to be split, also its
// wing in the other contour (if any) needs to be split at the same vertex.
// (BrepSplitEdge does this).
// Returns the wing leaving at the inserted vertex.
extern BREP_WING* BrepSplitWing(BREP_WING* wing, BREP_VERTEX* vertex);

// Join two edges at their common endpoint. The edges must be used
// in the same contours and must share an endpoint.
extern void BrepJoinEdges(BREP_EDGE* edge1, BREP_EDGE* edge2);

// Joins two wings at there common endpoint. The wings are supposed to
// be consecutive wings, i.e. wing1->next = wing2 (and wing2->prev = wing1)
// in a closed contour. The common vertex is wing2->vertex,
// the start vertex of the second wing (which is supposed to be
// the endpoint of the first wing as well). Returns the newly
// created wing.
extern BREP_WING* BrepJoinWings(BREP_WING* wing1, BREP_WING* wing2);

// wing1 and wing2 should be two wings belonging to the same contour.
// This routine creates a new edge connecting the start vertex of wing1
// and wing2, splitting the contour in two. A pointer to the newly created
// wing, connecting the start vertex of wing1 to the start vertex of wing2,
// is returned.
extern BREP_WING* BrepMakeEdgeSplitContour(BREP_WING* wing1, BREP_WING* wing2);

// Wing1 and wing2 are two wings belonging to a different contour. This
// routine creates a new edge connecting the starting vertices of
// wing1 and wing2 and merges to two contours to one. A pointer to
// the newly created wing, connecting the start vertex of wing1 to the
// start vertex of wing2, is returned.
extern BREP_WING* BrepMakeEdgeJoinContours(BREP_WING* wing1, BREP_WING* wing2);

// wing is a wing of an edge that is used in two different contours. This
// routine deletes the edge to which the wing belongs, inserting the reaminder
// of the "other" contour sharing the edge into the contour to which the wing
// belongs. The other contour is destroyed and the first is "closed" again
// in order to notify that it changed. Returns the wing from the first contour
// that comes before the first wing moved from the second to the first, or
// NULL if the first contour was a selfloop. In that case, the empty contour
// is deleted. In case of errors, wing is returned. Warning: if the face to which
// the other contour belongs, contains holes, these holes should probably be moved
// to the face to which 'wing' belongs.
extern BREP_WING* BrepDeleteEdgeJoinContours(BREP_WING* wing);

// 'wing' is a wing of an edge that is used two times (in opposite direction)
// in the same contour, i.o.w. a seam. This routine deletes this edge.
// If the wings happen to be consequtive wings, the wings are just deleted.
// If these are the only two wings in the contour, the contour is deleted as
// well and a NULL wing pointer is returned. If they are not consequtive, the
// contour is split in two contours. The wing before 'wing' will remain in
// the original contour and will be returned.
extern BREP_WING* BrepDeleteEdgeSplitContour(BREP_WING* wing);

// Creates a notch, connecting the start vertex of the wing with
// the specified vertex, ans inserts it into the contour to
// which the wing belongs. A pointer to the newly created
// wing, from the start vertex of the specified wing to the
// specified vertex, is returned.
extern BREP_WING* BrepMakeNotch(BREP_WING* wing, BREP_VERTEX* vertex);

// Creates a slit, connecting the two specified vertices,
// in the given face. A slit is a new contour containing
// two wings: one from v1 to v2 and the second
// from v2 back to v1. A pointer to the newly created
// wing, connecting v1 to v2, is returned.
extern BREP_WING* BrepMakeSlit(BREP_FACE* face, BREP_VERTEX* v1, BREP_VERTEX* v2);

// Replace the starting vertex of 'wing' by 'newvertex'. Assumes
// that 'wing' is correctly connected in a contour and it is not
// the only wing (that means: the contour is not a selfloop).
// Returns a pointer to the new wing, which is leaving at 'newvertex'.
extern BREP_WING* BrepWingReplaceVertex(BREP_WING* wing, BREP_VERTEX* newvertex);


//******************* Selectors and Iterators *************************
// adding and removing elements in the ring being iterated over is restricted
// to certain conditions: the first element being iterated over must remain
// in the ring and the next element of an element being passed must remain
// the same as well.

// Iterator over all wings (included in a contour) connecting the two
// given vertices
extern void BrepIterateWingsWithVertex(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2,
				       void (*func)(BREP_WING*));

// Returns the other wing in the edge
extern BREP_WING* BrepEdgeOtherWing(BREP_WING* wing);

// returns the next wing with same starting vertex as wing. The order is
// counterclockwise if the vertex is used in outer contours only and
// clockwise if used in inner contours only. Returns a null pointer
// if the current wing is not properly connected to possible
// other wings sharing the vertex, or the boundary of a partial
// shell was reached.
extern BREP_WING* BrepNextWingLeavingVertex(BREP_WING* wing);

// Looks whether an edge connecting the two vertices already
// exists, Returns it if it exists. Returns NULL if not.
extern BREP_EDGE* BrepFindEdge(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2);


//*********************** Destructors ****************************

// remove a wing from a contour, release the storage associated with the
// edge if it is not used in other contours as well. This routine will
// in general create a hole in a contour: the endpoint of a previous edge
// might not be the starting point of the next one where an edge has been
// deleted. The vertices are not deleted.
extern void BrepDestroyWing(BREP_WING* wing);


//********************* vertex octrees ***************************
// vertices can also be sorted and stored into an octree in order to make
// searching for a vertex significantly faster.

// Before trying to store vertices in an octree, first a routine should
// be specified to compare the client data of two vertices.
// The routine specified should return a code with the following meaning:
//
//	0:  x1 <= x2 , y1 <= y2, z1 <= z2 but not all are equal
//	1:  x1 >  x2 , y1 <= y2, z1 <= z2
//	2:  x1 <= x2 , y1 >  y2, z1 <= z2
//	3:  x1 >  x2 , y1 >  y2, z1 <= z2
//	4:  x1 <= x2 , y1 <= y2, z1 >  z2
//	5:  x1 >  x2 , y1 <= y2, z1 >  z2
//	6:  x1 <= x2 , y1 >  y2, z1 >  z2
//	7:  x1 >  x2 , y1 >  y2, z1 >  z2
//	8:  x1 == x2 , y1 == y2, z1 == z2
//
// in other words:
//
//	code&1 == 1 if x1 > x2 and 0 otherwise
//	code&2 == 1 if y1 > y2 and 0 otherwise
//	code&4 == 1 if z1 > z2 and 0 otherwise
//	code&8 == 1 if x1 == x2 and y1 == y2 and z1 == z2

// Looks up a vertex in the vertex octree, return NULL if not found
extern BREP_VERTEX* BrepFindVertex(BREP_VERTEX* vertex_data, BREP_VERTEX_OCTREE* vertices);

// Attaches a  BREP_VERTEX to the vertex octree
extern void BrepAttachVertex(BREP_VERTEX* vertex, BREP_VERTEX_OCTREE* vertices);

// Calls func for each BREP_VERTEX in the octree
extern void BrepIterateVertices(BREP_VERTEX_OCTREE* vertices, void (*func)(BREP_VERTEX*));

// For convenience, a vertex octree is created with each BREP_SOLID.
// The routines below allow to create and destroy vertex octrees not
// bound to a solid.

// Destroys all the vertices AND the vertex octree
extern void BrepDestroyVertices(BREP_VERTEX_OCTREE* vertices);

// Looks up the first vertex in the vertex octree at the location as specified in the
// given vertex data. There may be multiple vertices at the same location (e.g. having
// a different normal and/or a different name). These vertices should normally be
// stored as a suboctree of the octree containing all vertices. A pointer to the
// top of this suboctree is returned if there are vertices at the given location.
// NULL is returned if there are no vertices at the given location.
extern BREP_VERTEX_OCTREE* BrepFindVertexAtLocation(BREP_VERTEX* vertex, BREP_VERTEX_OCTREE* vertices);

// Iterator over all edge-wings between vertices
// at the locations specified by v1data and v2data
extern void BrepIterateWingsBetweenLocations(BREP_VERTEX* v1, BREP_VERTEX* v2,
	BREP_VERTEX_OCTREE* vertices, void (*func)(BREP_WING*));


#endif // _BREP_H_
