//
//  The BREP Library.
//  Copyright (C) 1996 Philippe Bekaert
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// C++ version: Romain Behar (romainbehar@yahoo.com), Feb 2002

// geometry.c: geometric data for topological entities such as vertices,
// edges, ...

#include <iostream>
#include <string.h>
#include <math.h>

#include "csg.h"

#include <stdlib.h>
#include <stdio.h>


// ************************* tolerances ***********************************

// Minimum (relative) vertex, edge and face tolerance
double vertex_tolerance = 1e-6;
double edge_tolerance = 1e-12;
double face_tolerance = 1e-12;

void SetMinimumVertexTolerance(double tol)
{
	vertex_tolerance = tol;
}

void SetMinimumEdgeTolerance(double tol)
{
	edge_tolerance = tol;
}

void SetMinimumFaceTolerance(double tol)
{
	face_tolerance = tol;
}


// ********************** Entity counts *******************************

static int TotalVertices = 0, vertexcount = 0;
static int TotalEdges = 0, edgecount = 0;
static int TotalContours = 0, contourcount = 0;
static int TotalFaces = 0, facecount = 0;
static int TotalShells = 0;
static int TotalSolids = 0;

int GetNumberOfVertices()
{ return TotalVertices; }

int GetNumberOfEdges()
{ return TotalEdges; }

int GetNumberOfContours()
{ return TotalContours; }

int GetNumberOfFaces()
{ return TotalFaces; }

int GetNumberOfShells()
{ return TotalShells; }

int GetNumberOfSolids()
{ return TotalSolids; }



// ********************** Vertex functions *******************************

// Vertex comparison flags:
//
// The following data may be taken into consideration when comparing vertices:
// - vertex location: normally always specified.
// - vertex normal: (0,0,0) if not specified, normalized vector if specified.
// - vertex name: NULL if not specified.
// What data should be taken into account when comparing vertices is
// controlled with the vertex compare flags VCMP_LOCATION,
// VCMP_NORMAL and VCMP_NAME.
//
// The order in which the data is taken into account by VertexCompare is
// as follows: first location is tested, than normals, than name (if relevant).
//
// When the normal or name is relevant, it also depends on the situation
// whether or not two vertices without normal (or name) and equal other data
// are to be considered equal or not. This is controlled with the flags
// VCMP_NO_NORMAL_IS_EQUAL_NORMAL and VCMP_NO_NAME_IS_EQUAL_NAME.
// When the normal (or name) is relevant, and only one of the two
// vertices has a normal (or name), the vertices are considered different.
//
// The flags can be |-ed together.
//
// Default vertex compare flags: vertex location, normal and name are
// all taken into account when comparing vertices. If two vertices to be
// compared have the same location, and the same normal,but both of them
// have no name (e.g. automatically created vertices by the MGF parser
// library, when discretising spheres ...) the vertices are to be
// considered equal

const unsigned int VCMP_LOCATION = 1;
const unsigned int VCMP_NORMAL = 2;
const unsigned int VCMP_NAME = 4;
const unsigned int VCMP_NO_NORMAL_IS_EQUAL_NORMAL = 8;
const unsigned int VCMP_NO_NAME_IS_EQUAL_NAME = 16;

static unsigned int vertex_compare_flags;

void SetVertexCompareFlags(unsigned int flags)
{
	vertex_compare_flags = flags;
}

// Compares two vertices and returns a code useful for ordering vertices
// into an octree. Returns 8 if the two vertices are to be considered
// the same vertices and 0-7, the index of an octree branch to be expored
// further, if not. This function is needed in order to sort created
// vertices into an octree for efficient later lookup. This is taken care of
// with the functions BrepAttachVertex() and BrepLookupVertex(). For each
// solid, a vertex octree is created at the time the solid is created
int VertexCompare(BREP_VERTEX* brepv1, BREP_VERTEX* brepv2)
{
	VERTEX* v1 = (VERTEX*)brepv1;
	VERTEX* v2 = (VERTEX*)brepv2;

	int code = 8;

	// First compare the coordinates
	if(vertex_compare_flags & VCMP_LOCATION)
		code = VertexCompareLocation(brepv1, brepv2);

	// Same coordinates, test the normals
	if(vertex_compare_flags & VCMP_NORMAL)
		{
			code = 0;

			if (v1->normal.x > v2->normal.x + EPSILON) code += 1;
			if (v1->normal.y > v2->normal.y + EPSILON) code += 2;
			if (v1->normal.z > v2->normal.z + EPSILON) code += 4;
			if(code != 0)
				// X1 > x2 || y1 > y2 || z1 > z2
				return code;

			if(v1->normal.x < v2->normal.x - EPSILON ||
				v1->normal.y < v2->normal.y - EPSILON ||
				v1->normal.z < v2->normal.z - EPSILON)
				// Not the same normal
				return code;

			if(!(vertex_compare_flags & VCMP_NO_NORMAL_IS_EQUAL_NORMAL))
				if(v1->normal.x == 0 && v1->normal.y == 0 && v1->normal.z == 0)
					// No valid normal => not the same vertex
					return 0;

			code = 8;
		}

	if(vertex_compare_flags & VCMP_NAME)
		{
			if(v1->name != NULL && v2->name != NULL)
				{
				// Both have a name, compare
				int cmp = strcmp(v1->name, v2->name);
				if(cmp < 0)
					return 1;

				if(cmp > 0)
					return 2;

				// Same name => same vertex
				return 8;
				}
			else
				{
				if(v1->name)
					// v1 has a name, v2 not
					return 3;

				if(v2->name)
					// v1 has no name, v2 has
					return 4;

				if(vertex_compare_flags & VCMP_NO_NAME_IS_EQUAL_NAME)
					// None has a name => same vertex
					return 8;

				// None has a name => not same vertex
				return 0;
				}
		}

	return code;
}

// This routine only compares the location of the two vertices.
// There may be multiple vertices at the same location,
// e.g. having a different normal and/or name. These vertices
// are considered different vertices by the vertex compare routine,
// but they are considered the same vertices by this one.
int VertexCompareLocation(BREP_VERTEX* brepv1, BREP_VERTEX* brepv2)
{
	VERTEX* v1 = (VERTEX*)brepv1;
	VERTEX* v2 = (VERTEX*)brepv2;

	// Two entities intersect if their tolerance region intersects, i.o.w. if
	// the distance between them is smaller than the sum of the two entity's
	// tolerances
	double tolerance = v1->tolerance + v2->tolerance;

	int code = 0;
	if(v1->point.x > v2->point.x + tolerance) code += 1;
	if(v1->point.y > v2->point.y + tolerance) code += 2;
	if(v1->point.z > v2->point.z + tolerance) code += 4;
	if(code !=0 )
		// x1 > x2 || y1 > y2 || z1 > z2
		return code;

	if(v1->point.x < v2->point.x - tolerance ||
		v1->point.y < v2->point.y - tolerance ||
		v1->point.z < v2->point.z - tolerance)
		// Not the same coordinates
		return code;

	return 8;
}

std::ostream& operator << (std::ostream& out, const VERTEX& vertex)
{
	out << "vertex id " << (vertex.id & ~IDMASK);
	out << ", name '" << (vertex.name ? vertex.name : "(null)") << "'";
	out << ", point = (" << vertex.point.x << "," << vertex.point.y << "," << vertex.point.z << ")";
	out << ", normal = (" << vertex.normal.x << "," << vertex.normal.y << "," << vertex.normal.z << ")";
	out << ", tolerance = " << vertex.tolerance << ", ";

	if(!(vertex.id & HIDDEN)) out << "not ";
	out << "hidden, ";
	if(!(vertex.id & PROCESSED)) out <<  "not ";
	out << "processed." << std::endl;

	return out;
}

VERTEX::VERTEX()
{
	id = ((vertexcount++)&~IDMASK) | VERTEXMASK;	// So it will be recognized as a vertex
	tolerance = vertex_tolerance;
	name = NULL;

#ifdef DEBUG
	std::cerr << "VertexCreateCallback:" << std::endl;
	std::cerr << *this << std::endl;
#endif

	TotalVertices++;
}

void VERTEX::DataCloseCallback(const bool done = false)
{
	ComputeTolerance();

#ifdef DEBUG
	std::cerr << "VertexCloseCallback:" << std::endl;
	std::cerr << *this << std::endl;
#endif
}

VERTEX::~VERTEX()
{
#ifdef DEBUG
	std::cerr << "VertexDestroyCallback:" << std::endl;
	std::cerr << "\t" << *this << std::endl;
#endif

	if(name)
		delete name;

	TotalVertices--;
}


// Computes a normals for the vertex if it doesn't have a normal yet
// unless normals in the input should be ignored
void VERTEX::ComputeNormal()
{
	// Don't recompute vertex normals that are given in the input,
	// unless they should be ignored
//	if(ignore_normals_on_input || VectorFloatEqual(normal, VECTOR(0, 0, 0), EPSILON))
//if() without statement in original BREP Library

	// Compute a normal for the vertex:
	// average normal of the faces sharing the vertex
	normal = VECTOR(0, 0, 0);

	BREP_WING_RING::iterator wingiterator;
	for(wingiterator = wing_ring.begin(); wingiterator != wing_ring.end(); wingiterator++)
		{
			BREP_WING* wing = *wingiterator;
			normal = normal + ((FACE *)(wing->contour->face))->normal;
		}

	normal.Normalize();
}

void VERTEX::ComputeTolerance()
{
	double max = fabs(point.x);
	double m = fabs(point.y);

	if(m > max)
		max = m;

	m = fabs(point.z);
	if(m > max)
		max = m;

	tolerance = vertex_tolerance * max;
}

// Enlarges the tolerance region of a vertex. Probably also a good
// place to merge the vertex with other approximately coinciding vertices
void VERTEX::EnlargeTolerance(double tol)
{
	if(tol > tolerance)
		tolerance = tol;
}


// ************************ Edge functions ****************************

EDGE::EDGE(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2) :
	BREP_EDGE(vertex1, vertex2)
{
	id = ((edgecount++) & ~IDMASK) | EDGEMASK;
	tolerance = edge_tolerance;

#ifdef DEBUG
	std::cerr << "EdgeCreateCallback:" << std::endl;
#endif

	TotalEdges++;
}

BREP_EDGE* CreateEdge(BREP_VERTEX* v1, BREP_VERTEX* v2)
{
	EDGE* edge = new EDGE(v1, v2);
	edge->DataCloseCallback();
	return edge;
}

void EDGE::DataCloseCallback(const bool done = false)
{
	double e;

#ifdef DEBUG
	std::cerr << "EdgeCloseCallback:" << std::endl;
#endif

	// Compute tolerance, extent and near as in M. Segal, SIGGRAPH '90 p 105
	extent = ((VERTEX*)(wing1->vertex))->tolerance;
	e = ((VERTEX*)(wing2->vertex))->tolerance;
	if(e > extent)
		extent = e;

	near = HUGE;

#ifdef DEBUG
	//std::cerr << *this;
#endif
}

EDGE::~EDGE()
{
#ifdef DEBUG
	std::cerr << "EdgeDestroyCallback:" << std::endl;
	std::cerr << "\t" << *this << std::endl;
#endif

	TotalEdges--;
}


std::ostream& operator << (std::ostream& out, const EDGE& edge)
{
	out << "edge id " << (edge.id & ~IDMASK)
		<< ", tolerance = " << edge.tolerance
		<< ", extent = " << edge.extent
		<< ", near = " << edge.near
		<< ", ";

	if(!(edge.id & HIDDEN)) out << "not ";
	out << "hidden, ";
	if(!(edge.id & PROCESSED)) out << "not ";
	out << "processed." << std::endl;

	return out;
}

/*
void EDGE::Print(std::ostream& out, const EDGE& edge)
{
	out << edge << std::endl;

	out << "\tvertex1 = " << *((VERTEX*)(wing[0].vertex)) << std::endl;
	out << "\tvertex2 = " << *((VERTEX*)(wing[1].vertex)) << std::endl;
}
*/

void PrintWing(std::ostream& out, BREP_WING* wing)
{
	BREP_EDGE* edge = wing->edge;

	out << *((EDGE*)edge);

	if(wing == edge->wing1)
		{
			out << ", wing 0" << std::endl;
			std::cerr << "\tvertex1 = " << *((VERTEX*)(edge->wing1->vertex)) << std::endl;
			std::cerr << "\tvertex2 = " << *((VERTEX*)(edge->wing2->vertex)) << std::endl;
		}
	else if(wing == edge->wing2)
		{
			out << ", wing 1" << std::endl;
			std::cerr << "\tvertex1 = " << *((VERTEX*)(edge->wing2->vertex)) << std::endl;
			std::cerr << "\tvertex2 = " << *((VERTEX*)(edge->wing1->vertex)) << std::endl;
		}
	else
		out << ", wing ???" << std::endl;

	if(wing->contour)
		{
			out << "\tbelongs to ";
			out << *((CONTOUR*)(wing->contour)) << std::endl;
		}
	else
		out << "\tdoesn't belong to a contour" << std::endl;
}

void PrintWingShort(std::ostream& out, BREP_WING *wing)
{
	if(wing == NULL)
		{
			out << "(null)";
			return;
		}

	out << "edge " << (((EDGE*)(wing->edge))->id & ~IDMASK)
		<< " wing " << ((wing == wing->edge->wing1) ? "0" :
		(wing == wing->edge->wing2) ? "1" : "ERROR")
		<< " contour ";
	if(wing->contour != NULL)
		out << (((CONTOUR*)(wing->contour))->id & ~IDMASK);
	else
		out << -1;
	out << ", vertex " << (((VERTEX*)(wing->vertex))->id & ~IDMASK)
		<< " -> ";
	if(wing->next != NULL)
		out << (((VERTEX*)(wing->next->vertex))->id & ~IDMASK);
	else
		out << -1;
	out << ", ";
}


// ********************** Contour functions ******************************

CONTOUR::CONTOUR(BREP_FACE* face) :
	BREP_CONTOUR(face)
{
	id = ((contourcount++)&~IDMASK) | CONTOURMASK;

#ifdef DEBUG
	std::cerr << "ContourCreateCallback:" << std::endl;
	std::cerr << *this << std::endl;
#endif

	TotalContours++;
}

BREP_CONTOUR* CreateContour(BREP_FACE* face)
{
	CONTOUR* contour = new CONTOUR(face);
	//contour->DataCloseCallback();
	return contour;
}

void CONTOUR::DataCloseCallback(const bool done = false)
{
#ifdef DEBUG
	std::cerr << "ContourCloseCallback:" << std::endl;
	std::cerr << *this << std::endl;
  //CheckContour(contour);
#endif

	ComputeNormal();
}

CONTOUR::~CONTOUR()
{
	TotalContours--;
}

bool CONTOUR::IsHole()
{
	FACE* f = (FACE*)(face);
	return (f->normal * normal) <= 0;
}

std::ostream& operator << (std::ostream& out, const CONTOUR& contour)
{
	out << "contour id " << (contour.id & ~IDMASK) << " normal = (" << contour.normal.x << "," << contour.normal.y << "," << contour.normal.z << "), ";
	if(!(contour.id & HIDDEN)) out << "not ";
	out << "hidden, ";
	if(!(contour.id & PROCESSED)) out << "not ";
	out << "processed." << std::endl;

	return out;
}

// Newells method for computing the normal on the plane defined by the contour.
// Ref: F. Tampieri, Graphics Gems III p 231
void CONTOUR::ComputeNormal()
{
	normal = VECTOR(0, 0, 0);

	VECTOR next = ((VERTEX*)(wings->vertex))->point;
	for(BREP_WING* wing = wings; wing; wing = (wing->next == wings ? NULL : wing->next))
		{
			VECTOR cur = next;
			next = ((VERTEX *)(wing->next->vertex))->point;

			normal.x += (cur.y - next.y) * (cur.z + next.z);
			normal.y += (cur.z - next.z) * (cur.x + next.x);
			normal.z += (cur.x - next.x) * (cur.y + next.y);
		}

	normal.Normalize();
}


// ************************** Face functions *****************************

FACE::FACE(BREP_SHELL* shell) :
	BREP_FACE(shell)
{
	id = ((facecount++)&~IDMASK) | FACEMASK;

	TotalFaces++;
}

void FACE::DataCloseCallback(const bool done = false)
{
#ifdef DEBUG
	std::cerr << "FaceCloseCallback:" << std::endl;
#endif

	// Compute the plane equation of the face with Newell's method
	ComputePlaneEquation();

#ifdef DEBUG
	std::cerr << *this << std::endl;
#endif
}

FACE::~FACE()
{
	TotalFaces--;
}


std::ostream& operator << (std::ostream& out, const FACE& face)
{
	out << "face id " << (face.id & ~IDMASK) << ", ";
	out << "normal = (" << face.normal.x << ", " << face.normal.y << ", " << face.normal.z << "), plane constant = " << face.d << ", tolerance = " << face.tolerance << ", extent = " << face.extent << ", near = " << face.near << ", ";

	if(!(face.id & HIDDEN)) out << "not ";
	out << "hidden, ";
	if(!(face.id & PROCESSED)) out << "not ";
	out << "processed." << std::endl;

	return out;
}

// For computing the plane equation of a face with Newells method,
// including face tolerance, see
// F. Tampieri, Graphics Gems III p 231
// M. Segal, IEEE CG&A Jan. 1988 p 53
static FACE* face_data;
static double dmin, dmax, emax;

static void do_close_face_vertex(BREP_VERTEX *vertex)
{
  VERTEX *vertex_data = (VERTEX*)(vertex);

	double d = -(face_data->normal * vertex_data->point);
	if(d < dmin) dmin = d;
	if(d > dmax) dmax = d;

	if(vertex_data->tolerance > emax)
		emax = vertex_data->tolerance;

	face_data->bounds.Enlarge(vertex_data->point);
}

// Computes the plane equation of the face using Newells method
void FACE::ComputePlaneEquation()
{
	face_data = this;
	normal = ((CONTOUR*)(outer_contour.front()))->normal;

	// Maximum vertex tolerance
	emax = 0;
	// Minimum plane constant
	dmin = HUGE;
	// Maximum plane constant
	dmax = -HUGE;

	BREP_CONTOUR_RING::iterator contour;
	for(contour = outer_contour.begin(); contour != outer_contour.end(); contour++)
		(*contour)->IterateVertices(do_close_face_vertex);

	d = (dmin + dmax)/2;
	tolerance = (dmax - dmin)/2 >= face_tolerance ? (dmax-dmin)/2 : face_tolerance;
	extent = face_data->tolerance + emax;
	near = HUGE;

	bounds.MIN_X -= extent;
	bounds.MIN_Y -= extent;
	bounds.MIN_Z -= extent;
	bounds.MAX_X += extent;
	bounds.MAX_Y += extent;
	bounds.MAX_Z += extent;
}

// Returns the signed distance from the vertex with given
// geometric data to the face with given geometric data.
// Fills in on what side (+1 or -1) of the plane of the face
// the vertex lies. If the vertex lies within the extent
// region of the plane (i.e. the distance to the plane is
// less than the extent of the plane), 0 is filled in in side
double FACE::SignedDistanceToPlane(VERTEX* v, int* side)
{
	double dist = (normal * v->point) + d;

	if(dist < -extent)
		*side = -1;
	else if(dist > extent)
		*side = +1;
	else
		*side = 0;

	return dist;
}

// v0->v1 and v1->v2 are two edges of the boundary of the face with given
// data. This routine returns TRUE if the corner at vertex v1 is a convex corner,
// and returns FALSE if it is a concave corner. Note that our conventions
// about edge directions are different than in Segals papers. In our case, the
// inside of a face is always on the left side of the boundary. Same situation
// for outer and inner contours
bool FACE::ConvexCorner(VERTEX* v0, VERTEX* v1, VERTEX* v2)
{
	VECTOR e1 = v1->point - v0->point;
	VECTOR e2 = v2->point - v1->point;
	VECTOR x = e1 ^ e2;
	return (x * normal) > 0;
}

// ShortestEdgeLength
static double minlen;

static void TestDecreaseEdgeLength(BREP_WING* wing)
{
	BREP_VERTEX* v1 = wing->vertex;
	BREP_VERTEX* v2 = wing->next->vertex;
	VECTOR p1 = ((VERTEX*)(v1))->point;
	VECTOR p2 = ((VERTEX*)(v2))->point;

	double len = VectorDist(p1, p2);
	if(len < minlen)
		minlen = len;
}

double FACE::ShortestEdgeLength()
{
	minlen = HUGE;
	outer_contour.front()->IterateWings(TestDecreaseEdgeLength);

	return minlen;
}



// ********************* Shell functions *********************************

SHELL::SHELL(BREP_SOLID* solid) :
	BREP_SHELL(solid)
{
	id = SHELLMASK;
	name = NULL;
	matname = NULL;

	TotalShells++;
}

void SHELL::DataCloseCallback(const bool done = false)
{
#ifdef DEBUG
	std::cerr << "ShellCloseCallback:" << std::endl;
	std::cerr << *this << std::endl;
#endif

	// CloseCallback()
	if(!done)
		CloseCallback();

	BREP_FACE_RING::iterator face;
	for(face = faces.begin(); face != faces.end(); face++)
		bounds.Enlarge(((FACE*)(*face))->bounds);
}

SHELL::~SHELL()
{
	if(name != NULL)
		delete name;

	if(matname != NULL)
		delete matname;

	TotalShells--;
}


std::ostream& operator << (std::ostream& out, const SHELL& shell)
{
	out << "shell id " << (shell.id & ~IDMASK) << ", ";
	//	<< "name = '" << (shell.name != NULL ? shell.name : "(null)") << "', "
	//	<< "material = '" << (shell.matname != NULL ? shell.matname : "(null)") << "', ";

	if(!(shell.id & HIDDEN)) out << "not ";
	out << "hidden, ";
	if(!(shell.id & PROCESSED)) out << "not ";
	out << "processed." << std::endl;

	return out;
}


// *********************** Solid functions ***************************

SOLID::SOLID()
{
	id = SOLIDMASK;
	filename = NULL;

	TotalSolids++;
}

void SOLID::DataCloseCallback(const bool done = false)
{
#ifdef DEBUG
	std::cerr << "SolidCloseCallback:" << std::endl;
	std::cerr << *this << std::endl;
#endif

	//CloseCallback();
	BREP_SHELL_RING::iterator shell;
	for(shell = shells.begin(); shell != shells.end(); shell++)
		((SHELL*)(*shell))->DataCloseCallback();

	for(shell = shells.begin(); shell != shells.end(); shell++)
		bounds.Enlarge(((SHELL*)(*shell))->bounds);
}

SOLID::~SOLID()
{
	if(filename)
		delete filename;

	TotalSolids--;
}


std::ostream& operator << (std::ostream& out, const SOLID& solid)
{
	out << "solid id " << (solid.id & ~IDMASK) << ", ";
	out << "filename = '" << (solid.filename != NULL ? solid.filename : "(null)") << "'";
	out << "bounds: " << solid.bounds;

	if(!(solid.id & HIDDEN)) out << "not ";
	out << "hidden, ";
	if(!(solid.id & PROCESSED)) out << "not ";
	out << "processed." << std::endl;

	return out;
}


// ****************** Entity functions ******************

std::ostream& operator << (std::ostream& out, ENTITY* entity)
{
	//if(mg_file)
	//	{
	//		out << "MGF file '" << (mg_file->fname != NULL ? mg_file->fname : "(null)")
	//			<< "' line '" << mg_file->lineno << "':" << std::endl;
	//	}

	if(entity == NULL)
		{
			out << "(no entity data)";
			return out;
		}

	switch(entity->EntityType())
		{
			case VERTEXMASK:	out << *((VERTEX*)entity); break;
			case EDGEMASK:		out << *((EDGE*)entity); break;
			case CONTOURMASK:	out << *((CONTOUR*)entity); break;
			case FACEMASK:		out << *((FACE*)entity); break;
			case SHELLMASK:		out << *((SHELL*)entity); break;
			case SOLIDMASK:		out << *((SOLID*)entity); break;
			default:
				out << "(unknown entity type, id = 0x" << std::hex << entity->id << std::dec << " nr = " << entity->EntityId() << ")";
		 }

	return out;
}


// ****************** brep library error reporting ***********************


// ******************* various initialisations **********************

void InitGeometry()
{
	TotalVertices = TotalEdges = TotalContours = TotalFaces = TotalShells = TotalSolids = 0;

	// Default vertex compare flags: vertex location, normal and name are
	// all taken into accoutn when comparing vertices. If two vertices to be
	// compared have the same location, and the same normal,but both of them
	// have no name (e.g. automatically created vertices by the MGF parser
	// library, when discretising spheres ...) the vertices are to be
	// considered equal
	SetVertexCompareFlags(VCMP_LOCATION | VCMP_NORMAL | VCMP_NAME | VCMP_NO_NAME_IS_EQUAL_NAME);
}



#include <stdlib.h>


void Error(char* routine, char* text)
{
	std::cerr << "Error";
	if(routine != NULL)
		std::cerr << " (in subroutine " << routine << ")";
	std::cerr << ": " << text << "." << std::endl;
}

void Fatal(int errcode, char* routine, char* text)
{
	std::cerr << "Fatal error";
	if(routine != NULL)
		std::cerr << " (in subroutine " << routine << ")";
	std::cerr << ": " << text << "." << std::endl;
	exit(errcode);
}

void Warning(char* routine, char* text)
{
	std::cerr << "Warning";
	if(routine != NULL)
		std::cerr << " (in subroutine " << routine << ")";
	std::cerr << ": " << text << "." << std::endl;
}




// Check for intersection between bounding box and the given ray.
// If there is an intersection between mindist and *maxdist along
// the ray, *maxdist is replaced with the distance to the point of
// intersection, and TRUE is returned.  Otherwise, FALSE is returned.
//
// If this routine is used to check for intersection with a volume
// rather than a "hollow" box, one should first determine if
// (ray->pos + mindist * ray->dir) is inside the bounding volume, and
// call BoundsIntersect() only if it is not.
//
// This routine was taken from rayshade [PhB].
//
int BOUNDINGBOX::BoundsIntersect(RAY* ray, double mindist, double *maxdist)
{
	double t, tmin, tmax;
	double dir, pos;

	tmax = *maxdist;
	tmin = mindist;

	dir = ray->dir.x;
	pos = ray->pos.x;

	if (dir < 0) {
		t = (MIN_X - pos) / dir;
		if (t < tmin)
			return false;
		if (t <= tmax)
			tmax = t;
		t = (MAX_X - pos) / dir;
		if (t >= tmin) {
			if (t > tmax)
				return false;
			tmin = t;
		}
	} else if (dir > 0.) {
		t = (MAX_X - pos) / dir;
		if (t < tmin)
			return false;
		if (t <= tmax)
			tmax = t;
		t = (MIN_X - pos) / dir;
		if (t >= tmin) {
			if (t > tmax)
				return false;
			tmin = t;
		}
	} else if (pos < MIN_X || pos > MAX_X)
		return false;

	dir = ray->dir.y;
	pos = ray->pos.y;

	if (dir < 0) {
		t = (MIN_Y - pos) / dir;
		if (t < tmin)
			return false;
		if (t <= tmax)
			tmax = t;
		t = (MAX_Y - pos) / dir;
		if (t >= tmin) {
			if (t > tmax)
				return false;
			tmin = t;
		}
	} else if (dir > 0.) {
		t = (MAX_Y - pos) / dir;
		if (t < tmin)
			return false;
		if (t <= tmax)
			tmax = t;
		t = (MIN_Y - pos) / dir;
		if (t >= tmin) {
			if (t > tmax)
				return false;
			tmin = t;
		}
	} else if (pos < MIN_Y || pos > MAX_Y)
		return false;

	dir = ray->dir.z;
	pos = ray->pos.z;

	if (dir < 0) {
		t = (MIN_Z - pos) / dir;
		if (t < tmin)
			return false;
		if (t <= tmax)
			tmax = t;
		t = (MAX_Z - pos) / dir;
		if (t >= tmin) {
			if (t > tmax)
				return false;
			tmin = t;
		}
	} else if (dir > 0.) {
		t = (MAX_Z - pos) / dir;
		if (t < tmin)
			return false;
		if (t <= tmax)
			tmax = t;
		t = (MIN_Z - pos) / dir;
		if (t >= tmin) {
			if (t > tmax)
				return false;
			tmin = t;
		}
	} else if (pos < MIN_Z || pos > MAX_Z)
		return false;

	// If tmin == mindist, then there was no "near"
	// intersection farther than EPSILON away
	if (tmin == mindist) {
		if (tmax < *maxdist) {
			*maxdist = tmax;
			return true;
		}
	} else {
		if (tmin < *maxdist) {
			*maxdist = tmin;
			return true;
		}
	}
	return false;	// Hit, but not closer than maxdist
}




// (This code has been copied from rayshade 4.0.6. and adapted)
//
// Quadrants are defined as:
//        |
//   1    |   0
//        |
// -------c--------
//        |
//   2    |   3
//        |
inline int quadrant(const VEC2D& p, const VEC2D& c)
{
	if(p.u < c.u)
		if(p.v < c.v)
			return 2;
		else
			return 1;
	else
		if(p.v < c.v)
			return 3;
		else
			return 0;
}


// Returns the winding number of the contour around the point, projected to 2D.
// This code was taken from Rayshade 4.0.6
static int ContourWinding(BREP_CONTOUR* contour, const VEC2D& center, unsigned int index)
{
	// Is the point inside the polygon?
	//
	// Compute the winding number by finding the quadrant each
	// polygon point lies in with respect to the the point in
	// question, and computing a "delta" (winding number).  If we
	// end up going around in a complete circle around
	// the point (winding number is non-zero at the end), then
	// we're inside.  Otherwise, the point is outside.
	//
	// Note that we can turn this into a 2D problem by projecting
	// all the points along the axis defined by poly->index,
	// the "dominant" part of the polygon's normal vector.
	if(contour->wings == NULL || (contour->wings->prev == contour->wings->next))
		{
			// Fewer than three vertices in the contour, return zero
			return 0;
		}

	int winding = 0;

	VEC2D last = VectorProject(((VERTEX*)(contour->wings->prev->vertex))->point, index);
	int lastquad = quadrant(last, center);

	BREP_WING* wing = contour->wings;
	do
		{
			VEC2D cur = VectorProject(((VERTEX*)(wing->vertex))->point, index);
			int quad = quadrant(cur, center);

			if(quad != lastquad)
				{
					if(((lastquad + 1) & 3) == quad)
						winding++;
					else if(((quad + 1) & 3) == lastquad)
						winding--;
					else
						{
							// Find where edge crosses, center's X axis
							double right = last.u - cur.u;
							double left = (last.v - cur.v) * (center.u - last.u);

							if(left + last.v * right > right * center.v)
								winding += 2;
							else
								winding -= 2;
						}
				}

			lastquad = quad;
			last = cur;
			wing = wing->next;
		}
	while (wing != contour->wings);

	return winding;
}


unsigned int VertexInsideContour(const VECTOR& point, BREP_CONTOUR* contour)
{
	FACE *face_data = (FACE*)(contour->face);

	unsigned int index = VectorDominantCoord(face_data->normal);
	VEC2D center = VectorProject(point, index);

	return ContourWinding(contour, center, index);
}


const double ON_THE_BORDER = 4*M_PI;

// Paulo Cezar Pinto Carvalho and Paulo Roma Cavalcanti, "Point in
// Polyhedron Testing Using Spherical Polygons", Graphics Gems V, p 42
static double GeoSolidAngle(VERTEX* v, BREP_CONTOUR* contour)
{
#ifdef DDEBUG
	fprintf(stderr, "---> GeoSolidAngleDebug:\n");
#endif

	// Contour with less than 3 vertices
	if(contour->wings == NULL || contour->wings->prev == contour->wings->next)
		{
#ifdef DDEBUG
			fprintf(stderr, "%s %d: < 3 verts.\n", __FILE__, __LINE__);
#endif
			return 0;
		}

	// Test whether v->point lies within the plane of the contour. If so, test
	// whether the point is inside of outside the contour. If outside, the solid angle
	// under which the contour is seen is zero. If inside, the solid angle is
	// 2pi or -2pi depending on the orientation of the contour around the vertex
	FACE* plane = (FACE *)(contour->face);
	if(fabs((plane->normal * v->point) + plane->d) < plane->extent + v->tolerance)
		{
			unsigned int winding = VertexInsideContour(v->point, contour);

#ifdef DDEBUG
    fprintf(stderr, "%s %d: vertex on contour plane\n", __FILE__, __LINE__);
    fprintf(stderr, "winding number = %d\n", winding);
#endif

			return (winding == 0 ? 0 : /* (winding > 0 ? 2*M_PI : -2*M_PI) */ ON_THE_BORDER);
		}

	VECTOR norm = ((CONTOUR*)(contour))->normal;
	double area = 0;
	int nr_verts = 0;

	VECTOR p2 = ((VERTEX*)(contour->wings->prev->vertex))->point;
	VECTOR p1 = ((VERTEX*)(contour->wings->vertex))->point;
	VECTOR a = p2 - p1;

#ifdef DDEBUG
  fprintf(stderr, "%s %d;\n\tp1 = (%f,%f,%f)\n\tp2 = (%f,%f,%f)\n\ta = (%f,%f,%f)\n", __FILE__, __LINE__,
	  p1.x, p1.y, p1.z, p2.x, p2.y, p2.z, a.x, a.y, a.z);
#endif

	BREP_WING* wing = contour->wings;
	VECTOR r1;
	for(; wing != NULL; wing = (wing->next == contour->wings ? NULL : wing->next), nr_verts++)
		{
#ifdef DDEBUG
    fprintf(stderr, "\tp1 = (%f,%f,%f)\n", p1.x, p1.y, p1.z);
#endif

			r1 = p1 - v->point;
			p2 = ((VERTEX*)(wing->next->vertex))->point;
			VECTOR b = p2 - p1;
			VECTOR n1 = a ^ r1;
			VECTOR n2 = r1 ^ b;

#ifdef DDEBUG
  fprintf(stderr, "%s %d;\n", __FILE__, __LINE__);
  fprintf(stderr, "\tp1 = (%f,%f,%f)\n", p1.x, p1.y, p1.z);
  fprintf(stderr, "\tp2 = (%f,%f,%f)\n", p2.x, p2.y, p2.z);
  fprintf(stderr, "\ta  = (%f,%f,%f)\n", a.x, a.y, a.z);
  fprintf(stderr, "\tb  = (%f,%f,%f)\n", b.x, b.y, b.z);
  fprintf(stderr, "\t*p = (%f,%f,%f)\n", v->point.x, v->point.y, v->point.z);
  fprintf(stderr, "\tr1 = (%f,%f,%f)\n", r1.x, r1.y, r1.z);
  fprintf(stderr, "\tn1 = (%f,%f,%f)\n", n1.x, n1.y, n1.z);
  fprintf(stderr, "\tn2 = (%f,%f,%f)\n", n2.x, n2.y, n2.z);
#endif

			double l1 = n1.Norm();
			double l2 = n2.Norm();
			double s = (n1 * n2) / (l1 * l2);
			double ang = acos(-1 >= (s < 1 ? s : 1) ? -1 : (s < 1 ? s : 1));

#ifdef DDEBUG
  fprintf(stderr, "%s %d; l1 = %f, l2 = %f, s = %f, ang = %f\n",
	  __FILE__, __LINE__, l1, l2, s, ang);
#endif

			// Determinant
			s = (b^a)*norm; // == b*(a^norm)
			area += s > -EPSILON ? M_PI - ang : M_PI + ang;

#ifdef DDEBUG
  fprintf(stderr, "%s %d; s = %f, s > 0.0 ? M_PI - ang : M_PI + ang = %f, area = %f\n",
	  __FILE__, __LINE__, s, s > 0.0 ? M_PI - ang : M_PI + ang, area);
#endif

			a = -b;
			p1 = p2;
		}

  area -= M_PI * (double)(nr_verts - 2);
#ifdef DDEBUG
  fprintf(stderr, "%s %d: nr_verts = %d, area = %f, DotProduct(norm, r1) = %f\n",
	  __FILE__, __LINE__, nr_verts, area, (norm * r1));
#endif

  area = ((norm * r1) > 0) ? -area : area;
#ifdef DDEBUG
  fprintf(stderr, "-> area = %f\n", area);
#endif
  return area;
}


unsigned int VertexInSolid(VERTEX* v, SOLID* solid)
{
#ifdef DEBUG
  fprintf(stderr, "===> VertexInSolid:\n");
  fprintf(stderr, "\tvertex = "); std::cerr << *((VERTEX*)(v)) << std::endl;
  fprintf(stderr, "\tsolid = "); std::cerr << *((SOLID*)(solid)) << std::endl;
#endif

	if(solid->bounds.OutOfBounds(v->point))
		{
#ifdef DEBUG
    fprintf(stderr, "vertex OUTSIDE bounding box of solid\n");
#endif
			return OUTMASK;
		}

	double Area = 0;
	BREP_SHELL_RING::iterator shell;
	for(shell = solid->shells.begin(); shell != solid->shells.end(); shell++)
		{
			BREP_FACE_RING::iterator face;
			for(face = (*shell)->faces.begin(); face != (*shell)->faces.end(); face++)
				{
					BREP_CONTOUR_RING::iterator contour;
					for(contour = (*face)->outer_contour.begin(); contour != (*face)->outer_contour.end(); contour++)
						{
							double area = GeoSolidAngle(v, *contour);
							if(area < -2*M_PI || area > 2*M_PI)
								{
									if(area != ON_THE_BORDER)
										Error("VertexInSolid", "something wrong with GeoSolidAngle()");

									return 0;
								}

							Area += area;

#ifdef DDEBUG
	fprintf(stderr, "Area = %12.4f\n", Area);
#endif
						}
				}
		}

#ifdef DEBUG
  fprintf(stderr, "--> vertex %s solid\n", Area > 2*M_PI || Area < -2*M_PI ? "INSIDE" : "OUTSIDE");
#endif

	return (Area > 2*M_PI || Area < -2*M_PI) ? INMASK : OUTMASK;
}


unsigned int ContourInSolid(BREP_CONTOUR* contour, SOLID* solid)
{
#ifdef DEBUG
  int count = 0;

  fprintf(stderr, "ContourInSolid:\n");
  fprintf(stderr, "\tsolid = "); std::cerr << *((SOLID*)(solid));
  fprintf(stderr, "\tcontour = "); std::cerr << *((CONTOUR*)(contour)) << std::endl;
  fprintf(stderr, "\tcontour vertices:\n");
  for(BREP_WING* wing = contour->wings; wing; wing = (wing->next == contour->wings ? NULL : wing->next)) {
    fprintf(stderr, "\t\tv%d = ", count++); std::cerr << *((VERTEX*)(wing->vertex)) << std::endl;
  }
#endif

	if(contour->wings == NULL)
		return OUTMASK;

	unsigned int mask = 0;
	BREP_WING* wing = contour->wings;
	do
		{
			// Take the midpoint of the wing
			VECTOR p = MidPoint(((VERTEX*)(wing->vertex))->point,
				((VERTEX*)(wing->next->vertex))->point);

			// Shift it a little bit into the face
			VECTOR u = ((VERTEX*)(wing->next->vertex))->point -
			((VERTEX*)(wing->vertex))->point;
			u = ((CONTOUR*)(contour))->normal ^ u;
			u.Normalize();

			VERTEX v = *((VERTEX*)(wing->vertex));
			v.point = p + (10*EPSILON)*u;
			v.tolerance = ((EDGE *)(wing->edge))->tolerance;

			mask = VertexInSolid(&v, solid);

#ifdef DEBUG
    fprintf(stderr, "mask = %s\n", mask == OUTMASK ? "OUTSIDE" : (mask == INMASK ? "INSIDE" : "UNKNOWN"));
#endif
			wing = wing->next;
		}
	while(wing != contour->wings && mask == 0);

	return mask;
}


/*

const int CULL_NONE = 0;
const int CULL_FRONT = 1;
const int CULL_BACK = 2;


static int icount = 0;
static int cull_flags = CULL_NONE;

static BREP_FACE *face1, *face2;

void DontIntersectFaces(BREP_FACE *f1, BREP_FACE *f2)
{
  face1 = f1;
  face2 = f2;
}

static unsigned RaytraceCull(unsigned flags)
{
  unsigned old_flags = cull_flags;
  cull_flags = flags;
  return old_flags;
}

// Returns winding number of intersection point in the face. The winding number
// * is zero for points outside the face and != zero for contained points
int IntersectionInsideFace(BREP_FACE *face, RAY *ray, double dist)
{
  FACE *face_data = (FACE *)(face);
  VEC2D center;
  int winding, index;

  switch ((index = VectorDominantCoord(face_data->normal))) {
  case XNORMAL:
    center.u = ray->pos.y + dist * ray->dir.y;
    center.v = ray->pos.z + dist * ray->dir.z;
    break;
  case YNORMAL:
    center.v = ray->pos.z + dist * ray->dir.z;
    center.u = ray->pos.x + dist * ray->dir.x;
    break;
  default:	// ZNORMAL
    center.u = ray->pos.x + dist * ray->dir.x;
    center.v = ray->pos.y + dist * ray->dir.y;
  }

  winding = 0;
	BREP_CONTOUR_RING::iterator contour;
	for(contour = face->outer_contour.begin(); contour != face->outer_contour.end(); contour++)
		{
			int w = ContourWinding(*contour, center, index);
			winding += w;
		}

  return winding;
}

static int BoundsIntersected(BOUNDINGBOX& bounds, RAY *ray, double mindist, double maxdist)
{
  VECTOR vtmp;

	vtmp = ray->pos + mindist*ray->dir;
	if(bounds.OutOfBounds(vtmp))
		return bounds.BoundsIntersect(ray, mindist, &maxdist);
  return true;
}

static BREP_FACE *FaceIntersect(BREP_FACE *face, RAY *ray, double mindist, double *maxdist)
{
  FACE *face_data = (FACE *)(face);
  double dist, d;

  if (face == face1 || face == face2)
    return NULL;

  // Test against face bounding box
  if (!BoundsIntersected(face_data->bounds, ray, mindist, *maxdist))
    return NULL;

  // Fill ray.pos in the face plane equation
  d = (face_data->normal * ray->pos) + face_data->d;
  if (cull_flags != CULL_NONE) {
    if ((d < 0. && (cull_flags & CULL_BACK)) ||
	(d > 0. && (cull_flags & CULL_FRONT)))
      return NULL;
  }

  // Compute intersection distance along the ray
  dist = (face_data->normal * ray->dir);
  if (fabs(dist) < EPSILON)
    return NULL;

  dist = - d / dist;
  if (dist < mindist || dist > *maxdist)
    return NULL;

  if (IntersectionInsideFace(face, ray, dist)) {
    *maxdist = dist;
    icount++;
    return face;
  }

  return NULL;
}

static BREP_FACE *ShellIntersect(BREP_SHELL *shell, RAY *ray, double mindist, double *maxdist,
		     int shadowtest)
{
  BREP_FACE *hit;

  if (shell->IS_HIDDEN())
    return NULL;

  if (!BoundsIntersected(((SHELL *)(shell->client_data))->bounds, ray, mindist, *maxdist))
    return NULL;

  hit = NULL;

	BREP_FACE_RING::iterator face;
	for(face = shell->faces.begin(); face != shell->faces.end(); face++)
		{
			BREP_FACE *f = FaceIntersect(*face, ray, mindist, maxdist);
			if(f)
				{
					if(shadowtest)
						return f;
					else
						hit = f;
				}
		}

  return hit;
}

static BREP_FACE *SolidIntersect(BREP_SOLID *solid, RAY *ray, double mindist, double *maxdist,
			  int shadowtest)
{
  BREP_FACE *hit;

  if (!BoundsIntersected(((SOLID *)(solid->client_data))->bounds, ray, mindist, *maxdist))
    return NULL;

  hit = NULL;
	BREP_SHELL_RING::iterator shell;
	for(solid->shells.begin(); shell != solid->shells.end(); shell++)
		{
			BREP_FACE *f = ShellIntersect(*shell, ray, mindist, maxdist, shadowtest);
			if(f)
				{
					if(shadowtest)
						return f;
					else
						hit = f;
				}
		}

  return hit;
}

#ifdef NEVER
static double WingLength(BREP_WING *wing)
{
  VERTEX *v1 = (VERTEX *)(wing->vertex->client_data),
         *v2 = (VERTEX *)(wing->next->vertex->client_data);

  return VectorDist(&(v1->point), &(v2->point))
         - ((EDGE *)(wing->edge->client_data))->extent;
}

static BREP_WING *FindGoodWing(BREP_SOLID *solid)
{
  BREP_SHELL *shell;
  BREP_FACE *face;
  BREP_CONTOUR *contour;
  BREP_WING *wing;

  for (shell = solid->shells; shell; shell = (shell->next == solid->shells ? NULL : shell->next)) {
    for (face = shell->faces; face; face = (face->next == shell->faces ? NULL : face->next)) {
      for (contour = face->outer_contour; contour; contour = (contour->next == face->outer_contour ? NULL : contour->next)) {
	for (wing = contour->wings; wing; wing = wing->next) {
	  if (WingLength(wing) > 0.)
	    return wing;
	}
      }
    }
  }

  return NULL;
}

// Simple application of raytracing: testing whether a vertex is inside the solid
int VertexInSolid(BREP_VERTEX *vertex, BREP_SOLID *solid)
{
  RAY ray;
  VECTOR p, u;
  BREP_WING *wing;
  double dist;

  VECTORCOPY(((VERTEX *)(vertex->client_data))->point, ray.pos);

  // an arbitrary ray direction suffices, but only for a closed solid. If the solid is
//   * not closed, the result would depend on the ray direction. To handle non-closed
//   * solids, we prefer a point-in-polyhedron inclusion test based on solid angles.

  // Find an arbitrary point on an arbitrary face of the solid
  wing = FindGoodWing(solid);
  if (!wing) {
    Error("VertexInSolid", "bad solid 1");
    return 0;
  }

  // Take the midpoint of the edge
  MIDPOINT(((VERTEX *)(wing->vertex->client_data))->point,
	   ((VERTEX *)(wing->next->vertex->client_data))->point,
	   p);

  // Move it a little inward to the face
  VECTORDIFF(((VERTEX *)(wing->next->vertex->client_data))->point,
	     ((VERTEX *)(wing->vertex->client_data))->point,
	     u);
  VECTORCROSSPRODUCT(((FACE *)(wing->contour->face->client_data))->normal,
		     u, u);
  VECTORNORMALIZE(u);
  VECTORSUMSCALED(p, 1.5 * ((EDGE *)(wing->edge->client_data))->extent, u, p);

  VECTORDIFF(p, ray.pos, ray.dir);
  dist = VECTORNORM(ray.dir) - EPSILON;
  VECTORSCALEINVERSE(dist, ray.dir, ray.dir);

  DontIntersectFaces(wing->contour->face, NULL);
  icount=1;	// One intersection with the face being excluded from testing

  dist = HUGE;
  SolidIntersect(solid, &ray, ((VERTEX *)(vertex->client_data))->tolerance, &dist, false);

  DontIntersectFaces(NULL, NULL);

  return icount&1 ? OUTMASK : INMASK;
}
#endif
*/



// isect: partitions pairs of faces into nonintersecting parts
//
// References:
//
// - Marc Segal and Carlo H. Sequin, "Partitioning Polyhedral Objects into
//      Nonintersecting Parts", IEEE Computer graphics and Applications,
//	Januari 1988 p 53
// - Marc Segal, "Using Tolerances to Guarantee Valid Polyhedral Modeling
//	Results", SIGGRAPH '90, p 105
//
//
// Known bugs:
//
// - endpoint determination may fail for small features, e.g. edges
//	with length of about the same order of magnitude as the vertex
//	tolerances
//
// In case of bugs: recompile this file with the symbol CHECK defined.
// Face pairs causing warnings or errors are dumped to an MGF file, which
// most of the time will be usable to investigate the problem with
// another instance of the program compiled after defining DEBUG symbol.
//
//
// Still to do:
//
// - feature merging (Segal, SIGGRAPH '90)
//
//
// Author:
//
//	Philippe Bekaert, Computer Graphics Research Group K.U.Leuven
//	Philippe.Bekaert@cs.kuleuven.ac.be
//	Aug, 29 1996
//
//
// Changelog:
//
// Sun Sep 22 23:23:37 1996  Philippe Bekaert  <philippe@strojka.cs.kuleuven.ac.be>
//
//      * non-coplanar faces containing contours that lie entirely in the
//	transversal face may cause problems. Solved??
//
// Wed Oct  9 18:10:11 1996  Philippe Bekaert  <philippe@flater.cs.kuleuven.ac.be>
//
//	* inside/outside marking of contours corrected.

#include <vector>
#include <algorithm>

#ifdef DEBUG
#include <iostream>
#include <stdio.h>
#endif


// Arbitrary, but not -1, 0 or 1
const int UNKNOWN_SIDE	= 10;

// Information about intersection points of two faces:

// Endpoint types: to be combined with |

// no faces need to be cut, only for merging
// with very nearby intersection points
const unsigned int VERTEX_ENDPOINT =	1;
// Only the "other" face needs to be cut
const unsigned int PHANTOM_ENDPOINT =	2;
// Both faces need to be partitioned
const unsigned int EXIT_ENDPOINT =	4;
const unsigned int ENTRY_ENDPOINT =	8;
// Deleted endpoints after merging
const unsigned int DELETED_ENDPOINT =	16;

class ENDPOINTstruct
{
public:

	// Returns +1 if p1 comes after p2, -1 if p1 preceeds p2
	// and 0 if they are exactly at the same position
	bool operator < (const ENDPOINTstruct& p2) const
	{
#ifdef DDEBUG
//		fprintf(stderr, "CompareEndpointPosition: endpoint %d vs. %d\n", p1-endpoints, p2-endpoints);
#endif
		return (ti < p2.ti);
	}


	// Parameter of endpoint on intersection line
	double ti;
	// Parameter of endpoint on contributing edge
	double te;
	// Intersection point coordinate tolerance
	double tol;
	// ENTRY_ENDPOINT or EXIT_ENDPOINT or ...
	unsigned int type;
	// Wing containing the intersection point
	BREP_WING* wing;

	// For debugging: source code line identifying the
	// situation how the endpoints was contributed
	int source;
};

// The endpoint list
std::vector<ENDPOINTstruct> endpoints;
typedef std::vector<ENDPOINTstruct>::iterator ENDPOINT;
const ENDPOINT NULLENDPOINT = endpoints.end();


// Intersection line data
class ISECT_LINE
{
public:
	// Parametric line equation is t.U + V, U normalized
	VECTOR U, V;

	// Tolerance
	 double tol;
};

static ISECT_LINE isect;


// Global parameters needed during partioning

// Pair of intersecting faces being processed
static FACE *faceA, *faceB;

// Wings leaving at the exit endpoint of
// the previous cutting interval
static BREP_WING *lastwingA, *lastwingB;

// Exit endpoint parameter on the intersecting
// line of the previous cutting interval
static double lastti;

// Minimum and maximum parameter along intersection
// line of endpoints contributed by faceA and faceB
static double timinA, timaxA, timinB, timaxB;

// See Segal, SIGGRAPH '90 p 111. The coplanarity factor indicates to what
// extent two faces are coplanar. Coplanar faces have a large coplanarity
// factor. For perpendicular faces, the coplanarity factor is sqrt(2). IN
// order to decide whether two faces are to be considered coplanar, their
// coplanarity factor is compared with a minimum coplanarity factor, which
// should be a number between 100 and 1000. Smaller values force the
// algorithm to deem pairs of faces that intersect in a relatively large
// dehedral angle to be coplanar
const double MIN_COPLANARITY_FACTOR = 500;
double min_coplanarity_factor = MIN_COPLANARITY_FACTOR;

// To set the minimum allowed coplanarity factor
void SetMinCoplanarityFactor(double factor)
{
	min_coplanarity_factor = factor;
}

// When CHECK is defined, pairs of faces that cause an error or warning
// are dumped in an MGF file, so the situation can be examined later on
#ifdef CHECK
#include <stdio.h>
#include <stdarg.h>
#include "writemgf.h"

static int check = false;

static int dumpid = 0;
static FILE *dumpf = NULL;

static void OpenDump()
{
  char fname[100];

  sprintf(fname, "dump%d.mgf", ++dumpid);
  dumpf = fopen(fname, "w");
  if (!dumpf) {
    fprintf(stderr, "Can't open dumpfile %s for writing.\n", fname);
  }
}

static void CloseDump()
{
	fclose(dumpf);
	dumpf = NULL;
}

static void DumpFace(FACE *face)
{
  if (!dumpf)
    return;

  fprintf(stderr, "Dumping face %s ...", face == faceA ? "A" : "B");

  InitMGFOutput(face->shell->solid, dumpf);
	OutputMaterial(((SHELL*)(face->shell))->matname,
		&(((SHELL*)(face->shell))->material));
  OutputFace(face);
  CloseMGFOutput(face->shell->solid);

  fprintf(stderr, "done.\n");
}

static void IsectWarning(char *routine, char *text, ...)
{
  va_list pvar;

  fprintf(stderr, "Warning");
  if (routine) fprintf(stderr, " (in subroutine %s)", routine);
  fprintf(stderr, ": ");

  va_start(pvar, text);
  vfprintf(stderr, text, pvar);
  va_end(pvar);

  fprintf(stderr, ".\n");

  if (!faceA && !faceB)
    return;

  OpenDump();

  fprintf(dumpf, "# Warning");
  if (routine) fprintf(dumpf, " (in subroutine %s)", routine);
  fprintf(dumpf, ": ");

  va_start(pvar, text);
  vfprintf(dumpf, text, pvar);
  va_end(pvar);

  fprintf(dumpf, ".\n");

  fprintf(dumpf, "# face A:\n");
  if (faceA) DumpFace(faceA);
  fprintf(dumpf, "# face B:\n");
  if (faceB) DumpFace(faceB);
  CloseDump();
}

#define Warning IsectWarning
#define Error IsectWarning

#endif // CHECK


// Computes the coplanarity factor of two planes
static double CoplanarityFactor(FACE *f1, FACE *f2)
{
	VECTOR normal = f1->normal ^ f2->normal;
	double s = normal.Norm();

	if(s < EPSILON)
		return HUGE;

	double c = 1 - s*s;
	if(c > 0)
		c = sqrt(c);
	else
		c = 0;

	double tol = sqrt(f1->extent * f1->extent + f2->extent * f2->extent + 2 * f1->extent * f2->extent * c) / s;

	return tol / (f1->extent >= f2->extent ? f1->extent : f2->extent);
}

// Computes the parametric equation P = V + tU of the
// intersection line of two non-parallel (face-)planes.
// Ref: Priamos Georgiades, "Plane-to-Plane Intersection",
// Graphics Gems III p 233
static ISECT_LINE ComputePlaneToPlaneIntersection(FACE* f1, FACE* f2)
{
	ISECT_LINE isect;
	isect.U = f1->normal ^ f2->normal;

	double s = (isect.U).Norm();
	double c = 1 - s*s;
	c = (c > 0 ? sqrt(c) : 0);
	isect.tol = sqrt(f1->extent * f1->extent + f2->extent * f2->extent + 2 * f1->extent * f2->extent * c) / s;
	double coplaf = isect.tol / (f1->extent >= f2->extent ? f1->extent : f2->extent);
	if(coplaf > min_coplanarity_factor)
		{
			char text[200];
			sprintf(text, "attempt to compute intersection line of parallel planes (coplanarity factor = %g, min. allowed = %g)", coplaf, min_coplanarity_factor);
			Warning("ComputePlaneToPlaneIntersection", text);
		}

	double maxabs = fabs(isect.U.x);
	int index = XNORMAL;
	double abs;
	if((abs = fabs(isect.U.y)) > maxabs) {maxabs = abs; index = YNORMAL;}
	if((abs = fabs(isect.U.z)) > maxabs) {maxabs = abs; index = ZNORMAL;}

  switch (index) {
  case XNORMAL:
    isect.V.x = 0;
    isect.V.y = (f1->normal.z * f2->d - f2->normal.z * f1->d) / isect.U.x;
    isect.V.z = (f2->normal.y * f1->d - f1->normal.y * f2->d) / isect.U.x;
    break;
  case YNORMAL:
    isect.V.y = 0;
    isect.V.z = (f1->normal.x * f2->d - f2->normal.x * f1->d) / isect.U.y;
    isect.V.x = (f2->normal.z * f1->d - f1->normal.z * f2->d) / isect.U.y;
    break;
  case ZNORMAL:
    isect.V.z = 0;
    isect.V.x = (f1->normal.y * f2->d - f2->normal.y * f1->d) / isect.U.z;
    isect.V.y = (f2->normal.x * f1->d - f1->normal.x * f2->d) / isect.U.z;
    break;
  default:
    Fatal(2, "PlaneToPlaneIntersection", "something impossible wrong");
  }

	(isect.U).InverseScale(s);

#ifdef DEBUG
fprintf(stderr, "PlaneToPlaneIntersection: U = (%f, %f, %f), V = (%f, %f, %f), tol = %g\n",
	isect.U.x, isect.U.y, isect.U.z,
	isect.V.x, isect.V.y, isect.V.z,
	isect.tol);
#endif

	return isect;
}


// Returns TRUE if the vertex approximately is on the intersection line.
// See Segal, SIGGRAPH '90 p 111
static int VertexOnIntersectionLine(VERTEX* v, double dist)
{
	if(fabs(dist) < v->tolerance + isect.tol)
		return true;
	else
		return false;

#ifdef NEVER
  double d, e;
  VECTOR x;

  // Numerically unstable: sometimes, about equal large numbers are subtracted
	x = v->point - isect.V;
  d = (isect.U * x);
  d *= -d;
  d += (x * x);
  e = isect.tol + v->tolerance;
  e *= e;
  return (d <= e);
#endif
}

// Computes the distance of the vertex to the plane of the face in *dist.
// Returns 0 if the vertex approximately lies on the intersection line
// of the two intersection faces being run and -1 or +1, the side w.r.t.
// the plane if not
static int VertexToFaceIntersection(BREP_VERTEX *v, FACE *f, double *dist)
{
	int side;
	*dist = f->SignedDistanceToPlane((VERTEX*)v, &side);

	if(VertexOnIntersectionLine((VERTEX*)v, *dist))
		side = 0;

	return side;
}

// Computes the parameter ti of an intersection point P = v1 + te.(v2-v1),
// on the intersection line ti.U + V of the planes of two intersecting faces
static double ProjectIntersection(VERTEX *v1, VERTEX *v2, double te)
{
	VECTOR x = v2->point - v1->point;
	x = v1->point + te*x;
	x = x - isect.V;
	return x * isect.U;
}

// Same as above, except that the intersection is at a vertex
static double ProjectVertexIntersection(VERTEX* v)
{
	VECTOR x = v->point - isect.V;
	return x * isect.U;
}

// The edge from v1 to v2 belonging to the face intersects another face. The
// intersection line of the face planes has direction vector U. This function
// returns ENTRY_ENDPOINT if the intersection point will be an entry endpoint
// of a cutting interval or EXIT_ENDPOINT if it will be an exit endpoint.
// It relies on the convention that the interior of a face is on the
// left side of each contour bounding the face, or: (v2-v1) x n, with v1 and v2
// the endpoints of a wing and n the normal to the face plane, points to the
// interior of the face. The intersection point will be an entry endpoint if
// this vector and U point to the same direction. It will be an exit endpoint
// if they point into opposite directions
static int TestEntryExit(VERTEX* v1, VERTEX* v2, FACE* f)
{
	VECTOR x = v2->point - v1->point;
	x = f->normal ^ x;

#ifdef DDDEBUG
  fprintf(stderr, "TestEntryExit: v1 = (%f,%f,%f), v2 = (%f,%f,%f), f->normal = (%f,%f,%f), x = (%f,%f,%f), dotp isect.U = %f\n",
	  v1->point.x, v1->point.y, v1->point.z,
	  v2->point.x, v2->point.y, v2->point.z,
	  f->normal.x, f->normal.y, f->normal.z,
	  x.x, x.y, x.z, (x * isect.U));
#endif

	return ((x * isect.U) > 0 ? ENTRY_ENDPOINT : EXIT_ENDPOINT);
}


// v0->v1 and v1->v2 are two edges along the intersection line of the planes
// of two faces. This function returns TRUE if the edges form a notch and
// FALSE if not. The two edges form a notch if they point into opposite
// direction. If the edges form a notch, v1 is an endpoint of a cutting interval
// along which the face containing the edges will need to be partitioned.
// In that caase, it is determined whether the endpoint is an entry or
// exit endpoint in the contributing face: v1 is
// an entry point if the edge v1->v2 points into opposite direction of the
// unit direction vector U of the intersection line. If v1 is an entry endpoint,
// ENTRY_ENDPOINT is filled in in type. EXIT_ENDPOINT if it is an exit endpoint.
// The value of type is unaltered if the edges do not form a notch, in which
// case v1 is not an endpoint
static int IsNotch(VERTEX* v0, VERTEX* v1, VERTEX* v2, int* type)
{
	VECTOR e1 = v1->point - v0->point;
	VECTOR e2 = v2->point - v1->point;

	if((e1 * e2) > 0)
		return false;

	*type = ((e2 * isect.U) < 0 ? ENTRY_ENDPOINT : EXIT_ENDPOINT);
	return true;
}

#ifdef DEBUG
static void PrintEndpointData(std::ostream& out, ENDPOINT p)
{
	if(p == NULLENDPOINT)
		{
			out << "(null)" << std::endl;
			return;
		}

	out << "intersection point " << (p - endpoints.begin()) << ": ti = " << (*p).ti
		<< ", tol = " << (*p).tol << ", te = " << (*p).te << ", source = " << (*p).source
		<< "," << std::endl << "\t type = ";

	//out << "0x" << hex << (*p).type << dec << " = ";
	if((*p).type & ENTRY_ENDPOINT) out << "ENTRY, ";
	if((*p).type & EXIT_ENDPOINT) out << "EXIT, ";
	if((*p).type & PHANTOM_ENDPOINT) out << "PHANTOM, ";
	if((*p).type & VERTEX_ENDPOINT) out << "VERTEX, ";
	if((*p).type & DELETED_ENDPOINT) out << "DELETED, ";

	PrintWingShort(out, (*p).wing);
	if(!(*p).wing || !(*p).wing->contour || !(*p).wing->next || !(*p).wing->contour->face)
		out << "ERROR: BAD WING: ";
	else if (faceA && (*p).wing->contour->face == faceA)
		out << "faceA, ";
	else if (faceB && (*p).wing->contour->face == faceB)
		out << "faceB, ";
	else if (faceA && faceB)
		out << "NOT faceA or faceB!";

	out << std::endl;
}
#endif


// Adds endpoint information to the list of endpoints
static void InsertEndpoint(double ti, double t, int type, BREP_WING* wing, int source)
{
	ENDPOINTstruct newendpoint;

	newendpoint.ti = ti;
	newendpoint.te = t;
	newendpoint.type = type;
	newendpoint.source = source;
	newendpoint.wing = wing;

	// See Segal SIGGRAPH '90 p 111
	newendpoint.tol = isect.tol;

	if(wing->contour->face == faceA)
		{
			if(ti < timinA) timinA = ti;
			if(ti > timaxA) timaxA = ti;
		}
	else
		{
			if(ti < timinB) timinB = ti;
			if(ti > timaxB) timaxB = ti;
		}

	endpoints.push_back(newendpoint);

#ifdef DEBUG
  fprintf(stderr, "%d: ", endpoints.size());
  PrintEndpointData(std::cerr, endpoints.end()-1);
#endif
}

// Wing is a "phantom" wing or a wing that results after splitting a contour.
// This routine will mark the contour to which wing belongs and the contour
// to which the wing in the reverse direction belongs whether they are
// inside or outside the other solid. This information is needed to do
// inside/outside classification for CSG operations
static void MarkContours(BREP_WING* wing)
{
  unsigned int mark, prevmark;

#ifdef DEBUG
  fprintf(stderr, "MarkContour:\n");
  fprintf(stderr, "\twing = "); PrintWingShort(std::cerr, wing); fprintf(stderr, "\n");
  fprintf(stderr, "\totherwing = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)); fprintf(stderr, "\n");
#endif // DEBUG

	FACE* main_face = (FACE*)wing->contour->face;
	FACE* transversal_face = (main_face == faceA ? faceB : faceA);

#ifdef DDEBUG
	std::cerr << "main_face = " << *main_face << std::endl;
	std::cerr << "transversal_face = " << *transversal_face << std::endl;
#endif

	VECTOR p1 = ((VERTEX*)(wing->vertex))->point;
	VECTOR p2 = ((VERTEX*)(wing->next->vertex))->point;
	VECTOR u = p2 - p1;
	VECTOR nm = ((FACE*)main_face)->normal;
	// Nm = ((CONTOUR *)(wing->contour->client_data))->normal;
	VECTOR s = nm ^ u;
	VECTOR nt = ((FACE*)transversal_face)->normal;

	if((s * nt) > 0)
		mark = OUTMASK;
	else
		mark = INMASK;

	prevmark = ((CONTOUR*)(wing->contour))->id & (OUTMASK | INMASK);

#ifdef DDEBUG
    fprintf(stderr, "p1 = (%f,%f,%f)\n", p1.x, p1.y, p1.z);
    fprintf(stderr, "p2 = (%f,%f,%f)\n", p2.x, p2.y, p2.z);
    fprintf(stderr, "u  = (%f,%f,%f)\n", u.x, u.y, u.z);
    fprintf(stderr, "nm  = (%f,%f,%f)\n", nm.x, nm.y, nm.z);
    fprintf(stderr, "s  = (%f,%f,%f)\n", s.x, s.y, s.z);
    fprintf(stderr, "nt  = (%f,%f,%f)\n", nt.x, nt.y, nt.z);
    fprintf(stderr, "DotProduct(s, nt) = %f ==> ",
	   (s * nt));
#endif
#ifdef DEBUG
    fprintf(stderr, "\tprevmark = %s%s -> mark = %s%s\n",
	    (prevmark & INMASK ? "INSIDE " : ""),
	    (prevmark & OUTMASK ? "OUTSIDE " : ""),
	    (mark & INMASK ? "INSIDE " : ""),
	    (mark & OUTMASK ? "OUTSIDE " : ""));
#endif

	((CONTOUR*)(wing->contour))->id &= ~(OUTMASK | INMASK);
	((CONTOUR*)(wing->contour))->id |= mark;

	if(BrepEdgeOtherWing(wing)->contour)
		{
			CONTOUR* ocontour_data = (CONTOUR*)(BrepEdgeOtherWing(wing)->contour);
			unsigned int omark = (mark == OUTMASK ? INMASK : OUTMASK);

#ifdef DEBUG
      fprintf(stderr, "\tocontour inout mask = %s%s -> omark = %s%s\n",
	      ocontour_data->id & OUTMASK ? "OUTSIDE" : "",
	      ocontour_data->id & INMASK ? "INSIDE" : "",
	      omark & OUTMASK ? "OUTSIDE" : "",
	      omark & INMASK ? "INSIDE" : "");
#endif

//    if (!(ocontour_data->id & (INMASK|OUTMASK))) {
//      ocontour_data->id |= omark;
//    } else if ((ocontour_data->id & (INMASK|OUTMASK)) != omark)
//      Warning("MarkContour", "contour %d: inconsistent inside/outside determination", ((CONTOUR*)(wing->contour))->EntityId());

			ocontour_data->id &= ~(INMASK|OUTMASK);
			ocontour_data->id |= omark;
		}
}

// in/out-marks the contour to which wing belongs
// if it has not been marked before
static void ConditionallyMarkContours(BREP_WING *wing)
{
	if((((CONTOUR*)(wing->contour))->id & (INMASK | OUTMASK)) == 0)
		MarkContours(wing);
}

// Test each wing of the main face, starting with first_wing, for intersection with
// the transversal face. Most of the intersection points will be an endpoint of a
// cutting interval, along which the faces will need to be partitioned. Some
// however will not. Record the endpoints with information which wing contributed
// it and whether it is an entry or exit point on the contributing face.
// "first_wing" is known to be the first wing contributing an intersection point
// (SMALLER INDENTATION HERE DUE TO LOSTS OF NESTINGS)
static void DetermineEndpoints(FACE* main_face, BREP_WING* first_wing, FACE *transversal_face)
{
#ifdef DEBUG
fprintf(stderr, "DetermineEndpoints:\n\tmain_face = %s: ",
	main_face == faceA ? "faceA" : "faceB");
	std::cerr << *main_face;
fprintf(stderr, "\n\ttransversal_face = %s: ",
	transversal_face == faceA ? "faceA" : "faceB");
	std::cerr << *transversal_face;
fprintf(stderr, "\n");
#endif

	// Start over looking for intersection from
	// the first wing of the first contour on
	// (Removed START_OVER #define/#endif from original BREP Library (Romain Behar)
	BREP_CONTOUR_RING::iterator contour;
	for(contour = main_face->outer_contour.begin(); contour != main_face->outer_contour.end(); contour++)
	{

		BREP_WING* wing = (*contour)->wings;

#ifdef DEBUG
    fprintf(stderr, "***** contour = %d:\n\twing = ", ((CONTOUR*)(*contour))->EntityId());
    PrintWingShort(std::cerr, wing);
    fprintf(stderr, "\n\tfirst wing of contour = ");
    PrintWingShort(std::cerr, (*contour)->wings);
    fprintf(stderr, "\n");
#endif

		double d2;
		int side2 = VertexToFaceIntersection(wing->vertex, transversal_face, &d2);

		// An impossible value indicating that we don't know
		// on what side of the transversal plane the starting
		// point of the previous wing lies --- this is only
		// needed in rare cases
		int side1 = UNKNOWN_SIDE;

		// Not in a phtanom interval
		bool firstphantom = false;
		bool inphantom = false;

		do
		{
			int side0 = side1;
			double d1 = d2;
			side1 = side2;
			side2 = VertexToFaceIntersection(wing->next->vertex, transversal_face, &d2);

			int type;
			double te, ti;

#ifdef DEBUG
      fprintf(stderr, "v%d (%f,%f,%f) '%s' -> v%d: d1=%g, side1=%d, d2=%g, side2=%d:\n",
	      ((VERTEX*)(wing->vertex))->id & ~IDMASK,
	      ((VERTEX*)(wing->vertex))->point.x,
	      ((VERTEX*)(wing->vertex))->point.y,
	      ((VERTEX*)(wing->vertex))->point.z,
	      (((VERTEX*)(wing->vertex))->name ?
	       ((VERTEX*)(wing->vertex))->name : "(null)"),
	      ((VERTEX*)(wing->next->vertex))->id & ~IDMASK,
	      d1, side1, d2, side2);
#endif

			if(side1 != side2 && (side1+side2) == 0)
			{
				// Intersection halfway the line segment
				// determine parameter of intersection point
				// on the intersection line of the main and
				// transversal face
				te = fabs(d1) / (fabs(d1) + fabs(d2));
				ti = ProjectIntersection((VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex), te);
				type = TestEntryExit((VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex), main_face);
				// Always an endpoint
				InsertEndpoint(ti, te, type, wing,  __LINE__);

			}
			else if(side1 == 0)
			{
				// Starting point lies in the
				// transversal plane
				te = 0;
				ti = ProjectVertexIntersection((VERTEX*)(wing->vertex));

				// We need to know whether the starting
				// vertex of the previous wing also
				// lies in the transversal plane or not
				if(side0 == UNKNOWN_SIDE)
				{
					double d0;
					side0 = VertexToFaceIntersection(wing->prev->vertex, transversal_face, &d0);

					// If the last wing lies inside the transversal plane,
					// determine whether a previous vertex has contributed
					// a phantom endpoint or not. If this "previous" vertex
					// contributes a phantom endpoint, it will be generated
					// later, when the vertex will be processed
					if(side0 == 0)
					{
						int prevside0, prevside1, prevside2, prevtype;
						double prevd0, prevd1, prevd2;
						BREP_WING* prevwing;

						// wing->vertex and wing->prev->vertex lie in the
						// transversal plane.
						// Search backward to find the last vertex that
						// is the top of a notch, or that doesn't lie in
						// the transversal plane. Normally, there is such
						// a vertex since the main face is known not to
						// be coplanar with the transversal face, and
						// contours are closed: so, either there is a
						// vertex in the contour that does not lie in the
						// transversal plane, or they all lie on the
						// intersection line, but then there must be a
						// notch (even two) somewhere. However, due to
						// the finite precision and extents, it might
						// happen that a contour of the main face is
						// judged to be coplanar with the transversal face
						prevside2 = side1;
						prevd2 = d1;

						prevside1 = side0;
						prevd1 = d0;
						prevwing = wing->prev;

						prevside0 = VertexToFaceIntersection(prevwing->prev->vertex, transversal_face, &prevd0);

						while(prevside0 == 0 && !IsNotch((VERTEX*)(prevwing->prev->vertex),
							(VERTEX*)(prevwing->vertex), (VERTEX*)(prevwing->next->vertex), &prevtype) && prevwing != wing)
						{
							prevside2 = prevside1;
							prevd2 = prevd1;

							prevside1 = prevside0;
							prevd1 = prevd0;
							prevwing = prevwing->prev;

							prevside0 = VertexToFaceIntersection(prevwing->prev->vertex, transversal_face, &prevd0);
						}

						if(prevwing == wing)
						{
							// Should not happen, but we prevent
							// infinite loops if it would happen
							Warning("DetermineEndpoints", "bad contour");
							continue;
						}

						if(prevside0 != 0)
						{
							inphantom = true;
						}
						else
						{
							// A previous vertex is the top of a notch, or the
							// the whole contour lies in the transversal plane
#ifdef DEBUG
	      fprintf(stderr, "previous vertex %d is top of a notch or the contour lies entirely in the transversal plane.\n",
		      ((VERTEX*)(prevwing->vertex))->id & ~IDMASK);
#endif

							// Find a vertex not lying in the transversal plane.
							while(prevside0 == 0 && prevwing != wing)
							{
								prevside2 = prevside1;
								prevd2 = prevd1;

								prevside1 = prevside0;
								prevd1 = prevd0;
								prevwing = prevwing->prev;

								prevside0 = VertexToFaceIntersection(prevwing->prev->vertex, transversal_face, &prevd0);
							}

							if(prevside0 != 0)
							{
								// We found a vertex outside the
								// transversal plane, so we have a notch
								inphantom = false;
							}
							else
							{
#ifdef DEBUG
		fprintf(stderr, "the contour lies entirely in the transversal plane ... we skip it.\n");
		continue;
#endif
								// Let's see what this does...
								inphantom = true;
							}
						}

						firstphantom = inphantom;
					}
				}

#ifdef DEBUG
      fprintf(stderr, "v%d <- : side0=%d, inphantom = %s\n",
	      ((VERTEX*)(wing->prev->vertex))->id & ~IDMASK,
	      side0, inphantom ? "TRUE" : "FALSE");
#endif

				if(side0 != 0 && side2 != 0)
				{
					// Both adjacent endpoints lie outside
					// the transversal plane
					if(side0 != side2)
					{
						// The previous and next endpoint lie
						// on opposite sides
						type = TestEntryExit((VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex), main_face);
						// Always an endpoint
						InsertEndpoint(ti, te, type|VERTEX_ENDPOINT, wing, __LINE__);
					}
					else
					{
						// The previous and next endpoint
						// lie on the same side
						if(!main_face->ConvexCorner((VERTEX*)(wing->prev->vertex), (VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex)))
						{
							// If the previous and current wing make
							// a convex corner at the starting
							// vertex of the current wing, no
							// endpoints are induced. If it is a
							// concave corner, two endpoints are
							// induced: one entry and one exit,
							// at the same physical location
							InsertEndpoint(ti, te, EXIT_ENDPOINT|VERTEX_ENDPOINT, wing, __LINE__);
							InsertEndpoint(ti, te, ENTRY_ENDPOINT|VERTEX_ENDPOINT, wing, __LINE__);
						}
						else
						{
							// Not an endpoint, no partitioning of
							// the main or traversal face is
							// necessary. The intersection point is
							// stored anyways in order to merge
							// merge with other intersection points
							// that lie within tolerance form it
							InsertEndpoint(ti, te, VERTEX_ENDPOINT, wing, __LINE__);
						}
					}
				}
				else if(side0 == 0)
				{
					// Previous endpoint lies in the transversal plane
					if(side2 != 0)
					{
						// Next endpoint doesn't lie in the
						// transversal plane

						// The previous edge lies in the transversal
						// plane, the current edge not. No
						// partitioning of the main face is necessary
						// along the previous wing, but partitioning
						// of the transversal plane might be,so we
						// add a phantom endpoint, induced by the
						// previous edge, to the endpoint list
						if(inphantom)
						{
							InsertEndpoint(ti, 1, PHANTOM_ENDPOINT|VERTEX_ENDPOINT, wing->prev, __LINE__);
							inphantom = false;
						}

						if(!main_face->ConvexCorner((VERTEX*)(wing->prev->vertex), (VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex)))
						{
							// If the corner at the starting vertex
							// of the current wing is not a convex
							// corner, the starting vertex is also
							// a real endpoint: the entry of exit
							// of a cutting interval along which
							// the main face *will* need to be
							// partitioned
							type = TestEntryExit((VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex), main_face);
							InsertEndpoint(ti, te, type|VERTEX_ENDPOINT, wing, __LINE__);
						}
						else
						{
							// No endpoint needs to be inserted.
							// Not even a VERTEX endpoint

						}
					}
					else
					{
						// Both previous and next lie in
						// the transversal plane
						if(IsNotch((VERTEX*)(wing->prev->vertex), (VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex), &type))
						{
							// Two wings along the intersection
							// line, going into opposite direction:
							// a notch

							// For a normal notch, inphantom should
							// be TRUE. For a slit however,
							// inphantom may be either TRUE or FALSE
							InsertEndpoint(ti, 1, PHANTOM_ENDPOINT|VERTEX_ENDPOINT, wing->prev, __LINE__);
							inphantom = 1 - inphantom;
							InsertEndpoint(ti, te, type|VERTEX_ENDPOINT, wing, __LINE__);
						}
						else
						{
						// Two wings along the intersection line,
						// going in the same direction.
						// No partitioning of the main face is
						// necessary. Partitioning of the transversal
						// face might be necessary, but the starting
						// vertex of the current wing is not an endpoint
						InsertEndpoint(ti, te, VERTEX_ENDPOINT, wing, __LINE__);
						}
					}
				}
				else if (side2 == 0)
				{
					// Current and next endpoint, i.o.w.
					// the current edge, lies in the
					// traversal plane (but the previous
					// edge does not)
					if(!main_face->ConvexCorner((VERTEX*)(wing->prev->vertex), (VERTEX*)(wing->vertex), (VERTEX*)(wing->next->vertex)))
					{

						// If the corner at the starting vertex of
						// the current wing is not a convex corner,
						// the starting vertex is a real endpoint:
						// the entry of exit of a cutting interval
						// along which the main face *will* need to
						// be partitioned
						type = TestEntryExit((VERTEX*)(wing->prev->vertex), (VERTEX*)(wing->vertex), main_face);
						InsertEndpoint(ti, 1, type|VERTEX_ENDPOINT, wing->prev, __LINE__);

					}
					else
					{
						// No endpoint needs to be inserted.
						// Not even a VERTEX endpoint
					}

					// The main face doesn't need to be cut for the
					// current wing, but the transversal face might
					// however. Insert the starting vector as a
					// phantom endpoint into the endpoint list
					if(!inphantom)
					{
						InsertEndpoint(ti, te, PHANTOM_ENDPOINT|VERTEX_ENDPOINT, wing, __LINE__);
						inphantom = true;
					}

					// The main face doesn't need to be cut, but we can
					// mark the contour as lying inside or outside the
					// other solid.
					// Out guess might not be correct, so, don't mark
					// the contour if already marked before. If the
					// contour was not yet marked, we might mark it
					// wrong if the contour contains wings on both sides
					// of the other solid. In that case, however,
					// the contour will be split later, and the
					// fragments marked correctly
					ConditionallyMarkContours(wing);
				}

			}
			else
			{
				// Either there is no intersection, or only the end
				// vertex lies inside the transversal plane.
				// In the latter case, the intersection will be handled
				// as an intersection of the next wing with the
				// transversal plane, or has been handled as an
				// intersection of the first wing in the contour
			}

			wing = wing->next;
		}
		while(wing != (*contour)->wings);

		if(inphantom != firstphantom)
			Warning("DetermineEndpoints", "something wrong with phantom endpoints");
	}
}

// Returns TRUE if the endpoints approximately coincide
static int CoincidingEndpoints(ENDPOINT p1, ENDPOINT p2)
{
	double d = (*p2).ti - (*p1).ti; if (d < 0) d = -d;
	double tol = (*p1).tol + (*p2).tol;
	return d <= tol;
}


// Align a series of endpoints that are within each others tolerance region
static void AlignEndpoints(int start, int end)
{
#ifdef DEBUG
	std::cerr << "AlignEndpoints:" << std::endl;
#endif

	// Find a good position on the intersection line
	// for alignment and corresponding tolerance region
	int nr_vertex_endpoints = 0;
	double t_align = 0;
	double tmin = HUGE;
	double tmax = -HUGE;

	ENDPOINT p = endpoints.begin() + start;
	for(int i = start; i <= end; i++, p++)
		{
			if((*p).type != VERTEX_ENDPOINT)
				{
					if((*p).ti - (*p).tol < tmin)
						tmin = (*p).ti - (*p).tol;
					if ((*p).ti + (*p).tol > tmax)
						tmax = (*p).ti + (*p).tol;
				}
			if((*p).type & VERTEX_ENDPOINT)
				{
					t_align += (*p).ti;
					nr_vertex_endpoints++;
				}
		}

	double tol;
	if(nr_vertex_endpoints == 0)
		{
			t_align = (tmin + tmax) * 0.5;
			tol = (tmax - tmin) * 0.5;
		}
	else
		{
			double m;
			t_align /= (double)nr_vertex_endpoints;
			tol = tmax - t_align;
			m = t_align - tmin;
			if (m > tol) tol = m;
		}

#ifdef DEBUG
  fprintf(stderr, "file %s line %d: %d vertex endpoints found, tmin = %g, tmax = %g., t_align = %g, tol = %g\n",
	  __FILE__, __LINE__,
	  nr_vertex_endpoints, tmin, tmax, t_align, tol);
#endif

	// Align the endpoints. Enlarge the tolerance region of vertex
	// endpoints to include the aligned intersection position
	p = endpoints.begin() + start;
	for(int i = start; i <= end; i++, p++)
		{
			if(p->type & VERTEX_ENDPOINT)
				{
					// Enlarge the tolerance region of the vertex to include the intersection point
					VERTEX* v;
					double d;

					if(p->te == 0)
						v = (VERTEX*)(p->wing->vertex);
					else
						{
							if(p->te != 1)
								Warning("MergeInterval", "vertex endppoint with te != 0 and != 1");
							v = (VERTEX*)(p->wing->next->vertex);
						}

					d = (t_align > p->ti ? t_align - p->ti : p->ti - t_align);
					p->tol = (d >= p->tol ? d : p->tol);
					if(p->tol > v->tolerance)
						v->EnlargeTolerance(p->tol);
				}
			else
				p->tol = tol;

			p->ti = t_align;
		}
}


// Secondary endpoint sorting:
// sorts the endpoints in the interval such that the endpoints induced by faceA
// all preceed the endpoints induced by faceB. Returns in firstA and lastA
// the index to the first and last endpoint induced by faceA and similar for faceB
static void SortEndpointFace(int start, int end,
			     int *firstA, int *lastA, int *firstB, int *lastB)
{
#ifdef DEBUG
  fprintf(stderr, "SortEndpointFace:\n");
#endif

	// *firstB is the index of the first endpoint induced by faceB
	*firstB = start;
	while(*firstB <= end && endpoints[*firstB].wing->contour->face != faceB)
		(*firstB)++;

	// *LastA is the index of the last endpoint induced by faceA
	*lastA = end;
	while(*lastA >= start && endpoints[*lastA].wing->contour->face != faceA)
		(*lastA)--;

	while(*lastA > *firstB)
		{
			std::swap(endpoints[*firstB], endpoints[*lastA]);

			while(*firstB <= end && endpoints[*firstB].wing->contour->face != faceB)
				(*firstB)++;
			while(*lastA >= start && endpoints[*lastA].wing->contour->face != faceA)
				(*lastA)--;
		}

	*firstA = start;
	*lastB = end;
}

// Exit endpoints must preceed phantom endpoints, phantom endpoints
// must preceed entry endpoints. This routine returns +1 if the
// order needs to be changed, 0 is the endpoints have the same
// type and -1 if the order is correct
static int CompareEndpointType(ENDPOINT p1, ENDPOINT p2)
{
	if((*p1).type == (*p2).type) return 0;
	if((*p1).type & EXIT_ENDPOINT) return -1;
	if((*p2).type & EXIT_ENDPOINT) return +1;
	if((*p1).type & PHANTOM_ENDPOINT) return -1;
	if((*p2).type & PHANTOM_ENDPOINT) return +1;
	if((*p1).type & ENTRY_ENDPOINT) return -1;
	if((*p2).type & ENTRY_ENDPOINT) return +1;

	// Other irrelevant endpoint type
	return 0;
}

// Tertiary sorting: first exit endpoints, then phantom endpoints,
// then entry endpoints. Note that the series to be sorted will
// be very small (at most a few: coinciding endpoints induced by
// the same face)
static void SortEndpointType(int first, int last)
{
#ifdef DEBUG
  fprintf(stderr, "SortEndpointType(%d, %d):\n", first, last);
#endif

	for(int i = first; i < last; i++)
		for(int j = i+1; j <= last; j++)
			if(CompareEndpointType(endpoints.begin()+i, endpoints.begin()+j) > 0)
				std::swap(endpoints[i], endpoints[j]);
}


// Annihilates pairs of endpoints that would cause zero length cutting intervals
static void AnnihilateEndpoints(int start, int end)
{
  int i, first_exit=end+1, last_exit=start-1, nr_exit = 0,
         first_phantom=end+1, last_phantom=start-1, nr_phantom = 0,
         first_entry=end+1, last_entry=start-1, nr_entry = 0;

#ifdef DEBUG
  fprintf(stderr, "AnnihilateEndpoints(%d, %d):\n", start, end);
#endif

  for (i=start; i<=end; i++) {
    if (endpoints[i].type & EXIT_ENDPOINT) {
      if (i < first_exit) first_exit = i;
      if (i > last_exit) last_exit = i;
      nr_exit ++;
    }
    else if (endpoints[i].type & PHANTOM_ENDPOINT) {
      if (i < first_phantom) first_phantom = i;
      if (i > last_phantom) last_phantom = i;
      nr_phantom ++;
    }
    else if (endpoints[i].type & ENTRY_ENDPOINT) {
      if (i < first_entry) first_entry = i;
      if (i > last_entry) last_entry = i;
      nr_entry ++;
    }
    else
      endpoints[i].type |= DELETED_ENDPOINT;
  }

  while (nr_phantom >= 2) {
    endpoints[first_phantom++].type |= DELETED_ENDPOINT;
    endpoints[first_phantom++].type |= DELETED_ENDPOINT;
    nr_phantom -= 2;

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: pair phantom endpoints annihilated\n", __FILE__, __LINE__);
#endif
  }

  while (nr_exit > 1 && nr_entry > 1) {
    endpoints[first_exit++].type |= DELETED_ENDPOINT;
    endpoints[first_entry++].type |= DELETED_ENDPOINT;
    nr_exit--; nr_entry--;

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: entry/exit endpoint pair annihilated\n", __FILE__, __LINE__);
#endif
  }
}


// Given start and end index of a series of approximately coinciding endpoints,
// align the endpoints, (secondary) sort them according to their contributing
// face, (tertiary) sort them according to their type and remove endpoints that
// would lead to zero length cutting intervals
static void MergeInterval(int start, int end)
{
	int firstA, firstB, lastA, lastB;

#ifdef DEBUG
  fprintf(stderr, "--- MergeInterval(%d,%d):\n", start, end);
#endif

	AlignEndpoints(start, end);
	SortEndpointFace(start, end, &firstA, &lastA, &firstB, &lastB);

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: after secondary sorting: firstA = %d, lastA = %d, firstB = %d, lastB = %d\n",
	  __FILE__, __LINE__, firstA, lastA, firstB, lastB);
#endif // DDEBUG

	SortEndpointType(firstA, lastA);
	SortEndpointType(firstB, lastB);

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: after tertiary sorting:\n", __FILE__, __LINE__);
  for(ENDPOINT p = endpoints.begin()+start; p != NULLENDPOINT && p != endpoints.begin()+end; p++)
    PrintEndpointData(std::cerr, p);
#endif

	AnnihilateEndpoints(firstA, lastA);
	AnnihilateEndpoints(firstB, lastB);

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: after annihilating:\n", __FILE__, __LINE__);
  for(ENDPOINT p = endpoints.begin()+start; p != NULLENDPOINT && p != endpoints.begin()+end; p++)
    PrintEndpointData(std::cerr, p);
#endif
}

// Aligns coinciding endpoints, (secondary) sorts them according to their
// contributing face, (tertiary) sorts them according to their type and
// removes endpoints that would lead to zero length cutting intervals
static void MergeEndpoints()
{
#ifdef DDEBUG
fprintf(stderr, "MergeSortAnnihilateEndpoints:\n");
#endif

	// Index of start of series of approximately coinciding endpoints
	unsigned int start = 0;
	ENDPOINT prev = endpoints.begin();
	ENDPOINT p = prev+1;
	for(unsigned int i = 1; i < endpoints.size(); prev = p, i++, p++)
		if(!CoincidingEndpoints(prev, p))
			{
				if(i-1 > start)
					MergeInterval(start, i-1);

				start = i;
			}

	if(endpoints.size()-1 > start)
		MergeInterval(start, endpoints.size()-1);
}


// Sorts the endpoints contributed by the intersecting faces
static void SortEndpoints()
{
#ifdef DEBUG
fprintf(stderr, "SortEndpoints: %d endpoints\n", endpoints.size());
#endif

	sort(endpoints.begin(), endpoints.end());

#ifdef DEBUG
  fprintf(stderr, "file %s line %d: after primary sorting:\n", __FILE__, __LINE__);
  for(ENDPOINT p = endpoints.begin(); p != NULLENDPOINT && p != endpoints.end(); p++)
	PrintEndpointData(std::cerr, p);
#endif

	MergeEndpoints();

#ifdef DEBUG
  fprintf(stderr, "file %s line %d: after merging etc...:\n", __FILE__, __LINE__);
  for(ENDPOINT p = endpoints.begin(); p != NULLENDPOINT && p != endpoints.end(); p++)
	PrintEndpointData(std::cerr, p);
#endif
}

// Given the endpoint data p for an endpoint, this routine computes what
// vertex data correspond to it, looks up whether a vertex with this data
// already exists in the solid to which the face belongs, and if not found,
// creates such a vertex. A pointer to the vertex is returned. The face does
// not have to be the same face as the face contributing the endpoint
static BREP_VERTEX* MakeVertex(FACE* face, ENDPOINT p)
{
#ifdef DEBUG
  fprintf(stderr, "MakeVertex:\n");
  fprintf(stderr, "\tface = %s\n", face == faceA ? "faceA" : (face == faceB ? "faceB" : "NOT faceA or faceB"));
  fprintf(stderr, "\t"); PrintEndpointData(std::cerr, p); fprintf(stderr, "\n");
#endif

	VERTEX* thevertex;
	if(((*p).type & VERTEX_ENDPOINT) && (*p).wing->contour->face == face)
		{
			if((*p).te == 0)
				thevertex = (VERTEX*)((*p).wing->vertex);
			else
				thevertex = (VERTEX*)((*p).wing->next->vertex);
		}
	else
		{
			// Endpoint somewhere halfway the edge or on the other face
			VERTEX* v1 = (VERTEX *)((*p).wing->vertex);
			VERTEX* v2 = (VERTEX *)((*p).wing->next->vertex);

			// Sometimes, MakeVertex() is called to create a new vertex at
			// the end of the previous cutting interval for a face. Return
			// the previous end vertex in this case instead of creating a
			// new vertex at the same position

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: face = %s, p->ti = %f, lastti = %f\n", __FILE__, __LINE__,
	  (face == faceA ? "faceA" : (face == faceB ? "faceB" : "face ???")),
	  (*p).ti, lastti);
  fprintf(stderr, "\tlastwingA = "); PrintWingShort(std::cerr, lastwingA); fprintf(stderr, "\n");
  fprintf(stderr, "\tlastwingB = "); PrintWingShort(std::cerr, lastwingB); fprintf(stderr, "\n");
#endif

			if((*p).ti == lastti)
				{
					if(face == faceA && lastwingA)
						{
#ifdef DEBUG
	fprintf(stderr, "file %s line %d: new vertex = lastwingA->vertex\n", __FILE__, __LINE__);
#endif

							return lastwingA->vertex;
						}
					else if(face == faceB && lastwingB)
						{
#ifdef DEBUG
	fprintf(stderr, "file %s line %d: new vertex = lastwingB->vertex\n", __FILE__, __LINE__);
#endif

							return lastwingB->vertex;
						}
				}

			//?? v.id = p - endpoints;
			VERTEX* v = new VERTEX();
			v->point = isect.V + ((*p).ti)*isect.U;

			if((*p).wing->contour->face == face)
				{
					// Compute a nice interpolated normal
					v->normal = (1-(*p).te)*v1->normal;
					v->normal = v->normal + ((*p).te)*v2->normal;
					(v->normal).Normalize();
				}
			else
				{
					if((*p).type & VERTEX_ENDPOINT)
						{
							if((*p).te == 0)
								v->point = v1->point;
							else
								v->point = v2->point;
						}

					v->normal = VECTOR(0, 0, 0);
				}

			v->tolerance = (*p).tol;

#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: vertex data = ", __FILE__, __LINE__);
    std::cerr << *v << std::endl;
#endif

			thevertex = (VERTEX*)BrepFindVertex(v, face->shell->solid->vertices);

#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: vertex %s.\n", __FILE__, __LINE__,
	    thevertex ? "found" : "not found --- installing");
#endif

			if(thevertex == NULL)
				{
					thevertex = v;
					thevertex->DataCloseCallback();
					BrepAttachVertex(thevertex, face->shell->solid->vertices);

#ifdef DEBUG
      fprintf(stderr, "file %s line %d: new ", __FILE__ ,__LINE__);
	std::cerr << *thevertex << std::endl;
#endif
				}
		}

	return thevertex;
}

static void PrintWingStderr(BREP_WING* wing)
{
	PrintWingShort(std::cerr, wing); fprintf(stderr, "\n");
}

// Creates a BREP_VERTEX corresponding to the intersection point and
// belonging to the solid that contains the face that contributed the
// intersection point. Unless this vertex is the endpoint of the
// contributing wing, the contributing edge is split (on both sides
// if used in two contours). The, possibly new, wing leaving at the
// intersection point is returned. This routine also replaces all
// other references to the split wing in the (global) endpoint list.
// Since the intersection point contributed by a wing is unique (within
// the tolerances of course), the edge line parameter (te) of the
// updates endpoints is set to zero: the intersection point is the starting
// point of the newly created wing that is referenced
static BREP_WING* MakeVertexSplitEdge(ENDPOINT p)
{
#ifdef DEBUG
  fprintf(stderr, "MakeVertexSplitEdge: ");
  PrintEndpointData(std::cerr, p);
#endif

	BREP_VERTEX* vertex = MakeVertex((FACE*)(*p).wing->contour->face, p);

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: the vertex:", __FILE__, __LINE__);
	std::cerr << *((VERTEX*)(vertex));
	std::cerr << std::endl << "\tWings leaving at the vertex:" << std::endl;
	//vertex->IterateWings(PrintWingStderr);
  fprintf(stderr, "file %s line %d: the wing:", __FILE__, __LINE__);
  PrintWing(std::cerr, (*p).wing);
#endif

	BREP_WING* wing;
	if(vertex == (*p).wing->vertex)
		wing = (*p).wing;
	else if (vertex == (*p).wing->next->vertex)
		wing = (*p).wing->next;
	else
		{
			BREP_WING* otherwing = BrepEdgeOtherWing((*p).wing);
			BREP_WING* newotherwing = NULL;

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: before BrepSplitWing(otherwing)\n", __FILE__, __LINE__);
  PrintWing(std::cerr, otherwing);
#endif

			if(otherwing->contour)
				{
					newotherwing = BrepSplitWing(otherwing, vertex);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings ('other' side):\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, newotherwing->prev);
    PrintWingShort(std::cerr, newotherwing);
    fprintf(stderr, "\n");
#endif
				}

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: before BrepSplitWing(p->wing)\n", __FILE__, __LINE__);
#endif
    wing = BrepSplitWing((*p).wing, vertex);
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings ('this' side):\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, wing->prev);
    PrintWingShort(std::cerr, wing);
    fprintf(stderr, "\n");
#endif

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: updating endpoint list.\n", __FILE__, __LINE__);
#endif

			// Replace all references to wings of the split edge in the endpoint list.
			for(unsigned int i = 0; i < endpoints.size(); i++)
				{
					if(endpoints[i].wing == (*p).wing)
						{
							endpoints[i].wing = wing;
							// Should be the same intersection point
							endpoints[i].te = 0;

#ifdef DDEBUG
	fprintf(stderr, "file %s line %d: endpoint %d updated ('this' side).\n", __FILE__, __LINE__, i);
#endif
						}
					else if(endpoints[i].wing == otherwing)
						{
							endpoints[i].wing = newotherwing;
							// Should be the same intersection point
							endpoints[i].te = 0;

#ifdef DDEBUG
	fprintf(stderr, "file %s line %d: endpoint %d updated ('other' side).\n", __FILE__, __LINE__, i);
#endif
						}
				}
		}

#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: wing = ", __FILE__, __LINE__);
  PrintWing(std::cerr, wing);
#endif

	return wing;
}

// Introduces a cut from the starting vertex of wing1 to the starting vertex of
// wing2. Assumes that the starting vertex of wing1 comes before the starting
// vertex of wing2 on the intersection line. Returns a wing leaving at the
// exit end of the cut. Also replaces contributing wings in the global
// endpoint list to belong to the right contour for further processing
static BREP_WING* MakeCut(BREP_WING* wing1, BREP_WING* wing2)
{
#ifdef DEBUG
  fprintf(stderr, "MakeCut:\n");
  fprintf(stderr, "\twing1       = "); PrintWingShort(std::cerr, wing1); fprintf(stderr, "\n");
  fprintf(stderr, "\twing2       = "); PrintWingShort(std::cerr, wing2); fprintf(stderr, "\n");
#endif

	BREP_WING* wing;
	if(wing1->contour == wing2->contour)
		{

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: before BrepMakeEdgeSplitContour():\n", __FILE__, __LINE__);
#endif
#ifdef DDEBUG
    fprintf(stderr, "\twing1->prev = "); PrintWingShort(std::cerr, wing1->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing1       = "); PrintWingShort(std::cerr, wing1); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2->prev = "); PrintWingShort(std::cerr, wing2->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2       = "); PrintWingShort(std::cerr, wing2); fprintf(stderr, "\n");
#endif

			wing = BrepMakeEdgeSplitContour(wing1, wing2);

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, wing);
    PrintWingShort(std::cerr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: after BrepMakeEdgeSplitContour():\n", __FILE__, __LINE__);
    fprintf(stderr, "\t  wing->prev = "); PrintWingShort(std::cerr, wing->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing       = "); PrintWingShort(std::cerr, wing); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing->next = "); PrintWingShort(std::cerr, wing->next); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->prev = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing       = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->next = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)->next); fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: checking wing1->contour:\n", __FILE__, __LINE__);
    //CheckContour(wing1->contour);
    fprintf(stderr, "file %s line %d: checking wing2->contour:\n", __FILE__, __LINE__);
    //CheckContour(wing2->contour);
    fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
    //CheckVertex(wing1->vertex);
    //CheckVertex(wing2->vertex);
#endif

			// Update the endpoints list: the exitant edge at p2 must be an
			// edge on the contour that is still to be processed. Failing to
			// do so may cause the split of contour to be merged again with
			// the original contour when processing the next cutting interval
			// if it starts at the exit endpoint of the one being processed.
			// This situation occurs may occur when an isolated vertex of one
			// face lies in the other face. Cfr. Segal IEEE CG&A jan. 88 p 60
			if(((FACE*)wing2->contour->face)->ConvexCorner((VERTEX*)(wing1->vertex), (VERTEX*)(wing2->vertex), (VERTEX*)(wing2->next->vertex)))
				{
					BREP_WING* owing = BrepEdgeOtherWing(wing);
					for(unsigned int i = 0; i < endpoints.size(); i++)
						{
							if(endpoints[i].wing == wing2)
								{
									endpoints[i].wing = owing;

#ifdef DEBUG
	fprintf(stderr, "file %s line %d: endpoint %d updated: wing ", __FILE__, __LINE__, i);
	PrintWingShort(std::cerr, wing2);
	fprintf(stderr, " -> ");
	PrintWingShort(std::cerr, owing);
	fprintf(stderr, "\n");
#endif
								}
						}

						wing2 = owing;
				}
		}
	else
		{
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: before BrepMakeEdgeJoinContours():\n", __FILE__, __LINE__);
#endif
#ifdef DDEBUG
    fprintf(stderr, "\twing1->prev = "); PrintWingShort(std::cerr, wing1->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing1       = "); PrintWingShort(std::cerr, wing1); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2->prev = "); PrintWingShort(std::cerr, wing2->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\twing2       = "); PrintWingShort(std::cerr, wing2); fprintf(stderr, "\n");
#endif

			wing = BrepMakeEdgeJoinContours(wing1, wing2);

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, wing);
    PrintWingShort(std::cerr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: after BrepMakeEdgeSplitContour():\n", __FILE__, __LINE__);
    fprintf(stderr, "\t  wing->prev = "); PrintWingShort(std::cerr, wing->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing       = "); PrintWingShort(std::cerr, wing); fprintf(stderr, "\n");
    fprintf(stderr, "\t  wing->next = "); PrintWingShort(std::cerr, wing->next); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->prev = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)->prev); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing       = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)); fprintf(stderr, "\n");
    fprintf(stderr, "\to wing->next = "); PrintWingShort(std::cerr, BrepEdgeOtherWing(wing)->next); fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: checking wing1->contour:\n", __FILE__, __LINE__);
    //CheckContour(wing1->contour);
    fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
    //CheckVertex(wing1->vertex);
    //CheckVertex(wing2->vertex);
#endif
		}

	MarkContours(wing);

	// The possibly updated wing leaving at the exit endpoint
	return wing2;
}


// Introduces a notch from the starting vertex of wing1 to v2. Assumes that
// the starting vertex of wing1 comes before v2 on the intersection line
static BREP_WING* MakeNotch1(BREP_WING* wing1, BREP_VERTEX* v2)
{
#ifdef DEBUG
  fprintf(stderr, "MakeNotch1:\n");
  fprintf(stderr, "\twing1       = "); PrintWingShort(std::cerr, wing1); fprintf(stderr, "\n");
  fprintf(stderr, "\tv2          = "); std::cerr << *((VERTEX*)(v2)) << std::endl;
#endif

	BREP_WING* wing = BrepMakeNotch(wing1, v2);

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, wing);
    PrintWingShort(std::cerr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: checking contour:\n", __FILE__, __LINE__);
  //CheckContour(wing1->contour);
  fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
  //CheckVertex(wing1->vertex);
  //CheckVertex(v2);
#endif

	MarkContours(wing);

	// Wing leaving at (exit endpoint) v2
	return wing->next;
}

// Introduces a notch from v1 to the starting vertex of wing2. Assumes that
// v1 comes before the starting vertex of wing2 on the intersection line
static BREP_WING* MakeNotch2(BREP_VERTEX* v1, BREP_WING* wing2)
{
#ifdef DEBUG
  fprintf(stderr, "MakeNotch2:\n");
  fprintf(stderr, "\tv1          = "); std::cerr << *((VERTEX*)(v1)) << std::endl;
  fprintf(stderr, "\twing2       = "); PrintWingShort(std::cerr, wing2); fprintf(stderr, "\n");
#endif

	BREP_WING* wing = BrepMakeNotch(wing2, v1);

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, wing);
    PrintWingShort(std::cerr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: checking contour:\n", __FILE__, __LINE__);
  //CheckContour(wing2->contour);
  fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
  //CheckVertex(v1);
  //CheckVertex(wing2->vertex);
#endif

	MarkContours(wing);
	return wing2;
}


// Introduces a slit in the face between the two vertices.
// Assumes that v1 comes before v2 on the intersection line
static BREP_WING* MakeSlit(FACE* face, BREP_VERTEX* v1, BREP_VERTEX* v2)
{
#ifdef DEBUG
  fprintf(stderr, "MakeSlit:\n");
  fprintf(stderr, "\tv1          = "); std::cerr << *((VERTEX*)(v1)) << std::endl;
  fprintf(stderr, "\tv2          = "); std::cerr << *((VERTEX*)(v2)) << std::endl;
#endif

	BREP_WING* wing = BrepMakeSlit(face, v1, v2);

#ifdef DEBUG
    fprintf(stderr, "file %s line %d: new wings:\n", __FILE__, __LINE__);
    PrintWingShort(std::cerr, wing);
    PrintWingShort(std::cerr, BrepEdgeOtherWing(wing));
    fprintf(stderr, "\n");
#endif
#ifdef DDEBUG
  fprintf(stderr, "file %s line %d: checking face:\n", __FILE__, __LINE__);
//  CheckFace(face);
  fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
  //CheckVertex(v1);
  //CheckVertex(v2);
#endif

	MarkContours(wing);

	// Wing leaving at (exit endpoint) v2 going to v1
	return wing->next;
}



// The routines Cut(), Notch() and Slit() introduce a cut, notch or slit in a
// face between endpoints p1 and p2. They return a BREP_WING leaving at the exit end
// of the cut, notch or slit, i.e. with highest parameter along the intersection
// line. This vertex may be the start vertex of a next cut, notch or slit.

// Both endpoints are contributed by the face. This routine will insert a cut in
// the face. Assumes that p1 preceeds p2, i.o.w. p1 is the entry endpoint and p2
// is the exit endpoint
static BREP_WING* Cut(FACE* face, ENDPOINT p1, ENDPOINT p2)
{
#ifdef DEBUG
  fprintf(stderr, "--- Cut:\n");
  fprintf(stderr, "face = "); std::cerr << *face << std::endl;
  fprintf(stderr, "p1 = "); PrintEndpointData(std::cerr, p1);
  fprintf(stderr, "p2 = "); PrintEndpointData(std::cerr, p2);
#endif

	BREP_WING* wing1 = MakeVertexSplitEdge(p1);
	if(wing1 == (*p2).wing)
		{
			// The point at which p1->wing was to be split
			// was its end vertex, so MakeVertexSplitEdge(p1)
			// returns p2->wing

#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: before MakeVertexSplitEdge():\n", __FILE__, __LINE__);
#endif

			BREP_WING* wing2 = MakeVertexSplitEdge(p2);

#ifdef DDEBUG
    fprintf(stderr, "file %s line %d: checking wing2->contour:\n", __FILE__, __LINE__);
    //CheckContour(wing2->contour);
    fprintf(stderr, "file %s line %d: checking vertices:\n", __FILE__, __LINE__);
    //CheckVertex(wing2->vertex);
    //CheckVertex(wing2->next->vertex);
    //CheckVertex(wing2->prev->vertex);
#endif

			return wing2;
		}

	BREP_WING* wing2 = MakeVertexSplitEdge(p2);

	if((*p1).ti < (*p2).ti)
		return MakeCut(wing1, wing2);
	else
		return MakeCut(wing2, wing1);
}


// Only the first endpoint was contributed by the face, the second not.
// This routine inserts a notch in the face, splitting p1->wing in two at
// the intersection point and connecting to the new vertex defined by p2
static BREP_WING* Notch(FACE* face, ENDPOINT p1, ENDPOINT p2)
{
#ifdef DEBUG
  fprintf(stderr, "--- Notch:\n");
  fprintf(stderr, "face = "); std::cerr << *face << std::endl;
  fprintf(stderr, "p1 = "); PrintEndpointData(std::cerr, p1);
  fprintf(stderr, "p2 = "); PrintEndpointData(std::cerr, p2);
#endif

	BREP_WING* wing = MakeVertexSplitEdge(p1);
	BREP_VERTEX* v = MakeVertex(face, p2);

	if((lastwingA && v == lastwingA->vertex) ||
		(lastwingB && v == lastwingB->vertex))
		{
			// A notch to the end vertex of the previous
			// cutting interval cuts the contour
			BREP_WING* wing2 = ((lastwingA && v == lastwingA->vertex) ? lastwingA : lastwingB);
			if(p1->ti < p2->ti)
				return MakeCut(wing, wing2);
			else
				return MakeCut(wing2, wing);
		}

	if((*p1).ti < (*p2).ti)
		return MakeNotch1(wing, v);
	else
		return MakeNotch2(v, wing);
}

// None of the endpoints was contributed by the face. This routine inserts a slit
// in the face. It is assumed that p1 preceeds p2 on the intersection line, i.o.w.
// p1 is the entry endpoint and p2 is the exit endpoint
static BREP_WING* Slit(FACE* face, ENDPOINT p1, ENDPOINT p2)
{
#ifdef DEBUG
  fprintf(stderr, "--- Slit:\n");
  fprintf(stderr, "face = "); std::cerr << *face << std::endl;
  fprintf(stderr, "p1 = "); PrintEndpointData(std::cerr, p1);
  fprintf(stderr, "p2 = "); PrintEndpointData(std::cerr, p2);
#endif

	BREP_VERTEX* v1 = MakeVertex(face, p1);
	BREP_VERTEX* v2 = MakeVertex(face, p2);

	if((lastwingA && v1 == lastwingA->vertex) ||
		(lastwingB && v1 == lastwingB->vertex))
		{
			// A slit including the endpoint of a previous
			// cutting interval is a notch
			BREP_WING *wing1 = ((lastwingA && v1 == lastwingA->vertex) ? lastwingA : lastwingB);
			return MakeNotch1(wing1, v2);
		}

	return MakeSlit(face, v1, v2);
}

// This routine inserts cuts, notches and slits in the faces as indicated by
// the endpoint multiplicities. There is at least one and at most two entry or exit
// endpoints for a cutting interval: at least from one and at most from both faces.
// If there is both an entry and exit endpoint for a given face, the face needs
// to be cut. If there is only one entry or exit endpoint for a face, a notch
// needs to be inserted. If a face didn't contribute neither an entry and neither
// an exit endpoint, a slit needs to be inserted in the face (In that case, the other
// face has contributed both entry and exit endpoint and will be cut). Phantom
// endpoints should come in pairs. If not, there is something wrong with the endpoint
// determination or sorting. A pair of phantom endpoints induces a cut, notch or slit
// in the face that didn't contribute the endpoints, but the face that contributed
// them is left intact. This routine is called with the last encountered entry and
// first encountered exit points on both faces at the start and end of a cutting
// interval.
static void CuttingInterval(ENDPOINT entryA, ENDPOINT entryB, ENDPOINT exitA, ENDPOINT exitB)
{
	// entry and exit endpoints of the cutting interval
	ENDPOINT ci_startA, ci_startB, ci_endA, ci_endB;
	ENDPOINT entry, exit;

#ifdef DEBUG
  fprintf(stderr, "--- CuttingInterval:\n");
  fprintf(stderr, "entryA = "); PrintEndpointData(std::cerr, entryA);
  fprintf(stderr, "entryB = "); PrintEndpointData(std::cerr, entryB);
  fprintf(stderr, "exitA = "); PrintEndpointData(std::cerr, exitA);
  fprintf(stderr, "exitB = "); PrintEndpointData(std::cerr, exitB);
#endif

	// Check if an entry and exit point is given for both faces.
	if(entryA == NULLENDPOINT || entryB == NULLENDPOINT || exitA == NULLENDPOINT || exitB == NULLENDPOINT)
		{
			Warning("CuttingInterval", "missing endpoints");
			return;
		}

	// Compare the last encountered entry points on face A and face B:
	// if they are at the same position, keep both as starting
	// points of a cutting interval. If not, save only the one with
	// largest parameter along the intersection line
	ci_startA = ci_startB = NULLENDPOINT;
	if((*entryA).ti >= (*entryB).ti)
		ci_startA = entryA;
	if((*entryB).ti >= (*entryA).ti)
		ci_startB = entryB;

	entry = (ci_startA != NULLENDPOINT ? ci_startA : ci_startB);

	// Compare the endpoints. Keep both if they are at the same position, or only
	// the one at with smallest parameter along the intersection line if different
	ci_endA = ci_endB = NULLENDPOINT;
	if((*exitA).ti <= (*exitB).ti)
		ci_endA = exitA;
	if((*exitB).ti <= (*exitA).ti)
		ci_endB = exitB;

	exit = (ci_endA != NULLENDPOINT ? ci_endA : ci_endB);

#ifdef DEBUG
  fprintf(stderr, "ci_startA = "); PrintEndpointData(std::cerr, ci_startA);
  fprintf(stderr, "ci_startB = "); PrintEndpointData(std::cerr, ci_startB);
  fprintf(stderr, "ci_endA = "); PrintEndpointData(std::cerr, ci_endA);
  fprintf(stderr, "ci_endB = "); PrintEndpointData(std::cerr, ci_endB);
#endif

	if(CoincidingEndpoints(entry, exit))
		{
			// Don't create zero length features
#ifdef DEBUG
    fprintf(stderr, "Coinciding endpoints ... \n");
#endif
			return;
		}

	if((*entry).ti > (*exit).ti)
		{
			Warning("CuttingInterval", "cutting interval determination incorrect");
			return;
		}

	// Check if phantom endpoints come in pairs
	if((entryA->type & PHANTOM_ENDPOINT) != (exitA->type & PHANTOM_ENDPOINT) ||
		(entryB->type & PHANTOM_ENDPOINT) != (exitB->type & PHANTOM_ENDPOINT))
		{
			Warning("CuttingInterval", "phantom endpoints should come in pairs");
		}

	// If not a pair of phantom endpoints, it must be
	// a pair of one entry and one exit endpoint
	else if (!(entryA->type & ENTRY_ENDPOINT) != !(exitA->type & EXIT_ENDPOINT) ||
		!(entryB->type & ENTRY_ENDPOINT) != !(exitB->type & EXIT_ENDPOINT))
		{
			Warning("CuttingInterval", "entry/exit property of endpoints incorrectly determined");
		}

#ifdef CHECK
  if (check) {
#ifdef DEBUG
    fprintf(stderr, "file %s line %d: just checking ...\n", __FILE__, __LINE__);
#endif
    return;
  }
#endif // CHECK

	// Partition face A
	if(!(entryA->type & PHANTOM_ENDPOINT) && !(exitA->type & PHANTOM_ENDPOINT))
		{
			if(ci_startA != NULLENDPOINT && ci_endA != NULLENDPOINT)
				lastwingA = Cut(faceA, ci_startA, ci_endA);
			else if(ci_startA != NULLENDPOINT)
				lastwingA = Notch(faceA, ci_startA, ci_endB);
			else if(ci_endA != NULLENDPOINT)
				lastwingA = Notch(faceA, ci_endA, ci_startB);
			else
				lastwingA = Slit(faceA, ci_startB, ci_endB);
		}

	// Partition face B
	if(!(entryB->type & PHANTOM_ENDPOINT) && !(exitB->type & PHANTOM_ENDPOINT))
		{
			if(ci_startB != NULLENDPOINT && ci_endB != NULLENDPOINT)
				lastwingB = Cut(faceB, ci_startB, ci_endB);
			else if(ci_startB != NULLENDPOINT)
				lastwingB = Notch(faceB, ci_startB, ci_endA);
			else if(ci_endB != NULLENDPOINT)
				lastwingB = Notch(faceB, ci_endB, ci_startA);
			else
				lastwingB = Slit(faceB, ci_startA, ci_endA);
		}

	lastti = (*exit).ti;

#ifdef DEBUG
  fprintf(stderr, "file %s line %d: after cutting:\n", __FILE__, __LINE__);
  fprintf(stderr, "\tlastwingA = "); PrintWingShort(std::cerr, lastwingA); fprintf(stderr, "\n");
  fprintf(stderr, "\tlastwingB = "); PrintWingShort(std::cerr, lastwingB); fprintf(stderr, "\n");
  fprintf(stderr, "\tlastti = %f\n", lastti);
#endif
}

// Traverses the intersection line, determines entry and exit
// endpoints of cutting intervals, and processes cutting intervals
static void TraverseIntersectionline()
{
#ifdef DEBUG
  fprintf(stderr, "TraverseIntersectionline:\n");
#endif

	// Initialize intersection line parameter of and wings in faceA and faceB
	// leaving at the exit endpoint of the previous cutting interval
	lastwingA = lastwingB = NULL;
	lastti = -HUGE;

	// Last encountered entry and exit endpoint on face A and B
	ENDPOINT entryA = NULLENDPOINT;
	ENDPOINT entryB = NULLENDPOINT;
	ENDPOINT exitA = NULLENDPOINT;
	ENDPOINT exitB = NULLENDPOINT;

	// Potential entry and exit endpoint for cutting interval
	ENDPOINT ci_startA = NULLENDPOINT;
	ENDPOINT ci_startB = NULLENDPOINT;
	ENDPOINT ci_endA = NULLENDPOINT;
	ENDPOINT ci_endB = NULLENDPOINT;

	// Flags indicating whether we are inside/outside faceA and faceB
	bool inA = false;
	bool inB = false;
	// Flag indicating whether we are currently in a cutting interval
	bool incut = false;

	// Loop over the sorted endpoints
	ENDPOINT endpoint = endpoints.begin();
	for(unsigned int i = 0; i < endpoints.size(); i++, endpoint++)
		{
#ifdef DEBUG
    fprintf(stderr, "--- i=%d ", i);
    PrintEndpointData(std::cerr, endpoint);
#endif

			// Skip deleted endpoints or purely VERTEX endpoints,
			// which are just there for alignement
			if((endpoint->type & DELETED_ENDPOINT) || (endpoint->type == VERTEX_ENDPOINT))
				continue;

			// Flip the state of the inA or inB flag, update the last
			// encountered entry or exit endpoint for the face accordingly.
			// Forget the last exit endpoint when an entry endpoint is
			// found and vice versa.
			if(endpoint->wing->contour->face == faceA)
				{
					inA = 1 - inA;
					if(inA)
						{
							entryA = endpoint;
							exitA = NULLENDPOINT;
						}
					else
						{
							entryA = NULLENDPOINT;
							exitA = endpoint;
						}
				}
			else if(endpoint->wing->contour->face == faceB)
				{
					inB = 1 - inB;
					if(inB)
						{
							entryB = endpoint;
							exitB = NULLENDPOINT;
						}
					else
						{
							entryB = NULLENDPOINT;
							exitB = endpoint;
						}
				}
			else
				{
					// This should never occur
					Warning("TraverseIntersectionline", "endpoint present from unknown face");
					// Ignore it
					continue;
				}

#ifdef DEBUG
    fprintf(stderr, "inA = %d, inB = %d, incut = %d.\n", inA, inB, incut);
#endif

			if(entryA != NULLENDPOINT && entryB != NULLENDPOINT)
				{
					// Start of a new cutting interval
					ci_startA = entryA;
					ci_startB = entryB;
					incut = true;
				}
			else if(incut && (exitA != NULLENDPOINT || exitB != NULLENDPOINT))
				{
					// End of cutting interval

					// Look ahead to find the first exit endpoint
					// induced by the other face
					ci_endA = exitA;
					for(unsigned int j = i+1; ci_endA == NULLENDPOINT && j < endpoints.size(); j++)
						{
							ENDPOINT endq = endpoints.begin()+j;
							if((*endq).wing->contour->face == faceA &&
								(((*endq).type & PHANTOM_ENDPOINT) || (endq->type & EXIT_ENDPOINT)))
								ci_endA = endq;
						}

					ci_endB = exitB;
					for(unsigned int j = i+1; ci_endB == NULLENDPOINT && j < endpoints.size(); j++)
						{
							ENDPOINT endq = endpoints.begin()+j;
							if((*endq).wing->contour->face == faceB &&
								(((*endq).type & PHANTOM_ENDPOINT) || (endq->type & EXIT_ENDPOINT)))
								ci_endB = endq;
						}

					CuttingInterval(ci_startA, ci_startB, ci_endA, ci_endB);
					incut = false;
				}
		}

	// After traversal of the intersection line,
	// we should be outside of both faces again
	if(inA || inB)
		Warning("TraverseIntersectionline", "something wrong with endpoint determination");
}

// Partitions a pair of possibly transversal intersecting
// faces along the intersection line
static void TransversalIntersection(FACE* _faceA, BREP_WING* first_wingA,
				    FACE* _faceB, BREP_WING* first_wingB)
{
#ifdef DEBUG
	std::cerr << "-----> TransversalIntersection():" << std::endl;
#endif

	// Global variables not to have to pass them as a parameter every time
	faceA = _faceA;
	faceB = _faceB;

	// Compute the parametric equation t.U + V and
	// tolerance of the face intersection line
	isect = ComputePlaneToPlaneIntersection(faceA, faceB);

	// Determine cutting interval endpoints = intersections
	// of an edge of one face with the plane of the other
	endpoints.clear();
	timinA = timinB = HUGE;
	timaxA = timaxB = -HUGE;
	DetermineEndpoints(faceA, first_wingA, faceB);
	DetermineEndpoints(faceB, first_wingB, faceA);

	// If all endpoints of one face come before the endpoints
	// contributed by the other face, there is no intersection
	if(timinA > timaxB || timinB > timaxA)
		{
#ifdef DEBUG
    fprintf(stderr, "timinA=%f > timaxB=%f || timinB=%f > timaxA=%f -> no intersection\n",
	    timinA, timaxB, timinB, timaxA);
#endif
			return;
		}

	// Sort endpoint list
	SortEndpoints();

	// Traverse intersection line, insert cuts, notches and
	// slits in the faces as indicated by the endpoints
#ifdef CHECK
  check = true;
  TraverseIntersectionline();
  check = false;
#endif // CHECK
	TraverseIntersectionline();
}

// faceA and faceB are two faces for which the coplanarity factor exceeds the
// allowed minimum. This routine will return TRUE if faceA really can
// be considered to be within the plane of faceB:  that is, when
// two vertices in faceA lies on opposite sides of the plane of faceB, or
// when any vertex of faceA lies within a distance equal to the extent of faceB
static int TestCoplanarIntersection(FACE* faceA, FACE* faceB)
{
#ifdef DEBUG
	std::cerr << "-----> TestCoplanarIntersection:" << std::endl;
#endif

	// Some value != 0
	int side = UNKNOWN_SIDE;

	// Check the signed distance of all vertices of faceA to the plane of faceB
	if(faceA->outer_contour.empty())
		// Face without contours, ignore it
		return false;

	BREP_CONTOUR_RING::iterator contour;
	for(contour = faceA->outer_contour.begin(); contour != faceA->outer_contour.end(); contour++)
		if((*contour)->wings)
			{
				BREP_WING* wing = (*contour)->wings;
				faceB->SignedDistanceToPlane((VERTEX*)(wing->vertex), &side);
				if(side == 0)
					// Vertex within the extent of the plane of faceB
					return true;

				wing = wing->next;

				while(wing != (*contour)->wings)
					{
						int prevside = side;
						faceB->SignedDistanceToPlane((VERTEX*)(wing->vertex), &side);

						if(side != prevside || side == 0)
						// The starting vertex of wing was on the
						// other side of the plane of faceB than
						// the starting vertex of wing->prev or
						// the vertex was within the extent of
						// the plane of faceB
							return true;

						wing = wing->next;
					}

				// There's no need to check the last wing:
				// if the starting vertex of the last wing is on
				// the other side of the starting vertex of the first,
				// there is a crossing in one of the other wings as well
			}

	return false;
}

// faceA and faceB are two faces for which the coplanarity factor is lower than the
// minimum allowed coplanarity factor. This routine classifies faceA w.r.t. faceB:
// If an edge of faceA has endpoints on different
// sides of the plane of faceB, a pointer to the crossing edge is returned.
// If faceA is entirely on the same side of the plane of faceB NULL is returned.
// Two faces faceA and faceB can only intersect transversally if both faces
// have an edge that crosses the plane of the other.
static BREP_WING* TestTransversalIntersection(FACE* faceA, FACE* faceB)
{
#ifdef DEBUG
  fprintf(stderr, "-----> TestTransversalIntersection:\n");
#endif

	// Some value != 0
	int side = UNKNOWN_SIDE;

	// Check the signed distance of all vertices of faceA to the plane of faceB
	BREP_CONTOUR_RING::iterator contour;
	if(faceA->outer_contour.empty())
		// Face without contours, ignore it
		return NULL;

	for(contour = faceA->outer_contour.begin(); contour != faceA->outer_contour.end(); contour++)
		{
			BREP_WING* wing;

			if((*contour)->wings)
				{
					wing = (*contour)->wings;
					faceB->SignedDistanceToPlane((VERTEX*)(wing->vertex), &side);
					if(side == 0)
						return wing;

					wing = wing->next;

					while(wing != (*contour)->wings)
						{
							int prevside = side;
							faceB->SignedDistanceToPlane((VERTEX*)(wing->vertex), &side);

							if(side == 0 || side != prevside)
								// The starting vertex of wing was
								// on the other side of the plane
								// of faceB than the starting
								// vertex of wing->prev
								return wing->prev;

							wing = wing->next;
						}

					// There's no need to check the last wing: if the starting
					// vertex of the last wing is on the other side of the
					// starting vertex of the first,
					// there is a crossing in one of the other wings as well
				}
		}

	// No need to classify faceB w.r.t. to faceA
	return NULL;
}


// Given two FACES, with appropriate geometric data (see geometry.[ch]), this
// routine splits the faces into nonintersecting parts. Plane equations and tolerances
// have been computed when reading in the objects, in CloseFaceCallback()
int PartitionFaces(FACE* faceA, FACE* faceB)
{
#ifdef DEBUG
fprintf(stderr, "=====> PartitionFaces:\n");
fprintf(stderr, "\tfaceA = "); std::cerr << *faceA << std::endl;
fprintf(stderr, "\tfaceB = "); std::cerr << *faceB << std::endl;
fprintf(stderr, "Coplanarity factor = %g\n", CoplanarityFactor(faceA, faceB));
#endif

//  if (faceA->ShortestEdgeLength() < minedgelength ||
//      faceB->ShortestEdgeLength() < minedgelength)
//    return NO_ISECT;	// Do not consider small faces

	if(CoplanarityFactor(faceA, faceB) > min_coplanarity_factor)
		{
			if(TestCoplanarIntersection(faceA, faceB) ||
				TestCoplanarIntersection(faceB, faceA))
				return COPLANAR_ISECT;
		}
	else
		{
			BREP_WING* wingA = TestTransversalIntersection(faceA, faceB);
			BREP_WING* wingB = TestTransversalIntersection(faceB, faceA);
			if(wingA != NULL && wingB != NULL)
				TransversalIntersection(faceA, wingA, faceB, wingB);

			return TRANSVERSAL_ISECT;
		}

	return NO_ISECT;
}




#include <string.h>
#include <time.h>
#include <vector>


clock_t start_clock;
// To print statistics etc ...
int verbose;
void (*csg_select)(SOLID*, SOLID*);
char* progname;

// Ignore normals in input file
int ignore_normals_on_input;
// Ignore materials when outputting MGF
int ignore_materials;

// The algorithm requires to remember coplanar face pairs until partitioning
// and inside/outside mark propagation has been done
typedef std::pair<BREP_FACE*, BREP_FACE*> COPLANAR_FACES;
typedef std::vector<COPLANAR_FACES> COPLANAR_FACES_LIST;
COPLANAR_FACES_LIST coplanar_faces;


// Given a pair of coplanar faces, IterateMatchingContours() calls a given
// function for each pair of matching contours. Two contours match after
// partitioning as soon as they have one wing connecting the same vertices
// and their enclosed region is on the same side.
// See M. Segal, IEEE CG&A Jan '88 p 61-62
static void (*matching_contour_func)(BREP_CONTOUR*, BREP_CONTOUR*);
static BREP_CONTOUR* matching_contourA;
static BREP_FACE* matching_contour_faceB;

static void do_matching_contours(BREP_WING* wingB)
{
	BREP_CONTOUR* contourB = wingB->contour;

	if(contourB == NULL || contourB->face != matching_contour_faceB)
		return;

	// wingB has endpoints at the same location as the first wing of
	// matching_contourA. The contours match if the enclosed region is on the same
	// side, i.o.w. the first vertex of wingB is at the same location as the
	// first vertex of the first wing of matching_contourA
	if(VertexCompareLocation(matching_contourA->wings->vertex, wingB->vertex) == 8)
		matching_contour_func(matching_contourA, contourB);
}

static void IterateMatchingContours(BREP_FACE* faceA, BREP_FACE* faceB,
				    void (*func)(BREP_CONTOUR* contourA, BREP_CONTOUR* contourB))
{
	void (*old_matching_contour_func)(BREP_CONTOUR*, BREP_CONTOUR*) = matching_contour_func;
	matching_contour_func = func;
	matching_contour_faceB = faceB;

	BREP_CONTOUR_RING::iterator contourA;
	for(contourA = faceA->outer_contour.begin(); contourA != faceA->outer_contour.end(); contourA++)
		{
			if((*contourA)->wings && (*contourA)->wings->next != (*contourA)->wings->prev)
				{
					matching_contourA = *contourA;
					BrepIterateWingsBetweenLocations((*contourA)->wings->vertex,
						(*contourA)->wings->next->vertex,
						faceB->shell->solid->vertices,
						do_matching_contours);
				}
		}

	matching_contour_func = old_matching_contour_func;
}

// Clears the contour masks OUTMASK, INMASK, WRITEMASK, DONTWRITEMASK, WRITEREVERSEMASK
static void unset_mark(BREP_CONTOUR* contour)
{
	((CONTOUR*)(contour))->id &= ~(INMASK|OUTMASK|WRITEMASK|DONTWRITEMASK|WRITEREVERSEMASK);
}

// Returns the inside/outside mask of the contour
static unsigned int inoutmark(BREP_CONTOUR* contour)
{
	return ((CONTOUR*)(contour))->id & (INMASK|OUTMASK);
}

// Marks a contour with the given mask
static void mark_contour(BREP_CONTOUR* contour, unsigned int mask)
{
#ifdef DDEBUG
	std::cerr << "mark_contour: contour = " << *((CONTOUR*)(contour));
	std::cerr << ", mask = "
		<< (mask & OUTMASK ? "OUTMASK" : "")
		<< (mask & INMASK ? "INMASK" : "")
		<< (mask & WRITEMASK ? "WRITEMASK" : "")
		<< (mask & DONTWRITEMASK ? "DONTWRITEMASK" : "")
		<< (mask & WRITEREVERSEMASK ? "WRITEREVERSEMASK" : "")
		<< std::endl;
#endif

	((CONTOUR*)(contour))->id |= mask;
}

// ContourA and contourB are matching contours in coplanar faces
// Keep one if they have the same orientation
static void keep_one_same(BREP_CONTOUR* contourA, BREP_CONTOUR* contourB)
{
#ifdef DEBUG
	std::cerr << "keep_one_same: contourA = " << *((CONTOUR*)(contourA))
		<< ", contourB = " << *((CONTOUR*)(contourB)) << std::endl;
#endif

	if(((CONTOUR*)(contourA))->normal * ((CONTOUR*)(contourB))->normal > 0)
		{
			// Same orientation
			unset_mark(contourA);
			mark_contour(contourA, WRITEMASK);
			unset_mark(contourB);
			mark_contour(contourB, DONTWRITEMASK);
		}
	else
		{
			// Opposite orientation
			unset_mark(contourA);
			mark_contour(contourA, DONTWRITEMASK);
			unset_mark(contourB);
			mark_contour(contourB, DONTWRITEMASK);
		}
}

// ContourA and contourB are matching contours in coplanar faces
// Keep one if they have opposite orientation.
static void keep_one_opposite(BREP_CONTOUR* contourA, BREP_CONTOUR* contourB)
{
#ifdef DEBUG
	std::cerr << "keep_one_opposite: contourA = " << *((CONTOUR*)(contourA))
		<< ", contourB = " << *((CONTOUR*)(contourB)) << std::endl;
#endif

	if(((CONTOUR*)(contourA))->normal * ((CONTOUR*)(contourB))->normal > 0)
		{
			// Same orientation
			unset_mark(contourA);
			mark_contour(contourA, DONTWRITEMASK);
			unset_mark(contourB);
			mark_contour(contourB, DONTWRITEMASK);
		}
	else
		{
			// Opposite orientation
			unset_mark(contourA);
			mark_contour(contourA, WRITEMASK);
			unset_mark(contourB);
			mark_contour(contourB, DONTWRITEMASK);
		}
}

// Partition pairs of faces belonging to shellA and shellB
static void partition_shells(BREP_SHELL* shellA, BREP_SHELL* shellB)
{
	BREP_FACE_RING::iterator faceA, faceB;
	for(faceA = shellA->faces.begin(); faceA != shellA->faces.end(); faceA++)
		for(faceB = shellB->faces.begin(); faceB != shellB->faces.end(); faceB++)
			{
				FACE* faA = (FACE*)((*faceA));
				FACE* faB = (FACE*)((*faceB));
				if(!DisjunctBounds(faA->bounds, faB->bounds))
					if(PartitionFaces(faA, faB) == COPLANAR_ISECT)
						// Record a pair of coplanar faces
						coplanar_faces.push_back(COPLANAR_FACES(*faceA, *faceB));
			}
}

/*
void do_isect(SOLID* solid)
{
  // COPLANAR_FACELIST* copl;

	BREP_SHELL_RING::iterator shellA;
	for(shellA = solid->shells.begin(); shellA != solid->shells.end(); shellA++)
		{
			BREP_SHELL_RING::iterator shellB = shellA;
			shellB++;
			for(; shellB != solid->shells.end(); shellB++)
				{
					SHELL* shA = (SHELL*)((*shellA)->client_data);
					SHELL* shB = (SHELL*)((*shellB)->client_data);
					if(!DisjunctBounds(shA->bounds, shB->bounds))
						partition_shells(*shellA, *shellB);
				}
		}

	for(cfaces = coplanar_faces.begin(); cfaces != coplanar_faces.end(); cfaces++)
		IterateMatchingContours((*cfaces).first, (*cfaces).second, keep_one_same);
}
*/

// inside/outside mark propagation: when a contour is cut during face partitioning,
// the cut contour parts are marked whether they are inside or outside the transversal
// solid. After partitioning, thi mark can be propagated to neighbouring contours,
// reducing the number of contours for which an explicit inside/outside test will
// be necessary
static void contour_propagate_mark(BREP_CONTOUR* contour);
static void wing_propagate_mark(BREP_WING* wing, unsigned int mark)
{
	BREP_CONTOUR* ocontour = BrepEdgeOtherWing(wing)->contour;
	if(ocontour == NULL || inoutmark(ocontour))
		return;

	mark_contour(ocontour, mark);
	contour_propagate_mark(ocontour);
}

static void contour_propagate_mark(BREP_CONTOUR* contour)
{
	unsigned int mark = inoutmark(contour);

	if(mark != 0)
		contour->IterateWings1A((void (*)(BREP_WING*, void*))wing_propagate_mark, (void*)mark);
}

// Propagates inside/outside mark of contours for which it was
// determined during PartitionFaces() to neighbouring contours
static void csg_propagate_mark(SOLID* solid)
{
	solid->IterateContours(contour_propagate_mark);
}

// inside/outside mark propagation only reduces the number of contours for which
// an explicit inside/outside test will be necessary. After propagation, this
// test is necessary for all contours that remained unmarked

static SOLID* global_othersolid;

static void contour_mark_unmarked(BREP_CONTOUR* contour)
{
	if (inoutmark(contour))
		return;

	unsigned int mark;
	if((mark = ContourInSolid(contour, global_othersolid)))
		{
			mark_contour(contour, mark);
			contour_propagate_mark(contour);
		}
	else
		{
			// The contour is most probably coplanar with a contour
			// of the other solid and will be handled later
			Warning("contour_mark_unmarked", "can't determine whether contour is inside or outside other solid");
		}
}

static void csg_mark_unmarked(SOLID* solid, SOLID* othersolid)
{
	global_othersolid = othersolid;
	solid->IterateContours(contour_mark_unmarked);
}


// Mark a contour that is outside the transversal solid for writing
static void mark_out(BREP_CONTOUR* contour)
{
	if(inoutmark(contour) & OUTMASK)
		mark_contour(contour, WRITEMASK);
	else
		mark_contour(contour, DONTWRITEMASK);
}

// Mark a contour that is inside the transversal solid for writing
static void mark_in(BREP_CONTOUR* contour)
{
	if(inoutmark(contour) & INMASK)
		mark_contour(contour, WRITEMASK);
	else
		mark_contour(contour, DONTWRITEMASK);
}

// Mark a contour that is inside the transversal solid for reverse writing
static void mark_in_reverse(BREP_CONTOUR* contour)
{
	if(inoutmark(contour) & INMASK)
		mark_contour(contour, WRITEREVERSEMASK);
	else
		mark_contour(contour, DONTWRITEMASK);
}

static void select_out(SOLID* solid)
{
	solid->IterateContours(mark_out);
}

static void select_in(SOLID* solid)
{
	solid->IterateContours(mark_in);
}

static void select_in_reverse(SOLID* solid)
{
	solid->IterateContours(mark_in_reverse);
}

static void select_union(SOLID* solidA, SOLID* solidB)
{
	select_out(solidA);
	select_out(solidB);

	COPLANAR_FACES_LIST::iterator cfaces;
	for(cfaces = coplanar_faces.begin(); cfaces != coplanar_faces.end(); cfaces++)
		IterateMatchingContours((*cfaces).first, (*cfaces).second, keep_one_same);
}

static void select_difference(SOLID* solidA, SOLID* solidB)
{
	select_out(solidA);
	select_in_reverse(solidB);

	COPLANAR_FACES_LIST::iterator cfaces;
	for(cfaces = coplanar_faces.begin(); cfaces != coplanar_faces.end(); cfaces++)
		IterateMatchingContours((*cfaces).first, (*cfaces).second, keep_one_opposite);
}

static void select_intersection(SOLID* solidA, SOLID* solidB)
{
	select_in(solidA);
	select_in(solidB);

	COPLANAR_FACES_LIST::iterator cfaces;
	for(cfaces = coplanar_faces.begin(); cfaces != coplanar_faces.end(); cfaces++)
		IterateMatchingContours((*cfaces).first, (*cfaces).second, keep_one_same);
}

static void mark_with_outmask(BREP_CONTOUR* contour)
{
	mark_contour(contour, OUTMASK);
}


// CSG main program
static void do_csg(SOLID* solidA, SOLID* solidB)
{
	solidA->IterateContours(unset_mark);
	solidB->IterateContours(unset_mark);

	if(!DisjunctBounds(((SOLID*)(solidA))->bounds, ((SOLID*)(solidB))->bounds))
		{
			BREP_SHELL_RING::iterator shellA;
			for(shellA = solidA->shells.begin(); shellA != solidA->shells.end(); shellA++)
				{
					BREP_SHELL_RING::iterator shellB;
					for(shellB = solidB->shells.begin(); shellB != solidB->shells.end(); shellB++)
						{
							SHELL* shA = (SHELL*)(*shellA);
							SHELL* shB = (SHELL*)(*shellB);
							if(!DisjunctBounds(shA->bounds, shB->bounds))
								partition_shells(*shellA, *shellB);
						}
				}
		}
	else
		{
			solidA->IterateContours(mark_with_outmask);
			solidB->IterateContours(mark_with_outmask);
		}

	csg_propagate_mark(solidA);
	csg_mark_unmarked(solidA, solidB);
	csg_propagate_mark(solidB);
	csg_mark_unmarked(solidB, solidA);

	csg_select(solidA, solidB);
}


// Prints some statistics in verbose mode
static void PrintStats()
{
	std::cerr << "Nr. of solids   = " << GetNumberOfSolids() << std::endl;
	std::cerr << "Nr. of solids   = " << GetNumberOfSolids() << std::endl;
	std::cerr << "Nr. of shells   = " << GetNumberOfShells() << std::endl;
	std::cerr << "Nr. of faces    = " << GetNumberOfFaces() << std::endl;
	std::cerr << "Nr. of contours = " << GetNumberOfContours() << std::endl;
	std::cerr << "Nr. of edges    = " << GetNumberOfEdges() << std::endl;
	std::cerr << "Nr. of vertices = " << GetNumberOfVertices() << std::endl;
	std::cerr << "Processing time: " << ((float)(clock() - start_clock)/(float)CLOCKS_PER_SEC) << " seconds" << std::endl;;
}

// Prints program usage
static void PrintUsage(std::ostream& out)
{
	out << "Usage: csg [options] fileA.mgf fileB.mgf\n\n";
	out << "Options:\n\n";
	out << "\t-q n: nr. quart circle divisions (n = positive integer, default = 5)\n";
	out << "\t-t x: vertex tolerance (x = positive real number, default = 1e-6)\n";
	out << "\t-p x: min. coplanarity factor (x = positive real number, default = 500)\n";
	out << "\t-v: verbose mode\n";
	out << "\t-d: compute CSG difference\n";
	out << "\t-u: compute CSG union\n";
	out << "\t-i: compute CSG intersection\n";
	out << "\t-I: ignore normals in input file\n";
	out << "\t-W: write faces without making separate faces for outer contours contained in holes\n";
	out << "\t-M: don't write materials\n";
	out << "\t-h: print usage\n\n";
}

// Sets defaults for various options
static void Defaults()
{
  //MGFSetNrQuartCircDivs(5);
  SetMinimumVertexTolerance(1e-6);
  SetMinimumEdgeTolerance(1e-12);
  SetMinimumFaceTolerance(1e-12);
  SetMinCoplanarityFactor(500);
  verbose = false;
  csg_select = select_union;
  ignore_normals_on_input = false;
  //MGFRespectVertexOrder(true);
  //write_as_is = false;
  ignore_materials = false;
}

// Parse command line options
static int GetOptions(int argc, char** argv)
{
	int i, j, n;
	float x;

#define next_arg_parm {i++; j=1;}
#define next_arg_no_parm {if (argv[i][j+1]) {j++;} else {i++; j=1;}}

  for (i=1, j=1; i<argc && argv[i][0] == '-';) {
    switch (argv[i][j]) {
    case 'q':
      if ((sscanf(&argv[i][j+1], "%d", &n) != 1 &&
	  (++i >= argc || sscanf(argv[i], "%d", &n) != 1)) ||
	  n <= 0)
	Fatal(1, NULL, "option -q requires a positive integer argument");
      else
	//MGFSetNrQuartCircDivs(n);
      next_arg_parm;
      break;
    case 't':
      if ((sscanf(&argv[i][j+1], "%f", &x) != 1 &&
	  (++i >= argc || sscanf(argv[i], "%f", &x) != 1)) ||
	  (x <= 0))
	Fatal(1, NULL, "option -t requires a positive real argument");
      else
	SetMinimumVertexTolerance((double)x);
      next_arg_parm;
      break;
    case 'p':
      if ((sscanf(&argv[i][j+1], "%f", &x) != 1 &&
	  (++i >= argc || sscanf(argv[i], "%f", &x) != 1)) ||
	  (x <= 0))
	Fatal(1, NULL, "option -p requires a positive real argument");
      else
	SetMinCoplanarityFactor((double)x);
      next_arg_parm;
      break;
    case 'v':
      verbose = true;
      next_arg_no_parm;
      break;
    case 'd':
      csg_select = select_difference;
      next_arg_no_parm;
      break;
    case 'u':
      csg_select = select_union;
      next_arg_no_parm;
      break;
    case 'i':
      csg_select = select_intersection;
      next_arg_no_parm;
      break;
    case 'I':
      ignore_normals_on_input = true;
      // MGFRespectVertexOrder(false);
      next_arg_no_parm;
      break;
    case 'W':
      //write_as_is = true;
      next_arg_no_parm;
      break;
    case 'M':
      ignore_materials = true;
      next_arg_no_parm;
      break;
    case 'h':
    case '?':
      PrintUsage(std::cerr);
      exit(0);
    default:
	char text[200];
	sprintf(text, "Invalid option '%c'.\n", argv[i][j]);
	Fatal(1, NULL, text);
    }
  }

  return i;
}

int csg_main(int argc, char** argv)
{
	Defaults();

	int i = GetOptions(argc, argv);
	if(i >= argc-1)
		{
			std::cerr << "Usage: " << progname << " [options] file1.mgf file2.mgf" << std::endl;
			exit(1);
		}

	start_clock = clock();

	InitGeometry();

	// Read the objects in the MGF files
	SOLID* solidA;// = ReadMGF(argv[i]);
	if(!solidA)
		exit(1);

	solidA->DataCloseCallback();

	if(verbose)
		{
			std::cerr << "------------- after reading " << argv[i] << " -------------- " << std::endl;
			PrintStats();
		}

	SOLID* solidB;// = ReadMGF(argv[i+1]);
	if(!solidB)
		exit(1);

	solidB->DataCloseCallback();

	if(verbose)
		{
			std::cerr << "------------- after reading " << argv[i+1] << " -------------- " << std::endl;
			PrintStats();
		}

		do_csg(solidA, solidB);

	if(verbose)
		{
			std::cerr << "------------- after doing CSG operation ------------------ " << std::endl;
			PrintStats();
		}

	std::cout << "# solidA: " << std::endl;
	//WriteMGF(solidA, stdout);

	std::cout << "# solidB: " << std::endl;
	//WriteMGF(solidB, stdout);

	return 0;
}

// Main program
/*
int main(int argc, char** argv)
{
	if((progname = strrchr(argv[0], '/')) != NULL)
		progname = progname+1;
	else
		progname = strdup(argv[0]);

	return csg_main(argc, argv);
}
*/







