//
//  The BREP Library.
//  Copyright (C) 1996 Philippe Bekaert
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// C++ version: Romain Behar (romainbehar@yahoo.com), Feb 2002

#ifndef _GEOMETRY_H_
#define _GEOMETRY_H_

#include <iostream>
#include <math.h>

#include "brep.h"




// In main.cpp
extern int ignore_normals_on_input;	// Ignore normals in input file
extern int ignore_materials;		// Ignore materials when outputting MGF

// In geometry.cpp
// Minimum (relative) vertex, edge and face tolerance
extern double vertex_tolerance,
              edge_tolerance,
              face_tolerance;




extern void Error(char *routine, char *text);
extern void Warning(char *routine, char *text);
extern void Fatal(int errcode, char *routine, char *text);



// Some math
const double EPSILON = 1.0e-6;

// Tests whether two floating point numbers are equal within the given tolerance
inline bool FloatEqual(double a, double b, double tolerance)
{
	return fabs(a - b) < tolerance;
}



// A vector in 3D space
class VECTOR
{
public:
	VECTOR(const double X = 0, const double Y = 0, const double Z = 0) :
		x(X),
		y(Y),
		z(Z)
	{}

	~VECTOR() {}

	VECTOR& operator = (const VECTOR& v);

	double Norm2()
	{
		return x*x+y*y+z*z;
	}

	double Norm()
	{
		return sqrt(Norm2());
	}

	void InverseScale(const double s)
	{
		if(fabs(s) < EPSILON)
			return;

		double is = 1/s;
		x *= is;
		y *= is;
		z *= is;
	}

	void Normalize()
	{
		double norm = Norm();
		InverseScale(norm);
	}

	double x;
	double y;
	double z;
};

const unsigned int XNORMAL = 0;
const unsigned int YNORMAL = 1;
const unsigned int ZNORMAL = 2;

inline VECTOR operator - (const VECTOR& a)
{ return VECTOR(-a.x, -a.y, -a.z); }

inline VECTOR& VECTOR::operator = (const VECTOR& v)
{ x = v.x; y = v.y; z = v.z; return *this; }

inline VECTOR operator + (const VECTOR& a, const VECTOR& b)
{ return VECTOR(a.x+b.x, a.y+b.y, a.z+b.z); }

inline VECTOR operator - (const VECTOR& a, const VECTOR& b)
{ return VECTOR(a.x-b.x, a.y-b.y, a.z-b.z); }

inline VECTOR operator * (const double d, const VECTOR& a)
{ return VECTOR(d*a.x, d*a.y, d*a.z); }

// Cross product
inline VECTOR operator ^ (const VECTOR& a, const VECTOR& b)
{ return VECTOR(a.y*b.z-a.z*b.y, a.z*b.x-a.x*b.z, a.x*b.y-a.y*b.x); }

// Dot product
inline double operator * (const VECTOR& a, const VECTOR& b)
{ return (a.x*b.x + a.y*b.y + a.z*b.z); }

// Scales a vector with the inverse of the real number s if not zero : (1/s)*v
inline VECTOR ScaleInverse(const double s, const VECTOR& v)
{
	double is = (s < -EPSILON || s > EPSILON) ? 1/s : 1;
	return is * v;
}

// Centre of two points
inline VECTOR MidPoint(const VECTOR& a, const VECTOR& b)
{ return 0.5*(a+b); }

// Distance between two points in 3D space
inline double VectorDist(const VECTOR& a, const VECTOR& b)
{
	return (b-a).Norm();
}

// Find the "dominant" part of the vector (eg patch-normal),
// this is used to turn the point-in-polygon test into a 2D problem
inline unsigned int VectorDominantCoord(const VECTOR& v)
{
	double ax = fabs(v.x);
	double ay = fabs(v.y);
	double az = fabs(v.z);

	if(ax >= ay && ax >= az)
		return XNORMAL;

	if(ay > ax && ay >= az)
		return YNORMAL;

	return ZNORMAL;
}

// Given a vector p in 3D space and an index i, which is
// XNORMAL, YNORMAL or ZNORMAL, projects the vector on
// the YZ, XZ or XY plane respectively

// 2d vector
class VEC2D
{
public:
	VEC2D(const double U = 0, const double V = 0) :
		u(U),
		v(V)
	{}

	double u;
	double v;
};

inline VEC2D VectorProject(const VECTOR& p, unsigned int index)
{
	if(index == XNORMAL)
		return VEC2D(p.y, p.z);

	if(index == YNORMAL)
		return VEC2D(p.x, p.z);

	// ZNORMAL:
	return VEC2D(p.x, p.y);
};

// Two vectors are equal if their components are equal within the given tolerance
inline bool VectorFloatEqual(const VECTOR& v, const VECTOR& w, double eps)
{
	return FloatEqual(v.x, w.x, eps)
		&& FloatEqual(v.y, w.y, eps)
		&& FloatEqual(v.z, w.z, eps);
}

// Tolerance value for e.g. a vertex position
inline const double VectorTolerance(const VECTOR& v)
{ return EPSILON * (fabs(v.x) + fabs(v.y) + fabs(v.z)); }


typedef struct RAY {
	VECTOR	pos, dir;
} RAY;




class BOUNDINGBOX
{
public:
	BOUNDINGBOX()
	{
		MIN_X = MIN_Y = MIN_Z = HUGE;
		MAX_X = MAX_Y = MAX_Z = -HUGE;
	}

	~BOUNDINGBOX()
	{
	}

	void Enlarge(const BOUNDINGBOX& extra)
	{
		MIN_X = MIN_X <= extra.MIN_X ? MIN_X : extra.MIN_X;
		MIN_Y = MIN_Y <= extra.MIN_Y ? MIN_Y : extra.MIN_Y;
		MIN_Z = MIN_Z <= extra.MIN_Z ? MIN_Z : extra.MIN_Z;

		MAX_X = MAX_X >= extra.MAX_X ? MAX_X : extra.MAX_X;
		MAX_Y = MAX_Y >= extra.MAX_Y ? MAX_Y : extra.MAX_Y;
		MAX_Z = MAX_Z >= extra.MAX_Z ? MAX_Z : extra.MAX_Z;
	}

	void Enlarge(const VECTOR& point)
	{
		MIN_X = MIN_X <= point.x ? MIN_X : point.x;
		MIN_Y = MIN_Y <= point.y ? MIN_Y : point.y;
		MIN_Z = MIN_Z <= point.z ? MIN_Z : point.z;

		MAX_X = MAX_X >= point.x ? MAX_X : point.x;
		MAX_Y = MAX_Y >= point.y ? MAX_Y : point.y;
		MAX_Z = MAX_Z >= point.z ? MAX_Z : point.z;
	}

	bool OutOfBounds(const VECTOR& p)
	{
		return
			p.x < MIN_X || p.x > MAX_X ||
			p.y < MIN_Y || p.y > MAX_Y ||
			p.z < MIN_Z || p.z > MAX_Z;
	}

	int BoundsIntersect(RAY* ray, double mindist, double *maxdist);

//#ifdef DEBUG
	friend std::ostream& operator << (std::ostream& out, const BOUNDINGBOX& bb)
	{
		out << "\tX: " << bb.MIN_X << " to " << bb.MAX_X << std::endl;
		out << "\tY: " << bb.MIN_Y << " to " << bb.MAX_Y << std::endl;
		out << "\tZ: " << bb.MIN_Z << " to " << bb.MAX_Z << std::endl;

		return out;
	}
//#endif

	// Bounding-box limits
	double MIN_X;
	double MIN_Y;
	double MIN_Z;
	double MAX_X;
	double MAX_Y;
	double MAX_Z;
};


// Test whether two boundingboxes are disjunct
inline bool DisjunctBounds(const BOUNDINGBOX& b1, const BOUNDINGBOX& b2)
{
	return (b1.MIN_X >= b2.MAX_X) || (b2.MIN_X >= b1.MAX_X)
		|| (b1.MIN_Y >= b2.MAX_Y) || (b2.MIN_Y >= b1.MAX_Y)
		|| (b1.MIN_Z >= b2.MAX_Z) || (b2.MIN_Z >= b1.MAX_Z);
}



// See Mark Segal, "Using Tolerances to Guarantee Valid Polyhedral Modeling
// Results", SIGGRAPH '90 p 105 for everything about tolerances, extents and
// near values and what they are used for.

// Struct with common first field of VERTEX, EDGE, CONTOUR, ...

// All entities have an unsigned integer used for identification and
// to hold flags for various other purposes, such as to indicate whether
// a contour lies on the inside or outside of another object after
// partitioning for e.g. CSG operations (see "various other masks" below)

// The IDMASK below is used to separate the type bit and other flags from
// the id-number of an entity. In order to provide as much as possible
// space for id-numbers of entities, entities for which most other flags
// are needed (contours) have the highest type bit. Doing so, about 16 Mo
// (2^24) topological entities of the same type are possible
const unsigned int IDMASK = 0xff000000;

// The most significant set bit indicates the type of an entity.
// The type bits are defined below
const unsigned int CONTOURMASK = 0x80000000;
const unsigned int VERTEXMASK = 0x40000000;
const unsigned int EDGEMASK = 0x20000000;
const unsigned int FACEMASK = 0x10000000;
const unsigned int SHELLMASK = 0x08000000;
const unsigned int SOLIDMASK = 0x04000000;

// Various masks for other purposes than identification of an entity: These masks
// set only bits that are less significant than the type bit of an entity.

// Mask indicating that an entity is hidden
const unsigned int HIDDEN = 0x02000000;

// Mask indicating that an entity has been processed
const unsigned int PROCESSED = 0x01000000;

// Masks to indicate whether a contour lies inside or outside another object after
// partitioning with PartitionFaces() in isect.c + for indicating whether the vertex
// data for a vertex have been written out already before in writemgf.c
const unsigned int OUTMASK = 0x04000000;
const unsigned int INMASK = 0x08000000;

// Masks indicating whether or not a contour needs
// to be written with the routines in writemgf.c
const unsigned int DONTWRITEMASK = 0x10000000;
const unsigned int WRITEMASK = 0x20000000;
const unsigned int WRITEREVERSEMASK = 0x40000000;


class ENTITY
{
public:
	// The id is always an unsigned int used for
	// identification and other purposes.
	// An unsigned int should contain at least
	// 4 bytes (32 bits)
	unsigned int id;

	// Returns the id number
	unsigned int EntityId()
	{
		return (id & ~IDMASK);
	}

	// Determines the entity type of an entity.
	// The type mask is returned
	unsigned int EntityType()
	{
		unsigned int typemask = 0x80000000;
		for(; typemask&IDMASK; typemask >>= 1)
			if(id & typemask)
				return typemask;

		return 0;
	}

	// Returns the entity type name of an entity
	char* EntityTypeName()
	{
		switch(EntityType())
			{
				case VERTEXMASK:	return "vertex";
				case EDGEMASK:		return "edge";
				case CONTOURMASK:	return "contour";
				case FACEMASK:		return "face";
				case SHELLMASK:		return "shell";
				case SOLIDMASK:		return "solid";
			}

		return "unknown entity type";
	}

	bool IS_HIDDEN()
	{
		return (id & HIDDEN) != 0;
	}

	bool IS_PROCESSED()
	{
		return (id & PROCESSED) != 0;
	}

	// Prints all geometric and type
	// information about the entity
	friend std::ostream& operator << (std::ostream& out, const ENTITY& entity);
};


// Geometric data for a vertex
class VERTEX :
	public BREP_VERTEX, public ENTITY
{
public:
	VERTEX();
	~VERTEX();
	void DataCloseCallback(const bool done = false);

	friend std::ostream& operator << (std::ostream& out, const VERTEX& vertex);

	void ComputeNormal();
	void ComputeTolerance();

	void EnlargeTolerance(double tol);

	// Coordinates in 3D space
	VECTOR point;

	// Normal vector to the surface at the vertex
	VECTOR normal;

	// Coordinate tolerance. Two vertices nearer than
	// this should be considered one vertex
	double tolerance;

	// Name for the vertex
	char* name;
};


// Geometric data for an edge
class EDGE :
	public BREP_EDGE, public ENTITY
{
public:
	EDGE(BREP_VERTEX* vertex1, BREP_VERTEX* vertex2);
	~EDGE();
	void DataCloseCallback(const bool done = false);

	friend std::ostream& operator << (std::ostream& out, const EDGE& edge);


	double tolerance;
	double extent;
	double near;
};

void PrintWing(std::ostream& out, BREP_WING* wing);
void PrintWingShort(std::ostream& out, BREP_WING* wing);


// Geometric data for a contour
class CONTOUR :
	public BREP_CONTOUR, public ENTITY
{
public:
	CONTOUR(BREP_FACE* face);
	~CONTOUR();
	void DataCloseCallback(const bool done = false);

	friend std::ostream& operator << (std::ostream& out, const CONTOUR& contour);

	void ComputeNormal();

	bool IsHole();


	VECTOR normal;
};


// Geometric data for a face
class FACE :
	public BREP_FACE, public ENTITY
{
public:
	FACE(BREP_SHELL* shell);
	~FACE();
	void DataCloseCallback(const bool done = false);

	friend std::ostream& operator << (std::ostream& out, const FACE& face);

	void ComputePlaneEquation();
	double SignedDistanceToPlane(VERTEX* v, int* side);
	bool ConvexCorner(VERTEX* v0, VERTEX* v1, VERTEX* v2);
	double ShortestEdgeLength();

	//
	double tolerance;
	double extent;
	double near;
	BOUNDINGBOX bounds;
	VECTOR normal;

	// Plane constant
	double d;
};


// Geometric data for a shell
class SHELL :
	public BREP_SHELL, public ENTITY
{
public:
	SHELL(BREP_SOLID* solid);
	~SHELL();
	void DataCloseCallback(const bool done = false);

	friend std::ostream& operator << (std::ostream& out, const SHELL& shell);


	char* name;
	char* matname;
	//C_MATERIAL material;
	//RGB_MATERIAL rgbmat;
	BOUNDINGBOX bounds;
};


// Geometric data for a solid
class SOLID :
	public BREP_SOLID, public ENTITY
{
public:
	SOLID();
	~SOLID();
	void DataCloseCallback(const bool done = false);

	friend std::ostream& operator << (std::ostream& out, const SOLID& solid);


	char* filename;
	BOUNDINGBOX bounds;
};


int GetNumberOfSolids();
int GetNumberOfShells();
int GetNumberOfContours();
int GetNumberOfFaces();
int GetNumberOfEdges();
int GetNumberOfVertices();



// ******************* various initialisations **********************

void InitGeometry();

void SetMinimumVertexTolerance(double tol);
void SetMinimumEdgeTolerance(double tol);
void SetMinimumFaceTolerance(double tol);

//extern SOLID* ReadMGF(char* filename);




// Simple application of raytracing: testing whether a vertex is inside
// the solid. This routine returns OUTMASK if the vertex is outside,
// INMASK if inside, and 0 if it can not be determined (serious error)
unsigned int ContourInSolid(BREP_CONTOUR* contour, SOLID* solod);

unsigned int VertexInSolid(VERTEX* vertex_data, SOLID* solid);
unsigned int VertexInsideContour(const VECTOR& point, BREP_CONTOUR* contour);



// faceA and faceB are to BREP_FACES with appropriate geometric data
// (see geometry.[ch]). This routine checks whether the faces are coplanar,
// intersect transversally or don't intersect at all. In the first case,
// COPLANAR_ISECT is returned. In the second case, TRANSVERSAL_ISECT.
// In the last case NO_ISECT. If the faces intersect
// transversally, the faces are partitionned along the intersection line.
// If a contour of a face is split in two, a flag OUTMASK or INMASK in the contours
// id is set indicating whether the contour is on the "outside" of the other face or
// on the "inside". The flags are defined in geometry.h. This flag can be
// used to classify objects w.r.t. each other for e.g. CSG operations
const int NO_ISECT = 0;
const int COPLANAR_ISECT = 1;
const int TRANSVERSAL_ISECT = 2;
int PartitionFaces(FACE* faceA, FACE* faceB);

// Sets the minimum allowed coplanarity factor of two faces (see Segal, SIGGRAPH '90
// p 111). The coplanarity factor indicates to what extent two faces are coplanar:
// truely coplanar faces have an infinitely large coplanarity factor. For
// perpendicular faces however, the coplanarity factor is sqrt(2). In
// order to decide whether two faces are to be considered coplanar, their
// coplanarity factor is compared with a minimum coplanarity factor, which
// should be a number between 100 and 1000. Smaller values force the algorithm
// to deem pairs of faces that intersect in a relatively large dehedral
// angle to be coplanar. 500 is a nice default value
extern void SetMinCoplanarityFactor(double factor);



#endif // _GEOMETRY_H_
