// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CMagnetiser K-3D object, which pulls geometry towards a point, animatedly
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "magnetiser.h"

#include <k3dchooser.h>
#include <k3deventconnector.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dieditable.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

/////////////////////////////////////////////////////////////////////////////
// CMagnetiser

CMagnetiser::CMagnetiser(sdpIUnknown* Initializer) :
	CMagnetiserBase(Initializer),
	m_MagnetFactor("Magnetfactor", Document(), Document()->StateRecorder(), 0),
	m_OffsetX("offsetx", Document(), Document()->StateRecorder(), 0),
	m_OffsetY("offsety", Document(), Document()->StateRecorder(), 0),
	m_OffsetZ("offsetz", Document(), Document()->StateRecorder(), 0),
	m_Radius("radius", Document(), Document()->StateRecorder(), 0),
	m_Profile("profile", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, "CMagnetiser", 0))
{
	AddVariable(&m_MagnetFactor);
	AddVariable(&m_OffsetX);
	AddVariable(&m_OffsetY);
	AddVariable(&m_OffsetZ);
	AddVariable(&m_Radius);
	AddVariable(&m_Profile);

	// Setup profile curve ...
	g_return_if_fail(m_Profile.Create("Magnet Profile", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, false));

	// Let the channel know that we're making changes ...
	k3d::Editing editing(m_Profile.Interface());

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(m_Profile.Interface());
	g_return_if_fail(channel);
	
	channel->SetSegmentCount(1);
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);
	g_assert( 4 == nodes.size() );
	
	k3dBezierChannelNodeArray::iterator node(nodes.begin());
	
	(*node)->SetPosition(sdpVector2(0.0, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.5, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.5, 0.0));
	node++;
	(*node)->SetPosition(sdpVector2(1.0, 0.0));
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CMagnetiser::EditObject()
{
	new CMagnetiserDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CMagnetiser::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	sdpString event(Event);
	if(event == k3dIChannel::Event_Changed)
		{
//			PopAnimate();
			k3dRedrawAll(Document(), false);
			return SDP_S_OK;
		}
	
	return CMagnetiserBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CMagnetiser::EditorPostRender(k3dIEditorEngine* const Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glColor3d(0.0, 1.0, 1.0);

	// Setup the tranformation matrix based on magnet position ...
	glMatrixMode(GL_MODELVIEW_MATRIX);
	glPushMatrix();
	glTranslated(m_OffsetX->Value1(m_CurrentTime, m_DeltaTime), m_OffsetY->Value1(m_CurrentTime, m_DeltaTime), m_OffsetZ->Value1(m_CurrentTime, m_DeltaTime));

	// Draw the magnet center point ...
	glPointSize(6.0f);
	glBegin(GL_POINTS);
		glVertex3d(0.0, 0.0, 0.0);
	glEnd();

	// Indicate the magnet's range ...
	static GLint slices = 32;
	const gdouble radius = m_Radius->Value1(m_CurrentTime, m_DeltaTime);
	
	GLUquadricObj* quadric = gluNewQuadric();
	g_return_val_if_fail(quadric, SDP_E_FAIL);
	gluQuadricDrawStyle(quadric, GLenum(GLU_SILHOUETTE));
	gluQuadricNormals(quadric, GLenum(GLU_SMOOTH));

	// X axis
	glPushMatrix();
	glRotated(90.0, 0.0, 1.0, 0.0);
	gluDisk(quadric, 0.0, radius, slices, 1);
	glPopMatrix();

	// Y axis
	glPushMatrix();
	glRotated(-90.0, 1.0, 0.0, 0.0);
	gluDisk(quadric, 0.0, radius, slices, 1);
	glPopMatrix();

	// Z axis
	gluDisk(quadric, 0.0, radius, slices, 1);
	gluDeleteQuadric(quadric);
	
	glPopMatrix();
	glPopAttrib();	// GL_ALL_ATTRIB_BITS

	return SDP_S_OK;
}

class CMagnetiserMagnetObject
{
protected:
	gdouble m_MagnetFactor;
	gdouble m_OffsetX;
	gdouble m_OffsetY;
	gdouble m_OffsetZ;
	gdouble m_Radius;
	k3dIChannel* m_Profile;
	
public:
	CMagnetiserMagnetObject( gdouble OffsetX, gdouble OffsetY, gdouble OffsetZ, gdouble MagnetFactor, gdouble Radius, k3dIChannel* Profile) :
		m_MagnetFactor(MagnetFactor),
		m_OffsetX(OffsetX),
		m_OffsetY(OffsetY),
		m_OffsetZ(OffsetZ),
		m_Radius(Radius),
		m_Profile(Profile)
	{
	}	
	
	bool HandleDescendantObject(k3dIObject* Object)
	{
		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		g_return_val_if_fail(locationserver, true);


		sdpVector3 coords;
		k3dILocation* location;		
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();

				sdpVector3 position( m_OffsetX, m_OffsetY, m_OffsetZ );
				sdpVector3 difference = position - coords;

				gdouble distance = difference.Length() / m_Radius;
				gdouble mix_factor;
				if( distance <= 1.0 )
					mix_factor = sdpMin( m_Profile->Value1(distance, 0.001) * m_MagnetFactor, 1.0 );
				else
					mix_factor = 0.0;
		
				coords += mix_factor * difference;	
			
				location->SetAnimXYZ(coords);
			}
		return true;
	}
};


///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CMagnetiser::PopAnimate()
{
g_assert_not_reached();
/*
	gdouble magnetfactor = m_MagnetFactor->Value1(m_CurrentTime, m_DeltaTime);
	gdouble offsetx = m_OffsetX->Value1(m_CurrentTime, m_DeltaTime);
	gdouble offsety = m_OffsetY->Value1(m_CurrentTime, m_DeltaTime);
	gdouble offsetz = m_OffsetZ->Value1(m_CurrentTime, m_DeltaTime);
	gdouble radius = m_Radius->Value1(m_CurrentTime, m_DeltaTime);
	
	// Apply a magnet functor to all our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CMagnetiserMagnetObject( offsetx, offsety, offsetz, magnetfactor, radius, m_Profile.Interface()));
*/
	
	return SDP_S_OK;
}

sdpResult CMagnetiser::InitComplete()
{
	if(0 == m_MagnetFactor.Object())
		g_return_val_if_fail(m_MagnetFactor.Create("Magnet Factor", 1.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_OffsetX.Object())
		g_return_val_if_fail(m_OffsetX.Create("Offset X", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_OffsetY.Object())
		g_return_val_if_fail(m_OffsetY.Create("Offset Y", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_OffsetZ.Object())
		g_return_val_if_fail(m_OffsetZ.Create("Offset Z", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Radius.Object())
		g_return_val_if_fail(m_Radius.Create("Radius", 4.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	
	return CMagnetiserBase::InitComplete();
}

//////////////////////////////////////////////////////////////////////////////////////
// CMagnetiserDialog

const sdpString CMagnetiserDialog::dialogname = "properties";
const sdpString CMagnetiserDialog::control_offsetx = "offsetx";
const sdpString CMagnetiserDialog::control_offsety = "offsety";
const sdpString CMagnetiserDialog::control_offsetz = "offsetz";
const sdpString CMagnetiserDialog::control_magnetfactor = "magnetfactor";
const sdpString CMagnetiserDialog::control_radius = "radius";
const sdpString CMagnetiserDialog::control_profile = "profile";

CMagnetiserDialog::CMagnetiserDialog(CMagnetiser* const Object) :
	CMagnetiserDialogBase(Object, true, false, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("experimentallib", "magnetiser.gtkml"));

	// Setup channel buttons ...
	Connectk3dChannelButton(control_magnetfactor, m_Object->m_MagnetFactor.Interface());
	Connectk3dChannelButton(control_offsetx, m_Object->m_OffsetX.Interface());
	Connectk3dChannelButton(control_offsety, m_Object->m_OffsetY.Interface());
	Connectk3dChannelButton(control_offsetz, m_Object->m_OffsetZ.Interface());
	Connectk3dChannelButton(control_radius, m_Object->m_Radius.Interface());

	// Setup profile channel ...
	Connectk3dChannelButton(control_profile, m_Object->m_Profile.Interface());

	RootWidget().Show();
}

