// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CGTSReader K-3D object, which reads GTS .gts files
		\author Tom Browder (tbrowder@home.com)
*/

#include "gtsreader.h"
#include <fstream>
#include <strstream>

#include <k3dcom.h>
#include <k3dmisc.h>

#include <k3diapplication.h>
#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dimouseeventobserver.h>
#include <k3diobjectcollection.h>
#include <k3dipoint.h>
#include <k3dipath.h>
#include <k3dipathpointcollection.h>

namespace
{

k3dIPoint* CreatePoint(k3dIMesh* const Mesh, const sdpVector3 Coords)
{
	// Sanity checks ...
	g_assert(Mesh);

	k3dIPoint* const point = Mesh->CreatePoint();
	g_return_val_if_fail(point, 0);

	k3d::Editing editing(point);
	point->Location()->SetLocalXYZ(Coords);

	return point;
}

k3dIPath* CreatePath(k3dIMesh* const Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	k3dIPath* const path = Mesh->CreatePath();
	g_return_val_if_fail(path, 0);

	return path;
}

typedef std::vector<char> LineBuffer;
bool dataline(std::ifstream& file, LineBuffer& linebuffer)
{
	// Strip comments (lines beginning with '#') and empty lines,
	// return true if buffer has been filled, return false on end of file
	while(!file.eof())
		{
			file.getline(&linebuffer[0], linebuffer.size());

			// Skip comments ...
			if(linebuffer[0] == '#')
				continue;

			// Skip blank lines
			sdpString s(&linebuffer[0]);
			s.trim();
			if(!s.size())
				continue;

			// Must qualify
			return true;
		}

	return false;
}

gulong k3d_index(const gulong& external_index)
{
	return external_index - 1;
}

} // namespace

///////////////////////////////////////////////////////////////////////////////
// CGTSReader

CGTSReader::CGTSReader(sdpIUnknown* Initializer) :
	CGTSReaderBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool CGTSReader::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	if(".gts" != filepath.right(4))
		return false;

	return true;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryReadFormat implementation

sdpResult CGTSReader::PreRead(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CGTSReader::ReadOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CGTSReader::ReadFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Try to open the input file ...
	std::ifstream file(FilePath);
	if(!file.good())
		{
			std::cerr << g_pretty_function() << ": error opening [" << FilePath << "]" << std::endl;
			g_return_val_if_fail(0, SDP_E_FAIL);
		}

	// Setup a line buffer ...
	LineBuffer linebuffer(1024);

	// Always use the default material
	// (later : use unique color for each mesh)
	k3dIMaterial* const defaultmaterial = k3dDefaultMaterial(Document());

	// Handle undo/redo state ...
	k3d::Editing editingdocobjects(Document()->Objects());

	// Return empty mesh if empty file
	if(!dataline(file, linebuffer))
		{
			// Create a new mesh object ...
			k3dIObject* const object = Document()->Objects()->CreateObject(k3d::classid::Mesh);
			g_return_val_if_fail(object, SDP_E_FAIL);

			k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(object);
			g_return_val_if_fail(mesh, SDP_E_FAIL);

			Document()->SetMouseFocus(dynamic_cast<k3dIMouseEventObserver*>(object));
			object->State()->SetState(k3dIState::SELECTED, true);

			object->SetName("Untitled empty GTS object");

			return SDP_S_OK;
		}

	// Extract point, edge and triangle numbers
	std::istrstream stream(&linebuffer[0]);
	guint np = 0;
	guint ne = 0;
	guint nt = 0;
	stream >> np >> ne >> nt;

	// The line just read in should be the first line of a component (surface)
	// of a properly formatted gts component file.
	//
	// File format is as follows : all the lines beginning with '#' are ignored.
	// The first data line of a surface contains three unsigned integers separated by
	// spaces : the number of vertices (nv or np), the number of edges (ne)
	//  and the number of triangles (nt), respectively.
	// Then follow :
	// * nv lines each containing the x, y and z coordinates of the vertices
	// * ne lines each containing two indices of the endpoints of an edge
	// * nt lines each containing the three ordered indices of the edges of a triangle
	// (coordinates and indices are space-separated)
	//
	// Notes:	1. All indices are numbered starting from 1
	//		2. The ordering of the edges defines the faces.
	//		   upon import.
	//		3. Importer handles multiple GTS objects in one file
	//		   as well as empty lines (those two features are not
	//		   specified in the GTS format)

	if (SDP_S_OK == k3dApplication()->Verbose())
		std::cerr << "DEBUG: np = " << np << "; ne = " << ne << "; nt = " << nt << std::endl;

	// Give objects unique names as we build them ...
	guint objectnumber = 0;

	// Travel through the stream ...
	while(np && ne && nt)
		{
			// Create a new mesh object ...
			k3dIObject* const object = Document()->Objects()->CreateObject(k3d::classid::Mesh);
			g_return_val_if_fail(object, SDP_E_FAIL);

			k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(object);
			g_return_val_if_fail(mesh, SDP_E_FAIL);
			k3d::Editing editingmesh(mesh);

			Document()->SetMouseFocus(dynamic_cast<k3dIMouseEventObserver*>(object));
			object->State()->SetState(k3dIState::SELECTED, true);
			object->SetName("GTS object " + sdpToString(objectnumber++));

			// Setup storage for vertices and edges
			//    sdpVector3Array geometric_vertices;
			sdpVector2Array geometric_edges;

			// ??? I need a vector of points to avoid dup points on read in:
			// An array of points
			std::vector<k3dIPoint*> pointvec;

			// Read points (equivalent to OBJ vertices); indexed from 1
			guint n;
			for(n = 0; n < np; n++)
				{
					// Get next data line
					if(!dataline(file, linebuffer))
						{
							std::cerr << "ERROR: Unexpected end of input file (file \""
								<< FilePath << "\", point #" << n << "." << std::endl;
							return SDP_E_FAIL;
						}

					// Extract a record type ...
					std::istrstream stream(&linebuffer[0]);

					sdpVector3 coords;
					stream >> coords;

					// Transform from GTS to K-3D rigth-handed system
					gdouble x = coords[0];
					gdouble y = coords[1];
					gdouble z = coords[2];

					gdouble xp = -x;
					gdouble yp =  z;
					gdouble zp = -y;

					coords[0] = xp;
					coords[1] = yp;
					coords[2] = zp;

					if (SDP_S_OK == k3dApplication()->Verbose())
						{
							std::cerr << "DEBUG: coords  = " << coords << std::endl;
						}

					k3dIPoint* const point = CreatePoint(mesh, coords);
					g_return_val_if_fail(point, SDP_E_FAIL);

					pointvec.push_back(point);
				}

			// Edges (no OBJ equivalent--need for determining face vertex indices)
			for(n = 0; n < ne; n++)
				{
					// Get next data line
					if(!dataline(file, linebuffer))
						{
							std::cerr << "ERROR: Unexpected end of input file (file \""
								<< FilePath << "\", edge #" << n << "." << std::endl;
							return SDP_E_FAIL;
						}

					// Extract a record type ...
					std::istrstream stream(&linebuffer[0]);

					sdpVector2 edge;
					stream >> edge;

					// transform indices from 1-base to 0-base
					edge[0] = k3d_index(gulong(edge[0]));
					edge[1] = k3d_index(gulong(edge[1]));

					geometric_edges.push_back(edge);
				}

			// Faces (triangles)(equivalent to OBJ faces)
			for(n = 0; n < nt; n++)
				{
					// Get next data line
					if(!dataline(file, linebuffer))
						{
							std::cerr << "ERROR: Unexpected end of input file (file \""
								<< FilePath << "\", triangle #" << n << "." << std::endl;
							return SDP_E_FAIL;
						}

					// Create a path ...
					k3dIPath* const path = CreatePath(mesh);
					g_return_val_if_fail(path, SDP_E_FAIL);
					k3d::Editing editingpath(path);

					// color?
					//path->SetColor(0, 1, 0); // rgb green?

					// Extract a record type ...
					std::istrstream stream(&linebuffer[0]);

					// Get the 3 edges indices
					gulong e0idx, e1idx, e2idx;
					stream >> e0idx >> e1idx >> e2idx;

					// Transform indices from 1-base to 0-base
					e0idx = k3d_index(e0idx);
					e1idx = k3d_index(e1idx);
					e2idx = k3d_index(e2idx);

					sdpVector2 e0 = geometric_edges[e0idx];
					sdpVector2 e1 = geometric_edges[e1idx];
					sdpVector2 e2 = geometric_edges[e2idx];

					gulong e0_lo = gulong(e0[0]);
					gulong e0_hi = gulong(e0[1]);
					gulong e1_lo = gulong(e1[0]);
					gulong e1_hi = gulong(e1[1]);
					gulong e2_lo = gulong(e2[0]);
					gulong e2_hi = gulong(e2[1]);

					gulong p[3];
					// Determine order of vertices
					if(e0_lo == e1_lo)
						{
							p[0] = e0_hi;
							p[1] = e0_lo;
							p[2] = e1_hi;
						}
					else if(e0_lo == e1_hi)
						{
							p[0] = e0_hi;
							p[1] = e0_lo;
							p[2] = e1_lo;
						}
					else if(e0_hi == e1_lo)
						{
							p[0] = e0_lo;
							p[1] = e0_hi;
							p[2] = e1_hi;
						}
					else// if(e0_hi == e1_hi)
						{
							p[0] = e0_lo;
							p[1] = e0_hi;
							p[2] = e1_lo;
						}

					// Sanity checks
					if (SDP_S_OK == k3dApplication()->Verbose()) {
					gulong p3 = (p[2] == e2_lo ? e2_hi : e2_lo); // should be same as p[0]
					std::cerr << "DEBUG: e0idx = " << e0idx << "; e1idx = " << e1idx << "; e2idx = " << e2idx  << std::endl;
					std::cerr << "       e0_lo = " << e0_lo << "; e0_hi = " << e0_hi << std::endl;
					std::cerr << "       e1_lo = " << e1_lo << "; e1_hi = " << e1_hi << std::endl;
					std::cerr << "       e2_lo = " << e2_lo << "; e2_hi = " << e2_hi << std::endl;
					std::cerr << "       p[0]  = " << p[0]  << "; p[1]  = " << p[1]  << std::endl;
					std::cerr << "       p[2]  = " << p[2]  << "; p[3]  = " << p3    << std::endl;
					g_assert(p[0] == p3);
					}

					// Decode edge vertices into points from point map
					for(int i = 0; i < 3; ++i)
						{
							k3dIPathPoint* const pathpoint = path->Points()->AddPoint(pointvec[p[i]]);
							g_return_val_if_fail(pointvec[p[i]], SDP_E_FAIL);
							g_return_val_if_fail(pathpoint, SDP_E_FAIL);
						}

					// All paths should have a material
					path->Materials()->SetMaterial(defaultmaterial);
				}

			// Finished with this object
			if (SDP_S_OK == k3dApplication()->Verbose())
				{
					std::cerr << "DEBUG: np = " << np << "; ne = " << ne << "; nt = " << nt  << std::endl;
				}

			// Go on for another one ?
			np = ne = nt = 0;
			if(dataline(file, linebuffer))
				{
					std::istrstream stream(&linebuffer[0]);
					stream >> np >> ne >> nt;
				}
		}

	return SDP_S_OK;
}


