// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CGTSWriter K-3D object, which exports GTS *.gts files
		\author Tom Browder (tbrowder@home.com)
*/

#include "gtswriter.h"

#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dpathcollection.h>
#include <k3dpathpointcollection.h>
#include <k3dpointcollection.h>

#include <k3diapplication.h>
#include <k3dihierarchy.h>
#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dipoint.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>

#include <fstream>
#include <strstream>

namespace
{

/// Tests to see if a group of paths are all triangles
class TriangleTest :
	public k3dIPathVisitor
{
public:
	TriangleTest() :
		m_AllTriangles(true)
	{
	}

	const bool AllTriangles() const
	{
		return m_AllTriangles;
	}

	void VisitPath(k3dIPath* const Path)
	{
		if(!m_AllTriangles)
			return;
			
		if(Path->Points()->PointCount() != 3)
			m_AllTriangles = false;
	}
	
	void operator()(k3dIPathCollection* const PathCollection)
	{
		if(!m_AllTriangles)
			return;
			
		PathCollection->VisitPaths(this);
	}

private:
	bool m_AllTriangles;
};

/// std::pair equivalent that maintains the order of its members
template<typename T1, typename T2>
class ordered_pair;

template<typename T1, typename T2>
bool operator<(const ordered_pair<T1,T2>& lhs, const ordered_pair<T1,T2>& rhs);

template<typename T1, typename T2>
class ordered_pair :
	public std::pair<T1, T2>
{
public:
	typedef T1 first_type;
	typedef T2 second_type;
	
	T1 first;
	T2 second;

	explicit ordered_pair()
	{
	}

	explicit ordered_pair(const T1& First, const T2& Second) :
		first(First < Second ? First : Second),
		second(First < Second ? Second : First)
	{
	}
	
	friend bool operator< <>(const ordered_pair& lhs, const ordered_pair& rhs);
};

template<typename T1, typename T2>
bool operator<(const ordered_pair<T1,T2>& lhs, const ordered_pair<T1,T2>& rhs)
{
	if(lhs.first != rhs.first)
		return lhs.first < rhs.first;
		
	return lhs.second < rhs.second;
}

typedef std::vector<k3dIPoint*> PointCollection;
typedef std::map<k3dIPoint*, gulong> PointMap;
typedef ordered_pair<k3dIPoint*, k3dIPoint*> Edge;
typedef std::set<Edge> EdgeCollection;
typedef std::map<Edge, gulong> EdgeMap;
typedef std::vector<Edge> Triangle;
typedef std::vector<Triangle> TriangleCollection;

class BuildTree
{
public:
	BuildTree(EdgeCollection& Edges, TriangleCollection& Triangles) :
		m_Edges(Edges),
		m_Triangles(Triangles)
	{
	}
	
	void operator()(k3dIPath* const Path)
	{
		// Create a collection of k3dIPathPoint references for this path
		typedef std::vector<k3dIPathPoint*> PathPoints;
		PathPoints pathpoints;
		k3d::copy(Path->Points(), std::back_inserter(pathpoints));

		// Creating edges ...
		Triangle triangle;
		for(gulong i = 0; i < pathpoints.size(); i++)
			{
				Edge edge(pathpoints[(i+0)%pathpoints.size()]->Point(), pathpoints[(i+1)%pathpoints.size()]->Point());
				m_Edges.insert(edge);
				triangle.push_back(edge);
			}
		m_Triangles.push_back(triangle);
	}
	
private:
	EdgeCollection& m_Edges;
	TriangleCollection& m_Triangles;
};

class WriteMesh
{
public:
	WriteMesh(std::ostream& Stream) :
		m_Stream(Stream)
	{
	}
	
	void operator()(k3dIMesh* const Mesh)
	{
		// Get the collection of points ...
		PointCollection points;
		k3d::copy(Mesh->Points(), std::back_inserter(points));

		// Create a mapping of points-to-one-based-indices ...
		PointMap pointmap;
		for(PointCollection::iterator point = points.begin(); point != points.end(); ++point)
			pointmap[*point] = pointmap.size();

		// Build the collection of edges & triangles ...
		EdgeCollection edges;
		TriangleCollection triangles;
		k3d::for_each(Mesh->Paths(), BuildTree(edges, triangles));
		
		// Create a mapping of edges-to-one-based-indices ...
		EdgeMap edgemap;
		for(EdgeCollection::iterator edge = edges.begin(); edge != edges.end(); ++edge)
			edgemap[*edge] = edgemap.size();

		// Write some comments
		k3dIObject* const object = dynamic_cast<k3dIObject*>(Mesh);
		m_Stream << "# object: " << object->Name() << std::endl;
		m_Stream << "# num points [" << points.size() << "] num edges [" << edges.size() << "] num triangles [" << triangles.size() << "]" << std::endl;

		// Data header
		m_Stream << points.size() << " " << edges.size() << " " << triangles.size() << std::endl;

		// Write points
		m_Stream << "# points" << std::endl;
		for(PointCollection::iterator point = points.begin(); point != points.end(); ++point)
			{
				// Transform coords back to GTS convention
				const sdpVector3 coords = (*point)->Location()->WorldXYZ();
				m_Stream << -coords[0] << " " << -coords[2] << " " << coords[1] << std::endl;
			}
			
		// Write edges
		m_Stream << "# edges" << std::endl;
		for(EdgeCollection::iterator edge = edges.begin(); edge != edges.end(); ++edge)
			m_Stream << pointmap[edge->first] << " " << pointmap[edge->second] << std::endl;
			
		// Write paths
		m_Stream << "# triangles" << std::endl;
		for(TriangleCollection::iterator triangle = triangles.begin(); triangle != triangles.end(); ++triangle)
			{
				for(Triangle::iterator edge = triangle->begin(); edge != triangle->end(); ++edge)
					m_Stream << edgemap[*edge] << " ";
				m_Stream << std::endl;
			}
	}
	
private:
	std::ostream& m_Stream;
};

} // namespace

///////////////////////////////////////////////////////////////
// CGTSWriter

CGTSWriter::CGTSWriter(sdpIUnknown* Initializer) :
	CGTSWriterBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool CGTSWriter::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	if(".*" == filepath.right(4))
		return true;

	return false;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryWriteFormat implementation

sdpResult CGTSWriter::PreWrite(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	TriangleTest test = k3d::for_each<k3dIPathCollection>(Document()->Objects(), TriangleTest());
	if(!test.AllTriangles())
		{
			k3dApplication()->Error("Document contains non-triangular paths.  Triangulate all paths before using GTS export.", "Export GTS File:");
			return SDP_S_FALSE;
		}

	return SDP_S_OK;
}

sdpResult CGTSWriter::WriteOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CGTSWriter::WriteFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// (GTS format is not a real one : it has been designed to store surfaces
	// as sets of points, edges and triangles, only;
	// each surface item can have optional arbitrary values, but no other
	// scene information can be found in a native GTS file)

	// Try to open the output file ...
	std::ofstream file(FilePath);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);
	file << "# Written by K-3D" << std::endl;

	// Get the set of available meshes ...
	typedef std::vector<k3dIMesh*> Meshes;
	Meshes meshes;
	k3d::copy<k3dIMesh>(Document()->Objects(), std::back_inserter(meshes));

	// Compute world coordinates
	Document()->Hierarchy()->SnapshotWorldCoordinates();

	// Write meshes to a file ...
	std::for_each(meshes.begin(), meshes.end(), WriteMesh(file));		

	return SDP_S_OK;
}


