// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CMaxReader K-3D object, which reads .3ds files from 3D Studio Max
		\author Ben Campbell (bcampbell@ntlworld.com)
*/

#include "maxreader.h"
#include "maxids.h"

#include <k3dcom.h>
#include <k3dmisc.h>

#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dimouseeventobserver.h>
#include <k3diobjectcollection.h>
#include <k3dipoint.h>
#include <k3dipath.h>
#include <k3dipathpointcollection.h>

#include <iostream>
#include <fstream>
#include <strstream>
#include <cstdio>
#include <sys/stat.h>

CMaxReader::CMaxReader(sdpIUnknown* Initializer) :
	CMaxReaderBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool CMaxReader::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	if(".3ds" == filepath.right(4))
		return true;

	return false;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryReadFormat implementation

sdpResult CMaxReader::PreRead(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaxReader::ReadOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}




sdpResult CMaxReader::ReadFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Keep track of editing state for undo/redo purposes ...
	k3d::Editing editing(Document()->Objects());

	// Allocate a buffer and load file
	std::ifstream file(FilePath, std::ios::in | std::ios::binary);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);
	file.seekg(0, std::ios::end);
	m_Bufsize = file.tellg();
	file.seekg(0);
	m_Buf = new guchar[m_Bufsize];
	file.read(reinterpret_cast<gchar*>(m_Buf), m_Bufsize);
	file.close();
	m_Ptr = m_Buf;

	sdpResult ret = LoadMesh();

	// Free buffer
	delete [] m_Buf;

	return ret;
}


sdpResult CMaxReader::GetChunkInfo()
{
	// Get chunk ID and chunk length
	if(m_Buf+m_Bufsize-m_Ptr < 6)
		return SDP_E_FAIL;

	m_ChunkId = GetShort();
	m_ChunkLength = GetLong() - 6;

	return SDP_S_OK;
}

void CMaxReader::SkipChunk()
{
	if(m_Ptr + m_ChunkLength <= m_Buf + m_Bufsize)
		m_Ptr += m_ChunkLength;
}

sdpResult CMaxReader::LoadMesh()
{
	// Check 3DS ID
	g_return_val_if_fail(GetChunkInfo() == SDP_S_OK, SDP_E_FAIL);
	g_return_val_if_fail(m_ChunkId == MAXID_MAIN, SDP_E_FAIL);

	// Get version and loop through contained objects
	gulong chunkbytes = m_ChunkLength;
	std::cerr << "main len: " << (chunkbytes+6) << std::endl;
	while(chunkbytes > 6)
		{
			g_return_val_if_fail(GetChunkInfo() == SDP_S_OK, SDP_E_FAIL);
			std::cerr << "Chunk 0x" << std::hex << m_ChunkId << " (" << std::dec << (m_ChunkLength+6) << " bytes)" << std::endl;

			chunkbytes -= m_ChunkLength + 6;

			switch(m_ChunkId)
				{
					case MAXID_VERSION:
						std::cerr << "  3DS Version: " << GetLong() << std::endl;
						break;
					case MAXID_OBJMESH:
						if(Handle_OBJMESH(m_ChunkLength) != SDP_S_OK)
							return SDP_E_FAIL;
						break;
					default:
						SkipChunk();
				}
		}

	return SDP_S_OK;
}


sdpString CMaxReader::GetString()
{
	sdpString s;
	char c;
	while((c = (char)(*m_Ptr++)))
	{
		s += c;
	}

	return s;
}


gushort CMaxReader::GetShort()
{
	g_assert(sdp::LittleEndian());
	if(m_Ptr+2 <= m_Buf+m_Bufsize)
	{
		gushort ret = m_Ptr[0] + (m_Ptr[1]*0x100);
		m_Ptr += 2;
		return ret;
	}
	else
		return 0;
}

gulong CMaxReader::GetLong()
{
	g_assert(sdp::LittleEndian());
	if(m_Ptr+4 <= m_Buf+m_Bufsize)
	{
		gulong ret = m_Ptr[0] +
			(m_Ptr[1]*0x100) +
			(m_Ptr[2]*0x10000) +
			(m_Ptr[3]*0x1000000);
		m_Ptr += 4;
		return ret;
	}
	else
		return 0;
}


float CMaxReader::GetFloat()
{
	gulong raw = GetLong();
	return *((float*)(&raw));
}



sdpResult CMaxReader::Handle_OBJMESH(int chunkbytes)
{
	std::cerr << "Handle_OBJMESH()" << std::endl;
	while(chunkbytes>=6)
		{
			g_return_val_if_fail(GetChunkInfo() == SDP_S_OK, SDP_E_FAIL);
			chunkbytes -= m_ChunkLength + 6;

			switch(m_ChunkId)
				{
					case MAXID_MATERIAL:
						std::cerr << " MAXID_MATERIAL (Skipping)" << std::endl;
						SkipChunk();
	/*
						sdpString materialname;
						while(1)
							{
								char c = stream.get();

								if(stream.eof())
									break;

								materialname += c;
							}

						if(materials.find(materialname) == materials.end())
							{
								k3dIObject* materialobject = Document()->CreateObject(sdpMaterial);
								g_return_val_if_fail(materialobject, SDP_E_FAIL);
								k3dIMaterial* material = dynamic_cast<k3dIMaterial*>(materialobject);
								g_return_val_if_fail(material, SDP_E_FAIL);

								materialobject->SetName(materialname);
								materials[materialname] = material;
							}

						currentmaterial = materials[materialname];

	*/
						break;
					case MAXID_ONEUNIT:
						std::cerr << " MAXID_ONEUNIT:  (" << GetFloat() << ")" << std::endl;
						break;
					case MAXID_OBJBLOCK:
						if(!Handle_OBJBLOCK(m_ChunkLength))
							return SDP_E_FAIL;
						break;
					default:
						std::cerr << " skipping unknown Chunk 0x" << std::hex << m_ChunkId << "(" << std::dec << (m_ChunkLength+6) << " bytes)" << std::endl;
						SkipChunk();
				}
		}

	return SDP_S_OK;
}



sdpResult CMaxReader::Handle_OBJBLOCK(int chunkbytes)
{
	sdpString s = GetString();
	chunkbytes -= (s.size()+1);

	k3dIObject* const object = Document()->Objects()->CreateObject(k3d::classid::Mesh);
	g_return_val_if_fail(object, SDP_E_FAIL);

	k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(object);
	g_return_val_if_fail(mesh, SDP_E_FAIL);
	Document()->SetMouseFocus(dynamic_cast<k3dIMouseEventObserver*>(object));
	object->State()->SetState(k3dIState::SELECTED, true);
	object->SetName(s.c_str());

	std::cerr << "  Object '" << s.c_str() << "'" << std::endl;

	k3d::Editing editing(mesh);

	while(chunkbytes >= 6)
		{
			g_return_val_if_fail(GetChunkInfo() == SDP_S_OK, SDP_E_FAIL);
			chunkbytes -= m_ChunkLength + 6;

			switch(m_ChunkId)
				{
					case MAXID_TRIMESH:
						std::cerr << " MAXID_TRIMESH" << std::endl;
						if(!Handle_TRIMESH(m_ChunkLength, mesh))
							return SDP_E_FAIL;

						mesh->CalculateNormals();

						break;
					case MAXID_CAMERA:
						std::cerr << " MAXID_CAMERA (skipping)" << std::endl;
						SkipChunk();
						break;
					default:
						std::cerr << " skipping unhandled Chunk 0x" << std::hex << m_ChunkId << "(" << std::dec << (m_ChunkLength+6) << " bytes)" << std::endl;
						SkipChunk();
				}
		}

	return SDP_S_OK;
}



sdpResult CMaxReader::Handle_TRIMESH(int chunkbytes, k3dIMesh* mesh)
{
	sdpVector3Array	verts;



	while(chunkbytes >= 6)
		{
			g_return_val_if_fail(GetChunkInfo() == SDP_S_OK, SDP_E_FAIL);
			chunkbytes -= m_ChunkLength + 6;

			switch(m_ChunkId)
				{
					case MAXID_VERTLIST:
					{
						std::cerr << "   MAXID_VERTLIST" << std::endl;
						verts.clear();

						gulong numverts = (gulong)GetShort();
						std::cerr << "    [" << numverts << " verts]" << std::endl;
						for(gulong i = 0; i < numverts; i++)
							{
								float x = GetFloat();
								float y = GetFloat();
								float z = GetFloat();

								verts.push_back(sdpVector3(x,y,z));
							}
						break;
					}
					case MAXID_FACELIST:
					{
						std::cerr << "   MAXID_FACELIST";
						int bytesleft = m_ChunkLength;

						gulong numfaces = (gulong)GetShort();
						bytesleft -= 2;
						std::cerr << "    [" << numfaces << " faces]" << std::endl;
						for(gulong facenum = 0; facenum < numfaces; facenum++)
							{
								k3dIPath* const path = mesh->CreatePath();
								g_return_val_if_fail(path, SDP_E_FAIL);
								k3d::Editing editingpath(path);

								int indices[3];
								indices[0] = (int)GetShort();
								indices[1] = (int)GetShort();
								indices[2] = (int)GetShort();
								GetShort();	// skip flags for now

								int q;
								for(q=0; q<3; ++q)
								{
									if(indices[q] >= (int)verts.size())
									{
										std::cerr << "vertex out-of-range" << std::endl;
										return SDP_E_FAIL;
									}

									k3dIPoint* const point = mesh->CreatePoint();
									g_return_val_if_fail(point, SDP_E_FAIL);
									k3d::Editing editingpoint(point);

									sdpVector3 pos = verts[ indices[q] ];
									
									point->Location()->SetLocalXYZ(pos);

		//							k3dIPathPoint* pathpoint =
										path->Points()->AddPoint(point);

								}

								// construct a face normal vector

		//						sdpVector3 a = verts[indices[1]] - verts[ indices[0] ];
		//						sdpVector3 b = verts[indices[1]] - verts[ indices[0] ];
		//						sdpVector3 n = a ^ b;	// a cross b
		//						n.Normalize();
		//						path->SetPathNormal(n);

								bytesleft -= 8;
							}

						while(bytesleft >= 6)
							{
								g_return_val_if_fail(GetChunkInfo() == SDP_S_OK, SDP_E_FAIL);
								bytesleft -= m_ChunkLength + 6;

								switch(m_ChunkId)
									{
										case MAXID_FACEMAT:
											{
												std::cerr << "    MAXID_FACEMAT" << std::endl;
												sdpString matname = GetString();
												std::cerr << "     name: " << matname.c_str() << std::endl;

												int numfaces = (int)GetShort();
												std::cerr << "     " << numfaces << " faces" << std::endl;
												int i;
												for(i = 0; i < numfaces; i++)
													GetShort();	// facenum
											}
											break;
										case MAXID_SMOOLIST:
											{
												// not sure if this is correct place...
												std::cerr << "    MAXID_SMOOLIST, child of FACELIST (skipping)" << std::endl;
												SkipChunk();
											}
											break;
										default:
											std::cerr << " skipping unhandled Chunk 0x" << std::hex << m_ChunkId << "(" << std::dec << (m_ChunkLength+6) << " bytes)" << std::endl;
											SkipChunk();
									}
							}
					}
						break;
					case MAXID_TRMATRIX:
						std::cerr << "   MAXID_TRMATRIX" << std::endl;

						float x,y,z;

						x=GetFloat();
						y=GetFloat();
						z=GetFloat();
						printf("    X axis: %f %f %f\n",x,y,z);
						x=GetFloat();
						y=GetFloat();
						z=GetFloat();
						printf("    Y axis: %f %f %f\n",x,y,z);
						x=GetFloat();
						y=GetFloat();
						z=GetFloat();
						printf("    Z axis: %f %f %f\n",x,y,z);
						x=GetFloat();
						y=GetFloat();
						z=GetFloat();
						printf("    Origin: %f %f %f\n",x,y,z);
						break;
					case MAXID_MAPLIST:
					{
						std::cerr << "   MAXID_MAPLIST" << std::endl;

						int numuvs = (int)GetShort();
						int i;

						std::cerr << "    [" << numuvs << " UV coords]" << std::endl;
						for(i=0; i<numuvs; ++i)
						{
							GetFloat();	// u
							GetFloat();	// v
						// TODO: add to mesh data
	/*
							pathpoint->Location()->SetTextureXYZ(Vertex(texture_vertices, texture_index));
							pathpoint->Location()->SetImplicitTextureXYZ(Vertex(texture_vertices, texture_index));
	*/
						}
						break;
					}
					case MAXID_SMOOLIST:
						// not sure if this is correct place...
						std::cerr << "   MAXID_SMOOLIST, child of TRIMESH (skipping)" << std::endl;
						SkipChunk();
						break;
					default:
						std::cerr << " skipping unhandled Chunk 0x" << std::hex << m_ChunkId << "(" << std::dec << (m_ChunkLength+6) << " bytes)" << std::endl;
						SkipChunk();
						break;
				}
		}

	return SDP_S_OK;
}



