// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the COBJReader K-3D object, which reads Wavefront .obj files
		\author Tim Shead (tshead@k-3d.com)
*/

#include "objreader.h"

#include <k3dcom.h>
#include <k3dmisc.h>

#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dimouseeventobserver.h>
#include <k3diobjectcollection.h>
#include <k3dipoint.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>

#include <fstream>
#include <strstream>

namespace
{

k3dIPoint* CreatePoint(k3dIMesh* const Mesh, const sdpVector3 Coords)
{
	// Sanity checks ...
	g_assert(Mesh);

	k3dIPoint* const point = Mesh->CreatePoint();
	g_return_val_if_fail(point, 0);

	k3d::Editing editing(point);
	point->Location()->SetLocalXYZ(Coords);

	return point;
}

k3dIPath* CreatePath(k3dIMesh* const Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	k3dIPath* const path = Mesh->CreatePath();
	g_return_val_if_fail(path, 0);

	return path;
}

sdpVector3 TextureVertex(const sdpVector3Array& Vertices, const glong Index)
{
	static sdpVector3 error;
	g_return_val_if_fail(Index, error);

	return (Index > 0) ? Vertices[Index-1] : Vertices[Vertices.size() + Index];
}

k3dIPoint* Vertex(const std::vector<k3dIPoint*>& Vertices, const glong Index)
{
	g_return_val_if_fail(Index, 0);

	return (Index > 0) ? Vertices[Index-1] : Vertices[Vertices.size() + Index];
}

} // namespace

///////////////////////////////////////////////////////////////////////////
// COBJReader

COBJReader::COBJReader(sdpIUnknown* Initializer) :
	COBJReaderBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool COBJReader::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	if(".obj" == filepath.right(4))
		return true;

	return false;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryReadFormat implementation

sdpResult COBJReader::PreRead(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult COBJReader::ReadOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult COBJReader::ReadFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Try to open the input file ...
	std::ifstream file(FilePath);
	if(!file.good())
		{
			std::cerr << g_pretty_function() << ": error opening [" << FilePath << "]" << std::endl;
			g_return_val_if_fail(0, SDP_E_FAIL);
		}

	// Create a mesh object ...
	k3d::Editing editingobjects(Document()->Objects());
	k3dIObject* const object = Document()->Objects()->CreateObject(k3d::classid::Mesh);
	g_return_val_if_fail(object, SDP_E_FAIL);

	k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(object);
	g_return_val_if_fail(mesh, SDP_E_FAIL);
	k3d::Editing editingmesh(mesh);

	Document()->SetMouseFocus(dynamic_cast<k3dIMouseEventObserver*>(object));
	object->State()->SetState(k3dIState::SELECTED, true);
	object->SetName("Untitled Wavefront .obj Object");

	// Setup storage for vertices ...
	std::vector<k3dIPoint*> geometric_points;
	sdpVector3Array texture_vertices;
	sdpVector3Array normal_vertices;
	sdpVector3Array parameter_vertices;
	std::map<sdpString, k3dIMaterial*> materials;

	k3dIMaterial* currentmaterial = k3dDefaultMaterial(Document());

	// Travel through the stream ...
	while(!file.eof())
		{
			// Grab one line at a time ...
			sdpString linebuffer;
			sdp::GetLine(file, linebuffer);

			// Skip empty lines ...
			if(0 == linebuffer.size())
				continue;

			// Skip comments ...
			if('#' == *linebuffer.begin())
				continue;

			// Extract a record type ...
			std::istrstream stream(linebuffer);

			std::string recordtype;
			stream >> recordtype;

			// Skip empty lines ...
			if(0 == recordtype.size())
				continue;

			// Geometric vertices ...
			if(recordtype == "v")
				{
					// Handle coordinates ...
					sdpVector3 coords;
					stream >> coords;
					k3dIPoint* const point = CreatePoint(mesh, coords);
					g_return_val_if_fail(point, SDP_E_FAIL);
					geometric_points.push_back(point);
				}
			// Texture vertices ...
			else if(recordtype == "vt")
				{
					sdpVector3 coords;
					stream >> coords;
					texture_vertices.push_back(coords);
				}
			// Vertex normals ...
			else if(recordtype == "vn")
				{
					sdpVector3 coords;
					stream >> coords;
					normal_vertices.push_back(coords);
				}
			// Parameter space vertices ...
			else if(recordtype == "vp")
				{
					sdpVector3 coords;
					stream >> coords;
					parameter_vertices.push_back(coords);
				}
			// Points ...
			else if(recordtype == "p")
				{
					gulong geometric_index = 0;
					stream >> geometric_index;
					g_return_val_if_fail(geometric_index, SDP_E_FAIL);
				}
			// Lines ...
			else if(recordtype == "l")
				{
					// Create a path ...
					k3dIPath* const path = CreatePath(mesh);
					g_return_val_if_fail(path, SDP_E_FAIL);
					k3d::Editing editingpath(path);

					while(!stream.eof())
						{
							gulong geometric_index = 0;
							stream >> geometric_index;
							g_return_val_if_fail(geometric_index, SDP_E_FAIL);

							k3dIPoint* const point = Vertex(geometric_points, geometric_index);
							g_return_val_if_fail(point, SDP_E_FAIL);

							k3dIPathPoint* const pathpoint = path->Points()->AddPoint(point);
							while(stream.peek() == '/')
								{
									stream.get();

									gulong texture_index = 0;
									stream >> texture_index;
									g_return_val_if_fail(texture_index, SDP_E_FAIL);

									pathpoint->Location()->SetTextureXYZ(TextureVertex(texture_vertices, texture_index));
									pathpoint->Location()->SetImplicitTextureXYZ(TextureVertex(texture_vertices, texture_index));

									break;
								}
						}
				}
			// Faces ...
			else if(recordtype == "f")
				{
					// Create a path ...
					k3dIPath* const path = CreatePath(mesh);
					g_return_val_if_fail(path, SDP_E_FAIL);
					k3d::Editing editingpath(path);

					while(!stream.eof())
						{
							gulong geometric_index = 0;
							stream >> geometric_index;
							g_return_val_if_fail(geometric_index, SDP_E_FAIL);

							k3dIPoint* const point = Vertex(geometric_points, geometric_index);
							g_return_val_if_fail(point, SDP_E_FAIL);

							k3dIPathPoint* const pathpoint = path->Points()->AddPoint(point);
							k3d::Editing editingpathpoint(pathpoint);
							if(stream.peek() == '/')
								{
									stream.get();

									if(stream.peek() != '/')
									{
									gulong texture_index = 0;
									stream >> texture_index;
									g_return_val_if_fail(texture_index, SDP_E_FAIL);

									pathpoint->Location()->SetTextureXYZ(TextureVertex(texture_vertices, texture_index));
									pathpoint->Location()->SetImplicitTextureXYZ(TextureVertex(texture_vertices, texture_index));
									}
								}
							if(stream.peek() == '/')
								{
									stream.get();

									gulong normal_index = 0;
									stream >> normal_index;
									g_return_val_if_fail(normal_index, SDP_E_FAIL);

									// We ignore normal vertices

									// to get coorect value of stream.eof() at endl
									stream.peek();
								}
						}

					path->Materials()->SetMaterial(currentmaterial);
					//path->FlipOrientation(); Probably not necessary
				}
			else if(recordtype == "o")
				{
					sdpString name;
					while(1)
						{
							char c = stream.get();

							if(stream.eof())
								break;

							name += c;
						}

					object->SetName(name.c_str());
				}
			else if(recordtype == "usemtl")
				{
					sdpString materialname;
					while(1)
						{
							char c = stream.get();

							if(stream.eof())
								break;

							materialname += c;
						}

					if(materials.find(materialname) == materials.end())
						{
							k3dIObject* const materialobject = Document()->Objects()->CreateObject(k3d::classid::Material);
							g_return_val_if_fail(materialobject, SDP_E_FAIL);
							k3dIMaterial* material = dynamic_cast<k3dIMaterial*>(materialobject);
							g_return_val_if_fail(material, SDP_E_FAIL);

							materialobject->SetName(materialname);
							materials[materialname] = material;
						}

					currentmaterial = materials[materialname];
				}
			else
				{
					std::cerr << "Unknown .OBJ record: " << recordtype << " will be ignored" << std::endl;
				}
		}

	return SDP_S_OK;
}

