// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the COBJWriter K-3D object, which exports Wavefront .obj files
		\author Tim Shead (tshead@k-3d.com)
*/

#include "objwriter.h"
#include <fstream>
#include <strstream>

#include <k3dmisc.h>
#include <k3dobjectcollection.h>

#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dipoint.h>
#include <k3dipointcollection.h>
#include <k3dipath.h>
#include <k3dipathcollection.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>


COBJWriter::COBJWriter(sdpIUnknown* Initializer) :
	COBJWriterBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// sdpIUnknown implementation

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool COBJWriter::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	if(".obj" == filepath.right(4))
		return true;

	return false;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryReadFormat implementation

sdpResult COBJWriter::PreWrite(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult COBJWriter::WriteOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult COBJWriter::WriteFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Try to open the output file ...
	std::ofstream file(FilePath);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);

	file << "# Written by K-3D" << std::endl;

	// Keep track of 1-based point indices as we write them ...
	typedef std::map<k3dIPoint*, gulong> PointMap;
	PointMap pointmap;
	gulong pointindex = 1;

	// Keep track of the material as we go ...
	k3dIMaterial* currentmaterial = 0;

	// Get the set of available meshes ...
	typedef std::vector<k3dIObject*> Meshes;
	Meshes meshes;
	k3d::copy_if(Document()->Objects(), std::back_inserter(meshes), k3d::Implements(sdp::TypeInfo<k3dIMesh>()));
		
	// For each mesh ...	
	for(Meshes::iterator meshobject = meshes.begin(); meshobject != meshes.end(); meshobject++)
		{
			k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(*meshobject);
			g_return_val_if_fail(mesh, SDP_E_FAIL);
			
			// Write points ...
			k3dIPointCollection* const pointserver = mesh->Points();
			for(K3DPOINTHANDLE pointhandle = pointserver->FirstPoint(); pointhandle; pointhandle = pointserver->NextPoint(pointhandle))
				{
					k3dIPoint* point = pointserver->Point(pointhandle);
					pointmap[point] = pointindex++;
					
					sdpVector3 coords = point->Location()->LocalXYZ();
					
					file << "v " << coords << std::endl;
				}
				
			// Write paths ...
			k3dIPathCollection* const pathserver = mesh->Paths();
			for(K3DPATHHANDLE pathhandle = pathserver->FirstPath(); pathhandle; pathhandle = pathserver->NextPath(pathhandle))
				{
					k3dIPath* const path = pathserver->Path(pathhandle);
					k3dIPathPointCollection* const pathpointserver = path->Points();
					if(0 == pathpointserver->PointCount())
						continue;

					// Handle materials as we go ...
					if(currentmaterial != path->Materials()->Material())
						{
							currentmaterial = path->Materials()->Material();
							if(currentmaterial)
								{
									k3dIObject* const object = dynamic_cast<k3dIObject*>(currentmaterial);
									g_assert(object);

									file << "usemtl " << object->Name() << std::endl;
								}
						}
															
					// Write the path ...
					file << "f";
					
					for(K3DPATHPOINTHANDLE pathpointhandle = pathpointserver->FirstPathPoint(); pathpointhandle; pathpointhandle = pathpointserver->NextPathPoint(pathpointhandle))
						file << " " << pointmap[pathpointserver->PathPoint(pathpointhandle)->Point()];
						
					file << std::endl;
				}
		}
	
	return SDP_S_OK;
}


