// K-3D
//
// .x (DirectX) file writer for K3D
//
// Adapted from GTS writer code by Adam Sakareassen
// Contact amsaka@ugrad.unimelb.edu.au
//
// Copyright  1995-2000, Timothy M. Shead
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CXWriter K-3D object, which exports .x files
		\author Adam Sakareassen (amsaka@ugrad.unimelb.edu.au)
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "xwriter.h"

#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dpathcollection.h>

#include <k3diapplication.h>
#include <k3dihierarchy.h>
#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3diobject.h>
#include <k3dipath.h>
#include <k3dipathcollection.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>
#include <k3dipoint.h>
#include <k3dipointcollection.h>

#include <fstream>
#include <strstream>

namespace
{

class TriangleTest :
	public k3dIPathVisitor
{
public:
	TriangleTest() :
		m_AllTriangles(true)
	{
	}

	const bool AllTriangles() const
	{
		return m_AllTriangles;
	}

	void VisitPath(k3dIPath* const Path)
	{
		if(!m_AllTriangles)
			return;
			
		if(Path->Points()->PointCount() != 3)
			m_AllTriangles = false;
	}
	
	void operator()(k3dIPathCollection* const PathCollection)
	{
		if(!m_AllTriangles)
			return;
			
		PathCollection->VisitPaths(this);
	}

private:
	bool m_AllTriangles;
};

} // namespace

///////////////////////////////////////////////////////////////
// CXWriter

CXWriter::CXWriter(sdpIUnknown* Initializer) :
	CXWriterBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool CXWriter::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	if(".x" == filepath.right(2))
		return true;

	return false;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryWriteFormat implementation

sdpResult CXWriter::PreWrite(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	TriangleTest test = k3d::for_each<k3dIPathCollection>(Document()->Objects(), TriangleTest());
	if(!test.AllTriangles())
		{
			k3dApplication()->Error("Document contains non-triangular paths.  Triangulate all paths before using .x export.", "Export .x File:");
			return SDP_S_FALSE;
		}

	return SDP_S_OK;
}

sdpResult CXWriter::WriteOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CXWriter::WriteFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Try to open the output file ...
	std::ofstream file(FilePath);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);


        //Write Header information

        file << "xof 0302txt 0032"  << std::endl;
        file << "Header {" << std::endl;
        file << " 1;" << std::endl;
        file << " 0;" << std::endl;
        file << " 1;" << std::endl;
        file << "}" << std::endl;

        file << "//Created using K-3D " << std::endl;
        file << "//Exporter plugin written by Adam Sakareassen"<< std::endl;
        file << "//Note: This file may have been created with unix style line endings"<< std::endl;
        file << "//      in this case you could use the utility unix2dos to convert to windows style"<< std::endl;


        //write template information
        file << ""<< std::endl;
        file << "template Header {"<< std::endl;
        file << " <3D82AB43-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " WORD major;"<< std::endl;
        file << " WORD minor;"<< std::endl;
        file << " DWORD flags;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Vector {"<< std::endl;
        file << " <3D82AB5E-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " FLOAT x;"<< std::endl;
        file << " FLOAT y;"<< std::endl;
        file << " FLOAT z;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Coords2d {"<< std::endl;
        file << "<F6F23F44-7686-11cf-8F52-0040333594A3>"<< std::endl;
        file << "FLOAT u;"<< std::endl;
        file << "FLOAT v;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Matrix4x4 {"<< std::endl;
        file << " <F6F23F45-7686-11cf-8F52-0040333594A3>"<< std::endl;
        file << "array FLOAT matrix[16];"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template ColorRGBA {"<< std::endl;
        file << " <35FF44E0-6C7C-11cf-8F52-0040333594A3>"<< std::endl;
        file << " FLOAT red;"<< std::endl;
        file << " FLOAT green;"<< std::endl;
        file << " FLOAT blue;"<< std::endl;
        file << " FLOAT alpha;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template ColorRGB {"<< std::endl;
        file << " <D3E16E81-7835-11cf-8F52-0040333594A3>"<< std::endl;
        file << " FLOAT red;"<< std::endl;
        file << " FLOAT green;"<< std::endl;
        file << " FLOAT blue;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template TextureFilename {"<< std::endl;
        file << " <A42790E1-7810-11cf-8F52-0040333594A3>"<< std::endl;
        file << " STRING filename;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Material {"<< std::endl;
        file << " <3D82AB4D-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " ColorRGBA faceColor;"<< std::endl;
        file << " FLOAT power;"<< std::endl;
        file << " ColorRGB specularColor;"<< std::endl;
        file << " ColorRGB emissiveColor;"<< std::endl;
        file << " [...]"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template MeshFace {"<< std::endl;
        file << " <3D82AB5F-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " DWORD nFaceVertexIndices;"<< std::endl;
        file << " array DWORD faceVertexIndices[nFaceVertexIndices];"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template MeshTextureCoords {"<< std::endl;
        file << " <F6F23F40-7686-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD nTextureCoords;"<< std::endl;
        file << " array Coords2d textureCoords[nTextureCoords];"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template MeshMaterialList {"<< std::endl;
        file << " <F6F23F42-7686-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD nMaterials;"<< std::endl;
        file << " DWORD nFaceIndexes;"<< std::endl;
        file << " array DWORD faceIndexes[nFaceIndexes];"<< std::endl;
        file << " [Material]"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template MeshNormals {"<< std::endl;
        file << " <F6F23F43-7686-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD nNormals;"<< std::endl;
        file << " array Vector normals[nNormals];"<< std::endl;
        file << " DWORD nFaceNormals;"<< std::endl;
        file << " array MeshFace faceNormals[nFaceNormals];"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Mesh {"<< std::endl;
        file << "<3D82AB44-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " DWORD nVertices;"<< std::endl;
        file << " array Vector vertices[nVertices];"<< std::endl;
        file << " DWORD nFaces;"<< std::endl;
        file << " array MeshFace faces[nFaces];"<< std::endl;
        file << " [...]"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template FrameTransformMatrix {"<< std::endl;
        file << " <F6F23F41-7686-11cf-8F52-0040333594A3>"<< std::endl;
        file << " Matrix4x4 frameMatrix;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Frame {"<< std::endl;
        file << "<3D82AB46-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " [...]"<< std::endl;
        file << "}"<< std::endl;
        file << "template FloatKeys {"<< std::endl;
        file << " <10DD46A9-775B-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD nValues;"<< std::endl;
        file << " array FLOAT values[nValues];"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template TimedFloatKeys {"<< std::endl;
        file << " <F406B180-7B3B-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD time;"<< std::endl;
        file << " FloatKeys tfkeys;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template AnimationKey {"<< std::endl;
        file << " <10DD46A8-775B-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD keyType;"<< std::endl;
        file << " DWORD nKeys;"<< std::endl;
        file << "array TimedFloatKeys keys[nKeys];"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template AnimationOptions {"<< std::endl;
        file << " <E2BF56C0-840F-11cf-8F52-0040333594A3>"<< std::endl;
        file << " DWORD openclosed;"<< std::endl;
        file << " DWORD positionquality;"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template Animation {"<< std::endl;
        file << " <3D82AB4F-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " [...]"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        file << "template AnimationSet {"<< std::endl;
        file << " <3D82AB50-62DA-11cf-AB39-0020AF71E433>"<< std::endl;
        file << " [Animation]"<< std::endl;
        file << "}"<< std::endl;
        file << ""<< std::endl;
        //template information complete


        // Get the set of available meshes ...
	typedef std::vector<k3dIMesh*> Meshes;
	Meshes meshes;
	k3d::copy<k3dIMesh>(Document()->Objects(), std::back_inserter(meshes));


	// Compute world coordinates
        Document()->Hierarchy()->SnapshotWorldCoordinates();

	// For each mesh ...
	for(Meshes::iterator meshiterator = meshes.begin(); meshiterator != meshes.end(); ++meshiterator)
		{
			k3dIMesh* const mesh = *meshiterator;
			k3dIObject* const object = dynamic_cast<k3dIObject*>(*meshiterator);

			// Write Mesh...
			sdpString oname = object->Name();
                        //All spaces are removed from the mesh name.
                        oname.replaceall(" ","_");

                        //file << "//Object: " << oname << std::endl;
			file << "Mesh "<< oname << "{" << std::endl;

	                // Keep track of 1-based point indices as we write them ...
	                typedef std::map<k3dIPoint*, gulong> PointMap;
	                PointMap pointmap;
	                gulong pointindex = 1;


                        // Write points ...
                        int firstpoint = 1;
			k3dIPointCollection* const pointcollection = mesh->Points();

                        gulong np = pointcollection->PointCount();    //get number of points
                        file << np << ";" <<std::endl;  //write the number of points

                        for(K3DPOINTHANDLE pointhandle = pointcollection->FirstPoint(); pointhandle; pointhandle = pointcollection->NextPoint(pointhandle))
			{
                                if (firstpoint == 0) {file << "," <<std::endl;}
                                firstpoint = 0;

                                k3dIPoint* point = pointcollection->Point(pointhandle);
			        pointmap[point] = pointindex++;

			        const sdpVector3 coords = point->Location()->WorldXYZ();

			        file << "" << coords[0]<<";" << coords[1] << ";"<<coords[2]  << ";";
			}
                        file << ";" << std::endl;

                        //Write Faces (Triangles written with references to above points)


                        // Write paths ...
			k3dIPathCollection* pathcollection = mesh->Paths();
                        const gulong nt = pathcollection->PathCount();
                        file << nt << ";" <<std::endl;    //write the total number of faces.

                        int firstpath = 1;
                        for(K3DPATHHANDLE pathhandle = pathcollection->FirstPath(); pathhandle; pathhandle = pathcollection->NextPath(pathhandle))
				{
                                        if (firstpath==0) file <<";,"<< std::endl;
                                        firstpath=0;


                                        k3dIPath* const path = pathcollection->Path(pathhandle);
					k3dIPathPointCollection* const pathpointcollection = path->Points();
					if(0 == pathpointcollection->PointCount())
						continue;

					// Write the path ...
					file << "3;";
                                        firstpoint = 1;
					for(K3DPATHPOINTHANDLE pathpointhandle = pathpointcollection->FirstPathPoint(); pathpointhandle; pathpointhandle = pathpointcollection->NextPathPoint(pathpointhandle))
                                        {
                                                if (firstpoint == 0) file << ",";
                                                firstpoint=0;
                                                file << "" << pointmap[pathpointcollection->PathPoint(pathpointhandle)->Point()] ;
                                        }

				}
                        file << ";;" << std::endl;
                        file << "}" << std::endl;  //end of mesh write




		}

	return SDP_S_OK;
}

