// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CJavaScriptEngine, an implementation of k3dIScriptEngine that supports the JavaScript language
		\author Tim Shead (tshead@k-3d.com)
*/

#include "javascriptengine.h"
#include "javascriptgtkmlcontainer.h"
#include "javascriptobjects.h"

#include <k3dicommandnode.h>

///////////////////////////////////////////////////////////////
// CJavaScriptEngine

CJavaScriptEngine::CJavaScriptEngine(sdpIUnknown* Initializer)
{
	// Sanity checks ...
	g_assert(Initializer);
	
	// Fire up the interpreter ...
	m_Runtime = JS_NewRuntime(0x100000);
	g_assert(m_Runtime);
}

CJavaScriptEngine::~CJavaScriptEngine()
{
	// Shut down the interpreter
	if(m_Runtime)
		JS_Finish(m_Runtime);
}

SDPCSTRING CJavaScriptEngine::Language()
{
	return "JavaScript";
}

sdpResult CJavaScriptEngine::Execute(SDPCSTRING Script, SDPCSTRING ScriptName)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Script, SDP_E_FAIL);
	g_return_val_if_string_fail(ScriptName, SDP_E_FAIL);
	
	// Create an execution context ...
	JSContext* const context = JS_NewContext(m_Runtime, 0x1000); 
	g_return_val_if_fail(context, SDP_E_FAIL);

	// Store a reference to ourselves in the interpreter context ...
	JS_SetContextPrivate(context, this);

	// Set the branch callback, so we can cancel scripts if necessary ...
	JS_SetBranchCallback(context, RawBranchCallback);

	// Set the error reporter so we can handle script errors as needed ...
	JS_SetErrorReporter(context, RawErrorReporter);

	// Create the global object ...
	static JSClass global_class = { "global", 0, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub };
	JSObject* const globalobject = JS_NewObject(context, &global_class, NULL, NULL); 
	g_return_val_if_fail(globalobject, SDP_E_FAIL);
	
	// Initialize core JS objects ...
	JS_InitStandardClasses(context, globalobject);

	// Create the application object ...
	CJavaScriptApplication::Create(context, globalobject);
	
	// Setup the GTKMLContainer class ...
	CJavaScriptGTKMLContainer::InitializeClass(context, globalobject);
	
	// Evaluate the script ...
	jsval result;
	JSBool returncode = JS_EvaluateScript(context, globalobject, Script, strlen(Script), ScriptName, 0, &result);
	
	// Cleanup the execution context ...
	JS_DestroyContext(context);
	
	// Return the results ...
	return (returncode == JS_TRUE) ? SDP_S_OK : SDP_E_FAIL;
}

static sdpResult BufferCopy(SDPSTRING Buffer, gulong& BufferLength, sdpString& Source)
{
	// Make sure we have space in the caller-supplied buffer
	g_return_val_if_fail(BufferLength > Source.size(), SDP_E_FAIL);

	strcpy(Buffer, Source);
	BufferLength = Source.size();

	return SDP_S_OK;
}

sdpResult CJavaScriptEngine::ConvertCommand(k3dICommandNode* CommandNode, SDPCSTRING Name, SDPCSTRING Arguments, SDPSTRING Buffer, gulong& BufferLength)
{
	// Sanity checks ...
	g_return_val_if_fail(CommandNode, SDP_E_FAIL);
	g_return_val_if_fail(Buffer, SDP_E_FAIL);
	g_return_val_if_fail(BufferLength, SDP_E_FAIL);

	sdpString path;
	for(k3dICommandNode* node = CommandNode; node->CommandNodeParent(); node = node->CommandNodeParent())
		{
			if(path.size())
				path = "/" + path;
				
			path = node->CommandNodeName() + path;
		}

	if(path.size())
		path = "/" + path;
		
	path = "." + path;

	sdpString result = "Application.CommandNode(\"" + path + "\").Command(\"" + sdpString(Name) + "\", \"" + sdpString(Arguments) + "\");";
	
	return BufferCopy(Buffer, BufferLength, result);
}

JSBool CJavaScriptEngine::RawBranchCallback(JSContext* Context, JSScript* Script)
{
	return reinterpret_cast<CJavaScriptEngine*>(JS_GetContextPrivate(Context))->BranchCallback(Context, Script);
}

JSBool CJavaScriptEngine::BranchCallback(JSContext* Context, JSScript* Script)
{
	return JS_TRUE;
}

void CJavaScriptEngine::RawErrorReporter(JSContext* Context, const char *Message, JSErrorReport *Report)
{
	reinterpret_cast<CJavaScriptEngine*>(JS_GetContextPrivate(Context))->ErrorReporter(Message, Report);
}

void CJavaScriptEngine::ErrorReporter(const char* Message, JSErrorReport* Report)
{
	std::cerr << "JavaScript error in " << sdpString(Report->filename) << ": " << Report->lineno+1 << std::endl;
	std::cerr << sdpString(Report->tokenptr) << std::endl;
	std::cerr << sdpString(Message) << std::endl;
}


