// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CJavaScriptGTKMLContainer class, which can instantiate GTKML user interfaces from JavaScript
		\author Tim Shead (tshead@k-3d.com)
*/

#include <javascriptgtkmlcontainer.h>

#include <jscntxt.h>
#include <sdpgtkevents.h>

#include <strstream>

///////////////////////////////////////////////////////////////
// CJavaScriptGTKMLContainer

JSClass CJavaScriptGTKMLContainer::Class = 
{
	"GTKMLContainer", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, Destructor
};

JSPropertySpec CJavaScriptGTKMLContainer::Properties[] =
{
//	{ "Mesh", PROPERTY_MESH, JSPROP_READONLY | JSPROP_PERMANENT, GetMesh, 0 },
//	{ "Points", PROPERTY_POINTS, JSPROP_READONLY | JSPROP_PERMANENT, GetPoints, 0 },
	{0}
};

JSFunctionSpec CJavaScriptGTKMLContainer::Methods[] = 
{
	{ "DoModal", JSDoModal, 0, 0, 0 },
	{ "Field", JSField, 1, 0, 0 },
	{0}
};

void CJavaScriptGTKMLContainer::InitializeClass(JSContext* Context, JSObject* Object)
{
	JS_InitClass(Context, Object, 0, &Class, Constructor, 1, Properties, Methods, 0, 0);
}

CJavaScriptGTKMLContainer* CJavaScriptGTKMLContainer::Container(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<CJavaScriptGTKMLContainer*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptGTKMLContainer::Constructor(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
//std::cout << "Constructor" << std::endl;

	// If called without new, replace obj with a new GTKMLContainer object
	if(!Context->fp->constructing)
		{
			Object = JS_NewObject(Context, &Class, 0, 0);
			if(!Object)
				return JS_FALSE;
				
			*Result = OBJECT_TO_JSVAL(Object);
		}

	// Create the actual container instance and associate it with the JS object ...
	JS_SetPrivate(Context, Object, new CJavaScriptGTKMLContainer());
	
	// Get a reference to the container instance ...
	CJavaScriptGTKMLContainer* const container = Container(Context, Object);
	g_return_val_if_fail(container, JS_FALSE);
	
	// Load GTKML from the supplied text string ...
	std::istrstream buffer(JS_GetStringBytes(JS_ValueToString(Context, argv[0])));
	container->Load(buffer, "JavaScript Source");
	
	// Ensure that we're called when the UI goes away ...
	container->MapEvent("delete-event", "delete", false, container->RootWidget(), true);

	// We should remain hidden until the user calls for us ...
	container->RootWidget().Hide();
			
	return JS_TRUE;
}

void CJavaScriptGTKMLContainer::Destructor(JSContext* Context, JSObject* Object)
{
//std::cout << "Destructor" << std::endl;

	delete Container(Context, Object);
}

JSBool CJavaScriptGTKMLContainer::JSDoModal(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Container(Context, Object)->DoModal();
	*Result = INT_TO_JSVAL(Container(Context, Object)->m_ResultCode);
	
	return JS_TRUE;
}

JSBool CJavaScriptGTKMLContainer::JSField(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	CJavaScriptGTKMLContainer* const container = Container(Context, Object);
	g_return_val_if_fail(container, JS_FALSE);
	
	*Result = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, container->Editable(JS_GetStringBytes(JS_ValueToString(Context, argv[0]))).GetText()));
			
	return JS_TRUE;
}

void CJavaScriptGTKMLContainer::DoModal()
{
	m_ResultCode = 0;
	
	RootWidget().Show();
	CJavaScriptGTKMLContainerBase::DoModal();
	RootWidget().Hide();
}

void CJavaScriptGTKMLContainer::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "delete")
		OnDelete(Event);
	else if(Event->Name() == "ok")
		OnOK();
	else if(Event->Name() == "cancel")
		OnCancel();
	else if(Event->Name() == "destroy")
		OnDestroy();
	else
		CJavaScriptGTKMLContainerBase::OnEvent(Event);
}

void CJavaScriptGTKMLContainer::OnDelete(sdpGtkEvent* const Event)
{
	// Turn it into a "cancel", instead ...
	static_cast<sdpGtkEventWidgetDeleteEvent*>(Event)->SetResult(true);
	
	OnCancel();
}

void CJavaScriptGTKMLContainer::OnOK()
{
	m_ResultCode = 1;
	CancelModal();
}

void CJavaScriptGTKMLContainer::OnCancel()
{
	m_ResultCode = 0;
	CancelModal();
}

void CJavaScriptGTKMLContainer::OnDestroy()
{
}

