// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements a scripting object model in JavaScript
		\author Tim Shead (tshead@k-3d.com)
		\author Dave Crane (ulunum@bigfoot.com)
*/

#include "javascriptobjects.h"

#include <k3dfilefilters.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3dobjectinfocollection.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3dicommandnode.h>
#include <k3didocument.h>
#include <k3dieditable.h>
#include <k3digeometryreadformat.h>
#include <k3digeometrywriteformat.h>
#include <k3dilocation.h>
#include <k3dimesh.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>
#include <k3dipath.h>
#include <k3dipathcollection.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>
#include <k3dipoint.h>
#include <k3dipointcollection.h>
#include <k3direnderfarm.h>
#include <k3direnderjob.h>
#include <k3direnderframe.h>
#include <k3ditransform.h>

namespace
{

JSObject* Convert(JSContext* Context, const sdpVector3& Source)
{
	jsval values[3];
	values[0] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source[0]));
	values[1] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source[1]));
	values[2] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source[2]));

	return JS_NewArrayObject(Context, 3, values);
}

JSObject* Convert(JSContext* Context, const sdpAngleAxis& Source)
{
	jsval values[4];
	values[0] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, sdpDegrees(Source.angle)));
	values[1] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source.axis[0]));
	values[2] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source.axis[1]));
	values[3] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source.axis[2]));

	return JS_NewArrayObject(Context, 4, values);
}

sdpVector3 ConvertVector3(JSContext* Context, jsval Value)
{
	sdpVector3 result;
	
	if(JSVAL_IS_OBJECT(Value))
		{
			jsval element;
			
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 0, &element);	JS_ValueToNumber(Context, element, &result[0]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 1, &element);	JS_ValueToNumber(Context, element, &result[1]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 2, &element);	JS_ValueToNumber(Context, element, &result[2]);
		}
	
	return result;
}

sdpAngleAxis ConvertAngleAxis(JSContext* Context, jsval Value)
{
	sdpAngleAxis result;
	
	if(JSVAL_IS_OBJECT(Value))
		{
			jsval element;
			
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 0, &element);	JS_ValueToNumber(Context, element, &result.angle);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 1, &element);	JS_ValueToNumber(Context, element, &result.axis[0]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 2, &element);	JS_ValueToNumber(Context, element, &result.axis[1]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 3, &element);	JS_ValueToNumber(Context, element, &result.axis[2]);
			
			result.angle = sdpRadians(result.angle);
		}
	
	return result;
}

/// Specialization of sdpToString() to handle JavaScript strings
const sdpString sdpToString(JSContext* const Context, jsval Value)
{
	if(JSVAL_IS_NULL(Value) || JSVAL_IS_VOID(Value))
		return sdpString();
		
	return sdpString(JS_GetStringBytes(JS_ValueToString(Context, Value)));
}

} // namespace

///////////////////////////////////////////////////////////////
// CJavaScriptEditable

JSClass CJavaScriptEditable::Skeleton::Class = 
{
	"Editable", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptEditable::Skeleton::Properties[] =
{
	{0}
};

JSFunctionSpec CJavaScriptEditable::Skeleton::Methods[] = 
{
	{ "PushEditing", Owner::PushEditing, 0, 0, 0 },
	{ "PopEditing", Owner::PopEditing, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptEditable::PushEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Interface(Context, Object)->PushEditing();
	return JS_TRUE;
}

JSBool CJavaScriptEditable::PopEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Interface(Context, Object)->PopEditing();
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptLocation

JSClass CJavaScriptLocation::Skeleton::Class = 
{
	"Location", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptLocation::Skeleton::Properties[] =
{
	{ "Local", 0, JSPROP_PERMANENT, Owner::GetLocal, Owner::SetLocal },
	{ "Animated", 0, JSPROP_PERMANENT, Owner::GetAnimated, Owner::SetAnimated },
	{ "World", 0, JSPROP_PERMANENT, Owner::GetWorld, Owner::SetWorld },
	{ "Camera", 0, JSPROP_PERMANENT, Owner::GetCamera, Owner::SetCamera },
	{ "Screen", 0, JSPROP_PERMANENT, Owner::GetScreen, Owner::SetScreen },
	{ "Normal", 0, JSPROP_PERMANENT, Owner::GetNormal, Owner::SetNormal },
	{ "Texture", 0, JSPROP_PERMANENT, Owner::GetTexture, Owner::SetTexture },
	{ "ImplicitTexture", 0, JSPROP_PERMANENT, Owner::GetImplicitTexture, Owner::SetImplicitTexture },
	{0}
};

JSFunctionSpec CJavaScriptLocation::Skeleton::Methods[] = 
{
	{0}
};

JSBool CJavaScriptLocation::GetLocal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->LocalXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetLocal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetLocalXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetAnimated(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->AnimXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetAnimated(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetAnimXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetWorld(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->WorldXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetWorld(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetWorldXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetCamera(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->CameraXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetCamera(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetCameraXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetScreen(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->ScreenXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetScreen(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetScreenXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetNormal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->NormalXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetNormal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetNormalXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->TextureXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetTextureXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetImplicitTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result = Interface(Context, Object)->ImplicitTextureXYZ();
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetImplicitTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Interface(Context, Object)->SetImplicitTextureXYZ(value);
	
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptPoint

JSClass CJavaScriptPoint::Skeleton::Class = 
{
	"Point", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptPoint::Skeleton::Properties[] =
{
	{ "Mesh", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetMesh, 0 },
	{ "Location", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetLocation, 0 },
	{ "Editable", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetEditable, 0 },
	{0}
};

JSFunctionSpec CJavaScriptPoint::Skeleton::Methods[] = 
{
	{ "PushEditing", Owner::PushEditing, 0, 0, 0 },
	{ "PopEditing", Owner::PopEditing, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptPoint::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPoint::GetLocation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(CJavaScriptLocation::Create(Interface(Context, Object)->Location(), Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPoint::GetEditable(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	return DynamicCast<k3dIEditable, CJavaScriptEditable>(Context, Object, ID, Value);
}

JSBool CJavaScriptPoint::PushEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PushEditing();
	return JS_TRUE;
}

JSBool CJavaScriptPoint::PopEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PopEditing();
	return JS_TRUE;
}


///////////////////////////////////////////////////////////////
// CJavaScriptPathPointCollection

JSClass CJavaScriptPathPointCollection::Skeleton::Class = 
{
	"PathPointServer", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptPathPointCollection::Skeleton::Properties[] =
{
	{0}
};

JSFunctionSpec CJavaScriptPathPointCollection::Skeleton::Methods[] = 
{
	{ "AddPoint", Owner::AddPoint, 1, 0, 0 },
	{ "RemovePoint", Owner::RemovePoint, 1, 0, 0 },
	{0}
};

JSBool CJavaScriptPathPointCollection::AddPoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIPoint* const point = CJavaScriptPoint::Interface(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(point, JS_FALSE);
	
	g_return_val_if_fail(Interface(Context, Object)->AddPoint(point), JS_FALSE);	
	return JS_TRUE;
}

JSBool CJavaScriptPathPointCollection::RemovePoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIPoint* const point = CJavaScriptPoint::Interface(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(point, JS_FALSE);
	
	g_return_val_if_fail(sdpSucceeded(Interface(Context, Object)->RemovePoint(point)), JS_FALSE);	
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptPath

JSClass CJavaScriptPath::Skeleton::Class = 
{
	"Path", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptPath::Skeleton::Properties[] =
{
	{ "Mesh", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetMesh, 0 },
	{ "Points", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetPoints, 0 },
	{ "Editable", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetEditable, 0 },
	{0}
};

JSFunctionSpec CJavaScriptPath::Skeleton::Methods[] = 
{
	{ "PushEditing", Owner::PushEditing, 0, 0, 0 },
	{ "PopEditing", Owner::PopEditing, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptPath::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPath::GetPoints(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(CJavaScriptPathPointCollection::Create(Interface(Context, Object)->Points(), Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPath::GetEditable(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	return DynamicCast<k3dIEditable, CJavaScriptEditable>(Context, Object, ID, Value);
}

JSBool CJavaScriptPath::PushEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PushEditing();
	return JS_TRUE;
}

JSBool CJavaScriptPath::PopEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PopEditing();
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptMesh

JSClass CJavaScriptMesh::Skeleton::Class = 
{
	"Mesh", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptMesh::Skeleton::Properties[] =
{
	{ "Points", 0, JSPROP_PERMANENT | JSPROP_READONLY, Owner::GetPoints, 0 },
	{ "Paths", 0, JSPROP_PERMANENT | JSPROP_READONLY, Owner::GetPoints, 0 },
	{0}
};

JSFunctionSpec CJavaScriptMesh::Skeleton::Methods[] = 
{
	{ "CreatePoint", Owner::CreatePoint, 0, 0, 0 },
	{ "CreatePath", Owner::CreatePath, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptMesh::GetPoints(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	// Convert the collection of points to a Javascript array ...
	std::vector<jsval> values;
	
	k3dIPointCollection* const pointcollection = Interface(Context, Object)->Points();
	for(K3DPOINTHANDLE pointhandle = pointcollection->FirstPoint(); pointhandle; pointhandle = pointcollection->NextPoint(pointhandle))
		{
			k3dIPoint* const point = pointcollection->Point(pointhandle);
			values.push_back(OBJECT_TO_JSVAL(CJavaScriptPoint::Create(point, Context, Object)));
		}
	
	if(values.size())
		*Value = OBJECT_TO_JSVAL(JS_NewArrayObject(Context, values.size(), &values.front()));
	else
		*Value = OBJECT_TO_JSVAL(JS_NewArrayObject(Context, 0, 0));

	return JS_TRUE;
}

JSBool CJavaScriptMesh::GetPaths(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	// Convert the collection of paths to a Javascript array ...
	std::vector<jsval> values;
	
	k3dIPathCollection* const pathcollection = Interface(Context,Object)->Paths();
	for(K3DPATHHANDLE pathhandle = pathcollection->FirstPath(); pathhandle; pathhandle = pathcollection->NextPath(pathhandle))
		{
			k3dIPath* const path = pathcollection->Path(pathhandle);
			values.push_back(OBJECT_TO_JSVAL(CJavaScriptPath::Create(path, Context, Object)));
		}		
	
	if(values.size())
		*Value = OBJECT_TO_JSVAL(JS_NewArrayObject(Context, values.size(), &values.front()));
	else
		*Value = OBJECT_TO_JSVAL(JS_NewArrayObject(Context, 0, 0));

	return JS_TRUE;
}

JSBool CJavaScriptMesh::CreatePoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIPoint* const point = Interface(Context, Object)->CreatePoint();
	g_return_val_if_fail(point, JS_FALSE);

	*Result = OBJECT_TO_JSVAL(CJavaScriptPoint::Create(point, Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptMesh::CreatePath(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIPath* const path = Interface(Context, Object)->CreatePath();
	g_return_val_if_fail(path, JS_FALSE);

	*Result = OBJECT_TO_JSVAL(CJavaScriptPath::Create(path, Context, Object));
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptObject

JSClass CJavaScriptObject::Skeleton::Class = 
{
	"Object", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptObject::Skeleton::Properties[] =
{
	{ "Document", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetDocument, 0 },
	{ "Name", 0, JSPROP_PERMANENT, Owner::GetName, Owner::SetName },
	{ "Position", 0, JSPROP_PERMANENT, Owner::GetPosition, Owner::SetPosition },
	{ "Orientation", 0, JSPROP_PERMANENT, Owner::GetOrientation, Owner::SetOrientation },
	{ "Scale", 0, JSPROP_PERMANENT, Owner::GetScale, Owner::SetScale },
	{ "Mesh", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetMesh, 0 },
	{ "Editable", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetEditable, 0 },
	{0}
};

JSFunctionSpec CJavaScriptObject::Skeleton::Methods[] = 
{
	{ "PushEditing", Owner::PushEditing, 0, 0, 0 },
	{ "PopEditing", Owner::PopEditing, 0, 0, 0 },
	{ "EditObject", Owner::EditObject, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptObject::GetDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Interface(Context, Object)->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Interface(Context, Object)->SetName(sdpToString(Context, *Value));
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetPosition(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = dynamic_cast<k3dITransform*>(Interface(Context, Object));
	if(transform)
		{
			const sdpVector3 position = transform->Position(k3dITransform::CURRENT_FRAME);
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, position));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetPosition(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = dynamic_cast<k3dITransform*>(Interface(Context, Object));
	if(transform)
		{
			const sdpVector3 position = ConvertVector3(Context, *Value);
			transform->SetPosition(k3dITransform::CURRENT_FRAME, position);
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetOrientation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = dynamic_cast<k3dITransform*>(Interface(Context, Object));
	if(transform)
		{
			const sdpAngleAxis rotation = transform->Orientation(k3dITransform::CURRENT_FRAME);
			*Value = OBJECT_TO_JSVAL(Convert(Context, rotation));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetOrientation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = dynamic_cast<k3dITransform*>(Interface(Context, Object));
	if(transform)
		{
			const sdpAngleAxis rotation = ConvertAngleAxis(Context, *Value);
			transform->SetOrientation(k3dITransform::CURRENT_FRAME, rotation);
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetScale(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = dynamic_cast<k3dITransform*>(Interface(Context, Object));
	if(transform)
		{
			const sdpVector3 scale = transform->Scale(k3dITransform::CURRENT_FRAME);
			*Value = OBJECT_TO_JSVAL(Convert(Context, scale));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetScale(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = dynamic_cast<k3dITransform*>(Interface(Context, Object));
	if(transform)
		{
			const sdpVector3 scale = ConvertVector3(Context, *Value);
			transform->SetScale(k3dITransform::CURRENT_FRAME, scale);
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dIMesh* const mesh = dynamic_cast<k3dIMesh*>(Interface(Context, Object));
	g_return_val_if_fail(mesh, JS_FALSE);
	
	*Value = OBJECT_TO_JSVAL(CJavaScriptMesh::Create(mesh, Context, Object));
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetEditable(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	return DynamicCast<k3dIEditable, CJavaScriptEditable>(Context, Object, ID, Value);
}

JSBool CJavaScriptObject::PushEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PushEditing();
	return JS_TRUE;
}

JSBool CJavaScriptObject::PopEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PopEditing();
	return JS_TRUE;
}

JSBool CJavaScriptObject::EditObject(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Interface(Context, Object)->EditObject();
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptCamera

JSClass CJavaScriptCamera::Skeleton::Class = 
{
	"Camera", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptCamera::Skeleton::Properties[] =
{
	{ "Document", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetDocument, 0 },
	{ "EditMode", 0, JSPROP_PERMANENT, Owner::GetEditMode, Owner::SetEditMode },
	{0}
};

JSFunctionSpec CJavaScriptCamera::Skeleton::Methods[] = 
{
	{ "Redraw", Owner::Redraw, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptCamera::GetDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptCamera::GetEditMode(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, sdpToString(Interface(Context, Object)->ActiveEditMode())));
	return JS_TRUE;
}

JSBool CJavaScriptCamera::SetEditMode(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	const sdpString mode(sdpToString(Context, *Value));
	if(mode == "point")
		Interface(Context, Object)->SetActiveEditMode(k3dICamera::POINT);
	else if(mode == "path")
		Interface(Context, Object)->SetActiveEditMode(k3dICamera::POLYGON);
	else
		g_return_val_if_fail(0, JS_FALSE);
		
	return JS_TRUE;
}

JSBool CJavaScriptCamera::Redraw(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Interface(Context, Object)->Redraw(false);
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptObjectCollection

JSClass CJavaScriptObjectCollection::Skeleton::Class = 
{
	"ObjectCollection", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptObjectCollection::Skeleton::Properties[] =
{
	{ "Editable", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetEditable, 0 },
	{0}
};

JSFunctionSpec CJavaScriptObjectCollection::Skeleton::Methods[] = 
{
	{ "PushEditing", Owner::PushEditing, 0, 0, 0 },
	{ "PopEditing", Owner::PopEditing, 0, 0, 0 },
	{ "Create", Owner::CreateObject, 1, 0, 0 },
	{ "Get", Owner::Get, 0, 0, 0 },
	{ "Object", Owner::Object, 1, 0, 0 },
	{ "Delete", Owner::Delete, 1, 0, 0 },
	{0}
};

JSBool CJavaScriptObjectCollection::GetEditable(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	return DynamicCast<k3dIEditable, CJavaScriptEditable>(Context, Object, ID, Value);
}

JSBool CJavaScriptObjectCollection::PushEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PushEditing();
	return JS_TRUE;
}

JSBool CJavaScriptObjectCollection::PopEditing(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	dynamic_cast<k3dIEditable*>(Interface(Context, Object))->PopEditing();
	return JS_TRUE;
}

JSBool CJavaScriptObjectCollection::CreateObject(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	// Look for object types by name ...
	std::vector<k3dIObjectInfo*> info;
	k3d::copy_if(k3dApplication()->Objects(), std::back_inserter(info), k3d::SameName(sdpToString(Context, argv[0])));
	
	// If we got at exactly one object type, create an instance of it ...
	g_return_val_if_fail(1 == info.size(), JS_FALSE);
	
	// Create an object ...
	k3dIObject* const object = Interface(Context, Object)->CreateObject((*info.begin())->ClassID());
	g_return_val_if_fail(object, JS_FALSE);
	
	// Return that baby...
	*Result = OBJECT_TO_JSVAL(CJavaScriptObject::Create(object, Context, Object));	
	return JS_TRUE;
}

JSBool CJavaScriptObjectCollection::Get(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	// Return the set of objects in this collection ...
	typedef std::vector<k3dIObject*> Objects;
	Objects objects;
	k3d::copy(Interface(Context, Object), std::back_inserter(objects));

	// Convert the set to a Javascript array ...
	std::vector<jsval> values;
	for(Objects::iterator object = objects.begin(); object != objects.end(); ++object)
		values.push_back(OBJECT_TO_JSVAL(CJavaScriptObject::Create(*object, Context, Object)));
	
	// Return that baby ...
	if(values.size())
		*Result = OBJECT_TO_JSVAL(JS_NewArrayObject(Context, values.size(), &values.front()));
	else
		*Result = OBJECT_TO_JSVAL(JS_NewArrayObject(Context, 0, 0));
		
	return JS_TRUE;
}

JSBool CJavaScriptObjectCollection::Object(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	// Look for objects by name ...
	std::vector<k3dIObject*> objects;
	k3d::copy_if(Interface(Context, Object), std::back_inserter(objects), k3d::SameName(sdpToString(Context, argv[0])));
	
	// If we found something return it ...
	g_return_val_if_fail(1 == objects.size(), JS_FALSE);
	
	*Result = OBJECT_TO_JSVAL(CJavaScriptObject::Create(*objects.begin(), Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptObjectCollection::Delete(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIObject* victim = 0;

	if(JSVAL_IS_STRING(argv[0]))
		{
			// Look for our victim by name ...
			std::vector<k3dIObject*> objects;
			k3d::copy_if(Interface(Context, Object), std::back_inserter(objects), k3d::SameName(sdpToString(Context, argv[0])));
	
			if(1 == objects.size());
				victim = *objects.begin();
		}
	else if(JSVAL_IS_OBJECT(argv[0]))
		{
			// We got a direct reference to the object ...
			victim = CJavaScriptObject::Interface(Context, JSVAL_TO_OBJECT(argv[0]));
		}

	Interface(Context, Object)->DeleteObject(victim);
	return JS_TRUE;
}


///////////////////////////////////////////////////////////////
// CJavaScriptDocument

JSClass CJavaScriptDocument::Skeleton::Class = 
{
	"Document", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptDocument::Skeleton::Properties[] =
{
	{ "Application", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetApplication, 0 },
	{ "Objects", 0, JSPROP_PERMANENT | JSPROP_READONLY, Owner::GetObjects, 0 },
	{0}
};

JSFunctionSpec CJavaScriptDocument::Skeleton::Methods[] = 
{
	{ "Import", Owner::Import, 2, 0, 0 },
	{ "Export", Owner::Export, 2, 0, 0 },
	{ "Save", Owner::Save, 1, 0, 0 },
	{ "StartChangeSet", Owner::StartStateChangeSet, 0, 0, 0 },
	{ "FinishChangeSet", Owner::FinishStateChangeSet, 1, 0, 0 },
	{ "RedrawAll", Owner::RedrawAll, 0, 0, 0 },
	{0}
};

JSBool CJavaScriptDocument::GetApplication(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptDocument::GetObjects(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(CJavaScriptObjectCollection::Create(Interface(Context, Object)->Objects(), Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptDocument::Import(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	const sdpString filepath(sdpToString(Context, argv[0]));
	g_return_val_if_fail(filepath.size(), JS_FALSE);
	
	const sdpString formatname(sdpToString(Context, argv[1]));

	k3d::Editing editing(Interface(Context, Object)->Objects());	
	k3d::auto_ptr<k3dIGeometryReadFormat> filter(
		formatname.size() ? 
		k3d::GetFileFilter<k3dIGeometryReadFormat>(Interface(Context, Object), formatname) :
		k3d::GetAutoFilter<k3dIGeometryReadFormat>(Interface(Context, Object), filepath));
	g_return_val_if_fail(filter.get(), JS_FALSE);
	
	return sdpSucceeded(Interface(Context, Object)->Import(filepath, filter.get())) ? JS_TRUE : JS_FALSE;
}

JSBool CJavaScriptDocument::Export(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	const sdpString filepath(sdpToString(Context, argv[0]));
	g_return_val_if_fail(filepath.size(), JS_FALSE);
	
	const sdpString formatname(sdpToString(Context, argv[1]));
	
	k3d::Editing editing(Interface(Context, Object)->Objects());	
	k3d::auto_ptr<k3dIGeometryWriteFormat> filter(
		formatname.size() ?
		k3d::GetFileFilter<k3dIGeometryWriteFormat>(Interface(Context, Object), formatname) :
		k3d::GetAutoFilter<k3dIGeometryWriteFormat>(Interface(Context, Object), filepath));
	g_return_val_if_fail(filter.get(), JS_FALSE);
	
	return sdpSucceeded(Interface(Context, Object)->Export(filepath, filter.get())) ? JS_TRUE : JS_FALSE;
}

JSBool CJavaScriptDocument::Save(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	return sdpSucceeded(Interface(Context, Object)->Save(sdpToString(Context, argv[0]))) ? JS_TRUE : JS_FALSE;
}

JSBool CJavaScriptDocument::StartStateChangeSet(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dStartStateChangeSet(Interface(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptDocument::FinishStateChangeSet(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dFinishStateChangeSet(Interface(Context, Object), sdpToString(Context, argv[0]));
	return JS_TRUE;
}

JSBool CJavaScriptDocument::RedrawAll(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dRedrawAll(Interface(Context, Object), false);
	return JS_TRUE;
}

/*
///////////////////////////////////////////////////////////////
// CJavaScriptRenderFrame

JSClass CJavaScriptRenderFrame::Skeleton::Class = 
{
	"RenderFrame", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptRenderFrame::Skeleton::Properties[] =
{
	{ "Job", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetJob, 0 },
	{ "Name", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetName, 0 },
	{ "Path", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetPath, 0 },
	{ "Priority", 0, JSPROP_PERMANENT, Owner::GetPriority, Owner::SetPriority },
	{ "ViewDuringRender", 0, JSPROP_PERMANENT, Owner::GetViewDuringRender, Owner::SetViewDuringRender },
	{ "ViewAfterRender", 0, JSPROP_PERMANENT, Owner::GetViewAfterRender, Owner::SetViewAfterRender },
	{ "Engine", 0, JSPROP_PERMANENT, Owner::GetEngine, Owner::SetEngine },
	{0}
};

JSFunctionSpec CJavaScriptRenderFrame::Skeleton::Methods[] = 
{
	{0}
};

JSBool CJavaScriptRenderFrame::GetJob(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Interface(Context, Object)->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetPath(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Interface(Context, Object)->Path()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetPriority(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = DOUBLE_TO_JSVAL(JS_NewDouble(Context, gdouble(Interface(Context, Object)->Priority())));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetPriority(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	gdouble priority = k3dIRenderFrame::MODERATE_PRIORITY;
	
	JS_ValueToNumber(Context, *Value, &priority);
	Interface(Context, Object)->SetPriority(guchar(priority));
	
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetViewDuringRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = BOOLEAN_TO_JSVAL(Interface(Context, Object)->ViewDuringRender());
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetViewDuringRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Interface(Context, Object)->SetViewDuringRender(JSVAL_TO_BOOLEAN(*Value) ? true : false);
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetViewAfterRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = BOOLEAN_TO_JSVAL(Interface(Context, Object)->ViewAfterRender());
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetViewAfterRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Interface(Context, Object)->SetViewAfterRender(JSVAL_TO_BOOLEAN(*Value) ? true : false);
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetEngine(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Interface(Context, Object)->Engine()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetEngine(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Interface(Context, Object)->SetEngine(sdpToString(Context, *Value));
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptRenderJob

JSClass CJavaScriptRenderJob::Skeleton::Class = 
{
	"RenderJob", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptRenderJob::Skeleton::Properties[] =
{
	{ "Application", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetApplication, 0 },
	{ "Name", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetName, 0 },
	{ "Path", 0, JSPROP_READONLY | JSPROP_PERMANENT, Owner::GetPath, 0 },
	{0}
};

JSFunctionSpec CJavaScriptRenderJob::Skeleton::Methods[] = 
{
	{ "CreateFrame", Owner::CreateFrame, 1, 0, 0 },
	{0}
};

JSBool CJavaScriptRenderJob::GetApplication(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptRenderJob::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Interface(Context, Object)->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderJob::GetPath(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Interface(Context, Object)->Path()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderJob::CreateFrame(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	sdpString framename;
	
	// Look for object types by name ...
	if(JSVAL_IS_STRING(argv[0]))
		framename = sdpToString(Context, argv[0]);
	
	// If we got a name ...
	if(framename.size())
		{
			k3dIRenderFrame* frame = Interface(Context, Object)->CreateFrame(framename);
			g_return_val_if_fail(frame, JS_FALSE);
			
			*Result = OBJECT_TO_JSVAL(CJavaScriptRenderFrame::Create(frame, Context, Object));
		}
		
	return JS_TRUE;
}
*/

///////////////////////////////////////////////////////////////
// CJavaScriptCommandNode

JSClass CJavaScriptCommandNode::Skeleton::Class = 
{
	"CommandNode", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptCommandNode::Skeleton::Properties[] =
{
	{0}
};

JSFunctionSpec CJavaScriptCommandNode::Skeleton::Methods[] = 
{
	{ "Command", Owner::Command, 2, 0, 0 },
	{0}
};

JSBool CJavaScriptCommandNode::Command(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	const sdpString command(sdpToString(Context, argv[0]));
	g_return_val_if_fail(command.size(), JS_FALSE);

	const sdpString arguments(sdpToString(Context, argv[1]));

	return (SDP_S_OK == Interface(Context, Object)->Command(command, arguments, 0)) ? JS_TRUE : JS_FALSE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptApplication

JSObject* CJavaScriptApplication::Create(JSContext* const Context, JSObject* const GlobalObject)
{
	// Sanity checks ...
	g_assert(Context);
	g_assert(GlobalObject);

	JSObject* result = JS_DefineObject(Context, GlobalObject, "Application", &Class, 0, 0);
	JS_SetPrivate(Context, result, k3dApplication());
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptApplication::Class = 
{
	"Application", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptApplication::Properties[] =
{
	{ "ActiveDocument", 0, JSPROP_READONLY | JSPROP_PERMANENT, GetActiveDocument, 0 },
	{ "BasePath", 0, JSPROP_READONLY | JSPROP_PERMANENT, GetBasePath, 0 },
	{0}
};

JSFunctionSpec CJavaScriptApplication::Methods[] = 
{
	{ "Close", Close, 0, 0, 0 },

	{ "NewDocument", NewDocument, 0, 0, 0 },
	{ "OpenDocument", OpenDocument, 1, 0, 0 },
	{ "CloseDocument", CloseDocument, 1, 0, 0 },

//	{ "CreateJob", CreateJob, 1, 0, 0 },
//	{ "StartJob", StartJob, 1, 0, 0 },
			
	{ "BrowserNavigate", BrowserNavigate, 1, 0, 0, },
	{ "Message", Message, 2, 0, 0 },
	{ "Query", Query, 5, 0, 0 },
	{ "Error", Error, 2, 0, 0 },
	{ "DebugOutput", DebugOutput, 1, 0, 0 },

	{ "CommandNode", CommandNode, 1, 0, 0 },

	{0}
};

JSBool CJavaScriptApplication::GetActiveDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dIDocument* document = k3dApplication()->ActiveDocument();
	if(document)
		*Value = OBJECT_TO_JSVAL(CJavaScriptDocument::Create(document, Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptApplication::GetBasePath(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, k3dApplication()->BasePath()));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Close(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dApplication()->Close();
	return JS_TRUE;
}

JSBool CJavaScriptApplication::NewDocument(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dIDocument* const document = k3dApplication()->CreateDocument("");
	if(document)
		*Result = OBJECT_TO_JSVAL(CJavaScriptDocument::Create(document, Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptApplication::OpenDocument(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dIDocument* const document = k3dApplication()->OpenDocument(sdpToString(Context, argv[0]));
	if(document)
		*Result = OBJECT_TO_JSVAL(CJavaScriptDocument::Create(document, Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptApplication::CloseDocument(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	if(JSVAL_IS_OBJECT(argv[0]))
		{
			k3dIDocument* const document = CJavaScriptDocument::Interface(Context, JSVAL_TO_OBJECT(argv[0]));
			if(document)
				k3dApplication()->CloseDocument(document);
		}

	return JS_TRUE;
}

JSBool CJavaScriptApplication::BrowserNavigate(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	k3dApplication()->BrowserNavigate(sdpToString(Context, argv[0]));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Message(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	k3dApplication()->Message(sdpToString(Context, argv[0]), sdpToString(Context, argv[1]));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Query(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	sdpStringArray buttons;
	for(uintN i = 2; i < argc; i++)
		buttons.push_back(sdpToString(Context, argv[i]));
	
	*Result = INT_TO_JSVAL(k3dApplication()->Query(sdpToString(Context, argv[0]), sdpToString(Context, argv[1]), 0, buttons));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Error(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	k3dApplication()->Error(sdpToString(Context, argv[0]), sdpToString(Context, argv[1]));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::DebugOutput(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	k3dApplication()->DebugOutput(sdpToString(Context, argv[0]));
	return JS_TRUE;
}

k3dICommandNode* GetNode(k3dICommandNode* Parent, SDPCSTRING NodeName)
{
	// Sanity checks ...
	g_assert(Parent);
	g_assert_string(NodeName);

	sdpString nodename(NodeName);

	for(K3DCOMMANDNODEHANDLE nodehandle = Parent->FirstCommandNode(); nodehandle; nodehandle = Parent->NextCommandNode(nodehandle))
		{
			k3dICommandNode* node = Parent->CommandNode(nodehandle);
			if(node->CommandNodeName() == nodename)
				return node;
		}

	return 0;
}

JSBool CJavaScriptApplication::CommandNode(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	sdpString nodepath = sdpToString(Context, argv[0]);
	if(0 == nodepath.size())
		{
			JS_ReportError(Context, "Empty command node path");
			return JS_FALSE;
		}
		
	k3dICommandNode* node = dynamic_cast<k3dICommandNode*>(k3dApplication());
	g_return_val_if_fail(node, JS_FALSE);

	node = node->CommandNodeByPath(nodepath);
	if(0 == node)
		{
			JS_ReportError(Context, "Could not find command node [%s]", SDPCSTRING(nodepath));
			return JS_FALSE;
		}
		
	*Result = OBJECT_TO_JSVAL(CJavaScriptCommandNode::Create(node, Context, Object));
	return JS_TRUE;
}


