// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CPLIBReader K-3D object, which reads PLIB .* files
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "plibreader.h"
#include <iostream>
#include <fstream>
#include <strstream>

#include <k3dcom.h>
#include <k3dmisc.h>

#include <k3dilocation.h>
#include <k3dimaterial.h>
#include <k3dimaterialcollection.h>
#include <k3dimesh.h>
#include <k3dimouseeventobserver.h>
#include <k3diobjectcollection.h>
#include <k3dipoint.h>
#include <k3dipath.h>
#include <k3dipathpointcollection.h>

#include <plib/ssg.h>

namespace
{

k3dIPoint* CreatePoint(k3dIMesh* const Mesh, const sdpVector3 Coords)
{
	// Sanity checks ...
	g_assert(Mesh);

	k3dIPoint* const point = Mesh->CreatePoint();
	g_return_val_if_fail(point, 0);

	point->Location()->SetLocalXYZ(Coords);

	return point;
}

k3dIPath* CreatePath(k3dIMesh* const Mesh)
{
	// Sanity checks ...
	g_assert(Mesh);

	k3dIPath* const path = Mesh->CreatePath();
	g_return_val_if_fail(path, 0);

	return path;
}

void LoadSSG(ssgEntity* node, std::vector<sdpVector3>& points, std::vector<gulong>& triangles)
{
	// Recursive function
	if(node->isAKindOf(ssgTypeBranch()))
		{
			// Branch node, check kids
			int nkids = ((ssgBranch*)node)->getNumKids();

			for(int i = 0; i < nkids; i++)
				LoadSSG(((ssgBranch*)node)->getKid(i), points, triangles);
		}

	if(!node->isAKindOf(ssgTypeLeaf()))
		return;


	ssgLeaf* leaf = (ssgLeaf*)node;
	int trianglenb = leaf->getNumTriangles();

	short v1, v2, v3, max = 0;
	for(int i = 0; i < trianglenb; i++)
		{
			leaf->getTriangle(i, &v1, &v2, &v3);
			triangles.push_back((gulong)v1);
			triangles.push_back((gulong)v2);
			triangles.push_back((gulong)v3);

//std::cerr << endl << " T : " << v1 << " " << v2 << " " << v3;
			if(v1 > max) max = v1;
			if(v2 > max) max = v2;
			if(v3 > max) max = v3;
		}

	for(int i = 0; i <= (int)max; i++)
		{
			float* vertex = leaf->getVertex(i);
			points.push_back(sdpVector3((gdouble)vertex[0], (gdouble)vertex[1], (gdouble)vertex[2]));
//std::cerr << endl << " V : " << points[i];
		}
}

} // namespace

///////////////////////////////////////////////////////////////
// CPLIBReader

CPLIBReader::CPLIBReader(sdpIUnknown* Initializer) :
	CPLIBReaderBase(Initializer)
{
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat implementation

bool CPLIBReader::QueryCanHandle(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	sdpString filepath(FilePath);
	filepath.makelower();
	
	if(".3ds" == filepath.right(4))
		return true;
	if(".ac" == filepath.right(3))
		return true;
	if(".ase" == filepath.right(4))
		return true;
	if(".atg" == filepath.right(4))
		return true;
	if(".dxf" == filepath.right(4))
		return true;
	if(".obj" == filepath.right(4))
		return true;
	if(".off" == filepath.right(4))
		return true;

	return false;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryReadFormat implementation

sdpResult CPLIBReader::PreRead(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CPLIBReader::ReadOptions(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CPLIBReader::ReadFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);

	// Init SSG library loader
	ssgLoaderOptions *loader_opts = new ssgLoaderOptions();
	ssgSetCurrentOptions(loader_opts);

	// Load file using corresponding function
	ssgEntity *obj = 0;
	sdpString filepath(FilePath);
	filepath.makelower();
	if(".3ds" == filepath.right(4))
		obj = ssgLoad3ds(FilePath, loader_opts);
	else
	if(".ac" == filepath.right(3))
		obj = ssgLoadAC3D(FilePath, loader_opts);
	else
	if(".ase" == filepath.right(4))
		obj = ssgLoadASE(FilePath, loader_opts);
	else
	if(".atg" == filepath.right(4))
		obj = ssgLoadATG(FilePath, loader_opts);
	else
	if(".dxf" == filepath.right(4))
		obj = ssgLoadDXF(FilePath, loader_opts);
	else
	if(".obj" == filepath.right(4))
		obj = ssgLoadOBJ(FilePath, loader_opts);
	else
	if(".off" == filepath.right(4))
		obj = ssgLoadOFF(FilePath, loader_opts);

	// Return if SSG loader failed
	g_return_val_if_fail(obj, SDP_E_FAIL);

	ssgFlatten(obj);

	// Declare mesh object pointers and an object counter
	k3dIObject* object;
	k3dIMesh* mesh;
	guint objectnumber = 0;

	// Always use the default material
	k3dIMaterial* defaultmaterial = k3dDefaultMaterial(Document());

//	while(travel through object list)
//		{
			// Create a new mesh object ...
			object = Document()->Objects()->CreateObject(k3d::classid::Mesh);
			g_return_val_if_fail(object, SDP_E_FAIL);

			mesh = dynamic_cast<k3dIMesh*>(object);
			g_return_val_if_fail(mesh, SDP_E_FAIL);

			Document()->SetMouseFocus(dynamic_cast<k3dIMouseEventObserver*>(object));
			object->State()->SetState(k3dIState::SELECTED, true);

			object->SetName("PLIB object " + sdpToString(objectnumber++));

			// Load geometry
			std::vector<sdpVector3> points;
			std::vector<gulong> triangles;
			LoadSSG(obj, points, triangles);

			// Build points
			std::vector<k3dIPoint*> k3dpoints;
			for(gulong n = 0; n < points.size(); n++)
				{
					k3dIPoint* point = CreatePoint(mesh, points[n]);
					g_return_val_if_fail(point, SDP_E_FAIL);
					k3dpoints.push_back(point);
				}

			// Build paths (triangles)
			for(gulong face = 0; face < triangles.size()/3; face++)
				{
					// Create a path ...
					k3dIPath* path = CreatePath(mesh);
					g_return_val_if_fail(path, SDP_E_FAIL);

					for(gulong i = 0; i < 3; i++)
						{
							k3dIPathPoint* pathpoint = path->Points()->AddPoint(k3dpoints[triangles[face*3+i]]);
							g_return_val_if_fail(pathpoint, SDP_E_FAIL);
						}

					// All paths should have a material
					path->Materials()->SetMaterial(defaultmaterial);
				}
//		}

	return SDP_S_OK;
}


