// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBooleanOperator K-3D object, which allows geometry to be rendered with boolean effects
		\author Tim Shead (tshead@k-3d.com)
*/

#include "booleanoperator.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3diriengine.h>

#include <sdpgtkevent.h>

CBooleanOperator::CBooleanOperator(sdpIUnknown* Initializer) :
	CBooleanOperatorBase(Initializer),
	m_Type("type", BOOLEAN_INTERSECTION, Document()->StateRecorder(), 0)
{
	AddVariable(&m_Type);
}

CBooleanOperator::~CBooleanOperator()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CBooleanOperator::EditObject()
{
	new CBooleanOperatorDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////
// k3dIRiRender overrides

namespace
{

class RenderSolid
{
public:
	void operator()(k3dIObject* const Object)
	{
		k3dIRiRender* const rirender = dynamic_cast<k3dIRiRender*>(Object);
		if(rirender)
			rirender->RiSetSolid(true);
	}
};

} // namespace

sdpResult CBooleanOperator::RiPushRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// Record the type of boolean ...
	switch(m_Type.Value())
		{
			case BOOLEAN_INTERSECTION:
				Engine->RiSolidBegin(k3dIRiEngine::RI_INTERSECTION);
				break;

			case BOOLEAN_UNION:
				Engine->RiSolidBegin(k3dIRiEngine::RI_UNION);
				break;

			case BOOLEAN_DIFFERENCE:
				Engine->RiSolidBegin(k3dIRiEngine::RI_DIFFERENCE);
				break;

			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}
		
	// Make sure our descendants know that they should render themselves as "solid" ...
	k3d::for_each_descendant(Hierarchy(), this, RenderSolid());

	return SDP_S_OK;
}

sdpResult CBooleanOperator::RiPopRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	Engine->RiSolidEnd();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

/////////////////////////////////////////////////////////////////////////////
// CBooleanOperatorDialog

const sdpString CBooleanOperatorDialog::dialogname = "properties";
const sdpString CBooleanOperatorDialog::control_booleantype = "booleantype";

CBooleanOperatorDialog::CBooleanOperatorDialog(CBooleanOperator* const Object) :
	CBooleanOperatorDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("sfxlib", "booleanoperator.gtkml"));

	// Setup choosers ...
	if(Getk3dChooser(control_booleantype))
		Getk3dChooser(control_booleantype)->Attach(new k3dChooserVariableDataSource<CBooleanOperator::BooleanType>(m_Object->m_Type), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Boolean Type", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}


