// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CFractalTerrain K-3D object, which creates fractal landscapes
		\author Romain Behar (romainbehar@yahoo.com)
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "fractalterrain.h"

#include <k3dorientationcontroller.h>
#include <k3doptions.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>
#include <k3dmisc.h>

#include <k3diapplication.h>
#include <k3dicamera.h>

#include <sdpgtkevents.h>

#include <sdpgl.h>

/*
///////////////////////////////////////////////////////////////////
// CFractalTerrainFlat

void CFractalTerrainFlat::InitializeTopology(const CFractalTerrain& Terrain)
{
}

void CFractalTerrainFlat::StartSurfaceDefinition(const CFractalTerrain& Terrain)
{
}

gdouble CFractalTerrainFlat::GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize)
{
	return 0.0;
}
*/

///////////////////////////////////////////////////////////////////
// CFractalTerrainFFT

// Generate random terrain

// Compute a random terrain with FFT
void CFractalTerrainFFT::Mountain(gulong N, gdouble d)
{
	gulong k, l;
	sdpComplex Tmp;


	m_Terrain = new fvector[N];
	for(k = 0; k < N; k++)
		m_Terrain[k] = new fftVector(N);

	// Set all values using a normal random variable (RNV),
	// Array[0][0] = 0, Array[0][N/2] = RNV, Array[N/2][0] = RNV,
	// Array[N/2][N/2]; fill up the array a normal random complex
	// variable (CRNV) following the rule : CRNVi = Conjugate(CRNVj)
	// when CRNVi and CRNVj are symmetrically placed around one
	// of the three real values :

	//  0  C  C  C R1  C  C  C  (N = 8)
	//  C  C  C C1  C  C  C  C
	//  C  C  C  C  C  C  C  C
	// C3  C  C  C  C  C  C  C
	// R2  C  C  C R3  C  C  C  C3 = Conjugate(C4), C1 = Conjugate(C2)
	// C4  C  C  C  C  C  C  C
	//  C  C  C  C  C  C  C  C
	//  C  C  C  C  C C2  C  C


	// 0, R1, R2, R3

	m_Terrain[0]->P[0] = sdpComplex(0);
	m_Terrain[0]->P[N/2] = sdpComplex(m_TerrainSeeds[0][N/2].normaldrand(0, N/2, d));
	m_Terrain[N/2]->P[0] = sdpComplex(m_TerrainSeeds[N/2][0].normaldrand(N/2, 0, d));
	m_Terrain[N/2]->P[N/2] = sdpComplex(m_TerrainSeeds[N/2][N/2].normaldrand(N/2, N/2, d));

	// Row[0] and Row[N/2]

	for(k = 1; k < N/2; k++)
	{
		Tmp = m_TerrainSeeds[0][k].cplxnormaldrand(0, k, d);
		m_Terrain[k]->P[0] = Tmp;
		m_Terrain[N-k]->P[0] = Tmp.Conjugate();

		Tmp = m_TerrainSeeds[N/2][k].cplxnormaldrand(N/2, k, d);
		m_Terrain[k]->P[N/2] = Tmp;
		m_Terrain[N-k]->P[N/2] = Tmp.Conjugate();
	}

	// Column[0] and Column[N/2]

	for(k = 1; k < N/2; k++)
	{
		Tmp = m_TerrainSeeds[k][0].cplxnormaldrand(k, 0, d);
		m_Terrain[0]->P[k] = Tmp;
		m_Terrain[0]->P[N-k] = Tmp.Conjugate();

		Tmp = m_TerrainSeeds[k][N/2].cplxnormaldrand(k, N/2, d);
		m_Terrain[N/2]->P[k] = Tmp;
		m_Terrain[N/2]->P[N-k] = Tmp.Conjugate();
	}

	// Fill up

	for(k = 1; k < N/2; k++)
		for(l = 1; l < N/2; l++)
		{
			Tmp = m_TerrainSeeds[k][l].cplxnormaldrand(k, l, d);
			// The square inside 0 R1 R2 R3 and symmtrical values
			m_Terrain[k]->P[l] = Tmp;
			m_Terrain[N-k]->P[N-l] = Tmp.Conjugate();

			Tmp = m_TerrainSeeds[N-k][l].cplxnormaldrand(k, l, d);
			// Bottom left inside square and symmetrical values
			m_Terrain[N-k]->P[l] = Tmp;
			m_Terrain[k]->P[N-l] = Tmp.Conjugate();
		}


	// 2D inverse FFT

	// Apply InvFFT on each row
	for(k = 0; k < N; k++)
		m_Terrain[k]->InvFFT();

	// Apply InvFFT on each column
	fftVector TmpFFT(N);
	for(l = 0; l < N; l++)
		{
			for(k = 0; k < N; k++) TmpFFT.P[k] = m_Terrain[k]->P[l];
			TmpFFT.InvFFT();
			for(k = 0; k < N; k++) m_Terrain[k]->P[l] = TmpFFT.P[k];
		}
}


void CFractalTerrainFFT::InitializeTopology(const CFractalTerrain& Terrain)
{
	gulong	N = gulong(pow(2, Terrain.m_Iterations.Value()));

	srand(static_cast<gulong>(Terrain.m_RandomSeed.Value()));

	m_TerrainSeeds.resize(N);
	for(gulong i = 0; i < N; i++)
		{
			m_TerrainSeeds[i].resize(N);
			for(gulong j = 0; j < N; j++)
				m_TerrainSeeds[i][j].Randomize();
		}
}

void CFractalTerrainFFT::StartSurfaceDefinition(const CFractalTerrain& Terrain)
{
	gulong	N = gulong(pow(2, Terrain.m_Iterations.Value()));

	// Build terrain using its fractal dimension and seeds
	Mountain(N, Terrain.m_FFTFractalDimension.Value());
}

gdouble CFractalTerrainFFT::GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize)
{
	// Sanity checks ...
	g_assert(XIndex < XSize);
	g_assert(YIndex < YSize);

	// 20 is an arbitrary constant to keep same y-scale ratio as other terrain algorithms
	return m_Terrain[YIndex]->P[XIndex].Real() / 20;
}


///////////////////////////////////////////////////////////////////
// CFractalTerrainHeterogeneousfBm

void CFractalTerrainHeterogeneousfBm::InitializeTopology(const CFractalTerrain& Terrain)
{
}

void CFractalTerrainHeterogeneousfBm::StartSurfaceDefinition(const CFractalTerrain& Terrain)
{
	m_ExponentArray.clear();

	gdouble frequency = 1.0;
	for(gulong i = 0; i <= Terrain.m_Octaves.Value(); i++)
		{
			m_ExponentArray.push_back(pow(frequency, -Terrain.m_FractalDimension.Value()));
			frequency *= Terrain.m_Lacunarity.Value();
		}
}

gdouble CFractalTerrainHeterogeneousfBm::GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize)
{
	// Sanity checks ...
	g_assert(XIndex < XSize);
	g_assert(YIndex < YSize);

	// Convert grid coordinates to the range [0, 1], [0, 1] ...
	gdouble x = gdouble(XIndex) / gdouble(XSize-1);
	gdouble y = gdouble(YIndex) / gdouble(YSize-1);

	// First, unscaled octave ...
	sdpVector3 point(x * Terrain.m_Frequency.Value(), y * Terrain.m_Frequency.Value(), Terrain.m_NoiseOffset.Value());
	gdouble value = Terrain.m_Offset.Value() + sdpNoise(point);
	point *= Terrain.m_Lacunarity.Value();

	// Spectral construction inner loop ...
	for(int i = 1; i < Terrain.m_Octaves.Value(); i++)
		{
			// Obtain displaced noise value ...
			gdouble increment = Terrain.m_Offset.Value() + sdpNoise(point);

			// Scale amplitude for this frequency ...
			increment *= m_ExponentArray[i];

			// Scale increment by current "altitude" ....
			increment *= value;

			// Add increment to value ...
			value += increment;

			// Raise spatial frequency ...
			point *= Terrain.m_Lacunarity.Value();
		}

	// Handle remainder in octaves ...
	gdouble remainder = Terrain.m_Octaves.Value() - gint(Terrain.m_Octaves.Value());
	if(remainder)
		{
			gdouble increment = Terrain.m_Offset.Value() + sdpNoise(point) * m_ExponentArray.back();
			value += remainder * increment * value;
		}

	return value;
}

///////////////////////////////////////////////////////////////////
// CFractalTerrain

CFractalTerrain::CFractalTerrain(sdpIUnknown* Initializer) :
	CFractalTerrainBase(Initializer),
	m_TerrainType("terraintype", HfBm, Document()->StateRecorder(), 0),
	m_XScale("xscale", 20.0, Document()->StateRecorder(), 0),
	m_YScale("yscale", 5.0, Document()->StateRecorder(), 0),
	m_ZScale("zscale", 20.0, Document()->StateRecorder(), 0),
	m_Iterations("iterations", 4, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_Smoothing("smoothing", true, Document()->StateRecorder(), 0),
	m_Sea("sea", false, Document()->StateRecorder(), 0),
	m_SeaLevel("sealevel", 0.0, Document()->StateRecorder(), 0),
	m_FractalDimension("fractaldimension", 0.25, Document()->StateRecorder(), 0),
	m_Lacunarity("lacunarity", 0.2, Document()->StateRecorder(), 0),
	m_Octaves("octaves", 4.0, Document()->StateRecorder(), 0),
	m_Offset("offset", 0.0, Document()->StateRecorder(), 0),
	m_Frequency("frequency", 6.0, Document()->StateRecorder(), 0),
	m_NoiseOffset("noiseoffset", 0.5, Document()->StateRecorder(), 0),
	m_FFTFractalDimension("fftfractaldimension", 2.1, Document()->StateRecorder(), 0),
	m_RandomSeed("randomseed", 123, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0)))
{
	AddVariable(&m_TerrainType);
	AddVariable(&m_XScale);
	AddVariable(&m_YScale);
	AddVariable(&m_ZScale);
	AddVariable(&m_Iterations);
	AddVariable(&m_Smoothing);
	AddVariable(&m_Sea);
	AddVariable(&m_SeaLevel);
	AddVariable(&m_FractalDimension);
	AddVariable(&m_Lacunarity);
	AddVariable(&m_Octaves);
	AddVariable(&m_Offset);
	AddVariable(&m_Frequency);
	AddVariable(&m_NoiseOffset);
	AddVariable(&m_FFTFractalDimension);
	AddVariable(&m_RandomSeed);
}

CFractalTerrain::~CFractalTerrain()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CFractalTerrain::EditObject()
{
	new CFractalTerrainDialog(this);
	return SDP_S_OK;
}

void CFractalTerrain::Update(bool CreateGeometry)
{
	switch(m_TerrainType.Value())
		{
			case HfBm:
				Update(CreateGeometry, m_HeterogeneousfBm);
				break;
			case FFT:
				Update(CreateGeometry, m_FFT);
				break;
			default:
				g_return_if_fail(0);
		}
}

void CFractalTerrain::Update(bool CreateGeometry, CFractalTerrainAlgorithm& Algorithm)
{
	// Calculate standard terrain parameters ...
	const gulong npoints = gulong(pow(2, m_Iterations.Value()));
	const gulong nsquares= npoints - 1;

	const gulong xpoints = npoints;
	const gulong zpoints = npoints;
	const gulong xsquares = nsquares;
	const gulong zsquares = nsquares;

	const gdouble xvoxelsize = m_XScale.Value() / gdouble(xsquares);
	const gdouble zvoxelsize = m_ZScale.Value() / gdouble(zsquares);
	const gdouble xoffset = m_XScale.Value() * 0.5;
	const gdouble zoffset = m_ZScale.Value() * 0.5;

	// Create geometry as needed ...
	if(CreateGeometry)
		{
			CreateTriangleGrid(xsquares, zsquares);
			for(gulong n = 0; n < (xsquares+1)*(zsquares+1); n++)
				m_Locations[n] = new k3dLocation();

			Algorithm.InitializeTopology(*this);
		}

	// Calculate surface elevations ...
	Algorithm.StartSurfaceDefinition(*this);

	gulong locationindex = 0;
	for(gulong z = 0; z < zpoints; z++)
		{
			for(gulong x = 0; x < xpoints; x++)
				{
					const gdouble elevation = Algorithm.GetSurfaceElevation(*this, x, z, xpoints, zpoints) * m_YScale.Value();

					const gdouble px = gdouble(x) * xvoxelsize - xoffset;
					const gdouble py = m_Sea.Value() && elevation < m_SeaLevel.Value() ? m_SeaLevel.Value() : elevation;
					const gdouble pz = gdouble(z) * zvoxelsize - zoffset;
					(m_Locations[locationindex])->m_LocalXYZ = sdpVector3(px, py, pz);
					locationindex++;
				}
		}

	if(m_Smoothing.Value())
		SetSmoothShading();
	else
		SetFlatShading();
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CFractalTerrain::InitComplete()
{
	Update(true);
	return CFractalTerrainBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CFractalTerrainDialog

const sdpString CFractalTerrainDialog::dialogname = "properties";
const sdpString CFractalTerrainDialog::control_xscale = "xscale";
const sdpString CFractalTerrainDialog::control_yscale = "yscale";
const sdpString CFractalTerrainDialog::control_zscale = "zscale";
const sdpString CFractalTerrainDialog::control_iterations = "iterations";
const sdpString CFractalTerrainDialog::control_smoothing = "smoothing";
const sdpString CFractalTerrainDialog::control_sea = "sea";
const sdpString CFractalTerrainDialog::control_sealevel = "sealevel";

const sdpString CFractalTerrainDialog::control_pageswitch = "pageswitch";

const sdpString CFractalTerrainDialog::control_fbm = "fbm";
const sdpString CFractalTerrainDialog::control_fft = "fft";

const sdpString CFractalTerrainDialog::control_fractaldimension = "fractaldimension";
const sdpString CFractalTerrainDialog::control_lacunarity = "lacunarity";
const sdpString CFractalTerrainDialog::control_octaves = "octaves";
const sdpString CFractalTerrainDialog::control_offset = "offset";
const sdpString CFractalTerrainDialog::control_frequency = "frequency";
const sdpString CFractalTerrainDialog::control_noiseoffset = "noiseoffset";

const sdpString CFractalTerrainDialog::control_fftfractaldimension = "fftfractaldimension";
const sdpString CFractalTerrainDialog::control_randomseed = "randomseed";

const sdpString CFractalTerrainDialog::control_position = "position";
const sdpString CFractalTerrainDialog::control_orientation = "orientation";
const sdpString CFractalTerrainDialog::control_scale = "scale";
const sdpString CFractalTerrainDialog::control_target = "target";

CFractalTerrainDialog::CFractalTerrainDialog(CFractalTerrain* const Object) :
	CFractalTerrainDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("sfxlib", "fractalterrain.gtkml"));

	// Setup check buttons ...
	Connectk3dCheckButton(control_smoothing, m_Object->m_Smoothing, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Smoothing", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton(control_sea, m_Object->m_Sea, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Sea", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Setup spin buttons ...
	Connectk3dSpinButton(control_xscale, m_Object->m_XScale, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain X Scale", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_yscale, m_Object->m_YScale, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Y Scale", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_zscale, m_Object->m_ZScale, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Z Scale", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_iterations, m_Object->m_Iterations, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Iterations", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_sealevel, m_Object->m_SeaLevel, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Sea Level", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_fractaldimension, m_Object->m_FractalDimension, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Fractal Dimension", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_lacunarity, m_Object->m_Lacunarity, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Lacunarity", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_octaves, m_Object->m_Octaves, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Octaves", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_offset, m_Object->m_Offset, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Offset", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_frequency, m_Object->m_Frequency, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Frequency", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_noiseoffset, m_Object->m_NoiseOffset, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Noise Offset", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_fftfractaldimension, m_Object->m_FFTFractalDimension, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain FFT Fractal Dimension", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_randomseed, m_Object->m_RandomSeed, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Fractal Terrain Random Seed", 0, new k3dActionUpdate<CFractalTerrain>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}

void CFractalTerrainDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_pageswitch)
		OnPageSwitch(Event);
	else
		CFractalTerrainDialogBase::OnEvent(Event);
}


// User switched from one tab to another ...
void CFractalTerrainDialog::OnPageSwitch(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Keep track of the current page ...
	sdpGtkEventNotebookSwitchPage* event = (sdpGtkEventNotebookSwitchPage*)Event;
	CFractalTerrain::TerrainType currentpage = CFractalTerrain::TerrainType(event->PageNumber());

	sdpString pagename;
	switch(currentpage)
		{
			case CFractalTerrain::HfBm:
				pagename = control_fbm;
				break;
			case CFractalTerrain::FFT:
				pagename = control_fft;
				break;
			default:
				g_return_if_fail(0);
		}

	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_pageswitch, pagename);

	// Modify the geometry ...
	k3dSetValue(m_Object->m_TerrainType, currentpage);
	m_Object->Update(true);
	k3dRedrawAll(m_Object->Document(), false);
}


