#ifndef FRACTALTERRAIN_H
#define FRACTALTERRAIN_H

// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CFractalTerrain K-3D object, which creates fractal landscapes
		\author Romain Behar (romainbehar@yahoo.com)
		\author Timothy M. Shead (tshead@k-3d.com)
*/

// Forward class declarations ...
class CFractalTerrain;

#include <k3dobjectdialog.h>
#include <k3dtrianglemesh.h>
#include <k3deditorrender.h>
#include <k3dvariable.h>

#include <k3dimaterial.h>

extern const sdp::ClassID sdpFractalTerrain;

/////////////////////////////////////////////////////////////////////////////
// CFractalTerrainDialog

typedef k3dObjectDialog CFractalTerrainDialogBase;

class CFractalTerrainDialog :
	public CFractalTerrainDialogBase
{
public:
	CFractalTerrainDialog(CFractalTerrain* const Object);

private:
	CFractalTerrain* const m_Object;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);
	void OnPageSwitch(sdpGtkEvent* Event);

	static const sdpString dialogname;
	static const sdpString control_xscale;
	static const sdpString control_yscale;
	static const sdpString control_zscale;
	static const sdpString control_iterations;
	static const sdpString control_smoothing;
	static const sdpString control_sea;
	static const sdpString control_sealevel;
	static const sdpString control_pageswitch;
	static const sdpString control_fbm;
	static const sdpString control_fft;
	static const sdpString control_fractaldimension;
	static const sdpString control_lacunarity;
	static const sdpString control_octaves;
	static const sdpString control_offset;
	static const sdpString control_frequency;
	static const sdpString control_noiseoffset;
	static const sdpString control_fftfractaldimension;
	static const sdpString control_randomseed;
	static const sdpString control_position;
	static const sdpString control_orientation;
	static const sdpString control_scale;
	static const sdpString control_target;
};

/////////////////////////////////////////////////////
// CFractalTerrainAlgorithm

/// An abstract interface for algorithms implementing different strategies for creating geometric terrains
class CFractalTerrainAlgorithm
{
public:
	/// Called whenever (topological) size of the terrain changes
	virtual void InitializeTopology(const CFractalTerrain& Terrain) = 0;
	/// Called before surface-definition begins
	virtual void StartSurfaceDefinition(const CFractalTerrain& Terrain) = 0;
	/// Called repeatedly during surface-definition to return the terrain elevation at a given sample point in the range [0, XSize), [0, YSize) ... results should (preferably) be normalized to the range [-1, 1]
	virtual gdouble GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize) = 0;
};

/*
/////////////////////////////////////////////////////
// CFractalTerrainFlat

/// A fractal terrain algorithm that returns a flat surface
class CFractalTerrainFlat : public CFractalTerrainAlgorithm
{
public:
	virtual ~CFractalTerrainFlat() {}

	void InitializeTopology(const CFractalTerrain& Terrain);
	void StartSurfaceDefinition(const CFractalTerrain& Terrain);
	gdouble GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize);
};
*/

/////////////////////////////////////////////////////
// CFractalTerrainHeterogeneousfBm

/// A fractal terrain algorithm that creates heterogeneous terrains based on fBM (from Perlin, "Texturing and Modeling, A Procedural Approach")
class CFractalTerrainHeterogeneousfBm : public CFractalTerrainAlgorithm
{
public:
	virtual ~CFractalTerrainHeterogeneousfBm() {}

	void InitializeTopology(const CFractalTerrain& Terrain);
	void StartSurfaceDefinition(const CFractalTerrain& Terrain);
	gdouble GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize);

protected:
	std::vector<gdouble> m_ExponentArray;
};

/////////////////////////////////////////////////////
// CFractalTerrainFFT

/// A fractal terrain algorithm that creates terrains out of a brownian movement using Fast Fourrier Transform (actually a 2-variable discrete inverse FFT is used)
class CFractalTerrainFFT : public CFractalTerrainAlgorithm
{
public:
	virtual ~CFractalTerrainFFT() {}

	void InitializeTopology(const CFractalTerrain& Terrain);
	void StartSurfaceDefinition(const CFractalTerrain& Terrain);
	gdouble GetSurfaceElevation(const CFractalTerrain& Terrain, const gulong XIndex, const gulong YIndex, const gulong XSize, const gulong YSize);

protected:
	std::vector<gdouble> m_ExponentArray;
// ---


	class terrainSeed
	{
	public:
		/// Store four random numbers
		sdpVector4 m_Seeds;

		terrainSeed() {}
		terrainSeed(sdpVector4 seed) { m_Seeds = seed; }

		// A random variable
		gdouble drand()
			{
				gdouble	d = (gdouble)rand()/(gdouble)RAND_MAX;
				if(d < 1e-6) return 1e-6;
				return d;
			}

		// A normal random variable
		gdouble normaldrand(int k, int l, gdouble d)
			{
				gdouble dk = (gdouble)k, dl = (gdouble)l;
				gdouble sigma = 1/exp((4-d)/2*log(dk*dk + dl*dl));
				return sigma*sqrt(-2*log(m_Seeds.n[0]))*cos(2*sdp::Pi()*m_Seeds.n[1]);
			}

		// A normal random complex variable
		sdpComplex cplxnormaldrand(int k, int l, gdouble d)
			{
				gdouble dk = (gdouble)k, dl = (gdouble)l;
				gdouble sigma = 1/exp((4-d)/2*log(dk*dk + dl*dl));
				return sigma*sdpComplex(sqrt(-2*log(m_Seeds.n[0]))*cos(2*sdp::Pi()*m_Seeds.n[1]), sqrt(-2*log(m_Seeds.n[2]))*cos(2*sdp::Pi()*m_Seeds.n[3]));
			}

		void Randomize()
			{
				for(gulong i = 0; i < 4; i++)
					m_Seeds.n[i] = drand();
			}
	};

	class fftVector
	{
	public:
		std::vector<sdpComplex> P;

		fftVector();
		fftVector(int Size)
			{
				m_Size = Size;

				P.resize(m_Size);
				Roots.resize(m_Size);
				Temp.resize(m_Size);


				sdpComplex Wj, Wt;

				Roots[0] = sdpComplex(1.0);
				Wt = sdpPolar(1.0, sdp::PiTimes2() / m_Size);
				Wj = Wt;
				for(int i = m_Size-1; i > 0; i--)
					{
						Roots[i] = Wt;
						Wt *= Wj;
					}
			}

		~fftVector()
			{
				P.clear();
				Roots.clear();
				Temp.clear();
			}

		void InvFFT()
			{
				RealFFT(m_Size, 0);
			}

		void RealFFT(gulong d, gulong k)
			{
				if(d >= 2)
					{
						gulong i, j;

						for(i = 0; i < d/2; i++)
							{
								j = k + 2*i;
								Temp[i] = P[j];
								Temp[i+d/2] = P[j+1];
							}

						for(i = 0; i < d; i++)
							P[k+i] = Temp[i];

						RealFFT(d/2, k);
						RealFFT(d/2, k+d/2);

						j = m_Size / d;
						for(i = 0; i < d/2; i++)
							{
								sdpComplex aux = Roots[i*j] * P[k+d/2+i];
								Temp[i] = P[k+i] + aux;
								Temp[i+d/2] = P[k+i] - aux;
							}

						for(i = 0; i < d; i++) P[k+i] = Temp[i];
					}
			}

	protected:
		int m_Size;

		std::vector<sdpComplex> Roots;
		std::vector<sdpComplex> Temp;
	};
	typedef fftVector *fvector;

	// Random seeds for a fractal terrain (built using rules)
	std::vector< std::vector<terrainSeed> > m_TerrainSeeds;

	// FFT fractal terrain
	fvector* m_Terrain;

	// Generate a fractal terrain using preset random values (m_TerrainSeeds)
	void Mountain(gulong N, gdouble d);

};

/////////////////////////////////////////////////////
// CFractalTerrain

typedef k3dTriangleMesh CFractalTerrainBase;
class CFractalTerrain :	public CFractalTerrainBase
{
friend class CFractalTerrainDialog;
friend class CFractalTerrainFlat;
friend class CFractalTerrainHeterogeneousfBm;
friend class CFractalTerrainFFT;
public:
	CFractalTerrain(sdpIUnknown* Initializer);
	~CFractalTerrain();

	// k3dIObject overrides ...
	sdpResult EditObject();

	// k3dIPersist overrides ...
	sdpResult InitComplete();

	// Where it all happens ...
	void Update(bool CreateGeometry);

protected:
	/// Stores the algorithm type to use
	typedef enum
	{
		HfBm,
		FFT,
	} TerrainType;

	friend std::ostream& operator<<(std::ostream& Stream, const TerrainType& Value)
	{
		switch(Value)
			{
				case HfBm:
					Stream << "hfbm";
					break;
				case FFT:
					Stream << "fft";
					break;
			}

		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, TerrainType& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "hfbm")
			Value = HfBm;
		else if(text == "fft")
			Value = FFT;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;

		return Stream;
	}


	k3dVariable<TerrainType> m_TerrainType;

	// Common data ...
	k3dVariable<gdouble> m_XScale;
	k3dVariable<gdouble> m_YScale;
	k3dVariable<gdouble> m_ZScale;
	k3dVariable<gdouble> m_Iterations;
	k3dVariable<bool> m_Smoothing;
	k3dVariable<bool> m_Sea;
	k3dVariable<gdouble> m_SeaLevel;

	// fBm terrain data ...
	k3dVariable<gdouble> m_FractalDimension;
	k3dVariable<gdouble> m_Lacunarity;
	k3dVariable<gdouble> m_Octaves;
	k3dVariable<gdouble> m_Offset;
	k3dVariable<gdouble> m_Frequency;
	k3dVariable<gdouble> m_NoiseOffset;

	// FFT terrain data ...
	k3dVariable<gdouble> m_FFTFractalDimension;
	k3dVariable<gdouble> m_RandomSeed;

	// Does the real work ...
	void Update(bool CreateGeometry, CFractalTerrainAlgorithm& Algorithm);

	// Available algorithms
//	CFractalTerrainFlat m_Flat;
	CFractalTerrainHeterogeneousfBm m_HeterogeneousfBm;
	CFractalTerrainFFT m_FFT;

};

#endif // FRACTALTERRAIN_H
