// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CImageLayer K-3D object, which renders 2D graphics either in front of or behind all 3D geometry.
		\author Tim Shead (tshead@k-3d.com)
*/

#include "imagelayer.h"

#include <k3dchannelbutton.h>
#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3drishadersetup.h>
#include <k3dshaderargumentspage.h>
#include <k3dshaderinfo.h>
#include <k3dshaderselector.h>

#include <k3diapplication.h>
#include <k3dieditorengine.h>
#include <k3dieditortexture.h>
#include <k3ditexture.h>
#include <k3diriengine.h>
#include <k3diritexture.h>
#include <k3diviewvolume.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

double CImageLayer::m_ForegroundCount = 0.0;
double CImageLayer::m_BackgroundCount = 0.0;

CImageLayer::CImageLayer(sdpIUnknown* Initializer) :
	CImageLayerBase(Initializer),
	m_Editor("editor", true, Document()->StateRecorder(), 0),
	m_Layer("layer", BACKGROUND, Document()->StateRecorder(), 0),
	m_Color("color", Document(), Document()->StateRecorder(), 0),
	m_Strength("strength", Document(), Document()->StateRecorder(), 0),
	m_Shader("shader", Document(), Document()->StateRecorder())
{
	AddVariable(&m_Editor);
	AddVariable(&m_Layer);
	AddVariable(&m_Color);
	AddVariable(&m_Strength);
	AddVariable(&m_Shader);
}

CImageLayer::~CImageLayer()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CImageLayer::EditObject()
{
	new CImageLayerDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// sdpsl::IShaderVisitor implementation

void CImageLayer::VisitShaders(sdpsl::IShaderVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);
	
	if(m_Shader.Shader())
		Visitor->VisitShader(m_Shader.Shader());
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender implementation

sdpResult CImageLayer::EditorPreRender(k3dIEditorEngine* const Engine)
{
	if(BACKGROUND == m_Layer.Value())	
		return DrawEditor(Engine);

	return SDP_S_FALSE;
}

sdpResult CImageLayer::EditorPostRender(k3dIEditorEngine* const Engine)
{
	if(FOREGROUND == m_Layer.Value())
		return DrawEditor(Engine);

	return SDP_S_FALSE;
}

sdpResult CImageLayer::DrawEditor(k3dIEditorEngine* const Engine)
{
	// If we aren't supposed to draw in the editor view, we're done ...
	if(false == m_Editor.Value())
		return SDP_S_FALSE;

	// If we don't have a texture assigned, we're done ...
	k3dITexture* const texture = m_Shader.PrimaryTexture();
	if(0 == texture)
		return SDP_S_FALSE;
		
	k3dIEditorTexture* const editortexture = dynamic_cast<k3dIEditorTexture*>(texture);
	g_return_val_if_fail(editortexture, SDP_E_FAIL);

	StoreOpenGLAttributes attributes;
		
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(-1.0, 1.0, -1.0, 1.0);

	// Get our color and strength ...
	sdpVector3 color;
	m_Color->Value3(m_CurrentTime, m_DeltaTime, color);
	gdouble strength = 0.25 * m_Strength->Value1(m_CurrentTime, m_DeltaTime);

	// Set ourselves up to draw an unlit, textured path with blending ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4d(color[0], color[1], color[2], strength);

	editortexture->EditorTextureSetup(Engine);

	glBegin(GL_POLYGON);

	glTexCoord2d(0.0, 0.0);
	glVertex3d(-1.0, 1.0, 1.0);
	
	glTexCoord2d(1.0, 0.0);
	glVertex3d(1.0, 1.0, 1.0);
	
	glTexCoord2d(1.0, 1.0);
	glVertex3d(1.0, -1.0, 1.0);
	
	glTexCoord2d(0.0, 1.0);
	glVertex3d(-1.0, -1.0, 1.0);

	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////
// k3dIRiRender implementation

sdpResult CImageLayer::RiPreRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	m_ForegroundCount = 0.0;
	m_BackgroundCount = 0.0;

	return SDP_S_OK;
}

sdpResult CImageLayer::RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;
		
	// If the user hasn't selected a shader, we're done ...
	if(0 == m_Shader.Shader())
		return SDP_S_FALSE;
		
	// Get data from the camera ...
	const gdouble nearplane = Engine->Camera()->ViewVolume()->NearPlane();
	const gdouble farplane = Engine->Camera()->ViewVolume()->FarPlane();
	const gdouble right = Engine->Camera()->ViewVolume()->Right();
	const gdouble top = Engine->Camera()->ViewVolume()->Top();
	
	const sdpVector3 cameracoords = k3dGetCameraToWorldMatrix(Engine->Camera()) * sdpVector3(0, 0, 0);

	// Reset the current transformation ...
	Engine->RiIdentity();
	Engine->RiTranslate(cameracoords[0], cameracoords[1], cameracoords[2]);
	Engine->RiRotate(90.0f, 1.0f, 0.0f, 0.0f);

	// Disable shadows ...
	k3dRiParameterList shadows;
	shadows.AddTokenValuePair(new k3dRiTokenValuePair("casts_shadows", new k3dRiStringArray("string", "none")));
	shadows.SetAttributes("render", Engine);

	// Set base color ...
	sdpVector3 color;
	m_Color->Value3(m_CurrentTime, m_DeltaTime, color);
	Engine->RiColor(color);

	// Set opacity ...
	sdpVector3 strength = m_Strength->Value1(m_CurrentTime, m_DeltaTime);
	Engine->RiOpacity(strength);

	// Setup the surface shader ...
	k3dRiSurfaceShader shader;
	m_Shader.InsertInto(Frame, Engine, m_CurrentTime, m_DeltaTime, shader);
	shader.Setup(Engine);

	// Draw a sphere
	gdouble radius;
	if(BACKGROUND == m_Layer.Value())
		{
			radius = (0.99 * farplane) - m_BackgroundCount;
			m_BackgroundCount += 1.0;
		}
	else
		{
			sdpVector3 fradius(right, top, nearplane);
			radius = (2.0 * fradius.Length()) + m_ForegroundCount;
			m_ForegroundCount += 0.01;
		}

	Engine->RiSphereV(radius, -radius, radius, 360.0, 0, 0, 0, 0);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CImageLayer::InitComplete()
{
	if(0 == m_Shader.Shader())
		m_Shader.SetShader("k3d_null");

	if(0 == m_Strength.Object())
		g_return_val_if_fail(sdpSucceeded(m_Strength.Create("ImageLayer Strength", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true)), SDP_E_FAIL);
	if(0 == m_Color.Object())
		g_return_val_if_fail(sdpSucceeded(m_Color.Create("ImageLayer Color", sdpVector3(1.0), true, true)), SDP_E_FAIL);

	return CImageLayerBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CImageLayerDialog

const sdpString CImageLayerDialog::dialogname = "properties";
const sdpString CImageLayerDialog::control_editor = "editor";
const sdpString CImageLayerDialog::control_layer = "layer";
const sdpString CImageLayerDialog::control_color = "color";
const sdpString CImageLayerDialog::control_strength = "strength";
const sdpString CImageLayerDialog::control_shaderselector = "shaderselector";
const sdpString CImageLayerDialog::control_shadercontainer = "shadercontainer";

CImageLayerDialog::CImageLayerDialog(CImageLayer* const Object) :
	CImageLayerDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("sfxlib", "imagelayer.gtkml"));

	// Setup check buttons ...
	Connectk3dCheckButton(control_editor, m_Object->m_Editor, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Visible in Editor", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup choosers ...
	if(Getk3dChooser(control_layer))
		Getk3dChooser(control_layer)->Attach(new k3dChooserVariableDataSource<CImageLayer::Layer>(m_Object->m_Layer), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Layer", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	

	// Setup channel buttons ...
	Connectk3dChannelButton(control_color, m_Object->m_Color.Interface());
	Connectk3dChannelButton(control_strength, m_Object->m_Strength.Interface());

	// Setup shader selector ...
	if(Getk3dShaderSelector("shaderselector"))
		Getk3dShaderSelector("shaderselector")->Attach(m_Object->Document(), &m_Object->m_Shader, true, sdpsl::IShader::SURFACE);

	if(Getk3dShaderInfo("shaderinfo"))
		Getk3dShaderInfo("shaderinfo")->Attach(m_Object->Document(), &m_Object->m_Shader);
	
	// Setup shader arguments pages ...		
	if(Getk3dShaderArgumentsPage("shaderarguments"))
		Getk3dShaderArgumentsPage("shaderarguments")->Attach(m_Object->Document(), &m_Object->m_Shader);

	Show();
}

void CImageLayerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	CImageLayerDialogBase::OnEvent(Event);
}


