// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CParticleBurst K-3D object, which creates a parametric particle system
		\author Tim Shead (tshead@k-3d.com)
*/

#include "particleburst.h"

#include <k3dchooser.h>
#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>
#include <k3dichannel.h>
#include <k3didocument.h>
#include <k3dihierarchy.h>
#include <k3diriengine.h>
#include <k3distate.h>
#include <k3ditexture.h>

#include <sdpgl.h>

////////////////////////////////////////////////////////////////////////////
// CParticle

CParticle::CParticle() :
	k3dState(k3dIState::VISIBLE)
{
}

sdpResult CParticle::SetState(gulong Mask, bool State)
{
	RawSetState(Mask, State);
	return SDP_S_OK;
}

k3dIState* CParticle::State()
{
	return this;
}

const sdpVector3& CParticle::LocalXYZ()
{
	return m_AnimXYZ;
}

const sdpVector3& CParticle::AnimXYZ()
{
	return m_AnimXYZ;
}

const sdpVector3& CParticle::WorldXYZ()
{
	return m_WorldXYZ;
}

const sdpVector3& CParticle::CameraXYZ()
{
	return m_CameraXYZ;
}

const sdpVector3& CParticle::ScreenXYZ()
{
	return m_ScreenXYZ;
}

namespace
{

/** \todo Get rid of this hack **/
sdpVector3 g_DummyCoordinate;

} // namespace

const sdpVector3& CParticle::NormalXYZ()
{
	return g_DummyCoordinate;
}

const sdpVector3& CParticle::TextureXYZ()
{
	return g_DummyCoordinate;
}

const sdpVector3& CParticle::ImplicitTextureXYZ()
{
	return g_DummyCoordinate;
}

void CParticle::SetLocalXYZ(const sdpVector3& Coords)
{
}

void CParticle::SetAnimXYZ(const sdpVector3& Coords)
{
	m_AnimXYZ = Coords;
}

void CParticle::SetWorldXYZ(const sdpVector3& Coords)
{
	m_WorldXYZ = Coords;
}

void CParticle::SetCameraXYZ(const sdpVector3& Coords)
{
	m_CameraXYZ = Coords;
}

void CParticle::SetScreenXYZ(const sdpVector3& Coords)
{
	m_ScreenXYZ = Coords;
}

void CParticle::SetNormalXYZ(const sdpVector3& Coords)
{
}

void CParticle::SetTextureXYZ(const sdpVector3& Coords)
{
}

void CParticle::SetImplicitTextureXYZ(const sdpVector3& Coords)
{
}

////////////////////////////////////////////////////////////////////////////
// CParticleBurst

const sdpString CParticleBurst::m_RecalculateParticles = "RecalculateParticles";
const sdpString CParticleBurst::m_CreateParticles = "CreateParticles";

CParticleBurst::CParticleBurst(sdpIUnknown* Initializer) :
	CParticleBurstBase(Initializer),

	m_PreviewColors("previewcolors", false, Document()->StateRecorder(), 0),	
//	m_ParticleType("particletype", POINTS, Document()->StateRecorder(), 0),

	m_Globaltime("globaltime", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Birthrate("birthrate", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_CreateParticles, 0)),
	m_Minlife("minlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_CreateParticles, 0)),
	m_Maxlife("maxlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_CreateParticles, 0)),
	m_Globalstrength("globalstrength", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),

	m_EmitterPositionSamples("emitterpositionsamples", 1.0, Document()->StateRecorder(), 0),
	m_BirthRateSamples("birthratesamples", 1.0, Document()->StateRecorder(), 0),
	m_RandomSeed("randomseed", 12345, Document()->StateRecorder(), 0),
	
	m_Color("color", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_ColorType("colortype", LIFETIME, Document()->StateRecorder(), 0),

	m_Minstrength("minstrength", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minstrengthend("minstrengthend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxstrength("maxstrength", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxstrengthend("maxstrengthend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Strengthlife("strengthlife", Document(), Document()->StateRecorder(), 0),
	m_StrengthRandom("strengthrandom", true, Document()->StateRecorder(), 0),
	m_StrengthRelative("strengthrelative", false, Document()->StateRecorder(), 0),
	m_StrengthLifetime("strengthlifetime", true, Document()->StateRecorder(), 0),

	m_Minsize("minsize", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minsizeend("minsizeend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxsize("maxsize", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxsizeend("maxsizeend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Sizelife("sizelife", Document(), Document()->StateRecorder(), 0),
	m_SizeRandom("sizerandom", true, Document()->StateRecorder(), 0),
	m_SizeRelative("sizerelative", false, Document()->StateRecorder(), 0),
	m_SizeLifetime("sizelifetime", true, Document()->StateRecorder(), 0),
	
	m_Mindirectionx("mindirectionx", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdirectionx("maxdirectionx", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindirectiony("mindirectiony", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdirectiony("maxdirectiony", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindirectionz("mindirectionz", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdirectionz("maxdirectionz", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_DirectionRandom("directionrandom", true, Document()->StateRecorder(), 0),
	m_DirectionEmitter("directionemitter", true, Document()->StateRecorder(), 0),

	m_Mindistance("mindistance", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindistanceend("mindistanceend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdistance("maxdistance", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdistanceend("maxdistanceend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Distancelife("distancelife", Document(), Document()->StateRecorder(), 0),
	m_DistanceRandom("distancerandom", true, Document()->StateRecorder(), 0),
	m_DistanceRelative("distancerelative", false, Document()->StateRecorder(), 0),
	m_DistanceLifetime("distancelifetime", true, Document()->StateRecorder(), 0),
	
	m_Mindriftx("mindriftx", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindriftxend("mindriftxend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdriftx("maxdriftx", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdriftxend("maxdriftxend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Driftxlife("driftxlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindrifty("mindrifty", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindriftyend("mindriftyend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdrifty("maxdrifty", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdriftyend("maxdriftyend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Driftylife("driftylife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindriftz("mindriftz", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mindriftzend("mindriftzend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdriftz("maxdriftz", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxdriftzend("maxdriftzend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Driftzlife("driftzlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_DriftRandom("driftrandom", true, Document()->StateRecorder(), 0),
	m_DriftRelative("driftrelative", false, Document()->StateRecorder(), 0),
	m_DriftLifetime("driftlifetime", true, Document()->StateRecorder(), 0),
			
	m_Mintwistx("mintwistx", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mintwistxend("mintwistxend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxtwistx("maxtwistx", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxtwistxend("maxtwistxend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Twistxlife("twistxlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mintwisty("mintwisty", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mintwistyend("mintwistyend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxtwisty("maxtwisty", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxtwistyend("maxtwistyend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Twistylife("twistylife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mintwistz("mintwistz", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Mintwistzend("mintwistzend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxtwistz("maxtwistz", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxtwistzend("maxtwistzend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Twistzlife("twistzlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_TwistRandom("twistrandom", true, Document()->StateRecorder(), 0),
	m_TwistRelative("twistrelative", false, Document()->StateRecorder(), 0),
	m_TwistLifetime("twistlifetime", true, Document()->StateRecorder(), 0),

/*	
	m_Minwidth("minwidth", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minwidthend("minwidthend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxwidth("maxwidth", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxwidthend("maxwidthend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Widthlife("widthlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minheight("minheight", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minheightend("minheightend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxheight("maxheight", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxheightend("maxheightend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Heightlife("heightlife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minangle("minangle", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Minangleend("minangleend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxangle("maxangle", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Maxangleend("maxangleend", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_Anglelife("anglelife", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIChannel::Event_Changed, m_RecalculateParticles, 0)),
	m_WidthRandom("widthrandom", true, Document()->StateRecorder(), 0),
	m_WidthRelative("widthrelative", false, Document()->StateRecorder(), 0),
	m_WidthLifetime("widthlifetime", true, Document()->StateRecorder(), 0),
	m_HeightRandom("heightrandom", true, Document()->StateRecorder(), 0),
	m_HeightRelative("heightrelative", false, Document()->StateRecorder(), 0),
	m_HeightLifetime("heightlifetime", true, Document()->StateRecorder(), 0),
	m_AngleAlign("anglealign", false, Document()->StateRecorder(), 0),
	m_AngleRandom("anglerandom", true, Document()->StateRecorder(), 0),
	m_AngleRelative("anglerelative", false, Document()->StateRecorder(), 0),
	m_AngleLifetime("anglelifetime", true, Document()->StateRecorder(), 0),
*/	
	
	m_InitializationComplete(false),
	m_Animate(true),
	m_Time(Document()->Time())
{
	// Sanity checks ...
	g_assert(m_Time);
	
//	AddVariable(&m_ParticleType);
	
	AddVariable(&m_Globaltime);
	AddVariable(&m_Birthrate);
	AddVariable(&m_Minlife);
	AddVariable(&m_Maxlife);
	AddVariable(&m_Globalstrength);

	AddVariable(&m_EmitterPositionSamples);
	AddVariable(&m_BirthRateSamples);
	AddVariable(&m_RandomSeed);
		
	AddVariable(&m_Color);
	AddVariable(&m_ColorType);

	AddVariable(&m_Minstrength);
	AddVariable(&m_Minstrengthend);
	AddVariable(&m_Maxstrength);
	AddVariable(&m_Maxstrengthend);
	AddVariable(&m_Strengthlife);
	AddVariable(&m_StrengthRandom);
	AddVariable(&m_StrengthRelative);
	AddVariable(&m_StrengthLifetime);
		
	AddVariable(&m_Minsize);
	AddVariable(&m_Minsizeend);
	AddVariable(&m_Maxsize);
	AddVariable(&m_Maxsizeend);
	AddVariable(&m_Sizelife);
	AddVariable(&m_SizeRandom);
	AddVariable(&m_SizeRelative);
	AddVariable(&m_SizeLifetime);
	
	AddVariable(&m_Mindirectionx);
	AddVariable(&m_Maxdirectionx);
	AddVariable(&m_Mindirectiony);
	AddVariable(&m_Maxdirectiony);
	AddVariable(&m_Mindirectionz);
	AddVariable(&m_Maxdirectionz);
	AddVariable(&m_DirectionRandom);
	AddVariable(&m_DirectionEmitter);
	
	AddVariable(&m_Mindistance);
	AddVariable(&m_Mindistanceend);
	AddVariable(&m_Maxdistance);
	AddVariable(&m_Maxdistanceend);
	AddVariable(&m_Distancelife);
	AddVariable(&m_DistanceRandom);
	AddVariable(&m_DistanceRelative);
	AddVariable(&m_DistanceLifetime);
	
	AddVariable(&m_Mindriftx);
	AddVariable(&m_Mindriftxend);
	AddVariable(&m_Maxdriftx);
	AddVariable(&m_Maxdriftxend);
	AddVariable(&m_Driftxlife);
	AddVariable(&m_Mindrifty);
	AddVariable(&m_Mindriftyend);
	AddVariable(&m_Maxdrifty);
	AddVariable(&m_Maxdriftyend);
	AddVariable(&m_Driftylife);
	AddVariable(&m_Mindriftz);
	AddVariable(&m_Mindriftzend);
	AddVariable(&m_Maxdriftz);
	AddVariable(&m_Maxdriftzend);
	AddVariable(&m_Driftzlife);
	AddVariable(&m_DriftRandom);
	AddVariable(&m_DriftRelative);
	AddVariable(&m_DriftLifetime);
	
	AddVariable(&m_Mintwistx);
	AddVariable(&m_Mintwistxend);
	AddVariable(&m_Maxtwistx);
	AddVariable(&m_Maxtwistxend);
	AddVariable(&m_Twistxlife);
	AddVariable(&m_Mintwisty);
	AddVariable(&m_Mintwistyend);
	AddVariable(&m_Maxtwisty);
	AddVariable(&m_Maxtwistyend);
	AddVariable(&m_Twistylife);
	AddVariable(&m_Mintwistz);
	AddVariable(&m_Mintwistzend);
	AddVariable(&m_Maxtwistz);
	AddVariable(&m_Maxtwistzend);
	AddVariable(&m_Twistzlife);
	AddVariable(&m_TwistRandom);
	AddVariable(&m_TwistRelative);
	AddVariable(&m_TwistLifetime);

/*	
	AddVariable(&m_Minwidth);
	AddVariable(&m_Minwidthend);
	AddVariable(&m_Maxwidth);
	AddVariable(&m_Maxwidthend);
	AddVariable(&m_Widthlife);
	AddVariable(&m_Minheight);
	AddVariable(&m_Minheightend);
	AddVariable(&m_Maxheight);
	AddVariable(&m_Maxheightend);
	AddVariable(&m_Heightlife);
	AddVariable(&m_Minangle);
	AddVariable(&m_Minangleend);
	AddVariable(&m_Maxangle);
	AddVariable(&m_Maxangleend);
	AddVariable(&m_Anglelife);
	AddVariable(&m_WidthRandom);
	AddVariable(&m_WidthRelative);
	AddVariable(&m_WidthLifetime);
	AddVariable(&m_HeightRandom);
	AddVariable(&m_HeightRelative);
	AddVariable(&m_HeightLifetime);
	AddVariable(&m_AngleAlign);
	AddVariable(&m_AngleRandom);
	AddVariable(&m_AngleRelative);
	AddVariable(&m_AngleLifetime);
*/
}

CParticleBurst::~CParticleBurst()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CParticleBurst::EditObject()
{
	new CParticleBurstDialog(this);
	return SDP_S_OK;
}

sdpResult CParticleBurst::ObjectDeleted()
{
	return CParticleBurstBase::ObjectDeleted();
}

sdpResult CParticleBurst::ObjectUnDeleted()
{
	return CParticleBurstBase::ObjectUnDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CParticleBurst::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	if(!m_InitializationComplete)
		return SDP_S_FALSE;

	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	const sdpString observerdata(ObserverData);

std::cout << "ObserveEvent() " << event << " - " << observerdata << std::endl;

	if(event == k3dIChannel::Event_Changed)
		{
			if(observerdata == m_RecalculateParticles)
				UpdateCalculateParticles();
			else if(observerdata == m_CreateParticles)
				UpdateCreateParticles();
			else
				std::cerr << "CParticleBurst::ObserveEvent(): Unknown observer data [" << observerdata << "]" << std::endl;
				
			k3dRedrawAll(Document(), false);
			return SDP_S_OK;
		}

	return CParticleBurstBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender implementation

sdpResult CParticleBurst::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	StoreOpenGLAttributes attributes;

	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glPointSize(2.0f);
	glDisable(GL_POINT_SMOOTH);

	const gdouble globaltime = m_Globaltime->Value1(m_CurrentTime, m_DeltaTime);

	// Setup colors ...
	const bool previewcolors = m_PreviewColors.Value();
	const bool lifetimecolors = LIFETIME == m_ColorType.Value();
	const bool randomcolors = RANDOM == m_ColorType.Value();
	
	sdpVector3 color(0, 0, 0);
	if(previewcolors)
		m_Color->Value3(globaltime, m_DeltaTime, color);
	glColor3dv(color);

	// Draw particles
	glBegin(GL_POINTS);
	const ParticleIterator end = m_Particles.end();
	for(ParticleIterator particle = m_Particles.begin(); particle != end; particle++)
		{
			// Skip dead and unborn particles ...
			if(!particle->IsAlive(globaltime))
				continue;

			if(previewcolors)
				{
					const gdouble particlesystemtime = globaltime - particle->m_Time[0];
					const gdouble particletime = particlesystemtime / (particle->m_Time[1] - particle->m_Time[0]);
							
					if(lifetimecolors)
						m_Color->Value3(particletime, m_DeltaTime, color);
					else if(randomcolors)
						color = particle->m_Color;
						
					glColor3dv(color);
				}
						
			glVertex3dv(particle->m_AnimXYZ);
		}
	glEnd();

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		DrawTransformationHandles(Engine, this, 1.0);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIRiRender implementation

sdpResult CParticleBurst::RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;
		
	double strength, particlestrength;
	double globaltime, particletime, particlesystemtime, particlesize;
	double strengthtime, sizetime;

	sdpVector3 pcolor;
	sdpVector3 opacity;

	globaltime = m_Globaltime->Value1(m_CurrentTime, m_DeltaTime);
	strength = m_Globalstrength->Value1(globaltime, m_DeltaTime);

	if(GLOBAL == m_ColorType.Value())
		{
			m_Color->Value3(globaltime, m_DeltaTime, pcolor);
		}

	ParticleIterator end = m_Particles.end();
	for(ParticleIterator particle = m_Particles.begin(); particle != end; particle++)
		{
			// Skip dead and unborn particles ...
			if(!particle->IsAlive(globaltime))
				continue;

			particlesystemtime = globaltime - particle->m_Time[0];
			particletime = particlesystemtime / (particle->m_Time[1] - particle->m_Time[0]);

			strengthtime = m_StrengthLifetime.Value() ? particletime : particlesystemtime;
			sizetime = m_SizeLifetime.Value() ? particletime : particlesystemtime;

			if(m_StrengthRelative.Value())
				particlestrength = m_Strengthlife->Value1(strengthtime, m_DeltaTime) * particle->m_Strength[1] + particle->m_Strength[0];
			else
				particlestrength = m_Strengthlife->Value1(strengthtime, m_DeltaTime) * (particle->m_Strength[1] - particle->m_Strength[0]) + particle->m_Strength[0];

			if(LIFETIME == m_ColorType.Value())
				m_Color->Value3(particletime, m_DeltaTime, pcolor);
			else if(RANDOM == m_ColorType.Value())
				pcolor = particle->m_Color;

			if(m_SizeRelative.Value())
				particlesize = m_Sizelife->Value1(sizetime, m_DeltaTime) * particle->m_Size[1] + particle->m_Size[0];
			else
				particlesize = m_Sizelife->Value1(sizetime, m_DeltaTime) * (particle->m_Size[1] - particle->m_Size[0]) + particle->m_Size[0];

			Engine->RiTransformBegin();
			Engine->RiIdentity();

			Engine->RiTranslate((float)particle->m_AnimXYZ[0], (float)particle->m_AnimXYZ[1], (float)particle->m_AnimXYZ[2]);
			Engine->RiScale((float)particlesize, (float)particlesize, (float)particlesize);

			Engine->RiColor(pcolor);

			opacity = particlestrength * strength;
			Engine->RiOpacity(opacity);

			Engine->RiSurfaceV("k3d_particleburstglow", 0, 0, 0, 0);
			Engine->RiSphereV(0.5, -0.5, 0.5, 360.0, 0, 0, 0, 0);

			Engine->RiTransformEnd();
		}

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dILocationCollection implementation

void CParticleBurst::VisitLocations(k3dILocationVisitor* const Visitor)
{
	// Sanity checks ...
	g_return_if_fail(Visitor);

//	std::for_each(m_Particles.begin(), m_Particles.end(), sdp::MakeVisitorAdapter(Visitor, &k3dILocationVisitor::VisitLocation));
	for(ParticleIterator particle = m_Particles.begin(); particle != m_Particles.end(); particle++)
		Visitor->VisitLocation(&(*particle));
}

gulong CParticleBurst::LocationCount()
{
	return m_Particles.size();
}

K3DLOCATIONHANDLE CParticleBurst::FirstLocation()
{
	return K3DLOCATIONHANDLE(m_Particles.size() ? 1 : 0);
}

K3DLOCATIONHANDLE CParticleBurst::NextLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_Particles.size(), 0);

	return K3DLOCATIONHANDLE(gulong(Location) < m_Particles.size() ? gulong(Location)+1 : 0);
}

k3dILocation* CParticleBurst::Location(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_Particles.size(), 0);

	return &m_Particles[gulong(Location)-1];
}

///////////////////////////////////////////////////////////////
// k3dIPersist implementation

sdpResult CParticleBurst::InitComplete()
{
	gdouble pi = sdp::Pi();
	gdouble pitimes2 = sdp::PiTimes2();
	if(0 == m_Globaltime.Object())
		{
			g_return_val_if_fail(m_Globaltime.Create("Particle Burst Global Time", 0.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Globaltime.Interface());
			k3dChannelSetLinearValue(m_Globaltime.Interface(), 0.0, 1.0);
		}
	if(0 == m_Birthrate.Object())
		g_return_val_if_fail(m_Birthrate.Create("Particle Burst Birthrate", 100.0, true, 0.0, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Minlife.Object())
		g_return_val_if_fail(m_Minlife.Create("Particle Burst Minimum Life", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Maxlife.Object())
		g_return_val_if_fail(m_Maxlife.Create("Particle Burst Maximum Life", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Globalstrength.Object())
		g_return_val_if_fail(m_Globalstrength.Create("Particle Burst Global Strength", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	if(0 == m_Color.Object())
		g_return_val_if_fail(m_Color.Create("Particle Burst Color", sdpVector3(1.0), true, true), SDP_E_FAIL);
		
	if(0 == m_Minstrength.Object())
		g_return_val_if_fail(m_Minstrength.Create("Particle Burst Start Strength Minimum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Minstrengthend.Object())
		g_return_val_if_fail(m_Minstrengthend.Create("Particle Burst End Strength Minimum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Maxstrength.Object())
		g_return_val_if_fail(m_Maxstrength.Create("Particle Burst Start Strength Maximum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Maxstrengthend.Object())
		g_return_val_if_fail(m_Maxstrengthend.Create("Particle Burst End Strength Maximum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Strengthlife.Object())
		{
			g_return_val_if_fail(m_Strengthlife.Create("Particle Burst Lifetime Strength", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Strengthlife.Interface());
			k3dChannelSetLinearValue(m_Strengthlife.Interface(), 0.0, 1.0);
		}

	if(0 == m_Minsize.Object())
		g_return_val_if_fail(m_Minsize.Create("Particle Burst Start Size Minimum", 0.2, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Minsizeend.Object())
		g_return_val_if_fail(m_Minsizeend.Create("Particle Burst End Size Minimum", 0.2, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Maxsize.Object())
		g_return_val_if_fail(m_Maxsize.Create("Particle Burst Start Size Maximum", 0.2, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Maxsizeend.Object())
		g_return_val_if_fail(m_Maxsizeend.Create("Particle Burst End Size Maximum", 0.2, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_Sizelife.Object())
		{
			g_return_val_if_fail(m_Sizelife.Create("Particle Burst Lifetime Size", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Sizelife.Interface());
			k3dChannelSetLinearValue(m_Sizelife.Interface(), 0.0, 1.0);
		}

	if(0 == m_Mindirectionx.Object())
		g_return_val_if_fail(m_Mindirectionx.Create("Particle Burst Start Direction X Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxdirectionx.Object())
		g_return_val_if_fail(m_Maxdirectionx.Create("Particle Burst Start Direction X Maximum", pi, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Mindirectiony.Object())
		g_return_val_if_fail(m_Mindirectiony.Create("Particle Burst Start Direction Y Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxdirectiony.Object())
		g_return_val_if_fail(m_Maxdirectiony.Create("Particle Burst Start Direction Y Maximum", pitimes2, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Mindirectionz.Object())
		g_return_val_if_fail(m_Mindirectionz.Create("Particle Burst Start Direction Z Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxdirectionz.Object())
		g_return_val_if_fail(m_Maxdirectionz.Create("Particle Burst Start Direction Z Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	
	if(0 == m_Mindistance.Object())
		g_return_val_if_fail(m_Mindistance.Create("Particle Burst Start Distance Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Mindistanceend.Object())
		g_return_val_if_fail(m_Mindistanceend.Create("Particle Burst End Distance Minimum", 10.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdistance.Object())
		g_return_val_if_fail(m_Maxdistance.Create("Particle Burst Start Distance Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdistanceend.Object())
		g_return_val_if_fail(m_Maxdistanceend.Create("Particle Burst End Distance Maximum", 10.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Distancelife.Object())
		{
			g_return_val_if_fail(m_Distancelife.Create("Particle Burst Lifetime Distance", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Distancelife.Interface());
			k3dChannelSetLinearValue(m_Distancelife.Interface(), 0.0, 1.0);
		}

	if(0 == m_Mindriftx.Object())
		g_return_val_if_fail(m_Mindriftx.Create("Particle Burst Start Drift X Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Mindriftxend.Object())
		g_return_val_if_fail(m_Mindriftxend.Create("Particle Burst End Drift X Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdriftx.Object())
		g_return_val_if_fail(m_Maxdriftx.Create("Particle Burst Start Drift X Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdriftxend.Object())
		g_return_val_if_fail(m_Maxdriftxend.Create("Particle Burst End Drift X Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Driftxlife.Object())
		{
			g_return_val_if_fail(m_Driftxlife.Create("Particle Burst Lifetime Drift X", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Driftxlife.Interface());
			k3dChannelSetLinearValue(m_Driftxlife.Interface(), 0.0, 1.0);
		}
	if(0 == m_Mindrifty.Object())
		g_return_val_if_fail(m_Mindrifty.Create("Particle Burst Start Drift Y Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Mindriftyend.Object())
		g_return_val_if_fail(m_Mindriftyend.Create("Particle Burst End Drift Y Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdrifty.Object())
		g_return_val_if_fail(m_Maxdrifty.Create("Particle Burst Start Drift Y Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdriftyend.Object())
		g_return_val_if_fail(m_Maxdriftyend.Create("Particle Burst End Drift Y Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Driftylife.Object())
		{
			g_return_val_if_fail(m_Driftylife.Create("Particle Burst Lifetime Drift Y", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Driftylife.Interface());
			k3dChannelSetLinearValue(m_Driftylife.Interface(), 0.0, 1.0);
		}
	if(0 == m_Mindriftz.Object())
		g_return_val_if_fail(m_Mindriftz.Create("Particle Burst Start Drift Z Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Mindriftzend.Object())
		g_return_val_if_fail(m_Mindriftzend.Create("Particle Burst End Drift Z Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdriftz.Object())
		g_return_val_if_fail(m_Maxdriftz.Create("Particle Burst Start Drift Z Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxdriftzend.Object())
		g_return_val_if_fail(m_Maxdriftzend.Create("Particle Burst End Drift Z Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Driftzlife.Object())
		{
			g_return_val_if_fail(m_Driftzlife.Create("Particle Burst Lifetime Drift Z", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Driftzlife.Interface());
			k3dChannelSetLinearValue(m_Driftzlife.Interface(), 0.0, 1.0);
		}

	if(0 == m_Mintwistx.Object())
		g_return_val_if_fail(m_Mintwistx.Create("Particle Burst Start Twist X Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Mintwistxend.Object())
		g_return_val_if_fail(m_Mintwistxend.Create("Particle Burst End Twist X Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxtwistx.Object())
		g_return_val_if_fail(m_Maxtwistx.Create("Particle Burst Start Twist X Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxtwistxend.Object())
		g_return_val_if_fail(m_Maxtwistxend.Create("Particle Burst End Twist X Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Twistxlife.Object())
		{
			g_return_val_if_fail(m_Twistxlife.Create("Particle Burst Lifetime Twist X", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Twistxlife.Interface());
			k3dChannelSetLinearValue(m_Twistxlife.Interface(), 0.0, 1.0);
		}
	if(0 == m_Mintwisty.Object())
		g_return_val_if_fail(m_Mintwisty.Create("Particle Burst Start Twist Y Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Mintwistyend.Object())
		g_return_val_if_fail(m_Mintwistyend.Create("Particle Burst End Twist Y Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxtwisty.Object())
		g_return_val_if_fail(m_Maxtwisty.Create("Particle Burst Start Twist Y Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxtwistyend.Object())
		g_return_val_if_fail(m_Maxtwistyend.Create("Particle Burst End Twist Y Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Twistylife.Object())
		{
			g_return_val_if_fail(m_Twistylife.Create("Particle Burst Lifetime Twist Y", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Twistylife.Interface());
			k3dChannelSetLinearValue(m_Twistylife.Interface(), 0.0, 1.0);
		}
	if(0 == m_Mintwistz.Object())
		g_return_val_if_fail(m_Mintwistz.Create("Particle Burst Start Twist Z Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Mintwistzend.Object())
		g_return_val_if_fail(m_Mintwistzend.Create("Particle Burst End Twist Z Minimum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxtwistz.Object())
		g_return_val_if_fail(m_Maxtwistz.Create("Particle Burst Start Twist Z Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxtwistzend.Object())
		g_return_val_if_fail(m_Maxtwistzend.Create("Particle Burst End Twist Z Maximum", 0.0, true, -1.0 * DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Twistzlife.Object())
		{
			g_return_val_if_fail(m_Twistzlife.Create("Particle Burst Lifetime Twist Z", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Twistzlife.Interface());
			k3dChannelSetLinearValue(m_Twistzlife.Interface(), 0.0, 1.0);
		}

/*
	if(0 == m_Minwidth.Object())
		g_return_val_if_fail(m_Minwidth.Create("Particle Burst Start Width Minimum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Minwidthend.Object())
		g_return_val_if_fail(m_Minwidthend.Create("Particle Burst End Width Minimum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxwidth.Object())
		g_return_val_if_fail(m_Maxwidth.Create("Particle Burst Start Width Maximum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxwidthend.Object())
		g_return_val_if_fail(m_Maxwidthend.Create("Particle Burst End Width Maximum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);

	if(0 == m_Widthlife.Object())
		{
			g_return_val_if_fail(m_Widthlife.Create("Particle Burst Lifetime Width", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Widthlife.Interface());
			k3dChannelSetLinearValue(m_Widthlife.Interface(), 0.0, 1.0);
		}

	if(0 == m_Minheight.Object())
		g_return_val_if_fail(m_Minheight.Create("Particle Burst Start Height Minimum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Minheightend.Object())
		g_return_val_if_fail(m_Minheightend.Create("Particle Burst End Height Minimum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxheight.Object())
		g_return_val_if_fail(m_Maxheight.Create("Particle Burst Start Height Maximum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Maxheightend.Object())
		g_return_val_if_fail(m_Maxheightend.Create("Particle Burst End Height Maximum", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);

	if(0 == m_Heightlife.Object())
		{
			g_return_val_if_fail(m_Heightlife.Create("Particle Burst Lifetime Height", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Heightlife.Interface());
			k3dChannelSetLinearValue(m_Heightlife.Interface(), 0.0, 1.0);
		}

	if(0 == m_Minangle.Object())
		g_return_val_if_fail(m_Minangle.Create("Particle Burst Start Angle Minimum", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Minangleend.Object())
		g_return_val_if_fail(m_Minangleend.Create("Particle Burst End Angle Minimum", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxangle.Object())
		g_return_val_if_fail(m_Maxangle.Create("Particle Burst Start Angle Maximum", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Maxangleend.Object())
		g_return_val_if_fail(m_Maxangleend.Create("Particle Burst End Angle Maximum", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);

	if(0 == m_Anglelife.Object())
		{
			g_return_val_if_fail(m_Anglelife.Create("Particle Burst Lifetime Angle", 1.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Anglelife.Interface());
			k3dChannelSetLinearValue(m_Anglelife.Interface(), 0.0, 1.0);
		}
*/

	// Force calculation of the particle system
	UpdateFull();

	m_InitializationComplete = true;

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate implementation

sdpResult CParticleBurst::PopAnimate()
{
	// If animation's been disabled, we're done ...
	if(!m_Animate)
		return SDP_S_OK;

	gdouble particlesystemtime, particletime;
	gdouble twisttime, distancetime, drifttime;
	sdpVector3 direction, twist;

	gdouble px, py;
	gdouble cosx, cosy, cosz;
	gdouble sinx, siny, sinz;

	// Calculate Global time
	gdouble globaltime = m_Globaltime->Value1(m_CurrentTime, m_DeltaTime);

	// For each particle ...
	ParticleIterator end = m_Particles.end();
	for(ParticleIterator particle = m_Particles.begin(); particle != end; particle++)
		{
			// Ignore dead and unborn particles
			if(particle->m_Time[0] > globaltime || particle->m_Time[1] < globaltime) continue;

			// Calculate particle times
			particlesystemtime = globaltime - particle->m_Time[0];
			particletime = particlesystemtime / (particle->m_Time[1] - particle->m_Time[0]);
			twisttime = m_TwistLifetime.Value() ? particletime : particlesystemtime;
			distancetime = m_DistanceLifetime.Value() ? particletime : particlesystemtime;
			drifttime = m_DriftLifetime.Value() ? particletime : particlesystemtime;

			// Calculate particle position
			particle->m_AnimXYZ = 0.0;

			if(m_DistanceRelative.Value())
				{
					particle->m_AnimXYZ[1] = particle->m_Distance[0] + m_Distancelife->Value1(distancetime, m_DeltaTime) * particle->m_Distance[1];
				}
			else
				{
					particle->m_AnimXYZ[1] = particle->m_Distance[0] + m_Distancelife->Value1(distancetime, m_DeltaTime) * (particle->m_Distance[1] - particle->m_Distance[0]);
				}

			direction = particle->m_DirectionXYZ;

			cosx = cos(direction[0]);
			sinx = sin(direction[0]);
			cosy = cos(direction[1]);
			siny = sin(direction[1]);
			cosz = cos(direction[2]);
			sinz = sin(direction[2]);

			py = particle->m_AnimXYZ[1];
			particle->m_AnimXYZ[1] = (cosx * particle->m_AnimXYZ[1]) - (sinx * particle->m_AnimXYZ[2]);
			particle->m_AnimXYZ[2] = (sinx * py) + (cosx * particle->m_AnimXYZ[2]);

			px = particle->m_AnimXYZ[0];
			particle->m_AnimXYZ[0] = (cosy * particle->m_AnimXYZ[0]) + (siny * particle->m_AnimXYZ[2]);
			particle->m_AnimXYZ[2] = (cosy * particle->m_AnimXYZ[2]) - (siny * px);

			px = particle->m_AnimXYZ[0];
			particle->m_AnimXYZ[0] = (cosz * particle->m_AnimXYZ[0]) - (sinz * particle->m_AnimXYZ[1]);
			particle->m_AnimXYZ[1] = (sinz * px) + (cosz * particle->m_AnimXYZ[1]);

			if(m_TwistRelative.Value())
				{
					twist[0] = particle->m_TwistXYZ[0][0] + m_Twistxlife->Value1(twisttime, m_DeltaTime) * particle->m_TwistXYZ[1][0];
					twist[1] = particle->m_TwistXYZ[0][1] + m_Twistylife->Value1(twisttime, m_DeltaTime) * particle->m_TwistXYZ[1][1];
					twist[2] = particle->m_TwistXYZ[0][2] + m_Twistzlife->Value1(twisttime, m_DeltaTime) * particle->m_TwistXYZ[1][2];
				}
			else
				{
					twist[0] = particle->m_TwistXYZ[0][0] + m_Twistxlife->Value1(twisttime, m_DeltaTime) * (particle->m_TwistXYZ[1][0] - particle->m_TwistXYZ[0][0]);
					twist[1] = particle->m_TwistXYZ[0][1] + m_Twistylife->Value1(twisttime, m_DeltaTime) * (particle->m_TwistXYZ[1][1] - particle->m_TwistXYZ[0][1]);
					twist[2] = particle->m_TwistXYZ[0][2] + m_Twistzlife->Value1(twisttime, m_DeltaTime) * (particle->m_TwistXYZ[1][2] - particle->m_TwistXYZ[0][2]);
				}

			cosx = cos(twist[0]);
			sinx = sin(twist[0]);
			cosy = cos(twist[1]);
			siny = sin(twist[1]);
			cosz = cos(twist[2]);
			sinz = sin(twist[2]);

			py = particle->m_AnimXYZ[1];
			particle->m_AnimXYZ[1] = (cosx * particle->m_AnimXYZ[1]) - (sinx * particle->m_AnimXYZ[2]);
			particle->m_AnimXYZ[2] = (sinx * py) + (cosx * particle->m_AnimXYZ[2]);

			px = particle->m_AnimXYZ[0];
			particle->m_AnimXYZ[0] = (cosy * particle->m_AnimXYZ[0]) + (siny * particle->m_AnimXYZ[2]);
			particle->m_AnimXYZ[2] = (cosy * particle->m_AnimXYZ[2]) - (siny * px);

			px = particle->m_AnimXYZ[0];
			particle->m_AnimXYZ[0] = (cosz * particle->m_AnimXYZ[0]) - (sinz * particle->m_AnimXYZ[1]);
			particle->m_AnimXYZ[1] = (sinz * px) + (cosz * particle->m_AnimXYZ[1]);


			if(m_DriftRelative.Value())
				{
					particle->m_AnimXYZ[0] += particle->m_DriftXYZ[0][0] + m_Driftxlife->Value1(drifttime, m_DeltaTime) * particle->m_DriftXYZ[1][0];
					particle->m_AnimXYZ[1] += particle->m_DriftXYZ[0][1] + m_Driftylife->Value1(drifttime, m_DeltaTime) * particle->m_DriftXYZ[1][1];
					particle->m_AnimXYZ[2] += particle->m_DriftXYZ[0][2] + m_Driftzlife->Value1(drifttime, m_DeltaTime) * particle->m_DriftXYZ[1][2];
				}
			else
				{
					particle->m_AnimXYZ[0] += particle->m_DriftXYZ[0][0] + m_Driftxlife->Value1(drifttime, m_DeltaTime) * (particle->m_DriftXYZ[1][0] - particle->m_DriftXYZ[0][0]);
					particle->m_AnimXYZ[1] += particle->m_DriftXYZ[0][1] + m_Driftylife->Value1(drifttime, m_DeltaTime) * (particle->m_DriftXYZ[1][1] - particle->m_DriftXYZ[0][1]);
					particle->m_AnimXYZ[2] += particle->m_DriftXYZ[0][2] + m_Driftzlife->Value1(drifttime, m_DeltaTime) * (particle->m_DriftXYZ[1][2] - particle->m_DriftXYZ[0][2]);
				}

			particle->m_AnimXYZ += particle->m_BirthXYZ;
		}

	return SDP_S_OK;
}

void CParticleBurst::CreateParticles()
{
//std::cout << "CParticleBurst::CreateParticles()" << std::endl;

	// Start over from scratch
	m_Particles.clear();

	gdouble totalframes = m_Time->Duration() * m_Time->FrameRate();
	gdouble dsample = (1.0 / m_BirthRateSamples.Value()) / totalframes;

	gdouble particlequantity = 0.0;
	for(gdouble j = dsample; j < 1.0; j += dsample)
		{
			gdouble pcount = m_Birthrate->Value1(j, dsample * 0.1) * dsample;
			particlequantity += pcount;

			gdouble i = 0.0;
			for(; particlequantity >= 1.0; particlequantity -= 1.0)
				{
					CParticle particle;
					if(pcount)
						{
							particle.m_Time[0] = j + (i++) * (dsample / pcount);
						}
					else
						{
							particle.m_Time[0] = j;
						}
					m_Particles.push_back(particle);
				}
		}
}

void CParticleBurst::RecalculateParticles()
{
//std::cout << "CParticleBurst::RecalculateParticles()" << std::endl;

	double minlife, maxlife, dlife;
	double minstrength, minstrengthend, maxstrength, maxstrengthend, dstrength, dstrengthend;
	double minsize, minsizeend, maxsize, maxsizeend, dsize, dsizeend;
	double mindistance, mindistanceend, maxdistance, maxdistanceend, ddistance, ddistanceend;
	double mindriftx, mindriftxend, maxdriftx, maxdriftxend, ddriftx, ddriftxend;
	double mindrifty, mindriftyend, maxdrifty, maxdriftyend, ddrifty, ddriftyend;
	double mindriftz, mindriftzend, maxdriftz, maxdriftzend, ddriftz, ddriftzend;
	double mindirectionx, maxdirectionx, ddirectionx;
	double mindirectiony, maxdirectiony, ddirectiony;
	double mindirectionz, maxdirectionz, ddirectionz;
	double mintwistx, mintwistxend, maxtwistx, maxtwistxend, dtwistx, dtwistxend;
	double mintwisty, mintwistyend, maxtwisty, maxtwistyend, dtwisty, dtwistyend;
	double mintwistz, mintwistzend, maxtwistz, maxtwistzend, dtwistz, dtwistzend;
//	double minwidth, minwidthend, maxwidth, maxwidthend, dwidth, dwidthend;
//	double minheight, minheightend, maxheight, maxheightend, dheight, dheightend;
//	double minangle, minangleend, maxangle, maxangleend, dangle, dangleend;

	// Reset the seed
	gulong seed = static_cast<gulong>(m_RandomSeed.Value());

	// For each particle ...
	gdouble index = 0.0;
	const gdouble particlecount = static_cast<gdouble>(m_Particles.size());
	ParticleIterator end = m_Particles.end();
	for(ParticleIterator particle = m_Particles.begin(); particle != end; particle++)
		{
			const gdouble fraction =  index / particlecount;

			// Find min / max values
			minlife = m_Minlife->Value1(fraction, m_DeltaTime);
			maxlife = m_Maxlife->Value1(fraction, m_DeltaTime);
			dlife = maxlife - minlife;
			minstrength = m_Minstrength->Value1(fraction, m_DeltaTime);
			maxstrength = m_Maxstrength->Value1(fraction, m_DeltaTime);
			dstrength = maxstrength - minstrength;
			minstrengthend = m_Minstrengthend->Value1(fraction, m_DeltaTime);
			maxstrengthend = m_Maxstrengthend->Value1(fraction, m_DeltaTime);
			dstrengthend = maxstrengthend - minstrengthend;
			minsize = m_Minsize->Value1(fraction, m_DeltaTime);
			maxsize = m_Maxsize->Value1(fraction, m_DeltaTime);
			dsize = maxsize - minsize;
			minsizeend = m_Minsizeend->Value1(fraction, m_DeltaTime);
			maxsizeend = m_Maxsizeend->Value1(fraction, m_DeltaTime);
			dsizeend = maxsizeend - minsizeend;
			mindistance = m_Mindistance->Value1(fraction, m_DeltaTime);
			maxdistance = m_Maxdistance->Value1(fraction, m_DeltaTime);
			ddistance = maxdistance - mindistance;
			mindistanceend = m_Mindistanceend->Value1(fraction, m_DeltaTime);
			maxdistanceend = m_Maxdistanceend->Value1(fraction, m_DeltaTime);
			ddistanceend = maxdistanceend - mindistanceend;
			mindriftx = m_Mindriftx->Value1(fraction, m_DeltaTime);
			maxdriftx = m_Maxdriftx->Value1(fraction, m_DeltaTime);
			ddriftx = maxdriftx - mindriftx;
			mindriftxend = m_Mindriftxend->Value1(fraction, m_DeltaTime);
			maxdriftxend = m_Maxdriftxend->Value1(fraction, m_DeltaTime);
			ddriftxend = maxdriftxend - mindriftxend;
			mindrifty = m_Mindrifty->Value1(fraction, m_DeltaTime);
			maxdrifty = m_Maxdrifty->Value1(fraction, m_DeltaTime);
			ddrifty = maxdrifty - mindrifty;
			mindriftyend = m_Mindriftyend->Value1(fraction, m_DeltaTime);
			maxdriftyend = m_Maxdriftyend->Value1(fraction, m_DeltaTime);
			ddriftyend = maxdriftyend - mindriftyend;
			mindriftz = m_Mindriftz->Value1(fraction, m_DeltaTime);
			maxdriftz = m_Maxdriftz->Value1(fraction, m_DeltaTime);
			ddriftz = maxdriftz - mindriftz;
			mindriftzend = m_Mindriftzend->Value1(fraction, m_DeltaTime);
			maxdriftzend = m_Maxdriftzend->Value1(fraction, m_DeltaTime);
			ddriftzend = maxdriftzend - mindriftzend;
			mindirectionx = m_Mindirectionx->Value1(fraction, m_DeltaTime);
			maxdirectionx = m_Maxdirectionx->Value1(fraction, m_DeltaTime);
			ddirectionx = maxdirectionx - mindirectionx;
			mindirectiony = m_Mindirectiony->Value1(fraction, m_DeltaTime);
			maxdirectiony = m_Maxdirectiony->Value1(fraction, m_DeltaTime);
			ddirectiony = maxdirectiony - mindirectiony;
			mindirectionz = m_Mindirectionz->Value1(fraction, m_DeltaTime);
			maxdirectionz = m_Maxdirectionz->Value1(fraction, m_DeltaTime);
			ddirectionz = maxdirectionz - mindirectionz;
			mintwistx = m_Mintwistx->Value1(fraction, m_DeltaTime);
			maxtwistx = m_Maxtwistx->Value1(fraction, m_DeltaTime);
			dtwistx = maxtwistx - mintwistx;
			mintwistxend = m_Mintwistxend->Value1(fraction, m_DeltaTime);
			maxtwistxend = m_Maxtwistxend->Value1(fraction, m_DeltaTime);
			dtwistxend = maxtwistxend - mintwistxend;
			mintwisty = m_Mintwisty->Value1(fraction, m_DeltaTime);
			maxtwisty = m_Maxtwisty->Value1(fraction, m_DeltaTime);
			dtwisty = maxtwisty - mintwisty;
			mintwistyend = m_Mintwistyend->Value1(fraction, m_DeltaTime);
			maxtwistyend = m_Maxtwistyend->Value1(fraction, m_DeltaTime);
			dtwistyend = maxtwistyend - mintwistyend;
			mintwistz = m_Mintwistz->Value1(fraction, m_DeltaTime);
			maxtwistz = m_Maxtwistz->Value1(fraction, m_DeltaTime);
			dtwistz = maxtwistz - mintwistz;
			mintwistzend = m_Mintwistzend->Value1(fraction, m_DeltaTime);
			maxtwistzend = m_Maxtwistzend->Value1(fraction, m_DeltaTime);
			dtwistzend = maxtwistzend - mintwistzend;

/*			
			minwidth = m_Minwidth->Value1(fraction, m_DeltaTime);
			maxwidth = m_Maxwidth->Value1(fraction, m_DeltaTime);
			dwidth = maxwidth - minwidth;
			minwidthend = m_Minwidthend->Value1(fraction, m_DeltaTime);
			maxwidthend = m_Maxwidthend->Value1(fraction, m_DeltaTime);
			dwidthend = maxwidthend - minwidthend;
			minheight = m_Minheight->Value1(fraction, m_DeltaTime);
			maxheight = m_Maxheight->Value1(fraction, m_DeltaTime);
			dheight = maxheight - minheight;
			minheightend = m_Minheightend->Value1(fraction, m_DeltaTime);
			maxheightend = m_Maxheightend->Value1(fraction, m_DeltaTime);
			dheightend = maxheightend - minheightend;
			minangle = m_Minangle->Value1(fraction, m_DeltaTime);
			maxangle = m_Maxangle->Value1(fraction, m_DeltaTime);
			dangle = maxangle - minangle;
			minangleend = m_Minangleend->Value1(fraction, m_DeltaTime);
			maxangleend = m_Maxangleend->Value1(fraction, m_DeltaTime);
			dangleend = maxangleend - minangleend;
*/

			// Calculate particle values
			particle->m_Time[1] = particle->m_Time[0] + dlife * (sdpRandom(seed)) + minlife;

			if(m_DistanceRandom.Value())
				{
					particle->m_Distance[0] = ddistance * (sdpRandom(seed)) + mindistance;
					particle->m_Distance[1] = ddistanceend * (sdpRandom(seed)) + mindistanceend;
				}
			else
				{
					particle->m_Distance[0] = ddistance * fraction + mindistance;
					particle->m_Distance[1] = ddistanceend * fraction + mindistanceend;
				}

			if(m_DriftRandom.Value())
				{
					particle->m_DriftXYZ[0][0] = ddriftx * (sdpRandom(seed)) + mindriftx;
					particle->m_DriftXYZ[1][0] = ddriftxend * (sdpRandom(seed)) + mindriftxend;
					particle->m_DriftXYZ[0][1] = ddrifty * (sdpRandom(seed)) + mindrifty;
					particle->m_DriftXYZ[1][1] = ddriftyend * (sdpRandom(seed)) + mindriftyend;
					particle->m_DriftXYZ[0][2] = ddriftz * (sdpRandom(seed)) + mindriftz;
					particle->m_DriftXYZ[1][2] = ddriftzend * (sdpRandom(seed)) + mindriftzend;
				}
			else
				{
					particle->m_DriftXYZ[0][0] = ddriftx * fraction + mindriftx;
					particle->m_DriftXYZ[1][0] = ddriftxend * fraction + mindriftxend;
					particle->m_DriftXYZ[0][1] = ddrifty * fraction + mindrifty;
					particle->m_DriftXYZ[1][1] = ddriftyend * fraction + mindriftyend;
					particle->m_DriftXYZ[0][2] = ddriftz * fraction + mindriftz;
					particle->m_DriftXYZ[1][2] = ddriftzend * fraction + mindriftzend;
				}

			if(m_DirectionRandom.Value())
				{
					particle->m_DirectionXYZ[0] = ddirectionx * (sdpRandom(seed)) + mindirectionx;
					particle->m_DirectionXYZ[1] = ddirectiony * (sdpRandom(seed)) + mindirectiony;
					particle->m_DirectionXYZ[2] = ddirectionz * (sdpRandom(seed)) + mindirectionz;
				}
			else
				{
					particle->m_DirectionXYZ[0] = ddirectionx * fraction + mindirectionx;
					particle->m_DirectionXYZ[1] = ddirectiony * fraction + mindirectiony;
					particle->m_DirectionXYZ[2] = ddirectionz * fraction + mindirectionz;
				}

			if(m_TwistRandom.Value())
				{
					particle->m_TwistXYZ[0][0] = dtwistx * (sdpRandom(seed)) + mintwistx;
					particle->m_TwistXYZ[1][0] = dtwistxend * (sdpRandom(seed)) + mintwistxend;
					particle->m_TwistXYZ[0][1] = dtwisty * (sdpRandom(seed)) + mintwisty;
					particle->m_TwistXYZ[1][1] = dtwistyend * (sdpRandom(seed)) + mintwistyend;
					particle->m_TwistXYZ[0][2] = dtwistz * (sdpRandom(seed)) + mintwistz;
					particle->m_TwistXYZ[1][2] = dtwistzend * (sdpRandom(seed)) + mintwistzend;
				}
			else
				{
					particle->m_TwistXYZ[0][0] = dtwistx * fraction + mintwistx;
					particle->m_TwistXYZ[1][0] = dtwistxend * fraction + mintwistxend;
					particle->m_TwistXYZ[0][1] = dtwisty * fraction + mintwisty;
					particle->m_TwistXYZ[1][1] = dtwistyend * fraction + mintwistyend;
					particle->m_TwistXYZ[0][2] = dtwistz * fraction + mintwistz;
					particle->m_TwistXYZ[1][2] = dtwistzend * fraction + mintwistzend;
				}

			if(m_StrengthRandom.Value())
				{
					particle->m_Strength[0] = dstrength * (sdpRandom(seed)) + minstrength;
					particle->m_Strength[1] = dstrengthend * (sdpRandom(seed)) + minstrengthend;
				}
			else
				{
					particle->m_Strength[0] = dstrength * fraction + minstrength;
					particle->m_Strength[1] = dstrengthend * fraction + minstrengthend;
				}

			if(m_SizeRandom.Value())
				{
					particle->m_Size[0] = dsize * (sdpRandom(seed)) + minsize;
					particle->m_Size[1] = dsizeend * (sdpRandom(seed)) + minsizeend;
				}
			else
				{
					particle->m_Size[0] = dsize * fraction + minsize;
					particle->m_Size[1] = dsizeend * fraction + minsizeend;
				}

/*
			if(m_WidthRandom.Value())
				{
					particle->m_Width[0] = dwidth * (sdpRandom(seed)) + minwidth;
					particle->m_Width[1] = dwidthend * (sdpRandom(seed)) + minwidthend;
				}
			else
				{
					particle->m_Width[0] = dwidth * fraction + minwidth;
					particle->m_Width[1] = dwidthend * fraction + minwidthend;
				}

			if(m_HeightRandom.Value())
				{
					particle->m_Height[0] = dheight * (sdpRandom(seed)) + minheight;
					particle->m_Height[1] = dheightend * (sdpRandom(seed)) + minheightend;
				}
			else
				{
					particle->m_Height[0] = dheight * fraction + minheight;
					particle->m_Height[1] = dheightend * fraction + minheightend;
				}

			if(m_AngleRandom.Value())
				{
					particle->m_Angle[0] = dangle * (sdpRandom(seed)) + minangle;
					particle->m_Angle[1] = dangleend * (sdpRandom(seed)) + minangleend;
				}
			else
				{
					particle->m_Angle[0] = dangle * fraction + minangle;
					particle->m_Angle[1] = dangleend * fraction + minangleend;
				}
*/

			m_Color->Value3(sdpRandom(seed), m_DeltaTime, particle->m_Color);

			index++;
		}
}

void CParticleBurst::RecalculateEmitterPosition()
{
//std::cout << "CParticleBurst::RecalculateEmitterPosition()" << std::endl;

	// Disable k3dIAnimate calculations
	m_Animate = false;

	double totalframes, dsample, j;

	sdpVector3 emittercoords;

	if(0.0 == m_EmitterPositionSamples.Value()) // Calculate particle positions exactly at birth for each particle
		{
			for(ParticleIterator particle = m_Particles.begin(); particle != m_Particles.end(); particle++)
				{
					m_Time->SetRenderTime(particle->m_Time[0]);
					Document()->Hierarchy()->Animate(particle->m_Time[0], 1.0);

//					m_Emitter->GetTargetCoordinates(emittercoords);

					particle->m_BirthXYZ = emittercoords;
				}
		}
	else // Calculate particle positions at some multiple of frame rate
		{
			totalframes = m_Time->Duration() * m_Time->FrameRate();
			dsample = (1.0 / m_EmitterPositionSamples.Value()) / (double)totalframes;

			ParticleIterator particle = m_Particles.begin();
			for(j = dsample; j < 1.0; j += dsample)
				{
					m_Time->SetRenderTime(j - dsample);
					Document()->Hierarchy()->Animate(m_Time->SetRenderTime(j - dsample), 1.0);

//					m_Emitter->GetTargetCoordinates(emittercoords);

					while(particle != m_Particles.end())
						{
							if(particle->m_Time[0] < j)
								{
									particle->m_BirthXYZ = emittercoords;
								}
							else
								{
									break;
								}

							particle++;
						}

					if(particle == m_Particles.end())
						break;
				}
		}

	// Enable k3dIAnimate calculations
	m_Animate = true;
}

void CParticleBurst::UpdateFull()
{
//std::cout << "CParticleBurst::UpdateFull()" << std::endl;

	CreateParticles();
	RecalculateParticles();
	RecalculateEmitterPosition();
}

void CParticleBurst::UpdateCalculateParticles()
{
//std::cout << "CParticleBurst::UpdateCalculateParticles()" << std::endl;

	RecalculateParticles();
}

void CParticleBurst::UpdateCreateParticles()
{
//std::cout << "CParticleBurst::UpdateCreateParticles()" << std::endl;

	CreateParticles();
	RecalculateParticles();
}

/////////////////////////////////////////////////////////////////////////////
// CParticleBurstDialog

const sdpString CParticleBurstDialog::dialogname = "properties";
const sdpString CParticleBurstDialog::control_position = "position";
const sdpString CParticleBurstDialog::control_orientation = "orientation";
const sdpString CParticleBurstDialog::control_scale = "scale";
const sdpString CParticleBurstDialog::control_target = "target";

CParticleBurstDialog::CParticleBurstDialog(CParticleBurst* const Object) :
	CParticleBurstDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("sfxlib", "particleburst.gtkml"));

	// Lifetime controls ...
	Connectk3dChannelButton("globaltime", m_Object->m_Globaltime.Interface());
	Connectk3dChannelButton("birthrate", m_Object->m_Birthrate.Interface());
	Connectk3dChannelButton("minlife", m_Object->m_Minlife.Interface());
	Connectk3dChannelButton("maxlife", m_Object->m_Maxlife.Interface());

	// Color controls ...
	if(Getk3dChooser("colortype"))
		Getk3dChooser("colortype")->Attach(
				new k3dChooserVariableDataSource<CParticleBurst::ColorType>(m_Object->m_ColorType),
				k3dApplication()->CommandRecorder(),
				0,
				"Color Type",
				0,
				new k3dActionRedrawAll(m_Object->Document(), false, 0),
				0);
	
	Connectk3dChannelButton("color", m_Object->m_Color.Interface());
	Connectk3dCheckButton("previewcolors", m_Object->m_PreviewColors, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Preview Colors", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Strength buttons
	Connectk3dChannelButton("minstrength", m_Object->m_Minstrength.Interface());
	Connectk3dChannelButton("minstrengthend", m_Object->m_Minstrengthend.Interface());
	Connectk3dChannelButton("maxstrength", m_Object->m_Maxstrength.Interface());
	Connectk3dChannelButton("maxstrengthend", m_Object->m_Maxstrengthend.Interface());
	Connectk3dChannelButton("strengthlife", m_Object->m_Strengthlife.Interface());
	Connectk3dCheckButton("strengthrandom", m_Object->m_StrengthRandom, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Random Strength", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("strengthrelative", m_Object->m_StrengthRelative, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Relative Strength", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("strengthlifetime", m_Object->m_StrengthLifetime, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lifetime Strength", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dChannelButton("globalstrength", m_Object->m_Globalstrength.Interface());

	// Size buttons
	Connectk3dChannelButton("minsize", m_Object->m_Minsize.Interface());
	Connectk3dChannelButton("minsizeend", m_Object->m_Minsizeend.Interface());
	Connectk3dChannelButton("maxsize", m_Object->m_Maxsize.Interface());
	Connectk3dChannelButton("maxsizeend", m_Object->m_Maxsizeend.Interface());
	Connectk3dChannelButton("sizelife", m_Object->m_Sizelife.Interface());
	Connectk3dCheckButton("sizerandom", m_Object->m_SizeRandom, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Random Size", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("sizerelative", m_Object->m_SizeRelative, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Relative Size", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("sizelifetime", m_Object->m_SizeLifetime, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lifetime Size", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Direction buttons
	Connectk3dChannelButton("mindirectionx", m_Object->m_Mindirectionx.Interface());
	Connectk3dChannelButton("maxdirectionx", m_Object->m_Maxdirectionx.Interface());
	Connectk3dChannelButton("mindirectiony", m_Object->m_Mindirectiony.Interface());
	Connectk3dChannelButton("maxdirectiony", m_Object->m_Maxdirectiony.Interface());
	Connectk3dChannelButton("mindirectionz", m_Object->m_Mindirectionz.Interface());
	Connectk3dChannelButton("maxdirectionz", m_Object->m_Maxdirectionz.Interface());
	Connectk3dCheckButton("directionrandom", m_Object->m_DirectionRandom, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Random Direction", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Distance buttons
	Connectk3dChannelButton("mindistance", m_Object->m_Mindistance.Interface());
	Connectk3dChannelButton("mindistanceend", m_Object->m_Mindistanceend.Interface());
	Connectk3dChannelButton("maxdistance", m_Object->m_Maxdistance.Interface());
	Connectk3dChannelButton("maxdistanceend", m_Object->m_Maxdistanceend.Interface());
	Connectk3dChannelButton("distancelife", m_Object->m_Distancelife.Interface());
	Connectk3dCheckButton("distancerandom", m_Object->m_DistanceRandom, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Random Distance", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("distancerelative", m_Object->m_DistanceRelative, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Relative Distance", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("distancelifetime", m_Object->m_DistanceLifetime, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lifetime Distance", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup drift buttons ...
	Connectk3dChannelButton("mindriftx", m_Object->m_Mindriftx.Interface());
	Connectk3dChannelButton("mindriftxend", m_Object->m_Mindriftxend.Interface());
	Connectk3dChannelButton("maxdriftx", m_Object->m_Maxdriftx.Interface());
	Connectk3dChannelButton("maxdriftxend", m_Object->m_Maxdriftxend.Interface());
	Connectk3dChannelButton("mindrifty", m_Object->m_Mindrifty.Interface());
	Connectk3dChannelButton("mindriftyend", m_Object->m_Mindriftyend.Interface());
	Connectk3dChannelButton("maxdrifty", m_Object->m_Maxdrifty.Interface());
	Connectk3dChannelButton("maxdriftyend", m_Object->m_Maxdriftyend.Interface());
	Connectk3dChannelButton("mindriftz", m_Object->m_Mindriftz.Interface());
	Connectk3dChannelButton("mindriftzend", m_Object->m_Mindriftzend.Interface());
	Connectk3dChannelButton("maxdriftz", m_Object->m_Maxdriftz.Interface());
	Connectk3dChannelButton("maxdriftzend", m_Object->m_Maxdriftzend.Interface());
	Connectk3dChannelButton("driftxlife", m_Object->m_Driftxlife.Interface());
	Connectk3dChannelButton("driftylife", m_Object->m_Driftylife.Interface());
	Connectk3dChannelButton("driftzlife", m_Object->m_Driftzlife.Interface());
	Connectk3dCheckButton("driftrandom", m_Object->m_DriftRandom, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Random Drift", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("driftrelative", m_Object->m_DriftRelative, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Relative Drift", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("driftlifetime", m_Object->m_DriftLifetime, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lifetime Drift", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup twist buttons ...
	Connectk3dChannelButton("mintwistx", m_Object->m_Mintwistx.Interface());
	Connectk3dChannelButton("mintwistxend", m_Object->m_Mintwistxend.Interface());
	Connectk3dChannelButton("maxtwistx", m_Object->m_Maxtwistx.Interface());
	Connectk3dChannelButton("maxtwistxend", m_Object->m_Maxtwistxend.Interface());
	Connectk3dChannelButton("mintwisty", m_Object->m_Mintwisty.Interface());
	Connectk3dChannelButton("mintwistyend", m_Object->m_Mintwistyend.Interface());
	Connectk3dChannelButton("maxtwisty", m_Object->m_Maxtwisty.Interface());
	Connectk3dChannelButton("maxtwistyend", m_Object->m_Maxtwistyend.Interface());
	Connectk3dChannelButton("mintwistz", m_Object->m_Mintwistz.Interface());
	Connectk3dChannelButton("mintwistzend", m_Object->m_Mintwistzend.Interface());
	Connectk3dChannelButton("maxtwistz", m_Object->m_Maxtwistz.Interface());
	Connectk3dChannelButton("maxtwistzend", m_Object->m_Maxtwistzend.Interface());
	Connectk3dChannelButton("twistxlife", m_Object->m_Twistxlife.Interface());
	Connectk3dChannelButton("twistylife", m_Object->m_Twistylife.Interface());
	Connectk3dChannelButton("twistzlife", m_Object->m_Twistzlife.Interface());
	Connectk3dCheckButton("twistrandom", m_Object->m_TwistRandom, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Twist Strength", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("twistrelative", m_Object->m_TwistRelative, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Twist Strength", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton("twistlifetime", m_Object->m_TwistLifetime, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Twist Strength", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

/*
	// Setup flare buttons ...
	Connectk3dChannelButton("minwidth", m_Object->m_Minwidth.Interface());
	Connectk3dChannelButton("minwidthend", m_Object->m_Minwidthend.Interface());
	Connectk3dChannelButton("maxwidth", m_Object->m_Maxwidth.Interface());
	Connectk3dChannelButton("maxwidthend", m_Object->m_Maxwidthend.Interface());
	Connectk3dChannelButton("minheight", m_Object->m_Minheight.Interface());
	Connectk3dChannelButton("minheightend", m_Object->m_Minheightend.Interface());
	Connectk3dChannelButton("maxheight", m_Object->m_Maxheight.Interface());
	Connectk3dChannelButton("maxheightend", m_Object->m_Maxheightend.Interface());
	Connectk3dChannelButton("minangle", m_Object->m_Minangle.Interface());
	Connectk3dChannelButton("minangleend", m_Object->m_Minangleend.Interface());
	Connectk3dChannelButton("maxangle", m_Object->m_Maxangle.Interface());
	Connectk3dChannelButton("maxangleend", m_Object->m_Maxangleend.Interface());
	Connectk3dChannelButton("widthlife", m_Object->m_Widthlife.Interface());
	Connectk3dChannelButton("heightlife", m_Object->m_Heightlife.Interface());
	Connectk3dChannelButton("anglelife", m_Object->m_Anglelife.Interface());
*/

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}


