#ifndef PARTICLEBURST_H
#define PARTICLEBURST_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CParticleBurst K-3D object, which creates a parametric particle system
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dstate.h>
#include <k3dtransformobject.h>
#include <k3dvariable.h>
#include <k3dobjectdialog.h>

#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3ditime.h>

#include <vector>

// Forward class declarations
class CParticleBurst;

/////////////////////////////////////////////////////////////////////////////
// CParticleBurstDialog

typedef k3dObjectDialog CParticleBurstDialogBase;

/// Encapsulates the ParticleBurst UI
class CParticleBurstDialog :
	public CParticleBurstDialogBase
{
public:
	CParticleBurstDialog(CParticleBurst* const Object);

protected:
	/// Stores a pointer to the owning object
	CParticleBurst* const m_Object;
	
	static const sdpString dialogname;
	static const sdpString control_position;
	static const sdpString control_orientation;
	static const sdpString control_scale;
	static const sdpString control_target;
};

////////////////////////////////////////////////////////////////
// CParticle

/// Encapsulates a single particle within a ParticleBurst particle system
class CParticle :
	public k3dILocation,
	public k3dState
{
public:
	CParticle();

	// k3dIState implementation ...
	sdpResult SetState(const gulong Mask, const bool State);

	// k3dILocation implementation ...
	k3dIState* State();
	const sdpVector3& LocalXYZ();
	const sdpVector3& AnimXYZ();
	const sdpVector3& WorldXYZ();
	const sdpVector3& CameraXYZ();
	const sdpVector3& ScreenXYZ();
	const sdpVector3& NormalXYZ();
	const sdpVector3& TextureXYZ();
	const sdpVector3& ImplicitTextureXYZ();
	void SetLocalXYZ(const sdpVector3& Coords);
	void SetAnimXYZ(const sdpVector3& Coords);
	void SetWorldXYZ(const sdpVector3& Coords);
	void SetCameraXYZ(const sdpVector3& Coords);
	void SetScreenXYZ(const sdpVector3& Coords);
	void SetNormalXYZ(const sdpVector3& Coords);
	void SetTextureXYZ(const sdpVector3& Coords);
	void SetImplicitTextureXYZ(const sdpVector3& Coords);

	/// Returns "true" iff the particle is alive at the given time
	bool IsAlive(gdouble Time) const
	{
		return (m_Time[0] <= Time) && (Time <= m_Time[1]);
	}

	sdpVector2 m_Time;
	sdpVector2 m_Strength;
	sdpVector3 m_Color;
	
	sdpVector2 m_Size;
	sdpVector3 m_DirectionXYZ;
	sdpVector2 m_Distance;
	sdpVector3 m_DriftXYZ[2];
	sdpVector3 m_TwistXYZ[2];
	
/*
	sdpVector2 m_Width;
	sdpVector2 m_Height;
	sdpVector2 m_Angle;
*/

	sdpVector3 m_BirthXYZ;
	sdpVector3 m_AnimXYZ;
	sdpVector3 m_WorldXYZ;
	sdpVector3 m_CameraXYZ;
	sdpVector3 m_ScreenXYZ;
};



/////////////////////////////////////////////////////////////
// CParticleBurst

typedef k3dTransformObject CParticleBurstBase;

class CParticleBurst :
	public CParticleBurstBase,
	public k3dILocationCollection
{
friend class CParticleBurstDialog;
public:
	CParticleBurst(sdpIUnknown* Initializer);
	~CParticleBurst();

	// k3dIObject implementation
	sdpResult EditObject();
	sdpResult ObjectDeleted();
	sdpResult ObjectUnDeleted();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// k3dIAnimate implementation
	sdpResult PopAnimate();

	// k3dIEditorRender implementation
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3dIRiRender implementation
	sdpResult RiRender(k3dIRenderFrame* const Frame, k3dIRiEngine* const Engine);

	// k3dILocationCollection implementation
	void VisitLocations(k3dILocationVisitor* const Visitor);
	gulong LocationCount();
	K3DLOCATIONHANDLE FirstLocation();
	K3DLOCATIONHANDLE NextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* Location(K3DLOCATIONHANDLE Location);

	// k3dIPersist implementation
	sdpResult InitComplete();

protected:
	void CreateParticles();
	void RecalculateParticles();
	void RecalculateEmitterPosition();

	void UpdateFull();
	void UpdateCreateParticles();
	void UpdateCalculateParticles();

	/// Enumerates how the particle colors are affected by the burst color channel
	typedef enum
	{
		RANDOM = 1,
		LIFETIME = 2,
		GLOBAL = 3,
	} ColorType;

	friend std::ostream& operator<<(std::ostream& Stream, const ColorType& Value)
	{
		switch(Value)
			{
				case RANDOM:
					Stream << "random";
					break;
				case LIFETIME:
					Stream << "lifetime";
					break;
				case GLOBAL:
					Stream << "global";
					break;
			}
	
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, ColorType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "random")
			Value = RANDOM;
		else if(text == "lifetime")
			Value = LIFETIME;
		else if(text == "global")
			Value = GLOBAL;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}

	// Editor controls ...
	k3dVariable<bool> m_PreviewColors;

	k3dScalarChannel m_Globaltime;
	k3dScalarChannel m_Birthrate;
	k3dScalarChannel m_Minlife;
	k3dScalarChannel m_Maxlife;
	k3dScalarChannel m_Globalstrength;

	k3dVariable<gdouble> m_EmitterPositionSamples;
	k3dVariable<gdouble> m_BirthRateSamples;
	k3dVariable<gulong> m_RandomSeed;

	k3dColorChannel m_Color;
	k3dVariable<ColorType> m_ColorType;
	
	k3dScalarChannel m_Minstrength;
	k3dScalarChannel m_Minstrengthend;
	k3dScalarChannel m_Maxstrength;
	k3dScalarChannel m_Maxstrengthend;
	k3dScalarChannel m_Strengthlife;
	k3dVariable<bool> m_StrengthRandom;
	k3dVariable<bool> m_StrengthRelative;
	k3dVariable<bool> m_StrengthLifetime;
	
	k3dScalarChannel m_Minsize;
	k3dScalarChannel m_Minsizeend;
	k3dScalarChannel m_Maxsize;
	k3dScalarChannel m_Maxsizeend;
	k3dScalarChannel m_Sizelife;
	k3dVariable<bool> m_SizeRandom;
	k3dVariable<bool> m_SizeRelative;
	k3dVariable<bool> m_SizeLifetime;

	k3dScalarChannel m_Mindirectionx;
	k3dScalarChannel m_Maxdirectionx;
	k3dScalarChannel m_Mindirectiony;
	k3dScalarChannel m_Maxdirectiony;
	k3dScalarChannel m_Mindirectionz;
	k3dScalarChannel m_Maxdirectionz;
	k3dVariable<bool> m_DirectionRandom;
	k3dVariable<bool> m_DirectionEmitter;

	k3dScalarChannel m_Mindistance;
	k3dScalarChannel m_Mindistanceend;
	k3dScalarChannel m_Maxdistance;
	k3dScalarChannel m_Maxdistanceend;
	k3dScalarChannel m_Distancelife;
	k3dVariable<bool> m_DistanceRandom;
	k3dVariable<bool> m_DistanceRelative;
	k3dVariable<bool> m_DistanceLifetime;

	k3dScalarChannel m_Mindriftx;
	k3dScalarChannel m_Mindriftxend;
	k3dScalarChannel m_Maxdriftx;
	k3dScalarChannel m_Maxdriftxend;
	k3dScalarChannel m_Driftxlife;
	k3dScalarChannel m_Mindrifty;
	k3dScalarChannel m_Mindriftyend;
	k3dScalarChannel m_Maxdrifty;
	k3dScalarChannel m_Maxdriftyend;
	k3dScalarChannel m_Driftylife;
	k3dScalarChannel m_Mindriftz;
	k3dScalarChannel m_Mindriftzend;
	k3dScalarChannel m_Maxdriftz;
	k3dScalarChannel m_Maxdriftzend;
	k3dScalarChannel m_Driftzlife;
	k3dVariable<bool> m_DriftRandom;
	k3dVariable<bool> m_DriftRelative;
	k3dVariable<bool> m_DriftLifetime;

	k3dScalarChannel m_Mintwistx;
	k3dScalarChannel m_Mintwistxend;
	k3dScalarChannel m_Maxtwistx;
	k3dScalarChannel m_Maxtwistxend;
	k3dScalarChannel m_Twistxlife;
	k3dScalarChannel m_Mintwisty;
	k3dScalarChannel m_Mintwistyend;
	k3dScalarChannel m_Maxtwisty;
	k3dScalarChannel m_Maxtwistyend;
	k3dScalarChannel m_Twistylife;
	k3dScalarChannel m_Mintwistz;
	k3dScalarChannel m_Mintwistzend;
	k3dScalarChannel m_Maxtwistz;
	k3dScalarChannel m_Maxtwistzend;
	k3dScalarChannel m_Twistzlife;
	k3dVariable<bool> m_TwistRandom;
	k3dVariable<bool> m_TwistRelative;
	k3dVariable<bool> m_TwistLifetime;

/*
	k3dScalarChannel m_Minwidth;
	k3dScalarChannel m_Minwidthend;
	k3dScalarChannel m_Maxwidth;
	k3dScalarChannel m_Maxwidthend;
	k3dScalarChannel m_Widthlife;
	k3dScalarChannel m_Minheight;
	k3dScalarChannel m_Minheightend;
	k3dScalarChannel m_Maxheight;
	k3dScalarChannel m_Maxheightend;
	k3dScalarChannel m_Heightlife;
	k3dScalarChannel m_Minangle;
	k3dScalarChannel m_Minangleend;
	k3dScalarChannel m_Maxangle;
	k3dScalarChannel m_Maxangleend;
	k3dScalarChannel m_Anglelife;
	k3dVariable<bool> m_WidthRandom;
	k3dVariable<bool> m_WidthRelative;
	k3dVariable<bool> m_WidthLifetime;
	k3dVariable<bool> m_HeightRandom;
	k3dVariable<bool> m_HeightRelative;
	k3dVariable<bool> m_HeightLifetime;
	k3dVariable<bool> m_AngleAlign;
	k3dVariable<bool> m_AngleRandom;
	k3dVariable<bool> m_AngleRelative;
	k3dVariable<bool> m_AngleLifetime;
*/

	typedef std::vector<CParticle> Particles;
	typedef Particles::iterator ParticleIterator;
	Particles m_Particles;

	bool m_InitializationComplete;
	bool m_Animate;

	k3dITime* m_Time;
	
private:
	static const sdpString m_RecalculateParticles;
	static const sdpString m_CreateParticles;
};

#endif // PARTICLEBURST_H
