// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CPolygonizer K-3D object, which polygonizes implicit objects
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "polygonizer.h"

#include <k3dcom.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3dchooser.h>

#include <k3diapplication.h>
#include <k3dieditorengine.h>
#include <k3diimplicitobject.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>

#include <sdpgtkevent.h>

CPolygonizer::CPolygonizer(sdpIUnknown* Initializer) :
	CPolygonizerBase(Initializer),
	m_VoxelSize("voxelsize", 1.0, Document()->StateRecorder(), 0),
	m_PolygonizationType("polygonizationtype", SurfacePolygonizer::MARCHINGCUBE, Document()->StateRecorder(), 0),
	m_LatticeXMin("latticexmin", -50, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMaximum<gdouble>(0, 0))),
	m_LatticeXMax("latticexmax", 50, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(0, 0))),
	m_LatticeYMin("latticeymin", -50, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMaximum<gdouble>(0, 0))),
	m_LatticeYMax("latticeymax", 50, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(0, 0))),
	m_LatticeZMin("latticezmin", -50, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMaximum<gdouble>(0, 0))),
	m_LatticeZMax("latticezmax", 50, Document()->StateRecorder(), new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(0, 0))),
	m_GridCenterX("gridcenterx", 0, Document()->StateRecorder(), 0),
	m_GridCenterY("gridcentery", 0, Document()->StateRecorder(), 0),
	m_GridCenterZ("gridcenterz", 0, Document()->StateRecorder(), 0)
{
	AddVariable(&m_VoxelSize);
	AddVariable(&m_PolygonizationType);
	AddVariable(&m_LatticeXMin);
	AddVariable(&m_LatticeXMax);
	AddVariable(&m_LatticeYMin);
	AddVariable(&m_LatticeYMax);
	AddVariable(&m_LatticeZMin);
	AddVariable(&m_LatticeZMax);
	AddVariable(&m_GridCenterX);
	AddVariable(&m_GridCenterY);
	AddVariable(&m_GridCenterZ);
}

CPolygonizer::~CPolygonizer()
{
}

//////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CPolygonizer::EditObject()
{
	new CPolygonizerDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CPolygonizer::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	sdpglTrace("CPolygonizer::EditorRender()");

	//const gdouble radius = m_Radius->Value1(m_CurrentTime, m_DeltaTime);
	//const gdouble height = m_Height->Value1(m_CurrentTime, m_DeltaTime);

	// We're lazy, so just draw a bounding-box right now ...
	//DrawBoundingBox(Engine, k3dBoundingBox3(sdpVector3(-radius, height, -radius), sdpVector3(radius, height, radius)));


	// Draw solid geometry ...
	if(SDP_S_OK == Engine->EditorQuerySolid())
		{
			// Offset solid geometry, so it's cleanly visible ...
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonOffset(1.0, 1.0);

			// Solid drawing ...
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			glEnable(GL_LIGHTING);
			glDisable(GL_TEXTURE_1D);
			glDisable(GL_TEXTURE_2D);

			if(SDP_S_OK == Engine->EditorQuerySolidCulling())
				glEnable(GL_CULL_FACE);
			else
				glDisable(GL_CULL_FACE);


			for(gulong p = 0; p < m_ImplicitPolygons.size(); p++)
				{
					std::vector<gulong> path = m_ImplicitPolygons[p];

					// Calculate normal
					gulong pointcount = path.size();
					if(pointcount < 3)
						continue;

					sdpVector3 normal = sdpVector3(0, 0, 0);
					sdpVector3 normal1;
					sdpVector3 normal2;

					for(gulong i = 2; i < pointcount; i++)
						{
							normal1 = m_ImplicitVertices[path[i]] - m_ImplicitVertices[path[i-1]];
							normal2 = m_ImplicitVertices[path[i-2]] - m_ImplicitVertices[path[i-1]];

							normal += normal1 ^ normal2;
						}

					normal1 = m_ImplicitVertices[path[0]] - m_ImplicitVertices[path[pointcount-1]];
					normal2 = m_ImplicitVertices[path[pointcount-2]] - m_ImplicitVertices[path[pointcount-1]];

					normal += normal1 ^ normal2;

					// Draw polygon
					glBegin(GL_POLYGON);
					for(gulong i = 0; i < path.size(); i++)
						{
							glNormal3dv(normal);
							glVertex3dv(m_ImplicitVertices[path[i]]);
						}
					glEnd();
				}

			glDisable(GL_POLYGON_OFFSET_FILL);
		}

	// Draw wireframe geometry ...
	if(SDP_S_OK == Engine->EditorQueryWireframe())
		{
			// Wireframe drawing ...
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glEnable(GL_LIGHTING);
			glDisable(GL_TEXTURE_1D);
			glDisable(GL_TEXTURE_2D);

			if(SDP_S_OK == Engine->EditorQueryWireframeCulling())
				glEnable(GL_CULL_FACE);
			else
				glDisable(GL_CULL_FACE);

			for(gulong p = 0; p < m_ImplicitPolygons.size(); p++)
				{
					std::vector<gulong> path = m_ImplicitPolygons[p];

					// Calculate normal
					gulong pointcount = path.size();
					if(pointcount < 3)
						continue;

					// Draw polygon
					glBegin(GL_POLYGON);
					for(gulong i = 0; i < path.size(); i++)
						glVertex3dv(m_ImplicitVertices[path[i]]);
					glEnd();
				}
		}

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		//DrawTransformationHandles(Engine, radius);
		DrawTransformationHandles(Engine, 10);

	return SDP_S_OK;
}



class CGetImplicitPoints
{
public:
	CGetImplicitPoints(PointList* list)
	{
		m_List = list;
	}

	void operator()(k3dIObject* const Object)
	{
		if(Object->ObjectInfo()->ClassID() != k3d::classid::ImplicitObject)
			return;

		k3dIImplicitObject* ImplicitObject = dynamic_cast<k3dIImplicitObject*>(Object);
		PointList objpoints = ImplicitObject->InsidePoints();
		for(gulong n = 0; n < objpoints.size(); n++)
			{
			m_List->push_back(objpoints[n]);
			}
		//copy(objpoints.begin(), objpoints.end(), m_List->begin());
	}

	PointList* m_List;
};


class CGetImplicitValue
{
public:
	CGetImplicitValue(const sdpVector3& position, gdouble* sum)
	{
		m_Position = position;
		m_Sum = sum;
	}

	void operator()(k3dIObject* const Object)
	{
		if(Object->ObjectInfo()->ClassID() != k3d::classid::ImplicitObject)
			return;

		k3dIImplicitObject* ImplicitObject = dynamic_cast<k3dIImplicitObject*>(Object);
		*m_Sum += ImplicitObject->ImplicitValue(m_Position);
	}

	sdpVector3 m_Position;
	gdouble* m_Sum;
};

gdouble CPolygonizer::MyImplicitFunction(const sdpVector3& position)
{
	gdouble sum = 0;

	// Get function value for each ImplicitObject descendant
	k3d::for_each_descendant(Hierarchy(), this, CGetImplicitValue(position, &sum));

	return sum;
}

class MyFunction :
	public ImplicitFunctor

{
public:
	MyFunction(CPolygonizer* Polygonizer) :
		m_Polygonizer(Polygonizer) {}

	gdouble ImplicitValue(const sdpVector3& Point)
	{
		return m_Polygonizer->MyImplicitFunction(Point);
	}

private:
	CPolygonizer* m_Polygonizer;
};

void CPolygonizer::Update(bool CreateGeometry)
{
	PointList ControlPoints;
	Hierarchy()->SnapshotWorldCoordinates();
	k3d::for_each_descendant(Hierarchy(), this, CGetImplicitPoints(&ControlPoints));

	// Clear geometry ...
	m_ImplicitVertices.clear();
	m_ImplicitPolygons.clear();

	// Set implicit surface functor ...
	MyFunction TheImplicitClass(this);
	TheImplicitClass.m_Vertices = &m_ImplicitVertices;
	TheImplicitClass.m_Polygons = &m_ImplicitPolygons;

	// Create polygonizer instance ...
	SurfacePolygonizer polygonizer(m_VoxelSize.Value(),
			(gint)m_LatticeXMin.Value(), (gint)m_LatticeXMax.Value(),
			(gint)m_LatticeYMin.Value(), (gint)m_LatticeYMax.Value(),
			(gint)m_LatticeZMin.Value(), (gint)m_LatticeZMax.Value(),
			sdpVector3(m_GridCenterX.Value(), m_GridCenterY.Value(), m_GridCenterZ.Value()), m_PolygonizationType.Value(),
			&TheImplicitClass);

	for(gulong nth = 0; nth < ControlPoints.size(); nth++)
		{
			// Check whether object has been polygonized ...
			if(!polygonizer.IsInside(ControlPoints[nth]))
				{
					// Polygonize current skeleton's implicit surface
					polygonizer.Polygonize(ControlPoints[nth]);
				}
		}

	Notify(k3dObject::Event_UserMessage1, sdpToString(m_ImplicitVertices.size()) + " Points, " + sdpToString(m_ImplicitPolygons.size()) + " Paths");
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CPolygonizer::InitComplete()
{
	Update(true);
	return CPolygonizerBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CPolygonizerDialog

const sdpString CPolygonizerDialog::dialogname = "properties";
const sdpString CPolygonizerDialog::control_voxelsize = "voxelsize";
const sdpString CPolygonizerDialog::control_polygonizationtype = "polygonizationtype";
const sdpString CPolygonizerDialog::control_latticexmin = "latticexmin";
const sdpString CPolygonizerDialog::control_latticexmax = "latticexmax";
const sdpString CPolygonizerDialog::control_latticeymin = "latticeymin";
const sdpString CPolygonizerDialog::control_latticeymax = "latticeymax";
const sdpString CPolygonizerDialog::control_latticezmin = "latticezmin";
const sdpString CPolygonizerDialog::control_latticezmax = "latticezmax";
const sdpString CPolygonizerDialog::control_gridcenterx = "gridcenterx";
const sdpString CPolygonizerDialog::control_gridcentery = "gridcentery";
const sdpString CPolygonizerDialog::control_gridcenterz = "gridcenterz";

CPolygonizerDialog::CPolygonizerDialog(CPolygonizer* const Object) :
	CPolygonizerDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("sfxlib", "polygonizer.gtkml"));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_voxelsize, m_Object->m_VoxelSize, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Voxel Size", 0, new k3dActionUpdate<CPolygonizer>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_latticexmin, m_Object->m_LatticeXMin, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lattice X Min", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_latticexmax, m_Object->m_LatticeXMax, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lattice X Max", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_latticeymin, m_Object->m_LatticeYMin, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lattice Y Min", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_latticeymax, m_Object->m_LatticeYMax, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lattice Y Max", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_latticezmin, m_Object->m_LatticeZMin, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lattice Z Min", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_latticezmax, m_Object->m_LatticeZMax, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Lattice Z Max", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_gridcenterx, m_Object->m_GridCenterX, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Grid Center X", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_gridcentery, m_Object->m_GridCenterY, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Grid Center Y", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_gridcenterz, m_Object->m_GridCenterZ, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Grid Center Z", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Setup choosers ...
	if(Getk3dChooser(control_polygonizationtype))
		Getk3dChooser(control_polygonizationtype)->Attach(new k3dChooserVariableDataSource<SurfacePolygonizer::PolygonizationType>(m_Object->m_PolygonizationType), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Polygonization", 0, new k3dActionUpdate<CPolygonizer>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Show();
}


