#ifndef POLYGONIZER_H
#define POLYGONIZER_H

// K-3D
// Copyright  1995 - 2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Defines the CPolygonizer K-3D object, which polygonizes implicit objects
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dobjectdialog.h>
#include <k3dtrianglemesh.h>
#include <k3deditorrender.h>
#include <k3dtool.h>
#include <k3dvariable.h>

#include "surfacepolygonizer.h"

extern const sdp::ClassID sdpPolygonizer;

// Forward class references ...
class CPolygonizer;

/////////////////////////////////////////////////////////////////////////////////////
// CPolygonizerDialog

typedef k3dObjectDialog CPolygonizerDialogBase;

/// Encapsulates the Polygonizer UI
class CPolygonizerDialog :
	public CPolygonizerDialogBase
{
public:
	CPolygonizerDialog(CPolygonizer* const Object);

protected:
	/// Stores a pointer to the owning object
	CPolygonizer* const m_Object;

	static const sdpString dialogname;
	static const sdpString control_polygonizationtype;
	static const sdpString control_voxelsize;
	static const sdpString control_latticexmin;
	static const sdpString control_latticexmax;
	static const sdpString control_latticeymin;
	static const sdpString control_latticeymax;
	static const sdpString control_latticezmin;
	static const sdpString control_latticezmax;
	static const sdpString control_gridcenterx;
	static const sdpString control_gridcentery;
	static const sdpString control_gridcenterz;
};

////////////////////////////////////////////////////////////////////////////////////
// CPolygonizer

typedef k3dTransformObject CPolygonizerBase;

class CPolygonizer :
	public CPolygonizerBase
{
friend class CPolygonizerDialog;
public:
	CPolygonizer(sdpIUnknown* Initializer);
	~CPolygonizer();

	// k3dIObject implementation
	sdpResult EditObject();

	// k3dIEditorRender implementation
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3dIPersist implementation
	sdpResult InitComplete();

	// Where it all happens ...
	void Update(bool CreateGeometry);

	gdouble MyImplicitFunction(const sdpVector3& position);

protected:

	/// Enumerates supported polygonization types
	friend std::ostream& operator<<(std::ostream& Stream, const SurfacePolygonizer::PolygonizationType& Value)
	{
		switch(Value)
			{
				case SurfacePolygonizer::MARCHINGCUBE:
					Stream << "marchingcube";
					break;
				case SurfacePolygonizer::TETRAHEDRAL:
					Stream << "tetrahedral";
					break;
			}

		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, SurfacePolygonizer::PolygonizationType& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "marchingcube")
			Value = SurfacePolygonizer::MARCHINGCUBE;
		else if(text == "tetrahedral")
			Value = SurfacePolygonizer::TETRAHEDRAL;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;

		return Stream;
	}

	// Implicit surface polygonization parameters
	k3dVariable<gdouble> m_VoxelSize;
	k3dVariable<SurfacePolygonizer::PolygonizationType> m_PolygonizationType;
	k3dVariable<gdouble> m_LatticeXMin;
	k3dVariable<gdouble> m_LatticeXMax;
	k3dVariable<gdouble> m_LatticeYMin;
	k3dVariable<gdouble> m_LatticeYMax;
	k3dVariable<gdouble> m_LatticeZMin;
	k3dVariable<gdouble> m_LatticeZMax;
	k3dVariable<gdouble> m_GridCenterX;
	k3dVariable<gdouble> m_GridCenterY;
	k3dVariable<gdouble> m_GridCenterZ;

	/// Store control points for the polygonizer to look for their implicit surface
	sdpVector3Array m_ControlPoints;

	/// Store implicit surface
	std::vector<sdpVector3> m_ImplicitVertices;
	std::vector< std::vector<gulong> > m_ImplicitPolygons;
};

#endif // POLYGONIZER_H
