// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSoftObject K-3D object, which generates soft objects
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "softobject.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>
#include <k3dicamera.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>

#include <sdpgtkevents.h>

#include <algorithm>
#include <iterator>
#include <iostream>


CSoftObject::CSoftObject(sdpIUnknown* Initializer) :
	CSoftObjectBase(Initializer),
	m_Radius("radius", 10.0, Document()->StateRecorder(), 0),
	m_Hardness("hardness", 1.5, Document()->StateRecorder(), 0),
	m_SkeletonType("skeletontype", SETOFPOINTS, Document()->StateRecorder(), 0),
	m_PotentialType("potentialtype", BLOB, Document()->StateRecorder(), 0)
{
	AddVariable(&m_Radius);
	AddVariable(&m_Hardness);
	AddVariable(&m_SkeletonType);
	AddVariable(&m_PotentialType);
}

CSoftObject::~CSoftObject()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSoftObject::EditObject()
{
	new CSoftObjectDialog(this);
	return SDP_S_OK;
}

void CSoftObject::Update(bool ModifyGeometry)
{
}


///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CSoftObject::EditorRender(k3dIEditorEngine* const Engine)
{
	// If we aren't visible, we're done ...
	if(!AdvancedQuery(k3dIState::VISIBLE))
		return SDP_S_FALSE;

	sdpglTrace("CSoftObject::EditorRender()");

	//const gdouble radius = m_Radius->Value1(m_CurrentTime, m_DeltaTime);
	//const gdouble height = m_Height->Value1(m_CurrentTime, m_DeltaTime);

	// We're lazy, so just draw a bounding-box right now ...
	//DrawBoundingBox(Engine, k3dBoundingBox3(sdpVector3(-radius, height, -radius), sdpVector3(radius, height, radius)));

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glColor3d(0.0, 0.0, 1.0);
	glPointSize(6.0f);

	// Draw a point to mark current object position
	glPushMatrix();
	glBegin(GL_POINTS);
		glVertex3d(0, 0, 0);
	glEnd();
	glPopMatrix();

	// If we're the active object, show the user ...
	if(Document()->MouseFocus() == this)
		//DrawTransformationHandles(Engine, radius);
		DrawTransformationHandles(Engine, 10);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CSoftObject::InitComplete()
{
	Update(true);
	return CSoftObjectBase::InitComplete();
}

///////////////////////////////////////////////////////////////
// k3dSoftObject overrides

class CSoftObjectGetPoints
{
public:
	CSoftObjectGetPoints(std::vector<sdpVector3>* pointlist)
	{
		m_PointList = pointlist;
	}

	void operator()(k3dIObject* const Object)
	{
		k3dILocationCollection* locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

		// For each location ...
		sdpVector3 coords;
		k3dILocation* location;
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->WorldXYZ();
				m_PointList->push_back(coords);
			}
	}

private:
	std::vector<sdpVector3>* m_PointList;
};

PointList CSoftObject::InsidePoints()
{
	m_ControlPoints.clear();
	k3d::for_each_descendant(Hierarchy(), this, CSoftObjectGetPoints(&m_ControlPoints));

	// Set default single point if there's no child object in the hierarchy
	if(m_ControlPoints.size() == 0)
		{
			sdpVector3 mypos = Position(k3dITransform::CURRENT_FRAME);
			m_ControlPoints.push_back(mypos);
		}

	return m_ControlPoints;
}

// Blinn object -> 'Blob'
gdouble BlobPotential(gdouble d, gdouble p)
{
	return 0.5*exp(p*(1-4*d*d));
}

// Kacic-Alesic and Wyvill function behaviour, by Carole Blanc and Christophe Schlick
gdouble FiniteSoftObjectPotential(gdouble d, gdouble p)
{
	if(d >= 1.0)
		return 0;

	gdouble d2 = d*d;
	if(d < 0.5)
		return 1 - 9*d2*d2/(p+(4.5-4*p)*d2);

	return (1-d2)*(1-d2)/(0.75-p+(1.5+4*p)*d2);
}

// Kacic-Alesic and Wyvill function behaviour, by Carole Blanc and Christophe Schlick
gdouble InfiniteSoftObjectPotential(gdouble d, gdouble p)
{
	gdouble d2 = d*d;
	if(d < 0.5)
		return 1 - 1/(2+p*(1-4*d2));

	return 1/(2+p*(-1+4*d2));
}


gdouble CSoftObject::ImplicitValue(const sdpVector3& position)
{
	gdouble value = 0;

	std::vector<sdpVector3>::iterator cpiterator;
	for(cpiterator = m_ControlPoints.begin(); cpiterator != m_ControlPoints.end(); cpiterator++)
		{
			sdpVector3 mypos = *cpiterator;

			switch(m_PotentialType.Value())
				{
					case CSoftObject::BLOB:
						value += BlobPotential((position-mypos).Length()/m_Radius.Value(), m_Hardness.Value());
						break;
					case CSoftObject::FINITESOFTOBJECT:
						value += FiniteSoftObjectPotential((position-mypos).Length()/m_Radius.Value(), m_Hardness.Value());
						break;
					case CSoftObject::INFINITESOFTOBJECT:
						value += InfiniteSoftObjectPotential((position-mypos).Length()/m_Radius.Value(), m_Hardness.Value());
						break;
				}
		}

	// Functions are normalized in the Wyvill's form : f(0.5) = 0.5
	return value - 0.5;
}

////////////////////////////////////////////////////////////////////////////
// CSoftObjectDialog

const sdpString CSoftObjectDialog::dialogname = "properties";
const sdpString CSoftObjectDialog::control_radius = "radius";
const sdpString CSoftObjectDialog::control_hardness = "hardness";
const sdpString CSoftObjectDialog::control_skeletontype = "skeletontype";
const sdpString CSoftObjectDialog::control_potentialtype = "potentialtype";

CSoftObjectDialog::CSoftObjectDialog(CSoftObject* const Object) :
	CSoftObjectDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("sfxlib", "softobject.gtkml"));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_radius, m_Object->m_Radius, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Radius", 0, new k3dActionUpdate<CSoftObject>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_hardness, m_Object->m_Hardness, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Hardness", 0, new k3dActionUpdate<CSoftObject>(m_Object, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Setup choosers ...
	if(Getk3dChooser(control_skeletontype))
		Getk3dChooser(control_skeletontype)->Attach(new k3dChooserVariableDataSource<CSoftObject::SkeletonType>(m_Object->m_SkeletonType), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Skeleton", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	if(Getk3dChooser(control_potentialtype))
		Getk3dChooser(control_potentialtype)->Attach(new k3dChooserVariableDataSource<CSoftObject::PotentialType>(m_Object->m_PotentialType), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Potential", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}


