#ifndef SOFTOBJECT_H
#define SOFTOBJECT_H

// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CSoftObject K-3D object, which generates soft objects
		\author Romain Behar (romainbehar@yahoo.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3deditorrender.h>
#include <k3dobjectdialog.h>
#include <k3diimplicitobject.h>
#include <k3dsoftobject.h>

#include <sdpgl.h>

extern const sdp::ClassID sdpImplicitObject;

// Forward class references ...
class CSoftObject;

/////////////////////////////////////////////////////////////////////////////
// CSoftObjectDialog

typedef k3dObjectDialog CSoftObjectDialogBase;

/// Encapsulates the SoftObject UI
class CSoftObjectDialog :
	public CSoftObjectDialogBase
{
public:
	CSoftObjectDialog(CSoftObject* const Object);

protected:
	/// Stores a pointer to the owning object
	CSoftObject* const m_Object;

	static const sdpString dialogname;
	static const sdpString control_radius;
	static const sdpString control_hardness;
	static const sdpString control_skeletontype;
	static const sdpString control_potentialtype;
};

/////////////////////////////////////////////////////////////////////////////
// CSoftObject

typedef k3dSoftObject CSoftObjectBase;

class CSoftObject :
	public CSoftObjectBase
{
friend class CSoftObjectDialog;
public:
	CSoftObject(sdpIUnknown* Initializer);
	~CSoftObject();

	// k3dIObject overrices ...
	sdpResult EditObject();

	// k3dIEditorRender implementation
	sdpResult EditorRender(k3dIEditorEngine* const Engine);

	// k3IPersist overrides ...
	sdpResult InitComplete();

	// k3dIImplicitObject overrides ...
	PointList InsidePoints();
	gdouble ImplicitValue(const sdpVector3& position);

	// Where it all happens ...
	void Update(bool ModifyGeometry);

protected:

	/// Enumerates supported skeleton types
	typedef enum
	{
		SETOFPOINTS,
		WIREFRAME,
		MESH
	} SkeletonType;

	friend std::ostream& operator<<(std::ostream& Stream, const SkeletonType& Value)
	{
		switch(Value)
			{
				case SETOFPOINTS:
					Stream << "setofpoints";
					break;
				case WIREFRAME:
					Stream << "wireframe";
					break;
				case MESH:
					Stream << "mesh";
					break;
			}

		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, SkeletonType& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "setofpoints")
			Value = SETOFPOINTS;
		else if(text == "wireframe")
			Value = WIREFRAME;
		else if(text == "mesh")
			Value = MESH;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;

		return Stream;
	};

	/// Enumerates supported potential function types
	typedef enum
	{
		BLOB,
		INFINITESOFTOBJECT,
		FINITESOFTOBJECT
	} PotentialType;

	friend std::ostream& operator<<(std::ostream& Stream, const PotentialType& Value)
	{
		switch(Value)
			{
				case BLOB:
					Stream << "blob";
					break;
				case INFINITESOFTOBJECT:
					Stream << "infinitesoftobject";
					break;
				case FINITESOFTOBJECT:
					Stream << "finitesoftobject";
					break;
			}

		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, PotentialType& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "blob")
			Value = BLOB;
		else if(text == "infinitesoftobject")
			Value = INFINITESOFTOBJECT;
		else if(text == "finitesoftobject")
			Value = FINITESOFTOBJECT;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;

		return Stream;
	};

	// Soft object parameters ...
	k3dVariable<gdouble> m_Radius;
	k3dVariable<gdouble> m_Hardness;
	k3dVariable<SkeletonType> m_SkeletonType;
	k3dVariable<PotentialType> m_PotentialType;

	// Control parameters ...
	std::vector<sdpVector3> m_ControlPoints;
};


#endif // SOFTOBJECT_H
