// A C++ Implicit Surface Polygonizer
// Copyright 2002, Romain Behar <romainbehar@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements SurfacePolygonizer, an implicit surface polygonizer
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "surfacepolygonizer.h"


namespace
{

// Number of iterations (convergence)
const int RES = 10;

// Directions
const gulong L = 0;	// left direction:	-x, -i
const gulong R = 1;	// right direction:	+x, +i
const gulong B = 2;	// bottom direction:	-y, -j
const gulong T = 3;	// top direction:	+y, +j
const gulong N = 4;	// near direction:	-z, -k
const gulong F = 5;	// far direction:	+z, +k

// Corners
const gulong LBN = 0;	// left bottom near
const gulong LBF = 1;	// left bottom far
const gulong LTN = 2;	// left top near
const gulong LTF = 3;	// left top far
const gulong RBN = 4;	// right bottom near
const gulong RBF = 5;	// right bottom far
const gulong RTN = 6;	// right top near
const gulong RTF = 7;	// right top far

} // namespace


// Initializer
SurfacePolygonizer::SurfacePolygonizer(const gdouble size, const gint xmin, const gint xmax, const gint ymin, const gint ymax, const gint zmin, const gint zmax, const sdpVector3& origin, const PolygonizationType& polygonization, ImplicitFunctor* implicitfunctor)
{
	// Lattice size, bounds and origin
	m_VoxelSize = size;
	m_MinCorner = Location(xmin, ymin, zmin);
	m_MaxCorner = Location(xmax, ymax, zmax);
	m_GridOrigin = origin;
	if(!(m_MinCorner <= PointLocation(m_GridOrigin) && PointLocation(m_GridOrigin) < m_MaxCorner))
		std::cerr << "Surface Polygonizer: grid origin must be in grid" << std::endl;

	// Implicit function
	m_FieldFunctor = implicitfunctor;

	// Init polygonizing algorithm
	m_Decomposition = polygonization;
	if(m_Decomposition == MARCHINGCUBE)
		{
			MakeCubeTable();
		}

	// Save vertices and polygon lists pointers
	m_Vertices = implicitfunctor->m_Vertices;
	m_Polygons = implicitfunctor->m_Polygons;
}


// Polygonize the implicit surface function
void SurfacePolygonizer::Polygonize(const sdpVector3& startingpoint)
{
	// Get a Location enclosing surface
	Location startinglocation = PointLocation(startingpoint);
	if(SurfaceLocation(startinglocation))
		{
			// Surface found, polygonize it
			PolygonizeSurface(startinglocation);
		}
	else
		{
			// Surface wasn't found, check around the grid boundary
		}
}

void SurfacePolygonizer::PolygonizeSurface(const Location& startinglocation)
{
	// Create initial cube
	SetCenter(startinglocation);
	Cube c(startinglocation);
	for(int n = 0; n < 8; n++)
		{
			c.corners[n] = SetCorner(startinglocation + Location(Bit(n, 2), Bit(n, 1), Bit(n, 0)));
		}

	// Push it on stack
	m_ActiveCubes.push(c);

	// Process active cubes till none left
	while(!m_ActiveCubes.empty())
		{
			Cube c = m_ActiveCubes.top();
			m_ActiveCubes.pop();

			// Polygonize
			switch(m_Decomposition)
				{
					case MARCHINGCUBE: MarchingCube(c); break;
					case TETRAHEDRAL: TetrahedralDecomposition(c); break;
					default: g_assert_not_reached();
				}

			// Test six face directions, maybe add to stack
			TestFace(c.l.Left(), c, L, LBN, LBF, LTN, LTF);
			TestFace(c.l.Right(), c, R, RBN, RBF, RTN, RTF);
			TestFace(c.l.Bottom(), c, B, LBN, LBF, RBN, RBF);
			TestFace(c.l.Top(), c, T, LTN, LTF, RTN, RTF);
			TestFace(c.l.Near(), c, N, LBN, LTN, RBN, RTN);
			TestFace(c.l.Far(), c, F, LBF, LTF, RBF, RTF);
		}
}


// Find a location enclosing surface
bool SurfacePolygonizer::SurfaceLocation(Location& startinglocation)
{
	Location loc2 = startinglocation;
	gdouble value2 = m_FieldFunctor->ImplicitValue(LocationPos(loc2));

	// Top
	do
		{
			Location loc1 = loc2;
			gdouble value1 = value2;

			loc2 = loc2.Top();
			value2 = m_FieldFunctor->ImplicitValue(LocationPos(loc2));

			if((value1*value2 < 0) || ((value1 == 0) && (value2 < 0)) || ((value2 == 0) && (value1 <0)))
				{
					startinglocation = loc1;
					return true;
				}
		}
	while(loc2 <= m_MaxCorner);

	// We reached the grid boundary: check the whole grid

	return false;
}


//**** Tetrahedral Polygonization ****

void SurfacePolygonizer::TetrahedralDecomposition(const Cube& c)
{
	// Decompose into tetrahedra and polygonize
	DoTet(c, LBN, LTN, RBN, LBF);
	DoTet(c, RTN, LTN, LBF, RBN);
	DoTet(c, RTN, LTN, LTF, LBF);
	DoTet(c, RTN, RBN, LBF, RBF);
	DoTet(c, RTN, LBF, LTF, RBF);
	DoTet(c, RTN, LTF, RTF, RBF);
}


// Triangulate the tetrahedron (b, c, d should appear clockwise when viewed from a)
void SurfacePolygonizer::DoTet(const Cube& cube1, int c1, int c2, int c3, int c4)
{
	Corner *a = cube1.corners[c1];
	Corner *b = cube1.corners[c2];
	Corner *c = cube1.corners[c3];
	Corner *d = cube1.corners[c4];

	int index = 0;
	bool apos, bpos, cpos, dpos;
	if(apos = (a->value >= 0))
		index += 8;
	if(bpos = (b->value >= 0))
		index += 4;
	if(cpos = (c->value >= 0))
		index += 2;
	if(dpos = (d->value >= 0))
		index += 1;

	// Index is now 4-bit number representing one of the 16 possible cases
	int e1, e2, e3, e4, e5, e6;
	if(apos != bpos)
		e1 = VerticeId(a, b);
	if(apos != cpos)
		e2 = VerticeId(a, c);
	if(apos != dpos)
		e3 = VerticeId(a, d);
	if(bpos != cpos)
		e4 = VerticeId(b, c);
	if(bpos != dpos)
		e5 = VerticeId(b, d);
	if(cpos != dpos)
		e6 = VerticeId(c, d);

	// 14 productive tetrahedral cases (0000 and 1111 do not yield polygons)
	switch(index)
		{
			case 1: SaveTriangle(e5, e6, e3); break;
			case 2: SaveTriangle(e2, e6, e4); break;
			case 3:	SaveTriangle(e3, e5, e4); SaveTriangle(e3, e4, e2); break;
			case 4: SaveTriangle(e1, e4, e5); break;
			case 5: SaveTriangle(e3, e1, e4); SaveTriangle(e3, e4, e6); break;
			case 6: SaveTriangle(e1, e2, e6); SaveTriangle(e1, e6, e5); break;
			case 7: SaveTriangle(e1, e2, e3); break;
			case 8: SaveTriangle(e1, e3, e2); break;
			case 9: SaveTriangle(e1, e5, e6); SaveTriangle(e1, e6, e2); break;
			case 10: SaveTriangle(e1, e3, e6); SaveTriangle(e1, e6, e4); break;
			case 11: SaveTriangle(e1, e5, e4); break;
			case 12: SaveTriangle(e3, e2, e4); SaveTriangle(e3, e4, e5); break;
			case 13: SaveTriangle(e6, e2, e4); break;
			case 14: SaveTriangle(e5, e3, e6); break;
		}
}


//**** Cubical Polygonization ****

namespace
{

// edge:                    LB, LT, LN, LF, RB, RT, RN, RF, BN, BF, TN, TF
const gulong corner1[12] = {LBN,LTN,LBN,LBF,RBN,RTN,RBN,RBF,LBN,LBF,LTN,LTF};
const gulong corner2[12] = {LBF,LTF,LTN,LTF,RBF,RTF,RTN,RTF,RBN,RBF,RTN,RTF};

} // namespace

// Triangulate the cube directly, without decomposition
void SurfacePolygonizer::MarchingCube(const Cube& cube1)
{
	gulong index = 0;
	for(gulong i = 0; i < 8; i++)
		if(cube1.corners[i]->value >= 0)
			index += 1 << i;

	std::vector< std::vector<gulong> > currentindex = m_CubeTable[index];
	for(gulong i = 0; i < currentindex.size(); i++)
		{
			gulong a, b, count = 0;

			for(gulong j = 0; j < currentindex[i].size(); j++)
				{
					Corner* c1 = cube1.corners[corner1[currentindex[i][j]]];
					Corner* c2 = cube1.corners[corner2[currentindex[i][j]]];
					gulong c = VerticeId(c1, c2);

					if(++count > 2)
						SaveTriangle(a, b, c);

					if(count < 3)
						a = b;
					b = c;
				}
		}
}

namespace
{

// Build Marching Cube Table

const gulong LB = 0;	// left bottom edge
const gulong LT = 1;	// left top edge
const gulong LN = 2;	// left near edge
const gulong LF = 3;	// left far edge
const gulong RB = 4;	// right bottom edge
const gulong RT = 5;	// right top edge
const gulong RN = 6;	// right near edge
const gulong RF = 7;	// right far edge
const gulong BN = 8;	// bottom near edge
const gulong BF = 9;	// bottom far edge
const gulong TN = 10;	// top near edge
const gulong TF = 11;	// top far edge

// Face on left when going from corner1 to corner2
const gulong leftface[12] = {B,  L,  L,  F,  R,  T,  N,  R,  N,  B,  T,  F};
// Face on right when going from corner1 to corner2
const gulong rightface[12] = {L,  T,  N,  L,  B,  R,  R,  F,  B,  F,  N,  T};

} // namespace

// Return next clockwise edge from given edge around given face
inline gulong SurfacePolygonizer::NextCWedge(gulong edge, gulong face)
{
	switch(edge)
		{
			case LB: return (face == L) ? LF : BN;
			case LT: return (face == L) ? LN : TF;
			case LN: return (face == L) ? LB : TN;
			case LF: return (face == L) ? LT : BF;
			case RB: return (face == R) ? RN : BF;
			case RT: return (face == R) ? RF : TN;
			case RN: return (face == R) ? RT : BN;
			case RF: return (face == R) ? RB : TF;
			case BN: return (face == B) ? RB : LN;
			case BF: return (face == B) ? LB : RF;
			case TN: return (face == T) ? LT : RN;
			case TF: return (face == T) ? RT : LF;
			default: g_assert_not_reached();
		}

	return LF;
}

// Create the 256 entry table for cubical polygonization
void SurfacePolygonizer::MakeCubeTable()
{
	for(gulong configuration = 0; configuration < 256; configuration++)
		{
			std::vector< std::vector<gulong> > triangles;

			gulong bits[8];
			for(gulong c = 0; c < 8; c++)
				bits[c] = Bit(configuration, c);

			bool done[12];
			for(gulong edge = 0; edge < 12; edge++) done[edge] = false;
			for(gulong edge = 0; edge < 12; edge++)
				if(!done[edge] && (bits[corner1[edge]] != bits[corner2[edge]]))
					{
						std::vector<gulong> triangle;

						// Get face that is to right of edge from pos to neg corner
						gulong face = bits[corner1[edge]] ? rightface[edge] : leftface[edge];

						gulong startingedge = edge, currentedge = edge;
						do
							{
							currentedge = NextCWedge(currentedge, face);
							done[currentedge] = true;

							if(bits[corner1[currentedge]] != bits[corner2[currentedge]])
								{
									triangle.push_back(currentedge);

									// face adjoining edge that is not the given face
									if(face == leftface[currentedge])
										face = rightface[currentedge];
									else
										face = leftface[currentedge];
								}
							}
						while(currentedge != startingedge);

						triangles.push_back(triangle);
					}

			m_CubeTable.push_back(triangles);
		}

	for(gulong i = 0; i < m_CubeTable.size(); i++)
		for(gulong j = 0; j < m_CubeTable[i].size(); j++)
			reverse(m_CubeTable[i][j].begin(), m_CubeTable[i][j].end());
}


//**** Storage ****

// Given cube at lattice (i, j, k), and four corners of face,
// if surface crosses face, compute other four corners of adjacent cube
// and add new cube to cube stack

void SurfacePolygonizer::TestFace(const Location& facelocation, Cube& old, int face, int c1, int c2, int c3, int c4)
{
	// No surface crossing?
	bool pos = old.corners[c1]->value >= 0;
	if((old.corners[c2]->value >= 0) == pos && (old.corners[c3]->value >= 0) == pos && (old.corners[c4]->value >= 0) == pos)
		return;

	// Out of bounds?
	if(!(m_MinCorner <= facelocation && facelocation < m_MaxCorner))
		return;

	// Already visited?
	if(SetCenter(facelocation))
		return;

	// Create new cube and add it to top of stack
	Cube newc(facelocation);

	const int facebit[6] = {2, 2, 1, 1, 0, 0};
	int bit = facebit[face];
	newc.corners[Flip(c1, bit)] = old.corners[c1];
	newc.corners[Flip(c2, bit)] = old.corners[c2];
	newc.corners[Flip(c3, bit)] = old.corners[c3];
	newc.corners[Flip(c4, bit)] = old.corners[c4];

	for(int n = 0; n < 8; n++)
		if(newc.corners[n] == NULL)
			newc.corners[n] = SetCorner(facelocation + Location(Bit(n, 2), Bit(n, 1), Bit(n, 0)));

	m_ActiveCubes.push(newc);
}


// Return the sdpVector3 corresponding to the Location
sdpVector3 SurfacePolygonizer::LocationPos(const Location& l)
{
	return m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i, (double)l.j, (double)l.k);
}

// Return the nearest location corresponding to the sdpVector3
SurfacePolygonizer::Location SurfacePolygonizer::PointLocation(const sdpVector3& p)
{
	sdpVector3 p2 = (p - m_GridOrigin)/m_VoxelSize;
	int li = (int)(p2.n[0]);
	int lj = (int)(p2.n[1]);
	int lk = (int)(p2.n[2]);

	return Location(li, lj, lk);
}

// Return the gradient at Location l
sdpVector3 SurfacePolygonizer::LocationGradient(const Location& l)
{
	gdouble delta = m_VoxelSize / 5;
	double gx = m_FieldFunctor->ImplicitValue(m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i + delta, (double)l.j, (double)l.k)) - m_FieldFunctor->ImplicitValue(m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i - delta, (double)l.j, (double)l.k));
	double gy = m_FieldFunctor->ImplicitValue(m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i, (double)l.j + delta, (double)l.k)) - m_FieldFunctor->ImplicitValue(m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i, (double)l.j - delta, (double)l.k));
	double gz = m_FieldFunctor->ImplicitValue(m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i, (double)l.j, (double)l.k + delta)) - m_FieldFunctor->ImplicitValue(m_GridOrigin + m_VoxelSize*sdpVector3((double)l.i, (double)l.j, (double)l.k - delta));

	return sdpVector3(gx, gy, gz);
}


// Return corner with the given lattice Location set (and cache) its function value
SurfacePolygonizer::Corner* SurfacePolygonizer::SetCorner(const Location& l)
{
	// Create new corner at Location l
	Corner* c = new Corner(l);
	c->p = LocationPos(l);

	// Find whether if has already been computed
	cornersiterator ci = m_Corners.find(l);
	if(ci != m_Corners.end())
		{
			c->value = (*ci).second->value;
			return c;
		}

	// Else compute value and gradient
	c->value = m_FieldFunctor->ImplicitValue(c->p);
	c->gradient = LocationGradient(l);

	// And save it
	m_Corners[l] = c;

	return c;
}


// Return true if already set, otherwise set and return false
bool SurfacePolygonizer::SetCenter(const Location& l)
{
	centersiterator ci = m_Centers.find(l);
	if(ci != m_Centers.end())
		{
			return true;
		}

	m_Centers[l] = true;

	return false;
}


// Save a triangle
void SurfacePolygonizer::SaveTriangle(gulong u, gulong v, gulong w)
{
	std::vector<gulong> triangle;
	triangle.push_back(u);
	triangle.push_back(v);
	triangle.push_back(w);

	m_Polygons->push_back(triangle);
}


// Save a quad
void SurfacePolygonizer::SaveQuad(gulong u, gulong v, gulong w, gulong x)
{
	std::vector<gulong> quad;
	quad.push_back(u);
	quad.push_back(v);
	quad.push_back(w);
	quad.push_back(x);

	m_Polygons->push_back(quad);
}


// Return index for vertex on edge
int SurfacePolygonizer::VerticeId(Corner *c1, Corner *c2)
// c1->value and c2->value are presumed of different sign
{
	int vid = m_Edges.GetValue(Edge(c1->l, c2->l));
	if(vid != -1)
		{
			// Has been previously computed, return saved index
			return vid;
		}

	// Compute index, save and return it
	sdpVector3 p;
	Converge(c1->p, c2->p, c1->value, p);
	m_Vertices->push_back(p);

	vid = m_Vertices->size() - 1;
	m_Edges.push_back(Edge(c1->l, c2->l, vid));

	return vid;
}

// From two points of differing sign, converge to zero crossing
void SurfacePolygonizer::Converge(const sdpVector3& p1, const sdpVector3& p2, double v, sdpVector3& p)
{
	sdpVector3 pos = p1;
	sdpVector3 neg = p2;

	if(v < 0)
		swap(pos, neg);

	p = 0.5 * (pos + neg);

	for(int iter = 0; iter < RES; iter++)
		{
			if(m_FieldFunctor->ImplicitValue(p) >= 0)
				pos = p;
			else
				neg = p;

			p = 0.5 * (pos + neg);
		}
}


// Tells whether Location belongs to a polygonized surface,
//  return surface gradient when answer is 'true'
bool SurfacePolygonizer::IsCorner(const Location& l, sdpVector3& returngradient)
{
	cornersiterator ci = m_Corners.find(l);
	if(ci != m_Corners.end())
		{
			returngradient = (*ci).second->gradient;
			return true;
		}

	return false;
}

// Check whether point is inside or outside computed polygonization
bool SurfacePolygonizer::IsInside(const sdpVector3& testpoint)
{
	sdpVector3 gradient;
	Location startlocation = PointLocation(testpoint);

	// Check 6 directions until we find either a polygonized surface
	//  or a boundary

	// Left
	Location move = startlocation;
	while(m_MinCorner <= move)
		{
		if(IsCorner(move, gradient))
			if(gradient.n[0] < 0)
				return false;
			else
				return true;

		move = move.Left();
		}
	// Right
	move = startlocation;
	while(move <= m_MaxCorner)
		{
		if(IsCorner(move, gradient))
			if(gradient.n[0] > 0)
				return false;
			else
				return true;

		move = move.Right();
		}
	// Bottom
	move = startlocation;
	while(m_MinCorner <= move)
		{
		if(IsCorner(move, gradient))
			if(gradient.n[1] < 0)
				return false;
			else
				return true;

		move = move.Bottom();
		}
	// Top
	move = startlocation;
	while(move <= m_MaxCorner)
		{
		if(IsCorner(move, gradient))
			if(gradient.n[1] > 0)
				return false;
			else
				return true;

		move = move.Top();
		}
	// Near
	move = startlocation;
	while(m_MinCorner <= move)
		{
		if(IsCorner(move, gradient))
			if(gradient.n[2] < 0)
				return false;
			else
				return true;

		move = move.Near();
		}
	// Far
	move = startlocation;
	while(move <= m_MaxCorner)
		{
		if(IsCorner(move, gradient))
			if(gradient.n[2] > 0)
				return false;
			else
				return true;

		move = move.Right();
		}

	// There is no surface around testpoint
	return false;
}


