#ifndef SURFACEPOLYGONIZER_H
#define SURFACEPOLYGONIZER_H

// A C++ Implicit Surface SurfacePolygonizer
// Copyright 2002, Romain Behar <romainbehar@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares SurfacePolygonizer, an implicit surface polygonizer
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <sdptypes.h>
#include <hash_map>

// It is based on the following program :
//
// C code from the article
// "An Implicit Surface SurfacePolygonizer"
// by Jules Bloomenthal, jbloom@beauty.gmu.edu
// in "Graphics Gems IV", Academic Press, 1994
//
// implicit.c
//     an implicit surface polygonizer, translated from Mesa
//     applications should call polygonize()
//
// Authored by Jules Bloomenthal, Xerox PARC.
// Copyright (c) Xerox Corporation, 1991.  All rights reserved.
// Permission is granted to reproduce, use and distribute this code for
// any and all purposes, provided that this notice appears in all copies.


// Implicit surface functor
//  ImplicitValue() returns positive or zero for inside, negative for outside

class ImplicitFunctor
{
public:
	virtual ~ImplicitFunctor() {}

	virtual gdouble ImplicitValue(const sdpVector3& Point) = 0;

	std::vector<sdpVector3>* m_Vertices;
	std::vector< std::vector<gulong> >* m_Polygons;
};


// SurfacePolygonizer implementation
class SurfacePolygonizer
{
public:
	typedef enum
	{
		MARCHINGCUBE,
		TETRAHEDRAL
	} PolygonizationType;


	SurfacePolygonizer(const gdouble size, const gint xmin, const gint xmax, const gint ymin, const gint ymax, const gint zmin, const gint zmax, const sdpVector3& origin, const PolygonizationType& polygonization, ImplicitFunctor* implicitfunctor);
	~SurfacePolygonizer() {}

	bool IsInside(const sdpVector3& testpoint);
	void Polygonize(const sdpVector3& startingpoint);

//protected:

	// Lattice position (centered on (0, 0, 0), signed values)
	class Location
	{
	public:
		Location(const int I = 0, const int J = 0, const int K = 0) :
			i(I),
			j(J),
			k(K)
		{
		}

		inline friend bool operator == (const Location& a, const Location& b)
		{
			return (a.i == b.i) && (a.j == b.j) && (a.k == b.k);
		}
		inline friend Location operator + (const Location& a, const Location& b)
		{
			return Location(a.i + b.i, a.j + b.j, a.k + b.k);
		}
		inline friend bool operator <= (const Location& a, const Location& b)
		{
			return (a.i <= b.i && a.j <= b.j && a.k <= b.k);
		}
		inline friend bool operator < (const Location& a, const Location& b)
		{
			return (a.i < b.i && a.j < b.j && a.k < b.k);
		}

		Location Left() { return Location(i-1, j, k); }
		Location Right() { return Location(i+1, j, k); }
		Location Bottom() { return Location(i, j-1, k); }
		Location Top() { return Location(i, j+1, k); }
		Location Near() { return Location(i, j, k-1); }
		Location Far() { return Location(i, j, k+1); }

		int i;
		int j;
		int k;
	};

	class std::hash<Location>
	{
	public:
		size_t operator() (const Location& Value) const
		{
			static const int HashBit = 5;
			static const int Mask = (1 << HashBit) - 1;
			return ((Value.i & Mask) << (HashBit*2)) | ((Value.j & Mask) << HashBit) | (Value.k & Mask);
		}
	};

	// Cube corner
	class Corner
	{
	public:
		Location l;
		sdpVector3 p;
		double value;
		sdpVector3 gradient;

		Corner(const Location& L) :
			l(L)
		{
		}

		inline friend bool operator == (const Corner& c1, const Corner& c2)
		{
			return (c1.l == c2.l) && (c1.p == c2.p) && (c1.value == c2.value);
		}
	};

	// Partitioning cell
	class Cube
	{
	public:
		Location l;
		Corner* corners[8];

		Cube(const Location& L) :
			l(L)
			{
				for(int i = 0; i < 8; i++)
					corners[i] = NULL;
			}
	};

	class Edge
	{
	public:
		Edge(const Location& L1, const Location& L2, const int VID = -1) :
			vid(VID)
		{
			if(L1.i > L2.i || (L1.i == L2.i && (L1.j > L2.j || (L1.j == L2.j && L1.k > L2.k))))
				{
					l1 = L2;
					l2 = L1;
				}
			else
				{
					l1 = L1;
					l2 = L2;
				}
		}

		inline friend bool operator == (const Edge& e1, const Edge& e2)
		{
			return (e1.l1 == e2.l1) && (e1.l2 == e2.l2) && (e1.vid == e2.vid);
		}

		Location l1;
		Location l2;
		int vid;
	};

	class EdgeHash
	{
	private:
		static const int HashBit = 5;
		static const int Mask = (1 << HashBit) - 1;
		static const int HashSize = 1 << (3 * HashBit);

		inline int HashFunc(const Location& l)
		{
			return ((((l.i & Mask) << HashBit) | (l.j & Mask)) << HashBit) | (l.k & Mask);
		}

	public:
		EdgeHash()
			{
			edges.resize(HashSize*2);
			}

		void push_back(const Edge& Value)
			{
			int index = HashFunc(Value.l1) + HashFunc(Value.l2);
			edges[index].push_back(Value);
			}

		int GetValue(const Edge& Value)
			{
				int index = HashFunc(Value.l1) + HashFunc(Value.l2);
				for(unsigned int n = 0; n < edges[index].size(); n++)
					{
						if(edges[index][n].l1 == Value.l1 && edges[index][n].l2 == Value.l2)
							return edges[index][n].vid;
					}

				return -1;
			}

	protected:
		std::vector< std::vector<Edge> > edges;
	};

private:
	ImplicitFunctor* m_FieldFunctor;

	void PolygonizeSurface(const Location& startinglocation);

	// Inline functions
	inline int Bit(int i, int n) { return (i>>n)&1; }
	inline int Flip(int i, int bit) { return i ^ (1 << bit); }

	sdpVector3 LocationPos(const Location& l);
	Location PointLocation(const sdpVector3& p);
	sdpVector3 LocationGradient(const Location& l);

	bool SurfaceLocation(Location& startinglocation);

	void TetrahedralDecomposition(const Cube& c);
	void DoTet(const Cube& cube1, int c1, int c2, int c3, int c4);

	void MakeCubeTable();
	void MarchingCube(const Cube& cube1);

	void TestFace(const Location& facelocation, Cube& old, int face, int c1, int c2, int c3, int c4);

	Corner* SetCorner(const Location& l);
	bool IsCorner(const Location& l, sdpVector3& returngradient);
	bool SetCenter(const Location& l);

	int VerticeId(Corner *c1, Corner *c2);
	void Converge(const sdpVector3& p1, const sdpVector3& p2, double v, sdpVector3& p);

	void SaveTriangle(gulong u, gulong v, gulong w);
	void SaveQuad(gulong u, gulong v, gulong w, gulong x);

	// Grid center ( Location(0, 0, 0) )
	sdpVector3 m_GridOrigin;

	// Width of the partitioning cube
	double m_VoxelSize;

	// Grid limit corners (left-bottom-near and right-top-far)
	Location m_MinCorner;
	Location m_MaxCorner;

	// Polygonization type
	PolygonizationType m_Decomposition;

	// Hash tables
	std::hash_map<Location, bool> m_Centers;
	typedef std::hash_map<Location, bool>::iterator centersiterator;
	std::hash_map<Location, Corner*> m_Corners;
	typedef std::hash_map<Location, Corner*>::iterator cornersiterator;
	EdgeHash m_Edges;

	// Active cube stack
	stack<Cube> m_ActiveCubes;

	// Surface storage
	std::vector<sdpVector3>* m_Vertices;
	std::vector< std::vector<gulong> >* m_Polygons;

	// Build fast Marching Cube tables
	std::vector <std::vector <std::vector <gulong> > > m_CubeTable;
	gulong NextCWedge(gulong edge, gulong face);
};

#endif // SURFACEPOLYGONIZER_H
