// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBend K-3D tool, which applies a "Bend" transformation to geometry
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "bend.h"

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

#include <math.h>

namespace
{

class BendLocations
{
public:
	BendLocations(const k3d::Axis Along, const k3d::Axis Around, const gdouble BendAngle, const gdouble Tightness, const gdouble JointPosition, const sdpVector3 Min, const sdpVector3 Max) :
		m_Along(Along),
		m_Around(Around),
		m_BendAngle(BendAngle),
		m_Tightness(Tightness),
		m_JointPosition(JointPosition),
		m_Min(Min),
		m_Max(Max),
		m_Size(Max - Min),
		m_Origin((Min + Max) * 0.5),
		m_Joint(sdpLerp(Min, Max, JointPosition)),
		m_JLower(sdpLerp(Min, Max, JointPosition - (1.0 - Tightness)))
	{
	}	
	
	void operator()(k3d::SelectedLocationProxy& Location)
	{
		sdpVector3 coords = Location.OldXYZ();

		if(m_Along == m_Around) 
			{
				Location->SetLocalXYZ(coords);
				return;
			}

		gdouble delta = 0;
		gdouble radius = 1.0 - m_Tightness;
		gdouble size = 0;
		sdpMatrix4 rotation;
		sdpVector3 axis;
		sdpVector3 offset;
		sdpVector3 pivot = m_Origin;
		switch(m_Along)
			{
				case k3d::X:
					if(m_Size[VX])
						delta = (coords[VX] - m_Min[VX]) / m_Size[VX];
					size = m_Size[VX];
					if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
					{
						coords[VX] = m_JLower[VX];
						pivot[VX] = m_JLower[VX];
					}
					else
					{
						m_Origin[VX] = m_Joint[VX];
					} 
					switch(m_Around)
						{
							case k3d::X:
								break;
							case k3d::Y:
								axis = sdpVector3(0,1,0);
								offset = sdpVector3(0,0,-1);
								break;
							case k3d::Z:
								axis = sdpVector3(0,0,1);
								offset = sdpVector3(0,1,0);
								break;
							default:
								g_return_if_fail(0);
						}
					break;
				case k3d::Y:
					if(m_Size[VY])
						delta = (coords[VY] - m_Min[VY]) / m_Size[VY];
					size = m_Size[VY];
					if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
					{
						coords[VY] = m_JLower[VY];
						pivot[VY] = m_JLower[VY];
					}
					else
					{
						m_Origin[VY] = m_Joint[VY];
					} 
					switch(m_Around)
						{
							case k3d::X:
								axis = sdpVector3(1,0,0);
								offset = sdpVector3(0,0,1);
								break;
							case k3d::Y:
								break;
							case k3d::Z:
								axis = sdpVector3(0,0,1);
								offset = sdpVector3(-1,0,0);
								break;
							default:
								g_return_if_fail(0);
						}
					break;
				case k3d::Z:
					if(m_Size[VZ])
						delta = (coords[VZ] - m_Min[VZ]) / m_Size[VZ];
					size = m_Size[VZ];
					if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
					{
						coords[VZ] = m_JLower[VZ];
						pivot[VZ] = m_JLower[VZ];
					}
					else
					{
						m_Origin[VZ] = m_Joint[VZ];
					} 
					switch(m_Around)
						{
							case k3d::X:
								axis = sdpVector3(1,0,0);
								offset = sdpVector3(0,-1,0);
								break;
							case k3d::Y:
								axis = sdpVector3(0,1,0);
								offset = sdpVector3(1,0,0);
								break;
							case k3d::Z:
								break;
							default:
								g_return_if_fail(0);
						}
					break;
				default:
					g_return_if_fail(0);
			}

		if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
		{
			gdouble d = (delta-(m_JointPosition-radius)) / (radius*2.0);
			gdouble factor = radius*size / tan(m_BendAngle*0.5);
			rotation = sdpRotation3D( d*m_BendAngle, axis );
			offset = offset * factor;
			coords -= pivot + offset;
			coords = rotation * coords;
			coords += pivot + offset;
		}
		else if( delta >= m_JointPosition+radius  )
		{
			rotation = sdpRotation3D( m_BendAngle, axis );
			coords -= m_Origin;
			coords = rotation * coords;
			coords += m_Origin;
		}
		
		Location->SetLocalXYZ(coords);
	}
	
private:
	const k3d::Axis m_Along;
	const k3d::Axis m_Around;
	const gdouble m_BendAngle;
	const gdouble m_Tightness;
	const gdouble m_JointPosition;
	const sdpVector3 m_Min;
	const sdpVector3 m_Max;
	const sdpVector3 m_Size;
	sdpVector3 m_Origin;
	const sdpVector3 m_Joint;
	const sdpVector3 m_JLower;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CBend

CBend::CBend(sdpIUnknown* Initializer) :
	CBendBase(Initializer),
	k3dMouseEventObserver("LMB-drag to bend selected Points"),
	m_BendAngle("bendangle", 0.0, 0, 0),
	m_Tightness("tightness", 0.6, 0, new k3dVariableRange<gdouble>(0.0, 1.0, 0)),
	m_JointPosition("jointposition", 0.5, 0, new k3dVariableRange<gdouble>(0.0, 1.0, 0)),
	m_Along("along", k3d::Y, 0, 0),
	m_Around("around", k3d::X, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CBend::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	std::copy(m_CurrentSelection.begin(), m_CurrentSelection.end(), k3d::Dimensions::Inserter(m_CurrentSelectionDimensions));

	return CBendBase::InitComplete();
}

sdpResult CBend::EditObject()
{
	new CBendDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CBend::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CBend::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = ((double)(CurrentX - LastX)) * 0.01;
	double yfactor = ((double)(CurrentY - LastY)) * 0.01;

	k3dSetValue(m_BendAngle, m_BendAngle.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);
	
	return SDP_S_OK;
}

sdpResult CBend::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CBend::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CBend::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), BendLocations(m_Along.Value(), m_Around.Value(), m_BendAngle.Value(), m_Tightness.Value(), m_JointPosition.Value(), m_CurrentSelectionDimensions.Min(), m_CurrentSelectionDimensions.Max()));
}

void CBend::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CBend::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Bend");
}

//////////////////////////////////////////////////////////////////////////////////////
// CBendDialog

const sdpString CBendDialog::dialogname = "properties";
const sdpString CBendDialog::control_along = "along";
const sdpString CBendDialog::control_around = "around";
const sdpString CBendDialog::control_bendangle = "bendangle";
const sdpString CBendDialog::control_tightness = "tightness";
const sdpString CBendDialog::control_jointposition = "jointposition";

CBendDialog::CBendDialog(CBend* const Object) :
	CBendDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "bend.gtkml"));

	// Setup choosers ...
	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), 0, "Along", new k3dActionMethodCall<CBend>(m_Object, &CBend::PushEditing, 0), new k3dActionMethodCall<CBend>(m_Object, &CBend::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBend>(m_Object, &CBend::PopEditing, 0));

	if(Getk3dChooser(control_around))
		Getk3dChooser(control_around)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Around), k3dApplication()->CommandRecorder(), 0, "Around", new k3dActionMethodCall<CBend>(m_Object, &CBend::PushEditing, 0), new k3dActionMethodCall<CBend>(m_Object, &CBend::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBend>(m_Object, &CBend::PopEditing, 0));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_bendangle, m_Object->m_BendAngle, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBend>(m_Object, &CBend::PushEditing, 0), new k3dActionMethodCall<CBend>(m_Object, &CBend::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBend>(m_Object, &CBend::PopEditing, 0));
	Connectk3dSpinButton(control_tightness, m_Object->m_Tightness, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBend>(m_Object, &CBend::PushEditing, 0), new k3dActionMethodCall<CBend>(m_Object, &CBend::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBend>(m_Object, &CBend::PopEditing, 0));
	Connectk3dSpinButton(control_jointposition, m_Object->m_JointPosition, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBend>(m_Object, &CBend::PushEditing, 0), new k3dActionMethodCall<CBend>(m_Object, &CBend::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBend>(m_Object, &CBend::PopEditing, 0));

	RootWidget().Show();
}


