// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBevel tool, which creates bevelled geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "bevel.h"

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dpathpointcollection.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

#include <numeric>

CBevel::PathProxy::PathProxy(CBevel& Object, k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PathProxyBase(Engine, Mesh),
	m_Object(Object)
{
}

CBevel::PathProxy::PathProxy(CBevel& Object, k3d::GeometryEngine& Engine, k3dIPath* const Path) :
	PathProxyBase(Engine, Path),
	m_Object(Object)
{
}

void CBevel::PathProxy::BuildGeometry()
{
	// Skip anything with < 3 sides ...
	if(m_Points.size() >= 3)
		{
			// Create a new point for each existing point ...
			for(PointCollection::iterator point = m_Points.begin(); point != m_Points.end(); ++point)
				m_NewPoints.push_back(&Engine().Point(Mesh()));

			// Create a new face path ...
			if(m_Object.m_CreateFaces.Value())
				{
					m_FacePaths.push_back(&Engine().Path(Mesh()));
					for(PointCollection::iterator newpoint = m_NewPoints.begin(); newpoint != m_NewPoints.end(); ++newpoint)
						m_FacePaths.back()->Actual()->Points()->AddPoint((*newpoint)->Actual());
				}

			// Create new edge paths ...
			if(m_Object.m_CreateEdges.Value())
				{
					for(gulong i = 0; i < m_Points.size(); i++)
						{
							m_EdgePaths.push_back(&Engine().Path(Mesh()));
					
							m_EdgePaths.back()->Actual()->Points()->AddPoint(m_Points[(i+0) % m_Points.size()]->Actual());
							m_EdgePaths.back()->Actual()->Points()->AddPoint(m_Points[(i+1) % m_Points.size()]->Actual());
							m_EdgePaths.back()->Actual()->Points()->AddPoint(m_NewPoints[(i+1) % m_NewPoints.size()]->Actual());
							m_EdgePaths.back()->Actual()->Points()->AddPoint(m_NewPoints[(i+0) % m_NewPoints.size()]->Actual());
						}
				}

/*										
			// Mark the original for deletion ...
			if(m_Object.m_DeleteOriginals.Value())
				MarkForDeletion();
*/
		}

	PathProxyBase::BuildGeometry();
}

void CBevel::PathProxy::UpdateGeometry()
{
	// Calculate some generic data about the path ...
	k3d::Dimensions dimensions;
	k3d::copy(Actual()->Points(), k3d::Dimensions::Inserter(dimensions));
	
	// Get the path center and normal vector ...
	const sdpVector3 center = dimensions.Center();
	const sdpVector3 normal = dimensions.Normal();
	
	// Calculate an offset perpendicular to the path face ...
	const sdpVector3 normaloffset = normal * m_Object.m_Offset.Value();

	// Set new point positions ...
	if(CBevel::SCALE == m_Object.m_InsetType.Value())
		{
			PointCollection::iterator newpoint = m_NewPoints.begin();
			for(PointCollection::iterator point = m_Points.begin(); point != m_Points.end(); ++point)
				(*newpoint++)->Actual()->Location()->SetLocalXYZ(sdpLerp(center, (*point)->Actual()->Location()->LocalXYZ(), m_Object.m_Scale.Value()) + normaloffset);
		}
	else
		{
			for(gulong i = 0; i < m_Points.size(); i++)
				{
				
					// Get the point we're going to offset from ...
					const sdpVector3 point2 = m_Points[i]->Actual()->Location()->LocalXYZ();

					// Look for the first point preceeding it by a non-zero distance ...
					sdpVector3 point1;
					for(gulong p1 = (i + m_Points.size()-1) % m_Points.size(); p1 != i; p1 = (p1 + m_Points.size()-1) % m_Points.size())
						{
							point1 = m_Points[p1]->Actual()->Location()->LocalXYZ();
							
							if(point1 != point2)
								break;
						}
					
					// Look for the first point following it by a non-zero distance ...
					sdpVector3 point3;
					for(gulong p3 = (i + 1) % m_Points.size(); p3 != i; p3 = (p3 + 1) % m_Points.size())
						{
							point3 = m_Points[p3]->Actual()->Location()->LocalXYZ();
							
							if(point3 != point2)
								break;
						}
				
					const sdpVector3 edge1 = point2 - point1;
					const sdpVector3 edge2 = point3 - point2;

					sdpVector3 edge3 = edge1 ^ normal; // Cross product
					sdpVector3 edge4 = edge2 ^ normal; // Cross product

					edge3.Normalize();
					edge4.Normalize();

					sdpVector3 inset = edge3 + edge4;
					inset.Normalize();
					inset *= m_Object.m_Inset.Value();

					// Calculate angle adjustment
					const gdouble theta = acos((edge1 * edge2) / (edge1.Length() * edge2.Length()));
					inset *= 1.0 / cos(theta * 0.5);

					m_NewPoints[i]->Actual()->Location()->SetLocalXYZ(m_Points[i]->Actual()->Location()->LocalXYZ() + normaloffset + inset);
				}
		}
	

	PathProxyBase::UpdateGeometry();
}

void CBevel::PathProxy::Cancel()
{
	m_FacePaths.clear();
	m_EdgePaths.clear();
	m_NewPoints.clear();
	
	PathProxyBase::Cancel();
}

/////////////////////////////////////////////////////////////////////////////
// CBevel

CBevel::CBevel(sdpIUnknown* Initializer) :
	CBevelBase(Initializer),
	k3dMouseEventObserver("LMB-drag to bevel selected Paths"),
	m_Offset("offset", -0.4, 0, 0),
	m_Inset("inset", 0.4, 0, 0),
	m_Scale("scale", 0.5, 0, 0),
	m_DeleteOriginals("deleteoriginals", true, 0, 0),
	m_CreateFaces("createfaces", true, 0, 0),
	m_CreateEdges("createedges", true, 0, 0),
	m_InsetType("insettype", INSET, 0, 0),
	m_Engine(
		Document(),
		new k3d::EngineMeshProxyFactory<k3d::EngineMeshProxy>(),
		new k3d::EnginePathProxyFactory2<CBevel, CBevel::PathProxy>(*this),
		new k3d::EnginePointProxyFactory<k3d::EnginePointProxy>(),
		new k3d::EngineEdgeProxyFactory<k3d::EngineEdgeProxy>()
		)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CBevel::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedPaths(Document()->Objects(), std::back_inserter(m_Selection));
	std::copy(m_Selection.begin(), m_Selection.end(), k3d::GeometryEngine::Inserter(m_Engine));

	return CBevelBase::InitComplete();
}

sdpResult CBevel::EditObject()
{
	new Properties(this);
	
	BuildGeometry();
	k3dRedrawAll(Document(), false);
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CBevel::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	return SDP_S_OK;
}

sdpResult CBevel::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = gdouble(CurrentX - LastX) * 0.01;
	double yfactor = gdouble(CurrentY - LastY) * 0.01;

	k3dSetValue(m_Offset, m_Offset.Value() - yfactor);
	k3dSetValue(m_Inset, m_Inset.Value() + xfactor);
	k3dSetValue(m_Scale, m_Scale.Value() + xfactor);

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CBevel::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	return SDP_S_OK;
}

void CBevel::BuildGeometry()
{
	// Get rid of any old geometry
	m_Engine.Cancel();

	// Create new geometry ...
	m_Engine.StartChangeSet();
	m_Engine.BuildGeometry();
	m_Engine.Synchronize();
	m_Engine.FinishChangeSet();
	
	// Update it ...
	m_Engine.UpdateGeometry();
	m_Engine.Manage();

	// Notify UI ...
	Notify(k3dObject::Event_UserMessage1, m_Engine.Description());
}

void CBevel::UpdateGeometry()
{
	m_Engine.UpdateGeometry();
}

void CBevel::CancelGeometry()
{
	m_Engine.Cancel();
}

void CBevel::FinishGeometry()
{
	m_Engine.Finish("Bevel Selected");
}

//////////////////////////////////////////////////////////////////////////////////////
// CBevel::Properties

CBevel::Properties::Properties(CBevel* const Object) :
	PropertiesBase(Object, true, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "bevel.gtkml"));

	// Connnect check buttons ...
	Connectk3dCheckButton("deleteoriginals", m_Object->m_DeleteOriginals, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton("createfaces", m_Object->m_CreateFaces, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton("createedges", m_Object->m_CreateEdges, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	
	// Connect spin buttons ...
	Connectk3dSpinButton("offset", m_Object->m_Offset, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("inset", m_Object->m_Inset, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("scale", m_Object->m_Scale, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Connect choosers ...
	if(Getk3dChooser("insettype"))
		Getk3dChooser("insettype")->Attach(new k3dChooserVariableDataSource<CBevel::InsetType>(m_Object->m_InsetType), k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CBevel>(m_Object, &CBevel::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Show();
}

void CBevel::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnOK();
}

void CBevel::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnCancel();
}


