// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBulge K-3D tool, which applies a "Bulge" transformation to geometry
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "bulge.h"

#include <k3dchooser.h>
#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dichannel.h>
#include <k3dieditable.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

namespace
{

class BulgeLocations
{
public:
	BulgeLocations(const k3d::Axis Along, const CBulge::Type Type, const bool DisplaceX, const bool DisplaceY, const bool DisplaceZ, const gdouble BulgeFactor, k3dIChannel* const Profile, const sdpVector3 Min, const sdpVector3 Max) :
		m_Along(Along),
		m_Type(Type),
		m_DisplaceX(DisplaceX),
		m_DisplaceY(DisplaceY),
		m_DisplaceZ(DisplaceZ),
		m_BulgeFactor(BulgeFactor),
		m_Profile(Profile),
		m_Size(Max - Min),
		m_Min(Min)
	{
	}	
	
	void operator()(k3d::SelectedLocationProxy& Location)
	{
		sdpVector3 coords = Location.OldXYZ();

		gdouble delta;
		gdouble bulge_amount;
		switch(m_Along)
			{
				case k3d::X:
					if(m_Size[0])
						delta = (coords[0] - m_Min[0]) / m_Size[0];
					break;
				case k3d::Y:
					if(m_Size[1])
						delta = (coords[1] - m_Min[1]) / m_Size[1];
					break;
				case k3d::Z:
					if(m_Size[2])
						delta = (coords[2] - m_Min[2]) / m_Size[2];
					break;
				default:
					g_return_if_fail(0);
			}

		bulge_amount = m_Profile->Value1(delta, 0.001);
		if( CBulge::RADIAL == m_Type )
			{
				gdouble distance = coords.Length();
				gdouble scale;
				if( 0 == distance )
					scale = 1.0;
				else
					scale = ( distance + m_BulgeFactor * bulge_amount ) / distance;

				if(m_DisplaceX && (k3d::X != m_Along))
					coords[0] *= scale;
		
				if(m_DisplaceY && (k3d::Y != m_Along))
					coords[1] *= scale;
		
				if(m_DisplaceZ && (k3d::Z != m_Along))
					coords[2] *= scale;
			}
		else
			{
				gdouble offset = m_BulgeFactor * bulge_amount;

				if(m_DisplaceX && (k3d::X != m_Along))
					coords[0] += offset;
		
				if(m_DisplaceY && (k3d::Y != m_Along))
					coords[1] += offset;
		
				if(m_DisplaceZ && (k3d::Z != m_Along))
					coords[2] += offset;
			}
	
		Location->SetLocalXYZ(coords);
	}
	
private:
	const k3d::Axis m_Along;
	const CBulge::Type m_Type;
	const bool m_DisplaceX;
	const bool m_DisplaceY;
	const bool m_DisplaceZ;
	const gdouble m_BulgeFactor;
	k3dIChannel* const m_Profile;
	const sdpVector3 m_Size;
	const sdpVector3 m_Min;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CBulge

CBulge::CBulge(sdpIUnknown* Initializer) :
	CBulgeBase(Initializer),
	k3dMouseEventObserver("LMB-drag to bulge selected Points"),
	m_BulgeFactor("bulgefactor", 1.0, 0, 0),
	m_DisplaceX("displacex", true, 0, 0),
	m_DisplaceY("displacey", true, 0, 0),
	m_DisplaceZ("displacez", true, 0, 0),
	m_Along("along", k3d::X, 0, 0),
	m_Type("type", RADIAL, 0, 0),
	m_Profile("profile", Document(), 0, new k3dEventConnector(this, k3dIChannel::Event_Changed, "CBulge", 0))
{
	// Create our profile curve, and set it to a sensible default ...
	m_Profile.Create("Bulge Profile", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, false);

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(m_Profile.Interface());
	g_return_if_fail(channel);
	
	// Let the channel know that we're making changes ...
	k3d::Editing editing(channel->Editable());

	channel->SetSegmentCount(2);
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);
	g_assert( 7 == nodes.size() );
	
	k3dBezierChannelNodeArray::iterator node(nodes.begin());
	
	(*node)->SetPosition(sdpVector2(0.0, 0.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.0, 0.5));
	node++;
	(*node)->SetPosition(sdpVector2(0.2, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.5, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.8, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(1.0, 0.5));
	node++;
	(*node)->SetPosition(sdpVector2(1.0, 0.0));
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CBulge::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	std::copy(m_CurrentSelection.begin(), m_CurrentSelection.end(), k3d::Dimensions::Inserter(m_CurrentSelectionDimensions));
	
	return CBulgeBase::InitComplete();
}

sdpResult CBulge::EditObject()
{
	new CBulgeDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CBulge::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CBulge::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = ((double)(CurrentX - LastX)) * 0.1;
	double yfactor = ((double)(CurrentY - LastY)) * 0.1;

	k3dSetValue(m_BulgeFactor, m_BulgeFactor.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);
	
	return SDP_S_OK;
}

sdpResult CBulge::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CBulge::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIChannel::Event_Changed)
		{
			PushEditing();
			UpdateGeometry();
			PopEditing();
			k3dRedrawAll(Document(), false);
			
			return SDP_S_OK;
		}
	
	return CBulgeBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

void CBulge::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CBulge::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), BulgeLocations(m_Along.Value(), m_Type.Value(), m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), m_BulgeFactor.Value(), m_Profile.Interface(), m_CurrentSelectionDimensions.Min(), m_CurrentSelectionDimensions.Max()));
}

void CBulge::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CBulge::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Bulge");
}

//////////////////////////////////////////////////////////////////////////////////////
// CBulgeDialog

const sdpString CBulgeDialog::dialogname = "properties";
const sdpString CBulgeDialog::control_displacex = "displacex";
const sdpString CBulgeDialog::control_displacey = "displacey";
const sdpString CBulgeDialog::control_displacez = "displacez";
const sdpString CBulgeDialog::control_along = "along";
const sdpString CBulgeDialog::control_bulgefactor = "bulgefactor";
const sdpString CBulgeDialog::control_profile = "profile";
const sdpString CBulgeDialog::control_type = "type";

CBulgeDialog::CBulgeDialog(CBulge* const Object) :
	CBulgeDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "bulge.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PushEditing, 0), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PopEditing, 0));
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PushEditing, 0), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PopEditing, 0));
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PushEditing, 0), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PopEditing, 0));

	// Setup choosers ...
	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), 0, "Along", new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PushEditing, 0), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PopEditing, 0));

	if(Getk3dChooser(control_type))
		Getk3dChooser(control_type)->Attach(new k3dChooserVariableDataSource<CBulge::Type>(m_Object->m_Type), k3dApplication()->CommandRecorder(), 0, "Type", new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PushEditing, 0), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PopEditing, 0));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_bulgefactor, m_Object->m_BulgeFactor, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PushEditing, 0), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CBulge>(m_Object, &CBulge::PopEditing, 0));
	
	// Setup profile channel ...
	Connectk3dChannelButton(control_profile, m_Object->m_Profile.Interface());

	RootWidget().Show();
}


