// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDisplace K-3D tool, which displaces geometry
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "displace.h"

#include <k3dchooser.h>
#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3dobjectselector.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>
#include <sdpgtkevents.h>

namespace
{

class DisplaceLocations
{
public:
	DisplaceLocations(CDisplaceFunction* const Function, const bool DisplaceX, const bool DisplaceY, const bool DisplaceZ, const gdouble Amplitude, const CDisplace::DisplaceType Type) :
		m_Function(Function),
		m_DisplaceX(DisplaceX),
		m_DisplaceY(DisplaceY),
		m_DisplaceZ(DisplaceZ),
		m_Amplitude(Amplitude),
		m_Type(Type)
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		sdpVector3 coords = Location.OldXYZ();

		if(CDisplace::LINEAR == m_Type)
			{
				gdouble offset = m_Amplitude * m_Function->Value( Location );

				if( m_DisplaceX )
					coords[0] += offset;
		
				if( m_DisplaceY )
					coords[1] += offset;
		
				if( m_DisplaceZ )
					coords[2] += offset;
			}
		else
			{
				gdouble distance = coords.Length();
				gdouble scale;
				if( 0 == distance )
					scale = 1.0;
				else
					scale = ( distance + m_Amplitude * m_Function->Value( Location ) ) / distance;

				if( m_DisplaceX )
					coords[0] *= scale;
		
				if( m_DisplaceY )
					coords[1] *= scale;
		
				if( m_DisplaceZ )
					coords[2] *= scale;
			}

		Location->SetLocalXYZ(coords);
	}
	
private:
	CDisplaceFunction* const m_Function;
	const bool m_DisplaceX;
	const bool m_DisplaceY;
	const bool m_DisplaceZ;
	const gdouble m_Amplitude;
	const CDisplace::DisplaceType m_Type;
};

} // namespace

////////////////////////////////////////////////////////////////////////////
// CDisplaceFunction

CDisplaceFunction::CDisplaceFunction()
{
	m_Type = FRACTAL;
	m_Fractal = NULL;
}

CDisplaceFunction::~CDisplaceFunction()
{
	if( m_Fractal )
		delete m_Fractal;
}

void CDisplaceFunction::SetFractal( const FractalType FType, const gulong Seed, const gdouble Smoothness, const gdouble Frequency, const gdouble Octaves )
{
	if( m_Fractal )
		delete m_Fractal;

	switch( FType )
		{
			case RANDOM:
				m_Fractal = new CRand();
				break;
			case NOISE:
				m_Fractal = new CNoise();
				break;
			case FBM:
				m_Fractal = new CFBm();
				break;
			case TURBULENCE:
				m_Fractal = new CTurbulence();
				break;
			case RIDGED:
				m_Fractal = new CRidged();
				break;
			default:
			g_return_if_fail(0);
		}

	m_Fractal->Seed( Seed );
	m_Fractal->Smoothness( Smoothness );
	m_Fractal->Octaves( Octaves );
	
	m_Frequency = Frequency;
	
	m_Type = FRACTAL;
}

void CDisplaceFunction::SetImage( const ImageType IType, k3dPixel* const Data, const gulong Width, const gulong Height )
{
	m_ImageType = IType;
	m_Data = Data;
	m_Width = Width;
	m_Height = Height;
	
	if( m_Data )
		m_Type = IMAGE;
	else
		m_Type = NONE;
}

void CDisplaceFunction::SetNone()
{
	m_Type = NONE;
}

gdouble CDisplaceFunction::Value(k3d::SelectedLocationProxy& Location)
{
	if( m_Type == FRACTAL )
		{
			if( !m_Fractal )
				return 0;
		
			sdpVector3 coords = Location.OldXYZ();
			
			return m_Fractal->Value( coords * m_Frequency );
		}
	else if( m_Type == IMAGE )
		{		
			sdpVector3 coords = Location->TextureXYZ();

			gulong i =  (gulong)( coords[0] * m_Width ) % m_Width;
			gulong j =  (gulong)( coords[1] * m_Height ) % m_Height;

			k3dPixel pixel = m_Data[ j*m_Width + i ];
			gdouble value = 0;

			switch( m_ImageType )
				{
					case AVERAGE:
						value = (pixel.Red + pixel.Green + pixel.Blue) / (3.0 * 255.0);
						break;
					case RED:
						value = pixel.Red / 255.0;
						break;
					case GREEN:
						value = pixel.Green / 255.0;
						break;
					case BLUE:
						value = pixel.Blue / 255.0;
						break;
					case ALPHA:
						value = pixel.Alpha / 255.0;
						break;
					default:
						g_return_val_if_fail(0, 0);
				}
			return value;
		}
	else
		{
			return 0;
		}
}

/////////////////////////////////////////////////////////////////////////////
// CDisplace

CDisplace::CDisplace(sdpIUnknown* Initializer) :
	CDisplaceBase(Initializer),
	k3dMouseEventObserver("LMB-drag to displace selected Points"),
	m_DisplaceX("displacex", true, 0, 0),
	m_DisplaceY("displacey", true, 0, 0),
	m_DisplaceZ("displacez", true, 0, 0),
	m_Amplitude("amplitude", 1.0, 0, 0),
	m_Smoothness( "smoothness", 0.5, 0, 0 ),
	m_Frequency( "frequency", 0.2, 0, 0 ),
	m_Octaves( "octaves", 4.0, 0, new k3dVariableTruncate( new k3dVariableMinimum<gdouble>(0.0, 0))),
	m_Seed("seed", 123457.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(0.0, 0))),
	m_FractalType("fractaltype", FBM, 0, 0),
	m_ImageType("imagetype", BLUE, 0, 0),
	m_Bitmap("bitmap", Document(), Document()->StateRecorder(), new k3dEventConnector(this, k3dIBitmap::Event_Changed, "CDisplace", 0)),
	m_DisplaceType("displacetype", RADIAL, 0, 0),
	m_Page(FRACTAL)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CDisplace::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	return CDisplaceBase::InitComplete();
}

sdpResult CDisplace::EditObject()
{
	new CDisplaceDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CDisplace::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIBitmap::Event_Changed)
		{
			PushEditing();
			UpdateGeometry();
			PopEditing();
			return SDP_S_OK;
		}
	
	return CDisplaceBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CDisplace::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CDisplace::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.1;
	double yfactor = (CurrentY - LastY) * 0.1;

	k3dSetValue(m_Amplitude, m_Amplitude.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CDisplace::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CDisplace::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CDisplace::UpdateGeometry()
{
	if( m_Page == FRACTAL )
		{
			m_Function.SetFractal( m_FractalType.Value(), static_cast<const gulong>(m_Seed.Value()), m_Smoothness.Value(), m_Frequency.Value(), m_Octaves.Value() );
		}
	else
		{
			if(m_Bitmap.Interface())
				{
					k3dPixel* const data = m_Bitmap.Interface()->Data();
					const gulong width = m_Bitmap.Interface()->Width();
					const gulong height = m_Bitmap.Interface()->Height();
		
					m_Function.SetImage( m_ImageType.Value(), data, width, height );
				}
			else
				{
					// No image 
					m_Function.SetNone();
				}
		}

	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), DisplaceLocations(&m_Function, m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), m_Amplitude.Value(), m_DisplaceType.Value()));
}

void CDisplace::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CDisplace::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Displace");
}

////////////////////////////////////////////////////////////////////////////
// CDisplaceDialog

const sdpString CDisplaceDialog::dialogname = "properties";
const sdpString CDisplaceDialog::control_amplitude = "amplitude";
const sdpString CDisplaceDialog::control_displacex = "displacex";
const sdpString CDisplaceDialog::control_displacey = "displacey";
const sdpString CDisplaceDialog::control_displacez = "displacez";
const sdpString CDisplaceDialog::control_smoothness = "smoothness";
const sdpString CDisplaceDialog::control_frequency = "frequency";
const sdpString CDisplaceDialog::control_octaves = "octaves";
const sdpString CDisplaceDialog::control_fractaltype = "fractaltype";
const sdpString CDisplaceDialog::control_randomseed = "randomseed";
const sdpString CDisplaceDialog::control_displacetype = "displacetype";
const sdpString CDisplaceDialog::control_imagetype = "imagetype";
const sdpString CDisplaceDialog::control_bitmap = "bitmap";
const sdpString CDisplaceDialog::control_pageswitch = "pageswitch";
const sdpString CDisplaceDialog::control_pagefractal = "pagefractal";
const sdpString CDisplaceDialog::control_pageimage = "pageimage";

CDisplaceDialog::CDisplaceDialog(CDisplace* const Object) :
	CDisplaceDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "displace.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));

	// Connect spin buttons
	Connectk3dSpinButton(control_amplitude, m_Object->m_Amplitude, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	Connectk3dSpinButton(control_smoothness, m_Object->m_Smoothness, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	Connectk3dSpinButton(control_frequency, m_Object->m_Frequency, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));	
	Connectk3dSpinButton(control_octaves, m_Object->m_Octaves, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));	
	Connectk3dSpinButton(control_randomseed, m_Object->m_Seed, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));

	// Initialise our page ...
	Notebook("tabs").SetPage(m_Object->m_Page);

	// Attach chooser 
	if( Getk3dChooser( control_fractaltype ) )
	{
		Getk3dChooser( control_fractaltype )->Attach(
			new k3dChooserVariableDataSource<FractalType>(m_Object->m_FractalType),
			k3dApplication()->CommandRecorder(),
			0,
			"Fractal Type",
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	}
	if( Getk3dChooser( control_imagetype ) )
	{
		Getk3dChooser( control_imagetype )->Attach(
			new k3dChooserVariableDataSource<ImageType>(m_Object->m_ImageType),
			k3dApplication()->CommandRecorder(),
			0,
			"Image Type",
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	}
	if( Getk3dChooser( control_displacetype ) )
	{
		Getk3dChooser( control_displacetype )->Attach(
			new k3dChooserVariableDataSource<CDisplace::DisplaceType>(m_Object->m_DisplaceType),
			k3dApplication()->CommandRecorder(),
			0,
			"Displace Type",
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0));
	}

	// Attach object selectors ...
	if(Getk3dObjectSelector(control_bitmap))
		Getk3dObjectSelector(control_bitmap)->Attach(
			&m_Object->m_Bitmap,
			k3dApplication()->CommandRecorder(),
			m_Object->Document()->StateRecorder(),
			"Select Bitmap",
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PushEditing, 0),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)),
			new k3dActionMethodCall<CDisplace>(m_Object, &CDisplace::PopEditing, 0),
			true,
			new k3dObjectSelectorInterfaceFilter(new sdp::TypeInfo<k3dIBitmap>()));

	CallUpdateControls();
	RootWidget().Show();
}

void CDisplaceDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_pageswitch)
		OnPageSwitch(Event);
	else
		CDisplaceDialogBase::OnEvent(Event);
}

void CDisplaceDialog::OnPageSwitch( sdpGtkEvent* Event )
{
	// Sanity checks ...
	g_assert( Event );
	
	// Keep track of the current page ...
	sdpGtkEventNotebookSwitchPage* event = (sdpGtkEventNotebookSwitchPage*)Event;
	CDisplace::Page currentpage = CDisplace::Page(event->PageNumber());
	
	sdpString pagename;
	switch( currentpage )
		{
			case CDisplace::FRACTAL:
				pagename = control_pagefractal;
				break;
			case CDisplace::IMAGE:
				pagename = control_pageimage;
				break;
			default:
				g_return_if_fail(0);
		}
	
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_pageswitch, pagename);
	
	m_Object->m_Page = currentpage;
	m_Object->UpdateGeometry();
	k3dRedrawAll(m_Object->Document(), false);
}


