#ifndef DISPLACE_H
#define DISPLACE_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CDisplace K-3D tool, which displaces geometry
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "fractals.h"

#include <k3dmouseeventobserver.h>
#include <k3dobjectdialog.h>
#include <k3dselection.h>
#include <k3dtool.h>

#include <k3dibitmap.h>

// Forward class references
class CDisplace;
class CDisplaceFunction;

/////////////////////////////////////////////////////////////////////////////
// CDisplacePage dialog

typedef k3dObjectDialog CDisplaceDialogBase;

class CDisplaceDialog :
	public CDisplaceDialogBase
{
public:
	CDisplaceDialog(CDisplace* const Object);

private:
	void OnEvent(sdpGtkEvent* Event);
	void OnPageSwitch(sdpGtkEvent* Event);

	CDisplace* const m_Object;

	static const sdpString dialogname;
	static const sdpString control_amplitude;
	static const sdpString control_displacex;
	static const sdpString control_displacey;
	static const sdpString control_displacez;
	static const sdpString control_smoothness;
	static const sdpString control_frequency;
	static const sdpString control_octaves;
	static const sdpString control_fractaltype;
	static const sdpString control_randomseed;
	static const sdpString control_displacetype;
	static const sdpString control_imagetype;
	static const sdpString control_bitmap;
	
	static const sdpString control_pageswitch;
	static const sdpString control_pagefractal;
	static const sdpString control_pageimage;
};

typedef enum
{
	RANDOM,
	NOISE,
	FBM,
	TURBULENCE,
	RIDGED,
} FractalType;

inline std::ostream& operator<<(std::ostream& Stream, const FractalType& Value)
{
	switch(Value)
		{
			case RANDOM:
				Stream << "random";
				break;
			case NOISE:
				Stream << "noise";
				break;
			case FBM:
				Stream << "fBm";
				break;
			case TURBULENCE:
				Stream << "turbulence";
				break;
			case RIDGED:
				Stream << "ridged";
				break;
			default:
				g_return_val_if_fail(0, Stream);
		}
	
	return Stream;
}
		
inline std::istream& operator>>(std::istream& Stream, FractalType& Value)
{
	sdpString text;
	Stream >> text;

	if( text == "random" )
		Value = RANDOM;
	else if( text == "noise" )
		Value = NOISE;
	else if( text == "fBm" )
		Value = FBM;
	else if( text == "turbulence" )
		Value = TURBULENCE;
	else if( text == "ridged" )
		Value = RIDGED;
	else
		std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
	
	return Stream;
}

typedef enum
{
	AVERAGE,
	RED,
	GREEN,
	BLUE,
	ALPHA,
} ImageType;

inline std::ostream& operator<<(std::ostream& Stream, const ImageType& Value)
{
	switch( Value )
		{
			case AVERAGE:
				Stream << "average";
				break;
			case RED:
				Stream << "red";
				break;
			case GREEN:
				Stream << "green";
				break;
			case BLUE:
				Stream << "blue";
				break;
			case ALPHA:
				Stream << "alpha";
				break;
		}
	
	return Stream;
}
		
inline std::istream& operator>>(std::istream& Stream, ImageType& Value)
{
	sdpString text;
	Stream >> text;

	if( text == "average" )
		Value = AVERAGE;
	else if( text == "red" )
		Value = RED;
	else if( text == "green" )
		Value = GREEN;
	else if( text == "blue" )
		Value = BLUE;
	else if( text == "alpha" )
		Value = ALPHA;
	else
		std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
	
	return Stream;
}

/////////////////////////////////////////////////////////////////////////////
// CDisplaceFunction

class CDisplaceFunction
{
public:
	CDisplaceFunction();
	~CDisplaceFunction();
	
	void SetFractal( const FractalType FType, const gulong Seed, const gdouble Smoothness, const gdouble Frequency, const gdouble Octaves );
	void SetImage( const ImageType IType, k3dPixel* const Data, const gulong Width, const gulong Height );
	void SetNone();
	
	gdouble Value(k3d::SelectedLocationProxy& Location);

protected:
	typedef enum
	{
		FRACTAL,
		IMAGE,
		NONE,
	} Type;
	
	Type m_Type;
	ImageType m_ImageType;
	
	CFractal* m_Fractal;
	gdouble m_Frequency;
	
	k3dPixel* m_Data;
	gulong m_Width;
	gulong m_Height;
};

/////////////////////////////////////////////////////////////////////////////
// CDisplace

typedef k3dTool CDisplaceBase;

class CDisplace :
	public CDisplaceBase,
	public k3dMouseEventObserver
{
friend class CDisplaceDialog;
public:
	CDisplace(sdpIUnknown* Initializer);

	// k3dIObject implementation
	sdpResult InitComplete();
	sdpResult EditObject();

	// k3dIEventObserver implementation
	sdpResult ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData);

	// k3dIMouseEventObserver implementation
	sdpResult OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

	void PushEditing();
	void UpdateGeometry();
	void CancelEditing();
	void PopEditing();

	typedef enum
	{
		LINEAR,
		RADIAL,
	} DisplaceType;

	friend std::ostream& operator<<(std::ostream& Stream, const DisplaceType& Value)
	{
		switch(Value)
			{
				case LINEAR:
					Stream << "linear";
					break;
				case RADIAL:
					Stream << "radial";
					break;
			}
			
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, DisplaceType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "linear")
			Value = LINEAR;
		else if(text == "radial")
			Value = RADIAL;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
		
	typedef enum
	{
		FRACTAL,
		IMAGE,
	} Page;
	
private:
	k3dVariable<bool> m_DisplaceX;
	k3dVariable<bool> m_DisplaceY;
	k3dVariable<bool> m_DisplaceZ;
	k3dVariable<gdouble> m_Amplitude;
	k3dVariable<gdouble> m_Smoothness;
	k3dVariable<gdouble> m_Frequency;
	k3dVariable<gdouble> m_Octaves;
	k3dVariable<gdouble> m_Seed;
	k3dVariable<FractalType> m_FractalType;
	k3dVariable<ImageType> m_ImageType;

	k3dObjectInterfaceVariable<k3dIBitmap> m_Bitmap;

	k3dVariable<DisplaceType> m_DisplaceType;
	Page m_Page;
	
	CDisplaceFunction m_Function;
	
	k3d::SelectedLocationArray m_CurrentSelection;
};



#endif // DISPLACE_H
