// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CExtrude, a K-3D tool for creating ducted solids from existing geometry
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "extrude.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>
#include <k3dicamera.h>
#include <k3dilocation.h>
#include <k3dimaterialcollection.h>
#include <k3dipathpointcollection.h>

#include <sdpgtkevents.h>
#include <sdpgl.h>

#include <algorithm>
#include <iterator>
#include <iostream>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CExtrude::PathProxy

CExtrude::PathProxy::PathProxy(CExtrude& Object, k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PathProxyBase(Engine, Mesh),
	m_Object(Object)
{
	TouchEdges();
}

CExtrude::PathProxy::PathProxy(CExtrude& Object, k3d::GeometryEngine& Engine, k3dIPath* const Path) :
	PathProxyBase(Engine, Path),
	m_Object(Object)
{
	TouchEdges();
}

void CExtrude::PathProxy::TouchEdges()
{
	for(EdgeCollection::iterator edge = m_Edges.begin(); edge != m_Edges.end(); ++edge)
		static_cast<EdgeProxy*>(*edge)->AddPath();
}

void CExtrude::PathProxy::BuildGeometry()
{
	// Create a new path for each segment in the resulting ducted solid ...
	const gulong segments = static_cast<gulong>(m_Object.m_Segments.Value());
	for(gulong i = 1; i <= segments; i++)
		{
			if(segments == i)
				{
					 if(!m_Object.m_CreateEndCap.Value())
						continue;
				}
			else
				{
					if(!m_Object.m_CreateLayers.Value())
						continue;	
				}
				
			k3d::EnginePathProxy& newpath = Engine().Path(Mesh());

			for(PointCollection::iterator point = m_Points.begin(); point != m_Points.end(); ++point)
				newpath.Actual()->Points()->AddPoint(static_cast<PointProxy*>(*point)->LayerPoint(i).Actual());
									
			m_LayerPaths.push_back(&newpath);
		}

	PathProxyBase::BuildGeometry();
}

void CExtrude::PathProxy::Cancel()
{
	m_LayerPaths.clear();
	PathProxyBase::Cancel();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CExtrude::PointProxy

CExtrude::PointProxy::PointProxy(CExtrude& Object, k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PointProxyBase(Engine, Mesh),
	m_Object(Object)
{
}

CExtrude::PointProxy::PointProxy(CExtrude& Object, k3d::GeometryEngine& Engine, k3dIPoint* const Point) :
	PointProxyBase(Engine, Point),
	m_Object(Object)
{
}

CExtrude::PointProxy& CExtrude::PointProxy::LayerPoint(const gulong Index)
{
	if(0 == m_LayerPoints.size())
		m_LayerPoints.push_back(this);
		
	while(m_LayerPoints.size() <= Index)
		m_LayerPoints.push_back(0);
		
	if(0 == m_LayerPoints[Index])
		m_LayerPoints[Index] = &Engine().Point(Mesh());
		
	return *static_cast<PointProxy*>(m_LayerPoints[Index]);
}

void CExtrude::PointProxy::UpdateGeometry()
{
	gulong index = 0;
	const gulong segments = static_cast<gulong>(m_Object.m_Segments.Value());
	const sdpVector3 originalposition = Actual()->Location()->LocalXYZ();
	
	for(PointCollection::iterator point = m_LayerPoints.begin(); point != m_LayerPoints.end(); ++point, ++index)
		{
			// The first point is this one, so skip it ...
			if(!index)
				continue;
				
			// Skip empty points ...
			if(0 == *point)
				continue;
				
			const gdouble percent = static_cast<gdouble>(index) / static_cast<gdouble>(segments);

			// Interpolate our transformation values ...
			const sdpVector3 offset = percent * sdpVector3(m_Object.m_OffsetX.Value(), m_Object.m_OffsetY.Value(), m_Object.m_OffsetZ.Value());
			const sdpVector3 scale = sdpLerp(sdpVector3(1, 1, 1), sdpVector3(m_Object.m_ScaleX.Value(), m_Object.m_ScaleY.Value(), m_Object.m_ScaleZ.Value()), percent);
			const sdpVector3 rotation = sdpLerp(sdpVector3(0, 0, 0), sdpVector3(m_Object.m_RotationX.Value(), m_Object.m_RotationY.Value(), m_Object.m_RotationZ.Value()), percent);

			// Create a transformation matrix ...
			const sdpMatrix4 matrix = sdpTranslation3D(offset) * sdpRotation3D(rotation) * sdpScaling3D(scale);

			// Calculate points ...
			(*point)->Actual()->Location()->SetLocalXYZ(matrix * originalposition);
		}

	PointProxyBase::UpdateGeometry();
}

void CExtrude::PointProxy::Cancel()
{
	m_LayerPoints.clear();
	PointProxyBase::Cancel();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// CExtrude::EdgeProxy

CExtrude::EdgeProxy::EdgeProxy(CExtrude& Object, k3d::GeometryEngine& Engine, k3d::EnginePointProxy& Point1, k3d::EnginePointProxy& Point2, const bool Original) :
	EdgeProxyBase(Engine, Point1, Point2, Original),
	m_Object(Object),
	m_PathCount(0)
{
}

void CExtrude::EdgeProxy::AddPath()
{
	// Keep track of how many paths share this edge, so we figure out if it's interior or exterior
	++m_PathCount;
}

void CExtrude::EdgeProxy::BuildGeometry()
{
	// See if we need to create geometry or not ...
	switch(m_PathCount)
		{
			case 1:
				if(!m_Object.m_ConnectExterior.Value())
					return;
				break;
			case 2:
				if(!m_Object.m_ConnectInterior.Value())
					return;
				break;
			default:
				g_return_if_fail(0); // Should never happen!
		}

	// Create a new path for each segment in the resulting ducted solid ...
	const gulong segments = static_cast<gulong>(m_Object.m_Segments.Value());
	for(gulong i = 0; i < segments; i++)
		{
			k3d::EnginePathProxy& newpath = Engine().Path(Point1().Mesh());

			newpath.Actual()->Points()->AddPoint(static_cast<PointProxy&>(Point1()).LayerPoint(i+1).Actual());
			newpath.Actual()->Points()->AddPoint(static_cast<PointProxy&>(Point2()).LayerPoint(i+1).Actual());
			newpath.Actual()->Points()->AddPoint(static_cast<PointProxy&>(Point2()).LayerPoint(i).Actual());
			newpath.Actual()->Points()->AddPoint(static_cast<PointProxy&>(Point1()).LayerPoint(i).Actual());
									
			m_LayerPaths.push_back(&newpath);
		}
	
	EdgeProxyBase::BuildGeometry();
}

void CExtrude::EdgeProxy::Cancel()
{
	m_LayerPaths.clear();
	EdgeProxyBase::Cancel();
}

///////////////////////////////////////////////////////////////
// CExtrude

CExtrude::CExtrude(sdpIUnknown* Initializer) :
	CExtrudeBase(Initializer),
	k3dMouseEventObserver("LMB-drag to offset/scale/rotate extrusion; control-LMB-drag to constrain to Active Axis."),
	m_Segments("segments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1.0, 0))),
	m_ConnectExterior("connectexterior", true, 0, 0),
	m_ConnectInterior("connectinterior", false, 0, 0),
	m_CreateLayers("createlayers", false, 0, 0),
	m_CreateEndCap("createendcap", true, 0, 0),
	m_Sensitivity("sensitivity", 0.1, 0, 0),
	m_OffsetX("offsetx", 0.0, 0, 0),
	m_OffsetY("offsety", 0.0, 0, 0),
	m_OffsetZ("offsetz", 10.0, 0, 0),
	m_ScaleX("scalex", 1.0, 0, 0),
	m_ScaleY("scaley", 1.0, 0, 0),
	m_ScaleZ("scalez", 1.0, 0, 0),
	m_RotationX("rotationx", 0.0, 0, 0),
	m_RotationY("rotationy", 0.0, 0, 0),
	m_RotationZ("rotationz", 0.0, 0, 0),
	m_InteractiveMode(EXTRUDE),
	m_Engine(
		Document(),
		new k3d::EngineMeshProxyFactory<k3d::EngineMeshProxy>(),
		new k3d::EnginePathProxyFactory2<CExtrude, CExtrude::PathProxy>(*this),
		new k3d::EnginePointProxyFactory2<CExtrude, CExtrude::PointProxy>(*this),
		new k3d::EngineEdgeProxyFactory2<CExtrude, CExtrude::EdgeProxy>(*this)
		)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CExtrude::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedPaths(Document()->Objects(), std::back_inserter(m_Selection));
	std::copy(m_Selection.begin(), m_Selection.end(), k3d::GeometryEngine::Inserter(m_Engine));

	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return CExtrudeBase::InitComplete();
}

sdpResult CExtrude::EditObject()
{
	new Properties(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CExtrude::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	return SDP_S_OK;
}

sdpResult CExtrude::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	gdouble xfactor = (CurrentX - LastX) * 0.01;
	gdouble yfactor = (CurrentY - LastY) * 0.01;

	switch(m_InteractiveMode)
		{
			case EXTRUDE:
				break;
			case OFFSETS:
				// If the user holds down either control key, only translate along the active axis ...
				if(ControlDown(Modifiers))
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_OffsetX, m_OffsetX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_OffsetY, m_OffsetY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_OffsetZ, m_OffsetZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				// Otherwise, translate along the two non-active axes ...
				else
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_OffsetZ, m_OffsetZ.Value() - xfactor);
									k3dSetValue(m_OffsetY, m_OffsetY.Value() - yfactor);
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_OffsetX, m_OffsetX.Value() - xfactor);
									k3dSetValue(m_OffsetZ, m_OffsetZ.Value() - yfactor);
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_OffsetX, m_OffsetX.Value() + xfactor);
									k3dSetValue(m_OffsetY, m_OffsetY.Value() - yfactor);
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				break;
			case SCALING:
				// If the user holds down either shift key, constrain scale factors to be equal ...
				if(ShiftDown(Modifiers))
					{
						k3dSetValue(m_ScaleX, m_ScaleX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						k3dSetValue(m_ScaleY, m_ScaleY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
					}
				// If the user holds down either control key, only scale along the active axis ...
				else if(ControlDown(Modifiers))
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_ScaleX, m_ScaleX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_ScaleY, m_ScaleY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				// Otherwise, scale along the two non-active axes ...
				else
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - xfactor);
									k3dSetValue(m_ScaleY, m_ScaleY.Value() - yfactor);
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_ScaleX, m_ScaleX.Value() - xfactor);
									k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - yfactor);
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_ScaleX, m_ScaleX.Value() - xfactor);
									k3dSetValue(m_ScaleY, m_ScaleY.Value() - yfactor);
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				break;
			case ROTATION:
				// Compute scaling factors for each mouse axis ...
				xfactor = sdpRadians(gdouble(CurrentX - LastX) * 0.1);
				yfactor = sdpRadians(gdouble(CurrentY - LastY) * 0.1);

				// If the user holds down either control key, only rotate along the active axis ...
				if(ControlDown(Modifiers))
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_RotationY, m_RotationY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_RotationZ, m_RotationZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_RotationX, m_RotationX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				// Otherwise, rotate along the two non-active axes ...
				else
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_RotationX, m_RotationX.Value() - xfactor);
									k3dSetValue(m_RotationZ, m_RotationZ.Value() - yfactor);
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_RotationY, m_RotationY.Value() - xfactor);
									k3dSetValue(m_RotationX, m_RotationX.Value() - yfactor);
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_RotationY, m_RotationY.Value() + xfactor);
									k3dSetValue(m_RotationZ, m_RotationZ.Value() - yfactor);
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				break;
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}

	// Update everything ...
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CExtrude::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	return SDP_S_OK;
}

void CExtrude::BuildGeometry()
{
	// Get rid of any old geometry
	m_Engine.Cancel();

	// Create new geometry ...
	m_Engine.StartChangeSet();
	m_Engine.BuildGeometry();
	m_Engine.Synchronize();
	m_Engine.FinishChangeSet();
	
	// Update it ...
	m_Engine.UpdateGeometry();
	m_Engine.Manage();

	// Notify UI ...
	Notify(k3dObject::Event_UserMessage1, m_Engine.Description());

std::cerr << g_pretty_function() << std::endl << m_Engine << std::endl;
}

void CExtrude::UpdateGeometry()
{
	m_Engine.UpdateGeometry();
}

void CExtrude::CancelGeometry()
{
	m_Engine.Cancel();
}

void CExtrude::FinishGeometry()
{
	m_Engine.Finish("Extrude Selected");
}

////////////////////////////////////////////////////////////////////////////
// CExtrude::Properties

CExtrude::Properties::Properties(CExtrude* const Object) :
	PropertiesBase(Object, true, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "extrude.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton("connectexterior", m_Object->m_ConnectExterior, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton("connectinterior", m_Object->m_ConnectInterior, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton("createlayers", m_Object->m_CreateLayers, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton("createendcap", m_Object->m_CreateEndCap, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Connect spin buttons ...
	Connectk3dSpinButton("segments", m_Object->m_Segments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	
	Connectk3dSpinButton("offsetx", m_Object->m_OffsetX, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("offsety", m_Object->m_OffsetY, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("offsetz", m_Object->m_OffsetZ, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("scalex", m_Object->m_ScaleX, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("scaley", m_Object->m_ScaleY, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("scalez", m_Object->m_ScaleZ, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("rotationx", m_Object->m_RotationX, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("rotationy", m_Object->m_RotationY, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("rotationz", m_Object->m_RotationZ, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CExtrude>(m_Object, &CExtrude::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Notebook("tabs").SetPage(m_Object->m_InteractiveMode);

	Show();
}

void CExtrude::Properties::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "pageswitch")
		OnPageSwitch(static_cast<sdpGtkEventNotebookSwitchPage*>(Event));
	else
		PropertiesBase::OnEvent(Event);
}

void CExtrude::Properties::OnPageSwitch(sdpGtkEventNotebookSwitchPage* const Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Keep track of the current page ...
	m_Object->m_InteractiveMode = static_cast<CExtrude::InteractiveModeType>(Event->PageNumber());
}

void CExtrude::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);

	PropertiesBase::OnOK();
}

void CExtrude::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);

	PropertiesBase::OnCancel();
}

