// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CFreehand, a K-3D tool for creating freehand geometry
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "freehand.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dboundingbox.h>

#include <k3distaterecorder.h>
#include <k3dicamera.h>
#include <k3dipathpointcollection.h>
#include <k3dicamera.h>
#include <k3dilocation.h>

#include <sdpgtkevent.h>

#include <iostream>

namespace
{

class FreehandPoints
{
public:
	FreehandPoints(sdpVector3Array& Points) :
		m_Points(Points),
		m_CurrentPoint(m_Points.begin())
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		// Sanity checks ...
		g_return_if_fail(m_CurrentPoint != m_Points.end());
		
		(*Point)->Location()->SetLocalXYZ(*m_CurrentPoint);
		
		m_CurrentPoint++;
	}
	
private:
	sdpVector3Array& m_Points;
	sdpVector3Array::iterator m_CurrentPoint;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CFreehand

CFreehand::CFreehand(sdpIUnknown* Initializer) :
	CFreehandBase(Initializer),
	k3dMouseEventObserver("LMB-click to create Points")
{
	// Make sure that we get called by the rendering engine ...
	SetState(k3dIState::VISIBLE, true);
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CFreehand::EditObject()
{
	new Properties(this);

	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender implementation

sdpResult CFreehand::EditorPostRender(k3dIEditorEngine* const Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glColor3d(0.0, 0.0, 1.0);
	glPointSize(6.0f);

	// Draw a plane to help the user orient themselves ...
	
	// First, create a bounding box ...
	k3dBoundingBox3 box;
	box.Expand(m_MouseCursor);
	
	for(gulong i = 0; i < m_PathPoints.size(); i++)
		box.Expand(m_PathPoints[i]);

	// Put a border around it ...
	const gdouble border = 1.0;
	
	sdpVector3 min = box.Min();
	sdpVector3 max = box.Max();
	
	min += sdpVector3(-border, -border, -border);
	max += sdpVector3(border, border, border);

	// Draw it ...			
	glBegin(GL_LINE_LOOP);
	
/*
	switch(Camera->ActiveAxis())
		{
			case k3dICamera::XAXIS:
				glVertex3d(0, min[1], max[2]);
				glVertex3d(0, max[1], max[2]);
				glVertex3d(0, max[1], min[2]);
				glVertex3d(0, min[1], min[2]);
				break;
				
			case k3dICamera::YAXIS:
				glVertex3d(min[0], 0, max[2]);
				glVertex3d(max[0], 0, max[2]);
				glVertex3d(max[0], 0, min[2]);
				glVertex3d(min[0], 0, min[2]);
				break;
				
			case k3dICamera::ZAXIS:
				glVertex3d(min[0], max[1], 0);
				glVertex3d(max[0], max[1], 0);
				glVertex3d(max[0], min[1], 0);
				glVertex3d(min[0], min[1], 0);
				break;
				
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}
*/

	glEnd();

	// Draw the vertices selected by the user
	glBegin(GL_POINTS);
	for(gulong j = 0; j < m_PathPoints.size(); j++)
		glVertex3dv(m_PathPoints[j]);
	glEnd();

	// Draw the current mouse cursor
	glColor3d(1.0, 0.0, 0.0);
	glBegin(GL_POINTS);
		glVertex3dv(m_MouseCursor);
	glEnd();

	glPopAttrib();	// GL_ALL_ATTRIB_BITS

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CFreehand::OnMouseMove(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CFreehand::OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);

	// Save our point
	m_PathPoints.push_back(m_MouseCursor);

	// Update the geometry
	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CFreehand::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);
	k3dRedrawAll(Document(), false);
	
	return SDP_S_OK;
}

sdpResult CFreehand::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	m_MouseCursor = ScreenToWorld(Camera, Modifiers, CurrentX, CurrentY);
	k3dRedrawAll(Document(), false);
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CFreehand message handlers

void CFreehand::BuildGeometry()
{
	// Get rid of any old geometry
	CancelGeometry();

	g_return_if_fail(m_PathPoints.size());

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(Document());
	k3d::CreateNewMesh(Document(), "Freehand", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(Document());

	// We add a single path and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewPath(**m_NewGeometry.Meshes().begin(), m_PathPoints.size(), m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry();
	m_NewGeometry.Manage();	
	
	Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CFreehand::UpdateGeometry()
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), FreehandPoints(m_PathPoints));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CFreehand::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CFreehand::FinishGeometry()
{
	m_NewGeometry.Finish(Document(), "Create Freehand Polygon");
}

sdpVector3 CFreehand::ScreenToWorld(k3dICamera* const Camera, const GdkModifierType Modifiers, const gdouble CurrentX, const gdouble CurrentY)
{
	// Sanity checks ...
	g_assert(Camera);

	// Turn mouse coordinates into a vector ...
	sdpVector3 mouseorigin, mousedirection;
	k3dMouseToWorldRay(Camera, sdpVector2(CurrentX, CurrentY), mouseorigin, mousedirection);

	// Intersect the vector with the active plane ...
	sdpVector3 result;

	switch(Camera->ActiveAxis())
		{
			case k3dICamera::XAXIS:
				sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), 0), mouseorigin, mousedirection, result);
				break;
			case k3dICamera::YAXIS:
				sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), 0), mouseorigin, mousedirection, result);
				break;
			case k3dICamera::ZAXIS:
				sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), 0), mouseorigin, mousedirection, result);
				break;
			default:
				g_return_val_if_fail(0, result);
		}

	return result;
}

//////////////////////////////////////////////////////////////////////////////////////
// CFreehand::Properties

CFreehand::Properties::Properties(CFreehand* const Object) :
	PropertiesBase(Object, true, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "freehand.gtkml"));

	Show();
}

void CFreehand::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnOK();
}

void CFreehand::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnCancel();
}


