// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CKnot, a K-3D tool which generates knot paths
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "knots.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dipathpointcollection.h>
#include <k3distaterecorder.h>
#include <k3dilocation.h>

#include <sdpgtkevent.h>

namespace
{

class KnotPoints
{
public:
	KnotPoints(const gdouble MeridianWraps, const gdouble LongitudinalWraps, const gdouble Scale) :
		m_MeridianWraps(MeridianWraps),
		m_LongitudinalWraps(LongitudinalWraps),
		m_Scale(Scale)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const gdouble percent = Point->Parameters()[0];
		const gdouble radians = percent * sdp::PiTimes2();
		const gdouble mu = radians * m_MeridianWraps;
		
		sdpVector3 coords;
		
		coords[0] = cos(mu) * (1 + cos(m_LongitudinalWraps * mu / m_MeridianWraps) / 2.0);
		coords[1] = sin(mu) * (1 + cos(m_LongitudinalWraps * mu / m_MeridianWraps) / 2.0);
		coords[2] = sin(m_LongitudinalWraps * mu / m_MeridianWraps) / 2.0;

		coords *= m_Scale;

		(*Point)->Location()->SetLocalXYZ(coords);
	}
	
private:
	const gdouble m_MeridianWraps;
	const gdouble m_LongitudinalWraps;
	const gdouble m_Scale;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CKnots

CKnots::CKnots(sdpIUnknown* Initializer) :
	CKnotsBase(Initializer),
	m_EdgeCount("edgecount", 100.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3.0, 0))),
	m_MeridianWraps("meridianwraps", 3.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1.0, 0))),
	m_LongitudinalWraps("longitudinalwraps", 4.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1.0, 0))),
	m_Scale("scale", 5.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CKnots::EditObject()
{
	new Properties(this);

	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;

}

void CKnots::BuildGeometry()
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong pointcount = static_cast<gulong>(m_EdgeCount.Value());
	g_return_if_fail(pointcount);
	
	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(Document());
	k3d::CreateNewMesh(Document(), "Knot", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(Document());

	// We add a single path and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewPath(**m_NewGeometry.Meshes().begin(), pointcount, m_NewGeometry);
	m_NewGeometry.PopEditing();
	
	UpdateGeometry();
	m_NewGeometry.Manage();	
	
	Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CKnots::UpdateGeometry()
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), KnotPoints(m_MeridianWraps.Value(), m_LongitudinalWraps.Value(), m_Scale.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CKnots::CancelGeometry()
{
	// Get rid of old geometry
	m_NewGeometry.Cancel();
}

void CKnots::FinishGeometry()
{
	// Make everything permanent
	m_NewGeometry.Finish(Document(), "Create Knot Curve");
}

/////////////////////////////////////////////////////////////////////////////
// CKnots::Properties

CKnots::Properties::Properties(CKnots* const Object) :
	PropertiesBase(Object, false, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)

{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "knots.gtkml"));

	// Connect spin buttons ...
	Connectk3dSpinButton("edgecount", m_Object->m_EdgeCount, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CKnots>(m_Object, &CKnots::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("meridianwraps", m_Object->m_MeridianWraps, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CKnots>(m_Object, &CKnots::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("longitudinalwraps", m_Object->m_LongitudinalWraps, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CKnots>(m_Object, &CKnots::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("scale", m_Object->m_Scale, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CKnots>(m_Object, &CKnots::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Show();
}

void CKnots::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnOK();
}

void CKnots::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnCancel();
}

