// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CLissaCurve K-3D tool, which generates geometry composed of Lissajous curves
		\author Tim Shead (tshead@k-3d.com)
*/

#include "lissacurve.h"

#include <k3dbuildgeometry.h>
#include <k3dmisc.h>
#include <k3dobjectdialog.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dipathpointcollection.h>
#include <k3distaterecorder.h>
#include <k3dilocation.h>

#include <sdpgtkevent.h>

namespace
{

/////////////////////////////////////////////////////////////////////////////
// LissaCurvePoints

class LissaCurvePoints
{
public:
	LissaCurvePoints(const gdouble XAmp, const gdouble XFreq, const gdouble XPhase, const gdouble YAmp, const gdouble YFreq, const gdouble YPhase, const gdouble ZAmp, const gdouble ZFreq, const gdouble ZPhase, const gdouble MAmp, const gdouble MFreq, const gdouble MPhase) :
		m_XAmp(XAmp),
		m_XFreq(XFreq),
		m_XPhase(XPhase),
		m_YAmp(YAmp),
		m_YFreq(YFreq),
		m_YPhase(YPhase),
		m_ZAmp(ZAmp),
		m_ZFreq(ZFreq),
		m_ZPhase(ZPhase),
		m_MAmp(MAmp),
		m_MFreq(MFreq),
		m_MPhase(MPhase)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const gdouble percent = Point->Parameters()[0];
		const gdouble radians = percent * sdp::PiTimes2();

		sdpVector3 coords(m_XAmp * sin(m_XFreq * radians + m_XPhase), m_YAmp * sin(m_YFreq * radians + m_YPhase), m_ZAmp * sin(m_ZFreq * radians + m_ZPhase));

		if(m_MAmp)
			coords *= 1.0 + (m_MAmp * sin(m_MFreq * radians + m_MPhase));

		(*Point)->Location()->SetLocalXYZ(coords);
	}
		
private:
	const gdouble m_XAmp;
	const gdouble m_XFreq;
	const gdouble m_XPhase;
	const gdouble m_YAmp;
	const gdouble m_YFreq;
	const gdouble m_YPhase;
	const gdouble m_ZAmp;
	const gdouble m_ZFreq;
	const gdouble m_ZPhase;
	const gdouble m_MAmp;
	const gdouble m_MFreq;
	const gdouble m_MPhase;
};
	
} // namespace

/////////////////////////////////////////////////////////////////////////////
// CLissaCurve

CLissaCurve::CLissaCurve(sdpIUnknown* Initializer) :
	CLissaCurveBase(Initializer),
	m_EdgeCount("edgecount", 100.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3.0, 0))),
	m_MAmp("mamp", 0.0, 0, 0),
	m_MFreq("mfreq", 20.0, 0, 0),
	m_MPhase("mphase", 0.0, 0, 0),
	m_XAmp("xamp", 5.0, 0, 0),
	m_XFreq("xfreq", 1.0, 0, 0),
	m_XPhase("xphase", 0.0, 0, 0),
	m_YAmp("yamp", 5.0, 0, 0),
	m_YFreq("yfreq", 1.0, 0, 0),
	m_YPhase("yphase", sdp::PiOver2(), 0, 0),
	m_ZAmp("zamp", 0.0, 0, 0),
	m_ZFreq("zfreq", 1.0, 0, 0),
	m_ZPhase("zphase", 0.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CLissaCurve::EditObject()
{
	new Properties(this);

	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

void CLissaCurve::BuildGeometry()
{
	// Get rid of any old geometry
	m_NewGeometry.Cancel();

	// Sanity checks ...
	const gulong pointcount = static_cast<gulong>(m_EdgeCount.Value());
	g_return_if_fail(pointcount);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(Document());
	k3d::CreateNewMesh(Document(), "LissaCurve", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(Document());
	
	// We add a single path and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewPath(**m_NewGeometry.Meshes().begin(), pointcount, m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry();
	m_NewGeometry.Manage();	
	
	Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CLissaCurve::UpdateGeometry()
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), LissaCurvePoints(m_XAmp.Value(), m_XFreq.Value(), m_XPhase.Value(), m_YAmp.Value(), m_YFreq.Value(), m_YPhase.Value(), m_ZAmp.Value(), m_ZFreq.Value(), m_ZPhase.Value(), m_MAmp.Value(), m_MFreq.Value(), m_MPhase.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CLissaCurve::CancelGeometry()
{
	// Get rid of old geometry
	m_NewGeometry.Cancel();
}

void CLissaCurve::FinishGeometry()
{
	// Make everything permanent
	m_NewGeometry.Finish(Document(), "Create LissaCurve");
}

///////////////////////////////////////////////////////////////////////////
// CLissaCurve::Properties

CLissaCurve::Properties::Properties(CLissaCurve* const Object) :
	CLissaCurve::PropertiesBase(Object, false, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "lissacurve.gtkml"));

	Connectk3dSpinButton("edgecount", m_Object->m_EdgeCount, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("xfreq", m_Object->m_XFreq, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("xamp", m_Object->m_XAmp, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("xphase", m_Object->m_XPhase, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("yfreq", m_Object->m_YFreq, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("yamp", m_Object->m_YAmp, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("yphase", m_Object->m_YPhase, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("zfreq", m_Object->m_ZFreq, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("zamp", m_Object->m_ZAmp, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("zphase", m_Object->m_ZPhase, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("mfreq", m_Object->m_MFreq, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("mamp", m_Object->m_MAmp, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("mphase", m_Object->m_MPhase, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CLissaCurve>(m_Object, &CLissaCurve::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Show();
}
	
void CLissaCurve::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	CLissaCurve::PropertiesBase::OnOK();
}
	
void CLissaCurve::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	CLissaCurve::PropertiesBase::OnCancel();
}
	
