// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CMagnet K-3D tool, which pulls geometry towards a point
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "magnet.h"

#include <k3dchooser.h>
#include <k3deventconnector.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>

#include <k3diapplication.h>
#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dieditable.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>
#include <sdpgl.h>

/////////////////////////////////////////////////////////////////////////////
// CMagnet

CMagnet::CMagnet(sdpIUnknown* Initializer) :
	CMagnetBase(Initializer),
	m_MagnetFactor("Magnetfactor", 1.0, 0, 0),
	m_Radius("radius", 4.0, 0, new k3dVariableMinimum<gdouble>(0.0, 0)),
	m_Profile("profile", Document(), 0, new k3dEventConnector(this, k3dIChannel::Event_Changed, "CMagnet", 0)),
	m_InteractiveUpdate(true)
{
	// Setup profile curve ...
	g_return_if_fail(m_Profile.Create("Magnet Profile", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, false));

	k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(m_Profile.Interface());
	g_return_if_fail(channel);
	
	// Let the channel know that we're making changes ...
	k3d::Editing editing(m_Profile.Interface());

	channel->SetSegmentCount(1);
	k3dBezierChannelNodeArray nodes;
	k3dBezierChannelNodes(channel, nodes);
	g_assert( 4 == nodes.size() );
	
	k3dBezierChannelNodeArray::iterator node(nodes.begin());
	
	(*node)->SetPosition(sdpVector2(0.0, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.5, 1.0));
	node++;
	(*node)->SetPosition(sdpVector2(0.5, 0.0));
	node++;
	(*node)->SetPosition(sdpVector2(1.0, 0.0));
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CMagnet::EditObject()
{
	new CMagnetDialog(this);
	
	Update( false );
	k3dRedrawAll(Document(), false);	
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEventObserver implementation

sdpResult CMagnet::ObserveEvent(k3dIEventSource* const Source, SDPCSTRING Event, SDPCSTRING EventData, SDPCSTRING ObserverData)
{
	// Sanity checks ...
	g_return_val_if_fail(Source, SDP_E_FAIL);
	g_return_val_if_string_fail(Event, SDP_E_FAIL);
	
	const sdpString event(Event);
	if(event == k3dIChannel::Event_Changed)
		{
			Update(false);
			k3dRedrawAll(Document(), false);
			return SDP_S_OK;
		}
	
	return CMagnetBase::ObserveEvent(Source, Event, EventData, ObserverData);
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CMagnet::EditorRender(k3dIEditorEngine* const Engine)
{
	// Sanity checks ...
	g_assert(Engine);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glColor3d(0.0, 1.0, 1.0);

	// Draw the magnet center point ...
	glPointSize(6.0f);
	glBegin(GL_POINTS);
		glVertex3d(0.0, 0.0, 0.0);
	glEnd();

	// Indicate the magnet's range ...
	static GLint slices = 32;
	const gdouble radius = m_Radius.Value();
	
	GLUquadricObj* quadric = gluNewQuadric();
	g_return_val_if_fail(quadric, SDP_E_FAIL);
	gluQuadricDrawStyle(quadric, GLenum(GLU_SILHOUETTE));
	gluQuadricNormals(quadric, GLenum(GLU_SMOOTH));

	// X axis
	glPushMatrix();
	glRotated(90.0, 0.0, 1.0, 0.0);
	gluDisk(quadric, 0.0, radius, slices, 1);
	glPopMatrix();

	// Y axis
	glPushMatrix();
	glRotated(-90.0, 1.0, 0.0, 0.0);
	gluDisk(quadric, 0.0, radius, slices, 1);
	glPopMatrix();

	// Z axis
	gluDisk(quadric, 0.0, radius, slices, 1);
	gluDeleteQuadric(quadric);
	
	glPopAttrib();	// GL_ALL_ATTRIB_BITS

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////////////
// k3dITransform overrides

sdpResult CMagnet::SetPosition(const ReferenceFrame Frame, gdouble Offsets[3])
{
	CMagnetBase::SetPosition(Frame, Offsets);
	
	if(CURRENT_FRAME == Frame && m_InteractiveUpdate)
		Update(false);
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

/*
sdpResult CMagnet::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// This is something of a hack, but I'm not sure how to do it better - we're overriding this so that the k3dTransformObject
	// doesn't have a chance to call k3dIDocument::FreezeDocument() ...

	m_Position[OLD_FRAME] = m_Position[CURRENT_FRAME];
	m_TransformType = NONE;
	
	return SDP_S_OK;
}
*/

class CMagnetFunctor
{
public:
	CMagnetFunctor(const sdpVector3 WorldOrigin, const gdouble MagnetFactor, const gdouble Radius, k3dIChannel* const Profile) :
		m_WorldOrigin(WorldOrigin),
		m_MagnetFactor(MagnetFactor),
		m_Radius(Radius),
		m_Profile(Profile)
	{
	}	
	
	bool HandleLocationCollection(k3dILocationCollection* LocationCollection)
	{
		m_MagnetOrigin = m_WorldOrigin;
		
		k3dIObject* const object = dynamic_cast<k3dIObject*>(LocationCollection);
		if(object)
			m_MagnetOrigin = k3dGetWorldToObjectMatrix(object) * m_MagnetOrigin;
	
		return true;
	}
	
	bool HandleLocation(k3dILocationCollection* LocationCollection, k3dILocation* Location)
	{
		sdpVector3 coords = Location->LocalXYZ();

		const sdpVector3 difference = m_MagnetOrigin - coords;

		const gdouble distance = difference.Length() / m_Radius;
		gdouble mix_factor;
		if( distance <= 1.0 )
			mix_factor = sdpMin( m_Profile->Value1(distance, 0.001) * m_MagnetFactor, 1.0 );
		else
			mix_factor = 0.0;
		
		coords += mix_factor * difference;	
	
		Location->SetLocalXYZ(coords);
		
		return true;
	}

protected:
	const sdpVector3 m_WorldOrigin;
	const gdouble m_MagnetFactor;
	const gdouble m_Radius;
	k3dIChannel* const m_Profile;
	
	sdpVector3 m_MagnetOrigin;
};

void CMagnet::Update(bool SaveUndoAction)
{
	// Record the action if requested ...
	if(SaveUndoAction)
		k3dStartStateChangeSet(Document());

/*
	CMagnetFunctor functor(k3dGetObjectToWorldMatrix(this) * sdpVector3(0, 0, 0), m_MagnetFactor.Value(), m_Radius.Value(), m_Profile.Interface());
	k3dForEachLocation(Document(), k3dIState::VISIBLESELECTED, functor);
*/

	// Finish recording the action if requested ...
	if(SaveUndoAction)
		k3dFinishStateChangeSet(Document(), "Magnet");
}

//////////////////////////////////////////////////////////////////////////////////////
// CMagnetDialog

const sdpString CMagnetDialog::dialogname = "properties";
const sdpString CMagnetDialog::control_magnetfactor = "magnetfactor";
const sdpString CMagnetDialog::control_radius = "radius";
const sdpString CMagnetDialog::control_profile = "profile";
const sdpString CMagnetDialog::control_position = "position";

CMagnetDialog::CMagnetDialog(CMagnet* const Object) :
	CMagnetDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "magnet.gtkml"));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_magnetfactor, m_Object->m_MagnetFactor, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate<CMagnet>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_radius, m_Object->m_Radius, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate<CMagnet>(m_Object, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Setup profile channel ...
	Connectk3dChannelButton(control_profile, m_Object->m_Profile.Interface());

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	RootWidget().Show();
}

void CMagnetDialog::OnOK()
{
	m_Object->m_InteractiveUpdate = false;
	m_Object->Update(true);
	k3dRedrawAll(m_Object->Document(), false);
	
	CMagnetDialogBase::OnOK();
}

void CMagnetDialog::OnCancel()
{
	m_Object->m_InteractiveUpdate = false;
	k3dRedrawAll(m_Object->Document(), false);
	
	CMagnetDialogBase::OnCancel();
}


