// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CMirror K-3D tool, which reflects geometry about an axis
		\author Tim Shead (tshead@k-3d.com)
*/

#include "mirror.h"

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>
#include <k3dimesh.h>
#include <k3dipath.h>
#include <k3dipathpointcollection.h>
#include <k3dipathpoint.h>

#include <sdpgtkevent.h>

namespace
{

class MirrorLocations
{
public:
	MirrorLocations(const k3d::Axis Axis) :
		m_Axis(Axis)
	{
	}
	
	void operator()(const k3d::SelectedLocationProxy& Location)
	{
		sdpVector3 coords = Location->LocalXYZ();
		coords[m_Axis] = -coords[m_Axis];
		Location->SetLocalXYZ(coords);
	}

private:
	const k3d::Axis m_Axis;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CMirror

CMirror::CMirror(sdpIUnknown* Initializer) :
	CMirrorBase(Initializer),
	m_Axis("axis", k3d::X, Document()->StateRecorder(), 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CMirror::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	
	PushEditing();
	
	return CMirrorBase::InitComplete();
}

sdpResult CMirror::EditObject()
{
	new CMirrorDialog(this);
	return SDP_S_OK;
}

void CMirror::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CMirror::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), MirrorLocations(m_Axis.Value()));
}

void CMirror::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CMirror::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Mirror");
}

//////////////////////////////////////////////////////////////////////////////////////
// CMirrorDialog

const sdpString CMirrorDialog::dialogname = "properties";
const sdpString CMirrorDialog::control_axis = "axis";
const sdpString CMirrorDialog::control_mirror = "mirror";

CMirrorDialog::CMirrorDialog(CMirror* const Object) :
	CMirrorDialogBase(Object, false, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "mirror.gtkml"));

	// Attach chooser 
	if(Getk3dChooser(control_axis))
		Getk3dChooser(control_axis)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Axis), k3dApplication()->CommandRecorder(), 0, "Axis", 0, 0, 0);

	// Setup command buttons ...
	AddCommand(control_mirror, control_mirror, CONTROL_BUTTON);
	
	Show();
}

void CMirrorDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_mirror)
		OnMirror();
	else
		CMirrorDialogBase::OnEvent(Event);
}

void CMirrorDialog::OnMirror()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_mirror, 0);

	m_Object->UpdateGeometry();
	k3dRedrawAll(m_Object->Document(), false);
}

void CMirrorDialog::OnOK()
{
	m_Object->PopEditing();
	CMirrorDialogBase::OnOK();
}

void CMirrorDialog::OnCancel()
{
	m_Object->CancelEditing();
	k3dRedrawAll(m_Object->Document(), false);
	CMirrorDialogBase::OnCancel();
}

