// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the functions necessary to export K-3D objects from the toollib module
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "bend.h"
#include "bevel.h"
#include "bulge.h"
#include "displace.h"
#include "extrude.h"
#include "freehand.h"
#include "knots.h"
#include "lissacurve.h"
#include "magnet.h"
#include "mirror.h"
#include "newell.h"
#include "pathextrude.h"
#include "primitives.h"
#include "rotate.h"
#include "scale.h"
#include "slant.h"
#include "smooth.h"
#include "snap.h"
#include "sphereize.h"
#include "subdivide.h"
#include "taper.h"
#include "translate.h"
#include "triangulate.h"
#include "twist.h"
#include "uvmap.h"
#include "wave.h"

#include <k3dcom.h>
#include <k3dmodule.h>

class ToolModule :
	public k3d::Module
{
public:
	sdpIUnknown* CreateInstance(const sdp::ClassID& ClassID, sdpIUnknown* Initializer)
	{
		if(ClassID == k3d::classid::Bend)
			return sdpCreateClassInstance<CBend>(Initializer);
		if(ClassID == k3d::classid::Bevel)
			return sdpCreateClassInstance<CBevel>(Initializer);
		if(ClassID == k3d::classid::Bulge)
			return sdpCreateClassInstance<CBulge>(Initializer);
		if(ClassID == k3d::classid::Displace)
			return sdpCreateClassInstance<CDisplace>(Initializer);
		if(ClassID == k3d::classid::Extrude)
			return sdpCreateClassInstance<CExtrude>(Initializer);
		if(ClassID == k3d::classid::Freehand)
			return sdpCreateClassInstance<CFreehand>(Initializer);
		if(ClassID == k3d::classid::LissaCurve)
			return sdpCreateClassInstance<CLissaCurve>(Initializer);
		if(ClassID == k3d::classid::Knots)
			return sdpCreateClassInstance<CKnots>(Initializer);
		if(ClassID == k3d::classid::Magnet)
			return sdpCreateClassInstance<CMagnet>(Initializer);
		if(ClassID == k3d::classid::Mirror)
			return sdpCreateClassInstance<CMirror>(Initializer);
		if(ClassID == k3d::classid::Newell)
			return sdpCreateClassInstance<CNewell>(Initializer);
		if(ClassID == k3d::classid::PathExtrude)
			return sdpCreateClassInstance<CPathExtrude>(Initializer);
		if(ClassID == k3d::classid::Primitives)
			return sdpCreateClassInstance<CPrimitives>(Initializer);
		if(ClassID == k3d::classid::Rotate)
			return sdpCreateClassInstance<CRotate>(Initializer);
		if(ClassID == k3d::classid::Scale)
			return sdpCreateClassInstance<CScale>(Initializer);
		if(ClassID == k3d::classid::Slant)
			return sdpCreateClassInstance<CSlant>(Initializer);
		if(ClassID == k3d::classid::Smooth)
			return sdpCreateClassInstance<CSmooth>(Initializer);
		if(ClassID == k3d::classid::Snap)
			return sdpCreateClassInstance<CSnap>(Initializer);
		if(ClassID == k3d::classid::Sphereize)
			return sdpCreateClassInstance<CSphereize>(Initializer);
		if(ClassID == k3d::classid::SubDivide)
			return sdpCreateClassInstance<CSubDivide>(Initializer);
		if(ClassID == k3d::classid::Taper)
			return sdpCreateClassInstance<CTaper>(Initializer);
		if(ClassID == k3d::classid::Twist)
			return sdpCreateClassInstance<CTwist>(Initializer);
		if(ClassID == k3d::classid::Translate)
			return sdpCreateClassInstance<CTranslate>(Initializer);
		if(ClassID == k3d::classid::Triangulate)
			return sdpCreateClassInstance<CTriangulate>(Initializer);
		if(ClassID == k3d::classid::UVMap)
			return sdpCreateClassInstance<CUVMap>(Initializer);
		if(ClassID == k3d::classid::Wave)
			return sdpCreateClassInstance<CWave>(Initializer);

		return 0;
	}

private:
	friend class sdpSingleton<ToolModule>;

	ToolModule() :
		k3d::Module("tools", VERSION)
	{
		// Bend ...
		sdpxml::Element& bend = RegisterObject( k3d::classid::Bend, "Bend");
		RegisterBookmark(bend, "Tools/Bend");

		// Bevel ...
		sdpxml::Element& bevel = RegisterObject( k3d::classid::Bevel, "Bevel");
		RegisterBookmark(bevel, "Tools/Bevel");

		// Bulge ...
		sdpxml::Element& bulge = RegisterObject( k3d::classid::Bulge, "Bulge");
		RegisterBookmark(bulge, "Tools/Bulge");

		// Displace ...
		sdpxml::Element& displace = RegisterObject( k3d::classid::Displace, "Displace");
		RegisterBookmark(displace, "Tools/Displace");

		// Extrude ...
		sdpxml::Element& extrude = RegisterObject( k3d::classid::Extrude, "Extrude");
		RegisterBookmark(extrude, "Tools/Extrude");

		// Freehand ...
		sdpxml::Element& freehand = RegisterObject( k3d::classid::Freehand, "Freehand");
		RegisterBookmark(freehand, "Tools/Freehand");

		// Knots ...
		sdpxml::Element& knots = RegisterObject( k3d::classid::Knots, "Knots");
		RegisterBookmark(knots, "Tools/Knots");

		// LissaCurve ...
		sdpxml::Element& lissacurve = RegisterObject( k3d::classid::LissaCurve, "LissaCurve");
		RegisterBookmark(lissacurve, "Tools/LissaCurve");

		// Magnet ...
		sdpxml::Element& magnet = RegisterObject( k3d::classid::Magnet, "Magnet");
		RegisterBookmark(magnet, "Tools/Magnet");
		RegisterIncomplete(magnet);

		// Mirror ...
		sdpxml::Element& mirror = RegisterObject( k3d::classid::Mirror, "Mirror");
		RegisterBookmark(mirror, "Tools/Mirror");

		// Newell ...
		sdpxml::Element& newell = RegisterObject( k3d::classid::Newell, "Newell");
		RegisterBookmark(newell, "Tools/Newell");

		// PathExtrude ...
		sdpxml::Element& pathextrude = RegisterObject( k3d::classid::PathExtrude, "PathExtrude");
		RegisterBookmark(pathextrude, "Tools/PathExtrude");

		// Primitives ...
		sdpxml::Element& primitives = RegisterObject( k3d::classid::Primitives, "Primitives");
		RegisterBookmark(primitives, "Tools/Primitives");
		RegisterIncomplete(primitives);

		// Rotate ...
		sdpxml::Element& rotate = RegisterObject(k3d::classid::Rotate, "RotatePoints");
		RegisterBookmark(rotate, "Tools/Rotate Points");

		// Scale ...
		sdpxml::Element& scale = RegisterObject(k3d::classid::Scale, "ScalePoints");
		RegisterBookmark(scale, "Tools/Scale Points");

		// Slant ...
		sdpxml::Element& slant = RegisterObject( k3d::classid::Slant, "Slant");
		RegisterBookmark(slant, "Tools/Slant");

		// Snap ...
		sdpxml::Element& snap = RegisterObject( k3d::classid::Snap, "Snap");
		RegisterBookmark(snap, "Tools/Snap");

		// Sphereize ...
		sdpxml::Element& sphereize = RegisterObject( k3d::classid::Sphereize, "Sphereize");
		RegisterBookmark(sphereize, "Tools/Sphereize");

		// Smooth ...
		sdpxml::Element& smooth = RegisterObject( k3d::classid::Smooth, "Smooth");
		RegisterBookmark(smooth, "Tools/Smooth");

		// SubDivide ...
		sdpxml::Element& subdivide = RegisterObject( k3d::classid::SubDivide, "SubDivide");
		RegisterBookmark(subdivide, "Tools/SubDivide");

		// Taper ...
		sdpxml::Element& taper = RegisterObject( k3d::classid::Taper, "Taper");
		RegisterBookmark(taper, "Tools/Taper");

		// Translate ...
		sdpxml::Element& translate = RegisterObject( k3d::classid::Translate, "TranslatePoints");
		RegisterBookmark(translate, "Tools/Translate Points");

		// Triangulate ...
		sdpxml::Element& triangulate = RegisterObject( k3d::classid::Triangulate, "Triangulate");
		RegisterBookmark(triangulate, "Tools/Triangulate");

		// Twist ...
		sdpxml::Element& twist = RegisterObject( k3d::classid::Twist, "Twist");
		RegisterBookmark(twist, "Tools/Twist");

		// UVMap ...
		sdpxml::Element& uvmap = RegisterObject( k3d::classid::UVMap, "UVMap");
		RegisterBookmark(uvmap, "Tools/UVMap");

		// Wave ...
		sdpxml::Element& wave = RegisterObject( k3d::classid::Wave, "Wave");
		RegisterBookmark(wave, "Tools/Wave");
	}
};

K3DMODULEPROCEDURES(ToolModule, sdpSingleton<ToolModule>);


