// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CPathExtrude, a K-3D tool for creating ducted solids from existing geometry
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "pathextrude.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>
#include <k3dicamera.h>
#include <k3dilocation.h>
#include <k3dimaterialcollection.h>
#include <k3dipathpointcollection.h>

#include <sdpgtkevents.h>
#include <sdpgl.h>

#include <algorithm>
#include <iterator>
#include <iostream>

///////////////////////////////////////////////////////////////
// CPathExtrude

CPathExtrude::CPathExtrude(sdpIUnknown* Initializer) :
	CPathExtrudeBase(Initializer),
	k3dMouseEventObserver("LMB-drag to offset/scale/rotate extrusion; control-LMB-drag to constrain to Active Axis"),
	m_Segments("segments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1.0, 0))),
	m_ConnectExterior("connectexterior", true, 0, 0),
	m_ConnectInterior("connectinterior", false, 0, 0),
	m_CreateLayers("createlayers", false, 0, 0),
	m_CreateEndCaps("createendcaps", false, 0, 0),
	m_Sensitivity("sensitivity", 0.1, 0, 0),
	m_OffsetX("offsetx", 0.0, 0, 0),
	m_OffsetY("offsety", 0.0, 0, 0),
	m_OffsetZ("offsetz", 0.0, 0, 0),
	m_ScaleX("scalex", 1.0, 0, 0),
	m_ScaleY("scaley", 1.0, 0, 0),
	m_ScaleZ("scalez", 1.0, 0, 0),
	m_RotationX("rotationx", 0.0, 0, 0),
	m_RotationY("rotationy", 0.0, 0, 0),
	m_RotationZ("rotationz", 0.0, 0, 0),
	m_InteractiveMode(EXTRUDE)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CPathExtrude::EditObject()
{
	new Properties(this);

/*
	Update(false, true);
*/
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

void CPathExtrude::SetShape()
{
/*
	m_ShapePaths.clear();
	m_ShapePoints.clear();

	k3dGetPaths(Document(), k3dIState::VISIBLESELECTED, m_ShapePaths);
	k3dGetPoints(Document(), k3dIState::VISIBLESELECTED, m_ShapePoints);
*/
}

void CPathExtrude::SetPath()
{
/*
	m_PathPaths.clear();
	m_PathPoints.clear();

	k3dGetPaths(Document(), k3dIState::VISIBLESELECTED, m_PathPaths);
	k3dGetPoints(Document(), k3dIState::VISIBLESELECTED, m_PathPoints);

	k3dSetValue<gdouble>(m_Segments, m_PathPoints.size());
	
	Notify(k3dIObject::Event_UpdateUI, 0);
*/
}

/*
void CPathExtrude::Update(bool SaveUndoAction, bool CreateGeometry)
{
	// Sanity checks ...
	const gulong segments = static_cast<gulong>(m_Segments.Value());
	g_return_if_fail(segments);

	// Remove old geometry as necessary ...
	if(CreateGeometry)
		DeleteAll();

	// If we don't have a shape, we're done ...
	if(0 == m_ShapePaths.size() || 0 == m_ShapePoints.size())
		return;
		
	// If we don't have a path, we're done ...
	if(0 == m_PathPaths.size() || 0 == m_PathPoints.size())
		return;

	// Create new geometry
	if(CreateGeometry)
		{
			if(SaveUndoAction)
				k3dStartStateChangeSet(Document());

			k3dMeshArray newmeshes;
			k3dPathArray newpaths;
			k3dPointArray newpoints;
		
			k3dDuctedSolid(Document(),
										m_ShapePaths,
										m_ShapePoints,
										segments,
										m_ConnectExterior.Value(),
										m_ConnectInterior.Value(),
										m_CreateLayers.Value(),
										m_CreateEndCaps.Value(),
										true,
										newmeshes,
										newpaths,
										newpoints,
										m_NewPoints);

			if(SaveUndoAction)
				k3dFinishStateChangeSet(Document(), "PathExtrude");

			gulong i;
			for(i = 0; i < newmeshes.size(); i++)
				ManageMesh(newmeshes[i], "PathExtrusion");

			for(i = 0; i < newpaths.size(); i++)
				BumpPathCount();
//				AddPath(newpaths[i]);
				
			for(i = 0; i < newpoints.size(); i++)
				BumpPointCount();
//				AddPoint(newpoints[i]);

			// Set materials (we can't use k3dPrimTool::SetDefaultMaterials() because the paths aren't managed)
			k3dIMaterial* defaultmaterial = k3dDefaultMaterial(Document());
			for(i = 0; i < newpaths.size(); i++)
				newpaths[i]->Materials()->SetMaterial(defaultmaterial);
		}

	// For each layer ...
	sdpVector3 lastx;
	for(gulong i = 0; i < segments; i++)
		{
			gulong prevlayer = (i-1) % segments;
			gulong currentlayer = i % segments;
			gulong nextlayer = (i+1) % segments;
		
			k3dIPoint* pointa = m_PathPoints[prevlayer];
			k3dIPoint* pointb = m_PathPoints[currentlayer];
			k3dIPoint* pointc = m_PathPoints[nextlayer];
		
			sdpVector3 locationa = pointa->Location()->LocalXYZ();
			sdpVector3 locationb = pointb->Location()->LocalXYZ();
			sdpVector3 locationc = pointc->Location()->LocalXYZ();

			// Calculate the spine-aligned cross-section plane (SCP) as a set of mutually-perpendicular unit vectors ...
			sdpVector3 ba = (locationa - locationb).Normalize();
			sdpVector3 bc = (locationc - locationb).Normalize();
						
			sdpVector3 z = (bc - ba).Normalize();
			sdpVector3 x = (bc ^ ba).Normalize();
			if(lastx * x < 0)
				x = -x;

			sdpVector3 y = (x ^ z).Normalize();

			lastx = x;

			sdpVector3 position = locationb;
																																				
			// Interpolate our transformation values ...
			gdouble percent = gdouble(i) / m_Segments.Value();
			sdpVector3 offset = percent * sdpVector3(m_OffsetX.Value(), m_OffsetY.Value(), m_OffsetZ.Value());
			sdpVector3 rotation = sdpLerp(sdpVector3(0, 0, 0), sdpVector3(m_RotationX.Value(), m_RotationY.Value(), m_RotationZ.Value()), percent);
			sdpVector3 scale = sdpLerp(sdpVector3(1, 1, 1), sdpVector3(m_ScaleX.Value(), m_ScaleY.Value(), m_ScaleZ.Value()), percent);

			// Create a transformation matrix ...
			sdpMatrix4 matrix = sdpTranslation3D(position);
			matrix = matrix * sdpTranslation3D(offset);
			matrix = matrix * sdpRotation3D(rotation);
			matrix = matrix * sdpScaling3D(scale);

			// Calculate points ...
			for(k3dDuctedSolidPointMapIterator point = m_NewPoints.begin(); point != m_NewPoints.end(); point++)
				{
					// Skip non-existent points ...
					if(0 == point->second[currentlayer])
						continue;
						
					sdpVector3 coords = point->first->Location()->LocalXYZ();
					coords = matrix * (x * coords[0]) + (y * coords[1]) + (z * coords[2]);
					
					// Store the new coordinates ...
					point->second[currentlayer]->Location()->SetLocalXYZ(coords);
				}
		}

	Notify(k3dIObject::Event_UpdateUI, 0);
}
*/

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

// Called while the user drags the left mouse button
sdpResult CPathExtrude::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	gdouble xfactor = (CurrentX - LastX) * 0.01;
	gdouble yfactor = (CurrentY - LastY) * 0.01;

	switch(m_InteractiveMode)
		{
			case EXTRUDE:
				break;
			case OFFSETS:
				// If the user holds down either control key, only translate along the active axis ...
				if(ControlDown(Modifiers))
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_OffsetX, m_OffsetX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_OffsetY, m_OffsetY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_OffsetZ, m_OffsetZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				// Otherwise, translate along the two non-active axes ...
				else
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_OffsetZ, m_OffsetZ.Value() - xfactor);
									k3dSetValue(m_OffsetY, m_OffsetY.Value() - yfactor);
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_OffsetX, m_OffsetX.Value() - xfactor);
									k3dSetValue(m_OffsetZ, m_OffsetZ.Value() - yfactor);
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_OffsetX, m_OffsetX.Value() + xfactor);
									k3dSetValue(m_OffsetY, m_OffsetY.Value() - yfactor);
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				break;
			case SCALING:
				// If the user holds down either shift key, constrain scale factors to be equal ...
				if(ShiftDown(Modifiers))
					{
						k3dSetValue(m_ScaleX, m_ScaleX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						k3dSetValue(m_ScaleY, m_ScaleY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
					}
				// If the user holds down either control key, only scale along the active axis ...
				else if(ControlDown(Modifiers))
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_ScaleX, m_ScaleX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_ScaleY, m_ScaleY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				// Otherwise, scale along the two non-active axes ...
				else
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - xfactor);
									k3dSetValue(m_ScaleY, m_ScaleY.Value() - yfactor);
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_ScaleX, m_ScaleX.Value() - xfactor);
									k3dSetValue(m_ScaleZ, m_ScaleZ.Value() - yfactor);
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_ScaleX, m_ScaleX.Value() - xfactor);
									k3dSetValue(m_ScaleY, m_ScaleY.Value() - yfactor);
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				break;
			case ROTATION:
				// Compute scaling factors for each mouse axis ...
				xfactor = sdpRadians(gdouble(CurrentX - LastX) * 0.1);
				yfactor = sdpRadians(gdouble(CurrentY - LastY) * 0.1);

				// If the user holds down either control key, only rotate along the active axis ...
				if(ControlDown(Modifiers))
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_RotationY, m_RotationY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_RotationZ, m_RotationZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_RotationX, m_RotationX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				// Otherwise, rotate along the two non-active axes ...
				else
					{
						switch(Camera->ActiveAxis())
							{
								case k3dICamera::XAXIS:
									k3dSetValue(m_RotationX, m_RotationX.Value() - xfactor);
									k3dSetValue(m_RotationZ, m_RotationZ.Value() - yfactor);
									break;
								case k3dICamera::YAXIS:
									k3dSetValue(m_RotationY, m_RotationY.Value() - xfactor);
									k3dSetValue(m_RotationX, m_RotationX.Value() - yfactor);
									break;
								case k3dICamera::ZAXIS:
									k3dSetValue(m_RotationY, m_RotationY.Value() + xfactor);
									k3dSetValue(m_RotationZ, m_RotationZ.Value() - yfactor);
									break;
								default:
									g_return_val_if_fail(0, SDP_E_FAIL);
							}
					}
				break;
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
				break;
		}

/*		
	// Update the display
	Update(false, false);
	k3dRedrawAll(Document(), false);
*/

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////////
// CPathExtrude::Properties

const sdpString CPathExtrude::Properties::dialogname = "properties";
const sdpString CPathExtrude::Properties::control_segments = "segments";
const sdpString CPathExtrude::Properties::control_connectexterior = "connectexterior";
const sdpString CPathExtrude::Properties::control_connectinterior = "connectinterior";
const sdpString CPathExtrude::Properties::control_createlayers = "createlayers";
const sdpString CPathExtrude::Properties::control_createendcaps = "createendcaps";
const sdpString CPathExtrude::Properties::control_offsetx = "offsetx";
const sdpString CPathExtrude::Properties::control_offsety = "offsety";
const sdpString CPathExtrude::Properties::control_offsetz = "offsetz";
const sdpString CPathExtrude::Properties::control_scalex = "scalex";
const sdpString CPathExtrude::Properties::control_scaley = "scaley";
const sdpString CPathExtrude::Properties::control_scalez = "scalez";
const sdpString CPathExtrude::Properties::control_rotationx = "rotationx";
const sdpString CPathExtrude::Properties::control_rotationy = "rotationy";
const sdpString CPathExtrude::Properties::control_rotationz = "rotationz";
const sdpString CPathExtrude::Properties::control_pageswitch = "pageswitch";
const sdpString CPathExtrude::Properties::control_tabs = "tabs";
const sdpString CPathExtrude::Properties::control_setshape = "setshape";
const sdpString CPathExtrude::Properties::control_setpath = "setpath";

CPathExtrude::Properties::Properties(CPathExtrude* const Object) :
	PropertiesBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "pathextrude.gtkml"));

	AddCommand(control_setshape, control_setshape, CONTROL_BUTTON);
	AddCommand(control_setpath, control_setpath, CONTROL_BUTTON);

/*
	// Connect check buttons ...
	Connectk3dCheckButton(control_connectexterior, m_Object->m_ConnectExterior, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton(control_connectinterior, m_Object->m_ConnectInterior, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton(control_createlayers, m_Object->m_CreateLayers, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dCheckButton(control_createendcaps, m_Object->m_CreateEndCaps, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Connect spin buttons ...
//	Connectk3dSpinButton(control_segments, m_Object->m_Segments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, true, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_offsetx, m_Object->m_OffsetX, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_offsety, m_Object->m_OffsetY, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_offsetz, m_Object->m_OffsetZ, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_scalex, m_Object->m_ScaleX, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_scaley, m_Object->m_ScaleY, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_scalez, m_Object->m_ScaleZ, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_rotationx, m_Object->m_RotationX, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_rotationy, m_Object->m_RotationY, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton(control_rotationz, m_Object->m_RotationZ, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPathExtrude>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
*/

	Notebook(control_tabs).SetPage(m_Object->InteractiveMode());

	Show();
}

void CPathExtrude::Properties::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_pageswitch)
		OnPageSwitch(Event);
	else if(Event->Name() == control_setshape)
		OnSetShape();
	else if(Event->Name() == control_setpath)
		OnSetPath();
	else
		PropertiesBase::OnEvent(Event);
}

void CPathExtrude::Properties::OnPageSwitch(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Keep track of the current page ...
	sdpGtkEventNotebookSwitchPage* event = (sdpGtkEventNotebookSwitchPage*)Event;
	m_Object->SetInteractiveMode(CPathExtrude::InteractiveModeType(event->PageNumber()));
}

void CPathExtrude::Properties::OnSetShape()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_setshape, 0);

/*
	m_Object->SetShape();
	k3dDeselectAll(m_Object->Document());
	
	m_Object->Update(false, true);
	k3dRedrawAll(m_Object->Document(), false);
*/
}

void CPathExtrude::Properties::OnSetPath()
{
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_setpath, 0);

/*
	m_Object->SetPath();
	k3dDeselectAll(m_Object->Document());

	m_Object->Update(false, true);
	k3dRedrawAll(m_Object->Document(), false);
*/
}


