// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CPrimitives K-3D tool, for creating simple geometric primitives
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dipathpointcollection.h>
#include <k3dilocation.h>

#include <sdpgtkevents.h>

#include <algorithm>
#include <strstream>

#include "primitives.h"
#include "kaleido.h"

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Disk

namespace
{

class DiskPoints
{
public:
	DiskPoints(const gdouble Radius, const gdouble Power) :
		m_Radius(Radius),
		m_Power(Power ? 1.0 / Power : 1.0)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const gdouble theta = Point->Parameters()[0] * sdp::PiTimes2();

		sdpVector3 coords(-cos(theta), sin(theta), 0);

		coords[0] = sdpSign(coords[0]) * pow(fabs(coords[0]), m_Power);
		coords[1] = sdpSign(coords[1]) * pow(fabs(coords[1]), m_Power);

		(*Point)->Location()->SetLocalXYZ(coords * m_Radius);
		(*Point)->Location()->SetTextureXYZ(sdpVector3((coords[0] * 0.5) + 0.5, 0.5 - (coords[1] * 0.5), 0));
		(*Point)->Location()->SetImplicitTextureXYZ(sdpVector3((coords[0] * 0.5) + 0.5, 0.5 - (coords[1] * 0.5), 0));
	}

private:
	const gdouble m_Radius;
	const gdouble m_Power;
};

} // namespace

CPrimitives::Disk::Disk(k3dIDocument* const Document) :
	m_Document(Document),
	m_Type(SOLID),
	m_Radius("radius", 5.0, 0, 0),
	m_Segments("segments", 32.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3, 0))),
	m_Power("power", 1.0, 0, 0)
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Disk::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong pointcount = static_cast<gulong>(m_Segments.Value());
	g_return_if_fail(pointcount);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Disk", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// We add a single path and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewPath(**m_NewGeometry.Meshes().begin(), pointcount, m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Disk::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), DiskPoints(m_Radius.Value(), m_Power.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Disk::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Disk::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Disk");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Cone

namespace
{

class ConePoints
{
public:
	ConePoints(const gdouble Length, const gdouble Radius, const gdouble LengthPower, const gdouble RadialPower) :
		m_Length(Length),
		m_Radius(Radius),
		m_LengthPower(LengthPower ? 1.0 / LengthPower : 1.0),
		m_RadialPower(RadialPower ? 1.0 / RadialPower : 1.0)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const sdpVector3 parameters(Point->Parameters());

		const gdouble theta = sdp::PiTimes2() * parameters[0];
		sdpVector3 coords(cos(theta), (0.5 - parameters[1]) * m_Length, sin(theta));

		coords[0] = sdpSign(coords[0]) * pow(fabs(coords[0]), m_RadialPower);
		coords[2] = sdpSign(coords[2]) * pow(fabs(coords[2]), m_RadialPower);

		gdouble lengthscale = parameters[1];
		lengthscale = sdpSign(lengthscale) * pow(fabs(lengthscale), m_LengthPower);

		coords[0] *= lengthscale;
		coords[2] *= lengthscale;

		coords[0] *= m_Radius;
		coords[2] *= m_Radius;

		(*Point)->Location()->SetLocalXYZ(coords);
	}

private:
	const gdouble m_Length;
	const gdouble m_Radius;
	const gdouble m_LengthPower;
	const gdouble m_RadialPower;
};

} // namespace

CPrimitives::Cone::Cone(k3dIDocument* const Document) :
	m_Document(Document),
	m_Length("length", 10.0, 0, 0),
	m_Radius("radius", 5.0, 0, 0),
	m_LengthPower("lengthpower", 1.0, 0, 0),
	m_RadialPower("radialpower", 1.0, 0, 0),
	m_LengthSegments("lengthsegments", 1.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RadialSegments("radialsegments", 32.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3, 0))),
	m_CreateBase(true)
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Cone::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong lengthsegments = static_cast<gulong>(m_LengthSegments.Value());
	const gulong radialsegments = static_cast<gulong>(m_RadialSegments.Value());

	g_return_if_fail(lengthsegments && radialsegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Cone", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// Add geometry and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewGrid(**m_NewGeometry.Meshes().begin(), lengthsegments, radialsegments, false, true, m_NewGeometry);
/*
	if(m_CreateBase)
		{
			k3dIPath* const base = k3d::CreateNewPath(Mesh, 0, m_NewGeometry);

		}
*/
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Cone::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), ConePoints(m_Length.Value(), m_Radius.Value(), m_LengthPower.Value(), m_RadialPower.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Cone::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Cone::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Cone");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Cylinder

namespace
{

class CylinderPoints
{
public:
	CylinderPoints(const gdouble Length, const gdouble Radius, const gdouble RadialPower) :
		m_Length(Length),
		m_Radius(Radius),
		m_RadialPower(RadialPower ? 1.0 / RadialPower : 1.0),
		m_Offset(sdpVector3(0.0, Length * 0.5, 0.0))
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const sdpVector3 parameters(Point->Parameters());

		const gdouble theta = sdp::PiTimes2() * parameters[0];
		sdpVector3 coords(cos(theta), (0.5 - parameters[1]) * m_Length, sin(theta));

		coords[0] = sdpSign(coords[0]) * pow(fabs(coords[0]), m_RadialPower);
		coords[2] = sdpSign(coords[2]) * pow(fabs(coords[2]), m_RadialPower);

		coords[0] *= m_Radius;
		coords[2] *= m_Radius;

		(*Point)->Location()->SetLocalXYZ(coords);
	}

private:
	const gdouble m_Length;
	const gdouble m_Radius;
	const gdouble m_RadialPower;
	const sdpVector3 m_Offset;
};

} // namespace

CPrimitives::Cylinder::Cylinder(k3dIDocument* const Document) :
	m_Document(Document),
	m_Length("length", 10.0, 0, 0),
	m_Radius("radius", 2.5, 0, 0),
	m_RadialPower("radialpower", 1.0, 0, 0),
	m_LengthSegments("lengthsegments", 1.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RadialSegments("radialsegments", 32.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3, 0))),
	m_CreateBase(true)
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Cylinder::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong lengthsegments = static_cast<gulong>(m_LengthSegments.Value());
	const gulong radialsegments = static_cast<gulong>(m_RadialSegments.Value());

	g_return_if_fail(lengthsegments && radialsegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Cylinder", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// We add geometry and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewGrid(**m_NewGeometry.Meshes().begin(), lengthsegments, radialsegments, false, true, m_NewGeometry);
/*
	if(m_CreateBase)
		{
			k3dIPath* const base = k3d::CreateNewPath(Mesh, 0, m_NewGeometry);
		}
*/
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Cylinder::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), CylinderPoints(m_Length.Value(), m_Radius.Value(), m_RadialPower.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Cylinder::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Cylinder::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Cylinder");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Grid

namespace
{

class GridPoints
{
public:
	GridPoints(const sdpVector2 Size) :
		m_Size(Size),
		m_Offset(Size * -0.5)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const sdpVector3 parameters(Point->Parameters());
		(*Point)->Location()->SetLocalXYZ(sdpVector3(sdpVector2(parameters[0] * m_Size[0], (1.0 - parameters[1]) * m_Size[1]) + m_Offset, 0.0));
	}

private:
	const sdpVector2 m_Size;
	const sdpVector2 m_Offset;
};

} // namespace

CPrimitives::Grid::Grid(k3dIDocument* const Document) :
	m_Document(Document),
	m_XLength("xlength", 10.0, 0, 0),
	m_YLength("ylength", 10.0, 0, 0),
	m_XSegments("xsegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_YSegments("ysegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0)))
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Grid::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong xsegments = static_cast<gulong>(m_XSegments.Value());
	const gulong ysegments = static_cast<gulong>(m_YSegments.Value());
	g_return_if_fail(xsegments);
	g_return_if_fail(ysegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Grid", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// We add geometry and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewGrid(**m_NewGeometry.Meshes().begin(), ysegments, xsegments, false, false, m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Grid::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), GridPoints(sdpVector2(m_XLength.Value(), m_YLength.Value())));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Grid::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Grid::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Grid");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Cube

namespace
{

class CubePoints
{
public:
	CubePoints(const sdpVector3 Size) :
		m_Size(Size),
		m_Offset(Size * -0.5)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const sdpVector3 parameters(Point->Parameters());
		(*Point)->Location()->SetLocalXYZ(sdpProduct(m_Size, parameters) + m_Offset);
	}

private:
	const sdpVector3 m_Size;
	const sdpVector3 m_Offset;
};

} // namespace

CPrimitives::Cube::Cube(k3dIDocument* const Document) :
	m_Document(Document),
	m_XLength("xlength", 10.0, 0, 0),
	m_YLength("ylength", 10.0, 0, 0),
	m_ZLength("zlength", 10.0, 0, 0),
	m_XSegments("xsegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_YSegments("ysegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_ZSegments("zsegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0)))
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Cube::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong xsegments = static_cast<gulong>(m_XSegments.Value());
	const gulong ysegments = static_cast<gulong>(m_YSegments.Value());
	const gulong zsegments = static_cast<gulong>(m_ZSegments.Value());
	g_return_if_fail(xsegments);
	g_return_if_fail(ysegments);
	g_return_if_fail(zsegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Cube", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// We add new geometry and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewCube(**m_NewGeometry.Meshes().begin(), xsegments, ysegments, zsegments, m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Cube::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), CubePoints(sdpVector3(m_XLength.Value(), m_YLength.Value(), m_ZLength.Value())));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Cube::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Cube::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Cube");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Sphere

namespace
{

class SpherePoints
{
public:
	SpherePoints(const sdpVector3 Size, const gdouble LengthPower, const gdouble RadialPower) :
		m_Size(Size * 0.5),
		m_LengthPower(LengthPower ? 1.0 / LengthPower : 0.0),
		m_RadialPower(RadialPower ? 1.0 / RadialPower : 0.0)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const sdpVector3 parameters(Point->Parameters());

		const gdouble theta = sdp::PiTimes2() * parameters[0];
		sdpVector3 coords(cos(theta), 0.5 - parameters[1], sin(theta));

		coords[0] = sdpSign(coords[0]) * pow(fabs(coords[0]), m_RadialPower);
		coords[2] = sdpSign(coords[2]) * pow(fabs(coords[2]), m_RadialPower);

		gdouble xfactor  = cos(sdp::Pi() * coords[1]);
		xfactor = sdpSign(xfactor) * pow(fabs(xfactor), m_LengthPower);

		gdouble yfactor = sin(sdp::Pi() * coords[1]);
		yfactor = sdpSign(yfactor) * pow(fabs(yfactor), m_LengthPower);

		coords[0] *= xfactor;
		coords[1] = yfactor;
		coords[2] *= xfactor;

		coords = sdpProduct(coords, m_Size);

		(*Point)->Location()->SetLocalXYZ(coords);
	}

private:
	const sdpVector3 m_Size;
	const gdouble m_LengthPower;
	const gdouble m_RadialPower;
};

} // namespace

CPrimitives::Sphere::Sphere(k3dIDocument* const Document) :
	m_Document(Document),
	m_XLength("xlength", 10.0, 0, 0),
	m_YLength("ylength", 10.0, 0, 0),
	m_ZLength("zlength", 10.0, 0, 0),
	m_LengthPower("lengthpower", 1.0, 0, 0),
	m_RadialPower("radialpower", 1.0, 0, 0),
	m_LengthSegments("lengthsegments", 12.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RadialSegments("radialsegments", 32.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3, 0)))
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Sphere::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong lengthsegments = static_cast<gulong>(m_LengthSegments.Value());
	const gulong radialsegments = static_cast<gulong>(m_RadialSegments.Value());

	g_return_if_fail(lengthsegments && radialsegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Sphere", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// We add new geometry and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewGrid(**m_NewGeometry.Meshes().begin(), lengthsegments, radialsegments, false, true, m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Sphere::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), SpherePoints(sdpVector3(m_XLength.Value(), m_YLength.Value(), m_ZLength.Value()), m_LengthPower.Value(), m_RadialPower.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Sphere::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Sphere::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Sphere");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Torus

namespace
{

class TorusPoints
{
public:
	TorusPoints(const gdouble InnerRadius, const gdouble OuterRadius, const gdouble RadialPower, const gdouble CrossPower) :
		m_RingRadius((InnerRadius + OuterRadius) * 0.5),
		m_CrossRadius((OuterRadius - InnerRadius) * 0.5),
		m_RadialPower(RadialPower ? 1.0 / RadialPower : 1.0),
		m_CrossPower(CrossPower ? 1.0 / CrossPower : 1.0)
	{
	}

	void operator()(k3d::NewPointProxy* const Point)
	{
		const sdpVector3 parameters(Point->Parameters());

		const gdouble theta = sdp::PiTimes2() * parameters[0];
		sdpVector3 coords(cos(theta), 0, sin(theta));
		coords[0] = sdpSign(coords[0]) * pow(fabs(coords[0]), m_RadialPower);
		coords[2] = sdpSign(coords[2]) * pow(fabs(coords[2]), m_RadialPower);

		const gdouble phi = sdp::PiTimes2() * (1.0 - parameters[1]);
		sdpVector2 crosscoords(cos(phi), sin(phi));
		crosscoords[0] = sdpSign(crosscoords[0]) * pow(fabs(crosscoords[0]), m_CrossPower);
		crosscoords[1] = sdpSign(crosscoords[1]) * pow(fabs(crosscoords[1]), m_CrossPower);

		const gdouble radius = m_RingRadius * (1.0 + (crosscoords[0] * m_CrossRadius / m_RingRadius));

		coords[0] *= radius;
		coords[1] = crosscoords[1] * m_CrossRadius;
		coords[2] *= radius;

		(*Point)->Location()->SetLocalXYZ(coords);
	}

private:
	const gdouble m_RingRadius;
	const gdouble m_CrossRadius;
	const gdouble m_RadialPower;
	const gdouble m_CrossPower;
};

} // namespace

CPrimitives::Torus::Torus(k3dIDocument* const Document) :
	m_Document(Document),
	m_InnerRadius("innerradius", 2.5, 0, 0),
	m_OuterRadius("outerradius", 5.0, 0, 0),
	m_RadialPower("radialpower", 1.0, 0, 0),
	m_CrossSectionPower("crosssectionpower", 1.0, 0, 0),
	m_RadialSegments("radialsegments", 32.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3, 0))),
	m_CrossSectionSegments("crosssectionsegments", 12.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(3, 0)))
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Torus::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong radialsegments = static_cast<gulong>(m_RadialSegments.Value());
	const gulong crosssectionsegments = static_cast<gulong>(m_CrossSectionSegments.Value());

	g_return_if_fail(radialsegments && crosssectionsegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Torus", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// We add a single path and shape it ...
	m_NewGeometry.PushEditing();
	k3d::CreateNewGrid(**m_NewGeometry.Meshes().begin(), crosssectionsegments, radialsegments, true, true, m_NewGeometry);
	m_NewGeometry.PopEditing();

	UpdateGeometry(Object);
	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Torus::UpdateGeometry(CPrimitives& Object)
{
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), TorusPoints(m_InnerRadius.Value(), m_OuterRadius.Value(), m_RadialPower.Value(), m_CrossSectionPower.Value()));
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));
}

void CPrimitives::Torus::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Torus::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Torus");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Cushion

namespace
{

typedef std::vector<gulong> CushionPath;
typedef std::vector<CushionPath> CushionPaths;

void AddCushionTriangle(const gulong v1, const gulong v2, const gulong v3, CushionPaths& Paths)
{
	CushionPath path;
	path.push_back(v1);
	path.push_back(v2);
	path.push_back(v3);

	Paths.push_back(path);
}

void AddCushionQuad(const gulong v1, const gulong v2, const gulong v3, const gulong v4, CushionPaths& Paths)
{
	CushionPath path;
	path.push_back(v1);
	path.push_back(v2);
	path.push_back(v3);
	path.push_back(v4);

	Paths.push_back(path);
}

} // namespace

CPrimitives::Cushion::Cushion(k3dIDocument* const Document) :
	m_Document(Document),
	m_LengthSegments("lengthsegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_RadialSegments("radialsegments", 5.0, 0, new k3dVariableTruncate(new k3dVariableMinimum<gdouble>(1, 0))),
	m_Diameter("diameter", 4.0, 0, 0),
	m_XLength("xlength", 8.0, 0, 0),
	m_YLength("ylength", 8.0, 0, 0),
	m_ZLength("zlength", 8.0, 0, 0)
{
	// Sanity checks ...
	g_assert(m_Document);
}

void CPrimitives::Cushion::BuildGeometry(CPrimitives& Object)
{
	// Get rid of any old geometry
	CancelGeometry();

	// Sanity checks ...
	const gulong lengthsegments = static_cast<gulong>(m_LengthSegments.Value());
	const gulong radialsegments = static_cast<gulong>(m_RadialSegments.Value());

	g_return_if_fail(lengthsegments && radialsegments);

	// Create a new mesh, keeping track of it for undo purposes ...
	m_NewGeometry.StartChangeSet(m_Document);
	k3d::CreateNewMesh(m_Document, "Cushion", m_NewGeometry);
	m_NewGeometry.FinishChangeSet(m_Document);

	// Calculate points on the surface of a sphere ...
	typedef std::vector<sdpVector3> SpherePoints;
	SpherePoints spherepoints;

	const gdouble spherediameter = m_Diameter.Value();
	const gulong N = lengthsegments*4;
	const gulong R = radialsegments*2-1;

	gulong i, j, n, c, d, e, f;

	for(j = 0; j < N; j++)
		for(i = 0; i < R; i++)
			{
				//gdouble  l = spherediameter/2 - (gdouble)(i+1)*spherediameter/((gdouble)R+1);
				gdouble l = sin(M_PI/2 -(gdouble)(i+1)*M_PI / (gdouble)(R+1))*spherediameter/2;
				gdouble r = sqrt(spherediameter*spherediameter/4 - l*l);

				spherepoints.push_back(sdpVector3( r*sin((gdouble)j*2*M_PI / (gdouble)N), l, r*cos((gdouble)j*2*M_PI / (gdouble)N)));
			}
	spherepoints.push_back(sdpVector3(0, spherediameter/2, 0));
	spherepoints.push_back(sdpVector3(0, -spherediameter/2, 0));

	// Cut sphere into 8 parts using distances ...
	const gdouble xlength = m_XLength.Value();
	const gdouble ylength = m_YLength.Value();
	const gdouble zlength = m_ZLength.Value();

	// Calculate individual cushion points ...
	typedef std::vector<sdpVector3> CushionPoints;
	CushionPoints cushionpoints;

	c = (R+1)/2;

	sdpVector3 tP = sdpVector3(xlength / 2.0, ylength / 2.0, zlength / 2.0);
	for(j = 0; j < N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j : j*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(xlength / 2.0, -ylength / 2.0, zlength / 2.0);
	for(j = 0; j < N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j : j*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	tP = sdpVector3(xlength / 2.0, ylength / 2.0, -zlength / 2.0);
	for(j = N / 4; j < N / 2+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j : j*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(xlength / 2.0, -ylength / 2.0, -zlength / 2.0);
	for(j = N / 4; j < N / 2+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j : j*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	tP = sdpVector3(-xlength / 2.0, ylength / 2.0, -zlength / 2.0);
	for(j = N / 2; j < 3*N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j : j*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(-xlength / 2.0, -ylength / 2.0, -zlength / 2.0);
	for(j = N / 2; j < 3*N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j : j*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	tP = sdpVector3(-xlength / 2.0, ylength / 2.0, zlength / 2.0);
	for(j = 3*N / 4; j <= N; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j % N :(j % N)*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(-xlength / 2.0, -ylength / 2.0, zlength / 2.0);
	for(j = 3*N / 4; j <= N; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j % N :(j % N)*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	// Calculate individual cushion paths
	CushionPaths cushionpaths;

	// Make sphere parts faces ...
	d = c*(N/4+1)+1;

	for(n = 0; n < 8; n++)
		{
			for(j = 0; j < N / 4; j++)
				for(i = 0; i < c-1; i++)
					AddCushionQuad(n*d + c*j + i, n*d + c*j + i+1, n*d + c*(j+1) + i+1, n*d + c*(j+1) + i, cushionpaths);

			for(j = 0; j < N / 4; j ++)
				if(n % 2 == 0)
					AddCushionTriangle(n*d + c*j, n*d + c*(j+1), (n+1) * d - 1, cushionpaths);
				else
					AddCushionTriangle(n*d + c*(j+1) - 1, (n+1) * d - 1, n*d + c*(j+2) - 1, cushionpaths);
		}

	// Link parts ...
	for(n = 0; n < 8; n++)
		{
			e = (n+1)*d - c-1;
			f = ((n+2) % 8)*d;

			for(j = 0; j < c-1; j++)
				AddCushionQuad(e+j, e+j+1, f+j+1, f+j, cushionpaths);
		}

	AddCushionQuad(0, d - 1, 7*d - 1, 7*d - c-1, cushionpaths);
	AddCushionQuad(d - c-1, 2*d, 3*d - 1, d - 1, cushionpaths);
	AddCushionQuad(3*d - c-1, 4*d, 5*d - 1, 3*d - 1, cushionpaths);
	AddCushionQuad(5*d - c-1, 6*d, 7*d - 1, 5*d - 1, cushionpaths);

	AddCushionQuad(d + c-1, 8*d - 2, 8*d - 1, 2*d - 1, cushionpaths);
	AddCushionQuad(2*d - 2, 2*d - 1, 4*d - 1, 3*d + c-1, cushionpaths);
	AddCushionQuad(4*d - 2, 4*d - 1, 6*d - 1, 5*d + c-1, cushionpaths);
	AddCushionQuad(6*d - 2, 6*d - 1, 8*d - 1, 7*d + c-1, cushionpaths);


	for(n = 0; n < 4; n++)
		{
			e = d*(n*2) + c-1;
			f = d*(n*2+1);

			for(j = 0; j < N / 4; j++)
				AddCushionQuad(e+j*c, f+j*c, f+(j+1)*c, e+(j+1)*c, cushionpaths);
		}

	AddCushionQuad(c-1, d*7 - 2, d*8 - c-1, d, cushionpaths);
	AddCushionQuad(d-2, d*2 - c-1, d*3, d*2 + c-1, cushionpaths);
	AddCushionQuad(d*3 - 2, d*4 - c-1, d*5, d*4 + c-1, cushionpaths);
	AddCushionQuad(d*5 - 2, d*6 - c-1, d*7, d*6 + c-1, cushionpaths);

	AddCushionQuad(d-1, d*3 - 1, d*5 - 1, d*7 - 1, cushionpaths);
	AddCushionQuad(d*2 - 1, d*8 - 1, d*6 - 1, d*4 - 1, cushionpaths);

	// Add geometry to the mesh ...
	m_NewGeometry.PushEditing();

	std::vector<k3dIPoint*> k3dpoints;
	for(gulong index = 0; index < cushionpoints.size(); index++)
		{
			k3dIPoint* const newpoint = (**m_NewGeometry.Meshes().begin())->CreatePoint();
			g_return_if_fail(newpoint);
			k3dpoints.push_back(newpoint);
			m_NewGeometry.Insert(newpoint, cushionpoints[index]);
		}

	// For each path ...
	for(gulong i = 0; i < cushionpaths.size(); i++)
		{
			k3dIPath* const path = (**m_NewGeometry.Meshes().begin())->CreatePath();
			g_return_if_fail(path);
			k3d::NewPathProxy* const pathproxy = m_NewGeometry.Insert(path, sdpVector3(0, 0, 0));
			pathproxy->PushEditing();

			// Attach points ...
			for(gulong j = 0; j < cushionpaths[i].size(); j++)
				{
				path->Points()->AddPoint(k3dpoints[cushionpaths[i][j]]);
				}
		}

	m_NewGeometry.PopEditing();

	// Calculate point positions ...
	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PushEditing));

	gulong index = 0;
	for(k3d::NewGeometryCollection::NewPoints::const_iterator point = m_NewGeometry.Points().begin(); point != m_NewGeometry.Points().end(); ++point)
		(**point)->Location()->SetLocalXYZ(cushionpoints[index++]);

	std::for_each(m_NewGeometry.Points().begin(), m_NewGeometry.Points().end(), std::mem_fun(&k3d::NewPointProxy::PopEditing));

	m_NewGeometry.Manage();

	Object.Notify(k3dObject::Event_UserMessage1, m_NewGeometry.Description());
}

void CPrimitives::Cushion::UpdateGeometry(CPrimitives& Object)
{
	// Currently, we haven't optimized cushions for updates, so just rebuild ...
	BuildGeometry(Object);
}

void CPrimitives::Cushion::CancelGeometry()
{
	m_NewGeometry.Cancel();
}

void CPrimitives::Cushion::FinishGeometry()
{
	m_NewGeometry.Finish(m_Document, "Create Cushion");
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives

CPrimitives::CPrimitives(sdpIUnknown* Initializer) :
	CPrimitivesBase(Initializer),
	m_Grid(Document()),
	m_Cube(Document()),
	m_Disk(Document()),
	m_Cone(Document()),
	m_Cylinder(Document()),
	m_Sphere(Document()),
	m_Torus(Document()),
	m_Cushion(Document()),
	m_CurrentPrimitive(&m_Grid)
/*
	m_PolyhedronNumber(1),
	m_PolyhedronDualCheck(false),
	m_PolyhedronLength("polyhedronlength", 10.0, 0, 0),
*/
{
}

/*
void CPrimitives::LoadPolyhedron(int PolyhedronIndex, bool Dual)
{
// B
	// Clear out old data ...
	m_PolyhedronPoints.clear();
	m_PolyhedronPaths.clear();
	m_PolyhedronScaling = 0.5;

	// Solve polyhedron (need_coordinates = 1, need_edgelist = 0, need_approx = 0, just_list = 0)
	Polyhedron* polyhedron = kaleido(const_cast<char*>(SDPCSTRING(sdpFormatString("#%d", PolyhedronIndex))), 1, 0, 0, 0);
	g_return_if_fail(polyhedron);

	double freeze = 0;
	double azimuth = 0;
	double elevation = 0;

	if(!Dual)
		{
			vrmodel(polyhedron, polyhedron->v, polyhedron->V, polyhedron->f, polyhedron->F, polyhedron->name, "", azimuth, elevation, freeze);
		}
	else
		{
			vrmodel(polyhedron, polyhedron->f, polyhedron->F, polyhedron->v, polyhedron->V, polyhedron->dual_name, "*", azimuth, elevation, freeze);
		}

	// Write the polyhedron geometric properties string
	m_PolyhedronProperties = "";
	if(Dual && (polyhedron->hemi || polyhedron->index == last_uniform - 1))
		{
			m_PolyhedronProperties += "infinite dual";

			// Infinite duals have to scaled to a unit sphere
			double maxlength = 0.0;
			for(gulong i = 0; i < m_PolyhedronPoints.size(); i++)
				{
					double length = m_PolyhedronPoints[i].Length();
					if(length > maxlength)
						maxlength = length;
				}
			m_PolyhedronScaling /= maxlength;
		}


	// Free the polyhedron data ...
	polyfree(polyhedron);
}

void CPrimitives::PolyhedronPoint(Kvector v)
{
	m_PolyhedronPoints.push_back(sdpVector3(v.x, v.y, v.z));
}

void CPrimitives::PolyhedronFace()
{
	m_PolyhedronPaths.push_back(PolyhedronPath());
}

void CPrimitives::PolyhedronPathPoint(int PointIndex)
{
	m_PolyhedronPaths.back().push_back(PointIndex);
}

void CPrimitives::PolyhedronTriangle(int PointIndex1, int PointIndex2, int PointIndex3)
{
	m_PolyhedronPaths.push_back(PolyhedronPath());

	m_PolyhedronPaths.back().push_back(PointIndex1);
	m_PolyhedronPaths.back().push_back(PointIndex2);
	m_PolyhedronPaths.back().push_back(PointIndex3);
}

void CPrimitives::PolyhedronQuad(int PointIndex1, int PointIndex2, int PointIndex3, int PointIndex4)
{
	m_PolyhedronPaths.push_back(PolyhedronPath());

	m_PolyhedronPaths.back().push_back(PointIndex1);
	m_PolyhedronPaths.back().push_back(PointIndex2);
	m_PolyhedronPaths.back().push_back(PointIndex3);
	m_PolyhedronPaths.back().push_back(PointIndex4);
}

int  CPrimitives::vrmodel(Polyhedron *P, Kvector *v, int V, Kvector *f, int F,	char *name, char *star,	double azimuth, double elevation, double freeze)
{
	int i, j, k, l, ll, ii, *hit;
	Kvector *temp;

// Rotate polyhedron
	rotframe (azimuth, elevation, freeze);
	Malloc (temp, V, Kvector)
	rotarray (temp, v, V);
	v = temp;
	Malloc (temp, F, Kvector)
	rotarray (temp, f, F);
	f = temp;



	for (i = 0; i < V; i++)
		PolyhedronPoint(v[i]);

// Auxiliary vertices (needed because current VRML browsers cannot handle
// non-simple paths, i.e., paths with self intersections):
// Each non-simple face is assigned an auxiliary vertex. By connecting it to
// the rest of the vertices the face is triangulated. The circum-center is used
// for the regular star faces of uniform pathhedra. The in-center is used for
// the pentagram (#79) and hexagram (#77) of the high-density snub duals, and
// for the pentagrams (#40, #58) and hexagram (#52) of the stellated duals with
// configuration (....)/2. Finally, the self-intersection of the crossed
// parallelogram is used for duals with form p q r| with an even denominator.

// This method do not work for the hemi-duals, whose faces are not star-shaped
// and have two self-intersections each.
// Thus, for each face we need six auxiliary vertices: The self intersections
// and the terminal points of the truncations of the infinite edges. The ideal
// vertices are listed, but are not used by the face-list.
// Note that the face of the last dual (#80) is octagonal, and constists of two
// quadrilaterals of the infinite type.

	if (*star && P->even != -1)
		Malloc (hit, F, int);
	for (i = 0; i < F; i++)
		if (!*star && (frac (P->n[P->ftype[i]]),
			frax.d != 1 && frax.d != frax.n - 1) ||
			*star && (P->K == 5 && P->D > 30 ||
			denominator (P->m[0]) != 1)) {
			// find the center of the face
			double h;
			if (!*star && P->hemi && !P->ftype[i])
				h = 0;
			else
				h = P->minr / dot (f[i],f[i]);
			PolyhedronPoint(scale (h, f[i]));
		} else if (*star && P->even != -1) {
			// find the self-intersection of a crossed parallelogram. hit is set if v0v1 intersects v2v3
			Kvector v0, v1, v2, v3, c0, c1, p;
			double d0, d1;
			v0 = v[P->incid[0][i]];
			v1 = v[P->incid[1][i]];
			v2 = v[P->incid[2][i]];
			v3 = v[P->incid[3][i]];
			d0 = sqrt (dot (diff (v0, v2), diff (v0, v2)));
			d1 = sqrt (dot (diff (v1, v3), diff (v1, v3)));
			c0 = scale (d1, sum (v0, v2));
			c1 = scale (d0, sum (v1, v3));
			p = scale (0.5 / (d0 + d1), sum (c0, c1));
			PolyhedronPoint(p);
			p = cross (diff (p, v2), diff (p, v3));
			hit[i] = (dot (p, p) < 1e-6);
		} else if (*star && P->hemi && P->index != last_uniform - 1) {
			// find the terminal points of the truncation and the
			// self-intersections.
			//  v23       v0       v21
			//  |  \     /  \     /  |
			//  |   v0123    v0321   |
			//  |  /     \  /     \  |
			//  v01       v2       v03

			Kvector v0, v1, v2, v3, v01, v03, v21, v23, v0123, v0321 ;
			Kvector u;
			double t = 1.5; //truncation adjustment factor
			j = !P->ftype[P->incid[0][i]];
			v0 = v[P->incid[j][i]];//real vertex
			v1 = v[P->incid[j+1][i]];//ideal vertex (unit Kvector)
			v2 = v[P->incid[j+2][i]];//real
			v3 = v[P->incid[(j+3)%4][i]];//ideal
			//compute intersections
			//	this uses the following linear algebra:
			//	v0123 = v0 + a v1 = v2 + b v3
			//	v0 x v3 + a (v1 x v3) = v2 x v3
			//	a (v1 x v3) = (v2 - v0) x v3
			//	a (v1 x v3) . (v1 x v3) =
			//		(v2 - v0) x v3 . (v1 x v3)
			u = cross (v1, v3);
			v0123 = sum (v0, scale (dot (cross (diff (v2, v0), v3),
				u) / dot (u,u), v1));
			v0321 = sum (v0, scale (dot (cross (diff (v0, v2), v1),
				u) / dot (u,u), v3));
			//compute truncations
			v01 = sum (v0 , scale (t, diff (v0123, v0)));
			v23 = sum (v2 , scale (t, diff (v0123, v2)));
			v03 = sum (v0 , scale (t, diff (v0321, v0)));
			v21 = sum (v2 , scale (t, diff (v0321, v2)));
			PolyhedronPoint(v01);
			PolyhedronPoint(v23);
			PolyhedronPoint(v0123);
			PolyhedronPoint(v03);
			PolyhedronPoint(v21);
			PolyhedronPoint(v0321);
		} else if (*star && P->index == last_uniform - 1) {
			// find the terminal points of the truncation and the
			// self-intersections.
			//  v23       v0       v21
			//  |  \     /  \     /  |
			//  |   v0123    v0721   |
			//  |  /     \  /     \  |
			//  v01       v2       v07
			//
			//  v65       v4       v67
			//  |  \     /  \     /  |
			//  |   v4365    v4567   |
			//  |  /     \  /     \  |
			//  v43       v6       v45

			Kvector v0, v1, v2, v3, v4, v5, v6, v7, v01, v07, v21, v23;
			Kvector v43, v45, v65, v67, v0123, v0721, v4365, v4567;
			double t = 1.5;//truncation adjustment factor
			Kvector u;
			for (j = 0; j < 8; j++)
				if (P->ftype[P->incid[j][i]] == 3)
					break;
			v0 = v[P->incid[j][i]];//real {5/3}
			v1 = v[P->incid[(j+1)%8][i]];//ideal
			v2 = v[P->incid[(j+2)%8][i]];//real {3}
			v3 = v[P->incid[(j+3)%8][i]];//ideal
			v4 = v[P->incid[(j+4)%8][i]];//real {5/2}
			v5 = v[P->incid[(j+5)%8][i]];//ideal
			v6 = v[P->incid[(j+6)%8][i]];//real {3/2}
			v7 = v[P->incid[(j+7)%8][i]];//ideal
			//compute intersections
			u = cross (v1, v3);
			v0123 = sum (v0, scale (dot (cross (diff (v2, v0), v3),
				u) / dot (u,u), v1));
			u = cross (v7, v1);
			v0721 = sum (v0, scale (dot (cross (diff (v2, v0), v1),
				u) / dot (u,u), v7));
			u = cross (v5, v7);
			v4567 = sum (v4, scale (dot (cross (diff (v6, v4), v7),
				u) / dot (u,u), v5));
			u = cross (v3, v5);
			v4365 = sum (v4, scale (dot (cross (diff (v6, v4), v5),
				u) / dot (u,u), v3));
			//compute truncations
			v01 = sum (v0 , scale (t, diff (v0123, v0)));
			v23 = sum (v2 , scale (t, diff (v0123, v2)));
			v07 = sum (v0 , scale (t, diff (v0721, v0)));
			v21 = sum (v2 , scale (t, diff (v0721, v2)));
			v45 = sum (v4 , scale (t, diff (v4567, v4)));
			v67 = sum (v6 , scale (t, diff (v4567, v6)));
			v43 = sum (v4 , scale (t, diff (v4365, v4)));
			v65 = sum (v6 , scale (t, diff (v4365, v6)));
			PolyhedronPoint(v01);
			PolyhedronPoint(v23);
			PolyhedronPoint(v0123);
			PolyhedronPoint(v07);
			PolyhedronPoint(v21);
			PolyhedronPoint(v0721);
			PolyhedronPoint(v45);
			PolyhedronPoint(v67);
			PolyhedronPoint(v4567);
			PolyhedronPoint(v43);
			PolyhedronPoint(v65);
			PolyhedronPoint(v4365);
		}


// Face list:
// Each face is printed in a separate line, by listing the indices of its
// vertices. In the non-simple case, the path is represented by the
// triangulation, each triangle consists of two polyhedron vertices and one
// auxiliary vertex.

	ii = V;
	for (i = 0; i < F; i++) {

		PolyhedronFace();

		if (*star) {
			if (P->K == 5 && P->D > 30 || denominator (P->m[0]) != 1) {
				for (j = 0; j < P->M - 1; j++)
					PolyhedronTriangle(P->incid[j][i], P->incid[j+1][i], ii);

				PolyhedronTriangle(P->incid[j][i], P->incid[0][i], ii++);

			} else if (P->even != -1) {
				if (hit[i])
				{
					PolyhedronTriangle(P->incid[3][i], P->incid[0][i], ii);
					PolyhedronTriangle(P->incid[1][i], P->incid[2][i], ii);
				}
				else
				{
					PolyhedronTriangle(P->incid[0][i], P->incid[1][i], ii);
					PolyhedronTriangle(P->incid[2][i], P->incid[3][i], ii);
				}
				ii++;
			} else if (P->hemi && P->index != last_uniform - 1) {
				j = !P->ftype[P->incid[0][i]];
				PolyhedronTriangle(ii, ii+1, ii+2);
				PolyhedronQuad(P->incid[j][i], ii+2, P->incid[j+2][i], ii+5);
				PolyhedronTriangle(ii+3, ii+4, ii+5);
				ii += 6;
			} else if (P->index == last_uniform - 1) {
				for (j = 0; j < 8; j++)
					if (P->ftype[P->incid[j][i]] == 3)
						break;
				PolyhedronTriangle(ii, ii+1, ii+2);
				PolyhedronQuad(P->incid[j][i], ii+2, P->incid[(j+2)%8][i], ii+5);
				PolyhedronTriangle(ii+3, ii+4, ii+5);

				PolyhedronTriangle(ii+6, ii+7, ii+8);
				PolyhedronQuad(P->incid[(j+4)%8][i], ii+8, P->incid[(j+6)%8][i], ii+11);
				PolyhedronTriangle(ii+9, ii+10, ii+11);
				ii += 12;
			} else {
				for (j = 0; j < P->M; j++)
					PolyhedronPathPoint(P->incid[j][i]);
			}
		} else {
			int split = (frac (P->n[P->ftype[i]]),
				 frax.d != 1 && frax.d != frax.n - 1);
			for (j = 0; j < V; j++) {
				for (k = 0; k < P->M; k++)
					if (P->incid[k][j] == i)
						break;
				if (k != P->M)
					break;
			}
			if (!split)
				PolyhedronPathPoint(j);
			ll = j;
			for (l = P->adj[k][j]; l != j; l = P->adj[k][l]) {
				for (k = 0; k < P->M; k++)
					if (P->incid[k][l] == i)
						break;
				if (P->adj[k][l] == ll) k = mod (k + 1 , P->M);
				if (!split)
					PolyhedronPathPoint(l);
				else
					PolyhedronTriangle(ll, l, ii);
				ll = l;
			}
			if (split)
				PolyhedronTriangle(ll, j, ii++);
		}

	}


// Face color indices - for pathhedra with multiple face types
// For non-simple faces, the index is repeated as many times as needed by the
// triangulation.

	free (v);
	free (f);
	return 1;
}

// Create a polyhedron
void CPrimitives::OnModifyPolyhedron(bool SaveUndoAction, bool CreateGeometry)
{
	// Create geometry as needed ...
	if(CreateGeometry)
		{
			// Make it
			LoadPolyhedron(m_PolyhedronNumber, m_PolyhedronDualCheck);

			// Name our mesh after the polyhedron type ...
			sdpString name = m_PolyhedronDualCheck ? uniform[m_PolyhedronNumber-1].dual : uniform[m_PolyhedronNumber-1].name;

			// Create a mesh ...
			if(SaveUndoAction)
				k3dStartStateChangeSet(Document());

			k3dIMesh* mesh = CreateMesh();
			g_assert(mesh);

			if(SaveUndoAction)
				k3dFinishStateChangeSet(Document(), "Create Polyhedron");

			k3d::Editing editing(mesh);
			ManageMesh(mesh, name);

			// Add geometry to the mesh ...
			AddGeometry(mesh, m_PolyhedronPoints.size(), m_PolyhedronPaths.size());

			// For each path ...
			for(gulong i = 0; i < m_PolyhedronPaths.size(); i++)
				{
					// Attach points ...
					for(gulong j = 0; j < m_PolyhedronPaths[i].size(); j++)
						{
							m_Paths[0][i]->Points()->AddPoint(m_Points[0][m_PolyhedronPaths[i][j]]);
						}
				}

			// Set materials
			SetDefaultMaterial();
		}

	// Calculate point positions ...
	for(gulong i = 0; i < m_PolyhedronPoints.size(); i++)
		{
			sdpVector3 coords = m_PolyhedronPoints[i] * m_PolyhedronScaling * m_PolyhedronLength.Value();

			m_Points[0][i]->Location()->SetLocalXYZ(coords);
		}
}

// Create a cushion ...
void CPrimitives::OnModifyCushion(bool SaveUndoAction, bool CreateGeometry)
{
	// Sanity checks ...
	const gulong cushionlengthsegments = static_cast<gulong>(m_CushionLengthSegments.Value());
	const gulong cushionradialsegments = static_cast<gulong>(m_CushionRadialSegments.Value());

	g_return_if_fail(cushionlengthsegments);
	g_return_if_fail(cushionradialsegments);


	// Build a sphere ...
	std::vector<sdpVector3> spherepoints;
	const gdouble spherediameter = static_cast<gdouble>(m_CushionDiameter.Value());
	const gulong N = cushionlengthsegments*4;
	const gulong R = cushionradialsegments*2-1;

	gulong i, j, n, c, d, e, f;

	for(j = 0; j < N; j++)
		for(i = 0; i < R; i++)
			{
				//gdouble  l = spherediameter/2 - (gdouble)(i+1)*spherediameter/((gdouble)R+1);
				gdouble l = sin(M_PI/2 -(gdouble)(i+1)*M_PI / (gdouble)(R+1))*spherediameter/2;
				gdouble r = sqrt(spherediameter*spherediameter/4 - l*l);

				spherepoints.push_back(sdpVector3( r*sin((gdouble)j*2*M_PI / (gdouble)N), l, r*cos((gdouble)j*2*M_PI / (gdouble)N)));
			}
	spherepoints.push_back(sdpVector3(0, spherediameter/2, 0));
	spherepoints.push_back(sdpVector3(0, -spherediameter/2, 0));


	// Cut sphere into 8 parts using distances ...
	const gdouble cushionxlength = static_cast<gdouble>(m_CushionXLength.Value());
	const gdouble cushionylength = static_cast<gdouble>(m_CushionYLength.Value());
	const gdouble cushionzlength = static_cast<gdouble>(m_CushionZLength.Value());

	cushionpoints.clear();
	c = (R+1)/2;

	sdpVector3 tP = sdpVector3(cushionxlength / 2.0, cushionylength / 2.0, cushionzlength / 2.0);
	for(j = 0; j < N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j : j*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(cushionxlength / 2.0, -cushionylength / 2.0, cushionzlength / 2.0);
	for(j = 0; j < N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j : j*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	tP = sdpVector3(cushionxlength / 2.0, cushionylength / 2.0, -cushionzlength / 2.0);
	for(j = N / 4; j < N / 2+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j : j*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(cushionxlength / 2.0, -cushionylength / 2.0, -cushionzlength / 2.0);
	for(j = N / 4; j < N / 2+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j : j*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	tP = sdpVector3(-cushionxlength / 2.0, cushionylength / 2.0, -cushionzlength / 2.0);
	for(j = N / 2; j < 3*N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j : j*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(-cushionxlength / 2.0, -cushionylength / 2.0, -cushionzlength / 2.0);
	for(j = N / 2; j < 3*N / 4+1; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j : j*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);

	tP = sdpVector3(-cushionxlength / 2.0, cushionylength / 2.0, cushionzlength / 2.0);
	for(j = 3*N / 4; j <= N; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R == 1 ? j % N :(j % N)*R+i] + tP);

	cushionpoints.push_back(spherepoints[N*R] + tP);

	tP = sdpVector3(-cushionxlength / 2.0, -cushionylength / 2.0, cushionzlength / 2.0);
	for(j = 3*N / 4; j <= N; j++)
		for(i = 0; i < c; i++)
			cushionpoints.push_back(spherepoints[R / 2 +(R == 1 ? j % N :(j % N)*R+i)] + tP);

	cushionpoints.push_back(spherepoints[N*R+1] + tP);


	// Create geometry as needed ...
	if(CreateGeometry)
		{
			if(SaveUndoAction)
				k3dStartStateChangeSet(Document());

			// Create a mesh ...
			k3dIMesh* mesh = CreateMesh();
			g_assert(mesh);

			if(SaveUndoAction)
				k3dFinishStateChangeSet(Document(), "Create Cushion");

			k3d::Editing editing(mesh);
			ManageMesh(mesh, "Cushion");

			// Clear paths
			m_CushionPaths.clear();

			// Make sphere parts faces ...
			d = c*(N/4+1)+1;

			for(n = 0; n < 8; n++)
			{

				for(j = 0; j < N / 4; j++)
					for(i = 0; i < c-1; i++)
						AddCushionQuad(n*d + c*j + i, n*d + c*j + i+1, n*d + c*(j+1) + i+1, n*d + c*(j+1) + i);

				for(j = 0; j < N / 4; j ++)
					if(n % 2 == 0)
						AddCushionTriangle(n*d + c*j, n*d + c*(j+1), (n+1) * d - 1);
					else
						AddCushionTriangle(n*d + c*(j+1) - 1, (n+1) * d - 1, n*d + c*(j+2) - 1);
			}

			// Link parts ...
			for(n = 0; n < 8; n++)
				{
					e = (n+1)*d - c-1;
					f = ((n+2) % 8)*d;

					for(j = 0; j < c-1; j++)
						AddCushionQuad(e+j, e+j+1, f+j+1, f+j);
				}

			AddCushionQuad(0, d - 1, 7*d - 1, 7*d - c-1);
			AddCushionQuad(d - c-1, 2*d, 3*d - 1, d - 1);
			AddCushionQuad(3*d - c-1, 4*d, 5*d - 1, 3*d - 1);
			AddCushionQuad(5*d - c-1, 6*d, 7*d - 1, 5*d - 1);

			AddCushionQuad(d + c-1, 8*d - 2, 8*d - 1, 2*d - 1);
			AddCushionQuad(2*d - 2, 2*d - 1, 4*d - 1, 3*d + c-1);
			AddCushionQuad(4*d - 2, 4*d - 1, 6*d - 1, 5*d + c-1);
			AddCushionQuad(6*d - 2, 6*d - 1, 8*d - 1, 7*d + c-1);


			for(n = 0; n < 4; n++)
				{
					e = d*(n*2) + c-1;
					f = d*(n*2+1);

					for(j = 0; j < N / 4; j++)
						AddCushionQuad(e+j*c, f+j*c, f+(j+1)*c, e+(j+1)*c);
				}

			AddCushionQuad(c-1, d*7 - 2, d*8 - c-1, d);
			AddCushionQuad(d-2, d*2 - c-1, d*3, d*2 + c-1);
			AddCushionQuad(d*3 - 2, d*4 - c-1, d*5, d*4 + c-1);
			AddCushionQuad(d*5 - 2, d*6 - c-1, d*7, d*6 + c-1);

			AddCushionQuad(d-1, d*3 - 1, d*5 - 1, d*7 - 1);
			AddCushionQuad(d*2 - 1, d*8 - 1, d*6 - 1, d*4 - 1);


			// Add geometry to the mesh ...
			AddGeometry(mesh, cushionpoints.size(), m_CushionPaths.size());

			// For each path ...
			for(gulong i = 0; i < m_CushionPaths.size(); i++)
				{
					// Attach points ...
					for(gulong j = 0; j < m_CushionPaths[i].size(); j++)
						m_Paths[0][i]->Points()->AddPoint(m_Points[0][m_CushionPaths[i][j]]);
				}

			// Set materials
			SetDefaultMaterial();
		}


	// Calculate point positions ...
	for(i = 0; i < cushionpoints.size(); i++)
		{
			sdpVector3 coords = cushionpoints[i];

			m_Points[0][i]->Location()->SetLocalXYZ(coords);
		}
}
*/

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CPrimitives::EditObject()
{
	new Properties(this);

	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

void CPrimitives::BuildGeometry()
{
	m_CurrentPrimitive->BuildGeometry(*this);
}

void CPrimitives::UpdateGeometry()
{
	m_CurrentPrimitive->UpdateGeometry(*this);
}

void CPrimitives::CancelGeometry()
{
	m_CurrentPrimitive->CancelGeometry();
}

void CPrimitives::FinishGeometry()
{
	m_CurrentPrimitive->FinishGeometry();
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives::Properties

const sdpString CPrimitives::Properties::control_pageswitch = "pageswitch";

const sdpString CPrimitives::Properties::control_grid = "grid";
const sdpString CPrimitives::Properties::control_cube = "cube";
const sdpString CPrimitives::Properties::control_disk = "disk";
const sdpString CPrimitives::Properties::control_cone = "cone";
const sdpString CPrimitives::Properties::control_cylinder = "cylinder";
const sdpString CPrimitives::Properties::control_cushion = "cushion";
const sdpString CPrimitives::Properties::control_sphere = "sphere";
const sdpString CPrimitives::Properties::control_torus = "torus";
const sdpString CPrimitives::Properties::control_polyhedron = "polyhedron";

CPrimitives::Properties::Properties(CPrimitives* const Object) :
	PropertiesBase(Object, false, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "primitives.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton("diskradius", m_Object->m_Disk.m_Radius, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("disksegments", m_Object->m_Disk.m_Segments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("diskpower", m_Object->m_Disk.m_Power, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("conelength", m_Object->m_Cone.m_Length, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("coneradius", m_Object->m_Cone.m_Radius, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("conelengthpower", m_Object->m_Cone.m_LengthPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("coneradialpower", m_Object->m_Cone.m_RadialPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("conelengthsegments", m_Object->m_Cone.m_LengthSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("coneradialsegments", m_Object->m_Cone.m_RadialSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("cylinderlength", m_Object->m_Cylinder.m_Length, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cylinderradius", m_Object->m_Cylinder.m_Radius, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cylinderradialpower", m_Object->m_Cylinder.m_RadialPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cylinderlengthsegments", m_Object->m_Cylinder.m_LengthSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cylinderradialsegments", m_Object->m_Cylinder.m_RadialSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("gridxlength", m_Object->m_Grid.m_XLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("gridylength", m_Object->m_Grid.m_YLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("gridxsegments", m_Object->m_Grid.m_XSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("gridysegments", m_Object->m_Grid.m_YSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("cubexlength", m_Object->m_Cube.m_XLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cubeylength", m_Object->m_Cube.m_YLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cubezlength", m_Object->m_Cube.m_ZLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cubexsegments", m_Object->m_Cube.m_XSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cubeysegments", m_Object->m_Cube.m_YSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cubezsegments", m_Object->m_Cube.m_ZSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("spherexlength", m_Object->m_Sphere.m_XLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("sphereylength", m_Object->m_Sphere.m_YLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("spherezlength", m_Object->m_Sphere.m_ZLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("spherelengthpower", m_Object->m_Sphere.m_LengthPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("sphereradialpower", m_Object->m_Sphere.m_RadialPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("spherelengthsegments", m_Object->m_Sphere.m_LengthSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("sphereradialsegments", m_Object->m_Sphere.m_RadialSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("torusinnerradius", m_Object->m_Torus.m_InnerRadius, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("torusouterradius", m_Object->m_Torus.m_OuterRadius, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("torusradialpower", m_Object->m_Torus.m_RadialPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("toruscrosssectionpower", m_Object->m_Torus.m_CrossSectionPower, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("torusradialsegments", m_Object->m_Torus.m_RadialSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("toruscrosssectionsegments", m_Object->m_Torus.m_CrossSectionSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	Connectk3dSpinButton("cushionlengthsegments", m_Object->m_Cushion.m_LengthSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cushionradialsegments", m_Object->m_Cushion.m_RadialSegments, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cushiondiameter", m_Object->m_Cushion.m_Diameter, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cushionxlength", m_Object->m_Cushion.m_XLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cushionylength", m_Object->m_Cushion.m_YLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	Connectk3dSpinButton("cushionzlength", m_Object->m_Cushion.m_ZLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CPrimitives>(m_Object, &CPrimitives::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Initialize our tab ...
	gint page = 0;
	CPrimitives::Primitive* const current = m_Object->m_CurrentPrimitive;
	if(current == &m_Object->m_Grid)
		page = 0;
	else if(current == &m_Object->m_Cube)
		page = 1;
	else if(current == &m_Object->m_Disk)
		page = 2;
	else if(current == &m_Object->m_Cone)
		page = 3;
	else if(current == &m_Object->m_Cylinder)
		page = 4;
	else if(current == &m_Object->m_Sphere)
		page = 5;
	else if(current == &m_Object->m_Torus)
		page = 6;
	else if(current == &m_Object->m_Cushion)
		page = 7;
//	else if(current == &m_Object->m_Polyhedron)
//		page = 8;

	Notebook("tabs").SetPage(page);

/*
	Connectk3dSpinButton("polyhedronlength", m_Object->m_PolyhedronLength, k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionUpdate2<CPrimitives>(m_Object, false, false, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);

	// Build pathhedra clist
	sdpGtkCList polyhedron_clist = CList("polyhedronchoice");

	polyhedron_clist.SetColumnTitle(0, "Nb");
	polyhedron_clist.SetColumnTitle(1, "Name");
	polyhedron_clist.ShowColumnTitles();

	sdpString number;
	const char* clistdata[2];

	for(gint j = 0; j < 80; j++)
		{
			number = sdpFormatString("%d", j+1);
			clistdata[0] = number;
			clistdata[1] = uniform[j].name;
			polyhedron_clist.Append(clistdata);

			number=sdpFormatString("%d*", j+1);
			clistdata[0] = number;
			clistdata[1] = uniform[j].dual;
			polyhedron_clist.Append(clistdata);
		}

	CList("polyhedronchoice").SetColumnWidth(0, 25);
*/

	Show();
}

void CPrimitives::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);

	PropertiesBase::OnOK();
}

void CPrimitives::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);

	PropertiesBase::OnCancel();
}

void CPrimitives::Properties::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "pageswitch")
		OnPageSwitch(dynamic_cast<sdpGtkEventNotebookSwitchPage*>(Event));
/*
	else if(Event->Name() == "polyhedronchoice")
		OnPolyhedronChoice(Event);
*/
	else
		PropertiesBase::OnEvent(Event);
}

sdpResult CPrimitives::Properties::Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);

	gint pagenumber = 0;
	if(Name == control_pageswitch)
		{
			if(Arguments == control_grid)
				pagenumber = 0;
			else if(Arguments == control_cube)
				pagenumber = 1;
			else if(Arguments == control_disk)
				pagenumber = 2;
			else if(Arguments == control_cone)
				pagenumber = 3;
			else if(Arguments == control_cylinder)
				pagenumber = 4;
			else if(Arguments == control_sphere)
				pagenumber = 5;
			else if(Arguments == control_torus)
				pagenumber = 6;
			else if(Arguments == control_cushion)
				pagenumber = 7;
			else if(Arguments == control_polyhedron)
				pagenumber = 8;
			else
				g_return_val_if_fail(0, SDP_E_FAIL);

			Notebook("tabs").InteractiveShowPage(pagenumber, k3dOptionsGetTutorialSpeed(k3dApplication()->Options()), true);

			return SDP_S_OK;
		}

	return PropertiesBase::Command(Name, Arguments, Result);
}

// User switched from one tab to another ...
void CPrimitives::Properties::OnPageSwitch(sdpGtkEventNotebookSwitchPage* const Event)
{
	// Sanity checks ...
	g_return_if_fail(Event);

	// Switch to a new primitive type ...
	sdpString pagename("");
	Primitive* newprimitive = 0;
	switch(Event->PageNumber())
		{
			case 0:
				pagename = control_grid;
				newprimitive = &m_Object->m_Grid;
				break;
			case 1:
				pagename = control_cube;
				newprimitive = &m_Object->m_Cube;
				break;
			case 2:
				pagename = control_disk;
				newprimitive = &m_Object->m_Disk;
				break;
			case 3:
				pagename = control_cone;
				newprimitive = &m_Object->m_Cone;
				break;
			case 4:
				pagename = control_cylinder;
				newprimitive = &m_Object->m_Cylinder;
				break;
			case 5:
				pagename = control_sphere;
				newprimitive = &m_Object->m_Sphere;
				break;
			case 6:
				pagename = control_torus;
				newprimitive = &m_Object->m_Torus;
				break;
			case 7:
				pagename = control_cushion;
				newprimitive = &m_Object->m_Cushion;
				break;
			case 8:
				pagename = control_polyhedron;
//				newprimitive = &m_Object->m_Polyhedron;
				break;
			default:
				g_return_if_fail(0);	// Did you add a new primitive type without updating this code?
		}

	// Sanity checks ...
	g_return_if_fail(pagename.size());
	g_return_if_fail(newprimitive);

	// Record this command for tutorials ...
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_pageswitch, pagename);

	// Switch to the new primitive type ...
	m_Object->CancelGeometry();
	m_Object->m_CurrentPrimitive = newprimitive;
	m_Object->BuildGeometry();

	k3dRedrawAll(m_Object->Document(), false);
}

/*
void CPrimitives::Properties::OnPolyhedronChoice(sdpGtkEvent* Event)
{
	sdpGtkEventCListSelectRow* event = static_cast<sdpGtkEventCListSelectRow*>(Event);

	guint row = event->Row();

	m_Object->m_PolyhedronNumber = row / 2 + 1;
	m_Object->m_PolyhedronDualCheck = (row % 2 == 1);
	m_Object->Update(false, true);
	k3dRedrawAll(m_Object->Document(), false);
}
*/


