#ifndef PRIMITIVES_H
#define PRIMITIVES_H

// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CPrimitives K-3D tool, for creating simple geometric primitives
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dbuildgeometry.h>
#include <k3dobjectdialog.h>
#include <k3dtool.h>

// Forward class declarations
class sdpGtkEventNotebookSwitchPage;

/////////////////////////////////////////////////////////////////////////////
// CPrimitives

typedef k3dTool CPrimitivesBase;

class CPrimitives :
	public CPrimitivesBase
{
public:
	CPrimitives(sdpIUnknown* Initializer);

	// k3dIObject implementation
	sdpResult EditObject();

	void BuildGeometry();
	void UpdateGeometry();
	void CancelGeometry();
	void FinishGeometry();

private:
	typedef k3dObjectDialog PropertiesBase;
	/// Provides a GUI for the CPrimitives tool
	class Properties :
		public PropertiesBase
	{
	public:
		Properties(CPrimitives* const Object);

		// k3dICommandNode implementation
		sdpResult Command(SDPCSTRING Name, SDPCSTRING Arguments, sdpIUnknown** Result);

	private:
		CPrimitives* const m_Object;

		void OnOK();
		void OnCancel();

		// Custom event handling ...
		void OnEvent(sdpGtkEvent* Event);
		void OnPageSwitch(sdpGtkEventNotebookSwitchPage* const Event);

//		void OnPolyhedronChoice(sdpGtkEvent* Event);

		static const sdpString control_pageswitch;

		static const sdpString control_disk;
		static const sdpString control_cone;
		static const sdpString control_cylinder;
		static const sdpString control_cube;
		static const sdpString control_grid;
		static const sdpString control_sphere;
		static const sdpString control_torus;
		static const sdpString control_polyhedron;
		static const sdpString control_cushion;
	};
	friend class Properties;

	/// Abstract interface for objects that create a specific type of primitive
	class Primitive
	{
	public:
		virtual ~Primitive() {}

		virtual void BuildGeometry(CPrimitives& Object) = 0;
		virtual void UpdateGeometry(CPrimitives& Object) = 0;
		virtual void CancelGeometry() = 0;
		virtual void FinishGeometry() = 0;
	};

	/// Creates a Disk-shaped polygon primitive
	class Disk :
		public Primitive
	{
	public:
		Disk(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		typedef enum
		{
			SOLID,
			TRIANGLES
		} Type;
		Type m_Type;

		k3dVariable<gdouble> m_Radius;
		k3dVariable<gdouble> m_Segments;
		k3dVariable<gdouble> m_Power;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Disk;

	/// Creates a Cone-shaped polygon primitive
	class Cone :
		public Primitive
	{
	public:
		Cone(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_Length;
		k3dVariable<gdouble> m_Radius;
		k3dVariable<gdouble> m_LengthPower;
		k3dVariable<gdouble> m_RadialPower;
		k3dVariable<gdouble> m_LengthSegments;
		k3dVariable<gdouble> m_RadialSegments;
		bool m_CreateBase;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Cone;

	/// Creates a Cylinder-shaped polygon primitive
	class Cylinder :
		public Primitive
	{
	public:
		Cylinder(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_Length;
		k3dVariable<gdouble> m_Radius;
		k3dVariable<gdouble> m_RadialPower;
		k3dVariable<gdouble> m_LengthSegments;
		k3dVariable<gdouble> m_RadialSegments;
		bool m_CreateBase;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Cylinder;

	/// Creates a Grid-shaped polygon primitive
	class Grid :
		public Primitive
	{
	public:
		Grid(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_XLength;
		k3dVariable<gdouble> m_YLength;
		k3dVariable<gdouble> m_XSegments;
		k3dVariable<gdouble> m_YSegments;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Grid;

	/// Creates a Cube-shaped polygon primitive
	class Cube :
		public Primitive
	{
	public:
		Cube(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_XLength;
		k3dVariable<gdouble> m_YLength;
		k3dVariable<gdouble> m_ZLength;
		k3dVariable<gdouble> m_XSegments;
		k3dVariable<gdouble> m_YSegments;
		k3dVariable<gdouble> m_ZSegments;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Cube;

	/// Creates a Sphere-shaped polygon primitive
	class Sphere :
		public Primitive
	{
	public:
		Sphere(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_XLength;
		k3dVariable<gdouble> m_YLength;
		k3dVariable<gdouble> m_ZLength;
		k3dVariable<gdouble> m_LengthPower;
		k3dVariable<gdouble> m_RadialPower;
		k3dVariable<gdouble> m_LengthSegments;
		k3dVariable<gdouble> m_RadialSegments;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Sphere;

	/// Creates a Torus-shaped polygon primitive
	class Torus :
		public Primitive
	{
	public:
		Torus(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_InnerRadius;
		k3dVariable<gdouble> m_OuterRadius;
		k3dVariable<gdouble> m_RadialPower;
		k3dVariable<gdouble> m_CrossSectionPower;
		k3dVariable<gdouble> m_RadialSegments;
		k3dVariable<gdouble> m_CrossSectionSegments;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Torus;

	/// Creates a Cushion-shaped polygon primitive
	class Cushion :
		public Primitive
	{
	public:
		Cushion(k3dIDocument* const Document);

		void BuildGeometry(CPrimitives& Object);
		void UpdateGeometry(CPrimitives& Object);
		void CancelGeometry();
		void FinishGeometry();

	private:
		friend class Properties;

		k3dIDocument* const m_Document;

		k3dVariable<gdouble> m_LengthSegments;
		k3dVariable<gdouble> m_RadialSegments;
		k3dVariable<gdouble> m_Diameter;
		k3dVariable<gdouble> m_XLength;
		k3dVariable<gdouble> m_YLength;
		k3dVariable<gdouble> m_ZLength;

		k3d::NewGeometryCollection m_NewGeometry;
	};
	friend class Cushion;

	// Supported primitive types ...
	Grid m_Grid;
	Cube m_Cube;
	Disk m_Disk;
	Cone m_Cone;
	Cylinder m_Cylinder;
	Sphere m_Sphere;
	Torus m_Torus;
	Cushion m_Cushion;

	/// Refers to the current Primitive object under construction
	Primitive* m_CurrentPrimitive;

/*
	gulong m_PolyhedronNumber;
	bool m_PolyhedronDualCheck;
	k3dVariable<gdouble> m_PolyhedronLength;


protected:
	// Special variables and methods to generate polyhedra ...
	typedef std::vector<sdpVector3> PolyhedronPoints;
	PolyhedronPoints m_PolyhedronPoints;

	sdpString m_PolyhedronProperties;
	double m_PolyhedronScaling;

	typedef std::vector<int> PolyhedronPath;
	typedef std::vector<PolyhedronPath> PolyhedronPaths;
	PolyhedronPaths m_PolyhedronPaths;

	void LoadPolyhedron(int PolyhedronIndex, bool Dual);
	void PolyhedronPoint(Kvector v);
	void PolyhedronFace();
	void PolyhedronPathPoint(int PointIndex);
	void PolyhedronTriangle(int PointIndex1, int PointIndex2, int PointIndex3);
	void PolyhedronQuad(int PointIndex1, int PointIndex2, int PointIndex3, int PointIndex4);

	int vrmodel( Polyhedron *P, Kvector *v, int V, Kvector *f, int F,	char *name, char *star, double azimuth, double elevation, double freeze);
*/
};


#endif // PRIMITIVES_H
