// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CScale K-3D tool, which applies a scaling transformation to locations
		\author Tim Shead (tshead@k-3d.com)
*/

#include "scale.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3distaterecorder.h>
#include <k3ditransform.h>

#include <sdpgtkevent.h>

namespace
{

class ScaleLocations
{
public:
	ScaleLocations(const gdouble XScale, const gdouble YScale, const gdouble ZScale) :
		m_Matrix(sdpScaling3D(sdpVector3(XScale, YScale, ZScale)))
	{
	}
	
	void operator()(k3d::SelectedLocationProxy& Location)
	{
		Location->SetLocalXYZ(m_Matrix * Location.OldXYZ());
	}
	
private:
	const sdpMatrix4 m_Matrix;
};

} // namespace


/////////////////////////////////////////////////////////////////////////////
// CScale

CScale::CScale(sdpIUnknown* Initializer) :
	CScaleBase(Initializer),
	k3dMouseEventObserver("LMB-drag to scale selected Points; control-LMB-drag to constrain to Active Axis"),
	m_X("x", 1.0, 0, 0),
	m_Y("y", 1.0, 0, 0),
	m_Z("z", 1.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CScale::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	return CScaleBase::InitComplete();
}

sdpResult CScale::EditObject()
{
	new CScaleDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CScale::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CScale::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	// If the user holds down either shift key, constrain scale factors to be equal ...
	if(ShiftDown(Modifiers))
		{
			k3dSetValue(m_X, m_X.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
			k3dSetValue(m_Y, m_Y.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
			k3dSetValue(m_Z, m_Z.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
		}
	// If the user holds down either control key, only scale along the active axis ...
	else if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_X, m_X.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_Y, m_Y.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_Z, m_Z.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, scale along the two non-active axes ...
	else
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_Z, m_Z.Value() - xfactor);
						k3dSetValue(m_Y, m_Y.Value() - yfactor);
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_X, m_X.Value() - xfactor);
						k3dSetValue(m_Z, m_Z.Value() - yfactor);
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_X, m_X.Value() - xfactor);
						k3dSetValue(m_Y, m_Y.Value() - yfactor);
						break;
					default:
						g_assert(0);
				}
		}

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CScale::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CScale::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CScale::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), ScaleLocations(m_X.Value(), m_Y.Value(), m_Z.Value()));
}

void CScale::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CScale::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Scale " + sdpToString(sdpVector3(m_X.Value(), m_Y.Value(), m_Z.Value())));
}

//////////////////////////////////////////////////////////////////////////////////////
// CScaleDialog

const sdpString CScaleDialog::dialogname = "properties";
const sdpString CScaleDialog::control_x = "x";
const sdpString CScaleDialog::control_y = "y";
const sdpString CScaleDialog::control_z = "z";

CScaleDialog::CScaleDialog(CScale* const Object) :
	CScaleDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "scale.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_x, m_Object->m_X, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CScale>(m_Object, &CScale::PushEditing, 0), new k3dActionMethodCall<CScale>(m_Object, &CScale::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CScale>(m_Object, &CScale::PopEditing, 0));
	Connectk3dSpinButton(control_y, m_Object->m_Y, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CScale>(m_Object, &CScale::PushEditing, 0), new k3dActionMethodCall<CScale>(m_Object, &CScale::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CScale>(m_Object, &CScale::PopEditing, 0));
	Connectk3dSpinButton(control_z, m_Object->m_Z, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CScale>(m_Object, &CScale::PushEditing, 0), new k3dActionMethodCall<CScale>(m_Object, &CScale::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CScale>(m_Object, &CScale::PopEditing, 0));

	Show();
}

