// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSlant class, which applies a "slant" transformation to an object's geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "slant.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

namespace
{

class SlantLocations
{
public:
	SlantLocations(const sdpVector3 Size, const k3dICamera::Axis Axis, const sdpVector3 Slant) :
		m_Size(Size),
		m_Axis(Axis),
		m_Slant(Slant)
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		double dx, dy, dz;

		sdpVector3 coords = Location.OldXYZ();

		switch(m_Axis)
			{
				case k3dICamera::XAXIS:
					dx = 0.0;
					dy = m_Size[2] ? coords[2] / m_Size[2] : coords[2];
					dz =  m_Size[1] ? coords[1] / m_Size[1] : coords[1];
					break;
				case k3dICamera::YAXIS:
					dx = m_Size[2] ? coords[2] / m_Size[2] : coords[2];
					dy = 0.0;
					dz = m_Size[0] ? coords[0] / m_Size[0] : coords[0];
					break;
				case k3dICamera::ZAXIS:
					dx = m_Size[1] ? coords[1] / m_Size[1] : coords[1];
					dy = m_Size[0] ? coords[0] / m_Size[0] : coords[0];
					dz = 0.0;
					break;
				default:
					g_return_if_fail(0);
			}

	/*
		if(m_Mirror)
			{
				dx = m_Xfalloff->GetValue1(fabs(dx), .0001);
				dy = m_Yfalloff->GetValue1(fabs(dy), .0001);
				dz = m_Zfalloff->GetValue1(fabs(dz), .0001);
			}
		else
			{
				dx = KSign<double>(dx) * m_Xfalloff->GetValue1(fabs(dx), .0001);
				dy = KSign<double>(dy) * m_Yfalloff->GetValue1(fabs(dy), .0001);
				dz = KSign<double>(dz) * m_Zfalloff->GetValue1(fabs(dz), .0001);
			}
	*/

		coords[0] += dx * m_Slant[0];
		coords[1] += dy * m_Slant[1];
		coords[2] += dz * m_Slant[2];

		Location->SetLocalXYZ(coords);
	}

private:
	const sdpVector3 m_Size;
	const k3dICamera::Axis m_Axis;
	const sdpVector3 m_Slant;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CSlant

CSlant::CSlant(sdpIUnknown* Initializer) :
	CSlantBase(Initializer),
	k3dMouseEventObserver("LMB-drag to slant selected Points; control-LMB-drag to constrain to Active Axis"),
	m_CameraActiveAxis(k3dICamera::ZAXIS),
	m_X("x", 0.0, 0, 0),
	m_Y("y", 0.0, 0, 0),
	m_Z("z", 0.0, 0, 0),
	m_Sensitivity("sensitivity", 0.1, 0, 0),
	m_Mirror(false)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSlant::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	std::copy(m_CurrentSelection.begin(), m_CurrentSelection.end(), k3d::Dimensions::Inserter(m_CurrentSelectionDimensions));

	return CSlantBase::InitComplete();
}

sdpResult CSlant::EditObject()
{
	new CSlantDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CSlant::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	// Save current active axis
	m_CameraActiveAxis = Camera->ActiveAxis();

	PushEditing();
	return SDP_S_OK;
}

sdpResult CSlant::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * m_Sensitivity.Value();
	double yfactor = (CurrentY - LastY) * m_Sensitivity.Value();

	// If the user holds down either control key, only translate along the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_X, m_X.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_Y, m_Y.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_Z, m_Z.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, translate along the two non-active axes ...
	else
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_Z, m_Z.Value() - xfactor);
						k3dSetValue(m_Y, m_Y.Value() - yfactor);
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_X, m_X.Value() - xfactor);
						k3dSetValue(m_Z, m_Z.Value() - yfactor);
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_X, m_X.Value() + xfactor);
						k3dSetValue(m_Y, m_Y.Value() - yfactor);
						break;
					default:
						g_assert(0);
				}
		}

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CSlant::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CSlant::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CSlant::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), SlantLocations(m_CurrentSelectionDimensions.Size(), m_CameraActiveAxis, sdpVector3(m_X.Value(), m_Y.Value(), m_Z.Value())));
}

void CSlant::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CSlant::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Slant");
}

//////////////////////////////////////////////////////////////////////////////////////
// CSlantDialog

const sdpString CSlantDialog::dialogname = "properties";
const sdpString CSlantDialog::control_x = "x";
const sdpString CSlantDialog::control_y = "y";
const sdpString CSlantDialog::control_z = "z";
const sdpString CSlantDialog::control_sensitivity = "sensitivity";

CSlantDialog::CSlantDialog(CSlant* const Object) :
	CSlantDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "slant.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_x, m_Object->m_X, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CSlant>(m_Object, &CSlant::PushEditing, 0), new k3dActionMethodCall<CSlant>(m_Object, &CSlant::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CSlant>(m_Object, &CSlant::PopEditing, 0));
	Connectk3dSpinButton(control_y, m_Object->m_Y, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CSlant>(m_Object, &CSlant::PushEditing, 0), new k3dActionMethodCall<CSlant>(m_Object, &CSlant::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CSlant>(m_Object, &CSlant::PopEditing, 0));
	Connectk3dSpinButton(control_z, m_Object->m_Z, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CSlant>(m_Object, &CSlant::PushEditing, 0), new k3dActionMethodCall<CSlant>(m_Object, &CSlant::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CSlant>(m_Object, &CSlant::PopEditing, 0));
	Connectk3dSpinButton(control_sensitivity, m_Object->m_Sensitivity, k3dApplication()->CommandRecorder(), 0, 0, 0, 0, 0);

	Show();
}


