// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSmooth K-3D tool, which "smooths" geometry by averaging point locations
		\author Tim Shead (tshead@k-3d.com)
*/

#include "smooth.h"

#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dimesh.h>
#include <k3dipath.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>
#include <k3dipoint.h>
#include <k3distate.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

namespace
{

class BuildLocationCollection
{
public:
	BuildLocationCollection(k3d::SelectedLocationArray& Collection) :
		m_Collection(Collection)
	{
	}

	void operator()(k3dIMesh* const Mesh)
	{
		typedef std::set<k3dIPath*> PathSet;
		typedef std::map<k3dIPoint*, PathSet> PointPathMap;
		typedef PointPathMap::iterator PointPathIterator;
		PointPathMap pointpathmap;
		
		// For each path in the mesh ...
		k3dIPathCollection* const paths = Mesh->Paths();
		for(K3DPATHHANDLE pathhandle = paths->FirstPath(); pathhandle; pathhandle = paths->NextPath(pathhandle))
			{
				// For each pathpoint in the path ...
				k3dIPath* const path = paths->Path(pathhandle);
				k3dIPathPointCollection* const pathpoints = path->Points();
				for(K3DPATHPOINTHANDLE pathpointhandle = pathpoints->FirstPathPoint(); pathpointhandle; pathpointhandle = pathpoints->NextPathPoint(pathpointhandle))
					{
						// If it refers to a selected point ...
						k3dIPathPoint* const pathpoint = pathpoints->PathPoint(pathpointhandle);
						if(SDP_S_OK == pathpoint->Point()->State()->QueryState(k3dIState::VISIBLESELECTED))
							{
								pointpathmap[pathpoint->Point()].insert(path);
							}
					}
			}

		// For each selected point ...
		for(PointPathIterator pointpath = pointpathmap.begin(); pointpath != pointpathmap.end(); pointpath++)
			{
				// Get the original coordinates ...
				sdpVector3 oldcoords = pointpath->first->Location()->LocalXYZ();
				
				// Calculate the average of all points in all paths that share this point ...
				gulong count = 0;
				sdpVector3 newcoords;
				for(PathSet::iterator path = pointpath->second.begin(); path != pointpath->second.end(); path++)
					{
						k3dIPathPointCollection* const pathpoints = (*path)->Points();
						for(K3DPATHPOINTHANDLE pathpointhandle = pathpoints->FirstPathPoint(); pathpointhandle; pathpointhandle = pathpoints->NextPathPoint(pathpointhandle))
							{
								sdpVector3 coords = pathpoints->PathPoint(pathpointhandle)->Point()->Location()->LocalXYZ();
								newcoords += coords;
								count++;
							}
					}
				newcoords /= count;

				
				// Insert the point into our collection ...
				m_Collection.push_back(k3d::SelectedLocationProxy(pointpath->first->Location(), oldcoords, newcoords));				
			}
	}
	
private:
	k3d::SelectedLocationArray& m_Collection;
};

class SmoothLocations
{
public:
	SmoothLocations(const gdouble SmoothingFactor) :
		m_SmoothingFactor(SmoothingFactor)
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		Location->SetLocalXYZ(sdpLerp(Location.OldXYZ(), Location.NewXYZ(), m_SmoothingFactor));
	}
	
private:
	const gdouble m_SmoothingFactor;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CSmooth

CSmooth::CSmooth(sdpIUnknown* Initializer) :
	CSmoothBase(Initializer),
	k3dMouseEventObserver("LMB-drag to smooth selected Paths"),
	m_SmoothingFactor("smoothing", 0.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CSmooth::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::for_each<k3dIMesh>(Document()->Objects(), BuildLocationCollection(m_CurrentSelection));
	return CSmoothBase::InitComplete();
}

sdpResult CSmooth::EditObject()
{
	new CSmoothDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CSmooth::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CSmooth::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	k3dSetValue(m_SmoothingFactor, m_SmoothingFactor.Value() + (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CSmooth::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CSmooth::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CSmooth::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), SmoothLocations(m_SmoothingFactor.Value()));
}

void CSmooth::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CSmooth::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Smooth " + sdpToString(m_SmoothingFactor.Value()));
}

//////////////////////////////////////////////////////////////////////////////////////
// CSmoothDialog

const sdpString CSmoothDialog::dialogname = "properties";
const sdpString CSmoothDialog::control_smoothingfactor = "smoothingfactor";

CSmoothDialog::CSmoothDialog(CSmooth* const Object) :
	CSmoothDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "smooth.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_smoothingfactor, m_Object->m_SmoothingFactor, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CSmooth>(m_Object, &CSmooth::PushEditing, 0), new k3dActionMethodCall<CSmooth>(m_Object, &CSmooth::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CSmooth>(m_Object, &CSmooth::PopEditing, 0));
	
	Show();
}


