// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSnap K-3D tool, which "snaps" a point to another point's location
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "snap.h"

#include <k3dmisc.h>
#include <k3dobjectcollection.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3dieditorengine.h>
#include <k3dilocation.h>
#include <k3dipoint.h>
#include <k3dipointcollection.h>
#include <k3distaterecorder.h>

#include <sdpgl.h>
#include <sdpgtkevent.h>
#include <iostream>

namespace
{

/// Functor for choosing points
class SelectPoint :
	public k3dIPointVisitor
{
public:
	SelectPoint(const sdpVector2 Mouse) :
		m_Mouse(Mouse),
		m_Point(0)
	{
	}

	k3dIPoint* Point() const
	{
		return m_Point;
	}

	void VisitPoint(k3dIPoint* const Point)
	{
		if(m_Point)
			return;

		if(SDP_S_OK != Point->State()->QueryState(k3dIState::VISIBLESELECTABLE))
			return;
									
		const sdpVector3 screencoords = Point->Location()->ScreenXYZ();

		if((fabs(screencoords[0] - m_Mouse[0]) < 4.0) && (fabs(screencoords[1] - m_Mouse[1]) < 4.0))
			m_Point = Point;
	}

	void operator()(k3dIPointCollection* const PointCollection)
	{
		if(m_Point)
			return;
			
		PointCollection->VisitPoints(this);
	}

private:
	const sdpVector2 m_Mouse;
	k3dIPoint* m_Point;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CSnap

CSnap::CSnap(sdpIUnknown* Initializer) :
	CSnapBase(Initializer),
	k3dMouseEventObserver("LMB-click Point to snap, then LMB-click Point to snap-to"),
	m_Point1(0),
	m_Point2(0)
{
	// Make sure that we get called by the rendering engine ...
	SetState(k3dIState::VISIBLE, true);
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CSnap::EditObject()
{
	new CSnapDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender implementation

sdpResult CSnap::EditorPostRender(k3dIEditorEngine* const Engine)
{
	// Sanity checks ...
	g_assert(Engine);
	
	// If nothing's selected, we're done!
	if(!m_Point1 && !m_Point2)
		return SDP_S_FALSE;

	// We need to work in screen coordinates ...
	Engine->Camera()->SnapshotScreenCoordinates();

	// Get window dimensions ...
	gdouble width = Engine->Camera()->Width();
	gdouble height = Engine->Camera()->Height();
	gdouble size = sdpMin(width, height);

	// Make sure they're non-zero (minimized window?) ...
	if(0.0 == size)
		return SDP_S_FALSE;
	
	// Draw selected points
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	// Push a new modelview matrix onto the stack ...
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, width, height, 0, -1, 1);
	glViewport(0, 0, GLint(width), GLint(height));

	glColor3d(1.0, 0.0, 0.0);
	glPointSize(6.0f);

	glBegin(GL_POINTS);
	
	if(m_Point1)
		{
			const sdpVector3 coords = m_Point1->Location()->ScreenXYZ();
			glVertex2d(coords[0], coords[1]);
		}
	if(m_Point2)
		{
			const sdpVector3 coords = m_Point2->Location()->ScreenXYZ();
			glVertex2d(coords[0], coords[1]);
		}

	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPopAttrib();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CSnap::OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	Camera->SnapshotScreenCoordinates();

	SelectPoint selection = k3d::for_each_if<k3dIPointCollection>(Document()->Objects(), SelectPoint(sdpVector2(CurrentX, CurrentY)), k3d::FlagIsSet<k3dIObject*>(k3dIState::VISIBLE));

	if(!selection.Point())
		return SDP_S_OK;

	if(m_Point1)
		{
			if(selection.Point() == m_Point1)
				{
					k3dApplication()->Error("You can't select the same point twice!", "Snap:");
					m_Point1 = 0;
					m_Point2 = 0;
				}
			else
				{	
					m_Point2 = selection.Point();
					PushEditing();
					UpdateGeometry();
					PopEditing();
				}
		}
	else
		{
			m_Point1 = selection.Point();
		}
		
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

void CSnap::PushEditing()
{
	// Sanity checks ...
	g_return_if_fail(m_Point1 && m_Point2);

	// Save it for posterity ...
	k3dStartStateChangeSet(Document());
	k3d::PushEditing(m_Point1);
}

void CSnap::UpdateGeometry()
{
	// Sanity checks ...
	g_return_if_fail(m_Point1 && m_Point2);
/*
	// Move the entire containing path ...
	sdpVector3 coords, coords2;
	if(m_MovePaths)
		{
			sdpVector3 start;
			m_Point1->Location()->GetOldXYZ(start);

			sdpVector3 end;
			m_Point2->Location()->GetOldXYZ(end);

			sdpVector3 delta = end - start;

			for(hr = m_PointMatchList->SetFirstMatch(); SDP_S_OK == hr; hr = m_PointMatchList->SetNextMatch())
				{
					m_PointLocation->GetOldXYZ(coords);
					coords[0] += dx;
					coords[1] += dy;
					coords[2] += dz;
					m_PointLocation->SetLocalXYZ(coords);
				}
		}
	// Move the solitary point ...
	else
*/
		{
			m_Point1->Location()->SetLocalXYZ(m_Point2->Location()->LocalXYZ());
		}

}

void CSnap::PopEditing()
{
	// Sanity checks ...
	g_return_if_fail(m_Point1 && m_Point2);

	k3d::PopEditing(m_Point1);
	k3dFinishStateChangeSet(Document(), "Snap");
	
	m_Point1 = 0;
	m_Point2 = 0;
}

////////////////////////////////////////////////////////////////////////////
// CSnapDialog

const sdpString CSnapDialog::dialogname = "properties";

CSnapDialog::CSnapDialog(CSnap* const Object) :
	CSnapDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "snap.gtkml"));

	Show();
}

