// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSphereize K-3D tool, which "sphereizes" objects by making them conform to the surface of a sphere
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "sphereize.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

namespace
{

class SphereizeLocations
{
public:
	SphereizeLocations(const gdouble SphereizeFactor, const gdouble TargetRadius) :
		m_SphereizeFactor(SphereizeFactor),
		m_TargetRadius(TargetRadius)
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		sdpVector3 coords = Location.OldXYZ();

		const gdouble radius = coords.Length();
		if(radius)
			{
				sdpVector3 normal(coords);
				normal.Normalize();
				
				coords += normal * (m_SphereizeFactor * (m_TargetRadius - radius));
				Location->SetLocalXYZ(coords);
			}
	}
	
private:
	const gdouble m_SphereizeFactor;
	const gdouble m_TargetRadius;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CSphereize

CSphereize::CSphereize(sdpIUnknown* Initializer) :
	CSphereizeBase(Initializer),
	k3dMouseEventObserver("LMB-drag to sphereize selected Points"),
	m_SphereizeFactor("sphereize_factor", 0.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CSphereize::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	std::copy(m_CurrentSelection.begin(), m_CurrentSelection.end(), k3d::Dimensions::Inserter(m_CurrentSelectionDimensions));

	return CSphereizeBase::InitComplete();
}

sdpResult CSphereize::EditObject()
{
	new CSphereizeDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CSphereize::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CSphereize::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	k3dSetValue(m_SphereizeFactor, m_SphereizeFactor.Value() + (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);
	
	return SDP_S_OK;
}

sdpResult CSphereize::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CSphereize::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CSphereize::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), SphereizeLocations(m_SphereizeFactor.Value(), m_CurrentSelectionDimensions.Extents().Length()));
}

void CSphereize::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CSphereize::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Sphereize " + sdpToString(m_SphereizeFactor.Value()));
}

//////////////////////////////////////////////////////////////////////////////////////
// CSphereizeDialog

const sdpString CSphereizeDialog::dialogname = "properties";
const sdpString CSphereizeDialog::control_spherefactor = "spherefactor";

CSphereizeDialog::CSphereizeDialog(CSphereize* const Object) :
	CSphereizeDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "sphereize.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_spherefactor, m_Object->m_SphereizeFactor, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CSphereize>(m_Object, &CSphereize::PushEditing, 0), new k3dActionMethodCall<CSphereize>(m_Object, &CSphereize::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CSphereize>(m_Object, &CSphereize::PopEditing, 0));
	
	Show();
}


