// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CSubDivide, a K-3D tool for subdividing paths
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "subdivide.h"

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dpathpointcollection.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>
#include <k3dipathpointcollection.h>
#include <k3dilocation.h>
#include <k3dpathpoint.h>

#include <sdpgtkevent.h>

namespace
{

///////////////////////////////////////////////////////////////
// CSubDivide::Path

CSubDivide::Path::Path(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PathBase(Engine, Mesh),
	m_CenterPoint(0)
{
}

CSubDivide::Path::Path(k3d::GeometryEngine& Engine, k3dIPath* const Path) :
	PathBase(Engine, Path),
	m_CenterPoint(0)
{
}

void CSubDivide::Path::Cancel()
{
	m_CenterPoint = 0;
}

k3d::EnginePointProxy& CSubDivide::Path::CenterPoint()
{
	if(0 == m_CenterPoint)
		{
			// Calculate some generic data about the path ...
			k3d::Dimensions dimensions;
			k3d::copy(Actual()->Points(), k3d::Dimensions::Inserter(dimensions));
	
			// Create a center point ...
			m_CenterPoint = &Engine().Point(Mesh());
			m_CenterPoint->Actual()->Location()->SetLocalXYZ(dimensions.Center());
		}
		
	return *m_CenterPoint;
}

///////////////////////////////////////////////////////////////
// CSubDivide::PathCenter

CSubDivide::PathCenter::PathCenter(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PathCenterBase(Engine, Mesh)
{
}

CSubDivide::PathCenter::PathCenter(k3d::GeometryEngine& Engine, k3dIPath* const Path) :
	PathCenterBase(Engine, Path)
{
}

void CSubDivide::PathCenter::BuildGeometry()
{
	// Skip anything with < 3 sides ...
	if(m_Points.size() < 3)
		return;
	
	// Create triangles between the center point and the edges ...
	for(gulong i = 0; i < m_Points.size(); i++)
		{
			k3d::EnginePathProxy& path = Engine().Path(Mesh());

			path.Actual()->Points()->AddPoint(m_Points[(i+0)%m_Points.size()]->Actual());
			path.Actual()->Points()->AddPoint(m_Points[(i+1)%m_Points.size()]->Actual());
			path.Actual()->Points()->AddPoint(CenterPoint().Actual());
		}

	// Mark the original for deletion ...
//	MarkForDeletion();
}

///////////////////////////////////////////////////////////////
// CSubDivide::PathCenterMidPoints

CSubDivide::PathCenterMidPoints::PathCenterMidPoints(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PathCenterMidPointsBase(Engine, Mesh)
{
}

CSubDivide::PathCenterMidPoints::PathCenterMidPoints(k3d::GeometryEngine& Engine, k3dIPath* const Path) :
	PathCenterMidPointsBase(Engine, Path)
{
}

void CSubDivide::PathCenterMidPoints::BuildGeometry()
{
	// Skip anything with < 3 sides ...
	if(m_Points.size() < 3)
		return;
	
	// Create quadrilaterals between the center point and edge midpoints ...
	for(gulong i = 0; i < m_Points.size(); i++)
		{
			k3d::EnginePointProxy* const point1 = m_Points[(i+1)%m_Points.size()];
			k3d::EnginePointProxy* const point2 = m_Points[(i+2)%m_Points.size()];
			k3d::EnginePointProxy* const point3 = m_Points[(i+3)%m_Points.size()];
			
			Edge& edge1 = static_cast<Edge&>(Engine().Edge(*point1, *point2, false));
			Edge& edge2 = static_cast<Edge&>(Engine().Edge(*point2, *point3, false));

			k3d::EnginePathProxy& path = Engine().Path(Mesh());
			
			path.Actual()->Points()->AddPoint(CenterPoint().Actual());
			path.Actual()->Points()->AddPoint(edge1.MidPoint().Actual());
			path.Actual()->Points()->AddPoint(point2->Actual());
			path.Actual()->Points()->AddPoint(edge2.MidPoint().Actual());
		}
		
	// Mark the original for deletion ...
//	MarkForDeletion();
}

///////////////////////////////////////////////////////////////
// CSubDivide::PathMidPoints

CSubDivide::PathMidPoints::PathMidPoints(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh, const bool Center, const bool Corners) :
	PathMidPointsBase(Engine, Mesh),
	m_Center(Center),
	m_Corners(Corners)
{
}

CSubDivide::PathMidPoints::PathMidPoints(k3d::GeometryEngine& Engine, k3dIPath* const Path, const bool Center, const bool Corners) :
	PathMidPointsBase(Engine, Path),
	m_Center(Center),
	m_Corners(Corners)
{
}

void CSubDivide::PathMidPoints::BuildGeometry()
{
	// Skip anything with < 3 sides ...
	if(m_Points.size() < 3)
		return;

	// Create an n-gon using edge midpoints ...
	if(m_Center)
		{
			k3d::EnginePathProxy& path = Engine().Path(Mesh());
			for(gulong i = 0; i < m_Points.size(); i++)
				{
					k3d::EnginePointProxy* const point1 = m_Points[(i+0)%m_Points.size()];
					k3d::EnginePointProxy* const point2 = m_Points[(i+1)%m_Points.size()];
			
					Edge& edge = static_cast<Edge&>(Engine().Edge(*point1, *point2, false));
			
					path.Actual()->Points()->AddPoint(edge.MidPoint().Actual());
				}
		}

	// Create triangles between edge midpoints and original points ...
	if(m_Corners)
		{
			for(gulong i = 0; i < m_Points.size(); i++)
				{
					k3d::EnginePointProxy* const point1 = m_Points[(i+0)%m_Points.size()];
					k3d::EnginePointProxy* const point2 = m_Points[(i+1)%m_Points.size()];
					k3d::EnginePointProxy* const point3 = m_Points[(i+2)%m_Points.size()];
			
					Edge& edge1 = static_cast<Edge&>(Engine().Edge(*point1, *point2, false));
					Edge& edge2 = static_cast<Edge&>(Engine().Edge(*point2, *point3, false));
			
					k3d::EnginePathProxy& path = Engine().Path(Mesh());
					path.Actual()->Points()->AddPoint(edge1.MidPoint().Actual());
					path.Actual()->Points()->AddPoint(point2->Actual());
					path.Actual()->Points()->AddPoint(edge2.MidPoint().Actual());
				}
		}

/*
	// Mark the original for deletion ...
	MarkForDeletion();
*/
}

///////////////////////////////////////////////////////////////
// CSubDivide::Edge

CSubDivide::Edge::Edge(k3d::GeometryEngine& Engine, k3d::EnginePointProxy& Point1, k3d::EnginePointProxy& Point2, const bool Original) :
	k3d::EngineEdgeProxy(Engine, Point1, Point2, Original),
	m_MidPoint(0)
{
}

void CSubDivide::Edge::Cancel()
{
	m_MidPoint = 0;
}

k3d::EnginePointProxy& CSubDivide::Edge::MidPoint()
{
	if(0 == m_MidPoint)
		{
			// Create a point ...
			m_MidPoint = &Engine().Point(Point1().Mesh());
			m_MidPoint->Actual()->Location()->SetLocalXYZ((Point1().Actual()->Location()->LocalXYZ() +Point2().Actual()->Location()->LocalXYZ()) * 0.5);
		}

	return *m_MidPoint;
}

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CSubDivide

CSubDivide::CSubDivide(sdpIUnknown* Initializer) :
	CSubDivideBase(Initializer),
	m_SubDivideType("Type", CENTERMIDPOINTS, 0, 0),
	m_Engine(
		Document(),
		new k3d::EngineMeshProxyFactory<k3d::EngineMeshProxy>(),
		new PathProxyFactory(*this),
		new k3d::EnginePointProxyFactory<k3d::EnginePointProxy>(),
		new k3d::EngineEdgeProxyFactory<CSubDivide::Edge>()
		)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CSubDivide::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedPaths(Document()->Objects(), std::back_inserter(m_Selection));

	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return CSubDivideBase::InitComplete();
}

sdpResult CSubDivide::EditObject()
{
	new Properties(this);
	return SDP_S_OK;
}

void CSubDivide::BuildGeometry()
{
	// Get rid of old geometry ...
	m_Engine.Cancel();
	m_Engine.Clear();
	std::copy(m_Selection.begin(), m_Selection.end(), k3d::GeometryEngine::Inserter(m_Engine));
	
	// Create new geometry ...
	m_Engine.StartChangeSet();
	m_Engine.BuildGeometry();
	m_Engine.Synchronize();
	m_Engine.FinishChangeSet();
	
	// Update it ...
	m_Engine.UpdateGeometry();
	m_Engine.Manage();

	// Notify the UI ...
	Notify(k3dObject::Event_UserMessage1, m_Engine.Description());
}

void CSubDivide::UpdateGeometry()
{
	m_Engine.UpdateGeometry();
}

void CSubDivide::CancelGeometry()
{
	m_Engine.Cancel();
}

void CSubDivide::FinishGeometry()
{
	m_Engine.Finish("Subdivide Selected");
}

////////////////////////////////////////////////////////////////////////////
// CSubDivide::Properties

CSubDivide::Properties::Properties(CSubDivide* const Object) :
	PropertiesBase(Object, false, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)

{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "subdivide.gtkml"));
	
	// Attach choosers ...
	if(Getk3dChooser("type"))
		Getk3dChooser("type")->Attach(new k3dChooserVariableDataSource<CSubDivide::SubDivideType>(m_Object->m_SubDivideType), k3dApplication()->CommandRecorder(), 0, 0, 0, new k3dActionMethodCall<CSubDivide>(m_Object, &CSubDivide::BuildGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), 0);
	
	Show();
}

void CSubDivide::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnOK();
}

void CSubDivide::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);
	
	PropertiesBase::OnCancel();
}


