#ifndef SUBDIVIDE_H
#define SUBDIVIDE_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares CSubDivide, a K-3D tool for subdividing paths
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dbuildgeometry.h>
#include <k3dobjectdialog.h>
#include <k3dtool.h>

/////////////////////////////////////////////////////////////////////////////
// CSubDivide

typedef k3dTool CSubDivideBase;

class CSubDivide :
	public CSubDivideBase
{
public:
	CSubDivide(sdpIUnknown* Initializer);

	//k3dIObject implementation
	sdpResult InitComplete();
	sdpResult EditObject();

	void BuildGeometry();
	void UpdateGeometry();
	void CancelGeometry();
	void FinishGeometry();
			
private:
	typedef k3dObjectDialog PropertiesBase;
	class Properties :
		public PropertiesBase
	{
	public:
		Properties(CSubDivide* const Object);

	private:
		CSubDivide* const m_Object;

		void OnOK();
		void OnCancel();
	};
	friend class Properties;

	///////////////////////////////////////////////////////////////////////////////
	// Path

	typedef k3d::EnginePathProxy PathBase;
	class Path :
		public PathBase
	{
	public:
		Path(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh);
		Path(k3d::GeometryEngine& Engine, k3dIPath* const Path);

		void Cancel();

		k3d::EnginePointProxy& CenterPoint();
	
	private:
		k3d::EnginePointProxy* m_CenterPoint;
	};

	///////////////////////////////////////////////////////////////////////////////
	// PathCenter

	typedef Path PathCenterBase;
	class PathCenter :
		public PathCenterBase
	{
	public:
		PathCenter(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh);
		PathCenter(k3d::GeometryEngine& Engine, k3dIPath* const Path);

		void BuildGeometry();
	};

	///////////////////////////////////////////////////////////////////////////////
	// PathCenterMidPoints

	typedef Path PathCenterMidPointsBase;
	class PathCenterMidPoints :
		public PathCenterMidPointsBase
	{
	public:
		PathCenterMidPoints(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh);
		PathCenterMidPoints(k3d::GeometryEngine& Engine, k3dIPath* const Path);

		void BuildGeometry();
	};

	///////////////////////////////////////////////////////////////////////////////
	// PathMidPoints

	typedef Path PathMidPointsBase;
	class PathMidPoints :
		public PathMidPointsBase
	{
	public:
		PathMidPoints(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh, const bool Center, const bool Corners);
		PathMidPoints(k3d::GeometryEngine& Engine, k3dIPath* const Path, const bool Center, const bool Corners);

		void BuildGeometry();
	
	private:
		const bool m_Center;
		const bool m_Corners;
	};

	//////////////////////////////////////////////////////////////////////////////
	// Edge

	class Edge :
		public k3d::EngineEdgeProxy
	{
	public:
		Edge(k3d::GeometryEngine& Engine, k3d::EnginePointProxy& Point1, k3d::EnginePointProxy& Point2, const bool Original);

		void Cancel();

		k3d::EnginePointProxy& MidPoint();
	
	private:
		k3d::EnginePointProxy* m_MidPoint;
	};
	
	////////////////////////////////////////////////////////////////////////////////////
	// PathProxyFactory
	
	class PathProxyFactory :
		public k3d::IEnginePathProxyFactory
	{
	public:
		PathProxyFactory(CSubDivide& Object) :
			m_Object(Object)
		{
		}
	
		k3d::EnginePathProxy* CreatePathProxy(k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh)
		{
			switch(m_Object.m_SubDivideType.Value())
				{
					case CSubDivide::CENTER:
						return new CSubDivide::PathCenter(Engine, Mesh);
					case CSubDivide::CENTERMIDPOINTS:
						return new CSubDivide::PathCenterMidPoints(Engine, Mesh);
					case CSubDivide::MIDPOINTS:
						return new CSubDivide::PathMidPoints(Engine, Mesh, true, true);
					case CSubDivide::MIDPOINTSCENTERONLY:
						return new CSubDivide::PathMidPoints(Engine, Mesh, true, false);
					case CSubDivide::MIDPOINTSCORNERSONLY:
						return new CSubDivide::PathMidPoints(Engine, Mesh, false, true);
				}

			g_return_val_if_fail(0, 0);				
		}
	
		k3d::EnginePathProxy* CreatePathProxy(k3d::GeometryEngine& Engine, k3dIPath* const Path)
		{
			switch(m_Object.m_SubDivideType.Value())
				{
					case CSubDivide::CENTER:
						return new CSubDivide::PathCenter(Engine, Path);
					case CSubDivide::CENTERMIDPOINTS:
						return new CSubDivide::PathCenterMidPoints(Engine, Path);
					case CSubDivide::MIDPOINTS:
						return new CSubDivide::PathMidPoints(Engine, Path, true, true);
					case CSubDivide::MIDPOINTSCENTERONLY:
						return new CSubDivide::PathMidPoints(Engine, Path, true, false);
					case CSubDivide::MIDPOINTSCORNERSONLY:
						return new CSubDivide::PathMidPoints(Engine, Path, false, true);
				}

			g_return_val_if_fail(0, 0);				
		}

	private:
		CSubDivide& m_Object;
	};
	friend class PathProxyFactory;
	
	/// Enumerates available subdivision algorithms
	typedef enum
	{
		CENTER,
		CENTERMIDPOINTS,
		MIDPOINTS,
		MIDPOINTSCENTERONLY,
		MIDPOINTSCORNERSONLY,
	} SubDivideType;

	friend std::ostream& operator<<(std::ostream& Stream, const SubDivideType& Value)
	{
		switch(Value)
			{
				case CENTER:
					Stream << "center";
					break;
				case CENTERMIDPOINTS:
					Stream << "centermidpoints";
					break;
				case MIDPOINTS:
					Stream << "midpoints";
					break;
				case MIDPOINTSCENTERONLY:
					Stream << "midpointscenteronly";
					break;
				case MIDPOINTSCORNERSONLY:
					Stream << "midpointscornersonly";
					break;
			}
	
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, SubDivideType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "center")
			Value = CENTER;
		else if(text == "centermidpoints")
			Value = CENTERMIDPOINTS;
		else if(text == "midpoints")
			Value = MIDPOINTS;
		else if(text == "midpointscenteronly")
			Value = MIDPOINTSCENTERONLY;
		else if(text == "midpointscornersonly")
			Value = MIDPOINTSCORNERSONLY;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;

		return Stream;
	}

	/// Stores the selected subdivision type
	k3dVariable<SubDivideType> m_SubDivideType;
	
	k3d::SelectedPathArray m_Selection;
	k3d::GeometryEngine m_Engine;
};

#endif // SUBDIVIDE_H

