// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTaper K-3D tool, which applies a "taper" transformation to geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "taper.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <sdpgtkevent.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3distaterecorder.h>

namespace
{

class TaperLocations
{
public:
	TaperLocations(const sdpVector3 Size, const k3dICamera::Axis Axis, const bool DisplaceX, const bool DisplaceY, const bool DisplaceZ, const gdouble TaperFactor) :
		m_Size(Size),
		m_Axis(Axis),
		m_DisplaceX(DisplaceX),
		m_DisplaceY(DisplaceY),
		m_DisplaceZ(DisplaceZ),
		m_TaperFactor(TaperFactor)
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		sdpVector3 coords = Location.OldXYZ();

		gdouble delta = 0.0;
		switch(m_Axis)
			{
				case k3dICamera::XAXIS:
					if(m_Size[0])
						delta = fabs(coords[0] / m_Size[0]);
					break;
				case k3dICamera::YAXIS:
					if(m_Size[1])
						delta = fabs(coords[1] / m_Size[1]);
					break;
				case k3dICamera::ZAXIS:
					if(m_Size[2])
						delta = fabs(coords[2] / m_Size[2]);
					break;
				default:
					g_return_if_fail(0);
			}

		const gdouble scale = 1.0 + (delta * (m_TaperFactor - 1.0));

		if(m_DisplaceX && (k3dICamera::XAXIS != m_Axis))
			coords[0] *= scale;

		if(m_DisplaceY && (k3dICamera::YAXIS != m_Axis))
			coords[1] *= scale;

		if(m_DisplaceZ && (k3dICamera::ZAXIS != m_Axis))
			coords[2] *= scale;

		Location->SetLocalXYZ(coords);
	}

private:
	const sdpVector3 m_Size;
	const k3dICamera::Axis m_Axis;
	const bool m_DisplaceX;
	const bool m_DisplaceY;
	const bool m_DisplaceZ;
	const gdouble m_TaperFactor;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CTaper

CTaper::CTaper(sdpIUnknown* Initializer) :
	CTaperBase(Initializer),
	k3dMouseEventObserver("LMB-drag to taper selected Points"),
	m_CameraActiveAxis(k3dICamera::ZAXIS),
	m_TaperFactor("taperfactor", 1.0, 0, 0),
	m_DisplaceX("displacex", true, 0, 0),
	m_DisplaceY("displacey", true, 0, 0),
	m_DisplaceZ("displacez", true, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTaper::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	std::copy(m_CurrentSelection.begin(), m_CurrentSelection.end(), k3d::Dimensions::Inserter(m_CurrentSelectionDimensions));

	return CTaperBase::InitComplete();
}

sdpResult CTaper::EditObject()
{
	new CTaperDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CTaper::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_CameraActiveAxis = Camera->ActiveAxis();

	PushEditing();
	return SDP_S_OK;
}

sdpResult CTaper::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = ((double)(CurrentX - LastX)) * 0.01;
	double yfactor = ((double)(CurrentY - LastY)) * 0.01;

	k3dSetValue(m_TaperFactor, m_TaperFactor.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CTaper::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CTaper::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CTaper::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), TaperLocations(m_CurrentSelectionDimensions.Extents(), m_CameraActiveAxis, m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), m_TaperFactor.Value()));
}

void CTaper::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CTaper::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Taper " + sdpToString(m_TaperFactor.Value()));
}

//////////////////////////////////////////////////////////////////////////////////////
// CTaperDialog

const sdpString CTaperDialog::dialogname = "properties";
const sdpString CTaperDialog::control_displacex = "displacex";
const sdpString CTaperDialog::control_displacey = "displacey";
const sdpString CTaperDialog::control_displacez = "displacez";
const sdpString CTaperDialog::control_taperfactor = "taperfactor";

CTaperDialog::CTaperDialog(CTaper* const Object) :
	CTaperDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "taper.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PushEditing, 0), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PopEditing, 0));
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PushEditing, 0), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PopEditing, 0));
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PushEditing, 0), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PopEditing, 0));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_taperfactor, m_Object->m_TaperFactor, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PushEditing, 0), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTaper>(m_Object, &CTaper::PopEditing, 0));

	Show();
}


