// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTranslate K-3D tool, which alters the position of objects or their geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "translate.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3distaterecorder.h>
#include <k3ditransform.h>
#include <k3disnaptarget.h>

#include <sdpgtkevent.h>

namespace
{

class TranslateLocations
{
public:
	TranslateLocations(const sdpVector3 Translate) :
		m_Matrix(sdpTranslation3D(Translate))
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		Location->SetLocalXYZ(m_Matrix * Location.OldXYZ());
	}

private:
	const sdpMatrix4 m_Matrix;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CTranslate

CTranslate::CTranslate(sdpIUnknown* Initializer) :
	CTranslateBase(Initializer),
	k3dMouseEventObserver("LMB-drag to translate selected Points; control-LMB-drag to constrain to Active Axis"),
	m_RawX("x", 0.0, 0, 0),
	m_RawY("y", 0.0, 0, 0),
	m_RawZ("z", 0.0, 0, 0),
	m_Sensitivity("sensitivity", 0.1, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CTranslate::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	return CTranslateBase::InitComplete();
}

sdpResult CTranslate::EditObject()
{
	new Properties(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CTranslate::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CTranslate::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * m_Sensitivity.Value();
	double yfactor = (CurrentY - LastY) * m_Sensitivity.Value();

	// If the user holds down either control key, only translate along the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_RawX, m_RawX.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_RawY, m_RawY.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_RawZ, m_RawZ.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, translate along the two non-active axes ...
	else
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_RawZ, m_RawZ.Value() - xfactor);
						k3dSetValue(m_RawY, m_RawY.Value() - yfactor);
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_RawX, m_RawX.Value() - xfactor);
						k3dSetValue(m_RawZ, m_RawZ.Value() - yfactor);
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_RawX, m_RawX.Value() + xfactor);
						k3dSetValue(m_RawY, m_RawY.Value() - yfactor);
						break;
					default:
						g_assert(0);
				}
		}

	// If the user holds down either alt key, try to "snap" to an object
	m_Translate = sdpVector3(m_RawX.Value(), m_RawY.Value(), m_RawZ.Value());
	if(Mod1Down(Modifiers) || k3dICamera::SNAP == Camera->ActiveSnapMode())
		{
			sdpString description;
			k3dSnap(Document(), m_Translate, m_Translate, description);
		}

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CTranslate::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CTranslate::UpdateControl()
{
	m_Translate = sdpVector3(m_RawX.Value(), m_RawY.Value(), m_RawZ.Value());
	UpdateGeometry();
}


void CTranslate::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CTranslate::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), TranslateLocations(m_Translate));
}

void CTranslate::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CTranslate::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Translate " + sdpToString(sdpVector3(m_RawX.Value(), m_RawY.Value(), m_RawZ.Value())));
}

//////////////////////////////////////////////////////////////////////////////////////
// CTranslate::Properties

const sdpString CTranslate::Properties::dialogname = "properties";
const sdpString CTranslate::Properties::control_x = "x";
const sdpString CTranslate::Properties::control_y = "y";
const sdpString CTranslate::Properties::control_z = "z";

CTranslate::Properties::Properties(CTranslate* const Object) :
	PropertiesBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "translate.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_x, m_Object->m_RawX, k3dApplication()->CommandRecorder(), 0, "", new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::PushEditing, 0), new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::UpdateControl, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::PopEditing, 0));
	Connectk3dSpinButton(control_y, m_Object->m_RawY, k3dApplication()->CommandRecorder(), 0, "", new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::PushEditing, 0), new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::UpdateControl, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::PopEditing, 0));
	Connectk3dSpinButton(control_z, m_Object->m_RawZ, k3dApplication()->CommandRecorder(), 0, "", new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::PushEditing, 0), new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::UpdateControl, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTranslate>(m_Object, &CTranslate::PopEditing, 0));
	Connectk3dSpinButton("sensitivity", m_Object->m_Sensitivity, k3dApplication()->CommandRecorder(), 0, 0, 0, 0, 0);

	Show();
}


