// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTriangulate tool, which triangulates geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "triangulate.h"

#include <k3dmisc.h>
#include <k3dpathpointcollection.h>

#include <k3diapplication.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

#include <numeric>


/// A negative-aware modulo
gint sdpModulo(gint Argument, gint Modulo)
{
	if(0 == Modulo)
		return 0;

	while(Argument < 0)
		Argument += Modulo;

	while(Argument >= Modulo)
		Argument -= Modulo;

	return Argument;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dTriangulatePoint2

class k3dTriangulatePoint2
{
public:
	k3dTriangulatePoint2(const sdpVector2 Coords, void* UserData) : m_Coords(Coords), m_UserData(UserData) {}

	const sdpVector2& Coords() const { return m_Coords; }
	void* UserData() const { return m_UserData; }

	gdouble operator[](int Index) const { return m_Coords[Index]; }

protected:
	sdpVector2 m_Coords;
	void* m_UserData;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// k3dTriangulatePoint2Array

typedef std::vector<k3dTriangulatePoint2> k3dTriangulatePoint2ArrayBase;

class k3dTriangulatePoint2Array : public k3dTriangulatePoint2ArrayBase
{
public:
	const k3dTriangulatePoint2& operator[](const gint Index) const { return k3dTriangulatePoint2ArrayBase::operator[](sdpModulo(Index, size())); }
};

// Calculate the area of a path ...
gdouble Area(const k3dTriangulatePoint2Array& Points)
{
	// Simplest case ... zero area
	if(Points.size() < 3)
		return 0.0;

	// Calculate the result by adding up determinants ...
	gdouble result = 0;

	gint pointsize = Points.size();
	for(gint i = 0; i < pointsize; i++)
		result += (Points[i][0] * Points[i+1][1]) - (Points[i+1][0] * Points[i][1]);

	return result * 0.5;
}

// Figure out whether an ear diagonal intersects any of the remaining path line segments ...
bool EarDiagonalIntersects(const k3dTriangulatePoint2Array& Points, const gint Index)
{
	gint pointsize = Points.size();
	for(gint i = 0; i < pointsize; i++)
		{
			// Don't test the ear- or adjacent segments ...
			gulong a = i;
			gulong b = sdpModulo(i+1, pointsize);
			gulong c = sdpModulo(Index-1, pointsize);
			gulong d = sdpModulo(Index+1, pointsize);

			if(a == c || a == d || b == c || b == d)
				continue;

			// See if the segment intersects the ear diagonal (and if so, how) ...
			sdpLineSegmentIntersectionType intersection = sdpLineSegmentIntersection(Points[i].Coords(), Points[i+1].Coords(), Points[Index-1].Coords(), Points[Index+1].Coords());

			switch(intersection)
				{
					case SDP_INTERIORINTERSECTION:
					case SDP_INTERIORCOLINEARPOINTS:
					case SDP_INTERIOREXTERIORCOLINEARPOINTS:
					case SDP_EXTERIORCOLINEARPOINTS:
						return true;
					case SDP_NOINTERSECTION:
					case SDP_EXTERIORCOLINEARPOINT:
					case SDP_INTERIORCOLINEARPOINT:
					case SDP_EXTERIORINTERSECTION:
					case SDP_EXTERIORCOLINEARPOINTSNOINTERSECTION:
						break;
					case SDP_ERROR:
						g_return_val_if_fail(0, true);
						break;
				}
		}

	return false;
}

// "Clip" an ear from the path ...
void ClipEar(k3dTriangulatePoint2Array& Points, const gint Index, k3dTriangulatePoint2Array& Results)
{
	Results.push_back(Points[Index-1]);
	Results.push_back(Points[Index]);
	Results.push_back(Points[Index+1]);

	Points.erase(Points.begin() + Index);
}

/// Apply an ear-clipping algorithm to the path contour ...
bool EarClipping(k3dTriangulatePoint2Array& Points, k3dTriangulatePoint2Array& Results)
{
	// Calculate the path's signed area ...
	gdouble area = Area(Points);
	// We use the sign of the area to determine what's "convex" ...
	const gdouble pathconvex = sdpSign(Area(Points));
	// We store the path's unsigned area to test against later ...
	gdouble patharea = fabs(area);

	while(Points.size() > 3)
		{
			gint i;
			gint pointsize = Points.size();
			for(i = 0; i < pointsize; i++)
				{
					// Calculate the ear's signed area ...
					area = sdpArea(Points[i-1].Coords(), Points[i].Coords(), Points[i+1].Coords());
					// We use the sign of the ear's area to test the ear for convexity ...
					const gdouble earconvex = sdpSign(area);
					// We use the ear's unsigned area to test against the path's area ...
					const gdouble eararea = fabs(area);

					// If the "ear" isn't convex, it can't be used ...
					if(earconvex != pathconvex)
						continue;

					// If the "ear" area is larger than the total path area, it obviously can't be used ...
					if(eararea > patharea)
						continue;

					// If the "ear" diagonal intersects any other line segments in the path, it can't be used ...
					if(EarDiagonalIntersects(Points, i))
						continue;

					// Clip the ear!
					ClipEar(Points, i, Results);

					// Keep track of decreasing path area ...
					patharea -= eararea;

					break;
				}

			// If the loop actually ended and we get this far, we couldn't find an ear to clip!
			g_return_val_if_fail(i < pointsize, false);
		}

	ClipEar(Points, 1, Results);

	return true;
}





CTriangulate::PathProxy::PathProxy(CTriangulate& Object, k3d::GeometryEngine& Engine, k3d::EngineMeshProxy& Mesh) :
	PathProxyBase(Engine, Mesh),
	m_Object(Object)
{
}

CTriangulate::PathProxy::PathProxy(CTriangulate& Object, k3d::GeometryEngine& Engine, k3dIPath* const Path) :
	PathProxyBase(Engine, Path),
	m_Object(Object)
{
}

void CTriangulate::PathProxy::BuildGeometry()
{
	// Skip anything with < 3 sides ...
	if(m_Points.size() >= 3)
		{
			// Compute path normal ...
			sdpVector3 normal = sdpVector3(0, 0, 0);

			sdpVector3 normal1;
			sdpVector3 normal2;

			gulong pointcount = m_Points.size();
			for(gulong i = 2; i < pointcount; i++)
				{
					normal1 = m_Points[i]->Actual()->Location()->LocalXYZ() - m_Points[i-1]->Actual()->Location()->LocalXYZ();
					normal2 = m_Points[i-2]->Actual()->Location()->LocalXYZ() - m_Points[i-1]->Actual()->Location()->LocalXYZ();

					normal += normal1 ^ normal2;
				}

			normal1 = m_Points[0]->Actual()->Location()->LocalXYZ() - m_Points[pointcount-1]->Actual()->Location()->LocalXYZ();
			normal2 = m_Points[pointcount-2]->Actual()->Location()->LocalXYZ() - m_Points[pointcount-1]->Actual()->Location()->LocalXYZ();

			normal += normal1 ^ normal2;
			normal.Abs();

			// See which axis to drop ...
			int dropaxis = 2;

			if(normal[0] >= normal[1] && normal[0] >= normal[2])
				dropaxis = 0;
			else if(normal[1] >= normal[0] && normal[1] >= normal[2])
				dropaxis = 1;

			sdpVector3 coords;
			k3dTriangulatePoint2Array points;
			for(gulong i = 0; i < m_Points.size(); i++)
				{
					coords = m_Points[i]->Actual()->Location()->LocalXYZ();
					points.push_back(k3dTriangulatePoint2(sdpVector2(coords, dropaxis), m_Points[i]));
				}

			/// Calculate a triangulation of 2D points ...
			k3dTriangulatePoint2Array results;
			EarClipping(points, results);

			k3dPathPointArray Results;
			for(gulong j = 0; j < results.size(); j++)
				Results.push_back(reinterpret_cast<k3dIPathPoint*>(results[j].UserData()));

			// For each result vertex ...
			for(gulong j = 0; j < Results.size(); j++)
				{
					// Allocate a new path every third vertex ...
					if(0 == j % 3)
						m_NewPaths.push_back(&Engine().Path(Mesh()));

					k3dIPathPoint* pathpoint = Results[j];
					g_assert(pathpoint);

					// Add the vertex to the current triangle ...
					m_NewPaths.back()->Actual()->Points()->AddPoint(pathpoint->Point());
				}

/*
			// Mark the original for deletion ...
			if(m_Object.m_DeleteOriginals.Value())
				MarkForDeletion();
*/
		}

	PathProxyBase::BuildGeometry();
}

void CTriangulate::PathProxy::UpdateGeometry()
{
cerr << "CTriangulate::UpdateGeometry()?!?" << endl;
	PathProxyBase::UpdateGeometry();
}

void CTriangulate::PathProxy::Cancel()
{
	m_NewPaths.clear();

	PathProxyBase::Cancel();
}

/////////////////////////////////////////////////////////////////////////////
// CTriangulate

CTriangulate::CTriangulate(sdpIUnknown* Initializer) :
	CTriangulateBase(Initializer),
	m_Engine(
		Document(),
		new k3d::EngineMeshProxyFactory<k3d::EngineMeshProxy>(),
		new k3d::EnginePathProxyFactory2<CTriangulate, CTriangulate::PathProxy>(*this),
		new k3d::EnginePointProxyFactory<k3d::EnginePointProxy>(),
		new k3d::EngineEdgeProxyFactory<k3d::EngineEdgeProxy>()
		)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CTriangulate::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedPaths(Document()->Objects(), std::back_inserter(m_Selection));
	std::copy(m_Selection.begin(), m_Selection.end(), k3d::GeometryEngine::Inserter(m_Engine));

	return CTriangulateBase::InitComplete();
}

sdpResult CTriangulate::EditObject()
{
	new Properties(this);

	BuildGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}


void CTriangulate::BuildGeometry()
{
	// Get rid of any old geometry
	m_Engine.Cancel();

	// Create new geometry ...
	m_Engine.StartChangeSet();
	m_Engine.BuildGeometry();
	m_Engine.Synchronize();
	m_Engine.FinishChangeSet();

	// Update it ...
	m_Engine.UpdateGeometry();
	m_Engine.Manage();

	// Notify UI ...
	Notify(k3dObject::Event_UserMessage1, m_Engine.Description());
}

void CTriangulate::CancelGeometry()
{
	m_Engine.Cancel();
}

void CTriangulate::FinishGeometry()
{
	m_Engine.Finish("Object Triangulated");
}

//////////////////////////////////////////////////////////////////////////////////////
// CTriangulate::Properties

CTriangulate::Properties::Properties(CTriangulate* const Object) :
	PropertiesBase(Object, true, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "triangulate.gtkml"));

	Show();
}

void CTriangulate::Properties::OnOK()
{
	m_Object->FinishGeometry();
	k3dRedrawAll(m_Object->Document(), false);

	PropertiesBase::OnOK();
}

void CTriangulate::Properties::OnCancel()
{
	m_Object->CancelGeometry();
	k3dRedrawAll(m_Object->Document(), false);

	PropertiesBase::OnCancel();
}

