// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTwist tool, which applies a "twist" transformation to selected geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "twist.h"

#include <k3dmisc.h>
#include <k3doptions.h>

#include <sdpgtkevent.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3distaterecorder.h>

namespace
{

class TwistLocations
{
public:
	TwistLocations(const sdpVector3 Size, const sdpVector3 Twist) :
		m_Size(Size),
		m_Twist(Twist)
	{
	}
	
	void operator()(k3d::SelectedLocationProxy& Location)
	{
		gdouble cosx, cosy, cosz;
		gdouble sinx, siny, sinz;
		gdouble delta, px, py;

		sdpVector3 coords = Location.OldXYZ();

		// Twist around X-Axis
		if(m_Twist[0] && m_Size[0])
			{
				delta = coords[0] / m_Size[0];
	//			delta = sdpSign<double>(delta) * m_Xfalloff->GetValue1(fabs(delta), .0001);

				cosx = cos(m_Twist[0] * delta);
				sinx = sin(m_Twist[0] * delta);

				py = coords[1];
				coords[1] = (cosx * coords[1]) - (sinx * coords[2]);
				coords[2] = (sinx * py) + (cosx * coords[2]);
			}

		// Twist around Y-Axis
		if(m_Twist[1] && m_Size[1])
			{
				delta = coords[1] / m_Size[1];
	//			delta = sdpSign<double>(delta) * m_Yfalloff->GetValue1(fabs(delta), .0001);

				cosy = cos(m_Twist[1] * delta);
				siny = sin(m_Twist[1] * delta);

				px = coords[0];
				coords[0] = (cosy * coords[0]) + (siny * coords[2]);
				coords[2] = (cosy * coords[2]) - (siny * px);
			}

		// Twist around Z-Axis
		if(m_Twist[2] && m_Size[2])
			{
				delta = coords[2] / m_Size[2];
	//			delta = sdpSign<double>(delta) * m_Zfalloff->GetValue1(fabs(delta), .0001);

				cosz = cos(m_Twist[2] * delta);
				sinz = sin(m_Twist[2] * delta);

				px = coords[0];
				coords[0] = (cosz * coords[0]) - (sinz * coords[1]);
				coords[1] = (sinz * px) + (cosz * coords[1]);
			}

		Location->SetLocalXYZ(coords);
	}

private:
	const sdpVector3 m_Size;
	const sdpVector3 m_Twist;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CTwist

CTwist::CTwist(sdpIUnknown* Initializer) :
	CTwistBase(Initializer),
	k3dMouseEventObserver("LMB-drag to twist selected Points; control-LMB-drag to constrain to Active Axis"),
	m_X("x", 0.0, 0, 0),
	m_Y("y", 0.0, 0, 0),
	m_Z("z", 0.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CTwist::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	std::copy(m_CurrentSelection.begin(), m_CurrentSelection.end(), k3d::Dimensions::Inserter(m_CurrentSelectionDimensions));
	
	return CTwistBase::InitComplete();
}

sdpResult CTwist::EditObject()
{
	new CTwistDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CTwist::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CTwist::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = sdpRadians(gdouble(CurrentX - LastX));
	double yfactor = sdpRadians(gdouble(CurrentY - LastY));

	// If the user holds down either control key, only translate along the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_X, m_X.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_Y, m_Y.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_Z, m_Z.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, translate along the two non-active axes ...
	else
		{
			switch(Camera->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_Z, m_Z.Value() - xfactor);
						k3dSetValue(m_Y, m_Y.Value() - yfactor);
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_X, m_X.Value() - xfactor);
						k3dSetValue(m_Z, m_Z.Value() - yfactor);
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_X, m_X.Value() + xfactor);
						k3dSetValue(m_Y, m_Y.Value() - yfactor);
						break;
					default:
						g_assert(0);
				}
		}

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CTwist::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CTwist::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CTwist::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), TwistLocations(m_CurrentSelectionDimensions.Extents(), sdpVector3(m_X.Value(), m_Y.Value(), m_Z.Value())));
}

void CTwist::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CTwist::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Twist");
}

//////////////////////////////////////////////////////////////////////////////////////
// CTwistDialog

const sdpString CTwistDialog::dialogname = "properties";
const sdpString CTwistDialog::control_x = "x";
const sdpString CTwistDialog::control_y = "y";
const sdpString CTwistDialog::control_z = "z";

CTwistDialog::CTwistDialog(CTwist* const Object) :
	CTwistDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "twist.gtkml"));

	// Connect spin buttons
	Connectk3dSpinButton(control_x, m_Object->m_X, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTwist>(m_Object, &CTwist::PushEditing, 0), new k3dActionMethodCall<CTwist>(m_Object, &CTwist::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTwist>(m_Object, &CTwist::PopEditing, 0));
	Connectk3dSpinButton(control_y, m_Object->m_Y, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTwist>(m_Object, &CTwist::PushEditing, 0), new k3dActionMethodCall<CTwist>(m_Object, &CTwist::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTwist>(m_Object, &CTwist::PopEditing, 0));
	Connectk3dSpinButton(control_z, m_Object->m_Z, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CTwist>(m_Object, &CTwist::PushEditing, 0), new k3dActionMethodCall<CTwist>(m_Object, &CTwist::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CTwist>(m_Object, &CTwist::PopEditing, 0));

	Show();
}


