// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//5H
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CUVMap K-3D tool, which is used to modify the texture-mapping coordinates of geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "uvmap.h"

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3dilocation.h>
#include <k3dimesh.h>
#include <k3dipath.h>
#include <k3dipathcollection.h>
#include <k3dipathpoint.h>
#include <k3dipathpointcollection.h>
#include <k3dipoint.h>
#include <k3dipointcollection.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

/////////////////////////////////////////////////////////////////////////////
// CUVMap

CUVMap::CUVMap(sdpIUnknown* Initializer) :
	CUVMapBase(Initializer),
	k3dMouseEventObserver("LMB-drag to pan texture; shift-LMB-drag to zoom"),
	m_MappingType("mappingtype", PROJECTION, Document()->StateRecorder(), 0),
	m_S0("s0", 0.0, 0, 0),
	m_S1("s1", 1.0, 0, 0),
	m_T0("t0", 0.0, 0, 0),
	m_T1("t1", 1.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject implementation

sdpResult CUVMap::InitComplete()
{
	// Record what was selected at the time we were created
//	k3d::GetSelectedLocations(Document(), std::back_inserter(m_CurrentSelection));
	return SDP_S_OK;
}

sdpResult CUVMap::EditObject()
{
	new Properties(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CUVMap::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	PushEditing();
	return SDP_S_OK;
}

sdpResult CUVMap::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	// If the user holds down either shift key, make it a "zoom" ...
	if(ShiftDown(Modifiers))
		{
			k3dSetValue(m_S0, m_S0.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
			k3dSetValue(m_S1, m_S1.Value() + (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
			k3dSetValue(m_T0, m_T0.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
			k3dSetValue(m_T1, m_T1.Value() + (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
		}
/*
	// If the user holds down either control key, only scale along the active axis ...
	else if(ControlDown(Modifiers))
		{
			switch(Document()->ActiveAxis())
				{
					case k3dICamera::XAXIS:
						k3dSetValue(m_X, m_X.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::YAXIS:
						k3dSetValue(m_Y, m_Y.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					case k3dICamera::ZAXIS:
						k3dSetValue(m_Z, m_Z.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
						break;
					default:
						g_assert(0);
				}
		}
*/
	// Otherwise, "pan" along the two axes ...
	else
		{
			k3dSetValue(m_S0, m_S0.Value() - xfactor);
			k3dSetValue(m_S1, m_S1.Value() - xfactor);
			k3dSetValue(m_T0, m_T0.Value() - yfactor);
			k3dSetValue(m_T1, m_T1.Value() - yfactor);
		}

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CUVMap::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CUVMap::PushEditing()
{
	k3dStartStateChangeSet(Document());
//	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CUVMap::UpdateGeometry()
{
//	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), ScaleLocations(m_X.Value(), m_Y.Value(), m_Z.Value()));
}

void CUVMap::CancelEditing()
{
//	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CUVMap::PopEditing()
{
//	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "UVMap");
}

class CUVMapUpdate
{
public:
	CUVMapUpdate(CUVMapAlgorithm& Algorithm, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1) :
		m_Algorithm(Algorithm),
		m_S0(S0),
		m_S1(S1),
		m_T0(T0),
		m_T1(T1)
	{
	}

	bool HandleMesh(k3dIMesh* Mesh)
	{
		m_Algorithm.HandleMesh(Mesh, m_S0, m_S1, m_T0, m_T1);
			
		k3dIPathCollection* paths = Mesh->Paths();
		for(K3DPATHHANDLE pathhandle = paths->FirstPath(); pathhandle; pathhandle = paths->NextPath(pathhandle))
			{
				k3dIPath* path = paths->Path(pathhandle);

				// Skip the path if it doesn't fit our criteria ...
				if(SDP_S_OK != path->State()->QueryState(k3dIState::VISIBLESELECTED))
					continue;

				if(!m_Algorithm.HandleMeshPath(path, m_S0, m_S1, m_T0, m_T1))
					break;
			}
		
		k3dIPointCollection* points = Mesh->Points();
		for(K3DPOINTHANDLE pointhandle = points->FirstPoint(); pointhandle; pointhandle = points->NextPoint(pointhandle))
			{
				k3dIPoint* point = points->Point(pointhandle);

				// Skip the point if it doesn't fit our criteria ...
				if(SDP_S_OK != point->State()->QueryState(k3dIState::VISIBLESELECTED))
					continue;

				if(!m_Algorithm.HandleMeshPoint(point, m_S0, m_S1, m_T0, m_T1))
					break;
			}
		
		return true;
	}
	
private:
	CUVMapAlgorithm& m_Algorithm;
	const gdouble m_S0;
	const gdouble m_S1;
	const gdouble m_T0;
	const gdouble m_T1;
};

/*
void CUVMap::Update(bool SaveUndoAction)
{
	// Record the action if requested ...
	if(SaveUndoAction)
		k3dStartStateChangeSet(Document());

	CUVMapAlgorithm* algorithm = &m_Projection;
	switch(m_MappingType.Value())
		{
			case PROJECTION:
				algorithm = &m_Projection;
				break;
			case CYLINDRICAL:
				algorithm = &m_Cylindrical;
				break;
			case SPHERICAL:
				algorithm = &m_Spherical;
				break;
			case CUBIC:
				algorithm = &m_Cubic;
				break;
			case POLYGON:
				algorithm = &m_Path;
				break;
			case IMPLICIT:
				algorithm = &m_Implicit;
				break;
			default:
				g_return_if_fail(0);
		}
		
	CUVMapUpdate functor(*algorithm, m_S0.Value(), m_S1.Value(), m_T0.Value(), m_T1.Value());
	k3dForEachMesh(Document(), k3dIState::VISIBLE, functor);

	// Finish recording the action if requested ...
	if(SaveUndoAction)
		k3dFinishStateChangeSet(Document(), "UVMap");
}
*/

///////////////////////////////////////////////////////////////
// CUVMapProjection

bool CUVMapProjection::HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
g_assert_not_reached();
//	m_Axis = Mesh->Document()->ActiveAxis();

	Mesh->Dimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
	m_Size = m_Max - m_Min;

	return true;
}

bool CUVMapProjection::HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	sdpVector3 coords = Point->Location()->LocalXYZ();
	
	sdpVector3 texturecoords;
	HandleCoordinates(coords, texturecoords, S0, S1, T0, T1);
	
	Point->Location()->SetTextureXYZ(texturecoords);
	
	return true;
}

bool CUVMapProjection::HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	sdpVector3 coords, texturecoords;
	k3dIPathPointCollection* pathpoints = Path->Points();
	for(K3DPATHPOINTHANDLE pathpointhandle = pathpoints->FirstPathPoint(); pathpointhandle; pathpointhandle = pathpoints->NextPathPoint(pathpointhandle))
		{
			k3dIPathPoint* point = pathpoints->PathPoint(pathpointhandle);
			coords = point->Location()->LocalXYZ();
			HandleCoordinates(coords, texturecoords, S0, S1, T0, T1);
			
			point->Location()->SetTextureXYZ(texturecoords);
		}
		
	return true;
}

void CUVMapProjection::HandleCoordinates(const sdpVector3& Coords, sdpVector3& Results, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	switch(m_Axis)
		{
			case k3dICamera::XAXIS:
				Results[0] = sdpLerp(S0, S1, ((m_Max[2] - Coords[2]) / m_Size[2]));
				Results[1] = sdpLerp(T0, T1, ((m_Max[1] - Coords[1]) / m_Size[1]));
				Results[2] = ((m_Max[0] - Coords[0]) / m_Size[0]);
				break;

			case k3dICamera::YAXIS:
				Results[0] = sdpLerp(S0, S1, ((Coords[0] - m_Min[0]) / m_Size[0]));
				Results[1] = sdpLerp(T0, T1, ((m_Max[2] - Coords[2]) / m_Size[2]));
				Results[2] = ((m_Max[1] - Coords[1]) / m_Size[1]);
				break;

			case k3dICamera::ZAXIS:
				Results[0] = sdpLerp(S0, S1, ((Coords[0] - m_Min[0]) / m_Size[0]));
				Results[1] = sdpLerp(T0, T1, ((m_Max[1] - Coords[1]) / m_Size[1]));
				Results[2] = ((m_Max[2] - Coords[2]) / m_Size[2]);
				break;

			default:
				g_return_if_fail(0); // unknown axis!
		}
}

///////////////////////////////////////////////////////////////
// CUVMapCylindrical

bool CUVMapCylindrical::HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
g_assert_not_reached();
//	m_Axis = Mesh->Document()->ActiveAxis();

	Mesh->Dimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
	m_Size = m_Max - m_Min;
//	m_Center = (m_Min + m_Max) * 0.5;
	
	return true;
}

bool CUVMapCylindrical::HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	sdpVector3 coords = Point->Location()->LocalXYZ();
	
	sdpVector3 texturecoords;
	HandleCoordinates(coords, 0, texturecoords, S0, S1, T0, T1);
	
	Point->Location()->SetTextureXYZ(texturecoords);
	return true;
}

bool CUVMapCylindrical::HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	// Record path quadrants
	gulong quadrants = 0;

	k3dIPathPointCollection* points = Path->Points();
	for(K3DPATHPOINTHANDLE pointhandle = points->FirstPathPoint(); pointhandle; pointhandle = points->NextPathPoint(pointhandle))
		{
			sdpVector3 coords = points->PathPoint(pointhandle)->Location()->LocalXYZ();

			switch(m_Axis)
				{
					case k3dICamera::XAXIS:
						quadrants |= ((coords[2] >= 0) ? (coords[1] >= 0 ? QUADRANT_II : QUADRANT_I) : (coords[1] >= 0 ? QUADRANT_III : QUADRANT_IV));
						break;

					case k3dICamera::YAXIS:
						quadrants |= ((coords[0] <= 0) ? (coords[2] <= 0 ? QUADRANT_II : QUADRANT_I) : (coords[2] <= 0 ? QUADRANT_III : QUADRANT_IV));
						break;

					case k3dICamera::ZAXIS:
						quadrants |= ((coords[0] <= 0) ? (coords[1] >= 0 ? QUADRANT_II : QUADRANT_I) : (coords[1] >= 0 ? QUADRANT_III : QUADRANT_IV));
						break;
				}
		}

	// Calculate texture coordinates
	for(K3DPATHPOINTHANDLE pointhandle2 = points->FirstPathPoint(); pointhandle2; pointhandle2 = points->NextPathPoint(pointhandle2))
		{
			sdpVector3 coords = points->PathPoint(pointhandle2)->Location()->LocalXYZ();
			
			sdpVector3 texturecoords;
			HandleCoordinates(coords, quadrants, texturecoords, S0, S1, T0, T1);
			
			points->PathPoint(pointhandle2)->Location()->SetTextureXYZ(texturecoords);
		}
	
	return true;
}

void CUVMapCylindrical::HandleCoordinates(const sdpVector3& Coords, const gulong Quadrants, sdpVector3& Results, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	switch(m_Axis)
		{
			case k3dICamera::XAXIS:
				{
					const gdouble theta = atan2(Coords[2], -Coords[1]) + (sdp::PiTimes2() * (theta < 0.0));
					Results[0] = (theta / sdp::PiTimes2());
					Results[1] = ((m_Max[0] - Coords[0]) / m_Size[0]);
					Results[2] = 0.0;
					if((theta <= sdp::Pi()) && (Quadrants & QUADRANT_I) && (Quadrants & QUADRANT_IV))
						Results[0] += 1.0;
						
					Results[0] = sdpLerp(S0, S1, Results[0]);
					Results[1] = sdpLerp(T0, T1, Results[1]);
				}
				break;

			case k3dICamera::YAXIS:
				{
					const gdouble theta = atan2(-Coords[0], Coords[2]) + (sdp::PiTimes2() * (theta < 0.0));
					Results[0] = (theta / sdp::PiTimes2());
					Results[1] = ((m_Max[1] - Coords[1]) / m_Size[1]);
					Results[2] = 0.0;
					if((theta <= sdp::Pi()) && (Quadrants & QUADRANT_I) && (Quadrants & QUADRANT_IV))
						Results[0] += 1.0;
						
					Results[0] = sdpLerp(S0, S1, Results[0]);
					Results[1] = sdpLerp(T0, T1, Results[1]);
				}
				break;

			case k3dICamera::ZAXIS:
				{
					const gdouble theta = atan2(-Coords[0], -Coords[1]) + (sdp::PiTimes2() * (theta < 0.0));
					Results[0] = (theta / sdp::PiTimes2());
					Results[1] = ((m_Max[2] - Coords[2]) / m_Size[2]);
					Results[2] = 0.0;
					if((theta <= sdp::Pi()) && (Quadrants & QUADRANT_I) && (Quadrants & QUADRANT_IV))
						Results[0] += 1.0;
						
					Results[0] = sdpLerp(S0, S1, Results[0]);
					Results[1] = sdpLerp(T0, T1, Results[1]);
				}
				break;
				
			default:
				g_return_if_fail(0); // Unknown axis!!!
		}
}

///////////////////////////////////////////////////////////////
// CUVMapSpherical

bool CUVMapSpherical::HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
g_assert_not_reached();
//	m_Axis = Mesh->Document()->ActiveAxis();

	Mesh->Dimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
	m_Size = m_Max - m_Min;
//	m_Center = (m_Min + m_Max) * 0.5;
	
	return true;
}

bool CUVMapSpherical::HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	sdpVector3 coords = Point->Location()->LocalXYZ();
	
	sdpVector3 texturecoords;
	HandleCoordinates(coords, 0, texturecoords, S0, S1, T0, T1);
	
	Point->Location()->SetTextureXYZ(texturecoords);
	return true;
}

bool CUVMapSpherical::HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	// Record path quadrants
	gulong quadrants = 0;

	k3dIPathPointCollection* points = Path->Points();
	for(K3DPATHPOINTHANDLE pointhandle = points->FirstPathPoint(); pointhandle; pointhandle = points->NextPathPoint(pointhandle))
		{
			sdpVector3 coords = points->PathPoint(pointhandle)->Location()->LocalXYZ();

			switch(m_Axis)
				{
					case k3dICamera::XAXIS:
						quadrants |= ((coords[2] >= 0) ? (coords[1] >= 0 ? QUADRANT_II : QUADRANT_I) : (coords[1] >= 0 ? QUADRANT_III : QUADRANT_IV));
						break;

					case k3dICamera::YAXIS:
						quadrants |= ((coords[0] <= 0) ? (coords[2] <= 0 ? QUADRANT_II : QUADRANT_I) : (coords[2] <= 0 ? QUADRANT_III : QUADRANT_IV));
						break;

					case k3dICamera::ZAXIS:
						quadrants |= ((coords[0] <= 0) ? (coords[1] >= 0 ? QUADRANT_II : QUADRANT_I) : (coords[1] >= 0 ? QUADRANT_III : QUADRANT_IV));
						break;
				}
		}

	// Calculate texture coordinates
	for(K3DPATHPOINTHANDLE pointhandle2 = points->FirstPathPoint(); pointhandle2; pointhandle2 = points->NextPathPoint(pointhandle2))
		{
			sdpVector3 coords = points->PathPoint(pointhandle2)->Location()->LocalXYZ();
			
			sdpVector3 texturecoords;
			HandleCoordinates(coords, quadrants, texturecoords, S0, S1, T0, T1);
			
			points->PathPoint(pointhandle2)->Location()->SetTextureXYZ(texturecoords);
		}
	
	return true;
}

void CUVMapSpherical::HandleCoordinates(const sdpVector3& Coords, const gulong Quadrants, sdpVector3& Results, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	switch(m_Axis)
		{
			case k3dICamera::XAXIS:
				{
					const gdouble theta = atan2(Coords[2], -Coords[1]) + (sdp::PiTimes2() * (theta < 0.0));
					const gdouble dist = sdpVector2(Coords[2], -Coords[1]).Length();
					const gdouble phi = atan2(dist, Coords[0]);
					Results[0] = theta / sdp::PiTimes2();
					Results[1] = phi / sdp::Pi();
					Results[2] = 0.0;
					if((theta <= sdp::Pi()) && (Quadrants & QUADRANT_I) && (Quadrants & QUADRANT_IV))
						Results[0] += 1.0;
						
					Results[0] = sdpLerp(S0, S1, Results[0]);
					Results[1] = sdpLerp(T0, T1, Results[1]);
				}
				break;

			case k3dICamera::YAXIS:
				{
					const gdouble theta = atan2(-Coords[0], Coords[2]) + (sdp::PiTimes2() * (theta < 0.0));
					const gdouble dist = sdpVector2(Coords[0], Coords[2]).Length();
					const gdouble phi = atan2(dist, Coords[1]);
					Results[0] = theta / sdp::PiTimes2();
					Results[1] = phi / sdp::Pi();
					Results[2] = 0.0;
					if((theta <= sdp::Pi()) && (Quadrants & QUADRANT_I) && (Quadrants & QUADRANT_IV))
						Results[0] += 1.0;
						
					Results[0] = sdpLerp(S0, S1, Results[0]);
					Results[1] = sdpLerp(T0, T1, Results[1]);
				}
				break;

			case k3dICamera::ZAXIS:
				{
					const gdouble theta = atan2(-Coords[0], -Coords[1]) + (sdp::PiTimes2() * (theta < 0.0));
					const gdouble dist = sdpVector2(Coords[0], -Coords[1]).Length();
					const gdouble phi = atan2(dist, Coords[2]);
					Results[0] = theta / sdp::PiTimes2();
					Results[1] = phi / sdp::Pi();
					Results[2] = 0.0;
					if((theta <= sdp::Pi()) && (Quadrants & QUADRANT_I) && (Quadrants & QUADRANT_IV))
						Results[0] += 1.0;
						
					Results[0] = sdpLerp(S0, S1, Results[0]);
					Results[1] = sdpLerp(T0, T1, Results[1]);
				}
				break;
		}
}

///////////////////////////////////////////////////////////////
// CUVMapCubic

bool CUVMapCubic::HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	Mesh->Dimensions(k3dIState::VISIBLESELECTED, m_Min, m_Max);
	m_Size = m_Max - m_Min;

	return true;
}

bool CUVMapCubic::HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	return true;
}

bool CUVMapCubic::HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	if(Path->Points()->PointCount() < 3)
		return true;

	// Calculate which axis this path should be on
	sdpVector3 normal;
	Path->PathNormal(normal);
	normal.Normalize();

	Axis axis;
	if(fabs(normal[1]) >= fabs(normal[0]))
		{
			if(fabs(normal[1]) >= fabs(normal[2]))
				{
					axis = (normal[1] >= 0) ? YAXIS : NYAXIS;
				}
			else
				{
					axis = (normal[2] >= 0) ? ZAXIS : NZAXIS;
				}
		}
	else
		{
			if(fabs(normal[2]) >= fabs(normal[0]))
				{
					axis = (normal[2] >= 0) ? ZAXIS : NZAXIS;
				}
			else
				{
					axis = (normal[0] >= 0) ? XAXIS : NXAXIS;
				}
		}

	// For each point in the path ...
	sdpVector3 coords, texturecoords;
	k3dIPathPointCollection* points = Path->Points();
	for(K3DPATHPOINTHANDLE point = points->FirstPathPoint(); point; point = points->NextPathPoint(point))
		{
			coords = points->PathPoint(point)->Location()->LocalXYZ();
			switch(axis)
				{
					case XAXIS:
						texturecoords[0] = ((m_Max[2] - coords[2]) / m_Size[2]);
						texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
						texturecoords[2] = ((m_Max[0] - coords[0]) / m_Size[0]);
						break;

					case NXAXIS:
						texturecoords[0] = ((coords[2] - m_Min[2]) / m_Size[2]);
						texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
						texturecoords[2] = ((m_Max[0] - coords[0]) / m_Size[0]);
						break;


					case YAXIS:
						texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
						texturecoords[1] = ((coords[2] - m_Min[2]) / m_Size[2]);
						texturecoords[2] = ((m_Max[1] - coords[1]) / m_Size[1]);
						break;

					case NYAXIS:
						texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
						texturecoords[1] = ((m_Max[2] - coords[2]) / m_Size[2]);
						texturecoords[2] = ((m_Max[1] - coords[1]) / m_Size[1]);
						break;

					case ZAXIS:
						texturecoords[0] = ((coords[0] - m_Min[0]) / m_Size[0]);
						texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
						texturecoords[2] = ((m_Max[2] - coords[2]) / m_Size[2]);
						break;

					case NZAXIS:
						texturecoords[0] = ((m_Max[0] - coords[0]) / m_Size[0]);
						texturecoords[1] = ((m_Max[1] - coords[1]) / m_Size[1]);
						texturecoords[2] = ((m_Max[2] - coords[2]) / m_Size[2]);
						break;

					default:
						g_assert(0);
						break;
				}

						
			texturecoords[0] = sdpLerp(S0, S1, texturecoords[0]);
			texturecoords[1] = sdpLerp(T0, T1, texturecoords[1]);

			points->PathPoint(point)->Location()->SetTextureXYZ(texturecoords);
		}

	return true;
}

///////////////////////////////////////////////////////////////
// CUVMapPath

bool CUVMapPath::HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	return true;
}

bool CUVMapPath::HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	return true;
}

bool CUVMapPath::HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	sdpVector3 texturecoords[4] =
	{
		sdpVector3(S1, T0, 0.0),
		sdpVector3(S0, T0, 0.0),
		sdpVector3(S0, T1, 0.0),
		sdpVector3(S1, T1, 0.0),
	};

	k3dIPathPointCollection* points = Path->Points();
	if(points->PointCount() != 4)
		return true;
		
	gulong index = 0;
	for(K3DPATHPOINTHANDLE pathpointhandle = points->FirstPathPoint(); pathpointhandle; pathpointhandle = points->NextPathPoint(pathpointhandle))
		points->PathPoint(pathpointhandle)->Location()->SetTextureXYZ(texturecoords[index++]);
	
	return true;
}

///////////////////////////////////////////////////////////////
// CUVMapImplicit

bool CUVMapImplicit::HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	return true;
}

bool CUVMapImplicit::HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	Point->Location()->SetTextureXYZ(Point->Location()->ImplicitTextureXYZ());
	return true;
}

bool CUVMapImplicit::HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1)
{
	sdpVector3 coords;	
	k3dIPathPointCollection* points = Path->Points();
	for(K3DPATHPOINTHANDLE point = points->FirstPathPoint(); point; point = points->NextPathPoint(point))
		{
			coords = points->PathPoint(point)->Location()->ImplicitTextureXYZ();
			
			coords[0] = sdpLerp(S0, S1, coords[0]);
			coords[1] = sdpLerp(T0, T1, coords[1]);
			
			points->PathPoint(point)->Location()->SetTextureXYZ(coords);
		}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////
// CUVMap::Properties

const sdpString CUVMap::Properties::control_mappingtype = "mappingtype";
const sdpString CUVMap::Properties::control_apply = "apply";

CUVMap::Properties::Properties(CUVMap* const Object) :
	PropertiesBase(Object, true, true, false, "properties", new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "uvmap.gtkml"));

	// Setup choosers ...
	if(Getk3dChooser(control_mappingtype))
		Getk3dChooser(control_mappingtype)->Attach(new k3dChooserVariableDataSource<CUVMap::MappingType>(m_Object->m_MappingType), k3dApplication()->CommandRecorder(), 0, "Texture Mapping Type", 0, 0, 0);

	// Connect spin buttons
	Connectk3dSpinButton("s0", m_Object->m_S0, k3dApplication()->CommandRecorder(), 0, "Texture S0 Coordinate", new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PushEditing, 0), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PopEditing, 0));
	Connectk3dSpinButton("s1", m_Object->m_S1, k3dApplication()->CommandRecorder(), 0, "Texture S1 Coordinate", new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PushEditing, 0), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PopEditing, 0));
	Connectk3dSpinButton("t0", m_Object->m_T0, k3dApplication()->CommandRecorder(), 0, "Texture T0 Coordinate", new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PushEditing, 0), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PopEditing, 0));
	Connectk3dSpinButton("t1", m_Object->m_T1, k3dApplication()->CommandRecorder(), 0, "Texture T1 Coordinate", new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PushEditing, 0), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CUVMap>(m_Object, &CUVMap::PopEditing, 0));

	// Setup buttons ...
	AddCommand(control_apply, control_apply, CONTROL_BUTTON);
	
	Show();
}

void CUVMap::Properties::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_apply)
		OnApply();
	else
		PropertiesBase::OnEvent(Event);
}

void CUVMap::Properties::OnApply()
{
	// Record this action for posterity (tutorials)
	k3dRecordCommand(this, USERINTERFACE, control_apply, 0);
	
	m_Object->PushEditing();
	m_Object->UpdateGeometry();
	m_Object->PopEditing();
	
	k3dRedrawAll(m_Object->Document(), false);
}


