#ifndef UVMAP_H
#define UVMAP_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Defines the CUVMap K-3D tool, which is used to modify the texture-mapping coordinates of geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dmouseeventobserver.h>
#include <k3dobjectdialog.h>
#include <k3dtool.h>

#include <k3dicamera.h>
#include <k3dimesh.h>
#include <k3dipath.h>
#include <k3dipoint.h>

/////////////////////////////////////////////////////////////////////////////
// CUVMapAlgorithm

/// Defines an abstract interface for UVMap algorithms
class CUVMapAlgorithm
{
public:
	virtual bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1) = 0;
	virtual bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1) = 0;
	virtual bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1) = 0;
};

/////////////////////////////////////////////////////////////////////////////

class CUVMapProjection :
	public CUVMapAlgorithm
{
public:
	bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	
private:
	void HandleCoordinates(const sdpVector3& Coords, sdpVector3& Results, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);

	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;

	k3dICamera::Axis m_Axis;
};

class CUVMapCylindrical :
	public CUVMapAlgorithm
{
public:
	bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);

private:
	void HandleCoordinates(const sdpVector3& Coords, const gulong Quadrants, sdpVector3& Results, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);

	typedef enum
	{
		QUADRANT_I = 1 << 0,
		QUADRANT_II = 1 << 1,
		QUADRANT_III = 1 << 2,
		QUADRANT_IV = 1 << 3
	} Quadrant;

	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;
//	sdpVector3 m_Center;

	k3dICamera::Axis m_Axis;
};

class CUVMapSpherical :
	public CUVMapAlgorithm
{
public:
	bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);

private:
	void HandleCoordinates(const sdpVector3& Coords, const gulong Quadrants, sdpVector3& Results, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);

	typedef enum
	{
		QUADRANT_I = 1 << 0,
		QUADRANT_II = 1 << 1,
		QUADRANT_III = 1 << 2,
		QUADRANT_IV = 1 << 3
	} Quadrant;

	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;
//	sdpVector3 m_Center;

	k3dICamera::Axis m_Axis;
};

class CUVMapCubic :
	public CUVMapAlgorithm
{
public:
	bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	
private:
	typedef enum
	{
		XAXIS,
		YAXIS,
		ZAXIS,
		NXAXIS,
		NYAXIS,
		NZAXIS
	} Axis;

	sdpVector3 m_Min;
	sdpVector3 m_Max;
	sdpVector3 m_Size;
};

class CUVMapPath :
	public CUVMapAlgorithm
{
public:
	bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
};

class CUVMapImplicit :
	public CUVMapAlgorithm
{
public:
	bool HandleMesh(k3dIMesh* Mesh, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPoint(k3dIPoint* Point, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
	bool HandleMeshPath(k3dIPath* Path, const gdouble S0, const gdouble S1, const gdouble T0, const gdouble T1);
};

/////////////////////////////////////////////////////////////////////////////
// CUVMap

typedef k3dTool CUVMapBase;

class CUVMap :
	public CUVMapBase,
	public k3dMouseEventObserver
{
public:
	CUVMap(sdpIUnknown* Initializer);

	// k3dIObject implementation
	sdpResult InitComplete();
	sdpResult EditObject();

	// k3dIMouseEventObserver implementation
	sdpResult OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

	void PushEditing();
	void UpdateGeometry();
	void CancelEditing();
	void PopEditing();

private:
	typedef k3dObjectDialog PropertiesBase;
	/// Provides a GUI for the CUVMap tool
	class Properties :
		public PropertiesBase
	{
	public:
		Properties(CUVMap* const Object);

	private:
		CUVMap* const m_Object;

		// Event handling
		void OnEvent(sdpGtkEvent* Event);
		void OnApply();
	
		static const sdpString control_mappingtype;
		static const sdpString control_apply;
	};
	friend class Properties;

	typedef enum
	{
		PROJECTION,
		CYLINDRICAL,
		SPHERICAL,
		CUBIC,
		POLYGON,
		IMPLICIT
	} MappingType;
	
	friend std::ostream& operator<<(std::ostream& Stream, const MappingType& Value)
	{
		switch(Value)
			{
				case PROJECTION:
					Stream << "projection";
					break;
				case CYLINDRICAL:
					Stream << "cylindrical";
					break;
				case SPHERICAL:
					Stream << "spherical";
					break;
				case CUBIC:
					Stream << "cubic";
					break;
				case POLYGON:
					Stream << "path";
					break;
				case IMPLICIT:
					Stream << "implicit";
					break;
			}
	
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, MappingType& Value)
	{
		sdpString text;
		Stream >> text;
		
		if(text == "projection")
			Value = PROJECTION;
		else if(text == "cylindrical")
			Value = CYLINDRICAL;
		else if(text == "spherical")
			Value = SPHERICAL;
		else if(text == "cubic")
			Value = CUBIC;
		else if(text == "path")
			Value = POLYGON;
		else if(text == "implicit")
			Value = IMPLICIT;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	/// Store the selected mapping type.
	k3dVariable<MappingType> m_MappingType;

	CUVMapProjection m_Projection;
	CUVMapCylindrical m_Cylindrical;
	CUVMapSpherical m_Spherical;
	CUVMapCubic m_Cubic;
	CUVMapPath m_Path;
	CUVMapImplicit m_Implicit;
	
	k3dVariable<gdouble> m_S0;
	k3dVariable<gdouble> m_S1;
	k3dVariable<gdouble> m_T0;
	k3dVariable<gdouble> m_T1;
};


#endif // UVMAP_H
