// K-3D
// Copyright  1995-2002, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CWave K-3D tool, which applies a "wave" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/


#include "wave.h"

#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dspinbutton.h>

#include <k3diapplication.h>
#include <k3dicamera.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

namespace
{

class WaveLocations
{
public:
	WaveLocations(const k3dICamera::Axis Axis, const bool DisplaceX, const bool DisplaceY, const bool DisplaceZ, const gdouble Amplitude, const gdouble Wavelength, const gdouble Phase, const bool FiniteRange, const gdouble Range) :
		m_Axis(Axis),
		m_DisplaceX(DisplaceX),
		m_DisplaceY(DisplaceY),
		m_DisplaceZ(DisplaceZ),
		m_Amplitude(Amplitude),
		m_Wavelength(Wavelength),
		m_Phase(Phase),
		m_FiniteRange(FiniteRange),
		m_Range(Range)
	{
	}

	void operator()(k3d::SelectedLocationProxy& Location)
	{
		const gdouble phase = m_Phase;
		const double frequency = m_Wavelength ? 1.0 / m_Wavelength : 0.0;

		sdpVector3 coords = Location.OldXYZ();

	//	coords -= m_Origin;
		double range = coords.Length();

		// Calculate displacement based on position in "wave" space ...
		double displacement;
		switch(m_Axis)
			{
				case k3dICamera::XAXIS:
					displacement = m_Amplitude * sin(sdp::PiTimes2() * (coords[0] * frequency) + phase);
					break;
				case k3dICamera::YAXIS:
					displacement = m_Amplitude * sin(sdp::PiTimes2() * (coords[1] * frequency) + phase);
					break;
				case k3dICamera::ZAXIS:
					displacement = m_Amplitude * sin(sdp::PiTimes2() * (coords[2] * frequency) + phase);
					break;
				default:
					g_assert(0);
					displacement = 0.0;
			}

		// Factor in the distance from origin
		if(m_FiniteRange && m_Range)
	//		displacement *= (range <= m_Range) ? m_RangeChannel->GetValue1(range / m_Range, .0001) : 0.0;
			displacement *= (range <= m_Range) ? range / m_Range : 0.0;

		// Figure out which direction to move the points ...
		sdpVector3 normal;

		// Spherical waves ...
		if(m_DisplaceX && m_DisplaceY && m_DisplaceZ)
			normal = coords;

		// Cylindrical waves ...
		else if(m_DisplaceX && m_DisplaceY)
			normal = sdpVector3(coords[0], coords[1], 0.0);
		else if(m_DisplaceY && m_DisplaceZ)
			normal = sdpVector3(0.0, coords[1], coords[2]);
		else if(m_DisplaceX && m_DisplaceZ)
			normal = sdpVector3(coords[0], 0.0, coords[2]);

		// Linear waves ...
		else if(m_DisplaceX)
			normal = sdpVector3(1.0, 0.0, 0.0);
		else if(m_DisplaceY)
			normal = sdpVector3(0.0, 1.0, 0.0);
		else if(m_DisplaceZ)
			normal = sdpVector3(0.0, 0.0, 1.0);

		normal.Normalize();

		coords += displacement * normal;
		Location->SetLocalXYZ(coords);
	}

private:
	const k3dICamera::Axis m_Axis;
	const bool m_DisplaceX;
	const bool m_DisplaceY;
	const bool m_DisplaceZ;
	const gdouble m_Amplitude;
	const gdouble m_Wavelength;
	const gdouble m_Phase;
	const bool m_FiniteRange;
	const gdouble m_Range;
};

} // namespace

/////////////////////////////////////////////////////////////////////////////
// CWave

CWave::CWave(sdpIUnknown* Initializer) :
	CWaveBase(Initializer),
	k3dMouseEventObserver("LMB-drag wave amplitude; shift-LMB-drag wavelength; control-LMB-drag phase"),
	m_CameraActiveAxis(k3dICamera::ZAXIS),
	m_DisplaceX("displacex", false, 0, 0),
	m_DisplaceY("displacey", true, 0, 0),
	m_DisplaceZ("displacez", false, 0, 0),
	m_Amplitude("amplitude", 0.0, 0, 0),
	m_Wavelength("wavelength", 10.0, 0, 0),
	m_Phase("phase", 0.0, 0, 0),
	m_FiniteRange("finiterange", false, 0, 0),
	m_Range("range", 20.0, 0, 0)
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CWave::InitComplete()
{
	// Record what was selected at the time we were created
	k3d::GetSelectedLocations(Document()->Objects(), std::back_inserter(m_CurrentSelection));
	return CWaveBase::InitComplete();
}

sdpResult CWave::EditObject()
{
	new CWaveDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMouseEventObserver implementation

sdpResult CWave::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	m_CameraActiveAxis = Camera->ActiveAxis();

	PushEditing();
	return SDP_S_OK;
}

sdpResult CWave::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute factors for each mouse axis ...
	double xfactor = gdouble(CurrentX - LastX) * 0.1;
	double yfactor = gdouble(CurrentY - LastY) * 0.1;

	if(ShiftDown(Modifiers))
		k3dSetValue(m_Wavelength, m_Wavelength.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
	else if(ControlDown(Modifiers))
		k3dSetValue(m_Phase, m_Phase.Value() - sdpRadians(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));
	else
		k3dSetValue(m_Amplitude, m_Amplitude.Value() - (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor));

	// Update the display
	UpdateGeometry();
	k3dRedrawAll(Document(), false);

	return SDP_S_OK;
}

sdpResult CWave::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	PopEditing();
	return SDP_S_OK;
}

void CWave::PushEditing()
{
	k3dStartStateChangeSet(Document());
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PushEditing);
}

void CWave::UpdateGeometry()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), WaveLocations(m_CameraActiveAxis, m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), m_Amplitude.Value(), m_Wavelength.Value(), m_Phase.Value(), m_FiniteRange.Value(), m_Range.Value()));
}

void CWave::CancelEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dCancelStateChangeSet(Document());
}

void CWave::PopEditing()
{
	std::for_each(m_CurrentSelection.begin(), m_CurrentSelection.end(), &k3d::PopEditing);
	k3dFinishStateChangeSet(Document(), "Wave");
}

//////////////////////////////////////////////////////////////////////////////////////
// CWaveDialog

const sdpString CWaveDialog::dialogname = "properties";
const sdpString CWaveDialog::control_displacex = "displacex";
const sdpString CWaveDialog::control_displacey = "displacey";
const sdpString CWaveDialog::control_displacez = "displacez";
const sdpString CWaveDialog::control_amplitude = "amplitude";
const sdpString CWaveDialog::control_wavelength = "wavelength";
const sdpString CWaveDialog::control_phase = "phase";

CWaveDialog::CWaveDialog(CWave* const Object) :
	CWaveDialogBase(Object, true, true, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("toollib", "wave.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CWave>(m_Object, &CWave::PushEditing, 0), new k3dActionMethodCall<CWave>(m_Object, &CWave::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CWave>(m_Object, &CWave::PopEditing, 0));
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CWave>(m_Object, &CWave::PushEditing, 0), new k3dActionMethodCall<CWave>(m_Object, &CWave::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CWave>(m_Object, &CWave::PopEditing, 0));
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CWave>(m_Object, &CWave::PushEditing, 0), new k3dActionMethodCall<CWave>(m_Object, &CWave::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CWave>(m_Object, &CWave::PopEditing, 0));

	// Connect spin buttons ...
	Connectk3dSpinButton(control_amplitude, m_Object->m_Amplitude, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CWave>(m_Object, &CWave::PushEditing, 0), new k3dActionMethodCall<CWave>(m_Object, &CWave::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CWave>(m_Object, &CWave::PopEditing, 0));
	Connectk3dSpinButton(control_wavelength, m_Object->m_Wavelength, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CWave>(m_Object, &CWave::PushEditing, 0), new k3dActionMethodCall<CWave>(m_Object, &CWave::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CWave>(m_Object, &CWave::PopEditing, 0));
	Connectk3dSpinButton(control_phase, m_Object->m_Phase, k3dApplication()->CommandRecorder(), 0, 0, new k3dActionMethodCall<CWave>(m_Object, &CWave::PushEditing, 0), new k3dActionMethodCall<CWave>(m_Object, &CWave::UpdateGeometry, new k3dActionRedrawAll(m_Object->Document(), false, 0)), new k3dActionMethodCall<CWave>(m_Object, &CWave::PopEditing, 0));

	Show();
}


