// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBender K-3D tool, which applies an animated "Bend" transformation to geometry
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "bender.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

#include <math.h>

/////////////////////////////////////////////////////////////////////////////
// CBender

CBender::CBender(sdpIUnknown* Initializer) :
	CBenderBase(Initializer),
	m_BendAngle("bendangle", Document(), Document()->StateRecorder(), 0),
	m_Tightness("tightness", Document(), Document()->StateRecorder(), 0),
	m_JointPosition("jointposition", Document(), Document()->StateRecorder(), 0),
	m_Along("along", k3d::Y, 0, 0),
	m_Around("around", k3d::X, 0, 0)
{
	AddVariable(&m_BendAngle);
	AddVariable(&m_Tightness);
	AddVariable(&m_JointPosition);
	AddVariable(&m_Along);
	AddVariable(&m_Around);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CBender::EditObject()
{
	new CBenderDialog(this);	
	return SDP_S_OK;
}

class CBenderBendObject
{
protected:
	k3d::Axis m_Along;
	k3d::Axis m_Around;
	gdouble m_BendAngle;
	gdouble m_Tightness;
	gdouble m_JointPosition;
	sdpVector3 m_Origin;
	sdpVector3 m_Min;
	sdpVector3 m_Size;
	sdpVector3 m_JLower;
	
public:
	CBenderBendObject(k3d::Axis Along, k3d::Axis Around, gdouble BendAngle, gdouble Tightness, gdouble JointPosition ) :
		m_Along(Along),
		m_Around(Around),
		m_BendAngle(BendAngle),
		m_Tightness(Tightness),
		m_JointPosition(JointPosition)
	{
	}	
	
	void operator()(k3dIObject* const Object)
	{
		sdpVector3 min, max, size, origin, joint, jlower;
		gdouble radius = 1.0 - m_Tightness;
		
		if( !k3dGetDimensions(Object, 0, min, max) )
			return;
		
		size = max - min;
		origin = size*0.5 + min;
		joint = sdpLerp( min, max, m_JointPosition );
		jlower = sdpLerp( min, max, m_JointPosition - radius );
		
		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver);
			return;

		if( m_Along == m_Around )
			return;
		
		sdpVector3 coords;
		gdouble delta = 0;
		gdouble length = 0;
		sdpMatrix4 rotation;
		sdpVector3 axis;
		sdpVector3 offset;
		sdpVector3 pivot = origin;

		k3dILocation* location;		
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
		{
			// Get location coordinates ...
			location = locationserver->Location(handle);
			coords = location->AnimXYZ();

			switch(m_Along)
				{
					case k3d::X:
						if(size[VX])
							delta = (coords[VX] - min[VX]) / size[VX];
						length = size[VX];
						if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
						{
							coords[VX] = jlower[VX];
							pivot[VX] = jlower[VX];
						}
						else
						{
							origin[VX] = joint[VX];
						} 
						switch(m_Around)
							{
								case k3d::X:
									break;
								case k3d::Y:
									axis = sdpVector3(0,1,0);
									offset = sdpVector3(0,0,-1);
									break;
								case k3d::Z:
									axis = sdpVector3(0,0,1);
									offset = sdpVector3(0,1,0);
									break;
								default:
									g_return_if_fail(0);
							}
						break;
					case k3d::Y:
						if(size[VY])
							delta = (coords[VY] - min[VY]) / size[VY];
						length = size[VY];
						if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
						{
							coords[VY] = jlower[VY];
							pivot[VY] = jlower[VY];
						}
						else
						{
							origin[VY] = joint[VY];
						} 
						switch(m_Around)
							{
								case k3d::X:
									axis = sdpVector3(1,0,0);
									offset = sdpVector3(0,0,1);
									break;
								case k3d::Y:
									break;
								case k3d::Z:
									axis = sdpVector3(0,0,1);
									offset = sdpVector3(-1,0,0);
									break;
								default:
									g_return_if_fail(0);
							}
						break;
					case k3d::Z:
						if(size[VZ])
							delta = (coords[VZ] - min[VZ]) / size[VZ];
						length = size[VZ];
						if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
						{
							coords[VZ] = jlower[VZ];
							pivot[VZ] = jlower[VZ];
						}
						else
						{
							origin[VZ] = joint[VZ];
						} 
						switch(m_Around)
							{
								case k3d::X:
									axis = sdpVector3(1,0,0);
									offset = sdpVector3(0,-1,0);
									break;
								case k3d::Y:
									axis = sdpVector3(0,1,0);
									offset = sdpVector3(1,0,0);
									break;
								case k3d::Z:
									break;
								default:
									g_return_if_fail(0);
							}
						break;
					default:
						g_return_if_fail(0);
				}
	
			if( delta > m_JointPosition-radius && delta < m_JointPosition+radius && m_BendAngle != 0.0 )
			{
				gdouble d = (delta-(m_JointPosition-radius)) / (radius*2.0);
				gdouble factor = radius*length / tan(m_BendAngle*0.5);
				rotation = sdpRotation3D( d*m_BendAngle, axis );
				offset = offset * factor;
				coords -= pivot + offset;
				coords = rotation * coords;
				coords += pivot + offset;
			}
			else if( delta >= m_JointPosition+radius  )
			{
				rotation = sdpRotation3D( m_BendAngle, axis );
				coords -= origin;
				coords = rotation * coords;
				coords += origin;
			}
			
			location->SetAnimXYZ(coords);
		}
	}
};


///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CBender::PopAnimate()
{
	// Calculate parameter values ...
	gdouble bendangle = m_BendAngle->Value1(m_CurrentTime, m_DeltaTime);
	gdouble tightness = m_Tightness->Value1(m_CurrentTime, m_DeltaTime);
	gdouble jointposition = m_JointPosition->Value1(m_CurrentTime, m_DeltaTime);
	
	k3d::for_each_descendant(Hierarchy(), this, CBenderBendObject(m_Along.Value(), m_Around.Value(), bendangle, tightness, jointposition));
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CBender::InitComplete()
{
	if(0 == m_BendAngle.Object())
		g_return_val_if_fail(m_BendAngle.Create("Bend Angle", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Tightness.Object())
		g_return_val_if_fail(m_Tightness.Create("Tightness", 0.6, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_JointPosition.Object())
		g_return_val_if_fail(m_JointPosition.Create("Joint Position", 0.5, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	return CBenderBase::InitComplete();
}


//////////////////////////////////////////////////////////////////////////////////////
// CBenderDialog

const sdpString CBenderDialog::dialogname = "properties";
const sdpString CBenderDialog::control_along = "along";
const sdpString CBenderDialog::control_around = "around";
const sdpString CBenderDialog::control_bendangle = "bendangle";
const sdpString CBenderDialog::control_tightness = "tightness";
const sdpString CBenderDialog::control_jointposition = "jointposition";

CBenderDialog::CBenderDialog(CBender* const Object) :
	CBenderDialogBase(Object, false, false, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "bender.gtkml"));

	// Setup choosers ...
	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Along", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dChooser(control_around))
		Getk3dChooser(control_around)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Around), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Around", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Connect spin buttons ...
	Connectk3dChannelButton(control_bendangle, m_Object->m_BendAngle.Interface());
	Connectk3dChannelButton(control_tightness, m_Object->m_Tightness.Interface());
	Connectk3dChannelButton(control_jointposition, m_Object->m_JointPosition.Interface());

	RootWidget().Show();
}

