// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CBulger K-3D object, which applies an animated "Bulge" transformation to geometry
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include "bulger.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dibezierchannel.h>
#include <k3dibezierchannelnode.h>
#include <k3dieditable.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>

/////////////////////////////////////////////////////////////////////////////
// CBulger

CBulger::CBulger(sdpIUnknown* Initializer) :
	CBulgerBase(Initializer),
	m_BulgeFactor("bulgefactor", Document(), Document()->StateRecorder(), 0),
	m_DisplaceX("displacex", true, Document()->StateRecorder(), 0),
	m_DisplaceY("displacey", true, Document()->StateRecorder(), 0),
	m_DisplaceZ("displacez", true, Document()->StateRecorder(), 0),
	m_Along("along", k3d::X, Document()->StateRecorder(), 0),
	m_Profile("profile", Document(), Document()->StateRecorder(), 0),
	m_DisplaceType("displacetype", RADIAL, Document()->StateRecorder(), 0)
{
	AddVariable(&m_BulgeFactor);
	AddVariable(&m_DisplaceX);
	AddVariable(&m_DisplaceY);
	AddVariable(&m_DisplaceZ);
	AddVariable(&m_Along);
	AddVariable(&m_Profile);
	AddVariable(&m_DisplaceType);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CBulger::EditObject()
{
	g_return_val_if_fail(new CBulgerDialog(this), SDP_E_FAIL);
	return SDP_S_OK;
}


class CBulgerBulgeObject
{
protected:
	const k3d::Axis m_Along;
	const CBulger::DisplaceType m_DisplaceType;
	const bool m_DisplaceX;
	const bool m_DisplaceY;
	const bool m_DisplaceZ;
	const gdouble m_BulgeFactor;
	k3dIChannel* const m_Profile;
	
public:
	CBulgerBulgeObject(const k3d::Axis Along, const CBulger::DisplaceType Type, const bool DisplaceX, const bool DisplaceY, const bool DisplaceZ, const gdouble BulgeFactor, k3dIChannel* const Profile) :
		m_Along(Along),
		m_DisplaceType(Type),
		m_DisplaceX(DisplaceX),
		m_DisplaceY(DisplaceY),
		m_DisplaceZ(DisplaceZ),
		m_BulgeFactor(BulgeFactor),
		m_Profile(Profile)
	{
		// Sanity checks
		g_assert(m_Profile);
	}	
	
	void operator()(k3dIObject* const Object)
	{
		sdpVector3 min, max, size;
		if( !k3dGetDimensions(Object, 0, min, max) )
			return;

		size = max - min;

		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

			
		sdpVector3 coords;
		gdouble delta = 0;
		gdouble Bulge_amount;
		k3dILocation* location;		
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();
						
				switch(m_Along)
					{
						case k3d::X:
							if(size[0])
								delta = (coords[0] - min[0]) / size[0];
							break;
						case k3d::Y:
							if(size[1])
								delta = (coords[1] - min[1]) / size[1];
							break;
						case k3d::Z:
							if(size[2])
								delta = (coords[2] - min[2]) / size[2];
							break;
						default:
							g_return_if_fail(0);
					}

				Bulge_amount = m_Profile->Value1(delta, 0.001);
				if( CBulger::RADIAL == m_DisplaceType )
					{
						gdouble distance = coords.Length();
						gdouble scale;
						if( 0 == distance )
							scale = 1.0;
						else
							scale = ( distance + m_BulgeFactor * Bulge_amount ) / distance;

						if(m_DisplaceX && (k3d::X != m_Along))
							coords[0] *= scale;
		
						if(m_DisplaceY && (k3d::Y != m_Along))
							coords[1] *= scale;
		
						if(m_DisplaceZ && (k3d::Z != m_Along))
							coords[2] *= scale;
					}
				else
					{
						gdouble offset = m_BulgeFactor * Bulge_amount;

						if(m_DisplaceX && (k3d::X != m_Along))
							coords[0] += offset;
		
						if(m_DisplaceY && (k3d::Y != m_Along))
							coords[1] += offset;
			
						if(m_DisplaceZ && (k3d::Z != m_Along))
							coords[2] += offset;
					}
						
			
				location->SetAnimXYZ(coords);
			}
	}
};


///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CBulger::PopAnimate()
{
	// Calculate Bulge amount ...
	gdouble bulgefactor = m_BulgeFactor->Value1(m_CurrentTime, m_DeltaTime);
	
	// Apply it to each of our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CBulgerBulgeObject(m_Along.Value(), m_DisplaceType.Value(), m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), bulgefactor, m_Profile.Interface()));
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CBulger::InitComplete()
{
	if(0 == m_BulgeFactor.Object())
		g_return_val_if_fail(m_BulgeFactor.Create("Bulge Factor", 1.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	if(0 == m_Profile.Object())
		{
			g_return_val_if_fail(m_Profile.Create("Bulger Profile", 0.0, false, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, false), SDP_E_FAIL);

			// Let the channel know that we're making changes ...
			k3d::Editing editing(m_Profile.Interface());

			// Setup initial curve ...
			k3dIBezierChannel* const channel = dynamic_cast<k3dIBezierChannel*>(m_Profile.Interface());
			g_return_val_if_fail(channel, SDP_E_FAIL);
			
			channel->SetSegmentCount(2);
			k3dBezierChannelNodeArray nodes;
			k3dBezierChannelNodes(channel, nodes);
			g_assert( 7 == nodes.size() );
			
			k3dBezierChannelNodeArray::iterator node(nodes.begin());
			
			(*node)->SetPosition(sdpVector2(0.0, 0.0));
			node++;
			(*node)->SetPosition(sdpVector2(0.0, 0.5));
			node++;
			(*node)->SetPosition(sdpVector2(0.2, 1.0));
			node++;
			(*node)->SetPosition(sdpVector2(0.5, 1.0));
			node++;
			(*node)->SetPosition(sdpVector2(0.8, 1.0));
			node++;
			(*node)->SetPosition(sdpVector2(1.0, 0.5));
			node++;
			(*node)->SetPosition(sdpVector2(1.0, 0.0));
		}		

	return CBulgerBase::InitComplete();
}


//////////////////////////////////////////////////////////////////////////////////////
// CBulgerDialog

const sdpString CBulgerDialog::dialogname = "properties";
const sdpString CBulgerDialog::control_displacex = "displacex";
const sdpString CBulgerDialog::control_displacey = "displacey";
const sdpString CBulgerDialog::control_displacez = "displacez";
const sdpString CBulgerDialog::control_along = "along";
const sdpString CBulgerDialog::control_bulgefactor = "bulgefactor";
const sdpString CBulgerDialog::control_profile = "profile";
const sdpString CBulgerDialog::control_displacetype = "displacetype";

CBulgerDialog::CBulgerDialog(CBulger* const Object) :
	CBulgerDialogBase(Object, false, false, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "bulger.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace X", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Y", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Z", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup choosers ...
	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Along", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
std::cout << control_displacetype << std::endl;
	if(Getk3dChooser(control_displacetype))
		Getk3dChooser(control_displacetype)->Attach(new k3dChooserVariableDataSource<CBulger::DisplaceType>(m_Object->m_DisplaceType), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "DisplaceType", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup channel buttons ...
	Connectk3dChannelButton(control_bulgefactor, m_Object->m_BulgeFactor.Interface());
	Connectk3dChannelButton(control_profile, m_Object->m_Profile.Interface());

	RootWidget().Show();
}


