// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CDisplacer K-3D object, which Displaces geometry in an animated sort of way
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "displacer.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3dobjectselector.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>
#include <k3distaterecorder.h>

#include <sdpgtkevent.h>
#include <sdpgtkevents.h>

/////////////////////////////////////////////////////////////////////////////
// CDisplacer

CDisplacer::CDisplacer(sdpIUnknown* Initializer) :
	CDisplacerBase(Initializer),
	m_DisplaceX("displacex", true, Document()->StateRecorder(), 0),
	m_DisplaceY("displacey", true, Document()->StateRecorder(), 0),
	m_DisplaceZ("displacez", true, Document()->StateRecorder(), 0),
	m_Amplitude("amplitude", Document(), Document()->StateRecorder(), 0),
	m_Smoothness( "smoothness", Document(), Document()->StateRecorder(), 0),
	m_Frequency( "frequency", Document(), Document()->StateRecorder(), 0),
	m_Octaves( "octaves", Document(), Document()->StateRecorder(), 0),
	m_Time("time", Document(), Document()->StateRecorder(), 0),
	m_Seed("seed", 77, Document()->StateRecorder(), 0),
	m_FractalType("fractaltype", FBM, Document()->StateRecorder(), 0),
	m_ImageType("imagetype", AVERAGE, Document()->StateRecorder(), 0),
	m_Bitmap("bitmap", Document(), Document()->StateRecorder(), 0),
	m_DisplaceType("displacetype", RADIAL, Document()->StateRecorder(), 0),
	m_DisplaceSource("displacesource", FRACTAL, Document()->StateRecorder(), 0)
{
	AddVariable(&m_DisplaceX);
	AddVariable(&m_DisplaceY);
	AddVariable(&m_DisplaceZ);
	AddVariable(&m_Amplitude);
	AddVariable(&m_Smoothness);
	AddVariable(&m_Frequency);
	AddVariable(&m_Octaves);
	AddVariable(&m_Time);
	AddVariable(&m_Seed);
	AddVariable(&m_FractalType);
	AddVariable(&m_ImageType);
	AddVariable(&m_Bitmap);
	AddVariable(&m_DisplaceType);
	AddVariable(&m_DisplaceSource);
	
	m_Function = new CDisplaceFunction;
}

CDisplacer::~CDisplacer()
{
	delete m_Function;
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CDisplacer::EditObject()
{
	g_return_val_if_fail(new CDisplacerDialog(this), SDP_E_FAIL);
	
	return SDP_S_OK;
}


class CDisplacerDisplaceObject
{
public:
	CDisplacerDisplaceObject( CDisplaceFunction* Function, bool DisplaceX, bool DisplaceY, bool DisplaceZ, gdouble Amplitude, CDisplacer::DisplaceType Type )
	{
		m_Function = Function;
	
		m_Amplitude = Amplitude;
		m_DisplaceX = DisplaceX;
		m_DisplaceY = DisplaceY;
		m_DisplaceZ = DisplaceZ;
		
		m_Type = Type;
		
	}

	CDisplacerDisplaceObject::~CDisplacerDisplaceObject()
	{
	}

	void operator()(k3dIObject* const Object)
	{
		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		g_return_if_fail(locationserver);
			
		sdpVector3 coords;
		k3dILocation* location;		
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();

				if(CDisplacer::LINEAR == m_Type)
					{
						gdouble offset = m_Amplitude * m_Function->Value( location );
	
						if( m_DisplaceX )
							coords[0] += offset;
		
						if( m_DisplaceY )
							coords[1] += offset;
		
						if( m_DisplaceZ )
							coords[2] += offset;
					}
				else
					{
						gdouble distance = coords.Length();
						gdouble scale;
						if( 0 == distance )
							scale = 1.0;
						else
							scale = ( distance + m_Amplitude * m_Function->Value( location ) ) / distance;
		
						if( m_DisplaceX )
							coords[0] *= scale;
		
						if( m_DisplaceY )
							coords[1] *= scale;
		
						if( m_DisplaceZ )
							coords[2] *= scale;
					}

				location->SetAnimXYZ(coords);
			}
	}
	
protected:
	gdouble m_Amplitude;
	bool m_DisplaceX;
	bool m_DisplaceY;
	bool m_DisplaceZ;
	CDisplacer::DisplaceType m_Type;
	gdouble m_Frequency;
	CDisplaceFunction* m_Function;
};


///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CDisplacer::PopAnimate()
{
	gdouble amplitude = m_Amplitude->Value1(m_CurrentTime, m_DeltaTime);
	gdouble smoothness = m_Smoothness->Value1(m_CurrentTime, m_DeltaTime);
	gdouble frequency = m_Frequency->Value1(m_CurrentTime, m_DeltaTime);
	gdouble octaves = m_Octaves->Value1(m_CurrentTime, m_DeltaTime);
//	gdouble time = m_Time->Value1(m_CurrentTime, m_DeltaTime);
	
	switch(m_DisplaceSource.Value())
		{
			case FRACTAL:
				m_Function->SetFractal( m_FractalType.Value(), static_cast<const gulong>(m_Seed.Value()), smoothness, frequency, octaves );
				break;
			case IMAGE:
				if(m_Bitmap.Interface())
					{
						k3dPixel* const data = m_Bitmap.Interface()->Data();
						const gulong width = m_Bitmap.Interface()->Width();
						const gulong height = m_Bitmap.Interface()->Height();
	
						m_Function->SetImage( m_ImageType.Value(), data, width, height );
					}
				else
					{
						// No image 
						m_Function->SetNone();
					}
				break;
			default:
				g_return_val_if_fail(0, SDP_E_FAIL);
		}
	
	// Dispalce each of our descendants ...
	k3d::for_each_descendant( Hierarchy(), this, CDisplacerDisplaceObject( m_Function, m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), amplitude, m_DisplaceType.Value()));

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CDisplacer::InitComplete()
{
	if( 0 == m_Amplitude.Object() )
		g_return_val_if_fail( m_Amplitude.Create("Amplitude", 1.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if( 0 == m_Smoothness.Object() )
		g_return_val_if_fail( m_Smoothness.Create("Smoothness", 0.5, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if( 0 == m_Frequency.Object() )
		g_return_val_if_fail( m_Frequency.Create("Frequency", 0.2, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if( 0 == m_Octaves.Object() )
		g_return_val_if_fail( m_Octaves.Create("Octaves", 4.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if( 0 == m_Time.Object() )
		g_return_val_if_fail( m_Time.Create("t Value", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	return CDisplacerBase::InitComplete();
}


////////////////////////////////////////////////////////////////////////////
// CDisplacerFunction

CDisplaceFunction::CDisplaceFunction()
{
	m_Type = FRACTAL;
	m_Fractal = NULL;
}

CDisplaceFunction::~CDisplaceFunction()
{
}

void CDisplaceFunction::SetFractal( const CDisplacer::FractalType FType, const gulong Seed, const gdouble Smoothness, const gdouble Frequency, const gdouble Octaves )
{
	if( m_Fractal )
		delete m_Fractal;

	switch( FType )
		{
			case CDisplacer::RANDOM:
				m_Fractal = new CRand();
				break;
			case CDisplacer::NOISE:
				m_Fractal = new CNoise();
				break;
			case CDisplacer::FBM:
				m_Fractal = new CFBm();
				break;
			case CDisplacer::TURBULENCE:
				m_Fractal = new CTurbulence();
				break;
			case CDisplacer::RIDGED:
				m_Fractal = new CRidged();
				break;
			default:
				g_return_if_fail(0);
		}

	m_Fractal->Seed( Seed );
	m_Fractal->Smoothness( Smoothness );
	m_Fractal->Octaves( Octaves );
	
	m_Frequency = Frequency;
	
	m_Type = FRACTAL;
}

void CDisplaceFunction::SetImage( const CDisplacer::ImageType IType, k3dPixel* const Data, const gulong Width, const gulong Height )
{
	m_ImageType = IType;
	m_Data = Data;
	m_Width = Width;
	m_Height = Height;
	
	if( m_Data )
		m_Type = IMAGE;
	else
		m_Type = NONE;
}

void CDisplaceFunction::SetNone()
{
	m_Type = NONE;
}

gdouble CDisplaceFunction::Value( k3dILocation* Location )
{
	if( m_Type == FRACTAL )
		{
			if( !m_Fractal )
				return 0;
		
			return m_Fractal->Value( Location->LocalXYZ() * m_Frequency );
		}
	else if( m_Type == IMAGE )
		{		
			sdpVector3 coords = Location->TextureXYZ();

			gulong i =  (gulong)( coords[0] * m_Width ) % m_Width;
			gulong j =  (gulong)( coords[1] * m_Height ) % m_Height;

			k3dPixel pixel = m_Data[ j*m_Width + i ];
			gdouble value = 0;

			switch( m_ImageType )
				{
					case CDisplacer::AVERAGE:
						value = (pixel.Red + pixel.Green + pixel.Blue) / (3.0 * 255.0);
						break;
					case CDisplacer::RED:
						value = pixel.Red / 255.0;
						break;
					case CDisplacer::GREEN:
						value = pixel.Green / 255.0;
						break;
					case CDisplacer::BLUE:
						value = pixel.Blue / 255.0;
						break;
					case CDisplacer::ALPHA:
						value = pixel.Alpha / 255.0;
						break;
					default:
						g_return_val_if_fail(0, 0);
				}
			return value;
		}
	else
		{
			return 0;
		}
}

////////////////////////////////////////////////////////////////////////////
// CDisplacerDialog

const sdpString CDisplacerDialog::dialogname = "properties";
const sdpString CDisplacerDialog::control_amplitude = "amplitude";
const sdpString CDisplacerDialog::control_displacex = "displacex";
const sdpString CDisplacerDialog::control_displacey = "displacey";
const sdpString CDisplacerDialog::control_displacez = "displacez";
const sdpString CDisplacerDialog::control_smoothness = "smoothness";
const sdpString CDisplacerDialog::control_frequency = "frequency";
const sdpString CDisplacerDialog::control_octaves = "octaves";
const sdpString CDisplacerDialog::control_time = "time";
const sdpString CDisplacerDialog::control_fractaltype = "fractaltype";
const sdpString CDisplacerDialog::control_randomseed = "randomseed";
const sdpString CDisplacerDialog::control_displacetype = "displacetype";
const sdpString CDisplacerDialog::control_imagetype = "imagetype";
const sdpString CDisplacerDialog::control_bitmap = "bitmap";
const sdpString CDisplacerDialog::control_pageswitch = "pageswitch";
const sdpString CDisplacerDialog::control_pagefractal = "pagefractal";
const sdpString CDisplacerDialog::control_pageimage = "pageimage";

CDisplacerDialog::CDisplacerDialog(CDisplacer* const Object) :
	CDisplacerDialogBase(Object, false, false, false, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "displacer.gtkml"));

	// Connect check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace X", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Y", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Z", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup channel buttons ...
	Connectk3dChannelButton(control_amplitude, m_Object->m_Amplitude.Interface());
	Connectk3dChannelButton(control_frequency, m_Object->m_Frequency.Interface());
	Connectk3dChannelButton(control_smoothness, m_Object->m_Smoothness.Interface());
	Connectk3dChannelButton(control_octaves, m_Object->m_Octaves.Interface());
//	Connectk3dChannelButton(control_time, m_Object->m_Time.Interface());

	// Attach chooser 
	if( Getk3dChooser( control_fractaltype ) )
	{
		Getk3dChooser( control_fractaltype )->Attach(
			new k3dChooserVariableDataSource<CDisplacer::FractalType>(m_Object->m_FractalType),
			k3dApplication()->CommandRecorder(),
			m_Object->Document()->StateRecorder(),
			"Fractal Type",
			0,
			new k3dActionRedrawAll(m_Object->Document(), false, 0),
			0);
	}
	if( Getk3dChooser( control_imagetype ) )
	{
		Getk3dChooser( control_imagetype )->Attach(
			new k3dChooserVariableDataSource<CDisplacer::ImageType>(m_Object->m_ImageType),
			k3dApplication()->CommandRecorder(),
			m_Object->Document()->StateRecorder(),
			"Image Type",
			0,
			new k3dActionRedrawAll(m_Object->Document(), false, 0),
			0);
	}
	if( Getk3dChooser( control_displacetype ) )
	{
		Getk3dChooser( control_displacetype )->Attach(
			new k3dChooserVariableDataSource<CDisplacer::DisplaceType>(m_Object->m_DisplaceType),
			k3dApplication()->CommandRecorder(),
			m_Object->Document()->StateRecorder(),
			"Displace Type",
			0,
			new k3dActionRedrawAll(m_Object->Document(), false, 0),
			0);
	}

	// Attach object selectors ...
	if(Getk3dObjectSelector(control_bitmap))
		Getk3dObjectSelector(control_bitmap)->Attach(
			&m_Object->m_Bitmap,
			k3dApplication()->CommandRecorder(),
			m_Object->Document()->StateRecorder(),
			"Select Bitmap",
			0,
			new k3dActionRedrawAll(m_Object->Document(), false, 0),
			0,
			true,
			new k3dObjectSelectorInterfaceFilter(new sdp::TypeInfo<k3dIBitmap>()));

	// Initialise our page ...
	BlockAllEvents();
	Notebook("tabs").SetPage(m_Object->m_DisplaceSource.Value());
	UnblockAllEvents();

	RootWidget().Show();
}

void CDisplacerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == control_pageswitch)
		OnPageSwitch(Event);
	else
		CDisplacerDialogBase::OnEvent(Event);
}

void CDisplacerDialog::OnPageSwitch( sdpGtkEvent* Event )
{
	// Sanity checks ...
	g_assert( Event );
	
	// Keep track of the current page ...
	sdpGtkEventNotebookSwitchPage* event = (sdpGtkEventNotebookSwitchPage*)Event;
	CDisplacer::DisplaceSource currentpage = CDisplacer::DisplaceSource(event->PageNumber());
	
	sdpString pagename;
	switch( currentpage )
		{
			case CDisplacer::FRACTAL:
				pagename = control_pagefractal;
				break;
			case CDisplacer::IMAGE:
				pagename = control_pageimage;
				break;
			default:
				g_return_if_fail(0);
		}
	
	k3dRecordCommand(this, k3dICommandNode::USERINTERFACE, control_pageswitch, pagename);
	
	k3dStartStateChangeSet(m_Object->Document());
	k3dSetValue(m_Object->m_DisplaceSource, currentpage);
	k3dFinishStateChangeSet(m_Object->Document(), "Displacement Source");
	
	k3dRedrawAll(m_Object->Document(), false);
}

