#ifndef DISPLACER_H
#define DISPLACER_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CDisplacer K-3D object, which displaces geometry in an animated sort of way
		\author Andy Gill (buzz@ucky.com)
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "fractals.h"

#include <k3dilocation.h>

#include <k3dpersistobject.h>
#include <k3dobjectdialog.h>
#include <k3dibitmap.h>

// Forward class references
class CDisplacer;
class CDisplaceFunction;

/////////////////////////////////////////////////////////////////////////////
// CDisplacerPage dialog

typedef k3dObjectDialog CDisplacerDialogBase;

class CDisplacerDialog :
	public CDisplacerDialogBase
{
public:
	CDisplacerDialog(CDisplacer* const Object);

protected:
	CDisplacer* const m_Object;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);
	void OnPageSwitch(sdpGtkEvent* Event);

	static const sdpString dialogname;
	static const sdpString control_amplitude;
	static const sdpString control_displacex;
	static const sdpString control_displacey;
	static const sdpString control_displacez;
	static const sdpString control_smoothness;
	static const sdpString control_frequency;
	static const sdpString control_octaves;
	static const sdpString control_time;
	static const sdpString control_fractaltype;
	static const sdpString control_randomseed;
	static const sdpString control_displacetype;
	static const sdpString control_imagetype;
	static const sdpString control_bitmap;
	
	static const sdpString control_pageswitch;
	static const sdpString control_pagefractal;
	static const sdpString control_pageimage;
};


/////////////////////////////////////////////////////////////////////////////
// CDisplacer

typedef k3dPersistObject CDisplacerBase;

class CDisplacer : public CDisplacerBase
{
friend class CDisplacerDialog;
public:
	CDisplacer(sdpIUnknown* Initializer);
	~CDisplacer();

	// k3dIObject overrides
	sdpResult EditObject();

	// k3dAnimate overrides
	sdpResult PopAnimate();
	
	// k3dPersist overrides
	sdpResult InitComplete();

	typedef enum
	{
		RANDOM,
		NOISE,
		FBM,
		TURBULENCE,
		RIDGED,
	} FractalType;

	friend std::ostream& operator<<(std::ostream& Stream, const FractalType& Value)
	{
		switch(Value)
			{
				case RANDOM:
					Stream << "random";
					break;
				case NOISE:
					Stream << "noise";
					break;
				case FBM:
					Stream << "fBm";
					break;
				case TURBULENCE:
					Stream << "turbulence";
					break;
				case RIDGED:
					Stream << "ridged";
					break;
			}
        	return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, FractalType& Value)
	{
		sdpString text;
		Stream >> text;
        	
		if( text == "random" )
			Value = RANDOM;
		else if( text == "noise" )
			Value = NOISE;
		else if( text == "fBm" )
			Value = FBM;
		else if( text == "turbulence" )
			Value = TURBULENCE;
		else if( text == "ridged" )
			Value = RIDGED;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}

	
	typedef enum
	{
		AVERAGE,
		RED,
		GREEN,
		BLUE,
		ALPHA,
	} ImageType;

	friend std::ostream& operator<<(std::ostream& Stream, const ImageType& Value)
	{
		switch( Value )
			{
				case AVERAGE:
					Stream << "average";
					break;
				case RED:
					Stream << "red";
					break;
				case GREEN:
					Stream << "green";
					break;
				case BLUE:
					Stream << "blue";
					break;
				case ALPHA:
					Stream << "alpha";
					break;
			}
		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, ImageType& Value)
	{
		sdpString text;
		Stream >> text;

		if( text == "average" )
			Value = AVERAGE;
		else if( text == "red" )
			Value = RED;
		else if( text == "green" )
			Value = GREEN;
		else if( text == "blue" )
			Value = BLUE;
		else if( text == "alpha" )
			Value = ALPHA;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	typedef enum
	{
		LINEAR,
		RADIAL,
	} DisplaceType;

	friend std::ostream& operator<<(std::ostream& Stream, const DisplaceType& Value)
	{
		switch(Value)
			{
				case LINEAR:
					Stream << "linear";
					break;
				case RADIAL:
					Stream << "radial";
					break;
			}
		return Stream;
	}
	
	friend std::istream& operator>>(std::istream& Stream, DisplaceType& Value)
	{
		sdpString text;
		Stream >> text;
	
		if(text == "linear")
			Value = LINEAR;
		else if(text == "radial")
			Value = RADIAL;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;
		
		return Stream;
	}
	
	typedef enum
	{
		FRACTAL,
		IMAGE,
	} DisplaceSource;

	friend std::ostream& operator<<(std::ostream& Stream, const DisplaceSource& Value)
	{
		switch(Value)
			{
				case FRACTAL:
					Stream << "fractal";
					break;
				case IMAGE:
					Stream << "image";
					break;
			}
		
		return Stream;
	}

	friend std::istream& operator>>(std::istream& Stream, DisplaceSource& Value)
	{
		sdpString text;
		Stream >> text;

		if(text == "fractal")
			Value = FRACTAL;
		else if(text == "image")
			Value = IMAGE;
		else
			std::cerr << g_pretty_function() << ": unknown enumeration [" << text << "]" << std::endl;

		return Stream;
	}
	
protected:
	k3dVariable<bool> m_DisplaceX;
	k3dVariable<bool> m_DisplaceY;
	k3dVariable<bool> m_DisplaceZ;
	k3dScalarChannel m_Amplitude;
	k3dScalarChannel m_Smoothness;
	k3dScalarChannel m_Frequency;
	k3dScalarChannel m_Octaves;
	k3dScalarChannel m_Time;
	k3dVariable<gdouble> m_Seed;
	k3dVariable<FractalType> m_FractalType;
	k3dVariable<ImageType> m_ImageType;

	k3dObjectInterfaceVariable<k3dIBitmap> m_Bitmap;

	k3dVariable<DisplaceType> m_DisplaceType;
	k3dVariable<DisplaceSource> m_DisplaceSource;
	
	CDisplaceFunction* m_Function;
};


/////////////////////////////////////////////////////////////////////////////
// CDisplaceFunction

class CDisplaceFunction
{
public:
	CDisplaceFunction();
	~CDisplaceFunction();
	
	void SetFractal( const CDisplacer::FractalType FType, const gulong Seed, const gdouble Smoothness, const gdouble Frequency, const gdouble Octaves );
	void SetImage( const CDisplacer::ImageType IType, k3dPixel* const Data, const gulong Width, const gulong Height );
	void SetNone();
	
	gdouble Value( k3dILocation* Location );

protected:
	typedef enum
	{
		FRACTAL,
		IMAGE,
		NONE,
	} Type;
	
	Type m_Type;
	CDisplacer::ImageType m_ImageType;
	
	CFractal* m_Fractal;
	gdouble m_Frequency;
	
	k3dPixel* m_Data;
	gulong m_Width;
	gulong m_Height;
};


#endif // DISPLACER_H
