// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CFractal class which encapsulates a fractal construction
		\author Andy Gill (buzz@ucky.com)
*/


#include "fractals.h"


/////////////////////////////////////////////////////////////////////////////
// CFractal

CFractal::CFractal()
{
	m_H = 0.5;
	m_Lacunarity = 2.0;
	m_Octaves = 4.0;
	m_Exponents = NULL;
}

// Provide a single smoothness parameter
void CFractal::Smoothness( gdouble smoothness )
{
	m_H = sdpMin( smoothness, 1.0 );
//	m_Lacunarity = smoothness*2.0 + 1.1;

	CalculateExponents();
}

void CFractal::Octaves( gdouble octaves )
{
	m_Octaves = octaves;
	
	CalculateExponents();
}

void CFractal::CalculateExponents()
{
	if( m_Exponents )
		delete [] m_Exponents;
	
	int num_of_exponents = (int)( ceil( m_Octaves )) + 1;
	m_Exponents = new gdouble[ num_of_exponents ];
	
	// Compute weight for each frequency
	gdouble freq = 1.0;
	for( int i = 0; i <= m_Octaves; i++ )
		{
			m_Exponents[i] = pow( freq, -m_H );
			freq *= m_Lacunarity;
		}

	CalculateBounds();
}


void CFractal::CalculateBounds()
{
	// Disable re-normalisation
	m_Min = 0.0;
	m_Max = 1.0;

	// Take a (random) cross section of values
	sdpVector3 p( 0, 1.173, -16.857 );
	gdouble min = 100;
	gdouble max = -100;
	for( int i=0; i<1000; i++ )
		{
			p[0] = i/359.45;
			
			gdouble v = Value( p );
			min = sdpMin( min, v );
			max = sdpMax( max, v );
		}

	// Reset re-normalise constants
	m_Min = min;
	m_Max = max;
}

void CFractal::Seed( gulong seed )
{
	// Get a random offset
	m_Seed = sdpRandom( seed ) * 10000.0;
}


/////////////////////////////////////////////////////////////////////////////
// CFBm

CFBm::CFBm()
{
	CalculateExponents();
}

gdouble CFBm::Value( const sdpVector3 Point )
{
	gdouble value = 0.0;
	sdpVector3 temp = Point + sdpVector3(m_Seed);
	
	for( int i = 0; i < m_Octaves; i++ )
		{
			value += sdpNoise( temp ) * m_Exponents[i];
			temp *= m_Lacunarity;
		}
	
	// Add in fractional octave
	gdouble remainder = m_Octaves - floor( m_Octaves );
	if( remainder != 0.0 )
		value += sdpNoise( temp ) * m_Exponents[ (int)ceil(m_Octaves) ] * remainder;
	
	// Scale value between [0,1]
	value = (value - m_Min) / (m_Max - m_Min);
	
	return value;
}


/////////////////////////////////////////////////////////////////////////////
// CTurbulence

CTurbulence::CTurbulence()
{
	CalculateExponents();
}

gdouble CTurbulence::Value( const sdpVector3 Point )
{
	gdouble value = 0.0;
	sdpVector3 temp = Point + sdpVector3(m_Seed);
	
	for( int i = 0; i < m_Octaves; i++ )
		{
			value += fabs( sdpNoise( temp ) ) * m_Exponents[i];
			temp *= m_Lacunarity;
		}
	
	// Add in fractional octave
	gdouble remainder = m_Octaves - floor( m_Octaves );
	if( remainder != 0.0 )
		value += fabs( sdpNoise( temp ) ) * m_Exponents[ (int)ceil(m_Octaves) ] * remainder;
	
	// Scale value between [0,1]
	value = (value - m_Min) / (m_Max - m_Min);

	return value;
}


/////////////////////////////////////////////////////////////////////////////
// CRidged

CRidged::CRidged()
{
	m_Offset = 1.0;
	m_Gain = 2.0;

	CalculateExponents();
}

gdouble CRidged::Value( const sdpVector3 Point )
{
	gdouble value = 0.0;
	gdouble signal;
	gdouble weight = 1.0;
	sdpVector3 temp = Point + sdpVector3(m_Seed);
	
		
	for( int i = 0; i < m_Octaves; i++ )
		{
			signal = m_Offset - fabs( sdpNoise( temp ) );
			value += signal * signal * weight * m_Exponents[i];
			
			temp *= m_Lacunarity;
			weight = sdpClamp( 0.0, signal*m_Gain, 1.0 );
		}
	
	// Add in fractional octave
	gdouble remainder = m_Octaves - floor( m_Octaves );
	if( remainder != 0.0 )
		{
			signal = m_Offset - fabs( sdpNoise( temp ) );
			value += signal * signal * weight * m_Exponents[ (int)ceil(m_Octaves) ] * remainder;
		}
		
	// Scale value between [0,1]
	value = (value - m_Min) / (m_Max - m_Min);

	return value;
}


/////////////////////////////////////////////////////////////////////////////
// CNoise

gdouble CNoise::Value( const sdpVector3 Point )
{
	gdouble value;

	value = sdpNoise( Point + sdpVector3(m_Seed) );
	return ( value + 1.0 ) / 2.0;
}


/////////////////////////////////////////////////////////////////////////////
// CRandom

gdouble CRand::Value( const sdpVector3 Point )
{
	return sdpRandom( m_Seed );
}

void CRand::Seed( gulong seed )
{
	m_Seed = seed;
}
