// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSlanter K-3D object, which applies an animated "slant" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "slanter.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>

#include <sdpgtkevent.h>

CSlanter::CSlanter(sdpIUnknown* Initializer) :
	CSlanterBase(Initializer),
	m_Displace("displace", k3d::X, Document()->StateRecorder(), 0),
	m_Along("along", k3d::Y, Document()->StateRecorder(), 0),
	m_Mirror("mirror", false, Document()->StateRecorder(), 0),
	m_SlantFactor("slantfactor", Document(), Document()->StateRecorder(), 0),
	m_SlantFalloff("slantfalloff", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_Displace);
	AddVariable(&m_Along);
	AddVariable(&m_Mirror);
	AddVariable(&m_SlantFactor);
	AddVariable(&m_SlantFalloff);
}

CSlanter::~CSlanter()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSlanter::EditObject()
{
	new CSlanterDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CSlanterSlantObject
{
public:
	CSlanterSlantObject(k3d::Axis Displace, k3d::Axis Along, bool Mirror, gdouble SlantFactor, k3dIChannel* SlantFalloff)
	{
		// Sanity checks ...
		g_assert(SlantFalloff);
		
		m_Displace = Displace;
		m_Along = Along;
		m_Mirror = Mirror;
		m_SlantFactor = SlantFactor;
		m_SlantFalloff = SlantFalloff;
	}

	void operator()(k3dIObject* const Object)
	{
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dGetDimensions(Object, 0, min, max))
			return;

		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble maxx = (max[0] - origin[0]) > (origin[0] - min[0]) ? max[0] - origin[0] : origin[0] - min[0];
		gdouble maxy = (max[1] - origin[1]) > (origin[1] - min[1]) ? max[1] - origin[1] : origin[1] - min[1];
		gdouble maxz = (max[2] - origin[2]) > (origin[2] - min[2]) ? max[2] - origin[2] : origin[2] - min[2];

		// Test to see that our slant axis isn't zero-length ...
		switch(m_Along)
			{
				case k3d::X:
					if(0 == maxx)
						return;
					break;
				case k3d::Y:
					if(0 == maxy)
						return;
					break;
				case k3d::Z:
					if(0 == maxz)
						return;
					break;
				default:
					g_return_if_fail(0);
			}

		// For each location ...
		gdouble delta;
		sdpVector3 coords;
		k3dILocation* location;
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();

				// Offset the origin ...
				coords -= origin;

				// Get our position along the slant axis ...
				switch(m_Along)
					{
						case k3d::X:
							delta = coords[0] / maxx;
							break;
						case k3d::Y:
							delta = coords[1] / maxy;
							break;
						case k3d::Z:
							delta = coords[2] / maxz;
							break;
						default:
							g_return_if_fail(0);
					}

				// Mirror it and handle falloff ...
				if(m_Mirror)
					delta = m_SlantFalloff->Value1(fabs(delta), 0.001);
				else
					delta = sdpSign(delta) * m_SlantFalloff->Value1(fabs(delta), 0.001);

				// Offset the coordinates ...
				switch(m_Displace)
					{
						case k3d::X:
							coords[0] += delta * m_SlantFactor;
							break;
						case k3d::Y:
							coords[1] += delta * m_SlantFactor;
							break;
						case k3d::Z:
							coords[2] += delta * m_SlantFactor;
							break;
						default:
							g_return_if_fail(0);
					}

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
	}
	
protected:
	k3d::Axis m_Displace;
	k3d::Axis m_Along;
	bool m_Mirror;
	gdouble m_SlantFactor;
	k3dIChannel* m_SlantFalloff;
};

sdpResult CSlanter::PopAnimate()
{
	// Calculate Rotation values ...
	double slantfactor = m_SlantFactor->Value1(m_CurrentTime, m_DeltaTime);
	
	// Apply it to each of our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CSlanterSlantObject(m_Displace.Value(), m_Along.Value(), m_Mirror.Value(), slantfactor, m_SlantFalloff.Interface()));

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CSlanter::InitComplete()
{
	if(0 == m_SlantFactor.Object())
		g_return_val_if_fail(m_SlantFactor.Create("Slant Factor", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);

	if(0 == m_SlantFalloff.Object())
		{
			g_return_val_if_fail(m_SlantFalloff.Create("Slant Falloff", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_SlantFalloff.Interface());
			k3dChannelSetLinearValue(m_SlantFalloff.Interface(), 0.0, 1.0);
		}

	return CSlanterBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CSlanterDialog

const sdpString CSlanterDialog::dialogname = "properties";
const sdpString CSlanterDialog::control_displace = "displace";
const sdpString CSlanterDialog::control_along = "along";
const sdpString CSlanterDialog::control_mirror = "mirror";
const sdpString CSlanterDialog::control_slantfactor = "slantfactor";
const sdpString CSlanterDialog::control_slantfalloff = "slantfalloff";

CSlanterDialog::CSlanterDialog(CSlanter* const Object) :
	CSlanterDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "slanter.gtkml"));

	// Setup choosers ...
	if(Getk3dChooser(control_displace))
		Getk3dChooser(control_displace)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Displace), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Along", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup check buttons ...
	Connectk3dCheckButton(control_mirror, m_Object->m_Mirror, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Mirror", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	
	// Setup channel buttons ...
	Connectk3dChannelButton(control_slantfactor, m_Object->m_SlantFactor.Interface());
	Connectk3dChannelButton(control_slantfalloff, m_Object->m_SlantFalloff.Interface());

	Show();
}

