// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSphereizer K-3D object, which applies an animated "sphereize" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "sphereizer.h"

#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3dilocation.h>
#include <k3dilocationcollection.h>

CSphereizer::CSphereizer(sdpIUnknown* Initializer) :
	CSphereizerBase(Initializer),
	m_SphereizeFactor("sphereizefactor", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_SphereizeFactor);
}

CSphereizer::~CSphereizer()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSphereizer::EditObject()
{
	new CSphereizerDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CSphereizerSphereizeObject
{
public:
	CSphereizerSphereizeObject(const gdouble SphereizeFactor) :
		m_SphereizeFactor(SphereizeFactor)
	{
	}

	void operator()(k3dIObject* const Object)
	{
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dGetDimensions(Object, 0, min, max))
			return;

		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble targetradius = (max - origin).Length();

		// For each location ...
		sdpVector3 coords, coords2;
		k3dILocation* location;
		gdouble radius;
		
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords= location->AnimXYZ();

				// Offset the origin ...
				coords -= origin;

				radius = coords.Length();
				if(radius)
					{
						coords2 = coords;
						coords += coords2.Normalize() * (m_SphereizeFactor * (targetradius - radius));
					}

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
	}
	
protected:
	const gdouble m_SphereizeFactor;
};

sdpResult CSphereizer::PopAnimate()
{
	// Modify our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CSphereizerSphereizeObject(m_SphereizeFactor->Value1(m_CurrentTime, m_DeltaTime)));
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CSphereizer::InitComplete()
{
	if(0 == m_SphereizeFactor.Object())
		g_return_val_if_fail(m_SphereizeFactor.Create("Sphereize Factor", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	return CSphereizerBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CSphereizerDialog

const sdpString CSphereizerDialog::dialogname = "properties";
const sdpString CSphereizerDialog::control_sphereizefactor = "sphereizefactor";

CSphereizerDialog::CSphereizerDialog(CSphereizer* const Object) :
	CSphereizerDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "sphereizer.gtkml"));

	Connectk3dChannelButton(control_sphereizefactor, m_Object->m_SphereizeFactor.Interface());

	CallUpdateControls();
	Show();
}

void CSphereizerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	CSphereizerDialogBase::OnEvent(Event);
}


