// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTaperer K-3D object, which applies an animated "taper" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "taperer.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>

#include <sdpgtkevent.h>

CTaperer::CTaperer(sdpIUnknown* Initializer) :
	CTapererBase(Initializer),
	m_DisplaceX("displacex", false, Document()->StateRecorder(), 0),
	m_DisplaceY("displacey", true, Document()->StateRecorder(), 0),
	m_DisplaceZ("displacez", true, Document()->StateRecorder(), 0),
	m_Along("along", k3d::X, Document()->StateRecorder(), 0),
	m_TaperFactor("taperfactor", Document(), Document()->StateRecorder(), 0),
	m_TaperFalloff("taperfalloff", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_DisplaceX);
	AddVariable(&m_DisplaceY);
	AddVariable(&m_DisplaceZ);
	AddVariable(&m_Along);
	AddVariable(&m_TaperFactor);
	AddVariable(&m_TaperFalloff);
}

CTaperer::~CTaperer()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTaperer::EditObject()
{
	new CTapererDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CTapererTaperObject
{
public:
	CTapererTaperObject(bool DisplaceX, bool DisplaceY, bool DisplaceZ, k3d::Axis Along, gdouble TaperFactor, k3dIChannel* TaperFalloff)
	{
		// Sanity checks ...
		g_assert(TaperFalloff);
		
		m_DisplaceX = DisplaceX;
		m_DisplaceY = DisplaceY;
		m_DisplaceZ = DisplaceZ;
		m_Along = Along;
		m_TaperFactor = TaperFactor;
		m_TaperFalloff = TaperFalloff;
	}

	void operator()(k3dIObject* const Object)
	{
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dGetDimensions(Object, 0, min, max))
			return;

		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble maxx = (max[0] - origin[0]) > (origin[0] - min[0]) ? max[0] - origin[0] : origin[0] - min[0];
		gdouble maxy = (max[1] - origin[1]) > (origin[1] - min[1]) ? max[1] - origin[1] : origin[1] - min[1];
		gdouble maxz = (max[2] - origin[2]) > (origin[2] - min[2]) ? max[2] - origin[2] : origin[2] - min[2];
		
		gdouble omtaperfactor = 1.0 - m_TaperFactor;

		// Test to see that our taper axis isn't zero-length ...
		switch(m_Along)
			{
				case k3d::X:
					if(0 == maxx)
						return;
					break;
				case k3d::Y:
					if(0 == maxy)
						return;
					break;
				case k3d::Z:
					if(0 == maxz)
						return;
					break;
				default:
					g_return_if_fail(0);
			}

		// For each location ...
		gdouble delta;
		gdouble scale;
		sdpVector3 coords;
		k3dILocation* location;
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();

				// Offset the origin ...
				coords -= origin;

				// Get our position along the slant axis ...
				switch(m_Along)
					{
						case k3d::X:
							delta = coords[0] / maxx;
							break;
						case k3d::Y:
							delta = coords[1] / maxy;
							break;
						case k3d::Z:
							delta = coords[2] / maxz;
							break;
						default:
							g_return_if_fail(0);
					}

				// Handle falloff ...
				delta = m_TaperFalloff->Value1(fabs(delta), 0.001);
				scale = delta * omtaperfactor + m_TaperFactor;

				// Offset the coordinates ...
				if(m_DisplaceX)
					coords[0] *= scale;
				if(m_DisplaceY)
					coords[1] *= scale;
				if(m_DisplaceZ)
					coords[2] *= scale;

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
	}
	
protected:
	bool m_DisplaceX;
	bool m_DisplaceY;
	bool m_DisplaceZ;
	k3d::Axis m_Along;
	gdouble m_TaperFactor;
	k3dIChannel* m_TaperFalloff;
};

sdpResult CTaperer::PopAnimate()
{
	// Calculate Rotation values ...
	gdouble taperfactor = m_TaperFactor->Value1(m_CurrentTime, m_DeltaTime);
	
	// Apply it to each of our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CTapererTaperObject(m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), m_Along.Value(), taperfactor, m_TaperFalloff.Interface()));

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CTaperer::InitComplete()
{
	if(0 == m_TaperFactor.Object())
		g_return_val_if_fail(m_TaperFactor.Create("Taper Factor", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);

	if(0 == m_TaperFalloff.Object())
		{
			g_return_val_if_fail(m_TaperFalloff.Create("Taper Falloff", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_TaperFalloff.Interface());
			k3dChannelSetLinearValue(m_TaperFalloff.Interface(), 1.0, 0.0);
		}
	
	return CTapererBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CTapererDialog

const sdpString CTapererDialog::dialogname = "properties";
const sdpString CTapererDialog::control_displacex = "displacex";
const sdpString CTapererDialog::control_displacey = "displacey";
const sdpString CTapererDialog::control_displacez = "displacez";
const sdpString CTapererDialog::control_along = "along";
const sdpString CTapererDialog::control_taperfactor ="taperfactor";
const sdpString CTapererDialog::control_taperfalloff = "taperfalloff";

CTapererDialog::CTapererDialog(CTaperer* const Object) :
	CTapererDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "taperer.gtkml"));

	// Setup choosers ...
	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Along", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace X", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Y", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Z", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup channel buttons ...
	Connectk3dChannelButton(control_taperfactor, m_Object->m_TaperFactor.Interface());
	Connectk3dChannelButton(control_taperfalloff, m_Object->m_TaperFalloff.Interface());

	CallUpdateControls();
	Show();
}


