// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTransformer class, which is used to animate the position, orientation, and scale of objects
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "transformer.h"

#include <k3dchooser.h>
#include <k3dmisc.h>
#include <k3doptions.h>
#include <k3dorientationcontroller.h>
#include <k3dpositioncontroller.h>
#include <k3dscalecontroller.h>
#include <k3dtargetcontroller.h>

#include <k3diapplication.h>
#include <k3dicurve.h>
#include <k3dilocation.h>
#include <k3dipoint.h>

#include <sdpgtkevent.h>

CTransformer::CTransformer(sdpIUnknown* Initializer) :
	CTransformerBase(Initializer),
	m_GlobalCycles("globalcycles", 1.0, Document()->StateRecorder(), 0),
	m_GlobalPhase("globalphase", 0.0, Document()->StateRecorder(), 0),
	m_EnableTranslation("enabletranslation", true, Document()->StateRecorder(), 0),
	m_AbsolutePath("absolutepath", true, Document()->StateRecorder(), 0),
	m_TranslateLast("translatelast", true, Document()->StateRecorder(), 0),
	m_TranslateType("translatetype", LINEAR, Document()->StateRecorder(), 0),
	m_XRotation("xrotation", Document(), Document()->StateRecorder(), 0),
	m_YRotation("yrotation", Document(), Document()->StateRecorder(), 0),
	m_ZRotation("zrotation", Document(), Document()->StateRecorder(), 0),
	m_XScale("xscale", Document(), Document()->StateRecorder(), 0),
	m_YScale("yscale", Document(), Document()->StateRecorder(), 0),
	m_ZScale("zscale", Document(), Document()->StateRecorder(), 0),
	m_XOffset("xoffset", Document(), Document()->StateRecorder(), 0),
	m_YOffset("yoffset", Document(), Document()->StateRecorder(), 0),
	m_ZOffset("zoffset", Document(), Document()->StateRecorder(), 0),
	m_Translation("translation", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_GlobalCycles);
	AddVariable(&m_GlobalPhase);
	AddVariable(&m_EnableTranslation);
	AddVariable(&m_AbsolutePath);
	AddVariable(&m_TranslateLast);
	AddVariable(&m_TranslateType);
	AddVariable(&m_Translation);
	AddVariable(&m_XRotation);
	AddVariable(&m_YRotation);
	AddVariable(&m_ZRotation);
	AddVariable(&m_XScale);
	AddVariable(&m_YScale);
	AddVariable(&m_ZScale);
	AddVariable(&m_XOffset);
	AddVariable(&m_YOffset);
	AddVariable(&m_ZOffset);
}

CTransformer::~CTransformer()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTransformer::EditObject()
{
	new CTransformerDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CTransformer::PopAnimate()
{
	// Calculate our global time ...
	gdouble globaltime = (m_CurrentTime * m_GlobalCycles.Value()) + m_GlobalPhase.Value();
	globaltime -= floor(globaltime);

	// Calculate transformations for our children ...
	SetPosition(k3dITransform::ANIMATED_CHILD_FRAME, GetPosition(globaltime, m_DeltaTime));
	
	sdpVector3 angles = GetOrientation(globaltime, m_DeltaTime);
	SetOrientation(k3dITransform::ANIMATED_CHILD_FRAME, sdpAngleAxis(sdpQuaternion(sdpEulerAngles(angles[2], angles[0], angles[1], sdpEulerAngles::ZXYstatic))));
	SetScale(k3dITransform::ANIMATED_CHILD_FRAME, GetScale(globaltime, m_DeltaTime));
		
	return CTransformerBase::PopAnimate();
}

sdpVector3 CTransformer::GetPointPosition(gdouble PathPercent)
{
	gulong pointcount = PointCount();
	if(0 == pointcount)
		return sdpVector3();

	gulong pointindex = 0;
	for(K3DPOINTHANDLE pointhandle = FirstPoint(); pointhandle; pointhandle = NextPoint(pointhandle))
		{
			if(gdouble(++pointindex) / gdouble(pointcount) < PathPercent)
				continue;
				
			sdpVector3 result = Point(pointhandle)->Location()->AnimXYZ();
			return result;
		}

#ifdef	FAIL_ON_RET
	sdpVector3 tmpResult;
	g_return_val_if_fail(0, tmpResult);
#else
	g_return_val_if_fail(0, sdpVector3());
#endif
}

sdpVector3 CTransformer::GetLinearPosition(gdouble PathPercent)
{
	if(0 == PointCount())
		return sdpVector3();
	
	// Get the first point ...
	sdpVector3 firstpoint = Point(FirstPoint())->Location()->AnimXYZ();
	if(1 == PointCount())
		return firstpoint;
	
	// Calculate the total distance around the path
	sdpVector3 point;
	sdpVector3 lastpoint = firstpoint;
	gdouble pathlength = 0;
	for(K3DPOINTHANDLE pointhandle1 = NextPoint(FirstPoint()); pointhandle1; pointhandle1 = NextPoint(pointhandle1))
		{
			point= Point(pointhandle1)->Location()->AnimXYZ();

			pathlength += (point - lastpoint).Length();
			lastpoint = point;
		}

	if(m_TranslateLast.Value())
		{
			pathlength += (firstpoint - lastpoint).Length();
		}

	// Calculate the desired current distance
	gdouble pathdistance = PathPercent * pathlength;

	// Find the location in space ...
	lastpoint = firstpoint;
	for(K3DPOINTHANDLE pointhandle2 = NextPoint(FirstPoint()); pointhandle2; pointhandle2 = NextPoint(pointhandle2))
		{
			sdpVector3 point = Point(pointhandle2)->Location()->AnimXYZ();

			gdouble segmentlength = (point - lastpoint).Length();

			if((pathdistance - segmentlength) < 0.0)
				return sdpLerp(lastpoint, point, pathdistance / segmentlength);

			pathdistance -= segmentlength;
			lastpoint = point;
		}
	
	gdouble segmentlength = (firstpoint - lastpoint).Length();
	return sdpLerp(lastpoint, firstpoint, pathdistance / segmentlength);
}

sdpVector3 CTransformer::GetNURBSPosition(gdouble PathPercent)
{
	return sdpVector3();
}

sdpVector3 CTransformer::GetPosition(gdouble Time, gdouble Delta)
{
	sdpVector3 result(m_XOffset->Value1(Time, Delta), m_YOffset->Value1(Time, Delta), m_ZOffset->Value1(Time, Delta));

	if(m_EnableTranslation.Value())
		{
			const gdouble pathpercent = m_Translation->Value1(Time, Delta);
			switch(m_TranslateType.Value())
				{
					case POINT:
						result += GetPointPosition(pathpercent);
						break;
					case LINEAR:
						result += GetLinearPosition(pathpercent);
						break;
					case NURBS:
						result += GetNURBSPosition(pathpercent);
						break;
					default:
						g_return_val_if_fail(0, result);
				}
		}

	return result;
}

sdpVector3 CTransformer::GetOrientation(gdouble Time, gdouble Delta)
{
	return sdpVector3(m_XRotation->Value1(Time, Delta), m_YRotation->Value1(Time, Delta), m_ZRotation->Value1(Time, Delta));
}

sdpVector3 CTransformer::GetScale(gdouble Time, gdouble Delta)
{
	return sdpVector3(m_XScale->Value1(Time, Delta), m_YScale->Value1(Time, Delta), m_ZScale->Value1(Time, Delta));
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CTransformer::InitComplete()
{
	if(0 == m_XRotation.Object())
		g_return_val_if_fail(m_XRotation.Create("X Rotation", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_YRotation.Object())
		g_return_val_if_fail(m_YRotation.Create("Y Rotation", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_ZRotation.Object())
		g_return_val_if_fail(m_ZRotation.Create("Z Rotation", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_XScale.Object())
		g_return_val_if_fail(m_XScale.Create("X Scale", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_YScale.Object())
		g_return_val_if_fail(m_YScale.Create("Y Scale", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_ZScale.Object())
		g_return_val_if_fail(m_ZScale.Create("Z Scale", 1.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
	if(0 == m_XOffset.Object())
		g_return_val_if_fail(m_XOffset.Create("X Offset", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_YOffset.Object())
		g_return_val_if_fail(m_YOffset.Create("Y Offset", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_ZOffset.Object())
		g_return_val_if_fail(m_ZOffset.Create("Z Offset", 0.0, true, -DBL_MAX, DBL_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Translation.Object())
		{
			g_return_val_if_fail(m_Translation.Create("Translation", 0.0, true, 0.0, 1.0, k3dIChannel::DATA_SCALAR, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Translation.Interface());
			k3dChannelSetLinearValue(m_Translation.Interface(), 0.0, 1.0);
		}

/*
	// Generate some data to test with ...
	k3dIPoint* point1 = CreatePoint();
	point1->Location()->SetLocalXYZ(sdpVector3(0, 0, 0));
	k3dIPoint* point2 = CreatePoint();
	point2->Location()->SetLocalXYZ(sdpVector3(10, 1, 0));
	k3dIPoint* point3 = CreatePoint();
	point3->Location()->SetLocalXYZ(sdpVector3(-5, 6, 0));
	k3dIPoint* point4 = CreatePoint();
	point4->Location()->SetLocalXYZ(sdpVector3(-10, 6, 0));
		
	k3dICurve* curve = CreateNURBSCurve();
	curve->Points()->AddPoint(point1);
	curve->Points()->AddPoint(point2);
	curve->Points()->AddPoint(point3);
	curve->Points()->AddPoint(point4);
	k3dMakeBezier(curve);
*/
		
	return CTransformerBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CTransformerDialog

const sdpString CTransformerDialog::dialogname = "properties";
const sdpString CTransformerDialog::control_translatelast = "translatelast";
const sdpString CTransformerDialog::control_translatetype = "translatetype";
const sdpString CTransformerDialog::control_translation = "translation";
const sdpString CTransformerDialog::control_xrotation = "xrotation";
const sdpString CTransformerDialog::control_yrotation = "yrotation";
const sdpString CTransformerDialog::control_zrotation = "zrotation";
const sdpString CTransformerDialog::control_xscale = "xscale";
const sdpString CTransformerDialog::control_yscale = "yscale";
const sdpString CTransformerDialog::control_zscale = "zscale";
const sdpString CTransformerDialog::control_xoffset = "xoffset";
const sdpString CTransformerDialog::control_yoffset = "yoffset";
const sdpString CTransformerDialog::control_zoffset = "zoffset";
const sdpString CTransformerDialog::control_position = "position";
const sdpString CTransformerDialog::control_orientation = "orientation";
const sdpString CTransformerDialog::control_scale = "scale";
const sdpString CTransformerDialog::control_target = "target";

CTransformerDialog::CTransformerDialog(CTransformer* const Object) :
	CTransformerDialogBase(Object, true, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "transformer.gtkml"));

	// Setup check buttons ...
	Connectk3dCheckButton(control_translatelast, m_Object->m_TranslateLast, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Translate Last", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);	

	// Setup channel buttons ...
	Connectk3dChannelButton(control_translation, m_Object->m_Translation.Interface());
	Connectk3dChannelButton(control_xrotation, m_Object->m_XRotation.Interface());
	Connectk3dChannelButton(control_yrotation, m_Object->m_YRotation.Interface());
	Connectk3dChannelButton(control_zrotation, m_Object->m_ZRotation.Interface());
	Connectk3dChannelButton(control_xscale, m_Object->m_XScale.Interface());
	Connectk3dChannelButton(control_yscale, m_Object->m_YScale.Interface());
	Connectk3dChannelButton(control_zscale, m_Object->m_ZScale.Interface());
	Connectk3dChannelButton(control_xoffset, m_Object->m_XOffset.Interface());
	Connectk3dChannelButton(control_yoffset, m_Object->m_YOffset.Interface());
	Connectk3dChannelButton(control_zoffset, m_Object->m_ZOffset.Interface());

	// Setup transformation controls ...
	if(Getk3dPositionController(control_position))
		Getk3dPositionController(control_position)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Position", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dOrientationController(control_orientation))
		Getk3dOrientationController(control_orientation)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Orientation", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dScaleController(control_scale))
		Getk3dScaleController(control_scale)->Attach(Object->Document(), m_Object, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Scale", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	if(Getk3dTargetController(control_target))
		Getk3dTargetController(control_target)->Attach(m_Object, &m_Object->m_Target, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Target", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup choosers ...
	if(Getk3dChooser(control_translatetype))
		Getk3dChooser(control_translatetype)->Attach(new k3dChooserVariableDataSource<CTransformer::TranslateType>(m_Object->m_TranslateType), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Translate Type", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	Show();
}


