// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTwister K-3D object, which applies an animated "twist" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "twister.h"

#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3dilocation.h>
#include <k3dilocationcollection.h>

CTwister::CTwister(sdpIUnknown* Initializer) :
	CTwisterBase(Initializer),
	m_X("x", Document(), Document()->StateRecorder(), 0),
	m_Y("y", Document(), Document()->StateRecorder(), 0),
	m_Z("z", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_X);
	AddVariable(&m_Y);
	AddVariable(&m_Z);
}

CTwister::~CTwister()
{
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTwister::EditObject()
{
	new CTwisterDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CTwisterTwistObject
{
public:
	CTwisterTwistObject(gdouble X, gdouble Y, gdouble Z)
	{
		m_X = X;
		m_Y = Y;
		m_Z = Z;
	}

	void operator()(k3dIObject* const Object)
	{
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dGetDimensions(Object, 0, min, max))
			return;

		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* const locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble maxx = (max[0] - origin[0]) > (origin[0] - min[0]) ? max[0] - origin[0] : origin[0] - min[0];
		gdouble maxy = (max[1] - origin[1]) > (origin[1] - min[1]) ? max[1] - origin[1] : origin[1] - min[1];
		gdouble maxz = (max[2] - origin[2]) > (origin[2] - min[2]) ? max[2] - origin[2] : origin[2] - min[2];

		// For each location ...
		sdpVector3 coords;
		k3dILocation* location;
		gdouble delta, px, py;
		gdouble cosx, cosy, cosz;
		gdouble sinx, siny, sinz;
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();

				// Offset the origin ...
				coords -= origin;

				// Twist around X-Axis
				if(m_X && maxx)
					{
						delta = coords[0] / maxx;
//						if(m_XFalloff()) delta = KSign<double>(delta) * m_Xfalloff->GetValue1(fabs(delta), .0001);

						cosx = cos(m_X * delta);
						sinx = sin(m_X * delta);

						py = coords[1];
						coords[1] = (cosx * coords[1]) - (sinx * coords[2]);
						coords[2] = (sinx * py) + (cosx * coords[2]);
					}

				// Twist around Y-Axis
				if(m_Y && maxy)
					{
						delta = coords[1] / maxy;
//						if(m_YFalloff()) delta = KSign<double>(delta) * m_Yfalloff->GetValue1(fabs(delta), .0001);

						cosy = cos(m_Y * delta);
						siny = sin(m_Y * delta);

						px = coords[0];
						coords[0] = (cosy * coords[0]) + (siny * coords[2]);
						coords[2] = (cosy * coords[2]) - (siny * px);
					}

				// Twist around Z-Axis
				if(m_Z && maxz)
					{
						delta = coords[2] / maxz;
//						if(m_ZFalloff()) delta = KSign<double>(delta) * m_Zfalloff->GetValue1(fabs(delta), .0001);

						cosz = cos(m_Z * delta);
						sinz = sin(m_Z * delta);

						px = coords[0];
						coords[0] = (cosz * coords[0]) - (sinz * coords[1]);
						coords[1] = (sinz * px) + (cosz * coords[1]);
					}

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
	}
	
protected:
	gdouble m_X;
	gdouble m_Y;
	gdouble m_Z;
};

sdpResult CTwister::PopAnimate()
{
	// Modify our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CTwisterTwistObject(m_X->Value1(m_CurrentTime, m_DeltaTime), m_Y->Value1(m_CurrentTime, m_DeltaTime), m_Z->Value1(m_CurrentTime, m_DeltaTime)));
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CTwister::InitComplete()
{
	if(0 == m_X.Object())
		g_return_val_if_fail(m_X.Create("Twister X Rotation", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Y.Object())
		g_return_val_if_fail(m_Y.Create("Twister Y Rotation", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
	if(0 == m_Z.Object())
		g_return_val_if_fail(m_Z.Create("Twister Z Rotation", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);

	return CTwisterBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CTwisterDialog

const sdpString CTwisterDialog::dialogname = "properties";
const sdpString CTwisterDialog::control_x = "x";
const sdpString CTwisterDialog::control_y = "y";
const sdpString CTwisterDialog::control_z = "z";

CTwisterDialog::CTwisterDialog(CTwister* const Object) :
	CTwisterDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "twister.gtkml"));

	// Setup channel buttons ...
	Connectk3dChannelButton(control_x, m_Object->m_X.Interface());
	Connectk3dChannelButton(control_y, m_Object->m_Y.Interface());
	Connectk3dChannelButton(control_z, m_Object->m_Z.Interface());

	CallUpdateControls();
	Show();
}

void CTwisterDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	CTwisterDialogBase::OnEvent(Event);
}


