// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CWaveizer K-3D object, which applies an animated "wave" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "waveizer.h"

#include <k3dchooser.h>
#include <k3dhierarchycollection.h>
#include <k3dmisc.h>
#include <k3doptions.h>

#include <k3diapplication.h>
#include <k3dilocation.h>
#include <k3dilocationcollection.h>

#include <sdpgtkevent.h>

CWaveizer::CWaveizer(sdpIUnknown* Initializer) :
	CWaveizerBase(Initializer),
	m_DisplaceX("displacex", false, Document()->StateRecorder(), 0),
	m_DisplaceY("displacey", true, Document()->StateRecorder(), 0),
	m_DisplaceZ("displacez", false, Document()->StateRecorder(), 0),
	m_Along("along", k3d::X, Document()->StateRecorder(), 0),
	m_Amplitude("amplitude", Document(), Document()->StateRecorder(), 0),
	m_Wavelength("wavelength", Document(), Document()->StateRecorder(), 0),
	m_Phase("phase", Document(), Document()->StateRecorder(), 0)
{
	AddVariable(&m_DisplaceX);
	AddVariable(&m_DisplaceY);
	AddVariable(&m_DisplaceZ);
	AddVariable(&m_Along);
	AddVariable(&m_Amplitude);
	AddVariable(&m_Wavelength);
	AddVariable(&m_Phase);
}

CWaveizer::~CWaveizer()
{
}

//////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CWaveizer::EditObject()
{
	new CWaveizerDialog(this);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CWaveizerWaveizeObject
{
public:
	CWaveizerWaveizeObject(bool DisplaceX, bool DisplaceY, bool DisplaceZ, k3d::Axis Along, gdouble Amplitude, gdouble Wavelength, gdouble Phase)
	{
		m_DisplaceX = DisplaceX;
		m_DisplaceY = DisplaceY;
		m_DisplaceZ = DisplaceZ;
		m_Along = Along;
		m_Amplitude = Amplitude;
		m_Wavelength = Wavelength;
		m_Phase = Phase;
	}

	void operator()(k3dIObject* const Object)
	{
		// Handle the possibility of a zero wavelength up front ...
		if(0 == m_Wavelength)
			return;
	
		// If the object doesn't support k3dILocationCollection, we're done ...
		k3dILocationCollection* locationserver = dynamic_cast<k3dILocationCollection*>(Object);
		if(!locationserver)
			return;

		// For each location ...
		sdpVector3 coords;
		k3dILocation* location;
		gdouble wavefactor;
		for(K3DLOCATIONHANDLE handle = locationserver->FirstLocation(); handle; handle = locationserver->NextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->Location(handle);
				coords = location->AnimXYZ();

				switch(m_Along)
					{
						case k3d::X:
							wavefactor = m_Amplitude * sin(sdp::PiTimes2() * (coords[0] / m_Wavelength) + m_Phase);
							break;
						case k3d::Y:
							wavefactor = m_Amplitude * sin(sdp::PiTimes2() * (coords[1] / m_Wavelength) + m_Phase);
							break;
						case k3d::Z:
							wavefactor = m_Amplitude * sin(sdp::PiTimes2() * (coords[2] / m_Wavelength) + m_Phase);
							break;
						default:
							g_return_if_fail(0);
					}

				// Spherical waves
				if(m_DisplaceX && m_DisplaceY && m_DisplaceZ)
					{
						coords += sdpVector3(coords).Normalize() * wavefactor;
					}
				// Cylindrical waves
				else if(m_DisplaceX && m_DisplaceY)
					{
						coords += sdpVector3(coords[0], coords[1], 0).Normalize() * wavefactor;
					}
				else if(m_DisplaceY && m_DisplaceZ)
					{
						coords += sdpVector3(0, coords[1], coords[2]).Normalize() * wavefactor;
					}
				else if(m_DisplaceX && m_DisplaceZ)
					{
						coords += sdpVector3(coords[0], 0, coords[2]).Normalize() * wavefactor;
					}
				// Linear waves
				else if(m_DisplaceX)
					{
						coords[0] += wavefactor;
					}
				else if(m_DisplaceY)
					{
						coords[1] += wavefactor;
					}
				else if(m_DisplaceZ)
					{
						coords[2] += wavefactor;
					}

				location->SetAnimXYZ(coords);
			}
	}

protected:
	bool m_DisplaceX;
	bool m_DisplaceY;
	bool m_DisplaceZ;

	k3d::Axis m_Along;
	
	gdouble m_Amplitude;
	gdouble m_Wavelength;
	gdouble m_Phase;
};

sdpResult CWaveizer::PopAnimate()
{
	// Calculate Rotation values ...
	gdouble amplitude = m_Amplitude->Value1(m_CurrentTime, m_DeltaTime);
	gdouble wavelength = m_Wavelength->Value1(m_CurrentTime, m_DeltaTime);
	gdouble phase = m_Phase->Value1(m_CurrentTime, m_DeltaTime);
	
	// Apply it to each of our descendants ...
	k3d::for_each_descendant(Hierarchy(), this, CWaveizerWaveizeObject(m_DisplaceX.Value(), m_DisplaceY.Value(), m_DisplaceZ.Value(), m_Along.Value(), amplitude, wavelength, phase));

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CWaveizer::InitComplete()
{
	if(0 == m_Amplitude.Object())
		g_return_val_if_fail(m_Amplitude.Create("Amplitude", 1.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Wavelength.Object())
		g_return_val_if_fail(m_Wavelength.Create("Wavelength", 10.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_DISTANCE, true), SDP_E_FAIL);
	if(0 == m_Phase.Object())
		{
			g_return_val_if_fail(m_Phase.Create("Phase", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, k3dIChannel::DATA_ANGLE, true), SDP_E_FAIL);
			
			k3d::Editing editing(m_Phase.Interface());
			k3dChannelSetLinearValue(m_Phase.Interface(), 0.0, sdp::PiTimes2());
		}

	return CWaveizerBase::InitComplete();
}

/////////////////////////////////////////////////////////////////////////////
// CWaveizerDialog

const sdpString CWaveizerDialog::dialogname = "properties";
const sdpString CWaveizerDialog::control_displacex = "displacex";
const sdpString CWaveizerDialog::control_displacey = "displacey";
const sdpString CWaveizerDialog::control_displacez = "displacez";
const sdpString CWaveizerDialog::control_along = "along";
const sdpString CWaveizerDialog::control_amplitude = "amplitude";
const sdpString CWaveizerDialog::control_wavelength = "wavelength";
const sdpString CWaveizerDialog::control_phase = "phase";

CWaveizerDialog::CWaveizerDialog(CWaveizer* const Object) :
	CWaveizerDialogBase(Object, false, false, true, dialogname, new k3dOptionsWindowGeometryStore(k3dApplication(), 0)),
	m_Object(Object)
{
	// Load the dialog template ...
	g_return_if_fail(LoadGTKMLTemplate("transformerlib", "waveizer.gtkml"));

	// Setup check buttons ...
	Connectk3dCheckButton(control_displacex, m_Object->m_DisplaceX, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace X", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacey, m_Object->m_DisplaceY, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Y", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);
	Connectk3dCheckButton(control_displacez, m_Object->m_DisplaceZ, k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Displace Z", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup choosers ...
	if(Getk3dChooser(control_along))
		Getk3dChooser(control_along)->Attach(new k3dChooserVariableDataSource<k3d::Axis>(m_Object->m_Along), k3dApplication()->CommandRecorder(), m_Object->Document()->StateRecorder(), "Along", 0, new k3dActionRedrawAll(m_Object->Document(), false, 0), 0);

	// Setup channel buttons ...
	Connectk3dChannelButton(control_amplitude, m_Object->m_Amplitude.Interface());
	Connectk3dChannelButton(control_wavelength, m_Object->m_Wavelength.Interface());
	Connectk3dChannelButton(control_phase, m_Object->m_Phase.Interface());

	Show();
}


