// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the renderframe application, which renders frames scheduled as part of a job with the virtual render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3doptions.h>
#include <k3direnderframe.h>

#ifndef SDPXML_H
#include <sdpxml.h>
#endif // SDPXML_H

#include <fstream>
#include <stdio.h>

#ifdef SDPWIN32
#define chdir SetCurrentDirectory
#endif // SDPWIN32

#ifdef SDPUNIX
#include <unistd.h>
#endif // SDPUNIX

#include <time.h>

/// Returns the current time, formatted as an sdpString
sdpString Now()
{
	// Get the current time ...
	time_t now_t;
	time(&now_t);
	
	tm* now_tm = localtime(&now_t);
	
	// Format it ...
	sdpString result;
	strftime(result.buffer(64), 64, "%m/%d/%Y %H:%M:%S", now_tm);
	result.resize(strlen(result));

	return result;
}

bool InputFile(sdpxml::Document& GlobalOptions, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(Element.Name() == "inputfile");
	
	return true;
}

bool OutputFile(sdpxml::Document& GlobalOptions, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(Element.Name() == "outputfile");
	
	return true;
}

bool RenderOperation(sdpxml::Document& GlobalOptions, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(Element.Name() == "renderoperation");

	// Load render data ...
	const sdpString type = sdpxml::GetAttribute<sdpString>(Element, "type", "");
	const sdpString engine = sdpxml::GetAttribute<sdpString>(Element, "engine", "");
	const sdpString sourcepath = sdpxml::GetAttribute<sdpString>(Element, "sourcepath", "");
	const bool visiblerender = sdpxml::GetAttribute(Element, "visiblerender", false);

	// Poke through global options and look for a render engine that matches ...
	sdpString commandline;
	k3dOptionsRenderEngines engines;
	k3dOptionsGetRenderEngines(GlobalOptions, engines);
	for(k3dOptionsRenderEngines::iterator engineiterator = engines.begin(); engineiterator != engines.end(); engineiterator++)
		{
			if((type == engineiterator->Type()) && (engine == engineiterator->Engine()))
				{
					commandline = engineiterator->Render();
					if(visiblerender && engineiterator->VisibleRender().size())
						commandline = engineiterator->VisibleRender();
					break;
				}
		}
	
	if(0 == commandline.size())
		{
			std::cerr << "Could not find requested render engine [" << type << "] [" << engine << "]" << std::endl;
			return false;
		}

	// Log the render output ...
	commandline += " 2> " + sdpString(k3dIRenderFrame::EngineErrorFile);
#ifndef SDPWIN32
	commandline += " | tee " + sdpString(k3dIRenderFrame::EngineLogFile);
#else // !SDPWIN32
	commandline += " > " + sdpString(k3dIRenderFrame::EngineLogFile);
#endif // SDPWIN32

	// Substitute the input source file path ...
	commandline.formattedreplace('%', "p", sourcepath);
	
	// Execute the command ...
	std::cout << Now() << " - Executing " << commandline << std::endl;
	system(commandline);
			
	return true;
}

bool CopyOperation(sdpxml::Document& GlobalOptions, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(Element.Name() == "copyoperation");

	// Extract copy data ...
	const sdpString from = sdpxml::GetAttribute<sdpString>(Element, "from", "");
	const sdpString to = sdpxml::GetAttribute<sdpString>(Element, "to", "");

	// Copy the file ...
#ifndef SDPWIN32
	sdpString copycommand = "cp " + from + " " + to;
#else // !SDPWIN32
	sdpString copycommand = "copy " + from + " " + to;
#endif // SDPWIN32

	system(copycommand);

	return true;
}

bool ViewOperation(sdpxml::Document& GlobalOptions, sdpxml::Element& Element)
{
	// Sanity checks ...
	g_assert(Element.Name() == "viewoperation");

	const sdpString path = sdpxml::GetAttribute<sdpString>(Element, "path", "");

	// Figure out their preferred bitmap viewer ...
	sdpxml::ElementPointerCollection viewers = sdpxml::XPathSelect(GlobalOptions, GlobalOptions, "application/helpers/command[@type=\"bitmapviewer\"]");
	if(0 == viewers.size())
		{
			std::cerr << "No bitmap viewer defined for viewing " << k3dIRenderFrame::OutputImageFile << std::endl;
			return false;
		}

	// View the image ...
	sdpString viewcommand = viewers.front()->Text();
	viewcommand.formattedreplace('%', "p", path);
	viewcommand += " &";
	
	system(viewcommand);
	
	return true;
}

/// Handles a single operation during processing for the frame ...
bool FrameOperation(sdpxml::Document& GlobalOptions, sdpxml::Element& Element)
{
	// Sanity checks ...
	const sdpString operation(Element.Name());

	if(operation == "inputfile")
		return InputFile(GlobalOptions, Element);
	else if(operation == "outputfile")
		return OutputFile(GlobalOptions, Element);
	else if(operation == "renderoperation")
		return RenderOperation(GlobalOptions, Element);
	else if(operation == "copyoperation")
		return CopyOperation(GlobalOptions, Element);
	else if(operation == "viewoperation")
		return ViewOperation(GlobalOptions, Element);
		
	return false;
}

/// Handles all processing for the given frame
bool RenderFrame(sdpxml::Document& GlobalOptions, SDPCSTRING FrameDirectory)
{
	// Sanity checks ...
	g_assert_string(FrameDirectory);

	// Setup some standard paths ...
	const sdpString controlpath = k3dIRenderFrame::ControlFile;
	const sdpString readypath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusReadyFile);
	const sdpString runningpath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusRunningFile);
	const sdpString errorpath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusErrorFile);
	const sdpString completepath = sdpAppendPath(FrameDirectory, k3dIRenderFrame::StatusCompleteFile);

	// Make sure the frame exists ...
	if(!sdpFileExists(FrameDirectory))
		{
			std::cerr << Now() << " - Missing Frame " << FrameDirectory << std::endl;
			return false;
		}

	// Setup our execution environment ...
	chdir(FrameDirectory);

	// Skip the frame if it's complete ...
	if(sdpFileExists(k3dIRenderFrame::StatusCompleteFile))
		return true;

	// Skip the frame if it errored out ...
	if(sdpFileExists(k3dIRenderFrame::StatusErrorFile))
		return true;

	// Skip the frame if it's running ...
	if(sdpFileExists(k3dIRenderFrame::StatusRunningFile))
		return true;

	// Make sure the frame is ready ...
	if(!sdpFileExists(k3dIRenderFrame::StatusReadyFile))
		{
			std::cerr << Now() << " - " << FrameDirectory << " is not ready" << std::endl;
			return false;
		}

	// Standard logging ...
	std::cout << Now() << " - Starting Frame " << FrameDirectory << std::endl;

	// Switch the frame status to running ...
	rename(k3dIRenderFrame::StatusReadyFile, k3dIRenderFrame::StatusRunningFile);

	// Load the frame options file ...
	sdpxml::Document frameoptions("empty");
	if(!frameoptions.Load(k3dIRenderFrame::ControlFile))
		{
			std::cerr << Now() << " - Missing Frame " << FrameDirectory << " Control File " << k3dIRenderFrame::ControlFile << std::endl;
			return false;
		}

	// Get the frame data ...
	sdpxml::ElementPointerCollection frame = sdpxml::XPathSelect(frameoptions, frameoptions, "frame");
	if(0 == frame.size())
		{
			std::cerr << Now() << " - Missing <frame> Data in Frame " << FrameDirectory << " Control File " << std::endl;
			return false;
		}
		
	for(sdpxml::ElementCollection::iterator operation = frame.front()->Children().begin(); operation != frame.front()->Children().end(); operation++)
		FrameOperation(GlobalOptions, *operation);
		
	// Switch the frame status to complete ...
	rename(k3dIRenderFrame::StatusRunningFile, k3dIRenderFrame::StatusCompleteFile);
	
	// Standard logging ...
	std::cout << Now() << " - Completed Frame " << FrameDirectory << std::endl;

	return true;
}

/// Prints usage info
void Usage()
{
	std::cout << "usage: renderframe [options]" << std::endl;
	std::cout << "       renderframe [optionspath] [directory ...]" << std::endl;
	std::cout << std::endl;
	std::cout << "  -h, --help               prints this help information and exits" << std::endl;
	std::cout << "      --version            prints program version information and exits" << std::endl;
	std::cout << std::endl;
}

/// Prints version info
void Version()
{
	std::cout << "K-3D RenderFrame Version " << VERSION << std::endl;
	std::cout << "Copyright (c) 1995-2000, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
	std::cout << "Licensed by the GNU General Public License.  See the COPYING file for details." << std::endl;
	std::cout << "K-3D Home Page: http://www.k-3d.com" << std::endl;
	std::cout << "Bug reports to k-3d-list@k-3d.com" << std::endl;
}

/// Program main
int main(int argc, char* argv[])
{
	// Put our arguments in a more useable form ...
	sdpStringArray arguments;
	for(int i = 1; i < argc; i++)
		arguments.push_back(argv[i]);

	// Print a "help" message ...
	if((arguments.end() != std::find(arguments.begin(), arguments.end(), "-h")) || (arguments.end() != std::find(arguments.begin(), arguments.end(), "--help")))
		{
			Usage();
			return 0;
		}

	// Print version data ...
	if(arguments.end() != std::find(arguments.begin(), arguments.end(), "--version"))
		{
			Version();
			return 0;
		}

	// Otherwise we should have a minimum of two arguments ...
	if(arguments.size() < 2)
		{
			Usage();
			return 1;
		}

	// Open the global options file ...
	sdpString globaloptionspath(arguments[0]);
	sdpxml::Document globaloptions("empty");
	if(!globaloptions.Load(globaloptionspath))
		{
			std::cerr << "renderframe: error opening global options file [" << globaloptionspath << "]" << std::endl;
			return 1;
		}

	// Each remaining argument should be a frame path to render ...						
	int result = 0;
	for(gulong j = 1; j < arguments.size(); j++)
		{
			if(!RenderFrame(globaloptions, arguments[j]))
				result = 1;
		}

	return result;
}


