// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the renderjob application, which runs jobs scheduled with the virtual render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpxml.h>

#include <k3direnderjob.h>

#include <fstream>

#include <stdio.h>

#ifdef SDPUNIX
#include <unistd.h>
#endif // SDPUNIX

#ifdef SDPWIN32
#define chdir SetCurrentDirectory
#endif // SDPWIN32

#include <time.h>

sdpString Now()
{
	// Get the current time ...
	time_t now_t;
	time(&now_t);
	
	tm* now_tm = localtime(&now_t);
	
	// Format it ...
	sdpString result;
	strftime(result.buffer(64), 64, "%m/%d/%Y %H:%M:%S", now_tm);
	result.resize(strlen(result));

	return result;
}

class RenderJobRenderFrame
{
public:
	RenderJobRenderFrame(SDPCSTRING GlobalOptionsPath) :
		m_GlobalOptionsPath(GlobalOptionsPath)
	{
		g_assert(m_GlobalOptionsPath.size());
	}

	bool HandleDirectory(SDPCSTRING DirectoryPath, SDPCSTRING Directory)
	{
		// Render the frame ...
		system("renderframe " + m_GlobalOptionsPath + " " + DirectoryPath);
		return true;
	}
	
protected:
	const sdpString m_GlobalOptionsPath;
};

bool RenderJob(SDPCSTRING GlobalOptionsPath, SDPCSTRING JobDirectory)
{
	// Sanity checks ...
	g_assert_string(GlobalOptionsPath);
	g_assert_string(JobDirectory);

	// Make sure the job exists ...
	if(!sdpFileExists(JobDirectory))
		{
			std::cerr << Now() << " - " << JobDirectory << " not found" << std::endl;
			return false;
		}

	// Setup our execution environment ...
	chdir(JobDirectory);

	// Skip the job if it's complete ...
	if(sdpFileExists(k3dIRenderJob::StatusCompleteFile))
		return true;

	// Skip the job if it errored out ...
	if(sdpFileExists(k3dIRenderJob::StatusErrorFile))
		return true;

	// Skip the job if it's running ...
	if(sdpFileExists(k3dIRenderJob::StatusRunningFile))
		return true;

	// Make sure the job is ready ...
	if(!sdpFileExists(k3dIRenderJob::StatusReadyFile))
		{
			std::cerr << "renderjob: Job " << JobDirectory << " is not ready" << std::endl;
			return false;
		}

	// Standard logging ...
	std::cout << Now() << " - Starting Job " << JobDirectory << std::endl;

	// Switch the job status to running ...
	rename(k3dIRenderJob::StatusReadyFile, k3dIRenderJob::StatusRunningFile);
	
	// For each directory in the job directory (non-recursive) ...
	RenderJobRenderFrame renderframe(GlobalOptionsPath);
	sdpForEachDirectory(JobDirectory, renderframe, false);

	// Switch the job status to complete ...
	rename(k3dIRenderJob::StatusRunningFile, k3dIRenderJob::StatusCompleteFile);
	
	// Standard logging ...
	std::cout << Now() << " - Completed Job " << JobDirectory << std::endl;
	
	return true;
}

/// Prints usage info
void Usage()
{
	std::cout << "usage: renderjob [options]" << std::endl;
	std::cout << "       renderjob [optionspath] [directory ...]" << std::endl;
	std::cout << std::endl;
	std::cout << "  -h, --help               prints this help information and exits" << std::endl;
	std::cout << "      --version            prints program version information and exits" << std::endl;
	std::cout << std::endl;
}

/// Prints version info
void Version()
{
	std::cout << "K-3D RenderJob Version " << VERSION << std::endl;
	std::cout << "Copyright (c) 1995-2000, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
	std::cout << "Licensed by the GNU General Public License.  See the COPYING file for details." << std::endl;
	std::cout << "K-3D Home Page: http://www.k-3d.com" << std::endl;
	std::cout << "Bug reports to k-3d-list@k-3d.com" << std::endl;
}

/// Program main
int main(int argc, char* argv[])
{
	// Put our arguments in a more useable form ...
	sdpStringArray arguments;
	for(int i = 1; i < argc; i++)
		arguments.push_back(argv[i]);

	// Print a "help" message ...
	if((arguments.end() != std::find(arguments.begin(), arguments.end(), "-h")) || (arguments.end() != std::find(arguments.begin(), arguments.end(), "--help")))
		{
			Usage();
			return 0;
		}

	// Print version data ...
	if(arguments.end() != std::find(arguments.begin(), arguments.end(), "--version"))
		{
			Version();
			return 0;
		}

	// Otherwise we should have a minimum of two arguments ...
	if(arguments.size() < 2)
		{
			Usage();
			return 1;
		}

	// Each remaining argument should be a job path to render ...						
	int result = 0;
	for(gulong j = 1; j < arguments.size(); j++)
		{
			if(!RenderJob(arguments[0], arguments[j]))
				result = 1;
		}

	return result;
}


